Require Import Coqlib.
Require Import Map.
Require Import Arith.
Require Import Machine.
Require Import Assertion.
Require Export Assertion_Lemmas3.
Require Import Omega.

Set Implicit Arguments.
Unset Strict Implicit.



(******************************************)

(* Lemmas about logic assertions          *)

(******************************************)

Lemma weaken_presv_pmstar_m :
  forall (p: Pred) (m m': MPred) (S: State),
    (p *<* m) S -> (forall H, m H -> m' H) -> (p *<* m') S.
Proof.
  intros.
  unfold pmstar.
  unfold pmstar in H.
  InvertAll.
  exists x. exists x0.
  split; auto.
Qed.

Lemma weaken_presv_pmstar_p :
  forall (p p' : Pred) (m : MPred) (S : State),
    (p *<* m) S -> (forall S, p S -> p' S) -> (p' *<* m) S.
Proof.
  intros.
  unfold pmstar. unfold pmstar in H.
  InvertAll.
  exists x. exists x0.
  split; auto.
Qed.

Lemma mtrue_weakest :
  forall m H, m H -> mTrue H.
Proof.
  intros.
  unfold mTrue.
  auto.
Qed.

Lemma pmstar_imprec_p :
  forall p m S,
    (p *<* m) S -> imprec p S.
Proof.
  intros.
  unfold imprec.
  apply weaken_presv_pmstar_m with m; auto.
  apply mtrue_weakest.
Qed.

Lemma imprec_presv_imprec :
  forall (p p' : Pred), (forall S, p S -> imprec p' S) -> (forall S, imprec p S -> imprec p' S).
Proof.
  intros.
  unfold imprec in H0.
  unfold pmstar in H0.
  InvertAll.
  assert (MemMod.subseteq x (S_Mem S)).
  eauto with PMap.
  induction S. induction a. induction a.
  apply imprec_monotone with x.
  apply H; auto.
  unfold S_Mem in H2. auto.
Qed.


Lemma imprec_monotone_S :
  forall (p : Pred) (S : State) (M : Mem),
    imprec p (S_Mem_Upd S M) -> MemMod.subseteq M (S_Mem S) -> imprec p S.
Proof.
  intros.
  induction S.
  induction a.
  induction a.
  apply imprec_monotone with M; auto.
Qed.

Definition monotonic_p (p : Pred) : Prop :=
  forall S M, p S -> MemMod.subseteq (S_Mem S) M -> p (S_Mem_Upd S M).


Lemma S_Mem_Upd_discard_interm:
  forall S x, S = S_Mem_Upd (S_Mem_Upd S x) (S_Mem S).
Proof.
  intros.
  induction S.
  induction a.
  induction a.
  UnfoldStateAll. auto.
Qed.
Hint Resolve S_Mem_Upd_discard_interm.

Lemma S_Mem_Upd_cancel:
  forall S x, x = S_Mem (S_Mem_Upd S x).
Proof.
  intros.
  UnfoldStateAll. auto.
Qed.
Hint Resolve S_Mem_Upd_cancel.


Lemma imprec_presv_imp_monotonic:
  forall (p q : Pred) , monotonic_p q -> (forall S, p S -> q S) -> (forall S, imprec p S -> q S).
Proof.
  intros.
  unfold imprec in H1.
  unfold pmstar in H1.
  InvertAll.
  assert (MemMod.subseteq x (S_Mem S)).
  eauto with PMap.
  assert (q (S_Mem_Upd S x)).
  apply H0; auto.
  unfold monotonic_p in H.
  assert (S = (S_Mem_Upd (S_Mem_Upd S x) (S_Mem S))).
  apply S_Mem_Upd_discard_interm.
  rewrite H6.
  apply H. auto.
  assert (x = S_Mem (S_Mem_Upd S x)).
  apply S_Mem_Upd_cancel.
  rewrite <- H7. auto.
Qed.

Lemma pTrue_m_mTrue :
  forall p m S, (p *<*m) S -> (m *^* mTrue) (S_Mem S).
Proof.
  intros.
  induction S. induction a. induction a.
  unfold pmstar in H. unfold mstar.
  InvertAll.
  exists x0. exists x.
  UnfoldStateAll.
  split. auto with PMap.
  split; auto.
  unfold mTrue. auto.
Qed.

Definition leqvM (m m' : MPred) :=
  forall h, m h <-> m' h.

Lemma leqvM_reflex :
  forall m, leqvM m m.
Proof.
  unfold leqvM. intros. split; auto.
Qed.
Hint Resolve leqvM_reflex.

Lemma leqvM_commut :
  forall m m', leqvM m m' -> leqvM m' m.
Proof.
  intros m m'. unfold leqvM.
  intros.
  assert (m h <-> m' h). apply H.
  inversion_clear H0.
  split; auto.
Qed.
Hint Resolve leqvM_commut.

Lemma leqvM_trans :
  forall m m' m'', leqvM m m' -> leqvM m' m'' -> leqvM m m''.
Proof.
  intros m m' m''. unfold leqvM. intros.
  assert (m h <-> m' h). apply H.
  assert (m' h <-> m'' h). apply H0.
  inversion_clear H1. inversion_clear H2.
  split; auto.
Qed.

Lemma mstar_eqv_subst_right:
  forall m m' m'', leqvM m' m'' -> leqvM (m *^* m') (m *^* m'').
Proof.
  intros m m' m''.
  unfold leqvM.
  intros.
  split.
  intro.
  unfold mstar in H0. unfold mstar.
  InvertAll.
  exists x. exists x0.
  split. auto. split. auto.
  assert (m' x0 <-> m'' x0). apply H.
  inversion_clear H2.
  apply H4; auto.
  unfold mstar. intros.
  InvertAll.
  exists x. exists x0.
  split; auto.
  split; auto.
  assert (m' x0 <-> m'' x0). apply H.
  inversion_clear H2.
  apply H5; auto.
Qed.
Hint Resolve mstar_eqv_subst_right.

Lemma mstar_commut2 :
  forall m m', leqvM (m *^* m') (m' *^* m).
Proof.
  unfold leqvM. intros.
  split; auto.
Qed.
Hint Resolve mstar_commut2.

Lemma mstar_eqv_subst_left:
  forall m m' m'', leqvM m m'' -> leqvM (m *^* m') (m'' *^* m').
Proof.
  intros.
  assert (leqvM (m *^* m') (m' *^* m)). auto.
  assert (leqvM (m' *^* m) (m' *^* m'')). auto.
  assert (leqvM (m *^* m') (m' *^* m'')).
  apply leqvM_trans with (m' *^* m); auto.
  assert (leqvM (m' *^* m'') (m'' *^* m')). auto.
  apply leqvM_trans with (m' *^* m''); auto.
Qed.
Hint Resolve mstar_eqv_subst_left.

Lemma mstar_assoc :
  forall m m' m'', leqvM (m *^* m' *^* m'') (m *^* (m' *^* m'')).
Proof.
unfold leqvM;intros.
split;unfold mstar;intros.
(*forward*)
repeat destruct H.
repeat destruct H0.
destruct H3.
exists x1;exists (MemMod.merge x2 x0).
subst.
split.
unfold MemMod.disjUnion.
auto with PMap.
split.
assumption.
exists x2;exists x0.
repeat split.
apply MemMod.disj_merge_disj_right with x1;assumption.
assumption.
assumption.
(*backward*)
repeat destruct H.
repeat destruct H0.
repeat destruct H2.
subst.
destruct H3.
exists (MemMod.merge x x1);exists x2.
repeat split.
auto with PMap.
auto with PMap.
exists x;exists x1.
repeat split;try assumption.
apply MemMod.disj_x_merge_disj_left with x2;assumption.
assumption.
Qed.

Lemma leqvM_app_subst:
  forall m m' H, leqvM m m' -> m H -> m' H.
Proof.
  intros.
  unfold leqvM in H0.
  assert (m H <-> m' H). apply H0.
  inversion_clear H2. auto.
Qed.

Lemma mstar_subst_weaker:
  forall (m m' m'': MPred) H, (m *^* m') H -> (forall H, m' H -> m'' H) -> (m *^* m'') H.
Proof.
  intros.
  unfold mstar in H0. unfold mstar.
  InvertAll.
  exists x. exists x0. split; auto.
Qed.

Lemma mstar_subst_weaker_left:
  forall (m m' m'': MPred) H, (m *^* m') H -> (forall H, m H -> m'' H) -> (m'' *^* m') H.
Proof.
  intros.
  unfold mstar in H0. unfold mstar.
  InvertAll.
  exists x. exists x0. split; auto.
Qed.

Lemma mstar_mTrue :
  forall m m' m'' H, (m *^* m' *^* m'') H -> (m' *^* mTrue) H.
Proof.
  intros.

  assert (leqvM (m *^* m') (m'*^*m)). auto.
  assert (leqvM (m *^* m' *^* m'') (m' *^* m *^* m'')).
  auto.
  assert (leqvM (m' *^* m *^* m'') (m' *^* (m *^* m''))).
  apply mstar_assoc.
  assert (leqvM (m *^* m' *^* m'') (m' *^* (m *^* m''))).
  apply leqvM_trans with (m' *^* m *^* m''); auto.
  assert ((m' *^* (m *^* m'')) H). apply leqvM_app_subst with (m *^* m' *^* m''); auto.
  apply mstar_subst_weaker with (m *^* m''). auto.
  apply mtrue_weakest.
Qed.

(******************)
(***added by wangwei***)


Lemma disjUnion_assoc_lr : forall M1 M2 M3 M12 M123,
  MemMod.disjUnion M1 M2 M12
  -> MemMod.disjUnion M12 M3 M123
  -> MemMod.disjUnion M1 (MemMod.merge M2 M3) M123.
Proof.
  unfold MemMod.disjUnion; intros.
  destruct H; destruct H0.
  subst.
  split.
    eauto with PMap.
  auto with PMap.
Qed.

Lemma disjUnion_assoc_l : forall M1 M2 M3 M12 M23 M123,
  MemMod.disjUnion M1 M2 M12
  -> MemMod.disjUnion M12 M3 M123
  -> M23 = MemMod.merge M2 M3
  -> MemMod.disjUnion M2 M3 M23
  /\ MemMod.disjUnion M1 M23 M123.
Proof.
  unfold MemMod.disjUnion; intros.
  destruct H; destruct H0.
  subst.
  split.
    split.
      eauto with PMap.
    trivial.
  split.
    eauto with PMap.
  auto with PMap.
Qed.

Lemma disjUnion_upd_left_disjUnion : forall M1 M2 M M1',
  MemMod.disjUnion M1 M2 M
  -> MemMod.eqDom M1 M1'
  -> MemMod.disjUnion M1' M2 (MemMod.merge M1' M2).
Proof.
  unfold MemMod.disjUnion.
  intros.
  destruct H.
  split_r.
  subst.
  eapply MemMod.eqDom_disj_presv; eauto.
Qed.

Lemma disjUnion_disjUnion_Inv_whole : forall M1 M2 M M',
  MemMod.disjUnion M1 M2 M 
  -> MemMod.disjUnion M1 M2 M'
  -> M = M'.
Proof.
  unfold MemMod.disjUnion.
  intros.
  destruct H; destruct H0.
  rewrite <- H2 in H1.
  trivial.
Qed.
Implicit Arguments disjUnion_disjUnion_Inv_whole [M1 M2 M M'].


Lemma merge_put_put_merge : forall l x M1 M2,
  MemMod.in_dom l M1
  -> MemMod.not_in_dom l M2
  -> MemMod.merge (MemMod.put M1 l x) M2 
  = MemMod.put (MemMod.merge M1 M2) l x.
Proof.
  unfold MemMod.in_dom, MemMod.not_in_dom, MemMod.merge, MemMod.put.
  intros.
  apply MemMod.extensional_eqMap.
  intros.
  destruct (MemMod.beq_A x0 l).
    trivial.
  destruct H.
  destruct (M1 x0); trivial.
Qed.

Lemma nat_eq_neq_dec:
  forall n n' : nat, n = n' \/ n <> n'.
Proof.
intros.
destruct (eq_nat_dec n n').
left;assumption.
right;assumption.
Qed.
