Require Import Coqlib.
Require Import Map.
Require Import Arith.
Require Import Machine.
Require Import Mem_Lemmas.
Require Import Assertion.
Require Import Omega.

Notation Local "[{ A || B = C }] " := (MemMod.disjUnion A B C) (left associativity, at level 68).

Notation Local " A ++ B " := (MemMod.merge A B) (left associativity, at level 67).

Set Implicit Arguments.
Unset Strict Implicit.

Lemma imprecM_weaker: forall M (A : MPred),
    A M  -> (imprecM A) M.
Proof.
  intros M A H.
  unfold imprecM.
  exists M; exists MemMod.emp.
  split.
    auto with PMap.
  unfold mTrue.
  auto.
Qed.

Lemma ptoxBlk_indom : forall size l offset M,
  ptoxBlk l size M
  -> offset <  size
  -> MemMod.in_dom (l+offset) M.
Proof.
  induction size; simpl; intros.
    inversion H0.
    destruct offset.
    unfold mstar in H.
    destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
    unfold ptox in HM1. 
    unfold pto in HM1.
    replace (l+0) with l; auto.
    apply MemMod.disjUnion_presv_in_dom with M1 M2; trivial.
    destruct HM1.
    unfold MemMod.sgltonM in H.
    destruct H.
    unfold MemMod.maps_to in H.
    unfold MemMod.in_dom.
    exists x; trivial.
  assert (offset < size).
    auto with arith.
  replace (l + S offset) with ((l+1) + offset).
  destruct H as [M1 [M2 [HJM [Hl H]]]].
  assert (MemMod.in_dom (l+1+offset) M2).
    apply IHsize; trivial.
    assert (MemMod.disjUnion M2 M1 M).
      auto with PMap.
    apply MemMod.disjUnion_presv_in_dom with M2 M1; trivial.
  rewrite <- plus_assoc.
  simpl.
  trivial.
Qed.

Lemma ptoxBlk_notindom: forall l s addr h,
  ptoxBlk l s h
  -> addr < l \/ addr >= l + s
  -> h addr = None.
Proof.
intros l s.
gen_clear l.
induction s; intros.
simpl in H.
unfold Aemp in H.
subst;trivial.
simpl in H.
destruct H as [h1 [h2 [HJh H]]].
destruct H.
assert (Hx : addr < l + 1 \/ addr >= l+1+s).
  omega.
assert (Hx0 := IHs (l+1) addr h2 H1 Hx).
assert (Hx1 : h1 addr = None).
unfold ptox in H.
inversion H.
unfold pto in H2.
unfold MemMod.sgltonM in H2.
destruct H2.
unfold MemMod.not_in_dom in H3.
apply H3.
omega.
substH HJh with (disjUnion_elim HJh addr).
rewrite Hx0 in HJh.
rewrite Hx1 in HJh.
destruct (h addr); try contradiction.
trivial.
Qed.

Lemma ptoxBlk_split : forall s1 s2 M l,
  ptoxBlk l (s1 + s2) M
  -> mstar (ptoxBlk l s1) (ptoxBlk (l+s1) s2) M.
Proof.
  induction s1; simpl; intros.
  unfold Aemp, MemMod.emp.
  exists MemMod.emp; exists M.
  split.
    auto with PMap.
    unfold MemMod.emp.
    split_l.
  replace (l+0) with l; auto with arith.
  destruct  H as [M1 [M2 [MJM [HM1 HM2]]]].
  replace (l + S s1) with (S l + s1); [idtac | omega].
  assert (Hx := IHs1 s2 M2 (l+1) HM2).
  apply mstar_assoc_right_left.
  unfold mstar.
  exists M1; exists M2; split_l.
  split_l.
  arith_rep (S l + s1) with (l+1+s1).
Qed.

Lemma ptoxBlk_merge : forall s1 s2 M l,
  mstar (ptoxBlk l s1) (ptoxBlk (l+s1) s2) M
  -> ptoxBlk l (s1 + s2) M.
Proof.
induction s1;simpl;intros.
unfold mstar, Aemp in H.
InvertAll.
subst.
assert (x0 = M).
auto with PMap.
subst.
arith_rep l with (l+0);assumption.
assert ( (ptox l *^* (ptoxBlk (l + 1) s1 *^* ptoxBlk (l + S s1) s2)) M).
apply mstar_assoc_left_right;assumption.
unfold mstar at 1 in H0.
InvertAll.
exists x;exists x0;intuition.
apply IHs1.
arith_rep (l+1+s1) with (l+S s1);assumption.
Qed.

Lemma precise_ptoxBlk : forall l s,
  precise (ptoxBlk l s).
Proof.
unfold precise.
intros.
apply MemMod.extensional_eqMap.
unfold MemMod.subseteq in * .
intro addr.
unfold MemMod.A in * .
unfold MemSpec.domain in * .
destruct (le_lt_dec l addr).
destruct (le_lt_dec (l+s) addr).
assert (M1 addr = None).
apply ptoxBlk_notindom with l s.
assumption.
omega.
assert (M2 addr = None).
apply ptoxBlk_notindom with l s.
assumption.
omega.
rewrite H3;rewrite H4;reflexivity.
assert (addr = l + (addr - l)).
omega.
rewrite H3 in l1.
assert (addr- l < s).
omega.
rewrite H3.
assert (M1 (l+(addr-l)) = M (l+(addr-l))).
apply H.
apply ptoxBlk_indom with s;assumption.
assert (M2 (l+(addr-l)) = M (l+(addr-l))).
apply H0.
apply ptoxBlk_indom with s;assumption.
rewrite H5;rewrite H6;reflexivity.
assert (M1 addr = None).
apply ptoxBlk_notindom with l s.
assumption.
omega.
assert (M2 addr = None).
apply ptoxBlk_notindom with l s.
assumption.
omega.
rewrite H3;rewrite H4;reflexivity.
Qed.

(******************)
(* leadsTo *)

Lemma leadsTo_intro : forall M addr x,
  M addr = Some x -> leadsTo addr x M.
Proof.
  unfold leadsTo.
  unfold imprecM.
  intros.
  unfold mstar, pto.
  exists (fun l : Label => if MemMod.beq_A l addr then Some x else None).
  exists (fun l : Label => if MemMod.beq_A l addr then None else M l).
  unfold MemMod.disjUnion.
  split.
    split.
      unfold MemMod.disjoint.
      intros l.
      unfold MemMod.not_in_dom.
      destruct (MemMod.beq_A l addr).
      right; trivial.
      auto.
    unfold MemMod.merge.
    apply (extensionality MemMod.A (option MemMod.B)
      M (fun x0 : MemMod.A =>
        match (if MemMod.beq_A x0 addr then Some x else None (A:=nat)) with
          | Some _ => if MemMod.beq_A x0 addr then Some x else None (A:=nat)
          | None => if MemMod.beq_A x0 addr then None (A:=nat) else M x0
        end)).
    intros l.
    destruct (MemMod.beq_A l addr).
    subst; trivial.
    trivial.
  split.
    unfold MemMod.sgltonM.
    split.
      unfold MemMod.maps_to.
      destruct (MemMod.beq_A addr addr).
        reflexivity.
        elim n;reflexivity.
      intros.
      unfold MemMod.not_in_dom.
      destruct (MemMod.beq_A x0 addr).
        elim H0;assumption.
        reflexivity.
    unfold mTrue.
  trivial.
Qed.

Lemma leadsTo_elim : forall addr x M,
  leadsTo addr x M -> M addr = Some x.
Proof.
  unfold leadsTo, imprecM, pto, mstar.
  intros.
  destruct H.
  destruct H.
  destruct H.
  destruct H0.
  unfold MemMod.sgltonM in H0.
  unfold MemMod.maps_to in H0.
  destruct H0.
  unfold mTrue in H1.
  unfold MemMod.disjUnion in H.
  destruct H.
  subst M.
  unfold MemMod.disjoint in H.
  unfold MemMod.not_in_dom in H.
  destruct (H addr).
    rewrite H0 in H3; discriminate H3.
  clear H.
  unfold MemMod.merge.
  rewrite H0.
  trivial.
Qed.
Implicit Arguments leadsTo_elim [addr x M].

Lemma leadsTo_Mono : forall addr x M1 M2 M,
  MemMod.disjUnion M1 M2 M
  -> leadsTo addr x M1
  -> leadsTo addr x M.
Proof.
  unfold leadsToW.
  intros.
  unfold imprecM in * .
  unfold mstar in * .
  destruct H0 as [M11 [M12 [HJM1 [HM2 _]]]].
  exists M11.
  exists (MemMod.merge M12 M2).
  rsplit.
  split_l.
  unfold mTrue; trivial.
  split.
    apply MemMod.disj_merge_A.
    destruct HJM1.
    trivial.
    destruct HJM1.
    subst M1.
    destruct H.
    eapply MemMod.disj_merge_disj_left; eauto.
  destruct H.
  subst M.
  destruct HJM1.
  subst M1.
  rewrite MemMod.merge_assoc.
  trivial.
Qed.
Implicit Arguments leadsTo_Mono [addr x M1 M2 M].

Lemma leadsTox_intro : forall addr M x,
  M addr = Some x -> leadsTox addr M.
Proof.
  intros.
  exists (fun l => if MemMod.beq_A l addr then Some x else None).
  exists (fun l => if MemMod.beq_A l addr then None else M l).
  split.
  unfold MemMod.disjUnion.
  unfold MemMod.disjoint.
  unfold MemMod.not_in_dom.
  split.
  intro x0.
  destruct (MemMod.beq_A x0 addr).
  right;reflexivity.
  left;reflexivity.
  unfold MemMod.merge.
  apply (extensionality MemMod.A (option MemMod.B) M
    (fun x0 : MemMod.A =>
    match (if MemMod.beq_A x0 addr then Some x else None (A:=MemMod.B)) with
    | Some _ => if MemMod.beq_A x0 addr then Some x else None (A:=MemMod.B)
    | None => if MemMod.beq_A x0 addr then None (A:=MemMod.B) else M x0
    end)).
  intro x0.
  destruct (MemMod.beq_A x0 addr).
  rewrite e;assumption.
  reflexivity.
  unfold ptox.
  split.
  exists x.
  unfold pto.
  unfold MemMod.sgltonM.
  unfold MemMod.maps_to.
  destruct (MemMod.beq_A addr addr).
  split.
  reflexivity.
  intros.
  unfold MemMod.not_in_dom.
  destruct (MemMod.beq_A x0 addr).
  elim H0;assumption.
  reflexivity.
  elim n;reflexivity.
  unfold mTrue;trivial.
Qed.

Lemma leadsTox_elim : forall addr M,
  leadsTox addr M -> exists x,M addr = Some x.
Proof.
  unfold leadsTox, imprecM, ptox, pto, mstar.
  intros.
  destruct H.
  destruct H.
  destruct H.
  destruct H0.
  unfold MemMod.sgltonM in H0.
  unfold MemMod.maps_to in H0.
  destruct H0.
  destruct H0.
  unfold mTrue in H1.
  unfold MemMod.disjUnion in H.
  destruct H.
  subst M.
  unfold MemMod.disjoint in H.
  unfold MemMod.not_in_dom in H.
  destruct (H addr).
    rewrite H0 in H3; discriminate H3.
  clear H.
  unfold MemMod.merge.
  rewrite H0.
  exists x1.
  trivial.
Qed.

(* leadsToW *)

Lemma leadsToW_intro : forall M l x0 x1,
  M l = Some x0
  -> M (l+1) = Some x1
  -> leadsToW l x0 M.
Proof.
  intros M l x0 x1 H1 H2.
  exists (fun addr : Label => if MemMod.beq_A l addr then Some x0 else if MemMod.beq_A (l+1) addr then Some x1 else None).
  exists (fun addr : Label => if MemMod.beq_A l addr then None else if MemMod.beq_A (l+1) addr then None else M addr).
  split.
    unfold MemMod.disjUnion.
    split.
      unfold MemMod.disjoint;unfold MemMod.not_in_dom.
      intro x.
      destruct (MemMod.beq_A l x). 
        right;reflexivity.
        destruct (MemMod.beq_A (l+1) x).
          right;reflexivity.
          left;reflexivity.
    apply (extensionality MemMod.A (option MemMod.B) 
      M  (MemMod.merge (fun addr : nat =>
        if MemMod.beq_A l addr
          then Some x0
          else if MemMod.beq_A (l + 1) addr then Some x1 else None (A:=MemMod.B))
      (fun addr : nat =>
        if MemMod.beq_A l addr
          then None (A:=MemMod.B)
          else if MemMod.beq_A (l + 1) addr then None (A:=MemMod.B) else M addr))).
    intro x.
    unfold MemMod.merge.
    destruct (MemMod.beq_A l x).
      rewrite <-e;assumption.
      destruct (MemMod.beq_A (l+1) x).
        rewrite <- e;assumption.
        reflexivity.
    split.
      exists (fun addr : Label => if MemMod.beq_A l addr then Some x0 else None).
      exists (fun addr : Label => if MemMod.beq_A (l+1) addr then Some x1 else None).
      split.
        unfold MemMod.disjUnion.
        split.
          unfold MemMod.disjoint.
          unfold MemMod.not_in_dom.
          intro x.
          destruct (MemMod.beq_A l x).
            destruct (MemMod.beq_A (l+1) x).
              rewrite <-e in e0.
              assert (1=l-l).
              apply plus_minus;rewrite e0;reflexivity.
              rewrite <- (minus_n_n l) in H.
              discriminate H.
              right;reflexivity.
            left;reflexivity.
          unfold MemMod.merge.
          apply (extensionality MemMod.A (option MemMod.B)
            (fun addr : nat =>
              if MemMod.beq_A l addr
                then Some x0
                else if MemMod.beq_A (l + 1) addr then Some x1 else None (A:=MemMod.B)) 
            (fun x : MemMod.A =>
              match (if MemMod.beq_A l x then Some x0 else None (A:=MemMod.B)) with
                | Some _ => if MemMod.beq_A l x then Some x0 else None (A:=MemMod.B)
                | None => if MemMod.beq_A (l + 1) x then Some x1 else None (A:=MemMod.B)
              end)).
          intro x.
          destruct (MemMod.beq_A l x);reflexivity.
        split.
          unfold pto;unfold MemMod.sgltonM.
          unfold MemMod.maps_to.
          destruct (MemMod.beq_A l l).
            split.
              reflexivity.
              intros x H3.
              unfold MemMod.not_in_dom.
              destruct (MemMod.beq_A l x).
                rewrite e0 in H3;elim H3;reflexivity.
                reflexivity.
              elim n;reflexivity.
          exists x1.
          unfold pto;unfold MemMod.sgltonM.
          unfold MemMod.maps_to.
          destruct (MemMod.beq_A (l+1) (l+1)).
            split.
              reflexivity.
              intros x H3.
              unfold MemMod.not_in_dom.
              destruct (MemMod.beq_A (l+1) x).
                rewrite e0 in H3;elim H3;reflexivity.
                reflexivity.
              elim n;reflexivity.
    unfold mTrue;trivial.
Qed.

Lemma leadsToW_elim : forall addr x M,
  leadsToW addr x M -> M addr = Some x.
Proof.
  unfold leadsToW, imprecM, pto, mstar.
  intros.
  destruct H.
  destruct H.
  destruct H.
  destruct H0.
  unfold MemMod.sgltonM in H0.
  unfold MemMod.maps_to in H0.
  destruct H0.
  unfold mTrue in H1.
  unfold MemMod.disjUnion in H.
  destruct H.
  subst M.
  unfold MemMod.disjoint in H.
  unfold MemMod.not_in_dom in H.
  destruct H0.
  unfold MemMod.disjoint, pto, ptox in H0.
  unfold MemMod.disjUnion, MemMod.sgltonM, pto in H0.
  unfold MemMod.disjoint, MemMod.merge, MemMod.maps_to in H0.
  unfold MemMod.sgltonM in H0.
  unfold MemMod.maps_to in H0.
  destruct H0.
  destruct H0.
  destruct H2.
  destruct H2.
  unfold MemMod.merge.
  subst x0.
  rewrite H2.
  trivial.
Qed.
Implicit Arguments leadsToW_elim [addr x M].

Lemma leadsToW_elim1 : forall l x0 M,
  leadsToW l x0 M
  -> exists x, M (l+1) = Some x.
Proof.
unfold leadsToW.
unfold imprecM.
unfold ptoW.
unfold mstar.
unfold pto.
unfold MemMod.sgltonM.
intros.
destruct H as [M1 [M2 [H3 [H4 H5]]]].
destruct H4 as [M3 [M4 [H6 [H7 H8]]]].
destruct H8 as [n [H9 H10]].
exists n.
unfold MemMod.disjUnion in * .
destruct H3;destruct H6.
rewrite H2 in H0.
rewrite H0.
unfold MemMod.merge.
destruct H7 as [H11 H12].
assert (M3 (l+1) = None).
  apply H12.
  intros FH.
  assert (1=l-l).
  apply plus_minus;rewrite FH;reflexivity.
  rewrite <- (minus_n_n l) in H3.
  discriminate H3.
rewrite H3.
unfold MemMod.maps_to in H9.
rewrite H9;reflexivity.
Qed.

Lemma leadsToW_unique : forall addr x x' M,
  leadsToW addr x M
  -> leadsToW addr x' M
  -> x = x'.
Proof.
  intros.
  assert (Hx := leadsToW_elim H0).
  rewrite (leadsToW_elim H) in Hx.
  injection Hx; trivial.
Qed.
Implicit Arguments leadsToW_unique [addr x x' M].

Lemma leadsToW_Mono : forall addr x M1 M2 M,
  MemMod.disjUnion M1 M2 M
  -> leadsToW addr x M1
  -> leadsToW addr x M.
Proof.
  unfold leadsToW.
  intros.
  unfold imprecM in * .
  unfold mstar in * . 
  destruct H0 as [M11 [M12 [HJM1 [HM2 _]]]].
  exists M11.
  exists (MemMod.merge M12 M2).
  rsplit.
  split_l.
  unfold mTrue; trivial.
  split.
    apply MemMod.disj_merge_A.
    destruct HJM1.
    trivial.
    destruct HJM1.
    subst M1.
    destruct H.
    eapply MemMod.disj_merge_disj_left; eauto.
  destruct H.
  subst M.
  destruct HJM1.
  subst M1.
  rewrite MemMod.merge_assoc.
  trivial.
Qed.
Implicit Arguments leadsToW_Mono [addr x M1 M2 M].

Lemma ptoW_leadsToW : forall M l v,
     (l |->w v) M
  -> (l |~>w v) M.
Proof.
  intros M l v H.
  unfold leadsToW.
  unfold imprecM.
  exists M; exists MemMod.emp.
  split.
    auto with PMap.
  unfold mTrue.
  split_lr.
Qed.
Implicit Arguments ptoW_leadsToW [M l v].

(* leadsToxW *)

Lemma leadsToxW_intro : forall M l x0 x1,
  M l = Some x0
  -> M (l+1) = Some x1
  -> leadsToxW l M.
Proof.
  intros M l x0 x1 H1 H2.
  exists (fun addr : Label => if MemMod.beq_A l addr then Some x0 else if MemMod.beq_A (l+1) addr then Some x1 else None).
  exists (fun addr : Label => if MemMod.beq_A l addr then None else if MemMod.beq_A (l+1) addr then None else M addr).
  split.
    unfold MemMod.disjUnion.
    split.
      unfold MemMod.disjoint;unfold MemMod.not_in_dom.
      intro x.
      destruct (MemMod.beq_A l x). 
        right;reflexivity.
        destruct (MemMod.beq_A (l+1) x).
          right;reflexivity.
          left;reflexivity.
    apply (extensionality MemMod.A (option MemMod.B) 
      M  (MemMod.merge (fun addr : nat =>
        if MemMod.beq_A l addr
          then Some x0
          else if MemMod.beq_A (l + 1) addr then Some x1 else None (A:=MemMod.B))
      (fun addr : nat =>
        if MemMod.beq_A l addr
          then None (A:=MemMod.B)
          else if MemMod.beq_A (l + 1) addr then None (A:=MemMod.B) else M addr))).
    intro x.
    unfold MemMod.merge.
    destruct (MemMod.beq_A l x).
      rewrite <-e;assumption.
      destruct (MemMod.beq_A (l+1) x).
        rewrite <- e;assumption.
        reflexivity.
    split.
      unfold ptoxW.
      exists (fun addr : Label => if MemMod.beq_A l addr then Some x0 else None).
      exists (fun addr : Label => if MemMod.beq_A (l+1) addr then Some x1 else None).
      split.
        unfold MemMod.disjUnion.
        split.
          unfold MemMod.disjoint.
          unfold MemMod.not_in_dom.
          intro x.
          destruct (MemMod.beq_A l x).
            destruct (MemMod.beq_A (l+1) x).
              rewrite <-e in e0.
              assert (1=l-l).
              apply plus_minus;rewrite e0;reflexivity.
              rewrite <- (minus_n_n l) in H.
              discriminate H.
              right;reflexivity.
            left;reflexivity.
          unfold MemMod.merge.
          apply (extensionality MemMod.A (option MemMod.B)
            (fun addr : nat =>
              if MemMod.beq_A l addr
                then Some x0
                else if MemMod.beq_A (l + 1) addr then Some x1 else None (A:=MemMod.B)) 
            (fun x : MemMod.A =>
              match (if MemMod.beq_A l x then Some x0 else None (A:=MemMod.B)) with
                | Some _ => if MemMod.beq_A l x then Some x0 else None (A:=MemMod.B)
                | None => if MemMod.beq_A (l + 1) x then Some x1 else None (A:=MemMod.B)
              end)).
          intro x.
          destruct (MemMod.beq_A l x);reflexivity.
        split.
          exists x0.
          unfold pto;unfold MemMod.sgltonM.
          unfold MemMod.maps_to.
          destruct (MemMod.beq_A l l).
            split.
              reflexivity.
              intros x H3.
              unfold MemMod.not_in_dom.
              destruct (MemMod.beq_A l x).
                rewrite e0 in H3;elim H3;reflexivity.
                reflexivity.
              elim n;reflexivity.
          exists x1.
          unfold pto;unfold MemMod.sgltonM.
          unfold MemMod.maps_to.
          destruct (MemMod.beq_A (l+1) (l+1)).
            split.
              reflexivity.
              intros x H3.
              unfold MemMod.not_in_dom.
              destruct (MemMod.beq_A (l+1) x).
                rewrite e0 in H3;elim H3;reflexivity.
                reflexivity.
              elim n;reflexivity.
    unfold mTrue;trivial.
Qed.

Lemma leadsToxW_elim : forall l M,
  leadsToxW l M
  -> exists x, M l = Some x.
Proof.
unfold leadsToxW.
unfold imprecM.
unfold ptoxW.
unfold mstar.
unfold ptox.
unfold pto.
unfold MemMod.sgltonM.
intros.
destruct H as [M1 [M2 [H3 [H4 H5]]]].
destruct H4 as [M3 [M4 [H6 [H7 H8]]]].
destruct H7 as [n H9].
exists n.
unfold MemMod.disjUnion in * .
destruct H3;destruct H6.
rewrite H2 in H0.
rewrite H0.
unfold MemMod.merge.
destruct H9.
unfold MemMod.maps_to in H3.
rewrite H3.
reflexivity.
Qed.

Lemma leadsToxW_elim1 : forall l M,
  leadsToxW l M
  -> exists x, M (l+1) = Some x.
Proof.
unfold leadsToxW.
unfold imprecM.
unfold ptoxW.
unfold mstar.
unfold ptox.
unfold pto.
unfold MemMod.sgltonM.
intros.
destruct H as [M1 [M2 [H3 [H4 H5]]]].
destruct H4 as [M3 [M4 [H6 [H7 H8]]]].
destruct H8 as [n [H9 H10]].
exists n.
unfold MemMod.disjUnion in * .
destruct H3;destruct H6.
rewrite H2 in H0.
rewrite H0.
unfold MemMod.merge.
destruct H7 as [n1 [H11 H12]].
assert (M3 (l+1) = None).
  apply H12.
  intros FH.
  assert (1=l-l).
  apply plus_minus;rewrite FH;reflexivity.
  rewrite <- (minus_n_n l) in H3.
  discriminate H3.
rewrite H3.
unfold MemMod.maps_to in H9.
rewrite H9;reflexivity.
Qed.

Lemma leadsToW_leadsToxW :
  forall M l x,
    leadsToW l x M
    -> leadsToxW l M.
Proof.
  unfold leadsToW, leadsToxW.
  unfold ptoW, ptoxW.
  unfold imprecM.
  unfold mstar, pto, ptox, mTrue.
  intros.
  destruct H as [M1 [M2 [HJM [H _]]]].
  destruct H as [M11 [M12 [HJM1 [Hpto1 Hpto2]]]].
  exists M1; exists M2.
  split_l.
  split_r.
  exists M11; exists M12.
  split_l.
  split.
    exists x; trivial.
  trivial.
Qed.
Implicit Arguments leadsToW_leadsToxW [M l x].

Lemma leadsToxW_leadsToW :
  forall M l,
    leadsToxW l M
    -> exists x, leadsToW l x M.
Proof.
  unfold leadsToW, leadsToxW.
  unfold ptoW, ptoxW.
  unfold imprecM.
  unfold mstar, pto, ptox, mTrue.
  intros.
  destruct H as [M1 [M2 [HJM [H _]]]].
  destruct H as [M11 [M12 [HJM1 [Hpto1 Hpto2]]]].
  destruct Hpto1 as [x Hpto1].
  exists x.
  exists M1; exists M2.
  split_l.
  split_r.
  exists M11; exists M12.
  auto.
Qed.
Implicit Arguments leadsToxW_leadsToW [M l].

Lemma leadsToxW_Mono : forall l M1 M2 M,
  MemMod.disjUnion M1 M2 M
  -> leadsToxW l M1
  -> leadsToxW l M.
Proof.
  unfold leadsToxW.
  unfold imprecM, mstar, mTrue.
  intros.
  destruct H0 as [M11 [M12 [HJM1 [Hblk Ht]]]].
  exists M11; exists (MemMod.merge M12 M2).
  rsplit.
  split_lr.
  destruct H.
  destruct HJM1. 
  subst.
  split.
    apply MemMod.disj_merge_A.
    trivial.
    eapply MemMod.disj_merge_disj_left; eauto.
  rewrite MemMod.merge_assoc.
  trivial.
Qed.

Lemma leadsToxW_eqDom : forall M M' l,
  MemMod.eqDom M M'
  -> leadsToxW l M
  -> leadsToxW l M'.
Proof.
  unfold MemMod.eqDom.
  unfold MemMod.sub_dom.
  unfold MemMod.in_dom.
  intros M M' l H H0.
  destruct H as [H1 H2].
  destruct (H1 l (leadsToxW_elim H0)) as [x1 H3].
  destruct (H1 (l+1) (leadsToxW_elim1 H0)) as [x2 H4].
  apply leadsToxW_intro with x1 x2;assumption.
Qed.

(* leadsToBlk *)

Lemma ptoxBlk_intro : forall size l M,
  (forall addr, (addr < l -> M addr = None)
    /\ (addr >= l /\ addr < (l+size) -> exists x, M addr = Some x)
    /\ (addr >= (l + size) -> M addr = None))
  -> ptoxBlk l size M.
Proof.
  induction size.
    intros.
    simpl.
    unfold Aemp.
    unfold MemMod.emp.
    apply (extensionality MemMod.A (option MemMod.B) M 
      (fun _ : MemMod.A => None)).
    intros addr.
    substH H with (H addr).
    destruct H as [H1 [H2 H3]].
    destruct (le_lt_dec l addr).
    assert (addr >= l + 0).
      omega.
    apply H3; trivial.
    apply H1; trivial.
  intros.
  simpl.
  InvertAll.
  assert (l>=l/\l<l+S size).
    omega.
  destruct (H l) as [H1 [H2 H3]].
  destruct (H2 H0).
  exists (fun addr => if MemMod.beq_A addr l then Some x else None).
  exists (fun x => if MemMod.beq_A x l then None else M x).
  split.
    unfold MemMod.disjUnion.
    split.
      unfold MemMod.disjoint.
      intro x0.
      unfold MemMod.not_in_dom.
      destruct (MemMod.beq_A x0 l).
      right;reflexivity.
      left;reflexivity.
      unfold MemMod.merge.
      apply (extensionality MemMod.A (option MemMod.B) M
        (fun x0 : MemMod.A =>
          match (if MemMod.beq_A x0 l then Some x else None (A:=MemMod.B)) with
            | Some _ => if MemMod.beq_A x0 l then Some x else None (A:=MemMod.B)
            | None => if MemMod.beq_A x0 l then None (A:=MemMod.B) else M x0
          end)).
      intro x0.
      destruct (MemMod.beq_A x0 l).
      rewrite e;assumption.
      reflexivity.
  split.
    unfold ptox, pto.
    exists x.
    unfold MemMod.sgltonM.
    unfold MemMod.maps_to.
    split.
      destruct (MemMod.beq_A l l).
      reflexivity.
      elim n; reflexivity.
    intros.
    unfold MemMod.not_in_dom.
    destruct (MemMod.beq_A x0 l).
    elim H5;assumption.
    reflexivity.
  apply IHsize.
  intros.
  clear H0 H1 H2 H3.
  split; intros.
    substH H with (H addr).
    InvertAll.
    destruct (MemMod.beq_A addr l).
    reflexivity.
    apply H1.
    assert (addr <= l).
    omega.
    destruct (le_lt_eq_dec addr l H2).
    assumption.
    elim n;assumption.
  split; intros.
    assert (exists x , M addr = Some x).
    destruct (H addr) as [H1 [H2 H3]].
    apply H2.
    destruct H0.
    split;omega.
    destruct H1.
    exists x0.
    destruct (MemMod.beq_A addr l).
    destruct H0.
    rewrite e in H0.
    elim (le_Sn_n l).
    omega.
    assumption.
    destruct (MemMod.beq_A addr l).
    reflexivity.
    destruct (H addr) as [H1 [H2 H3]].
    apply H3.
    omega.
Qed.
    
Lemma lt_ge_dec : forall n low high,
  {n < low} + {n >=low /\ n < high} + {n>=high}.
Proof.
intros.
destruct (le_lt_dec low n).
destruct (le_lt_dec high n).
right.
omega.
left;right;omega.
left;left;assumption.
Qed.

Lemma leadsToBlk_intro : forall M l size,
  (forall off, off < size -> leadsTox (l+off) M)
  -> leadsToBlk l size M.
Proof.
  intros.
  unfold leadsToBlk.
  unfold imprecM.
  unfold mstar.  
  exists (fun addr =>
    match (lt_ge_dec addr l (l+size)) with
      | inleft (left _) => None
      | inleft (right _) => M addr
      | inright _ => None
    end).
  exists (fun addr => 
    match (lt_ge_dec addr l (l+size)) with
      | inleft (left _) => M addr
      | inleft (right _) => None
      | inright _ => M addr
    end).
  split.
  split.
  unfold MemMod.disjoint.
  intro x.
  unfold MemMod.not_in_dom.
  destruct (lt_ge_dec x l (l+size)).
  destruct s.
  left;reflexivity.
  right;reflexivity.
  left;reflexivity.
  unfold MemMod.merge.
  apply (extensionality MemMod.A (option MemMod.B) M
    (fun x : MemMod.A =>
    match
      match lt_ge_dec x l (l + size) with
      | inleft s => if s then None (A:=MemMod.B) else M x
      | inright _ => None (A:=MemMod.B)
      end
    with
    | Some _ =>
        match lt_ge_dec x l (l + size) with
        | inleft s => if s then None (A:=MemMod.B) else M x
        | inright _ => None (A:=MemMod.B)
        end
    | None =>
        match lt_ge_dec x l (l + size) with
        | inleft s => if s then M x else None (A:=MemMod.B)
        | inright _ => M x
        end
    end)).
  intro x.
  destruct (lt_ge_dec x l (l+size)).
  destruct s.
  reflexivity.
  destruct (M x);reflexivity.
  reflexivity.
split.
  apply ptoxBlk_intro.
  split;intros;destruct (lt_ge_dec addr l (l+size)).
    destruct s.
    reflexivity.
    destruct a.
    elim (le_not_lt l addr);omega.
    reflexivity.
    destruct s.
    split;intros.
    elim (le_not_lt l addr);omega.
    reflexivity.
    split;intros.
    assert (leadsTox (l+(addr-l)) M).
    apply H;omega.
    replace (l+(addr-l)) with addr in H1.
    apply leadsTox_elim;assumption.
    omega.
    elim (le_not_lt (l+size) addr);omega.
    split.
    intro.
    elim (le_not_lt (l+size) addr);omega.
    intros;reflexivity.
  unfold mTrue;trivial.
Qed.

Lemma leadsToBlk_elim : forall l offset size M,
  leadsToBlk l size M
  -> offset <  size
  -> leadsTox (l+offset) M.
Proof.
  intros.
  unfold leadsToBlk in H.
  destruct H as [M1 [M2 [HJM H]]].
  destruct H.
  assert (H2 := ptoxBlk_indom H H0).
  unfold MemMod.in_dom in H2.
  destruct H2.
  apply leadsTox_intro with x.
  unfold MemMod.disjUnion in HJM.
  destruct HJM.
  subst.
  unfold MemMod.merge.
  rewrite H2.
  reflexivity.
Qed.

Lemma leadsToBlk_Mono : forall size l M1 M2 M,
  MemMod.disjUnion M1 M2 M
  -> leadsToBlk l size M1
  -> leadsToBlk l size M.
Proof.
  unfold leadsToBlk.
  unfold imprecM, mstar, mTrue.
  intros.
  destruct H0 as [M11 [M12 [HJM1 [Hblk Ht]]]].
  exists M11; exists (MemMod.merge M12 M2).
  rsplit.
  split_lr.
  destruct H.
  destruct HJM1. 
  subst.
  split.
    apply MemMod.disj_merge_A.
    trivial.
    eapply MemMod.disj_merge_disj_left; eauto.
  rewrite MemMod.merge_assoc.
  trivial.
Qed.

Lemma leadsToBlk_shrink_low : forall l size off M,
  leadsToBlk l size M
  -> off  <= size
  -> leadsToBlk (l+off) (size - off) M.
Proof.
  unfold leadsToBlk.
  intros.
  destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
  arith_rep size with (off + (size - off)) in HM1.
  assert (Hx := ptoxBlk_split HM1).
  destruct Hx as [M11 [M12 [HJM1 [HM11 HM12]]]].
  unfold imprecM.
  exists M12; exists (MemMod.merge M11 M2).
  split.
    disjUnion shift_l.
    merge commut M12.
    merge rewrite HJM1.
    trivial.
  auto.
Qed.

Lemma leadsToBlk_shrink_high : forall l size off M,
  leadsToBlk l size M
  -> off  <= size
  -> leadsToBlk l off M.
Proof.
  unfold leadsToBlk.
  intros.
  destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
  arith_rep size with (off + (size - off)) in HM1.
  assert (Hx := ptoxBlk_split HM1).
  destruct Hx as [M11 [M12 [HJM1 [HM11 HM12]]]].
  unfold imprecM.
  exists M11; exists (MemMod.merge M12 M2).
  split.
    disjUnion shift_l.
    merge rewrite HJM1.
    trivial.
  auto.
Qed.

(* MemUpd16 *)

Lemma leadsToW_MemUpd16_leadsToW : forall l  x' M M',
  MemUpd16 M l x' M'
  -> leadsToW l x' M'.
Proof.
  unfold MemUpd16.
  intros.
  destruct H.
  subst.
  unfold MemMod.put.
  apply leadsToW_intro with x.
  destruct (MemMod.beq_A l (l+1)).
  substH e with (plus_minus l l 1 e).
  rewrite <- (minus_n_n l) in e.
  discriminate e.
  destruct (MemMod.beq_A l l).
  reflexivity.
  elim n0;reflexivity.
  destruct (MemMod.beq_A (l+1) (l+1)).
    reflexivity.
    elim n ;reflexivity.
Qed.

(* non-touch *)
Lemma leadsToW_MemUpd16_leadsToW_nt : forall l l' x x' M M',
  leadsToW l x M
  -> l' <l-1 \/ l' > l+1
  -> MemUpd16 M l' x' M'
  -> leadsToW l x M'.
Proof.
unfold MemUpd16.
intros.
assert (H2:=leadsToW_elim H).
assert (H3:=leadsToW_elim1 H).
destruct H3.
destruct H1.
rewrite H1.
unfold MemMod.put.
apply leadsToW_intro with x0.
destruct (MemMod.beq_A l (l'+1)).
rewrite e in H0.
destruct H0.
replace (l'+1-1) with l' in H0.
elim (lt_irrefl l');assumption.
omega.
assert (~(l'+1+1<l')).
apply lt_asym.
omega.
elim H4;omega.
destruct (MemMod.beq_A l l').
rewrite e in H0.
assert (~( l' < l' - 1 \/ l' > l' + 1)).
omega.
elim H4;assumption.
assumption.
destruct (MemMod.beq_A (l+1) (l'+1)).
assert (l=l').
omega.
rewrite H4 in H0.
assert (~( l' < l' - 1 \/ l' > l' + 1)).
omega.
elim H5;assumption.
destruct (MemMod.beq_A (l+1) l').
rewrite <- e in H0.
assert (~(l+1<l-1\/ l+1>l+1)).
omega.
elim H4;assumption.
assumption.
Qed.

(*these two is unnecessary**)

Lemma leadsToxW_MemUpd16_leadsToxW : forall l x M M',
  leadsToxW l M
  -> MemUpd16 M  l x M'
  -> leadsToxW l M'.
Proof.
intros.
assert (leadsToW l x M').
apply leadsToW_MemUpd16_leadsToW with M;assumption.
apply leadsToW_leadsToxW with x;assumption.
Qed.

Lemma leadsToxW_MemUpd16_leadsToW : forall l x M M',
  leadsToxW l M
  -> MemUpd16 M l x M'
  -> leadsToW l x M'.
Proof.
intros.
apply leadsToW_MemUpd16_leadsToW with M;assumption.
Qed.

(* non-touch *)
Lemma leadsToxW_MemUpd16_leadsToxW_nt : forall l l' x M M',
  leadsToxW l M
  -> l' <l-1 \/ l' > l+1
  -> MemUpd16 M l' x M'
  -> leadsToxW l M'.
Proof.
intros.
 destruct (leadsToxW_leadsToW H).
assert (H3:=leadsToW_MemUpd16_leadsToW_nt H2 H0 H1).
apply leadsToW_leadsToxW with x0;assumption.
Qed.

Lemma leadsToxW_MemUpd16_eqDom : forall l x M M',
  leadsToxW l M
  -> MemUpd16 M l x M'
  -> MemMod.eqDom M M'.
Proof.
  intros.
  destruct (leadsToxW_elim H) as [x0 H2].
  unfold MemUpd16 in H0. 
  destruct H0.
  subst M'.
  split; intros.
    assert (MemMod.sub_dom M (MemMod.put M l x)).
      auto with PMap.
    assert (MemMod.sub_dom (MemMod.put M l x) (MemMod.put (MemMod.put M l x) (l + 1) x1)).
      auto with PMap.
      eapply MemMod.subDom_trans; eauto.
  assert (MemMod.in_dom l M).
    unfold MemMod.in_dom.
    exists x0; trivial.
  assert (MemMod.sub_dom (MemMod.put M l x) M).
    auto with PMap.
  destruct (leadsToxW_elim1 H) as [x2 H3].
  assert (MemMod.in_dom (l+1) (MemMod.put M l x)).
    unfold MemMod.put.
    unfold MemMod.in_dom in H2 |- * .
    exists x2.
    destruct (MemMod.beq_A (l + 1) l); trivial.
    assert (forall n, ~ (n + 1 = n)).
      induction n; simpl.
      auto with arith.
      intro Hx.
      injection Hx.
      intro Hy.
      destruct (IHn Hy).
    destruct (H4 l e).
  assert (MemMod.sub_dom (MemMod.put (MemMod.put M l x) (l + 1) x1) (MemMod.put M l x)).
    auto with PMap.
  eapply MemMod.subDom_trans; eauto.
Qed.
Implicit Arguments leadsToxW_MemUpd16_eqDom [l x M M'].

Lemma leadsToxW_MemUpd16_Mono : forall l x M1 M2 M M1',
  MemMod.disjUnion M1 M2 M
  -> leadsToxW l M1
  -> MemUpd16 M1 l x M1'
  -> MemMod.disjUnion M1' M2 (MemMod.merge M1' M2)
    /\ MemUpd16 M l x (MemMod.merge M1' M2).
Proof.
  unfold MemUpd16.
  intros.
  split.
    destruct H1.
    subst M1'.
    rewrite <- MemMod.put_merge_merge_put.
    rewrite <- MemMod.put_merge_merge_put.
    apply MemMod.extend_presv_disjUnion_left.
    apply MemMod.extend_presv_disjUnion_left.
    destruct H; subst; split; auto.
    destruct H.
    assert (MemMod.in_dom l M1).
      destruct (leadsToxW_elim H0).
      exists x1; trivial.
    unfold MemMod.disjoint in H.
    destruct (H l); trivial.
    unfold MemMod.in_dom, MemMod.not_in_dom in * .
    destruct H2.
    rewrite H2 in H3; discriminate.
    assert (MemMod.in_dom (l+1) M1).
      destruct (leadsToxW_elim1 H0).
      exists x1; trivial.
    destruct H.
    unfold MemMod.disjoint in H.
    destruct (H (l+1)); trivial.
    unfold MemMod.in_dom, MemMod.not_in_dom in * .
    destruct H1.
    rewrite H1 in H3; discriminate.
  destruct H1.
  subst M1'.
  exists x0.
  destruct H.
  subst M.
  rewrite <- MemMod.put_merge_merge_put.
  rewrite <- MemMod.put_merge_merge_put.
  trivial.
Qed.
Implicit Arguments leadsToxW_MemUpd16_Mono [l x M1 M2 M M1'].

Lemma leadsToxW_MemUpd16_DMono : forall l x M1 M2 M M',
  MemMod.disjUnion M1 M2 M
  -> leadsToxW l M1
  -> MemUpd16 M l x M'
  -> exists M1', MemMod.disjUnion M1' M2 M'
    /\ MemUpd16 M1 l x M1'.
Proof.
  unfold MemUpd16.
  intros.
  destruct H1 as [x1 H1].
  exists (MemMod.put (MemMod.put M1 l x) (l+1) x1).
  destruct (leadsToxW_elim H0).
  destruct (leadsToxW_elim1 H0).
  assert (MemMod.in_dom (l+1) (MemMod.put M1 l x)).
    unfold MemMod.put.
    unfold MemMod.in_dom in H3 |- * .
    exists x2.
    destruct (MemMod.beq_A (l + 1) l); trivial.
    assert (forall n, ~ (n + 1 = n)).
      induction n; simpl.
      auto with arith.
      intro Hx.
      injection Hx.
      intro Hy.
      destruct (IHn Hy).
    destruct (H4 l e).
  split.
    subst M'.
    apply MemMod.extend_presv_disjUnion_left.
    apply MemMod.extend_presv_disjUnion_left.
    trivial.
    destruct H.
    unfold MemMod.disjoint in H.
    destruct (H l); trivial.
    unfold MemMod.in_dom, MemMod.not_in_dom in * .
    rewrite H2 in H5; discriminate.
    destruct H.
    unfold MemMod.disjoint in H.
    destruct (H (l+1)); trivial.
    unfold MemMod.in_dom, MemMod.not_in_dom in * .
    rewrite H3 in H5; discriminate.
  exists x1.
  trivial.
Qed.
Implicit Arguments leadsToxW_MemUpd16_DMono [l x M1 M2 M M'].

Lemma leadsToxW_MemUpd16_DMonoR : forall l x M1 M2 M M',
  MemMod.disjUnion M2 M1 M
  -> leadsToxW l M1
  -> MemUpd16 M l x M'
  -> exists M1', MemMod.disjUnion M2 M1' M'
    /\ MemUpd16 M1 l x M1'.
Proof.
  intros.
  assert (MemMod.disjUnion M1 M2 M); auto with PMap.
  destruct (leadsToxW_MemUpd16_DMono H2 H0 H1).
  destruct H3.
  exists x0; auto with PMap.
Qed.
Implicit Arguments leadsToxW_MemUpd16_DMonoR [l x M1 M2 M M'].

Lemma disjoint_put_disjoint : forall l x M1 M2,
  MemMod.in_dom l M1
  -> MemMod.disjoint M1 M2
  -> MemMod.disjoint (MemMod.put M1 l x) M2.
Proof.
  unfold MemMod.in_dom, MemMod.disjoint, MemMod.put, MemMod.not_in_dom.
  intros.
  destruct H.
  destruct (MemMod.beq_A x0 l).
    subst x0.
    destruct (H0 l).
      rewrite H in H1.
      discriminate H1.
    right; trivial.
  destruct (H0 x0).
    left; trivial.
  right; trivial.
Qed.

Lemma indom_indom_put : forall l l' x' M,
  MemMod.in_dom l M
  -> MemMod.in_dom l (MemMod.put M l' x').
Proof.
  unfold MemMod.in_dom, MemMod.put.
  intros.
  destruct H.
  destruct (MemMod.beq_A l l').
    exists x'; trivial.
  exists x; trivial.
Qed.

(*****added by wangwei******)
(** the leadsToBlk_MemUpd16_eqDom lemma can be omitted **)
(** because of this lemma *******************************)
Lemma leadsToBlk_MemUpd16_eqDom2 : forall l l1 x M M' size,
  leadsToBlk l size M
  -> size >=2
  -> l1>=l
  -> l1<=l+size-2
  -> MemUpd16 M l1 x M'
  -> MemMod.eqDom M M'.
Proof.
unfold MemUpd16.
intros.
unfold MemMod.eqDom.
unfold MemMod.sub_dom.
unfold MemMod.in_dom.
destruct H3.
subst.
unfold MemMod.put.
split.
(*** --> **)
intros.
destruct H3.
destruct (eq_nat_dec x1 (l1+1)).
exists x0.
rewrite e.
destruct (MemMod.beq_A (l1+1) (l1+1)).
reflexivity.
elim n;reflexivity.
destruct (eq_nat_dec x1 l1).
exists x.
destruct (MemMod.beq_A x1 (l1+1)).
elim n;assumption.
destruct (MemMod.beq_A x1 l1).
reflexivity.
elim n1;assumption.
exists x2.
destruct (MemMod.beq_A x1 (l1+1)).
elim n;assumption.
destruct (MemMod.beq_A x1 l1).
elim n0;assumption.
assumption.
(*** <-- ***)
intros.
destruct H3.
destruct (MemMod.beq_A x1 (l1+1)).
rewrite e.
apply leadsTox_elim.
replace (l1+1) with (l+ (l1+1-l)).
apply leadsToBlk_elim with size.
assumption.
omega.
omega.
destruct (MemMod.beq_A x1 l1).
rewrite e.
apply leadsTox_elim.
replace l1 with (l+(l1-l)).
apply leadsToBlk_elim with size.
assumption.
omega.
omega.
exists x2.
assumption.
Qed.

(***added by wangwei ***)
Lemma leadsToBlk_MemUpd16_leadsToBlk: forall l x M M' size,
  leadsToBlk l size M
  -> size >=2
  -> MemUpd16 M l x M'
  -> leadsToBlk l size M'.
Proof.
intros.
assert (MemMod.eqDom M M').
apply leadsToBlk_MemUpd16_eqDom2 with l l x size;try assumption.
auto.
omega.
apply leadsToBlk_intro.
intros.
assert (leadsTox (l+off) M).
apply leadsToBlk_elim with size;assumption.
assert (MemMod.in_dom (l+off) M').
assert (MemMod.in_dom (l+off) M).
unfold MemMod.in_dom.
apply leadsTox_elim;assumption.
destruct H2.
apply H2;assumption.
destruct H5.
apply leadsTox_intro with x0.
assumption.
Qed.

Lemma leadsToBlk_MemUpd16_Mono : forall l x size M1 M2 M M1',
  MemMod.disjUnion M1 M2 M  
  -> leadsToBlk l size M1
  -> MemUpd16 M1 l x M1'
  -> size >= 2
  -> MemMod.disjUnion M1' M2 (MemMod.merge M1' M2)
  /\ leadsToBlk l size M1'
  /\ MemUpd16 M l x (MemMod.merge M1' M2).
Proof.
intros.
destruct H.
assert (MemMod.eqDom M1 M1').
apply leadsToBlk_MemUpd16_eqDom2 with l l x size;try assumption.
auto.
omega.
split.
split.
apply eqDom_disjoint_left with M1;assumption.
reflexivity.
split.
apply leadsToBlk_MemUpd16_leadsToBlk with x M1;assumption.
unfold MemUpd16 in * .
destruct H1.
exists x0.
subst.
apply (extensionality MemMod.A (option MemMod.B)).
intro x1.
unfold MemMod.merge.
unfold MemMod.put.
destruct (MemMod.beq_A x1 (l+1)).
reflexivity.
destruct (MemMod.beq_A x1 l).
reflexivity.
reflexivity.
Qed.
Implicit Arguments leadsToBlk_MemUpd16_Mono [l x size M1 M2 M M1'].

Lemma leadsToBlk_MemUpd16_DMono : forall l x M1 M2 M M',
  MemMod.disjUnion M1 M2 M
  -> leadsToBlk l 2 M1
  -> MemUpd16 M l x M'
  -> exists M1', MemMod.disjUnion M1' M2 M'
    /\ MemUpd16 M1 l x M1'.
Proof.
intros.
destruct H1.
exists (MemMod.put (MemMod.put M1 l x) (l+1) x0).
split.
destruct H.
subst.
split.
apply MemMod.update_presv_disj_left.
apply MemMod.update_presv_disj_left.
assumption.
unfold MemMod.in_dom.
apply leadsTox_elim.
replace l with (l+0).
apply leadsToBlk_elim with 2.
assumption.
auto.
auto.
assert (MemMod.in_dom (l+1) M1).
unfold MemMod.in_dom.
apply leadsTox_elim.
apply leadsToBlk_elim with 2.
assumption.
auto.
destruct H1.
exists x1.
unfold MemMod.put.
destruct (MemMod.beq_A (l+1) l).
assert (l+1<>l).
omega.
elim H2;assumption.
assumption.
apply (extensionality MemMod.A (option MemMod.B)). 
intro x1.
unfold MemMod.put.
unfold MemMod.merge.
destruct (MemMod.beq_A x1 (l+1)).
reflexivity.
destruct (MemMod.beq_A x1 l).
reflexivity.
reflexivity.
exists x0.
reflexivity.
Qed.
Implicit Arguments leadsToBlk_MemUpd16_DMono [l x M1 M2 M M'].

Lemma leadsToBlk_MemUpd16_eqDom : forall l x M M',
  leadsToBlk l 2 M
  -> MemUpd16 M l x M'
  -> MemMod.eqDom M M'.
Proof.
intros.
apply leadsToBlk_MemUpd16_eqDom2 with l l x 2;
try assumption.
auto.
auto.
omega.
Qed.

Implicit Arguments leadsToBlk_MemUpd16_eqDom [l x M M'].


(*******************************************)

(* unique *)

(*******************************************)

Lemma leadsTo_unique : forall addr x x' M,
  leadsTo addr x M
  -> leadsTo addr x' M
  -> x = x'.
Proof.
  intros.
  assert (Hx := leadsTo_elim H0).
  rewrite (leadsTo_elim H) in Hx.
  injection Hx; trivial.
Qed.
Implicit Arguments leadsTo_unique [addr x x' M].


Lemma leadsToW_unique_disjUnion_l : forall addr x x' M1 M2 M,
  MemMod.disjUnion M1 M2 M
  -> leadsToW addr x M1
  -> leadsToW addr x' M
  -> x = x'.
Proof.
  intros.
  assert (Hx := leadsToW_elim H0).
  assert (Hx' := leadsToW_elim H1).
  destruct H.
  subst M.
  unfold MemMod.merge in Hx'.
  rewrite Hx in Hx'.
  injection Hx'; trivial.
Qed.
Implicit Arguments leadsToW_unique_disjUnion_l [addr x x' M1 M2 M].

(************************************************)
(*  stkSpace                                    *)

Lemma precise_stkSpacePrec : forall sp size, precise (stkSpacePrec sp size).
Proof.
  unfold stkSpacePrec.
  unfold precise.
  intros sp size.
  intros.
  destruct H1; destruct H2.
  assert (Hx := @precise_ptoxBlk (sp - size) size).
  unfold precise in Hx.
  eapply Hx; eauto.
Qed.
Implicit Arguments precise_stkSpacePrec [sp size].

Lemma stkSpace_stkSpacePrec : forall sp size M,
  stkSpace sp size M
  -> mstar mTrue (stkSpacePrec sp size) M.
Proof.
intros.
unfold stkSpacePrec.
unfold stkSpace in H.
unfold leadsToBlk in H.
unfold imprecM in H.
unfold mstar in * .
InvertAll.
exists x0;exists x.
intuition.
Qed.
Implicit Arguments stkSpace_stkSpacePrec [sp size M].

Lemma stkSpacePrec_stkSpace : forall sp size M,
  stkSpacePrec sp size M
  -> stkSpace sp size M.
Proof.
intros.
unfold stkSpace.
unfold stkSpacePrec in H.
destruct H.
intuition.
unfold leadsToBlk.
unfold imprecM.
unfold mstar.
exists M;exists MemMod.emp.
intuition.
unfold mTrue;trivial.
Qed.

Implicit Arguments stkSpacePrec_stkSpace [sp size M].

Lemma stkSpace_weaken : forall M sp s1 s2,
  stkSpace sp s1 M
  -> s2 <= s1 
  -> stkSpace sp s2 M.
Proof.
unfold stkSpace.
intros.
destruct H.
intuition.
replace s2 with (s1 - (s1-s2)).
replace (sp-(s1-(s1-s2)))with ((sp-s1)+(s1-s2)).
apply leadsToBlk_shrink_low.
assumption.
omega.
omega.
omega.
Qed.
Implicit Arguments stkSpace_weaken [M sp s1 s2].

Lemma stkSpacePrec_split : forall sp s1 s2 M,
  stkSpacePrec sp (s1 + s2) M
  -> mstar (stkSpacePrec sp s2) (stkSpacePrec (sp-s2) s1) M.
Proof.
unfold stkSpacePrec. 
intros.
destruct H.
assert ((ptoxBlk (sp -(s1+s2)) s1  *^* ptoxBlk (sp-(s1+s2)+s1) s2 ) M).
apply ptoxBlk_split;assumption.
unfold mstar in H1;InvertAll.
exists x0;exists x;intuition.
arith_rep (sp-s2) with (sp-(s1+s2)+s1).
arith_rep (sp-s2-s1) with (sp-(s1+s2)).
Qed.

Implicit Arguments stkSpacePrec_split [sp s1 s2 M].

Lemma stkSpacePrec_merge : forall sp s1 s2 M1 M2,
  stkSpacePrec sp s2 M1 
  -> stkSpacePrec (sp-s2) s1 M2
  -> stkSpacePrec sp (s1 + s2) (MemMod.merge M1 M2).
Proof.
unfold stkSpacePrec.
intros.
InvertAll.
intuition.
apply ptoxBlk_merge.
assert (MemMod.disjoint M2 M1).
unfold MemMod.disjoint.
unfold MemMod.not_in_dom.
intros.
destruct (le_lt_dec (sp-s2) x).
left;apply ptoxBlk_notindom with (sp-s2-s1) s1.
assumption.
right;omega.
right;apply ptoxBlk_notindom with (sp-s2) s2.
assumption.
left;assumption.
exists M2;exists M1;intuition.
arith_rep (sp-(s1+s2)) with (sp-s2-s1).
arith_rep (sp-(s1+s2)+s1) with (sp-s2).
Qed.

Implicit Arguments stkSpacePrec_merge [sp s1 s2 M1 M2].

Lemma stkSpacePrec_downup : forall sp size d M,
  stkSpacePrec sp (size + d) M
  -> stkSpacePrec (sp - d + d) (size + d) M.
Proof.
unfold stkSpacePrec.
intros.
intuition.
arith_rep (sp-d+d-(size+d)) with (sp-(size+d)).
Qed.

Implicit Arguments stkSpacePrec_downup [sp size d M].

Lemma stkSpacePrec_leadsToxW : forall M sp s l,
  stkSpacePrec sp s M
  -> s >= 2
  -> l >= (sp - s)
  -> l <= sp - 2
  -> leadsToxW l M.
Proof.
unfold stkSpacePrec.
intros.
destruct H.
assert (MemMod.in_dom ((sp-s)+(l-(sp-s))) M).
apply ptoxBlk_indom with s.
assumption.
omega.
assert (MemMod.in_dom ((sp-s)+(l+1-(sp-s))) M).
apply ptoxBlk_indom with s.
assumption.
omega.
replace (sp-s+(l-(sp-s))) with l in H4.
replace (sp-s+(l+1-(sp-s))) with (l+1) in H5.
destruct H4.
destruct H5.
apply leadsToxW_intro with x x0;assumption.
omega.
omega. 
Qed.

Lemma stkSpacePrec_leadsToBlk : forall M sp s,
  stkSpacePrec sp s M
  -> leadsToBlk (sp - s) s M.
Proof.
unfold stkSpacePrec. 
intros.
unfold leadsToBlk.
unfold imprecM.
destruct H.
exists M;exists MemMod.emp.
split.
auto with PMap.
split.
assumption.
unfold mTrue.
trivial.
Qed.

Lemma stkSpacePrec_ptoxBlk : forall M sp s,
  stkSpacePrec sp s M
  -> ptoxBlk (sp - s) s M.
Proof.
unfold stkSpacePrec.
intros.
destruct H.
assumption.
Qed.

Lemma stkSpacePrec_MemUpd16_eqDom : forall M sp s l x M',
  stkSpacePrec sp s M
  -> s >= 2
  -> l >= (sp - s)
  -> l <= sp - 2
  -> MemUpd16 M l x M'
  -> MemMod.eqDom M M'.
Proof.
unfold stkSpacePrec.
intros.
destruct H.
apply leadsToBlk_MemUpd16_eqDom2 with (sp-s) l x s;try assumption.
unfold leadsToBlk.
apply imprecM_weaker.
assumption.
omega.
Qed.

Lemma stkSpacePrec_stkSpacePrec_eqDom : forall M M' sp s,
  stkSpacePrec sp s M
  -> stkSpacePrec sp s M'
  -> MemMod.eqDom M M'.
Proof.
intros.
unfold MemMod.eqDom.
unfold MemMod.sub_dom.
destruct H.
destruct H0.
split.
intro.
destruct (lt_ge_dec x (sp-s) sp).
destruct s0.
assert (M x = None).
apply ptoxBlk_notindom with (sp-s) s.
assumption.
left;assumption.
unfold MemMod.in_dom.
intro.
destruct H4.
rewrite H3 in H4;discriminate H4.
assert (MemMod.in_dom (sp-s+(x-(sp-s))) M').
apply ptoxBlk_indom with s.
assumption.
omega.
replace (sp-s+(x-(sp-s))) with x in H3.
intro;assumption.
destruct a.
apply le_plus_minus.
omega.
assert (M x = None).
apply ptoxBlk_notindom with (sp-s) s.
assumption.
right.
omega.
unfold MemMod.in_dom.
intro.
destruct H4.
rewrite H3 in H4;discriminate H4.
intro.
destruct (lt_ge_dec x (sp-s) sp).
destruct s0.
assert (M' x = None).
apply ptoxBlk_notindom with (sp-s) s.
assumption.
left;assumption.
unfold MemMod.in_dom.
intro.
destruct H4.
rewrite H3 in H4;discriminate H4.
assert (MemMod.in_dom (sp-s+(x-(sp-s))) M).
apply ptoxBlk_indom with s.
assumption.
omega.
replace (sp-s+(x-(sp-s))) with x in H3.
intro;assumption.
destruct a.
apply le_plus_minus.
omega.
assert (M' x = None).
apply ptoxBlk_notindom with (sp-s) s.
assumption.
right.
omega.
unfold MemMod.in_dom.
intro.
destruct H4.
rewrite H3 in H4;discriminate H4.
Qed.

Lemma stkSpacePrec_MemUpd16_stkSpacePrec : forall M sp s l x M',
  stkSpacePrec sp s M
  -> s >= 2
  -> l >= (sp - s)
  -> l <= sp - 2
  -> MemUpd16 M l x M'
  -> stkSpacePrec sp s M'.
Proof.
intros.
intuition.
assert (MemMod.eqDom M M').
apply stkSpacePrec_MemUpd16_eqDom with sp s l x;
try assumption.
unfold stkSpacePrec in * .
destruct H.
intuition.
apply ptoxBlk_intro.
intros.
split.
intro.
assert (MemMod.not_in_dom addr M').
apply MemMod.eqDom_notin_presv with M.
assumption.
unfold MemMod.not_in_dom.
apply ptoxBlk_notindom with (sp-s) s.
assumption.
left;assumption.
unfold MemMod.not_in_dom in H7;assumption.
split.
intros.
destruct H4.
apply H4.
replace addr with ((sp-s)+(addr - (sp-s))).
apply ptoxBlk_indom with s.
assumption.
destruct H6.
clear H6.
omega.
destruct H6.
omega.
intro.
assert (MemMod.not_in_dom addr M').
apply MemMod.eqDom_notin_presv with M.
assumption.
unfold MemMod.not_in_dom.
apply ptoxBlk_notindom with (sp-s) s.
assumption.
right;assumption.
unfold MemMod.not_in_dom in H7;assumption.
Qed.

Lemma stkSpace_Mono : forall sp size M1 M2 M,
  MemMod.disjUnion M1 M2 M
  -> stkSpace sp size M1
  -> stkSpace sp size M.
Proof.
intros.
unfold stkSpace in * .
destruct H0;intuition.
apply leadsToBlk_Mono with M1 M2;auto with PMap.
Qed.
Implicit Arguments stkSpace_Mono [sp size M1 M2 M].

Lemma stkSpace_Mono_r : forall sp size M1 M2 M,
  MemMod.disjUnion M1 M2 M
  -> stkSpace sp size M2
  -> stkSpace sp size M.
Proof.
intros.
unfold stkSpace in * .
destruct H0;intuition.
apply leadsToBlk_Mono with M2 M1;auto with PMap.
Qed.
Implicit Arguments stkSpace_Mono [sp size M1 M2 M].


(********************************************************)

(* The two lemmas below are used in Itr_soundness file. *)


Lemma leadsToxW_leadsToW_DMono : forall l v M1 M2 M,
  MemMod.disjUnion M1 M2 M
  -> leadsToxW l M1
  -> leadsToW l v M
  -> leadsToW l v M1.
Proof.
intros.
assert (H2:=leadsToxW_leadsToW H0).
destruct H2.
assert (x=v).
apply leadsToW_unique_disjUnion_l with l M1 M2 M;
assumption.
rewrite <- H3;assumption.
Qed.
Implicit Arguments leadsToxW_leadsToW_DMono [l v M1 M2 M].

Lemma leadsToBlk_leadsToxW : forall l size M,
  leadsToBlk l size M
  -> size > 1
  -> leadsToxW l M.
Proof.
intros.
assert (leadsTox l M).
replace l with (l+0).
apply leadsToBlk_elim with size.
assumption.
omega.
omega.
assert (leadsTox (l+1) M).
apply leadsToBlk_elim with size.
assumption.
auto.
substH H1 with (leadsTox_elim H1).
substH H2 with (leadsTox_elim H2).
destruct H1;destruct H2.
apply leadsToxW_intro with x x0;assumption.
Qed.
Implicit Arguments leadsToBlk_leadsToxW [l size M].

Unset Implicit Arguments.

Lemma nat_eq_neq_dec:
  forall n n' : nat, n = n' \/ n <> n'.
Proof.
intros.
destruct (eq_nat_dec n n').
left;assumption.
right;assumption.
Qed.


Lemma mstar_presv_uniqueM:
forall m1 m2, uniqueM m1 -> uniqueM m2 -> uniqueM (mstar m1 m2).
Proof.
  intros.
  unfold uniqueM in *.
  intros.
  unfold mstar in *.
  InvertAll.
  assert (x = x1). apply H; auto.
  assert (x0 = x2). apply H0; auto.
  inversion_clear H3. inversion_clear H4.
  subst M M'.
  subst x x0. auto.
Qed.


Lemma pto_uniqueM:
  forall l v, uniqueM (l |-> v).
Proof.
  unfold uniqueM.
  unfold pto.
  unfold MemMod.sgltonM.
  unfold MemMod.not_in_dom.
  intros.
  unfold MemMod.maps_to in *.
  InvertAll.
  apply MemMod.extensional_eqMap.
  intro.
  assert (x = l \/ x <> l).
  apply nat_eq_neq_dec.
  inversion H.
  subst x. rewrite H0. rewrite H1. auto.
  assert (M x = None). apply H3; auto.
  rewrite H5. assert (M' x = None).
  apply H2; auto.
  rewrite H6. auto.
Qed.

Lemma pto_precise:
forall l v, precise (l |-> v).
Proof.
  intros.
  unfold precise.
  intros.
  assert (uniqueM (l |-> v)).
  apply pto_uniqueM.
  unfold uniqueM in H3.
  apply H3; auto.
Qed.

