Require Import Coqlib.
Require Import Map.
Require Import Arith.
Require Import Machine.
Require Import Assertion.
Require Import Omega.

Require Import Mem_Lemmas.
Require Import Assertion_Lemmas.

Set Implicit Arguments.
Unset Strict Implicit.

(* ************************************************)
(* portage                                        *)

Definition mjudge (h:Mem) (A:MPred) := A h.

Notation "M ||- A" := (mjudge M A) (at level 80).

Notation Local " A ** B " := (mstar A B) 
  (at level 67, right associativity).

Notation Local "A ==> B" :=
  (forall M, M ||- A -> M ||- B) (at level 13).

Notation Local "A <==> B" :=
  (forall M, M ||- A <-> M ||- B) (at level 13).

Implicit Types A B C D : MPred.

Implicit Types M h : Mem.

Notation Local sconj := mstar (only parsing).

Notation Local join := MemMod.disjUnion (only parsing).

Notation Local emp := MemMod.emp (only parsing).

Notation Local meq := mEq (only parsing).

Notation Local unit := Aemp (only parsing).

Lemma join_comm : forall M1 M2 M,
  join M1 M2 M -> join M2 M1 M.
Proof.
  intros M1 M2 M.
  auto with PMap.
Qed.

Definition bang (A:Prop) : MPred :=
  fun h => unit h /\ A.

Require Import List.
Fixpoint itr (tdata : Set) (p : tdata -> MPred) (lst : list tdata) 
  {struct lst} : MPred :=
  match lst with
    | nil => unit
    | cons data lst' => (p data) ** (itr p lst')
  end.

Notation Local "! A" := (bang A) (at level 55).

(* Notation Local pto := pto. *)

(* Notation Local ptoW := ptoW. *)

(* Notation Local ptoxW := ptoxW. *)

Notation Local set := MemMod.put.

Notation Local top := mTrue.

(* Parameter sig : forall (l : Label) (v : nat), Mem. *)


(* ************************************************)
(* mjudge                                         *)

Lemma mjudge_intro : forall M A,
  A M -> M ||- A.
Proof.
  unfold mjudge; trivial.
Qed.

Lemma mjudge_elim : forall M A,
  M ||- A -> A M.
Proof.
  unfold mjudge; trivial.
Qed.

(* ************************************************)
(* sconj                                          *)

Lemma sconj_intro : forall ha hb h A B,
  ha ||- A -> hb ||- B -> join ha hb h ->
    h ||- A ** B.
Proof.
  intros.
  exists ha; exists hb; auto.
Qed.
Implicit Arguments sconj_intro [ha hb h A B].

Lemma sconj_elim : forall M A B,
  M ||- A ** B
    -> exists ha, exists hb, join ha hb M
      /\ (ha ||- A) /\ (hb ||- B).
Proof.
  unfold sconj; trivial.
Qed.

Lemma sconj_sym : forall h A B,
  h ||- A ** B -> h ||- B ** A. 
Proof.  
  intros h A B [ha [hb [J [Ha Hb]]]].
  apply (sconj_intro Hb Ha (join_comm J)).
Qed.

Lemma sconj_mono_left : forall h A B,
  h ||- (A ** B) -> 
    forall A', (A ==> A') -> 
      h ||- (A' ** B).
Proof.
  destruct 1 as [ha [hb [J [HA HB]]]].
  exists ha; exists hb.
  intuition.
  apply H.
  apply HA.
Qed.

Lemma sconj_mono_right : forall h A B,
  h ||- (A ** B) -> 
    forall B', (B ==> B') -> 
      h ||- (A ** B').
Proof.
  destruct 1 as [ha [hb [J [HA HB]]]].
  exists ha; exists hb.
  intuition.
  apply H.
  apply HB.
Qed.

Lemma sconj_mono : forall h A A' B B',
  h ||- A ** B -> 
    (A ==> A') -> (B ==> B') -> 
    h ||- A' ** B'.
Proof.
  intros.
  assert (h ||- (A' ** B)).
  eapply sconj_mono_left; [eauto | assumption].
  eapply sconj_mono_right; [eauto | assumption].
Qed.

Lemma sconj_eqv_left : forall M A B,
  M ||- A ** B -> forall A',
    (A <==> A')
  -> M ||- A' ** B.
Proof.
  intros M A A' B H H0.
  apply sconj_mono_left with A; trivial.
  intro M'.
  destruct (H0 M'); trivial.
Qed.
Implicit Arguments sconj_eqv_left [M A B].

Lemma sconj_assoc_left : forall A B C h,
  h ||- (A ** B) ** C -> h ||- A ** (B ** C).
Proof.
  intros.
  destruct H as [hab [hc [Jabc [HAB HC]]]].
  destruct HAB as [ha [hb [Jab [HA HB]]]].
  apply sconj_intro with ha (MemMod.merge hb hc).
  trivial.
  apply sconj_intro with hb hc; trivial.
  split_r.
  disjoint lift_l.
  split_r.
    disjoint des.
    disjoint lift_l.
  destruct Jabc.
  subst h.
  destruct Jab.
  subst hab.
  merge assoc ha.
  trivial.
Qed.

Lemma sconj_assoc_right : forall A B C h,
  h ||- A ** (B ** C) -> h ||- (A ** B) ** C.
Proof.
  intros.
  apply sconj_sym.
  apply sconj_assoc_left.
  exact (sconj_sym (sconj_assoc_left (sconj_sym H))).
Qed.

(* ************************************************)
(* eqv                                           *)
(*
Lemma eqv_imp : forall A A',
  A <==> A' -> A ==> A'.
Proof.
  intros A A' H M.
  destruct (H M); auto.
Qed.
Implicit Arguments eqv_imp [A A'].

Lemma eqv_refl : forall A, A <==> A.
Proof.
  intros; split; auto.
Qed.

Lemma eqv_sym :
  forall A B, A <==> B -> B <==> A.
Proof.
  intros; split; destruct (H M); auto.
Qed.
Hint Resolve eqv_sym : splogic.

Lemma eqv_trans :
  forall A B C, A <==> B -> B <==> C -> A <==> C.
Proof.
  intros A B C H1 H2; split; destruct (H1 M); destruct (H2 M); auto.
Qed.

Lemma eqv_elim : forall M A A',
    A <==> A' 
  -> M ||- A 
  -> M ||- A'.
Proof.
  intros M A A' H H0; destruct (H M); auto.
Qed.
Implicit Arguments eqv_elim [M A A'].

Lemma eqv_elim1 : forall M A A' B,
    A <==> A' 
  -> M ||- A ** B
  -> M ||- A' ** B.
Proof.
  intros M A A' B H H0.
  destruct (sconj_elim H0) as [M1 [M2 [HJM [HM1 HM2]]]].
  destruct (H M1).
  apply sconj_intro with M1 M2; auto.
Qed.
Implicit Arguments eqv_elim1 [M A A' B].
*)

(* ************************************************)
(* unit                                           *)

Lemma unit_intro : emp ||- unit.
Proof.
  unfold mjudge, unit, Aemp. 
  trivial.
Qed.

Lemma unit_intro_left : forall h A,
  h ||- A -> h ||- unit ** A.
Proof.
  intros.
  eapply sconj_intro.
  2: apply H.
  apply unit_intro.
  unfold join.
  auto with PMap.
Qed.
Implicit Arguments unit_intro_left [h A].

Lemma unit_intro_right : forall h A,
  h ||- A -> h ||- A ** unit.
Proof.
  intros.
  apply sconj_sym.
  apply (unit_intro_left H).
Qed.

Lemma unit_elim : forall h,
  h ||- unit -> h = emp.
Proof.
  intros.
  symmetry.
  unfold emp, unit in H.
  rewrite H.
  trivial.
Qed.

Lemma join_emp_l : forall m1 m : Mem, join emp m1 m -> m1 = m.
Proof.
 intros.
 auto with PMap.
Qed.

Lemma unit_elim_left : forall h A,
  h ||- unit ** A -> h ||- A.
Proof.
  intros.
  destruct H as [h1 [h2 [HJ [H1 H2]]]].
  unfold unit in H1.
  rewrite H1 in HJ.
  rewrite (join_emp_l HJ) in H2.
  trivial.
Qed.
Implicit Arguments unit_elim_left [h A].

Lemma unit_elim_right : forall h A,
  h ||- A ** unit -> h ||- A.
Proof.
  intros.
  apply (unit_elim_left (sconj_sym H)).
Qed.
Implicit Arguments unit_elim_right [h A].

(* ************************************************)
(* top                                            *)

Lemma top_intro : forall h, h ||- top.
Proof.
  unfold mjudge, top, top. 
  tauto.
Qed.

Lemma top_intro_left : forall h A,
  h ||- A -> h ||- top ** A.
Proof.
  intros.
  assert (join emp h h).
  unfold join,emp; auto with PMap.
  apply (sconj_intro (top_intro emp) H H0).
Qed.
Implicit Arguments top_intro_left [h A].

Lemma top_intro_left1 : forall M A B,
  M ||- B ** A -> M ||- top ** A.
Proof.
  intros.
  destruct (sconj_elim H) as [M1 [M2 [HJM [HM1 HM2]]]].
  apply sconj_intro with M1 M2; auto.
  apply top_intro; auto.
Qed.
Implicit Arguments top_intro_left1 [M A B].

Lemma top_intro_right : forall M A,
  M ||- A -> M ||- A ** top.
Proof.
  intros M A H.
  substH H with (top_intro_left H).
  apply sconj_sym; trivial.
Qed.
Implicit Arguments top_intro_right [M A].

Lemma top_intro_right1 : forall M A B,
  M ||- A ** B -> M ||- A ** top.
Proof.
  intros M A B H.
  destruct (sconj_elim H) as [M1 [M2 [HJM [HM1 HM2]]]].
  apply sconj_intro with M1 M2; auto.
  apply top_intro; auto.
Qed.
Implicit Arguments top_intro_right1 [M A B].

(* *********************************************** *)
(* bang                                            *)

Lemma bang_intro : forall h (A:Prop),
  h ||- unit -> A ->
    h ||- ! A.
Proof.
  unfold bang.
  intros h A Hp Aok.
  constructor; assumption.
Qed.

Lemma bang_elim1 : forall h P,
  h ||- ! P -> P.
Proof.
  unfold mjudge, bang.
  intuition.
Qed.

Lemma bang_elim2 : forall h P,
    h ||- ! P -> h ||- unit.
Proof.
  unfold mjudge, bang.
  intuition.
Qed.

Lemma bang_elim : forall h P,
  h ||- ! P ->
    h ||- unit /\ P.
Proof.
  intros h P Hp.
  constructor.
  exact (bang_elim2 Hp).
  exact (bang_elim1 Hp).
Qed.

Lemma bang_elim_left : forall h P B,
    h ||- ! P ** B ->
      P /\ h ||- B.
Proof.
  intros h A B H.
  assert (A /\ h ||- unit ** B).
  destruct H as [ha [hb [J [HA HB]]]].
  constructor.
  apply (bang_elim1 HA).
  assert (HA' := bang_elim2 HA).
  eapply sconj_intro; eauto.
  intuition.
  intuition.
  apply unit_elim_left; trivial.
Qed.

Lemma bang_elim_left1 : forall h P B,
  h ||- ! P ** B ->
    P.
Proof.
  intros h A B H.
  destruct (bang_elim_left H).
  assumption.
Qed.

Lemma bang_elim_left2 : forall h P B,
  h ||- ! P ** B ->
    h ||- B.
Proof.
  intros h A B H.
  destruct (bang_elim_left H).
  assumption.
Qed.

Lemma bang_elim_right : forall h P B,
    h ||- B ** ! P ->
      P /\ h ||- B.
Proof.
  intros h A B H.
  substH H with (sconj_sym H).
  apply bang_elim_left; auto.
Qed.

Lemma bang_intro_left : forall h (P:Prop) B,
  P -> h ||- B ->
    h ||- ! P ** B.
Proof.
  intros h A B HA HB.
  unfold bang.
  assert (Q := unit_intro_left HB).
  eapply sconj_mono; eauto.
  intros M H.
  unfold mjudge in * .
  auto.
Qed.

Lemma bang_intro_right : forall h (P:Prop) B,
  P -> h ||- B ->
    h ||- B ** ! P.
Proof.
  intros.
  apply sconj_sym.
  apply bang_intro_left; auto.
Qed.

(* *********************************************** *)
(* pto                                             *)

Lemma pto_intro : forall M l w,
  (forall l', (l' = l -> M l' = Some w)
    /\ (l' <> l -> M l' = None)) 
  -> M ||- l |-> w.
Proof.
  intros M l w.
  unfold mjudge, pto, MemMod.sgltonM.
  unfold MemMod.maps_to, MemMod.not_in_dom.
  intros.
  split.
    substH H with (H l).
    destruct H.
    apply H; trivial.
  intros l' H0.
  substH H with (H l').
  destruct H.
  apply H1; trivial.
Qed.

Lemma pto_elim : forall M l w,
  (M ||- l |-> w)
  -> (forall l', (l' = l -> M l' = Some w)
    /\ (l' <> l -> M l' = None)).
Proof.
  intros M l w.
  unfold mjudge, pto, MemMod.sgltonM.
  unfold MemMod.maps_to, MemMod.not_in_dom.
  intros.
  split.
    intros.
    subst l'.
    destruct H; trivial.
  intros.
  destruct H.
  apply H1; trivial.
Qed.

Lemma pto_get_eq : forall M l w,
  M ||- l |-> w -> 
    M l = Some w.
Proof.
  intros M l w H.
  substH H with (pto_elim H).
  destruct (H l).
  apply H0; trivial.
Qed.

Lemma pto_get_eq1 : forall M l w A,
  M ||- l |-> w ** A -> 
    M l = Some w.
Proof.
  intros M l w A H.
  destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
  substH HJM with (disjUnion_elim HJM l).
  rewrite (pto_get_eq HM1) in HJM.
  destruct (M2 l).
    contradiction.
  destruct (M l); auto.
  contradiction.
Qed.

Lemma pto_update : forall M l v1 v2,
  M ||- l |-> v1 ->
    set M l v2 ||- l |-> v2.
Proof.
  intros M l v1 v2 H.
  apply pto_intro.
  substH H with (pto_elim H).
  intros l'.
  substH H with (H l').
  unfold MemMod.put.
  destruct (MemMod.beq_A l' l).
  split; intro; trivial.
  destruct (H0 e).
  split; intro; trivial.
  destruct (n H0).
  destruct H.
  auto.
Qed.

Lemma pto_update1 : forall M l v1 A,
  M ||- (l |-> v1 ** A) -> forall v2,
    set M l v2 ||- (l |-> v2 ** A).
Proof.
  intros M l v1 A H v2.
  destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
  exists (set M1 l v2); exists M2.
  split.
    apply MemMod.update_presv_disjUnion_left; trivial.
    substH HM1 with (pto_get_eq HM1).
    unfold MemMod.in_dom.
    exists v1; trivial.
  split_r.
  apply mjudge_elim.
  apply pto_update with v1; trivial.
Qed.

Lemma pto_pto_eq : forall h x v1 v2,
  h ||- x |-> v1 ->
  h ||- x |-> v2 ->
   v1 = v2.
Proof.
  intros M l v1 v2 H1 H2.
  substH H1 with (pto_get_eq H1).
  substH H2 with (pto_get_eq H2).
  rewrite H1 in H2.
  injection H2; trivial.
Qed.

Lemma pto_pto_eq1 : forall h x v1 v2 A B,
  h ||- x |-> v1 ** A ->
    h ||- x |-> v2 ** B ->
      v1 = v2.
Proof.
  intros h x v1 v2 A B H1 H2.
  assert (Some v1 = Some v2).
  rewrite <- (pto_get_eq1 H1).
  rewrite <- (pto_get_eq1 H2).
  reflexivity.
  injection H; auto.
Qed.

(* *********************************************** *)
(* ptox                                             *)

Lemma pto_ptox : forall M l v,
  M ||- l |-> v
  -> M ||- l |->_.
Proof.
  intros M l w H.
  unfold ptox.
  unfold mjudge.
  exists w; trivial.
Qed.

Lemma pto_ptox1 : forall M l v A,
  M ||- l |-> v ** A
  -> M ||- l |->_ ** A.
Proof.
  unfold ptox.
  intros M l v A H.
  destruct (sconj_elim H) as [M1 [M2 [HJM [HM1 HM2]]]].
  apply sconj_intro with M1 M2; auto.
  unfold mjudge in * .
  exists v; trivial.
Qed.

Lemma ptox_pto : forall M l,
  (M ||- l |->_)
  -> exists v, M ||- l |-> v.
Proof.
  intros M l H.
  unfold ptox in H.
  unfold mjudge in H.
  apply H.
Qed.

Lemma ptox_pto1 : forall M l A,
  (M ||- l |->_ ** A)
  -> exists v, M ||- l |-> v ** A.
Proof.
  intros M l A H.
  destruct (sconj_elim H) as [M1 [M2 [HJM [HM1 HM2]]]].
  destruct (ptox_pto HM1).
  exists x.
  eapply sconj_intro; eauto.
Qed.

(* add by Xinyu *)
Lemma ptox_elim1 : forall M l A,
  (M ||- l |->_ *^* A)
  -> exists w, M ||- l |-> w *^* A.
Proof.
  apply ptox_pto1.
Qed.

Lemma ptox_update : forall M l v,
  M ||- l |->_
  -> set M l v ||- l |-> v.
Proof.
  intros M l v H.
  destruct (ptox_pto H).
  apply pto_update with x; trivial.
Qed.

Lemma ptox_update1 : forall M l A,
  M ||- l |->_ ** A
  -> forall v, set M l v ||- l |-> v ** A.
Proof.
  intros M l A H v.
  destruct (ptox_pto1 H).
  apply pto_update1 with x; trivial.
Qed.

(* ********************************************** *)
(* leadsTo                                        *)

Lemma leadsTo_pto : forall M l v,
  M ||- l |~> v
  -> M ||- l |-> v ** top.
Proof.
  intros M l v H.
  unfold mjudge, leadsTo in H.
  unfold imprecM in H.
  apply mjudge_intro; trivial.
Qed.
Hint Resolve leadsTo_pto : splogic.

Lemma leadsTo_pto1 : forall M l v A,
  M ||- l |~> v ** A
  -> M ||- l |-> v ** top ** A.
Proof.
  intros M l v A H.
  unfold leadsTo in H.
  unfold imprecM in H.
  apply sconj_assoc_left; trivial.
Qed.

Lemma pto_leadsTo : forall M l v,
  M ||- l |-> v
  -> M ||- l |~> v.
Proof.
  intros M l v H.
  unfold leadsTo, imprecM.
  eapply top_intro_right; eauto.
Qed.
Hint Resolve pto_leadsTo : splogic.

Lemma pto_leadsTo1 : forall M l v A,
  M ||- l |-> v ** A
  -> M ||- l |~> v.
Proof.
  intros M l v A H.
  unfold leadsTo, imprecM.
  eapply top_intro_right1; eauto.
Qed.

Lemma leadsTo_mono : forall M l v A,
  M ||- l |~> v ** A
  -> M ||- l |~> v.
Proof.
  intros M l v A H.
  substH H with (leadsTo_pto1 H).
  eapply pto_leadsTo1; eauto.
Qed.
Hint Resolve leadsTo_mono : splogic.

Lemma leadsTo_get_eq : forall M l v,
  M ||- l |~> v
  -> M l = Some v.
Proof.
  intros M l v H.
  substH H with (leadsTo_pto H).
  eapply pto_get_eq1; eauto.
Qed.

Lemma leadsTo_update : forall M l v1 v2,
  M ||- l |~> v1 ->
    set M l v2 ||- l |~> v2.
Proof.
  intros M l v1 v2 H.
  substH H with (leadsTo_pto H).
  apply pto_leadsTo1 with top.
  eapply pto_update1; eauto.
Qed.

(* ********************************************** *)
(* leadsTox                                       *)

Lemma leadsTox_leadsTo : forall M l,
  M ||- l |~>_ 
  -> exists v, M ||- l |~> v.
Proof.
  unfold leadsTox.
  intros M l H.
  destruct (sconj_elim H) as [M1 [M2 [HJM [HM1 HM2]]]].
  substH HM1 with (ptox_pto HM1).
  destruct HM1 as [v HM1].
  exists v.
  apply pto_leadsTo1 with top.
  apply sconj_intro with M1 M2; auto.
Qed.

Lemma leadsTo_leadsTox : forall M l v,
  M ||- l |~> v 
  -> M ||- l |~>_.
Proof.
  intros M l v H.
  substH H with (leadsTo_pto H).
  destruct (sconj_elim H) as [M1 [M2 [HJM [HM1 HM2]]]].
  unfold leadsTox.
  unfold imprecM.
  apply sconj_intro with M1 M2; auto.
  apply pto_ptox with v; trivial.
Qed.

(* under construction *)

(* ************************************************)
(* WORD-wise                                      *)

(* ************************************************)
(* setW                                           *)

Notation Local junk := 0.

Notation Local setW := (MemUpdW) (only parsing).

(* ************************************************)
(* ptoW                                          *)

Lemma ptoW_pto : forall M l v,
  (M ||- l |->w v) 
  -> M ||- l |-> v ** l+1 |->_ .
Proof.
  intros M l v H.
  unfold ptoW in H.
  destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
  exists M1; exists M2; split_l.
  split_lr.
Qed.
Hint Resolve ptoW_pto : splogic.

Lemma ptoW_pto1 : forall M l v A,
  (M ||- l |->w v ** A) 
  -> M ||- (l |-> v ** l+1 |->_ ) ** A .
Proof.
  intros M l v A H.
  destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
  substH HM1 with (ptoW_pto HM1).
  apply sconj_intro with M1 M2; auto.
Qed.

Lemma pto_ptoW : forall M l v,
  M ||- l |-> v ** l+1 |->_
  -> (M ||- l |->w v).
Proof.
  intros M l w H.
  unfold ptoW.
  destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
  exists M1; exists M2; auto.
Qed.
Hint Resolve pto_ptoW : splogic.

Lemma pto_ptoW1 : forall M l v A,
  M ||- (l |-> v ** l+1 |->_ ) ** A
  -> (M ||- l |->w v ** A).
Proof.
  intros M l w A H.
  destruct (sconj_elim H) as [M1 [M2 [HJM [HM1 HM2]]]].
  apply sconj_intro with M1 M2; auto.
Qed.

Lemma ptoW_get_eq : forall M l v,
  M ||- l |->w v -> 
    M l = Some v.
Proof.
  intros M l v H.
  substH H with (ptoW_pto H).
  eapply pto_get_eq1; eauto.
Qed.

Lemma ptoW_get_eq1 : forall M l v A,
  M ||- l |->w v ** A -> 
    M l = Some v.
Proof.
  intros M l v A H.
  substH H with (ptoW_pto1 H).
  substH H with (sconj_assoc_left H).
  eapply pto_get_eq1; eauto.
Qed.

Lemma ptoW_ptoW_eq : forall M l v1 v2,
  M ||- l |->w v1 ->
  M ||- l |->w v2 ->
    v1 = v2.
Proof.
  intros M l v1 v2 H1 H2.
  substH H1 with (ptoW_get_eq H1).
  substH H2 with (ptoW_get_eq H2).
  rewrite H1 in H2.
  injection H2; auto.
Qed.

Lemma ptoW_ptoW_eq1 : forall M l v1 v2 A B,
  M ||- l |->w v1 ** A ->
  M ||- l |->w v2 ** B ->
      v1 = v2.
Proof.
  intros h x v1 v2 A B H1 H2.
  assert (Some v1 = Some v2).
  rewrite <- (ptoW_get_eq1 H1).
  rewrite <- (ptoW_get_eq1 H2).
  reflexivity.
  injection H; auto.
Qed.

(* temporary *)
Lemma ptoW_pto_eq : forall M l v1 v2,
  M ||- l |->w v1 ->
  M ||- l |-> v2 ->
    v1 = v2.
Proof.
  intros M l v1 v2 H1 H2.
  substH H1 with (ptoW_get_eq H1).
  substH H2 with (pto_get_eq H2).
  rewrite H1 in H2.
  injection H2; auto.
Qed.

(* temporary *)
Lemma ptoW_pto_eq1 : forall M l v1 v2 A B,
  M ||- l |->w v1 ** A ->
  M ||- l |-> v2 ** B ->
      v1 = v2.
Proof.
  intros h x v1 v2 A B H1 H2.
  assert (Some v1 = Some v2).
  rewrite <- (ptoW_get_eq1 H1).
  rewrite <- (pto_get_eq1 H2).
  reflexivity.
  injection H; auto.
Qed.

Lemma ptoW_update : forall M l v1,
  M ||- l |->w v1 -> forall v2 v',
    setW M l v2 v' ||- l |->w v2.
Proof.
  intros M l v1 H v2 v'.
  unfold MemUpdW.
  substH H with (ptoW_pto H).
  unfold ptoW.
  substH H with (sconj_sym H).
  apply sconj_sym.
  apply pto_ptox1 with v'.
  apply ptox_update1; auto.
  substH H with (sconj_sym H).
  apply sconj_sym.
  apply pto_update1 with v1; auto.
Qed.

Lemma ptoW_update1 : forall M l v1 A,
  M ||- l |->w v1 *^* A -> forall v2 v',
    setW M l v2 v' ||- l |->w v2 *^* A.
Proof.
  intros M l v1 A H v2 v'.
  substH H with (ptoW_pto1 H).
  unfold setW.
  apply pto_ptoW1.
  apply sconj_assoc_right.
  apply sconj_sym.
  apply sconj_assoc_right.
  assert (H1 := sconj_assoc_left H).
  substH H1 with (sconj_sym H1).
  substH H1 with (sconj_assoc_left H1).
  apply pto_ptox1 with v'.
  apply ptox_update1.
  apply sconj_sym.
  apply sconj_assoc_right.
  apply sconj_sym.
  apply sconj_assoc_right.
  substH H with (sconj_assoc_left H).
  apply pto_update1 with v1; trivial.
Qed.

(* add by Xinyu *)
Lemma ptoW_elim  : forall M l v,
  (M ||- l |->w v)
  -> exists x, M ||- l |-> v ** l+1 |-> x.
Proof.
  intros M l v H.
  substH H with (ptoW_pto H).
  substH H with (sconj_sym H).
  substH H with (ptox_pto1 H).
  destruct H.
  exists x.
  apply sconj_sym; trivial.
Qed.

(* add by Xinyu *)
Lemma ptoW_elim1 : forall M l v m,
  (M ||- l |->w v *^* m) 
  -> exists x, M ||- l |-> v ** l+1 |-> x *^* m.
Proof.
  intros M l v m H.
  substH H with (ptoW_pto1 H).
  substH H with (sconj_assoc_left H).
  substH H with (sconj_sym H).
  substH H with (sconj_assoc_left H).
  substH H with (ptox_pto1 H).
  destruct H.
  exists x.
  apply sconj_assoc_right.
  apply sconj_sym.
  apply sconj_assoc_right; trivial.
Qed.

(* add by Xinyu *)
Lemma ptoW_intro : forall M l v x,
  M ||- l |-> v ** l+1 |-> x
  -> (M ||- l |->w v).
Proof.
  intros M l v x H.
  apply pto_ptoW.
  apply sconj_sym.
  apply pto_ptox1 with x.
  apply sconj_sym; trivial.
Qed.

(* add by Xinyu *)
Lemma ptoW_intro1 : forall M l v x m,
  M ||- l |-> v ** l+1 |-> x *^* m
  -> M ||- l |->w v *^* m.
Proof.
  intros M l w x m H.
  apply pto_ptoW1.
  apply sconj_assoc_right.
  apply sconj_sym.
  apply sconj_assoc_right.
  apply pto_ptox1 with x.
  apply sconj_assoc_left.
  apply sconj_sym.
  apply sconj_assoc_left; trivial.
Qed.

(**********************************************)
(* ptoxW                                      *)

Lemma ptoxW_ptoW : forall M l,
  M ||- l |->w_ -> exists x, M ||- l |->w x.
Proof.
  intros M l H.
  unfold ptoxW in H.
  substH H with (ptox_pto1 H).
  destruct H.
  exists x.
  unfold ptoW.
  trivial.
Qed.

Lemma ptoxW_ptoW1 : forall M l A,
  M ||- l |->w_ ** A -> exists x, M ||- l |->w x ** A.
Proof.
  intros M l A H.
  destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
  destruct (ptoxW_ptoW HM1).
  exists x.
  exists M1; exists M2; split_l.
  auto.
Qed.

Lemma ptoW_ptoxW : forall M l x,
  M ||- l |->w x -> M ||- l |->w_ .
Proof.
  intros M l x H.
  unfold ptoW, ptoxW, ptox, pto in * .
  destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
  exists M1; exists M2; split_l.
  split_r.
  exists x; trivial.
Qed.

Lemma ptoW_ptoxW1 : forall M l x A,
  M ||- l |->w x ** A -> M ||- l |->w_ ** A.
Proof.
  intros M l x A H.
  destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
  exists M1; exists M2; split_l.
  split_r.
  apply ptoW_ptoxW with x; trivial.
Qed.

Lemma ptoxW_update : forall M l,
  M ||- l |->w_ -> forall v2 v',
    MemUpdW M l v2 v' ||- l |->w v2.
Proof.
  intros M l H v2 v'.
  substH H with (ptoxW_ptoW H).
  destruct H.
  apply ptoW_update with x; trivial.
Qed.

Lemma ptoxW_update1 : forall M l A,
  M ||- l |->w_ *^* A 
  -> forall v2 v',
  setW M l v2 v' ||- l |->w v2 *^* A.
Proof.
  intros M l A H v2 v'.
  substH H with (ptoxW_ptoW1 H).
  destruct H.
  apply ptoW_update1 with x; trivial.
Qed.

(*********************************)
(* leadsToW                      *)

Lemma ptoW_leadsToW : forall M l v,
     M ||- l |->w v
  -> M ||- l |~>w v.
Proof.
  intros M l v H.
  unfold leadsToW.
  unfold imprecM.
  exists M; exists MemMod.emp.
  split.
    auto with PMap.
  unfold mTrue.
  split_lr.
Qed.
Implicit Arguments ptoW_leadsToW [M l v].
Hint Resolve ptoW_leadsToW : splogic.

Lemma leadsToW_ptoW : forall M l v,
     M ||- l |~>w v
  -> M ||- l |->w v ** top.
Proof.
  unfold leadsToW.
  unfold imprecM.
  trivial.
Qed.
Implicit Arguments leadsToW_ptoW [M l v].
Hint Resolve leadsToW_ptoW : splogic.

Lemma ptoW_leadsToW1 : forall M l v A,
     M ||- l |->w v ** A
  -> M ||- l |~>w v.
Proof.
  intros M l v A H.
  unfold leadsToW.
  unfold imprecM.
  destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
  exists M1; exists M2; split_l.
  unfold mTrue.
  auto.
Qed.
Implicit Arguments ptoW_leadsToW1 [M l v A].
Hint Resolve ptoW_leadsToW1 : splogic.

Lemma leadsToW_leadsTo : forall M l v,
  M ||- l |~>w v
    -> M ||- l |~> v
      /\ M ||- (l + 1) |~>_.
Proof.
  intros M l v H.
  substH H with (leadsToW_ptoW H).
  substH H with (ptoW_pto1 H). 
  substH H with (sconj_assoc_left H).
  split.
    eapply pto_leadsTo1; eauto.
  substH H with (sconj_sym H).
  substH H with (sconj_assoc_left H).
  substH H with (ptox_pto1 H).
  destruct H.
  apply leadsTo_leadsTox with x. 
  eapply pto_leadsTo1; eauto.
Qed.
Implicit Arguments leadsToW_leadsTo [M l v].
Hint Resolve leadsToW_leadsTo : splogic.

Lemma ptoW_leadsToW_eq1 : forall M l v v' A, 
     M ||- l |->w v ** A
  -> M ||- l |~>w v'
  -> v = v'.
Proof.
  intros M l v v' A H H0.
  substH H0 with (leadsToW_leadsTo H0).
  destruct H0.
  substH H with (ptoW_pto1 H).
  substH H0 with (leadsTo_pto H0).
  substH H with (sconj_assoc_left H).
  eapply pto_pto_eq1; eauto.
Qed.
Implicit Arguments ptoW_leadsToW_eq1 [M l v v' A].

(* temporary *)
Lemma pto_leadsToW_eq1 : forall M l v v' A, 
     M ||- l |-> v ** A
  -> M ||- l |~>w v'
  -> v = v'.
Proof.
  intros M l v v' A H H0.
  substH H0 with (leadsToW_leadsTo H0).
  destruct H0.
  substH H0 with (leadsTo_pto H0).
  eapply pto_pto_eq1; eauto.
Qed.
Implicit Arguments ptoW_leadsToW_eq1 [M l v v' A].

Lemma leadsToW_get_eq : forall M l v, 
     M ||- l |~>w v
  -> M l = Some v.
Proof.
  intros.
  substH H with (leadsToW_ptoW H).
  apply ptoW_get_eq1 with top; trivial.
Qed.
Implicit Arguments leadsToW_get_eq [M l v].

Lemma leadsToW_leadsToW_eq : forall M l v v',
  M ||- l |~>w v
  -> M ||- l |~>w v'
  -> v = v'.
Proof.
  unfold mjudge; intros. 
  substH H with (leadsToW_ptoW H).
  substH H0 with (leadsToW_ptoW H0).
  eapply ptoW_ptoW_eq1; eauto.
Qed.
Implicit Arguments leadsToW_leadsToW_eq [M l v v'].

Lemma leadsToW_update : forall M l v1,
  M ||- l |~>w v1 -> forall v2 v',
    setW M l v2 v' ||- l |~>w v2.
Proof.
  intros M l v1 H v2 v'.
  substH H with (leadsToW_ptoW H).
  apply ptoW_leadsToW1 with top; trivial.
  apply ptoW_update1 with v1; trivial.
Qed.

(**********************************)
(* leadsToxW                      *)

Lemma leadsToW_leadsToxW : forall M l v,
  M ||- l |~>w v
  -> M ||- l|~>w_.
Proof.
  unfold leadsToxW.
  intros M l v H.
  substH H with (leadsToW_ptoW H).
  unfold imprecM.
  apply ptoW_ptoxW1 with v; trivial.
Qed.
Implicit Arguments leadsToW_leadsToxW [M l v].

Lemma leadsToxW_leadsToW : forall M l,
  M ||- l |~>w_
  -> exists v, M ||- l|~>w v.
Proof.
  unfold leadsToxW.
  intros M l H.
  unfold imprecM in H.
  substH H with (ptoxW_ptoW1 H).
  destruct H.
  exists x.
  eapply ptoW_leadsToW1; eauto.
Qed.
Implicit Arguments leadsToxW_leadsToW [M l].

Lemma ptoW_leadsToxW1 : forall M l v A,
     M ||- l |->w v *^* A
  -> M ||- l |~>w_ .
Proof.
  intros M l v A H.
  unfold leadsToxW.
  unfold imprecM.
  destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
  exists M1; exists M2; split_l.
  unfold mTrue.
  split_r.
  apply ptoW_ptoxW with v; trivial.
Qed.
Implicit Arguments ptoW_leadsToxW1 [M l v A].

Lemma ptoxW_leadsToxW1 : forall M l A,
     M ||- l |->w_ *^* A
  -> M ||- l |~>w_ .
Proof.
  intros M l A H.
  unfold leadsToxW.
  unfold imprecM.
  destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
  exists M1; exists M2; split_l.
  unfold mTrue.
  auto.
Qed.
Implicit Arguments ptoxW_leadsToxW1 [M l A].

Lemma leadsToxW_update : forall M l,
  M ||- l |~>w_ -> forall v2 v',
    setW M l v2 v' ||- l |~>w v2.
Proof.
  intros M l H v2 v'.
  substH H with (leadsToxW_leadsToW H).
  destruct H as [v H].
  apply leadsToW_update with v; trivial.
Qed.
Implicit Arguments leadsToxW_update [M l].

(* ************************************************)
(* meq                                            *)

Lemma meq_intro : forall M,
  M ||- meq M.
Proof.
  unfold mjudge, meq; auto.
Qed.
Hint Resolve meq_intro : splogic.

Lemma meq_elim : forall M M',
  M ||- meq M' -> M = M'.
Proof.
  unfold mjudge, meq; auto.
Qed.

Lemma meq_update : forall M l v x,
  setW M l v x ||- meq (setW M l v x).
Proof.
  intros; apply meq_intro; trivial.
Qed.

Lemma meq_update1 : forall M1 M l v x A,
  (M ||- meq M1 ** A)
  -> (M1 ||- l |~>w_)
  -> setW M l v x ||- meq (setW M1 l v x) ** A.
Proof.
  unfold meq.
  intros M' M l v x A H H0.
  destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
  subst M'.
  substH H0 with (leadsToxW_leadsToW H0).
  destruct H0.
  substH H with (leadsToW_ptoW H).
  substH H with (ptoW_pto1 H).
  assert (MemMod.in_dom l M1).
    substH H with (sconj_assoc_left H).
    exists x0.
    eapply pto_get_eq1; eauto.
  assert (MemMod.in_dom (l+1) (set M1 l v)).
    substH H with (sconj_assoc_left H).
    substH H with (sconj_sym H).
    substH H with (sconj_assoc_left H).
    substH H with (ptox_pto1 H).
    destruct H.
    exists x1.
    substH H with (pto_get_eq1 H).
    unfold set.
    destruct (MemMod.beq_A (l+1) l); trivial.
    apply False_ind; omega.
  exists (MemUpdW M1 l v x); exists M2.
  split; auto.
  unfold MemUpdW.
  apply MemMod.update_presv_disjUnion_left; trivial.
  apply MemMod.update_presv_disjUnion_left; trivial.
Qed.

(*********************************)
(* ptoxBlk                       *)

Lemma imp_ptoxBlk_unit : forall l,
  (ptoxBlk l 0) ==> (unit).
Proof.
  unfold ptoxBlk; intros l; trivial.
Qed.

Lemma imp_unit_ptoxBlk : forall l,
  (unit) ==> (ptoxBlk l 0).
Proof.
  unfold ptoxBlk; intros l; trivial.
Qed.

Hint Resolve imp_ptoxBlk_unit imp_unit_ptoxBlk : splogic.

Lemma imp_ptoxBlk_ptox : forall l,
  (ptoxBlk l 1) ==> (ptox l).
Proof.
  unfold ptoxBlk; intro l; intros.
  apply unit_elim_right; trivial.
Qed.

Lemma imp_ptox_ptoxBlk : forall l,
  (ptox l) ==> (ptoxBlk l 1).
Proof.
  unfold ptoxBlk; intro l; intros.
  apply unit_intro_right; trivial.
Qed.

Hint Resolve imp_ptoxBlk_ptox imp_ptox_ptoxBlk : splogic.

Lemma imp_ptoxBlk_ptoxW : forall l,
  (ptoxBlk l 2) ==> (ptoxW l).
Proof.
  unfold ptoxBlk, ptoxW; intro l; intros.
  apply unit_elim_right; trivial.
  apply sconj_assoc_right; trivial.
Qed.

Lemma imp_ptoxW_ptoxBlk : forall l,
  (ptoxW l) ==> (ptoxBlk l 2).
Proof.
  unfold ptoxBlk, ptoxW; intro l; intros.
  apply sconj_assoc_left; trivial.
  apply unit_intro_right; trivial.
Qed.

Hint Resolve imp_ptoxBlk_ptoxW imp_ptoxW_ptoxBlk : splogic.

Lemma ptoxBlk_split : forall s1 s2 M l,
  M ||- ptoxBlk l (s1 + s2)
  -> M||- (ptoxBlk l s1) ** (ptoxBlk (l+s1) s2).
Proof.
  induction s1; simpl; intros.
  unfold Aemp, MemMod.emp.
  exists MemMod.emp; exists M.
  split.
    auto with PMap.
    unfold MemMod.emp.
    split_l.
  replace (l+0) with l; auto with arith.
  destruct  H as [M1 [M2 [MJM [HM1 HM2]]]].
  replace (l + S s1) with (S l + s1); [idtac | omega].
  assert (Hx := IHs1 s2 M2 (l+1) HM2).
  apply sconj_assoc_right.
  unfold mstar.
  exists M1; exists M2; split_l.
  split_l.
  arith_rep (S l + s1) with (l+1+s1).
Qed.

Lemma ptoxBlk_merge : forall s1 s2 M l,
  M||- (ptoxBlk l s1) ** (ptoxBlk (l+s1) s2)
  -> M ||- ptoxBlk l (s1 + s2).
Proof.
induction s1;simpl;intros.
unfold mjudge, sconj, Aemp in H.
InvertAll.
subst.
assert (x0 = M).
auto with PMap.
subst.
arith_rep l with (l+0); assumption.
assert (M ||- (ptox l *^* (ptoxBlk (l + 1) s1 *^* ptoxBlk (l + S s1) s2))).
apply sconj_assoc_left; assumption.
unfold mjudge, mstar at 1 in H0.
InvertAll.
exists x;exists x0;intuition.
apply IHs1.
arith_rep (l+1+s1) with (l+S s1); assumption.
Qed.

Lemma imp_ptoxBlk_split : forall l s1 s2,
  (ptoxBlk l (s1 + s2)) ==> ((ptoxBlk l s1) ** (ptoxBlk (l+s1) s2)).
Proof.
  intros l s1 s2 M.
  intro H.
  apply ptoxBlk_split; trivial.
Qed.

Lemma imp_ptoxBlk_merge : forall l s1 s2,
  ((ptoxBlk l s1) ** (ptoxBlk (l+s1) s2)) ==> (ptoxBlk l (s1 + s2)).
Proof.
  intros l s1 s2 M.
  intro H.
  apply ptoxBlk_merge; trivial.
Qed.

Hint Resolve imp_ptoxBlk_split imp_ptoxBlk_merge : splogic.

(* ******************************************************)
(* Basic tactics                                        *)
(* ******************************************************)

Ltac hyprep H y :=
  let H' := fresh in
    (assert (H' := y); try (clear H; rename H' into H)).

Ltac sep_mjudge :=
  apply mjudge_elim.

Ltac sep_mjudgeH H :=
  hyprep H (mjudge_intro H).

Ltac sep_turnon :=
  match goal with
    | H : unit ?M |- _ => sep_mjudgeH H; sep_turnon
    | H : top ?M |- _ => sep_mjudgeH H; sep_turnon
    | H : (sconj ?A ?B) ?M |- _ => sep_mjudgeH H; sep_turnon
    | H : (ptox ?l) ?M |- _ => sep_mjudgeH H; sep_turnon
    | H : (pto ?l ?x) ?M |- _ => sep_mjudgeH H; sep_turnon
    | H : (leadsTo ?l ?x) ?M |- _ => sep_mjudgeH H; sep_turnon
    | H : (leadsTox ?l) ?M |- _ => sep_mjudgeH H; sep_turnon
    | H : (ptoxW ?l) ?M |- _ => sep_mjudgeH H; sep_turnon
    | H : (ptoW ?l ?x) ?M |- _ => sep_mjudgeH H; sep_turnon
    | H : (leadsToW ?l ?x) ?M |- _ => sep_mjudgeH H; sep_turnon
    | H : (leadsToxW ?l) ?M |- _ => sep_mjudgeH H; sep_turnon
    | H : (stkSpacePrec ?sp ?size) ?M |- _ => sep_mjudgeH H; sep_turnon
    | H : (stkSpace ?sp ?size) ?M |- _ => sep_mjudgeH H; sep_turnon
    | _ => (try sep_mjudge); idtac
  end.

Ltac sep_turnoff := unfold mjudge in * .

Tactic Notation "sep" "on" := sep_turnon.

Tactic Notation "sep" "off" := sep_turnoff.

(*
Ltac sepelim Hp :=
  match type of Hp with
    | _ ||- unit => hyprep Hp (unit_elim Hp); subst h
    | _ ||- bang _ => 
      let A := fresh "A" in let Hp' := fresh "Hp" in
        (destruct (bang_elim Hp) as  [Hp' A]; 
          hyprep Hp' (unit_elim Hp'); subst h; clear Hp')
(*     | _ ||- lex _ =>  *)
(*       let x := fresh "x" in let Hp' := fresh "Hp'" in *)
(*      (destruct (lex_elim Hp) as [x Hp'];  *)
(*        clear Hp; rename Hp' into Hp; sepelim Hp) *)

    | _ ||- sconj (bang _) _ => 
      let A := fresh "A" in let Hp' := fresh "Hp" in
        (destruct (bang_elim_left Hp) as  [A Hp']; 
          clear Hp; rename Hp' into Hp; sepelim Hp)
    | _ ||- sconj _ (bang _) => 
      let A := fresh "A" in let Hp' := fresh "Hp" in
        (destruct (bang_elim_right Hp) as  [A Hp']; 
          clear Hp; rename Hp' into Hp; sepelim Hp)

    | _ ||- sconj unit _ => hyprep Hp (unit_elim_left Hp); sepelim Hp
    | _ ||- sconj _ unit => hyprep Hp (unit_elim_right Hp); sepelim Hp
    | _ ||- sconj (sconj _ _) _ => hyprep Hp (sconj_assoc_left Hp); sepelim Hp

    | _ ||- sconj _ _ =>  
      let h1 := fresh "h" in let h2 := fresh "h" in let J := fresh "J" in 
        let Hp1 := fresh "Hp" in let Hp2 := fresh "Hp" in
          (destruct Hp as [h1 [h2 [J [Hp1 Hp2]]]];
            sepelim Hp1; sepelim Hp2;
            assert (Hp := sconj_intro Hp1 Hp2 J);
              clear h1 h2 J Hp1 Hp2)
    | _ => idtac
  end.
*)

Ltac assocl H := 
  match type of H with 
    | _ ||- sconj (ptoxW ?l) _ => idtac
    | _ ||- sconj (ptoW ?l ?x) _ => idtac
    | _ ||- sconj (sconj _ _) _ => hyprep H (sconj_assoc_left H); assocl H
    | _ => idtac
  end.

(* Ltac assocl H := *)
(*   repeat (hyprep H (sconj_assoc_left H)). *)

Ltac assocr :=
  match goal with 
    | |- _ ||- sconj (ptoxW ?l) _ => idtac
    | |- _ ||- sconj (ptoW ?l ?w) _ => idtac
    | |- _ ||- sconj (sconj _ _) _ => apply sconj_assoc_right; assocr
    | _ => idtac
  end.

(* Ltac assocr := *)
(*   repeat (apply sconj_assoc_right). *)

Ltac sep_inj := 
  match goal with

    | H : ?M ||- (pto ?l ?v),
      H0 : ?M ||- (pto ?l ?v2) |- _
        => generalize (pto_pto_eq H H0)

    | H : ?M ||- sconj (pto ?l ?v) _,
      H0 : ?M ||- sconj (pto ?l ?v2) _ |- _
        => generalize (pto_pto_eq1 H H0)

    | H : ?M ||- (ptoW ?l ?v),
      H0 : ?M ||- (ptoW ?l ?v2) |- _
        => generalize (ptoW_ptoW_eq H H0)

    | H : ?M ||- sconj (ptoW ?l ?v) _,
      H0 : ?M ||- sconj (ptoW ?l ?v2) _ |- _
        => generalize (ptoW_ptoW_eq1 H H0)

(* temporary *)
    | H : ?M ||- (ptoW ?l ?v),
      H0 : ?M ||- (pto ?l ?v2) |- _
        => generalize (ptoW_pto_eq H H0)
(* temporary *)
    | H : ?M ||- sconj (ptoW ?l ?v) _,
      H0 : ?M ||- sconj (pto ?l ?v2) _ |- _
        => generalize (ptoW_pto_eq1 H H0)

    | H : ?M ||- sconj (ptoW ?l ?v) _,
      H0 : ?M ||- leadsToW ?l ?v2 |- _
        => generalize (ptoW_leadsToW_eq1 H H0)
(* temporary *)
    | H : ?M ||- sconj (pto ?l ?v) _,
      H0 : ?M ||- leadsToW ?l ?v2 |- _
        => generalize (pto_leadsToW_eq1 H H0)

    | H : ?M ||- leadsToW ?l ?v,
      H0 : ?M ||- leadsToW ?l ?v2 |- _
        => generalize (leadsToW_leadsToW_eq H H0)
  end.

Tactic Notation "sep" "inj" := sep_inj.

Ltac sep_update :=
  match goal with
    | |- MemUpdW ?M ?l ?v ?junk ||- ptoW ?l ?v 
      => apply ptoxW_update

    | |- MemUpdW ?M ?l ?v ?junk ||- ptoxW ?l 
      => apply ptoW_ptoxW with v; apply ptoxW_update

    | |- MemUpdW ?M ?l ?v ?junk ||- sconj (ptoW ?l ?v) ?B => 
      apply ptoxW_update1

    | |- MemUpdW ?M ?l ?v ?junk ||- sconj (ptoxW ?l) ?B => 
      apply ptoW_ptoxW1 with v; apply ptoxW_update1

    | |- MemUpdW ?M ?l ?v ?junk ||- (leadsToW ?l ?v)
      => apply leadsToxW_update

    | |- MemUpdW ?M ?l ?v ?junk ||- (leadsToxW ?l)
      => apply leadsToW_leadsToxW with v; apply leadsToxW_update

    | |- MemUpdW ?M ?l ?v ?junk ||- sconj (meq (MemUpdW ?M1 ?l ?v ?junk)) ?B
      => apply meq_update1; auto

    | _ => fail 1 "Error: format"
  end.

Tactic Notation "sep" "upd" := sep_update.

(**************************************************************)
(* memory shift                                               *)
(**************************************************************)

Ltac sep_lift n :=
  match eval compute in n with
    | O => assocr
    | _ => assocr; apply sconj_sym; sep_lift (pred n)
  end.

Ltac sep_liftH H n :=
  match eval compute in n with
    | O => assocl H
    | _ => assocl H; hyprep H (sconj_sym H); sep_liftH H (pred n)
  end.

Ltac sep_shift_l n :=
  match eval compute in n with
    | 0 => apply sconj_sym
    | _ => sep_lift (pred n); apply sconj_sym
  end.
  
Tactic Notation "sep" "lift" constr (n) :=
  sep_lift n.

Tactic Notation "sep" "lift" constr (n) "in" hyp (H):=
  sep_liftH H n.

Tactic Notation "sep" "lift" constr (n) "in" hyp (H) "as" hyp (Hx):=
  generalize H; intro Hx; sep_liftH Hx n.

Ltac sep_searchH H C n :=
  match eval compute in n with
    | O => assocl H; 
      match type of H with
        | _ ||- C ** ?B => idtac
        | _ ||- ?A ** ?B => fail 1 "Error: Not found"
        | _ => fail 1 "Error:"
      end
    | _ => assocl H; 
      match type of H with 
        | _ ||- C ** ?B => idtac
        | _ ||- ?A ** ?B => hyprep H (sconj_sym H); sep_searchH H C (pred n)
        | _ => fail 1 "Error:"
      end
  end.

Ltac sep_search C n :=
  match eval compute in n with
    | O => assocr; 
      match goal with
        | |- _ ||- sconj C ?B => idtac
        | |- _ ||- sconj ?A ?B => fail 1 "Error: Not found"
        | _ => fail 1 "Error:"
      end
    | _ => assocr; 
      match goal with 
        | |- _ ||- sconj C ?B => idtac
        | |- _ ||- sconj ?A ?B => apply sconj_sym; sep_search C (pred n)
        | _ => fail 1 "Error:"
      end
  end.

Tactic Notation "sep" "search" constr (C) "in" hyp (H) "with" constr (n) :=
  sep_searchH H C n.

Tactic Notation "sep" "search" constr (C) "in" hyp (H) :=
  sep_searchH H C 10.

Tactic Notation "sep" "search" constr (C) "with" constr (n) :=
  sep_search C n.

Tactic Notation "sep" "search" constr (C) :=
  sep_search C 30.

(**************************************************************)
(* stack operation                                            *)
(**************************************************************)

Lemma stkSpacePrec_stkSpace : forall M sp size,
  M ||- (stkSpacePrec sp size)
  -> M ||- stkSpace sp size.
Proof.
  intros.
  unfold stkSpace.
  unfold stkSpacePrec in H.
  unfold mjudge in * .
  destruct H.
  split_l.
  unfold leadsToBlk.
  unfold imprecM.
  exists M; exists MemMod.emp.
  split.
    auto with PMap.
  split_l.
  unfold mTrue; trivial.
Qed.
Implicit Arguments stkSpacePrec_stkSpace [M sp size].

Lemma stkSpacePrec_stkSpace1 : forall M sp size A,
  M ||- (stkSpacePrec sp size *^* A)
  -> M ||- stkSpace sp size.
Proof.
  intros M sp size A H.
  unfold stkSpace.
  unfold stkSpacePrec in H.
  unfold mjudge in * .
  destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
  destruct HM1.
  split_l.
  unfold leadsToBlk.
  unfold imprecM.
  exists M1; exists M2; split_l.
  split.
    auto with PMap.
  unfold mTrue; trivial.
Qed.
Implicit Arguments stkSpacePrec_stkSpace1 [M sp size A].

Lemma stkSpacePrec_ptoxW :forall M sp,
  M ||- stkSpacePrec sp 2
  -> M ||- ptoxW (sp - 2).
Proof.
  intros M sp H.
  unfold stkSpacePrec in H.
  unfold ptoxBlk in H.
  unfold ptoxW.
  unfold mjudge in * .
  unfold Aemp in H.
  destruct H.
  substH H0 with (sconj_assoc_right H0).
  unfold mjudge in H0.
  destruct H0 as [M1 [M2 [HJM [HM1 HM2]]]].
  subst M2.
  assert (M1 = M).
    auto with PMap.
  subst M1.
  trivial.
Qed.

Lemma ptoxW_stkSpacePrec :forall M l,
  M ||- ptoxW l
  -> M ||- stkSpacePrec (l+2) 2.
Proof.
  intros M l H.
  unfold stkSpacePrec, ptoxW in * .
  simpl.
  unfold mjudge.
  split.
    auto with arith.
  assert (l+2-2 = l).
    omega.
  rewrite H0.
  apply sconj_assoc_left.
  exists M; exists MemMod.emp.
  split.
    auto with PMap.
  unfold Aemp.
  split_r.
  trivial.
Qed.

Lemma stkSpacePrec_unit : forall M sp,
  M ||- stkSpacePrec sp 0
  -> M ||- unit.
Proof.
  intros M sp H.
  unfold stkSpacePrec in H.
  simpl in H.
  unfold mjudge in * .
  destruct H; trivial.
Qed.

Lemma stkSpacePrec_unit1 : forall M sp A,
  M ||- (stkSpacePrec sp 0 ** A)
  -> M ||- A.
Proof.
  intros M sp A H.
  unfold stkSpacePrec in H.
  destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
  destruct HM1.
  simpl in H0.
  unfold Aemp in H0.
  subst M1.
  assert (M2 = M).
    auto with PMap.
  subst M2.
  trivial.
Qed.

Lemma unit_stkSpacePrec : forall M sp,
  M ||- unit
  -> M ||- stkSpacePrec sp 0.
Proof.
  intros M sp H.
  unfold stkSpacePrec.
  unfold mjudge in * .
  simpl.
  auto with arith.
Qed.

Lemma unit_stkSpacePrec_unit : forall M sp A,
  M ||- A
  -> M ||- sconj (stkSpacePrec sp 0) A.
Proof.
  intros M sp A H.
  exists MemMod.emp; exists M.
  split.
    auto with PMap.
  split.
    apply unit_stkSpacePrec.
    unfold mjudge, Aemp; trivial.
  trivial.
Qed.


(*************************)
(* ** stack split     ** *)

Lemma stkSpacePrec_splitW : forall M sp size, 
  M ||- stkSpacePrec (sp+2) (size+2)
    -> M ||- stkSpacePrec sp size *^* 
      (sp |->w_ ).
Proof.
  intros M sp size H.
  substH H with (stkSpacePrec_split H).
  arith_rep (sp + 2 - 2) with sp in H.
  substH H with (sconj_sym H).
  apply (sconj_mono_right H).
  intros Mx HMx.
  arith_rep sp with (sp + 2 - 2).
  apply stkSpacePrec_ptoxW; trivial.
Qed.
Implicit Arguments stkSpacePrec_splitW [M sp size].

Lemma stkSpacePrec_splitWh : forall M sp size, 
  M ||- stkSpacePrec sp size
    -> size >= 2
    -> M ||- stkSpacePrec (sp - 2) (size - 2) *^* 
      (sp - 2 |->w_ ).
Proof.
  intros M sp size H H0.
  assert (sp >= 2).
    destruct H; omega.
  arith_rep sp with (sp - 2 + 2) in H. 
  arith_rep size with (size - 2 + 2) in H.
  apply stkSpacePrec_splitW; trivial.
Qed.
Implicit Arguments stkSpacePrec_splitWh [M sp size].

Lemma stkSpacePrec_splitWh1 : forall M sp size A, 
  M ||- stkSpacePrec sp size *^* A
    -> size >= 2
    -> M ||- stkSpacePrec (sp - 2) (size - 2) *^* 
      (sp -2 |->w_ *^* A).
Proof.
  intros.
  destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
  apply sconj_assoc_left.
  exists M1; exists M2; split_l.
  split_r.
  apply stkSpacePrec_splitWh; trivial.
Qed.
Implicit Arguments stkSpacePrec_splitWh1 [M sp size A].

Lemma stkSpacePrec_splitW1 : forall M sp size A, 
  M ||- stkSpacePrec (sp + 2) (size + 2) *^* A
    -> M ||- stkSpacePrec sp size *^* 
      (sp |->w_ *^* A).
Proof.
  intros M sp size A H.
  destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
  apply sconj_assoc_left.
  exists M1; exists M2; split_l.
  split_r.
  apply stkSpacePrec_splitW; trivial.
Qed.
Implicit Arguments stkSpacePrec_splitW1 [M sp size A].

Lemma stkSpacePrec_mergeW : forall M sp size,
  M ||- stkSpacePrec (sp - 2) (size - 2) *^* sp - 2 |->w_ 
    -> sp >= 2
    -> size >= 2
    ->  M ||- stkSpacePrec sp size.
Proof.
  intros M sp size H H0 H1.
  arith_rep sp with (sp - 2 + 2).  
  arith_rep size with (size - 2 + 2).
  destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
  assert (MemMod.disjUnion M2 M1 M).
    auto with PMap.
  destruct H.
  subst M.
  unfold mjudge.
  apply stkSpacePrec_merge.
  apply ptoxW_stkSpacePrec; trivial.
  arith_rep (sp - 2 + 2 - 2) with (sp - 2).
Qed.
Implicit Arguments stkSpacePrec_mergeW [M sp size].

Lemma stkSpacePrec_mergeW1 : forall M sp size A, 
  M ||- stkSpacePrec (sp - 2) (size - 2) *^* 
    (sp - 2 |->w_ *^* A)
    -> sp >= 2
    -> size >= 2
    ->  M ||- stkSpacePrec sp size *^* A.
Proof.
  intros M sp size A H1 H2 H3.
  substH H1 with (sconj_assoc_right H1).
  destruct H1 as [M1 [M2 [HJM [HM1 HM2]]]].
  exists M1; exists M2; split_l.
  split_r.
  apply stkSpacePrec_mergeW; trivial.
Qed.
Implicit Arguments stkSpacePrec_mergeW1 [M sp size A].

Lemma stkSpacePrec_mergeWh : forall M sp size,
  M ||- stkSpacePrec sp size *^* sp |->w_
    ->  M ||- stkSpacePrec (sp + 2) (size + 2).
Proof.
  intros.
  unfold mjudge.
  substH H with (sconj_sym H).
  destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
  destruct HJM; subst M.
  apply stkSpacePrec_merge.
  apply ptoxW_stkSpacePrec; trivial.
  arith_rep (sp + 2 - 2) with sp.
Qed.
Implicit Arguments stkSpacePrec_mergeWh [M sp size].

Lemma stkSpacePrec_mergeWh1 : forall M sp size A, 
  M ||- stkSpacePrec sp size *^* 
    (sp |->w_ *^* A)
    ->  M ||- stkSpacePrec (sp + 2) (size + 2) *^* A.
Proof.
  intros.
  substH H with (sconj_assoc_right H).
  destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
  exists M1; exists M2; split_l.
  split_r.
  apply stkSpacePrec_mergeWh; trivial.
Qed.
Implicit Arguments stkSpacePrec_mergeWh1 [M sp size A].

Ltac stk_splitw :=
  match goal with 
    | |- _ ||- stkSpacePrec (?sp + 2) (?size + 2) => apply stkSpacePrec_mergeWh
    | |- _ ||- sconj (stkSpacePrec (?sp + 2) (?size +2)) _ => apply stkSpacePrec_mergeWh1
    | |- _ ||- stkSpacePrec ?sp ?size => 
      apply stkSpacePrec_mergeW; 
        [ idtac 
          | simpl; auto with arith; omega || fail 1 "Error: top address of stake should be >= 2" 
          | simpl; auto with arith; omega || fail 1 "Error: size of stake should be >= 2" 
        ]
    | |- _ ||- sconj (stkSpacePrec ?sp ?size) _ => 
      apply stkSpacePrec_mergeW1;  
        [ idtac 
          | simpl; auto with arith; omega || fail 1 "Error: top address of stake should be >= 2"
          | simpl; auto with arith; omega || fail 1 "Error: size of stake should be >= 2"
        ]
    | _ => fail 1 "Error: cannot split the current goal"
  end.

Ltac stk_splitwH H :=
  match type of H with
    | _ ||- stkSpacePrec (?sp + 2) (?size + 2) 
      => hyprep H (stkSpacePrec_splitW H)
    | _ ||- sconj (stkSpacePrec (?sp + 2) (?size +2)) _ 
      => hyprep H (stkSpacePrec_splitW1 H)
    | _ ||- stkSpacePrec ?sp ?size
      => let Hx := fresh "H" in 
        (assert (Hx : size >= 2); 
          [ simpl; auto with arith; omega || fail 1 "Error: size of stake should be >= 2" | idtac];
          hyprep H (stkSpacePrec_splitWh H Hx); clear Hx)
    | _ ||- sconj (stkSpacePrec ?sp ?size) _ 
      => let Hx := fresh "H" in 
        (assert (Hx : size >= 2); 
          [ simpl; auto with arith; omega || fail 1 "Error: size of stake should be >= 2" | idtac];
            hyprep H (stkSpacePrec_splitWh1 H Hx); clear Hx)
    | _ => fail 1 "Error: cannot split the current hypothesis"
  end.

Lemma stkSpacePrec_split : forall M sp size k, 
  M ||- stkSpacePrec (sp+k) (size+k)
    -> M ||- stkSpacePrec (sp+k) k *^* stkSpacePrec sp size.
Proof.
  intros M sp size k H.
  substH H with (stkSpacePrec_split H).
  arith_rep (sp + k - k) with sp in H.
Qed.
Implicit Arguments stkSpacePrec_split [M sp size k].

Lemma stkSpacePrec_splith : forall M sp size k, 
  M ||- stkSpacePrec sp size
    -> k <= size
    -> M ||- stkSpacePrec (sp - k) (size - k) *^* stkSpacePrec sp k.
Proof.
  intros M sp size k H H0.
  assert (sp >= k).
    destruct H; omega.
  apply sconj_sym.
  arith_rep sp with (sp - k + k).
  arith_rep (sp - k + k - k) with (sp - k).
  apply stkSpacePrec_split; trivial.
  arith_rep (sp - k + k) with sp.
  arith_rep (size - k + k) with size.
Qed.
Implicit Arguments stkSpacePrec_splith [M sp size k].

Lemma stkSpacePrec_splith1 : forall M sp size k A, 
  M ||- stkSpacePrec sp size *^* A
    -> k <= size
    -> M ||- stkSpacePrec sp k *^* stkSpacePrec (sp - k) (size - k) *^* A.
Proof.
  intros M sp size k A H1 H2.
  destruct H1 as [M1 [M2 [HJM [HM1 HM2]]]].
  exists M1; exists M2; split_l.
  split_r.
  apply sconj_sym.
  apply stkSpacePrec_splith; trivial.
Qed.
Implicit Arguments stkSpacePrec_splith1 [M sp size k A].

Lemma stkSpacePrec_split1 : forall M sp size k A, 
  M ||- stkSpacePrec (sp + k) (size + k) *^* A
    -> M ||- stkSpacePrec (sp+k) k *^* stkSpacePrec sp size *^* A.
Proof.
  intros M sp size k A H.
  destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
  exists M1; exists M2; split_l.
  split_r.
  apply stkSpacePrec_split; trivial.
Qed.
Implicit Arguments stkSpacePrec_split1 [M sp size k A].

Lemma stkSpacePrec_merge : forall M sp size k,
  M ||- stkSpacePrec (sp - k) (size - k) *^* stkSpacePrec sp k
    -> sp >= k
    -> size >= k 
    ->  M ||- stkSpacePrec sp size.
Proof.
  intros M sp size k H H0 H1.
  arith_rep sp with (sp - k + k).  
  arith_rep size with (size - k + k).
  destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
  assert (MemMod.disjUnion M2 M1 M).
    auto with PMap.
  destruct H.
  subst M.
  unfold mjudge.
  apply stkSpacePrec_merge.
  arith_rep (sp - k + k) with sp.
  arith_rep (sp - k + k - k) with (sp - k).
Qed.
Implicit Arguments stkSpacePrec_merge [M sp size k].

Lemma stkSpacePrec_merge1 : forall M sp size k A, 
  M ||- stkSpacePrec (sp - k) (size - k) *^* stkSpacePrec sp k *^* A
    -> sp >= k
    -> size >= k 
    ->  M ||- stkSpacePrec sp size *^* A.
Proof.
  intros M sp size k A H H0 H1.
  destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
  exists M1; exists M2; split_l.
  split_r.
  apply (stkSpacePrec_merge HM1 H0 H1); trivial.
Qed.
Implicit Arguments stkSpacePrec_merge1 [M sp size k A].

Lemma stkSpacePrec_mergeh : forall M sp size k,
  M ||- stkSpacePrec sp size *^* stkSpacePrec (sp+k) k
    ->  M ||- stkSpacePrec (sp + k) (size + k).
Proof.
  intros M sp size k H.
  unfold mjudge.
  substH H with (sconj_sym H).
  destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
  destruct HJM; subst M.
  apply Assertion.stkSpacePrec_merge; trivial.
  arith_rep (sp + k - k) with sp.
Qed.
Implicit Arguments stkSpacePrec_mergeh [M sp size k].

Lemma stkSpacePrec_mergeh1 : forall M sp size k A, 
  M ||- stkSpacePrec sp size *^* stkSpacePrec (sp+k) k *^* A
    ->  M ||- stkSpacePrec (sp + k) (size + k) *^* A.
Proof.
  intros M sp size k A H.
  destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
  exists M1; exists M2; split_l.
  split_r.
  apply stkSpacePrec_mergeh; trivial.
Qed.
Implicit Arguments stkSpacePrec_mergeh1 [M sp size k A].

Ltac stk_split k :=
  match goal with 
    | |- _ ||- stkSpacePrec (?sp + k) (?size + k) => apply stkSpacePrec_mergeh
    | |- _ ||- sconj (stkSpacePrec (?sp + k) (?size +k)) _ => apply stkSpacePrec_mergeh1
    | |- _ ||- stkSpacePrec ?sp ?size 
      => apply stkSpacePrec_merge with k;
        [ idtac 
          | simpl; auto with arith; omega || fail 1 "Error: top address of stake should be >= k" 
          | simpl; auto with arith; omega || fail 1 "Error: size of stake should be >= k" 
        ]
    | |- _ ||- sconj (stkSpacePrec ?sp ?size) _ 
      => apply stkSpacePrec_merge1 with k;
        [ idtac 
          | simpl; auto with arith; omega || fail 1 "Error: top address of stake should be >= k" 
          | simpl; auto with arith; omega || fail 1 "Error: size of stake should be >= k" 
        ]
    | _ => fail 1 "Error: cannot split the current goal"
  end.

Ltac stk_splitH H k :=
  match type of H with
    | _ ||- stkSpacePrec (?sp + k) (?size + k) 
      => hyprep H (stkSpacePrec_split H)
    | _ ||- sconj (stkSpacePrec (?sp + k) (?size + k)) _ 
      => hyprep H (stkSpacePrec_split1 H)
    | _ ||- stkSpacePrec ?sp ?size
      => let Hx := fresh "H" in 
        (assert (Hx : size >= k); 
          [ simpl; auto with arith; omega || fail 1 "Error: size of stake should be >= k" | hyprep H (stkSpacePrec_splith H Hx); clear Hx])
    | _ ||- sconj (stkSpacePrec ?sp ?size) _ 
      => let Hx := fresh "H" in 
        (assert (Hx : size >= k); 
          [ simpl; auto with arith; omega || fail 1 "Error: size of stake should be >= k" | hyprep H (stkSpacePrec_splith1 H Hx); clear Hx])
    | _ => fail 1 "Error: cannot split the current hypothesis"
  end.

Tactic Notation "sep" "stksplit" := stk_splitw.

Tactic Notation "sep" "stksplit" "in" hyp (H) := stk_splitwH H.

Tactic Notation "sep" "stksplit" constr (k) := stk_split k.

Tactic Notation "sep" "stksplit" constr (k) "in" hyp (H) := stk_splitH H k.

(**************************************************************)
(* memory re-order macth                                      *)
(**************************************************************)

Ltac matchhp ig ih sg sh Hp :=
  match goal with
    | |- _ ||- top => apply top_intro (* true *)
    | |- ?h ||- ?A => 
      match type of Hp with
        | h ||- A => exact Hp (* assumption *)
      end
      
    | |- _ ||- unit => try (apply unit_intro)
    | |- _ ||- bang _ => apply bang_intro; auto

    | |- ?h ||- ptoxW ?l => (* match: ptoW |- ptoxW *)
      match type of Hp with
        | h ||- ptoW l ?v => 
          apply (ptoW_ptoxW Hp); trivial; fail 1 "Error : unknown"
      end

    | |- _ ||- sconj _ unit => apply unit_intro_right; 
      matchhp (pred ig) sh (pred sg) sh Hp
    | |- _ ||- sconj unit _ => apply unit_intro_left;
      matchhp (pred ig) sh (pred sg) sh Hp

    | |- _ ||- sconj (bang _) _ => apply bang_intro_left; 
      [auto | matchhp (pred ig) sh (pred sg) sh Hp]
    | |- _ ||- sconj _ (bang _) => apply bang_intro_right; 
      [auto | matchhp (pred ig) sh (pred sg) sh Hp]
      
(*     | [NI : ~ set_in ?x ?s |- *)
(*       ?h ||- fas ?P (set_add ?x ?s)] => apply fas_intro1;  *)
(*     [assocr; matchhp (S ig) sh (S sg) sh | assumption] *)

(*     | [NI : ~ set_in ?x ?s |- *)
(*       ?h ||- sconj (fas ?P (set_add ?x ?s)) _] => apply fas_intro;  *)
(*      [assocr; matchhp (S ig) sh (S sg) sh | assumption] *)
      
    | |- _ ||- sconj _ (bang _) => apply bang_intro_right; 
      [auto | matchhp (pred ig) sh (pred sg) sh Hp]

(*     | |- ?h ||- sconj (ptoW ?l ?v) ?C => (\* match left pto *\) *)
(*       match type of Hp with *)
(*         | h ||- sconj (ptoW ptoW l ?v2) ?B => *)
(*           let h1 := fresh "h" in *)
(*             (apply (sconj_mono_left Hp); [] *)
(*               (clear h Hp; intros h Hp) || (clear Hp; intros h1 Hp); *)
(*                 assocl Hp; assocr; *)
(*                 matchhp (pred ig) (pred sh) (pred sg) (pred sh) Hp) *)
(*       end *)

    | |- ?h ||- sconj ?A ?C => (* match left *)
      match type of Hp with
        | h ||- sconj A ?B => 
          let h1 := fresh "h" in
            (apply (sconj_mono_right Hp);
              (clear h Hp; intros h Hp) || (clear Hp; intros h1 Hp);
                assocl Hp; assocr;
                matchhp (pred ig) (pred sh) (pred sg) (pred sh) Hp)
      end

    | |- ?h ||- sconj (ptoxW ?l) ?C => (* match left: ptoW |- ptoxW *)
      match type of Hp with
        | h ||- sconj (ptoW l ?v) ?B => 
          let h1 := fresh "h" in
            (apply (sconj_mono (A':=ptoxW l)(B':=C) Hp);
              (clear h Hp; intros h Hp) || (clear Hp; intros h1 Hp);
                [ apply (ptoW_ptoxW Hp); fail 1 "Error : unknown"
                  | 
                  assocl Hp; assocr;
                  matchhp (pred ig) (pred sh) (pred sg) (pred sh) Hp ])
      end

    | |- ?h ||- sconj ?C ?B => (* match right *)
      match type of Hp with
        | h ||- sconj ?A B => 
          let h' := fresh "h" in
            (apply (sconj_mono_left Hp);
              (clear h Hp; intros h Hp) || (clear Hp; intros h' Hp);
                assocl Hp; assocr;
                matchhp (pred ig) (pred sh) (pred sg) (pred sh) Hp)
      end

    | |- ?h ||- sconj ?C ?D => (* shift *)
      match type of Hp with
        | h ||- sconj ?A ?B => 
          match eval compute in ig with
            | O => (* no more shift on goal *)
              match eval compute in ih with
            | O => idtac (* finished *)
                | _ => (* shift hyp, continue with matching *)
                  hyprep Hp (sconj_sym Hp); assocl Hp;
                  matchhp sg (pred ih) sg sh Hp
              end
            | _ => (* shift goal *)
              apply sconj_sym; assocr;  matchhp (pred ig) ih sg sh Hp
          end
      end
    | _ => match type of Hp with
             | _ ||- sconj _ unit => hyprep Hp (unit_elim_right Hp); 
               matchhp ig (pred sh) sg (pred sh) Hp
             | _ ||- sconj unit _ => hyprep Hp (unit_elim_left Hp);
               matchhp ig (pred sh) sg (pred sh) Hp
             | _ => idtac
           end
(*     | _ => idtac *)
  end.

Ltac simplsep sizeg sizeh Hp :=
  match type of Hp with
    | ?h ||- _ =>
      match goal with
        | |- h ||- _ =>
          assocl Hp; assocr; matchhp sizeg sizeh sizeg sizeh Hp
      end
 end.

Tactic Notation "sep" "matchH" hyp (H) "with" constr (size) :=
  simplsep size size H.


(**************************************************************)
(* memory macth                                               *)
(**************************************************************)

Ltac sep_trivial :=
  match goal with 

    | |- emp ||- unit =>
      apply unit_intro; trivial

    | |- ?M ||- top =>
      apply top_intro; trivial

    | H: ?M ||- pto ?l ?v 
      |- ?M ||- ptox ?l => apply pto_ptox with v; trivial

    | H : ?M ||- stkSpacePrec ?sp 2
       |- ?M ||- ptoxW (?sp-2) =>
        apply (stkSpacePrec_ptoxW H); trivial

    | H : ?M ||- ptoxW ?l 
       |- ?M ||- stkSpacePrec (?l+2) 2 =>
        apply (ptoxW_stkSpacePrec H); trivial

    | H : ?M ||- (stkSpacePrec ?sp ?size)
       |- ?M ||- stkSpace ?sp ?size =>
        apply (stkSpacePrec_stkSpace H); trivial

    | H : ?M ||- sconj (stkSpacePrec ?sp ?size) _
       |- ?M ||- stkSpace ?sp ?size =>
        apply (stkSpacePrec_stkSpace1 H); trivial

    | H : ?M ||- sconj (ptoW ?l ?v) _
       |- ?M ||- leadsToW ?l ?v => (* match: ptoW |- leadsToW *)
          apply (ptoW_leadsToW1 H); trivial

    | H : ?M ||- sconj (ptoW ?l ?v) _
       |- ?M ||- leadsToxW ?l => (* match: ptoW, ptoxW |- leadsToxW *)
          apply (ptoW_leadsToxW1 H); trivial

    | H : ?M ||- sconj (ptoxW ?l) _
       |- ?M ||- leadsToxW ?l => (* match: ptoW, ptoxW |- leadsToxW *)
          apply (ptoxW_leadsToxW1 H); trivial

    | H: ?M ||- ptoW ?l ?v 
      |- ?M ||- ptoxW ?l => apply ptoW_ptoxW with v; trivial

    | H: ?M ||- leadsToW ?l ?v 
      |- ?M ||- leadsToxW ?l => apply leadsToW_leadsToxW with v; trivial

    | H: ?M ||- ?A
      |- ?M ||- ?B => auto with splogic; fail

    | |- ?A ==> ?B => intros; auto with splogic; fail

    | _ => trivial
  end. 

Tactic Notation "sep" "trivial" := sep_trivial.

Ltac sep_simpl :=
  match goal with 

(* unit *)
    | H : ?M ||- unit ** ?A
       |- _  => 
        hyprep H (unit_elim_left H)

    |  |- ?M ||- unit ** ?A =>
        apply unit_intro_left

    | H : ?M ||- ?A ** unit
       |- _  => 
        hyprep H (unit_elim_right H)

    |  |- ?M ||- ?A ** unit =>
        apply unit_intro_right

(* stkSpacePrec empty *)
    | H : ?M ||- (stkSpacePrec ?sp 0) ** ?A
       |- _  =>
        hyprep H (stkSpacePrec_unit1 H)

    | H : ?M ||- stkSpacePrec ?sp 0
       |- _  =>
        hyprep H (stkSpacePrec_unit H)

    | |- ?M ||- sconj (stkSpacePrec ?sp 0) _ =>
      apply stkSpacePrec_unit1

    | |- ?M ||- stkSpacePrec ?sp 0 =>
      apply stkSpacePrec_unit

    | _ => trivial
  end.

Tactic Notation "sep" "simpl" := sep_simpl.


(* obselete *)
Ltac sep_elim_left H :=
  match type of H with
    
    | ?M ||- unit =>
      hyprep H (unit_elim H); subst M
    
    | ?M ||- (ptoxW ?l) =>
      let v:= fresh "v" in
      (hyprep H (ptoxW_ptoW H); destruct H as [v H])

    | ?M ||- sconj (ptoxW ?l) ?A =>
      let v:= fresh "v" in
      (hyprep H (ptoxW_ptoW1 H); destruct H as [v H])

    | _ => fail 1 "Format Error : M |- l |-> _ * _ "
  end.

Tactic Notation "sep" "elim_left" hyp (H) := sep_elim_left H.

(* new *)

Ltac sep_match_left H :=
  match goal with 
    | |- ?M ||- sconj ?A' ?B' =>
      match type of H with
        | M ||- sconj ?A ?B => 
          let M1 := fresh "M" in
            (apply (sconj_mono (A':=A') (B':=B') H); 
              (clear M H; intros M H; sep_trivial) 
              || (clear H; intros M1 H; sep_trivial)) 
      end
  end.

Tactic Notation "sep" "match_left" hyp (H) := sep_match_left H.

Ltac sep_subst_left A :=
  match goal with
    | |- ?M ||- ?A' ** ?B' =>
      apply sconj_mono_left with A; [ idtac | sep_trivial ]
  end.

Ltac sep_subst_leftH H A' :=
  match type of H with
    | ?M ||- ?A ** ?B =>
      let Hx := fresh "Hx" in 
        (assert (Hx : A ==> A'); 
          [sep_trivial 
            | hyprep H (sconj_mono_left H (A':=A') Hx); clear Hx])
  end.

Tactic Notation "sep" "subst_left" "with" constr (A) "in" hyp (H) := sep_subst_leftH H A.

Tactic Notation "sep" "subst_left" "with" constr (A) := sep_subst_left A.

(* new *)
Ltac sep_elim H :=
  match type of H with
    
    | ?M ||- unit =>
      hyprep H (unit_elim H); subst M

    | ?M ||- unit ** ?A =>
      hyprep H (unit_elim_left H)
    
    | ?M ||- (ptoW ?l ?v) =>
      hyprep H (ptoW_pto H)

    | ?M ||- (ptoW ?l ?v) ** ?A =>
      hyprep H (ptoW_pto1 H)

    | ?M ||- (ptoxW ?l) =>
      let v:= fresh "v" in
      (hyprep H (ptoxW_ptoW H); destruct H as [v H])

    | ?M ||- sconj (ptoxW ?l) ?A =>
      let v:= fresh "v" in
      (hyprep H (ptoxW_ptoW1 H); destruct H as [v H])

    | ?M ||- (leadsToxW ?l) =>
      let v:= fresh "v" in
      (hyprep H (leadsToxW_leadsToW H); destruct H as [v H])

    | _ => fail 1 "Format Error : M |- l |-> _ * _ "
  end.

Tactic Notation "sep" "elim" hyp (H) := sep_elim H.

Tactic Notation "sep" "destruct" hyp (H) := sep_elim H.

Ltac sep_exists_left v :=
  match goal with

    | |- ?M ||- sconj (ptoxW ?l) ?A =>
      apply ptoW_ptoxW1 with v

    | |- ?M ||- ptoxW ?l =>
      apply ptoW_ptoxW with v

    | |- ?M ||- leadsToxW ?l =>
      apply leadsToW_leadsToxW with v
    | _ => fail 1 "Format Error : M |- l |-> _ * _ "
  end.

Tactic Notation "sep" "exists_left" constr (v) := sep_exists_left v.

Tactic Notation "sep" "exists" constr (v) := sep_exists_left v.

(* some tactics added by Xinyu *)
Ltac sep_leqv_subst m :=
  match goal with
    | |- ?M ||- ?A *^* ?B =>
      let Hx := fresh in
        (assert (Hx : leqvM m A);
          [(auto || idtac) | (apply (leqvM_app_subst (H := M) (mstar_eqv_subst_left B Hx)); auto; clear Hx) || fail 1 "failed" ])
    | _ => idtac
  end.

Tactic Notation "sep" "leqv_subst" "with" constr (m) := sep_leqv_subst m.

Ltac sep_leqv_substH H m :=
  match type of H with
    | ?M ||- ?A *^* ?B =>
      let Hx := fresh in
        (assert (Hx : leqvM A m); 
        [(auto || idtac) | hypreplace H (leqvM_app_subst (mstar_eqv_subst_left B Hx) H)])
    | _ => idtac
  end.

Tactic Notation "sep" "leqv_subst" "in" hyp (H) "with" constr (m) := sep_leqv_substH H m.

Ltac sep_weaker_subst H m :=
  match type of H with
    | ?M ||- ?A *^* ?B =>
      let Hx := fresh in
        (assert (Hx : forall H, A H -> m H);
          [(auto || idtac) | hypreplace H (mstar_subst_weaker_left H Hx)])
    | _ => idtac
  end.

Tactic Notation "sep" "weaker_subst" "in" hyp (H) "with" constr (m) := sep_weaker_subst H m.

Lemma stkSpace_sp_ge_sz:
forall M sp sz,
  M ||- stkSpace sp sz -> sp >= sz.
Proof.
  intros.
  sep off. unfold stkSpace in H.
  inversion H; auto.
Qed.

Lemma stkSpace_sp_ge_sz1:
forall M sp sz B,
  M ||- stkSpace sp sz *^* B -> sp >= sz.
Proof.
  intros.
  unfold mstar in H.
  sep off.
  InvertAll.
  apply stkSpace_sp_ge_sz with x; auto.
Qed.

Lemma stkSpacePrec_sp_ge_sz:
forall M sp sz,
  M ||- stkSpacePrec sp sz -> sp >= sz.
Proof.
  intros. sep off.
  unfold stkSpacePrec in H.
  inversion H; auto.
Qed.


Lemma stkSpacePrec_sp_ge_sz1:
forall M sp sz B,
  M ||- stkSpacePrec sp sz *^* B -> sp >= sz.
Proof.
  intros. sep off.
  unfold mstar in H.
  InvertAll.
  apply stkSpacePrec_sp_ge_sz with x; auto.
Qed.


Ltac sep_update2 :=
  match goal with
    | |- MemUpdW ?M ?l ?v ?junk ||- ptoW ?l ?v 
      => apply ptoxW_update
    | |- MemUpdW ?M ?l ?v ?junk ||- ptoxW ?l 
      => apply ptoW_ptoxW with v; apply ptoxW_update
    | |- MemUpdW ?M ?l ?v ?junk ||- (ptoW ?l ?v) *^* ?B => 
      apply ptoxW_update1
    | |- MemUpdW ?M ?l ?v ?junk ||- (ptoxW ?l) *^* ?B => 
      apply ptoW_ptoxW1 with v; apply ptoxW_update1
    | |- MemUpdW ?M ?l ?v ?junk ||- (leadsToW ?l ?v)
      => apply leadsToxW_update
    | |- MemUpdW ?M ?l ?v ?junk ||- (leadsToxW ?l)
      => apply leadsToW_leadsToxW with v; apply leadsToxW_update
    | |- MemMod.put ?M ?l ?v ||- (pto ?l ?v) *^* ?B 
      => apply ptox_update1
    | _ => fail 1 "Error: format"
  end.


Tactic Notation "sep" "upd2" := sep_update2.


Lemma ptoW_precise:
forall l v, precise (l |->w v).
Proof.
  intros.
  unfold precise.
  intros.
  assert (exists x, M1 ||- l |-> v *^* l + 1 |-> x).
  apply ptoW_elim; auto.
  assert (exists x, M2 ||- l |-> v *^* l + 1 |-> x).
  apply ptoW_elim; auto.
  inversion_clear H3. inversion_clear H4.
  assert (((l |-> v) *^* (l + 1 |-> x) *^* mTrue) M).
  assert (exists M1', MemMod.disjUnion M1 M1' M).
  auto with PMap.
  inversion_clear H4.
  exists M1. exists x1. split; auto.
  split; auto.
  unfold top. auto.
  assert (((l |-> v) *^* (l + 1 |-> x0) *^* mTrue) M).
  assert (exists M2', MemMod.disjUnion M2 M2' M).
  auto with PMap.
  inversion_clear H6.
  exists M2. exists x1. split; auto.
  split; auto.
  unfold top. auto.
  assert (x = x0).
  sep on.
  sep lift 1 in H6.
  sep lift 1 in H4.
  apply pto_pto_eq1 with M (l+1) (sconj top (l |-> v)) (sconj top (l |-> v)); auto.
  subst x0.
  sep off.
  assert (precise (sconj (l |-> v) (l + 1 |-> x))).
  apply mstar_presv_prec.
  apply pto_precise. apply pto_precise.
  unfold precise in H7.
  apply H7 with M; auto.
Qed.

Lemma ptoW_indom1:
forall M l v A,
  (l |->w v *^* A) M -> MemMod.in_dom l M /\ MemMod.in_dom (l+1) M.
Proof.
  intros.
  unfold mstar in H.
  InvertAll.
  assert (MemMod.in_dom l x /\ MemMod.in_dom (l+1) x).
  apply ptoW_indom with v; auto.
  inversion_clear H1.
  inversion_clear H0.
  subst M.
  split; apply MemMod.merge_presv_indom_left; auto.
Qed.

Lemma ptoxW_indom1:
forall M l A,
  (l |->w_ *^* A) M -> MemMod.in_dom l M /\ MemMod.in_dom (l+1) M.
Proof.
  intros.
  assert (exists v, M ||- (l |->w v *^* A)).
  apply ptoxW_ptoW1. auto.
  inversion_clear H0.
  apply ptoW_indom1 with x A; auto.
Qed.
