Require Import Coqlib.
Require Import Map.
Require Import Arith.
Require Import Machine.
Require Import Mem_Lemmas.
Require Import Assertion.
Require Import Omega.

Lemma prepost_weaken_pre : forall (mP mP' mQ: MPred) S S',
  prepost mP mQ S S'
  -> (forall M, mP M -> mP' M)
  -> prepost mP' mQ S S'.
Proof.
  unfold prepost.
  intros mP mP' mQ S S' HSS' H.
  destruct HSS' as [M1 [M2 [M1' [HM1 [HM1' [HJM HJM']]]]]].
  exists M1; exists M2; exists M1'.
  auto.
Qed.
Implicit Arguments prepost_weaken_pre [mP mP' mQ S S'].

Lemma prepost_weaken_post : forall (mP mQ mQ': MPred) S S',
  prepost mP mQ S S'
  -> (forall M, mQ M -> mQ' M)
  -> prepost mP mQ' S S'.
Proof.
  unfold prepost.
  intros mP mQ mQ' S S' HSS' H.
  destruct HSS' as [M1 [M2 [M1' [HM1 [HM1' [HJM HJM']]]]]].
  exists M1; exists M2; exists M1'.
  auto.
Qed.
Implicit Arguments prepost_weaken_post [mP mQ mQ' S S'].

Lemma precise_Aemp : precise Aemp.
Proof.
  unfold precise, Aemp.
  intros M1 M2 M Hsub1 Hsub2 HM1 HM2.
  subst M1 M2.
  trivial.
Qed.

Lemma prec_disjUnion_inv : forall M M1 M2 M1' M2' (Inv : MPred),
  MemMod.disjUnion M1 M2 M
  -> MemMod.disjUnion M1' M2' M
  -> Inv M2
  -> Inv M2'
  -> precise Inv
  -> M1 = M1' /\ M2 = M2'.
Proof.
  intros.
  unfold precise in H3.
  assert (MemMod.subseteq M2 M).
    eapply MemMod.disjUnion_subseteq_right; eauto.
  assert (MemMod.subseteq M2' M).
    eapply MemMod.disjUnion_subseteq_right; eauto.
  assert (Heq := H3 _ _ _ H4 H5 H1 H2).
  split_r.
  eapply MemMod.disjUnion_unique with (M5:=M); eauto.
  auto with PMap.
  auto with PMap.
Qed.
Implicit Arguments prec_disjUnion_inv [M M1 M2 M1' M2' Inv].

Lemma prec_disjUnion_inv_left : forall M M1 M2 M1' M2' (Inv : MPred),
  MemMod.disjUnion M1 M2 M
  -> MemMod.disjUnion M1' M2' M
  -> Inv M1
  -> Inv M1'
  -> precise Inv
  -> M1 = M1' /\ M2 = M2'.
Proof.
  intros.
  assert (MemMod.disjUnion M2 M1 M).
    auto with PMap.
  assert (MemMod.disjUnion M2' M1' M).
    auto with PMap.
  assert (Hx := prec_disjUnion_inv H4 H5 H1 H2 H3).
  destruct Hx; auto.
Qed.
Implicit Arguments prec_disjUnion_inv_left [M M1 M2 M1' M2' Inv].

Lemma prepost_elim : forall mP mQ A S S',
  prepost mP mQ S S'
  -> precise mP
  -> (mP *^* A) (S_Mem S)
  -> (mQ *^* A) (S_Mem S').
Proof.
  unfold prepost.
  intros mP mQ A S S' HSS' HmP HS.
  destruct HSS' as [M1 [M2 [M1' [HM1 [HM1' [HJM HJM']]]]]].
  destruct HS as [M1x [M2x [HJMx [HM1x HM2x]]]].
  destruct (prec_disjUnion_inv_left HJM HJMx HM1 HM1x HmP).
  subst M1x M2x; clear HJMx.
  exists M1'; exists M2; auto.
Qed.
Implicit Arguments prepost_elim [mP mQ A S S'].

Lemma prepost_elim_emp : forall S S',
  prepost Aemp Aemp S S'
  -> (S_Mem S') = (S_Mem S).
Proof.
  intros S S'.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  simplS.
  unfold prepost, Aemp.
  intros H.
  destruct H as [M1 [M2 [M1' H]]].
  destruct H as [HM1 [HM1' [HJM HJM']]].
  subst M1 M1'.
  simplS in * .
  rewrite (MemMod.emp_disjUnion_unique (MemMod.disjUnion_commut HJM)) in HJM'.
  rewrite (MemMod.emp_disjUnion_unique (MemMod.disjUnion_commut HJM')).
  trivial.
Qed.

Lemma prepost_elim_emp_M : forall M R fl isr M' R' fl' isr',
  prepost Aemp Aemp (M, R, fl, isr) (M', R', fl', isr')
  -> M' = M.
Proof.
  intros M R fl isr M' R' fl' isr'.
  unfold prepost, Aemp.
  intros H.
  destruct H as [M1 [M2 [M1' H]]].
  destruct H as [HM1 [HM1' [HJM HJM']]].
  subst M1 M1'.
  simplS in * .
  rewrite (MemMod.emp_disjUnion_unique (MemMod.disjUnion_commut HJM)) in HJM'.
  rewrite (MemMod.emp_disjUnion_unique (MemMod.disjUnion_commut HJM')).
  trivial.
Qed.
