Add LoadPath "..".

Require Import Coqlib.
Require Import MultiMapT.
Require Import EqNat.

Require Import Machine.
Require Import Assertion.
Require Import Operational_Semantics.

Notation LangID := nat   (only parsing).

Inductive Spec0 : Type :=
  packS0 : LangID -> Spec -> Spec0.

Module OSpec0 <: ONE_TYPE.
  Definition B := Spec0.
End OSpec0.

(* psi : PSpec *)
Module OPSpec := MultiMap OSpec0.
Definition PSpec := OPSpec.map.

Definition emptCT := OPSpec.emp_map.
Definition consCT : flab -> Spec0 -> PSpec -> PSpec :=
  fun f theta Psi => OPSpec.merge (OPSpec.sig_map f theta) Psi.

(* look up code heap type *)
Definition lookupCT : PSpec -> flab -> Spec0 -> Prop :=
  fun Psi f s =>   OPSpec.lookup Psi f s.

(* assertion be in domian PSpec type *)
Definition inCT : PSpec -> flab -> Prop :=
  fun Psi f => OPSpec.indom Psi f.

(** Subset of PSpec, subsetCT Psi Psi' means: Psi is a subset of Psi' **)
Definition subsetCT : PSpec -> PSpec -> Prop :=
  fun Psi Psi' => OPSpec.sub Psi Psi'.

Definition Assert := PSpec -> State -> Prop.

Definition ITP := Spec -> Assert.

Definition LDict := LangID -> OptionT ITP.

(** Subset of LDict, subsetD D D' means: D is a subset of D' **)
Definition subsetD : LDict -> LDict -> Prop :=
  fun D D' => forall (rho : LangID) (itp : ITP),
  D rho = SomeT1 _ itp -> D' rho = SomeT1 _ itp.

Definition ext_dict (D : LDict) (rho : LangID) (itp : ITP) : LDict :=
  fun (rho' : LangID) =>
    if (beq_nat rho' rho) then SomeT1 _ itp else D rho'.

Lemma ext_dict_subsetD : forall D rho itp,
  D rho = NoneT1 ITP
  -> subsetD D (ext_dict D rho itp).
Proof.
  unfold subsetD, ext_dict; intros.
  compare rho0 rho; intros.
  subst rho0.
  rewrite H in H0.
  discriminate.
  generalize (refl_equal (beq_nat rho0 rho)).
  pattern rho0 at 2.
  destruct (beq_nat rho0 rho); intros.
    rewrite (beq_nat_eq _ _ H1) in n.
    contradiction n; trivial.
  trivial.
Qed.

Parameter itr_spec : Spec.
Parameter itr_ID : LangID.
Parameter itr_Itp : ITP.

Definition itr_theta := packS0 itr_ID itr_spec.

Definition AFalse := fun (Psi : PSpec) (S : State) => False.

Definition Itp0 (D : LDict) (theta : Spec0) : Assert :=
  match theta with
    packS0 rho s =>
        match D rho with
          NoneT1 => AFalse
        | SomeT1 itp => itp s
        end
  end.

Definition Mono_CT_assert (a : Assert) : Prop :=
  forall Psi S, a Psi S -> forall Psi', subsetCT Psi Psi' -> a Psi' S.

Definition Mono_CT_PSpec (D : LDict) (Psi : PSpec) : Prop :=
  forall f theta0, lookupCT Psi f theta0 -> Mono_CT_assert (Itp0 D theta0).

Definition Mono_CT_Itp (D : LDict) := 
  forall theta0, Mono_CT_assert (Itp0 D theta0).

Definition WFcmd_aux (D : LDict) (Psi : PSpec) (theta : Spec0) (f : flab) (c : comm) : Prop :=
  forall Psi' S,
    (Itp0 D theta) Psi' S
    -> subsetCT Psi Psi' 
    -> (exists S', nextS f c S S') 
        /\ (forall f' S', nextS f c S S' -> nextPC c (S_Mem S) (S_RF S) (S_Flag S) f f' 
               -> exists theta', lookupCT Psi' f' theta' /\ Itp0 D theta' Psi' S')
        /\ (~ itr_enabled S 
            \/ (exists S', itr_nextS f S S' /\ forall S', itr_nextS f S S' 
               -> exists theta', lookupCT Psi' h_entry theta' /\ Itp0 D theta' (consCT f theta Psi') S')
           ).

(*
Definition WFcmd_aux (D : LDict) (Psi : PSpec) (a : Assert) (f : flab) (c : comm) : Prop :=
  forall Psi' S,
    a Psi' S
    -> subsetCT Psi Psi' 
    -> (exists S', nextS f c S S') 
        /\ (forall f' S', nextS f c S S' -> nextPC c (S_Mem S) (S_RF S) (S_Flag S) f f' 
               -> exists theta, lookupCT Psi' f' theta /\ Itp0 D theta Psi' S')
        /\ (~ itr_enabled S 
            \/ (exists S', itr_nextS f S S' /\ forall S', itr_nextS f S S' 
               -> exists theta, lookupCT Psi' h_entry theta /\ Itp0 D theta Psi' S')
           ).
*)


Definition WFcmd (D : LDict) (Psi : PSpec) (theta : Spec0) (f : flab) (c : comm) : Prop :=
  WFcmd_aux D Psi theta f c /\ Mono_CT_assert (Itp0 D theta).

(*
Definition WFcmd (D : LDict) (Psi : PSpec) (a : Assert) (f : flab) (c : comm) : Prop :=
  WFcmd_aux D Psi a f c /\ Mono_CT_assert a.
*)

Inductive WFiseq (D : LDict) (Psi : PSpec) : Spec0 -> flab -> iseq -> Prop :=
    wfis_seq : 
      forall (theta : Spec0) (f: flab) (i : instr) (is : iseq) (theta' : Spec0), 
        WFcmd D (consCT (f+1) theta' Psi) theta f (comml i)
        -> WFiseq D Psi theta' (f+1) is
        -> WFiseq D Psi theta f (seq i is)
  | wfis_end :
      forall (theta : Spec0) (f: flab) (si : sinstr),
        WFcmd D Psi theta f (comms si)
        -> WFiseq D Psi theta f (sseq si).


(*
Inductive WFiseq (D : LDict) (Psi : PSpec) : Assert -> flab -> iseq -> Prop :=
    wfis_seq : 
      forall (a : Assert) (f: flab) (i : instr) (is : iseq) (theta : Spec0), 
        WFcmd D (consCT (f+1) theta Psi) a f (comml i)
        -> WFiseq D Psi (Itp0 D theta) (f+1) is
        -> WFiseq D Psi a f (seq i is)
  | wfis_end :
      forall (a : Assert) (f: flab) (si : sinstr),
        WFcmd D Psi a f (comms si)
        -> WFiseq D Psi a f (sseq si).
*)


Definition WFcdhp (D : LDict) (Psi : PSpec) (C : CodeHeap) (Psi' : PSpec) : Prop :=
  forall (f : flab) (theta : Spec0),
    lookupCT Psi' f theta 
    -> exists isq : iseq, lookupC C f isq 
                          /\ WFiseq D Psi theta f isq.

(*
Definition WFcdhp (D : LDict) (Psi : PSpec) (C : CodeHeap) (Psi' : PSpec) : Prop :=
  forall (f : flab) (theta : Spec0),
    lookupCT Psi' f theta -> exists isq : iseq, lookupC C f isq /\ 
      WFiseq D Psi (Itp0 D theta) f isq.
*)

Definition WFworld (D : LDict) (Psi : PSpec) (W : World) : Prop :=
  match W with
    (X, pc) =>
    match X with
      (ch, st) =>
          lookupCT Psi h_entry itr_theta 
          /\ D itr_ID = SomeT1 _ itr_Itp
          /\ exists Psi', WFcdhp D Psi ch Psi' 
                    /\ subsetCT Psi Psi' 
                    /\ Mono_CT_PSpec D Psi'
                    /\ exists theta, Itp0 D theta Psi' st
                    /\ exists isq, lookupC ch pc isq
                    /\ WFiseq D Psi theta pc isq
    end
  end.

(*
Definition WFworld (D : LDict) (Psi : PSpec) (W : World) : Prop :=
  match W with
    (X, pc) =>
    match X with
      (ch, st) =>
          lookupCT Psi h_entry itr_theta 
          /\ D itr_ID = SomeT1 _ itr_Itp
          /\ exists Psi', WFcdhp D Psi ch Psi' 
                    /\ subsetCT Psi Psi' 
                    /\ Mono_CT_PSpec D Psi'
                    /\ exists a, a Psi' st
                    /\ exists isq, lookupC ch pc isq
                    /\ WFiseq D Psi a pc isq
    end
  end.
*)

(*****************************************************************************)
(*** Mono related Lemma                                                     **)
(*****************************************************************************)
Lemma WFiseq_mono_a:
  forall D Psi f s isq,
    WFiseq D Psi s f isq
    -> Mono_CT_assert (Itp0 D s).
Proof.
intros D Psi f s isq WFI.
inversion WFI.
(** sequece instructions **)
rename H into WFc.
unfold WFcmd in WFc.
InvertAll; auto.
rename H into WFc.
unfold WFcmd in WFc.
InvertAll; auto.
Qed.

Lemma MonoCT_psi_extend:
  forall (D:LDict) (Psi:PSpec) (f:flab) (s:Spec0),
    Mono_CT_PSpec D Psi
    -> Mono_CT_assert (Itp0 D s)
    -> Mono_CT_PSpec D (consCT f s Psi).
Proof.
unfold Mono_CT_PSpec.
intros D Psi f s monoCT monoa.
intros f' s' lkpCT. 
unfold consCT in lkpCT.
assert (lookupCT (OPSpec.sig_map f s) f' s' \/ lookupCT Psi f' s').
apply OPSpec.lookup_merge_elim; auto.
rename H into lkpCT'.
inversion lkpCT'.
(** f = f' **)
rename H into lkpCTf.
inversion lkpCTf.
rename H into pair.
inversion pair.
rename H1 into eqs.
rewrite <- eqs; auto.
absurd (False); auto.
(** f' in Psi **)
apply monoCT with f'; auto.
Qed.

(*****************************************************************************)
(*** PSPec related Lemma                                                    **)
(*****************************************************************************)
(** Lemma for subset Code Specification **)

Lemma Psi_extend_subsetCT:
  forall Psi Psi' pc s,
    subsetCT Psi Psi'  -> subsetCT Psi (consCT pc s Psi').
Proof.
unfold subsetCT, consCT.
intros.
apply OPSpec.sub_trans with Psi'; auto.
apply OPSpec.sub_merge_r.
Qed.

Lemma subsetCT_refl:
  forall Psi, subsetCT Psi Psi.
Proof.
  unfold subsetCT.
  intros.
  apply OPSpec.sub_refl.
Qed.

Lemma subsetCT_trans:
  forall Psi1 Psi2 Psi3,
    subsetCT Psi1 Psi2 -> subsetCT Psi2 Psi3 -> subsetCT Psi1 Psi3.
Proof.
unfold subsetCT.
intros Psi1 Psi2 Psi3 sub1 sub2 f theta lkpCT.
apply sub2.
apply sub1; auto.
Qed.

Lemma Psi_extend2_subsetCT:
  forall Psi Psi' pc s,
    subsetCT Psi Psi'
    -> subsetCT (consCT pc s Psi) (consCT pc s Psi').
Proof.
unfold subsetCT, consCT.
intros.
apply OPSpec.sub_merge_sub; auto.
Qed.

Lemma lookupCT_subsetCT : forall Psi Psi' l theta0,
  lookupCT Psi l theta0
  -> subsetCT Psi Psi'
  -> lookupCT Psi' l theta0.
Proof.
  unfold lookupCT, subsetCT.
  intros Psi Psi' l theta0 Hlk Hsubset.
  eapply OPSpec.lookup_sub_lookup; eauto.
Qed.

Lemma lookupCT_consCT : forall Psi l theta0,
  lookupCT (consCT l theta0 Psi) l theta0.
Proof.
  unfold lookupCT, consCT.
  intros Psi l theta0.
  apply OPSpec.lookup_merge_intro_l.
  unfold OPSpec.sig_map, OPSpec.lookup.
  simpl.
  left; trivial.
Qed.

Lemma consCT_subsetCT:
  forall Psi pc s,
    subsetCT Psi (consCT pc s Psi).
Proof.
unfold subsetCT, consCT.
intros.
apply OPSpec.sub_merge_r.
Qed.

Lemma Psi_extend3_subsetCT : forall Psi Psi' f s,
  lookupCT Psi' f s
  -> subsetCT Psi Psi'
  -> subsetCT (consCT f s Psi) Psi'.
Proof.
unfold subsetCT,lookupCT.
unfold OPSpec.sub       .
intros.
unfold OPSpec.lookup.
destruct H1.
rewrite <- H1.
apply H.
apply H0.
apply H1.
Qed.

Lemma lookupCT_lookupCT_consCT : forall Psi f f' s s',
  lookupCT Psi f s
  -> lookupCT (consCT f' s' Psi) f s.
Proof.
intros.
apply (lookupCT_subsetCT Psi (consCT f' s' Psi) f s).
apply H.
apply consCT_subsetCT.
Qed.

(** OCAP WF Isequence when Psi enlarge with (f,s) **)

Lemma WFcmd_Mono_CT : forall D Psi Psi' theta f c,
  subsetCT Psi Psi'
  -> WFcmd D Psi theta f c
  -> WFcmd D Psi' theta f c.
Proof.
  unfold WFcmd, WFcmd_aux.
  intros D Psi Psi' theta f c Hsub [H1 H2].
  split_r.
  intros.
  apply H1; trivial.
  eapply subsetCT_trans; eauto.
Qed.

Lemma WFcdhp_psi_extend:
  forall D Psi Psi' ch f s isq,
    lookupC ch f isq
    -> WFiseq D Psi s f isq
    -> WFcdhp D Psi ch Psi' 
    -> WFcdhp D Psi ch (consCT f s Psi').
Proof.
unfold WFcdhp, consCT.
intros D Psi Psi' ch f s isq lkpC2 WFi WFc f' s' lkpCT.
assert (lookupCT (OPSpec.sig_map f s) f' s' \/ lookupCT Psi' f' s').
apply OPSpec.lookup_merge_elim; auto.
rename H into lkpCTor.
inversion lkpCTor.
rename H into lkpCTsig.
inversion lkpCTsig.
rename H into pair.
inversion pair.
rename H0 into eqf.
rename H1 into eqs.
rewrite <- eqf.
rewrite <- eqs.
exists isq. split; auto.
absurd (False); auto.
apply WFc; auto.
Qed.

Hint Resolve Psi_extend_subsetCT.
Hint Resolve subsetCT_refl.
Hint Resolve Psi_extend2_subsetCT.
