Add LoadPath "..".

Require Import Coqlib.
Require Import MultiMapT.
Require Import EqNat.

Require Import Machine.
Require Import Assertion.
Require Import Operational_Semantics.

Require Import CAP0.

Definition compatible_Dict (D D' : LDict) : Prop :=
  forall (rho : LangID),
    (D rho = NoneT1 _)\/ (D' rho = NoneT1 _) \/
      (exists itp,
        D rho = SomeT1 _ itp /\ D' rho = SomeT1 _ itp).

Definition union_Dict (D D' : LDict) : LDict :=
  fun (rho : LangID) =>
    match D rho with
      NoneT1  => D' rho
    | SomeT1 itp => D rho
    end.

Definition Describe_theta (D : LDict) (theta : Spec0) : Prop :=
  exists rho, exists a,
    theta = packS0 rho a /\ (exists itp, D rho = SomeT1 ITP itp).

Definition Describe_Psi (D : LDict) (Psi : PSpec) : Prop :=
  forall f rho a,
    lookupCT Psi f (packS0 rho a) /\ (exists itp, D rho = SomeT1 ITP itp).

Definition union_Psi (Psi Psi':PSpec) : PSpec :=
  OPSpec.merge Psi Psi'.

(*****************************************************************************)
(*** Dictionary related lemmas                                              **)
(*****************************************************************************)
(** (D1 U D2) is the super set of D1 **)
Lemma subsetD_Union_Left:
  forall D1 D2,
    compatible_Dict D1 D2
    -> subsetD D1 (union_Dict D1 D2).
Proof.
intros D1 D2 cpt_D.
unfold subsetD.
intros.
unfold union_Dict.
rewrite H.
auto.
Qed.

(** (D1 U D2) is the super set of D2 **)
Lemma subsetD_Union_Right:
  forall D1 D2,
    compatible_Dict D1 D2
    -> subsetD D2 (union_Dict D1 D2).
Proof.
intros D1 D2 cpt_D.
unfold compatible_Dict in cpt_D.
unfold subsetD.
intros.
assert (D1 rho = NoneT1 ITP \/ 
        D2 rho = NoneT1 ITP \/
          (exists itp : ITP, D1 rho = SomeT1 ITP itp /\ D2 rho = SomeT1 ITP itp)).
apply cpt_D.
inversion H0.
  unfold union_Dict.
  rewrite H1.
  auto.
inversion H1.
  rewrite H2 in H.
  inversion H.

  InvertAll.
  unfold union_Dict.
  induction (D1 rho); auto.
  rewrite H2.
  rewrite <- H4.
  auto.
Qed.

(** OCAP WF command when Dictionary enlarge **)
Lemma WFstate_dict_enlarge:
  forall D1 D2 theta Psi st,
    (Itp0 D1 theta) Psi st
    -> subsetD D1 D2
    -> (Itp0 D2 theta) Psi st.
Proof.
unfold subsetD, Itp0.
intros D1 D2 theta Psi st aS subD.
induction theta.
assert (D1 n = NoneT1 ITP \/ exists itp, D1 n = SomeT1 ITP itp).
induction (D1 n).
left; auto.
right. exists t; auto.
rename H into orItp.
inversion orItp.
rename H into nonItp.
rewrite nonItp in aS.
inversion aS.
InvertAll.
rename x into itp.
rename H0 into smItp.
rewrite smItp in aS.
assert (D2 n = SomeT1 ITP itp).
apply subD. auto.
rename H into smItp'.
rewrite smItp'; auto.
Qed.

(** OCAP WF command when Dictionary enlarge **)
Lemma WFstate_dict_reduce:
  forall D1 D2 rho,
    (exists itp, D1 rho = SomeT1 ITP itp) /\ subsetD D1 D2
    -> forall Psi st a,
      (Itp0 D2 (packS0 rho a)) Psi st
      -> (Itp0 D1 (packS0 rho a)) Psi st.
Proof.
unfold subsetD, Itp0.
intros D1 D2 rho.
intros [[itp desD] subC] Psi st a.
assert (D2 rho = SomeT1 ITP itp).
apply subC; auto.
rename H into desD'.
rewrite desD.
rewrite desD'.
auto.
Qed.

Lemma Mono_dict_extend:
  forall D1 D2 rho,
    (exists itp, D1 rho = SomeT1 ITP itp) /\ subsetD D1 D2
    -> forall a, Mono_CT_assert (Itp0 D1 (packS0 rho a))
      -> Mono_CT_assert (Itp0 D2 (packS0 rho a)).
Proof.
unfold Mono_CT_assert.
intros D1 D2 rho.
intros D1rho a aSp Psi st aS Psi' subCT.
apply WFstate_dict_enlarge with D1.
apply aSp with Psi.
apply WFstate_dict_reduce with D2; auto.
auto.
InvertAll.
auto.
Qed.

(** OCAP WF command when Dictionary enlarge **)
Lemma WFcmd_dict_enlarge:
  forall D1 D2 rho,
    (exists itp, D1 rho = SomeT1 ITP itp) /\ subsetD D1 D2
    -> forall  Psi f theta c, 
      WFcmd D1 Psi (packS0 rho theta) f c
      -> WFcmd D2 Psi (packS0 rho theta) f c.
Proof.
unfold WFcmd.
unfold WFcmd_aux.
intros D1 D2 rho. 
intros itp_subD Psi f theta c.
intros [WFca mono].
split.
intros Psi' st aS subCT.
substH WFca with (WFca Psi' st (WFstate_dict_reduce _ _ _ itp_subD _ _ _ aS) subCT).
destruct WFca as [nS [paS pitrS]].
split_l.
split.
intros f' st' nS' nPC'.
assert ( exists theta : Spec0, lookupCT Psi' f' theta /\ Itp0 D1 theta Psi' st').
apply paS; auto.
destruct H as [theta' [lkpCT aS']].
exists theta'.
split_l.
InvertAll.
apply WFstate_dict_enlarge with D1; auto.
inversion_clear pitrS.
left; auto.
destruct H as [st' [nS_itr pitrS]].
right.
exists st'.
split; auto.
intros.
substH pitrS with (pitrS _ H).
destruct pitrS as [theta' [lkpCT aS_itr]].
exists theta'.
split_l.
InvertAll.
apply WFstate_dict_enlarge with D1; auto.
apply  Mono_dict_extend with D1; auto.
Qed.

(** OCAP WF Isequence when Dictional enlarge **)
Lemma WFiseq_dict_enlarge:
  forall D D' Psi,
    Describe_Psi D Psi /\ subsetD D D'
    -> (forall isq rho theta f, 
      WFiseq D Psi (packS0 rho theta) f isq
      -> WFiseq D' Psi (packS0 rho theta) f isq).
Proof.
intros D D' Psi [disD subD]. 
induction isq.
intros rho a f WFi.
unfold Describe_Psi in disD.
assert (lookupCT Psi f (packS0 rho a) /\
         (exists itp : ITP, D rho = SomeT1 ITP itp)).
apply disD.
destruct H as [lkpCT smD].
inversion_clear WFi.
assert (WFcmd D' (consCT (f + 1) theta' Psi) (packS0 rho a) f (comml i)).
apply WFcmd_dict_enlarge with D; auto.
induction theta'.
assert (WFiseq D' Psi (packS0 n s) (f + 1) isq).
apply IHisq; auto.
rename H1 into WFc'.
rename H2 into WFi'.
gen_clear WFc' WFi'.
apply wfis_seq.

intros rho a f WFi.
unfold Describe_Psi in disD.
assert (lookupCT Psi f (packS0 rho a) /\
         (exists itp : ITP, D rho = SomeT1 ITP itp)).
apply disD.
destruct H as [lkpCT smD].
inversion_clear WFi.
assert (WFcmd D' Psi (packS0 rho a) f (comms s)).
apply WFcmd_dict_enlarge with D; auto.
apply wfis_end; auto.
Qed.

(** OCAP WF Codeheap when Psi enlarge **)
Lemma WFcdhp_dict_enlarge:
  forall D1 D2 Psi,
    Describe_Psi D1 Psi /\ subsetD D1 D2
    -> forall  Psi' ch, WFcdhp D1 Psi ch Psi' 
      -> WFcdhp D2 Psi ch Psi'.
Proof.
unfold WFcdhp.
intros D1 D2 Psi [desD subD] Psi' ch.
intros WFC f theta lkpCT.
assert (exists isq : iseq, lookupC ch f isq /\ WFiseq D1 Psi theta f isq).
apply WFC; auto.
destruct H as [isq [lkpC WFi]].
exists isq.
split_l.
induction theta.
apply WFiseq_dict_enlarge with D1; auto.
Qed.

(*****************************************************************************)
(*** PSPec related lemma for link                                           **)
(*****************************************************************************)
(** if (l theta) in (Psi1 U Psi2), it in Psi1 or Psi2 or in both of Psi1 and Psi2**)
Lemma lkpCT_Union:
  forall Psi1 Psi2 l theta,
    lookupCT (union_Psi Psi1 Psi2) l theta
    ->  lookupCT Psi1 l theta \/ lookupCT Psi2 l theta.
Proof.
unfold lookupCT.
intros.
apply OPSpec.lookup_merge_elim.
auto.
Qed.

(** (Psi1 U Psi2) is the super set of Psi1 **)
Lemma subsetCT_Union_Left:
  forall Psi1 Psi2,
    subsetCT Psi1 (union_Psi Psi1 Psi2).
Proof.
unfold subsetCT, union_Psi.
intros.
apply OPSpec.sub_merge_l.
Qed.

(** (Psi1 U Psi2) is the super set of Psi2 **)
Lemma subsetCT_Union_Right:
  forall Psi1 Psi2,
    subsetCT Psi2 (union_Psi Psi1 Psi2).
Proof.
unfold subsetCT, union_Psi.
intros.
apply OPSpec.sub_merge_r.
Qed.

(** OCAP WF command when Psi enlarge **)
Lemma WFcmd_psi_enlarge:
  forall D Psi Psi' f a c,
    subsetCT Psi Psi'
    -> WFcmd D Psi  a f c
    -> WFcmd D Psi' a f c.
Proof.
unfold WFcmd.
unfold WFcmd_aux.
intros D Psi Psi' f a c subCT.
intros.
InvertAll.
split_r.
intros.
apply H0; auto.
apply subsetCT_trans with Psi'; auto.
Qed.

(** OCAP WF Isequence when Psi enlarge **)
Lemma WFiseq_psi_enlarge:
  forall D Psi Psi',
    subsetCT Psi Psi'
    -> (forall isq a f, 
      WFiseq D Psi a f isq
      -> WFiseq D Psi' a f isq).
Proof.
unfold subsetCT.
intros D Psi Psi' subCT. 
induction isq.
intros a f WFi.
inversion WFi.
apply wfis_seq with theta'; trivial.
apply WFcmd_psi_enlarge with (consCT (f + 1) theta' Psi); trivial.
apply Psi_extend2_subsetCT; auto.
apply IHisq; trivial.
intros a f WFi.
inversion_clear WFi.
assert (WFcmd D Psi' a f (comms s)).
apply WFcmd_psi_enlarge with Psi; auto.
apply wfis_end; auto.
Qed.

(** OCAP WF Codeheap when Psi enlarge **)
Lemma WFcdhp_psi_enlarge:
  forall D Psi1 Psi2 Psi' ch,
    subsetCT Psi1 Psi2
    -> WFcdhp D Psi1 ch Psi' 
    -> WFcdhp D Psi2 ch Psi'.
Proof.
unfold WFcdhp.
intros D Psi1 Psi2 Psi' ch subCT.
intros WFc f theta lkpCT.
assert (exists isq : iseq, lookupC ch f isq /\ WFiseq D Psi1 theta f isq).
apply WFc; auto.
InvertAll.
rename x into isq.
rename H into lkpC.
rename H1 into WFi.
exists isq.
split_l.
apply WFiseq_psi_enlarge with Psi1; auto.
Qed.

(** Lemma 3.3 in OCAP paper **)
Lemma WFcdhp_Enlarge:
  forall D D' Psi Psi' C Psi'',
    WFcdhp D Psi C Psi'' 
    -> subsetCT Psi Psi'
    -> subsetD D D'
    -> Describe_Psi D Psi
    -> Describe_Psi D' Psi'
    -> WFcdhp D' Psi' C Psi''.
Proof.
intros D D' Psi Psi' C Psi'' WFc subCT subD desD desD'.
apply WFcdhp_psi_enlarge with Psi; auto.
apply WFcdhp_dict_enlarge with D; auto.
Qed.

Lemma Describe_dict_union_left:
  forall D1 D2 Psi,
    compatible_Dict D1 D2
    -> Describe_Psi D1 Psi
    -> Describe_Psi (union_Dict D1 D2) Psi.
Proof.
intros D1 D2 Psi.
unfold Describe_Psi.
intros comp_dict des1 f rho a.
substH des1 with (des1 f rho a).
destruct des1 as [lkpCT [itp d1rho]].
unfold compatible_Dict in comp_dict.
assert (D1 rho = NoneT1 ITP \/
              D2 rho = NoneT1 ITP \/
              (exists itp : ITP,
                 D1 rho = SomeT1 ITP itp /\ D2 rho = SomeT1 ITP itp)).
apply comp_dict.
inversion_clear H.
rewrite d1rho in H0.
inversion H0.

inversion_clear H0.
unfold union_Dict.
rewrite d1rho.
split_l.
exists itp; auto.

destruct H as [itp1 [d1rho' d2rho]].
split_l.
exists itp1.
unfold union_Dict.
rewrite d1rho'.
auto.
Qed.

Lemma Describe_dict_union_right:
  forall D1 D2 Psi,
    compatible_Dict D1 D2
    -> Describe_Psi D2 Psi
    -> Describe_Psi (union_Dict D1 D2) Psi.
Proof.
intros D1 D2 Psi.
unfold Describe_Psi.
intros comp_dict des1 f rho a.
substH des1 with (des1 f rho a).
destruct des1 as [lkpCT [itp d2rho]].
unfold compatible_Dict in comp_dict.
assert (D1 rho = NoneT1 ITP \/
              D2 rho = NoneT1 ITP \/
              (exists itp : ITP,
                 D1 rho = SomeT1 ITP itp /\ D2 rho = SomeT1 ITP itp)).
apply comp_dict.
inversion_clear H.
unfold union_Dict.
rewrite H0.
split_l.
exists itp; auto.

inversion_clear H0.
rewrite d2rho in H.
inversion H.

destruct H as [itp1 [d1rho d2rho']].
split_l.
exists itp1.
unfold union_Dict.
rewrite d1rho.
auto.
Qed.

Lemma Describe_psi_union:
  forall D Psi1 Psi2,
       Describe_Psi D Psi1
    -> Describe_Psi D Psi2
    -> Describe_Psi D (union_Psi Psi1 Psi2).
Proof.
intros D Psi1 Psi2.
unfold Describe_Psi.
unfold lookupCT.
intros des1 des2 f rho a.
substH des1 with (des1 f rho a).
destruct des1 as [lkpCT1 dPsi1].
substH des2 with (des2 f rho a).
destruct des2 as [lkpCT2 dPsi2].
split_r.
apply OPSpec.lookup_merge_intro.
left; auto.
Qed.

Lemma Describe_enlarge:
  forall D1 D2 Psi1 Psi2,
    compatible_Dict D1 D2
    -> Describe_Psi D1 Psi1
    -> Describe_Psi D2 Psi2
    -> Describe_Psi (union_Dict D1 D2) (union_Psi Psi1 Psi2).
Proof.
intros D1 D2 Psi1 Psi2.
intros comp_dict des1 des2.
apply Describe_psi_union.
apply Describe_dict_union_left; auto.
apply Describe_dict_union_right; auto.
Qed.

(****************************************************************************)
(******* Code heap link lemma                                            ****)
(****************************************************************************)
Lemma WFcdhp_link: 
  forall D_1 D_2 Psi_1 Psi_1' Psi_2 Psi_2' C_1 C_2 C,
    WFcdhp D_1 Psi_1 C_1 Psi_1'
    -> WFcdhp D_2 Psi_2 C_2 Psi_2'
    -> compatible_Dict D_1 D_2
    -> CHMod.disjUnion C_1 C_2 C
    -> Describe_Psi D_1 Psi_1
    -> Describe_Psi D_2 Psi_2
    -> WFcdhp (union_Dict D_1 D_2) (union_Psi Psi_1 Psi_2) C (union_Psi Psi_1' Psi_2').
Proof.
intros D_1 D_2 Psi_1 Psi_1' Psi_2 Psi_2' C_1 C_2 C.
intros wfc1 wfc2 comp_dict union_ch des1 des2.
assert ( WFcdhp (union_Dict D_1 D_2) (union_Psi Psi_1 Psi_2) C_1 Psi_1').
apply WFcdhp_Enlarge with D_1 Psi_1; auto.
  apply subsetCT_Union_Left.
  apply subsetD_Union_Left; auto.
  apply Describe_enlarge; auto.
assert ( WFcdhp (union_Dict D_1 D_2) (union_Psi Psi_1 Psi_2) C_2 Psi_2').
apply WFcdhp_Enlarge with D_2 Psi_2; auto.
  apply subsetCT_Union_Right.
  apply subsetD_Union_Right; auto.
  apply Describe_enlarge; auto.
clear wfc1 wfc2.
rename H into wfc1.
rename H0 into wfc2.

unfold WFcdhp.
intros l theta lkpCT.
assert (lookupCT Psi_1' l theta \/ lookupCT Psi_2' l theta).
apply lkpCT_Union; auto.
inversion H.
  unfold WFcdhp in wfc1.
  assert (exists isq : iseq,
           lookupC C_1 l isq /\
           WFiseq (union_Dict D_1 D_2) (union_Psi Psi_1 Psi_2)
             theta l isq).
  apply wfc1; auto.
  InvertAll.
  exists x.
  split.
  apply lkpC_DisjUnion_Left with C_1 C_2.
  split; auto.
  auto.

  unfold WFcdhp in wfc2.
  assert (exists isq : iseq,
           lookupC C_2 l isq /\
           WFiseq (union_Dict D_1 D_2) (union_Psi Psi_1 Psi_2)
             theta l isq).
  apply wfc2; auto.
  InvertAll.
  exists x.
  split.
  apply lkpC_DisjUnion_Right with C_1 C_2.
  split; auto.
  auto.
Qed.
