Add LoadPath "..".

Require Import Coqlib.

Require Import Machine.
Require Import Assertion.
Require Import Operational_Semantics.
Require Import CAP0.


Lemma existsNextPC : forall pc i S, exists pc', nextPC (comml i) (S_Mem S) (S_RF S) (S_Flag S) pc pc'.
Proof.
  induction i;  intros. 

  exists (pc+1). constructor.
  exists (pc+1). constructor.
  exists n. constructor.
  assert (S_Flag S _ZF = true \/ S_Flag S _ZF = false).
  induction (S_Flag S _ZF); auto.
  inversion_clear H.
  exists (pc+1). constructor. auto.
  exists n. constructor. auto.
  exists (pc+1). constructor.
  exists (pc+1). constructor.
  exists (pc+1). constructor.
  exists (pc+1). constructor.
  exists (pc+1). constructor.
  exists (pc+1). constructor.
  exists (pc+1). constructor.
  exists (pc+1). constructor.
  exists (pc+1). constructor.
  exists (pc+1). constructor.
  exists (pc+1). constructor.
  exists (pc+1). constructor.
Qed.
Hint Resolve existsNextPC.

Lemma existsNS_existsNPC:
forall pc s S,
  (exists S', nextS pc (comms s) S S') ->
      exists pc', nextPC (comms s) (S_Mem S) (S_RF S) (S_Flag S) pc pc'.
Proof.
  intros.
  induction s.
  exists n. constructor.
  inversion H.
  rename x into S'.
  inversion H0.
  inversion H1. rename x into pc'.
  exists pc'.
  apply nextPC_retw with (S_RF S SP); auto.
  rewrite H4. auto.
  unfold popW in H3. inversion_clear H3.
  rewrite <- H4. unfold S_RF. auto.
  
  inversion H.
  rename x into S'. inversion H0.
  assert (exists pc', (sp |~>w pc') M).
  apply leadsToInterK; auto.  
  inversion H7. exists x.
  apply nextPC_iret with sp; auto.
Qed.
Hint Resolve existsNS_existsNPC.


Lemma CAP0_Progress:
  forall D Psi W,
    WFworld D Psi W -> exists W', stepInstr W W'.
Proof.
intros.
(** unfold WFworld **)
unfold WFworld in H.
induction W as ((ch,st),pc).
InvertAll.
rename x into Psi'.
rename x1 into isq.
rename x0 into theta.
rename H  into itrD.
rename H0 into lkpCT.
rename H1 into subCT.
rename H2 into WFC.
rename H3 into monoCT.
rename H5 into aS.
rename H6 into lkpC.
rename H7 into WFI.

induction st as (((h,R),fl),isr).
unfold stepInstr.

(** inductive over instructions **)
induction isq.
clear IHisq.
inversion_clear lkpC.

inversion_clear WFI.
rename H1 into WFc.
unfold WFcmd in WFc.
InvertAll.
rename H1 into WFc.
unfold WFcmd_aux in WFc.
rename H3 into mono_a.
unfold Mono_CT_assert in mono_a.

rename theta' into theta0.
assert 
((exists S' : State, nextS pc (comml i) (h,R,fl,isr) S') /\
        (forall (f' : PC) (S' : State),
         nextS pc (comml i) (h,R,fl,isr) S' ->
         nextPC (comml i) (S_Mem (h,R,fl,isr)) (S_RF (h,R,fl,isr)) (S_Flag (h,R,fl,isr)) pc f' ->
         exists theta' : Spec0,
           lookupCT (consCT (pc+1) theta0 Psi') f' theta' /\ Itp0 D theta' (consCT (pc+1) theta0 Psi') S') /\
        (~ itr_enabled (h,R,fl,isr) \/
         (exists S' : State,
            itr_nextS pc (h,R,fl,isr) S' /\
            (forall S'0 : State,
             itr_nextS pc (h,R,fl,isr) S'0 ->
             exists theta' : Spec0,
               lookupCT (consCT (pc+1) theta0 Psi') h_entry theta' /\
               Itp0 D theta' (consCT pc theta (consCT (pc+1) theta0 Psi')) S'0)))).
apply WFc; auto.
apply mono_a with Psi'; auto.
InvertAll.
rename H1 into Simp.
rename H5 into itrnS.
rename H4 into nS.
clear Simp itrnS WFc.

assert (exists pc', nextPC (comml i) (S_Mem (h, R, fl, isr)) (S_RF (h, R, fl, isr))
          (S_Flag (h, R, fl, isr)) pc pc').
auto.
inversion H1.
rename x0 into pc'.
exists (ch, x, pc').
split; auto.
exists (comml i).
split; auto.

(** specicial instructions **)
inversion_clear lkpC.
inversion_clear WFI.
rename H0 into WFc.
unfold WFcmd in WFc.
InvertAll.
rename H0 into WFc.
unfold WFcmd_aux in WFc.

assert ((exists S' : State, nextS pc (comms s) (h,R,fl,isr) S') /\
        (forall (f' : PC) (S' : State),
         nextS pc (comms s) (h,R,fl,isr) S' ->
         nextPC (comms s) (S_Mem (h,R,fl,isr)) (S_RF (h,R,fl,isr)) (S_Flag (h,R,fl,isr)) pc f' ->
         exists theta' : Spec0,
           lookupCT Psi' f' theta' /\ Itp0 D theta' Psi' S') /\
        (~ itr_enabled (h,R,fl,isr) \/
         (exists S' : State,
            itr_nextS pc (h,R,fl,isr) S' /\
            (forall S'0 : State,
             itr_nextS pc (h,R,fl,isr) S'0 ->
             exists theta' : Spec0,
               lookupCT Psi' h_entry theta' /\
               Itp0 D theta' (consCT pc theta Psi') S'0)))).
apply WFc; auto.

InvertAll.
rename H0 into Simp.
rename H1 into mono_a.
rename H4 into itrnS.
rename H3 into nS.

assert (exists pc', nextPC (comms s) (S_Mem (h, R, fl, isr)) (S_RF (h, R, fl, isr))
           (S_Flag (h, R, fl, isr)) pc pc').
apply existsNS_existsNPC; auto. exists x. auto.
inversion H0. exists (ch, x, x0).
split; auto.
exists (comms s). split; auto.
Qed.


Lemma CAP0_Preservation:
  forall D Psi W W',
    WFworld D Psi W -> step W W' -> WFworld D Psi W'.
Proof.
intros D Psi W W' WFW stepW.
unfold WFworld in WFW.
induction W as ((ch,st),pc).
induction W' as ((ch',st'), pc').
InvertAll.
rename x into Psi'.
rename x1 into isq.
rename x0 into theta.

rename H into lkpCTh. (** lookup h_entry into CT **)
rename H0 into subCT.
rename H1 into itrD.   (** interrupt ID in Dict **)
rename H2 into WFC.
rename H3 into monoCT.
rename H5 into aS.
rename H6 into lkpCpc. (** lookup pc into C **)
rename H7 into WFIseq.

(** INDUCTION INSTR/INTERRUPT **)
inversion_clear stepW.
(** INSTRUCTION *)
rename H into step1.
unfold stepInstr in step1.
InvertAll.
rename x into c.
rename H into equCh. rewrite <- equCh. clear equCh.
rename H0 into pc_inC.
rename H1 into nPC.
rename H3 into nS.

(** INDUCTION OVER ISEQ **)
induction WFIseq. clear IHWFIseq.
rename H into WFc.
unfold WFcmd in WFc.
InvertAll.
rename H0 into mono_a.
unfold Mono_CT_assert in mono_a.
rename H into WFc.

inversion_clear lkpCpc. (** lookupC **)
assert (comml i = c).
apply l_inC_unique with ch f; auto.
rewrite <- H1 in nPC.
rewrite <- H1 in nS.

clear pc_inC H. (** l_inC **)
clear H1 c.     (** c=comml i/c **)


(** WFPROG **)
unfold WFworld.
(** WFC/Mono_CT/subCT/interrupt entry/itrD **)
split; auto.
split; auto.

unfold WFcmd_aux in WFc.
(** WFcmd: aS -> a'S' **)
rename st into S0.
rename theta' into theta0.
assert ((exists S' : State, nextS f (comml i) S0 S') /\
        (forall (f' : PC) (S' : State),
         nextS f (comml i) S0 S' ->
         nextPC (comml i) (S_Mem S0) (S_RF S0) (S_Flag S0) f f' ->
         exists theta' : Spec0,
           lookupCT (consCT (f+1) theta0 Psi') f' theta' /\ Itp0 D theta' (consCT (f+1) theta0 Psi') S') /\
        (~ itr_enabled S0 \/
         (exists S' : State,
            itr_nextS f S0 S' /\
            (forall S'0 : State,
             itr_nextS f S0 S'0 ->
             exists theta' : Spec0,
               lookupCT (consCT (f+1) theta0 Psi') h_entry theta' /\
               Itp0 D theta' (consCT f theta (consCT (f+1) theta0 Psi')) S'0)))).
apply WFc; auto.
apply mono_a with Psi'; auto.
rename theta0 into theta'.
rename S0 into st.
InvertAll.
rename H3 into aS_itr.
clear WFc aS_itr.
rename H2 into nS_WFc.
clear nS_WFc. clear x.
rename H into Simp.

assert (exists theta0 : Spec0,
           lookupCT (consCT (f + 1) theta' Psi') pc' theta0 /\
           Itp0 D theta0 (consCT (f + 1) theta' Psi') st').
apply Simp; auto.

exists (consCT (f+1) theta' Psi').
 split.
 apply WFcdhp_psi_extend with is; auto.
 split; auto.
 split.
 apply MonoCT_psi_extend; auto.
 apply WFiseq_mono_a with Psi (f+1) is; auto.

InvertAll.
rename x into theta''.
rename H  into lkpCT.
rename H2 into aS'.
exists theta''.
split; auto.

unfold WFcdhp in WFC.
(** INDUCTION OVER SIMPLE INSTRUCTION *)
induction i.
(** _cli ***********************************************************)
(*******************************************************************)
  inversion nPC. (** nextPC **)
  rename H4 into getPC'.
  rewrite <- getPC' in lkpCT.
  unfold lookupCT in lkpCT.
  inversion lkpCT.
  inversion H.
  subst.
  exists is.
   split; auto.

  apply WFC; subst; auto.
(** _sti ***********************************************************)
(*******************************************************************)
 inversion nPC. (** nextPC **)
  rename H4 into getPC'.
  rewrite <- getPC' in lkpCT.
  inversion lkpCT.
  inversion H.
  rename H5 into eqTheta.
  subst.
  exists is.
   split; auto.

  apply WFC; subst; auto.
(** _call n ********************************************************)
(*******************************************************************)
  inversion nPC. (** nextPC **)
  rename H5 into getPC'.
  subst.
  inversion lkpCT.
  inversion H.
  subst.
  exists is.
   split; auto.

  apply WFC; auto.

(** _jne n ********************************************************)
(*******************************************************************)
  inversion nPC. (** nextPC **)
  (** DO not jump **)
  subst.
  inversion lkpCT.
  inversion H.
  subst.
  exists is.
   split; auto.

  apply WFC; auto.
  (** DO jump **)
  subst.
  inversion lkpCT.
  inversion H.
  subst.
  exists is.
   split; auto.

  apply WFC; auto.

(** _movwp o r *****************************************************)
(*******************************************************************)
  inversion nPC. (** nextPC **)
  subst.
  inversion lkpCT.
  inversion H.
  subst.
  exists is.
   split; auto.

  apply WFC; auto.

(** _movwld a r ****************************************************)
(*******************************************************************)
  inversion nPC. (** nextPC **)
  subst.
  inversion lkpCT.
  inversion H.
  subst.
  exists is.
   split; auto.

  apply WFC; auto.

(** _movwst o a ****************************************************)
(*******************************************************************)
  inversion nPC. (** nextPC **)
  subst.
  inversion lkpCT.
  inversion H.
  subst.
  exists is.
   split; auto.

  apply WFC; auto.

(** _pushwp o ******************************************************)
(*******************************************************************)
  inversion nPC. (** nextPC **)
  subst.
  inversion lkpCT.
  inversion H.
  subst.
  exists is.
   split; auto.

  apply WFC; auto.

(** _pushwm a ******************************************************)
(*******************************************************************)
  inversion nPC. (** nextPC **)
  subst.
  inversion lkpCT.
  inversion H.
  subst.
  exists is.
   split; auto.

  apply WFC; auto.

(** _pushfw ********************************************************)
(*******************************************************************)
  inversion nPC. (** nextPC **)
  subst.
  inversion lkpCT.
  inversion H.
  subst.
  exists is.
   split; auto.

  apply WFC; auto.

(** _popw r ********************************************************)
(*******************************************************************)
  inversion nPC. (** nextPC **)
  subst.
  inversion lkpCT.
  inversion H.
  subst.
  exists is.
   split; auto.

  apply WFC; auto.

(** _popfw  ********************************************************)
(*******************************************************************)
  inversion nPC. (** nextPC **)
  subst.
  inversion lkpCT.
  inversion H.
  subst.
  exists is.
   split; auto.

  apply WFC; auto.

(** _cmpw o r ******************************************************)
(*******************************************************************)
  inversion nPC. (** nextPC **)
  subst.
  inversion lkpCT.
  inversion H.
  subst.
  exists is.
   split; auto.

  apply WFC; auto.

(** _addw o r ******************************************************)
(*******************************************************************)
  inversion nPC. (** nextPC **)
  subst.
  inversion lkpCT.
  inversion H.
  subst.
  exists is.
   split; auto.

  apply WFC; auto.

(** _subw o r ******************************************************)
(*******************************************************************)
  inversion nPC. (** nextPC **)
  subst.
  inversion lkpCT.
  inversion H.
  subst.
  exists is.
   split; auto.

  apply WFC; auto.

(** _outb_eoi ******************************************************)
(*******************************************************************)
  inversion nPC. (** nextPC **)
  subst.
  inversion lkpCT.
  inversion H.
  subst.
  exists is.
   split; auto.

  apply WFC; auto.

(*******************************************************************)
(** Sepcicial Instructions: jmp, retw, iret ************************)
rename H into WFc.
unfold WFcmd in WFc.
InvertAll.
rename H0 into mono_a.
rename H into WFc.

inversion_clear lkpCpc. (** lookupC **)
assert (comms si = c).
apply l_inC_unique with ch f; auto.
rewrite <- H0 in nPC.
rewrite <- H0 in nS.

clear pc_inC H. (** l_inC **)
clear H0 c.     (** c=comml i/c **)

(** WFPROG **)
unfold WFworld.
(** interrupt entry/itrD **)
split; auto.
split; auto.
exists Psi'.
 split; auto.
 split; auto.
 split; auto.

unfold WFcmd_aux in WFc.
(** WFcmd: aS -> a'S' **)
rename st into S.
assert ((exists S' : State, nextS f (comms si) S S') /\
        (forall (f' : PC) (S' : State),
         nextS f (comms si) S S' ->
         nextPC (comms si) (S_Mem S) (S_RF S) (S_Flag S) f f' ->
         exists theta' : Spec0,
           lookupCT Psi' f' theta' /\ Itp0 D theta' Psi' S') /\
        (~ itr_enabled S \/
         (exists S' : State,
            itr_nextS f S S' /\
            (forall S'0 : State,
             itr_nextS f S S'0 ->
             exists theta' : Spec0,
               lookupCT Psi' h_entry theta' /\
               Itp0 D theta' (consCT f theta Psi') S'0)))).
apply WFc; auto.
rename S into st.
InvertAll.
rename H2 into aS_itr.
clear WFc aS_itr.
rename H1 into nS_WFc.
clear nS_WFc. clear x.
rename H into Simp.

assert (exists theta : Spec0,
           lookupCT Psi' pc' theta /\
           Itp0 D theta Psi' st').
apply Simp; auto.

InvertAll.
rename x into theta'.
rename H  into lkpCT.
rename H1 into aS'.
exists theta'.
split. auto.
unfold WFcdhp in WFC.
apply WFC.
auto.

(*******************************************************************)
(** Interrupt ******************************************************)
rename H into stepI.
unfold stepItr in stepI.
InvertAll.
rename H into equCh. rewrite <- equCh. clear equCh.
rename H3 into h_inC.
rename H1 into itrEN.
rename H0 into itrnS.
rewrite h_inC.

unfold WFcdhp in WFC.

(** WFPROG **)
unfold WFworld.
(** interrupt entry/itrD/WFcdhp/subsetCT/Mono_CT_PSpec **)
split; auto.
split; auto.

(** WFiseq ==> a'S' **)
assert (WFiseq D Psi theta pc isq). auto.
rename H into WFIseq2.
induction WFIseq. clear IHWFIseq.
rename H into WFc.
unfold WFcmd in WFc.
InvertAll.
rename H0 into mono_a.
unfold Mono_CT_assert in mono_a.
rename H into WFc.
unfold WFcmd_aux in WFc.

exists (consCT f theta (consCT (f+1) theta' Psi')).
 split.
 assert (WFcdhp D Psi ch (consCT (f + 1) theta' Psi')).
 inversion lkpCpc.
 apply WFcdhp_psi_extend with is; auto.
 apply WFcdhp_psi_extend with (seq i is); auto.

 split; auto.
 split.
 assert (Mono_CT_PSpec D (consCT (f + 1) theta' Psi')).
 apply MonoCT_psi_extend; auto.
 apply WFiseq_mono_a with Psi (f+1) is; auto.
 apply MonoCT_psi_extend; auto.

rename st into S0.
rename theta' into theta0.
assert ((exists S' : State, nextS f (comml i) S0 S') /\
        (forall (f' : PC) (S' : State),
         nextS f (comml i) S0 S' ->
         nextPC (comml i) (S_Mem S0) (S_RF S0) (S_Flag S0) f f' ->
         exists theta' : Spec0,
           lookupCT (consCT (f+1) theta0 Psi') f' theta' /\ Itp0 D theta' (consCT (f+1) theta0 Psi') S') /\
        (~ itr_enabled S0 \/
         (exists S' : State,
            itr_nextS f S0 S' /\
            (forall S'0 : State,
             itr_nextS f S0 S'0 ->
             exists theta' : Spec0,
               lookupCT (consCT (f+1) theta0 Psi') h_entry theta' /\
               Itp0 D theta' (consCT f theta (consCT (f+1) theta0 Psi')) S'0)))).
apply WFc; auto.
apply mono_a with Psi'; auto.
rename theta0 into theta'.
rename S0 into st.
InvertAll.
rename H into Simp.
rename H2 into aS_itr.
clear WFc Simp.
rename H1 into nS_WFc.
clear nS_WFc. clear x.


inversion_clear aS_itr.
absurd (itr_enabled st); auto.
InvertAll.
rename H1 into a'S'.
rename H into itrnS'.
rename x into st0.
clear itrnS'. clear st0.

assert (exists theta0 : Spec0,
           lookupCT (consCT (f + 1) theta' Psi') h_entry theta0 /\
           Itp0 D theta0 (consCT f theta (consCT (f + 1) theta' Psi')) st').
apply a'S'; auto.
InvertAll.
rename x into itr_theta'.
exists itr_theta'.
split. auto.

rename H into lkpCTitr.
inversion lkpCTitr.
inversion H.
rename H3 into eqTheta.
rewrite <- eqTheta.
inversion_clear lkpCpc.
exists is. split; auto.

apply WFC.
auto.

(** specicial instructions for Interrupt **)
rename H into WFc.
unfold WFcmd in WFc.
InvertAll.
rename H0 into mono_a.
rename H into WFc.

unfold WFcmd_aux in WFc.
(** WFcmd: aS -> a'S' **)
rename st into S.
assert ((exists S' : State, nextS f (comms si) S S') /\
        (forall (f' : PC) (S' : State),
         nextS f (comms si) S S' ->
         nextPC (comms si) (S_Mem S) (S_RF S) (S_Flag S) f f' ->
         exists theta' : Spec0,
           lookupCT Psi' f' theta' /\ Itp0 D theta' Psi' S') /\
        (~ itr_enabled S \/
         (exists S' : State,
            itr_nextS f S S' /\
            (forall S'0 : State,
             itr_nextS f S S'0 ->
             exists theta' : Spec0,
               lookupCT Psi' h_entry theta' /\
               Itp0 D theta' (consCT f theta Psi') S'0)))).
apply WFc; auto.
rename S into st.
InvertAll.
rename H into Simp.
rename H2 into aS_itr.
clear WFc Simp.
rename H1 into nS_WFc.
clear nS_WFc. clear x.

inversion_clear aS_itr.
absurd (itr_enabled st); auto.
InvertAll.
rename H1 into a'S'.
rename H into itrnS'.
clear itrnS'. clear x.

assert (exists itr_theta' : Spec0,
           lookupCT Psi' h_entry itr_theta' /\ Itp0 D itr_theta' (consCT f theta Psi') st').
apply a'S'; auto.
InvertAll.
rename x into itr_theta'.

exists (consCT f theta Psi').
split.
 apply WFcdhp_psi_extend with (sseq si); auto.
 split; auto.
 split.
 apply MonoCT_psi_extend; auto.
exists itr_theta'.
split. auto.

apply WFC.
auto.
Qed.


(** Soundness of CAP0: safe run N steps***************************)
Lemma CAP0_Soundness_n:
  forall (n:nat) (D:LDict) (Psi:PSpec) (W:World),
    WFworld D Psi W -> (exists W', stepN n W W').
Proof.
induction n.
 (** stepN 0 W W **)
 intros.
  exists W.
  constructor; auto.

 (** stepN n W W' **)
 intros.
 (** step once: W |--> W1, CASE 1, instructions**)
 assert (exists W1, stepInstr W W1).
  apply CAP0_Progress with D Psi; auto.
 InvertAll. 
 rename x into W1.

 (** use IHn : W1 |==>n W' **)
 assert (exists W' : World, stepN n W1 W').
  apply IHn with D Psi.
   apply CAP0_Preservation with W; auto.
    constructor; auto.
 InvertAll.
 exists x.
 constructor.
 exists W1.
 split.
  constructor; auto.
  auto.
Qed.

(** Soundness of CAP0: after N steps, Interrupt OK **********)
Lemma CAP0_Soundness_nI:
  forall (n:nat) (D:LDict) (Psi:PSpec) (W:World),
    WFworld D Psi W
    -> (forall W', stepN n W W' -> exists W'', stepInstr W' W'').
Proof.
induction n.
 (** stepN 0 W W **)
 intros D Psi W WFW W' stepN.
 inversion stepN.
 rename H into eqWW'.
 rewrite eqWW'.
  apply CAP0_Progress with D Psi; auto.

 (** stepN n W W' **)
 intros D Psi W WFW W' stepN.
 inversion stepN.
 (** step once: W |--> W1, CASE 1, instructions**)
 InvertAll. 
 rename x into W1.

 (** use IHn : W1 |==>n W' **)
 apply IHn with D Psi W1.
 apply CAP0_Preservation with W; auto.
 auto.
Qed.

(** Soundness of CAP0 **************************************)
Lemma CAP0_Soundness:
  forall (n:nat) (D:LDict) (Psi:PSpec) (W:World),
    WFworld D Psi W
    -> (exists W', stepN n W W')
    /\ (forall W', stepN n W W' -> exists W'', stepInstr W' W'').
Proof.
split.
 apply CAP0_Soundness_n  with D Psi; auto.
 apply CAP0_Soundness_nI with D Psi; auto.
Qed.
