Require Import Coqlib.
Require Import Map.
Require Import Arith.
Require Import Machine.
Require Import Assertion.
Require Import Omega.

Import MemMod.

Set Implicit Arguments.
Unset Strict Implicit.

(* merge *)

Lemma merge_sym : forall M1 M2,
  disjoint M1 M2 -> merge M1 M2 = merge M2 M1.
Proof.
  unfold disjoint;unfold merge;unfold not_in_dom.
  intros;apply extensional_eqMap.
  intro;assert (M1 x = None \/ M2 x = None).
  auto.
  elim H0;intro;rewrite H1.
  case (M2 x);[trivial | trivial].
  case (M1 x);[trivial | trivial].
Qed.

Lemma merge_assoc : forall M1 M2 M3,
  merge (merge M1 M2) M3 = merge M1 (merge M2 M3).
Proof.
intros;unfold merge.
apply extensional_eqMap.
intro;case (M1 x);trivial.
Qed.

Lemma disjUnion_intro : forall M1 M2 M,
  (forall addr, match M1 addr, M2 addr, M addr with
                    | Some x1, None, Some x2 => x1 = x2
                    | None, Some x1, Some x2 => x1 = x2
                    | None, None, None => True
                    | _, _, _ => False
               end)
  ->   MemMod.disjUnion M1 M2 M.
Proof.
  intros M1 M2 M H.
  unfold disjUnion.
  unfold merge, disjoint.
  split.
    intros x.
    substH H with (H x).
    unfold not_in_dom.
    destruct (M1 x); destruct (M2 x); destruct (M x); tauto.
  apply extensional_eqMap.
  intros x.
  substH H with (H x).
  destruct (M1 x); destruct (M2 x); destruct (M x); auto; try contradiction.
Qed.

Lemma disjUnion_elim : forall M1 M2 M,
  MemMod.disjUnion M1 M2 M
  -> forall addr, match M1 addr, M2 addr, M addr with
                    | Some x1, None, Some x2 => x1 = x2
                    | None, Some x1, Some x2 => x1 = x2
                    | None, None, None => True
                    | _, _, _ => False
                  end.
Proof.
  intros M1 M2 M H.
  unfold disjUnion in H.
  unfold merge, disjoint in H.
  destruct H.
  subst M.
  unfold not_in_dom in H.
  intros l.
  substH H with (H l).
  destruct H.
    rewrite H.
    destruct (M2 l); trivial.
  rewrite H.
  destruct (M1 l); trivial.
Qed.

Lemma disjUnion_sym : forall M1 M2 M3, 
  disjUnion M1 M2 M3
  -> disjUnion M2 M1 M3.
Proof.
unfold disjUnion;unfold disjoint.
intros.
destruct H as [H H1].
split.
intro.
assert (not_in_dom x M1 \/ not_in_dom x M2).
apply H.
tauto.
assert (merge M1 M2 = merge M2 M1).
apply merge_sym.
unfold disjoint;apply H.
rewrite H1;apply H0.
Qed.

Lemma disjUnion_update_left : forall l v1 v2 M1 M2 M,
  M1 l = Some v1
  -> disjUnion M1 M2 M
  -> disjUnion (put M1 l v2) M2 (put M l v2).
Proof.
  intros l v1 v2 M1 M2 M H0 H1.
  substH H1 with (disjUnion_elim H1).
  apply disjUnion_intro.
  intros addr.
  substH H1 with (H1 addr).
  unfold put.
  destruct (beq_A addr l).
    subst l.
    destruct (M1 addr); destruct (M2 addr);
      destruct (M addr); trivial; try discriminate.
  destruct (M1 addr); destruct (M2 addr);
    destruct (M addr); trivial; try discriminate.
Qed.

Lemma disjUnion_assoc_left_right : forall M1 M2 M3 M4,
  disjUnion (merge M1 M2) M3 M4
  -> disjoint M1 M2
  -> disjUnion M1 (merge M2 M3) M4.
Proof.
intros;apply disjUnion_assoc_left_right;trivial.
Qed.

Lemma disjUnion_assoc_right_left : forall M1 M2 M3 M4,
  disjUnion M1 (merge M2 M3) M4
  -> disjoint M2 M3
  -> disjUnion (merge M1 M2) M3 M4.
Proof.
intros.
apply disjUnion_assoc_right_left;trivial.
Qed.

Lemma disjoint_merge_intro_left : forall M1 M2 M3,
  disjoint M1 M3
  -> disjoint M2 M3
  -> disjoint (merge M1 M2) M3.
Proof.
unfold disjoint;unfold merge;unfold not_in_dom.
intros.
assert (M1 x = None \/ M3 x = None).
trivial.
elim H1;intro;rewrite H2;[trivial | right;trivial].
Qed.

Lemma disjoint_merge_intro_right : forall M1 M2 M3,
  disjoint M1 M2
  -> disjoint M1 M3
  -> disjoint M1 (merge M2 M3).
Proof.
unfold disjoint;unfold merge;unfold not_in_dom.
intros.
assert (M1 x = None \/ M2 x = None).
trivial.
elim H1.
intro;left;trivial.
intro H2.
assert (M1 x = None \/ M3 x = None).
trivial.
elim H3.
intro;left;trivial.
intro;right;rewrite H2;trivial.
Qed.

Lemma disjUnion_disjoint_left : forall M1 M2 M3 Mx My,
  disjUnion M1 Mx M3
  -> disjUnion M3 M2 My
  -> disjoint M1 M2.
Proof.
unfold disjUnion;unfold disjoint;unfold not_in_dom.
intros;destruct H as [H H1];destruct H0 as [H0 H2].
rewrite H1 in H0.
assert (merge M1 Mx x = None \/ M2 x = None).
trivial.
elim H3.
intros.
left;unfold merge in H4.
assert ((exists b , M1 x  = Some b)\/ M1 x = None).
case (M1 x);[intro;left;exists b;trivial  |  right;trivial].
elim H5.
intro.
destruct H6;rewrite H6 in H4;rewrite H4 in H6.
assumption.
trivial.
intro;right;trivial.
Qed.

Lemma disjUnion_left_disjoint_left : forall M1 M2 M3 Mx,
  disjUnion M1 Mx M3
  -> disjoint M3 M2
  -> disjoint M1 M2.
Proof.
unfold disjUnion;unfold disjoint;unfold not_in_dom.
intros.
destruct H.
assert (M3 x = None \/ M2 x = None).
trivial.
elim H2.
rewrite H1;unfold merge.
case (M1 x).
intros;left;assumption.
intro;left;trivial.
intro;right;assumption.
Qed.


Lemma disjUnion_right_disjoint_left : forall M1 M2 M3 Mx,
  disjUnion Mx M1 M3
  -> disjoint M3 M2
  -> disjoint M1 M2.
Proof.
unfold disjUnion;unfold disjoint;unfold not_in_dom.
intros.
destruct H.
assert (M3 x = None \/ M2 x = None).
trivial.
elim H2.
rewrite H1;unfold merge.
assert (Mx x = None \/ M1 x = None).
trivial.
elim H3.
intro.
rewrite H4;intro;left;assumption.
intros;left;assumption.
intros;right;assumption.
Qed.


Lemma disjUnion_right_disjoint_right : forall M1 M2 M3 Mx,
  disjUnion Mx M2 M3
  -> disjoint M1 M3
  -> disjoint M1 M2.
Proof.
unfold disjUnion;unfold disjoint;unfold not_in_dom.
intros.
destruct H.
assert (M1 x = None \/ M3 x = None).
trivial.
elim H2.
intro;left;assumption.
assert (Mx x = None \/ M2 x = None).
trivial.
rewrite H1;unfold merge.
elim H3.
intro H4;rewrite H4;right;assumption.
intros;right;assumption.
Qed.


Lemma disjUnion_left_disjoint_right : forall M1 M2 M3 Mx,
  disjUnion M2 Mx M3
  -> disjoint M1 M3
  -> disjoint M1 M2.
Proof.
unfold disjUnion;unfold disjoint;unfold not_in_dom.
intros.
assert (M1 x = None \/ M3 x = None);trivial.
elim H1.
intro;left;assumption.
destruct H.
rewrite H2;unfold merge.
assert (M2 x = None \/ Mx x = None);trivial.
elim H3.
intros;right;assumption.
intro;case (M2 x);intros;right;trivial.
Qed.


Lemma disjUnion_disjoint : forall M1 M2 M3,
  disjUnion M1 M2 M3
  -> disjoint M1 M2.
Proof.
unfold disjUnion;intros;destruct H;trivial.
Qed.

Lemma disjoint_sym : forall M1 M2,
  disjoint M1 M2 -> disjoint M2 M1.
Proof.
unfold disjoint.
intros.
assert (not_in_dom x M1 \/ not_in_dom x M2);trivial.
elim H0.
intro;right;assumption.
intro;left;assumption.
Qed.

Lemma eqDom_disjoint_left : forall M1 M2 M1',
  disjoint M1 M2 -> eqDom M1 M1' -> disjoint M1' M2.
Proof.
unfold disjoint;unfold eqDom;unfold sub_dom;unfold not_in_dom.
intros.
assert (M1 x = None \/ M2 x = None);trivial.
destruct H0.
elim H1.
intro.
left.
assert (in_dom x M1 -> in_dom x M1');auto.
assert (in_dom x M1' -> in_dom x M1);auto.
unfold in_dom in H4;unfold in_dom in H5.
induction (M1' x).
assert (exists b : B,M1 x = Some b).
apply H5.
exists a.
auto.
inversion H6.
rewrite H3 in H7.
inversion H7.
trivial.
intro;right;assumption.
Qed.

Lemma eqDom_disjoint_right : forall M1 M2 M2',
  disjoint M1 M2 -> eqDom M2 M2' -> disjoint M1 M2'.
Proof.
unfold disjoint;unfold eqDom;unfold sub_dom;unfold not_in_dom.
intros.
assert (M1 x = None \/ M2 x = None);trivial.
destruct H0.
elim H1.
intro.
left.
auto.
assert (in_dom x M2 -> in_dom x M2');auto.
assert (in_dom x M2' -> in_dom x M2);auto.
unfold in_dom in H3;unfold in_dom in H4.
intro.
right.
induction (M2' x).
assert (exists b : B,M2 x = Some b).
apply H4.
exists a;auto.
inversion H6.
rewrite H5 in H7.
inversion H7.
trivial.
Qed.


Lemma disjUnion_merge : forall M1 M2 M3,
  MemMod.disjUnion M1 M2 M3 -> MemMod.merge M1 M2 = M3.
Proof.
unfold disjUnion;intros M1 M2 M3 H;destruct H as [disjH EQ];rewrite EQ;trivial.
Qed.

Ltac disjoint_trivial :=
      match goal with 
        | H : MemMod.disjoint ?M1 ?M2 
          |-  MemMod.disjoint ?M1 ?M2 
            => exact H
        | H : MemMod.disjoint ?M2 ?M1 
          |-  MemMod.disjoint ?M1 ?M2 
          => apply disjoint_sym; trivial
        | H : MemMod.disjUnion ?M1 ?M2 ?M3
          |-  MemMod.disjoint ?M1 ?M2 
          => apply (@disjUnion_disjoint M1 M2 M3 H); trivial
        | H : MemMod.disjUnion ?M2 ?M1 ?M3
          |-  MemMod.disjoint ?M1 ?M2
          => apply disjoint_sym;
            apply (@disjUnion_disjoint M2 M1 M3 H); trivial
        | _ => trivial
      end.

Tactic Notation "disjoint" "trivial" := disjoint_trivial.

Ltac disjoint_des := 
  match goal with 
    | |- MemMod.disjoint (MemMod.merge ?M1 ?M2) ?M3 => 
        apply disjoint_merge_intro_left; disjoint_des
    | |- MemMod.disjoint ?M1 (MemMod.merge ?M2 ?M3) => 
        apply disjoint_merge_intro_right; disjoint_des
    | |- MemMod.disjoint ?M1 ?M2 => disjoint_trivial
    | _ => idtac
  end.

Tactic Notation "disjoint" "des" := disjoint_des.

Ltac disjoint_lift_l := 
  match goal with
    | H: MemMod.disjUnion ?M1 ?Mx ?M3
      |- MemMod.disjoint ?M1 ?M2 => 
        apply (@disjUnion_left_disjoint_left M1 M2 M3 Mx H); disjoint_trivial
    | H: MemMod.disjUnion ?Mx ?M1 ?M3
      |- MemMod.disjoint ?M1 ?M2 => 
        apply (@disjUnion_right_disjoint_left M1 M2 M3 Mx H); disjoint_trivial
  end.

Tactic Notation "disjoint" "lift_l" := disjoint_lift_l.

Ltac disjoint_lift_l_1 Mx :=
  match goal with
    | H: MemMod.disjUnion ?M1 Mx ?M3
      |- MemMod.disjoint ?M1 ?M2 => 
        apply (@disjUnion_left_disjoint_left M1 M2 M3 Mx H); disjoint_trivial
    | H: MemMod.disjUnion Mx ?M1 ?M3
      |- MemMod.disjoint ?M1 ?M2 => 
        apply (@disjUnion_right_disjoint_left M1 M2 M3 Mx H); disjoint_trivial
    | |- MemMod.disjoint ?M1 ?M2 => 
        apply (@disjUnion_right_disjoint_left M1 M2 (MemMod.merge M1 Mx) Mx); disjoint_trivial
  end.

Tactic Notation  "disjoint" "lift_l" constr (Mx) := disjoint_lift_l_1 Mx.

Ltac disjoint_lift_r :=
  match goal with
    | H: MemMod.disjUnion ?M2 ?Mx ?M3
      |- MemMod.disjoint ?M1 ?M2 => 
        apply (@disjUnion_left_disjoint_right M1 M2 M3 Mx H); disjoint_trivial
    | H: MemMod.disjUnion ?Mx ?M2 ?M3
      |- MemMod.disjoint ?M1 ?M2 => 
        apply (@disjUnion_right_disjoint_right M1 M2 M3 Mx H); disjoint_trivial
  end.

Tactic Notation "disjoint" "lift_r" := disjoint_lift_r.

Ltac disjoint_lift_r_1 Mx :=
  match goal with
    | H: MemMod.disjUnion ?M2 Mx ?M3
      |- MemMod.disjoint ?M1 ?M2 => 
        apply (@disjUnion_left_disjoint_right M1 M2 M3 Mx H); disjoint_trivial
    | H: MemMod.disjUnion Mx ?M2 ?M3
      |- MemMod.disjoint ?M1 ?M2 => 
        apply (@disjUnion_right_disjoint_right M1 M2 M3 Mx H); disjoint_trivial
    | |- MemMod.disjoint ?M1 ?M2 => 
        apply (@disjUnion_left_disjoint_right M1 M2 (MemMod.merge M2 Mx) Mx); disjoint_trivial
  end.

Tactic Notation  "disjoint" "lift_r" constr (Mx) := disjoint_lift_r_1 Mx.

Tactic Notation  "disjoint" "lift_lr" := disjoint_lift_l; disjoint_lift_r.

Tactic Notation  "disjoint" "lift_llr" := disjoint_lift_l; disjoint_lift_l; disjoint_lift_r.

Tactic Notation  "disjoint" "lift_lrr" := disjoint_lift_l; disjoint_lift_r; disjoint_lift_r.


Ltac disjoint_replace_l M1 :=
  match goal with 
    | |- MemMod.disjoint ?M1' ?M2 
      => apply (@eqDom_disjoint_left M1 M2 M1'); [disjoint_trivial | trivial]
  end.

Ltac disjoint_replace_r M2 :=
  match goal with 
    | |- MemMod.disjoint ?M1 ?M2' 
      => apply (@eqDom_disjoint_right M1 M2 M2'); [disjoint_trivial | trivial]
  end.

Tactic Notation "disjoint" "rep_l" constr (M1) := disjoint_replace_l M1.

Tactic Notation "disjoint" "rep_r" constr (M2) := disjoint_replace_r M2.

Ltac disjoint_solve_left :=
  match goal with 

    | H: MemMod.disjUnion ?M1 ?Mx ?M3,
      H1: MemMod.disjUnion ?M3 ?M2 ?My
      |- MemMod.disjoint ?M1 ?M2 => 
        apply (@disjUnion_disjoint_left M1 M2 M3 Mx My H H1); trivial

    | H: MemMod.disjUnion ?M1 ?Mx ?M3,
      H1: MemMod.disjUnion ?M2 ?M3 ?My
      |- MemMod.disjoint ?M1 ?M2 => 
        apply (@disjUnion_disjoint_left M1 M2 M3 Mx My H (@disjUnion_sym M2 M3 My H1)); 
          disjoint_trivial
        
    | H: MemMod.disjUnion ?Mx ?M1 ?M3,
      H1: MemMod.disjUnion ?M3 ?M2 ?My
      |- MemMod.disjoint ?M1 ?M2 => 
        apply (@disjUnion_disjoint_left M1 M2 M3 Mx My (@disjUnion_sym Mx M1 M3 H) H1); 
          disjoint_trivial
        
    | H: MemMod.disjUnion ?Mx ?M1 ?M3,
      H1: MemMod.disjUnion ?M2 ?M3 ?My
      |- MemMod.disjoint ?M1 ?M2 => 
        apply (@disjUnion_disjoint_left M1 M2 M3 Mx My (@disjUnion_sym Mx M1 M3 H) (@disjUnion_sym M2 M3 My H1)); 
          disjoint_trivial
          
    | _ => fail
  end.

Ltac disjoint_solve := (disjoint_trivial; disjoint_solve_left) || (apply disjoint_sym; (disjoint_solve_left)).

Tactic Notation "disjoint" "solve_left" := disjoint_solve_left.

Tactic Notation "disjoint" "solve" := disjoint_solve.

(******* merge ********)

Ltac merge_rew H := 
  match type of H with
    | MemMod.disjUnion ?M1 ?M2 ?Mx =>
      match goal with
        | |- context [MemMod.merge M1 M2] => rewrite (disjUnion_merge H)
      end
  end.
                         
Ltac merge_rewH H H0 := 
  match type of H with
    | MemMod.disjUnion ?M1 ?M2 ?Mx =>
      match goal with
        | H0 : context [MemMod.merge M1 M2] 
          |- _ => rewrite (disjUnion_merge H) in H0
      end
  end.

Tactic Notation "merge" "rewrite" hyp (Hx) := merge_rew Hx.

Tactic Notation "merge" "rewrite" hyp (Hx) "in" hyp (H) := merge_rewH Hx H.

Ltac merge_commut M1 :=
  match goal with
    | |- context [MemMod.merge M1 ?M2] 
      => rewrite (@merge_sym M1 M2); [idtac | try disjoint_solve]
    | |- context [MemMod.merge ?M2 M1] 
      => rewrite (@merge_sym M2 M1); [idtac | try disjoint_solve]
    | _ => fail
  end.

Tactic Notation "merge" "commut" constr (M1) := merge_commut M1.

Ltac merge_assoc M1 :=
  match goal with
    | |- context [MemMod.merge M1 (MemMod.merge ?M2 ?M3)] => 
      rewrite <- (@merge_assoc M1 M2 M3)
    | |- context [MemMod.merge (MemMod.merge ?M2 ?M3) ?M1] => 
      rewrite (@merge_assoc M2 M3 M1)
    | _ => fail
  end.

Tactic Notation "merge" "assoc" constr (M1) := merge_assoc M1.


(******* disjUnion ********)

Ltac disjUnion_shift_r :=
  match goal with
    | |- MemMod.disjUnion (MemMod.merge ?M1 ?M2) ?M3 ?M =>
      apply (@disjUnion_assoc_right_left M1 M2 M3 M); [idtac | try disjoint_solve]
    | |- _ => fail
  end.

Ltac disjUnion_shift_l :=
  match goal with
    | |- MemMod.disjUnion ?M1 (MemMod.merge ?M2 ?M3) ?M =>
      apply (@disjUnion_assoc_left_right M1 M2 M3 M); [idtac | try disjoint_solve]
    | |- _ => fail
  end.


Tactic Notation "disjUnion" "shift_r" := disjUnion_shift_r.

Tactic Notation "disjUnion" "shift_l" := disjUnion_shift_l.

(* Ltac disjUnion_replace_l M1 := *)
(*   match goal with  *)
(*     | |- MemMod.disjUnion ?M1' ?M2 ?M3  *)
(*       => apply (@eqDom_disjUnion_left M1 M2 M3 M1'); [trivial | trivial] *)
(*   end. *)

(* Ltac disjUnion_replace_r M2 := *)
(*   match goal with  *)
(*     | |- MemMod.disjUnion ?M1 ?M2' ?M3 *)
(*       => apply (@eqDom_disjUnion_right M1 M2 M3 M2'); [trivial | trivial] *)
(*   end. *)

(* Tactic Notation "disjUnion" "rep_l" constr (M1) := disjUnion_replace_l M1. *)

(* Tactic Notation "disjUnion" "rep_r" constr (M2) := disjUnion_replace_r M2. *)

Unset Implicit Arguments.
