(****** Operational Semantics ******)

Require Import Coqlib.
Require Import Map.
Require Import Arith.
Require Import Machine.
Require Import Assertion.
Require Import Assertion_Lemmas.
Require Import Operational_Semantics.

Lemma itr_enabled_dec : forall S,
  itr_enabled S \/ ~ itr_enabled S.
Proof.
  intros [[[M R] fl] isr].
  generalize (refl_equal (fl _IF)); pattern fl at 1;
    destruct (fl _IF); intro Hfl.
  destruct isr.
  right.
  intros [Hx Hy].
  simplS in * .
  discriminate.
  left.
  unfold itr_enabled; simplS.
  auto.
  right.
  intros [Hx Hy].
  simplS in * .
  rewrite Hfl in Hx.
  discriminate.
Qed.

(**********************************************************

  nextS nextPC Lemmas 

***********************************************************)

Lemma nextS_inversion : forall pc c M R fl isr M' R' fl' isr',
  nextS pc c (M, R, fl, isr) (M', R', fl', isr')
  -> match c with 
       | comms _retw => (M=M') 
                          /\ (exists pc', leadsToW (R SP) pc' M 
                                /\ R' = RF_update R SP (R SP + 2))
                          /\ fl=fl' /\ isr=isr'
       | comms (_jmp f) => (M'=M) /\ (R'=R)/\ fl'=fl /\ isr'=isr
       | comml (_cli) => (M'=M) /\ (R'=R)/\ fl'=(Clear_Flag fl _IF) /\ isr'=isr
       | comml (_sti) => (M'=M) /\ (R'=R)/\ fl'=(Set_Flag fl _IF) /\ isr'=isr

       | comml (_call f) => (R SP >= 2 /\ leadsToBlk (R SP - 2) 2 M /\ MemUpd16 M (R SP - 2) (pc+1) M') 
                            /\ (R'= RF_update R SP (R SP - 2))/\ fl=fl' /\ isr=isr'

       | comml (_jne f) => (M'=M) /\ (R'=R)/\ fl'=fl /\ isr'=isr
       | comml (_movwp o r) => (M'=M) /\ (R'=RF_update R r (opr_eval R o))
                         /\ fl'=fl /\ isr'=isr
       | comml (_movwld a r) => (M'=M) 
                          /\ (exists x, leadsToW (addr_eval R a) x M /\ (R'=RF_update R r x))
                         /\ fl'=fl /\ isr'=isr
       | comml (_movwst o a) => (leadsToxW (addr_eval R a) M 
                              /\ MemUpd16 M (addr_eval R a) (opr_eval R o) M')
                            /\ R'=R /\ fl'=fl /\ isr'=isr
       | comml (_pushwp o) => (R SP >= 2 /\ leadsToBlk (R SP - 2) 2 M /\ MemUpd16 M (R SP -2) (opr_eval R o) M')
                            /\ R' = RF_update R SP (R SP -2) /\ fl' = fl /\ isr'=isr
       | comml (_pushwm a) => (exists x, leadsToW (addr_eval R a) x M 
                                /\ R SP >= 2 /\ leadsToBlk (R SP - 2) 2 M /\ MemUpd16 M (R SP -2) x M')
                            /\ R' = RF_update R SP (R SP -2) 
                            /\ fl' = fl /\ isr'=isr
       | comml (_pushfw) => (R SP >= 2 /\ leadsToBlk (R SP - 2) 2 M /\ MemUpd16 M (R SP -2) (FR2nat fl) M')
                            /\ R' = RF_update R SP (R SP -2) /\ fl' = fl /\ isr'=isr
       | comml (_popw r) => (M'=M)
                            /\ (exists x, leadsToW (R SP) x M 
                                 /\ R' = RF_update (RF_update R SP (R SP +2)) r x)
                            /\ fl' = fl /\ isr'=isr
       | comml _popfw => (M'=M) /\ (R' = RF_update R SP (R SP +2))
                         /\ (exists flnum, leadsToW (R SP) flnum M /\ fl' = nat2FR flnum)
                         /\ isr'=isr
       | comml (_cmpw o r) => (M'=M)/\(R'=R)
                               /\ ((opr_eval R o = R r) /\ (fl' = Set_Flag fl _ZF)
                                 \/ (opr_eval R o <> R r) /\ (fl' = Clear_Flag fl _ZF))
                               /\ isr'=isr
       | comml (_addw o r) => (M'=M)/\(R'=RF_update R r (opr_eval R o + R r))
                               /\ (trash_fl fl fl' _ZF)
                               /\ isr'=isr
       | comml (_subw o r) => (M'=M)/\(R'=RF_update R r (R r - opr_eval R o))
                               /\ (trash_fl fl fl' _ZF)
                               /\ isr'=isr
       | comml _outb_eoi => (M'=M) /\ (R'=R /\ R AX = 32 /\ R DX = 32) /\ (fl'=fl) /\ (isr'=false)
       | comms (_iret) => (M'=M /\ leadsToBlk (R SP) 6 M) 
                             /\ R'=RF_update R SP (R SP + 6)
                             /\ (exists flnum, fl' = nat2FR flnum /\ leadsToW (R SP + 4) flnum M)
                             /\ isr' = isr
     end.
Proof.
  induction c as [ i | s]; [destruct i | destruct s]; trivial.
  (* case c = cli *)
  intros.
  inversion_clear H; auto.
  (* case c = sti *)
  intros.
  inversion_clear H; auto.
  (* case c = _call f *)
  intros.
  inversion_clear H.
  destruct ((fun x => H0) 0).
  destruct H1.
  destruct H.
  subst; auto.
  (* case c = jne f *)
  intros.
  inversion_clear H; auto.
  (* case c = movwp o r *)
  intros.
  inversion_clear H; subst; auto.
  (* case c = movwld a r *)
  intros.
  inversion_clear H.
  subst.
  split_l. 
  split.
  exists x; split_lr.
  split_lr.
  (* case c = movwst o a *)
  intros.
  inversion_clear H; subst; auto.
  (* case c = pushwp o *)
  intros.
  inversion_clear H.
  inversion_clear H1.
  inversion_clear H.
  subst.
  destruct H2.
  auto.
  (* case c = pushwm a *)
  intros.
  inversion_clear H.
  inversion_clear H2.
  inversion_clear H.
  subst.
  destruct H3.
  split.
    exists x.
    split_l; split_l; split_lr.
  split_l;  split_lr.
  (* case c = pushfw *)
  intros.
  inversion_clear H.
  inversion_clear H0.
  inversion_clear H.
  destruct H1.
  auto.
  (* case c = popw r *)
  intros.
  inversion_clear H.
  inversion_clear H0.
  subst.
  split_l.
  split.
    exists x.
    split_lr.
  split_lr.
  (* case c = popfw *)
  intros.
  inversion_clear H.
  inversion_clear H0.
  subst.
  split_l.
  split_l.
  split_r.
  exists flnum; split_lr.
  (* case c = cmpw o r *)
  intros.
  inversion_clear H.
  destruct H0.
   destruct H.
   subst.
   split_l.
   split_l.
   split_r.
   left; split_lr.
  destruct H.
  subst.
  split_l.
  split_l.
  split_r.
  right; split_lr.
  (* case c = addw o r *)
  intros.
  inversion_clear H.
  subst.
  split_l.
  split_l.
  split_lr.
  (* case c = subw o r *)
  intros.
  inversion_clear H.
  subst.
  split_l.
  split_l.
  split_lr.
  (* case c = outb_eoi *)
  intros.
  inversion_clear H.
  split_l.
  split.
    split_l; split_lr.
  split_lr.
  (* case c = jmp f *)
  intros.
  inversion_clear H.
  split_l.
  split_l; split_lr.
  (* case c = _retw *)
  intros.
  inversion H; auto.
  (* case c = iret *)
  intros.
  inversion_clear H; subst; intuition.
  exists flnum; auto.
Qed.
Implicit Arguments nextS_inversion [pc c M R fl isr M' R' fl' isr'].

Lemma nextS_injection : 
  forall c pc M R fl isr M' R' fl' isr' M'' R'' fl'' isr'',
    nextS pc c (M,R,fl,isr) (M',R',fl',isr')
    -> nextS pc c (M,R,fl,isr) (M'',R'',fl'',isr'')
    -> match c with 
         | comml (_call f) => 
             (MemUpd16 M (R SP-2) (pc+1) M'' /\ R'=R'')/\(fl'=fl''/\isr'=isr'')
         | comml (_movwst o a) => 
             (MemUpd16 M (addr_eval R a) (opr_eval R o) M'' /\ R'=R'')/\(fl'=fl''/\isr'=isr'')
         | comml (_pushwp o) => 
             (MemUpd16 M (R SP-2) (opr_eval R o) M'' /\ R'=R'')/\(fl'=fl''/\isr'=isr'')
         | comml (_pushwm a) => 
             ((exists x, leadsToW (addr_eval R a) x M /\ MemUpd16 M (R SP-2) x M'')
               /\ R'=R'')
             /\(fl'=fl''/\isr'=isr'')
         | comml (_pushfw) => 
             (MemUpd16 M (R SP-2) (FR2nat fl) M'' /\ R'=R'')
             /\(fl'=fl''/\isr'=isr'')
         | comml (_addw o r) => (M'=M''/\R'=R'')/\ (trash_fl fl fl'' _ZF /\isr'=isr'')
         | comml (_subw o r) => (M'=M''/\R'=R'')/\ (trash_fl fl fl'' _ZF /\isr'=isr'')
         | _ => (M'=M''/\R'=R'')/\(fl'=fl''/\isr'=isr'')
       end.
Proof.
  intros c pc M R fl isr M' R' fl' isr' M'' R'' fl'' isr''.
  intros HnextS' HnextS''.
  destruct c as [i | i]; destruct i;
    inversion HnextS'; subst; 
      inversion HnextS''; subst; 
        try (split; split_lr; fail).
  inversion H2; inversion H3.
    destruct H0; destruct H4.
    subst; split; split_lr.
  rewrite (leadsToW_unique H11 H12); split;split_lr.
  inversion H10; inversion H11.
    destruct H0; destruct H2.
    subst; split; split_lr.
  inversion H11; inversion H13.
    destruct H0; destruct H2.
    subst.
    assert (Heq := leadsToW_unique H10 H12).
    subst x0.
    split;split_lr.
    exists x; split_lr.
  inversion H2; inversion H3.
    destruct H0; destruct H4.
    subst; split; split_lr.
  inversion H4; inversion H3.
    subst.
    assert (Heq := leadsToW_unique H H1); subst.
    split; split_lr.
  inversion H4; inversion H3.
    subst.
    assert (Heq := leadsToW_unique H H1); subst.
    split; split_lr.
  destruct H2 as [Hx | Hx]; destruct H1 as [Hy | Hy];
    destruct Hx; destruct Hy;
    subst; split; split_lr; contradiction.
  destruct H3; destruct H2.
    inversion H; inversion H0.
    subst. 
    assert (Heq := leadsToW_unique H1 H3); subst.
    split; split_lr.
  rewrite (leadsToW_unique H10 H12); split;split_lr.
Qed.
Implicit Arguments nextS_injection [c pc M R fl isr M' R' fl' isr' M'' R'' fl'' isr''].

(******************************************************)

Lemma nextPC_inversion : forall pc c M R fl pc',
  nextPC c M R fl pc pc'
  -> match c with 
       | comml (_call f) => pc' = f
       | comms _retw => leadsToW (R SP) pc' M
       | comms _iret => leadsToW (R SP) pc' M /\ leadsToBlk (R SP) 6 M
       | comms (_jmp f) => pc' = f
       | comml (_jne f) => (fl _ZF = true -> pc'=pc+1) /\ (fl _ZF = false -> pc'=f)
       | comml _ => pc' = pc+1
     end.
Proof.
  intros pc c M R fl pc' H.
  destruct c as [i | i]; destruct i;
    inversion H; subst; auto.
  split; intro; trivial.
  rewrite H1 in H0; discriminate.
  split; intro; trivial.
  rewrite H1 in H0; discriminate.
Qed.
Implicit Arguments nextPC_inversion [pc c M R fl pc'].

Lemma nextPC_injection : forall c M R fl pc pc' pc'',
  nextPC c M R fl pc pc'
  -> nextPC c M R fl pc pc''
  -> pc' = pc''.
Proof.
  intros c M R fl pc pc' pc'' HnextPC' HnextPC''.
  destruct c as [i | i]; destruct i;
    inversion HnextPC'; subst; 
      inversion HnextPC''; subst; trivial.
  rewrite H0 in H1; discriminate.
  rewrite H0 in H1; discriminate.
  rewrite (leadsToW_unique H0 H1); trivial.
  rewrite (leadsToW_unique H1 H3); trivial.
Qed.
Implicit Arguments nextPC_injection [c M R fl pc pc' pc''].


(******************************************************)

Lemma nextS_Mono : forall pc c M1 M R fl isr M1' R' fl' isr' M2,
  MemMod.disjUnion M1 M2 M
  -> nextS pc c (M1, R, fl, isr) (M1', R', fl', isr')
  -> exists M', MemMod.disjUnion M1' M2 M'
    /\ nextS pc c (M, R, fl, isr) (M', R', fl', isr').
Proof.
  intros.
  induction c as [i|i]; destruct i; 
    destruct (nextS_inversion H0) as [HM [HR [Hfl Hisr]]]; subst.
  exists M; split_l.
  constructor.
  exists M; split_l.
  constructor.
  unfold stkSpace in HM.

  exists (MemMod.merge M1' M2).
  destruct HM as [H1 [H2 H3]].
  destruct (leadsToBlk_MemUpd16_Mono H H2 H3) as [H4 [H5 H6]].
  auto with arith.
  split_l.
  constructor.
  unfold pushW.
  unfold stkSpace.
  split.
    split_lr.
    eapply leadsToBlk_Mono; eauto.
  split.
    destruct (leadsToBlk_MemUpd16_Mono H H2 H3) as [H7 [H8 H9]]; auto.
  trivial.

  exists M; split_l.
  constructor.

  exists M; split_l.
  destruct (nextS_inversion H0) as [HM [HR [Hfl Hisr]]].
  eapply nextS_movwp; eauto.
  destruct HR.
  
  exists M; split_l.
  destruct H1.
  eapply nextS_movwld; eauto.
  eapply leadsToW_Mono; eauto.

  exists (MemMod.merge M1' M2).
  destruct HM as [H1 H2].
  destruct (leadsToxW_MemUpd16_Mono H H1 H2).
  split_l.
  eapply nextS_movwst; eauto.
  eapply leadsToxW_Mono; eauto.
  
  exists (MemMod.merge M1' M2).
  destruct HM as [H1 [H2 H3]].
  destruct (leadsToBlk_MemUpd16_Mono H H2 H3) as [H4 [H5 H6]].
  auto.
  split_l.
  eapply nextS_pushwp; eauto.
  split.
    split_lr.
    eapply leadsToBlk_Mono; eauto.
  split_lr.

  exists (MemMod.merge M1' M2).
  destruct HM as [x [H1 [H2 [H3 H4]]]].
  destruct (leadsToBlk_MemUpd16_Mono H H3 H4) as [H5 [H6 H7]].
  auto.
  split_l.
  apply nextS_pushwm with (addr_eval R a) x; eauto.
    eapply leadsToW_Mono; eauto.
  split.
    split_lr.
    eapply leadsToBlk_Mono; eauto.
  split_lr.

  exists (MemMod.merge M1' M2).
  destruct HM as [H1 [H2 H3]].
  destruct (leadsToBlk_MemUpd16_Mono H H2 H3) as [H4 [H5 H6]].
  auto.
  split_l.
  eapply nextS_pushfw; eauto.
  split.
    split_lr.
    eapply leadsToBlk_Mono; eauto.
  split_lr.

  exists M; split_l.
  destruct HR as [x [H1 H2]].
  eapply nextS_popw; eauto.
  split_r.
  eapply leadsToW_Mono; eauto.

  exists M; split_l.
  destruct Hfl as [x [H1 H2]].
  eapply nextS_popfw; eauto.
  split_r.
  eapply leadsToW_Mono; eauto.

  exists M; split_l.
  eapply nextS_cmpw; eauto.

  exists M; split_l.
  eapply nextS_addw; eauto.

  exists M; split_l.
  eapply nextS_subw; eauto.

  exists M; split_l.
  destruct HR as [H1 [H2 H3]]; subst.
  eapply nextS_EOI; eauto.

  exists M; split_l.
  eapply nextS_jmp; eauto.

  exists M; split_l.
  eapply nextS_retw; eauto.
  destruct HR.
  exists x; unfold popW; auto.
  destruct H1; split_r.
  eapply leadsToW_Mono; eauto.

  destruct HM.
  subst.
  exists M; split_l.
  destruct Hfl as [fln [H1 H3]].
  eapply nextS_iret; eauto.
  replace (R SP + 0) with (R SP); auto.
  eapply leadsToBlk_Mono; eauto.
  replace (R SP + 0) with (R SP); auto.
  eapply leadsToW_Mono; eauto.
Qed.
Implicit Arguments nextS_Mono [pc c M1 M R fl isr M1' R' fl' isr' M2].


Require Omega.

Lemma itr_nextS_Mono : forall pc M1 M2 M R fl isr M1' R' fl' isr',
  MemMod.disjUnion M1 M2 M
  -> itr_nextS pc (M1, R, fl, isr) (M1',R',fl',isr')
  -> exists M' : MemMod.Map,
    MemMod.disjUnion M1' M2 M'
    /\ itr_nextS pc (M, R, fl, isr) (M', R', fl', isr').
Proof.
  intros pc M1 M2 M R fl isr M1' R' fl' isr'.
  unfold itr_nextS.
  intros HJM HnextS.
  exists (MemMod.merge M1' M2).
  destruct HnextS as [sp [fln [cs [M1'' [M1''' HnextS]]]]].
  destruct HnextS as [Hsp [Hk [HR' [Hfl' [Hisr' [Hfln [Hupd [Hupd' Hupd'']]]]]]]].
  subst sp.
  destruct ((fun x => Hk) 0) as [Hsp Hlds].
  destruct (leadsToBlk_MemUpd16_Mono HJM Hlds Hupd) as [HJ'' [Hlds'' HM1'']].
  auto with arith.
  assert (Hldsx : leadsToBlk (R SP - 4) (6 - 2) M1'').
    replace (R SP - 4) with (R SP - 6 + 2); [ idtac | omega].
    apply leadsToBlk_shrink_low; auto with arith.
  assert (Hldsy : leadsToBlk (R SP - 4) 4 M1'').
    apply leadsToBlk_shrink_high with 4; auto with arith.
  destruct (leadsToBlk_MemUpd16_Mono HJ'' Hldsy Hupd') as [HJ''' [Hlds''' HM1''']].
  auto with arith.
  clear Hldsx Hldsy.
  assert (Hldsx : leadsToBlk (R SP - 2) (4 - 2) M1''').
    replace (R SP - 2) with (R SP - 4 + 2); [ idtac | omega].
    apply leadsToBlk_shrink_low; auto with arith.
  destruct (leadsToBlk_MemUpd16_Mono HJ''' Hldsx Hupd'') as [HJ' [Hlds' HM1']].
  auto with arith.
  split_l.
  exists (R SP);exists fln; exists cs.
  exists (MemMod.merge M1'' M2); exists (MemMod.merge M1''' M2).
  split_l. 
  split. 
    eapply stkSpace_Mono; eauto.
  split_l; split_l; split_l; split_l.
  split_l; split_lr.
Qed.
Implicit Arguments itr_nextS_Mono [pc M1 M2 M R fl isr M1' R' fl' isr'].


(**********************************************)


Lemma nextS_DMono : forall pc c M1 M2 M R fl isr M' R' fl' isr' S0,
  MemMod.disjUnion M1 M2 M
  -> nextS pc c (M1, R, fl, isr) S0
  -> nextS pc c (M, R, fl, isr) (M', R', fl', isr')
  -> exists M1',
    MemMod.disjUnion M1' M2 M'
    /\ nextS pc c (M1, R, fl, isr) (M1', R', fl', isr').
Proof.
  intros pc c M1 M2 M R fl isr M' R' fl' isr' S0.
  intros HJM HnextS0 HnextS.
  destruct S0 as [[[M10 R0] fl0] isr0].
  induction c as [i|i]; destruct i;
    (destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]]; subst);
    (destruct (nextS_inversion HnextS0) as [HM0 [HR0 [Hfl0 Hisr0]]]; subst).
(* case i = cli *)
  exists M1; split_l.
  constructor.
(* case i = sti *)
  exists M1; split_l.
  constructor.
(* case i = call f *)
  destruct HM as [Hsp [HlM HM]].
  destruct HM0 as [Hsp'[HlM1 HM1]].
  destruct (leadsToBlk_MemUpd16_DMono HJM HlM1 HM)
    as [M1' [HJM' H3]].
  exists M1'; split_l.
  constructor; unfold pushW, stkSpace; auto.
(* case i = jne f *)
  exists M1; split_l.
  constructor.
(* case i = movwp o r *)
  exists M1; split_l.
  econstructor; eauto.
(* case i = movwld a r *)
  exists M1; split_l.
  destruct HR as [x [H1 H2]].
  destruct HR0 as [x0 [H3 H4]].
  assert (leadsToW (addr_eval R a) x M1).
    assert (H:=leadsToW_Mono HJM H3).
    assert (Heq := leadsToW_unique H1 H).
    subst x0.
    trivial.
  apply nextS_movwld with (addr_eval R a) x; trivial.
(* case i = movwst o a *)
  destruct HM as [H1 H2].
  destruct HM0 as [H3 H4].
  destruct (leadsToxW_MemUpd16_DMono HJM H3 H2)
    as [M1' [HJM' H5]].
  exists M1'; split_l.
  eapply nextS_movwst; eauto.
(* case i = pushwp o *)
  destruct HM as [Hsp [H1 H2]].
  destruct HM0 as [Hsp' [H3 H4]].
  destruct (leadsToBlk_MemUpd16_DMono HJM H3 H2)
    as [M1' [HJM' H5]].
  exists M1'; split_l.
  apply nextS_pushwp with (opr_eval R o); trivial.
  unfold pushW, stkSpace.
  split.
    split_lr.
  split_lr.
(* case i = pushwm a *)
  destruct HM as [x [H1 [H2 [H3 H4]]]].
  destruct HM0 as [x0 [H10 [H20 [H30 H40]]]].
  assert (Heq := leadsToW_unique H1 (leadsToW_Mono HJM H10)).
  subst x0.
  destruct (leadsToBlk_MemUpd16_DMono HJM H30 H4)
    as [M1' [HJM' H8]].
  exists M1'; split_l.
  apply nextS_pushwm with (addr_eval R a) x; trivial.
  unfold pushW, stkSpace.
  split.
    split_lr.
  split_lr.
(* case i = pushfw  *)
  destruct HM as [H1 [H2 H3]].
  destruct HM0 as [H10 [H20 H30]].
  destruct (leadsToBlk_MemUpd16_DMono HJM H20 H3)
    as [M1' [HJM' H8]].
  exists M1'; split_l.
  apply nextS_pushfw.
  unfold pushW, stkSpace.
  split.
    split_lr.
  split_lr.
(* case i = popw r *)
  destruct HR as [x [H1 H2]].
  destruct HR0 as [x0 [H10 H20]].
  assert (Heq := leadsToW_unique H1 (leadsToW_Mono HJM H10)).
  subst.
  exists M1; split_l.
  apply nextS_popw with (RF_update R SP (R SP + 2)) x0; trivial.
  unfold popW; auto.
(* case i = popfw *)
  destruct Hfl as [x [H1 H2]].
  destruct Hfl0 as [x0 [H10 H20]].
  assert (Heq := leadsToW_unique H1 (leadsToW_Mono HJM H10)).
  subst.
  exists M1; split_l.
  apply nextS_popfw with x0; trivial.
  unfold popW; auto.
(* case i = cmpw o r *)
  exists M1; split_l.
  apply nextS_cmpw; trivial.
(* case i = addw o r *)
  exists M1; split_l.
  eapply nextS_addw; eauto.
(* case i = subw o r *)
  exists M1; split_l.
  eapply nextS_subw; eauto.
(* case i = outb eoi *)
  exists M1; split_l.
  destruct HR as [H1 [H2 H3]].
  destruct HR0 as [H10 [H20 H30]].
  subst.
  eapply nextS_EOI; eauto.
(* case i = jmp f *)
  exists M1; split_l.
  eapply nextS_jmp; eauto.
(* case i = retw *)
  destruct HR as [pcx [H1 H2]].
  destruct HR0 as [pcx0 [H10 H20]].
  exists M10; split_l.
  eapply nextS_retw; eauto.
  assert (Heq := leadsToW_unique H1 (leadsToW_Mono HJM H10)).
  subst.
  exists pcx0; unfold popW, stkSpace; auto.
(* case i = iret *)
  destruct HM as [H1 H2].
  destruct HM0 as [H10 H20].
  destruct Hfl as [x [Hfl H3]].
  destruct Hfl0 as [x0 [Hfl0 H30]].
  subst.
  exists M1; split_l.
  eapply nextS_iret; eauto.
  replace (R SP + 0) with (R SP) ; auto.
  replace (R SP + 0) with (R SP) ; auto.
  assert (Heq := leadsToW_unique H3 (leadsToW_Mono HJM H30)).
  subst; trivial.
Qed.
Implicit Arguments nextS_DMono [pc c M1 M2 M R fl isr M' R' fl' isr' S0].

Lemma MemUpd16_3_DMono : forall M1 M2 M M'1 M'2 l x1 x2 x3 M',
  MemMod.disjUnion M1 M2 M
  -> leadsToBlk l 6 M1
  -> MemUpd16 M l x1 M'1
  -> MemUpd16 M'1 (l+2) x2 M'2
  -> MemUpd16 M'2 (l + 4) x3 M'
  -> exists M1', exists M1'1, exists M1'2,
    MemMod.disjUnion M1' M2 M'
    /\ MemUpd16 M1 l x1 M1'1
    /\ MemUpd16 M1'1 (l+2) x2 M1'2
    /\ MemUpd16 M1'2 (l + 4) x3 M1'.
Proof. 
intros.
assert (leadsToBlk l 2 M1).
apply leadsToBlk_shrink_high with 6;intuition.
destruct (leadsToBlk_MemUpd16_DMono H H4 H1)
    as [m'1 [H5 H6]].
assert (6>=2).
omega.
destruct (leadsToBlk_MemUpd16_Mono H H0 H6 H7) as [H01 [H02 H03]].
assert (leadsToBlk (l+2) (6-2) m'1).
apply leadsToBlk_shrink_low;intuition.
arith_rep (6-2) with 4 in H8.
assert (leadsToBlk (l+2) 2 m'1).
apply leadsToBlk_shrink_high with 4;intuition.
destruct (leadsToBlk_MemUpd16_DMono H5 H9 H2)
    as [m'2 [H10 H11]].
assert (4>=2).
omega.
destruct (leadsToBlk_MemUpd16_Mono H5 H8 H11 H12) as [H001 [H002 H003]].
assert (leadsToBlk ((l+2)+2) (4-2) m'2).
apply leadsToBlk_shrink_low;intuition.
arith_rep (4-2) with 2 in H13.
arith_rep (l+2+2) with (l+4) in H13.
destruct (leadsToBlk_MemUpd16_DMono H10 H13 H3)
    as [m1' [H14 H15]].
exists m1';exists m'1;exists m'2.
intuition.
Qed.

Implicit Arguments MemUpd16_3_DMono [M1 M2 M M'1 M'2 l x1 x2 x3 M'].

Lemma itr_nextS_DMono : forall pc M1 M2 M R fl isr M' R' fl' isr',
  MemMod.disjUnion M1 M2 M
  -> stkSpace (R SP) 6 M1
  -> itr_nextS pc (M, R, fl, isr) (M',R',fl',isr')
  -> exists M1' : MemMod.Map,
    MemMod.disjUnion M1' M2 M'
    /\ itr_nextS pc (M1, R, fl, isr) (M1', R', fl', isr').
Proof.
  intros pc M1 M2 M R fl isr M' R' fl' isr'.
  intros HJM Hstk HnextS.
  unfold itr_nextS in HnextS.
  destruct HnextS as [sp [fln [cs [M'' [M''' H ]]]]].
  destruct H as [Hsp [Hstk2 [HR' [Hfl' [Hisr' [Hfln [Hupd1 [Hupd2 Hupd3]]]]]]]].
  unfold stkSpace in Hstk.
  destruct Hstk.
  subst sp.
  assert (Hx0 : R SP - 4 = R SP - 6 + 2).
    omega.
  rewrite Hx0 in Hupd2.
  assert (Hx1 : R SP - 2 = R SP - 6 + 4).
    omega.
  rewrite Hx1 in Hupd3.
  destruct (MemUpd16_3_DMono HJM H0 Hupd1 Hupd2 Hupd3) 
    as [M1' [M1'' [M1''' [HJM' [Hupd1' [Hupd2' Hupd3']]]]]].
  exists M1'; split_l.
  unfold itr_nextS.
  exists (R SP); exists fln; exists cs; exists M1''; exists M1'''.
  split_l.
  split. split; auto.
  split_l; split_l; split_l; split_l.
  rewrite Hx0.
  rewrite Hx1.
  auto.
Qed.
Implicit Arguments itr_nextS_DMono [pc M1 M2 M R fl isr M' R' fl' isr'].

(******************************************************)


Lemma nextS_eqDom : forall pc c M R fl isr M' R' fl' isr',
  nextS pc c (M, R, fl, isr) (M', R', fl', isr')
  -> MemMod.eqDom M M'.
Proof.
  induction c as [ i | i]; destruct i; simpl; intros; 
    destruct (nextS_inversion H) as [HM [HR [Hfl Hisr]]]; subst; auto with PMap.
  destruct HM as [H1 [H2 H3]].
  eapply leadsToBlk_MemUpd16_eqDom; eauto.
  destruct HM as [H1 H2].
  eapply leadsToxW_MemUpd16_eqDom; eauto.
  destruct HM as [H1 [H2 H3]].
  eapply leadsToBlk_MemUpd16_eqDom; eauto.
  destruct HM as [x [H1 [H2 [H3 H4]]]].
  eapply leadsToBlk_MemUpd16_eqDom; eauto.
  destruct HM as [H1 [H2 H3]].
  eapply leadsToBlk_MemUpd16_eqDom; eauto.
  destruct HM.
  subst M'.
  auto with PMap.
Qed.
Implicit Arguments nextS_eqDom [pc c M R fl isr M' R' fl' isr'].

Lemma itr_nextS_eqDom : forall pc M R fl isr M' R' fl' isr',
  itr_nextS pc (M, R, fl, isr) (M', R', fl', isr')
  -> MemMod.eqDom M M'.
Proof.
  unfold itr_nextS.
  intros pc M R fl isr M' R' fl' isr' H.
  destruct H as [sp [fln [cs [M'1 [M'2 [Hsp [Hstk [HR' [Hfl' [Hisr' [Hfln H]]]]]]]]]]].
  destruct H as [H1 [H2 H3]].
  unfold stkSpace in Hstk.
  destruct Hstk.
  assert (leadsToBlk (sp - 4) 4 M).
    arith_rep 4 with (6 -2).
    arith_rep (sp - (6 - 2)) with (sp - 6 + 2).
    apply leadsToBlk_shrink_low; trivial.
    omega.
  assert (leadsToBlk (sp - 2) 2 M).
    arith_rep 2 with (6 -4).
    arith_rep (sp - (6 - 4)) with (sp - 6 + 4).
    apply leadsToBlk_shrink_low; trivial.
    omega.
  assert (leadsToxW (sp - 6) M).
    apply leadsToBlk_leadsToxW with 6; trivial.
    omega.
  assert (leadsToxW (sp - 4) M).
    apply leadsToBlk_leadsToxW with 4; trivial.
    omega.
  assert (leadsToxW (sp - 2) M).
    apply leadsToBlk_leadsToxW with 2; trivial.
    omega.
  apply MemMod.eqDom_trans with M'1.
  apply leadsToxW_MemUpd16_eqDom with (sp-6) pc; trivial.
  apply MemMod.eqDom_trans with M'2.
  apply leadsToxW_MemUpd16_eqDom with (sp-4) cs; trivial.
  apply leadsToxW_MemUpd16_leadsToxW_nt with (sp-6) pc M; trivial.
    left; omega.
  apply leadsToxW_MemUpd16_eqDom with (sp-2) fln; trivial.
  apply leadsToxW_MemUpd16_leadsToxW_nt with (sp-4) cs M'1; trivial.
  apply leadsToxW_MemUpd16_leadsToxW_nt with (sp-6) pc M; trivial.
    left; omega.
    left; omega.
Qed.
Implicit Arguments itr_nextS_eqDom [pc M R fl isr M' R' fl' isr'].
