(****** Operational Semantics ******)

Require Import Coqlib.
Require Import Map.
Require Import Arith.
Require Import Machine.
Require Import Assertion.
Require Import Assertion_Lemmas.
Require Import Assertion_Lemmas2.
Require Import Operational_Semantics.
Require Import OpSem_Lemmas.

(**************************************************
 * Properties of operational semantics of instrs
 **************************************************)
Lemma ptoxBlk_split : forall s1 s2 M l,
  ptoxBlk l (s1 + s2) M
  -> mstar (ptoxBlk l s1) (ptoxBlk (l+s1) s2) M.
(** added by wanghaibo **)
Proof.
  induction s1; simpl; intros.
  unfold Aemp, MemMod.emp.
  exists MemMod.emp; exists M.
  split.
    auto with PMap.
    unfold MemMod.emp.
    split_l.
  replace (l+0) with l; auto with arith.
  destruct  H as [M1 [M2 [MJM [HM1 HM2]]]].
  replace (l + S s1) with (S l + s1); [idtac | omega].
  assert (Hx := IHs1 s2 M2 (l+1) HM2).
  apply mstar_assoc_right_left.
  unfold mstar.
  exists M1; exists M2; split_l.
  split_l.
  arith_rep (S l + s1) with (l+1+s1).
Qed.

Lemma stkSpace_rev_mono:
  forall sp n n' M,
    stkSpace sp n M -> n >= n' -> stkSpace sp n' M.
Proof.
unfold stkSpace.
intros.
destruct H as [Ineq Mp].
split.
omega.
unfold leadsToBlk.
unfold imprecM.
assert (leadsToBlk  (sp - n)((n - n') + n') M).
assert (n - n' + n' = n).
omega.
rewrite H;assumption.
destruct H as [M1 [M2 [Mdis [MH MH1]]]].
assert ((ptoxBlk (sp - n) (n - n') *^* ptoxBlk (sp - n +(n - n')) n') M1).
eapply ptoxBlk_split.
assumption.
assert (sp - n + (n - n') = sp - n').
omega.
rewrite H1 in H.
destruct H as [M3 [M4 [Mdis1 [MH2 MH3]]]].
exists M4.
exists (MemMod.merge M3 M2).
split.
assert (MemMod.disjUnion M4 M3 M1).
apply (MemMod.disjUnion_commut Mdis1).
eapply disjUnion_assoc_lr.
apply H.
assumption.
split.
assumption.
unfold mTrue;trivial.
Qed.


(* movwp *)
Lemma trivial_enable_movwp:
forall p f o r, enable p (nextS f (comml (_movwp o r))).
Proof.
unfold enable.
intros.
destruct S as [[[M R] F]  b].
exists (M, RF_update R r (opr_eval R o) , F, b).
eapply nextS_movwp.
reflexivity.
trivial.
Qed.
Hint Resolve trivial_enable_movwp.


Lemma frame_movwp:
  forall f o r, frame (nextS f (comml (_movwp o r))).
Proof.
unfold frame.
intros.
exists M1.
split;destruct H1 as [M H1];inversion H1;subst;inversion H;subst;simpl.
apply H0.
unfold S_Mem_Upd.
simpl.
eapply nextS_movwp.
reflexivity.
trivial.
Qed.

(* movwld *)
Lemma enable_movwld:
  forall (p : Pred) a r f, 
    (forall S, p S -> leadsToxW (addr_eval (S_RF S) a) (S_Mem S)) -> enable p (nextS f (comml (_movwld a r))).
Proof.
unfold enable.
intros.
destruct S as [[[M R] F] b].
assert ((addr_eval (S_RF (M, R, F, b)) a |~>w_) (S_Mem (M, R, F, b))).
apply (H (M, R, F, b));trivial.
simpl in H1.
destruct H1 as [M1 [M2 [Mdis [MH1 MH2]]]].
destruct MH1 as [M11 [M22 [Mdis' [MH11 MH22]]]].
destruct MH11.
exists (M,RF_update R r x,F,b).
eapply nextS_movwld.
reflexivity.
exists M1. exists M2.
split;[assumption | split].
exists M11. exists M22.
split;[assumption | split].
apply H1.
assumption.
trivial.
trivial.
Qed.

Lemma safe_mono_movwld:
  forall f a r, safe_mono (nextS f (comml (_movwld a r))).
Proof.
unfold safe_mono. unfold enable.
intros.
destruct H0 as [M1 [M2 [Mdis [MH MH1]]]].
destruct S as [[[M R] F] b].
simpl in Mdis.
unfold S_Mem_Upd in MH.
simpl in MH.
assert (exists S' : State, nextS f (comml (_movwld a r)) (M1, R, F, b) S').
apply H;apply MH.
destruct H0 as [S' H0].
inversion H0.
subst.
exists (M, RF_update R r x, F, b).
assert ((addr_eval R a |~>w x) M).
destruct H10 as [M3 [M4 [Mdis1 [MH2 MH3]]]].
exists M3. exists (MemMod.merge M4 M2).
split.
eapply disjUnion_assoc_lr.
apply Mdis1.
apply Mdis.
split.
apply MH2.
trivial.
eapply nextS_movwld.
reflexivity.
apply H1.
reflexivity.
Qed.

Lemma disjUnion_mapto : forall M1 M2 M l x, 
  MemMod.disjUnion M1 M2 M ->  MemMod.maps_to M1 l x -> MemMod.maps_to M l x.
Proof.
(** added by wang **)
unfold MemMod.maps_to.
intros.
destruct H.
unfold MemMod.merge in H1.
rewrite H1.
rewrite H0.
trivial.
Qed.

Lemma disjUnion_leadsToxW : forall M1 M2 M l , MemMod.disjUnion M1 M2 M ->(l |~>w_) M1 -> (l |~>w_) M.
Proof.
(** added by wang **)
intros.
destruct H0 as [M3 [M4 [Mdis [MH MH1]]]].
exists M3. exists (MemMod.merge M4 M2).
split.
eapply disjUnion_assoc_lr.
apply Mdis.
apply H.
split;[apply MH | trivial].
Qed.


Lemma disjUnion_leadsToW : forall M1 M2 M x l , MemMod.disjUnion M1 M2 M ->(l |~>w x) M1 -> (l |~>w x) M.
Proof.
(** added by wang **)
intros.
destruct H0 as [M3 [M4 [Mdis [MH MH1]]]].
exists M3. exists (MemMod.merge M4 M2).
split.
eapply disjUnion_assoc_lr.
apply Mdis.
apply H.
split;[apply MH | trivial].
Qed.

Lemma frame_movwld:
  forall f a r, frame (nextS f (comml (_movwld a r))).
Proof.
unfold frame.
intros.
exists M1.
destruct S as [[[M R] F] b].
inversion H.
subst.
unfold S_Mem_Upd.
simpl.
simpl in H0. 
inversion H1.
unfold S_Mem_Upd in H2.
simpl in H2.
inversion H2.
subst.
split.
apply H0.
destruct H13 as [M3 [M4 [Mdis [MH MH1]]]].
eapply nextS_movwld.
reflexivity.
exists M3. exists M4.
split;[apply Mdis |split;[apply MH | trivial]].
destruct MH as [M5 [M6 [Mdis1 [MH2 MH3]]]]. 
destruct MH2.

assert (MemMod.maps_to M3 (addr_eval R a) x1).
eapply disjUnion_mapto. apply Mdis1. apply H3.
assert (MemMod.maps_to M1 (addr_eval R a) x1).
eapply disjUnion_mapto. apply Mdis. apply H5.
assert (MemMod.maps_to M (addr_eval R a) x1).
eapply disjUnion_mapto. apply H0. apply H6.

destruct H11 as [M7 [M8 [Mdis2 [MH4 MH5]]]].
destruct MH4 as [M9 [M0 [Mdis3 [MH6 MH7]]]].

assert (MemMod.maps_to M7 (addr_eval R a) x).
eapply disjUnion_mapto. apply Mdis3. destruct MH6. apply H8.
assert (MemMod.maps_to M (addr_eval R a) x).
eapply disjUnion_mapto. apply Mdis2. apply H8.

unfold MemMod.maps_to in H7.
unfold MemMod.maps_to in H9.
rewrite H7 in H9.
inversion H9.
trivial.
Qed.

(* movwst *)
Lemma enable_movwst:
  forall (p : Pred) o a f, 
    (forall S, p S -> leadsToxW (addr_eval (S_RF S) a) (S_Mem S)) -> enable p (nextS f (comml (_movwst o a))).
Proof.
unfold enable.
intros.
destruct S as [[[M R] F] b].
exists (MemMod.put (MemMod.put M (addr_eval R a)(opr_eval R o)) (addr_eval R a + 1) f, R, F, b).
eapply nextS_movwst.
reflexivity.
reflexivity.
apply (H (M, R, F, b));apply H0.
exists f.
trivial.
Qed.

Lemma safe_mono_movwst:
  forall f o a, safe_mono (nextS f (comml (_movwst o a))).
Proof.
unfold safe_mono. unfold enable.
intros.
destruct S as [[[M R] F] b].
exists (MemMod.put (MemMod.put M (addr_eval R a)(opr_eval R o)) (addr_eval R a + 1) f, R, F, b).
eapply nextS_movwst.
reflexivity.
reflexivity.
destruct H0 as [M1 [M2 [Mdis [MH MH1]]]].
assert (exists S' : State, nextS f (comml (_movwst o a)) (M1, R, F, b) S').
apply H.
unfold S_Mem_Upd in MH.
simpl in MH.
apply MH.
destruct H0 as [S' H0].
inversion H0.
subst.
simpl in Mdis.
eapply disjUnion_leadsToxW.
apply Mdis.
apply H11.
exists f.
trivial.
Qed.

Lemma disjoint_put_disjoint : forall l x M1 M2,
  MemMod.in_dom l M1
  -> MemMod.disjoint M1 M2
  -> MemMod.disjoint (MemMod.put M1 l x) M2.
(** copy from Assertion_Lemma3 **)
Proof.
  unfold MemMod.in_dom, MemMod.disjoint, MemMod.put, MemMod.not_in_dom.
  intros.
  destruct H.
  destruct (MemMod.beq_A x0 l).
    subst x0.
    destruct (H0 l).
      rewrite H in H1.
      discriminate H1.
    right; trivial.
  destruct (H0 x0).
    left; trivial.
  right; trivial.
Qed.

Lemma leadsToxW_indom : forall M l ,leadsToxW l M
  -> (MemMod.in_dom l M /\ MemMod.in_dom (l + 1) M).
(** added by wang **)
Proof.
intros.
destruct H as [M1 [M2 [Mdis [MH MH1]]]].
destruct MH as [M3 [M4 [Mdis1 [MH2 MH3]]]].
destruct MH2.
assert (ptoW l x M1).
exists M3. exists M4.
auto.
assert (MemMod.in_dom l M1 /\ MemMod.in_dom (l + 1) M1).
eapply ptoW_indom.
apply H0.
elim H1.
intros.
split.
destruct H2. exists x0.
assert (MemMod.maps_to M l x0).
eapply disjUnion_mapto.
apply Mdis. apply H2. apply H4.
destruct H3.
exists x0.
assert (MemMod.maps_to M (l + 1) x0).
eapply disjUnion_mapto.
apply Mdis. apply H3. apply H4.
Qed.

Lemma leadsToW_leadsToxW :
  forall M l x,
    leadsToW l x M
    -> leadsToxW l M.
Proof.
(** copy from Assertion_Lemma3 **)
  unfold leadsToW, leadsToxW.
  unfold ptoW, ptoxW.
  unfold imprecM.
  unfold mstar, pto, ptox, mTrue.
  intros.
  destruct H as [M1 [M2 [HJM [H _]]]].
  destruct H as [M11 [M12 [HJM1 [Hpto1 Hpto2]]]].
  exists M1; exists M2.
  split_l.
  split_r.
  exists M11; exists M12.
  split_l.
  split.
    exists x; trivial.
  trivial.
Qed.

Lemma indom_indom_put : forall l l' x' M,
  MemMod.in_dom l M
  -> MemMod.in_dom l (MemMod.put M l' x').
(** copy from Assertion_Lemma3 **)
Proof.
  unfold MemMod.in_dom, MemMod.put.
  intros.
  destruct H.
  destruct (MemMod.beq_A l l').
    exists x'; trivial.
  exists x; trivial.
Qed.


Lemma frame_movwst:
  forall f o a, frame (nextS f (comml (_movwst o a))).
Proof.
unfold frame.
unfold S_Mem_Upd.
unfold enableS.
intros.
destruct S as [[[M R] F] b].
simpl in H1.
inversion H. 
subst.
simpl.
simpl in H0.
destruct H1.
inversion H1.
subst.
destruct H13.
exists (MemMod.put (MemMod.put M1 (addr_eval R a) (opr_eval R o))
         (addr_eval R a + 1) x).
split.
assert (MemMod.in_dom (addr_eval R a) M1 
  /\ MemMod.in_dom (addr_eval R a +1) M1).
eapply leadsToxW_indom.
apply H14.
split.
inversion H3.
assert (MemMod.disjoint (MemMod.put M1 (addr_eval R a) (opr_eval R o)) M2).
apply disjoint_put_disjoint. apply H4.
destruct H0.
apply H0.
apply disjoint_put_disjoint. apply indom_indom_put.
apply H5. apply H6. 
rewrite H2.
destruct H0.
rewrite H4.
assert (MemMod.put (MemMod.merge M1 M2) (addr_eval R a) (opr_eval R o) 
  = MemMod.merge (MemMod.put M1  (addr_eval R a) (opr_eval R o) ) M2).
apply MemMod.put_merge_merge_put.
rewrite H5.
apply MemMod.put_merge_merge_put.
eapply nextS_movwst.
reflexivity. reflexivity. apply H14.
exists x. trivial.
Qed.

(* addw *)
Lemma trivial_enable_addw:
forall p f o r, enable p (nextS f (comml (_addw o r))).
Proof.
unfold enable.
intros.
destruct S as [[[M R] F] b].
exists (M, RF_update R r ((opr_eval R o) + R r),Set_Flag F _ZF, b).
eapply nextS_addw.
reflexivity. reflexivity. reflexivity.
left. auto.
Qed.

Lemma frame_addw:
  forall f o r, frame (nextS f (comml (_addw o r))).
Proof.
unfold frame.
intros.
destruct S as [[[M R] F] b].
unfold S_Mem_Upd in H1.
unfold S_Mem_Upd.
inversion H.
subst.
exists M1. 
split. apply H0.
eapply nextS_addw.
reflexivity. reflexivity. reflexivity.
apply H13.
Qed.

(* cmpw *)
Lemma trivial_enable_cmpw:
forall p f o r, enable p (nextS f (comml (_cmpw o r))).
Proof.
unfold enable.
intros.
destruct S as [[[M R] F] b].
assert ({opr_eval R o = R r} + {opr_eval R o <> R r}).
eapply MemMod.beq_A.
elim H0.
intro.
exists (M, R, Set_Flag F  _ZF, b).
eapply nextS_cmpw.
auto.
intro.
exists (M, R, Clear_Flag F  _ZF, b).
eapply nextS_cmpw.
auto.
Qed.

Lemma frame_cmpw:
  forall f o r, frame (nextS f (comml (_cmpw o r))).
Proof.
unfold frame.
intros.
destruct S as [[[M R] F] b].
inversion H.
subst.
exists M1.
unfold S_Mem_Upd.
simpl in H0. simpl.
split. apply H0.
eapply nextS_cmpw.
apply H10.
Qed.
(* pushwp *)

Lemma leadsToBlk_indom : forall l size M off, leadsToBlk l size M -> off < size -> MemMod.in_dom (l + off) M.
Proof.
(** added by wanghaibo **)
intros.
assert (leadsToBlk (l + off)(size - off) M).
assert (off <= size).
omega.
apply leadsToBlk_shrink_low;assumption.
destruct H1 as [M1 [M2 [SH [SH1 SH2]]]].
unfold ptoxBlk in SH1.
assert (size - off = S (size - off - 1)).
omega.
rewrite H1 in SH1.
destruct SH1 as [M3 [SH3 [SH4 [[n [SH5 SH5']] SH6]]]].
assert (MemMod.in_dom (l + off) M3).
apply (MemMod.mapsto_indom SH5).
destruct SH;destruct SH4.
assert (MemMod.in_dom (l + off) M1).
subst.
apply MemMod.merge_presv_indom_left;assumption.
rewrite H4;apply MemMod.merge_presv_indom_left;assumption.
Qed.

Lemma frame_pushwp:
  forall f o, frame (nextS f (comml (_pushwp o))).
Proof.
unfold frame. unfold S_Mem_Upd.
intros.
destruct S as [[[M R] F] b].
inversion H. subst.
simpl in H0. simpl. simpl in H1.
destruct H10 as [MH [MH1 MH2]].
destruct MH1 as [x MH3].
exists (MemMod.put (MemMod.put M1 (R SP - 2) (opr_eval R o)) (R SP - 2 + 1) x).
rewrite MH3.
inversion H1. inversion H2. subst.
destruct H11. destruct H3.
split.
split.
assert (MemMod.in_dom (R SP - 2 + 1) M1).
eapply leadsToBlk_indom.
apply H5.
omega.
apply disjoint_put_disjoint. apply indom_indom_put.
apply H6.
assert (MemMod.in_dom (R SP - 2 + 0) M1).
eapply leadsToBlk_indom.
apply H5. omega.
apply disjoint_put_disjoint.
assert (R SP -2 + 0 = R SP -2).
omega.
rewrite <- H8.
apply H7.
destruct H0.
apply H0.
destruct H0.
rewrite H6.
assert (MemMod.put (MemMod.merge M1 M2) (R SP - 2) (opr_eval R o) 
  = MemMod.merge (MemMod.put M1  (R SP -2) (opr_eval R o) ) M2).
apply MemMod.put_merge_merge_put.
rewrite H7.
apply MemMod.put_merge_merge_put.
eapply nextS_pushwp.
reflexivity.
unfold pushW.
split.
split;[apply H3 | apply H5].
split;[exists x;trivial | trivial].
Qed.


Lemma leadsToBlk_Mono : forall size l M1 M2 M,
  MemMod.disjUnion M1 M2 M
  -> leadsToBlk l size M1
  -> leadsToBlk l size M.
(** copy from Assertion_Lemmas3 **)
Proof.
  unfold leadsToBlk.
  unfold imprecM, mstar, mTrue.
  intros.
  destruct H0 as [M11 [M12 [HJM1 [Hblk Ht]]]].
  exists M11; exists (MemMod.merge M12 M2).
  rsplit.
  split_lr.
  destruct H.
  destruct HJM1. 
  subst.
  split.
    apply MemMod.disj_merge_A.
    trivial.
    eapply MemMod.disj_merge_disj_left; eauto.
  rewrite MemMod.merge_assoc.
  trivial.
Qed.

Lemma safe_mono_pushwp:
  forall f o, safe_mono (nextS f (comml (_pushwp o))).
Proof.
unfold safe_mono. unfold enable.
intros.
destruct S as [[[M R] F] b].
exists (MemMod.put (MemMod.put M (R SP - 2) (opr_eval R o)) (R SP - 2 + 1) f, RF_update R SP (R SP - 2), F, b).
eapply nextS_pushwp.
reflexivity.
destruct H0 as [M1 [M2 [Mdis [MH MH1]]]].
unfold pushW.
unfold S_Mem_Upd in MH.
simpl in MH.
simpl in Mdis.
split.
assert (exists S' : State, nextS f (comml (_pushwp o)) (M1, R, F, b) S').
apply H.
apply MH.
destruct H0 as [S' H0].
inversion H0.
subst.
inversion H9.
destruct H1.
split.
apply H1.
eapply leadsToBlk_Mono.
apply Mdis. apply H3.
unfold MemUpd16.
split;[exists f;reflexivity | reflexivity].
Qed.


Lemma enable_pushwp:
  forall (p : Pred) n f opr,
    (forall S, p S -> stkSpace (S_RF S SP) n (S_Mem S))
    -> n >= 2
    -> enable p (nextS f (comml (_pushwp opr))).
Proof.
  intros.
  unfold enable.
  intros.
  induction S as (((M, R), fl), isr).
  exists (MemMod.put (MemMod.put M (R SP - 2) (opr_eval R opr)) (R SP - 2 + 1) 0, RF_update R SP (R SP - 2), fl, isr).
  apply nextS_pushwp with (opr_eval R opr); auto.
  unfold pushW.
  split.
    apply stkSpace_rev_mono with n; auto.
    assert (stkSpace (S_RF (M, R, fl, isr) SP) n (S_Mem (M, R, fl, isr))).
      apply H; auto.
    UnfoldStateAll. auto.
  split; auto.
  unfold MemUpd16.
  exists 0. auto.
Qed.


(* pushwm *)
Lemma safe_mono_pushwm:
  forall f addr, safe_mono (nextS f (comml (_pushwm addr))).
Proof.
unfold safe_mono. unfold enable.
intros.
destruct H0 as [M1 [M2 [Mdis [MH MH1]]]].
destruct S as [[[M R] F] b].
simpl in Mdis.
unfold S_Mem_Upd in MH.
simpl in MH.
assert (exists S' : State, nextS f (comml (_pushwm addr)) (M1, R, F, b) S').
apply H;apply MH.
destruct H0 as [S' H0].
inversion H0.
subst.
inversion H9.
destruct H1 as [M3 [Mdis1 [MH2 MH3]]].
inversion MH2 as [M4 [M5 [Mdis2 [MH4 MH5]]]].
destruct MH5.
exists (MemMod.put (MemMod.put M (R SP - 2) x) (R SP - 2 + 1) x1, RF_update R SP (R SP - 2), F, b).
eapply nextS_pushwm.
reflexivity.
exists x0.
exists (MemMod.merge M3 M2).
split.
eapply disjUnion_assoc_lr.
apply Mdis1. apply Mdis. split.
apply MH2.
trivial.
unfold pushW.
split.
inversion H10. inversion H2.
split.
apply H4.
eapply leadsToBlk_Mono.
apply Mdis. apply H5.
split;[exists x1;reflexivity | reflexivity].
Qed.


Lemma frame_pushwm:
  forall f addr, frame (nextS f (comml (_pushwm addr))).
Proof.
unfold frame. unfold S_Mem_Upd. unfold enableS.
intros.
destruct S as [[[M R] F] b].
simpl in H0,H1.
inversion H.
subst.
destruct H1 as [S' MH].
inversion MH.
subst.
destruct H11 as [H1 [H2 H3]].
destruct H12 as [MH1 [MH2 MH3]].
assert ((addr_eval R addr |~>w x0) M).
eapply disjUnion_leadsToW. apply H0. apply H9.
assert (x = x0).
eapply (leadsToW_unique  H10 H4).
inversion H2.
rewrite H6.
exists (MemMod.put (MemMod.put M1 (R SP - 2) x)(R SP - 2 +1) x1).
simpl.
split.
destruct H0 as [H0 H0']. 
split.
inversion MH1.
inversion H8.

apply disjoint_put_disjoint.
assert (MemMod.in_dom (R SP - 2 + 1) M1).
eapply leadsToBlk_indom. apply H8.
omega.
apply indom_indom_put. apply H12.
apply disjoint_put_disjoint.
assert (R SP - 2 = R SP - 2 + 0).
omega.
rewrite H12.
eapply leadsToBlk_indom. apply H8.
omega. apply H0.
assert (MemMod.put (MemMod.merge M1 M2) (R SP - 2) x 
  = MemMod.merge (MemMod.put M1  (R SP -2) x ) M2).
apply MemMod.put_merge_merge_put.
rewrite H0'.
rewrite H7.
apply MemMod.put_merge_merge_put.
eapply nextS_pushwm. reflexivity. apply H9.
split.
apply MH1.
rewrite H5.
split.
exists x1. reflexivity.
apply H3.
Qed.


Lemma enable_pushwm:
  forall (p : Pred) n f addr,
    (forall S, p S -> stkSpace (S_RF S SP) n (S_Mem S)
      /\ leadsToxW (addr_eval (S_RF S) addr) (S_Mem S))
    -> n >= 2
    -> enable p (nextS f (comml (_pushwm addr))).
Proof.
  intros.
  unfold enable.
  intros.
  destruct S as [[[M R] fl] isr].
  substH H with (H _ H1).
  simplS in * .
  destruct H as [H2 H3].
  substH H3 with (leadsToxW_leadsToW H3).
  destruct H3 as [x H3].
  exists (MemMod.put (MemMod.put M (R SP - 2) x) (R SP - 2 + 1) 0, RF_update R SP (R SP - 2), fl, isr).
  apply nextS_pushwm with (addr_eval R addr) x; auto.
  unfold pushW.
  split.
    apply stkSpace_rev_mono with n; auto.
  split; auto.
  unfold MemUpd16.
  exists 0. auto.
Qed.

(* popw *)
Lemma enable_popw:
  forall (p : Pred) f r,
    (forall S, p S -> ((S_RF S SP) |~>w_ ) (S_Mem S))
    -> enable p (nextS f (comml (_popw r))).
Proof.
unfold enable.
intros.
destruct S as [[[M R] F] b].
assert ((S_RF (M, R, F, b) SP |~>w_) (S_Mem (M, R, F, b))).
apply H. apply H0.
simpl in H1.
inversion H1 as [M1 [M2 [Mdis [MH1 MH2]]]].
destruct MH1 as [M3 [M4 [Mdis1 [MH3 MH4]]]].
destruct MH3.
exists (M, RF_update (RF_update R SP (R SP + 2)) r x, F, b).
eapply nextS_popw.
split.
assert ((R SP |~>w x) M).
eapply disjUnion_leadsToW.
apply Mdis.
exists (MemMod.merge M3 M4).
exists MemMod.emp.
inversion Mdis1.
rewrite <- H4.
split.
apply MemMod.unit_emp_disjUnion.
split.
exists M3. exists M4.
split;[apply Mdis1 | split; [apply H2 | apply MH4]].
trivial.
apply H3.
reflexivity.
reflexivity.
Qed.


Lemma safe_mono_popw:
  forall f r, safe_mono (nextS f (comml (_popw r))).
Proof.
unfold safe_mono. unfold enable.
intros.
destruct S as [[[M R] F] b].
destruct H0 as [M1 [M2 [Mdis [MH1 MH2]]]].
unfold S_Mem_Upd in MH1.
simpl in MH1, Mdis.
assert (exists S' : State, nextS f (comml (_popw r)) (M1, R, F, b) S').
apply H. apply MH1.
destruct H0.
inversion H0.
subst.
inversion H8.
exists (M, RF_update (RF_update R SP (R SP + 2)) r x0, F, b).
eapply nextS_popw.
split.
assert ((R SP |~>w x0)M).
eapply disjUnion_leadsToW. apply Mdis. apply H1.
apply H3.
reflexivity.
reflexivity.
Qed.

Lemma frame_popw:
  forall f r, frame (nextS f (comml (_popw r))).
Proof.
unfold frame. unfold enableS. unfold S_Mem_Upd.
intros.
destruct S as [[[M R] F] b].
simpl in H0,H1. 
inversion H.
subst.
exists M1.
simpl.
split. apply H0.
destruct H1.
inversion H1.
subst.
inversion H10.
eapply nextS_popw.
split.
apply H2.
reflexivity.
inversion H9.
rewrite H5.
assert ((R SP |~>w x1) M).
eapply disjUnion_leadsToW.
apply H0.
apply H2.
assert (x = x1).
eapply leadsToW_unique.
apply H4. apply H6.
rewrite H7.
reflexivity.
Qed.

Lemma frame_call:
  forall f f', frame (nextS f (comml (_call f'))).
Proof.
unfold frame. unfold enableS.
intros.
destruct S as [[[M R] F] b].
destruct S' as [[[M' R'] F'] b'].
UnfoldStateAll.
destruct H1 as [S H1].
apply nextS_DMono with M S; auto.
Qed.
