Require Import Coqlib.
Require Import Arith.
Require Import Machine.

Lemma beq_sym : forall r r', beq_reg r r' = beq_reg r' r.
Proof.
  double induction r r'; simpl; intros; auto.
Qed.

Lemma beq_refl : forall r , beq_reg r r = true.
Proof.
  induction r; simpl; intros; auto.
Qed.

Lemma beq_req_true : forall r r', r = r' -> beq_reg r r' = true.
  intros.
  rewrite H.
  induction r';auto.
Qed.

Lemma beq_true_req : forall r r', beq_reg r r' = true -> r = r'.
Proof.
  double induction r r'; simpl; intros; discriminate || trivial.
Qed.

Lemma beq_rneq_false : forall r r', r <> r' -> beq_reg r r' = false.
  intros.
  induction r;
    induction r';
      tauto.
Qed.

Lemma beq_reg_dec : forall r r', 
  {beq_reg r r' = true} + {beq_reg r r' = false}.
Proof.
  double induction r r'; simpl; intros; auto.
Qed.

Ltac simplbreg :=
  match goal with 
    
    | H : beq_reg ?r ?r = false |- _ =>
        rewrite (beq_req_true r r (refl_equal r)) in H;
          discriminate H; 
            simplbreg
        
    | H : context [ beq_reg ?r ?r] |- _ =>
        rewrite (beq_req_true r r (refl_equal r)) in H;
          simplbreg
   
    | |- context [ beq_reg ?r ?r] =>
        rewrite (beq_req_true r r (refl_equal r));
          simplbreg

    | H : beq_reg ?rx ?ry = ?f
      |- context [beq_reg ?rx ?ry] => 
        rewrite H; simplbreg
    | H : beq_reg ?rx ?ry = ?f
      |- context [beq_reg ?ry ?rx] => 
        rewrite (beq_sym rx ry) in H; 
          rewrite H; 
           simplbreg

    | H : beq_reg ?rx ?ry = ?f,
      H1 : context [beq_reg ?rx ?ry] |- _ =>
        rewrite H in H1; simplbreg
    | H : beq_reg ?rx ?ry = ?f,
      H1 : context [beq_reg ?ry ?rx] |- _ =>
        rewrite (beq_sym rx ry) in H; 
          rewrite H in H1;
            simplbreg

    | H : beq_reg ?rx ?ry = true
      |- _ =>
        let Heq := fresh "Heq" in
        (generalize (beq_true_req rx ry H); 
          intro Heq; subst rx); simplbreg

    | [H : ?x = ?x |- _ ] => clear H; simplbreg
    | [H : true = false |- _ ] => discriminate H
    | [H : false = true |- _ ] => discriminate H
    | _ => idtac

  end.

Ltac solvebreg :=
  intros; 
  match goal with 
    | |- context [ beq_reg ?r ?r] =>
        rewrite (beq_req_true r r (refl_equal r));
          reflexivity

    | H : beq_reg ?rx ?ry = ?f 
      |- beq_reg ?rx ?ry = ?f =>
        assumption

    | H : beq_reg ?rx ?ry = ?f 
      |- beq_reg ?ry ?rx = ?f =>
        rewrite (beq_sym rx ry) in H;  assumption
    
    | H : beq_reg ?rx ?ry = true
      |- _ =>
        let Heq := fresh "Heq" in
        (generalize (beq_true_req rx ry H); 
          intro Heq; subst rx);  solvebreg
        
    | H : beq_reg ?r ?r = false
      |- _ =>
        rewrite (beq_req_true r r (refl_equal r)) in H;
          discriminate H
            
    | |- beq_reg ?rx ?ry = ?f =>
        simpl; reflexivity

    | _ => fail 1 "The goal cannot be solved. "
end.

Ltac decbeqreg rx ry :=
  let Hb := fresh "Hb" in
  (destruct (beq_reg_dec rx ry) as [Hb | Hb]; simplbreg).

(********************)

Lemma RF_update_RF_update : forall R r v v',
  RF_update (RF_update R r v') r v = RF_update R r v.
Proof.
intros.
apply (extensionality reg nat).
intro r0.
unfold RF_update.
generalize (beq_reg r0 r).
induction b;reflexivity.
Qed.

Lemma RF_update_RF : forall R r v,
  RF_update R r v r = v.
Proof.
unfold RF_update.
intros.
replace (beq_reg r r) with true.
reflexivity.
apply beq_reg_refl.
Qed.

Lemma RF_update_prev : forall R r,
  RF_update R r (R r) = R.
Proof.
intros.
apply (extensionality reg nat).
intro r0;unfold RF_update.
destruct (reg_eq_dec r0 r).
subst.
replace (beq_reg r r) with true;
[reflexivity|apply beq_reg_refl].
replace (beq_reg r0 r) with false;
[reflexivity|apply flag_neq_beq_false;assumption ].
Qed.

Lemma RF_update_RF_neq : forall R r r' v,
  beq_reg r' r = false ->
  RF_update R r v r' = R r'.
Proof.
unfold RF_update.
intros.
rewrite H.
trivial.
Qed.

Lemma set_get_eq : 
  forall R r w, (RF_update R r w) r = w.
Proof.
  intros.
  unfold RF_update.
  rewrite (beq_req_true r r (refl_equal _)).
  trivial.
Qed.

Lemma set_get_neq : forall R r r' w,
  beq_reg r r' = false 
  -> (RF_update R r' w) r 
  = R r.
Proof.
  intros.
  unfold RF_update.
  decbeqreg r r'; trivial.
Qed.

Lemma set_set_eq : forall R r w w',
  RF_update (RF_update R r w') r w
  = RF_update R r w.
Proof.
  intros.
  unfold RF_update.
  apply (extensionality reg nat).
  intro r'.
  decbeqreg r' r; trivial.
Qed.

Ltac reg_upd_simpl :=
  match goal with

    | [ |- context [(RF_update ?R ?r ?v) ?r] ] =>
      rewrite (set_get_eq R r v); reg_upd_simpl

    | [ |- context [(RF_update ?R ?r ?v) ?r] ] =>
      rewrite set_get_eq; reg_upd_simpl

    | [ H : context [(RF_update ?R ?r ?v) ?r] |- _ ] =>
      rewrite set_get_eq in H; reg_upd_simpl

    | [ |- context [(RF_update ?R ?r2 ?v) ?r1] ] =>
      rewrite set_get_neq; [idtac | solvebreg]; reg_upd_simpl

    | [ H: context [(RF_update ?R ?r2 ?v) ?r1] |- _ ] =>
      rewrite set_get_neq in H; [idtac | solvebreg]; reg_upd_simpl

    | [ |- context [(RF_update (RF_update ?s ?r2 ?v2) ?r1 ?v1)] ] =>
      rewrite set_set_eq; reg_upd_simpl

    | [ H: context [(RF_update (RF_update ?s ?r2 ?v2) ?r1 ?v1)] |- _ ] =>
      rewrite set_set_eq in H; reg_upd_simpl

    | _ => idtac
  end.

Tactic Notation "reg" "simpl" := reg_upd_simpl.
