Add LoadPath "..".
Add LoadPath "../CAP0".

Require Import Arith.
Require Import Coqlib.
Require Import Map.
Require Import Machine.
Require Import Assertion.
Require Import Assertion_Lemmas.
Require Import Assertion_Lemmas4.
Require Import Operational_Semantics.
Require Import OpSem_Lemmas.
Require Import OpSem_Lemmas2.
Require Import Reg.
Require Import CAP0.
Require Import SCAP.
Require Import AIM.
Require Import QLib.
Require Import SCAP_READY.
Require Import SCAP_IDLE.
Require Import SCAP_Itr.
Require Import Linking.
Require Import Linking_Lemmas.
Require Import Linking_Lemmas2.
Require Import Linking_Lemmas3.


Require Import EqNat.
Require Import Omega.
Require Import List.
Require Import Arith_Lemmas.


Lemma imprec_presv_ih_enable_sw:
  forall p, ih_enable_sw p -> ih_enable_sw (imprec p).
Proof.
  intro p.
  unfold ih_enable_sw.
  intro.
  set (Fm1 := (fun S0 => (INV0 *^* stkSpacePrec (S_RF S0 SP) (_SW_Stk + 2)))).
  set (p0 := (fun S0 => (pTrue *<* (Fm1 S0)) S0)).
  assert (monotonic_p p0).

  unfold monotonic_p.
  intros.
  induction S as (((M0, R0), fl0), isr0). unfold p0. 
  unfold p0 in H0. simplS. unfold Fm1 in *. simplS in H0.
  unfold S_Mem in H1.
  assert (exists M0', MemMod.disjUnion M0 M0' M).
  apply MemMod.subseteq_split; auto.
  unfold pmstar in H0. simplS in H0. InvertAll.
  unfold pmstar. simplS. exists (MemMod.merge x0 x). exists x1.
  unfold pTrue.
  split; auto.
  cut (MemMod.disjUnion x1 (MemMod.merge x0 x) M). auto with PMap.
  apply MemMod.disjUnion_assoc_left_right; auto.
  inversion_clear H2; auto with PMap.
  assert (MemMod.disjUnion x1 x0 M0). auto with PMap.
  inversion H4. rewrite H7 in H3. auto.

  set (q := (fun S0 => S_ISR S0 = false)).
  assert (monotonic_p q). unfold q. auto.
  set (q' := (fun S => p0 S /\ q S)).
  assert (forall S0 : State, imprec p S0 -> q' S0).
  assert (monotonic_p q').
  unfold q'. apply monotonic_conj; auto.
  apply (imprec_presv_imp_monotonic H2 H); auto.
  auto.
Qed.
Hint Resolve imprec_presv_ih_enable_sw.


Lemma IH_SW_progress:
  forall psi pc theta Psi a S,
    IH_WFcmd psi pc theta (comml (_call _SWITCH))
    -> a = Itp_AIM rho_ih theta
    -> a Psi S
    -> exists S', nextS pc (comml (_call _SWITCH)) S S'.
Proof.
  intros.
  induction theta.
  assert (imprec p S).
  rewrite H0 in H1.
  apply Itp_AIM_imprecp with g n Psi rho_ih; auto.
  unfold IH_WFcmd in H.
  inversion_clear H.
  inversion_clear H3.
  clear H.
  unfold IH_Call_SW in H4.
  InvertAll.
  assert (ih_enable_sw (imprec p)).
  auto.
  unfold ih_enable_sw in H6.
  assert ((pTrue *<* (INV0 *^* stkSpacePrec (S_RF S SP) (_SW_Stk + 2))) S /\
       S_ISR S = false).
  apply H6; auto.
  inversion_clear H8.
  apply stkSpace_call_prog with (_SW_Stk + 2).
  assert (((INV0 *^* stkSpacePrec (S_RF S SP) (_SW_Stk + 2)) *^* mTrue) (S_Mem S)).
  apply pTrue_m_mTrue with pTrue; auto.
  apply mstar_mTrue with INV0 mTrue; auto.
  omega.
  inversion H3.
  absurd (prim_entry _SWITCH); auto. unfold prim_entry. left. auto.
Qed.

Lemma enable_sw_nxt_stkspace_ih:
forall p pc S S' M1 M2,
  ih_enable_sw p
  -> nextS pc (comml (_call _SWITCH)) S S'
  -> MemMod.disjUnion M1 M2 (S_Mem S)
  -> imprec p (S_Mem_Upd S M1)
  -> exists M1',
        MemMod.disjUnion M1' M2 (S_Mem S')
        /\ (S_RF S' SP |->w (pc+1) *^* stkSpacePrec (S_RF S' SP) 8 *^* mTrue) M1'.
Proof.
  intros.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  UnfoldStateAll.
  unfold ih_enable_sw in H.
  assert ((mTrue *^*
             (INV0 *^* stkSpacePrec (R SP) (_SW_Stk + 2))) M1).
  unfold imprec in H2.
  unfold pmstar in H2.
  InvertAll.
  simplS in H3. simplS in H2.
  assert ((pTrue *<* (INV0 *^* stkSpacePrec (S_RF (x, R, fl, isr) SP) (_SW_Stk + 2)))
              (x, R, fl, isr) /\ S_ISR (x, R, fl, isr) = false).
  apply H; auto.
  inversion_clear H4. clear H7. rename H6 into H4.
  unfold S_RF in H4.
  unfold pmstar in H4.
  InvertAll.
  exists (MemMod.merge x1 x0). exists x2. 
  split. unfold S_Mem in H6.
  assert (MemMod.disjUnion x2 (MemMod.merge x1 x0) M1).
  apply MemMod.disjUnion_assoc_left_right; auto.
  inversion_clear H6. auto with PMap.
  assert (MemMod.disjUnion x2 x1 x); auto with PMap.
  inversion_clear H7.
  rewrite <- H10. auto. auto with PMap.
  unfold mTrue. auto.
  clear H.
  assert (frame (nextS pc (comml (_call _SWITCH)))).
  apply frame_call.
  unfold frame in H.
  assert (exists M1', 
        MemMod.disjUnion M1' M2 (S_Mem (M', R', fl', isr')) /\
        nextS pc (comml (_call _SWITCH)) (S_Mem_Upd (M, R, fl, isr) M1) (S_Mem_Upd (M', R', fl', isr') M1')).
  apply H; auto.
  clear H.
  UnfoldStateAll.
  unfold enableS.
  exists (MemUpdW M1 (R SP - 2) (pc+1) 0, RF_update R SP (R SP - 2), fl, isr).
  apply nextS_call.
  unfold pushW.
  split.
  sep on.
  sep lift 2 in H3.
  sep stksplit 2 in H3. sep lift 0 in H3.
  sep trivial.
  split; auto.
  apply  MemUpdW_MemUpd16 with 0; auto.
  clear H.
  UnfoldStateAll.
  InvertAll.
  exists x.
  split_l.
  destruct (nextS_inversion H5).
  InvertAll.
  subst R'. reg simpl.
  substH H11 with (MemUpd16_MemUpdW H11).
  InvertAll.
  subst x.
  sep on.
  sep lift 0.
  sep upd.
  sep lift 2 in H3.
  sep stksplit in H3.
  sep lift 1 in H3.
  sep match_left H3; auto.
  assert (_SW_Stk + 2 - 2 = _SW_Stk). auto.
  rewrite H7 in H3.
  sep lift 2 in H3.
  sep match_left H3; auto.
Qed.


Lemma enter_SW_pre_ih:
forall psi pc theta Psi'0 S S',
  IH_WFcmd psi pc theta (comml (_call _SWITCH))
  -> Itp_AIM rho_ih theta Psi'0 S
  -> nextS pc (comml (_call _SWITCH)) S S'
  -> SW_Prd S'.
Proof.
  intros.
  unfold SW_Prd.
  induction theta.
  unfold Itp_AIM in H0.
  InvertAll. rename x into T. rename x0 into B. rename x1 into RdyQ. rename x2 into tid.
  exists T. exists RdyQ. exists tid.
  rename n into tag.
  unfold IH_WFcmd in H.
  inversion_clear H.
  inversion_clear H2. clear H.
  unfold IH_Call_SW in H3.
  InvertAll.
  clear H.
  induction S as (((M, R), fl), isr).
  exists (S_RF S' SP).
  unfold pmstar in H0.
  InvertAll.
  rename x1 into M1. rename x2 into M2.
  assert (exists tcb, TQMod.maps_to T tid tcb).
  apply WF_AIM_World_currTCB with Psi'0 (S_Mem_Upd (M, R, fl, isr) M1) B RdyQ rho_ih (Ast p g tag); auto.
  inversion_clear H5.
  rename x1 into tcb. exists tcb. exists (pc+1).
  split.

Lemma enter_SW_preMem_ih:
forall pc S S' p g tag M1 M2 Psi T B RdyQ tid,
  nextS pc (comml (_call _SWITCH)) S S'
  -> ih_enable_sw p
  -> MemMod.disjUnion M1 M2 (S_Mem S)
  -> WF_AIM_World Psi (S_Mem_Upd S M1) T B RdyQ tid
         rho_ih (Ast p g tag)
  -> (embed_curr_TCB T tid *^* (embed_RdyQ T RdyQ *^* embed_WQSet T B)) M2
  -> exists junk, (SW_MPre T RdyQ tid (S_RF S' SP) (pc+1) junk *^* mTrue) (S_Mem S').
Proof.
  intros.
  assert (imprec p (S_Mem_Upd S M1)).
  apply WF_AIM_World_imprecp with Psi T B RdyQ tid rho_ih g tag; auto.
  unfold SW_MPre.
  assert (exists M1',
        MemMod.disjUnion M1' M2 (S_Mem S')
        /\ (S_RF S' SP |->w (pc+1) *^* stkSpacePrec (S_RF S' SP) 8 *^* mTrue) M1').
  apply enable_sw_nxt_stkspace_ih with p S M1; auto.
  InvertAll.
  rename x into M1'.
  assert (exists junk, (S_RF S' SP |-> pc + 1 *^* S_RF S' SP + 1 |-> junk *^* stkSpacePrec (S_RF S' SP) 8 *^* mTrue) M1').
  sep on.
  sep lift 0 in H7.
  assert (exists junk,  
           M1' ||- S_RF S' SP |-> pc + 1 *^* S_RF S' SP + 1 |-> junk *^* (stkSpacePrec (S_RF S' SP) 8 *^* mTrue)).
  apply ptoW_elim1; auto.
  inversion H6. exists x. sep on.
  sep matchH H8 with 4.
  inversion_clear H6. clear H7. exists x.

  apply merge_SW_pre with M1' M2 B; auto.
Qed.

  apply enter_SW_preMem_ih with (M, R, fl, isr) p g tag M1 M2 Psi'0 B; auto.
  split; auto.
  split; auto.

  apply WF_AIM_World_RdyQinT with Psi'0 (S_Mem_Upd (M, R, fl, isr) M1) B tid rho_ih (Ast p g tag); auto.
  split; auto.
  split.
  apply WF_AIM_World_cthNotRdy with Psi'0 (S_Mem_Upd (M, R, fl, isr) M1) T B rho_ih (Ast p g tag); auto.
  apply WF_AIM_World_WFTQ with Psi'0 (S_Mem_Upd (M, R, fl, isr) M1) B RdyQ tid rho_ih (Ast p g tag); auto.
  inversion_clear H2.
  absurd (prim_entry _SWITCH); auto.
  unfold prim_entry. left. auto.
Qed.


Lemma enable_sw_stkspace_ih:
forall p S,
  ih_enable_sw p -> p S -> 
      exists M1, exists M2, 
        MemMod.disjUnion M1 M2 (S_Mem S) 
        /\ stkSpacePrec (S_RF S SP) (8 + 2) M1 
        /\ (INV0 *^* mTrue) M2.
Proof.
  intros.
  unfold ih_enable_sw in H.
  assert ((pTrue *<* (INV0 *^* stkSpacePrec (S_RF S SP) (_SW_Stk + 2))) S /\ S_ISR S = false).
  apply H; auto.
  inversion_clear H1. clear H3. rename H2 into H1.
  induction S as (((M, R), fl), isr).
  simplS. simplS in H1.
  assert ((mTrue *^* (INV0 *^* stkSpacePrec (R SP) (_SW_Stk + 2))) M).
  unfold pmstar in H1.
  InvertAll.
  exists x. exists x0. unfold mTrue.
  auto.
  sep on.
  sep lift 2 in H2.
  sep off.
  unfold mstar at 1 in H2.
  InvertAll.
  exists x. exists x0.
  split_l. split_l.
  sep on. sep matchH H5 with 2.
Qed.


Lemma decompose_World_ih:
  forall Psi T B RdyQ tid p g tag M1 R fl isr,
    WF_AIM_World Psi (M1, R, fl, isr) T B RdyQ tid rho_ih (Ast p g tag)
    -> exists TM, exists cthM, exists M12, exists cthM1, exists cthM2,
        MemMod.disjUnion M12 cthM M1
        /\ MemMod.disjUnion cthM1 cthM2 cthM
        /\ WF_currTID tid T B RdyQ rho_ih
        /\ WF_TWQ T B RdyQ
        /\ exists_Idle T RdyQ tid
        /\ WF_WQ_WQSpec B Delta
        /\ MemSplit (ThrdMemMod.remove TM tid) M12
        /\ ThrdMemMod.maps_to TM tid cthM
        /\ WF_Curr_Thrd Psi (S_Mem_Upd (M1, R, fl, isr) cthM) rho_ih (Ast p g tag) T tid
        /\ WF_ReadyQ Psi T RdyQ TM
        /\ WF_WQSet Psi T B TM
        /\ p (cthM1, R, fl, isr)
        /\ INV1 cthM2.
Proof.
  intros.
  unfold WF_AIM_World in H.
  InvertAll.
  rename x into TM. rename x0 into cthM. exists TM. exists cthM.
  assert (exists M12, MemMod.disjUnion M12 cthM M1 /\ MemSplit (ThrdMemMod.remove TM tid) M12).
  apply memsplit_thrd_remove; auto.
  InvertAll.
  rename x into M12. exists M12.
  assert (pmstar p INV1 (cthM, R, fl, isr)).
  unfold WF_Curr_Thrd in H5.
  assert (beq_nat rho_ih rho_ih = true). simpl. auto.
  rewrite H9 in H5. 
  InvertAll. unfold itp_IH_theta in H13.
  inversion_clear H13.
  inversion_clear H14.
  unfold lift_IH_PG in H13.
  InvertAll. UnfoldStateAll. auto.
  InvertAll. UnfoldStateAll. auto.
  unfold pmstar in H9.
  InvertAll.
  rename x into cthM1. rename x0 into cthM2.
  exists cthM1. exists cthM2.
  split; auto. split; auto. split; auto. split; auto.
  split; auto. split; auto. split; auto. split; auto.
Qed.

Lemma WFcallsw_enablesw_ih:
forall psi pc p g tag,
  IH_WFcmd psi pc (Ast p g tag) (comml (_call _SWITCH))
  -> ih_enable_sw p.
Proof.
  intros.
  unfold IH_WFcmd in H.
  inversion_clear H.
  unfold IH_Call_SW in H0.
  InvertAll. auto.
  absurd (prim_entry _SWITCH); auto.
  inversion H0. auto.
  unfold prim_entry. left. auto.
Qed.

Lemma call_sw_flags_false_ih:
forall p g tag psi Psi pc M M' R fl isr T tid,
  WF_Curr_Thrd Psi (M, R, fl, isr) rho_ih (Ast p g tag) T tid
  -> IH_WFcmd psi pc (Ast p g tag) (comml (_call _SWITCH))
  -> p (M', R, fl, isr)
  -> (fl _IF = false) /\ (isr = false).
Proof.
  intros.
  unfold IH_WFcmd in H0.
  inversion_clear H0.
  unfold IH_Call_SW in H2.
  InvertAll.
  unfold ih_enable_sw in H4.
  assert ((pTrue *<* (INV0 *^* stkSpacePrec (S_RF (M', R, fl, isr) SP) (_SW_Stk + 2))) (M', R, fl, isr) /\
       S_ISR (M', R, fl, isr) = false).
  apply H4; auto.
  inversion_clear H6. unfold S_ISR in H9. 
  split; auto.
  unfold WF_Curr_Thrd in H.
  assert (beq_nat rho_ih rho_ih = true). simpl. auto.
  rewrite H6 in H.
  InvertAll. 
  unfold itp_IH_theta in H12.
  inversion_clear H12.
  unfold S_Flag in H11. auto.
  inversion H2. absurd (prim_entry _SWITCH).
  auto. unfold prim_entry. left. auto.
Qed.

Lemma call_sw_nextPC_spec_ih:
forall psi pc p g tag,
    IH_WFcmd psi pc (Ast p g tag) (comml (_call _SWITCH))
    -> exists p', exists g',
         SCAP_lookupCT psi (pc+1) (Ast p' g' tag)
         /\ (forall pc', pimp (p |> (wrap_fun ih_gy pc _SWITCH pc')) p')
         /\ (forall pc', gimp (p >o (wrap_fun ih_gy pc _SWITCH pc') oo g') g).
Proof.
  intros.
  unfold IH_WFcmd in H. inversion_clear H.
  unfold IH_Call_SW in H0.
  InvertAll. exists x. exists x0.
  split; auto.
  inversion_clear H0.
  absurd (prim_entry _SWITCH); auto.
  unfold prim_entry. left. auto.
Qed.
Hint Resolve call_sw_nextPC_spec_ih.

Lemma psi_sw_lookupNPC_ih:
forall psi Psi Psi' Psi'0 p' g' tag pc,
  lift_psi psi rho_ih Psi
  -> Psi' = ext_Prim_Specs Psi
  -> subsetCT Psi' Psi'0
  -> SCAP_lookupCT psi pc (Ast p' g' tag)
  -> lookupCT Psi'0 pc (packS0 rho_ih (Ast p' g' tag)).
Proof.
  intros.
  subst Psi'.
  apply subsetCT_presv_loopup with (ext_Prim_Specs Psi); auto.
  unfold lift_psi in H.
  assert (lookupCT Psi pc (packS0 rho_ih (Ast p' g' tag))).
  apply H; auto.
  assert (subsetCT Psi (ext_Prim_Specs Psi)).
  unfold ext_Prim_Specs.
  apply Psi_extend_subsetCT.
  apply Psi_extend_subsetCT.
  apply Psi_extend_subsetCT.
  apply subsetCT_refl.
  apply subsetCT_presv_loopup with Psi; auto.
Qed.


Lemma enter_SW_grt_auxb0_ih:
forall cthM1 R fl isr M cthM11 cthM12 S' cthM11'' pc pc' M' R' fl' isr' junk S0,
  MemMod.disjUnion cthM11 cthM12 cthM1 
  -> (INV0 *^* mTrue) cthM12
  -> stkSpacePrec (R SP) (8 + 2) cthM11
  -> (S_RF S' SP - 2 |-> pc + 1 *^* S_RF S' SP - 2 + 1 |-> junk *^*
        stkSpacePrec (S_RF S' SP - 2) 8) cthM11''
  -> nextS pc (comml (_call _SWITCH)) (M, R, fl, isr) (M', R', fl', isr')
  -> nextS pc' (comms _retw) S0 S'
  -> S_RF S0 SP = R' SP
  -> MemMod.subseteq (MemMod.merge cthM12 cthM11'') (S_Mem S')
  -> isr = S_ISR S'
  -> fl _IF = S_Flag S' _IF
  -> MemMod.disjoint cthM11'' cthM12
  -> wrap_fun ih_gy pc _SWITCH pc' (cthM1, R, fl, isr)
      (S_Mem_Upd S' (MemMod.merge cthM12 cthM11'')).
Proof.
  intros.
  unfold wrap_fun.
  unfold conjG.
  exists (S_Mem_Upd S0 (MemMod.merge cthM12 cthM11'')).
  split.
  exists (MemMod.merge cthM12 (MemUpdW cthM11 (R SP - 2) (pc+1) junk), R', fl', isr').
  split.

  apply step_call_hyp1 with M M' _SW_Stk; auto.
  unfold ih_gy. exists (R' SP). split.
  assert (R SP - 2 = R' SP).
  apply enter_SW_grt_auxb0_hyp0 with pc M fl isr M' fl' isr' _SWITCH; auto.
  rewrite H10.
  assert (S_RF S' SP = S_RF S0 SP + 2).
  apply enter_SW_grt_auxb0_hyp1 with pc'; auto.
  rewrite H5 in H11.
Lemma enter_SW_grt_auxb0_hyp2_ih:
forall cthM12 cthM11 cthM11'' S' R' pc fl' isr' S0 sp junk,
  (INV0 *^* mTrue) cthM12
  -> stkSpacePrec sp (8 + 2) cthM11
  -> (S_RF S' SP - 2 |-> pc *^* S_RF S' SP - 2 + 1 |-> junk *^*
        stkSpacePrec (S_RF S' SP - 2) 8) cthM11''
  -> S_RF S' SP = (R' SP) + 2
  -> sp - 2 = (R' SP)
  -> MemMod.disjoint cthM11 cthM12
  -> MemMod.disjoint cthM11'' cthM12
  -> prepost (INV0 *^* stkSpacePrec (R' SP) _SW_Stk)
     (INV0 *^* stkSpacePrec (R' SP) _SW_Stk)
     (MemMod.merge cthM12 (MemUpdW cthM11 (R' SP) pc junk), R', fl',
     isr') (S_Mem_Upd S0 (MemMod.merge cthM12 cthM11'')).
Proof.
  unfold prepost.
  intros. induction S0 as (((M0, R0), fl0), isr0).
  simplS. clear M0 R0 fl0 isr0.
  destruct H as [cthM12_1 [chtM12_2 [HcthM12split [HcthM12_1 HcthM12_2]]]].
  set (cthM11' := MemUpdW cthM11 (R' SP) pc junk).
  assert (sp >= 2).
  apply ge_trans with (8 + 2); auto with arith.
  apply stkSpacePrec_sp_ge_sz with cthM11; auto.
  assert (sp = R' SP + 2). rewrite <- H3.
  assert (sp - 2 + 2 = sp). auto. auto.
  rewrite H6 in H0.
  assert ((R' SP |-> pc *^* R' SP + 1 |-> junk *^*
               stkSpacePrec (R' SP) 8) cthM11').
  unfold cthM11'.
  apply enter_SW_grt_auxb0_hyp21. auto.
  rewrite H2 in H1.
  assert (R' SP + 2 - 2 = R' SP). auto.
  rewrite H8 in H1.
  sep on.
  sep lift 2 in H7. sep lift 2 in H1.
  destruct H7 as [cthM11'_1 [cthM11'_2 [Hsplit11' [HM11'_1 HM11'_2]]]].
  destruct H1 as [cthM11''_1 [cthM11''_2 [Hsplit11'' [HM11''_1 HM11''_2]]]].
  exists (MemMod.merge cthM12_1 cthM11'_1).
  exists (MemMod.merge chtM12_2 cthM11'_2).
  exists (MemMod.merge cthM12_1 cthM11''_1).
  split.
  exists cthM12_1. exists cthM11'_1.
  split; auto.
  split; auto.
  apply enter_SW_grt_auxb0_hyp22 with cthM11 cthM12 cthM11' chtM12_2 cthM11'_2 (R' SP) pc junk; auto.
  split.
  exists cthM12_1. exists cthM11''_1.
  split; auto.
  split; auto.
  apply enter_SW_grt_auxb0_hyp23 with cthM12 cthM11'' chtM12_2 cthM11''_2; auto.

  split.
  apply enter_SW_grt_auxb0_hyp24 with cthM11 (R' SP) pc junk; auto.
  assert (cthM11'_2 = cthM11''_2).
  apply enter_SW_grt_auxb0_hyp26 with (R' SP) pc junk; auto.
  rewrite H1. apply enter_SW_grt_auxb0_hyp25; auto.
Qed.

  apply enter_SW_grt_auxb0_hyp2_ih with S' (R SP); auto.
  unfold MemMod.disjUnion in H. inversion H. auto.
  assert ((S_Flag S0 _IF = S_Flag S' _IF) /\ (S_ISR S0 = S_ISR S')).
  apply enter_SW_grt_auxb0_hyp3 with pc'; auto.
  inversion_clear H10.
  assert (S_Flag (M, R, fl, isr) _IF = S_Flag (M', R', fl', isr') _IF 
             /\ S_ISR (M, R, fl, isr) = S_ISR (M', R', fl', isr')).
  apply enter_SW_grt_auxb0_hyp4 with pc _SWITCH; auto.
  inversion_clear H10.
  split.
  assert (S_Flag (S_Mem_Upd S0 (MemMod.merge cthM12 cthM11'')) = S_Flag S0).
  auto. rewrite H10. rewrite H11. rewrite <- H8.
  UnfoldStateAll. auto.
  split.
  assert (S_ISR (S_Mem_Upd S0 (MemMod.merge cthM12 cthM11'')) = S_ISR S0).
  auto. rewrite H10. rewrite H12. rewrite <- H7. UnfoldStateAll.
  auto. split.
  unfold S_RF. auto.
  assert (S_RF (S_Mem_Upd S0 (MemMod.merge cthM12 cthM11'')) = S_RF S0).
  auto. rewrite H10.  rewrite H5. unfold S_RF. auto.
  assert (S_RF S' SP = S_RF S0 SP + 2).
  apply enter_SW_grt_auxb0_hyp1 with pc'; auto.
  rewrite H10 in H2.
  assert (S_RF S0 SP + 2 - 2 = S_RF S0 SP).
  omega.
  rewrite H11 in H2.
  apply enter_SW_grt_auxb0_hyp5 with (pc + 1) junk (stkSpacePrec (S_RF S0 SP) 8); auto.
Qed.


Lemma enter_SW_grt_auxb_ih:
forall (p p' : Pred) cthM1 R fl isr M cthM11 cthM12 S' cthM11'' pc pc' M' R' fl' isr' junk S0,
  p (cthM1, R, fl, isr)
  -> (forall pc' : nat, pimp (p |> wrap_fun ih_gy pc _SWITCH pc') p')
  -> MemMod.disjUnion cthM11 cthM12 cthM1 
  -> (INV0 *^* mTrue) cthM12
  -> stkSpacePrec (R SP) (8 + 2) cthM11
  -> ((S_RF S' SP - 2 |-> pc + 1) *^* (S_RF S' SP - 2 + 1 |-> junk) *^* stkSpacePrec (S_RF S' SP - 2) 8) cthM11''
  -> nextS pc (comml (_call _SWITCH)) (M, R, fl, isr) (M', R', fl', isr')
  -> nextS pc' (comms _retw) S0 S'
  -> S_RF S0 SP = R' SP
  -> MemMod.subseteq (MemMod.merge cthM12 cthM11'') (S_Mem S')
  -> isr = S_ISR S'
  -> fl _IF = S_Flag S' _IF
  -> MemMod.disjoint cthM11'' cthM12
  -> p' (S_Mem_Upd S' (MemMod.merge cthM12 cthM11'')).
Proof.
  intros.
  unfold pimp in H0.
  apply H0 with pc'.
  unfold linkPG. exists (cthM1, R, fl, isr).
  split; auto.
  apply enter_SW_grt_auxb0_ih with M cthM11 M' R' fl' isr' junk S0; auto.
Qed.


Lemma enter_SW_grt_hype_ih:
forall cthM'1 cthM'2 cthM11'' cthM12 cthM2 cthM',
  INV1 cthM'2
  -> MemMod.disjUnion cthM'1 cthM'2 cthM'
  -> disjUnionList (cthM2 :: cthM12 :: cthM11'' :: nil) cthM'
  -> INV1 cthM2
  -> (cthM2 = cthM'2) /\ (cthM'1 = MemMod.merge cthM12 cthM11'').
Proof.
  intros.
  assert (exists cthM'1', MemMod.disjUnion cthM2 cthM'1' cthM' /\ disjUnionList (cthM12 :: cthM11'' :: nil) cthM'1').
  unfold disjUnionList in *.
  inversion_clear H1.
  exists x.
  inversion_clear H3. split; auto.
  InvertAll. rename x into cthM'1'.
  assert (cthM2 = cthM'2).
  assert (precise INV1).
  apply Prec1.
  unfold precise in H4.
  apply H4 with cthM'.
  apply MemMod.disjUnion_subseteq_left with cthM'1'; auto.
  apply MemMod.disjUnion_subseteq_right with cthM'1; auto.
  auto. auto.

  split_l.
  assert (cthM'1' = cthM'1).
  rewrite <- H4 in H0.
  apply MemMod.disjUnion_unique with cthM2 cthM2 cthM'; auto.
  auto with PMap.
  rewrite <- H6.
  unfold disjUnionList in H5.
  InvertAll.
  subst x0.
  assert (cthM11'' = x). auto with PMap.
  subst x.
  inversion_clear H8. auto.
Qed.



Lemma enter_SW_grt2_aux2_hyp0_ih:
forall INVM cthM0 cthM12 M22 M12 cthM2 M0' M2' S0 M21' cthM_oldStk,
  MemMod.disjUnion INVM cthM0 cthM12
  -> disjUnionList (M22 :: M12 :: cthM2 :: cthM12 :: nil) M2'
  -> MemMod.disjUnion M0' M2' (S_Mem S0)
  -> MemMod.disjUnion M21' cthM_oldStk M0'
  -> MemMod.disjoint cthM0 cthM_oldStk.
Proof.
  intros.
  assert (MemMod.disjoint (MemMod.merge cthM0 cthM2) cthM_oldStk).
  apply enter_SW_grt2_aux2_hyp0 with INVM cthM12 M22 M12 M0' M2' S0 M21'; auto.
  apply MemMod.disj_merge_disj_left with cthM2; auto.
Qed.


Lemma enter_SW_grt2_aux2_hyp1_ih:
forall pc' M0' M2' S0 S' M22 M12 cthM2 cthM12 M21' cthM_oldStk M122 cthMR INVM cthM0,
  MemMod.disjUnion M0' M2' (S_Mem S0)
  -> disjUnionList (M22 :: M12 :: cthM2 :: cthM12 :: nil) M2'
  -> MemMod.disjUnion M21' cthM_oldStk M0'
  -> MemMod.disjUnion M122 cthMR M12
  -> MemMod.disjUnion INVM cthM0 cthM12
  -> nextS pc' (comms _retw) S0 S'
  -> disjUnionList 
       (M22 :: cthMR :: INVM :: M21' :: MemMod.merge cthM0 cthM_oldStk :: cthM2 :: M122 :: nil) 
          (S_Mem S').
Proof.
  intros.
  induction S0 as (((Mx, Rx), flx), isrx).
  induction S' as (((My, Ry), fly), isry).
  UnfoldStateAll.
  destruct (nextS_inversion H4).
  subst My. clear H4 H6.
  assert ((eqM M21' *^* eqM cthM_oldStk *^* (DUL2mstar (M22 :: M12 :: cthM2 :: cthM12 :: nil))) Mx).
  exists M0'. exists M2'.
  split_l. split.
  exists M21'. exists cthM_oldStk. unfold eqM. auto.
  apply DUL2mstar_left; auto.
  unfold DUL2mstar in H4.
  sep on.
  sep lift 5 in H4.
  assert (forall M, eqM cthM12 M -> (eqM INVM *^* eqM cthM0) M).
  intros. unfold eqM in H5. subst M. 
  exists INVM. exists cthM0. unfold eqM. auto.
  sep weaker_subst in H4 with (eqM INVM *^* eqM cthM0).
  clear H5 H6.
  sep lift 5 in H4.
  assert (forall M, eqM M12 M -> (eqM M122 *^* eqM cthMR) M).
  intros. unfold eqM in H5. subst M. 
  exists M122. exists cthMR. unfold eqM. auto.
  sep weaker_subst in H4 with (eqM M122 *^* eqM cthMR).
  sep off.
  cut (DUL2mstar (M22 :: cthMR :: INVM :: M21' :: MemMod.merge cthM0 cthM_oldStk :: cthM2 :: M122 :: nil) Mx).
  apply DUL2mstar_right.
  unfold DUL2mstar.
  sep on.
  sep matchH H4 with 8.
  sep lift 2 in H4.
  sep off.
  unfold eqM.
  unfold mstar in H4.
  unfold eqM in H4.
  InvertAll.
  subst x x1. 
  unfold Aemp in H11. subst x2.
  assert (cthM0 = x0). auto with PMap.
  subst x0.
  inversion_clear H7.
  subst h. auto with PMap.
Qed.


Lemma enter_SW_grt2_hyp3_ih:
forall M22 cthMR INVM M21' cthM_old M122 S' cthM2,
    disjUnionList
          (M22 :: cthMR :: INVM :: M21' :: cthM_old :: cthM2 :: M122 :: nil)
          (S_Mem S')
    -> exists M'0, exists M'1, 
        MemMod.disjUnion M'0 M'1 (S_Mem S')
        /\ MemMod.disjUnion M21' M22 M'1
        /\ disjUnionList (cthMR :: INVM :: cthM_old :: cthM2 :: M122 :: nil) M'0.
Proof.
  intros.
  induction S' as (((Mx, Rx), flx), isrx).
  UnfoldStateAll.
  assert ((eqM M21' *^* eqM M22 *^* (DUL2mstar (cthMR :: INVM :: cthM_old :: cthM2 :: M122 :: nil))) Mx).
  assert (DUL2mstar (M22 :: cthMR :: INVM :: M21' :: cthM_old :: cthM2 :: M122 :: nil) Mx).
  apply DUL2mstar_left; auto.
  unfold DUL2mstar in *.
  sep on.
  sep matchH H0 with 8.
  unfold mstar at 1 in H0.
  InvertAll.
  exists x0.  exists x.
  split. auto with PMap.
  split.
  unfold mstar in H0. unfold eqM in H0.
  InvertAll.
  subst x1 x2. auto.
  apply DUL2mstar_right; auto.
Qed.



Lemma enter_SW_grt2_hyp7_ih:
forall cthMR INVM cthM_old M122 M'0 TM TM' tid tid' cthM2,
  disjUnionList (cthMR :: INVM :: cthM_old :: cthM2 :: M122 :: nil) M'0
  -> MemSplit (ThrdMemMod.remove (ThrdMemMod.remove TM tid) tid') M122
  -> TM' = ThrdMemMod.put
           (ThrdMemMod.put TM tid'
              (MemMod.merge (MemMod.merge cthMR INVM) cthM2)) tid cthM_old
  -> tid <> tid'
  -> MemSplit TM' M'0.
Proof.
  intros.
  rewrite H1.
  assert (DUL2mstar (cthMR :: INVM :: cthM_old :: cthM2 :: M122 :: nil) M'0).
  apply DUL2mstar_left; auto.
  unfold DUL2mstar in H3.
  sep on.
  assert (M'0 ||- eqM cthM_old *^* (eqM M122 *^* (eqM cthMR *^* eqM INVM *^* eqM cthM2))).
  sep matchH H3 with 5.
  clear H3. rename H4 into H3.
  sep off.
  unfold mstar in H3.
  unfold eqM in H3.
  InvertAll.
  subst x x1 x4 x5 x6.
  apply enter_SW_grt_hyp6 with x0; auto with PMap.

  assert (ThrdMemMod.remove (ThrdMemMod.put TM tid' (MemMod.merge (MemMod.merge cthMR INVM) cthM2)) tid
           = ThrdMemMod.put (ThrdMemMod.remove TM tid) tid' (MemMod.merge (MemMod.merge cthMR INVM) cthM2)).
  auto with PMap.
  rewrite H3. clear H3.
  apply enter_SW_grt_hyp6 with M122; auto.
  inversion_clear H9. subst x3.
  inversion_clear H7.
  subst x2. auto.
Qed.


Lemma enter_SW_grt2_hypa_ih:
forall cthMR INVM M22 M21' cthM_old M122 cthM2 S',
    disjUnionList
          (M22 :: cthMR :: INVM :: M21' :: cthM_old :: cthM2 :: M122 :: nil)
          (S_Mem S')
    -> MemMod.disjUnion (MemMod.merge INVM cthM2) cthMR
           (MemMod.merge (MemMod.merge cthMR INVM) cthM2).
Proof.
  intros.
  induction S' as (((Mx, Rx), flx), isrx).
  simplS in H.
  assert (DUL2mstar (M22 :: cthMR :: INVM :: M21' :: cthM_old :: cthM2 :: M122 :: nil) Mx).
  apply DUL2mstar_left; auto.
  unfold DUL2mstar in H0.
  assert ((eqM INVM *^* eqM cthM2 *^* eqM cthMR *^* mTrue) Mx).
  sep on.
  sep matchH H0 with 4.
  unfold mstar at 1 in H1.
  InvertAll.
  assert ((eqM cthMR *^* eqM INVM *^* eqM cthM2) x).
  sep on. sep matchH H1 with 3.
  unfold mstar at 1 in H1.
  InvertAll.
  unfold eqM in H7. subst x2.
  unfold mstar in H1.
  unfold eqM in H1. 
  InvertAll.
  subst x2 x3.
  inversion_clear H6.
  subst x1.
  clear H4 H2 x0.
  unfold mstar in H3. unfold eqM in H3.
  InvertAll.
  subst x1 x2 x3.
  inversion_clear H4.
  subst x0.
  inversion_clear H2. subst x. auto.
Qed.


Lemma enter_SW_grt2_hypa_ih2:
forall cthMR INVM M22 M21' cthM_old M122 cthM2 S',
    disjUnionList
          (M22 :: cthMR :: INVM :: M21' :: cthM_old :: cthM2 :: M122 :: nil)
          (S_Mem S')
    -> MemMod.disjoint INVM cthM2.
Proof.
  intros.
  induction S' as (((M, R), fl), isr).
  simplS in H.
  assert (DUL2mstar (M22 :: cthMR :: INVM :: M21' :: cthM_old :: cthM2 :: M122 :: nil) M).
  apply DUL2mstar_left. auto.
  clear H.
  assert ((eqM INVM *^* eqM cthM2 *^* mTrue) M).
  unfold DUL2mstar in H0.
  sep on.
  sep matchH H0 with 3.
  unfold mstar in H.
  unfold eqM in H.
  InvertAll.
  subst x1 x2.
  inversion_clear H2; auto.
Qed.


Lemma enter_SW_grt2_auxc_ih_hyp0:
forall M0 cthM_old M'2 M01 M02,
  MemMod.disjUnion M0 cthM_old M'2
  -> MemMod.disjUnion M01 M02 M0
  -> MemMod.disjUnion (MemMod.merge M01 cthM_old) M02 M'2.
Proof.
  intros.
  assert ((eqM M01 *^* eqM M02 *^* eqM cthM_old) M'2).
  exists M0. exists cthM_old.
  split_l.
  split. exists M01. exists M02.
  unfold eqM. auto.
  unfold eqM. auto.
  assert ((eqM M01 *^* eqM cthM_old *^* eqM M02) M'2).
  sep on.
  sep matchH H1 with 3.
  unfold mstar in H2. unfold eqM in H2.
  InvertAll.
  subst x0 x1 x2.
  inversion_clear H4. subst x. auto.
Qed.

Lemma enter_SW_grt2_hype_ih:
forall M'2 M'20 M'21 M0 M01 M02 cthM_old,
  MemMod.disjUnion M'20 M'21 M'2
  -> INV1 M'21
  -> MemMod.disjUnion M0 cthM_old M'2
  -> MemMod.disjUnion M01 M02 M0
  -> INV1 M02
  -> M'20 = MemMod.merge M01 cthM_old.
Proof.
  intros.
  assert (MemMod.disjUnion (MemMod.merge M01 cthM_old) M02 M'2).
  apply enter_SW_grt2_auxc_ih_hyp0 with M0; auto.
  assert (precise INV1). apply Prec1.
  assert (M'21 = M02).
  unfold precise in H5.
  apply H5 with M'2; auto.
  apply MemMod.disjUnion_subseteq_right with M'20; auto.
  apply MemMod.disjUnion_subseteq_right with (MemMod.merge M01 cthM_old); auto.
  apply MemMod.disjUnion_unique with M'21 M02 M'2; auto with PMap.
Qed.

Lemma enter_SW_grt2_hyp2_ih:
forall T T' B RdyQ' tid tid' sp sp' tcb tcb' pc pc' S0 S' M0' M2' M22 M12 cthM2 cthM12 cthMR TM junk,
  SW_Post_aux T' RdyQ' tid' sp' sp tcb' pc junk (S_Mem_Upd S0 M0')
  -> nextS pc' (comms _retw) S0 S'
  -> MemMod.disjUnion M0' M2' (S_Mem S0)
  -> disjUnionList (M22 :: M12 :: cthM2 :: cthM12 :: nil) M2'
  -> embed_WQSet T B M22
  -> MemSplit (ThrdMemMod.remove TM tid) M12
  -> ThrdMemMod.maps_to TM tid' cthMR
  -> tid <> tid'
  -> INV1 cthM2
  -> (INV0 *^* mTrue) cthM12
  -> T' = TQMod.put T tid (sp, TCB_Status tcb, pc)
  -> exists INV0M, exists cthM0, exists M21', exists M122, exists cthM_old, exists cthM_oldStk,
        MemMod.disjUnion INV0M cthM0 cthM12 
        /\ INV0 INV0M
        /\ (embed_curr_TCB T' tid' *^* embed_RdyQ T' RdyQ') M21'
        /\ MemMod.disjUnion M122 cthMR M12
        /\ MemSplit (ThrdMemMod.remove (ThrdMemMod.remove TM tid) tid') M122
        /\ MemMod.disjUnion cthM0 cthM_oldStk cthM_old
        /\ (sp |-> pc *^* sp + 1 |-> junk *^* stkSpacePrec sp 8) cthM_oldStk
        /\ disjUnionList (M22 :: cthMR :: INV0M :: M21' :: cthM_old :: cthM2 :: M122 :: nil) (S_Mem S').
Proof.
  intros.
  unfold mstar at 1 in H8. InvertAll.
  rename x into INV0M. rename x0 into cthM0. exists INV0M. exists cthM0.
  unfold SW_Post_aux in H.
  InvertAll.
  unfold SW_MPost in H11.
  assert ((embed_curr_TCB T' tid' *^* embed_RdyQ T' RdyQ' *^* (sp |-> pc *^* sp + 1 |-> junk *^*
           stkSpacePrec sp 8)) (S_Mem (S_Mem_Upd S0 M0'))).
  sep on.
  sep matchH H11 with 4.
  clear H11.
  unfold mstar at 1 in H15.
  InvertAll.
  rename x into M21'. exists M21'.
  rename x0 into cthM_oldStk.
  assert (exists M122, MemMod.disjUnion M122 cthMR M12 /\
                        MemSplit (ThrdMemMod.remove (ThrdMemMod.remove TM tid) tid') M122).
  apply memsplit_thrd_remove; auto.
  unfold ThrdMemMod.maps_to. unfold ThrdMemMod.remove.
  induction (ThrdMemMod.beq_A tid' tid).
  absurd (tid' <> tid); auto. unfold ThrdMemMod.maps_to in H5. auto.
  InvertAll. rename x into M122.
  exists M122.
  exists (MemMod.merge cthM0 cthM_oldStk).
  exists cthM_oldStk.
  split; auto. split; auto. split; auto.
  split; auto. split; auto.
  split. unfold MemMod.disjUnion. split; auto.
  apply enter_SW_grt2_aux2_hyp0_ih with INV0M cthM12 M22 M12 cthM2 M0' M2' S0 M21'; auto.
  split. auto.
  apply enter_SW_grt2_aux2_hyp1_ih with pc' M0' M2' S0 M12 cthM12; auto.
Qed.

Lemma enter_SW_grt2_auxc_ih_aux4_hyp1:
forall M'2 M0 cthM_old M01 M02 Mx cthM_oldStk oldStk1 oldStk2,
  MemMod.disjUnion M0 cthM_old M'2
  -> MemMod.disjUnion M01 M02 M0
  -> MemMod.disjUnion Mx cthM_oldStk cthM_old
  -> MemMod.disjUnion oldStk1 oldStk2 cthM_oldStk
  -> MemMod.disjoint M01 oldStk1.
Proof.
  intros.
  assert (MemMod.disjoint M01 cthM_old).
  inversion_clear H0. subst M0.
  inversion_clear H.
  apply MemMod.disj_merge_disj_left with M02; auto.
  inversion_clear H1.
  subst cthM_old.
  assert (MemMod.disjoint M01 cthM_oldStk).
  apply MemMod.disj_x_merge_disj_right with Mx; auto.
  inversion_clear H2.
  rewrite H6 in H1.
  apply MemMod.disj_x_merge_disj_left with oldStk2; auto.
Qed.


Lemma enter_SW_grt2_auxc_ih_aux4_hyp2:
forall M'2 M0 cthM_old M01 M02 cthM0 cthM_oldStk oldStk1 oldStk2,
  MemMod.disjUnion M0 cthM_old M'2
  -> MemMod.disjUnion M01 M02 M0
  -> MemMod.disjUnion cthM0 cthM_oldStk cthM_old
  -> MemMod.disjUnion oldStk1 oldStk2 cthM_oldStk
  -> MemMod.disjUnion (MemMod.merge M01 oldStk1) (MemMod.merge oldStk2 cthM0)
     (MemMod.merge M01 cthM_old).
Proof.
  intros.
  assert ((eqM M01 *^* (eqM cthM0 *^* (eqM oldStk1 *^* eqM oldStk2))) 
             (MemMod.merge M01 cthM_old)).
  exists M01. exists cthM_old.
  split.
  split; auto.
  inversion_clear H0. subst M0.
  inversion H.
  apply MemMod.disj_merge_disj_left with M02; auto.
  split.
  unfold eqM. auto.
  exists cthM0. exists cthM_oldStk.
  split_l.
  unfold eqM. split_l.
  unfold mstar. exists oldStk1. exists oldStk2. auto.
  assert ((eqM M01 *^* eqM oldStk1 *^* (eqM oldStk2 *^* eqM cthM0))
             (MemMod.merge M01 cthM_old)).
  sep on.
  sep matchH H3 with 4.
  unfold mstar in H4. unfold eqM in H4.
  InvertAll.
  subst x1 x2 x3 x4.
  inversion_clear H8. subst x.
  inversion_clear H6. subst x0. auto.
Qed.



Lemma enter_SW_grt2_auxc_ih_aux4_hyp3:
forall M'2 M0 M01 M02 cthM_old cthM0 cthM_oldStk sp w junk m,
  MemMod.disjUnion cthM0 cthM_oldStk cthM_old
  -> MemMod.disjUnion M0 cthM_old M'2
  -> MemMod.disjUnion M01 M02 M0
  -> (sp |-> w *^* sp + 1 |-> junk *^* m) cthM_oldStk
  -> (sp |~>w w) (MemMod.merge M01 cthM_old).
Proof.
  intros.
  assert ((eqM M01 *^* (eqM cthM0 *^* eqM cthM_oldStk)) 
             (MemMod.merge M01 cthM_old)).
  exists M01. exists cthM_old.
  split.
  split; auto.
  inversion_clear H1. subst M0.
  inversion H0.
  apply MemMod.disj_merge_disj_left with M02; auto.
  split.
  unfold eqM. auto.
  exists cthM0. exists cthM_oldStk.
  unfold eqM. auto.
  sep on.
  sep lift 2 in H3.
  assert (forall M, eqM cthM_oldStk M -> (sp |->w w *^* m) M).
  intros.
  unfold eqM in H4. subst M.
  apply ptoW_intro1 with junk; auto.
  sep weaker_subst in H3 with (sp |->w w *^* m).
  sep on. sep lift 0 in H3.
  sep trivial.
Qed.

Lemma enter_SW_grt2_auxc_ih_aux4:
forall pc pc' M R fl isr M' R' fl' isr' cthM1 cthM11 cthM12 INVM cthM0 M0 M01 M02 cthM_oldStk cthM_old R0 fl0 junk M'2,
  nextS pc (comml (_call _SWITCH)) (M, R, fl, isr) (M', R', fl', isr')
  -> MemMod.disjUnion cthM11 cthM12 cthM1
  -> stkSpacePrec (R SP) (8 + 2) cthM11
  -> MemMod.disjUnion INVM cthM0 cthM12
  -> (R' SP |-> pc + 1 *^* R' SP + 1 |-> junk *^* stkSpacePrec (R' SP) 8) cthM_oldStk
  -> MemMod.disjUnion cthM0 cthM_oldStk cthM_old
  -> INV0 INVM
  -> INV0 M01
  -> INV1 M02
  -> MemMod.disjUnion M0 cthM_old M'2
  -> MemMod.disjUnion M01 M02 M0
  -> fl0 _IF = false
  -> fl _IF = false
  -> isr = false
  -> R0 SP = R' SP + 2
  -> wrap_fun ih_gy pc _SWITCH pc' (cthM1, R, fl, isr)
     (MemMod.merge M01 cthM_old, R0, fl0, false).
Proof.
  intros.
  unfold wrap_fun.
  unfold conjG at 1.
  exists (MemMod.merge M01 cthM_old, RF_update R0 SP (R0 SP - 2), fl0, false).
  split.
  unfold conjG.
  assert (nextS pc (comml (_call _SWITCH)) (cthM1, R, fl, isr) (MemUpdW cthM1 (R SP - 2) (pc+1) junk, RF_update R SP (R SP - 2), fl, isr)).
  apply nextS_call.
  unfold pushW.
  split.
  assert ((mTrue *^* stkSpacePrec (R SP) (8 + 2)) cthM1).
  exists cthM12. exists cthM11. split; auto with PMap.
  unfold mTrue. split; auto.
  sep on. sep lift 1 in H14.
  sep stksplit 2 in H14.
  sep lift 0 in H14. sep trivial.
  split; auto.
  apply MemUpdW_MemUpd16 with junk; auto.
  set (cthM' := MemUpdW cthM1 (R SP - 2) (pc + 1) junk) in *.
  exists (cthM', RF_update R SP (R SP - 2), fl, isr).
  split; auto.
  assert (R SP - 2 = R' SP).
  apply enter_SW_grt_auxb0_hyp0 with pc M fl isr M' fl' isr' _SWITCH; auto.
  unfold ih_gy.
  exists (R SP - 2). simplS. reg simpl.

  assert (R SP = R' SP + 2).
  rewrite <- H15.
  assert (R SP >= 2).
  apply ge_trans with (8 + 2).
  sep on.
  apply stkSpacePrec_sp_ge_sz with cthM11; auto.
  auto with arith.
  assert (R SP - 2 + 2 = R SP). auto. auto. rename H16 into HRSP'.

  split.
  unfold prepost. simplS.
  sep on.
  sep stksplit in H1.
  destruct H1 as [cthM11_1 [cthM11_2 [HcthM11split [HcthM11_1 HcthM11_2]]]].

  exists (MemMod.merge INVM cthM11_1).
  exists (MemMod.merge (MemUpdW cthM11_2 (R SP - 2) (pc+1) junk) cthM0).
  sep lift 2 in H3.
  unfold mstar at 1 in H3.
  sep off.
  InvertAll.
  rename x into oldStk1. rename x0 into oldStk2.
  exists (MemMod.merge M01 oldStk1).

  split.
  exists INVM. exists cthM11_1. split; auto.
  split; auto.
  apply enter_BLK_grt2_auxc_hyp1 with cthM1 cthM11 cthM12 cthM11_2 cthM0; auto.

  split.
  exists M01. exists oldStk1.
  rewrite <- H15 in H3. split; auto.
  split; auto.

  apply enter_SW_grt2_auxc_ih_aux4_hyp1 with M'2 M0 cthM_old M02 cthM0 cthM_oldStk oldStk2; auto.

  split. 
  assert (MemMod.disjUnion (MemMod.merge (MemUpdW cthM11_2 (R SP - 2) (pc + 1) junk) cthM0)
                             (MemMod.merge INVM cthM11_1) cthM').

  apply enter_SW_grt2_auxc_auxx with cthM1 cthM11 cthM12; auto.
  auto with PMap.

  assert (MemUpdW cthM11_2 (R SP - 2) (pc + 1) junk = oldStk2).
  rewrite H15 in HcthM11_2. rewrite H15.
  apply enter_SW_grt2_auxc_hypx; auto.
  rewrite H16.
  apply enter_SW_grt2_auxc_ih_aux4_hyp2 with M'2 M0 M02 cthM_oldStk; auto.
  rewrite <- H10 in H11. split_l. split_l.
  split_l. rewrite <- H13 in HRSP'. rewrite HRSP'. auto.

  apply nextS_retw. exists (pc+1).
  unfold popW.
  reg simpl.
  split.
  rewrite H13. assert (R' SP + 2 - 2 = R' SP). auto.
  rewrite H14.
  apply enter_SW_grt2_auxc_ih_aux4_hyp3 
    with M'2 M0 M02 cthM0 cthM_oldStk junk (stkSpacePrec (R' SP) 8);
  auto.

  rewrite H13. assert (R' SP + 2 - 2 = R' SP). auto.
  rewrite H14. rewrite <- H13.
  assert (RF_update R0 SP (R0 SP) = R0).
  apply RF_update_prev. auto.
Qed.



Lemma enter_SW_grt2_auxc_ih:
forall cthM1 cthM2 cthM R fl isr cthM11 cthM12 INVM cthM0 cthM_oldStk pc junk
            M0 M'2 cthM_old R0 fl0 (p p'': Pred) M M' R' fl' isr',
  MemMod.disjUnion cthM1 cthM2 cthM
  -> p (cthM1, R, fl, isr)
  -> INV1 cthM2
  -> MemMod.disjUnion cthM11 cthM12 cthM1
  -> stkSpacePrec (R SP) (8 + 2) cthM11
  -> MemMod.disjUnion INVM cthM0 cthM12
  -> INV0 INVM
  -> MemMod.disjUnion cthM0 cthM_oldStk cthM_old
  -> (R' SP |-> pc + 1 *^* R' SP + 1 |-> junk *^* stkSpacePrec (R' SP) 8) cthM_oldStk
  -> (INV0 *^* INV1) M0
  -> MemMod.disjUnion M0 cthM_old M'2
  -> (forall pc' : nat, pimp (p |> wrap_fun ih_gy pc _SWITCH pc') p'')
  -> fl0 _IF = false
  -> fl _IF = false
  -> isr = false
  -> nextS pc (comml (_call _SWITCH)) (M, R, fl, isr) (M', R', fl', isr')
  -> R0 SP = R' SP + 2
  -> (p'' *<* INV1) (S_Mem_Upd (cthM_old, R0, fl0, false) M'2).
Proof.
  intros.
  unfold pmstar.
  simplS.

  unfold mstar in H8.
  InvertAll. rename x into M01. rename x0 into M02.
  exists (MemMod.merge M01 cthM_old).
  exists M02.
  split.
  apply enter_SW_grt2_auxc_ih_hyp0 with M0; auto.
  split; auto.
  unfold pimp in H10.
  apply H10 with 0.
  unfold linkPG.
  exists (cthM1, R, fl, isr).
  split; auto.
  apply enter_SW_grt2_auxc_ih_aux4 with M M' R' fl' isr' cthM11 cthM12 INVM cthM0 M0 M02 cthM_oldStk junk M'2; auto.
Qed.

Lemma enter_SW_grt2_hypK_ih:
forall cthM2 INVM cthMR M22 M21' cthM_old M122 S',
    disjUnionList
          (M22 :: cthMR :: INVM :: M21' :: cthM_old :: cthM2 :: M122 :: nil)
          (S_Mem S')
  -> MemMod.disjoint cthM2 INVM.
Proof.
  intros.
  assert (MemMod.disjoint INVM cthM2).
  apply enter_SW_grt2_hypa_ih2 with cthMR M22 M21' cthM_old M122 S'; auto.
  auto with PMap.
Qed.


Lemma enter_SW_grt2_hypJ_ih:
forall cthM' cthM2 INVM cthMR M22 M21' cthM_old M122 S',
  cthM' = MemMod.merge (MemMod.merge cthMR INVM) cthM2
  -> disjUnionList
          (M22 :: cthMR :: INVM :: M21' :: cthM_old :: cthM2 :: M122 :: nil)
          (S_Mem S')
  -> MemMod.disjUnion (MemMod.merge cthM2 INVM) cthMR cthM'.
Proof.
  intros.
  subst cthM'.
  assert (MemMod.merge cthM2 INVM = MemMod.merge INVM cthM2).
  assert (MemMod.disjoint cthM2 INVM).
  apply enter_SW_grt2_hypK_ih with cthMR M22 M21' cthM_old M122 S'; auto.
  auto with PMap.
  rewrite H.
  apply enter_SW_grt2_hypa_ih with M22 M21' cthM_old M122 S'; auto.
Qed.



Lemma enter_SW_ih_prdaux:
forall M1' cthM11' M21 R' fl' isr' pc junk Psi'0 S T B RdyQ tid rho theta tcb,
  MemMod.disjUnion cthM11' M21 M1'
  -> ((R' SP) |-> pc *^* (R' SP) + 1 |-> junk *^* stkSpacePrec (R' SP) 8) cthM11'
  -> (embed_curr_TCB T tid *^* embed_RdyQ T RdyQ) M21
  -> WF_AIM_World Psi'0 S T B RdyQ tid rho theta
  -> T tid = Some tcb
  -> SW_Prd_aux T RdyQ tid (R' SP) tcb pc junk (M1', R', fl', isr').
Proof.
  intros.
  unfold SW_Prd_aux.
  split. unfold S_Mem. 
  apply mem_compose_c with cthM11' M21; auto.
  split; auto.
  split. 
  apply WF_AIM_World_RdyQinT with Psi'0 S B tid rho theta.
  auto.
  split; auto.
  split.
  unfold WF_AIM_World in H2.
  InvertAll.
  unfold WF_currTID in H4.
  InvertAll. auto.
  apply WF_AIM_World_WFTQ with Psi'0 S B RdyQ tid rho theta.
  auto.
Qed.

Lemma enter_SW_ih_grt:
forall D psi pc theta D' Psi Psi' Psi'0 S S',
  IH_WFcmd psi pc theta (comml (_call _SWITCH))
  -> Minimal_D D
  -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
  -> lift_psi psi rho_ih Psi
  -> Psi' = ext_Prim_Specs Psi
  -> subsetCT Psi' Psi'0
  -> Itp_AIM rho_ih theta Psi'0 S
  -> nextS pc (comml (_call _SWITCH)) S S'
  -> WFST D rho_prim 0 Psi'0 (scap_lift_g SW_grt S').
Proof.
  intros.
  unfold WFST.
  intros.
  unfold scap_lift_g in H7.
  InvertAll.
  induction S' as (((M', R'), fl'), isr').
  induction S as (((M, R), fl), isr).
  induction theta.
  rename n into tag.
  unfold Itp_AIM in H5.
  unfold pmstar in H5.
  InvertAll.
  rename x into T. rename x0 into B. rename x1 into RdyQ. rename x2 into tid.
  rename x3 into M1. rename x4 into M2.
  assert ((embed_curr_TCB T tid *^* embed_RdyQ T RdyQ *^* embed_WQSet T B) M2).
  apply leqvM_app_subst with (embed_curr_TCB T tid *^* (embed_RdyQ T RdyQ *^* embed_WQSet T B)); auto.
  apply leqvM_commut.
  apply mstar_assoc.
  unfold mstar at 1 in H11. InvertAll.
  rename x into M21. rename x0 into M22. clear H12.
  assert (exists TM, exists cthM, exists M12, exists cthM1, exists cthM2,
        MemMod.disjUnion M12 cthM M1
        /\ MemMod.disjUnion cthM1 cthM2 cthM
        /\ WF_currTID tid T B RdyQ rho_ih
        /\ WF_TWQ T B RdyQ
        /\ exists_Idle T RdyQ tid
        /\ WF_WQ_WQSpec B Delta
        /\ MemSplit (ThrdMemMod.remove TM tid) M12
        /\ ThrdMemMod.maps_to TM tid cthM
        /\ WF_Curr_Thrd Psi'0 (S_Mem_Upd (M1, R, fl, isr) cthM) rho_ih (Ast p g tag) T tid
        /\ WF_ReadyQ Psi'0 T RdyQ TM
        /\ WF_WQSet Psi'0 T B TM
        /\ p (cthM1, R, fl, isr)
        /\ INV1 cthM2).
  apply decompose_World_ih; auto.
  InvertAll. rename x into TM. rename x0 into cthM. rename x1 into M12. rename x2 into cthM1. rename x3 into cthM2.
  assert (WQMod.not_in_dom tid RdyQ). unfold WF_currTID in H16. InvertAll. auto.
  assert (exists M1, exists M2, 
            MemMod.disjUnion M1 M2 (S_Mem (cthM1, R, fl, isr))
         /\ stkSpacePrec (S_RF (cthM1, R, fl, isr) SP) (8 + 2) M1 
         /\ (INV0 *^* mTrue) M2).

  apply enable_sw_stkspace_ih with p; auto.
  apply WFcallsw_enablesw_ih with psi pc g tag; auto.
  InvertAll. rename x into cthM11. rename x0 into cthM12.
  unfold S_Mem in H29. unfold S_RF in H28.
  assert (disjUnionList (M2 :: M12 :: cthM2 :: cthM12 :: cthM11 :: nil) M).
  apply mem_list_compose_a with cthM1 cthM M1; auto.

  assert 
    (exists junk,
        ((R' SP |-> (pc+1)) *^* (R' SP + 1 |-> junk) *^* (stkSpacePrec (R' SP) 8)) (MemUpdW cthM11 (R' SP) (pc + 1) junk)
        /\ disjUnionList (M2 :: M12 :: cthM2 :: cthM12 :: (MemUpdW cthM11 (R' SP) (pc + 1) junk) :: nil) M').
  apply call_sw_frame with M R fl fl' isr isr'; auto.
  InvertAll. rename x into junk1.
  set (cthM11' := MemUpdW cthM11 (R' SP) (pc + 1) junk1) in *.
  assert (exists M1', exists M2', 
           MemMod.disjUnion cthM11' M21 M1' 
           /\ disjUnionList (M22 :: M12 :: cthM2 :: cthM12 :: nil) M2'
           /\ MemMod.disjUnion M1' M2' M').
  apply mem_list_compose_b with M2; auto.
  InvertAll.  rename x into M1'. rename x0 into M2'.
  assert (exists tcb, T tid = Some tcb).
  unfold WF_currTID in H16.
  InvertAll.
  unfold TQMod.in_dom in H40. auto.
  inversion_clear H36. rename x into tcb.
  assert (exists M0' : MemMod.Map,
         exists RdyQ' : WQ,
           exists tid' : TID,
             exists sp' : nat,
               exists tcb' : Thread,
                 MemMod.disjUnion M0' M2' (S_Mem S0) /\
                 addThrd RdyQ tid = addThrd RdyQ' tid' /\
                 SW_Post_aux (TQMod.put T tid ((R' SP), TCB_Status tcb, pc+1)) 
                              RdyQ' tid' sp' (R' SP) tcb' (pc+1) junk1 (S_Mem_Upd S0 M0')).
  apply H8 with M1' T tcb; auto.
  assert (S_Mem_Upd (M', R', fl', isr') M1' = (M1', R', fl', isr')). auto.
  rewrite H36.
  apply enter_SW_ih_prdaux 
   with cthM11' M21 Psi'0 (S_Mem_Upd (M, R, fl, isr) M1) B rho_ih (Ast p g tag); auto.
  clear H8.
  InvertAll.
  rename x into M0'. rename x0 into RdyQ'. rename x1 into tid'.
  rename x2 into sp'. rename x3 into tcb'.
  assert ((fl _IF = false) /\ (isr = false)).
  apply call_sw_flags_false_ih with p g tag psi Psi'0 pc cthM cthM1 R T tid; auto.
  inversion_clear H39.
  assert (tid = tid' \/ tid <> tid').
  apply nat_eq_neq_dec.
  inversion_clear H39.

  (* tid = tid' *)
  (* clear H24 H27 H26 cthM11. *)
  assert (RdyQ = RdyQ').
  assert (WQMod.not_in_dom tid' RdyQ').
  unfold SW_Post_aux in H40. InvertAll. auto.
  apply unique_remain with tid tid'; auto.
  exists (pc + 1).
  rewrite <- H39 in H40.
  rewrite <- H43 in H40. clear H43 H39 H8 tid' RdyQ'.
  assert (S_RF S0 SP = R' SP).
  assert (sp' = R' SP).
  apply enter_SW_grt_aux60 with T RdyQ tid tcb (pc+1) junk1 tcb' S0 M0'; auto.
  unfold SW_Post_aux in H40. InvertAll.
  rewrite H8 in H44. auto.
  split.
  apply enter_SW_grt_aux7 with T RdyQ tid (R' SP) tcb junk1 sp' tcb' M0' M2'; auto.
  assert (exists p', exists g',
            SCAP_lookupCT psi (pc+1) (Ast p' g' tag)
         /\ (forall pc', pimp (p |> (wrap_fun ih_gy pc _SWITCH pc')) p')
         /\ (forall pc', gimp (p >o (wrap_fun ih_gy pc _SWITCH pc') oo g') g)).
  auto.
  InvertAll. rename x into p'. rename x0 into g'.
  exists (packS0 rho_ih (Ast p' g' tag)).
  split.
  apply psi_sw_lookupNPC_ih with psi Psi Psi'; auto.

  intros.
  unfold Itp0.
  assert (D rho_ih = SomeT1 _ (Itp_AIM rho_ih)).
  unfold Minimal_D in H0. InvertAll. auto.
  rewrite H46.
  unfold Itp_AIM.
  set (T' := TQMod.put T tid (R' SP, TCB_Status tcb, pc + 1)) in * |- *.
  exists T'. exists B. exists RdyQ. exists tid.

  assert (exists M21_0, exists cthM11'',
            ((S_RF S' SP - 2) |-> (pc+1) *^* (S_RF S' SP - 2 + 1) |-> junk1 *^* stkSpacePrec (S_RF S' SP - 2) 8) cthM11''
            /\ (embed_curr_TCB T' tid *^* embed_RdyQ T' RdyQ) M21_0
            /\ disjUnionList (M21_0 :: M22 :: M12 :: cthM2 :: cthM12 :: cthM11'' :: nil) (S_Mem S')).
  apply enter_SW_grt_aux8 with T (R' SP) sp' tcb tcb' pc' S0 M0' M2'; auto.
  InvertAll. rename x into M21_0. rename x0 into cthM11''.
  unfold pmstar.
  assert (exists M'1, exists M'2, 
            disjUnionList (M12 :: cthM2 :: cthM12 :: cthM11'' :: nil) M'1
            /\ MemMod.disjUnion M21_0 M22 M'2
            /\ MemMod.disjUnion M'1 M'2 (S_Mem S')).
  apply enter_SW_grt_hyp4.
  auto.
  InvertAll.
  rename x into M'1. rename x0 into M'2.
  exists M'1. exists M'2. split; auto.
  split.
  unfold WF_AIM_World.
  split. unfold T'.
  apply WF_currTID_ext_T with tcb; auto.
  split. unfold T'.
  apply exists_Idle_ext_T; auto.
  split. unfold T'.
  apply WF_TWQ_ext_T; auto.
  exists tcb. auto.
  split; auto.
  assert (exists cthM', MemMod.disjUnion M12 cthM' M'1 
                        /\ disjUnionList (cthM2 :: cthM12 :: cthM11'' :: nil) cthM').
  auto.
  InvertAll.
  rename x into cthM'.
  exists (ThrdMemMod.put TM tid cthM'). exists cthM'.
  split.
  assert (S_Mem (S_Mem_Upd S' M'1) = M'1). auto.
  rewrite H54.
  apply enter_SW_grt_hyp6 with M12; auto.
  split. auto with PMap.

  split.
  assert (S_ISR (M, R, fl, isr) = S_ISR S' /\ S_Flag (M, R, fl, isr) _IF = S_Flag S' _IF).
  apply enter_SW_grt_hyp9 with pc pc' _SWITCH (M', R', fl', isr') S0; auto.
  unfold S_ISR at 1 in H54. unfold S_Flag at 1 in H54.
  inversion_clear H54.
  unfold WF_Curr_Thrd.
  simpl.
  unfold WF_Curr_Thrd in H22. simpl in H22.
  InvertAll.
  exists (R' SP, TCB_Status tcb, pc + 1).
  exists x0.
  split. unfold T'. auto with PMap.
  split. unfold TQMod.maps_to in H54.
  rewrite H38 in H54. inversion H54.
  rewrite H22. unfold TCB_Status. auto.
  apply weaken_itp_IH_theta with p g (cthM, R, fl, isr); auto.
  rewrite H41 in H57.
  rewrite H57. UnfoldStateAll. auto.
  intro. clear H58.
  assert (p' (S_Mem_Upd S' (MemMod.merge cthM12 cthM11''))).
  apply enter_SW_grt_auxb_ih with p cthM1 R fl isr M cthM11 pc pc' M' R' fl' isr' junk1 S0; auto.

  apply enter_SW_grt_hypc with M21_0 M22 M12 cthM2; auto.
  apply enter_SW_grt_hypd with M21_0 M22 M12 cthM2 S'; auto.
  unfold pmstar.
  exists (MemMod.merge cthM12 cthM11'').
  exists cthM2.
  split. apply enter_SW_grt_hypa; auto.
  split; auto.
  assert (S_Mem_Upd (S_Mem_Upd S' M'1) cthM' = S_Mem_Upd S' cthM').
  auto.
  rewrite H58. clear H58.
  unfold IH_ItpG. intros.
  clear H54 H22 H59 x x0.
  InvertAll.
  rename x into cthM'1. rename x0 into cthM'2. rename x1 into SM1. rename x2 into SM2.
  exists cthM1. exists cthM2. exists SM1. exists SM2.
  split. unfold S_Mem. auto. (* H14 *)
  split. auto. split. auto. split. auto.
  unfold gimp in H45.
  apply H45 with pc'.
  unfold conjG.
  exists (S_Mem_Upd (S_Mem_Upd S' cthM') cthM'1).
  split; auto.
  assert ((cthM2 = cthM'2) /\ (cthM'1 = MemMod.merge cthM12 cthM11'')).
  apply enter_SW_grt_hype_ih with cthM'; auto.
  inversion_clear H60. rewrite H63.
  unfold conjPG. split; auto.
  assert (S_Mem_Upd (cthM, R, fl, isr) cthM1 = (cthM1, R, fl, isr)).
  auto. rewrite H60. clear H60.
  assert (S_Mem_Upd (S_Mem_Upd S' cthM') (MemMod.merge cthM12 cthM11'')
            = S_Mem_Upd S' (MemMod.merge cthM12 cthM11'')). auto.
  rewrite H60. clear H60.
  apply enter_SW_grt_auxb0_ih with M cthM11 M' R' fl' isr' junk1 S0; auto.
  apply enter_SW_grt_hypc with M21_0 M22 M12 cthM2; auto.
  apply enter_SW_grt_hypd with M21_0 M22 M12 cthM2 S'; auto.
  split.
  apply enter_SW_grt_hypf with T (R' SP, TCB_Status tcb, pc + 1); auto.
  apply enter_SW_grt_hypg with T (R' SP, TCB_Status tcb, pc + 1); auto.
  unfold WF_currTID in H16. InvertAll. auto.
  apply leqvM_app_subst with 
   (embed_curr_TCB T' tid *^* embed_RdyQ T' RdyQ *^* embed_WQSet T' B); auto.
  apply mstar_assoc. unfold mstar at 1. exists M21_0. exists M22.
  split; auto. split; auto.
  apply enter_SW_grt_hyph with T tid (R' SP, TCB_Status tcb, pc + 1); auto.
  unfold WF_currTID in H16. InvertAll. auto.

  (* tid <> tid' *)
  assert (WQMod.not_in_dom tid' RdyQ').
  unfold SW_Post_aux in H40.
  InvertAll. auto.
  assert (WQMod.in_dom tid RdyQ' /\ WQMod.in_dom tid' RdyQ).
  apply enter_SW_grt2_hyp0; auto.
  inversion_clear H44.
  assert (WF_ReadyQ Psi'0 T RdyQ TM). auto.
  unfold WF_ReadyQ in H23.
  assert (exists tcb', TQMod.maps_to T tid' tcb' /\
          (TCB_Status tcb' = READY \/ TCB_Status tcb' = IDLE) /\
          (exists M : Mem,
             ThrdMemMod.maps_to TM tid' M /\
             (TCB_SP tcb' |~>w TCB_PC tcb') M /\
             (WF_Rdy_Thrd Psi'0 M tcb' \/ WF_Rdy_Thrd_IH Psi'0 M tcb'))).
  apply H23. auto.
  InvertAll. rename x into tcb''.
  rename x0 into cthMR.

  induction tcb'' as ((sp'', status''), pc'').
  exists pc''.
  assert (tcb' = (sp'', status'', pc'') /\ sp' = sp'' /\ S_RF S0 SP = sp').
  unfold SW_Post_aux in H40.
  InvertAll.
  assert (tcb' = (sp'', status'', pc'')).
  assert (TQMod.maps_to T tid' tcb').
  unfold TQMod.maps_to in H40.
  unfold TQMod.put in H40. induction (TQMod.beq_A tid' tid).
  absurd (tid' <> tid); auto. unfold TQMod.maps_to. auto.
  unfold TQMod.maps_to in H55. unfold TQMod.maps_to in H47.
  rewrite H55 in H47. inversion H47. auto.
  split; auto.
  split; auto.
  rewrite H55 in H53. auto.
  inversion_clear H51. inversion_clear H54.

  unfold TCB_PC in H49. unfold TCB_SP in H49.
  split.
  apply enter_SW_grt2_hyp1 with TM tid tid' cthMR sp'' M0' M2' M22 M12 cthM2 cthM12; auto.
  rewrite <- H51. auto.
  inversion_clear H52.

  (* not in interrupt handler *)
  unfold WF_Rdy_Thrd in H54.
  InvertAll.
  rename x into rho'. rename x0 into p'. rename x1 into g'.
  rename x2 into tag'. clear H56.
  exists (packS0 rho' (Ast p' g' tag')).
  split; auto.
  intros.
  unfold Itp0.
  assert (D rho' = SomeT1 _ (Itp_AIM rho')).
  apply non_ih_spec_miniD; auto.
  rewrite H58. clear H58.
  unfold Itp_AIM.
  exists (TQMod.put T tid (R' SP, TCB_Status tcb, pc + 1)).
  set (T' := TQMod.put T tid (R' SP, TCB_Status tcb, pc + 1)).
  exists B. exists RdyQ'. exists tid'.
  clear H23.


  assert (exists INVM, exists cthM0, exists M21', exists M122, exists cthM_old, exists cthM_oldStk,
        MemMod.disjUnion INVM cthM0 cthM12 
        /\ INV0 INVM
        /\ (embed_curr_TCB T' tid' *^* embed_RdyQ T' RdyQ') M21'
        /\ MemMod.disjUnion M122 cthMR M12
        /\ MemSplit (ThrdMemMod.remove (ThrdMemMod.remove TM tid) tid') M122
        /\ MemMod.disjUnion cthM0 cthM_oldStk cthM_old
        /\ (R' SP |-> (pc+1) *^* R' SP + 1 |-> junk1 *^* stkSpacePrec (R' SP) 8) cthM_oldStk
        /\ disjUnionList (M22 :: cthMR :: INVM :: M21' :: cthM_old :: cthM2 :: M122 :: nil) (S_Mem S')).
  apply enter_SW_grt2_hyp2_ih with T B sp' tcb tcb' pc' S0 M0' M2'; auto.
  InvertAll.
  rename x into INVM. rename x0 into cthM0. rename x1 into M21'. rename x2 into M122.
  rename x3 into cthM_old. rename x4 into cthM_oldStk.
  unfold pmstar.
  assert (
      exists M'0, exists M'1, 
        MemMod.disjUnion M'0 M'1 (S_Mem S')
        /\ MemMod.disjUnion M21' M22 M'1
        /\ disjUnionList (cthMR :: INVM :: cthM_old :: cthM2 :: M122 :: nil) M'0
  ).
  apply enter_SW_grt2_hyp3_ih; auto.
  InvertAll. rename x into M'0. rename x0 into M'1.
  exists M'0. exists M'1. split; auto.
  split.
  unfold WF_AIM_World.
  split.
  apply enter_SW_grt2_hyp4 with T RdyQ tid (R' SP, TCB_Status tcb, pc + 1) (sp'', status'', pc'') rho_ih; 
  auto.

  split.

  apply enter_SW_grt2_hyp5 with T RdyQ tid (R' SP) (pc+1) tcb; auto.
  split.
  apply enter_SW_grt2_aux6 with RdyQ tid tid' tcb (R' SP, TCB_Status tcb, pc + 1) T rho_ih; auto.
  split; auto.
  exists (ThrdMemMod.put (ThrdMemMod.put TM tid' (MemMod.merge (MemMod.merge cthMR INVM) cthM2)) tid cthM_old).
  set (TM' := ThrdMemMod.put (ThrdMemMod.put TM tid' (MemMod.merge (MemMod.merge cthMR INVM) cthM2)) tid cthM_old).
  exists (MemMod.merge (MemMod.merge cthMR INVM) cthM2). set (cthM' := (MemMod.merge (MemMod.merge cthMR INVM) cthM2)).
  split. assert (S_Mem (S_Mem_Upd S' M'0) = M'0). auto.
  rewrite H68.
  apply enter_SW_grt2_hyp7_ih with cthMR INVM cthM_old M122 TM tid tid' cthM2; auto.
  split. unfold cthM'. unfold TM'.
  unfold ThrdMemMod.put. unfold ThrdMemMod.maps_to.
  induction (ThrdMemMod.beq_A tid' tid).
  absurd (tid' <> tid); auto. induction (ThrdMemMod.beq_A tid' tid').
  auto. absurd (tid' = tid'); auto.
  split.
  unfold WF_Curr_Thrd.
  assert (beq_nat rho' rho_ih = false).
  apply neq_beq_false. inversion_clear H52. rewrite H68. discriminate.
  rewrite H68. discriminate. rewrite H68.
  unfold itp_NH_theta.
  assert (S_ISR S' = false).
  apply enter_SW_grt2_hyp8 with pc pc' M R fl isr M' R' fl' isr' S0 _SWITCH; auto.

  split. 
  UnfoldStateAll. auto.
  assert ((INV0 *^* INV1 -* itp_NH_theta rho' (Ast p' g' tag') Psi'0)
          (cthMR, S_RF S', S_Flag S', false)).
  apply H59.
  rewrite H51 in H55.
  apply enter_SW_grt2_hyp9 with S0 pc'; auto.
  assert (S_ISR (M, R, fl, isr) = S_ISR S' /\ S_Flag (M, R, fl, isr) _IF = S_Flag S' _IF).
  apply enter_SW_grt_hyp9 with pc pc' _SWITCH (M', R', fl', isr') S0; auto.
  inversion_clear H71.
  rewrite <- H73. unfold S_Flag. auto.
  assert (itp_NH_theta rho' (Ast p' g' tag') Psi'0 (cthM', S_RF S', S_Flag S', false)).
  unfold cthM'.
  unfold pm_mwand in H71.
  assert ((MemMod.merge (MemMod.merge cthMR INVM) cthM2, S_RF S', S_Flag S', false) =
           S_Mem_Upd (cthMR, S_RF S', S_Flag S', false) (MemMod.merge (MemMod.merge cthMR INVM) cthM2)).
  auto.
  rewrite H72. apply H71 with (MemMod.merge INVM cthM2); auto.
  unfold S_Mem.
  apply enter_SW_grt2_hypa_ih with M22 M21' cthM_old M122 S'; auto.
  unfold mstar. exists INVM. exists cthM2.
  split; auto.
  unfold MemMod.disjUnion. split; auto.
  apply enter_SW_grt2_hypa_ih2 with cthMR M22 M21' cthM_old M122 S'; auto.
  assert (S_Mem_Upd (S_Mem_Upd S' M'0) cthM' = (cthM', S_RF S', S_Flag S', false)).
  induction S' as (((M'', R''), fl''), isr'').
  UnfoldStateAll.
  rewrite H70. auto. rewrite H73.
  clear H73. unfold itp_NH_theta in H72.
  inversion_clear H72. auto.

  split.
  unfold WF_ReadyQ.
  intros.
  assert (tid0 = tid \/ tid0 <> tid). apply nat_eq_neq_dec.
  inversion_clear H70.
  exists (R' SP, TCB_Status tcb, pc + 1). rewrite H71.
  split; auto. unfold T'.
  unfold TQMod.put. unfold TQMod.maps_to.
  induction (TQMod.beq_A tid tid). auto. absurd (tid = tid); auto.
  split.
  unfold TCB_Status at 1. unfold TCB_Status at 2.
  unfold WF_currTID in H16. InvertAll.
  unfold TQMod.maps_to in H74. rewrite H74 in H38.
  inversion H38. rewrite <- H77. auto.
  exists cthM_old.
  split. unfold TM'.
  unfold ThrdMemMod.put at 1. unfold ThrdMemMod.maps_to.
  induction (ThrdMemMod.beq_A tid tid); auto.
  absurd (tid = tid); auto.
  split.
  unfold TCB_SP. unfold TCB_PC.
  apply enter_SW_grt2_hypb with cthM0 cthM_oldStk junk1; auto.
  right.
  unfold WF_Rdy_Thrd_IH.
  assert (exists p', exists g',
         SCAP_lookupCT psi (pc+1) (Ast p' g' tag)
         /\ (forall pc', pimp (p |> (wrap_fun ih_gy pc _SWITCH pc')) p')
         /\ (forall pc', gimp (p >o (wrap_fun ih_gy pc _SWITCH pc') oo g') g)).
  apply call_sw_nextPC_spec_ih; auto.
  InvertAll. rename x into p''. rename x0 into g''.
  exists p''. exists g''. exists tag.
  split. apply psi_sw_lookupNPC_ih with psi Psi Psi'; auto.
  split. apply enter_SW_grt2_hypb with cthM0 cthM_oldStk junk1; auto.
  intros.
  unfold pm_mwand. unfold S_Mem.
  intros.
  apply weaken_itp_IH_theta with p g (cthM, R, fl, isr); auto.
  unfold WF_Curr_Thrd in H22. simpl in H22.
  InvertAll.
  unfold TQMod.maps_to in H78. rewrite H38 in H78. inversion H78.
  rewrite H22. UnfoldStateAll. auto.
  intro. clear H78.

  apply enter_SW_grt2_auxc_ih with cthM1 cthM2 cthM R fl isr cthM11 cthM12 INVM cthM0 cthM_oldStk pc junk1
        M0 p M M' R' fl' isr'; auto.

(* IH_ItpG *)
  intros.
  unfold gimp in H74.
  unfold IH_ItpG.
  unfold IH_ItpG in H78.
  InvertAll. rename x into M'20. rename x0 into M'21.
  rename x1 into SM1. rename x2 into SM2.
  exists cthM1. exists cthM2. exists SM1. exists SM2.
  split; auto.
  split; auto.
  split; auto.
  split; auto.
  apply H74 with pc'.
  unfold conjG.
  exists (S_Mem_Upd (S_Mem_Upd (cthM_old, R0, fl0, false) M'2) M'20).
  split; auto.
  unfold conjPG.
  split; auto.
  unfold mstar in H77. InvertAll.
  rename x into M01. rename x0 into M02.
  assert (M'20 = MemMod.merge M01 cthM_old).
  apply enter_SW_grt2_hype_ih with M'2 M'21 M0 M02; auto.
  assert (S_Mem_Upd (cthM, R, fl, isr) cthM1 = (cthM1, R, fl, isr)).
  UnfoldStateAll. auto. rewrite H86. clear H86.
  assert (S_Mem_Upd (S_Mem_Upd (cthM_old, R0, fl0, false) M'2) M'20 = (M'20, R0, fl0, false)).
  UnfoldStateAll. auto. rewrite H86. clear H86.
  rewrite H84.
  apply enter_SW_grt2_auxc_ih_aux4 with M M' R' fl' isr' cthM11 cthM12 INVM cthM0 M0 M02 cthM_oldStk junk1 M'2; auto.

  
 (* tid0 <> tid : other ready threads *)
  unfold WF_ReadyQ in H44.
  assert (exists tcb : Thread,
          TQMod.maps_to T tid0 tcb /\
          (TCB_Status tcb = READY \/ TCB_Status tcb = IDLE) /\
          (exists M : Mem,
             ThrdMemMod.maps_to TM tid0 M /\
             (TCB_SP tcb |~>w TCB_PC tcb) M /\
             (WF_Rdy_Thrd Psi'0 M tcb \/ WF_Rdy_Thrd_IH Psi'0 M tcb))).
  apply H44; auto.

  apply enter_SW_grt2_hypf with RdyQ' tid tid'; auto.
  InvertAll.
  exists x.
  split. unfold T'.
  unfold TQMod.put. unfold TQMod.maps_to.
  induction (TQMod.beq_A tid0 tid).
  absurd (tid0 <> tid); auto. unfold TQMod.maps_to in H69.
  auto. split; auto.
  exists x0.
  split; auto.
  unfold TM'.
  unfold ThrdMemMod.put at 1. unfold ThrdMemMod.maps_to.
  induction (ThrdMemMod.beq_A tid0 tid). absurd (tid0 <> tid); auto.
  unfold ThrdMemMod.put.
  induction (ThrdMemMod.beq_A tid0 tid').
  rewrite <- a in H39.
  absurd (~ WQMod.in_dom tid0 RdyQ'); auto with PMap.
  unfold ThrdMemMod.maps_to in H73. auto.
  apply enter_SW_grt2_hypg with T TM RdyQ RdyQ' tid tid' (R' SP, TCB_Status tcb, pc + 1) cthM_old 
            (MemMod.merge (MemMod.merge cthMR INVM) cthM2) rho_ih; auto.
  apply enter_SW_grt2_hyph with M21' M22 T (R' SP, TCB_Status tcb, pc + 1) RdyQ tid rho_ih; auto.

  (* tid' is in interrupt hander *)
  unfold WF_Rdy_Thrd_IH in H54.
  InvertAll.
  rename x into p'. rename x0 into g'.
  rename x1 into tag'.
  exists (packS0 rho_ih (Ast p' g' tag')).
  split; auto.
  intros.
  unfold Itp0.
  assert (D rho_ih = SomeT1 _ (Itp_AIM rho_ih)).
  unfold Minimal_D in H0. InvertAll. auto.
  rewrite H58. clear H58.
  unfold Itp_AIM.
  exists (TQMod.put T tid (R' SP, TCB_Status tcb, pc + 1)).
  set (T' := TQMod.put T tid (R' SP, TCB_Status tcb, pc + 1)).
  exists B. exists RdyQ'. exists tid'.
  clear H23.

  assert (exists INVM, exists cthM0, exists M21', exists M122, exists cthM_old, exists cthM_oldStk,
        MemMod.disjUnion INVM cthM0 cthM12 
        /\ INV0 INVM
        /\ (embed_curr_TCB T' tid' *^* embed_RdyQ T' RdyQ') M21'
        /\ MemMod.disjUnion M122 cthMR M12
        /\ MemSplit (ThrdMemMod.remove (ThrdMemMod.remove TM tid) tid') M122
        /\ MemMod.disjUnion cthM0 cthM_oldStk cthM_old
        /\ (R' SP |-> (pc+1) *^* R' SP + 1 |-> junk1 *^* stkSpacePrec (R' SP) 8) cthM_oldStk
        /\ disjUnionList (M22 :: cthMR :: INVM :: M21' :: cthM_old :: cthM2 :: M122 :: nil) (S_Mem S')).
  apply enter_SW_grt2_hyp2_ih with T B sp' tcb tcb' pc' S0 M0' M2'; auto.
  InvertAll.
  rename x into INVM. rename x0 into cthM0. rename x1 into M21'. rename x2 into M122.
  rename x3 into cthM_old. rename x4 into cthM_oldStk.
  unfold pmstar.
  assert (
      exists M'0, exists M'1, 
        MemMod.disjUnion M'0 M'1 (S_Mem S')
        /\ MemMod.disjUnion M21' M22 M'1
        /\ disjUnionList (cthMR :: INVM :: cthM_old :: cthM2 :: M122 :: nil) M'0
  ).
  apply enter_SW_grt2_hyp3_ih; auto.
  InvertAll. rename x into M'0. rename x0 into M'1.
  exists M'0. exists M'1. split; auto.
  split.
  unfold WF_AIM_World.
  split.
  apply enter_SW_grt2_hyp4 with T RdyQ tid (R' SP, TCB_Status tcb, pc + 1) (sp'', status'', pc'') rho_ih; 
  auto.

  split.
  apply enter_SW_grt2_hyp5 with T RdyQ tid (R' SP) (pc+1) tcb; auto.
  split.
  apply enter_SW_grt2_aux6 with RdyQ tid tid' tcb (R' SP, TCB_Status tcb, pc + 1) T rho_ih; auto.
  split; auto.
  exists (ThrdMemMod.put (ThrdMemMod.put TM tid' (MemMod.merge (MemMod.merge cthMR INVM) cthM2)) tid cthM_old).
  set (TM' := ThrdMemMod.put (ThrdMemMod.put TM tid' (MemMod.merge (MemMod.merge cthMR INVM) cthM2)) tid cthM_old).
  exists (MemMod.merge (MemMod.merge cthMR INVM) cthM2). set (cthM' := (MemMod.merge (MemMod.merge cthMR INVM) cthM2)).
  split. assert (S_Mem (S_Mem_Upd S' M'0) = M'0). auto.
  rewrite H67.
  apply enter_SW_grt2_hyp7_ih with cthMR INVM cthM_old M122 TM tid tid' cthM2; auto.
  split. unfold cthM'. unfold TM'.
  unfold ThrdMemMod.put. unfold ThrdMemMod.maps_to.
  induction (ThrdMemMod.beq_A tid' tid).
  absurd (tid' <> tid); auto. induction (ThrdMemMod.beq_A tid' tid').
  auto. absurd (tid' = tid'); auto.
  split.
  unfold WF_Curr_Thrd. simpl.
  assert ((INV0 *^* INV1 -* itp_IH_theta (Ast p' g' tag') status'' Psi'0)
          (cthMR, S_RF S', S_Flag S', false)).
  apply H57.
  rewrite H51 in H55.
  apply enter_SW_grt2_hyp9 with S0 pc'; auto.

  apply enter_SW_grt2_hypI with pc pc' M R fl isr M' R' fl' isr' S0; auto.
  exists (sp'', status'', pc''). exists status''.
  split.
  unfold TQMod.maps_to.
  unfold T'. unfold TQMod.put.
  induction (TQMod.beq_A tid' tid).
  absurd (tid' <> tid); auto. auto.
  split; auto.
  unfold pm_mwand in H67.
  assert (S_ISR S' = false).
  apply enter_SW_grt2_hyp8 with pc pc' M R fl isr M' R' fl' isr' S0 _SWITCH; auto.
  assert (itp_IH_theta (Ast p' g' tag') status'' Psi'0
          (S_Mem_Upd (cthMR, S_RF S', S_Flag S', false) cthM')).
  apply H67 with (MemMod.merge cthM2 INVM); auto.
  unfold S_Mem.
  apply enter_SW_grt2_hypJ_ih with M22 M21' cthM_old M122 S'; auto.
  unfold mstar.

  exists INVM. exists cthM2.
  split; auto.
  assert (MemMod.disjUnion cthM2 INVM (MemMod.merge cthM2 INVM)).
  unfold MemMod.disjUnion. split; auto.
  apply enter_SW_grt2_hypK_ih with cthMR M22 M21' cthM_old M122 S'; auto.
  auto with PMap.
  UnfoldStateAll. rewrite H69. auto.

  split.
  unfold WF_ReadyQ.
  intros.
  assert (tid0 = tid \/ tid0 <> tid). apply nat_eq_neq_dec.
  inversion_clear H69.
  exists (R' SP, TCB_Status tcb, pc + 1). rewrite H70.
  split; auto. unfold T'.
  unfold TQMod.put. unfold TQMod.maps_to.
  induction (TQMod.beq_A tid tid). auto. absurd (tid = tid); auto.
  split.
  unfold TCB_Status at 1. unfold TCB_Status at 2.
  unfold WF_currTID in H16. InvertAll.
  unfold TQMod.maps_to in H73. rewrite H73 in H38.
  inversion H38. rewrite <- H76. auto.
  exists cthM_old.
  split. unfold TM'.
  unfold ThrdMemMod.put at 1. unfold ThrdMemMod.maps_to.
  induction (ThrdMemMod.beq_A tid tid); auto.
  absurd (tid = tid); auto.
  split.
  unfold TCB_SP. unfold TCB_PC.
  apply enter_SW_grt2_hypb with cthM0 cthM_oldStk junk1; auto.
  right.
  unfold WF_Rdy_Thrd_IH.
  assert (exists p', exists g',
         SCAP_lookupCT psi (pc+1) (Ast p' g' tag)
         /\ (forall pc', pimp (p |> (wrap_fun ih_gy pc _SWITCH pc')) p')
         /\ (forall pc', gimp (p >o (wrap_fun ih_gy pc _SWITCH pc') oo g') g)).
  apply call_sw_nextPC_spec_ih; auto.
  InvertAll. rename x into p''. rename x0 into g''.
  exists p''. exists g''. exists tag.
  split. apply psi_sw_lookupNPC_ih with psi Psi Psi'; auto.
  split. apply enter_SW_grt2_hypb with cthM0 cthM_oldStk junk1; auto.
  intros.
  unfold pm_mwand. unfold S_Mem.
  intros.
  apply weaken_itp_IH_theta with p g (cthM, R, fl, isr); auto.
  unfold WF_Curr_Thrd in H22. simpl in H22.
  InvertAll.
  unfold TQMod.maps_to in H77. rewrite H38 in H77. inversion H77.
  rewrite H22. UnfoldStateAll. auto.
  intro. clear H77.

  apply enter_SW_grt2_auxc_ih with cthM1 cthM2 cthM R fl isr cthM11 cthM12 INVM cthM0 cthM_oldStk pc junk1
        M0 p M M' R' fl' isr'; auto.



(* IH_ItpG *)
  intros.
  unfold gimp in H73.
  unfold IH_ItpG.
  unfold IH_ItpG in H77.
  InvertAll. rename x into M'20. rename x0 into M'21.
  rename x1 into SM1. rename x2 into SM2.
  exists cthM1. exists cthM2. exists SM1. exists SM2.
  split; auto.
  split; auto.
  split; auto.
  split; auto.
  apply H73 with pc'.
  unfold conjG.
  exists (S_Mem_Upd (S_Mem_Upd (cthM_old, R0, fl0, false) M'2) M'20).
  split; auto.
  unfold conjPG.
  split; auto.
  unfold mstar in H76. InvertAll.
  rename x into M01. rename x0 into M02.
  assert (M'20 = MemMod.merge M01 cthM_old).
  apply enter_SW_grt2_hype_ih with M'2 M'21 M0 M02; auto.
  assert (S_Mem_Upd (cthM, R, fl, isr) cthM1 = (cthM1, R, fl, isr)).
  UnfoldStateAll. auto. rewrite H85. clear H85.
  assert (S_Mem_Upd (S_Mem_Upd (cthM_old, R0, fl0, false) M'2) M'20 = (M'20, R0, fl0, false)).
  UnfoldStateAll. auto. rewrite H85. clear H85.
  rewrite H83.
  apply enter_SW_grt2_auxc_ih_aux4 with M M' R' fl' isr' cthM11 cthM12 INVM cthM0 M0 M02 cthM_oldStk junk1 M'2; auto.

  
 (* tid0 <> tid : other ready threads *)
  unfold WF_ReadyQ in H44.
  assert (exists tcb : Thread,
          TQMod.maps_to T tid0 tcb /\
          (TCB_Status tcb = READY \/ TCB_Status tcb = IDLE) /\
          (exists M : Mem,
             ThrdMemMod.maps_to TM tid0 M /\
             (TCB_SP tcb |~>w TCB_PC tcb) M /\
             (WF_Rdy_Thrd Psi'0 M tcb \/ WF_Rdy_Thrd_IH Psi'0 M tcb))).
  apply H44; auto.

  apply enter_SW_grt2_hypf with RdyQ' tid tid'; auto.
  InvertAll.
  exists x.
  split. unfold T'.
  unfold TQMod.put. unfold TQMod.maps_to.
  induction (TQMod.beq_A tid0 tid).
  absurd (tid0 <> tid); auto. unfold TQMod.maps_to in H69.
  auto. split; auto.
  exists x0.
  split; auto.
  unfold TM'.
  unfold ThrdMemMod.put at 1. unfold ThrdMemMod.maps_to.
  induction (ThrdMemMod.beq_A tid0 tid). absurd (tid0 <> tid); auto.
  unfold ThrdMemMod.put.
  induction (ThrdMemMod.beq_A tid0 tid').
  rewrite <- a in H39.
  absurd (~ WQMod.in_dom tid0 RdyQ'); auto with PMap.
  unfold ThrdMemMod.maps_to in H73. auto.
  apply enter_SW_grt2_hypg with T TM RdyQ RdyQ' tid tid' (R' SP, TCB_Status tcb, pc + 1) cthM_old 
            (MemMod.merge (MemMod.merge cthMR INVM) cthM2) rho_ih; auto.
  apply enter_SW_grt2_hyph with M21' M22 T (R' SP, TCB_Status tcb, pc + 1) RdyQ tid rho_ih; auto.
Qed.


Lemma Itp_AIM_sw_ih_disable:
  forall theta Psi S,
    Itp_AIM rho_ih theta Psi S -> S_Flag S _IF = false.
Proof.
  intros.
  unfold Itp_AIM in H.
  InvertAll.
  unfold pmstar in H. InvertAll.
  unfold WF_AIM_World in H.
  InvertAll.
  unfold WF_Curr_Thrd in H7.
  assert (beq_nat rho_ih rho_ih = true). simpl. auto.
  rewrite H9 in H7.
  unfold itp_IH_theta in H7.
  InvertAll.
  UnfoldStateAll. auto.
Qed.

Lemma enter_SW_disable_ih:
forall pc theta Psi'0 S S',
  Itp_AIM rho_ih theta Psi'0 S
  -> nextS pc (comml (_call _SWITCH)) S S'
  -> S_Flag S' _IF = false.
Proof.
  intros.
  assert (S_Flag S _IF = false).
  apply Itp_AIM_sw_ih_disable with theta Psi'0; auto.
  assert ((S_Flag S _IF = S_Flag S' _IF) /\ (S_ISR S = S_ISR S')).
  apply  enter_SW_grt_auxb0_hyp4 with pc _SWITCH; auto.
  inversion H2.
  rewrite <- H3. auto.
Qed.
  


Lemma enter_SW_ih:
forall D psi pc theta D' Psi Psi' Psi'0 a S S',
  IH_WFcmd psi pc theta (comml (_call _SWITCH))
  -> Minimal_D D
  -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
  -> lift_psi psi rho_ih Psi
  -> Psi' = ext_Prim_Specs Psi
  -> a = Itp_AIM rho_ih theta
  -> subsetCT Psi' Psi'0
  -> a Psi'0 S
  -> nextS pc (comml (_call _SWITCH)) S S'
  -> Itp0 D' SWITCH_Spec0 Psi'0 S'.
Proof.
  intros.
  unfold Itp0.
  unfold SWITCH_Spec0.
  rewrite H1.
  simpl.
  rewrite H4 in H6.
  clear H4 a.
  split.
  apply enter_SW_pre_ih with psi pc theta Psi'0 S; auto.
  split.

  apply enter_SW_disable_ih with pc theta Psi'0 S; auto.
  exists 0.
  apply enter_SW_ih_grt with psi pc theta D' Psi Psi' S; auto.
Qed.


Lemma IH_SW_Presv:
  forall D psi pc theta D' Psi Psi' Psi'0 a S S' f',
    IH_WFcmd psi pc theta (comml (_call _SWITCH))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_ih Psi
    -> Psi' = ext_Prim_Specs Psi
    -> a = Itp_AIM rho_ih theta
    -> subsetCT Psi' Psi'0
    -> a Psi'0 S
    -> nextS pc (comml (_call _SWITCH)) S S'
    -> nextPC (comml (_call _SWITCH)) (S_Mem S) (S_RF S) (S_Flag S) pc f'
    -> exists theta0 : Spec0, lookupCT Psi'0 f' theta0 /\ Itp0 D' theta0 Psi'0 S'.
Proof.
  intros.
  assert (lookupCT Psi'0 _SWITCH SWITCH_Spec0).
  apply subsetCT_presv_lookupCT with Psi'; auto.
  rewrite H3. auto.
  inversion H8. clear H14 H13 H12 H10 H11 fl R M f pc0.
  exists SWITCH_Spec0.
  split; auto.
  apply enter_SW_ih with D psi pc theta Psi Psi' a S; auto.
Qed.



Lemma IH_SW_Itp :
  forall D psi pc theta D' Psi Psi',
    IH_WFcmd psi pc theta (comml (_call _SWITCH))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_ih Psi
    -> Psi' = ext_Prim_Specs Psi
    (* -> a = Itp_AIM rho_ready theta *)
    -> WFcmd D' Psi' (packS0 rho_ih theta) pc (comml (_call _SWITCH)).
Proof.
  intros.
  unfold WFcmd.
  split.
  unfold WFcmd_aux.
  intros.
  assert (Itp_AIM rho_ih theta Psi'0 S).
  unfold Itp0 in H4.
  rewrite H1 in H4.
  unfold ext_dict in H4.
  assert (beq_nat rho_ih rho_prim = false). simpl. auto.
  rewrite H6 in H4.
  unfold Minimal_D in H0.
  InvertAll.
  rewrite H9 in H4. auto.
  split.
  apply IH_SW_progress with psi theta Psi'0 (Itp_AIM rho_ih theta); auto.
  split.
  intros.
  apply IH_SW_Presv with D psi pc theta Psi Psi' (Itp_AIM rho_ih theta) S; auto.
  left.

  apply call_sw_disable_itr_ih with psi Psi'0 pc theta; auto.

  rewrite H1.
  unfold Itp0.
  unfold ext_dict.
  assert (beq_nat rho_ih rho_prim = false).
  simpl. auto.
  rewrite H4.
  unfold Minimal_D in H0. InvertAll.
  rewrite H7.
  apply Itp_AIM_Mono_CT.
Qed.
