Add LoadPath "..".
Add LoadPath "../CAP0".

Require Import Arith.
Require Import Coqlib.
Require Import Map.
Require Import Machine.
Require Import Mem_Lemmas.
Require Import Assertion.
Require Import Assertion_Lemmas.
Require Import Assertion_Lemmas4.
Require Import Assertion_Lemmas3.
Require Import Operational_Semantics.
Require Import OpSem_Lemmas.
Require Import OpSem_Lemmas2.
Require Import Reg.
Require Import CAP0.
Require Import SCAP.
Require Import AIM.
Require Import QLib.
Require Import SCAP_READY.
Require Import SCAP_IDLE.
Require Import SCAP_Itr.
Require Import Linking.

Require Import EqNat.
Require Import Omega.
Require Import List.
Require Import Arith_Lemmas.


Lemma embed_WQSet_inv_cons : 
  forall T B l Q M,
    embed_WQSet T (WQSetMod.put B l Q) M
    -> WQSetMod.not_in_dom l B
    -> exists M1, exists M2, 
      MemMod.disjUnion M1 M2 M
      /\ embed_WQSet T B M1
      /\ ptoTQ l (filterTQ T Q) M2.
Proof.
  intros T B' l Q M H.
  generalize (refl_equal ( WQSetMod.put B' l Q)).
  pattern B' at 2.
  set (B := (WQSetMod.put B' l Q)) in * .
  simpl.
  generalize B B' T l Q M H.
  subst B.
  clear T B' l Q M H.
  intros B B' T l Q M H.
  generalize B' l Q; clear B' l Q.
  induction H.
 
  intros.
  apply False_ind.
  assert (WQSetMod.emp l = WQSetMod.put B' l Q l).
    rewrite H; trivial.
  unfold WQSetMod.emp, WQSetMod.put in H1.
  destruct (WQSetMod.beq_A l l).
  discriminate.
  destruct (n (refl_equal _)).

  intros B' l' Q' H3 H4.
  destruct (WQSetMod.beq_A l' l).
  subst l'.
  assert (Q = Q').
    assert ((WQSetMod.put B l Q) l = (WQSetMod.put B' l Q') l).
      rewrite H3; trivial.
    unfold WQSetMod.put in H5.
    destruct (WQSetMod.beq_A l l).
    injection H5; trivial.
    destruct (n (refl_equal _)).
  subst Q'.
  assert (B' = B).
    apply WQSetMod.extensional_eqMap.
    intros x.
    assert (WQSetMod.put B l Q x = WQSetMod.put B' l Q x).
      rewrite H3; trivial.
    unfold WQSetMod.put in H5.
    destruct (WQSetMod.beq_A x l).
      subst x.
      unfold WQSetMod.not_in_dom in * .
      rewrite H0; rewrite H4; trivial.
      rewrite H5; trivial.
  subst B'.
  exists M; exists M'.
  split_l.
  split_lr.

  assert (B = WQSetMod.put (WQSetMod.remove B' l) l' Q').
    rewrite <- WQSetMod.remove_ext_ext_remove; trivial.
    rewrite <- H3.
    apply WQSetMod.extensional_eqMap.
    unfold WQSetMod.not_in_dom in H0.
    unfold WQSetMod.put, WQSetMod.remove.
    intro x.
    destruct (WQSetMod.beq_A x l); trivial.
    subst x; trivial.
  assert (IH := IHembed_WQSet _ _ _ H5).
  assert (Hx : WQSetMod.not_in_dom l' (WQSetMod.remove B' l)).
    apply WQSetMod.remove_shrinks_dom; trivial.
  substH IH with (IH Hx).
  clear IHembed_WQSet.
  destruct IH as [M1 [M2 [HJM [HM1 HM2]]]].
  exists (MemMod.merge M1 M'); exists M2.
  split.
    disjUnion shift_r.
    merge commut M2.
    disjUnion shift_l.
    elim HJM; intros; subst; trivial.
  assert (Hy : WQSetMod.maps_to B' l Q).
    assert ((WQSetMod.put B l Q) l = (WQSetMod.put B' l' Q') l).
      rewrite H3; trivial.
    unfold WQSetMod.maps_to.
    unfold WQSetMod.put in H6.
    destruct (WQSetMod.beq_A l l').
      subst l'; destruct (n (refl_equal _)).
    destruct (WQSetMod.beq_A l l).
      auto.
    destruct (n1 (refl_equal _)).
  split_r.
  assert (Hz : B' = WQSetMod.put (WQSetMod.remove B' l) l Q).
    rewrite WQSetMod.put_cancel_remove.
    unfold WQSetMod.maps_to, WQSetMod.put in * .
    apply WQSetMod.extensional_eqMap.
    intros x.
    destruct (WQSetMod.beq_A x l); trivial.
    subst x; trivial.
  rewrite Hz.
  apply wq_cons with M1 M'; auto.
  apply WQSetMod.nid_remove.
  split_r.
  disjoint lift_l.
Qed.

Lemma memsplit_subheap:
  forall TM tid M0 M,
    ThrdMemMod.maps_to TM tid M0 -> MemSplit TM M -> MemMod.subseteq M0 M.
Proof.
  intros.
  unfold ThrdMemMod.maps_to in H.
  induction H0.
  unfold ThrdMemMod.emp in H.
  inversion H.
  unfold ThrdMemMod.put in H.
  destruct (ThrdMemMod.beq_A tid tid0).
  inversion H.
  rewrite <- H4.
  apply MemMod.disjUnion_subseteq_right with M';assumption.
  substH IHMemSplit with (IHMemSplit H).
  apply MemMod.subseteq_transitivity with M'.
  assumption.
  apply MemMod.disjUnion_subseteq_left with M'';assumption.
Qed.

Lemma neq_beq_false:
  forall x y, x <> y -> beq_nat x y = false.
Proof.
  double induction x y.
  intro H;elim H;reflexivity.
  intros.
  simpl;reflexivity.
  intros.
  simpl;reflexivity.
  intros.
  simpl;apply H0.
  intro P;apply H1;rewrite P;reflexivity.
Qed.

Lemma WF_Curr_Thrd_pinv:
  forall Psi S rho p g tag T tid,
    WF_Curr_Thrd Psi S rho (Ast p g tag) T tid -> rho <> rho_ih -> NH_pinvstar p S.
Proof.
  intros.
  unfold WF_Curr_Thrd in H.
  assert (beq_nat rho rho_ih = false).
  apply neq_beq_false; auto.
  rewrite H1 in H.
  unfold itp_NH_theta in H. InvertAll. auto.
Qed.

Lemma WF_AIM_World_pinv:
  forall S Psi T B RdyQ tid rho p g tag,
    rho <> rho_ih -> WF_AIM_World Psi S T B RdyQ tid rho (Ast p g tag) -> imprec (NH_pinvstar p) S.
Proof.
  intros.
  unfold WF_AIM_World in H0.
  InvertAll.
  assert (MemMod.subseteq x0 (S_Mem S)).
  apply memsplit_subheap with x tid; auto.
  apply imprec_monotone_S with x0; auto.
  apply imprec_weaker.
  apply WF_Curr_Thrd_pinv with Psi rho g tag T tid; auto.
Qed.


Lemma Itp_AIM_pinv :
  forall p g tag Psi S rho,
    Itp_AIM rho (Ast p g tag) Psi S -> rho <> rho_ih -> imprec (NH_pinvstar p) S.
Proof.
  intros.
  unfold Itp_AIM in H.
  InvertAll.
  assert (imprec (fun S => WF_AIM_World Psi S x x0 x1 x2 rho (Ast p g tag)) S).
  apply pmstar_imprec_p with (embed_curr_TCB x x2 *^* (embed_RdyQ x x1 *^* embed_WQSet x x0)); auto.
  apply imprec_presv_imprec with (fun S : State =>
          WF_AIM_World Psi S x x0 x1 x2 rho (Ast p g tag)); auto.
  intros.
  apply WF_AIM_World_pinv with Psi x x0 x1 x2 rho g tag; auto.
Qed.

Lemma pinv_imp_p :
  forall p S, (NH_pinvstar p) S ->  (imprec p) S.
Proof.
  intros.
  unfold NH_pinvstar.
  apply pmstar_imprec_p with (NH_Inv (S_Flag S)); auto.
Qed.
Hint Resolve pinv_imp_p.



Lemma pflag_monotonic:
  forall b fl, monotonic_p (fun S => S_Flag S fl = b).
Proof.
  intro.
  unfold monotonic_p.
  intros.
  induction S. induction a. induction a.
  UnfoldStateAll. auto.
Qed.
Hint Resolve pflag_monotonic.

Lemma pisr_monotonic:
  forall b, monotonic_p (fun S => S_ISR S = b).
Proof.
  intro.
  unfold monotonic_p.
  induction S as (((M, R), fl), isr).
  UnfoldStateAll. auto.
Qed.
Hint Resolve pisr_monotonic.

Lemma ptrue_monotonic:
  forall m, monotonic_p (pTrue *<* m).
Proof.
  unfold monotonic_p, pTrue, pmstar.
  intros m S M H H0.
  destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
  simplS.
  destruct (MemMod.subseteq_split H0) as [M3 H].
  exists (MemMod.merge M1 M3); exists M2.
  split.
  elim H; intros.
  elim HJM; intros.
  split.
    disjoint des.
    disjoint lift_r.
  subst M; rewrite H4.
  merge assoc M3.
  merge assoc M2.
  merge commut M2.
  trivial.
  auto.
Qed.



Lemma monotonic_conj:
  forall p q, monotonic_p p -> monotonic_p q -> monotonic_p (fun S => p S /\ q S).
Proof.
  unfold monotonic_p.
  intros p q H H0.
  intros S Mx H1 H2.
  destruct H1; split; auto.
Qed.

Lemma stkSpace_ra :
  forall sp n m M, (stkSpacePrec sp n *^* m) M -> n >= 2 -> stkSpace sp 2 M.
Proof.
  intros sp n m M H H0.
  sep on.
  sep stksplit 2 in H.
  apply stkSpacePrec_stkSpace1 with mTrue.
  sep matchH H with 2.
Qed.


Lemma stkSpace_call_prog:
  forall S n pc f, (stkSpacePrec (S_RF S SP) n *^* mTrue) (S_Mem S) -> n >= 2
    -> exists S' : State, nextS pc (comml (_call f)) S S'.
Proof.
  intros.
  induction S. induction a. induction a.
  exists (MemMod.put (MemMod.put a (b1 SP - 2) (pc + 1)) (b1 SP - 2 + 1) 0, RF_update b1 SP (b1 SP-2), b0, b).
  apply nextS_call; auto.
  unfold pushW.
  unfold S_RF in H. unfold S_Mem in H.
  split.
  apply stkSpace_ra with n mTrue; auto.
  split; auto.
  unfold MemUpd16.
  exists 0. auto.
Qed.

Lemma enable_sw_prog:
  forall p S pc f, imprec p S -> enable_sw p
    -> exists S' : State, nextS pc (comml (_call f)) S S'.
Proof.
  intros.
  unfold enable_sw in H0.
  assert (monotonic_p (fun S => (pTrue*<*(INV0 *^* INV1 *^* stkSpacePrec (S_RF S SP) (_SW_Stk + 2))) S)).
  unfold monotonic_p.
  intros.
  induction S0 as (((M0, R0), fl0), isr0). simplS. simplS in H1.
  unfold S_Mem in H2.
  assert (exists M0', MemMod.disjUnion M0 M0' M).
  apply MemMod.subseteq_split; auto.
  unfold pmstar in H1. simplS in H1. InvertAll.
  unfold pmstar. simplS. exists (MemMod.merge x0 x). exists x1.
  unfold pTrue.
  split; auto.
  cut (MemMod.disjUnion x1 (MemMod.merge x0 x) M). auto with PMap.
  apply MemMod.disjUnion_assoc_left_right; auto.
  inversion_clear H3; auto with PMap.
  assert (MemMod.disjUnion x1 x0 M0). auto with PMap.
  inversion H5. rewrite H8 in H4. auto.
  assert (forall S0 : State,
       imprec p S0 ->
       (pTrue *<* (INV0 *^* INV1 *^* stkSpacePrec (S_RF S0 SP) (_SW_Stk + 2)))
         S0).
  apply (imprec_presv_imp_monotonic H1 H0).
  assert ((pTrue *<* (INV0 *^* INV1 *^* stkSpacePrec (S_RF S SP) (_SW_Stk + 2)))
         S).
  apply H2; auto.
  clear H2 H1 H0 H.
  apply stkSpace_call_prog with (_SW_Stk + 2); auto.
  assert (((INV0 *^* INV1 *^* stkSpacePrec (S_RF S SP) (_SW_Stk + 2)) *^* mTrue) (S_Mem S)).
  apply pTrue_m_mTrue with pTrue; auto.
  apply mstar_mTrue with (INV0 *^* INV1) mTrue; auto.
  omega.
Qed.

Lemma NH_SW_prog_aux :
  forall psi pc theta Psi rho S,
    NH_Call_Prim psi pc _SWITCH theta -> Itp_AIM rho theta Psi S -> rho <> rho_ih 
    -> exists S', nextS pc (comml (_call _SWITCH)) S S'.
Proof.
  intros.
  induction theta.
  unfold NH_Call_Prim in H.
  InvertAll.
  inversion H2.
  inversion H4.
  rewrite <- H10 in H6.
  clear H8 H7 H12 H11 H10 H4 H5 H2 p0 p' g0 g' tag H.
  assert (imprec (NH_pinvstar p) S).
  apply Itp_AIM_pinv with g n Psi rho; auto.
  assert (imprec p S).
  apply imprec_presv_imprec with (NH_pinvstar p); auto.
  apply enable_sw_prog with p; auto.
  assert (_SWITCH <> _BLOCK). auto.
  contradiction.
  assert (_SWITCH <> _UNBLK). auto.
  contradiction.
Qed.

Lemma WF_Curr_Thrd_imprecp:
  forall Psi S rho p g tag T tid,
    WF_Curr_Thrd Psi S rho (Ast p g tag) T tid -> imprec p S.
Proof.
  intros.
  unfold WF_Curr_Thrd in H.
  assert (rho = rho_ih \/ rho <> rho_ih).
  apply nat_eq_neq_dec.
  inversion_clear H0.
  rewrite H1 in H. assert (beq_nat rho_ih rho_ih = true). simpl. auto.
  rewrite H0 in H.
  InvertAll.
  unfold itp_IH_theta in H4.
  inversion_clear H4.
  inversion_clear H5.
  unfold lift_IH_PG in H4.
  InvertAll.
  apply pmstar_imprec_p with INV1; auto.
  InvertAll.
  apply pmstar_imprec_p with INV1; auto.
  assert (beq_nat rho rho_ih = false).
  apply neq_beq_false; auto.
  rewrite H0 in H.
  unfold itp_NH_theta in H. InvertAll. 
  unfold NH_pinvstar in H.
  apply pmstar_imprec_p with (NH_Inv (S_Flag S)); auto.
Qed.


Lemma WF_AIM_World_imprecp:
  forall S Psi T B RdyQ tid rho p g tag,
    WF_AIM_World Psi S T B RdyQ tid rho (Ast p g tag) -> imprec p S.
Proof.
  intros.
  unfold WF_AIM_World in H.
  InvertAll.
  assert (MemMod.subseteq x0 (S_Mem S)).
  apply memsplit_subheap with x tid; auto.
  apply imprec_monotone_S with x0; auto.
  apply WF_Curr_Thrd_imprecp with Psi rho g tag T tid; auto.
Qed.


Lemma Itp_AIM_imprecp :
  forall p g tag Psi S rho,
    Itp_AIM rho (Ast p g tag) Psi S -> imprec p S.
Proof.
  intros.
  unfold Itp_AIM in H.
  InvertAll.
  assert (imprec (fun S => WF_AIM_World Psi S x x0 x1 x2 rho (Ast p g tag)) S).
  apply pmstar_imprec_p with (embed_curr_TCB x x2 *^* (embed_RdyQ x x1 *^* embed_WQSet x x0)); auto.
  apply imprec_presv_imprec with (fun S : State =>
          WF_AIM_World Psi S x x0 x1 x2 rho (Ast p g tag)); auto.
  intros.
  apply WF_AIM_World_imprecp with Psi x x0 x1 x2 rho g tag; auto.
Qed.

Lemma call_sw_disable_itr_ih:
forall psi Psi'0 pc theta S,
  IH_WFcmd psi pc theta (comml (_call _SWITCH)) 
  -> Itp_AIM rho_ih theta Psi'0 S
  -> ~ itr_enabled S.
Proof.
  intros.
  intro.
  unfold itr_enabled in H1.
  unfold Itp_AIM in H0.
  InvertAll.
  unfold pmstar in H0.
  InvertAll.
  unfold WF_AIM_World in H0.
  InvertAll.
  unfold WF_Curr_Thrd in H10.
  assert (beq_nat rho_ih rho_ih = true). simpl. auto.
  rewrite H12 in H10.
  InvertAll.
  unfold itp_IH_theta in H16.
  inversion_clear H16.
  induction S as (((M, R), fl), isr).
  UnfoldStateAll.
  rewrite H15 in H2. inversion H2.
Qed.

Lemma subsetCT_presv_lookupCT :
  forall Psi Psi' f theta0,
    subsetCT Psi Psi' -> lookupCT Psi f theta0 -> lookupCT Psi' f theta0.
Proof.
  unfold subsetCT. unfold lookupCT.
  intros.
  apply OPSpec.lookup_sub_lookup with Psi; auto.
Qed.

Lemma get_SW_Spec: forall Psi, lookupCT (ext_Prim_Specs Psi) _SWITCH SWITCH_Spec0.
Proof.
  intro.
  unfold ext_Prim_Specs.
  unfold lookupCT.
  unfold consCT.
  apply OPSpec.lookup_merge_intro.
  right.
  apply OPSpec.lookup_merge_intro.
  right.
  apply OPSpec.lookup_merge_intro. left.
  unfold OPSpec.lookup.
  unfold ListT.In.
  unfold OPSpec.sig_map. left. auto.
Qed.
Hint Resolve get_SW_Spec.

Lemma get_BLK_Spec: forall Psi, lookupCT (ext_Prim_Specs Psi) _BLOCK BLK_Spec0.
Proof.
  intro.
  unfold ext_Prim_Specs.
  unfold lookupCT.
  unfold consCT.
  apply OPSpec.lookup_merge_intro.
  right.
  apply OPSpec.lookup_merge_intro.
  left.
  unfold OPSpec.lookup.
  unfold ListT.In.
  unfold OPSpec.sig_map. left. auto.
Qed.
Hint Resolve get_BLK_Spec.

Lemma enter_SW_grt_auxb0_hyp4:
  forall pc' f S0 S',
    nextS pc' (comml (_call f)) S0 S'
    -> (S_Flag S0 _IF = S_Flag S' _IF) /\ (S_ISR S0 = S_ISR S').
Proof.
  intros pc' f S0 S' H.
  destruct S0 as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion H) as [HM' [HR' [Hfl' Hisr']]].
  simplS.
  subst fl isr; auto.
Qed.

Lemma WF_AIM_World_currTCB:
forall Psi S T B RdyQ tid rho theta,
  WF_AIM_World Psi S T B RdyQ tid
         rho theta -> exists tcb, TQMod.maps_to T tid tcb.
Proof.
  intros.
  unfold WF_AIM_World in H.
  InvertAll.
  unfold WF_currTID in H0.
  InvertAll.
  unfold TQMod.in_dom in H9.
  unfold TQMod.maps_to. auto.
Qed.

Lemma enbedBlkQ_inWQSet:
forall B T w BlkQ M,
  embed_WQSet T B M 
  -> B w = Some BlkQ 
  -> (ptoTQ w (filterTQ T BlkQ) *^* mTrue) M.
Proof.
  intros.
  induction H.
  unfold WQSetMod.emp in H0. inversion H0.
  unfold WQSetMod.put in H0.
  induction (WQSetMod.beq_A w l).
  inversion H0.
  rewrite H5 in H2.
  exists M'. exists M. split; auto with PMap.
  split; auto. rewrite a. auto.
  unfold mTrue. auto.
  assert ((ptoTQ w (filterTQ T BlkQ) *^* mTrue) M).
  apply IHembed_WQSet; auto.
  assert ((ptoTQ l (filterTQ T Q) *^* (ptoTQ w (filterTQ T BlkQ) *^* mTrue)) M'').
  exists M'. exists M. split; auto with PMap.
  sep on.
  sep lift 1 in H5.
  sep matchH H5 with 2.
Qed.

Lemma merge_SW_pre:
forall S' pc M1' M2 T tid RdyQ B junk,
  (S_RF S' SP |-> pc *^* S_RF S' SP + 1 |-> junk *^* stkSpacePrec (S_RF S' SP) 8 *^* mTrue) M1'
  -> (embed_curr_TCB T tid *^* (embed_RdyQ T RdyQ *^* embed_WQSet T B)) M2
  -> MemMod.disjUnion M1' M2 (S_Mem S')
  -> (embed_curr_TCB T tid *^* embed_RdyQ T RdyQ *^* S_RF S' SP |-> pc *^* S_RF S' SP + 1 |-> junk *^* 
        stkSpacePrec (S_RF S' SP) 8 *^* mTrue) (S_Mem S').
Proof.
  intros.
  induction S' as (((M, R), fl), isr).
  UnfoldStateAll.
  rename M1' into M1.
  assert (((R SP |-> pc  *^* R SP + 1 |-> junk *^* stkSpacePrec (R SP) 8 *^* mTrue) *^*
              (embed_curr_TCB T tid *^* (embed_RdyQ T RdyQ *^* embed_WQSet T B))) M).
  exists M1. exists M2.
  split; auto.
  sep on.
  sep lift 4 in H2.
  sep lift 0.
  sep match_left H2; auto.
  sep lift 0 in H2.
  sep matchH H2 with 5.
Qed.



Lemma WF_AIM_World_RdyQinT:
forall Psi S T B RdyQ tid rho theta,
   WF_AIM_World Psi S T B RdyQ tid rho theta
   -> subsetQ_TQ RdyQ T.
Proof.
  intros.
  unfold subsetQ_TQ.
  intros.
  unfold WF_AIM_World in H.
  InvertAll.
  unfold WF_ReadyQ in H7.
  assert (exists tcb : Thread,
         TQMod.maps_to T tid0 tcb /\
         (TCB_Status tcb = READY \/ TCB_Status tcb = IDLE) /\
         (exists M : Mem,
            ThrdMemMod.maps_to x tid0 M /\
            (TCB_SP tcb |~>w TCB_PC tcb) M /\
            (WF_Rdy_Thrd Psi M tcb \/ WF_Rdy_Thrd_IH Psi M tcb))).
  apply H7; auto.
  InvertAll.
  unfold TQMod.in_dom.
  exists x1. auto.
Qed.

Lemma WF_AIM_World_cthNotRdy:
forall Psi S T B RdyQ tid rho theta,
   WF_AIM_World Psi S T B RdyQ tid rho theta
   -> WQMod.not_in_dom tid RdyQ.
Proof.
  intros.
  unfold WF_AIM_World in H.
  destruct H as [H H1].
  unfold WF_currTID in H.
  destruct H as [H2 [H3 H4]].
  assumption.
Qed.

Lemma WF_AIM_World_WFTQ:
forall Psi S T B RdyQ tid rho theta,
   WF_AIM_World Psi S T B RdyQ tid rho theta
   -> WF_TQ T.
Proof.
  intros.
  unfold WF_AIM_World in H.
  InvertAll.
  unfold WF_TWQ in H1.
  InvertAll. auto.
Qed.

Lemma WF_AIM_World_existsIdle:
forall Psi S T B RdyQ tid rho theta,
   WF_AIM_World Psi S T B RdyQ tid rho theta
   -> exists_Idle T RdyQ tid.
Proof.
  intros.
  unfold WF_AIM_World in H.
  InvertAll. auto.
Qed.

Lemma WF_AIM_World_BlkQinT:
forall Psi S T B RdyQ tid rho theta w BlkQ,
   WF_AIM_World Psi S T B RdyQ tid rho theta
   -> B w = Some BlkQ
   -> subsetQ_TQ BlkQ T.
Proof.
  intros.
  unfold subsetQ_TQ.
  intros.
  unfold WF_AIM_World in H.
  InvertAll.
  unfold WF_WQSet in H10.
  assert (exists tcb : Thread,
          exists M : Mem,
            TQMod.maps_to T tid0 tcb /\
            ThrdMemMod.maps_to x tid0 M /\ WF_Wait_Thrd Psi M tcb w).
  apply H10 with BlkQ; auto.
  InvertAll.
  unfold TQMod.in_dom. exists x1. auto.
Qed.

Fixpoint disjUnionList (l : list Mem) (M: Mem) {struct l}: Prop :=
  match l with
    nil => (M = MemMod.emp)
  | (M0 :: l') => exists M', disjUnionList l' M' /\ MemMod.disjUnion M0 M' M
  end.

(* disjUnionList ==> eqM list   by guoyu *)
Lemma disjUnionList_elim : forall l M,
  disjUnionList l M
  -> M ||- itr mEq l.
Proof.
  induction l; simpl; intros.
    subst M.
    sep trivial.
  destruct H as [M' [H1 H2]].
  exists a; exists M'; split_l.
  split. 
    unfold mEq; trivial.
  sep on.
  apply IHl; trivial.
Qed.
Implicit Arguments disjUnionList_elim [l M].

Lemma disjUnionList_intro : forall l M,
  M ||- itr mEq l
  -> disjUnionList l M.
Proof.
  induction l; simpl; intros.
    sep elim H; trivial.
  destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
  unfold mEq in HM1.
  subst a.
  exists M2; auto.
Qed.
Implicit Arguments disjUnionList_intro [l M].


(* disjUnionList ==> eqM list   by Xinyu *)
Definition eqM (M M': Mem) : Prop := M = M'.

Fixpoint DUL2mstar (dul : list Mem) {struct dul} : MPred :=
  match dul with
    nil => Aemp
  | M :: dul' => eqM M *^* (DUL2mstar dul')
  end.

Lemma DUL2mstar_left:
forall dul M,
  disjUnionList dul M -> DUL2mstar dul M.
Proof.
  intro dul.
  induction dul.
  unfold disjUnionList.
  unfold DUL2mstar.
  unfold Aemp. auto.
  unfold disjUnionList.
  fold disjUnionList.
  unfold DUL2mstar. fold DUL2mstar.
  intros.
  inversion_clear H.
  exists a. exists x. inversion_clear H0.
  split_l. unfold eqM. split_l.
  apply IHdul; auto.
Qed.

Lemma DUL2mstar_right:
forall dul M,
   DUL2mstar dul M -> disjUnionList dul M.
Proof.
  intro dul.
  induction dul.
  unfold disjUnionList.
  unfold DUL2mstar.
  unfold Aemp. auto.
  unfold disjUnionList.
  fold disjUnionList.
  unfold DUL2mstar. fold DUL2mstar.
  intros.
  unfold mstar in H.
  InvertAll.
  unfold eqM in H.
  subst x.
  exists x0.
  split; auto.
Qed.

Lemma mem_list_compose_a:
forall cthM11 cthM12 cthM1 cthM2 cthM M12 M1 M2 M,
  MemMod.disjUnion cthM11 cthM12 cthM1
  -> MemMod.disjUnion cthM1 cthM2 cthM
  -> MemMod.disjUnion M12 cthM M1
  -> MemMod.disjUnion M1 M2 M
  -> disjUnionList (M2 :: M12 :: cthM2 :: cthM12 :: cthM11 :: nil) M.
Proof.
  intros.
  unfold disjUnionList.
  exists M1. split; auto with PMap.
  exists cthM. split; auto.
  exists cthM1. split; auto with PMap.
  exists cthM11. split; auto with PMap.
  exists MemMod.emp. split; auto with PMap.
Qed.

Lemma ptoxW_elim1 : forall M l m,
  (M ||- l |->w_  *^* m) 
  -> M ||- l |->_ *^* l+1 |->_ *^* m.
Proof.
  intros M l m H.
  assert (exists x, M ||- l |->w x *^* m).
  apply ptoxW_ptoW1; auto.
  inversion_clear H0.
  assert (exists y, M ||- l |-> x *^* l + 1 |-> y *^* m).
  apply ptoW_elim1; auto.
  inversion H0.
  sep matchH H2 with 3.
  assert (h ||- l |->_ *^* l + 1 |-> x0).
  apply pto_ptox1 with x; auto.
  sep lift 1 in H3.
  sep lift 1.
  apply pto_ptox1 with x0; auto.
Qed.

Lemma call_sw_frame:
  forall M M' R R' fl fl' isr isr' M2 M12 cthM2 cthM12 cthM11 pc,
  disjUnionList (M2 :: M12 :: cthM2 :: cthM12 :: cthM11 :: nil) M
  -> stkSpacePrec (R SP) (8 + 2) cthM11
  -> nextS pc (comml (_call _SWITCH)) (M, R, fl, isr) (M', R', fl', isr')
  -> (exists junk,
        (R' SP |-> (pc+1) *^* R' SP + 1 |-> junk *^* (stkSpacePrec (R' SP) 8)) (MemUpdW cthM11 (R' SP) (pc + 1) junk)
        /\ disjUnionList (M2 :: M12 :: cthM2 :: cthM12 :: (MemUpdW cthM11 (R' SP) (pc + 1) junk) :: nil) M').
Proof.
  intros.
  destruct (nextS_inversion H1).
  clear H1.
  InvertAll.
  substH H7 with (MemUpd16_MemUpdW H7).
  destruct H7 as [junk H7].
  exists junk.
  split.
  sep on.
  unfold MemUpdW.
  sep lift 1.
  sep upd2.
  sep lift 2.
  sep upd2.
  sep stksplit in H0.
  sep lift 1 in H0.
  sep lift 2. sep lift 1 in H0.
  subst R'. reg simpl. auto.
  substH H with (disjUnionList_elim H).
  apply disjUnionList_intro.
  simpl.
  simpl in H.
  subst M'.
  sep lift 4.
  subst R'.
  reg simpl.
  apply meq_update1.
  sep matchH H with 4.
  sep on.
  sep stksplit in H0.
  sep lift 1 in H0.
  sep trivial.
Qed.

 

Lemma mem_list_compose_b:
forall M' M21 M22 M2 M12 cthM2 cthM12 cthM11',
  MemMod.disjUnion M21 M22 M2
  -> disjUnionList (M2 :: M12 :: cthM2 :: cthM12 :: cthM11' :: nil) M'
  -> exists M1', exists M2', 
           MemMod.disjUnion cthM11' M21 M1' 
           /\ disjUnionList (M22 :: M12 :: cthM2 :: cthM12 :: nil) M2'
           /\ MemMod.disjUnion M1' M2' M'.
Proof.
   intros. 
   assert (disjUnionList (M21 :: cthM11' :: M22 :: M12 :: cthM2 :: cthM12 :: nil) M'). 
   substH H0 with (disjUnionList_elim H0). 
   apply disjUnionList_intro. 
   unfold itr in * . 
   sep matchH H0 with 6. 
   exists M21;exists M22. 
   unfold mEq.
   intuition. 
   assert (M' ||- (mEq M2)). 
   sep matchH H0 with 2. 
   unfold mEq in H1. 
   sep off. 
   subst M2;assumption. 
   clear H0. 
   simpl in H1. 
   destruct H1 as [MM1 [H1 H2]]. 
   destruct H1 as [MM2 [H1 H3]]. 
   exists (MemMod.merge M21 cthM11');exists MM2. 
   intuition. 
   assert (MemMod.disjoint cthM11' M21). 
   disjoint lift_l. 
   split. 
   assumption. 
   auto with PMap. 
   destruct H3. 
   apply MemMod.disjUnion_assoc_right_left. 
   assumption. 
   subst MM1;assumption. 
Qed. 
  
Lemma mem_compose_c:
forall cthM11' M21 M1' sp pc junk T tid RdyQ,
  MemMod.disjUnion cthM11' M21 M1'
  -> (sp |-> pc *^* sp + 1 |-> junk *^* stkSpacePrec sp 8) cthM11'
  -> (embed_curr_TCB T tid *^* embed_RdyQ T RdyQ) M21
  -> SW_MPre T RdyQ tid sp pc junk M1'.
Proof. 
   intros.
   sep on. 
   assert (P := sconj_intro H0 H1 H). 
   unfold SW_MPre. 
   sep matchH P with 5. 
Qed. 

Lemma unique_remain:
  forall tid tid' RdyQ RdyQ',
    WQMod.not_in_dom tid RdyQ -> WQMod.not_in_dom tid' RdyQ'
      -> addThrd RdyQ tid = addThrd RdyQ' tid'
      -> tid = tid'
      -> RdyQ = RdyQ'.
Proof.
  intros.
  rewrite H2 in H ;rewrite H2 in H1.
(*
  inclear H2.
*)
  apply WQMod.extensional_eqMap.
  intro x.
  unfold addThrd in H1.
  destruct (WQMod.beq_A tid' x). 
  subst.
  unfold WQMod.not_in_dom in * .
  rewrite H;rewrite H0;reflexivity.
  substH H with (WQMod.put_noninterfere RdyQ dummy_sglton n).
  substH H0 with (WQMod.put_noninterfere RdyQ' dummy_sglton n).
  rewrite <- H;rewrite <- H0;rewrite H1;reflexivity.
Qed.

Lemma enter_SW_grt_aux60:
forall T RdyQ tid sp tcb pc junk sp' tcb' S0 M0', 
  SW_Post_aux (TQMod.put T tid (sp, TCB_Status tcb, pc)) RdyQ
          tid sp' sp tcb' pc junk (S_Mem_Upd S0 M0')
          -> sp' = sp.
Proof.
  intros.
  unfold SW_Post_aux in H.
  InvertAll.
  assert (tcb' = (sp, TCB_Status tcb, pc)).
  assert (TQMod.maps_to (TQMod.put T tid (sp, TCB_Status tcb, pc)) tid (sp, TCB_Status tcb, pc)).
  auto with PMap.
  unfold TQMod.maps_to in H3. unfold TQMod.maps_to in H. rewrite H3 in H.
  inversion_clear H. auto.
  rewrite H3 in H1. unfold TCB_SP in H1. auto.
Qed.

Lemma enter_sw_grt_hyp1:
forall T RdyQ tid sp pc junk S0 M0' M2',
  SW_MPost T RdyQ tid sp pc junk (S_Mem (S_Mem_Upd S0 M0'))
  -> S_RF (S_Mem_Upd S0 M0') SP = sp
  -> MemMod.disjUnion M0' M2' (S_Mem S0)
  -> Ret_Addr pc S0.
Proof.
  unfold SW_MPost;  unfold Ret_Addr.
  simpl.
  intros.
  apply nextPC_retw with sp.
  assumption.
  sep on.
  assert (M0' ||- sp |->w pc *^* (embed_curr_TCB T tid *^* embed_RdyQ T RdyQ *^* stkSpacePrec sp 8)).
  apply ptoW_intro1 with junk.
  sep matchH H with 5.
  clear H.
  assert ((sp |~>w pc) M0').
  apply ptoW_leadsToW1 with 
    (embed_curr_TCB T tid *^* embed_RdyQ T RdyQ *^* stkSpacePrec sp 8); auto.
  sep off.
  eapply leadsToW_Mono; eauto.
Qed.

Lemma enter_SW_grt_aux7:
forall T RdyQ tid sp tcb pc junk sp' tcb' S0 M0' M2', 
  SW_Post_aux (TQMod.put T tid (sp, TCB_Status tcb, pc)) RdyQ
          tid sp' sp tcb' pc junk (S_Mem_Upd S0 M0')
          -> MemMod.disjUnion M0' M2' (S_Mem S0)
          (* -> (S_RF S0 SP = sp) : need to add this line *)
          -> Ret_Addr pc S0.
Proof.
  intros.
  assert (sp' = sp).
  apply enter_SW_grt_aux60 with T RdyQ tid tcb pc junk tcb' S0 M0'; auto.
  unfold SW_Post_aux in H. InvertAll.
  apply enter_sw_grt_hyp1 with (TQMod.put T tid (sp, TCB_Status tcb, pc)) RdyQ tid sp junk M0' M2'; auto. 
  rewrite H1 in H4. auto.
Qed.

Lemma enter_SW_grt_hyp2:
  forall T RdyQ tid sp pc junk M,
    SW_MPost T RdyQ tid sp pc junk M
    -> exists M21_0, exists cthM11_0,
            (embed_curr_TCB T tid *^*
               embed_RdyQ T RdyQ) M21_0
            /\ (sp |-> pc *^* sp + 1 |-> junk *^* stkSpacePrec sp 8) cthM11_0
            /\ MemMod.disjUnion M21_0 cthM11_0 M.
Proof.
  intros.
  unfold SW_MPost in H.
  sep on.
  assert (M ||- (embed_curr_TCB T tid *^* embed_RdyQ T RdyQ) *^* 
    (sp |-> pc *^* sp+1 |-> junk *^* stkSpacePrec sp 8)).
  sep matchH H with 5.
  destruct H0 as [M1 [M2 P]].
  exists M1;exists M2;intuition.
Qed.

Lemma disjUnionList_split2:
forall M1 M2 M ML,
  disjUnionList (M1 :: M2 ::ML) M
  -> exists M'1, exists M'2, 
            disjUnionList ML M'1
            /\ MemMod.disjUnion M1 M2 M'2
            /\ MemMod.disjUnion M'1 M'2 M.
Proof.
  intros.
  destruct H as [M1' [H H1]].
  destruct H as [M2' [H H2]].
  exists M2';exists (MemMod.merge M1 M2).
  intuition.
  split.
  disjoint lift_r.
  reflexivity.
  apply (MemMod.disjUnion_commut).
  split.
  apply MemMod.disj_merge_B.
  disjoint lift_r.
  destruct H2;assumption.
  assert (MemMod.merge M1 (MemMod.merge M2 M2') = (MemMod.merge (MemMod.merge M1 M2) M2')).
  auto with PMap.
  rewrite <- H0.
  destruct H2;destruct H1.
  rewrite <- H3;rewrite <- H4.
  reflexivity.
Qed.

Lemma disjUnionList_merge2:
forall M1 M2 M ML M1' M2',
disjUnionList ML M1'
  -> MemMod.disjUnion M1 M2 M2'
  -> MemMod.disjUnion M2' M1' M
  -> disjUnionList (M1 :: M2 :: ML) M.
Proof.
  intros.
  exists (MemMod.merge M2 M1').
  split.
  exists M1'.
  intuition.
  split.
  disjoint lift_l.
  reflexivity.
  split.
  apply MemMod.disj_merge_A.
  destruct H0;assumption.
  disjoint lift_l.
  assert (MemMod.merge (MemMod.merge M1 M2) M1' = MemMod.merge M1 (MemMod.merge M2 M1')).
  auto with PMap.
  rewrite <- H2.
  destruct H0;destruct H1.
  rewrite <- H3; rewrite <- H4.
  reflexivity.
Qed.


Lemma enter_SW_grt_hyp4:
forall M21_0 M22 M12 cthM2 cthM12 cthM11'' S',
  disjUnionList (M21_0 :: M22 :: M12 :: cthM2 :: cthM12 :: cthM11'' :: nil) (S_Mem S')
  -> exists M'1, exists M'2, 
            disjUnionList (M12 :: cthM2 :: cthM12 :: cthM11'' :: nil) M'1
            /\ MemMod.disjUnion M21_0 M22 M'2
            /\ MemMod.disjUnion M'1 M'2 (S_Mem S').
Proof.
  intros.
  apply disjUnionList_split2.
  assumption.
Qed.

Lemma enter_SW_grt_aux5:
forall M12 cthM2 cthM12 cthM11'' M'1,
  disjUnionList (M12 :: cthM2 :: cthM12 :: cthM11'' :: nil) M'1
  -> exists cthM', MemMod.disjUnion M12 cthM' M'1 
                        /\ disjUnionList (cthM2 :: cthM12 :: cthM11'' :: nil) cthM'.
Proof.
  intros.
  unfold disjUnionList in H.
  inversion H.
  exists x. inversion H0.
  split; auto.
Qed.
Hint Resolve enter_SW_grt_aux5.

Lemma enter_SW_grt_hyp6:
forall TM tid M12 cthM' M'1,
  MemSplit (ThrdMemMod.remove TM tid) M12
  -> MemMod.disjUnion M12 cthM' M'1
  -> MemSplit (ThrdMemMod.put TM tid cthM') M'1.
Proof.
  intros.
  assert (ThrdMemMod.put TM tid cthM'= ThrdMemMod.put (ThrdMemMod.remove TM tid) tid cthM').
  auto with PMap.
  rewrite H1.
  apply MS_split with M12;try assumption.
  auto with PMap.
Qed.

Lemma enter_SW_grt_aux8:
forall T T' RdyQ tid sp sp' tcb tcb' pc pc' junk S0 S' M0' M2' M22 M12 cthM2 cthM12,
  SW_Post_aux T' RdyQ tid sp' sp tcb' pc junk (S_Mem_Upd S0 M0')
  -> nextS pc' (comms _retw) S0 S'
  -> MemMod.disjUnion M0' M2' (S_Mem S0)
  -> disjUnionList (M22 :: M12 :: cthM2 :: cthM12 :: nil) M2'
  -> T' = TQMod.put T tid (sp, TCB_Status tcb, pc)
  -> exists M21_0, exists cthM11'',
            ((S_RF S' SP - 2) |-> pc *^* (S_RF S' SP - 2 + 1) |-> junk *^* stkSpacePrec (S_RF S' SP - 2) 8) cthM11''
            /\ (embed_curr_TCB T' tid *^* embed_RdyQ T' RdyQ) M21_0
            /\ disjUnionList (M21_0 :: M22 :: M12 :: cthM2 :: cthM12 :: cthM11'' :: nil) (S_Mem S').
Proof.
  intros.
  assert (sp' = sp).
  apply enter_SW_grt_aux60 with T RdyQ tid tcb pc junk tcb' S0 M0'; auto.
  rewrite H3 in H. auto.
  unfold SW_Post_aux in H.
  InvertAll.
  assert (exists M21_0, exists cthM11_0,
            (embed_curr_TCB T' tid *^*
               embed_RdyQ T' RdyQ) M21_0
            /\ (sp |-> pc *^* sp+1 |-> junk *^* stkSpacePrec sp 8) cthM11_0
            /\ MemMod.disjUnion M21_0 cthM11_0 M0').
  apply enter_SW_grt_hyp2; auto.
  InvertAll. rename x into M21_0. rename x0 into cthM11_0.
  exists M21_0.
  induction S0 as (((Mx, Rx), flx), isrx).
  induction S' as (((My, Ry), fly), isry).
  UnfoldStateAll.
  subst sp'.
  destruct (nextS_inversion H0).
  subst My. clear H0.
  exists cthM11_0.
  InvertAll.
  subst Ry. reg simpl.
  assert (Rx SP + 2 -  2 = Rx SP). auto.
  rewrite H4. rewrite H7.
  split_l.
  split_l.
  assert ((eqM M21_0 *^* eqM cthM11_0 *^* (DUL2mstar (M22 :: M12 :: cthM2 :: cthM12 :: nil))) Mx).
  exists M0'. exists M2'.
  split_l.
  split. exists M21_0. exists cthM11_0. unfold eqM. auto.
  apply DUL2mstar_left; auto.
  cut (DUL2mstar (M21_0 :: M22 :: M12 :: cthM2 :: cthM12 :: cthM11_0 :: nil) Mx).
  intro.
  apply DUL2mstar_right; auto.
  unfold DUL2mstar in *.
  sep on.
  sep matchH H14 with 7.
Qed.



(* chaged because of the change of WF_currTID *)
Lemma WF_currTID_ext_T :
forall T B RdyQ tid tcb tcb' rho,
  WF_currTID tid T B RdyQ rho
  -> T tid = Some tcb 
  -> TCB_Status tcb = TCB_Status tcb'
  -> WF_currTID tid (TQMod.put T tid tcb') B RdyQ rho.
Proof.
  unfold WF_currTID.
  intros.
  InvertAll.
  split; auto. split; auto. split. auto with PMap.
  exists tcb'.
  rewrite <- H1.
  split; auto.
  split.
  unfold TQMod.maps_to in H5.
  rewrite H5 in H0. inversion H0. rewrite <- H8. auto.
  unfold rho_tcb_consistent in H7.
  unfold rho_tcb_consistent.
  rewrite <- H1.
  unfold TQMod.maps_to in H5.
  rewrite H5 in H0. inversion H0.
  rewrite H8 in H7. auto.
Qed.



Lemma WF_TWQ_ext_T:
forall T B RdyQ tid tcb,
  WF_TWQ T B RdyQ -> (exists tcb, T tid = Some tcb) -> WQMod.not_in_dom tid RdyQ -> WF_TWQ (TQMod.put T tid tcb) B RdyQ.
Proof.
  unfold WF_TWQ.
  intros.
  split; auto.
  unfold WF_TQ.
  intros.
  InvertAll.
  unfold WF_TQ in H3.
  apply H3. assert (tid0 = tid \/ tid0 <> tid). apply nat_eq_neq_dec.
  inversion_clear H0.
  rewrite H6. unfold TQMod.in_dom. exists x. auto.
  unfold TQMod.in_dom in H2.
  InvertAll.
  unfold TQMod.put in H0.
  induction (TQMod.beq_A tid0 tid). contradiction.
  unfold TQMod.in_dom. exists x0. auto.
  InvertAll. split; auto.
Qed.

Lemma exists_Idle_ext_T:
forall T RdyQ tid tcb sp pc,
  exists_Idle T RdyQ tid 
  -> T tid = Some tcb 
  -> WQMod.not_in_dom tid RdyQ
  -> exists_Idle (TQMod.put T tid (sp, TCB_Status tcb, pc)) RdyQ tid.
Proof.
  unfold exists_Idle.
  intros.
  inversion_clear H. left.
  inversion_clear H2. inversion_clear H.
  unfold TQMod.maps_to in H2.
  rewrite H2 in H0. inversion H0.
  rewrite H4 in H3.
  rewrite H3.
  exists (sp, IDLE, pc). split; auto with PMap.
  right.
  InvertAll.
  exists x. exists x0. split; auto.
  split; auto.
  unfold TQMod.maps_to.
  unfold TQMod.maps_to in H2.
  cut (tid <> x). intro.
  unfold TQMod.put.
  induction (TQMod.beq_A x tid).
  absurd (x <> tid); auto. auto.
  unfold WQMod.not_in_dom in H1.
  unfold WQMod.in_dom in H.
  assert (tid = x \/ tid <> x). apply nat_eq_neq_dec.
  inversion_clear H3; auto.
  rewrite H5 in H1.
  rewrite H1 in H.
  InvertAll. inversion_clear H3.
Qed.

Lemma enter_SW_grt_hyp9:
forall pc pc' f S S1 S2 S3,
  nextS pc (comml (_call f)) S S1
  -> S_ISR S1 = S_ISR S2
  -> S_Flag S1 _IF = S_Flag S2 _IF
  -> nextS pc' (comms _retw) S2 S3
  -> S_ISR S = S_ISR S3 /\ S_Flag S _IF = S_Flag S3 _IF.
Proof.
  intros.
  inversion H;subst.
  inversion H2;subst.
  simpl in * .
  intuition.
Qed.

Lemma step_call_hyp:
forall pc f M R fl isr M' R' fl' isr' cthM11 cthM12 cthM1 n,
  nextS pc (comml (_call f)) (M, R, fl, isr) (M', R', fl', isr')
  -> stkSpacePrec (R SP) (n + 2) cthM11
  -> MemMod.disjUnion cthM11 cthM12 cthM1
  -> nextS pc (comml (_call f)) (cthM1, R, fl, isr)
     (MemMod.merge cthM12 (MemMod.put cthM11 (R SP - 2) (pc + 1)), R', fl', isr').
Proof.
  intros.
  inversion H;subst.
  apply nextS_call.
  unfold pushW in * .
  destruct H5 as [P1 [P2 P3]].
  assert (stkSpace (R SP) 2 cthM11).
  apply stkSpace_weaken with (n+2).
  apply stkSpacePrec_stkSpace;assumption.
  omega.
  assert (stkSpace (R SP) 2 cthM1).
  apply stkSpace_Mono with cthM11 cthM12.
  assumption.
  assumption.
  intuition.
  destruct H3.
  assert (2>1).
  omega.
  substH H4 with (leadsToBlk_leadsToxW H4 H5).
  substH H4 with (leadsToxW_elim1 H4).
  destruct H4.
  exists x.
  assert (MemMod.put (MemMod.put cthM1 (R SP - 2) (pc + 1)) (R SP - 2 + 1) x = MemMod.put cthM1 (R SP - 2) (pc + 1)).
  apply MemMod.extensional_eqMap;intro.
  unfold MemMod.put.
  destruct (MemMod.beq_A x0 (R SP - 2 + 1)).
  destruct (MemMod.beq_A x0 (R SP -2)).
  rewrite <- e0 in e.
  assert (x0 <> x0 +1).
  omega.
  contradiction.
  rewrite e;symmetry;assumption.
  reflexivity.
  rewrite H6.
  symmetry.
  destruct H1.
  subst.
  assert (MemMod.disjUnion (MemMod.put cthM11 (R SP - 2) (pc + 1)) cthM12 
   (MemMod.put (MemMod.merge cthM11 cthM12) (R SP - 2) (pc + 1))).
  split.
  apply MemMod.disjoint_commut.
  apply MemMod.update_presv_disj_right.
  apply MemMod.disjoint_commut;assumption.
  destruct H2.
  substH H7 with (leadsToBlk_leadsToxW H7 H5).
  substH H7 with (leadsToxW_elim H7).
  unfold MemMod.in_dom;assumption.
  apply MemMod.put_merge_merge_put.
  substH H7 with (MemMod.disjUnion_commut H7).
  destruct H7;assumption.
Qed.

Lemma enter_SW_grt_auxb0_hyp0:
forall pc M R fl isr M' R' fl' isr' f,
  nextS pc (comml (_call f)) (M, R, fl, isr) (M', R', fl', isr')
  -> R SP - 2 = R' SP.
Proof.
  intros.
  inversion H;subst.
  unfold pushW in H3.
  destruct H3 as [H4 [H5 H6]].
  subst.
  symmetry.
  apply RF_update_RF.
Qed.

Lemma enter_SW_grt_auxb0_hyp1:
forall pc S0 S',
  nextS pc (comms _retw) S0 S' -> S_RF S' SP = S_RF S0 SP + 2.
Proof.
  intros.
  inversion H;subst.
  destruct H0 as [pc' [H1 H2]].
  subst.
  simpl.
  apply RF_update_RF.
Qed.


Lemma enter_SW_grt_auxb0_hyp3:
  forall pc' S0 S',
    nextS pc' (comms _retw) S0 S'
    -> (S_Flag S0 _IF = S_Flag S' _IF) /\ (S_ISR S0 = S_ISR S').
Proof.
  intros.
  inversion H;subst.
  simpl.
  intuition.
Qed.

Lemma enter_SW_grt_auxb0_hyp5:
  forall pc S0 S' M M' x y m,
    nextS pc (comms _retw) S0 S'
    -> MemMod.disjoint M M'
    -> (S_RF S0 SP |-> x *^* S_RF S0 SP + 1 |-> y *^* m) M
    -> nextS pc (comms _retw) (S_Mem_Upd S0 (MemMod.merge M' M))
          (S_Mem_Upd S' (MemMod.merge M' M)).
Proof.
  intros.
  inversion H;subst.
  unfold S_Mem_Upd.
  simpl.
  apply nextS_retw.
  destruct H2 as [pc' [P1 P2]].
  exists x.
  split.
  assert (MemMod.disjUnion M M' (MemMod.merge M' M)).
  split; auto with PMap.
  apply leadsToW_Mono with M M'. auto.
  simplS in H1.
  assert ((R SP |->w x *^* m) M).
  apply ptoW_intro1 with y; auto.
  sep on. sep trivial.
  auto.
Qed.

Lemma enter_BLK_grt2_auxc_hyp01:
forall l m M,
  (l |->w_ *^* m) M -> MemMod.in_dom l M.
Proof.
  intros.
  assert (MemMod.in_dom l M /\ MemMod.in_dom (l+1) M).
  apply ptoxW_indom1 with m; auto.
  inversion H0; auto.
Qed.

Lemma enter_BLK_grt2_auxc_hyp02:
forall l m M,
  (l |->w_ *^* m) M -> MemMod.in_dom (l+1) M.
  intros.
  assert (MemMod.in_dom l M /\ MemMod.in_dom (l+1) M).
  apply ptoxW_indom1 with m; auto.
  inversion H0; auto.
Qed.

Lemma step_call_hyp1:
forall pc f M R fl isr M' R' fl' isr' cthM11 cthM12 cthM1 n junk,
  nextS pc (comml (_call f)) (M, R, fl, isr) (M', R', fl', isr')
  -> stkSpacePrec (R SP) (n + 2) cthM11
  -> MemMod.disjUnion cthM11 cthM12 cthM1
  -> nextS pc (comml (_call f)) (cthM1, R, fl, isr)
     (MemMod.merge cthM12 (MemUpdW cthM11 (R SP - 2) (pc + 1) junk), R', fl', isr').
Proof.
  intros.
  destruct (nextS_inversion H).
  InvertAll. subst isr' fl'.
  apply nextS_call.
  unfold pushW.
  split.
  assert ((stkSpacePrec (R SP) (n + 2) *^* eqM cthM12) cthM1).
  exists cthM11. exists cthM12. split_l.
  unfold eqM. split; auto.
  sep on.
  sep stksplit 2 in H3. sep lift 0 in H3.
  sep trivial.
  subst R'. split; auto.
  assert (MemMod.disjUnion cthM12 cthM11 cthM1). auto with PMap.
  clear H1.
  assert (MemMod.disjUnion cthM12 (MemUpdW cthM11 (R SP - 2) (pc + 1) junk) (MemUpdW cthM1 (R SP - 2) (pc + 1) junk)).
  sep on.
  sep stksplit in H0.
  sep lift 1 in H0.
  assert (MemMod.in_dom (R SP - 2) cthM11).
  apply enter_BLK_grt2_auxc_hyp01 with (stkSpacePrec (R SP - 2) (n + 2 - 2)). auto.
  assert (MemMod.in_dom (R SP - 2 + 1) cthM11).
  apply enter_BLK_grt2_auxc_hyp02 with (stkSpacePrec (R SP - 2) (n + 2 - 2)). auto.
  sep off.
  unfold MemUpdW.
  apply MemMod.update_presv_disjUnion_right.
  apply MemMod.update_presv_disjUnion_right. auto.
  auto. unfold MemMod.in_dom. unfold MemMod.put.
  induction (MemMod.beq_A (R SP - 2 + 1) (R SP - 2)).
  exists (pc+1). auto.
  auto.
  inversion_clear H1.
  rewrite <- H6.
  apply MemUpdW_MemUpd16 with junk; auto.
Qed.

Lemma enter_SW_grt_auxb0_hyp21:
forall sp n M w w1,
  stkSpacePrec (sp + 2) (n + 2) M
  -> (sp |-> w *^* sp + 1 |-> w1 *^* stkSpacePrec sp n) (MemUpdW M sp w w1).
Proof.
  intros.
  sep on.
  unfold MemUpdW.
  sep lift 1.
  sep upd2.
  sep lift 2.
  sep upd2.
  sep stksplit in H.
  sep lift 2.
  sep match_left H; auto.
Qed.
  


Lemma enter_SW_grt_auxb0_hyp22:
forall cthM11 cthM12 cthM11' cthM12_1 cthM12_2 cthM11'_1 cthM11'_2 sp pc junk,
  MemMod.disjUnion cthM12_1 cthM12_2 cthM12
  -> MemMod.disjoint cthM11 cthM12
  -> cthM11' = MemUpdW cthM11 sp pc junk
  -> MemMod.disjUnion cthM11'_1 cthM11'_2 cthM11'
  -> stkSpacePrec (sp + 2) (8 + 2) cthM11
  -> MemMod.disjoint cthM12_1 cthM11'_1.
Proof.
  intros.
  assert (MemMod.disjUnion cthM11' cthM12 (MemUpdW (MemMod.merge cthM11 cthM12) sp pc junk)).
  subst cthM11'.
  unfold MemUpdW.
  assert (MemMod.in_dom sp cthM11 /\ MemMod.in_dom (sp+1) cthM11).
  sep on.
  sep stksplit in H3.
  sep lift 1 in H3.
  apply ptoxW_indom1 with (stkSpacePrec sp 8); auto.
  inversion_clear H1.
  apply MemMod.update_presv_disjUnion_left.
  apply MemMod.update_presv_disjUnion_left.
  split; auto. auto.
  unfold MemMod.in_dom. unfold MemMod.put.
  induction (MemMod.beq_A (sp + 1) sp); auto.
  exists pc. auto.
  set (Mx := (MemUpdW (MemMod.merge cthM11 cthM12) sp pc junk)) in *.
  assert ((eqM cthM11'_1 *^* eqM cthM11'_2 *^* (eqM cthM12_1 *^* eqM cthM12_2)) Mx).
  exists cthM11'. exists cthM12.
  split_l. split. exists cthM11'_1. exists cthM11'_2. unfold eqM. auto.
  exists cthM12_1. exists cthM12_2. unfold eqM. auto.
  assert ((eqM cthM12_1 *^* eqM cthM11'_1 *^* (eqM cthM11'_2 *^* eqM cthM12_2)) Mx).
  sep on. sep matchH H5 with 4.
  unfold mstar in H6.
  unfold eqM in H6.
  InvertAll.
  subst x3 x4. auto.
  inversion H10; auto.
Qed.



Lemma enter_SW_grt_auxb0_hyp23:
forall cthM12 cthM11'' cthM12_1 cthM12_2 cthM11''_1 cthM11''_2,
  MemMod.disjUnion cthM12_1 cthM12_2 cthM12
  -> MemMod.disjoint cthM11'' cthM12
  -> MemMod.disjUnion cthM11''_1 cthM11''_2 cthM11''
  -> MemMod.disjoint cthM12_1 cthM11''_1.
Proof.
  intros.
  inversion_clear H1.
  subst cthM11''.
  assert (MemMod.disjoint cthM11''_1 cthM12).
  apply MemMod.disj_merge_disj_left with cthM11''_2; auto.
  inversion_clear H.
  subst cthM12.
  apply MemMod.disj_merge_disj_left with cthM12_2; auto with PMap.
Qed.


Lemma enter_SW_grt_auxb0_hyp24:
forall cthM11 cthM12 cthM11' cthM12_1 cthM12_2 cthM11'_1 cthM11'_2 sp pc junk,
  MemMod.disjUnion cthM12_1 cthM12_2 cthM12
  -> MemMod.disjoint cthM11 cthM12
  -> cthM11' = MemUpdW cthM11 sp pc junk
  -> MemMod.disjUnion cthM11'_1 cthM11'_2 cthM11'
  -> stkSpacePrec (sp + 2) (8 + 2) cthM11
  ->  MemMod.disjUnion (MemMod.merge cthM12_1 cthM11'_1)
       (MemMod.merge cthM12_2 cthM11'_2) (MemMod.merge cthM12 cthM11').
Proof.
  intros.
  assert (MemMod.disjUnion cthM11' cthM12 (MemUpdW (MemMod.merge cthM11 cthM12) sp pc junk)).
  subst cthM11'.
  unfold MemUpdW.
  assert (MemMod.in_dom sp cthM11 /\ MemMod.in_dom (sp+1) cthM11).
  sep on.
  sep stksplit in H3.
  sep lift 1 in H3.
  apply ptoxW_indom1 with (stkSpacePrec sp 8); auto.
  inversion_clear H1.
  apply MemMod.update_presv_disjUnion_left.
  apply MemMod.update_presv_disjUnion_left; auto.
  split; auto.
  unfold MemMod.in_dom. unfold MemMod.put.
  induction (MemMod.beq_A (sp + 1) sp); auto.
  exists pc. auto.
  set (Mx := (MemUpdW (MemMod.merge cthM11 cthM12) sp pc junk)) in *.
  assert ((eqM cthM11'_1 *^* eqM cthM11'_2 *^* (eqM cthM12_1 *^* eqM cthM12_2)) Mx).
  exists cthM11'. exists cthM12.
  split_l. split. exists cthM11'_1. exists cthM11'_2. unfold eqM. auto.
  exists cthM12_1. exists cthM12_2. unfold eqM. auto.
  assert ((eqM cthM12_1 *^* eqM cthM11'_1 *^* (eqM cthM12_2 *^* eqM cthM11'_2)) Mx).
  sep on. sep matchH H5 with 4.
  assert (MemMod.disjUnion cthM12 cthM11' Mx); auto with PMap.
  inversion_clear H7. rewrite <- H9.
  unfold mstar in H6.
  unfold eqM in H6.
  InvertAll.
  subst x1 x2 x3 x4.
  inversion_clear H10. inversion_clear H12.
  subst x0 x. auto.
Qed.


Lemma enter_SW_grt_auxb0_hyp25:
forall cthM12 cthM11'' cthM12_1 cthM12_2 cthM11''_1 cthM11''_2,
  MemMod.disjUnion cthM12_1 cthM12_2 cthM12
  -> MemMod.disjoint cthM11'' cthM12
  -> MemMod.disjUnion cthM11''_1 cthM11''_2 cthM11''
  -> MemMod.disjUnion (MemMod.merge cthM12_1 cthM11''_1)
     (MemMod.merge cthM12_2 cthM11''_2) (MemMod.merge cthM12 cthM11'').
Proof.
  intros.
  assert (MemMod.disjUnion cthM12 cthM11'' (MemMod.merge cthM12 cthM11'')).
  split; auto with PMap.
  assert ((eqM cthM12_1 *^* eqM cthM12_2 *^* (eqM cthM11''_1 *^* eqM cthM11''_2)) (MemMod.merge cthM12 cthM11'')).
  exists cthM12. exists cthM11''.
  split_l.
  split. exists cthM12_1. exists cthM12_2. unfold eqM. auto.
         exists cthM11''_1. exists cthM11''_2. unfold eqM. auto.
  assert ((eqM cthM12_1 *^* eqM cthM11''_1 *^* (eqM cthM12_2 *^* eqM cthM11''_2))
              (MemMod.merge cthM12 cthM11'')).
  sep on.
  sep matchH H3 with 4.
  clear H3. unfold mstar in H4.
  unfold eqM in H4.
  InvertAll.
  subst x1 x2 x3 x4.
  inversion_clear H5. inversion_clear H7.
  subst x x0. auto.
Qed.


Lemma enter_SW_grt_auxb0_hyp26:
forall sp pc junk cthM11'_2 cthM11''_2,
  (sp |-> pc *^* sp + 1 |-> junk) cthM11'_2
  -> (sp |-> pc *^* sp + 1 |-> junk) cthM11''_2
  -> cthM11'_2 = cthM11''_2.
Proof.
  intros.
  assert (uniqueM (sp |-> pc *^* sp + 1 |-> junk)).
  apply mstar_presv_uniqueM; apply pto_uniqueM.
  unfold uniqueM in H1.
  apply H1; auto.
Qed.

Lemma enter_SW_grt_auxb0_hyp2:
forall cthM12 cthM11 cthM11'' S' R' pc fl' isr' S0 sp junk,
  (INV0 *^* INV1 *^* mTrue) cthM12
  -> stkSpacePrec sp (8 + 2) cthM11
  -> (S_RF S' SP - 2 |-> pc *^* S_RF S' SP - 2 + 1 |-> junk *^*
        stkSpacePrec (S_RF S' SP - 2) 8) cthM11''
  -> S_RF S' SP = (R' SP) + 2
  -> sp - 2 = (R' SP)
  -> MemMod.disjoint cthM11 cthM12
  -> MemMod.disjoint cthM11'' cthM12
  -> prepost (INV0 *^* INV1 *^* stkSpacePrec (R' SP) _SW_Stk)
     (INV0 *^* INV1 *^* stkSpacePrec (R' SP) _SW_Stk)
     (MemMod.merge cthM12 (MemUpdW cthM11 (R' SP) pc junk), R', fl',
     isr') (S_Mem_Upd S0 (MemMod.merge cthM12 cthM11'')).
Proof.
  unfold prepost.
  intros. induction S0 as (((M0, R0), fl0), isr0).
  simplS. clear M0 R0 fl0 isr0.
  destruct H as [cthM12_1 [chtM12_2 [HcthM12split [HcthM12_1 HcthM12_2]]]].
  set (cthM11' := MemUpdW cthM11 (R' SP) pc junk).
  assert (sp >= 2).
  apply ge_trans with (8 + 2); auto with arith.
  apply stkSpacePrec_sp_ge_sz with cthM11; auto.
  assert (sp = R' SP + 2). rewrite <- H3.
  assert (sp - 2 + 2 = sp). auto. auto.
  rewrite H6 in H0.
  assert ((R' SP |-> pc *^* R' SP + 1 |-> junk *^*
               stkSpacePrec (R' SP) 8) cthM11').
  unfold cthM11'.
  apply enter_SW_grt_auxb0_hyp21. auto.
  rewrite H2 in H1.
  assert (R' SP + 2 - 2 = R' SP). auto.
  rewrite H8 in H1.
  sep on.
  sep lift 2 in H7. sep lift 2 in H1.
  destruct H7 as [cthM11'_1 [cthM11'_2 [Hsplit11' [HM11'_1 HM11'_2]]]].
  destruct H1 as [cthM11''_1 [cthM11''_2 [Hsplit11'' [HM11''_1 HM11''_2]]]].
  exists (MemMod.merge cthM12_1 cthM11'_1).
  exists (MemMod.merge chtM12_2 cthM11'_2).
  exists (MemMod.merge cthM12_1 cthM11''_1).
  split.
  exists cthM12_1. exists cthM11'_1.
  split; auto.
  split; auto.
  apply enter_SW_grt_auxb0_hyp22 with cthM11 cthM12 cthM11' chtM12_2 cthM11'_2 (R' SP) pc junk; auto.
  split.
  exists cthM12_1. exists cthM11''_1.
  split; auto.
  split; auto.
  apply enter_SW_grt_auxb0_hyp23 with cthM12 cthM11'' chtM12_2 cthM11''_2; auto.

  split.
  apply enter_SW_grt_auxb0_hyp24 with cthM11 (R' SP) pc junk; auto.
  assert (cthM11'_2 = cthM11''_2).
  apply enter_SW_grt_auxb0_hyp26 with (R' SP) pc junk; auto.
  rewrite H1. apply enter_SW_grt_auxb0_hyp25; auto.
Qed.

Lemma enter_SW_grt_auxb0:
forall cthM1 R fl isr M cthM11 cthM12 S' cthM11'' pc pc' M' R' fl' isr' junk S0,
  MemMod.disjUnion cthM11 cthM12 cthM1 
  -> (INV0 *^* INV1 *^* mTrue) cthM12
  -> stkSpacePrec (R SP) (8 + 2) cthM11
  -> (S_RF S' SP - 2 |-> pc + 1 *^* S_RF S' SP - 2 + 1 |-> junk *^*
        stkSpacePrec (S_RF S' SP - 2) 8) cthM11''
  -> nextS pc (comml (_call _SWITCH)) (M, R, fl, isr) (M', R', fl', isr')
  -> nextS pc' (comms _retw) S0 S'
  -> S_RF S0 SP = R' SP
  -> MemMod.subseteq (MemMod.merge cthM12 cthM11'') (S_Mem S')
  -> isr = S_ISR S'
  -> fl _IF = S_Flag S' _IF
  -> MemMod.disjoint cthM11'' cthM12
  -> wrap_fun gy pc _SWITCH pc' (cthM1, R, fl, isr)
      (S_Mem_Upd S' (MemMod.merge cthM12 cthM11'')).
Proof.
  intros.
  unfold wrap_fun.
  unfold conjG.
  exists (S_Mem_Upd S0 (MemMod.merge cthM12 cthM11'')).
  split.
  exists (MemMod.merge cthM12 (MemUpdW cthM11 (R SP - 2) (pc+1) junk), R', fl', isr').
  split.

  apply step_call_hyp1 with M M' _SW_Stk; auto.
  unfold gy. exists (R' SP). 
  split.

  assert (R SP - 2 = R' SP).
  apply enter_SW_grt_auxb0_hyp0 with pc M fl isr M' fl' isr' _SWITCH; auto.
  rewrite H10.
  assert (S_RF S' SP = S_RF S0 SP + 2).
  apply enter_SW_grt_auxb0_hyp1 with pc'; auto.
  rewrite H5 in H11.

  apply enter_SW_grt_auxb0_hyp2 with S' (R SP); auto.
  unfold MemMod.disjUnion in H. inversion H. auto.
  assert ((S_Flag S0 _IF = S_Flag S' _IF) /\ (S_ISR S0 = S_ISR S')).
  apply enter_SW_grt_auxb0_hyp3 with pc'; auto.
  inversion_clear H10.
  assert (S_Flag (M, R, fl, isr) _IF = S_Flag (M', R', fl', isr') _IF 
             /\ S_ISR (M, R, fl, isr) = S_ISR (M', R', fl', isr')).
  apply enter_SW_grt_auxb0_hyp4 with pc _SWITCH; auto.
  inversion_clear H10.
  split.
  assert (S_Flag (S_Mem_Upd S0 (MemMod.merge cthM12 cthM11'')) = S_Flag S0).
  auto. rewrite H10. rewrite H11. rewrite <- H8.
  UnfoldStateAll. auto.
  split.
  assert (S_ISR (S_Mem_Upd S0 (MemMod.merge cthM12 cthM11'')) = S_ISR S0).
  auto. rewrite H10. rewrite H12. rewrite <- H7. UnfoldStateAll.
  auto. split.
  unfold S_RF. auto.
  assert (S_RF (S_Mem_Upd S0 (MemMod.merge cthM12 cthM11'')) = S_RF S0).
  auto. rewrite H10.  rewrite H5. unfold S_RF. auto.
  assert (S_RF S' SP = S_RF S0 SP + 2).
  apply enter_SW_grt_auxb0_hyp1 with pc'; auto.
  rewrite H10 in H2.
  assert (S_RF S0 SP + 2 - 2 = S_RF S0 SP). auto.
  rewrite H11 in H2.
  apply enter_SW_grt_auxb0_hyp5 with (pc + 1) junk (stkSpacePrec (S_RF S0 SP) 8); auto.
Qed.

Lemma enter_SW_grt_auxb:
forall (p p' : Pred) cthM1 R fl isr M cthM11 cthM12 S' cthM11'' pc pc' M' R' fl' isr' junk S0,
  p (cthM1, R, fl, isr)
  -> (forall pc' : nat, pimp (p |> wrap_fun gy pc _SWITCH pc') p')
  -> MemMod.disjUnion cthM11 cthM12 cthM1 
  -> (INV0 *^* INV1 *^* mTrue) cthM12
  -> stkSpacePrec (R SP) (8 + 2) cthM11
  -> ((S_RF S' SP - 2 |-> pc + 1) *^* (S_RF S' SP - 2 + 1 |-> junk) *^* stkSpacePrec (S_RF S' SP - 2) 8) cthM11''
  -> nextS pc (comml (_call _SWITCH)) (M, R, fl, isr) (M', R', fl', isr')
  -> nextS pc' (comms _retw) S0 S'
  -> S_RF S0 SP = R' SP
  -> MemMod.subseteq (MemMod.merge cthM12 cthM11'') (S_Mem S')
  -> isr = S_ISR S'
  -> fl _IF = S_Flag S' _IF
  -> MemMod.disjoint cthM11'' cthM12
  -> p' (S_Mem_Upd S' (MemMod.merge cthM12 cthM11'')).
Proof.
  intros.
  unfold pimp in H0.
  apply H0 with pc'.
  unfold linkPG. exists (cthM1, R, fl, isr).
  split; auto.
  apply enter_SW_grt_auxb0 with M cthM11 M' R' fl' isr' junk S0; auto.
Qed.

Lemma enter_SW_grt_hypa:
forall cthM2 cthM12 cthM11'' cthM',
   disjUnionList (cthM2 :: cthM12 :: cthM11'' :: nil) cthM'
   -> MemMod.disjUnion (MemMod.merge cthM12 cthM11'') cthM2 cthM'.
Proof.
  intros.
  cut (MemMod.disjUnion cthM2 (MemMod.merge cthM12 cthM11'') cthM').
  intro.
  auto with PMap.
  unfold disjUnionList in H.
  InvertAll.
  subst x1. assert (cthM11'' = x0). auto with PMap.
  subst x0.
  inversion_clear H2.
  subst x. auto.
Qed.

Lemma enter_SW_grt_hypc:
forall M21_0 M22 M12 cthM2 cthM12 cthM11'' S',
  disjUnionList
          (M21_0 :: M22 :: M12 :: cthM2 :: cthM12 :: cthM11'' :: nil)
          (S_Mem S')
  -> MemMod.subseteq (MemMod.merge cthM12 cthM11'') (S_Mem S').
Proof.
  intros.
  unfold disjUnionList in H.
  induction S' as (((M, R), fl), isr).
  UnfoldStateAll.
  InvertAll.
  subst x4.
  assert (cthM11'' = x3). auto with PMap.
  subst x3. clear H6.
  inversion_clear H5. subst x2.
  clear H.
  apply MemMod.subseteq_transitivity with x.
  apply MemMod.subseteq_transitivity with x0.
  apply MemMod.subseteq_transitivity with x1.
  apply MemMod.disjUnion_subseteq_right with cthM2; auto.
  apply MemMod.disjUnion_subseteq_right with M12; auto.
  apply MemMod.disjUnion_subseteq_right with M22; auto.
  apply MemMod.disjUnion_subseteq_right with M21_0; auto.
Qed.


Lemma enter_SW_grt_hypd:
forall M21_0 M22 M12 cthM2 cthM12 cthM11'' S',
  disjUnionList
          (M21_0 :: M22 :: M12 :: cthM2 :: cthM12 :: cthM11'' :: nil)
          (S_Mem S')
  -> MemMod.disjoint cthM11'' cthM12.
Proof.
  intros.
  unfold disjUnionList in H.
  induction S' as (((M, R), fl), isr).
  UnfoldStateAll.
  InvertAll.
  subst x4.
  assert (cthM11'' = x3). auto with PMap.
  subst x3. clear H6.
  inversion_clear H5. auto with PMap.
Qed.


Lemma enter_SW_grt_hypf:
forall Psi'0 T RdyQ TM T' tid tcb cthM',
  WF_ReadyQ Psi'0 T RdyQ TM
  -> T' = TQMod.put T tid tcb
  -> WQMod.not_in_dom tid RdyQ
  -> WF_ReadyQ Psi'0 T' RdyQ (ThrdMemMod.put TM tid cthM').
Proof.
  intros.
  unfold WF_ReadyQ in * .
  intros.
  subst T'.
  intros.
  unfold TQMod.put.
  unfold ThrdMemMod.put.
  unfold ThrdMemMod.maps_to in * .
  unfold  TQMod.maps_to in * .
  unfold TQMod.beq_A.
  unfold TQSpec.beq_Domain.
  unfold ThrdMemMod.beq_A.
  unfold ThrdMemSpec.beq_Domain.
  destruct (eq_nat_dec tid0 tid).
  rewrite e in H2;destruct H2.
  rewrite H1 in H0;inversion H0.
  apply H;assumption.
Qed.

Lemma enter_SW_grt_hypg:
forall Psi'0 T B TM T' tid tcb cthM',
  WF_WQSet Psi'0 T B TM
  -> unblocked tid B
  -> T' = TQMod.put T tid tcb
  -> WF_WQSet Psi'0 T' B (ThrdMemMod.put TM tid cthM').
Proof.
  intros.
  unfold WF_WQSet in * .
  unfold unblocked in H0.
  intros.
  substH H0 with (H0 l Q H2).
  subst T'.
  unfold TQMod.put.
  unfold ThrdMemMod.put.
  unfold ThrdMemMod.maps_to in * .
  unfold  TQMod.maps_to in * .
  unfold TQMod.beq_A.
  unfold TQSpec.beq_Domain.
  unfold ThrdMemMod.beq_A.
  unfold ThrdMemSpec.beq_Domain.
  destruct (eq_nat_dec tid0 tid).
  rewrite e in H3;destruct H3.
  rewrite H0 in H1;inversion H1.
  apply H with Q;assumption.
Qed.

Lemma enter_SW_grt_hyph:
forall T T' B M tid tcb,
  embed_WQSet T B M
  -> unblocked tid B
  -> T' = TQMod.put T tid tcb
  -> embed_WQSet T' B M.
Proof.
  intros.
  subst T'.
  induction H.
  apply wq_emp.
  assert (unblocked tid B).
  unfold unblocked in * .
  intros.
  apply H0 with w.
  unfold WQSetMod.maps_to in * .
  unfold WQSetMod.put.
  destruct (WQSetMod.beq_A w l).
  rewrite e in H4;rewrite H1 in H4;inversion H4.
  assumption.
  apply wq_cons with M M';try assumption.
  apply IHembed_WQSet.
  assumption.
  assert (filterTQ T Q = filterTQ (TQMod.put T tid tcb) Q).
  apply TQMod.extensional_eqMap.
  intros.
  unfold TQMod.put.
  unfold filterTQ.
  unfold unblocked in H0.
  destruct (TQMod.beq_A x tid).
  unfold WQSetMod.maps_to in H0.
  subst x.
  assert (Q tid = None).
  apply H0 with l.
  unfold WQSetMod.put.
  destruct (WQSetMod.beq_A l l).
  reflexivity.
  elim n;reflexivity.
  rewrite H5.
  reflexivity.
  reflexivity.
  rewrite H5 in H2;assumption.
Qed.


Lemma enter_SW_grt2_hyp0:
forall RdyQ RdyQ' tid tid',
  addThrd RdyQ tid = addThrd RdyQ' tid'
  -> WQMod.not_in_dom tid RdyQ
  -> WQMod.not_in_dom tid' RdyQ'
  -> tid <> tid'
  -> (WQMod.in_dom tid RdyQ' /\ WQMod.in_dom tid' RdyQ).
Proof.
  intros.
  unfold addThrd in H.
  split.
  assert (WQMod.put RdyQ tid dummy_sglton tid =  WQMod.put RdyQ' tid' dummy_sglton tid).
  rewrite H;reflexivity.
  assert ( WQMod.put RdyQ' tid' dummy_sglton tid = RdyQ' tid).
  apply WQMod.put_noninterfere.
  intro;apply H2;symmetry;assumption.
  rewrite H4 in H3.
  unfold WQMod.put in H3.
  destruct (WQMod.beq_A tid tid).
  exists dummy_sglton;symmetry;assumption.
  elim n;reflexivity.
  assert (WQMod.put RdyQ tid dummy_sglton tid' =  WQMod.put RdyQ' tid' dummy_sglton tid').
  rewrite H;reflexivity.
  assert (WQMod.put RdyQ tid dummy_sglton tid' = RdyQ tid').
  apply WQMod.put_noninterfere.
  assumption.
  rewrite H4 in H3.
  unfold WQMod.put in H3.
  destruct (WQMod.beq_A tid' tid').
  exists dummy_sglton;assumption.
  elim n;reflexivity.
Qed.

Lemma enter_SW_grt2_hyp1:
forall TM tid tid' cthMR sp'' pc'' S0 M0' M2' M22 M12 cthM2 cthM12,
  ThrdMemMod.maps_to TM tid' cthMR
  -> (sp'' |~>w pc'') cthMR
  -> S_RF S0 SP = sp''
  -> MemMod.disjUnion M0' M2' (S_Mem S0)
  -> disjUnionList (M22 :: M12 :: cthM2 :: cthM12 :: nil) M2'
  -> MemSplit (ThrdMemMod.remove TM tid) M12
  -> tid <> tid'
  -> Ret_Addr pc'' S0.
Proof.
  intros.
  induction S0 as (((Mx, Rx), flx), isrx).
  UnfoldStateAll.
  unfold Ret_Addr.
  intro pc.
  simplS.
  apply nextPC_retw with sp''; auto.
  apply leadsToW_Mono with M2' M0'; auto.
  auto with PMap.
  assert (DUL2mstar (M22 :: M12 :: cthM2 :: cthM12 :: nil) M2').
  apply DUL2mstar_left; auto.
  unfold DUL2mstar in H6.
  sep on.
  sep lift 1 in H6.
  unfold mstar at 1 in H6. sep off.
  InvertAll.
  unfold eqM in H6. subst x.
  apply leadsToW_Mono with M12 x0; auto.
  assert (MemMod.subseteq cthMR M12).
  apply memsplit_subheap with (ThrdMemMod.remove TM tid) tid'; auto.
  unfold ThrdMemMod.maps_to. unfold ThrdMemMod.remove.
  induction (ThrdMemMod.beq_A tid' tid); auto.
  absurd (tid' <> tid); auto.
  assert (exists M12_2, MemMod.disjUnion cthMR M12_2 M12).
  auto with PMap. inversion_clear H8.
  apply leadsToW_Mono with cthMR x; auto.
Qed.

Lemma non_ih_spec_miniD:
  forall rho D,
    (rho = rho_ready \/ rho = rho_idle) 
    -> Minimal_D D
    -> D rho = SomeT1 _ (Itp_AIM rho).
Proof.
  intros.
  inversion_clear H.
  rewrite H1.
  unfold Minimal_D in H0.
  InvertAll. auto.
  rewrite H1. unfold Minimal_D in H0. InvertAll. auto.
Qed.

Lemma minimalD_elim_Rdy:
  forall D, Minimal_D D -> D rho_ready = SomeT1 _ (Itp_AIM rho_ready).
Proof.
  intros.
  unfold Minimal_D in H.
  InvertAll. auto.
Qed.

Lemma minimalD_elim_idle:
  forall D, Minimal_D D -> D rho_idle = SomeT1 _ (Itp_AIM rho_idle).
Proof.
  intros.
  unfold Minimal_D in H.
  InvertAll. auto.
Qed.

Lemma minimalD_elim_ih:
  forall D, Minimal_D D -> D rho_ih = SomeT1 _ (Itp_AIM rho_ih).
Proof.
  intros.
  unfold Minimal_D in H.
  InvertAll. auto.
Qed.

Lemma enter_SW_grt2_aux2_hyp0:
forall INVM cthM0 cthM12 M22 M12 cthM2 M0' M2' S0 M21' cthM_oldStk,
  MemMod.disjUnion INVM cthM0 cthM12
  -> disjUnionList (M22 :: M12 :: cthM2 :: cthM12 :: nil) M2'
  -> MemMod.disjUnion M0' M2' (S_Mem S0)
  -> MemMod.disjUnion M21' cthM_oldStk M0'
  -> MemMod.disjoint (MemMod.merge cthM0 cthM2) cthM_oldStk.
Proof.
  intros.
  induction S0 as (((M, R), fl), isr).
  simplS in H1.
  assert ((eqM M21' *^* eqM cthM_oldStk *^* (DUL2mstar (M22 :: M12 :: cthM2 :: cthM12 :: nil))) M).
  exists M0'. exists M2'.
  split_l. split.
  exists M21'. exists cthM_oldStk. unfold eqM. auto.
  apply DUL2mstar_left; auto.
  unfold DUL2mstar in H3.
  sep on.
  sep lift 5 in H3.
  assert (forall M, eqM cthM12 M -> (eqM INVM *^* eqM cthM0) M).
  intros. unfold eqM in H4. subst M0. 
  exists INVM. exists cthM0. unfold eqM. auto.
  sep weaker_subst in H3 with (eqM INVM *^* eqM cthM0).
  assert ((eqM cthM0 *^* eqM cthM2 *^* eqM cthM_oldStk *^* mTrue) M).
  sep on.
  sep matchH H3 with 4.
  unfold mstar in H6.
  unfold eqM in H6.
  InvertAll.
  subst x2 x3 x4.
  inversion_clear H10.
  subst x1.
  sep off.
  inversion_clear H8; auto.
Qed.


Lemma enter_SW_grt2_aux2_hyp1:
forall pc' M0' M2' S0 S' M22 M12 cthM2 cthM12 M21' cthM_oldStk M122 cthMR INVM cthM0,
  MemMod.disjUnion M0' M2' (S_Mem S0)
  -> disjUnionList (M22 :: M12 :: cthM2 :: cthM12 :: nil) M2'
  -> MemMod.disjUnion M21' cthM_oldStk M0'
  -> MemMod.disjUnion M122 cthMR M12
  -> MemMod.disjUnion INVM cthM0 cthM12
  -> nextS pc' (comms _retw) S0 S'
  -> disjUnionList 
       (M22 :: cthMR :: INVM :: M21' :: MemMod.merge (MemMod.merge cthM0 cthM2) cthM_oldStk ::M122 :: nil) 
          (S_Mem S').
Proof.
  intros.
  induction S0 as (((Mx, Rx), flx), isrx).
  induction S' as (((My, Ry), fly), isry).
  UnfoldStateAll.
  destruct (nextS_inversion H4).
  subst My. clear H4 H6.
  assert ((eqM M21' *^* eqM cthM_oldStk *^* (DUL2mstar (M22 :: M12 :: cthM2 :: cthM12 :: nil))) Mx).
  exists M0'. exists M2'.
  split_l. split.
  exists M21'. exists cthM_oldStk. unfold eqM. auto.
  apply DUL2mstar_left; auto.
  unfold DUL2mstar in H4.
  sep on.
  sep lift 5 in H4.
  assert (forall M, eqM cthM12 M -> (eqM INVM *^* eqM cthM0) M).
  intros. unfold eqM in H5. subst M. 
  exists INVM. exists cthM0. unfold eqM. auto.
  sep weaker_subst in H4 with (eqM INVM *^* eqM cthM0).
  clear H5 H6.
  sep lift 5 in H4.
  assert (forall M, eqM M12 M -> (eqM M122 *^* eqM cthMR) M).
  intros. unfold eqM in H5. subst M. 
  exists M122. exists cthMR. unfold eqM. auto.
  sep weaker_subst in H4 with (eqM M122 *^* eqM cthMR).
  sep off.
  cut (DUL2mstar (M22 :: cthMR :: INVM :: M21' :: MemMod.merge (MemMod.merge cthM0 cthM2) cthM_oldStk ::M122 :: nil) Mx).
  apply DUL2mstar_right.
  unfold DUL2mstar.
  sep on.
  sep matchH H4 with 7.
  sep lift 3 in H4.
  sep off.
  unfold eqM.
  unfold mstar in H4.
  unfold eqM in H4.
  InvertAll.
  subst x x1 x3. 
  unfold Aemp in H13. subst x4.
  assert (cthM0 = x2). auto with PMap.
  subst x2.
  inversion_clear H8.
  assert (MemMod.merge cthM2 cthM0 = MemMod.merge cthM0 cthM2); auto with PMap.
  rewrite H8 in H9.
  assert (MemMod.disjUnion x0 cthM_oldStk h). auto with PMap.
  rewrite H9 in H11.
  inversion_clear H11; auto.
Qed.



Lemma enter_SW_grt2_hyp4_hyp0:
forall RdyQ RdyQ' tid tid' B,
  addThrd RdyQ tid = addThrd RdyQ' tid'
  -> tid <> tid'
  -> disj_RQ_WQ B RdyQ
  -> unblocked tid' B.
Proof.
  unfold unblocked;intros.
  unfold disj_RQ_WQ in H1.
  substH H1 with (H1 w Q H2).
  assert (WQMod.in_dom tid' RdyQ).
  unfold addThrd in H.
  assert ( WQMod.put RdyQ tid dummy_sglton tid' = RdyQ tid').
  apply WQMod.put_noninterfere;assumption.
  rewrite H in H3.
  unfold WQMod.put in H3.
  destruct (WQMod.beq_A tid' tid').
  exists dummy_sglton.
  symmetry;assumption.
  elim n;reflexivity.
  apply WQMod.disj_exclusive_left with RdyQ.
  apply WQMod.disjoint_commut;assumption.
  assumption.
Qed.

(* changed because of the change of WF_currTID *)
Lemma enter_SW_grt2_hyp4:
forall T T' B RdyQ RdyQ' tid tid' tcb tcb' rho rho',
  tid <> tid'
  -> addThrd RdyQ tid = addThrd RdyQ' tid'
  -> WF_currTID tid T B RdyQ rho
  -> T' = TQMod.put T tid tcb
  -> WQMod.not_in_dom tid' RdyQ'
  -> WF_TWQ T B RdyQ
  -> TQMod.maps_to T tid' tcb'
  -> TCB_Status tcb' = READY \/
        TCB_Status tcb' = IDLE
  -> rho_tcb_consistent tcb' rho'
  -> WF_currTID tid' T' B RdyQ' rho'.
Proof.
  intros.
  unfold WF_currTID.
  unfold WF_TWQ in H4.
  unfold WF_currTID in H1.
  InvertAll.
  split.
  apply enter_SW_grt2_hyp4_hyp0 with RdyQ RdyQ' tid; auto.
  split; auto.
  rewrite H2. unfold TQMod.in_dom. unfold TQMod.put.
  induction (TQMod.beq_A tid' tid).
  absurd (tid' <> tid); auto.
  split.
  exists tcb'. auto. exists tcb'.
  unfold TQMod.maps_to.
  induction (TQMod.beq_A tid' tid).
  contradiction. split; auto.
Qed.

Lemma enter_SW_grt2_hyp5_hyp0:
forall tid'' RdyQ tid tid' RdyQ',
  WQMod.in_dom tid'' RdyQ
  -> addThrd RdyQ tid = addThrd RdyQ' tid'
  -> tid' <> tid''
  -> WQMod.in_dom tid'' RdyQ'.
Proof.
  intros.
  unfold addThrd in H0.
  assert (WQMod.put RdyQ' tid' dummy_sglton tid'' = RdyQ' tid'').
  apply WQMod.put_noninterfere;assumption.
  rewrite <- H0 in H2.
  destruct H.
  unfold WQMod.put in H2.
  destruct (WQMod.beq_A tid'' tid).
  exists dummy_sglton.
  symmetry;assumption.
  exists x.
  rewrite H in H2;symmetry;assumption.
Qed.

Lemma enter_SW_grt2_hyp5:
forall T T' RdyQ RdyQ' tid tid' sp pc tcb,
  exists_Idle T RdyQ tid
  -> addThrd RdyQ tid = addThrd RdyQ' tid'
  -> T tid = Some tcb
  (* -> T tid' = Some tcb' *)
  -> tid <> tid'
  -> WQMod.not_in_dom tid' RdyQ'
  -> WQMod.not_in_dom tid RdyQ
  -> WQMod.in_dom tid RdyQ'
  -> WQMod.in_dom tid' RdyQ
  -> T' = TQMod.put T tid (sp, TCB_Status tcb, pc)
  -> exists_Idle T' RdyQ' tid'.
Proof.
  intros.
  unfold exists_Idle.
  unfold exists_Idle in H.
  inversion_clear H.
  InvertAll.
  assert (x = tcb). unfold TQMod.maps_to in H8.
  rewrite H8 in H1. inversion H1. auto.
  rewrite H in H9.
  rewrite H9 in H7.
  right. exists tid. exists (sp, IDLE, pc).
  split; auto. split; auto. rewrite H7.
  unfold TQMod.maps_to. unfold TQMod.put.
  induction (TQMod.beq_A tid tid). auto.
  absurd (tid = tid); auto.
  InvertAll. rename x into tid''. rename x0 into tcb''.
  assert (tid' = tid'' \/ tid' <> tid'').
  apply nat_eq_neq_dec. inversion_clear H9.
  left. exists tcb''. split. rewrite <- H11 in H8.
  rewrite H7. unfold TQMod.put. unfold TQMod.maps_to.
  induction (TQMod.beq_A tid' tid).
  absurd (tid' <> tid); auto. auto. auto.
  right. exists tid''. exists tcb''.
  split.
  apply enter_SW_grt2_hyp5_hyp0 with RdyQ tid tid'; auto.
  split; auto.
  rewrite H7. unfold TQMod.maps_to. unfold TQMod.put.
  induction (TQMod.beq_A tid'' tid).
  rewrite a in H. absurd (~ WQMod.in_dom tid RdyQ); auto with PMap.
  auto.
Qed.

Lemma enter_SW_grt2_aux6_hyp0:
forall B RdyQ RdyQ' tid tid',
  disj_RQ_WQ B RdyQ
  -> unblocked tid B
  -> addThrd RdyQ tid = addThrd RdyQ' tid'
  -> tid <> tid'
  -> disj_RQ_WQ B RdyQ'.
Proof.
  unfold addThrd;unfold disj_RQ_WQ.
  intros.
  substH H with (H l Q H3).
  unfold unblocked in H0.
  substH H0 with (H0 l Q H3).
  substH H with (WQMod.disjoint_commut H).
  assert (WQMod.disjoint Q (WQMod.put RdyQ tid dummy_sglton)).
  apply WQMod.disjoint_commut.
  apply WQMod.extend_presv_disj_left;assumption.
  rewrite H1 in H4.
  unfold WQMod.put in H4.
  unfold WQMod.disjoint in * .
  intro.
  destruct (H4 x).
  left;assumption.
  right.
  unfold WQMod.not_in_dom in * .
  destruct (WQMod.beq_A x tid' ).
  inversion H5.
  assumption.
Qed.

(* changed because of the change of WF_currTID *)
Lemma enter_SW_grt2_aux6:
forall RdyQ RdyQ' tid tid' tcb tcb' B T T' rho,
  addThrd RdyQ tid = addThrd RdyQ' tid'
  -> TQMod.maps_to T tid tcb
  -> WF_currTID tid T B RdyQ rho
  -> tid <> tid'
  -> WF_TWQ T B RdyQ
  -> T' = TQMod.put T tid tcb'
  -> WF_TWQ T' B RdyQ'.
Proof.
  intros.
  unfold WF_TWQ.
  split.
  rewrite H4. unfold WF_TWQ in H3. InvertAll.
  unfold WF_TQ. unfold WF_TQ in H5.
  intros.
  apply H5.
  assert (tid0 = tid \/ tid0 <> tid).
  apply nat_eq_neq_dec. inversion_clear H8.
  rewrite H9. unfold TQMod.in_dom. exists tcb. auto.
  unfold TQMod.put in H6. unfold TQMod.in_dom in H6.
  induction (TQMod.beq_A tid0 tid).
  contradiction. auto.
  unfold WF_TWQ in H3. InvertAll. split; auto.
  unfold WF_currTID in H1. InvertAll.
  apply enter_SW_grt2_aux6_hyp0 with RdyQ tid tid'; auto.
Qed.

Lemma enter_SW_grt2_hyp7:
forall cthMR INVM cthM_old M122 M'0 TM TM' tid tid',
  disjUnionList (cthMR :: INVM :: cthM_old :: M122 :: nil) M'0
  -> MemSplit (ThrdMemMod.remove (ThrdMemMod.remove TM tid) tid') M122
  -> TM' = ThrdMemMod.put (ThrdMemMod.put TM tid' (MemMod.merge cthMR INVM))
           tid cthM_old
  -> tid <> tid'
  -> MemSplit TM' M'0.
Proof.
  intros.
  rewrite H1.
  assert (DUL2mstar (cthMR :: INVM :: cthM_old :: M122 :: nil) M'0).
  apply DUL2mstar_left; auto.
  unfold DUL2mstar in H3.
  sep on.
  sep lift 2 in H3.
  sep off.
  unfold mstar in H3.
  unfold eqM in H3.
  InvertAll.
  subst x x3 x4 x5.
  unfold Aemp in H12.
  subst x6.
  assert (M122 = x1). auto with PMap.
  subst x1. clear H9.
  apply enter_SW_grt_hyp6 with x0; auto with PMap.

  assert (ThrdMemMod.remove (ThrdMemMod.put TM tid' (MemMod.merge cthMR INVM)) tid
           = ThrdMemMod.put (ThrdMemMod.remove TM tid) tid' (MemMod.merge cthMR INVM)).
  auto with PMap.
  rewrite H3. clear H3.
  inversion_clear H7.
  subst x2.
  apply enter_SW_grt_hyp6 with M122; auto.
Qed.


Lemma enter_SW_grt2_hyp8:
forall pc pc' M R fl isr M' R' fl' isr' S0 S' f,
  nextS pc (comml (_call f)) (M, R, fl, isr) (M', R', fl', isr')
  -> S_ISR (M', R', fl', isr') = S_ISR S0
  -> isr = false
  -> nextS pc' (comms _retw) S0 S'
  -> S_ISR S' = false.
Proof.
  intros.
  inversion H;subst.
  inversion H2;subst.
  simpl in H0;symmetry;assumption.
Qed.

Lemma enter_SW_grt2_hyp9:
forall S0 S' sp'' pc',
  S_RF S0 SP = sp''
  -> nextS pc' (comms _retw) S0 S'
  -> S_RF S' SP = sp'' + 2.
Proof.
  intros.
  inversion H0;subst.
  simpl.
  destruct H1 as [pc [P1 P2]].
  subst R'.
  rewrite (RF_update_RF R SP (R SP + 2)).
  reflexivity.
Qed.

Lemma enter_SW_grt2_hypa:
forall cthMR INVM M22 M21' cthM_old M122 S',
    disjUnionList
          (M22 :: cthMR :: INVM :: M21' :: cthM_old :: M122 :: nil)
          (S_Mem S')
    -> MemMod.disjoint cthMR INVM.
Proof.
  intros.
  induction S' as (((M, R), fl), isr).
  simplS in H.
  assert (DUL2mstar (M22 :: cthMR :: INVM :: M21' :: cthM_old :: M122 :: nil) M).
  apply DUL2mstar_left. auto.
  clear H.
  assert ((eqM cthMR *^* eqM INVM *^* mTrue) M).
  unfold DUL2mstar in H0.
  sep on.
  sep matchH H0 with 3.
  unfold mstar in H.
  unfold eqM in H.
  InvertAll.
  subst x1 x2.
  inversion_clear H2; auto.
Qed.

Lemma enter_SW_grt2_hypb:
forall M cthM_oldStk cthM_old R' pc junk,
  MemMod.disjUnion M cthM_oldStk cthM_old
  -> (R' SP |-> pc + 1 *^* R' SP + 1 |-> junk *^* stkSpacePrec (R' SP) 8) cthM_oldStk
  -> (R' SP |~>w pc + 1) cthM_old.
Proof.
  intros.
  apply leadsToW_Mono with cthM_oldStk M. auto with PMap.
  sep on. 
  assert (cthM_oldStk ||- R' SP |->w pc + 1 *^* stkSpacePrec (R' SP) 8).
  apply ptoW_intro1 with junk; auto.
  sep trivial.
Qed.



Lemma enter_SW_grt2_hypf:
forall RdyQ RdyQ' tid tid' tid0,
  WQMod.in_dom tid0 RdyQ'
  -> tid0 <> tid
  -> addThrd RdyQ tid = addThrd RdyQ' tid'
  -> WQMod.in_dom tid0 RdyQ.
Proof.
  intros.
  apply enter_SW_grt2_hyp5_hyp0 with RdyQ' tid' tid;
    [idtac|symmetry|intro;apply H0;symmetry];assumption.
Qed.

(* changed because of the change of WF_currTID *)
Lemma enter_SW_grt2_hypg:
forall Psi T B TM TM' RdyQ RdyQ' tid tid' T' tcb M M' rho,
  WF_WQSet Psi T B TM
  -> addThrd RdyQ tid = addThrd RdyQ' tid'
  -> WF_currTID tid T B RdyQ rho
  -> WF_TWQ T B RdyQ
  -> T' = TQMod.put T tid tcb
  -> TM' = ThrdMemMod.put (ThrdMemMod.put TM tid' M') tid M
  -> tid <> tid'
  -> WF_WQSet Psi T' B TM'.
Proof.
  unfold WF_WQSet.
  intros.
  assert (
      exists tcb : Thread,
        exists M : Mem,
          TQMod.maps_to T tid0 tcb /\
          ThrdMemMod.maps_to TM tid0 M /\ WF_Wait_Thrd Psi M tcb l).
  apply H with Q; auto.
  InvertAll. exists x. exists x0.
  assert (tid <> tid0).
  assert (tid = tid0 \/ tid <> tid0).
  apply nat_eq_neq_dec.
  inversion_clear H10; auto.
  subst tid0.
  unfold WF_currTID in H1.
  InvertAll.
  unfold unblocked in H10.
  assert (WQMod.not_in_dom tid Q).
  apply H10 with l; auto.
  absurd (~ WQMod.in_dom tid Q); auto with PMap.
  split.
  subst T'.
  unfold TQMod.maps_to. unfold TQMod.put.
  induction (TQMod.beq_A tid0 tid); auto.
  absurd (tid0 <> tid); auto.
  split_r.
  subst TM'.
  assert (tid' <> tid0).
  assert (tid' = tid0 \/ tid' <> tid0). 
  apply nat_eq_neq_dec.
  inversion_clear H4; auto.
  clear H.
  unfold WF_TWQ in H2.
  InvertAll.
  unfold disj_RQ_WQ in H13.
  assert (WQMod.disjoint Q RdyQ).
  apply H13 with l; auto.
  unfold WQMod.disjoint in H4.
  assert (WQMod.not_in_dom tid' Q \/ WQMod.not_in_dom tid' RdyQ).
  auto.
  inversion_clear H14.
  rewrite H12 in H15.
  absurd (~ WQMod.in_dom tid0 Q); auto with PMap.
  assert (WQMod.in_dom tid' RdyQ).
  unfold addThrd in H0.
  assert (WQMod.maps_to (WQMod.put RdyQ' tid' dummy_sglton) tid' dummy_sglton).
  auto with PMap.
  rewrite <- H0 in H14.
  unfold WQMod.put in H14. unfold WQMod.maps_to in H14.
  induction (WQMod.beq_A tid' tid).
  absurd (tid' <> tid); auto.
  unfold WQMod.in_dom. exists dummy_sglton. auto.
  absurd (~ WQMod.in_dom tid' RdyQ); auto with PMap.
  unfold ThrdMemMod.maps_to. unfold ThrdMemMod.put.
  induction (ThrdMemMod.beq_A tid0 tid).
  absurd (tid0 <> tid); auto.
  induction (ThrdMemMod.beq_A tid0 tid'); auto.
  absurd (tid0 <> tid'); auto.
Qed.


(* changed because of the change of WF_currTID *)
Lemma enter_SW_grt2_hyph:
forall M'1 M21' M22 T' tid' RdyQ' T B tcb RdyQ tid rho,
  MemMod.disjUnion M21' M22 M'1
  -> (embed_curr_TCB T' tid' *^* embed_RdyQ T' RdyQ') M21'
  -> embed_WQSet T B M22
  -> T' = TQMod.put T tid tcb
  -> WF_currTID tid T B RdyQ rho
  -> (embed_curr_TCB T' tid' *^* (embed_RdyQ T' RdyQ' *^* embed_WQSet T' B))
     M'1.
Proof.
  intros.
  assert (embed_WQSet T' B M22).
  destruct H3.
  apply enter_SW_grt_hyph with T tid tcb;assumption.
  sep on.
  assert (M'1||- (embed_curr_TCB T' tid' *^* embed_RdyQ T' RdyQ' *^* embed_WQSet T' B)).
  apply sconj_intro with M21' M22;assumption.
  sep matchH H5 with 3.
Qed.

Lemma enter_SW_grt2_hypI:
forall pc pc' M R fl isr M' R' fl' isr' S0 S',
  nextS pc (comml (_call _SWITCH)) (M, R, fl, isr) (M', R', fl', isr')
  -> S_Flag (M', R', fl', isr') _IF = S_Flag S0 _IF
  -> fl _IF = false
  -> nextS pc' (comms _retw) S0 S'
  -> S_Flag S' _IF  = false.
Proof.
  intros.
  inversion H;subst.
  inversion H2;subst.
  destruct H6.
  simpl in H0.
  simpl.
  rewrite H0 in H1;assumption.
Qed.

Lemma rho_tcb_consistent_ih:
  forall tcb, rho_tcb_consistent tcb rho_ih.
Proof.
  intro.
  unfold rho_tcb_consistent.
  split. intro. inversion H. intro. inversion H.
Qed.
Hint Resolve rho_tcb_consistent_ih.


Lemma enter_SW_grt2_hyp3:
forall M22 cthMR INVM M21' cthM_old M122 S',
    disjUnionList
          (M22 :: cthMR :: INVM :: M21' :: cthM_old :: M122 :: nil)
          (S_Mem S')
    -> exists M'0, exists M'1, 
        MemMod.disjUnion M'0 M'1 (S_Mem S')
        /\ MemMod.disjUnion M21' M22 M'1
        /\ disjUnionList (cthMR :: INVM :: cthM_old :: M122 :: nil) M'0.
Proof.
  intros.
  induction S' as (((Mx, Rx), flx), isrx).
  UnfoldStateAll.
  assert ((eqM M21' *^* eqM M22 *^* (DUL2mstar (cthMR :: INVM :: cthM_old :: M122 :: nil))) Mx).
  assert (DUL2mstar (M22 :: cthMR :: INVM :: M21' :: cthM_old :: M122 :: nil) Mx).
  apply DUL2mstar_left; auto.
  unfold DUL2mstar in *.
  sep on.
  sep matchH H0 with 7.
  unfold mstar at 1 in H0.
  InvertAll.
  exists x0.  exists x.
  split. auto with PMap.
  split.
  unfold mstar in H0. unfold eqM in H0.
  InvertAll.
  subst x1 x2. auto.
  apply DUL2mstar_right; auto.
Qed.

Lemma subsetCT_presv_loopup:
forall theta pc Psi Psi',
  subsetCT Psi Psi' -> lookupCT Psi pc theta -> lookupCT Psi' pc theta.
Proof.
  unfold lookupCT.
  unfold subsetCT.
  intros.
  apply OPSpec.lookup_sub_lookup with Psi; auto.
Qed.

Lemma psi_sw_lookupNPC:
forall rho psi Psi Psi' Psi'0 p' g' tag pc,
  rho = rho_ready \/ rho = rho_idle
  -> lift_psi psi rho Psi
  -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
  -> subsetCT Psi' Psi'0
  -> SCAP_lookupCT psi pc (Ast p' g' tag)
  -> lookupCT Psi'0 pc (packS0 rho (Ast p' g' tag)).
Proof.
  intros.
  subst Psi'.
  apply subsetCT_presv_loopup with (ext_IH_Spec (ext_Prim_Specs Psi)); auto.
  unfold lift_psi in H0.
  assert (lookupCT Psi pc (packS0 rho (Ast p' g' tag))).
  apply H0; auto.
  assert (subsetCT Psi (ext_IH_Spec (ext_Prim_Specs Psi))).
  unfold ext_Prim_Specs.
  unfold ext_IH_Spec.
  apply Psi_extend_subsetCT.
  apply Psi_extend_subsetCT.
  apply Psi_extend_subsetCT.
  apply Psi_extend_subsetCT.
  apply subsetCT_refl.
  apply subsetCT_presv_loopup with Psi; auto.
Qed.

Lemma enter_SW_grt2_auxc_hyp0:
forall cthM0 cthM2 cthM_oldStk cthM_old M0 M'2,
  MemMod.disjUnion (MemMod.merge cthM0 cthM2) cthM_oldStk cthM_old
  -> MemMod.disjUnion M0 cthM_old M'2
  -> MemMod.disjoint cthM0 cthM2
  -> MemMod.disjUnion (MemMod.merge (MemMod.merge cthM0 M0) cthM_oldStk) cthM2
     M'2.
Proof.
  intros.
  assert ((eqM M0 *^* (eqM cthM0 *^* eqM cthM2 *^* eqM cthM_oldStk)) M'2).
  exists M0. exists cthM_old.
  split_l. split. unfold eqM. auto.
  exists (MemMod.merge cthM0 cthM2). exists cthM_oldStk.
  split_l. unfold eqM. split_r.
  exists cthM0. exists cthM2.
  split; auto.
  split; auto.
  assert ((eqM cthM0 *^* eqM M0 *^* eqM cthM_oldStk *^* eqM cthM2) M'2).
  sep on.
  sep matchH H2 with 4.
  unfold mstar in H3. unfold eqM in H3.
  InvertAll.
  subst x0 x2 x3 x4.
  inversion_clear H7.
  subst x1.
  inversion_clear H5. subst x. auto.
Qed.

Lemma enter_SW_grt2_hype:
forall M0 cthM_old M'2 cthM0 cthM2 cthM_oldStk M'20 M'21 fl fl0,
  MemMod.disjUnion M0 cthM_old M'2
  -> MemMod.disjUnion (MemMod.merge cthM0 cthM2) cthM_oldStk cthM_old
  -> MemMod.disjUnion M'20 M'21 M'2
  -> NH_Inv fl0 M'21
  -> NH_Inv fl cthM2
  -> fl _IF = false
  -> fl0 _IF = false
  -> M'20 = MemMod.merge (MemMod.merge cthM0 M0) cthM_oldStk.
Proof.
  intros.
  unfold NH_Inv in * .
  rewrite H4 in H3;rewrite H5 in H2.
  rewrite H2 in H1.
  rewrite H3 in H0.
  destruct H1.
  rewrite (MemMod.unit_emp_merge M'20) in H6.
  rewrite (MemMod.unit_emp_merge cthM0) in H0.
  subst.
  assert (MemMod.merge cthM0 M0 = MemMod.merge M0 cthM0).
  apply MemMod.disj_merge_commut.
  disjoint lift_l.
  rewrite H6.
  destruct H0;destruct H. 
  rewrite (MemMod.merge_assoc M0 cthM0 cthM_oldStk).
  rewrite <- H7;rewrite <- H8;reflexivity.
Qed.

Lemma WF_AIM_World_RdyNotBlk:
  forall Psi'0 S T B RdyQ tid theta w BlkQ rho,
     WF_AIM_World Psi'0 S T B RdyQ tid rho theta
  -> B w = Some BlkQ
  -> WQMod.disjoint RdyQ BlkQ.
Proof.
  intros.
  unfold WF_AIM_World in H.
  InvertAll.
  unfold WF_TWQ in H2.
  InvertAll.
  unfold disj_RQ_WQ in H11.
  assert (WQMod.disjoint BlkQ RdyQ).
  apply H11 with w; auto.
  auto with PMap.
Qed.

(*
Lemma call_fun_frame:
  forall M M' R R' fl fl' isr isr' M2 M12 cthM2 cthM12 cthM11 pc f w stkspace,
  disjUnionList (M2 :: M12 :: cthM2 :: cthM12 :: cthM11 :: nil) M
  -> (R SP |->w w *^* stkSpacePrec (R SP) (stkspace + 2)) cthM11
  -> nextS pc (comml (_call f)) (M, R, fl, isr) (M', R', fl', isr')
  -> exists cthM11',
        ((R' SP |->w (pc+1)) *^* (R' SP + 2 |->w w) *^* (stkSpacePrec (R' SP) stkspace)) cthM11'
        /\ disjUnionList (M2 :: M12 :: cthM2 :: cthM12 :: cthM11' :: nil) M'.
Proof.
  intros.
  substH H1 with (nextS_inversion H1).
  simpl in H1.
  destruct H1 as [P1 [P2 [P3 P4]]].
  destruct P1 as [P11 [P12 P13]].
  destruct P13.
  exists ( MemUpdW cthM11 (R SP - 2) (pc + 1) x).
  assert ( (R SP-2 |->w pc + 1 *^* R SP |->w w *^* stkSpacePrec (R SP-2) stkspace)
     (MemUpdW cthM11 (R SP - 2) (pc + 1) x)).
  sep on.
  sep lift 0.
  sep upd.
  sep matchH H0 with 3.
  sep stksplit in H0.
  arith_rep (stkspace +2-2) with stkspace in H0.
  split.
  subst R'.
  rewrite (RF_update_RF R SP (R SP - 2)).
  arith_rep (R SP - 2 + 2) with (R SP).
  substH H with (disjUnionList_elim H).
  apply disjUnionList_intro.
  unfold itr in * .
  subst M'.
  sep lift 4.
  replace (MemMod.put (MemMod.put M (R SP - 2) (pc + 1)) (R SP - 2 + 1) x) with 
    ((MemUpdW M (R SP - 2) (pc + 1) x)).
  sep upd.
  sep matchH H with 6.
  sep on.
  sep lift 1 in H0.
  sep stksplit in H0.
  sep lift 1 in H0.
  sep trivial.
  trivial.
Qed.
*)

Lemma call_fun_frame2:
  forall M M' R R' fl fl' isr isr' M2 M12 cthM2 cthM12 cthM11 pc f w stkspace,
  disjUnionList (M2 :: M12 :: cthM2 :: cthM12 :: cthM11 :: nil) M
  -> (R SP |->w w *^* stkSpacePrec (R SP) (stkspace + 2)) cthM11
  -> nextS pc (comml (_call f)) (M, R, fl, isr) (M', R', fl', isr')
  -> exists junk1,
        (R' SP |-> (pc+1) *^* R' SP + 1 |-> junk1 *^* (R' SP + 2 |->w w) *^* (stkSpacePrec (R' SP) stkspace)) (MemUpdW cthM11 (R' SP) (pc + 1) junk1)
        /\ disjUnionList (M2 :: M12 :: cthM2 :: cthM12 :: (MemUpdW cthM11 (R' SP) (pc + 1) junk1) :: nil) M'.
Proof.
  intros.
  assert ((eqM cthM11 *^* DUL2mstar (M2 :: M12 :: cthM2 :: cthM12 :: nil)) M).
  assert (DUL2mstar (M2 :: M12 :: cthM2 :: cthM12 :: cthM11 :: nil) M).
  apply DUL2mstar_left; auto.
  unfold DUL2mstar in *.
  sep on. sep matchH H2 with 6.
  unfold mstar at 1 in H2.
  InvertAll.
  unfold eqM in H2.
  subst x.

  destruct (nextS_inversion H1).
  InvertAll.
  substH H10 with (MemUpd16_MemUpdW H10).
  destruct H10 as [junk H10].
  exists junk.

  split.
  unfold MemUpdW.
  sep on.
  sep lift 1. sep upd2.
  sep lift 3. sep upd2.
  subst R'. reg simpl.
  assert (R SP - 2 + 2 = R SP). auto.
  rewrite H6.
  sep lift 1 in H0.
  sep stksplit in H0.
  assert (stkspace + 2 -  2= stkspace). auto.
  rewrite H9 in H0.
  sep lift 3.
  sep match_left H0; auto.
  assert (MemMod.disjUnion (MemUpdW cthM11 (R' SP) (pc+1) junk) x0 M').
  subst R'. reg simpl.
  assert (MemMod.in_dom (R SP - 2) cthM11 /\ MemMod.in_dom (R SP - 2 + 1) cthM11).
  sep on.
  sep lift 1 in H0.
  sep stksplit in H0.
  sep lift 1 in H0.
  apply ptoxW_indom1 with 
    (R SP |->w w *^* stkSpacePrec (R SP - 2) (stkspace + 2 - 2)); auto.
  inversion_clear H6.
  subst M'.
  unfold MemUpdW.
  apply MemMod.update_presv_disjUnion_left.
  apply MemMod.update_presv_disjUnion_left; auto.
  unfold MemMod.in_dom. unfold MemMod.put.
  induction (MemMod.beq_A (R SP - 2 + 1) (R SP - 2)); auto.
  exists (pc+1). auto.
  set (cthM11' := MemUpdW cthM11 (R' SP) (pc + 1) junk) in *.
  cut (DUL2mstar (M2 :: M12 :: cthM2 :: cthM12 :: cthM11' :: nil) M').
  intro.
  apply DUL2mstar_right; auto.
  unfold DUL2mstar.
  sep on.
  sep lift 4.
  exists cthM11'. exists x0.
  split_l. split. unfold eqM. auto.
  unfold DUL2mstar in H5.
  sep on.
  sep matchH H5 with 5.
Qed.



Lemma split_blk_Q:
forall T tid RdyQ B M w BlkQ,
  (embed_curr_TCB T tid *^* (embed_RdyQ T RdyQ *^* embed_WQSet T B)) M
  -> B w = Some BlkQ
  -> ((embed_curr_TCB T tid *^* embed_RdyQ T RdyQ *^* (ptoTQ w (filterTQ T BlkQ))) *^* (embed_WQSet T (WQSetMod.remove B w))) M.
Proof.
  intros.
  sep on.
  sep matchH H with 3.
  sep off.
  assert (B = (WQSetMod.put (WQSetMod.remove B w) w BlkQ)).
  rewrite (WQSetMod.put_cancel_remove B w BlkQ).
  apply WQSetMod.extensional_eqMap.
  intro.
  unfold WQSetMod.put.
  destruct (WQSetMod.beq_A x w).
  rewrite e;assumption.
  reflexivity.
  rewrite H1 in H.
  assert (WQSetMod.not_in_dom w (WQSetMod.remove B w)).
  auto with PMap.
  substH H with (embed_WQSet_inv_cons T  (WQSetMod.remove B w) w BlkQ M  H H2).
  destruct H as [M1 [M2 [P1 [P2 P3]]]].
  exists M2;exists M1.
  intuition.
Qed.

Lemma enable_unblk_monotonic:
  forall w wqa, monotonic_p (enable_unblk w wqa).
Proof.
  intros.
  unfold enable_unblk.
  set (p0 := fun S0 => (pTrue *<*
          (S_RF S0 SP |->w w *^*
           (wqa *^* stkSpacePrec (S_RF S0 SP) (_UNBLK_Stk + 2)))) S0).
  set (p1 := fun (S0 : State) => Delta w = SomeT wqa).
  assert (monotonic_p p0).

  unfold monotonic_p.
  intros.
  induction S as (((M0, R0), fl0), isr0). unfold p0. 
  unfold p0 in H. simplS. simplS in H.
  unfold S_Mem in H0.
  assert (exists M0', MemMod.disjUnion M0 M0' M).
  apply MemMod.subseteq_split; auto.
  unfold pmstar in H. simplS in H. InvertAll.
  unfold pmstar. simplS. exists (MemMod.merge x0 x). exists x1.
  unfold pTrue.
  split; auto.
  cut (MemMod.disjUnion x1 (MemMod.merge x0 x) M). auto with PMap.
  apply MemMod.disjUnion_assoc_left_right; auto.
  inversion_clear H1; auto with PMap.
  assert (MemMod.disjUnion x1 x0 M0). auto with PMap.
  inversion H3. rewrite H6 in H2. auto.

  assert (monotonic_p p1).
  unfold monotonic_p.
  intros.
  unfold p1. unfold p1 in H0. auto.
  assert (monotonic_p (fun S => p1 S /\ p0 S)).
  unfold monotonic_p.
  intros.
  InvertAll.
  split.
  unfold monotonic_p in H0. apply H0; auto. unfold monotonic_p in H. apply H; auto.
  unfold p0 in H1. unfold p1 in H1. auto.
Qed.
Hint Resolve enable_unblk_monotonic.

Lemma enter_BLK_grt2_auxc_hyp1:
forall M M1 M2 M11 M12 M21 M22,
  MemMod.disjUnion M1 M2 M
  -> MemMod.disjUnion M11 M12 M1
  -> MemMod.disjUnion M21 M22 M2
  -> MemMod.disjoint M21 M11.
Proof.
  intros.
  assert (MemMod.disjoint M1 M21).
  inversion_clear H1.
  subst M2.
  inversion_clear H.
  apply MemMod.disj_x_merge_disj_left with M22; auto.
  inversion_clear H0. subst M1.
  cut (MemMod.disjoint M11 M21). intro. auto with PMap.
  apply MemMod.disj_merge_disj_left with M12; auto.
Qed.


Lemma enter_BLK_grt2_auxc_hyp2:
forall M M1 M2 M21 M22 M221 M222,
  MemMod.disjUnion M1 M2 M
  -> MemMod.disjUnion M21 M22 M2
  -> MemMod.disjUnion M221 M222 M22
  -> MemMod.disjoint M1 M221.
Proof.
  intros.
  assert (MemMod.disjoint M1 M22).
  inversion_clear H0. subst M2.
  apply MemMod.disj_x_merge_disj_right with M21; auto.
  inversion H; auto.
  inversion H1. subst M22.
  apply MemMod.disj_x_merge_disj_left with M222; auto.
Qed.

Lemma enter_BLK_grt2_auxc_hyp4:
forall M'2 M0 cthM_old cthM0 cthM2 cthM_oldStk oldStk1 oldStk2,
  MemMod.disjUnion (MemMod.merge cthM0 cthM2) cthM_oldStk cthM_old
  -> MemMod.disjUnion M0 cthM_old M'2
  -> MemMod.disjUnion oldStk1 oldStk2 cthM_oldStk
  -> MemMod.disjUnion (MemMod.merge M0 oldStk1) (MemMod.merge oldStk2 cthM0)
     (MemMod.merge (MemMod.merge cthM0 M0) cthM_oldStk).
Proof.
  intros.
  assert ((eqM cthM0 *^* eqM M0 *^* (eqM oldStk1 *^* eqM oldStk2)) 
             (MemMod.merge (MemMod.merge cthM0 M0) cthM_oldStk)).
  exists (MemMod.merge cthM0 M0).
  exists cthM_oldStk.
  split.
  split; auto.
  apply MemMod.disj_merge_B.
  inversion_clear H.
  apply MemMod.disj_merge_disj_left with cthM2; auto.
  inversion_clear H.
  subst cthM_old.
  apply MemMod.disj_x_merge_disj_right with (MemMod.merge cthM0 cthM2); auto.
  inversion H0; auto.
  split.
  exists cthM0. exists M0.
  unfold eqM. split; auto.
  split; auto.
  assert (MemMod.disjoint M0 (MemMod.merge cthM0 cthM2)).
  inversion_clear H.
  subst cthM_old.
  apply MemMod.disj_x_merge_disj_left with cthM_oldStk; auto.
  inversion H0; auto.
  assert (MemMod.disjoint M0 cthM0).
  apply MemMod.disj_x_merge_disj_left with cthM2; auto.
  auto with PMap.
  exists oldStk1. exists oldStk2. unfold eqM. auto.
  set (Mx := MemMod.merge (MemMod.merge cthM0 M0) cthM_oldStk) in *.
  assert ((eqM M0 *^* eqM oldStk1 *^* (eqM oldStk2 *^* eqM cthM0)) Mx).
  sep on.
  sep matchH H2 with 4.
  unfold mstar in H3. unfold eqM in H3.
  InvertAll.
  subst x1 x2 x3 x4.
  inversion_clear H7. inversion_clear H5.
  subst x x0. auto.
Qed.


Lemma enter_BLK_grt2_auxc_hyp5:
forall M'2 M0 cthM_old cthM0 cthM2 cthM_oldStk sp w junk m,
  MemMod.disjUnion (MemMod.merge cthM0 cthM2) cthM_oldStk cthM_old
  -> MemMod.disjUnion M0 cthM_old M'2
  -> (sp |-> w *^* sp + 1 |-> junk *^* m) cthM_oldStk
  -> (sp |~>w w) (MemMod.merge (MemMod.merge cthM0 M0) cthM_oldStk).
Proof.
  intros.
  assert ((sp |->w w *^* m) cthM_oldStk).
  apply ptoW_intro1 with junk; auto. clear H1.
  assert ((eqM cthM0 *^* eqM M0 *^* eqM cthM_oldStk) 
             (MemMod.merge (MemMod.merge cthM0 M0) cthM_oldStk)).
  exists (MemMod.merge cthM0 M0).
  exists cthM_oldStk.
  split.
  split; auto.
  apply MemMod.disj_merge_B.
  inversion_clear H.
  apply MemMod.disj_merge_disj_left with cthM2; auto.
  inversion_clear H.
  subst cthM_old.
  apply MemMod.disj_x_merge_disj_right with (MemMod.merge cthM0 cthM2); auto.
  inversion H0; auto.
  split.
  exists cthM0. exists M0.
  unfold eqM. split; auto.
  split; auto.
  assert (MemMod.disjoint M0 (MemMod.merge cthM0 cthM2)).
  inversion_clear H.
  subst cthM_old.
  apply MemMod.disj_x_merge_disj_left with cthM_oldStk; auto.
  inversion H0; auto.
  assert (MemMod.disjoint M0 cthM0).
  apply MemMod.disj_x_merge_disj_left with cthM2; auto.
  auto with PMap.
  unfold eqM. auto.
  sep on.
  sep lift 2 in H1.
  assert (forall M, eqM cthM_oldStk M -> (sp |->w w *^* m) M).
  intros. unfold eqM in H3. subst M. auto.
  sep weaker_subst in H1 with (sp |->w w *^* m).
  sep on.
  sep lift 0 in H1.
  sep trivial.
Qed.


Lemma enter_SW_grt2_auxc_auxx:
forall cthM1 cthM11 cthM12 cthM11_1 cthM11_2 INVM cthM0 cthM' sp pc junk,
  MemMod.disjUnion cthM11 cthM12 cthM1
  -> MemMod.disjUnion INVM cthM0 cthM12
  -> MemMod.disjUnion cthM11_1 cthM11_2 cthM11
  -> (sp - 2 |->w_ ) cthM11_2
  -> cthM' = MemUpdW cthM1 (sp - 2) (pc + 1) junk
  -> MemMod.disjUnion (MemMod.merge (MemUpdW cthM11_2 (sp - 2) (pc+1) junk) cthM0) (MemMod.merge INVM cthM11_1) cthM'.
Proof.
  intros.

  assert ((eqM cthM11_1 *^* eqM cthM11_2 *^* (eqM INVM *^* eqM cthM0)) cthM1).
  exists cthM11. exists cthM12. split; auto.
  split. exists cthM11_1. exists cthM11_2. unfold eqM.
  auto. exists INVM. exists cthM0. unfold eqM. auto.

  assert (MemMod.disjUnion (MemUpdW (MemMod.merge cthM11_2 cthM0) (sp-2) (pc+1) junk) (MemMod.merge INVM cthM11_1) cthM').
  subst cthM'.
  unfold MemUpdW.
  apply MemMod.update_presv_disjUnion_left.
  apply MemMod.update_presv_disjUnion_left.

  assert ((eqM cthM11_2 *^* eqM cthM0 *^* (eqM INVM *^* eqM cthM11_1)) cthM1).
  sep on. sep matchH H4 with 4.
  unfold mstar at 1 in H3. InvertAll.
  unfold mstar in H3. InvertAll.
  unfold eqM in H3, H9. subst x1 x2.
  inversion_clear H6. subst x. 
  unfold mstar in H7.
  InvertAll. unfold eqM in H7, H9. subst x x1.
  inversion_clear H6. subst x0. auto.
  apply MemMod.merge_presv_indom_left.
  eapply enter_BLK_grt2_auxc_hyp01; eauto.
  assert (MemMod.in_dom (sp - 2 + 1) (MemMod.merge cthM11_2 cthM0)).
  apply MemMod.merge_presv_indom_left.
  eapply enter_BLK_grt2_auxc_hyp02; eauto.
  unfold MemMod.in_dom. unfold MemMod.put.
  induction (MemMod.beq_A (sp - 2 + 1) (sp - 2)). exists (pc+1). auto.
  auto.
  assert (MemMod.disjUnion (MemUpdW cthM11_2 (sp - 2) (pc + 1) junk)
           cthM0 (MemUpdW (MemMod.merge cthM11_2 cthM0) (sp - 2) (pc + 1) junk)).
  unfold MemUpdW.
  apply MemMod.update_presv_disjUnion_left.
  apply MemMod.update_presv_disjUnion_left.
  split; auto.
  assert ((eqM cthM11_2 *^* eqM cthM0 *^* (eqM INVM *^* eqM cthM11_1)) cthM1).
  sep on. sep matchH H4 with 4.
  unfold mstar at 1 in H6. InvertAll.
  unfold mstar in H6. InvertAll.
  unfold eqM in H6, H11. subst x1 x2.
  inversion_clear H8. auto.
  eapply enter_BLK_grt2_auxc_hyp01; eauto.
  assert (MemMod.in_dom (sp - 2 + 1) cthM11_2).
  eapply enter_BLK_grt2_auxc_hyp02; eauto.
  unfold MemMod.in_dom. unfold MemMod.put.
  induction (MemMod.beq_A (sp - 2 + 1) (sp - 2)). exists (pc+1). auto.
  auto.
  inversion_clear H6.
  rewrite H8 in H5. auto.
Qed.

Lemma enter_SW_grt2_auxc_hypx:
forall sp w junk oldStk2 cthM11_2,
  (sp |-> w *^* sp + 1 |-> junk) oldStk2
  -> (sp |->w_ ) cthM11_2
  -> MemUpdW cthM11_2 sp w junk = oldStk2.
Proof.
  intros.
  assert ((sp |-> w *^* sp+1 |-> junk) (MemUpdW cthM11_2 sp w junk)).
  sep on.
  sep lift 1.
  unfold MemUpdW.
  sep upd2. sep lift 1.
  sep upd2.
  auto.
  assert (uniqueM (sp |-> w *^* sp + 1 |-> junk)).
  apply mstar_presv_uniqueM; apply pto_uniqueM.
  unfold uniqueM in H2.
  apply H2; auto.
Qed.


Lemma ext_unblockedT_presv_WQSet:
  forall T tid tcb B M,
    embed_WQSet T B M
    -> unblocked tid B
    -> embed_WQSet (TQMod.put T tid tcb) B M.
Proof.
  intros T tid tcb B M H.
  induction H.
  intro. apply wq_emp.
  intro.
  apply wq_cons with M M'; auto.
  apply IHembed_WQSet.
  unfold unblocked in *.
  intros.
  apply H3 with w.
  unfold WQSetMod.maps_to. unfold WQSetMod.put.
  induction (WQSetMod.beq_A w l).
  subst w.
  unfold WQSetMod.not_in_dom in H0.
  unfold WQSetMod.maps_to in H4. rewrite H4 in H0.
  inversion H0. auto.
  assert (filterTQ T Q = filterTQ (TQMod.put T tid tcb) Q).
  apply TQMod.extensional_eqMap.
  unfold filterTQ.
  intro x.
  assert ((exists y, Q x = Some y) \/ Q x = None).
  induction (Q x); auto. left. exists a. auto.
  inversion_clear H4.
  inversion_clear H5. rewrite H4.
  unfold TQMod.put.
  induction (TQMod.beq_A x tid).
  subst x.
  unfold unblocked in H3.
  assert (WQMod.not_in_dom tid Q).
  apply H3 with l.
  auto with PMap. unfold WQMod.not_in_dom in H5.
  rewrite H5 in H4. inversion H4. auto.
  rewrite H5. auto.
  rewrite <- H4. auto.
Qed.

Lemma enable_sw_nxt_stkspace:
forall p pc S S' M1 M2,
  enable_sw p
  -> nextS pc (comml (_call _SWITCH)) S S'
  -> MemMod.disjUnion M1 M2 (S_Mem S)
  -> imprec p (S_Mem_Upd S M1)
  -> exists M1',
        MemMod.disjUnion M1' M2 (S_Mem S')
        /\ (S_RF S' SP |->w (pc+1) *^* stkSpacePrec (S_RF S' SP) 8 *^* mTrue) M1'.
Proof.
  intros.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  UnfoldStateAll.
  unfold enable_sw in H.
  assert ((mTrue *^*
             (INV0 *^* INV1 *^* stkSpacePrec (R SP) (_SW_Stk + 2))) M1).
  unfold imprec in H2.
  unfold pmstar in H2.
  InvertAll.
  simplS in H3. simplS in H2.
  assert ((pTrue *<* (INV0 *^* INV1 *^* stkSpacePrec (S_RF (x, R, fl, isr) SP) (_SW_Stk + 2)))
              (x, R, fl, isr)).
  apply H; auto.
  unfold S_RF in H4.
  unfold pmstar in H4.
  InvertAll.
  exists (MemMod.merge x1 x0). exists x2. 
  split. unfold S_Mem in H6.
  assert (MemMod.disjUnion x2 (MemMod.merge x1 x0) M1).
  apply MemMod.disjUnion_assoc_left_right; auto.
  inversion_clear H6. auto with PMap.
  assert (MemMod.disjUnion x2 x1 x); auto with PMap.
  inversion_clear H7.
  rewrite <- H10. auto. auto with PMap.
  unfold mTrue. auto.
  clear H.
  assert (frame (nextS pc (comml (_call _SWITCH)))).
  apply frame_call.
  unfold frame in H.
  assert (exists M1', 
        MemMod.disjUnion M1' M2 (S_Mem (M', R', fl', isr')) /\
        nextS pc (comml (_call _SWITCH)) (S_Mem_Upd (M, R, fl, isr) M1) (S_Mem_Upd (M', R', fl', isr') M1')).
  apply H; auto.
  clear H.
  UnfoldStateAll.
  unfold enableS.
  exists (MemUpdW M1 (R SP - 2) (pc+1) 0, RF_update R SP (R SP - 2), fl, isr).
  apply nextS_call.
  unfold pushW.
  split.
  sep on.
  sep lift 3 in H3.
  sep stksplit 2 in H3. sep lift 0 in H3.
  sep trivial.
  split; auto.
  apply  MemUpdW_MemUpd16 with 0; auto.
  clear H.
  UnfoldStateAll.
  InvertAll.
  exists x.
  split_l.
  destruct (nextS_inversion H5).
  InvertAll.
  subst R'. reg simpl.
  substH H11 with (MemUpd16_MemUpdW H11).
  InvertAll.
  subst x.
  sep on.
  sep lift 3 in H3.
  sep stksplit in H3.
  sep lift 1 in H3.
  assert (MemUpdW M1 (R SP - 2) (pc + 1) x0 ||- 
          R SP - 2 |->w (pc+1) *^* 
           (mTrue *^* (INV0 *^* INV1) *^*
            stkSpacePrec (R SP - 2) (_SW_Stk + 2 - 2))).
  sep upd. auto.
  assert (MemUpdW M1 (R SP - 2) (pc + 1) x0
         ||- R SP - 2 |->w pc + 1 *^* stkSpacePrec (R SP - 2) (_SW_Stk + 2 - 2) *^* 
             ((INV0 *^* INV1) *^* mTrue)).
  sep matchH H7 with 5.
  destruct H10 as [M1x [M1y [Hsplit [HM1x HM1y]]]].
  exists M1x. exists M1y.
  split; auto.
  unfold mTrue. split; auto.
Qed.



Lemma enable_sw_stkspace:
forall p S,
  enable_sw p -> p S -> 
      exists M1, exists M2, 
        MemMod.disjUnion M1 M2 (S_Mem S) 
        /\ stkSpacePrec (S_RF S SP) (8 + 2) M1 
        /\ ((INV0 *^*INV1) *^* mTrue) M2.
Proof.
  intros.
  unfold enable_sw in H.
  assert ((pTrue *<* (INV0 *^* INV1 *^* stkSpacePrec (S_RF S SP) (_SW_Stk + 2))) S).
  apply H; auto.
  induction S as (((M, R), fl), isr).
  simplS. simplS in H1.
  assert ((mTrue *^* (INV0 *^* INV1 *^* stkSpacePrec (R SP) (_SW_Stk + 2))) M).
  unfold pmstar in H1.
  InvertAll.
  exists x. exists x0. unfold mTrue.
  auto.
  sep on.
  sep lift 3 in H2.
  sep off.
  unfold mstar at 1 in H2.
  InvertAll.
  exists x. exists x0.
  split_l. split_l.
  sep on. sep matchH H5 with 3.
Qed.


Lemma enter_SW_grt_hype:
forall S' cthM'1 cthM'2 cthM11'' cthM12 cthM2 cthM' fl,
  NH_Inv (S_Flag S') cthM'2
  -> fl _IF = S_Flag S' _IF
  -> MemMod.disjUnion cthM'1 cthM'2 cthM'
  -> disjUnionList (cthM2 :: cthM12 :: cthM11'' :: nil) cthM'
  -> NH_Inv fl cthM2
  -> (cthM2 = cthM'2) /\ (cthM'1 = MemMod.merge cthM12 cthM11'').
Proof.
  intros.
  assert (exists cthM'1', MemMod.disjUnion cthM2 cthM'1' cthM' /\ disjUnionList (cthM12 :: cthM11'' :: nil) cthM'1').
  unfold disjUnionList in *.
  inversion_clear H2.
  exists x.
  inversion_clear H4. split; auto.
  InvertAll. rename x into cthM'1'.
  assert (cthM2 = cthM'2).
  unfold NH_Inv in *.
  rewrite <- H0 in H.
  induction (fl _IF).
  assert (precise (INV0 *^* INV1)).
  apply mstar_presv_prec; auto.
  apply Prec0. apply Prec1.
  unfold precise in H5.
  apply H5 with cthM'.
  apply MemMod.disjUnion_subseteq_left with cthM'1'; auto.
  apply MemMod.disjUnion_subseteq_right with cthM'1; auto.
  auto. auto.
  assert (precise Aemp).
  auto. unfold precise in H5.
  apply H5 with cthM'; auto.
  apply MemMod.disjUnion_subseteq_left with cthM'1'; auto.
  apply MemMod.disjUnion_subseteq_right with cthM'1; auto.
  split_l.
  assert (cthM'1' = cthM'1).
  rewrite <- H5 in H1.
  apply MemMod.disjUnion_unique with cthM2 cthM2 cthM'; auto.
  auto with PMap.
  rewrite <- H7.
  unfold disjUnionList in H6.
  InvertAll.
  subst x0.
  assert (cthM11'' = x). auto with PMap.
  subst x.
  inversion_clear H9. auto.
Qed.

