Add LoadPath "..".
Add LoadPath "../CAP0".

Require Import Arith.
Require Import Coqlib.
Require Import Map.
Require Import Machine.
Require Import Assertion.
Require Import Assertion_Lemmas.
Require Import Operational_Semantics.
Require Import OpSem_Lemmas.

Require Import CAP0.
Require Import SCAP.
Require Import AIM.
Require Import QLib.
Require Import SCAP_READY.
Require Import SCAP_IDLE.
Require Import SCAP_Itr.

Require Import EqNat.

Require Import Linking.

Lemma Prd_Itr_Stk_elim : forall Psi rho p g tag M R fl isr,
  Prd_Itr_Stk p
  -> itr_enabled (M, R, fl, isr)
  -> itp_NH_theta rho (Ast p g tag) Psi (M, R, fl, isr)
  -> stkSpace (R SP) (IH_Stk+6) M.
Proof.
  intros Psi rho p g tag M R fl isr Hpstk Hen Hitp.
  unfold itp_NH_theta in Hitp.
  destruct Hitp as [Hisrf [Hp Hwfst]].
  unfold NH_pinvstar, pmstar, mstar in Hp.
  destruct Hp as [M1 [M2 [HMJM [HpM1 HM2]]]].
  unfold Prd_Itr_Stk in Hpstk.
  simplS in HpM1.
  substH Hpstk with (Hpstk _ HpM1).
  assert (H : itr_enabled (M1, R, fl, isr)).
    simplS in * ; auto.
  substH Hpstk with (Hpstk H); clear H.
  destruct Hpstk as [M11 [M12 [HJM1 [HM11 HM12]]]].
  simplS in * .
  substH HM11 with (stkSpacePrec_stkSpace HM11).
  apply stkSpace_Mono with M1 M2; trivial.
  apply stkSpace_Mono with M12 M11; trivial.
  auto with PMap.
Qed.
Implicit Arguments Prd_Itr_Stk_elim [Psi rho p g tag M R fl isr].

Lemma NH_WFST_Mono_CT : forall rho Psi Psi' p,
  NH_WFST rho Psi p
  -> subsetCT Psi Psi'
  -> NH_WFST rho Psi' p.
Proof.
  unfold NH_WFST, NH_WFST_aux; fold NH_WFST_aux.
  intros rho Psi Psi' p Hwfst Hsub.
  destruct Hwfst as [n Hwfst].
  exists n.
  generalize p Hwfst; clear p Hwfst.
  induction n; simpl; intros p Hwfst.
    intros S' Hp.
    substH Hwfst with (Hwfst S' Hp).
    destruct Hwfst as [ra [Hra [p' [g' [tag' [Hlk [Htag' Hwfst]]]]]]].
    exists ra; split_l.
    exists p'; exists g'; exists tag'.
    split.
      apply Hsub; trivial.
    split_lr.
  simpl in * .
  intros S' Hp.
  substH Hwfst with (Hwfst S' Hp).
  destruct Hwfst as [ra [Hra [p' [g' [tag' [Hlk Hwfst]]]]]].
  exists ra; split_l.
  exists p'; exists g'; exists tag'.
  split.
    apply Hsub; trivial.
  intros pc S'' HnextS'.
  substH Hwfst with (Hwfst pc S'' HnextS').
  destruct Hwfst as [Hp' Hg'].
  split_l.
  apply IHn; trivial.
Qed.
Implicit Arguments NH_WFST_Mono_CT [rho Psi Psi' p].

Lemma itp_NH_theta_Mono_CT : forall rho theta,
  Mono_CT_assert (itp_NH_theta rho theta).
Proof.
  unfold Mono_CT_assert, itp_NH_theta.
  intros rho [p g tag] Psi S H Psi' Hsub.
  destruct H as [Hisr [Hp Hwfst]].
  split_l.
  split_l.
  intros Htag.
  substH Hwfst with (Hwfst Htag).
  eapply NH_WFST_Mono_CT; eauto.
Qed.
Implicit Arguments itp_NH_theta_Mono_CT [rho theta].

Lemma itp_IH_theta_Mono_CT : forall theta status,
  Mono_CT_assert (itp_IH_theta theta status).
Proof.
  unfold Mono_CT_assert, itp_IH_theta.
  intros [p g tag] status Psi S H Psi' Hsub.
  destruct H as [Hif [H | [Htag [Hp Hwfst]]]].
    split_l.
    left.
    unfold lift_IH_PG in * .
    destruct H as [Htag [Hp H]].
    split_l; split_l.
    intros S' Hg.
    substH H with (H S' Hg).
    destruct H as [S'' [HnextS' [ra [Hra [rho [theta [Hconsis2 [Hrho [Hlk H]]]]]]]]].
    exists S''; split_l.
    exists ra; split_l.
    exists rho; exists theta; split_l.
    split. auto.
    split.
      apply Hsub; trivial.
    intros pc' S''' HnextS''.
    substH H with (H pc' S''' HnextS'').
    apply (itp_NH_theta_Mono_CT Psi S''' H Psi' Hsub).
  split_l.
  right.
  split_l.
  split_l.
  destruct Hwfst as [n Hwfst].
  exists n.
  generalize g S Hwfst; clear Hif Htag Hp g S Hwfst.
  induction n; simpl; intros g S Hwfst.
    intros S' Hg.
    substH Hwfst with (Hwfst S' Hg).
    destruct Hwfst as [ra [Hra [theta' [Hlk Hwfst]]]].
    exists ra; split_l.
    exists theta'.
    split.
      apply Hsub; trivial.
    intros pc' S'' HnextS'.
    substH Hwfst with (Hwfst pc' S'' HnextS').
    destruct theta' as [p' g' tag'].
    unfold lift_IH_PG in * .
    destruct Hwfst as [Htag' [Hp' Hwfst]].
    split_l; split_l.
    intros S''' Hg'.
    substH Hwfst with (Hwfst S''' Hg').
    destruct Hwfst as [S'''' [HS'''' [ra' [Hra' [rho [theta'' [Hrho [Hconsis2 [Hlk'' Hwfst]]]]]]]]].
    exists S''''; split_l.
    exists ra'; split_l.
    exists rho; exists theta''; split_l.
    split; auto.
    split.
      apply Hsub; trivial.
    intros pc'' Sx HnextS'''.
    substH Hwfst with (Hwfst pc'' Sx HnextS''').
    apply (itp_NH_theta_Mono_CT Psi Sx Hwfst Psi' Hsub).
  intros S' Hg.
  substH Hwfst with (Hwfst S' Hg).
  destruct Hwfst as [ra [Hra [p' [g' [tag' [Hlk [Htag' Hwfst]]]]]]].
  exists ra; split_l.
  exists p'; exists g'; exists tag'.
  split.
    apply Hsub; trivial.
  split_l.
  intros pc S'' HnextS'.
  substH Hwfst with (Hwfst pc S'' HnextS').
  destruct Hwfst as [Hp' Hg'].
  split_l.
  apply IHn; trivial.
Qed.
Implicit Arguments itp_IH_theta_Mono_CT [theta].

(* changed by Xinyu, because of the change of
   WF_ReadyQ *)
Lemma WF_ReadyQ_Mono_CT : forall Psi T RdyQ TM,
  WF_ReadyQ Psi T RdyQ TM
  -> forall Psi', subsetCT Psi Psi'
    -> WF_ReadyQ Psi' T RdyQ TM.
Proof.
  intros.
  unfold WF_ReadyQ.
  intros tid Hindom.
  substH H with (H tid Hindom).
  destruct H as [tcb [Hmto [Htcb H]]].
  destruct H as [M [HM [Htcbsp [Hth | Hih]]]].
  exists tcb; split_l; split_l.
  exists M; split_l; split_l.
  left.
  destruct tcb as [[l status] pc].
  unfold WF_Rdy_Thrd in * .
  destruct Hth as [rho [p [g [tag [Hrho [Hconsis [Hlds [Hlk Hth]]]]]]]].
  exists rho; exists p; exists g; exists tag.
  split_l; split_l; split_l.
  split.
    apply H0; trivial.
  intros R fl Hsp Hfl.
  substH Hth with (Hth R fl Hsp Hfl).
  unfold pm_mwand in * .
  simplS in * .
  intros M1 M' HJM' HM1.
  substH Hth with (Hth M1 M' HJM' HM1).
  apply (itp_NH_theta_Mono_CT Psi (M', R, fl, false) Hth Psi' H0).
  exists tcb; split_l; split_l. 
  exists M; split_l; split_l.
  right.
  destruct tcb as [[l status] pc].
  unfold WF_Rdy_Thrd_IH in * .
  destruct Hih as [p [g [tag [Hlk [Hlds Hih]]]]].
  exists p; exists g; exists tag.
  split.
    apply H0; trivial.
  split_l.
  intros R fl HR Hfl.
  substH Hih with (Hih R fl HR Hfl).
  unfold pm_mwand in * .
  intros M1 M' HJM' HM1.
  substH Hih with (Hih M1 M' HJM' HM1).
  apply (itp_IH_theta_Mono_CT status Psi (S_Mem_Upd (M, R, fl, false) M') Hih Psi' H0).
Qed.

Lemma WF_WQSet_Mono_CT : forall Psi T Bset TM,
  WF_WQSet Psi T Bset TM
  -> forall Psi', subsetCT Psi Psi'
    -> WF_WQSet Psi' T Bset TM.
Proof.
  intros Psi T Bset TM H Psi' Hsub.
  unfold WF_WQSet in * .
  intros l Q tid Hmapsto Hindom.
  substH H with (H l Q tid Hmapsto Hindom).
  clear Hmapsto Hindom.
  destruct H as [tcb [M [Htcb [HM Hl]]]].
  exists tcb; exists M. 
  split_l; split_l.
  destruct tcb as [[id status] pc].
  unfold WF_Wait_Thrd in * .
  destruct Hl as [Hst [HidM Hl]].
  destruct Hl as [wqa [Hwqa [p [g [tag [Hlk H]]]]]].
  split_l; split_l.
  exists wqa; split_l.
  exists p; exists g; exists tag.
  split.
    apply Hsub; trivial.
  intros R fl HR Hfl.
  substH H with (H R fl HR Hfl).
  unfold pm_mwand in * .
  intros M1 M' HJM' HM1.
  substH H with (H M1 M' HJM' HM1).
  apply (itp_NH_theta_Mono_CT Psi (S_Mem_Upd (M, R, fl, false) M') H Psi' Hsub).
Qed.

Lemma Itp_AIM_Mono_CT : forall rho theta,
  Mono_CT_assert (Itp_AIM rho theta).
Proof.
  unfold Mono_CT_assert, Itp_AIM.
  intros rho [p g tag] Psi S H Psi' Hsub.
  destruct S as [[[M R] fl] isr].
  destruct H as [T [B [RdyQ [tid H]]]].
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar in * .
  destruct H as [M1 [M2 [HJM H]]].
  destruct H as [H1 H2].
  exists M1; exists M2; split_l; split_r.
  simplS in * .
  clear M M2 HJM H2.
  unfold WF_AIM_World in * .
  simplS in * .
  destruct H1 as [H1 [H2 [H3 [H4 H]]]].
  split_l; split_l; split_l; split_l.
  clear H1 H2 H3 H4.
  destruct H as [TM [Mcth [HSplit [HTM [H1 [H2 H]]]]]].
  exists TM; exists Mcth; split_l; split_l.
  clear HSplit HTM.
  split.
    unfold WF_Curr_Thrd in * .
    destruct (beq_nat rho rho_ih).
    InvertAll.
    exists x. exists x0. split_l. split_l.
      apply (itp_IH_theta_Mono_CT x0 Psi (Mcth, R, fl, isr) H4 Psi' Hsub).
    apply (itp_NH_theta_Mono_CT Psi (Mcth, R, fl, isr) H1 Psi' Hsub).
  split.
    clear rho p g tag R fl isr B M1 Mcth H1 H.
    eapply WF_ReadyQ_Mono_CT; eauto.
  clear p g tag R fl isr RdyQ M1 H1 H2.
  eapply WF_WQSet_Mono_CT; eauto.
Qed.
Implicit Arguments Itp_AIM_Mono_CT [rho theta].

Lemma itp_NH_theta_Mono_CT_consCT : forall f theta0 Psi rho theta S,
  itp_NH_theta rho theta Psi S
  -> itp_NH_theta rho theta (consCT f theta0 Psi) S.
Proof.
  intros.
  assert (Hsub : subsetCT Psi (consCT f theta0 Psi)).
    auto.
  apply (itp_NH_theta_Mono_CT Psi S H _ Hsub).
Qed.

Lemma itp_IH_theta_Mono_CT_consCT : forall f theta0 status Psi theta S,
  itp_IH_theta theta status Psi S
  -> itp_IH_theta theta status (consCT f theta0 Psi) S.
Proof.
  intros.
  assert (Hsub : subsetCT Psi (consCT f theta0 Psi)).
    auto.
  apply (itp_IH_theta_Mono_CT status Psi S H _ Hsub).
Qed.

Lemma WF_ReadyQ_Mono_CT_consCT : forall f theta0 Psi T RdyQ TM,
  WF_ReadyQ Psi T RdyQ TM
  -> WF_ReadyQ (consCT f theta0 Psi) T RdyQ TM.
Proof.
  intros.
  assert (Hsub : subsetCT Psi (consCT f theta0 Psi)).
    auto.
  eapply WF_ReadyQ_Mono_CT; eauto.
Qed.

Lemma WF_WQSet_Mono_CT_consCT : forall f theta0 Psi T Bset TM,
  WF_WQSet Psi T Bset TM
  -> WF_WQSet (consCT f theta0 Psi) T Bset TM.
Proof.
  intros.
  assert (Hsub : subsetCT Psi (consCT f theta0 Psi)).
    auto.
  eapply WF_WQSet_Mono_CT; eauto.
Qed.

Lemma memsplit_thrd_remove:
  forall tm m, MemSplit tm m ->
    forall tid m', ThrdMemMod.maps_to tm tid m' 
      -> exists m'', MemMod.disjUnion m'' m' m /\ MemSplit (ThrdMemMod.remove tm tid) m''.
Proof.
  intros tm m H.
  induction H.
  intros.
  unfold ThrdMemMod.maps_to in H.
  unfold ThrdMemMod.emp in H.
  inversion H.
  intros.
  unfold ThrdMemMod.maps_to in H2.
  unfold ThrdMemMod.put in H2.
  induction (ThrdMemMod.beq_A tid0 tid).
  assert (M'' = m'). inversion_clear H2; auto.
  rewrite H3 in H.
  exists M'. split; auto.
  rewrite a.
  assert (ThrdMemMod.remove (ThrdMemMod.put TM tid M'') tid = ThrdMemMod.remove TM tid).
  auto with PMap.
  rewrite H4.
  assert (TM = ThrdMemMod.remove TM tid).
  auto with PMap.
  rewrite <- H5. auto.

  assert (exists m'', MemMod.disjUnion m'' m' M' /\ MemSplit (ThrdMemMod.remove TM tid0) m'').
  apply IHMemSplit.
  unfold ThrdMemMod.maps_to.
  auto.
  InvertAll.
  exists (MemMod.merge x M'').
  assert (MemMod.disjUnion m' (MemMod.merge x M'') M).
  apply MemMod.disjUnion_assoc_left_right.
  unfold MemMod.disjUnion in H3.
  inversion_clear H3. auto with PMap.
  unfold MemMod.disjUnion in H3.
  inversion_clear H3. 
  assert (MemMod.merge x m' = MemMod.merge m' x).
  auto with PMap.
  rewrite <- H3. rewrite <- H6. auto.
  split.
  auto with PMap.
  assert (tid <> tid0). auto.
  assert (ThrdMemMod.remove (ThrdMemMod.put TM tid M'') tid0 = ThrdMemMod.put (ThrdMemMod.remove TM tid0) tid M'').
  auto with PMap.
  rewrite H7.
  apply MS_split with x; auto.
  unfold MemMod.disjUnion. split; auto.
  unfold MemMod.disjUnion in H. inversion_clear H.
  unfold MemMod.disjUnion in H3. inversion_clear H3.
  rewrite H10 in H8.
  apply MemMod.disj_merge_disj_left with m'; auto.
  auto with PMap.
Qed.

Lemma MemSplit_Inv : forall TM M Mc tid,
  MemSplit TM M
  -> ThrdMemMod.maps_to TM tid Mc
  -> exists Mo, MemMod.disjUnion Mc Mo M /\ MemSplit (ThrdMemMod.remove TM tid) Mo.
Proof.
  intros.
  destruct (memsplit_thrd_remove _ _ H tid Mc H0).
  destruct H1.
  exists x.
  split_r.
  auto with PMap.
Qed.
Implicit Arguments MemSplit_Inv [TM M Mc tid].

Lemma nextS_DMono2 : forall pc c tid Mcth TM Mth Mq M M'
  R fl isr R' fl' isr' S0,
  ThrdMemMod.maps_to TM tid Mcth
  -> MemSplit TM Mth
  -> MemMod.disjUnion Mth Mq M
  -> nextS pc c (Mcth, R, fl, isr) S0
  -> nextS pc c (M, R, fl, isr) (M', R', fl', isr')
  -> exists Mcth', exists Mth', exists TM',
    ThrdMemMod.maps_to TM' tid Mcth'
    /\ MemSplit TM' Mth'
    /\ MemMod.disjUnion Mth' Mq M'
    /\ nextS pc c (Mcth, R, fl, isr) (Mcth', R', fl', isr')
    /\ (forall tidx Mx, tidx <> tid 
        -> ThrdMemMod.maps_to TM tidx Mx 
        -> ThrdMemMod.maps_to TM' tidx Mx).
Proof.
  intros pc c tid Mcth TM Mth Mq M M'.
  intros R fl isr R' fl' isr' S0.
  intros HtoMcth HsplitTM HJM HnextS0 HnextS.
  destruct S0 as [[[M0 R0] fl0] isr0].
  destruct (MemSplit_Inv HsplitTM HtoMcth) as [Moth [HJMth HsplitMoth]].
  assert (Hx : MemMod.disjUnion Mcth (MemMod.merge Moth Mq) M).
    destruct (@disjUnion_assoc_l Mcth Moth Mq Mth (MemMod.merge Moth Mq) M)
      as [HJMcth2Mo HJMth']; trivial.
  assert (exists Mcth' : MemMod.Map,
    MemMod.disjUnion Mcth' (MemMod.merge Moth Mq) M' /\
    nextS pc c (Mcth, R, fl, isr) (Mcth', R', fl', isr')).
    eapply (nextS_DMono Hx HnextS0 HnextS); eauto.
  destruct H as [Mcth' [HJM' HnextS1]].
  clear M0 R0 fl0 isr0 HnextS0.
  exists Mcth'.
  exists (MemMod.merge Mcth' Moth).
  exists (ThrdMemMod.put TM tid Mcth').
  assert (HeqMcth : MemMod.eqDom Mcth Mcth').
    eapply nextS_eqDom; eauto.
  split_l.
  split.
    assert (H : MemSplit (ThrdMemMod.put (ThrdMemMod.remove TM tid) tid Mcth') (MemMod.merge Mcth' Moth)).
      apply (MS_split (MemMod.merge Mcth' Moth) Moth Mcth' (ThrdMemMod.remove TM tid) tid); 
        auto with PMap.
      apply MemMod.disjUnion_commut.
      eapply disjUnion_upd_left_disjUnion; eauto.
    assert (Hy : ThrdMemMod.put (ThrdMemMod.remove TM tid) tid Mcth'
      = ThrdMemMod.put TM tid Mcth').
      auto with PMap.
    rewrite Hy in H; trivial.
  split.
    assert (H3 : MemMod.eqDom Mcth Mcth').
      eauto with PMap.
    set (Mth' := (MemMod.merge Mcth' Moth)).
    assert (HJMth' : MemMod.disjUnion Mcth' Moth Mth').
      split_r.
      destruct HJMth.
      eauto with PMap.
    assert (H4 : MemMod.eqDom Mth Mth').
      eauto with PMap.
    split.
      destruct HJM.
      eauto with PMap.
    destruct HJM'.
    unfold Mth'.
    subst M'.
    rewrite MemMod.merge_assoc; trivial.
  split_l.
  intros.
  unfold ThrdMemMod.maps_to, ThrdMemMod.put in H0 |- * .
  destruct (ThrdMemMod.beq_A tidx tid).
    destruct (H e).
  trivial.
Qed.
Implicit Arguments nextS_DMono2 [pc c tid Mcth TM Mth Mq M M'
  R fl isr R' fl' isr' S0].

Lemma nextS_Mono2 : forall Mcth TM tid Mth Mq M Mcth'
  pc c R fl isr R' fl' isr',
  MemSplit TM Mth
  -> MemMod.disjUnion Mth Mq M
  -> ThrdMemMod.maps_to TM tid Mcth
  -> nextS pc c (Mcth, R, fl, isr) (Mcth', R', fl', isr')
  -> exists M', nextS pc c (M, R, fl, isr) (M', R', fl', isr').
Proof.
  intros Mcth TM tid Mth Mq M Mcth' pc c R fl isr R' fl' isr'.
  intros Hsplit HJM Htid HnextS.
  destruct (MemSplit_Inv Hsplit Htid) as [Moth [HJMth _]].
  assert (Hx : MemMod.disjUnion Mcth (MemMod.merge Moth Mq) M).
    destruct (@disjUnion_assoc_l Mcth Moth Mq Mth (MemMod.merge Moth Mq) M)
      as [HJMcth2Mo HJMth']; trivial.
  destruct (nextS_Mono Hx HnextS) as [M' [_ Hy]]. 
  exists M'; trivial.
Qed.
Implicit Arguments nextS_Mono2
  [Mcth TM tid Mth Mq M Mcth' pc c R fl isr R' fl' isr'].

Lemma NH_WFST_aux_strengthen :
  forall rho n Psi S0 g S1 g',
    NH_WFST_aux rho n Psi (NH_ItpG g S0)
    -> pimp (NH_ItpG g' S1) (NH_ItpG g S0)
    -> NH_WFST_aux rho n Psi (NH_ItpG g' S1).
Proof.
  destruct n; simpl in * ; intros.
    apply H.
    apply H0; trivial.
  apply H.
  apply H0; trivial.
Qed.
Implicit Arguments NH_WFST_aux_strengthen [rho n Psi S0 g S1 g'].

Lemma NH_WFST_strengthen :
  forall rho Psi S0 g S1 g',
    NH_WFST rho Psi (NH_ItpG g S0)
    -> pimp (NH_ItpG g' S1) (NH_ItpG g S0)
    -> NH_WFST rho Psi (NH_ItpG g' S1).
Proof.
  unfold NH_WFST; intros.
  destruct H.
  exists x.
  eapply NH_WFST_aux_strengthen; eauto.
Qed.
Implicit Arguments NH_WFST_strengthen [rho Psi S0 g S1 g'].

Lemma IH_WFST_strengthen :
  forall n Psi S0 g S1 g' status,
    IH_WFST n Psi (IH_ItpG g S0) status
    -> pimp (IH_ItpG g' S1) (IH_ItpG g S0)
    -> IH_WFST n Psi (IH_ItpG g' S1) status.
Proof.
  unfold IH_WFST.
  destruct n; simpl in * ; intros.
    apply H.
    apply H0; trivial.
  apply H.
  apply H0; trivial.
Qed. 
Implicit Arguments IH_WFST_strengthen [n Psi S0 g S1 g' status].

Lemma WF_CurrTID_nh_ih : forall tid T B RdyQ rho,
  WF_currTID tid T B RdyQ rho
  -> WF_currTID tid T B RdyQ rho_ih.
Proof.
  unfold WF_currTID; intros tid T B RdyQ rho H.
  destruct H as [HB [HRQ [HT [tcb [Hmto [Htcb Hrho]]]]]].
  split_l; split_l; split_l.
  exists tcb; split_l; split_l.
  unfold rho_tcb_consistent.
  split; intro H; discriminate H.
Qed.
Implicit Arguments WF_CurrTID_nh_ih [tid T B RdyQ rho].

Lemma WF_CurrTID_ih_nh : forall tid tcb T B RdyQ rho status,
  WF_currTID tid T B RdyQ rho_ih
  -> TQMod.maps_to T tid tcb
  -> TCB_Status tcb = status
  -> rho = rho_ready \/ rho = rho_idle
  -> rho_tcb_consistent2 status rho
  -> WF_currTID tid T B RdyQ rho.
Proof.
  unfold WF_currTID; intros tid tcb T B RdyQ rho status H Hmto Htcb Hrho Hst.
  destruct H as [HB [HRQ [HT [tcb' [Hmto' [Htcb' H]]]]]].
  assert (tcb' = tcb).
    unfold TQMod.maps_to in * . 
    rewrite Hmto' in Hmto. 
    injection Hmto; auto.
  subst tcb'; clear Hmto' Htcb'.
  destruct H.
  unfold rho_tcb_consistent2 in Hst.
  destruct Hst.
  subst status.
  split_l; split_l; split_l.
  exists tcb; split_l. 
  split.
    destruct Hrho; auto.
  unfold rho_tcb_consistent.
  destruct Hrho as [Hrho | Hrho]; subst rho;
  split; intro; auto.
Qed.
Implicit Arguments WF_CurrTID_ih_nh [tid tcb T B RdyQ rho status].

(* chaged by Xinyu : added rho because of the change of
   WF_currTID and WF_ReadyQ *)
Lemma WF_ReadyQ_untouch : forall Psi T RdyQ TM TM' tid Bset rho,
  WF_ReadyQ Psi T RdyQ TM
  -> (forall tidx Mx, tidx <> tid 
        -> ThrdMemMod.maps_to TM tidx Mx 
        -> ThrdMemMod.maps_to TM' tidx Mx)
  -> WF_currTID tid T Bset RdyQ rho
  -> WF_ReadyQ Psi T RdyQ TM'.
Proof.
  unfold WF_ReadyQ, WF_currTID.
  intros.
  substH H with (H tid0 H2). 
  destruct H as [tcb [Hmto [Htcb [M [HmtoM [Htcbsp Hih]]]]]].
  exists tcb; split_l; split_l.
  exists M.
  split.
    apply H0; trivial.
    destruct H1.
    destruct H1.
    intro Hf.
    unfold WQMod.not_in_dom in H1.
    unfold WQMod.in_dom in H2.
    destruct H2.
    subst tid0.
    rewrite H1 in H2.
    discriminate H2.
  split_lr.
Qed.
Implicit Arguments WF_ReadyQ_untouch [Psi T RdyQ TM TM' tid Bset rho].

(* changed by Xinyu *)
Lemma WF_WQSet_untouch : forall Psi T Bset RdyQ TM TM' tid rho,
  WF_WQSet Psi T Bset TM
  -> (forall tidx Mx, tidx <> tid 
        -> ThrdMemMod.maps_to TM tidx Mx 
        -> ThrdMemMod.maps_to TM' tidx Mx)
  -> WF_currTID tid T Bset RdyQ rho
  -> WF_WQSet Psi T Bset TM'.
Proof.
  unfold WF_WQSet, WF_currTID.
  intros.
  substH H with (H l Q tid0 H2 H3).
  destruct H as [tcb [M [H4 [H5 H6]]]].
  exists tcb; exists M; split_l.
  split_r.
  destruct H1.
  destruct H1.
  apply H0; trivial.
  intro Hf.
  subst tid0.
  unfold unblocked in H.
  unfold WQMod.not_in_dom in H.
  unfold WQMod.in_dom in H3.
  destruct H3.
  rewrite (H _ _ H2) in H3.
  discriminate H3.
Qed.
Implicit Arguments WF_WQSet_untouch [Psi T Bset RdyQ TM TM' tid rho].

Lemma itr_nextS_Mono2 : forall Mcth TM tid Mth Mq M pc R fl isr Mcth' R' fl' isr',
  MemSplit TM Mth
  -> MemMod.disjUnion Mth Mq M
  -> ThrdMemMod.maps_to TM tid Mcth
  -> itr_nextS pc (Mcth, R, fl, isr) (Mcth', R', fl', isr')
  -> exists M', 
    itr_nextS pc (M, R, fl, isr) (M', R', fl', isr').
Proof.
  intros Mcth TM tid Mth Mq M pc R fl isr Mcth' R' fl' isr'.
  intros Hsplit HJM Htid HnextS.
  destruct (MemSplit_Inv Hsplit Htid) as [Moth [HJMth _]].
  assert (Hx : MemMod.disjUnion Mcth (MemMod.merge Moth Mq) M).
    destruct (@disjUnion_assoc_l Mcth Moth Mq Mth (MemMod.merge Moth Mq) M)
      as [HJMcth2Mo HJMth']; trivial.
  destruct (itr_nextS_Mono Hx HnextS) as [M' [_ Hy]].
  exists M'; trivial.
Qed.
Implicit Arguments itr_nextS_Mono2 [Mcth TM tid Mth Mq M pc R fl isr 
  Mcth' R' fl' isr'].

Lemma itr_nextS_DMono2 : forall pc tid Mcth TM Mth Mq M M'
  R fl isr R' fl' isr',
  ThrdMemMod.maps_to TM tid Mcth
  -> MemSplit TM Mth
  -> MemMod.disjUnion Mth Mq M
  -> stkSpace (R SP) (IH_Stk+6) Mcth
  -> itr_nextS pc (M, R, fl, isr) (M', R', fl', isr')
  -> exists Mcth', exists Mth', exists TM',
    ThrdMemMod.maps_to TM' tid Mcth'
    /\ MemSplit TM' Mth'
    /\ MemMod.disjUnion Mth' Mq M'
    /\ itr_nextS pc (Mcth, R, fl, isr) (Mcth', R', fl', isr')
    /\ (forall tidx Mx, tidx <> tid 
        -> ThrdMemMod.maps_to TM tidx Mx 
        -> ThrdMemMod.maps_to TM' tidx Mx).
Proof.
  intros pc tid Mcth TM Mth Mq M M'.
  intros R fl isr R' fl' isr'.
  intros HtoMcth HsplitTM HJM Hstk HnextS.
  destruct (MemSplit_Inv HsplitTM HtoMcth) as [Moth [HJMth HsplitMoth]].
  assert (Hx : MemMod.disjUnion Mcth (MemMod.merge Moth Mq) M).
    destruct (@disjUnion_assoc_l Mcth Moth Mq Mth (MemMod.merge Moth Mq) M)
      as [HJMcth2Mo HJMth']; trivial.
  assert (Hstk' : stkSpace (R SP) 6 Mcth).
    eapply stkSpace_weaken;eauto.
    omega.
  assert (exists Mcth' : MemMod.Map,
    MemMod.disjUnion Mcth' (MemMod.merge Moth Mq) M' /\
    itr_nextS pc (Mcth, R, fl, isr) (Mcth', R', fl', isr')).
    eapply (itr_nextS_DMono Hx Hstk' HnextS); eauto.
  destruct H as [Mcth' [HJM' HnextS1]].
  exists Mcth'.
  exists (MemMod.merge Mcth' Moth).
  exists (ThrdMemMod.put TM tid Mcth').
  assert (HeqMcth : MemMod.eqDom Mcth Mcth').
    eapply itr_nextS_eqDom; eauto.
  split_l.
  split.
    assert (H : MemSplit (ThrdMemMod.put (ThrdMemMod.remove TM tid) tid Mcth') (MemMod.merge Mcth' Moth)).
      apply (MS_split (MemMod.merge Mcth' Moth) Moth Mcth' (ThrdMemMod.remove TM tid) tid); 
        auto with PMap.
      apply MemMod.disjUnion_commut.
      eapply disjUnion_upd_left_disjUnion; eauto.
    assert (Hy : ThrdMemMod.put (ThrdMemMod.remove TM tid) tid Mcth'
      = ThrdMemMod.put TM tid Mcth').
      auto with PMap.
    rewrite Hy in H; trivial.
  split.
    assert (H3 : MemMod.eqDom Mcth Mcth').
      eauto with PMap.
    set (Mth' := (MemMod.merge Mcth' Moth)).
    assert (HJMth' : MemMod.disjUnion Mcth' Moth Mth').
      split_r.
      destruct HJMth.
      eauto with PMap.
    assert (H4 : MemMod.eqDom Mth Mth').
      eauto with PMap.
    split.
      destruct HJM.
      eauto with PMap.
    destruct HJM'.
    unfold Mth'.
    subst M'.
    rewrite MemMod.merge_assoc; trivial.
  split_l.
  intros.
  unfold ThrdMemMod.maps_to, ThrdMemMod.put in H0 |- * .
  destruct (ThrdMemMod.beq_A tidx tid).
    destruct (H e).
  trivial.
Qed.
Implicit Arguments itr_nextS_DMono2 [pc tid Mcth TM Mth Mq M M' R fl isr R' fl' isr'].

Lemma Ret_Addr_Mono : forall f M1 M2 M R fl isr,
  MemMod.disjUnion M1 M2 M
  -> Ret_Addr f (M1, R, fl, isr)
  -> Ret_Addr f (M, R, fl, isr).
Proof.
  unfold Ret_Addr;  intros.
  substH H0 with (H0 pc).
  inversion_clear H0.
  apply nextPC_retw with sp; trivial.
  eapply leadsToW_Mono; eauto.
Qed.
Implicit Arguments Ret_Addr_Mono [f M1 M2 M R fl isr].
