Add LoadPath "..".
Add LoadPath "../CAP0".

Require Import Arith.
Require Import Coqlib.
Require Import Map.
Require Import Machine.
Require Import Assertion.
Require Import Assertion_Lemmas.
Require Import Assertion_Lemmas4.
Require Import Operational_Semantics.
Require Import OpSem_Lemmas.

Require Import CAP0.
Require Import SCAP.
Require Import AIM.
Require Import QLib.
Require Import SCAP_READY.
Require Import SCAP_IDLE.
Require Import SCAP_Itr.

Require Import EqNat.

Require Import Linking.
Require Import Linking_Lemmas.
Require Import Linking_Lemmas2.

Lemma weaken_lift_IH_PG:
forall Psi p g tag p' g' S S' status,
  lift_IH_PG (Ast p g tag) Psi status S
  -> ((p *<* INV1) S -> (p' *<* INV1) S')
  -> (forall S'', IH_ItpG g' S' S'' -> IH_ItpG g S S'')
  -> lift_IH_PG (Ast p' g' tag) Psi status S'.
Proof.
  intros.
  unfold lift_IH_PG.
  unfold lift_IH_PG in H.
  InvertAll.
  split; auto.
Qed.

Lemma weaken_itp_IH_theta:
forall Psi p g tag p' g' S S' status,
  itp_IH_theta (Ast p g tag) status Psi S
  -> S_Flag S' _IF = false
  -> ((p *<* INV1) S -> (p' *<* INV1) S')
  -> (forall S'', IH_ItpG g' S' S'' -> IH_ItpG g S S'')
  -> itp_IH_theta (Ast p' g' tag) status Psi S'.
Proof.
  intros.
  unfold itp_IH_theta.
  split; auto.
  unfold itp_IH_theta in H.
  InvertAll. inversion H4.
  left.
  apply weaken_lift_IH_PG with p g S; auto.
  right.
  InvertAll.
  split; auto.
  split; auto.
  exists x.
  apply IH_WFST_strengthen with S g; auto.
Qed.

Lemma decompose_World:
  forall rho Psi T B RdyQ tid p g tag M1 R fl isr,
    rho = rho_ready \/ rho = rho_idle
    -> WF_AIM_World Psi (M1, R, fl, isr) T B RdyQ tid rho (Ast p g tag)
    -> exists TM, exists cthM, exists M12, exists cthM1, exists cthM2,
        MemMod.disjUnion M12 cthM M1
        /\ MemMod.disjUnion cthM1 cthM2 cthM
        /\ WF_currTID tid T B RdyQ rho
        /\ WF_TWQ T B RdyQ
        /\ exists_Idle T RdyQ tid
        /\ WF_WQ_WQSpec B Delta
        /\ MemSplit (ThrdMemMod.remove TM tid) M12
        /\ ThrdMemMod.maps_to TM tid cthM
        /\ WF_Curr_Thrd Psi (S_Mem_Upd (M1, R, fl, isr) cthM) rho (Ast p g tag) T tid
        /\ WF_ReadyQ Psi T RdyQ TM
        /\ WF_WQSet Psi T B TM
        /\ p (cthM1, R, fl, isr)
        /\ NH_Inv fl cthM2
        /\ (tag = 1 -> NH_WFST rho Psi (NH_ItpG g (cthM, R, fl, isr))).
Proof.
  intros rho Psi T B RdyQ tid p g tag M1 R fl isr Hrho; intros.
  unfold WF_AIM_World in H.
  InvertAll.
  rename x into TM. rename x0 into cthM. exists TM. exists cthM.
  assert (exists M12, MemMod.disjUnion M12 cthM M1 /\ MemSplit (ThrdMemMod.remove TM tid) M12).
  apply memsplit_thrd_remove; auto.
  InvertAll.
  rename x into M12. exists M12.
  assert (NH_pinvstar p (cthM, R, fl, isr)).
  apply WF_Curr_Thrd_pinv with Psi rho g tag T tid; auto.
  destruct Hrho; subst rho; auto.
  unfold NH_pinvstar in H9.
  unfold pmstar in H9. InvertAll.
  rename x into cthM1. rename x0 into cthM2.
  exists cthM1. exists cthM2.
  split; auto. split; auto. split; auto. split; auto.
  split; auto. split; auto. split; auto. split; auto.
  split; auto. split; auto. split; auto. split; auto.
  split; auto.
  unfold WF_Curr_Thrd in H5.
  destruct Hrho; subst rho; simpl in H5.
  unfold itp_NH_theta in H5.
  InvertAll.
  UnfoldStateAll. auto.
  unfold itp_NH_theta in H5.
  InvertAll.
  UnfoldStateAll. auto.
Qed.


(***********************************************)
(* lemmas r.t. filterTQ by guoyu               *)

Lemma filterTQ_put : forall T Q tid tcb,
  (fun l : nat =>
    if eq_nat_dec tid l 
      then Some tcb 
      else filterTQ T Q l) = 
  (filterTQ (TQMod.put T tid tcb) (addThrd Q tid)).
Proof.
  intros T Q tid tcb.
  apply (extensionality nat (option Thread) (fun l : nat =>
    if eq_nat_dec tid l 
      then Some tcb 
      else filterTQ T Q l)
    (filterTQ (TQMod.put T tid tcb) (addThrd Q tid))).
  intros tidx.
  unfold filterTQ, TQMod.put, addThrd.
  unfold WQMod.put.
  unfold WQMod.beq_A, TQMod.beq_A.
  unfold WQSpec.beq_Domain, TQSpec.beq_Domain.
  destruct (eq_nat_dec tid tidx).
    subst tidx.
    destruct (eq_nat_dec tid tid); trivial.
    destruct (n (refl_equal _)).
  destruct (eq_nat_dec tidx tid); trivial.
  subst tidx.
  destruct (n (refl_equal _)).
Qed.

Lemma filterTQ_unfold : forall T Q tid tcb,
  TQMod.maps_to (filterTQ T Q) tid tcb
  -> TQMod.maps_to T tid tcb.
Proof.
  unfold TQMod.maps_to.
  unfold filterTQ.
  intros T Q tid tcb H.
  destruct (Q tid); trivial.
  discriminate.
Qed.

Lemma filterTQ_unfold2 : forall T Q tid tcb,
  TQMod.maps_to (filterTQ T Q) tid tcb
  -> Q tid = Some dummy_sglton.
Proof.
  unfold TQMod.maps_to, WQMod.in_dom.
  unfold filterTQ.
  intros T Q tid tcb H.
  generalize (refl_equal (Q tid)).
  pattern Q at 1.
  destruct (Q tid); intro Hx.
    induction b.
    trivial.
  discriminate.
Qed.

Lemma filterTQ_not_empty : forall T RdyQ tid tcb,
  exists_Idle T RdyQ tid
  -> TCB_Status tcb = READY
  -> TQMod.maps_to T tid tcb
  -> (forall l : TQMod.A, filterTQ T RdyQ l = None)
  -> False.
Proof.
  intros T RdyQ tid tcb Hex Htcb Hmto Hemp.
  unfold exists_Idle in Hex.
  destruct Hex.
    destruct H as [tcbx [H0 H1]].
    unfold TQMod.maps_to in * .
    rewrite H0 in Hmto.
    injection Hmto.
    intros H; subst tcbx.
    rewrite Htcb in H1.
    discriminate H1.
  unfold filterTQ in Hemp.
  destruct H as [tid' [tcb' [Hin [Hmto' Htcb']]]].
  substH Hemp with (Hemp tid').
  unfold WQMod.in_dom in Hin.
  destruct Hin.
  rewrite H in Hemp.
  unfold TQMod.maps_to in Hmto'.
  rewrite Hmto' in Hemp.
  discriminate.
Qed.

Lemma filterTQ_prev : forall T Q tid tcb T',
  subsetQ_TQ Q T
  -> WQMod.not_in_dom tid Q
  -> T' = TQMod.put T tid tcb
  -> filterTQ T' Q = filterTQ T Q.
Proof.
  intros T RdyQ tid tcb T' H H0 H1.
  unfold subsetQ_TQ in H.
  unfold WQMod.not_in_dom in H0.
  unfold filterTQ.
  apply (extensionality TQMod.A (option TQMod.B)
    (fun tid0 : TID =>
    match RdyQ tid0 with
    | Some _ => T' tid0
    | None => None (A:=TQMod.B)
    end)
    (fun tid0 : TID =>
    match RdyQ tid0 with
    | Some _ => T tid0
    | None => None (A:=TQMod.B)
    end)).
  intros tidx.
  destruct (TQMod.beq_A tid tidx).
    subst tidx.
    rewrite H0.
    trivial.
  destruct (RdyQ tidx).
    subst T'.
    rewrite TQMod.put_noninterfere; trivial.
  trivial.
Qed.

Lemma filterTQ_prev2 : forall T Q Q' tid tcb T',
  subsetQ_TQ Q T
  -> Q' = WQMod.remove Q tid
  -> T' = TQMod.put T tid tcb
  -> filterTQ T' Q' = filterTQ T Q'.
Proof.
  intros T Q Q' tid tcb T' H H0 H1.
  apply filterTQ_prev with tid tcb; trivial.
    unfold subsetQ_TQ in * .
    intros.
    subst Q'.
    apply H.
    unfold WQMod.in_dom, WQMod.remove in H2 |- * .
    destruct H2.
    exists x.
    destruct (WQMod.beq_A tid0 tid).
    discriminate.
    trivial.
  subst Q'.
  apply WQMod.nid_remove.
Qed.

(**********************************************)
(* local spec should satify spec in Linking.v *)

Lemma ptoTCB_intro1 : forall M ptr tcb tst A,
  M ||- (ptr |->w (TCB_SP tcb) 
    *^* ptr + 2 |->w tst) *^* A
  -> nat2Status tst = Some (TCB_Status tcb)
  -> M ||- ptoTCB ptr tcb *^* A.
Proof.
  intros M ptr tcb tst A H H0.
  destruct (sconj_elim H) as [M1 [M2 [HJM [HM1 HM2]]]].
  exists M1; exists M2; split_l; split_r.
  unfold ptoTCB.
  destruct tcb as [[sp' tst'] pc'].
  exists tst; auto.
Qed.

Lemma ptoTCB_elim1 : forall M ptr tcb A,
  M ||- ptoTCB ptr tcb *^* A
  -> exists tst, 
    (M ||- (ptr |->w (TCB_SP tcb) 
      *^* ptr + 2 |->w tst) *^* A)
    /\ nat2Status tst = Some (TCB_Status tcb).
Proof.
  intros M ptr tcb A H.
  destruct (sconj_elim H) as [M1 [M2 [HJM [HM1 HM2]]]].
  unfold ptoTCB in HM1.
  destruct tcb as [[sp' tst'] pc'].
  sep off.
  destruct HM1 as [w [HM1 Hw]].
  exists w.
  unfold TCB_Status, TCB_SP.
  split_r.
  exists M1; exists M2; auto.
Qed.
Implicit Arguments ptoTCB_elim1 [M ptr tcb A].

Lemma embed_curr_TCB_intro1 : forall M T tid tcb A,
  TQMod.maps_to T tid tcb
  -> M ||- (curr_thrd |->w tid 
      *^* tid |->w_ 
      *^* tid + 2 |->w_ 
      *^* ptoTCB (tid + 4) tcb) *^* A
  -> M ||- embed_curr_TCB T tid *^* A.
Proof.
  intros M T tid tcb A H H0.
  destruct (sconj_elim H0) as [M1 [M2 [HJM [HM1 HM2]]]].
  exists M1; exists M2; split_l; split_r.
  unfold embed_curr_TCB.
  exists tcb; split_l.
  sep on.
  sep matchH HM1 with 4.
Qed.

Lemma embed_curr_TCB_elim1 : forall M T tid A,
  M ||- embed_curr_TCB T tid *^* A
  -> exists tcb,
    TQMod.maps_to T tid tcb
    /\ M ||- (curr_thrd |->w tid 
      *^* tid |->w_ 
      *^* tid + 2 |->w_ 
      *^* ptoTCB (tid + 4) tcb) *^* A.
Proof.
  intros M T tid A H.
  destruct (sconj_elim H) as [M1 [M2 [HJM [HM1 HM2]]]].
  unfold embed_curr_TCB in HM1.
  sep off.
  destruct HM1 as [tcb [Htcb HM1]].
  exists tcb; split_l.
  exists M1; exists M2; split_l; split_r.
  sep on.
  sep matchH HM1 with 6.
Qed.

Lemma ptoTQ_elim :
   forall Q ptr M,
     M ||- (ptoTQ ptr Q)
     -> exists head, exists tail, exists prev,
       M ||- (ptr |->w head *^* ptr + 2 |->w tail *^* 
         Queue Thread ptoTCB head tail prev 0 Q).
Proof.
  intros.
  sep off.
  unfold ptoTQ , ptoQ in H.
  destruct H as [head [tail [prev H]]].
  exists head; exists tail; exists prev.
  assumption.
Qed.
Implicit Arguments ptoTQ_elim [Q ptr M].

Lemma ptoTQ_elim1 :   
   forall Q ptr M A,
     M ||- (ptoTQ ptr Q *^* A)
     -> exists head, exists tail, exists prev,
       M ||- (ptr |->w head *^* ptr + 2 |->w tail *^* 
         Queue Thread ptoTCB head tail prev 0 Q *^* A).
Proof.
  intros.
  destruct H as [M1 [M2 [Hdis [HQ HA]]]].
  destruct (ptoTQ_elim HQ) as [head [tail [prev H]]].
  exists head; exists tail; exists prev.
  exists M1;exists M2.
  intuition.
Qed.
Implicit Arguments ptoTQ_elim1 [Q ptr M A].

Lemma ptoTQ_intro :
   forall head tail prev Q ptr M,
     M ||- (ptr |->w head *^* ptr + 2 |->w tail *^* 
       Queue Thread ptoTCB head tail prev 0 Q)
     -> M ||- (ptoTQ ptr Q ).
Proof.
  unfold ptoTQ.
  unfold ptoQ.
  sep off.
  intros.
  exists head; exists tail; exists prev.
  assumption.
Qed.
Implicit Arguments ptoTQ_intro [head tail prev Q ptr M ].

Lemma ptoTQ_intro1 :   
   forall head tail prev Q ptr M A,
     M ||- (ptr |->w head *^* ptr + 2 |->w tail *^* 
       Queue Thread ptoTCB head tail prev 0 Q *^* A)
     -> M ||- (ptoTQ ptr Q *^* A).
Proof.
  intros.
  destruct H as [M1 [M2 [Hdis [HQ HA]]]].
  substH HQ with (ptoTQ_intro HQ).
  exists M1; exists M2;intuition.
Qed.
Implicit Arguments ptoTQ_intro1 [head tail prev Q ptr M A].

