(*  -----------by wangwei *)
Add LoadPath "..".

Require Import Coqlib.
Require Import Map.
Require Import Arith.
Require Import Machine.
Require Import Assertion.
Require Import Assertion_Lemmas.
Require Import Assertion_Lemmas3.
Require Import Assertion_Lemmas4.
Require Import Operational_Semantics.
Require Import Mem_Lemmas.
Require Import QLib.

Lemma Queue_tail0_head0 :
  forall Node inMemNode head prev next Q M,
    Queue Node inMemNode head 0 prev next Q M
    -> head = 0.
Proof.
  intros Node inMemNode head prev next Q M HQ.
  inversion HQ. 
  reflexivity.
  inversion H0.
  inversion H3.
Qed.
Implicit Arguments Queue_tail0_head0 [Node inMemNode head prev next Q M].

Lemma Queue_tail0_head0_1 :
  forall Node inMemNode head prev next Q M A,
    M ||- (Queue Node inMemNode head 0 prev next Q) *^* A
    -> head = 0.
Proof.
  intros  Node inMemNode head prev next Q M A HQM.
  destruct HQM as [M1 [M2 [HdisjU [HQ HA]]]].
  eapply Queue_tail0_head0; eauto.
Qed.
Implicit Arguments Queue_tail0_head0_1 [Node inMemNode head prev next Q M A].

Lemma Queue_head0_tail0_pre :
  forall  Node inMemNode head tail prev next Q M,
    Queue Node inMemNode head tail prev next Q M
    -> head = 0
    -> tail = 0.
Proof.
  intros.
  induction H.
  reflexivity.
  auto.
  rewrite (IHQueue H0) in *.
  inversion H3.
Qed.

Lemma  Queue_head0_tail0:
  forall Node inMemNode tail prev next Q M,
    Queue Node inMemNode 0 tail prev next Q M
    -> tail = 0.
Proof.
  intros Node inMemNode tail prev next Q M HQ.
  eapply Queue_head0_tail0_pre; eauto.
Qed.
Implicit Arguments Queue_head0_tail0 [tail prev next Q M].

Lemma  Queue_head0_tail0_1:
  forall Node inMemNode tail prev next Q M A,
    M ||- (Queue Node inMemNode 0 tail prev next Q) *^* A
    -> tail = 0.
Proof.
  intros Node inMemNode tail prev next Q M A HM.
  destruct HM as [M1 [M2 [HdisjU [HQ HA]]]].
  eapply Queue_head0_tail0;eauto.
Qed.
Implicit Arguments Queue_head0_tail0_1 [Node inMemNode tail prev next Q M A].

Lemma Queue_ptr0_emp:
  forall Node inMemNode prev next Q M,
    Queue Node inMemNode 0 0 prev next Q M
      -> M = MemMod.emp /\ Q = (fun l:nat => None(A:=Node)).
Proof.
  intros Node inMemNode prev next Q M HQ.
  inversion HQ; subst.
  intuition.
  inversion H0.
  inversion H3.
Qed.
Implicit Arguments Queue_ptr0_emp [Node inMemNode prev next Q M].

Lemma Queue_ptr0_emp_1:
  forall Node inMemNode prev next Q M A,
    M ||- (Queue Node inMemNode 0 0 prev next Q) *^* A
      -> M ||- A /\ Q = (fun l:nat => None(A:=Node)).
Proof.
  intros Node inMemNode prev next Q M A HM.
  destruct HM as [M1 [M2 [Hdis [HQ HA]]]].
  destruct (Queue_ptr0_emp HQ).
  subst M1.
  substH Hdis with (MemMod.disjUnion_commut Hdis).
  substH Hdis with (MemMod.emp_disjUnion_unique Hdis).
  subst M2.
  intuition.
Qed.
Implicit Arguments Queue_ptr0_emp_1 [Node inMemNode prev next Q M A].

Lemma difMem_label_unequal:
    forall Node inMemNode l1 l2 prev1 prev2 next1 next2 tcb1 tcb2 M1 M2,
      ptoNode Node inMemNode l1 prev1 next1 tcb1 M1
      -> ptoNode Node inMemNode  l2 prev2 next2 tcb2 M2
      -> MemMod.disjoint M1 M2
      -> l1 <>l2.
Proof.
  intros.
  unfold ptoNode in * .
  unfold ptoW in *.
  sep on.
  sep lift 0 in H0; sep lift 0 in H.
  sep off.
  destruct H as [M11 [M12 [P11 [P12 P13]]]].
  destruct H0 as [M21 [M22 [P21 [P22 P23]]]].
  destruct P22; destruct P12.
  intro e.
  unfold MemMod.maps_to in * .
  assert (MemMod.disjoint M11 M21).
  disjoint lift_lr.
  unfold MemMod.disjoint in * .
  unfold MemMod.not_in_dom in * .
  subst l1.
  substH H4 with (H4 l2).
  destruct H4.
  rewrite H4 in H2; inversion H2.
  rewrite H4 in H; inversion H.  
Qed.
Implicit Arguments difMem_label_unequal [Node inMemNode l1 l2 prev1 prev2 next1 next2 tcb1 tcb2 M1 M2].

Lemma Queue_unemp_get : 
  forall Node inMemNode head tail prev next Q M,
    M ||- Queue Node inMemNode head tail prev next Q
    -> head > 0
    -> (exists Q', exists tcb, exists next', exists next'', exists tail',
      M ||- (ptoNode Node inMemNode  head prev next' tcb *^*
        Queue Node inMemNode next'' tail' head next Q' )
      /\ Q = (fun l' => if eq_nat_dec head l' then Some tcb else Q' l')/\
      Q' head = None (A := Node)/\
    (tail' = tail /\ tail' > 0 /\ next'' = next' \/ tail' = 0/\ tail = head /\ next'' = 0 /\ next' = next)).
Proof. 
  intros Node inMemNode head tail prev next Q M.
  intros HQ Hhead.
  assert (Htail : tail > 0).
    destruct tail.
    rewrite (Queue_tail0_head0 HQ) in Hhead.
    inversion Hhead.
    omega.
  clear Hhead.
  induction HQ.
  (* empty queue *)
  inversion Htail.
  (* singlton queue *)
  exists (fun l :nat => None (A := Node)); exists node; exists next; exists 0; exists 0.
  intuition.
  exists M;exists MemMod.emp.
  intuition.
  apply Q_emp.
  (* nodes' number  >= 2 *)
  substH IHHQ with (IHHQ H1).
  clear H1 H2 HQ.
  destruct IHHQ as [Q' [tcb [next1 [next2 [tail' [HM1 [HQQ' [HQ'head Hdis]]]]]]]].
  destruct Hdis as [[Ht't [Ht' Hnext21]] | [Ht'0 [Hth [Hnext2 Hnext1]]]].
    (* nodes' number > 2*)
    subst next2.
    exists (fun l' => if eq_nat_dec l l' then Some node else Q' l');
      exists tcb; exists next1;exists next1; exists l.
    destruct HM1 as [MM1 [MM2 [PP1 [PP2 PP3]]]].
    assert (MemMod.disjoint MM1 M2).
    disjoint lift_l.
    assert (UnE := difMem_label_unequal PP2 H0 H1).
    intuition.
    exists MM1; exists (MemMod.merge MM2 M2).
    intuition.
    destruct PP1.
    apply MemMod.disjUnion_assoc_left_right.
    assumption.
    rewrite <- H3;assumption.
    subst tail'.
    apply (Q_cons_more (Node := Node) (inMemNode := inMemNode) next1 tail head Q' MM2 M2 (MemMod.merge MM2 M2) l next' node);
      try assumption.
    split.
    disjoint lift_l.
    reflexivity.
    apply extensionality.
    intro x.
    subst Q.
    destruct (eq_nat_dec l x); destruct (eq_nat_dec head x);
      try reflexivity.
    subst l head;elim UnE;reflexivity.
    destruct (eq_nat_dec l head).
    elim UnE;auto.
    assumption.
    (* nodes' number = 2 *)
    subst tail'.
    destruct HM1 as [MM1 [MM2 [PP1 [PP2 PP3]]]].
    subst next2.
    destruct (Queue_ptr0_emp PP3).
    subst MM2 Q'.
    substH PP1 with (MemMod.emp_disjUnion_unique PP1).
    subst MM1.
    assert (HH := H).
    destruct HH.
    assert (UnE := difMem_label_unequal PP2 H0 H1).
    exists (fun l' :nat => if eq_nat_dec l l' then Some node else None); exists tcb; exists l; exists l; exists l.
    intuition.
    exists M1; exists M2.
    subst next1 Q tail.
    intuition.
    apply (Q_cons_fst (Node:=Node) (inMemNode:=inMemNode) l head next' node M2); assumption.
    apply extensionality.
    intro x.
    subst Q.
    destruct (eq_nat_dec l x); destruct (eq_nat_dec head x); 
    try reflexivity.
    rewrite e in UnE; rewrite e0 in UnE;elim UnE; reflexivity.
    destruct (eq_nat_dec l head).
    elim UnE;auto.
    reflexivity.
Qed.
Implicit Arguments  Queue_unemp_get [Node inMemNode head tail prev next Q M].

Lemma Queue_unemp_get_1 : 
  forall Node inMemNode head tail prev next Q M A,
    M ||- (Queue Node inMemNode head tail prev next Q ) *^* A
    -> head > 0
    -> (exists Q', exists tcb, exists next', exists next'', exists tail',
      M ||- (ptoNode Node inMemNode  head prev next' tcb *^*
        Queue Node inMemNode next'' tail' head next Q' ) *^* A
      /\ Q = (fun l' => if eq_nat_dec head l' then Some tcb else Q' l')/\
      Q' head = None (A := Node)/\
    (tail' = tail /\ tail' > 0 /\ next'' = next' \/ tail' = 0/\ tail = head /\ next'' = 0 /\ next' = next)).
Proof.
  intros Node inMemNode head tail prev next Q M A.
  intros HM Hhead.
  destruct HM as [M00 [M11 [HdisjU [HQ HA]]]].
  substH HQ with (Queue_unemp_get HQ Hhead).
  destruct HQ as [Q' [tcb [next' [next'' [tail' [HM00 [HQQ' [HQ'h Htail']]]]]]]].
  exists Q'; exists tcb; exists next'; exists next''; exists tail'.
  split.
  exists M00; exists M11.
  intuition.
  intuition.
Qed.
Implicit Arguments  Queue_unemp_get_1 [Node inMemNode head tail prev next Q M A].

Lemma Queue_emp_intro_1 : 
  forall Node inMemNode prev next M A, 
    M ||- A ->
    M ||- Queue Node inMemNode 0 0 prev next (fun l => None) *^* A.
Proof.
  intros.
  exists MemMod.emp; exists M.
  intuition.
  apply Q_emp.
Qed.
Implicit Arguments  Queue_emp_intro_1 [Node inMemNode prev next M A].

Lemma Queue_cons_fst_intro_1 : 
  forall Node inMemNode l prev next node M A,
  M ||- (ptoNode Node inMemNode l prev next node) *^* A
  -> l > 0
  -> M ||- Queue Node inMemNode l l prev next
     (fun l' => if eq_nat_dec l l' then Some node else None) *^* A.
Proof.
  intros.
  destruct H as [M1 [M2 [Hdis [Hpto HA]]]]. 
  exists M1; exists M2.
  intuition.
  apply (Q_cons_fst (Node:=Node) (inMemNode:=inMemNode) l prev next node M1);
    assumption.
Qed.

Lemma Queue_cons_more_intro_1 :
  forall Node inMemNode head tail prev Q M l next' node A,
    M ||- (Queue Node inMemNode head tail prev l Q) *^*
      (ptoNode Node inMemNode l tail next' node) *^* A
    -> tail > 0
    -> l > 0
    -> M ||- Queue Node inMemNode head l prev next' 
    (fun l' => if eq_nat_dec l l' then Some node else Q l') *^* A.
Proof.
  intros.
  destruct H as [M0 [MA [Hdis1 [HM0 HA]]]].
  destruct HM0 as [M1 [M2 [Hdis2 [HM1 HM2]]]].
  exists M0; exists MA.
  intuition.
  apply (Q_cons_more (Node:=Node) (inMemNode:=inMemNode) head tail prev Q M1 M2 M0 l next' node ); assumption.
Qed.

Lemma Queue_unemp_split : 
  forall Node inMemNode head tail prev next Q M,
    M ||- Queue Node inMemNode head tail prev next Q
    -> tail > 0
    -> (exists Q', exists tcb, exists prev', exists head', exists tail',
      M ||- (ptoNode Node inMemNode tail prev' next tcb *^*
        Queue Node inMemNode head' tail' prev tail Q' )
      /\ Q = (fun l' => if eq_nat_dec tail l' then Some tcb else Q' l')/\
    (head' = head /\ head' > 0 /\ prev' = tail' \/ head' = 0/\ tail = head /\  prev' = prev)).
Proof.
  intros.
  inversion H; subst.
  inversion H0.
  exists (fun l : nat => None (A:= Node)); exists node; exists prev;
    exists 0; exists 0.
  intuition.
  sep lift 1.
  apply Queue_emp_intro_1;assumption.
  exists Q0; exists node; exists tail0; exists head; exists tail0.
  intuition.
  sep lift 1.
  exists M1; exists M2.
  intuition.
  left.
  intuition.
  destruct head.
  sep on.
  assert (tail0 = 0).
  eapply Queue_head0_tail0; eauto.
  rewrite H6 in H4;inversion H4.
  omega.
Qed.
Implicit Arguments Queue_unemp_split [Node inMemNode head tail prev next Q M].

Lemma Queue_unemp_split_1 : 
  forall Node inMemNode head tail prev next Q M A,
    M ||- Queue Node inMemNode head tail prev next Q *^* A
    -> tail > 0
    -> (exists Q', exists tcb, exists prev', exists head', exists tail',
      M ||- (ptoNode Node inMemNode tail prev' next tcb *^*
        Queue Node inMemNode head' tail' prev tail Q' ) *^* A
      /\ Q = (fun l' => if eq_nat_dec tail l' then Some tcb else Q' l')/\
    (head' = head /\ head' > 0 /\ prev' = tail' \/ head' = 0/\ tail = head /\  prev' = prev)).
Proof.
  intros.
  destruct H as [M1 [M2 [Hdis [HM HA]]]].
  substH HM with (Queue_unemp_split HM H0).
  destruct HM as [Q' [tcb [prev' [head' [tail' [P1 [P2 P3]]]]]]].
  exists Q'; exists tcb; exists prev'; exists head'; exists tail'.
  split.
  exists M1; exists M2.
  intuition.
  intuition.
Qed.
Implicit Arguments Queue_unemp_split_1 [Node inMemNode head tail prev next Q M A].

Lemma Queue_upd_next_1 :
  forall Node inMemNode head tail prev next next' jk Q M A,
    M ||- Queue Node inMemNode head tail prev next Q *^* A
    -> tail > 0
    -> MemMod.put (MemMod.put M (tail + 2) next') (tail + 2 + 1) jk
    ||- Queue Node inMemNode head tail prev next' Q *^* A.
Proof.
  intros.
  substH H with (Queue_unemp_split_1 H H0).
  destruct H as [Q' [tcb [prev' [head' [tail' [HM [HQQ' Hsplit]]]]]]].
  destruct Hsplit as [[Hhead' [Hht Hprev']]| [Hhead' [Hht Hprev']]].
  subst head' prev' Q.
  assert (Htail' : tail' > 0).
  destruct tail'.
  sep lift 1 in HM.
  substH HM with (Queue_tail0_head0_1 HM).
  rewrite HM in Hht; inversion Hht.
  omega.
  apply Queue_cons_more_intro_1 with tail'.
  unfold ptoNode in * .
  assert (MemUpdW M (tail + 2) next' jk 
    ||- Queue Node inMemNode head tail' prev tail Q' *^*
      (tail |->w tail' *^* tail + 2 |->w next' *^* inMemNode (tail + 4) tcb) *^*
      A). 
  sep lift 2.
  sep upd.
  sep matchH HM with 10.
  auto.
  assumption.
  assumption.
  subst head' tail prev'.
  sep lift 1 in HM.
  assert (Htail' := Queue_head0_tail0_1 HM).
  subst tail' Q.
  destruct ( Queue_ptr0_emp_1 HM).
  subst Q'.
  apply Queue_cons_fst_intro_1.
  unfold ptoNode in * .
  assert (MemUpdW M (head + 2) next' jk 
    ||- head |->w prev *^* head + 2 |->w next' *^* inMemNode (head + 4) tcb *^*
       A).
  sep lift 1.
  sep upd.
  sep matchH H with 10.
  auto.
  assumption.
Qed.

Implicit Arguments Queue_upd_next_1 [Node inMemNode head tail prev next Q M A].
