Add LoadPath "..".

Require Import Coqlib.
Require Import Arith.
Require Import Machine.
Require Import Assertion.
Require Import Operational_Semantics.

Definition ptoNode (Node : Set) (inMem : Label -> Node -> MPred) (l prev next : Label) (node : Node) : MPred :=
  mstar (mstar (ptoW l prev) (ptoW (l+2) next)) (inMem (l+4) node).

(* Queue Node inMemNode head tail head_prev tail_post Q M *)
(* head : head of queue;  tail : tail of queue *)
(* head_prev : prev field in the head node     *)
(* tail_next : next field in the tail node     *)
(* Q : corresponding abstract set of Node      *)
(* Inductive Queue (Node : Set) (inMemNode : Label -> Node -> MPred) :  *)
(*                        Label -> Label -> Label -> Label -> (Label -> option Node) -> Mem -> Prop := *)
(*     Q_emp       : forall prev next, Queue Node inMemNode 0 0 prev next (fun l => None) MemMod.emp *)
(*   | Q_cons_fst  : forall l prev next node Q M prev' next', *)
(*                        mstar (Queue Node inMemNode 0 0 prev' next' Q) (ptoNode Node inMemNode l prev next node) M *)
(*                        -> l > 0 *)
(*                        -> Queue Node inMemNode l l prev next (fun l' => if eq_nat_dec l l' then Some node else Q l') M *)
(*   | Q_cons_more : forall head tail prev Q M l next' node, *)
(*                        mstar (Queue Node inMemNode head tail prev l Q) (ptoNode Node inMemNode l tail next' node) M *)
(*                        -> tail > 0 *)
(*                        -> l > 0 *)
(*                        -> Queue Node inMemNode head l prev next' (fun l' => if eq_nat_dec l l' then Some node else Q l') M. *)

Inductive Queue (Node : Set) (inMemNode : Label -> Node -> MPred) : 
  Label -> Label -> Label -> Label -> (Label -> option Node) -> Mem -> Prop :=

  Q_emp       : forall prev next, 
            Queue Node inMemNode 0 0 prev next (fun l => None) MemMod.emp

| Q_cons_fst  :  forall l prev next node M ,
  (ptoNode Node inMemNode l prev next node) M
  -> l > 0
  -> Queue Node inMemNode l l prev next
     (fun l' => if eq_nat_dec l l' then Some node else None) M

| Q_cons_more : forall head tail prev Q M1 M2 M l next' node,
  MemMod.disjUnion M1 M2 M
  -> (Queue Node inMemNode head tail prev l Q) M1
  -> (ptoNode Node inMemNode l tail next' node) M2
  -> tail > 0
  -> l > 0
  -> Queue Node inMemNode head l prev next' 
  (fun l' => if eq_nat_dec l l' then Some node else Q l') M.

Implicit Arguments Q_emp [Node inMemNode]. 
Implicit Arguments Q_cons_fst [Node inMemNode].
Implicit Arguments Q_cons_more [Node inMemNode].

Definition ptoQ (Node : Set) (inMemNode : Label -> Node -> MPred) (l : Label) (Q : Label -> option Node) : MPred :=
  fun (M : Mem) =>
  exists head, exists tail, exists prev, 
    mstar (mstar (ptoW l head) (ptoW (l+2) tail)) (Queue Node inMemNode head tail prev 0 Q) M.

(*
Definition enQueue_pre 
  (Node : Set) (inMemNode : Label -> Node -> MPred) (sp l ln : Label) (node : Node) (Q : Label -> option Node) := 
  fun (M : Mem) => 
    exists prev, exists next,
      ln > 0 /\
      mstar (mstar (mstar (mstar (ptoW (sp + 2) ln) 
                                 (ptoW (sp + 4) l)) 
                                 (ptoxW (sp - 2))) 
                                 (ptoQ Node inMemNode l Q))
                                 (ptoNode Node inMemNode ln prev next node)
      M.


Definition p_enQueue : Pred :=
  fun (S : State) =>
    exists Node, exists inMemNode, exists sp, exists l, exists Q, exists ln, exists node,
      ((enQueue_pre Node inMemNode sp l ln node Q) *^* mTrue) (S_Mem S) /\ (S_RF S SP = sp) /\ (sp >= 2).

Definition enQueue_post
  (Node : Set) (inMemNode : Label -> Node -> MPred) (sp l ln: Label) (node : Node) (Q : Label -> option Node) := 
  fun (M : Mem) => 
      mstar (mstar (mstar (ptoW (sp + 2) ln) 
                          (ptoW (sp + 4) l)) 
                          (ptoxW (sp - 2))) 
                          (ptoQ Node inMemNode l (fun l => if eq_nat_dec ln l then Some node else Q l))
      M.


Definition g_enQueue : Grt :=
  fun (S S' : State) =>
    forall Node inMemNode sp l ln node Q,
      forall M M0,
        MemMod.disjUnion M M0 (S_Mem S) 
	-> S_RF S SP = sp
        -> enQueue_pre Node inMemNode sp l ln node Q M
        -> exists M',
             MemMod.disjUnion M' M0 (S_Mem S')
             /\ enQueue_post Node inMemNode sp l ln node Q M'
    /\ S_RF S SP = S_RF S' SP /\ S_RF S BP = S_RF S' BP /\ S_Flag S _IF = S_Flag S' _IF.


Definition deQueue_pre 
  (Node : Set) (inMemNode : Label -> Node -> MPred) (sp l : Label) (Q : Label -> option Node) :=
  fun (M : Mem) =>
      mstar (mstar (ptoW (sp+2) l) (ptoxW (sp - 2))) (ptoQ Node inMemNode l Q) M.

Definition p_deQueue : Pred :=
  fun (S : State) =>
    exists Node, exists inMemNode, exists sp, exists l, exists Q,
      ((deQueue_pre Node inMemNode sp l Q) *^* mTrue) (S_Mem S) /\ (S_RF S SP = sp) /\ (sp >=2).

Definition deQueue_post_a
  (Node : Set) (inMemNode : Label -> Node -> MPred) (sp l ln : Label) (node : Node) (Q : Label -> option Node) := 
  fun (M : Mem) => 
    exists prev, exists next,
      mstar (mstar (mstar (ptoW (sp + 2) l) 
                          (ptoxW (sp - 2)))
                          (ptoQ Node inMemNode l Q))
                          (ptoNode Node inMemNode ln prev next node)
      M.


Definition g_deQueue : Grt :=
  fun (S S' : State) =>
    forall Node inMemNode sp l Q,
      forall M M0,
        MemMod.disjUnion M M0 (S_Mem S)
        -> S_RF S SP = sp 
        -> deQueue_pre Node inMemNode sp l Q M
        -> exists M',
             MemMod.disjUnion M' M0 (S_Mem S')
             /\ ((forall l, Q l = None) /\ deQueue_pre Node inMemNode sp l Q M' /\ S_RF S' AX = 0
               \/ exists ln, exists node, exists Q', (Q' ln = None)
                 /\ (fun l' => if eq_nat_dec ln l' then Some node else Q' l') = Q
                 /\ deQueue_post_a Node inMemNode sp l ln node Q' M'
                 /\ S_RF S' AX = ln
                )
    /\ S_RF S SP = S_RF S' SP /\ S_RF S BP = S_RF S' BP /\ S_Flag S _IF = S_Flag S' _IF.
*)