Add LoadPath "..".
Add LoadPath "../CAP0".

Require Import Arith.
Require Import Coqlib.
Require Import Map.
Require Import Machine.
Require Import Assertion.
Require Import Assertion_Lemmas.
Require Import Assertion_Lemmas4.
Require Import Operational_Semantics.
Require Import OpSem_Lemmas.
Require Import OpSem_Lemmas2.
Require Import Reg.
Require Import CAP0.
Require Import SCAP.
Require Import AIM.
Require Import QLib.
Require Import SCAP_READY.
Require Import SCAP_IDLE.
Require Import SCAP_Itr.
Require Import Linking.
Require Import Linking_Lemmas.
Require Import Linking_Lemmas2.
Require Import Linking_Lemmas3.

Require Import EqNat.
Require Import Omega.
Require Import List.
Require Import Arith_Lemmas.

Lemma prim_BLK:
  prim_entry _BLOCK.
Proof.
  unfold prim_entry.
  right.
  left. auto.
Qed.
Hint Resolve prim_BLK.

Lemma enable_blk_monotonic:
  forall w wqa, monotonic_p (enable_blk w wqa).
Proof.
  intros.
  unfold enable_blk.
  set (p0 := fun S0 => (pTrue *<*
          (S_RF S0 SP |->w w *^*
           ((INV0 *^* INV1) *^* stkSpacePrec (S_RF S0 SP) (_BLK_Stk + 2)))) S0).
  set (p1 := fun (S0 : State) => Delta w = SomeT wqa).
  assert (monotonic_p p0).

  unfold monotonic_p.
  intros.
  induction S as (((M0, R0), fl0), isr0). unfold p0. 
  unfold p0 in H. simplS. simplS in H.
  unfold S_Mem in H0.
  assert (exists M0', MemMod.disjUnion M0 M0' M).
  apply MemMod.subseteq_split; auto.
  unfold pmstar in H. simplS in H. InvertAll.
  unfold pmstar. simplS. exists (MemMod.merge x0 x). exists x1.
  unfold pTrue.
  split; auto.
  cut (MemMod.disjUnion x1 (MemMod.merge x0 x) M). auto with PMap.
  apply MemMod.disjUnion_assoc_left_right; auto.
  inversion_clear H1; auto with PMap.
  assert (MemMod.disjUnion x1 x0 M0). auto with PMap.
  inversion H3. rewrite H6 in H2. auto.

  assert (monotonic_p p1).
  unfold monotonic_p.
  intros.
  unfold p1. unfold p1 in H0. auto.
  assert (monotonic_p (fun S => p1 S /\ p0 S)).
  unfold monotonic_p.
  intros.
  InvertAll.
  split.
  unfold monotonic_p in H0. apply H0; auto. unfold monotonic_p in H. apply H; auto.
  unfold p0 in H1. unfold p1 in H1. auto.
Qed.
Hint Resolve enable_blk_monotonic.


Lemma enable_blk_prog:
  forall w wqa S pc f, enable_blk w wqa S
    -> exists S' : State, nextS pc (comml (_call f)) S S'.
Proof.
  intros.
  unfold enable_blk in H.
  InvertAll.
  apply stkSpace_call_prog with (_BLK_Stk + 2); auto.
  assert (((S_RF S SP |->w w *^*
           (INV0 *^* INV1 *^* stkSpacePrec (S_RF S SP) (_BLK_Stk + 2))) *^* mTrue) (S_Mem S)).
  apply pTrue_m_mTrue with pTrue; auto.
  assert (((INV0 *^* INV1 *^* stkSpacePrec (S_RF S SP) (_BLK_Stk + 2)) *^* mTrue) (S_Mem S)).
  apply mstar_mTrue with (S_RF S SP |->w w) mTrue; auto.
  apply mstar_mTrue with (INV0 *^* INV1) mTrue; auto.
  omega.
Qed.


Lemma NH_Call_blk_enable:
forall psi pc p g n S,
  NH_Call_Prim psi pc _BLOCK (Ast p g n)
  -> imprec p S 
  -> exists w, exists wqa, enable_blk w wqa S.
Proof.
  intros.
  unfold NH_Call_Prim in H.
  InvertAll.
  inversion H1.
  absurd (_BLOCK <> _SWITCH); auto.
  inversion H3.
  rewrite <- H7 in H5. rewrite <- H8 in H5.
  rewrite <- H9 in H4.
  clear H7 H8 H9 H3 H tag p0 g0.
  induction S as (((M, R), fl), isr). 
  unfold imprec in H0.
  unfold pmstar in H0.
  InvertAll. rename x into M1. rename x0 into M2.
  UnfoldStateAll.
  assert (exists w : nat,
    exists wqa : WQassert,
      enable_blk w wqa (M1, R, fl, isr) /\
      (forall pc' : nat, pimp (wrap_fun (gblk wqa) pc _BLOCK pc' (M1, R, fl, isr)) p') /\
      (forall pc' : nat,
        pimp (wrap_fun (gblk wqa) pc _BLOCK pc' (M1, R, fl, isr) |> g') (g (M1, R, fl, isr)))).
  apply H5; auto.
  InvertAll. rename x into w. rename x0 into wqa.
  exists w. exists wqa.
  assert (monotonic_p (enable_blk w wqa)). auto.
  unfold monotonic_p in H8.
  assert (enable_blk w wqa (S_Mem_Upd (M1, R, fl, isr) M)). apply H8; auto.
  unfold S_Mem. apply MemMod.disjUnion_subseteq_left with M2; auto.
  UnfoldStateAll. auto.

  absurd (_BLOCK <> _UNBLK); auto.
Qed.

Lemma NH_BLK_prog_aux :
  forall psi pc theta Psi rho S,
    NH_Call_Prim psi pc _BLOCK theta -> Itp_AIM rho theta Psi S -> rho <> rho_ih 
    -> exists S', nextS pc (comml (_call _BLOCK)) S S'.
Proof.
  intros.
  induction theta.
  assert (imprec (NH_pinvstar p) S).
  apply Itp_AIM_pinv with g n Psi rho; auto.
  assert (imprec p S).
  apply imprec_presv_imprec with (NH_pinvstar p); auto.
  assert (exists w, exists wqa, enable_blk w wqa S).
  apply NH_Call_blk_enable with psi pc p g n; auto.
  InvertAll.
  apply enable_blk_prog with x x0; auto.
Qed.



Lemma NH_BLK_progress:
  forall psi pc theta Psi a S,
    NH_WFcmd psi pc theta (comml (_call _BLOCK))
    -> a = Itp_AIM rho_ready theta
    -> a Psi S
    -> exists S', nextS pc (comml (_call _BLOCK)) S S'.
Proof.
  intros.
  unfold NH_WFcmd in H.
  induction theta.
  unfold NH_WFcmd_aux in H.
  inversion_clear H.
  clear H3.
  inversion H2.
  inversion_clear H. clear H3.
  apply NH_BLK_prog_aux with psi (Ast p g n) Psi rho_ready; auto.
  rewrite H0 in H1. auto.
  inversion_clear H. absurd (prim_entry _BLOCK); auto.
Qed.

Lemma WF_AIM_World_blk:
forall Psi'0 S T B RdyQ tid p g tag w wqa,
  WF_AIM_World Psi'0 S T B RdyQ tid rho_ready (Ast p g tag)
  -> enable_blk w wqa S
  -> exists BlkQ, 
      B w = Some BlkQ
      /\ Delta w = SomeT wqa 
                    /\ (pmstar pTrue (mstar (ptoW (S_RF S SP) w)
                                           (mstar (mstar INV0 INV1) 
                                                  (stkSpacePrec (S_RF S SP) (_BLK_Stk + 2))
                                           )
                                   )
                     ) S.
Proof.
  intros.
  unfold enable_blk in H0.
  InvertAll.
  unfold WF_AIM_World in H.
  InvertAll.
  unfold WF_WQ_WQSpec in H4.
  assert (WQSetMod.in_dom w B).
  apply H4 with wqa; auto.
  unfold WQSetMod.in_dom in H9.
  InvertAll.
  exists x1. auto.
Qed.


Lemma WF_AIM_World_Rdy_StatusRdy:
forall Psi'0 S T B RdyQ tid theta tcb,
   WF_AIM_World Psi'0 S T B RdyQ tid rho_ready theta
   -> TQMod.maps_to T tid tcb
   -> TCB_Status tcb = READY.
Proof.
  intros.
  unfold WF_AIM_World in H.
  InvertAll.
  unfold WF_currTID in H1.
  InvertAll.
  unfold TQMod.maps_to in H12. unfold TQMod.maps_to in H0.
  rewrite H12 in H0. inversion H0.
  rewrite H15 in H14.
  unfold rho_tcb_consistent in H14.
  inversion_clear H14.
  apply H13; auto.
Qed.

Lemma WF_AIM_World_cthNotBlk:
  forall Psi'0 S T B RdyQ tid theta w BlkQ rho,
     WF_AIM_World Psi'0 S T B RdyQ tid rho theta
  -> B w = Some BlkQ
  -> WQMod.not_in_dom tid BlkQ.
Proof.
  intros.
  unfold WF_AIM_World in H.
  InvertAll.
  unfold WF_currTID in H1.
  InvertAll.
  unfold unblocked in H8.
  apply H8 with w; auto.
Qed.

Lemma enable_blk_nxt_stkspace:
forall pc S S' M1 M2 f w,
  ((pTrue *<*
        (S_RF (S_Mem_Upd S M1) SP |->w w *^*
         (INV0 *^* INV1 *^*
          stkSpacePrec (S_RF (S_Mem_Upd S M1) SP) (_BLK_Stk + 2))))
         (S_Mem_Upd S M1))
  -> nextS pc (comml (_call f)) S S'
  -> MemMod.disjUnion M1 M2 (S_Mem S)
  -> exists M1',
        MemMod.disjUnion M1' M2 (S_Mem S')
        /\ (S_RF S' SP |->w (pc+1) *^* (S_RF S' SP + 2 |->w w) *^* stkSpacePrec (S_RF S' SP) 10 *^* mTrue) M1'.
Proof.
  intros.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  UnfoldStateAll.
  assert ((mTrue *^*
       (R SP |->w w *^*
        (INV0 *^* INV1 *^* stkSpacePrec (R SP) (_BLK_Stk + 2)))) M1).
  unfold pmstar in H.
  InvertAll.
  exists x. exists x0. split_l.
  split. unfold mTrue. auto.
  auto. clear H.
  assert (frame (nextS pc (comml (_call f)))).
  apply frame_call.
  unfold frame in H.
  assert (exists M1', 
        MemMod.disjUnion M1' M2 (S_Mem (M', R', fl', isr')) /\
        nextS pc (comml (_call f)) (S_Mem_Upd (M, R, fl, isr) M1) (S_Mem_Upd (M', R', fl', isr') M1')).
  apply H; auto.
  clear H.
  UnfoldStateAll.
  unfold enableS.
  exists (MemUpdW M1 (R SP - 2) (pc+1) 0, RF_update R SP (R SP - 2), fl, isr).
  apply nextS_call.
  unfold pushW.
  split.
  sep on.
  sep lift 4 in H2.
  sep stksplit 2 in H2. sep lift 0 in H2.
  sep trivial.
  split; auto.
  apply  MemUpdW_MemUpd16 with 0; auto.
  clear H.
  UnfoldStateAll.
  InvertAll.
  exists x.
  split_l.
  destruct (nextS_inversion H4).
  InvertAll.
  subst R'. reg simpl.
  substH H10 with (MemUpd16_MemUpdW H10).
  InvertAll.
  subst x.
  sep on.
  sep lift 0.
  sep upd.
  assert (R SP - 2 + 2 = R SP). auto.
  rewrite H6.
  sep lift 4 in H2.
  sep stksplit in H2.
  assert (_BLK_Stk + 2 - 2 = _BLK_Stk). auto.
  rewrite H9 in H2.
  sep matchH H2 with 4.
Qed.

Lemma enable_blk_nxt_stkspace1:
forall pc S S' M1 M2 f w,
  ((pTrue *<*
        (S_RF (S_Mem_Upd S M1) SP |->w w *^*
         (INV0 *^* INV1 *^*
          stkSpacePrec (S_RF (S_Mem_Upd S M1) SP) (_BLK_Stk + 2))))
         (S_Mem_Upd S M1))
  -> nextS pc (comml (_call f)) S S'
  -> MemMod.disjUnion M1 M2 (S_Mem S)
  -> exists M1', exists junk1, exists junk2,
        MemMod.disjUnion M1' M2 (S_Mem S')
        /\ (S_RF S' SP |-> (pc+1) *^* S_RF S' SP + 1 |-> junk1 
            *^* (S_RF S' SP + 2 |-> w) *^* S_RF S' SP + 3 |-> junk2 *^* stkSpacePrec (S_RF S' SP) 10 *^* mTrue) M1'.
Proof.
  intros.
  assert (exists M1',
            MemMod.disjUnion M1' M2 (S_Mem S')
            /\ (S_RF S' SP |->w (pc+1) *^* (S_RF S' SP + 2 |->w w) *^* stkSpacePrec (S_RF S' SP) 10 *^* mTrue) M1').
  apply enable_blk_nxt_stkspace with S M1 f; auto.
  destruct H2 as [M1' [Hsplit Hmem]].
  exists M1'.
  sep on. sep lift 0 in Hmem.
  assert (exists junk1,
            M1' ||- S_RF S' SP |-> pc + 1 *^* S_RF S' SP + 1 |-> junk1 *^*
             (S_RF S' SP + 2 |->w w *^*
              (stkSpacePrec (S_RF S' SP) 10 *^* mTrue))).
  apply ptoW_elim1. auto.
  clear Hmem. destruct H2 as [junk1 Hmem].
  exists junk1.
  sep lift 2 in Hmem.
  assert (exists junk2,
            M1' ||- S_RF S' SP + 2 |-> w *^* S_RF S' SP + 2 + 1 |-> junk2 *^*
             (stkSpacePrec (S_RF S' SP) 10 *^* mTrue *^*
              (S_RF S' SP |-> pc + 1 *^* S_RF S' SP + 1 |-> junk1))).
  apply ptoW_elim1; auto.
  assert (S_RF S' SP + 2 + 1 = S_RF S' SP + 3). omega.
  rewrite H3 in H2.
  inversion H2.
  exists x. split_l.
  sep on.
  sep matchH H4 with 6.
Qed.




Lemma merge_BLK_pre:
forall pc S' M1' w T tid RdyQ B BlkQ M2 x0 x1,
  (S_RF S' SP |-> pc *^* S_RF S' SP + 1 |-> x0 *^*
      S_RF S' SP + 2 |-> w *^* S_RF S' SP + 3 |-> x1 *^*
        stkSpacePrec (S_RF S' SP) 10 *^* mTrue) M1'
  -> (embed_curr_TCB T tid *^* (embed_RdyQ T RdyQ *^* embed_WQSet T B)) M2
  -> B w = Some BlkQ
  -> MemMod.disjUnion M1' M2 (S_Mem S')
  -> (BLK_MPre T RdyQ BlkQ tid (S_RF S' SP) w pc x0 x1 *^* mTrue) (S_Mem S').
Proof.
  intros.
  induction S' as (((M, R), fl), isr).
  UnfoldStateAll.
  rename M1' into M1.
  unfold BLK_MPre.
  assert (((R SP |-> pc  *^* R SP + 1 |-> x0 *^* R SP + 2 |-> w *^*
            R SP + 3 |-> x1 *^* stkSpacePrec (R SP) 10 *^* mTrue) *^*
              (embed_curr_TCB T tid *^* (embed_RdyQ T RdyQ *^* embed_WQSet T B))) M).
  exists M1. exists M2.
  split; auto.
  assert (forall M, embed_WQSet T B M -> (ptoTQ w (filterTQ T BlkQ) *^* mTrue) M).
  intros.
  apply enbedBlkQ_inWQSet with B; auto.
  sep on.
  sep lift 8 in H3.
  sep weaker_subst in H3 with (ptoTQ w (filterTQ T BlkQ) *^* mTrue); auto.
  clear H5.
  sep on.
  sep lift 8 in H3.
  sep lift 0.
  sep match_left H3; auto.
  sep lift 0 in H3.
  sep match_left H3; auto.
  sep lift 0 in H3.
  sep match_left H3; auto.
  sep lift 1 in H3.
  sep match_left H3; auto.
  sep matchH H3 with 5.
Qed.


Lemma enter_BLK_preMem:
forall pc S S' p g tag M1 M2 Psi T B RdyQ tid wqa w BlkQ,
  nextS pc (comml (_call _BLOCK)) S S'
  -> MemMod.disjUnion M1 M2 (S_Mem S)
  -> WF_AIM_World Psi (S_Mem_Upd S M1) T B RdyQ tid
         rho_ready (Ast p g tag)
  -> (embed_curr_TCB T tid *^* (embed_RdyQ T RdyQ *^* embed_WQSet T B)) M2
  -> Delta w = SomeT wqa
  -> (pTrue *<*
         (S_RF (S_Mem_Upd S M1) SP |->w w *^*
          (INV0 *^* INV1 *^*
           stkSpacePrec (S_RF (S_Mem_Upd S M1) SP) (_BLK_Stk + 2))))
          (S_Mem_Upd S M1)
  -> B w = Some BlkQ
  -> exists junk1, exists junk2, (BLK_MPre T RdyQ BlkQ tid (S_RF S' SP) w (pc+1) junk1 junk2 *^* mTrue) (S_Mem S').
Proof.
  intros.
  assert (exists M1', exists junk1, exists junk2,
            MemMod.disjUnion M1' M2 (S_Mem S')
            /\ (S_RF S' SP |-> (pc+1) *^* S_RF S' SP + 1 |-> junk1 *^* S_RF S' SP + 2 |-> w *^* S_RF S' SP + 3 |-> junk2 
              *^* stkSpacePrec (S_RF S' SP) 10 *^* mTrue) M1').
  apply enable_blk_nxt_stkspace1 with S M1 _BLOCK; auto.
  InvertAll.
  rename x into M1'. exists x0. exists x1.
  apply merge_BLK_pre with M1' B M2; auto.
Qed.


Lemma enter_BLK_pre:
forall psi pc theta Psi'0 S S',
  NH_WFcmd psi pc theta (comml (_call _BLOCK))
  -> Itp_AIM rho_ready theta Psi'0 S
  -> nextS pc (comml (_call _BLOCK)) S S'
  -> BLK_Prd S'.
Proof.
  intros.
  unfold BLK_Prd.
  induction theta.
  unfold Itp_AIM in H0.
  InvertAll. rename x into T. rename x0 into B. rename x1 into RdyQ. rename x2 into tid.
  exists T. exists RdyQ.
  unfold NH_WFcmd in H.
  unfold NH_WFcmd_aux in H.
  InvertAll. clear H. clear H4.
  inversion_clear H2.
  inversion_clear H. clear H2.
  unfold NH_Call_Prim in H3.
  inversion_clear H3.
  inversion H.
  absurd (_BLOCK <> _SWITCH); auto.
  inversion H4. rewrite <- H8 in H6. rewrite <- H9 in H6.
  clear H9 H8 H10 H3 H4 H5 tag p0 g0.
  rename n into tag.
  induction S as (((M, R), fl), isr).
  unfold pmstar in H0.
  InvertAll.
  rename x into M1. rename x0 into M2.

  assert (exists w, exists wqa, enable_blk w wqa (M1, R, fl, isr)).
  apply NH_Call_blk_enable with psi pc p g tag; auto.
  unfold NH_Call_Prim. split; auto.
  assert (imprec (NH_pinvstar p) (M1, R, fl, isr)).
  apply WF_AIM_World_pinv with Psi'0 T B RdyQ tid rho_ready g tag; auto.
  apply imprec_presv_imprec with (NH_pinvstar p); auto.
  InvertAll. rename x into w. rename x0 into wqa.

  assert (exists BlkQ, B w = Some BlkQ
                                 /\ Delta w = SomeT wqa 
                                 /\ (pmstar pTrue (mstar (ptoW (S_RF (M1, R, fl, isr) SP) w)
                                                         (mstar (mstar INV0 INV1) 
                                                             (stkSpacePrec (S_RF (M1, R, fl, isr) SP) (_BLK_Stk + 2))
                                                          )
                                                   )
                                    ) (M1, R, fl, isr)
         ).
  apply WF_AIM_World_blk with Psi'0 T RdyQ tid p g tag; auto.
  InvertAll. rename x into BlkQ.
  exists BlkQ.
  exists tid.
  exists (S_RF S' SP).
  exists w.
  assert (exists tcb, TQMod.maps_to T tid tcb).
  apply WF_AIM_World_currTCB with Psi'0 (S_Mem_Upd (M, R, fl, isr) M1) B RdyQ rho_ready (Ast p g tag); auto.
  inversion_clear H9.
  rename x into tcb. exists tcb. exists (pc+1).
  split.
  apply enter_BLK_preMem with (M, R, fl, isr) p g tag M1 M2 Psi'0 B wqa; auto.
  split. apply WF_AIM_World_existsIdle with Psi'0 (S_Mem_Upd (M, R, fl, isr) M1) B rho_ready (Ast p g tag); auto.
  split; auto.
  split. apply WF_AIM_World_RdyQinT with Psi'0 (S_Mem_Upd (M, R, fl, isr) M1) B tid rho_ready (Ast p g tag); auto.
  split.
  apply WF_AIM_World_BlkQinT with Psi'0 (S_Mem_Upd (M, R, fl, isr) M1) B RdyQ tid rho_ready (Ast p g tag) w; auto.
  split.
  apply WF_AIM_World_Rdy_StatusRdy with Psi'0 (M1, R, fl, isr) T B RdyQ tid (Ast p g tag); auto.
  split; auto.
  split.
  apply WF_AIM_World_cthNotRdy with Psi'0 (S_Mem_Upd (M, R, fl, isr) M1) T B rho_ready (Ast p g tag); auto.
  split.
  apply WF_AIM_World_cthNotBlk with Psi'0 (S_Mem_Upd (M, R, fl, isr) M1) T B RdyQ (Ast p g tag) w rho_ready; auto.
  split.
  apply WF_AIM_World_RdyNotBlk with Psi'0 (M1, R, fl, isr) T B tid (Ast p g tag) w rho_ready; auto.
  apply WF_AIM_World_WFTQ with Psi'0 (S_Mem_Upd (M, R, fl, isr) M1) B RdyQ tid rho_ready (Ast p g tag); auto.
  assert (_BLOCK <> _UNBLK). auto.
  contradiction.
  inversion_clear H.
  absurd (prim_entry _BLOCK); auto.
Qed.


Lemma enter_BLK_disable:
forall psi pc theta Psi'0 S S',
  NH_WFcmd psi pc theta (comml (_call _BLOCK))
  -> Itp_AIM rho_ready theta Psi'0 S
  -> nextS pc (comml (_call _BLOCK)) S S'
  -> S_Flag S' _IF = false.
Proof.
  intros.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  unfold NH_WFcmd in H.
  induction theta.
  destruct H as [HWFcmd [Htmp1 Htmp2]].
  clear Htmp1 Htmp2.
  unfold NH_WFcmd_aux in HWFcmd.
  inversion_clear HWFcmd.
  destruct H as [Hprim HWFcmd].
  clear Hprim.
  unfold NH_Call_Prim in HWFcmd.
  inversion_clear HWFcmd.
  clear H.
  assert (fl _IF = false).
  assert (imprec (NH_pinvstar p) (M, R, fl, isr)).
  apply Itp_AIM_pinv with g n Psi'0 rho_ready. auto.
  discriminate.
  unfold imprec in H. unfold pmstar in H.
  InvertAll.
  unfold NH_pinvstar in H.
  simplS in H.
  unfold pmstar in H.
  InvertAll. simplS in H.
  assert (S_Flag (x1, R, fl, isr) _IF = false).
  apply H2; auto.
  UnfoldStateAll; auto. clear H0 H2.
  destruct (nextS_inversion H1).
  InvertAll.
  subst fl'.
  UnfoldStateAll. auto.

  inversion_clear H.
  assert (prim_entry _BLOCK); auto.
  contradiction.
Qed.
  
Lemma WF_AIM_World_blk2:
forall Psi'0 S T B RdyQ tid p g tag w wqa S',
  WF_AIM_World Psi'0 S T B RdyQ tid rho_ready (Ast p g tag)
  -> enable_blk w wqa S'
  -> exists BlkQ, 
      B w = Some BlkQ
      /\ Delta w = SomeT wqa 
                    /\ (pmstar pTrue (mstar (ptoW (S_RF S' SP) w)
                                           (mstar (mstar INV0 INV1) 
                                                  (stkSpacePrec (S_RF S' SP) (_BLK_Stk + 2))
                                           )
                                   )
                     ) S'.
Proof.
  intros.
  unfold enable_blk in H0.
  InvertAll.
  unfold WF_AIM_World in H.
  InvertAll.
  unfold WF_WQ_WQSpec in H4.
  assert (WQSetMod.in_dom w B).
  apply H4 with wqa; auto.
  unfold WQSetMod.in_dom in H9.
  InvertAll.
  exists x1. auto.
Qed.

Lemma enter_BLK_grt_hyp0:
forall sp w cthM1 R fl isr,
  (pTrue *<*
         (sp |->w w *^*
          (INV0 *^* INV1 *^* stkSpacePrec sp (_BLK_Stk + 2))))
          (cthM1, R, fl, isr)
  -> exists cthM11, exists cthM12,
            MemMod.disjUnion cthM11 cthM12 cthM1
            /\ (sp |->w w *^* stkSpacePrec sp (_BLK_Stk + 2)) cthM11
            /\ (INV0 *^* INV1 *^* mTrue) cthM12.
Proof.
  intros.
  assert (((sp |->w w *^* stkSpacePrec sp (_BLK_Stk + 2)) *^* (INV0 *^* INV1 *^* mTrue)) cthM1).
  assert ((mTrue *^* (sp |->w w *^* (INV0 *^* INV1 *^* stkSpacePrec sp (_BLK_Stk + 2)))) cthM1).
  unfold pmstar in H.
  InvertAll.
  exists x. exists x0.
  split_l. split; auto.
  sep on.

  sep lift 1 in H0.
  sep matchH H0 with 5.
  unfold mstar at 1 in H0.
  InvertAll.
  exists x. exists x0. split_l.
  split; auto.
Qed.

Lemma mem_compose_c_blk:
forall cthM11' M21 M1' sp pc T tid RdyQ w BlkQ junk1 junk2,
  MemMod.disjUnion cthM11' M21 M1'
  -> (sp |-> pc *^* sp + 1 |-> junk1 *^* (sp + 2 |-> w) *^* sp + 3 |-> junk2 *^*
         stkSpacePrec sp _BLK_Stk) cthM11'
  -> (embed_curr_TCB T tid *^* embed_RdyQ T RdyQ *^* ptoTQ w (filterTQ T BlkQ)) M21
  -> BLK_MPre T RdyQ BlkQ tid sp w pc junk1 junk2 M1'.
Proof.
  intros.
  assert (((sp |-> pc *^* sp + 1 |-> junk1 *^* sp + 2 |-> w *^* sp + 3 |-> junk2 *^* stkSpacePrec sp _BLK_Stk) *^* 
           (embed_curr_TCB T tid *^* embed_RdyQ T RdyQ *^*
        ptoTQ w (filterTQ T BlkQ))) M1').
  exists cthM11'. exists M21. auto.
  unfold BLK_MPre.
  sep on.
  sep matchH H2 with 8.
Qed.

Lemma enter_BLK_grt2_hyp2:
forall T T' B RdyQ' BlkQ' tid tid' sp sp' w tcb' pc pc' S0 S' M0' M2' M22 M12 cthM2 cthM12 cthMR TM junk1 junk2,
  BLK_Post_aux T' RdyQ' BlkQ' tid' sp' sp w tcb' pc junk1 junk2 (S_Mem_Upd S0 M0')
  -> nextS pc' (comms _retw) S0 S'
  -> MemMod.disjUnion M0' M2' (S_Mem S0)
  -> disjUnionList (M22 :: M12 :: cthM2 :: cthM12 :: nil) M2'
  -> embed_WQSet T (WQSetMod.remove B w) M22
  -> MemSplit (ThrdMemMod.remove TM tid) M12
  -> ThrdMemMod.maps_to TM tid' cthMR
  -> tid <> tid'
  -> NH_Inv (S_Flag S0) cthM2
  -> (INV0 *^* INV1 *^* mTrue) cthM12
  -> T' = TQMod.put T tid (sp, BLOCKED, pc)
  -> exists INVM, exists cthM0, exists M21', exists M122, exists cthM_old, exists cthM_oldStk,
        MemMod.disjUnion INVM cthM0 cthM12 
        /\ (INV0 *^* INV1) INVM
        /\ (embed_curr_TCB T' tid' *^* embed_RdyQ T' RdyQ' *^* ptoTQ w (filterTQ T' BlkQ')) M21'
        /\ MemMod.disjUnion M122 cthMR M12
        /\ MemSplit (ThrdMemMod.remove (ThrdMemMod.remove TM tid) tid') M122
        /\ MemMod.disjUnion (MemMod.merge cthM0 cthM2) cthM_oldStk cthM_old
        /\ (sp |-> pc *^* sp+1 |-> junk1 *^* sp + 2 |-> w *^* sp+3 |-> junk2 *^* stkSpacePrec sp _BLK_Stk) cthM_oldStk
        /\ disjUnionList (M22 :: cthMR :: INVM :: M21' :: cthM_old :: M122 :: nil) (S_Mem S').
Proof.
  intros.
  unfold mstar at 1 in H8. InvertAll.
  rename x into INVM. rename x0 into cthM0. exists INVM. exists cthM0.
  unfold BLK_Post_aux in H.
  InvertAll.
  unfold BLK_MPost in H16.
  assert ((embed_curr_TCB T' tid' *^* embed_RdyQ T' RdyQ' *^* ptoTQ w (filterTQ T' BlkQ') *^* 
           (sp |-> pc *^* sp+1 |-> junk1 *^* sp + 2 |-> w *^* sp+3 |-> junk2 *^* stkSpacePrec sp 10)) (S_Mem (S_Mem_Upd S0 M0'))).
  induction S0 as (((Mx, Rx), flx), isrx).
  simplS in H16. simplS.
  sep on.
  sep matchH H16 with 8.

  unfold mstar at 1 in H15.
  InvertAll.
  rename x into M21'. exists M21'.
  rename x0 into cthM_oldStk.
  assert (exists M122, MemMod.disjUnion M122 cthMR M12 /\
                        MemSplit (ThrdMemMod.remove (ThrdMemMod.remove TM tid) tid') M122).
  apply memsplit_thrd_remove; auto.
  unfold ThrdMemMod.maps_to. unfold ThrdMemMod.remove.
  induction (ThrdMemMod.beq_A tid' tid).
  absurd (tid' <> tid); auto. unfold ThrdMemMod.maps_to in H5. auto.
  InvertAll. rename x into M122.
  exists M122.
  exists (MemMod.merge (MemMod.merge cthM0 cthM2) cthM_oldStk).
  exists cthM_oldStk.
  split; auto. split; auto. split; auto.
  split; auto. split; auto.
  split. unfold MemMod.disjUnion. split; auto.
  apply enter_SW_grt2_aux2_hyp0 with INVM cthM12 M22 M12 M0' M2' S0 M21'; auto.
  split. auto.
  apply enter_SW_grt2_aux2_hyp1 with pc' M0' M2' S0 M12 cthM12; auto.
Qed.


Lemma ext_B_presv_wffTID:
forall T' B B' RdyQ' BlkQ w tid tid' rho',
  WF_currTID tid' T' B RdyQ' rho'
  -> tid <> tid'
  -> B' = WQSetMod.put B w (addThrd BlkQ tid)
  -> B w = Some BlkQ
  -> WF_currTID tid' T' B' RdyQ' rho'.
Proof.
  intros.
  unfold WF_currTID in *.
  destruct H as [Hunblk [HnotRdy [HinT Htcb]]].
  split.
  unfold unblocked in *.
  subst B'.
  unfold WQSetMod.maps_to.
  unfold WQSetMod.put.
  intros.
  induction (WQSetMod.beq_A w0 w).
  inversion H. clear H H3 Q.
  unfold addThrd.
  unfold WQMod.not_in_dom.
  unfold WQMod.put.
  induction (WQMod.beq_A tid' tid).
  absurd (tid' <> tid); auto.
  assert (WQMod.not_in_dom tid' BlkQ).
  apply Hunblk with w.
  unfold WQSetMod.maps_to. auto.
  unfold WQMod.not_in_dom in H. auto.
  apply Hunblk with w0; auto.
  split_l. split_l.
  destruct Htcb as [tcb [HtcbinT [Hstatus Hconsist]]].
  exists tcb. split_l.
  split; auto.
Qed.

Lemma enter_BLK_grt2_hyp4_hyp0:
forall RdyQ RdyQ' tid tid' B,
  RdyQ = addThrd RdyQ' tid'
  -> tid <> tid'
  -> disj_RQ_WQ B RdyQ
  -> unblocked tid' B.
Proof.
  intros.
  unfold unblocked.
  unfold disj_RQ_WQ in H1.
  intros.
  assert (WQMod.disjoint Q RdyQ).
  apply H1 with w; auto.
  unfold WQMod.disjoint in H3.
  assert (WQMod.not_in_dom tid' Q \/ WQMod.not_in_dom tid' RdyQ).
  apply H3; auto.
  clear H3. inversion_clear H4. auto.
  unfold WQMod.not_in_dom in H3.
  subst RdyQ.
  unfold addThrd in H3.
  unfold WQMod.put in H3.
  induction (WQMod.beq_A tid' tid'). inversion H3.
  absurd (tid' = tid'); auto.
Qed.



Lemma enter_BLK_grt2_hyp4:
forall T T' B RdyQ RdyQ' tid tid' tcb tcb' rho rho',
  tid <> tid'
  -> RdyQ = addThrd RdyQ' tid'
  -> WF_currTID tid T B RdyQ rho
  -> T' = TQMod.put T tid tcb
  -> WQMod.not_in_dom tid' RdyQ'
  -> WF_TWQ T B RdyQ
  -> TQMod.maps_to T tid' tcb'
  -> TCB_Status tcb' = READY \/
        TCB_Status tcb' = IDLE
  -> rho_tcb_consistent tcb' rho'
  -> WF_currTID tid' T' B RdyQ' rho'.
Proof.
  intros.
  unfold WF_currTID.
  unfold WF_TWQ in H4.
  unfold WF_currTID in H1.
  InvertAll.
  split.
  apply enter_BLK_grt2_hyp4_hyp0 with RdyQ RdyQ' tid; auto.
  split; auto.
  rewrite H2. unfold TQMod.in_dom. unfold TQMod.put.
  induction (TQMod.beq_A tid' tid).
  absurd (tid' <> tid); auto.
  split.
  exists tcb'. auto. exists tcb'.
  unfold TQMod.maps_to.
  induction (TQMod.beq_A tid' tid).
  contradiction. split; auto.
Qed.

Lemma enter_BLK_grt2_hyp5:
forall T T' RdyQ RdyQ' tid tid' sp pc tcb,
  exists_Idle T RdyQ tid
  -> RdyQ = addThrd RdyQ' tid'
  -> T tid = Some tcb
  -> tid <> tid'
  -> WQMod.not_in_dom tid' RdyQ'
  -> WQMod.not_in_dom tid RdyQ
  -> T' = TQMod.put T tid (sp, BLOCKED, pc)
  -> TCB_Status tcb = READY
  -> exists_Idle T' RdyQ' tid'.
Proof.
  intros.
  unfold exists_Idle.
  unfold exists_Idle in H.
  inversion_clear H.
  InvertAll.
  assert (x = tcb). unfold TQMod.maps_to in H7.
  rewrite H7 in H1. inversion H1. auto.
  rewrite H in H8.
  rewrite H8 in H6. inversion H6.
  InvertAll. rename x into tid''. rename x0 into tcb''.
  assert (tid' = tid'' \/ tid' <> tid'').
  apply nat_eq_neq_dec. inversion_clear H8.
  left. exists tcb''. split. rewrite <- H10 in H7.
  rewrite H5. unfold TQMod.put. unfold TQMod.maps_to.
  induction (TQMod.beq_A tid' tid).
  absurd (tid' <> tid); auto. auto. auto.
  right. exists tid''. exists tcb''.
  assert (WQMod.in_dom tid'' RdyQ').
  unfold WQMod.in_dom in H.
  rewrite H0 in H. unfold addThrd in H.
  unfold WQMod.put in H.
  induction (WQMod.beq_A tid'' tid').
  absurd (tid'' <> tid'); auto.
  inversion H. exists x. auto.
  split; auto.
  split; auto.
  rewrite H5. unfold TQMod.maps_to. unfold TQMod.put.
  induction (TQMod.beq_A tid'' tid).
  rewrite a in H8.
  assert (WQMod.in_dom tid RdyQ).
  rewrite H0. unfold addThrd.
  unfold WQMod.in_dom. unfold WQMod.put.
  induction (WQMod.beq_A tid tid').
  contradiction. auto.
  absurd (~ WQMod.in_dom tid RdyQ); auto with PMap. auto.
Qed.


Lemma extT_presv_wffTQ:
forall T T' tid tcb tcb',
  WF_TQ T
  -> T tid = Some tcb
  -> T' = TQMod.put T tid tcb'
  -> WF_TQ T'.
Proof.
  intros.
  unfold WF_TQ in *.
  intros.
  subst T'.
  unfold TQMod.put in H2. unfold TQMod.in_dom in H2.
  induction (TQMod.beq_A tid0 tid).
  subst tid0.
  apply H; auto. unfold TQMod.in_dom.
  exists tcb; auto.
  apply H; auto.
Qed.


Lemma extB_presv_wffWQSet2:
forall B B' tid w BlkQ,
  WF_WQSet2 B
  -> unblocked tid B
  -> B w = Some BlkQ
  -> B' = WQSetMod.put B w (addThrd BlkQ tid)
  -> WF_WQSet2 B'.
Proof.
  intros.
  unfold WF_WQSet2 in *.
  rewrite H2.
  unfold WQSetMod.put.
  unfold WQSetMod.maps_to.
  intros.
  induction (WQSetMod.beq_A l w).
  induction (WQSetMod.beq_A l' w).
  rewrite <- a0 in a. contradiction.
  inversion H3.
  unfold WQMod.disjoint.
  unfold addThrd.
  unfold WQMod.put.
  unfold WQMod.not_in_dom.
  intro.
  induction (WQMod.beq_A x tid).
  right.
  unfold unblocked in H0.
  assert (WQMod.not_in_dom tid Q').
  apply H0 with l'; auto.
  rewrite a0.
  auto.
  assert (WQMod.disjoint BlkQ Q').
  apply H with w l'; auto.
  unfold WQMod.disjoint in H6.
  unfold WQMod.not_in_dom in H6.
  apply H6.

  induction (WQSetMod.beq_A l' w).
  inversion H4.
  unfold WQMod.disjoint.
  unfold addThrd.
  unfold WQMod.put.
  unfold WQMod.not_in_dom.
  intro.
  induction (WQMod.beq_A x tid).
  left.
  unfold unblocked in H0.
  assert (WQMod.not_in_dom tid Q).
  apply H0 with l; auto.
  rewrite a0.
  auto.
  assert (WQMod.disjoint BlkQ Q).
  apply H with w l; auto.
  unfold WQMod.disjoint in H6.
  unfold WQMod.not_in_dom in H6.
  assert (BlkQ x = None \/ Q x = None).
  apply H6. inversion H8; auto.
  apply H with l l'; auto.
Qed.

Lemma extB_RdyQ_presv_disjRQWQ:
forall B B' RdyQ RdyQ' tid tid' w BlkQ,
  disj_RQ_WQ B RdyQ
  -> WQMod.not_in_dom tid RdyQ
  -> RdyQ = addThrd RdyQ' tid'
  -> B w = Some BlkQ
  -> B' = WQSetMod.put B w (addThrd BlkQ tid)
  -> disj_RQ_WQ B' RdyQ'.
Proof.
  intros.
  unfold disj_RQ_WQ in *.
  rewrite H3.
  unfold WQSetMod.maps_to. unfold WQSetMod.put.
  intros.
  induction (WQSetMod.beq_A l w).
  inversion H4.
  assert (WQMod.disjoint BlkQ RdyQ).
  apply H with w; auto.
  unfold WQMod.disjoint.
  unfold addThrd.
  unfold WQMod.not_in_dom. unfold WQMod.put.
  intro.
  induction (WQMod.beq_A x tid). right.
  rewrite a0.
  rewrite H1 in H0.
  unfold WQMod.not_in_dom in H0.
  unfold addThrd in H0.
  unfold WQMod.put in H0.
  induction (WQMod.beq_A tid tid').
  inversion H0. auto.
  unfold WQMod.disjoint in H5.
  assert (WQMod.not_in_dom x BlkQ \/ WQMod.not_in_dom x RdyQ).
  apply H5.
  inversion_clear H7; auto.
  unfold WQMod.not_in_dom in H8. right. rewrite H1 in H8.
  unfold addThrd in H8. unfold WQMod.put in H8.
  induction (WQMod.beq_A x tid'). inversion H8. auto.

  assert (WQMod.disjoint Q RdyQ).
  apply H with l; auto.
  unfold WQMod.disjoint in *.
  intro.
  assert (WQMod.not_in_dom x Q \/ WQMod.not_in_dom x RdyQ).
  apply H5.
  inversion_clear H6; auto.
  right.
  rewrite H1 in H7.
  unfold WQMod.not_in_dom in H7.
  unfold addThrd in H7. unfold WQMod.put in H7.
  induction (WQMod.beq_A x tid'). inversion H7. auto.
Qed.
  


(* changed because of the change of WF_currTID *)
Lemma enter_BLK_grt2_aux6:
forall RdyQ RdyQ' tid tid' tcb tcb' B B' T T' rho w BlkQ,
  RdyQ = addThrd RdyQ' tid'
  -> T tid = Some tcb
  -> WF_currTID tid T B RdyQ rho
  -> WF_TWQ T B RdyQ
  -> B w = Some BlkQ
  -> B' = WQSetMod.put B w (addThrd BlkQ tid)
  -> T' = TQMod.put T tid tcb'
  -> WQMod.not_in_dom tid RdyQ
  -> WF_TWQ T' B' RdyQ'.
Proof.
  intros.
  unfold WF_TWQ.
  split.
  apply extT_presv_wffTQ with T tid tcb tcb'; auto.
  unfold WF_TWQ in H2. InvertAll. auto.
  unfold WF_TWQ in H2. InvertAll.
  split.
  apply extB_presv_wffWQSet2 with B tid w BlkQ; auto.
  unfold WF_currTID in H1. InvertAll. auto.
  apply extB_RdyQ_presv_disjRQWQ with B RdyQ tid tid' w BlkQ; auto.
Qed.

Lemma extB_presv_WF_WQ_WQSpec:
forall B B' Delta w BlkQ BlkQ',
  WF_WQ_WQSpec B Delta
  -> B' = WQSetMod.put B w BlkQ'
  -> B w = Some BlkQ
  -> WF_WQ_WQSpec B' Delta.
Proof.
  intros.
  unfold WF_WQ_WQSpec in *.
  intros.
  rewrite H0.
  unfold WQSetMod.in_dom.
  unfold WQSetMod.put.
  induction (WQSetMod.beq_A l w).
  exists BlkQ'. auto.
  assert (WQSetMod.in_dom l B).
  apply H with wqa; auto. auto.
Qed.



Lemma enter_BLK_grt2_hypd0:
forall cthM1 cthM11 cthM12 R fl isr pc f,
  MemMod.disjUnion cthM11 cthM12 cthM1
  -> stkSpace (R SP) 2 cthM11
  -> exists cthM1', 
            MemUpd16 cthM1 (R SP - 2) (pc+1) cthM1'
            /\ nextS pc (comml (_call f)) (cthM1, R, fl, isr) (cthM1', RF_update R SP (R SP - 2), fl, isr).
Proof.
  intros.
  assert (stkSpace (R SP) 2 cthM1).
  apply stkSpace_Mono with cthM11 cthM12; auto.
  exists (MemUpdW cthM1 (R SP - 2) (pc + 1) 0).
  split. apply MemUpdW_MemUpd16 with 0; auto.
  apply nextS_call; auto.
  unfold pushW. split_l.
  split; auto.
  apply MemUpdW_MemUpd16 with 0; auto.
Qed.

Lemma enter_BLK_grt2_auxc_hyp3:
forall sp sp' w0 w junk1 junk2 oldStk2 cthM11_2,
  (sp' |-> w0 *^* sp' + 1 |-> junk1 *^* sp' + 2 |-> w *^*
         sp' + 3 |-> junk2) oldStk2
  -> (sp - 2 |->w_ *^* (sp |-> w *^* sp + 1 |-> junk2))
                cthM11_2
  ->  sp = sp' + 2
  -> MemUpdW cthM11_2 (sp - 2) w0 junk1 = oldStk2.
Proof.
  intros.
  rewrite H1 in H0.
  assert (sp' + 2 - 2 = sp'). auto.
  rewrite H1. rewrite H2. rewrite H2 in H0.
  assert (sp' + 2 + 1 = sp' + 3). auto.
  rewrite H3 in H0. clear H1 H2 H3. clear sp.
  assert ((sp' |-> w0 *^* sp' + 1 |-> junk1 *^* sp' + 2 |-> w *^*
              sp' + 3 |-> junk2) (MemUpdW cthM11_2 sp' w0 junk1)).
  sep on.
  unfold MemUpdW.
  sep lift 1.
  sep upd2. sep lift 3.
  sep upd2.
  assert (exists x, cthM11_2 ||- sp' |->w x *^* (sp' + 2 |-> w *^* sp' + 3 |-> junk2)).
  apply ptoxW_ptoW1; auto.
  inversion H1. clear H1.
  assert (exists y, cthM11_2 ||- sp' |-> x *^* sp' + 1 |-> y *^* (sp' + 2 |-> w *^* sp' + 3 |-> junk2)).
  apply ptoW_elim1; auto. inversion_clear H1.
  sep matchH H3 with 4.
  unfold mstar in H3. sep off.
  InvertAll. exists x1. exists x2. unfold ptox.
  split_l. split. exists x. auto. exists x0. auto.
  assert (uniqueM (sp' |-> w0 *^* sp' + 1 |-> junk1 *^* sp' + 2 |-> w *^*
        sp' + 3 |-> junk2)).
  apply mstar_presv_uniqueM. apply mstar_presv_uniqueM.
  apply mstar_presv_uniqueM; apply pto_uniqueM.
  apply pto_uniqueM. apply pto_uniqueM.
  unfold uniqueM in H2.
  apply H2; auto.
Qed.



Lemma enter_BLK_grt2_auxc_hyp0:
forall cthM1 cthM11 cthM12 cthM11_1 cthM11_2 INVM cthM0 cthM' sp w pc junk1 junk2,
  MemMod.disjUnion cthM11 cthM12 cthM1
  -> MemMod.disjUnion INVM cthM0 cthM12
  -> MemMod.disjUnion cthM11_1 cthM11_2 cthM11
  -> (sp - 2 |->w_ *^* (sp |-> w *^* sp + 1 |-> junk2)) cthM11_2
  -> cthM' = MemUpdW cthM1 (sp - 2) (pc + 1) junk1
  -> MemMod.disjUnion (MemMod.merge (MemUpdW cthM11_2 (sp - 2) (pc+1) junk1) cthM0) (MemMod.merge INVM cthM11_1) cthM'.
Proof.
  intros.

  assert ((eqM cthM11_1 *^* eqM cthM11_2 *^* (eqM INVM *^* eqM cthM0)) cthM1).
  exists cthM11. exists cthM12. split; auto.
  split. exists cthM11_1. exists cthM11_2. unfold eqM.
  auto. exists INVM. exists cthM0. unfold eqM. auto.

  assert (MemMod.disjUnion (MemUpdW (MemMod.merge cthM11_2 cthM0) (sp-2) (pc+1) junk1) (MemMod.merge INVM cthM11_1) cthM').
  subst cthM'.
  unfold MemUpdW.
  apply MemMod.update_presv_disjUnion_left.
  apply MemMod.update_presv_disjUnion_left.

  assert ((eqM cthM11_2 *^* eqM cthM0 *^* (eqM INVM *^* eqM cthM11_1)) cthM1).
  sep on. sep matchH H4 with 4.
  unfold mstar at 1 in H3. InvertAll.
  unfold mstar in H3. InvertAll.
  unfold eqM in H3, H9. subst x1 x2.
  inversion_clear H6. subst x. 
  unfold mstar in H7.
  InvertAll. unfold eqM in H7, H9. subst x x1.
  inversion_clear H6. subst x0. auto.
  apply MemMod.merge_presv_indom_left.
  eapply enter_BLK_grt2_auxc_hyp01; eauto.
  assert (MemMod.in_dom (sp - 2 + 1) (MemMod.merge cthM11_2 cthM0)).
  apply MemMod.merge_presv_indom_left.
  eapply enter_BLK_grt2_auxc_hyp02; eauto.
  unfold MemMod.in_dom. unfold MemMod.put.
  induction (MemMod.beq_A (sp - 2 + 1) (sp - 2)). exists (pc+1). auto.
  auto.
  assert (MemMod.disjUnion (MemUpdW cthM11_2 (sp - 2) (pc + 1) junk1)
           cthM0 (MemUpdW (MemMod.merge cthM11_2 cthM0) (sp - 2) (pc + 1) junk1)).
  unfold MemUpdW.
  apply MemMod.update_presv_disjUnion_left.
  apply MemMod.update_presv_disjUnion_left.
  split; auto.
  assert ((eqM cthM11_2 *^* eqM cthM0 *^* (eqM INVM *^* eqM cthM11_1)) cthM1).
  sep on. sep matchH H4 with 4.
  unfold mstar at 1 in H6. InvertAll.
  unfold mstar in H6. InvertAll.
  unfold eqM in H6, H11. subst x1 x2.
  inversion_clear H8. auto.
  eapply enter_BLK_grt2_auxc_hyp01; eauto.
  assert (MemMod.in_dom (sp - 2 + 1) cthM11_2).
  eapply enter_BLK_grt2_auxc_hyp02; eauto.
  unfold MemMod.in_dom. unfold MemMod.put.
  induction (MemMod.beq_A (sp - 2 + 1) (sp - 2)). exists (pc+1). auto.
  auto.
  inversion_clear H6.
  rewrite H8 in H5. auto.
Qed.

Lemma enter_BLK_grt2_auxc_aux4:
forall pc pc' M R fl isr M' R' fl' isr' cthM1 cthM11 cthM12 INVM cthM0 M0 cthM_oldStk R0 fl0 w wqa junk1 junk2 cthM2 cthM_old M'2,
  nextS pc (comml (_call _BLOCK)) (M, R, fl, isr) (M', R', fl', isr')
  -> MemMod.disjUnion cthM11 cthM12 cthM1
  -> (R SP |-> w *^* R SP + 1 |-> junk2 *^* stkSpacePrec (R SP) (_BLK_Stk + 2)) cthM11
  -> MemMod.disjUnion INVM cthM0 cthM12
  -> (R' SP |-> pc + 1 *^* R' SP + 1 |-> junk1 *^* R' SP + 2 |-> w *^* R' SP + 3 |-> junk2 *^* 
        stkSpacePrec (R' SP) _BLK_Stk) cthM_oldStk
  -> (INV0 *^* INV1) INVM
  -> (INV0 *^* INV1 *^* wqa) M0
  -> MemMod.disjUnion M0 cthM_old M'2
  -> MemMod.disjUnion (MemMod.merge cthM0 cthM2) cthM_oldStk cthM_old
  -> fl0 _IF = false
  -> fl _IF = false
  -> isr = false
  -> R0 SP = R' SP + 2
  -> wrap_fun (gblk wqa) pc _BLOCK pc' (cthM1, R, fl, isr)
     (MemMod.merge (MemMod.merge cthM0 M0) cthM_oldStk, R0, fl0, false).
Proof.
  intros.
  unfold wrap_fun.
  unfold conjG.  
  exists (MemMod.merge (MemMod.merge cthM0 M0) cthM_oldStk, RF_update R0 SP (R0 SP - 2), fl0, false).
  split.

  assert (nextS pc (comml (_call _BLOCK)) (cthM1, R, fl, isr) (MemUpdW cthM1 (R SP - 2) (pc+1) junk1, RF_update R SP (R SP - 2), fl, isr)).
  apply nextS_call.
  unfold pushW.
  split.
  assert ((mTrue *^* (R SP |-> w *^* R SP + 1 |-> junk2 *^*
                          stkSpacePrec (R SP) (_BLK_Stk + 2))) cthM1).
  exists cthM12. exists cthM11. split; auto with PMap.
  unfold mTrue. split; auto.
  sep on. sep lift 3 in H12.
  sep stksplit 2 in H12.
  sep lift 0 in H12. sep trivial.
  split; auto.
  apply MemUpdW_MemUpd16 with junk1; auto.
  set (cthM' := MemUpdW cthM1 (R SP - 2) (pc + 1) junk1) in *.
  exists (cthM', RF_update R SP (R SP - 2), fl, isr).
  split; auto.
  assert (R SP - 2 = R' SP).
  apply enter_SW_grt_auxb0_hyp0 with pc M fl isr M' fl' isr' _BLOCK; auto.
  unfold gblk.
  exists (R SP - 2). simplS. reg simpl.

  assert (R SP = R' SP + 2).
  rewrite <- H13.
  assert (R SP >= 2).
  apply ge_trans with (_BLK_Stk + 2).
  sep on. sep lift 2 in H1.
  apply stkSpacePrec_sp_ge_sz1 with cthM11 (R SP |-> w *^* R SP + 1 |-> junk2); auto.
  auto with arith.
  assert (R SP - 2 + 2 = R SP). auto. auto. rename H14 into HRSP'.

  split.

  unfold prepost. simplS.
  sep on.
  sep lift 2 in H1.
  sep stksplit in H1.
  destruct H1 as [cthM11_1 [cthM11_2 [HcthM11split [HcthM11_1 HcthM11_2]]]].

  exists (MemMod.merge INVM cthM11_1).
  exists (MemMod.merge (MemUpdW cthM11_2 (R SP - 2) (pc+1) junk1) cthM0).
  sep lift 4 in H3.
  unfold mstar at 1 in H3.
  sep off.
  InvertAll.
  rename x into oldStk1. rename x0 into oldStk2.
  exists (MemMod.merge M0 oldStk1).

  split.
  exists INVM. exists cthM11_1. split; auto.
  split; auto.
  apply enter_BLK_grt2_auxc_hyp1 with cthM1 cthM11 cthM12 cthM11_2 cthM0; auto.

  split.
  exists M0. exists oldStk1.
  rewrite <- H13 in H3. split; auto.
  split; auto.
  apply enter_BLK_grt2_auxc_hyp2 with M'2 cthM_old (MemMod.merge cthM0 cthM2) cthM_oldStk oldStk2; auto.

  split. 
  assert (MemMod.disjUnion (MemMod.merge (MemUpdW cthM11_2 (R SP - 2) (pc + 1) junk1) cthM0)
                             (MemMod.merge INVM cthM11_1) cthM').
  apply enter_BLK_grt2_auxc_hyp0 with cthM1 cthM11 cthM12 w junk2; auto.
  auto with PMap.


  assert (MemUpdW cthM11_2 (R SP - 2) (pc + 1) junk1 = oldStk2).
  apply enter_BLK_grt2_auxc_hyp3 with (R' SP) w junk2; auto.
  rewrite H14.

  apply enter_BLK_grt2_auxc_hyp4 with M'2 cthM_old cthM2; auto.

  rewrite <- H11 in HRSP'. rewrite <- HRSP'. split; auto.
  rewrite <- H8 in H9. auto.

  apply nextS_retw. exists (pc+1).
  unfold popW.
  reg simpl.
  split.
  rewrite H11. assert (R' SP + 2 - 2 = R' SP). auto.
  rewrite H12.

  sep on.
  assert (cthM_oldStk ||- R' SP |-> pc + 1 *^* R' SP + 1 |-> junk1 *^* (R' SP + 2 |-> w *^*
                             R' SP + 3 |-> junk2 *^* stkSpacePrec (R' SP) _BLK_Stk)).
  sep matchH H3 with 5.
  sep off.
  apply enter_BLK_grt2_auxc_hyp5 with M'2 cthM_old cthM2 junk1
            (R' SP + 2 |-> w *^* R' SP + 3 |-> junk2 *^*
             stkSpacePrec (R' SP) _BLK_Stk); auto.
  rewrite H11. assert (R' SP + 2 - 2 = R' SP). auto.
  rewrite H12. rewrite <- H11.
  assert (RF_update R0 SP (R0 SP) = R0).
  apply RF_update_prev. auto.
Qed.

Lemma enter_BLK_grt2_auxc:
forall cthM1 cthM2 cthM R fl isr cthM11 cthM12 INVM cthM0 cthM_oldStk pc junk1 junk2
            M0 M'2 cthM_old R0 fl0 (p p': Pred) M M' R' fl' isr' w wqa,
  MemMod.disjUnion cthM1 cthM2 cthM
  -> p (cthM1, R, fl, isr)
  -> NH_Inv fl cthM2
  -> MemMod.disjUnion cthM11 cthM12 cthM1
  -> (R SP |-> w *^* R SP + 1 |-> junk2 *^* stkSpacePrec (R SP) (_BLK_Stk + 2)) cthM11
  -> MemMod.disjUnion INVM cthM0 cthM12
  -> (INV0 *^* INV1) INVM
  -> MemMod.disjUnion (MemMod.merge cthM0 cthM2) cthM_oldStk cthM_old
  -> (R' SP |-> pc + 1 *^* R' SP + 1 |-> junk1 *^* R' SP + 2 |-> w *^* R' SP + 3 |-> junk2 *^*
         stkSpacePrec (R' SP) _BLK_Stk) cthM_oldStk
  -> (INV0 *^* INV1 *^* wqa) M0
  -> MemMod.disjUnion M0 cthM_old M'2
  -> (forall pc' : nat, pimp (wrap_fun (gblk wqa) pc _BLOCK pc' (cthM1, R, fl, isr)) p')
  -> fl0 _IF = false
  -> fl _IF = false
  -> nextS pc (comml (_call _BLOCK)) (M, R, fl, isr) (M', R', fl', isr')
  -> R0 SP = R' SP + 2
  -> isr = false
  -> NH_pinvstar p' (S_Mem_Upd (cthM_old, R0, fl0, false) M'2).
Proof.
  intros.
  unfold NH_pinvstar.
  unfold pmstar.
  exists (MemMod.merge (MemMod.merge cthM0 M0) cthM_oldStk).
  exists cthM2.
  split.
  assert (S_Mem (S_Mem_Upd (cthM_old, R0, fl0, false) M'2) = M'2).
  auto.
  rewrite H16. clear H16.
  apply enter_SW_grt2_auxc_hyp0 with cthM_old; auto.

  assert (MemMod.disjoint cthM12 cthM2).
  apply MemMod.disj_merge_disj_right with cthM11.
  inversion_clear H2. subst cthM1. inversion_clear H. auto.
  apply MemMod.disj_merge_disj_right with INVM.
  inversion_clear H4. subst cthM12. auto.

  split.
  simplS.
  unfold pimp in H10.
  apply H10 with 0.


  apply enter_BLK_grt2_auxc_aux4 with M M' R' fl' isr' cthM11 cthM12 INVM w junk1 junk2 cthM2 cthM_old M'2; auto.

  simplS.
  unfold NH_Inv. rewrite H11. unfold NH_Inv in H1. rewrite H12 in H1.
  auto.
Qed.


Lemma enter_BLK_grt2_hyp8:
forall T RdyQ RdyQ' tid' B BlkQ w,
  RdyQ = addThrd RdyQ' tid'
  -> WF_TWQ T B RdyQ
  -> B w = Some BlkQ
  -> WQMod.not_in_dom tid' BlkQ.
Proof.
  intros.
  unfold WF_TWQ in H0.
  InvertAll.
  unfold disj_RQ_WQ in H4.
  assert (WQMod.disjoint BlkQ RdyQ).
  apply H4 with w; auto.
  unfold WQMod.disjoint in H3.
  assert (WQMod.not_in_dom tid' BlkQ \/ WQMod.not_in_dom tid' RdyQ).
  apply H3. inversion_clear H5; auto.
  assert (WQMod.in_dom tid' RdyQ).
  unfold addThrd in H. rewrite H.
  unfold WQMod.in_dom. unfold WQMod.put.
  induction (WQMod.beq_A tid' tid'); auto. exists dummy_sglton. auto.
  absurd (tid' = tid'); auto.
  absurd (~ WQMod.in_dom tid' RdyQ);
  auto with PMap.
Qed.

Lemma enter_BLK_grt2_hyp9:
forall T B RdyQ Q tid l rho,
  WF_currTID tid T B RdyQ rho
  -> WQSetMod.maps_to B l Q
  -> WQMod.not_in_dom tid Q.
Proof.
  intros.
  unfold WF_currTID in H.
  InvertAll.
  unfold unblocked in H1.
  apply H1 with l; auto.
Qed.

Lemma enter_BLK_grt2_hypa:
forall M'1 M21' M22 T T' B B' RdyQ RdyQ' BlkQ BlkQ' tid tid' w tcb,
  MemMod.disjUnion M21' M22 M'1
  -> (embed_curr_TCB T' tid' *^* embed_RdyQ T' RdyQ' *^*
         ptoTQ w (filterTQ T' BlkQ')) M21'
  -> embed_WQSet T (WQSetMod.remove B w) M22
  -> B' = WQSetMod.put B w BlkQ'
  -> B w = Some BlkQ
  -> BlkQ' = addThrd BlkQ tid
  -> T' = TQMod.put T tid tcb
  -> RdyQ = addThrd RdyQ' tid'
  -> WF_currTID tid T B RdyQ rho_ready
  -> (embed_curr_TCB T' tid' *^* (embed_RdyQ T' RdyQ' *^* embed_WQSet T' B')) M'1.
Proof.
  intros.
  assert ((embed_curr_TCB T' tid' *^* embed_RdyQ T' RdyQ' 
              *^* ptoTQ w (filterTQ T' BlkQ') *^* embed_WQSet T (WQSetMod.remove B w)) M'1).
  exists M21'. exists M22. split; auto.
  sep on.
  sep lift 0 in H8.
  sep match_left H8; auto.
  sep match_left H8; auto.
  sep off.
  unfold mstar in H8. InvertAll.
  assert (embed_WQSet T' (WQSetMod.put (WQSetMod.remove B w) w BlkQ') M).
  apply wq_cons with x0 x; auto with PMap.
  subst T'.
  apply ext_unblockedT_presv_WQSet; auto.
  unfold WF_currTID in H7.
  InvertAll.
  unfold unblocked in *.
  intros.
  apply H5 with w0.
  unfold WQSetMod.remove in H14.
  unfold WQSetMod.maps_to in H14.
  induction (WQSetMod.beq_A w0 w).
  inversion H14. auto.

  assert (WQSetMod.put (WQSetMod.remove B w) w BlkQ' = WQSetMod.put B w BlkQ').
  auto with PMap.
  subst B'.
  rewrite <- H12. auto.
Qed.



Lemma enable_blk_unique_args:
forall S w w' wqa wqa', enable_blk w wqa S -> enable_blk w' wqa' S -> (w = w' /\ wqa = wqa').
Proof.
  unfold enable_blk.
  intros.
  InvertAll.
  induction S as (((M, R), fl), isr).
  UnfoldStateAll.
  assert ((mTrue *^* 
          (R SP |->w w' *^*
             (INV0 *^* INV1 *^* stkSpacePrec (R SP) (_BLK_Stk + 2)))) M).
  unfold pmstar in H2.
  InvertAll. exists x. exists x0. split; auto.
  assert ((mTrue *^* 
          (R SP |->w w *^*
             (INV0 *^* INV1 *^* stkSpacePrec (R SP) (_BLK_Stk + 2)))) M).
  unfold pmstar in H3.
  InvertAll. exists x. exists x0. split; auto.
  assert (w = w').
  sep on.
  sep lift 1 in H4.
  sep lift 1 in H.
  apply ptoW_ptoW_eq1 with M (R SP) 
           (INV0 *^* INV1 *^* stkSpacePrec (R SP) (_BLK_Stk + 2) *^* mTrue)
           (INV0 *^* INV1 *^* stkSpacePrec (R SP) (_BLK_Stk + 2) *^* mTrue); auto.
  subst w. split_l.
  rewrite H0 in H1. inversion H1. auto.
Qed.
          


Lemma enter_BLK_grt2_auxc_hyp6:
forall cthM11 cthM11' sp sp' pc junk1 w junk2,
  (sp |->w w *^* stkSpacePrec sp (_BLK_Stk + 2)) cthM11
  -> cthM11' = MemUpdW cthM11 sp' (pc + 1) junk1
  -> (sp' |-> pc + 1 *^* sp' + 1 |-> junk1 *^* sp' + 2 |-> w *^*
        sp' + 3 |-> junk2 *^* stkSpacePrec sp' _BLK_Stk) cthM11'
  -> sp = sp' + 2
  -> (sp |-> w *^* sp + 1 |-> junk2 *^* stkSpacePrec sp (_BLK_Stk + 2)) cthM11.
Proof.
  intros.
  subst sp.
  sep on.
  sep lift 1 in H.
  sep stksplit in H.
  sep lift 1 in H.
  destruct H as [cthM11_1 [cthM11_2 [HcthM11split [HcthM11_1 HcthM11_2]]]].
  assert (MemMod.disjUnion (MemUpdW cthM11_1 sp' (pc+1) junk1) cthM11_2 cthM11').
  subst cthM11'.
  unfold MemUpdW.
  assert (MemMod.in_dom sp' cthM11_1 /\ MemMod.in_dom (sp' + 1) cthM11_1).
  assert (exists v, cthM11_1 ||- sp' |->w v).
  apply ptoxW_ptoW; auto.
  sep off.
  inversion H.
  apply ptoW_indom with x; auto.
  inversion_clear H.
  apply MemMod.update_presv_disjUnion_left.
  apply MemMod.update_presv_disjUnion_left; auto.
  unfold MemMod.in_dom. unfold MemMod.put.
  induction (MemMod.beq_A (sp' + 1) sp').
  exists (pc+1); auto.
  auto.
  assert (cthM11' ||- sp' |-> pc + 1 *^* sp' + 1 |-> junk1 *^* (sp' + 2 |-> w *^*
           sp' + 3 |-> junk2 *^* stkSpacePrec sp' _BLK_Stk)).
  sep matchH H1 with 5. clear H1.
  destruct H2 as [cthM11'1 [cthM11'2 [Hsplit' [H11'1 H11'2]]]].
  assert (cthM11_2 = cthM11'2).
  assert ((sp' + 2 |->w w *^* stkSpacePrec sp' _BLK_Stk) cthM11'2).
  apply ptoW_intro1 with junk2; auto.
  assert (sp' + 2 + 1 = sp' + 3). auto.
  rewrite H1. auto.
  assert (precise (sp' + 2 |->w w *^* stkSpacePrec sp' _BLK_Stk)).
  apply mstar_presv_prec.
  apply ptoW_precise. apply precise_stkSpacePrec.
  unfold precise in H2.
  apply H2 with cthM11'; auto.
  apply MemMod.disjUnion_subseteq_right with (MemUpdW cthM11_1 sp' (pc + 1) junk1); auto.
  apply MemMod.disjUnion_subseteq_right with cthM11'1; auto.
  rewrite <- H1 in H11'2.
  sep lift 2.
  sep stksplit.
  sep lift 1.
  exists cthM11_1. exists cthM11_2.
  split; auto. split; auto.
  assert (sp' + 2 + 1 = sp' + 3). auto.
  rewrite H2. auto.
Qed.



(* >>>>>>>>>>>>>> *)
Lemma enter_BLK_grt:
forall D psi pc theta D' Psi Psi' Psi'0 S S',
  NH_WFcmd psi pc theta (comml (_call _BLOCK))
  -> Minimal_D D
  -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
  -> lift_psi psi rho_ready Psi
  -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
  -> subsetCT Psi' Psi'0
  -> Itp_AIM rho_ready theta Psi'0 S
  -> nextS pc (comml (_call _BLOCK)) S S'
  -> WFST D rho_prim 0 Psi'0 (scap_lift_g BLK_grt S').
Proof.
  intros.
  unfold WFST.
  intros.
  unfold scap_lift_g in H7.
  InvertAll.
  induction S' as (((M', R'), fl'), isr').
  induction S as (((M, R), fl), isr).
  unfold S_ISR at 1 in H7.
  unfold S_Flag at 1 in H10.
  induction theta.
  rename n into tag.
  unfold Itp_AIM in H5.
  unfold pmstar in H5.
  InvertAll.
  rename x into T. rename x0 into B. rename x1 into RdyQ. rename x2 into tid.
  rename x3 into M1. rename x4 into M2.

  assert (exists TM, exists cthM, exists M12, exists cthM1, exists cthM2,
        MemMod.disjUnion M12 cthM M1
        /\ MemMod.disjUnion cthM1 cthM2 cthM
        /\ WF_currTID tid T B RdyQ rho_ready
        /\ WF_TWQ T B RdyQ
        /\ exists_Idle T RdyQ tid
        /\ WF_WQ_WQSpec B Delta
        /\ MemSplit (ThrdMemMod.remove TM tid) M12
        /\ ThrdMemMod.maps_to TM tid cthM
        /\ WF_Curr_Thrd Psi'0 (S_Mem_Upd (M1, R, fl, isr) cthM) rho_ready (Ast p g tag) T tid
        /\ WF_ReadyQ Psi'0 T RdyQ TM
        /\ WF_WQSet Psi'0 T B TM
        /\ p (cthM1, R, fl, isr)
        /\ NH_Inv fl cthM2
        /\ (tag = 1 -> NH_WFST rho_ready Psi'0 (NH_ItpG g (cthM, R, fl, isr)))).
  apply decompose_World; auto.
  InvertAll. rename x into TM. rename x0 into cthM. rename x1 into M12. rename x2 into cthM1. rename x3 into cthM2.
(* (M, R, fl, isr) ---> call ----> (M', R', fl', isr')    *)
(* (M', R', fl', isr') ---> BLK_grt ---> S0               *)

(* M: H9 : MemMod.disjUnion M1 M2 M                       *)
(* M1 : H5  *)
(* M2 : H12 *)
(* M1 : H11 : MemMod.disjUnion M12 cthM M1                *)
(* M12 : H18  *)
(* cthM : H19 *)
(* cthM : H13 : MemMod.disjUnion cthM1 cthM2 cthM         *)
(* cthM1 : H23 *)
(* cthM2 : H24 *)

  (* tid not in RdyQ *)
  assert (WQMod.not_in_dom tid RdyQ). unfold WF_currTID in H14. InvertAll. auto.

  unfold NH_WFcmd in H.
  inversion_clear H. clear H28.
  induction H27.
  inversion_clear H. clear H27.

  assert (exists w, exists wqa, enable_blk w wqa (cthM1, R, fl, isr)).
  apply NH_Call_blk_enable with psi pc p g tag; auto.
  InvertAll. rename x into w. rename x0 into wqa.
  unfold NH_Call_Prim in H28.
  inversion_clear H28.
  inversion H27.
  absurd (_BLOCK <> _SWITCH); auto.
  clear H28. inversion H30.
  subst p0 g0 tag0. clear H30. clear H27.

  (* BlkQ *)
  assert (exists BlkQ, 
          B w = Some BlkQ
          /\ Delta w = SomeT wqa 
          /\ (pmstar pTrue (mstar (ptoW (S_RF (cthM1, R, fl, isr) SP) w)
                                  (mstar (mstar INV0 INV1) 
                                         (stkSpacePrec (S_RF (cthM1, R, fl, isr) SP) (_BLK_Stk + 2))
                                  )
                           )
              ) (cthM1, R, fl, isr)).
  apply WF_AIM_World_blk2 with Psi'0 (M1, R, fl, isr) T RdyQ tid p g tag; auto.
  InvertAll. rename x into BlkQ.
  unfold S_RF in H33.


  assert (exists cthM11, exists cthM12,
            MemMod.disjUnion cthM11 cthM12 cthM1
            /\ (R SP |->w w *^* stkSpacePrec (R SP) (_BLK_Stk + 2)) cthM11
            /\ (INV0 *^* INV1 *^* mTrue) cthM12).
  apply enter_BLK_grt_hyp0 with R fl isr; auto.
  InvertAll.
  rename x into cthM11. rename x0 into cthM12.
(* cthM1 : H34 : MemMod.disjUnion cthM11 cthM12 cthM1 *)
(* cthM11 : H30 *)
(* cthM12 : H36 *)

  assert (disjUnionList (M2 :: M12 :: cthM2 :: cthM12 :: cthM11 :: nil) M).
  apply mem_list_compose_a with cthM1 cthM M1; auto.
  assert 
    (exists junk,
        (R' SP |-> (pc+1) *^* R' SP + 1 |-> junk *^* (R' SP + 2 |->w w) *^* (stkSpacePrec (R' SP) _BLK_Stk)) (MemUpdW cthM11 (R' SP) (pc + 1) junk)
        /\ disjUnionList (M2 :: M12 :: cthM2 :: cthM12 :: (MemUpdW cthM11 (R' SP) (pc + 1) junk) :: nil) M').

  apply call_fun_frame2 with M R fl fl' isr isr' _BLOCK; auto.
  InvertAll. rename x into junk1. set (cthM11' := MemUpdW cthM11 (R' SP) (pc + 1) junk1) in *.
  assert (((embed_curr_TCB T tid *^* embed_RdyQ T RdyQ *^* (ptoTQ w (filterTQ T BlkQ))) *^* (embed_WQSet T (WQSetMod.remove B w))) M2).
  apply split_blk_Q; auto.
  unfold mstar at 1 in H38.
  InvertAll. rename x into M21. rename x0 into M22.
  assert (exists M1', exists M2', 
           MemMod.disjUnion cthM11' M21 M1' 
           /\ disjUnionList (M22 :: M12 :: cthM2 :: cthM12 :: nil) M2'
           /\ MemMod.disjUnion M1' M2' M').
  apply mem_list_compose_b with M2; auto.
  InvertAll.  rename x into M1'. rename x0 into M2'.

  (* tid and tcb *)
  assert (exists tcb, T tid = Some tcb /\ TCB_Status tcb = READY).
  unfold WF_currTID in H14.
  InvertAll.
  exists x. unfold rho_tcb_consistent in H50.
  inversion_clear H50.
  split. unfold TQMod.maps_to in H48. auto.
  apply H49; auto.
  InvertAll. rename x into tcb.

  assert (exists junk2, 
           (R' SP |-> pc + 1 *^* R' SP + 1 |-> junk1 *^* R' SP + 2 |-> w *^* R' SP + 3 |-> junk2 *^*
             stkSpacePrec (R' SP) _BLK_Stk) cthM11').
  sep on.
  sep lift 2 in H37.
  assert (exists junk, 
            cthM11' ||- R' SP + 2 |-> w *^* R' SP + 2 + 1|-> junk *^*
            (stkSpacePrec (R' SP) _BLK_Stk *^*
             (R' SP |-> pc + 1 *^* R' SP + 1 |-> junk1))).
  apply ptoW_elim1; auto.
  inversion H46. exists x.
  sep on. sep matchH H48 with 5.
  assert (R' SP + 2 + 1 = R' SP + 3). omega.
  rewrite H49 in H48. auto.
  substH H37 with H46. clear H46.
  destruct H37 as [junk2 H37].

  unfold BLK_grt in H8.
  assert (exists M0' : MemMod.Map,
         exists RdyQ' : WQ,
           exists tid' : TID,
             exists sp' : nat,
               exists tcb' : Thread,
                 MemMod.disjUnion M0' M2' (S_Mem S0) /\
                 RdyQ = addThrd RdyQ' tid' /\
                 BLK_Post_aux (TQMod.put T tid ((R' SP), BLOCKED, pc+1)) 
                              RdyQ' (addThrd BlkQ tid) tid' sp' (R' SP) w tcb' (pc+1) junk1 junk2 (S_Mem_Upd S0 M0')).
  apply H8 with M1' T BlkQ tid tcb; auto.
  unfold BLK_Prd_aux.
  split. assert (S_Mem (S_Mem_Upd (M', R', fl', isr') M1') = M1'). auto.
  rewrite H46.
  apply mem_compose_c_blk with cthM11' M21; auto.
  split; auto.
  split. auto.
  split. 
  apply WF_AIM_World_RdyQinT with Psi'0 (S_Mem_Upd (M, R, fl, isr) M1) B tid rho_ready (Ast p g tag).
  auto.
  split.
  apply WF_AIM_World_BlkQinT with Psi'0 (S_Mem_Upd (M, R, fl, isr) M1) B RdyQ tid rho_ready (Ast p g tag) w; auto.
  split; auto.
  split; auto.
  split. auto.
  split.
  apply WF_AIM_World_cthNotBlk with Psi'0 (S_Mem_Upd (M, R, fl, isr) M1) T B RdyQ (Ast p g tag) w rho_ready; auto.
  split.
  apply WF_AIM_World_RdyNotBlk with Psi'0 (M1, R, fl, isr) T B tid (Ast p g tag) w rho_ready; auto.
  apply WF_AIM_World_WFTQ with Psi'0 (S_Mem_Upd (M, R, fl, isr) M1) B RdyQ tid rho_ready (Ast p g tag); auto.
  clear H8.
  InvertAll.
  rename x into M0'. rename x0 into RdyQ'. rename x1 into tid'.
  rename x2 into sp'. rename x3 into tcb'.

  (* fl and isr *)
  assert ((fl _IF = false) /\ (isr = false)).
  assert (S_Flag (cthM1, R, fl, isr) _IF = false).
  apply H29; auto. split. unfold S_Flag in H48. auto.
  unfold WF_Curr_Thrd in H20.
  assert (beq_nat rho_ready rho_ih = false). simpl. auto.
  rewrite H50 in H20. unfold itp_NH_theta in H20.
  InvertAll. UnfoldStateAll. auto.
  inversion_clear H48.

  assert (tid = tid' \/ tid <> tid').
  apply nat_eq_neq_dec.
  inversion_clear H48.

  (* tid = tid' : contradition *)
  assert (WQMod.in_dom tid' RdyQ).
  unfold addThrd in H8.
  rewrite H8.
  unfold WQMod.put.
  unfold WQMod.in_dom.
  induction (WQMod.beq_A tid' tid'). exists dummy_sglton. auto.
  absurd (tid' = tid'); auto.
  rewrite <- H52 in H48.
  absurd (~ WQMod.in_dom tid RdyQ); auto with PMap.

  (* tid <> tid' *)
  assert (WQMod.in_dom tid' RdyQ).
  unfold addThrd in H8.
  rewrite H8.
  unfold WQMod.put.
  unfold WQMod.in_dom.
  induction (WQMod.beq_A tid' tid'). exists dummy_sglton. auto.
  absurd (tid' = tid'); auto.
  assert (WF_ReadyQ Psi'0 T RdyQ TM). auto.
  unfold WF_ReadyQ in H21.
  assert (exists tcb : Thread,
          TQMod.maps_to T tid' tcb /\
          (TCB_Status tcb = READY \/ TCB_Status tcb = IDLE) /\
          (exists M : Mem,
             ThrdMemMod.maps_to TM tid' M /\
             (TCB_SP tcb |~>w TCB_PC tcb) M /\
             (WF_Rdy_Thrd Psi'0 M tcb \/ WF_Rdy_Thrd_IH Psi'0 M tcb))).
  apply H21. auto.
  InvertAll. rename x into tcb''.
  rename x0 into cthMR. clear H21.

  induction tcb'' as ((sp'', status''), pc'').
  exists pc''.
  assert (tcb' = (sp'', status'', pc'') /\ sp' = sp'' /\ S_RF S0 SP = sp').
  unfold BLK_Post_aux in H49.
  InvertAll.
  assert (tcb' = (sp'', status'', pc'')).
  assert (TQMod.maps_to T tid' tcb').
  unfold TQMod.maps_to in H21.
  unfold TQMod.put in H21. induction (TQMod.beq_A tid' tid).
  absurd (tid' <> tid); auto. unfold TQMod.maps_to. auto.
  unfold TQMod.maps_to in H61. unfold TQMod.maps_to in H54.
  rewrite H61 in H54. inversion H54. auto.
  split; auto.
  split; auto.
  rewrite H61 in H49. auto.
  inversion_clear H21. inversion_clear H60.

  unfold TCB_PC in H56. unfold TCB_SP in H56.
  split.
  apply enter_SW_grt2_hyp1 with TM tid tid' cthMR sp'' M0' M2' M22 M12 cthM2 cthM12; auto.
  rewrite <- H21. auto.

  inversion_clear H59.

  (* not in interrupt handler *)
  unfold WF_Rdy_Thrd in H60.
  InvertAll.
  rename x into rho'. rename x0 into pR. rename x1 into gR.
  rename x2 into tagR. clear H62.
  exists (packS0 rho' (Ast pR gR tagR)).
  split; auto.
  intros.
  unfold Itp0.
  assert (D rho' = SomeT1 _ (Itp_AIM rho')).
  apply non_ih_spec_miniD; auto.
  rewrite H64. clear H64.
  unfold Itp_AIM.
  exists (TQMod.put T tid (R' SP, BLOCKED, pc + 1)).
  set (T' := TQMod.put T tid (R' SP, BLOCKED, pc + 1)).
  exists (WQSetMod.put B w (addThrd BlkQ tid)). exists RdyQ'. exists tid'.
  set (BlkQ' := addThrd BlkQ tid).
  set (B' := WQSetMod.put B w BlkQ').

  assert (exists INVM, exists cthM0, exists M21', exists M122, exists cthM_old, exists cthM_oldStk,
        MemMod.disjUnion INVM cthM0 cthM12 
        /\ (INV0 *^* INV1) INVM
        /\ (embed_curr_TCB T' tid' *^* embed_RdyQ T' RdyQ' *^* ptoTQ w (filterTQ T' BlkQ')) M21'
        /\ MemMod.disjUnion M122 cthMR M12
        /\ MemSplit (ThrdMemMod.remove (ThrdMemMod.remove TM tid) tid') M122
        /\ MemMod.disjUnion (MemMod.merge cthM0 cthM2) cthM_oldStk cthM_old
        /\ (R' SP |-> (pc+1) *^* R' SP + 1 |-> junk1 
              *^* R' SP + 2 |-> w *^* R' SP + 3 |-> junk2 *^* stkSpacePrec (R' SP) _BLK_Stk) cthM_oldStk
        /\ disjUnionList (M22 :: cthMR :: INVM :: M21' :: cthM_old :: M122 :: nil) (S_Mem S')).

  apply enter_BLK_grt2_hyp2 with T B sp' tcb' pc' S0 M0' M2'; auto.
  unfold NH_Inv.
  rewrite <- H10.
  assert (S_Flag (M, R, fl, isr) _IF = S_Flag (M', R', fl', isr') _IF 
             /\ S_ISR (M, R, fl, isr) = S_ISR (M', R', fl', isr')).
  apply enter_SW_grt_auxb0_hyp4 with pc _BLOCK; auto.
  inversion_clear H64.
  unfold S_Flag in H66. rewrite <- H66.
  unfold NH_Inv in H24. auto.
  InvertAll.
  rename x into INVM. rename x0 into cthM0. rename x1 into M21'. rename x2 into M122.
  rename x3 into cthM_old. rename x4 into cthM_oldStk.
  unfold pmstar.
  assert (
      exists M'0, exists M'1, 
        MemMod.disjUnion M'0 M'1 (S_Mem S')
        /\ MemMod.disjUnion M21' M22 M'1
        /\ disjUnionList (cthMR :: INVM :: cthM_old :: M122 :: nil) M'0
  ).
  apply enter_SW_grt2_hyp3; auto.
  InvertAll. rename x into M'0. rename x0 into M'1.
  exists M'0. exists M'1. split; auto.
  split.
  unfold WF_AIM_World.
  split.
  assert (WF_currTID tid' T' B RdyQ' rho').
  apply enter_BLK_grt2_hyp4 with T RdyQ tid (R' SP, BLOCKED, pc + 1) (sp'', status'', pc'') rho_ready; 
  auto.
  unfold BLK_Post_aux in H49.
  InvertAll. auto.
  apply ext_B_presv_wffTID with B BlkQ w tid; auto.
  split.
  apply enter_BLK_grt2_hyp5 with T RdyQ tid (R' SP) (pc+1) tcb; auto.
  unfold BLK_Post_aux in H49. InvertAll. auto.
  split.
  apply enter_BLK_grt2_aux6 with RdyQ tid tid' tcb (R' SP, BLOCKED, pc + 1) B T rho_ready w BlkQ; auto.
  split.
  apply extB_presv_WF_WQ_WQSpec with B w BlkQ BlkQ'; auto.
  exists (ThrdMemMod.put (ThrdMemMod.put TM tid' (MemMod.merge cthMR INVM)) tid cthM_old).
  set (TM' := ThrdMemMod.put (ThrdMemMod.put TM tid' (MemMod.merge cthMR INVM)) tid cthM_old).
  exists (MemMod.merge cthMR INVM). set (cthM' := MemMod.merge cthMR INVM).
  split. assert (S_Mem (S_Mem_Upd S' M'0) = M'0). auto.
  rewrite H75.
  apply enter_SW_grt2_hyp7 with cthMR INVM cthM_old M122 TM tid tid'; auto.
  split. unfold cthM'. unfold TM'.
  unfold ThrdMemMod.put. unfold ThrdMemMod.maps_to.
  induction (ThrdMemMod.beq_A tid' tid).
  absurd (tid' <> tid); auto. induction (ThrdMemMod.beq_A tid' tid').
  auto. absurd (tid' = tid'); auto.
  split.
  unfold WF_Curr_Thrd.
  assert (beq_nat rho' rho_ih = false).
  apply neq_beq_false. inversion_clear H59. rewrite H75. discriminate.
  rewrite H75. discriminate. rewrite H75.
  unfold itp_NH_theta.
  assert (S_ISR S' = false).
  apply enter_SW_grt2_hyp8 with pc pc' M R fl isr M' R' fl' isr' S0 _BLOCK; auto.
  split. 
  UnfoldStateAll. auto.
  assert ((INV0 *^* INV1 -* itp_NH_theta rho' (Ast pR gR tagR) Psi'0)
          (cthMR, S_RF S', S_Flag S', false)).
  apply H65.
  rewrite H21 in H61.
  apply enter_SW_grt2_hyp9 with S0 pc'; auto.
  assert (S_ISR (M, R, fl, isr) = S_ISR S' /\ S_Flag (M, R, fl, isr) _IF = S_Flag S' _IF).
  apply enter_SW_grt_hyp9 with pc pc' _BLOCK (M', R', fl', isr') S0; auto.
  inversion_clear H78.
  rewrite <- H80. unfold S_Flag. auto.
  assert (itp_NH_theta rho' (Ast pR gR tagR) Psi'0 (cthM', S_RF S', S_Flag S', false)).
  unfold cthM'.
  unfold pm_mwand in H78.
  assert ((MemMod.merge cthMR INVM, S_RF S', S_Flag S', false) =
           S_Mem_Upd (cthMR, S_RF S', S_Flag S', false) (MemMod.merge cthMR INVM)).
  auto.
  rewrite H79. apply H78 with INVM; auto.
  unfold S_Mem.
  assert (MemMod.disjUnion cthMR INVM (MemMod.merge cthMR INVM)).
  unfold MemMod.disjUnion. split; auto.
  apply enter_SW_grt2_hypa with M22 M21' cthM_old M122 S'; auto.
  auto with PMap.
  assert (S_Mem_Upd (S_Mem_Upd S' M'0) cthM' = (cthM', S_RF S', S_Flag S', false)).
  induction S' as (((M'', R''), fl''), isr'').
  UnfoldStateAll. auto.
  rewrite H77. auto. rewrite H80.
  clear H80. unfold itp_NH_theta in H79.
  inversion_clear H79. auto.

  split.
  unfold WF_ReadyQ.
  intros.
  assert (tid0 = tid \/ tid0 <> tid). apply nat_eq_neq_dec.
  inversion_clear H77.
  (* tid0 = tid: contradiciton *)
  assert (WQMod.in_dom tid0 RdyQ).
  rewrite H8.
  unfold addThrd. unfold WQMod.in_dom, WQMod.put.
  induction (WQMod.beq_A tid0 tid').
  rewrite a in H75.
  assert (WQMod.not_in_dom tid' RdyQ').
  unfold BLK_Post_aux in H49. InvertAll. auto.
  absurd (~ WQMod.in_dom tid' RdyQ'); auto with PMap.
  auto. rewrite H78 in H77.
  absurd (~WQMod.in_dom tid RdyQ); auto with PMap.
  
 (* tid0 <> tid : other ready threads *)
  assert (WQMod.not_in_dom tid' RdyQ').
  unfold BLK_Post_aux in H49. InvertAll. auto.
  unfold WF_ReadyQ in H53.
  assert (exists tcb : Thread,
          TQMod.maps_to T tid0 tcb /\
          (TCB_Status tcb = READY \/ TCB_Status tcb = IDLE) /\
          (exists M : Mem,
             ThrdMemMod.maps_to TM tid0 M /\
             (TCB_SP tcb |~>w TCB_PC tcb) M /\
             (WF_Rdy_Thrd Psi'0 M tcb \/ WF_Rdy_Thrd_IH Psi'0 M tcb))).
  apply H53; auto.
  rewrite H8. unfold addThrd.
  unfold WQMod.in_dom, WQMod.put.
  induction (WQMod.beq_A tid0 tid').
  rewrite a in H75. 
  absurd (~ WQMod.in_dom tid' RdyQ'); auto with PMap.
  auto.
  InvertAll.
  exists x.
  split. unfold T'.
  unfold TQMod.put, TQMod.maps_to.
  induction (TQMod.beq_A tid0 tid).
  absurd (tid0 <> tid); auto. 
  unfold TQMod.maps_to in H79. auto. 
  split; auto.
  exists x0.
  split; auto.
  unfold TM'.
  unfold ThrdMemMod.put at 1. unfold ThrdMemMod.maps_to.
  induction (ThrdMemMod.beq_A tid0 tid). absurd (tid0 <> tid); auto.
  unfold ThrdMemMod.put.
  induction (ThrdMemMod.beq_A tid0 tid').
  rewrite a in H75.
  absurd (~ WQMod.in_dom tid' RdyQ'); auto with PMap.
  unfold ThrdMemMod.maps_to in H82. auto.
  unfold WF_WQSet.
  intros.
  assert (l = w \/ l <> w). apply nat_eq_neq_dec.
  inversion H78.
  (* l = w *)
  clear H78. rewrite H79 in H75.
  unfold B' in H75.
  unfold WQSetMod.maps_to, WQSetMod.put in H75.
  induction (WQSetMod.beq_A w w).
  inversion H75. rewrite <- H80 in H77.
  rewrite H79.
  clear H80 H79 H75 a Q l.
  unfold BlkQ' in H77.
  assert (tid0 = tid \/ tid0 <> tid). apply nat_eq_neq_dec.
  inversion_clear H75.

  (* tid0 = tid *)
  exists (R' SP, BLOCKED, pc + 1).
  exists cthM_old.
  rewrite H78.
  split. unfold T'.
  unfold TQMod.put, TQMod.maps_to.
  induction (TQMod.beq_A tid tid). auto. absurd (tid = tid); auto.
  split. unfold TM'.
  unfold ThrdMemMod.put at 1. unfold ThrdMemMod.maps_to.
  induction (ThrdMemMod.beq_A tid tid); auto.
  absurd (tid = tid); auto.
  unfold WF_Wait_Thrd.
  split; auto.
  split.
  assert ((R' SP |-> pc + 1 *^* R' SP + 1 |-> junk1 *^* (R' SP + 2 |-> w *^*
         R' SP + 3 |-> junk2 *^* stkSpacePrec (R' SP) _BLK_Stk)) cthM_oldStk).
  sep on.
  sep matchH H71 with 5.
  sep on.
  assert (cthM_oldStk ||- R' SP |->w pc + 1 *^*
                         (R' SP + 2 |-> w *^* R' SP + 3 |-> junk2 *^*
                           stkSpacePrec (R' SP) _BLK_Stk)).
  apply ptoW_intro1 with junk1; auto.
  clear H75 H71.
  assert (cthM_old ||- mTrue *^* 
                           (R' SP |->w pc + 1 *^*
                             (R' SP + 2 |-> w *^* R' SP + 3 |-> junk2 *^*
                               stkSpacePrec (R' SP) _BLK_Stk))).
  exists (MemMod.merge cthM0 cthM2). exists cthM_oldStk.
  split; auto. split_r. unfold mTrue. auto.
  sep lift 1 in H71. sep trivial.

  exists wqa. split; auto.
  exists p'. exists g'. exists tag.
  split.
  apply psi_sw_lookupNPC with psi Psi Psi'; auto.
  intros.
  unfold pm_mwand. unfold S_Mem.
  intros.
  unfold itp_NH_theta.
  split.
  UnfoldStateAll. auto.
  split.

  apply enter_BLK_grt2_auxc with cthM1 cthM2 cthM R fl isr cthM11 cthM12 INVM cthM0 cthM_oldStk pc
        junk1 junk2 M0 p M M' R' fl' isr' w wqa; auto.

  apply enter_BLK_grt2_auxc_hyp6 with cthM11' (R' SP) pc junk1; auto.
  assert (R SP - 2 = R' SP).
  eapply enter_SW_grt_auxb0_hyp0; eauto.
  assert (R SP >= 2). apply ge_trans with (_BLK_Stk + 2); auto with arith.
  unfold mstar in H30. InvertAll. apply stkSpacePrec_sp_ge_sz with x0; auto.
  rewrite <- H82. assert (R SP - 2 + 2 = R SP). auto.
  auto.

  assert (exists w : nat,
          exists wqa : WQassert,
            enable_blk w wqa (cthM1, R, fl, isr) /\
            (forall pc' : nat, pimp (wrap_fun (gblk wqa) pc _BLOCK pc' (cthM1, R, fl, isr)) p') /\
            (forall pc' : nat,
             pimp (wrap_fun (gblk wqa) pc _BLOCK pc' (cthM1, R, fl, isr) |> g') (g (cthM1, R, fl, isr)))).
  apply H32; auto. InvertAll.
  assert (w = x /\ wqa = x0). apply enable_blk_unique_args with (cthM1, R, fl, isr); auto.
  inversion_clear H84. rewrite H87. auto.

(* WFST *)
  intro.
  assert (NH_WFST rho_ready Psi'0 (NH_ItpG g (cthM, R, fl, isr))).
  apply H26. auto.
  clear H26. clear H82.
  apply NH_WFST_strengthen with (cthM, R, fl, isr) g; auto.
  unfold pimp.
  intros.
  unfold NH_ItpG.
  unfold NH_ItpG in H26.
  InvertAll. rename x into M'20. rename x0 into M'21.
  rename x1 into SM1. rename x2 into SM2.
  exists cthM1. exists cthM2. exists SM1. exists SM2.
  split; auto.
  split; auto.
  split. unfold S_Flag. auto.
  split. auto.
  assert (exists w : nat,
          exists wqa : WQassert,
            enable_blk w wqa (cthM1, R, fl, isr) /\
            (forall pc' : nat, pimp (wrap_fun (gblk wqa) pc _BLOCK pc' (cthM1, R, fl, isr)) p') /\
            (forall pc' : nat,
             pimp (wrap_fun (gblk wqa) pc _BLOCK pc' (cthM1, R, fl, isr) |> g') (g (cthM1, R, fl, isr)))).
  apply H32; auto. InvertAll.
  assert (w = x /\ wqa = x0). apply enable_blk_unique_args with (cthM1, R, fl, isr); auto.
  inversion_clear H89.
  rewrite <- H92 in H90. clear H92 H91 H86 H88 x0 x.

  unfold pimp in H90. apply H90 with pc'. clear H90.
  unfold linkPG.
  exists (S_Mem_Upd (S_Mem_Upd (cthM_old, R0, fl0, false) M'2) M'20).
  split; auto.
  assert (M'20 = MemMod.merge (MemMod.merge cthM0 M0) cthM_oldStk).
  apply enter_SW_grt2_hype with cthM_old M'2 cthM2 M'21 fl fl0; auto.
  assert (S_Mem_Upd (S_Mem_Upd (cthM_old, R0, fl0, false) M'2) M'20 = (M'20, R0, fl0, false)).
  UnfoldStateAll. auto. rewrite H88. clear H88.
  rewrite H86.
  apply enter_BLK_grt2_auxc_aux4 with M M' R' fl' isr' cthM11 cthM12 INVM w junk1 junk2 cthM2 cthM_old M'2; auto.

  apply enter_BLK_grt2_auxc_hyp6 with cthM11' (R' SP) pc junk1; auto.
  assert (R SP - 2 = R' SP).
  eapply enter_SW_grt_auxb0_hyp0; eauto.
  assert (R SP >= 2). apply ge_trans with (_BLK_Stk + 2); auto with arith.
  unfold mstar in H30. InvertAll. apply stkSpacePrec_sp_ge_sz with x0; auto.
  rewrite <- H88. assert (R SP - 2 + 2 = R SP). auto.
  auto.
  (* end of WFST *)

  assert (WQMod.in_dom tid0 BlkQ).
  unfold addThrd in H77.
  unfold WQMod.put, WQMod.in_dom in H77.
  induction (WQMod.beq_A tid0 tid). contradiction. auto.
  unfold WF_WQSet in H22.
  assert (exists tcb : Thread,
          exists M : Mem,
            TQMod.maps_to T tid0 tcb /\
            ThrdMemMod.maps_to TM tid0 M /\ WF_Wait_Thrd Psi'0 M tcb w).
  apply H22 with BlkQ; auto.
  InvertAll. exists x. exists x0.
  split. unfold T'. unfold TQMod.maps_to, TQMod.put.
  induction (TQMod.beq_A tid0 tid). 
  contradiction. auto.
  split; auto. unfold TM'.
  unfold ThrdMemMod.put at 1. unfold ThrdMemMod.maps_to.
  induction (ThrdMemMod.beq_A tid0 tid). contradiction.
  unfold ThrdMemMod.put. induction (ThrdMemMod.beq_A tid0 tid').
  assert (WQMod.not_in_dom tid' BlkQ).
  apply enter_BLK_grt2_hyp8 with T RdyQ RdyQ' B w; auto.
  rewrite <- a in H81.
  absurd (~ WQMod.in_dom tid0 BlkQ); auto with PMap. auto.
  absurd (w = w); auto. 
  clear H78.
  assert (WQSetMod.maps_to B l Q).
  unfold B' in H75.
  unfold WQSetMod.maps_to, WQSetMod.put in H75.
  induction (WQSetMod.beq_A l w). contradiction.
  auto.
  unfold WF_WQSet in H22.
  assert (exists tcb : Thread,
          exists M : Mem,
            TQMod.maps_to T tid0 tcb /\
            ThrdMemMod.maps_to TM tid0 M /\ WF_Wait_Thrd Psi'0 M tcb l).
  apply H22 with Q; auto.
  InvertAll. exists x. exists x0.
  assert (tid0 = tid \/ tid0 <> tid).
  apply nat_eq_neq_dec.
  inversion_clear H82.
  assert (WQMod.not_in_dom tid Q).
  apply enter_BLK_grt2_hyp9 with T B RdyQ l rho_ready; auto.
  rewrite <- H84 in H82.
  absurd (~ WQMod.in_dom tid0 Q); auto with PMap.
  split. unfold T'.
  unfold TQMod.maps_to, TQMod.put.
  induction (TQMod.beq_A tid0 tid). contradiction.
  auto.
  split; auto.
  unfold TM'.
  unfold ThrdMemMod.put at 1. unfold ThrdMemMod.maps_to.
  induction (ThrdMemMod.beq_A tid0 tid). contradiction.
  unfold ThrdMemMod.put.
  induction (ThrdMemMod.beq_A tid0 tid').
  assert (WQMod.not_in_dom tid' Q).
  apply enter_BLK_grt2_hyp8 with T RdyQ RdyQ' B l; auto.
  rewrite <- a in H82.
  absurd (~ WQMod.in_dom tid0 Q); auto with PMap. auto.
  apply enter_BLK_grt2_hypa with M21' M22 T B RdyQ BlkQ BlkQ' tid w (R' SP, BLOCKED, pc+1); auto.

  (* tid' is in interrupt hander *)
  unfold WF_Rdy_Thrd_IH in H60.
  InvertAll.
  rename x into p''. rename x0 into g''.
  rename x1 into tag''.
  exists (packS0 rho_ih (Ast p'' g'' tag'')).
  split; auto.
  intros.
  unfold Itp0.
  assert (D rho_ih = SomeT1 _ (Itp_AIM rho_ih)).
  unfold Minimal_D in H0. InvertAll. auto.
  rewrite H64. clear H64.
  unfold Itp_AIM.
  exists (TQMod.put T tid (R' SP, BLOCKED, pc + 1)).
  set (T' := TQMod.put T tid (R' SP, BLOCKED, pc + 1)).
  exists (WQSetMod.put B w (addThrd BlkQ tid)). exists RdyQ'. exists tid'.
  set (BlkQ' := addThrd BlkQ tid).
  set (B' := WQSetMod.put B w BlkQ').

  assert (exists INVM, exists cthM0, exists M21', exists M122, exists cthM_old, exists cthM_oldStk,
        MemMod.disjUnion INVM cthM0 cthM12 
        /\ (INV0 *^* INV1) INVM
        /\ (embed_curr_TCB T' tid' *^* embed_RdyQ T' RdyQ' *^* ptoTQ w (filterTQ T' BlkQ')) M21'
        /\ MemMod.disjUnion M122 cthMR M12
        /\ MemSplit (ThrdMemMod.remove (ThrdMemMod.remove TM tid) tid') M122
        /\ MemMod.disjUnion (MemMod.merge cthM0 cthM2) cthM_oldStk cthM_old
        /\ (R' SP |-> (pc+1) *^* R' SP + 1 |-> junk1 *^* R' SP + 2 |-> w *^* R' SP + 3 |-> junk2 *^* stkSpacePrec (R' SP) _BLK_Stk) cthM_oldStk
        /\ disjUnionList (M22 :: cthMR :: INVM :: M21' :: cthM_old :: M122 :: nil) (S_Mem S')).

  apply enter_BLK_grt2_hyp2 with T B sp' tcb' pc' S0 M0' M2'; auto.
  unfold NH_Inv.
  rewrite <- H10.
  assert (S_Flag (M, R, fl, isr) _IF = S_Flag (M', R', fl', isr') _IF 
             /\ S_ISR (M, R, fl, isr) = S_ISR (M', R', fl', isr')).
  apply enter_SW_grt_auxb0_hyp4 with pc _BLOCK; auto.
  inversion_clear H64.
  unfold S_Flag in H65. rewrite <- H65.
  unfold NH_Inv in H24. auto.
  InvertAll.
  rename x into INVM. rename x0 into cthM0. rename x1 into M21'. rename x2 into M122.
  rename x3 into cthM_old. rename x4 into cthM_oldStk.
  unfold pmstar.
  assert (
      exists M'0, exists M'1, 
        MemMod.disjUnion M'0 M'1 (S_Mem S')
        /\ MemMod.disjUnion M21' M22 M'1
        /\ disjUnionList (cthMR :: INVM :: cthM_old :: M122 :: nil) M'0
  ).
  apply enter_SW_grt2_hyp3; auto.
  InvertAll. rename x into M'0. rename x0 into M'1.
  exists M'0. exists M'1. split; auto.
  split.
  unfold WF_AIM_World.
  assert (WQMod.not_in_dom tid' RdyQ').
  unfold BLK_Post_aux in H49. InvertAll. auto.

  split.
  assert (WF_currTID tid' T' B RdyQ' rho_ih).
  apply enter_BLK_grt2_hyp4 with T RdyQ tid (R' SP, BLOCKED, pc + 1) (sp'', status'', pc'') rho_ready; 
  auto.
  apply ext_B_presv_wffTID with B BlkQ w tid; auto.
  split.
  apply enter_BLK_grt2_hyp5 with T RdyQ tid (R' SP) (pc+1) tcb; auto.
  split.
  apply enter_BLK_grt2_aux6 with RdyQ tid tid' tcb (R' SP, BLOCKED, pc + 1) B T rho_ready w BlkQ; auto.
  split.
  apply extB_presv_WF_WQ_WQSpec with B w BlkQ BlkQ'; auto.
  exists (ThrdMemMod.put (ThrdMemMod.put TM tid' (MemMod.merge cthMR INVM)) tid cthM_old).
  set (TM' := ThrdMemMod.put (ThrdMemMod.put TM tid' (MemMod.merge cthMR INVM)) tid cthM_old).
  exists (MemMod.merge cthMR INVM). set (cthM' := MemMod.merge cthMR INVM).
  split. assert (S_Mem (S_Mem_Upd S' M'0) = M'0). auto.
  rewrite H76.
  apply enter_SW_grt2_hyp7 with cthMR INVM cthM_old M122 TM tid tid'; auto.
  split. unfold cthM'. unfold TM'.
  unfold ThrdMemMod.put. unfold ThrdMemMod.maps_to.
  induction (ThrdMemMod.beq_A tid' tid).
  absurd (tid' <> tid); auto. induction (ThrdMemMod.beq_A tid' tid').
  auto. absurd (tid' = tid'); auto.
  split.
  unfold WF_Curr_Thrd.
  assert (beq_nat rho_ih rho_ih = true). simpl. auto.
  rewrite H76.

  assert ((INV0 *^* INV1 -* itp_IH_theta (Ast p'' g'' tag'') status'' Psi'0)
          (cthMR, S_RF S', S_Flag S', false)).
  apply H63.
  rewrite H21 in H61.
  apply enter_SW_grt2_hyp9 with S0 pc'; auto.
  assert (S_ISR (M, R, fl, isr) = S_ISR S' /\ S_Flag (M, R, fl, isr) _IF = S_Flag S' _IF).
  apply enter_SW_grt_hyp9 with pc pc' _BLOCK (M', R', fl', isr') S0; auto.
  inversion_clear H77.
  rewrite <- H79. unfold S_Flag. auto.
  exists (sp'', status'', pc''). exists status''.
  split.
  unfold TQMod.maps_to.
  unfold T'. unfold TQMod.put.
  induction (TQMod.beq_A tid' tid).
  absurd (tid' <> tid); auto. auto.
  split; auto.

  unfold pm_mwand in H77.
  assert (S_ISR S' = false).
  apply enter_SW_grt2_hyp8 with pc pc' M R fl isr M' R' fl' isr' S0 _BLOCK; auto.
  assert (itp_IH_theta (Ast p'' g'' tag'') status'' Psi'0
          (S_Mem_Upd (cthMR, S_RF S', S_Flag S', false) cthM')).
  apply H77 with INVM; auto.
  unfold S_Mem. unfold cthM'.
  assert (MemMod.disjUnion cthMR INVM (MemMod.merge cthMR INVM)).
  unfold MemMod.disjUnion. split; auto.
  apply enter_SW_grt2_hypa with M22 M21' cthM_old M122 S'; auto.
  auto with PMap.
  rewrite <- H78 in H79. UnfoldStateAll. auto.

  split.

  (* WF_ReadyQ *)
  unfold WF_ReadyQ.
  intros.
  assert (tid0 = tid \/ tid0 <> tid). apply nat_eq_neq_dec.
  inversion_clear H77.

  (* tid0 = tid: contradiciton *)
  assert (WQMod.in_dom tid0 RdyQ).
  rewrite H8.
  unfold addThrd. unfold WQMod.in_dom, WQMod.put.
  induction (WQMod.beq_A tid0 tid').
  rewrite a in H76.
  assert (WQMod.not_in_dom tid' RdyQ').
  unfold BLK_Post_aux in H50. InvertAll. auto.
  absurd (~ WQMod.in_dom tid' RdyQ'); auto with PMap.
  auto. rewrite H78 in H77.
  absurd (~WQMod.in_dom tid RdyQ); auto with PMap.

 (* tid0 <> tid : other ready threads *)
  assert (WQMod.not_in_dom tid' RdyQ').
  unfold BLK_Post_aux in H49. InvertAll. auto.
  unfold WF_ReadyQ in H53.
  assert (exists tcb : Thread,
          TQMod.maps_to T tid0 tcb /\
          (TCB_Status tcb = READY \/ TCB_Status tcb = IDLE) /\
          (exists M : Mem,
             ThrdMemMod.maps_to TM tid0 M /\
             (TCB_SP tcb |~>w TCB_PC tcb) M /\
             (WF_Rdy_Thrd Psi'0 M tcb \/ WF_Rdy_Thrd_IH Psi'0 M tcb))).
  apply H53; auto.
  rewrite H8. unfold addThrd.
  unfold WQMod.in_dom, WQMod.put.
  induction (WQMod.beq_A tid0 tid').
  rewrite a in H76. 
  absurd (~ WQMod.in_dom tid' RdyQ'); auto with PMap.
  auto.
  InvertAll.
  exists x.
  split. unfold T'.
  unfold TQMod.put, TQMod.maps_to.
  induction (TQMod.beq_A tid0 tid).
  absurd (tid0 <> tid); auto. 
  unfold TQMod.maps_to in H79. auto. 
  split; auto.
  exists x0.
  split; auto.
  unfold TM'.
  unfold ThrdMemMod.put at 1. unfold ThrdMemMod.maps_to.
  induction (ThrdMemMod.beq_A tid0 tid). absurd (tid0 <> tid); auto.
  unfold ThrdMemMod.put.
  induction (ThrdMemMod.beq_A tid0 tid').
  rewrite a in H76.
  absurd (~ WQMod.in_dom tid' RdyQ'); auto with PMap.
  unfold ThrdMemMod.maps_to in H82. auto.
  unfold WF_WQSet.
  intros.
  assert (l = w \/ l <> w). apply nat_eq_neq_dec.
  inversion H78.
  (* l = w *)
  clear H78. rewrite H79 in H76.
  unfold B' in H76.
  unfold WQSetMod.maps_to, WQSetMod.put in H76.
  induction (WQSetMod.beq_A w w).
  inversion H76. rewrite <- H80 in H77.
  rewrite H79.
  clear H80 H79 H76 a Q l.
  unfold BlkQ' in H77.
  assert (tid0 = tid \/ tid0 <> tid). apply nat_eq_neq_dec.
  inversion_clear H76.

  (* tid0 = tid *)
  exists (R' SP, BLOCKED, pc + 1).
  exists cthM_old.
  rewrite H78.
  split. unfold T'.
  unfold TQMod.put, TQMod.maps_to.
  induction (TQMod.beq_A tid tid). auto. absurd (tid = tid); auto.
  split. unfold TM'.
  unfold ThrdMemMod.put at 1. unfold ThrdMemMod.maps_to.
  induction (ThrdMemMod.beq_A tid tid); auto.
  absurd (tid = tid); auto.
  unfold WF_Wait_Thrd.
  split; auto.
  split.

  assert ((R' SP |-> pc + 1 *^* R' SP + 1 |-> junk1 *^* (R' SP + 2 |-> w *^*
         R' SP + 3 |-> junk2 *^* stkSpacePrec (R' SP) _BLK_Stk)) cthM_oldStk).
  sep on.
  sep matchH H70 with 5.
  sep on.
  assert (cthM_oldStk ||- R' SP |->w pc + 1 *^*
                         (R' SP + 2 |-> w *^* R' SP + 3 |-> junk2 *^*
                           stkSpacePrec (R' SP) _BLK_Stk)).
  apply ptoW_intro1 with junk1; auto.
  assert (cthM_old ||- mTrue *^* 
                           (R' SP |->w pc + 1 *^*
                             (R' SP + 2 |-> w *^* R' SP + 3 |-> junk2 *^*
                               stkSpacePrec (R' SP) _BLK_Stk))).
  exists (MemMod.merge cthM0 cthM2). exists cthM_oldStk.
  split; auto. split_r. unfold mTrue. auto.
  sep lift 1 in H80. sep trivial.

  exists wqa. split; auto.
  exists p'. exists g'. exists tag.
  split.
  apply psi_sw_lookupNPC with psi Psi Psi'; auto.
  intros.
  unfold pm_mwand. unfold S_Mem.
  intros.
  unfold itp_NH_theta.
  split.
  UnfoldStateAll. auto.
  split.

  apply enter_BLK_grt2_auxc with cthM1 cthM2 cthM R fl isr cthM11 cthM12 INVM cthM0 cthM_oldStk pc
        junk1 junk2 M0 p M M' R' fl' isr' w wqa; auto.

  apply enter_BLK_grt2_auxc_hyp6 with cthM11' (R' SP) pc junk1; auto.
  assert (R SP - 2 = R' SP).
  eapply enter_SW_grt_auxb0_hyp0; eauto.
  assert (R SP >= 2). apply ge_trans with (_BLK_Stk + 2); auto with arith.
  unfold mstar in H30. InvertAll. apply stkSpacePrec_sp_ge_sz with x0; auto.
  rewrite <- H82. assert (R SP - 2 + 2 = R SP). auto.
  auto.

  assert (exists w : nat,
          exists wqa : WQassert,
            enable_blk w wqa (cthM1, R, fl, isr) /\
            (forall pc' : nat, pimp (wrap_fun (gblk wqa) pc _BLOCK pc' (cthM1, R, fl, isr)) p') /\
            (forall pc' : nat,
             pimp (wrap_fun (gblk wqa) pc _BLOCK pc' (cthM1, R, fl, isr) |> g') (g (cthM1, R, fl, isr)))).
  apply H32; auto. InvertAll.
  assert (w = x /\ wqa = x0). apply enable_blk_unique_args with (cthM1, R, fl, isr); auto.
  inversion_clear H84. rewrite H87. auto.

(* WFST *)
  intro.
  assert (NH_WFST rho_ready Psi'0 (NH_ItpG g (cthM, R, fl, isr))).
  apply H26. auto.
  clear H26. clear H82.
  apply NH_WFST_strengthen with (cthM, R, fl, isr) g; auto.
  unfold pimp.
  intros.
  unfold NH_ItpG.
  unfold NH_ItpG in H26.
  InvertAll. rename x into M'20. rename x0 into M'21.
  rename x1 into SM1. rename x2 into SM2.
  exists cthM1. exists cthM2. exists SM1. exists SM2.
  split; auto.
  split; auto.
  split. unfold S_Flag. auto.
  split. auto.
  assert (exists w : nat,
          exists wqa : WQassert,
            enable_blk w wqa (cthM1, R, fl, isr) /\
            (forall pc' : nat, pimp (wrap_fun (gblk wqa) pc _BLOCK pc' (cthM1, R, fl, isr)) p') /\
            (forall pc' : nat,
             pimp (wrap_fun (gblk wqa) pc _BLOCK pc' (cthM1, R, fl, isr) |> g') (g (cthM1, R, fl, isr)))).
  apply H32; auto. InvertAll.
  assert (w = x /\ wqa = x0). apply enable_blk_unique_args with (cthM1, R, fl, isr); auto.
  inversion_clear H89. rewrite <- H92 in H90. clear H92 H91 H86 H88 x x0.

  unfold pimp in H90. apply H90 with pc'.

  unfold linkPG.
  exists (S_Mem_Upd (S_Mem_Upd (cthM_old, R0, fl0, false) M'2) M'20).
  split; auto.
  assert (M'20 = MemMod.merge (MemMod.merge cthM0 M0) cthM_oldStk).
  apply enter_SW_grt2_hype with cthM_old M'2 cthM2 M'21 fl fl0; auto.
  assert (S_Mem_Upd (S_Mem_Upd (cthM_old, R0, fl0, false) M'2) M'20 = (M'20, R0, fl0, false)).
  UnfoldStateAll. auto. rewrite H88. clear H88.
  rewrite H86.
  apply enter_BLK_grt2_auxc_aux4 with M M' R' fl' isr' cthM11 cthM12 INVM w junk1 junk2 cthM2 cthM_old M'2; auto.

  apply enter_BLK_grt2_auxc_hyp6 with cthM11' (R' SP) pc junk1; auto.
  assert (R SP - 2 = R' SP).
  eapply enter_SW_grt_auxb0_hyp0; eauto.
  assert (R SP >= 2). apply ge_trans with (_BLK_Stk + 2); auto with arith.
  unfold mstar in H30. InvertAll. apply stkSpacePrec_sp_ge_sz with x0; auto.
  rewrite <- H88. assert (R SP - 2 + 2 = R SP). auto.
  auto.
  (* end of WFST *)

  assert (WQMod.in_dom tid0 BlkQ).
  unfold addThrd in H77.
  unfold WQMod.put, WQMod.in_dom in H77.
  induction (WQMod.beq_A tid0 tid). contradiction. auto.
  unfold WF_WQSet in H22.
  assert (exists tcb : Thread,
          exists M : Mem,
            TQMod.maps_to T tid0 tcb /\
            ThrdMemMod.maps_to TM tid0 M /\ WF_Wait_Thrd Psi'0 M tcb w).
  apply H22 with BlkQ; auto.
  InvertAll. exists x. exists x0.
  split. unfold T'. unfold TQMod.maps_to, TQMod.put.
  induction (TQMod.beq_A tid0 tid). 
  contradiction. auto.
  split; auto. unfold TM'.
  unfold ThrdMemMod.put at 1. unfold ThrdMemMod.maps_to.
  induction (ThrdMemMod.beq_A tid0 tid). contradiction.
  unfold ThrdMemMod.put. induction (ThrdMemMod.beq_A tid0 tid').
  assert (WQMod.not_in_dom tid' BlkQ).
  apply enter_BLK_grt2_hyp8 with T RdyQ RdyQ' B w; auto.
  rewrite <- a in H81.
  absurd (~ WQMod.in_dom tid0 BlkQ); auto with PMap. auto.
  absurd (w = w); auto. 
  clear H78.
  assert (WQSetMod.maps_to B l Q).
  unfold B' in H76.
  unfold WQSetMod.maps_to, WQSetMod.put in H76.
  induction (WQSetMod.beq_A l w). contradiction.
  auto.
  unfold WF_WQSet in H22.
  assert (exists tcb : Thread,
          exists M : Mem,
            TQMod.maps_to T tid0 tcb /\
            ThrdMemMod.maps_to TM tid0 M /\ WF_Wait_Thrd Psi'0 M tcb l).
  apply H22 with Q; auto.
  InvertAll. exists x. exists x0.
  assert (tid0 = tid \/ tid0 <> tid).
  apply nat_eq_neq_dec.
  inversion_clear H82.
  assert (WQMod.not_in_dom tid Q).
  apply enter_BLK_grt2_hyp9 with T B RdyQ l rho_ready; auto.
  rewrite <- H84 in H82.
  absurd (~ WQMod.in_dom tid0 Q); auto with PMap.
  split. unfold T'.
  unfold TQMod.maps_to, TQMod.put.
  induction (TQMod.beq_A tid0 tid). contradiction.
  auto.
  split; auto.
  unfold TM'.
  unfold ThrdMemMod.put at 1. unfold ThrdMemMod.maps_to.
  induction (ThrdMemMod.beq_A tid0 tid). contradiction.
  unfold ThrdMemMod.put.
  induction (ThrdMemMod.beq_A tid0 tid').
  assert (WQMod.not_in_dom tid' Q).
  apply enter_BLK_grt2_hyp8 with T RdyQ RdyQ' B l; auto.
  rewrite <- a in H82.
  absurd (~ WQMod.in_dom tid0 Q); auto with PMap. auto.
  apply enter_BLK_grt2_hypa with M21' M22 T B RdyQ BlkQ BlkQ' tid w (R' SP, BLOCKED, pc+1); auto.
  absurd (_BLOCK <> _UNBLK). auto. auto.
  inversion_clear H.
  absurd (prim_entry _BLOCK); auto.
Qed.


Lemma enter_BLK:
forall D psi pc theta D' Psi Psi' Psi'0 a S S',
  NH_WFcmd psi pc theta (comml (_call _BLOCK))
  -> Minimal_D D
  -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
  -> lift_psi psi rho_ready Psi
  -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
  -> a = Itp_AIM rho_ready theta
  -> subsetCT Psi' Psi'0
  -> a Psi'0 S
  -> nextS pc (comml (_call _BLOCK)) S S'
  -> Itp0 D' BLK_Spec0 Psi'0 S'.
Proof.
  intros.
  unfold Itp0.
  unfold BLK_Spec0.
  rewrite H1.
  simpl.
  rewrite H4 in H6.
  clear H4 a.
  split.
  apply enter_BLK_pre with psi pc theta Psi'0 S; auto.
  split.
  apply enter_BLK_disable with psi pc theta Psi'0 S; auto.
  exists 0.
  apply enter_BLK_grt with psi pc theta D' Psi Psi' S; auto.
Qed.

Lemma NH_BLK_Presv:
  forall D psi pc theta D' Psi Psi' Psi'0 a S S' f',
    NH_WFcmd psi pc theta (comml (_call _BLOCK))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_ready Psi
    -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
    -> a = Itp_AIM rho_ready theta
    -> subsetCT Psi' Psi'0
    -> a Psi'0 S
    -> nextS pc (comml (_call _BLOCK)) S S'
    -> nextPC (comml (_call _BLOCK)) (S_Mem S) (S_RF S) (S_Flag S) pc f'
    -> exists theta0 : Spec0, lookupCT Psi'0 f' theta0 /\ Itp0 D' theta0 Psi'0 S'.
Proof.
  intros.
  assert (lookupCT Psi'0 _BLOCK BLK_Spec0).
  apply subsetCT_presv_lookupCT with Psi'; auto.
  rewrite H3.
  unfold ext_IH_Spec, ext_Prim_Specs. 
  apply lookupCT_subsetCT with (consCT _BLOCK BLK_Spec0 Psi). 
  apply lookupCT_consCT.
  apply Psi_extend_subsetCT.
  apply Psi_extend_subsetCT.
  apply Psi_extend2_subsetCT.
  apply consCT_subsetCT.
  inversion H8. clear H14 H13 H12 H10 H11 fl R M f pc0.
  exists BLK_Spec0.
  split; auto.
  apply enter_BLK with D psi pc theta Psi Psi' a S; auto.
Qed.

Lemma call_blk_disable_itr:
forall psi Psi'0 pc theta S,
  NH_WFcmd psi pc theta (comml (_call _BLOCK)) 
  -> Itp_AIM rho_ready theta Psi'0 S
  -> ~ itr_enabled S.
Proof.
  intros.
  intro.
  unfold NH_WFcmd in H.
  induction theta.
  destruct H as [HWFcmd [Htmp1 Htmp2]].
  clear Htmp1 Htmp2.
  unfold NH_WFcmd_aux in HWFcmd.
  inversion_clear HWFcmd.
  destruct H as [Hprim HWFcmd].
  clear Hprim.
  unfold NH_Call_Prim in HWFcmd.
  inversion_clear HWFcmd.
  clear H.
  assert (S_Flag S _IF = false).
  assert (imprec (NH_pinvstar p) S).
  apply Itp_AIM_pinv with g n Psi'0 rho_ready. auto.
  discriminate.
  unfold imprec in H. unfold pmstar in H.
  InvertAll.
  induction S as (((M, R), fl), isr).
  simplS. simplS in H.
  unfold NH_pinvstar in H.
  unfold pmstar in H.
  InvertAll. simplS in H.
  assert (S_Flag (x1, R, fl, isr) _IF = false).
  apply H2; auto.
  UnfoldStateAll; auto.
  unfold itr_enabled in H1.
  InvertAll.
  rewrite H3 in H. inversion H.

  inversion_clear H.
  assert (prim_entry _BLOCK); auto.
Qed.



Lemma NH_BLK_Itp :
  forall D psi pc theta D' Psi Psi',
    NH_WFcmd psi pc theta (comml (_call _BLOCK))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_ready Psi
    -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
    (* -> a = Itp_AIM rho_ready theta *)
    -> WFcmd D' Psi' (packS0 rho_ready theta) pc (comml (_call _BLOCK)).
Proof.
  intros.
  unfold WFcmd.
  split.
  unfold WFcmd_aux.
  intros.
  assert (Itp_AIM rho_ready theta Psi'0 S).
  unfold Itp0 in H4.
  rewrite H1 in H4.
  unfold ext_dict in H4.
  assert (beq_nat rho_ready rho_prim = false). simpl. auto.
  rewrite H6 in H4.
  unfold Minimal_D in H0.
  InvertAll.
  rewrite H7 in H4. auto.
  split.
  apply NH_BLK_progress with psi theta Psi'0 (Itp_AIM rho_ready theta); auto.
  split.
  intros.

  apply NH_BLK_Presv with D psi pc theta Psi Psi' (Itp_AIM rho_ready theta) S; auto.
  left.


  apply call_blk_disable_itr with psi Psi'0 pc theta; auto.
  rewrite H1.  
  unfold Itp0.
  unfold ext_dict.
  assert (beq_nat rho_ready rho_prim = false).
  simpl. auto.
  rewrite H4.
  unfold Minimal_D in H0. InvertAll.
  rewrite H5.
  apply Itp_AIM_Mono_CT.
Qed.
