Add LoadPath "..".
Add LoadPath "../CAP0".

Require Import Arith.
Require Import Coqlib.
Require Import Map.
Require Import Machine.
Require Import Assertion.
Require Import Assertion_Lemmas.
Require Import Assertion_Lemmas4.
Require Import Operational_Semantics.
Require Import OpSem_Lemmas.
Require Import OpSem_Lemmas2.
Require Import Reg.
Require Import CAP0.
Require Import SCAP.
Require Import AIM.
Require Import QLib.
Require Import SCAP_READY.
Require Import SCAP_IDLE.
Require Import SCAP_Itr.
Require Import Linking.
Require Import Linking_Lemmas.
Require Import Linking_Lemmas2.
Require Import Linking_Lemmas3.

Require Import EqNat.
Require Import Omega.
Require Import List.
Require Import Arith_Lemmas.


Lemma NH_SW_progress:
  forall psi pc theta Psi a S,
    NH_WFcmd psi pc theta (comml (_call _SWITCH))
    -> a = Itp_AIM rho_ready theta
    -> a Psi S
    -> exists S', nextS pc (comml (_call _SWITCH)) S S'.
Proof.
  intros.
  unfold NH_WFcmd in H.
  induction theta.
  unfold NH_WFcmd_aux in H.
  inversion_clear H.
  inversion H2.
  inversion_clear H.
  apply NH_SW_prog_aux with psi (Ast p g n) Psi rho_ready; auto.
  rewrite H0 in H1. auto.
  inversion_clear H.
  unfold prim_entry in H4.
  assert (_SWITCH = _SWITCH \/ _SWITCH = _BLOCK \/ _SWITCH = _UNBLK).
  left. auto. contradiction.
Qed.



Lemma enter_SW_preMem:
forall pc S S' p g tag M1 M2 Psi T B RdyQ tid,
  nextS pc (comml (_call _SWITCH)) S S'
  -> enable_sw p
  -> MemMod.disjUnion M1 M2 (S_Mem S)
  -> WF_AIM_World Psi (S_Mem_Upd S M1) T B RdyQ tid
         rho_ready (Ast p g tag)
  -> (embed_curr_TCB T tid *^* (embed_RdyQ T RdyQ *^* embed_WQSet T B)) M2
  -> exists junk, (SW_MPre T RdyQ tid (S_RF S' SP) (pc+1) junk *^* mTrue) (S_Mem S').
Proof.
  intros.
  assert (imprec (NH_pinvstar p) (S_Mem_Upd S M1)).
  apply WF_AIM_World_pinv with Psi T B RdyQ tid rho_ready g tag; auto.
  assert (imprec p (S_Mem_Upd S M1)). 
  apply imprec_presv_imprec with (NH_pinvstar p); auto.
  unfold SW_MPre.
  assert (exists M1',
        MemMod.disjUnion M1' M2 (S_Mem S')
        /\ (S_RF S' SP |->w (pc+1) *^* stkSpacePrec (S_RF S' SP) 8 *^* mTrue) M1').
  apply enable_sw_nxt_stkspace with p S M1; auto.
  InvertAll.
  rename x into M1'.
  assert (exists junk, (S_RF S' SP |-> pc + 1 *^* S_RF S' SP + 1 |-> junk *^* stkSpacePrec (S_RF S' SP) 8 *^* mTrue) M1').
  sep on.
  sep lift 0 in H8.
  assert (exists junk,  
           M1' ||- S_RF S' SP |-> pc + 1 *^* S_RF S' SP + 1 |-> junk *^* (stkSpacePrec (S_RF S' SP) 8 *^* mTrue)).
  apply ptoW_elim1; auto.
  inversion H7. exists x. sep on.
  sep matchH H9 with 4.
  inversion_clear H7. clear H8. exists x.
  apply merge_SW_pre with M1' M2 B; auto.
Qed.




Lemma enter_SW_pre:
forall psi pc theta Psi'0 S S',
  NH_WFcmd psi pc theta (comml (_call _SWITCH))
  -> Itp_AIM rho_ready theta Psi'0 S
  -> nextS pc (comml (_call _SWITCH)) S S'
  -> SW_Prd S'.
Proof.
  intros.
  unfold SW_Prd.
  induction theta.
  unfold Itp_AIM in H0.
  InvertAll. rename x into T. rename x0 into B. rename x1 into RdyQ. rename x2 into tid.
  exists T. exists RdyQ. exists tid.
  rename n into tag.
  unfold NH_WFcmd in H.
  unfold NH_WFcmd_aux in H.
  InvertAll.
  inversion_clear H2.
  inversion_clear H3. clear H2.
  unfold NH_Call_Prim in H5.
  inversion_clear H5.
  inversion H2.
  inversion H6. rewrite <- H12 in H8.
  clear H9 H10 H12 H13 H14 H5 H6 H7 tag0 p0 g0 p' g'.
  induction S as (((M, R), fl), isr).
  exists (S_RF S' SP).
  unfold pmstar in H0.
  InvertAll.
  rename x into M1. rename x0 into M2.
  assert (exists tcb, TQMod.maps_to T tid tcb).
  apply WF_AIM_World_currTCB with Psi'0 (S_Mem_Upd (M, R, fl, isr) M1) B RdyQ rho_ready (Ast p g tag); auto.
  inversion_clear H6.
  rename x into tcb. exists tcb. exists (pc+1).
  split.
  apply enter_SW_preMem with (M, R, fl, isr) p g tag M1 M2 Psi'0 B; auto.
  split; auto.
  split; auto.
  apply WF_AIM_World_RdyQinT with Psi'0 (S_Mem_Upd (M, R, fl, isr) M1) B tid rho_ready (Ast p g tag); auto.
  split; auto.
  split.
  apply WF_AIM_World_cthNotRdy with Psi'0 (S_Mem_Upd (M, R, fl, isr) M1) T B rho_ready (Ast p g tag); auto.
  apply WF_AIM_World_WFTQ with Psi'0 (S_Mem_Upd (M, R, fl, isr) M1) B RdyQ tid rho_ready (Ast p g tag); auto.
  assert (_SWITCH <> _BLOCK). auto.
  contradiction.
  assert (_SWITCH <> _UNBLK). auto. contradiction.
  assert (prim_entry _SWITCH).
  unfold prim_entry. left. auto.
  inversion_clear H3. contradiction.
Qed.



Lemma enter_SW_disable:
forall psi pc theta Psi'0 S S',
  NH_WFcmd psi pc theta (comml (_call _SWITCH))
  -> Itp_AIM rho_ready theta Psi'0 S
  -> nextS pc (comml (_call _SWITCH)) S S'
  -> S_Flag S' _IF = false.
Proof.
  intros.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  unfold NH_WFcmd in H.
  induction theta.
  destruct H as [HWFcmd [Htmp1 Htmp2]].
  clear Htmp1 Htmp2.
  unfold NH_WFcmd_aux in HWFcmd.
  inversion_clear HWFcmd.
  destruct H as [Hprim HWFcmd].
  clear Hprim.
  unfold NH_Call_Prim in HWFcmd.
  inversion_clear HWFcmd.
  clear H.
  assert (fl _IF = false).
  assert (imprec (NH_pinvstar p) (M, R, fl, isr)).
  apply Itp_AIM_pinv with g n Psi'0 rho_ready. auto.
  discriminate.
  unfold imprec in H. unfold pmstar in H.
  InvertAll.
  unfold NH_pinvstar in H.
  simplS in H.
  unfold pmstar in H.
  InvertAll. simplS in H.
  assert (S_Flag (x1, R, fl, isr) _IF = false).
  apply H2; auto.
  UnfoldStateAll; auto. clear H0 H2.
  destruct (nextS_inversion H1).
  InvertAll.
  subst fl'.
  UnfoldStateAll. auto.

  inversion_clear H.
  assert (prim_entry _SWITCH). unfold prim_entry.
  left. auto.
  contradiction.
Qed.


Lemma S_Mem_Upd_discard_interm2 :
forall S M1 M2,
  (S_Mem_Upd (S_Mem_Upd S M2) M1) = (S_Mem_Upd S M1).
Proof.
  induction S as (((M, R) , fl),isr).
  intros.
  UnfoldStateAll. auto.
Qed.
Hint Resolve S_Mem_Upd_discard_interm2.



Lemma WFcallsw_enablesw:
forall psi pc p g tag,
  NH_WFcmd psi pc (Ast p g tag) (comml (_call _SWITCH))
  -> enable_sw p.
Proof.
  intros.
  unfold NH_WFcmd in H.
  InvertAll.
  unfold NH_WFcmd_aux in H0.
  inversion_clear H0.
  inversion_clear H1.
  inversion_clear H3.
  inversion H1. inversion H5.
  auto. 
  assert (_SWITCH <> _BLOCK). auto. contradiction.
  assert (_SWITCH <> _UNBLK). auto. contradiction.
  assert (prim_entry _SWITCH). unfold prim_entry. left. auto.
  inversion H1. contradiction.
Qed.

Lemma call_sw_nextPC_spec:
forall psi pc p g tag,
    NH_WFcmd psi pc (Ast p g tag) (comml (_call _SWITCH))
    -> exists p', exists g',
         SCAP_lookupCT psi (pc+1) (Ast p' g' tag)
         /\ (forall pc', pimp (p |> (wrap_fun gy pc _SWITCH pc')) p')
         /\ (forall pc', gimp (p >o (wrap_fun gy pc _SWITCH pc') oo g') g).
Proof.
  intros.
  unfold NH_WFcmd in H. inversion_clear H.
  unfold NH_WFcmd_aux in H0.
  inversion_clear H0.
  inversion_clear H.
  inversion_clear H2. inversion H.
  exists p'. exists g'.
  inversion H4.
  split; auto.
  absurd (_SWITCH <> _BLOCK); auto.
  absurd (_SWITCH <> _UNBLK); auto.
  inversion_clear H.
  absurd (prim_entry _SWITCH); auto.
  unfold prim_entry. left. auto.
Qed.
Hint Resolve call_sw_nextPC_spec.


Lemma minimalD_containsRdy:
  forall D, Minimal_D D -> D rho_ready = SomeT1 _ (Itp_AIM rho_ready).
Proof.
  intros.
  unfold Minimal_D in H.
  InvertAll. auto.
Qed.
Hint Resolve minimalD_containsRdy.


Lemma call_sw_flags_false:
forall p g tag psi Psi pc M M' R fl isr T tid,
  WF_Curr_Thrd Psi (M, R, fl, isr) rho_ready (Ast p g tag) T tid
  -> NH_WFcmd psi pc (Ast p g tag) (comml (_call _SWITCH))
  -> p (M', R, fl, isr)
  -> (fl _IF = false) /\ (isr = false).
Proof.
  intros.
  unfold NH_WFcmd in H0.
  InvertAll. unfold NH_WFcmd_aux in H2.
  inversion_clear H2.
  inversion_clear H3. inversion_clear H5.
  assert (S_Flag (M', R, fl, isr) _IF = false).
  apply H6; auto.
  unfold S_Flag in H5. split; auto.
  unfold WF_Curr_Thrd in H.
  assert (beq_nat rho_ready rho_ih = false).
  apply neq_beq_false. discriminate. rewrite H7 in H.
  unfold itp_NH_theta in H.
  unfold S_ISR in H. InvertAll. auto.
  InvertAll.
  absurd (prim_entry _SWITCH); auto.
  unfold prim_entry. left. auto.
Qed.




Lemma enter_SW_grt2_hyp2:
forall T T' B RdyQ' tid tid' sp sp' tcb tcb' pc pc' S0 S' M0' M2' M22 M12 cthM2 cthM12 cthMR TM junk,
  SW_Post_aux T' RdyQ' tid' sp' sp tcb' pc junk (S_Mem_Upd S0 M0')
  -> nextS pc' (comms _retw) S0 S'
  -> MemMod.disjUnion M0' M2' (S_Mem S0)
  -> disjUnionList (M22 :: M12 :: cthM2 :: cthM12 :: nil) M2'
  -> embed_WQSet T B M22
  -> MemSplit (ThrdMemMod.remove TM tid) M12
  -> ThrdMemMod.maps_to TM tid' cthMR
  -> tid <> tid'
  -> NH_Inv (S_Flag S0) cthM2
  -> (INV0 *^* INV1 *^* mTrue) cthM12
  -> T' = TQMod.put T tid (sp, TCB_Status tcb, pc)
  -> exists INVM, exists cthM0, exists M21', exists M122, exists cthM_old, exists cthM_oldStk,
        MemMod.disjUnion INVM cthM0 cthM12 
        /\ (INV0 *^* INV1) INVM
        /\ (embed_curr_TCB T' tid' *^* embed_RdyQ T' RdyQ') M21'
        /\ MemMod.disjUnion M122 cthMR M12
        /\ MemSplit (ThrdMemMod.remove (ThrdMemMod.remove TM tid) tid') M122
        /\ MemMod.disjUnion (MemMod.merge cthM0 cthM2) cthM_oldStk cthM_old
        /\ (sp |-> pc *^* sp + 1 |-> junk *^* stkSpacePrec sp 8) cthM_oldStk
        /\ disjUnionList (M22 :: cthMR :: INVM :: M21' :: cthM_old :: M122 :: nil) (S_Mem S').
Proof.
  intros.
  unfold mstar at 1 in H8. InvertAll.
  rename x into INVM. rename x0 into cthM0. exists INVM. exists cthM0.
  unfold SW_Post_aux in H.
  InvertAll.
  unfold SW_MPost in H11.
  assert ((embed_curr_TCB T' tid' *^* embed_RdyQ T' RdyQ' *^* (sp |-> pc *^* sp + 1 |-> junk *^*
           stkSpacePrec sp 8)) (S_Mem (S_Mem_Upd S0 M0'))).
  sep on.
  sep matchH H11 with 4.
  clear H11.
  unfold mstar at 1 in H15.
  InvertAll.
  rename x into M21'. exists M21'.
  rename x0 into cthM_oldStk.
  assert (exists M122, MemMod.disjUnion M122 cthMR M12 /\
                        MemSplit (ThrdMemMod.remove (ThrdMemMod.remove TM tid) tid') M122).
  apply memsplit_thrd_remove; auto.
  unfold ThrdMemMod.maps_to. unfold ThrdMemMod.remove.
  induction (ThrdMemMod.beq_A tid' tid).
  absurd (tid' <> tid); auto. unfold ThrdMemMod.maps_to in H5. auto.
  InvertAll. rename x into M122.
  exists M122.
  exists (MemMod.merge (MemMod.merge cthM0 cthM2) cthM_oldStk).
  exists cthM_oldStk.
  split; auto. split; auto. split; auto.
  split; auto. split; auto.
  split. unfold MemMod.disjUnion. split; auto.
  apply enter_SW_grt2_aux2_hyp0 with INVM cthM12 M22 M12 M0' M2' S0 M21'; auto.
  split. auto.
  apply enter_SW_grt2_aux2_hyp1 with pc' M0' M2' S0 M12 cthM12; auto.
Qed.




Lemma enter_SW_grt2_auxc_aux4:
forall pc pc' M R fl isr M' R' fl' isr' cthM1 cthM11 cthM12 INVM cthM0 M0 cthM_oldStk R0 fl0 junk cthM2 cthM_old M'2,
  nextS pc (comml (_call _SWITCH)) (M, R, fl, isr) (M', R', fl', isr')
  -> MemMod.disjUnion cthM11 cthM12 cthM1
  -> stkSpacePrec (R SP) (8 + 2) cthM11
  -> MemMod.disjUnion INVM cthM0 cthM12
  -> (R' SP |-> pc + 1 *^* R' SP + 1 |-> junk *^* stkSpacePrec (R' SP) 8) cthM_oldStk
  -> (INV0 *^* INV1) INVM
  -> (INV0 *^* INV1) M0
  -> MemMod.disjUnion M0 cthM_old M'2
  -> MemMod.disjUnion (MemMod.merge cthM0 cthM2) cthM_oldStk cthM_old
  -> fl0 _IF = false
  -> fl _IF = false
  -> isr = false
  -> R0 SP = R' SP + 2
  -> wrap_fun gy pc _SWITCH pc' (cthM1, R, fl, isr)
     (MemMod.merge (MemMod.merge cthM0 M0) cthM_oldStk, R0, fl0, false).
Proof.
  intros.
  unfold wrap_fun.
  unfold conjG.  
  exists (MemMod.merge (MemMod.merge cthM0 M0) cthM_oldStk, RF_update R0 SP (R0 SP - 2), fl0, false).
  split.

  assert (nextS pc (comml (_call _SWITCH)) (cthM1, R, fl, isr) (MemUpdW cthM1 (R SP - 2) (pc+1) junk, RF_update R SP (R SP - 2), fl, isr)).
  apply nextS_call.
  unfold pushW.
  split.
  assert ((mTrue *^* stkSpacePrec (R SP) (8 + 2)) cthM1).
  exists cthM12. exists cthM11. split; auto with PMap.
  unfold mTrue. split; auto.
  sep on. sep lift 1 in H12.
  sep stksplit 2 in H12.
  sep lift 0 in H12. sep trivial.
  split; auto.
  apply MemUpdW_MemUpd16 with junk; auto.
  set (cthM' := MemUpdW cthM1 (R SP - 2) (pc + 1) junk) in *.
  exists (cthM', RF_update R SP (R SP - 2), fl, isr).
  split; auto.
  assert (R SP - 2 = R' SP).
  apply enter_SW_grt_auxb0_hyp0 with pc M fl isr M' fl' isr' _SWITCH; auto.
  unfold gy.
  exists (R SP - 2). simplS. reg simpl.

  assert (R SP = R' SP + 2).
  rewrite <- H13.
  assert (R SP >= 2).
  apply ge_trans with (8 + 2).
  sep on.
  apply stkSpacePrec_sp_ge_sz with cthM11; auto.
  auto with arith.
  assert (R SP - 2 + 2 = R SP). auto. auto. rename H14 into HRSP'.

  split.

  unfold prepost. simplS.
  sep on.
  sep stksplit in H1.
  destruct H1 as [cthM11_1 [cthM11_2 [HcthM11split [HcthM11_1 HcthM11_2]]]].

  exists (MemMod.merge INVM cthM11_1).
  exists (MemMod.merge (MemUpdW cthM11_2 (R SP - 2) (pc+1) junk) cthM0).
  sep lift 2 in H3.
  unfold mstar at 1 in H3.
  sep off.
  InvertAll.
  rename x into oldStk1. rename x0 into oldStk2.
  exists (MemMod.merge M0 oldStk1).

  split.
  exists INVM. exists cthM11_1. split; auto.
  split; auto.
  apply enter_BLK_grt2_auxc_hyp1 with cthM1 cthM11 cthM12 cthM11_2 cthM0; auto.

  split.
  exists M0. exists oldStk1.
  rewrite <- H13 in H3. split; auto.
  split; auto.
  apply enter_BLK_grt2_auxc_hyp2 with M'2 cthM_old (MemMod.merge cthM0 cthM2) cthM_oldStk oldStk2; auto.

  split. 
  assert (MemMod.disjUnion (MemMod.merge (MemUpdW cthM11_2 (R SP - 2) (pc + 1) junk) cthM0)
                             (MemMod.merge INVM cthM11_1) cthM').


  apply enter_SW_grt2_auxc_auxx with cthM1 cthM11 cthM12; auto.
  auto with PMap.

  assert (MemUpdW cthM11_2 (R SP - 2) (pc + 1) junk = oldStk2).
  rewrite H13 in HcthM11_2. rewrite H13.

  apply enter_SW_grt2_auxc_hypx; auto.

  rewrite H14.
  apply enter_BLK_grt2_auxc_hyp4 with M'2 cthM_old cthM2; auto.

  rewrite <- H11 in HRSP'. rewrite <- HRSP'. split; auto.
  rewrite <- H8 in H9. auto.

  apply nextS_retw. exists (pc+1).
  unfold popW.
  reg simpl.
  split.
  rewrite H11. assert (R' SP + 2 - 2 = R' SP). auto.
  rewrite H12.
  apply enter_BLK_grt2_auxc_hyp5 with M'2 cthM_old cthM2 junk
            (stkSpacePrec (R' SP) 8); auto.

  rewrite H11. assert (R' SP + 2 - 2 = R' SP). auto.
  rewrite H12. rewrite <- H11.
  assert (RF_update R0 SP (R0 SP) = R0).
  apply RF_update_prev. auto.
Qed.

Lemma enter_SW_grt2_auxc:
forall cthM1 cthM2 cthM R fl isr cthM11 cthM12 INVM cthM0 cthM_oldStk pc junk
            M0 M'2 cthM_old R0 fl0 (p p'': Pred) M M' R' fl' isr',
  MemMod.disjUnion cthM1 cthM2 cthM
  -> p (cthM1, R, fl, isr)
  -> NH_Inv fl cthM2
  -> MemMod.disjUnion cthM11 cthM12 cthM1
  -> stkSpacePrec (R SP) (8 + 2) cthM11
  -> MemMod.disjUnion INVM cthM0 cthM12
  -> (INV0 *^* INV1) INVM
  -> MemMod.disjUnion (MemMod.merge cthM0 cthM2) cthM_oldStk cthM_old
  -> (R' SP |-> pc + 1 *^* R' SP + 1 |-> junk *^* stkSpacePrec (R' SP) 8) cthM_oldStk
  -> (INV0 *^* INV1) M0
  -> MemMod.disjUnion M0 cthM_old M'2
  -> (forall pc' : nat, pimp (p |> wrap_fun gy pc _SWITCH pc') p'')
  -> fl0 _IF = false
  -> fl _IF = false
  -> nextS pc (comml (_call _SWITCH)) (M, R, fl, isr) (M', R', fl', isr')
  (* -> nextS pc' (comms _retw) S0 S' *)
  -> R0 SP = R' SP + 2
  -> isr = false
  -> NH_pinvstar p'' (S_Mem_Upd (cthM_old, R0, fl0, false) M'2).
Proof.
  intros.
  unfold NH_pinvstar.
  unfold pmstar.
  exists (MemMod.merge (MemMod.merge cthM0 M0) cthM_oldStk).
  exists cthM2.
  split.
  assert (S_Mem (S_Mem_Upd (cthM_old, R0, fl0, false) M'2) = M'2).
  auto.
  rewrite H16. clear H16.
  apply enter_SW_grt2_auxc_hyp0 with cthM_old; auto.
  assert (MemMod.disjoint cthM12 cthM2).
  apply MemMod.disj_merge_disj_right with cthM11.
  inversion_clear H2. subst cthM1. inversion_clear H. auto.
  apply MemMod.disj_merge_disj_right with INVM.
  inversion_clear H4. subst cthM12. auto.
  split.
  simplS.
  unfold pimp in H10.
  apply H10 with 0.
  unfold linkPG.
  exists (cthM1, R, fl, isr).
  split; auto.

  apply enter_SW_grt2_auxc_aux4 with M M' R' fl' isr' cthM11 cthM12 INVM junk cthM2 cthM_old M'2; auto.
  assert (S_Flag (S_Mem_Upd (cthM_old, R0, fl0, false) M'2)
             = fl0).
  UnfoldStateAll. auto.
  rewrite H16.
  unfold NH_Inv. rewrite H11.
  unfold NH_Inv in H1. rewrite H12 in H1. auto.
Qed.

(* >>>>>>>>>>>>>>>> *)
Lemma enter_SW_grt:
forall D psi pc theta D' Psi Psi' Psi'0 S S',
  NH_WFcmd psi pc theta (comml (_call _SWITCH))
  -> Minimal_D D
  -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
  -> lift_psi psi rho_ready Psi
  -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
  -> subsetCT Psi' Psi'0
  -> Itp_AIM rho_ready theta Psi'0 S
  -> nextS pc (comml (_call _SWITCH)) S S'
  -> WFST D rho_prim 0 Psi'0 (scap_lift_g SW_grt S').
Proof.
  intros.
  unfold WFST.
  intros.
  unfold scap_lift_g in H7.
  InvertAll.
  induction S' as (((M', R'), fl'), isr').
  induction S as (((M, R), fl), isr).
  induction theta.
  rename n into tag.
  unfold Itp_AIM in H5.
  unfold pmstar in H5.
  InvertAll.
  rename x into T. rename x0 into B. rename x1 into RdyQ. rename x2 into tid.
  rename x3 into M1. rename x4 into M2.
  assert ((embed_curr_TCB T tid *^* embed_RdyQ T RdyQ *^* embed_WQSet T B) M2).
  apply leqvM_app_subst with (embed_curr_TCB T tid *^* (embed_RdyQ T RdyQ *^* embed_WQSet T B)); auto.
  apply leqvM_commut.
  apply mstar_assoc.
  unfold mstar at 1 in H11. InvertAll.
  rename x into M21. rename x0 into M22. clear H12.

  assert (exists TM, exists cthM, exists M12, exists cthM1, exists cthM2,
        MemMod.disjUnion M12 cthM M1
        /\ MemMod.disjUnion cthM1 cthM2 cthM
        /\ WF_currTID tid T B RdyQ rho_ready
        /\ WF_TWQ T B RdyQ
        /\ exists_Idle T RdyQ tid
        /\ WF_WQ_WQSpec B Delta
        /\ MemSplit (ThrdMemMod.remove TM tid) M12
        /\ ThrdMemMod.maps_to TM tid cthM
        /\ WF_Curr_Thrd Psi'0 (S_Mem_Upd (M1, R, fl, isr) cthM) rho_ready (Ast p g tag) T tid
        /\ WF_ReadyQ Psi'0 T RdyQ TM
        /\ WF_WQSet Psi'0 T B TM
        /\ p (cthM1, R, fl, isr)
        /\ NH_Inv fl cthM2
        /\ (tag = 1 -> NH_WFST rho_ready Psi'0 (NH_ItpG g (cthM, R, fl, isr)))).
  apply decompose_World; auto.
  InvertAll. rename x into TM. rename x0 into cthM. rename x1 into M12. rename x2 into cthM1. rename x3 into cthM2.
  assert (WQMod.not_in_dom tid RdyQ). unfold WF_currTID in H16. InvertAll. auto.
  assert (exists M1, exists M2, 
            MemMod.disjUnion M1 M2 (S_Mem (cthM1, R, fl, isr))
         /\ stkSpacePrec (S_RF (cthM1, R, fl, isr) SP) (8 + 2) M1 
         /\ ((INV0 *^*INV1) *^* mTrue) M2).
  apply enable_sw_stkspace with p; auto.
  apply WFcallsw_enablesw with psi pc g tag; auto.
  InvertAll. rename x into cthM11. rename x0 into cthM12.
  unfold S_Mem in H30. unfold S_RF in H29.
  assert (disjUnionList (M2 :: M12 :: cthM2 :: cthM12 :: cthM11 :: nil) M).
  apply mem_list_compose_a with cthM1 cthM M1; auto.
  assert 
    (exists junk,
        (R' SP |-> (pc+1) *^* R' SP + 1 |-> junk *^* (stkSpacePrec (R' SP) 8)) (MemUpdW cthM11 (R' SP) (pc + 1) junk)
        /\ disjUnionList (M2 :: M12 :: cthM2 :: cthM12 :: (MemUpdW cthM11 (R' SP) (pc + 1) junk) :: nil) M').
  apply call_sw_frame with M R fl fl' isr isr'; auto.
  InvertAll. rename x into junk1.
  set (cthM11' := MemUpdW cthM11 (R' SP) (pc + 1) junk1) in *.
  assert (exists M1', exists M2', 
           MemMod.disjUnion cthM11' M21 M1' 
           /\ disjUnionList (M22 :: M12 :: cthM2 :: cthM12 :: nil) M2'
           /\ MemMod.disjUnion M1' M2' M').
  apply mem_list_compose_b with M2; auto.
  InvertAll.  rename x into M1'. rename x0 into M2'.
  assert (exists tcb, T tid = Some tcb).
  unfold WF_currTID in H16.
  InvertAll.
  unfold TQMod.in_dom in H39. auto.
  inversion_clear H37. rename x into tcb.
  assert (exists M0' : MemMod.Map,
         exists RdyQ' : WQ,
           exists tid' : TID,
             exists sp' : nat,
               exists tcb' : Thread,
                 MemMod.disjUnion M0' M2' (S_Mem S0) /\
                 addThrd RdyQ tid = addThrd RdyQ' tid' /\
                 SW_Post_aux (TQMod.put T tid ((R' SP), TCB_Status tcb, pc+1)) 
                              RdyQ' tid' sp' (R' SP) tcb' (pc+1) junk1 (S_Mem_Upd S0 M0')).
  apply H8 with M1' T tcb; auto.
  unfold SW_Prd_aux.
  split. assert (S_Mem (S_Mem_Upd (M', R', fl', isr') M1') = M1'). auto.
  rewrite H37.
  apply mem_compose_c with cthM11' M21; auto.
  split; auto.
  split. 
  apply WF_AIM_World_RdyQinT with Psi'0 (S_Mem_Upd (M, R, fl, isr) M1) B tid rho_ready (Ast p g tag).
  auto.
  split; auto. split; auto.
  apply WF_AIM_World_WFTQ with Psi'0 (S_Mem_Upd (M, R, fl, isr) M1) B RdyQ tid rho_ready (Ast p g tag).
  auto.
  clear H8.
  InvertAll.
  rename x into M0'. rename x0 into RdyQ'. rename x1 into tid'.
  rename x2 into sp'. rename x3 into tcb'.
  assert ((fl _IF = false) /\ (isr = false)).
  apply call_sw_flags_false with p g tag psi Psi'0 pc cthM cthM1 R T tid; auto.
  inversion_clear H40.
  assert (tid = tid' \/ tid <> tid').
  apply nat_eq_neq_dec.
  inversion_clear H40.

  (* tid = tid' *)
  (* clear H24 H27 H26 cthM11. *)
  assert (RdyQ = RdyQ').
  assert (WQMod.not_in_dom tid' RdyQ').
  unfold SW_Post_aux in H41. InvertAll. auto.
  apply unique_remain with tid tid'; auto.
  exists (pc + 1).
  rewrite <- H40 in H41.
  rewrite <- H44 in H41. clear H44 H40 H8 tid' RdyQ'.
  assert (S_RF S0 SP = R' SP).
  assert (sp' = R' SP).
  apply enter_SW_grt_aux60 with T RdyQ tid tcb (pc+1) junk1 tcb' S0 M0'; auto.
  unfold SW_Post_aux in H41. InvertAll.
  rewrite H8 in H45. auto.
  split.
  apply enter_SW_grt_aux7 with T RdyQ tid (R' SP) tcb junk1 sp' tcb' M0' M2'; auto.
  assert (exists p', exists g',
            SCAP_lookupCT psi (pc+1) (Ast p' g' tag)
         /\ (forall pc', pimp (p |> (wrap_fun gy pc _SWITCH pc')) p')
         /\ (forall pc', gimp (p >o (wrap_fun gy pc _SWITCH pc') oo g') g)).
  auto.
  InvertAll. rename x into p'. rename x0 into g'.
  exists (packS0 rho_ready (Ast p' g' tag)).
  split.
  apply psi_sw_lookupNPC with psi Psi Psi'; auto.

  intros.
  unfold Itp0.
  assert (D rho_ready = SomeT1 _ (Itp_AIM rho_ready)).
  auto.
  rewrite H47.
  unfold Itp_AIM.
  set (T' := TQMod.put T tid (R' SP, TCB_Status tcb, pc + 1)) in * |- *.
  exists T'. exists B. exists RdyQ. exists tid.

  assert (exists M21_0, exists cthM11'',
            ((S_RF S' SP - 2) |-> (pc+1) *^* (S_RF S' SP - 2 + 1) |-> junk1 *^* stkSpacePrec (S_RF S' SP - 2) 8) cthM11''
            /\ (embed_curr_TCB T' tid *^* embed_RdyQ T' RdyQ) M21_0
            /\ disjUnionList (M21_0 :: M22 :: M12 :: cthM2 :: cthM12 :: cthM11'' :: nil) (S_Mem S')).
  apply enter_SW_grt_aux8 with T (R' SP) sp' tcb tcb' pc' S0 M0' M2'; auto.
  InvertAll. rename x into M21_0. rename x0 into cthM11''.
  unfold pmstar.
  assert (exists M'1, exists M'2, 
            disjUnionList (M12 :: cthM2 :: cthM12 :: cthM11'' :: nil) M'1
            /\ MemMod.disjUnion M21_0 M22 M'2
            /\ MemMod.disjUnion M'1 M'2 (S_Mem S')).
  apply enter_SW_grt_hyp4.
  auto.
  InvertAll.
  rename x into M'1. rename x0 into M'2.
  exists M'1. exists M'2. split; auto.
  split.
  unfold WF_AIM_World.
  split. unfold T'.
  apply WF_currTID_ext_T with tcb; auto.
  split. unfold T'.
  apply exists_Idle_ext_T; auto.
  split. unfold T'.
  apply WF_TWQ_ext_T; auto.
  exists tcb. auto.
  split; auto.
  assert (exists cthM', MemMod.disjUnion M12 cthM' M'1 
                        /\ disjUnionList (cthM2 :: cthM12 :: cthM11'' :: nil) cthM').
  auto.
  InvertAll.
  rename x into cthM'.
  exists (ThrdMemMod.put TM tid cthM'). exists cthM'.
  split.
  assert (S_Mem (S_Mem_Upd S' M'1) = M'1). auto.
  rewrite H55.
  apply enter_SW_grt_hyp6 with M12; auto.
  split. auto with PMap.

  split.
  assert (S_ISR (M, R, fl, isr) = S_ISR S' /\ S_Flag (M, R, fl, isr) _IF = S_Flag S' _IF).
  apply enter_SW_grt_hyp9 with pc pc' _SWITCH (M', R', fl', isr') S0; auto.
  unfold S_ISR at 1 in H55. unfold S_Flag at 1 in H55.
  inversion_clear H55.
  unfold WF_Curr_Thrd.
  assert (beq_nat rho_ready rho_ih = false).
  apply neq_beq_false. discriminate.
  rewrite H55. clear H55.
  unfold itp_NH_theta.
  split.
  rewrite H43 in H57. rewrite H57. auto.
  split.
  unfold NH_pinvstar.
  unfold pmstar.
  exists (MemMod.merge cthM12 cthM11'').
  exists cthM2. 
  split. assert (MemMod.disjUnion (MemMod.merge cthM12 cthM11'') cthM2 cthM').
  apply enter_SW_grt_hypa; auto.
  UnfoldStateAll. auto.
  split; auto.
  
  assert (S_Mem_Upd (S_Mem_Upd (S_Mem_Upd S' M'1) cthM')
        (MemMod.merge cthM12 cthM11'') = S_Mem_Upd S' (MemMod.merge cthM12 cthM11'')).
  auto. rewrite H55. clear H55.
  apply enter_SW_grt_auxb with p cthM1 R fl isr M cthM11 pc pc' M' R' fl' isr' junk1 S0; auto.
  apply enter_SW_grt_hypc with M21_0 M22 M12 cthM2; auto.
  apply enter_SW_grt_hypd with M21_0 M22 M12 cthM2 S'; auto.
  assert (S_Flag (S_Mem_Upd (S_Mem_Upd S' M'1) cthM') = S_Flag S'). auto.
  rewrite H55. unfold NH_Inv. rewrite <- H58. 
  unfold NH_Inv in H26. auto.
  intro Tag1.
  assert (NH_WFST rho_ready Psi'0 (NH_ItpG g (cthM, R, fl, isr))).
  apply H28. auto.
  apply NH_WFST_strengthen with (cthM, R, fl, isr) g; auto.
  unfold pimp.
  intros.
  unfold NH_ItpG.
  assert ((S_Mem_Upd (S_Mem_Upd S' M'1) cthM') = (S_Mem_Upd S' cthM')).
  auto. rewrite H60 in H59. clear H60.

  unfold NH_ItpG in H59.
  InvertAll.
  rename x into cthM'1. rename x0 into cthM'2. rename x1 into SM1. rename x2 into SM2.
  exists cthM1. exists cthM2. exists SM1. exists SM2.
  split. unfold S_Mem. auto. (* H14 *)
  split. auto. split. auto. split. auto.
  unfold gimp in H46.
  apply H46 with pc'.
  unfold conjG.
  exists (S_Mem_Upd (S_Mem_Upd S' cthM') cthM'1).
  split; auto.
  assert ((cthM2 = cthM'2) /\ (cthM'1 = MemMod.merge cthM12 cthM11'')).
  apply enter_SW_grt_hype with S' cthM' fl; auto.
  inversion_clear H63. rewrite H66.
  unfold conjPG. split; auto.
  assert (S_Mem_Upd (cthM, R, fl, isr) cthM1 = (cthM1, R, fl, isr)).
  auto. rewrite H63. clear H63.
  assert (S_Mem_Upd (S_Mem_Upd S' cthM') (MemMod.merge cthM12 cthM11'')
            = S_Mem_Upd S' (MemMod.merge cthM12 cthM11'')). auto.
  rewrite H63. clear H63.
  apply enter_SW_grt_auxb0 with M cthM11 M' R' fl' isr' junk1 S0; auto.
  apply enter_SW_grt_hypc with M21_0 M22 M12 cthM2; auto.
  apply enter_SW_grt_hypd with M21_0 M22 M12 cthM2 S'; auto.
  split.
  apply enter_SW_grt_hypf with T (R' SP, TCB_Status tcb, pc + 1); auto.
  apply enter_SW_grt_hypg with T (R' SP, TCB_Status tcb, pc + 1); auto.
  unfold WF_currTID in H16. InvertAll. auto.
  apply leqvM_app_subst with 
   (embed_curr_TCB T' tid *^* embed_RdyQ T' RdyQ *^* embed_WQSet T' B); auto.
  apply mstar_assoc. unfold mstar at 1. exists M21_0. exists M22.
  split; auto. split; auto.
  apply enter_SW_grt_hyph with T tid (R' SP, TCB_Status tcb, pc + 1); auto.
  unfold WF_currTID in H16. InvertAll. auto.

  (* tid <> tid' *)
  assert (WQMod.not_in_dom tid' RdyQ').
  unfold SW_Post_aux in H41.
  InvertAll. auto.
  assert (WQMod.in_dom tid RdyQ' /\ WQMod.in_dom tid' RdyQ).
  apply enter_SW_grt2_hyp0; auto.
  inversion_clear H45.
  assert (WF_ReadyQ Psi'0 T RdyQ TM). auto.
  unfold WF_ReadyQ in H23.
  assert (exists tcb', TQMod.maps_to T tid' tcb' /\
          (TCB_Status tcb' = READY \/ TCB_Status tcb' = IDLE) /\
          (exists M : Mem,
             ThrdMemMod.maps_to TM tid' M /\
             (TCB_SP tcb' |~>w TCB_PC tcb') M /\
             (WF_Rdy_Thrd Psi'0 M tcb' \/ WF_Rdy_Thrd_IH Psi'0 M tcb'))).
  apply H23. auto.
  InvertAll. rename x into tcb''.
  rename x0 into cthMR.

  induction tcb'' as ((sp'', status''), pc'').
  exists pc''.
  assert (tcb' = (sp'', status'', pc'') /\ sp' = sp'' /\ S_RF S0 SP = sp').
  unfold SW_Post_aux in H41.
  InvertAll.
  assert (tcb' = (sp'', status'', pc'')).
  assert (TQMod.maps_to T tid' tcb').
  unfold TQMod.maps_to in H41.
  unfold TQMod.put in H41. induction (TQMod.beq_A tid' tid).
  absurd (tid' <> tid); auto. unfold TQMod.maps_to. auto.
  unfold TQMod.maps_to in H56. unfold TQMod.maps_to in H48.
  rewrite H56 in H48. inversion H48. auto.
  split; auto.
  split; auto.
  rewrite H56 in H54. auto.
  inversion_clear H52. inversion_clear H55.

  unfold TCB_PC in H50. unfold TCB_SP in H50.
  split.
  apply enter_SW_grt2_hyp1 with TM tid tid' cthMR sp'' M0' M2' M22 M12 cthM2 cthM12; auto.
  rewrite <- H52. auto.
  inversion_clear H53.

  (* not in interrupt handler *)
  unfold WF_Rdy_Thrd in H55.
  InvertAll.
  rename x into rho'. rename x0 into p'. rename x1 into g'.
  rename x2 into tag'. clear H57.
  exists (packS0 rho' (Ast p' g' tag')).
  split; auto.
  intros.
  unfold Itp0.
  assert (D rho' = SomeT1 _ (Itp_AIM rho')).
  apply non_ih_spec_miniD; auto.
  rewrite H59. clear H59.
  unfold Itp_AIM.
  exists (TQMod.put T tid (R' SP, TCB_Status tcb, pc + 1)).
  set (T' := TQMod.put T tid (R' SP, TCB_Status tcb, pc + 1)).
  exists B. exists RdyQ'. exists tid'.
  clear H23.

  assert (exists INVM, exists cthM0, exists M21', exists M122, exists cthM_old, exists cthM_oldStk,
        MemMod.disjUnion INVM cthM0 cthM12 
        /\ (INV0 *^* INV1) INVM
        /\ (embed_curr_TCB T' tid' *^* embed_RdyQ T' RdyQ') M21'
        /\ MemMod.disjUnion M122 cthMR M12
        /\ MemSplit (ThrdMemMod.remove (ThrdMemMod.remove TM tid) tid') M122
        /\ MemMod.disjUnion (MemMod.merge cthM0 cthM2) cthM_oldStk cthM_old
        /\ (R' SP |-> (pc+1) *^* R' SP + 1 |-> junk1 *^* stkSpacePrec (R' SP) 8) cthM_oldStk
        /\ disjUnionList (M22 :: cthMR :: INVM :: M21' :: cthM_old :: M122 :: nil) (S_Mem S')).
  apply enter_SW_grt2_hyp2 with T B sp' tcb tcb' pc' S0 M0' M2'; auto.
  unfold NH_Inv.
  rewrite <- H10. unfold S_Flag.
  assert (S_Flag (M, R, fl, isr) _IF = S_Flag (M', R', fl', isr') _IF 
             /\ S_ISR (M, R, fl, isr) = S_ISR (M', R', fl', isr')).
  apply enter_SW_grt_auxb0_hyp4 with pc _SWITCH; auto.
  inversion_clear H23.
  unfold S_Flag in H59. rewrite <- H59.
  unfold NH_Inv in H26. auto.
  InvertAll.
  rename x into INVM. rename x0 into cthM0. rename x1 into M21'. rename x2 into M122.
  rename x3 into cthM_old. rename x4 into cthM_oldStk.
  unfold pmstar.
  assert (
      exists M'0, exists M'1, 
        MemMod.disjUnion M'0 M'1 (S_Mem S')
        /\ MemMod.disjUnion M21' M22 M'1
        /\ disjUnionList (cthMR :: INVM :: cthM_old :: M122 :: nil) M'0
  ).
  apply enter_SW_grt2_hyp3; auto.
  InvertAll. rename x into M'0. rename x0 into M'1.
  exists M'0. exists M'1. split; auto.
  split.
  unfold WF_AIM_World.
  split.

  apply enter_SW_grt2_hyp4 with T RdyQ tid (R' SP, TCB_Status tcb, pc + 1) (sp'', status'', pc'') rho_ready; 
  auto.

  split.

  apply enter_SW_grt2_hyp5 with T RdyQ tid (R' SP) (pc+1) tcb; auto.
  split.
  apply enter_SW_grt2_aux6 with RdyQ tid tid' tcb (R' SP, TCB_Status tcb, pc + 1) T rho_ready; auto.
  split; auto.
  exists (ThrdMemMod.put (ThrdMemMod.put TM tid' (MemMod.merge cthMR INVM)) tid cthM_old).
  set (TM' := ThrdMemMod.put (ThrdMemMod.put TM tid' (MemMod.merge cthMR INVM)) tid cthM_old).
  exists (MemMod.merge cthMR INVM). set (cthM' := MemMod.merge cthMR INVM).
  split. assert (S_Mem (S_Mem_Upd S' M'0) = M'0). auto.
  rewrite H69.
  apply enter_SW_grt2_hyp7 with cthMR INVM cthM_old M122 TM tid tid'; auto.
  split. unfold cthM'. unfold TM'.
  unfold ThrdMemMod.put. unfold ThrdMemMod.maps_to.
  induction (ThrdMemMod.beq_A tid' tid).
  absurd (tid' <> tid); auto. induction (ThrdMemMod.beq_A tid' tid').
  auto. absurd (tid' = tid'); auto.
  split.
  unfold WF_Curr_Thrd.
  assert (beq_nat rho' rho_ih = false).
  apply neq_beq_false. inversion_clear H53. rewrite H69. discriminate.
  rewrite H69. discriminate. rewrite H69.
  unfold itp_NH_theta.
  assert (S_ISR S' = false).
  apply enter_SW_grt2_hyp8 with pc pc' M R fl isr M' R' fl' isr' S0 _SWITCH; auto.
  split. 
  UnfoldStateAll. auto.
  assert ((INV0 *^* INV1 -* itp_NH_theta rho' (Ast p' g' tag') Psi'0)
          (cthMR, S_RF S', S_Flag S', false)).
  apply H60.
  rewrite H52 in H56.
  apply enter_SW_grt2_hyp9 with S0 pc'; auto.
  assert (S_ISR (M, R, fl, isr) = S_ISR S' /\ S_Flag (M, R, fl, isr) _IF = S_Flag S' _IF).
  apply enter_SW_grt_hyp9 with pc pc' _SWITCH (M', R', fl', isr') S0; auto.
  inversion_clear H72.
  rewrite <- H74. unfold S_Flag. auto.
  assert (itp_NH_theta rho' (Ast p' g' tag') Psi'0 (cthM', S_RF S', S_Flag S', false)).
  unfold cthM'.
  unfold pm_mwand in H72.
  assert ((MemMod.merge cthMR INVM, S_RF S', S_Flag S', false) =
           S_Mem_Upd (cthMR, S_RF S', S_Flag S', false) (MemMod.merge cthMR INVM)).
  auto.
  rewrite H73. apply H72 with INVM; auto.
  unfold S_Mem.
  assert (MemMod.disjUnion cthMR INVM (MemMod.merge cthMR INVM)).
  unfold MemMod.disjUnion. split; auto.
  apply enter_SW_grt2_hypa with M22 M21' cthM_old M122 S'; auto.
  auto with PMap.
  assert (S_Mem_Upd (S_Mem_Upd S' M'0) cthM' = (cthM', S_RF S', S_Flag S', false)).
  induction S' as (((M'', R''), fl''), isr'').
  UnfoldStateAll. auto.
  rewrite H71. auto. rewrite H74.
  clear H74. unfold itp_NH_theta in H73.
  inversion_clear H73. auto.

  split.
  unfold WF_ReadyQ.
  intros.
  assert (tid0 = tid \/ tid0 <> tid). apply nat_eq_neq_dec.
  inversion_clear H71.

  (* tid0 = tid *)
  exists (R' SP, TCB_Status tcb, pc + 1). rewrite H72.
  split; auto. unfold T'.
  unfold TQMod.put. unfold TQMod.maps_to.
  induction (TQMod.beq_A tid tid). auto. absurd (tid = tid); auto.
  split.
  unfold TCB_Status at 1. unfold TCB_Status at 2.
  unfold WF_currTID in H16. InvertAll.
  unfold TQMod.maps_to in H75. rewrite H75 in H39.
  inversion H39. rewrite <- H78. auto.
  exists cthM_old.
  split. unfold TM'.
  unfold ThrdMemMod.put at 1. unfold ThrdMemMod.maps_to.
  induction (ThrdMemMod.beq_A tid tid); auto.
  absurd (tid = tid); auto.
  split.
  unfold TCB_SP. unfold TCB_PC.
  apply enter_SW_grt2_hypb with (MemMod.merge cthM0 cthM2) cthM_oldStk junk1; auto.
  left.
  unfold WF_Rdy_Thrd. exists rho_ready.
  assert (exists p', exists g',
         SCAP_lookupCT psi (pc+1) (Ast p' g' tag)
         /\ (forall pc', pimp (p |> (wrap_fun gy pc _SWITCH pc')) p')
         /\ (forall pc', gimp (p >o (wrap_fun gy pc _SWITCH pc') oo g') g)).
  apply call_sw_nextPC_spec; auto.
  InvertAll. rename x into p''. rename x0 into g''.
  exists p''. exists g''. exists tag.
  split. left. auto.
  split.
  unfold WF_currTID in H16.
  InvertAll.
  unfold TQMod.maps_to in H78. rewrite H78 in H39. inversion H39. rewrite <- H81.
  unfold rho_tcb_consistent in H80. unfold rho_tcb_consistent. auto.
  split.
  apply enter_SW_grt2_hypb with (MemMod.merge cthM0 cthM2) cthM_oldStk junk1; auto.
  split.
  apply psi_sw_lookupNPC with psi Psi Psi'; auto.
  intros.
  unfold pm_mwand. unfold S_Mem.
  intros.
  unfold itp_NH_theta.
  split.
  UnfoldStateAll. auto.
  split.
  apply enter_SW_grt2_auxc with cthM1 cthM2 cthM R fl isr cthM11 cthM12 INVM cthM0 cthM_oldStk pc junk1
        M0 p M M' R' fl' isr'; auto.

(* WFST *)
  intro.
  assert (NH_WFST rho_ready Psi'0 (NH_ItpG g (cthM, R, fl, isr))).
  apply H28. auto.
  clear H28. clear H79.
  apply NH_WFST_strengthen with (cthM, R, fl, isr) g; auto.
  unfold pimp.
  intros.
  unfold gimp in H75.
  unfold NH_ItpG.
  unfold NH_ItpG in H28.
  InvertAll. rename x into M'20. rename x0 into M'21.
  rename x1 into SM1. rename x2 into SM2.
  exists cthM1. exists cthM2. exists SM1. exists SM2.
  split; auto.
  split; auto.
  split. unfold S_Flag. auto.
  split. auto.
  apply H75 with pc'.
  unfold conjG.
  exists (S_Mem_Upd (S_Mem_Upd (cthM_old, R0, fl0, false) M'2) M'20).
  split; auto.
  unfold conjPG.
  split; auto.
  assert (M'20 = MemMod.merge (MemMod.merge cthM0 M0) cthM_oldStk).
  apply enter_SW_grt2_hype with cthM_old M'2 cthM2 M'21 fl fl0; auto.
  assert (S_Mem_Upd (cthM, R, fl, isr) cthM1 = (cthM1, R, fl, isr)).
  UnfoldStateAll. auto. rewrite H85. clear H85.
  assert (S_Mem_Upd (S_Mem_Upd (cthM_old, R0, fl0, false) M'2) M'20 = (M'20, R0, fl0, false)).
  UnfoldStateAll. auto. rewrite H85. clear H85.
  rewrite H83.
  apply enter_SW_grt2_auxc_aux4 with M M' R' fl' isr' cthM11 cthM12 INVM junk1 cthM2 cthM_old M'2; auto.
  
 (* tid0 <> tid : other ready threads *)
  unfold WF_ReadyQ in H45.
  assert (exists tcb : Thread,
          TQMod.maps_to T tid0 tcb /\
          (TCB_Status tcb = READY \/ TCB_Status tcb = IDLE) /\
          (exists M : Mem,
             ThrdMemMod.maps_to TM tid0 M /\
             (TCB_SP tcb |~>w TCB_PC tcb) M /\
             (WF_Rdy_Thrd Psi'0 M tcb \/ WF_Rdy_Thrd_IH Psi'0 M tcb))).
  apply H45; auto.
  apply enter_SW_grt2_hypf with RdyQ' tid tid'; auto.
  InvertAll.
  exists x.
  split. unfold T'.
  unfold TQMod.put. unfold TQMod.maps_to.
  induction (TQMod.beq_A tid0 tid).
  absurd (tid0 <> tid); auto. unfold TQMod.maps_to in H71.
  auto. split; auto.
  exists x0.
  split; auto.
  unfold TM'.
  unfold ThrdMemMod.put at 1. unfold ThrdMemMod.maps_to.
  induction (ThrdMemMod.beq_A tid0 tid). absurd (tid0 <> tid); auto.
  unfold ThrdMemMod.put.
  induction (ThrdMemMod.beq_A tid0 tid').
  rewrite <- a in H40.
  absurd (~ WQMod.in_dom tid0 RdyQ'); auto with PMap.
  unfold ThrdMemMod.maps_to in H75. auto.
  apply enter_SW_grt2_hypg with T TM RdyQ RdyQ' tid tid' (R' SP, TCB_Status tcb, pc + 1) cthM_old 
            (MemMod.merge cthMR INVM) rho_ready; auto.
  apply enter_SW_grt2_hyph with M21' M22 T (R' SP, TCB_Status tcb, pc + 1) RdyQ tid rho_ready; auto.

  (* tid' is in interrupt hander *)
  unfold WF_Rdy_Thrd_IH in H55.
  InvertAll.
  rename x into p'. rename x0 into g'.
  rename x1 into tag'.
  exists (packS0 rho_ih (Ast p' g' tag')).
  split; auto.
  intros.
  unfold Itp0.
  assert (D rho_ih = SomeT1 _ (Itp_AIM rho_ih)).
  unfold Minimal_D in H0. InvertAll. auto.
  rewrite H59. clear H59.
  unfold Itp_AIM.
  exists (TQMod.put T tid (R' SP, TCB_Status tcb, pc + 1)).
  set (T' := TQMod.put T tid (R' SP, TCB_Status tcb, pc + 1)).
  exists B. exists RdyQ'. exists tid'.
  clear H23.

  assert (exists INVM, exists cthM0, exists M21', exists M122, exists cthM_old, exists cthM_oldStk,
        MemMod.disjUnion INVM cthM0 cthM12 
        /\ (INV0 *^* INV1) INVM
        /\ (embed_curr_TCB T' tid' *^* embed_RdyQ T' RdyQ') M21'
        /\ MemMod.disjUnion M122 cthMR M12
        /\ MemSplit (ThrdMemMod.remove (ThrdMemMod.remove TM tid) tid') M122
        /\ MemMod.disjUnion (MemMod.merge cthM0 cthM2) cthM_oldStk cthM_old
        /\ (R' SP |-> (pc+1) *^* R' SP + 1 |-> junk1 *^* stkSpacePrec (R' SP) 8) cthM_oldStk
        /\ disjUnionList (M22 :: cthMR :: INVM :: M21' :: cthM_old :: M122 :: nil) (S_Mem S')).
  apply enter_SW_grt2_hyp2 with T B sp' tcb tcb' pc' S0 M0' M2'; auto.
  unfold NH_Inv.
  rewrite <- H10. unfold S_Flag.
  assert (S_Flag (M, R, fl, isr) _IF = S_Flag (M', R', fl', isr') _IF 
             /\ S_ISR (M, R, fl, isr) = S_ISR (M', R', fl', isr')).
  apply enter_SW_grt_auxb0_hyp4 with pc _SWITCH; auto.
  inversion_clear H23. unfold S_Flag in H59. rewrite <- H59.
  unfold NH_Inv in H26. auto.
  InvertAll.
  rename x into INVM. rename x0 into cthM0. rename x1 into M21'. rename x2 into M122.
  rename x3 into cthM_old. rename x4 into cthM_oldStk.

  unfold pmstar.
  assert (
      exists M'0, exists M'1, 
        MemMod.disjUnion M'0 M'1 (S_Mem S')
        /\ MemMod.disjUnion M21' M22 M'1
        /\ disjUnionList (cthMR :: INVM :: cthM_old :: M122 :: nil) M'0
  ).
  apply enter_SW_grt2_hyp3; auto.
  InvertAll. rename x into M'0. rename x0 into M'1.
  exists M'0. exists M'1. split; auto.
  split.
  unfold WF_AIM_World.
  split.

  apply enter_SW_grt2_hyp4 with T RdyQ tid (R' SP, TCB_Status tcb, pc + 1) (sp'', status'', pc'') rho_ready; 
  auto. auto.

  split.
  apply enter_SW_grt2_hyp5 with T RdyQ tid (R' SP) (pc+1) tcb; auto.
  split.
  apply enter_SW_grt2_aux6 with RdyQ tid tid' tcb (R' SP, TCB_Status tcb, pc + 1) T rho_ready; auto.
  split; auto.
  exists (ThrdMemMod.put (ThrdMemMod.put TM tid' (MemMod.merge cthMR INVM)) tid cthM_old).
  set (TM' := ThrdMemMod.put (ThrdMemMod.put TM tid' (MemMod.merge cthMR INVM)) tid cthM_old).
  exists (MemMod.merge cthMR INVM). set (cthM' := MemMod.merge cthMR INVM).
  split. assert (S_Mem (S_Mem_Upd S' M'0) = M'0). auto.
  rewrite H68.
  apply enter_SW_grt2_hyp7 with cthMR INVM cthM_old M122 TM tid tid'; auto.
  split. unfold cthM'. unfold TM'.
  unfold ThrdMemMod.put. unfold ThrdMemMod.maps_to.
  induction (ThrdMemMod.beq_A tid' tid).
  absurd (tid' <> tid); auto. induction (ThrdMemMod.beq_A tid' tid').
  auto. absurd (tid' = tid'); auto.
  split.
  unfold WF_Curr_Thrd.
  simpl.
  assert (S_Flag S' _IF = false).
  apply enter_SW_grt2_hypI with pc pc' M R fl isr M' R' fl' isr' S0; auto.

  assert ((INV0 *^* INV1 -* itp_IH_theta (Ast p' g' tag') status'' Psi'0)
          (cthMR, S_RF S', S_Flag S', false)).
  apply H58.
  rewrite H52 in H56.
  apply enter_SW_grt2_hyp9 with S0 pc'; auto. auto.
  exists (sp'', status'', pc''). exists status''.
  split.
  unfold T'. unfold TQMod.maps_to. unfold TQMod.put.
  induction (TQMod.beq_A tid' tid). absurd (tid' <> tid); auto.
  auto.
  split; auto.
  unfold pm_mwand in H70.
  assert (S_ISR S' = false).
  apply enter_SW_grt2_hyp8 with pc pc' M R fl isr M' R' fl' isr' S0 _SWITCH; auto.
  assert (itp_IH_theta (Ast p' g' tag') status'' Psi'0
          (S_Mem_Upd (cthMR, S_RF S', S_Flag S', false) cthM')).
  apply H70 with INVM; auto.
  unfold S_Mem. unfold cthM'.
  assert (MemMod.disjUnion cthMR INVM (MemMod.merge cthMR INVM)).
  unfold MemMod.disjUnion. split; auto.
  apply enter_SW_grt2_hypa with M22 M21' cthM_old M122 S'; auto.
  auto with PMap.
  rewrite <- H71 in H72. UnfoldStateAll. auto.

  split.
  unfold WF_ReadyQ.
  intros.
  assert (tid0 = tid \/ tid0 <> tid). apply nat_eq_neq_dec.
  inversion_clear H70.
  exists (R' SP, TCB_Status tcb, pc + 1). rewrite H71.
  split; auto. unfold T'.
  unfold TQMod.put. unfold TQMod.maps_to.
  induction (TQMod.beq_A tid tid). auto. absurd (tid = tid); auto.
  split.
  unfold TCB_Status at 1. unfold TCB_Status at 2.
  unfold WF_currTID in H16. InvertAll.
  unfold TQMod.maps_to in H74. rewrite H74 in H39.
  inversion H39. rewrite <- H77. auto.
  exists cthM_old.
  split. unfold TM'.
  unfold ThrdMemMod.put at 1. unfold ThrdMemMod.maps_to.
  induction (ThrdMemMod.beq_A tid tid); auto.
  absurd (tid = tid); auto.
  split.
  unfold TCB_SP. unfold TCB_PC.
  apply enter_SW_grt2_hypb with (MemMod.merge cthM0 cthM2) cthM_oldStk junk1; auto.
  left.
  unfold WF_Rdy_Thrd. exists rho_ready.
  assert (exists p', exists g',
         SCAP_lookupCT psi (pc+1) (Ast p' g' tag)
         /\ (forall pc', pimp (p |> (wrap_fun gy pc _SWITCH pc')) p')
         /\ (forall pc', gimp (p >o (wrap_fun gy pc _SWITCH pc') oo g') g)).
  apply call_sw_nextPC_spec; auto.
  InvertAll. rename x into p''. rename x0 into g''.
  exists p''. exists g''. exists tag.
  split. left. auto.
  split.
  unfold WF_currTID in H16. InvertAll.
  unfold TQMod.maps_to in H77. rewrite H77 in H39.
  inversion H39. rewrite <- H80. unfold rho_tcb_consistent in H79.
  unfold rho_tcb_consistent. auto.
  split.
  apply enter_SW_grt2_hypb with (MemMod.merge cthM0 cthM2) cthM_oldStk junk1; auto.
  split.
  apply psi_sw_lookupNPC with psi Psi Psi'; auto.
  intros.
  unfold pm_mwand. unfold S_Mem.
  intros.
  unfold itp_NH_theta.
  split.
  UnfoldStateAll. auto.
  split.
  apply enter_SW_grt2_auxc with cthM1 cthM2 cthM R fl isr cthM11 cthM12 INVM cthM0 cthM_oldStk pc junk1
        M0 p M M' R' fl' isr'; auto.

(* WFST *)
  intro.
  assert (NH_WFST rho_ready Psi'0 (NH_ItpG g (cthM, R, fl, isr))).
  apply H28. auto.
  clear H28. clear H78.
  apply NH_WFST_strengthen with (cthM, R, fl, isr) g; auto.
  unfold pimp.
  intros.
  unfold gimp in H74.
  unfold NH_ItpG.
  unfold NH_ItpG in H28.
  InvertAll. rename x into M'20. rename x0 into M'21.
  rename x1 into SM1. rename x2 into SM2.
  exists cthM1. exists cthM2. exists SM1. exists SM2.
  split; auto.
  split; auto.
  split. unfold S_Flag. auto.
  split. auto.
  apply H74 with pc'.
  unfold conjG.
  exists (S_Mem_Upd (S_Mem_Upd (cthM_old, R0, fl0, false) M'2) M'20).
  split; auto.
  unfold conjPG.
  split; auto.
  assert (M'20 = MemMod.merge (MemMod.merge cthM0 M0) cthM_oldStk).
  apply enter_SW_grt2_hype with cthM_old M'2 cthM2 M'21 fl fl0; auto.
  assert (S_Mem_Upd (cthM, R, fl, isr) cthM1 = (cthM1, R, fl, isr)).
  UnfoldStateAll. auto. rewrite H84. clear H84.
  assert (S_Mem_Upd (S_Mem_Upd (cthM_old, R0, fl0, false) M'2) M'20 = (M'20, R0, fl0, false)).
  UnfoldStateAll. auto. rewrite H84. clear H84.
  rewrite H82.
  apply enter_SW_grt2_auxc_aux4 with M M' R' fl' isr' cthM11 cthM12 INVM junk1 cthM2 cthM_old M'2; auto.
  
 (* tid0 <> tid : other ready threads *)
  unfold WF_ReadyQ in H45.
  assert (exists tcb : Thread,
          TQMod.maps_to T tid0 tcb /\
          (TCB_Status tcb = READY \/ TCB_Status tcb = IDLE) /\
          (exists M : Mem,
             ThrdMemMod.maps_to TM tid0 M /\
             (TCB_SP tcb |~>w TCB_PC tcb) M /\
             (WF_Rdy_Thrd Psi'0 M tcb \/ WF_Rdy_Thrd_IH Psi'0 M tcb))).
  apply H45; auto.
  apply enter_SW_grt2_hypf with RdyQ' tid tid'; auto.
  InvertAll.
  exists x.
  split. unfold T'.
  unfold TQMod.put. unfold TQMod.maps_to.
  induction (TQMod.beq_A tid0 tid).
  absurd (tid0 <> tid); auto. unfold TQMod.maps_to in H70.
  auto. split; auto.
  exists x0.
  split; auto.
  unfold TM'.
  unfold ThrdMemMod.put at 1. unfold ThrdMemMod.maps_to.
  induction (ThrdMemMod.beq_A tid0 tid). absurd (tid0 <> tid); auto.
  unfold ThrdMemMod.put.
  induction (ThrdMemMod.beq_A tid0 tid').
  rewrite <- a in H40.
  absurd (~ WQMod.in_dom tid0 RdyQ'); auto with PMap.
  unfold ThrdMemMod.maps_to in H74. auto.
  apply enter_SW_grt2_hypg with T TM RdyQ RdyQ' tid tid' (R' SP, TCB_Status tcb, pc + 1) cthM_old 
            (MemMod.merge cthMR INVM) rho_ready; auto.
  apply enter_SW_grt2_hyph with M21' M22 T (R' SP, TCB_Status tcb, pc + 1) RdyQ tid rho_ready; auto.
Qed.




Lemma enter_SW:
forall D psi pc theta D' Psi Psi' Psi'0 a S S',
  NH_WFcmd psi pc theta (comml (_call _SWITCH))
  -> Minimal_D D
  -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
  -> lift_psi psi rho_ready Psi
  -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
  -> a = Itp_AIM rho_ready theta
  -> subsetCT Psi' Psi'0
  -> a Psi'0 S
  -> nextS pc (comml (_call _SWITCH)) S S'
  -> Itp0 D' SWITCH_Spec0 Psi'0 S'.
Proof.
  intros.
  unfold Itp0.
  unfold SWITCH_Spec0.
  rewrite H1.
  simpl.
  rewrite H4 in H6.
  clear H4 a.
  split.
  apply enter_SW_pre with psi pc theta Psi'0 S; auto.
  split.
  apply enter_SW_disable with psi pc theta Psi'0 S; auto.
  exists 0.
  apply enter_SW_grt with psi pc theta D' Psi Psi' S; auto.
Qed.

Lemma NH_SW_Presv:
  forall D psi pc theta D' Psi Psi' Psi'0 a S S' f',
    NH_WFcmd psi pc theta (comml (_call _SWITCH))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_ready Psi
    -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
    -> a = Itp_AIM rho_ready theta
    -> subsetCT Psi' Psi'0
    -> a Psi'0 S
    -> nextS pc (comml (_call _SWITCH)) S S'
    -> nextPC (comml (_call _SWITCH)) (S_Mem S) (S_RF S) (S_Flag S) pc f'
    -> exists theta0 : Spec0, lookupCT Psi'0 f' theta0 /\ Itp0 D' theta0 Psi'0 S'.
Proof.
  intros.
  assert (lookupCT Psi'0 _SWITCH SWITCH_Spec0).
  apply subsetCT_presv_lookupCT with Psi'; auto.
  rewrite H3. 
  unfold ext_IH_Spec, ext_Prim_Specs. 
  apply lookupCT_subsetCT with (consCT _SWITCH SWITCH_Spec0 Psi). 
  apply lookupCT_consCT.
  apply Psi_extend_subsetCT.
  apply Psi_extend_subsetCT.
  apply Psi_extend_subsetCT.
  apply subsetCT_refl.
  inversion H8. clear H14 H13 H12 H10 H11 fl R M f pc0.
  exists SWITCH_Spec0.
  split; auto.
  apply enter_SW with D psi pc theta Psi Psi' a S; auto.
Qed.

Lemma call_sw_disable_itr:
forall psi Psi'0 pc theta S,
  NH_WFcmd psi pc theta (comml (_call _SWITCH)) 
  -> Itp_AIM rho_ready theta Psi'0 S
  -> ~ itr_enabled S.
Proof.
  intros.
  unfold NH_WFcmd in H.
  induction theta.
  InvertAll.
  unfold NH_WFcmd_aux in H1.
  inversion_clear H1.
  unfold NH_Call_Prim in H2.
  InvertAll.
  assert (monotonic_p (fun S => S_Flag S _IF = false)). auto.
  assert (forall S, imprec p S -> (fun S => S_Flag S _IF = false) S).
  apply (imprec_presv_imp_monotonic H4 H5); auto.
  assert (imprec p S).
  apply imprec_presv_imprec with (NH_pinvstar p); auto.
  apply Itp_AIM_pinv with g n Psi'0 rho_ready; auto.
  assert ((fun S0 : State => S_Flag S0 _IF = false) S).
  apply (H6 S H7). intro.
  unfold itr_enabled in H9.
  assert (S_Flag S _IF = false). auto.
  inversion H9. rewrite H11 in H10. inversion H10.
  inversion H2. absurd (prim_entry _SWITCH); auto.
  unfold prim_entry. left. auto.
Qed.

Lemma NH_SW_Itp :
  forall D psi pc theta D' Psi Psi',
    NH_WFcmd psi pc theta (comml (_call _SWITCH))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_ready Psi
    -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
    (* -> a = Itp_AIM rho_ready theta *)
    -> WFcmd D' Psi' (packS0 rho_ready theta) pc (comml (_call _SWITCH)).
Proof.
  intros.
  unfold WFcmd.
  split.
  unfold WFcmd_aux.
  intros.
  assert (Itp_AIM rho_ready theta Psi'0 S).
  unfold Itp0 in H4.
  rewrite H1 in H4.
  unfold ext_dict in H4.
  assert (beq_nat rho_ready rho_prim = false). simpl. auto.
  rewrite H6 in H4.
  unfold Minimal_D in H0.
  InvertAll.
  rewrite H7 in H4. auto.
  split.
  apply NH_SW_progress with psi theta Psi'0 (Itp_AIM rho_ready theta); auto.
  split.
  intros.
  apply NH_SW_Presv with D psi pc theta Psi Psi' (Itp_AIM rho_ready theta) S; auto.
  left.
  apply call_sw_disable_itr with psi Psi'0 pc theta; auto.
  rewrite H1.  
  unfold Itp0.
  unfold ext_dict.
  assert (beq_nat rho_ready rho_prim = false).
  simpl. auto.
  rewrite H4.
  unfold Minimal_D in H0. InvertAll.
  rewrite H5.
  apply Itp_AIM_Mono_CT.
Qed.

