Add LoadPath "..".
Add LoadPath "../CAP0".

Require Import Arith.
Require Import Coqlib.
Require Import Map.
Require Import Machine.
Require Import Assertion.
Require Import Assertion_Lemmas.
Require Import Assertion_Lemmas4.
Require Import Operational_Semantics.
Require Import OpSem_Lemmas.
Require Import OpSem_Lemmas2.
Require Import Reg.
Require Import CAP0.
Require Import SCAP.
Require Import AIM.
Require Import QLib.
Require Import SCAP_READY.
Require Import SCAP_IDLE.
Require Import SCAP_Itr.
Require Import Linking.
Require Import Linking_Lemmas.
Require Import Linking_Lemmas2.
Require Import Linking_Lemmas3.

Require Import EqNat.
Require Import Omega.
Require Import List.
Require Import Arith_Lemmas.



Lemma prim_UNBLK:
  prim_entry _UNBLK.
Proof.
  unfold prim_entry.
  right. right. auto.
Qed.
Hint Resolve prim_UNBLK.


Lemma enable_unblk_prog:
  forall w wqa S pc f, enable_unblk w wqa S
    -> exists S' : State, nextS pc (comml (_call f)) S S'.
Proof.
  intros.
  unfold enable_unblk in H.
  InvertAll.
  apply stkSpace_call_prog with (_UNBLK_Stk + 2); auto.
  assert (((S_RF S SP |->w w *^*
           (wqa *^* stkSpacePrec (S_RF S SP) (_UNBLK_Stk + 2))) *^* mTrue) (S_Mem S)).
  apply pTrue_m_mTrue with pTrue; auto.
  assert (((wqa *^* stkSpacePrec (S_RF S SP) (_UNBLK_Stk + 2)) *^* mTrue) (S_Mem S)).
  apply mstar_mTrue with (S_RF S SP |->w w) mTrue; auto.
  apply mstar_mTrue with wqa mTrue; auto.
  omega.
Qed.


Lemma NH_Call_unblk_enable:
forall psi pc p g n S,
  NH_Call_Prim psi pc _UNBLK (Ast p g n)
  -> imprec p S 
  -> exists w, exists wqa, enable_unblk w wqa S.
Proof.
  intros.
  unfold NH_Call_Prim in H.
  InvertAll.
  inversion H1.
  absurd (_UNBLK <> _SWITCH); auto.
  absurd (_UNBLK <> _BLOCK); auto.
  inversion H3.
  rewrite <- H7 in H5. rewrite <- H8 in H5.
  rewrite <- H9 in H4.
  clear H7 H8 H9 H3 H tag p0 g0.
  induction S as (((M, R), fl), isr). 
  unfold imprec in H0.
  unfold pmstar in H0.
  InvertAll. rename x into M1. rename x0 into M2.
  UnfoldStateAll.
  assert (exists w : nat,
    exists wqa : WQassert,
      enable_unblk w wqa (M1, R, fl, isr) /\
      (forall pc' : nat, pimp (wrap_fun (gunblk wqa) pc _UNBLK pc' (M1, R, fl, isr)) p') /\
      (forall pc' : nat,
        pimp (wrap_fun (gunblk wqa) pc _UNBLK pc' (M1, R, fl, isr) |> g') (g (M1, R, fl, isr)))).
  apply H5; auto.
  InvertAll. rename x into w. rename x0 into wqa.
  exists w. exists wqa.
  assert (monotonic_p (enable_unblk w wqa)). auto.
  unfold monotonic_p in H8.
  assert (enable_unblk w wqa (S_Mem_Upd (M1, R, fl, isr) M)). apply H8; auto.
  unfold S_Mem. apply MemMod.disjUnion_subseteq_left with M2; auto.
  UnfoldStateAll. auto.
Qed.

Lemma NH_UNBLK_prog_aux :
  forall psi pc theta Psi rho S,
    NH_Call_Prim psi pc _UNBLK theta -> Itp_AIM rho theta Psi S -> rho <> rho_ih 
    -> exists S', nextS pc (comml (_call _UNBLK)) S S'.
Proof.
  intros.
  induction theta.
  assert (imprec (NH_pinvstar p) S).
  apply Itp_AIM_pinv with g n Psi rho; auto.
  assert (imprec p S).
  apply imprec_presv_imprec with (NH_pinvstar p); auto.
  assert (exists w, exists wqa, enable_unblk w wqa S).
  apply NH_Call_unblk_enable with psi pc p g n; auto.
  InvertAll.
  apply enable_unblk_prog with x x0; auto.
Qed.

Lemma NH_UNBLK_progress:
  forall psi pc theta Psi a S,
    NH_WFcmd psi pc theta (comml (_call _UNBLK))
    -> a = Itp_AIM rho_ready theta
    -> a Psi S
    -> exists S', nextS pc (comml (_call _UNBLK)) S S'.
Proof.
  intros.
  unfold NH_WFcmd in H.
  induction theta.
  unfold NH_WFcmd_aux in H.
  inversion_clear H.
  clear H3.
  inversion_clear H2.
  inversion_clear H. clear H2.
  apply NH_UNBLK_prog_aux with psi (Ast p g n) Psi rho_ready; auto.
  rewrite H0 in H1. auto.
  inversion_clear H. 
  absurd (prim_entry _UNBLK); auto.
Qed.

Lemma WF_AIM_World_unblk:
forall Psi'0 S T B RdyQ tid p g tag w wqa,
  WF_AIM_World Psi'0 S T B RdyQ tid rho_ready (Ast p g tag)
  -> enable_unblk w wqa S
  -> exists BlkQ, 
      B w = Some BlkQ
      /\ Delta w = SomeT wqa 
                    /\ (pmstar pTrue (mstar (ptoW (S_RF S SP) w)
                                           (mstar wqa 
                                                  (stkSpacePrec (S_RF S SP) (_UNBLK_Stk + 2))
                                           )
                                   )
                     ) S.
Proof.
  intros.
  unfold enable_unblk in H0.
  InvertAll.
  unfold WF_AIM_World in H.
  InvertAll.
  unfold WF_WQ_WQSpec in H4.
  assert (WQSetMod.in_dom w B).
  apply H4 with wqa; auto.
  unfold WQSetMod.in_dom in H9.
  InvertAll.
  exists x1. auto.
Qed.


Lemma enable_unblk_nxt_stkspace0:
forall pc S S' M1 M2 f w wqa,
  ((pTrue *<*
        (S_RF (S_Mem_Upd S M1) SP |->w w *^*
         (wqa *^*
          stkSpacePrec (S_RF (S_Mem_Upd S M1) SP) (_UNBLK_Stk + 2))))
         (S_Mem_Upd S M1))
  -> nextS pc (comml (_call f)) S S'
  -> MemMod.disjUnion M1 M2 (S_Mem S)
  -> exists M1',
        MemMod.disjUnion M1' M2 (S_Mem S')
        /\ ((S_RF S' SP + 2 |->w w) *^* stkSpacePrec (S_RF S' SP) _UNBLK_Stk *^* mTrue) M1'.
Proof.
  intros.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  UnfoldStateAll.
  assert ((mTrue *^*
       (R SP |->w w *^* (wqa *^* stkSpacePrec (R SP) (_UNBLK_Stk + 2)))) M1).
  unfold pmstar in H.
  InvertAll.
  exists x. exists x0. split_l.
  split. unfold mTrue. auto.
  auto. clear H.
  assert (frame (nextS pc (comml (_call f)))).
  apply frame_call.
  unfold frame in H.
  assert (exists M1', 
        MemMod.disjUnion M1' M2 (S_Mem (M', R', fl', isr')) /\
        nextS pc (comml (_call f)) (S_Mem_Upd (M, R, fl, isr) M1) (S_Mem_Upd (M', R', fl', isr') M1')).
  apply H; auto.
  clear H.
  UnfoldStateAll.
  unfold enableS.
  exists (MemUpdW M1 (R SP - 2) (pc+1) 0, RF_update R SP (R SP - 2), fl, isr).
  apply nextS_call.
  unfold pushW.
  split.
  sep on.
  sep lift 3 in H2.
  sep stksplit 2 in H2. sep lift 0 in H2.
  sep trivial.
  split; auto.
  apply  MemUpdW_MemUpd16 with 0; auto.
  clear H.
  UnfoldStateAll.
  InvertAll.
  exists x.
  split_l.
  destruct (nextS_inversion H4).
  InvertAll.
  subst R'. reg simpl.
  substH H10 with (MemUpd16_MemUpdW H10).
  InvertAll.
  subst x.
  sep on.
  sep lift 3 in H2.
  sep stksplit in H2.
  sep lift 1 in H2.
  assert (MemUpdW M1 (R SP - 2) (pc + 1) x0 ||- 
          R SP - 2 |->w (pc+1) *^* 
           (mTrue *^* R SP |->w w *^* wqa *^*
            stkSpacePrec (R SP - 2) (_UNBLK_Stk + 2 - 2))).
  sep upd. auto.
  assert (R SP - 2 + 2 = R SP). auto.
  rewrite H9.
  assert (MemUpdW M1 (R SP - 2) (pc + 1) x0
         ||- R SP |->w w *^* stkSpacePrec (R SP - 2) (_UNBLK_Stk + 2 - 2) *^* 
             (R SP - 2 |->w pc + 1 *^* wqa *^* mTrue)).
  sep matchH H6 with 5.
  destruct H10 as [M1x [M1y [Hsplit [HM1x HM1y]]]].
  exists M1x. exists M1y.
  split; auto.
  unfold mTrue. split; auto.
Qed.

Lemma enable_unblk_nxt_stkspace:
forall pc S S' M1 M2 f w wqa,
  ((pTrue *<*
        (S_RF (S_Mem_Upd S M1) SP |->w w *^*
         (wqa *^*
          stkSpacePrec (S_RF (S_Mem_Upd S M1) SP) (_UNBLK_Stk + 2))))
         (S_Mem_Upd S M1))
  -> nextS pc (comml (_call f)) S S'
  -> MemMod.disjUnion M1 M2 (S_Mem S)
  -> exists M1', exists junk,
        MemMod.disjUnion M1' M2 (S_Mem S')
        /\ ((S_RF S' SP + 2 |-> w) *^* (S_RF S' SP + 3 |-> junk) *^* stkSpacePrec (S_RF S' SP) _UNBLK_Stk *^* mTrue) M1'.
Proof.
  intros.
  assert (exists M1',
        MemMod.disjUnion M1' M2 (S_Mem S')
        /\ ((S_RF S' SP + 2 |->w w) *^* stkSpacePrec (S_RF S' SP) _UNBLK_Stk *^* mTrue) M1').
  apply enable_unblk_nxt_stkspace0 with pc S M1 f wqa; auto.
  destruct H2 as [M1' [Hsplit Hmem]].
  exists M1'.
  sep on. sep lift 0 in Hmem.
  assert (exists junk1,
            M1' ||- S_RF S' SP + 2 |-> w *^* S_RF S' SP + 2 + 1 |-> junk1 *^*
             (stkSpacePrec (S_RF S' SP) _UNBLK_Stk *^* mTrue)).
  apply ptoW_elim1. auto.
  clear Hmem. destruct H2 as [junk1 Hmem].
  exists junk1.
  assert (S_RF S' SP + 2 + 1 = S_RF S' SP + 3). auto.
  rewrite H2 in Hmem.
  split; auto.
Qed.




Lemma merge_UNBLK_pre:
forall S' M1' w T tid RdyQ B BlkQ M2 junk,
  (S_RF S' SP + 2 |-> w *^* S_RF S' SP + 3 |-> junk *^*
        stkSpacePrec (S_RF S' SP) _UNBLK_Stk *^* mTrue) M1'
  -> (embed_curr_TCB T tid *^* (embed_RdyQ T RdyQ *^* embed_WQSet T B)) M2
  -> B w = Some BlkQ
  -> MemMod.disjUnion M1' M2 (S_Mem S')
  -> (UNBLK_MPre T RdyQ BlkQ (S_RF S' SP) w junk *^* mTrue) (S_Mem S').
Proof.
  intros.
  induction S' as (((M, R), fl), isr).
  UnfoldStateAll.
  rename M1' into M1.
  unfold UNBLK_MPre.
  assert (((R SP + 2 |-> w *^*
            R SP + 3 |-> junk *^* stkSpacePrec (R SP) _UNBLK_Stk *^* mTrue) *^*
              (embed_curr_TCB T tid *^* (embed_RdyQ T RdyQ *^* embed_WQSet T B))) M).
  exists M1. exists M2.
  split; auto.
  assert (forall M, embed_WQSet T B M -> (ptoTQ w (filterTQ T BlkQ) *^* mTrue) M).
  intros.
  apply enbedBlkQ_inWQSet with B; auto.
  sep on.
  sep lift 6 in H3.
  sep weaker_subst in H3 with (ptoTQ w (filterTQ T BlkQ) *^* mTrue); auto.
  clear H5.
  sep on.
  sep lift 7 in H3.
  sep lift 0.
  sep match_left H3; auto.
  sep lift 0 in H3.
  sep match_left H3; auto.
  sep lift 1 in H3.
  sep match_left H3; auto.
  sep match_left H3; auto.
  sep matchH H3 with 2.
Qed.

Lemma enter_UNBLK_preMem:
forall pc S S' p g tag M1 M2 Psi T B RdyQ tid wqa w BlkQ,
  nextS pc (comml (_call _UNBLK)) S S'
  (* -> enable_blk p wqa *)
  -> MemMod.disjUnion M1 M2 (S_Mem S)
  -> WF_AIM_World Psi (S_Mem_Upd S M1) T B RdyQ tid
         rho_ready (Ast p g tag)
  -> (embed_curr_TCB T tid *^* (embed_RdyQ T RdyQ *^* embed_WQSet T B)) M2
  -> Delta w = SomeT wqa
  -> (pTrue *<*
         (S_RF (S_Mem_Upd S M1) SP |->w w *^*
          (wqa *^*
           stkSpacePrec (S_RF (S_Mem_Upd S M1) SP) (_UNBLK_Stk + 2))))
          (S_Mem_Upd S M1)
  -> B w = Some BlkQ
  -> exists junk, (UNBLK_MPre T RdyQ BlkQ (S_RF S' SP) w junk *^* mTrue) (S_Mem S').
Proof.
  intros.
  assert (exists M1', exists junk,
            MemMod.disjUnion M1' M2 (S_Mem S')
            /\ ((S_RF S' SP + 2 |-> w) *^* (S_RF S' SP + 3 |-> junk) *^* stkSpacePrec (S_RF S' SP) _UNBLK_Stk *^* mTrue) M1').

  apply enable_unblk_nxt_stkspace with pc S M1 _UNBLK wqa; auto.
  InvertAll.
  rename x into M1'. rename x0 into junk. exists junk.
  apply merge_UNBLK_pre with M1' tid B M2; auto.
Qed.


Lemma enter_UNBLK_pre:
forall psi pc theta Psi'0 S S',
  NH_WFcmd psi pc theta (comml (_call _UNBLK))
  -> Itp_AIM rho_ready theta Psi'0 S
  -> nextS pc (comml (_call _UNBLK)) S S'
  -> UNBLK_Prd S'.
Proof.
  intros.
  unfold UNBLK_Prd.
  induction theta.
  unfold Itp_AIM in H0.
  InvertAll. rename x into T. rename x0 into B. rename x1 into RdyQ. rename x2 into tid.
  exists T. exists RdyQ.
  unfold NH_WFcmd in H.
  unfold NH_WFcmd_aux in H.
  InvertAll. clear H. clear H4.
  inversion_clear H2.
  inversion_clear H. clear H2.
  unfold NH_Call_Prim in H3.
  inversion_clear H3.
  inversion H.
  absurd (_UNBLK <> _SWITCH); auto.
  absurd (_UNBLK <> _BLOCK); auto.
  inversion H4. rewrite <- H8 in H6. rewrite <- H9 in H6.
  clear H9 H8 H10 H3 H4 H5 tag p0 g0.
  rename n into tag.
  induction S as (((M, R), fl), isr).
  unfold pmstar in H0.
  InvertAll.
  rename x into M1. rename x0 into M2.

  assert (exists w, exists wqa, enable_unblk w wqa (M1, R, fl, isr)).
  apply NH_Call_unblk_enable with psi pc p g tag; auto.
  unfold NH_Call_Prim. split; auto.
  assert (imprec (NH_pinvstar p) (M1, R, fl, isr)).
  apply WF_AIM_World_pinv with Psi'0 T B RdyQ tid rho_ready g tag; auto.
  apply imprec_presv_imprec with (NH_pinvstar p); auto.
  InvertAll. rename x into w. rename x0 into wqa.

  assert (exists BlkQ, B w = Some BlkQ
                                 /\ Delta w = SomeT wqa 
                                 /\ (pmstar pTrue (mstar (ptoW (S_RF (M1, R, fl, isr) SP) w)
                                                         (mstar wqa 
                                                             (stkSpacePrec (S_RF (M1, R, fl, isr) SP) (_UNBLK_Stk + 2))
                                                          )
                                                   )
                                    ) (M1, R, fl, isr)
         ).
  apply WF_AIM_World_unblk with Psi'0 T RdyQ tid p g tag; auto.
  InvertAll. rename x into BlkQ.
  exists BlkQ.
  exists (S_RF S' SP).
  exists w.
  split.
  apply enter_UNBLK_preMem with pc (M, R, fl, isr) p g tag M1 M2 Psi'0 B tid wqa; auto.

  split. apply WF_AIM_World_RdyQinT with Psi'0 (S_Mem_Upd (M, R, fl, isr) M1) B tid rho_ready (Ast p g tag); auto.
  split.
  apply WF_AIM_World_BlkQinT with Psi'0 (S_Mem_Upd (M, R, fl, isr) M1) B RdyQ tid rho_ready (Ast p g tag) w; auto.
  split; auto.
  split.
  apply WF_AIM_World_RdyNotBlk with Psi'0 (M1, R, fl, isr) T B tid (Ast p g tag) w rho_ready; auto.
  apply WF_AIM_World_WFTQ with Psi'0 (S_Mem_Upd (M, R, fl, isr) M1) B RdyQ tid rho_ready (Ast p g tag); auto.
  inversion_clear H.
  absurd (prim_entry _UNBLK); auto.
Qed.

Lemma enter_UNBLK_disable:
forall psi pc theta Psi'0 S S',
  NH_WFcmd psi pc theta (comml (_call _UNBLK))
  -> Itp_AIM rho_ready theta Psi'0 S
  -> nextS pc (comml (_call _UNBLK)) S S'
  -> S_Flag S' _IF = false.
Proof.
  intros.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  unfold NH_WFcmd in H.
  induction theta.
  destruct H as [HWFcmd [Htmp1 Htmp2]].
  clear Htmp1 Htmp2.
  unfold NH_WFcmd_aux in HWFcmd.
  inversion_clear HWFcmd.
  destruct H as [Hprim HWFcmd].
  clear Hprim.
  unfold NH_Call_Prim in HWFcmd.
  inversion_clear HWFcmd.
  clear H.
  assert (fl _IF = false).
  assert (imprec (NH_pinvstar p) (M, R, fl, isr)).
  apply Itp_AIM_pinv with g n Psi'0 rho_ready. auto.
  discriminate.
  unfold imprec in H. unfold pmstar in H.
  InvertAll.
  unfold NH_pinvstar in H.
  simplS in H.
  unfold pmstar in H.
  InvertAll. simplS in H.
  assert (S_Flag (x1, R, fl, isr) _IF = false).
  apply H2; auto.
  UnfoldStateAll; auto. clear H0 H2.
  destruct (nextS_inversion H1).
  InvertAll.
  subst fl'.
  UnfoldStateAll. auto.

  inversion_clear H.
  assert (prim_entry _UNBLK); auto.
  contradiction.
Qed.

Lemma WF_AIM_World_unblk2:
forall Psi'0 S T B RdyQ tid p g tag w wqa S',
  WF_AIM_World Psi'0 S T B RdyQ tid rho_ready (Ast p g tag)
  -> enable_unblk w wqa S'
  -> exists BlkQ, 
      B w = Some BlkQ
      /\ Delta w = SomeT wqa 
                    /\ (pmstar pTrue (mstar (ptoW (S_RF S' SP) w)
                                           (mstar wqa
                                                  (stkSpacePrec (S_RF S' SP) (_UNBLK_Stk + 2))
                                           )
                                   )
                     ) S'.
Proof.
  intros.
  unfold enable_unblk in H0.
  InvertAll.
  unfold WF_AIM_World in H.
  InvertAll.
  unfold WF_WQ_WQSpec in H4.
  assert (WQSetMod.in_dom w B).
  apply H4 with wqa; auto.
  unfold WQSetMod.in_dom in H9.
  InvertAll.
  exists x1. auto.
Qed.


Lemma enter_UNBLK_grt_hyp0:
forall sp w cthM1 R fl isr wqa,
  (pTrue *<*
         (sp |->w w *^*
          (wqa *^* stkSpacePrec sp (_UNBLK_Stk + 2))))
          (cthM1, R, fl, isr)
  -> exists cthM11, exists cthM12,
            MemMod.disjUnion cthM11 cthM12 cthM1
            /\ (sp |->w w *^* stkSpacePrec sp (_UNBLK_Stk + 2)) cthM11
            /\ (wqa *^* mTrue) cthM12.
Proof.
  intros.
  assert ((sp |->w w *^* stkSpacePrec sp (_UNBLK_Stk + 2) *^* (wqa *^* mTrue)) cthM1).
  assert ((mTrue *^* (sp |->w w *^* (wqa *^* stkSpacePrec sp (_UNBLK_Stk + 2)))) cthM1).
  unfold pmstar in H.
  InvertAll.
  exists x. exists x0.
  split_l. split; auto.
  sep on.

  sep lift 1 in H0.
  sep matchH H0 with 5.
  unfold mstar at 1 in H0.
  InvertAll.
  exists x. exists x0. split_l.
  split; auto.
Qed.

Lemma enter_UNBLK_grt_hyp1:
forall sp pc w junk cthM11',
  (sp |-> pc *^* sp+1 |-> junk *^* (sp + 2 |->w w) *^*
         stkSpacePrec sp _UNBLK_Stk) cthM11'
  -> exists cthM11'1, exists cthM11'2,
      MemMod.disjUnion cthM11'1 cthM11'2 cthM11'
      /\ ((sp + 2 |->w w) *^* (stkSpacePrec sp _UNBLK_Stk)) cthM11'1
      /\ (sp |-> pc *^* sp+1 |-> junk) cthM11'2.
Proof.
  intros.
  assert ((sp + 2 |->w w *^* stkSpacePrec sp _UNBLK_Stk *^* (sp |-> pc *^* sp + 1 |-> junk)) cthM11').
  sep on.
  sep matchH H with 4.
  unfold mstar at 1 in H0.
  InvertAll.
  exists x. exists x0. split; auto.
Qed.

Lemma split_blk_Q_2:
forall T tid RdyQ B M w BlkQ,
  (embed_curr_TCB T tid *^* (embed_RdyQ T RdyQ *^* embed_WQSet T B)) M
  -> B w = Some BlkQ
  -> ((embed_RdyQ T RdyQ *^* (ptoTQ w (filterTQ T BlkQ))) *^* (embed_curr_TCB T tid *^* embed_WQSet T (WQSetMod.remove B w))) M.
Proof.
  intros.
  assert (((embed_curr_TCB T tid *^* embed_RdyQ T RdyQ *^* (ptoTQ w (filterTQ T BlkQ))) *^* (embed_WQSet T (WQSetMod.remove B w))) M).
  apply split_blk_Q. auto. auto.
  sep on.
  sep matchH H1 with 4.
Qed.


Lemma mem_list_compose_d:
forall M' M21 M22 M2 M12 cthM2 cthM12 cthM11' cthM11'1 cthM11'2,
  MemMod.disjUnion M21 M22 M2
  -> disjUnionList (M2 :: M12 :: cthM2 :: cthM12 :: cthM11' :: nil) M'
  -> MemMod.disjUnion cthM11'1 cthM11'2 cthM11'
  -> exists M1', exists M2', 
           MemMod.disjUnion cthM11'1 M21 M1' 
           /\ disjUnionList (M22 :: M12 :: cthM2 :: cthM12 :: cthM11'2 :: nil) M2'
           /\ MemMod.disjUnion M1' M2' M'.
Proof.
  intros.
  assert (DUL2mstar (M2 :: M12 :: cthM2 :: cthM12 :: cthM11' :: nil) M').
  apply DUL2mstar_left; auto.
  unfold DUL2mstar in H2.
  assert ((eqM M21 *^* eqM M22 *^* 
            (eqM M12 *^* (eqM cthM2 *^* (eqM cthM12 *^* (eqM cthM11' *^* Aemp))))) M').
  unfold mstar at 1 in H2.
  InvertAll.
  exists x. exists x0. split_l. split_r.
  unfold eqM in H2. subst x.
  exists M21. exists M22. unfold eqM.
  split; auto. clear H2.
  assert ((eqM cthM11'1 *^* eqM cthM11'2 *^* 
             (eqM M21 *^* eqM M22 *^* eqM M12 *^* eqM cthM2 *^* eqM cthM12 *^* Aemp)) M').
  assert ((eqM cthM11' *^*
            (eqM M21 *^* eqM M22 *^* eqM M12 *^* eqM cthM2 *^* eqM cthM12 *^* Aemp)) M').
  sep on. sep matchH H3 with 7.
  clear H3.
  unfold mstar at 1 in H2.
  InvertAll.
  exists x. exists x0. split_l. split_r.
  unfold eqM in H2. subst x.
  exists cthM11'1. exists cthM11'2. unfold eqM. auto.
  clear H3.
  assert ((eqM cthM11'1 *^* eqM M21 *^*
             (eqM M22 *^* eqM M12 *^* eqM cthM2 *^* eqM cthM12 *^* eqM cthM11'2 *^*
                 Aemp)) M').
  sep on.
  sep matchH H2 with 8. clear H2.
  unfold mstar at 1 in H3. InvertAll.
  exists x.
  exists x0.
  split.
  unfold mstar in H3. unfold eqM in H3.
  InvertAll.
  subst x1 x2. auto.
  split_r.
  assert (DUL2mstar (M22 :: M12 :: cthM2 :: cthM12 :: cthM11'2 :: nil) x0).
  unfold DUL2mstar.
  sep on.
  sep matchH H5 with 6.
  apply DUL2mstar_right; auto.
Qed.


Lemma mem_compose_c_unblk:
forall cthM11'1 M21 M1' sp T RdyQ w junk BlkQ,
  MemMod.disjUnion cthM11'1 M21 M1'
  -> ((sp + 2 |-> w) *^* (sp + 3 |-> junk) *^*
         stkSpacePrec sp _UNBLK_Stk) cthM11'1
  -> (embed_RdyQ T RdyQ *^* ptoTQ w (filterTQ T BlkQ)) M21
  -> UNBLK_MPre T RdyQ BlkQ sp w junk M1'.
Proof.
  intros.
  unfold UNBLK_MPre.
  assert (((sp + 2 |-> w *^* sp + 3 |-> junk *^* stkSpacePrec sp _UNBLK_Stk) *^* 
             (embed_RdyQ T RdyQ *^* ptoTQ w (filterTQ T BlkQ))) M1').
  exists cthM11'1. exists M21.
  split; auto.
  sep on.
  sep matchH H2 with 5.
Qed.


Lemma enter_UNBLK_grt_hyp3:
forall S0 pc junk cthM11'2 M22 M12 cthM2 cthM12 M2' M0',
  (S_RF S0 SP |-> pc *^* S_RF S0 SP+1 |-> junk) cthM11'2
  -> disjUnionList (M22 :: M12 :: cthM2 :: cthM12 :: cthM11'2 :: nil) M2'
  -> MemMod.disjUnion M0' M2' (S_Mem S0)
  -> Ret_Addr pc S0.
Proof.
  intros.
  induction S0 as (((M0, R0), fl0), isr0).
  unfold S_Mem in H1.
  simplS in H.
  unfold Ret_Addr.
  intro.
  simplS.
  apply nextPC_retw with (R0 SP); auto.
  assert ((R0 SP |->w pc) cthM11'2).
  apply ptoW_intro with junk; auto.
  assert (DUL2mstar (M22 :: M12 :: cthM2 :: cthM12 :: cthM11'2 :: nil) M2').
  apply DUL2mstar_left; auto.
  assert ((eqM M0' *^* DUL2mstar (M22 :: M12 :: cthM2 :: cthM12 :: cthM11'2 :: nil)) M0).
  exists M0'. exists M2'.
  split_l. split_r. unfold eqM. auto.
  clear H3.
  unfold DUL2mstar in H4. clear H.
  assert (forall M, eqM cthM11'2 M -> (R0 SP |->w pc) M).
  intros.
  unfold eqM in H. subst M. auto.
  sep on.
  sep lift 5 in H4.
  sep weaker_subst in H4 with (R0 SP |->w pc).
  sep on.
  sep trivial.
Qed.

Lemma enter_UNBLK_grt_hyp4:
forall m1 m2 m3 m4 m5 M0',
  (m1 *^* m2 *^* m3 *^* m4 *^* m5) M0'
  -> exists M0'1, exists M0'2, 
            MemMod.disjUnion M0'1 M0'2 M0'
            /\ (m1 *^* m2) M0'1
            /\ (m3 *^* m4 *^* m5) M0'2.
Proof.
  intros.
  assert ((m1 *^* m2 *^* (m3 *^* m4 *^* m5)) M0').
  sep on.
  sep matchH H with 5.
  unfold mstar at 1 in H0.
  InvertAll.
  exists x. exists x0.
  split; auto.
Qed.

Lemma enter_UNBLK_grt_hyp5:
forall S0 S' M0' M2' pc' M22 M12 cthM2 cthM12 cthM11'2 M0'1 M0'2,
  MemMod.disjUnion M0' M2' (S_Mem S0)
  -> nextS pc' (comms _retw) S0 S'
  -> disjUnionList (M22 :: M12 :: cthM2 :: cthM12 :: cthM11'2 :: nil) M2'
  -> MemMod.disjUnion M0'1 M0'2 M0'
  -> exists M'1, exists M'2,
         MemMod.disjUnion M0'1 M22 M'2
         /\ disjUnionList (M12 :: cthM2 :: cthM12 :: M0'2 :: cthM11'2 :: nil) M'1
         /\ MemMod.disjUnion M'1 M'2 (S_Mem S').
Proof.
  intros.
  induction S0 as (((Mx, Rx), flx), isrx).
  induction S' as (((My, Ry), fly), isry).
  UnfoldStateAll.
  destruct (nextS_inversion H0).
  clear H4 H0 Ry fly isry.
  subst My.
  assert (DUL2mstar (M22 :: M12 :: cthM2 :: cthM12 :: cthM11'2 :: nil) M2').
  apply DUL2mstar_left; auto.
  assert ((eqM M0'1 *^* eqM M0'2 *^* DUL2mstar (M22 :: M12 :: cthM2 :: cthM12 :: cthM11'2 :: nil)) Mx).
  exists M0'. exists M2'.
  split_l. split_r. exists M0'1. exists M0'2.
  unfold eqM. auto.
  clear H0.
  unfold DUL2mstar in H3.
  assert ((eqM M0'1 *^* eqM M22 *^* (eqM M12 *^* (eqM cthM2 *^* (eqM cthM12 *^* (eqM M0'2 *^* (eqM cthM11'2 *^* Aemp)))))) Mx).
  sep on.
  sep matchH H3 with 8. clear H3.
  unfold mstar at 1 in H0.
  InvertAll.
  exists x0. exists x.
  split.
  unfold mstar in H0.
  unfold eqM in H0.
  InvertAll.
  subst x1 x2. auto.
  split.
  assert (DUL2mstar (M12 :: cthM2 :: cthM12 :: M0'2 :: cthM11'2 :: nil) x0).
  unfold DUL2mstar.
  auto.
  apply DUL2mstar_right. auto.
  auto with PMap.
Qed.

Lemma enter_UNBLK_grt_hyp6:
forall cthM2 cthM12 M0'2 cthM11'2 cthM',
  disjUnionList (cthM2 :: cthM12 :: M0'2 :: cthM11'2 :: nil) cthM'
  -> MemMod.disjUnion (MemMod.merge cthM12 (MemMod.merge M0'2 cthM11'2)) cthM2
     cthM'.
Proof.
  intros.
  assert (DUL2mstar (cthM2 :: cthM12 :: M0'2 :: cthM11'2 :: nil) cthM').
  apply DUL2mstar_left; auto.
  unfold DUL2mstar in H0.
  unfold mstar at 1 in H0.
  InvertAll.
  unfold eqM in H0.
  subst x.
  unfold mstar at 1 in H3.
  InvertAll.
  unfold eqM in H2. subst x.
  inversion_clear H0.
  subst x0.
  unfold mstar at 1 in H4.
  InvertAll.
  unfold eqM in H3. subst x.
  inversion_clear H0.
  subst x1.
  unfold mstar in H5.
  InvertAll.
  unfold Aemp in H6. subst x1.
  assert (x = x0). auto with PMap.
  subst x0.
  unfold eqM in H4. subst x. auto with PMap.
Qed.


Lemma step_call_hyp_2:
forall pc f M R fl isr M' R' fl' isr' cthM11 cthM12 cthM1 n w junk,
  nextS pc (comml (_call f)) (M, R, fl, isr) (M', R', fl', isr')
  -> (R SP |->w w *^* stkSpacePrec (R SP) (n + 2)) cthM11
  -> MemMod.disjUnion cthM11 cthM12 cthM1
  -> nextS pc (comml (_call f)) (cthM1, R, fl, isr)
     (MemMod.merge cthM12 (MemUpdW cthM11 (R SP - 2) (pc + 1) junk), R', fl', isr').
Proof.
  intros.
  assert (MemMod.disjUnion cthM12 (MemUpdW cthM11 (R SP - 2) (pc + 1) junk) (MemUpdW cthM1 (R SP - 2) (pc+1) junk)).
  assert (MemMod.in_dom (R SP - 2) cthM11 /\ MemMod.in_dom (R SP - 2 + 1) cthM11).
  sep on.
  sep lift 1 in H0.
  sep stksplit in H0.
  sep lift 1 in H0.
  apply ptoxW_indom1 with (R SP |->w w *^* stkSpacePrec (R SP - 2) (n + 2 - 2)); auto.
  inversion_clear H2.
  unfold MemUpdW.
  apply MemMod.update_presv_disjUnion_right.
  apply MemMod.update_presv_disjUnion_right; auto.
  auto with PMap.
  unfold MemMod.in_dom. unfold MemMod.put.
  induction (MemMod.beq_A (R SP - 2 + 1) (R SP - 2)).
  exists (pc+1). auto. auto.
  inversion_clear H2.
  rewrite <- H4.
  clear H3 H4.
  destruct (nextS_inversion H).
  InvertAll.
  subst fl' isr'. subst R'.
  apply nextS_call.
  unfold pushW.
  split.
  assert ((R SP |->w w *^* stkSpacePrec (R SP) (n + 2) *^* mTrue) cthM1).
  exists cthM11. exists cthM12.
  unfold mTrue. split; auto.
  sep on.
  sep lift 1 in H3.
  sep stksplit 2 in H3. sep lift 0 in H3.
  sep trivial.
  split; auto.
  apply MemUpdW_MemUpd16 with junk; auto.
Qed.


Lemma enter_UNBLK_grt_auxb0_hyp5:
forall pc pc' S0 S' cthM2 cthM12 M0'2 cthM11'2 cthM',
  nextS pc' (comms _retw) S0 S'
  -> (S_RF S0 SP |->w pc) cthM11'2
  -> disjUnionList (cthM2 :: cthM12 :: M0'2 :: cthM11'2 :: nil) cthM'
  -> nextS pc' (comms _retw)
     (S_Mem_Upd S0 (MemMod.merge cthM12 (MemMod.merge M0'2 cthM11'2)))
     (S_Mem_Upd S' (MemMod.merge cthM12 (MemMod.merge M0'2 cthM11'2))).
Proof.
  intros.
  induction S0 as (((Mx, Rx), flx), isrx).
  induction S' as (((My, Ry), fly), isry).
  simplS.
  unfold S_RF in H0.
  destruct (nextS_inversion H).
  InvertAll.
  subst Ry fly isry.
  apply nextS_retw.
  exists pc.
  unfold popW.
  split_r.
  assert ((eqM cthM12 *^* (eqM M0'2 *^* eqM cthM11'2)) (MemMod.merge cthM12 (MemMod.merge M0'2 cthM11'2))).
  assert (DUL2mstar (cthM2 :: cthM12 :: M0'2 :: cthM11'2 :: nil) cthM').
  apply DUL2mstar_left; auto.
  unfold DUL2mstar in H3.
  unfold mstar at 1 in H3. InvertAll.
  assert ((eqM cthM12 *^* (eqM M0'2 *^* (eqM cthM11'2 *^* Aemp))) x1); auto.
  unfold mstar at 1 in H6.
  InvertAll.
  unfold eqM in H6. subst x2.
  unfold mstar at 1 in H10. InvertAll.
  unfold eqM in H9. subst x2. inversion_clear H6.
  subst x3.
  unfold mstar in H11.
  InvertAll.
  unfold Aemp in H12. subst x3.
  assert (x2 = x4). auto with PMap.
  subst x4.
  unfold eqM in H10. subst x2.
  inversion_clear H8.
  subst x1.
  sep on.
  sep matchH H7 with 3.
  assert ((eqM cthM11'2 *^* (eqM cthM12 *^* eqM M0'2)) 
            (MemMod.merge cthM12 (MemMod.merge M0'2 cthM11'2))).
  sep on. sep matchH H3 with 3. clear H3.
  assert (forall M, eqM cthM11'2 M -> (Rx SP |->w pc) M).
  intros.
  unfold eqM in H3. subst M. auto.
  sep on.
  sep weaker_subst in H5 with (Rx SP |->w pc).
  sep on.
  sep trivial.
Qed.


Lemma enter_UNBLK_grt_hyp8:
forall S' cthM'1 cthM'2 cthM11'2 cthM12 cthM2 cthM' M0'2 fl,
  NH_Inv (S_Flag S') cthM'2
  -> fl _IF = S_Flag S' _IF
  -> MemMod.disjUnion cthM'1 cthM'2 cthM'
  -> disjUnionList (cthM2 :: cthM12 :: M0'2 :: cthM11'2 :: nil) cthM'
  -> NH_Inv fl cthM2
  -> (cthM2 = cthM'2) /\ (cthM'1 = MemMod.merge cthM12 (MemMod.merge M0'2 cthM11'2)).
Proof.
  intros.
  assert (exists cthM'1', MemMod.disjUnion cthM2 cthM'1' cthM' /\ disjUnionList (cthM12 :: M0'2 :: cthM11'2 :: nil) cthM'1').
  unfold disjUnionList in *.
  inversion_clear H2.
  exists x.
  inversion_clear H4. split; auto.
  InvertAll. rename x into cthM'1'.
  assert (cthM2 = cthM'2).
  unfold NH_Inv in *.
  rewrite <- H0 in H.
  induction (fl _IF).
  assert (precise (INV0 *^* INV1)).
  apply mstar_presv_prec; auto.
  apply Prec0. apply Prec1.
  unfold precise in H5.
  apply H5 with cthM'.
  apply MemMod.disjUnion_subseteq_left with cthM'1'; auto.
  apply MemMod.disjUnion_subseteq_right with cthM'1; auto.
  auto. auto.
  assert (precise Aemp).
  auto. unfold precise in H5.
  apply H5 with cthM'; auto.
  apply MemMod.disjUnion_subseteq_left with cthM'1'; auto.
  apply MemMod.disjUnion_subseteq_right with cthM'1; auto.
  split_l.
  assert (cthM'1' = cthM'1).
  rewrite <- H5 in H1.
  apply MemMod.disjUnion_unique with cthM2 cthM2 cthM'; auto.
  auto with PMap.
  rewrite <- H7.
  unfold disjUnionList in H6.
  InvertAll.
  subst x1.
  assert (cthM11'2 = x0). auto with PMap.
  subst x0.
  inversion_clear H10.
  subst x.
  inversion_clear H9. auto.
Qed.


Lemma enter_UNBLK_grt_hypa:
forall M'2 M0'1 M22 T RdyQ w BlkQ B  tid,
  MemMod.disjUnion M0'1 M22 M'2
  -> (embed_RdyQ T RdyQ *^* ptoTQ w (filterTQ T BlkQ)) M0'1
  -> (embed_curr_TCB T tid *^* embed_WQSet T (WQSetMod.remove B w)) M22
  -> B w = Some BlkQ
  -> (embed_curr_TCB T tid *^* (embed_RdyQ T RdyQ *^* embed_WQSet T B)) M'2.
Proof.
  intros.
  assert ((embed_RdyQ T RdyQ *^* ptoTQ w (filterTQ T BlkQ) *^*
              (embed_curr_TCB T tid *^* embed_WQSet T (WQSetMod.remove B w))) M'2).
  exists M0'1. exists M22. auto.
  sep on.
  sep lift 2 in H3.
  sep match_left H3; auto.
  sep lift 1 in H3.
  sep match_left H3; auto.
  sep off.
  assert (embed_WQSet T (WQSetMod.put (WQSetMod.remove B w) w BlkQ) M).
  unfold mstar in H3. InvertAll.
  apply wq_cons with x0 x; auto with PMap.
  assert (B = WQSetMod.put (WQSetMod.remove B w) w BlkQ).
  apply WQSetMod.extensional_eqMap.
  intro.
  unfold WQSetMod.put.
  induction (WQSetMod.beq_A x w).
  subst x. auto.
  unfold WQSetMod.remove.
  induction (WQSetMod.beq_A x w).
  contradiction. auto.
  rewrite H5. auto.
Qed.


Lemma enter_UNBLK_grt_aux5:
forall M12 cthM2 cthM12 M0'2 cthM11'2 M'1,
  disjUnionList (M12 :: cthM2 :: cthM12 :: M0'2 :: cthM11'2 :: nil) M'1
  -> exists cthM', MemMod.disjUnion M12 cthM' M'1 
                        /\ disjUnionList (cthM2 :: cthM12 :: M0'2 :: cthM11'2 :: nil) cthM'.
Proof.
  intros.
  unfold disjUnionList in H.
  inversion H.
  exists x. inversion H0.
  split; auto.
Qed.


Lemma enter_UNBLK_grt_auxb0_hyp21:
forall cthM11 cthM11' cthM12 sp w pc junk1 cthM11'1 cthM11'2 cthM11'1_1 cthM11'1_2,
  MemMod.disjUnion cthM11'1 cthM11'2 cthM11'
  -> MemMod.disjoint cthM11 cthM12
  -> cthM11' = MemUpdW cthM11 sp pc junk1
  -> (sp + 2 |->w w *^* stkSpacePrec (sp + 2) (_UNBLK_Stk + 2)) cthM11
  -> MemMod.disjUnion cthM11'1_1 cthM11'1_2 cthM11'1
  -> MemMod.disjUnion cthM11'1_1
     (MemMod.merge (MemMod.merge cthM11'1_2 cthM11'2) cthM12)
     (MemMod.merge cthM12 cthM11').
Proof.
  intros.
  assert (MemMod.disjoint cthM11' cthM12).
  rewrite H1. unfold MemUpdW.
  assert (MemMod.in_dom sp cthM11 /\ MemMod.in_dom (sp + 1) cthM11).
  sep on.
  sep lift 1 in H2.
  sep stksplit in H2.
  sep lift 1 in H2.
  apply ptoxW_indom1 with (sp + 2 |->w w *^* stkSpacePrec sp _UNBLK_Stk); auto. 
  inversion_clear H4.
  apply MemMod.update_presv_disj_left; auto.
  apply MemMod.update_presv_disj_left; auto.
  unfold MemMod.put. unfold MemMod.in_dom.
  induction (MemMod.beq_A (sp + 1) sp).
  exists pc; auto. auto.
  assert ((eqM cthM11'1_1 *^* (eqM cthM11'1_2 *^* eqM cthM11'2 *^* eqM cthM12)) (MemMod.merge cthM12 cthM11')).
  sep lift 3.
  exists cthM12. exists cthM11'.
  split. split; auto with PMap.
  split. unfold eqM. auto.
  sep lift 2.
  exists cthM11'2. exists cthM11'1.
  split; auto with PMap.
  split. unfold eqM. auto.
  exists cthM11'1_1. exists cthM11'1_2. unfold eqM.
  auto.
  unfold mstar at 1 in H5.
  InvertAll.
  unfold eqM in H5. subst x.
  unfold mstar in H8.
  InvertAll.
  unfold eqM in H7, H11. subst x2 x3.
  unfold eqM in H9. subst x1.
  inversion_clear H8.
  subst x.
  inversion_clear H5.
  subst x0.
  auto.
Qed.
  
Lemma enter_UNBLK_grt_auxb0_hyp22:
forall cthM' cthM2 cthM12 M0'2 cthM11'2 M0'2_1 M0'2_2,
  disjUnionList (cthM2 :: cthM12 :: M0'2 :: cthM11'2 :: nil) cthM'
  -> MemMod.disjUnion M0'2_1 M0'2_2 M0'2
  -> MemMod.disjUnion M0'2_1
     (MemMod.merge (MemMod.merge M0'2_2 cthM11'2) cthM12)
     (MemMod.merge cthM12 (MemMod.merge M0'2 cthM11'2)).
Proof.
  intros.
  assert (exists Mx, MemMod.disjUnion cthM2 Mx cthM' /\ disjUnionList (cthM12 :: M0'2 :: cthM11'2 :: nil) Mx).
  unfold disjUnionList in H.
  inversion_clear H.
  exists x. inversion_clear H1.
  unfold disjUnionList. auto.
  InvertAll. rename x into Mx.
  assert (disjUnionList (cthM12 :: M0'2 :: cthM11'2 :: nil) Mx). auto.
  assert ((eqM M0'2_1 *^* (eqM M0'2_2 *^* eqM cthM11'2 *^* eqM cthM12)) Mx).
  assert (DUL2mstar (cthM12 :: M0'2 :: cthM11'2 :: nil) Mx).
  apply DUL2mstar_left; auto.
  unfold DUL2mstar in H4.
  sep on.
  sep lift 1 in H4.
  assert ((eqM M0'2_1 *^* eqM M0'2_2) M0'2).
  exists M0'2_1. exists M0'2_2. unfold eqM. auto.
  assert (forall M, eqM M0'2 M -> (eqM M0'2_1 *^* eqM M0'2_2) M).
  intros.
  unfold eqM in H6. subst M. auto.
  sep weaker_subst in H4 with (eqM M0'2_1 *^* eqM M0'2_2).
  sep on.
  sep matchH H4 with 5.
  unfold disjUnionList in H2.
  InvertAll.
  rewrite H2 in H8.
  assert (cthM11'2 = x0). auto with PMap.
  subst x0.
  inversion_clear H7.
  subst x.
  inversion_clear H6.
  rewrite <- H9.
  unfold mstar in H4.
  unfold eqM in H4. InvertAll.
  subst x4 x5 x3 x.
  inversion_clear H12. subst x2.
  inversion_clear H10. subst x0. auto.
Qed.

Lemma enter_UNBLK_grt_auxb0_hyp23:
forall cthM' cthM2 cthM12 M0'2 cthM11'2 sp pc junk1,
  disjUnionList (cthM2 :: cthM12 :: M0'2 :: cthM11'2 :: nil) cthM'
  -> (sp |-> pc + 1 *^* sp + 1 |-> junk1) cthM11'2
  -> (sp |~>w pc+1) (MemMod.merge cthM12 (MemMod.merge M0'2 cthM11'2)).
Proof.
  intros.
  assert (cthM11'2 ||- sp |->w pc+1).
  apply ptoW_intro with junk1; auto.
  assert (exists Mx, MemMod.disjUnion cthM2 Mx cthM' /\ disjUnionList (cthM12 :: M0'2 :: cthM11'2 :: nil) Mx).
  unfold disjUnionList in H.
  inversion_clear H.
  exists x. inversion_clear H2.
  unfold disjUnionList. auto.
  InvertAll. rename x into Mx.
  assert (disjUnionList (cthM12 :: M0'2 :: cthM11'2 :: nil) Mx). auto.
  assert ((eqM cthM12 *^* eqM M0'2 *^* eqM cthM11'2) Mx).
  assert (DUL2mstar (cthM12 :: M0'2 :: cthM11'2 :: nil) Mx).
  apply DUL2mstar_left; auto.
  unfold DUL2mstar in H5.
  sep on.
  sep matchH H5 with 3; auto.
  assert (forall M, eqM cthM11'2 M -> (sp |->w pc + 1) M).
  intros.
  unfold eqM in H6. subst M. auto.
  sep on.
  sep lift 2 in H5.
  sep weaker_subst in H5 with (sp |->w pc + 1).
  unfold disjUnionList in H3.
  InvertAll.
  rewrite H3 in H11.
  assert (cthM11'2 = x0). auto with PMap.
  subst x0.
  inversion_clear H10.
  subst x.
  inversion_clear H9.
  rewrite <- H12.
  sep on.
  sep trivial.
Qed.


Lemma enter_UNBLK_grt_auxb0:
forall cthM1 R fl isr M cthM11 cthM12 S' pc pc' M' R' fl' isr' S0 wqa w M0'2 cthM11'2 cthM2 cthM' cthM11' cthM11'1 junk1 junk2,
  MemMod.disjUnion cthM11 cthM12 cthM1 
  -> (R SP |->w w *^* stkSpacePrec (R SP) (_UNBLK_Stk + 2)) cthM11
  -> cthM11' = MemUpdW cthM11 (R' SP) (pc + 1) junk1
  -> (R' SP + 2 |-> w *^* R' SP + 3 |-> junk2 *^* stkSpacePrec (R' SP) 10) M0'2
  -> (R' SP + 2 |-> w *^* R' SP + 3 |-> junk2 *^*
         stkSpacePrec (R' SP) _UNBLK_Stk) cthM11'1
  -> (R' SP |-> pc + 1 *^* R' SP + 1 |-> junk1) cthM11'2
  -> MemMod.disjUnion cthM11'1 cthM11'2 cthM11'
  -> nextS pc (comml (_call _UNBLK)) (M, R, fl, isr) (M', R', fl', isr')
  -> nextS pc' (comms _retw) S0 S'
  -> S_RF S0 SP = R' SP
  -> isr = S_ISR S'
  -> fl _IF = S_Flag S' _IF
  -> disjUnionList (cthM2 :: cthM12 :: M0'2 :: cthM11'2 :: nil) cthM'
  -> S_RF S0 AX = 0
  -> wrap_fun (gunblk wqa) pc _UNBLK pc' (cthM1, R, fl, isr)
      (S_Mem_Upd S' (MemMod.merge cthM12 (MemMod.merge M0'2 cthM11'2))).
Proof.
  intros.
  unfold wrap_fun.
  unfold conjG.
  exists (S_Mem_Upd S0 (MemMod.merge cthM12 (MemMod.merge M0'2 cthM11'2))).
  split.
  exists (MemMod.merge cthM12 (MemUpdW cthM11 (R SP - 2) (pc+1) junk1), R', fl', isr').
  split.
  apply step_call_hyp_2 with M M' _UNBLK_Stk w; auto.
  unfold gunblk. exists (R' SP). split.
  assert (R SP - 2 = R' SP).
  apply enter_SW_grt_auxb0_hyp0 with pc M fl isr M' fl' isr' _UNBLK; auto.
  rewrite H13.
  rewrite <- H1.
  assert (S_RF S' SP = S_RF S0 SP + 2).
  apply enter_SW_grt_auxb0_hyp1 with pc'; auto.
  rewrite H8 in H14.
  assert (S_RF
           (S_Mem_Upd S0 (MemMod.merge cthM12 (MemMod.merge M0'2 cthM11'2))) = S_RF S0).
  UnfoldStateAll. auto.
  rewrite H15.

  unfold prepost.
  induction S0 as (((M0, R0), fl0), isr0).
  simplS. simplS in H12. simplS in H15. rewrite H12. clear H7 H8 H12 H15 M0 R0 fl0 isr0.
  assert (R SP >= 2).
  apply ge_trans with (_UNBLK_Stk + 2); auto with arith.
  sep on.
  sep lift 1 in H0.
  apply stkSpacePrec_sp_ge_sz1 with cthM11 (R SP |->w w); auto.
  assert (R SP = R' SP + 2).
  rewrite <- H13. assert (R SP - 2 + 2 = R SP). auto. auto.

  sep on.
  sep lift 2 in H2.
  sep lift 2 in H3.
  destruct H3 as [cthM11'1_1 [cthM11'1_2 [Hsplit_11'1 [HM11'1_1 HM11'1_2]]]].
  destruct H2 as [M0'2_1 [M0'2_2 [Hsplit0'2 [HM0'2_1 HM0'2_2]]]].
  exists cthM11'1_1.
  exists (MemMod.merge (MemMod.merge cthM11'1_2 cthM11'2) cthM12).
  exists M0'2_1.
  split.
  exists MemMod.emp. exists cthM11'1_1.
  split; auto with PMap. split_r.
  unfold choice. left. unfold Aemp.
  split; auto.
  split; auto.
  rewrite H8 in H0.

  split.
  apply enter_UNBLK_grt_auxb0_hyp21 with cthM11 (R' SP) w (pc + 1) junk1 cthM11'1; auto.
  inversion_clear H; auto.
  assert (cthM11'1_2 = M0'2_2).
  assert (R' SP + 2 + 1 = R' SP + 3). omega.
  apply enter_SW_grt_auxb0_hyp26 with (R' SP + 2) w junk2; auto.
  rewrite H2. auto. rewrite H2. auto. rewrite H2.
  apply enter_UNBLK_grt_auxb0_hyp22 with cthM' cthM2; auto.
  induction S0 as (((M0, R0), fl0), isr0).
  simplS.
  unfold S_RF in H8.
  induction S' as (((Mx, Rx), flx), isrx).
  UnfoldStateAll.
  destruct (nextS_inversion H6).
  InvertAll.
  subst fl' isr'.
  destruct (nextS_inversion H7).
  InvertAll. subst fl0 isr0.
  auto.

  induction S0 as (((M0, R0), fl0), isr0).
  induction S' as (((Mx, Rx), flx), isrx).
  simplS.
  destruct (nextS_inversion H7). InvertAll.
  subst flx isrx.
  apply nextS_retw.
  exists (pc+1).
  unfold popW. split; auto.
  unfold S_RF in H8.  rewrite H8.
  apply enter_UNBLK_grt_auxb0_hyp23 with cthM' cthM2 junk1; auto.
Qed.

Lemma enter_UNBLK_grt_aux7:
forall (p p' : Pred) cthM1 R fl isr M cthM11 cthM12 S' pc pc' M' R' fl' isr' S0 wqa w M0'2 cthM11'2 cthM2 cthM' cthM11' cthM11'1 junk1 junk2,
  p (cthM1, R, fl, isr)
  -> (forall pc' : nat, pimp (wrap_fun (gunblk wqa) pc _UNBLK pc' (cthM1, R, fl, isr)) p')
  -> MemMod.disjUnion cthM11 cthM12 cthM1 
  -> (R SP |->w w *^* stkSpacePrec (R SP) (_UNBLK_Stk + 2)) cthM11
  -> cthM11' = MemUpdW cthM11 (R' SP) (pc + 1) junk1
  -> (R' SP + 2 |-> w *^* R' SP + 3 |-> junk2 *^* stkSpacePrec (R' SP) 10) M0'2
  -> (R' SP + 2 |-> w *^* R' SP + 3 |-> junk2 *^*
         stkSpacePrec (R' SP) _UNBLK_Stk) cthM11'1
  -> (R' SP |-> pc + 1 *^* R' SP + 1 |-> junk1) cthM11'2
  -> MemMod.disjUnion cthM11'1 cthM11'2 cthM11'
  -> nextS pc (comml (_call _UNBLK)) (M, R, fl, isr) (M', R', fl', isr')
  -> nextS pc' (comms _retw) S0 S'
  -> S_RF S0 SP = R' SP
  -> isr = S_ISR S'
  -> fl _IF = S_Flag S' _IF
  -> disjUnionList (cthM2 :: cthM12 :: M0'2 :: cthM11'2 :: nil) cthM'
  -> S_RF S0 AX = 0
  -> p' (S_Mem_Upd S' (MemMod.merge cthM12 (MemMod.merge M0'2 cthM11'2))).
Proof.
  intros.
  unfold pimp in H0.
  apply H0 with pc'.
  apply enter_UNBLK_grt_auxb0 with M cthM11 M' R' fl' isr' S0 w cthM2 cthM' cthM11' cthM11'1 junk1 junk2; auto.
Qed.

Lemma enable_unblk_unique_args:
forall S w w' wqa wqa', enable_unblk w wqa S -> enable_unblk w' wqa' S -> (w = w' /\ wqa = wqa').
Proof.
  unfold enable_unblk.
  intros.
  InvertAll.
  induction S as (((M, R), fl), isr).
  UnfoldStateAll.
  assert ((mTrue *^* 
          (R SP |->w w' *^*
             (wqa' *^* stkSpacePrec (R SP) (_UNBLK_Stk + 2)))) M).
  unfold pmstar in H2.
  InvertAll. exists x. exists x0. split; auto.
  assert ((mTrue *^* 
          (R SP |->w w *^*
             (wqa *^* stkSpacePrec (R SP) (_UNBLK_Stk + 2)))) M).
  unfold pmstar in H3.
  InvertAll. exists x. exists x0. split; auto.
  assert (w = w').
  sep on.
  sep lift 1 in H4.
  sep lift 1 in H.
  apply ptoW_ptoW_eq1 with M (R SP) 
           (wqa *^* stkSpacePrec (R SP) (_UNBLK_Stk + 2) *^* mTrue)
           (wqa' *^* stkSpacePrec (R SP) (_UNBLK_Stk + 2) *^* mTrue); auto.
  subst w. split_l.
  rewrite H0 in H1. inversion H1. auto.
Qed.



Lemma enter_UNBLK_grt2_hyp0:
forall T tidR RdyQ tid BlkQ B w Psi'0 TM rho,
  WQMod.in_dom tidR BlkQ
  -> B w = Some BlkQ
  -> WF_TWQ T B RdyQ
  -> WF_WQSet Psi'0 T B TM
  -> WF_currTID tid T B RdyQ rho
  -> exists tcbR, TQMod.maps_to T tidR tcbR /\ WQMod.not_in_dom tidR RdyQ /\ tidR <> tid.
Proof.
  intros.
  unfold WF_TWQ in H1.
  InvertAll.
  unfold WF_WQSet in H2.
  assert (exists tcb : Thread,
           exists M : Mem,
           TQMod.maps_to T tidR tcb /\
           ThrdMemMod.maps_to TM tidR M /\ WF_Wait_Thrd Psi'0 M tcb w).
  apply H2 with BlkQ; auto.
  InvertAll. rename x into tcbR.
  exists tcbR. split_l.
  split.
  unfold disj_RQ_WQ in H6.
  assert (WQMod.disjoint BlkQ RdyQ).
  apply H6 with w; auto.
  unfold WQMod.disjoint in H8.
  assert (WQMod.not_in_dom tidR BlkQ \/ WQMod.not_in_dom tidR RdyQ).
  auto. inversion_clear H10; auto.
  unfold WQMod.not_in_dom in H11.
  unfold WQMod.in_dom in H. rewrite H11 in H.
  inversion H. inversion H10.
  (* tidR <> tid *)
  unfold WF_currTID in H3.
  InvertAll.
  unfold unblocked in H8.
  assert (tidR = tid \/ tidR <> tid).
  apply nat_eq_neq_dec.
  inversion_clear H13; auto.
  rewrite <- H15 in H8.
  assert (WQMod.not_in_dom tidR BlkQ).
  apply H8 with w; auto.
  absurd (~ WQMod.in_dom tidR BlkQ); auto with PMap.
Qed.


Lemma enter_UNBLK_grt2_hyp1:
forall tid tidR T T' B B' RdyQ RdyQ' BlkQ BlkQ' w rho tcbR,
  WF_currTID tid T B RdyQ rho
  -> T' = TQMod.put T tidR tcbR
  -> B' = WQSetMod.put B w BlkQ'
  -> BlkQ' = WQMod.remove BlkQ tidR
  -> RdyQ' = addThrd RdyQ tidR
  -> B w = Some BlkQ
  -> WQMod.not_in_dom tidR RdyQ
  -> WQMod.in_dom tidR BlkQ
  -> WF_currTID tid T' B' RdyQ' rho.
Proof.
  intros.
  unfold WF_currTID.
  split.
Lemma removeB_presv_unblocked:
  forall tid B BlkQ tidR w,
    unblocked tid B -> B w = Some BlkQ -> unblocked tid (WQSetMod.put B w (WQMod.remove BlkQ tidR)).
Proof.
  unfold unblocked.
  intros.
  unfold WQSetMod.maps_to in H1.
  unfold WQSetMod.put in H1.
  induction (WQSetMod.beq_A w0 w).
  inversion H1.
  unfold WQMod.not_in_dom.
  unfold WQMod.remove.
  induction (WQMod.beq_A tid tidR); auto.
  assert (WQMod.not_in_dom tid BlkQ).
  apply H with w; auto. auto.
  apply H with w0; auto.
Qed.
  subst B'.
  subst BlkQ'.
  unfold WF_currTID in H.
  InvertAll.
  apply removeB_presv_unblocked; auto.

  assert (tidR <> tid).

Lemma currTID_not_blocked_TID:
forall B BlkQ tidR tid w,
  unblocked tid B -> B w = Some BlkQ -> WQMod.in_dom tidR BlkQ -> tidR <> tid.
Proof.
  intros.
  assert (tidR = tid \/ tidR <> tid).
  apply nat_eq_neq_dec.
  inversion_clear H2; auto.
  subst tid.
  unfold unblocked in H.
  assert (WQMod.not_in_dom tidR BlkQ).
  apply H with w; auto.
  absurd (~ WQMod.in_dom tidR BlkQ); auto with PMap.
Qed.
  apply currTID_not_blocked_TID with B BlkQ w; auto.
  unfold WF_currTID in H.
  InvertAll. auto.

  split.
  subst RdyQ'.
  unfold addThrd. unfold WQMod.not_in_dom.
  unfold WQMod.put.
  induction (WQMod.beq_A tid tidR).
  absurd (tid <> tidR); auto.
  unfold WF_currTID in H.
  InvertAll. auto.

  split.
  subst T'.
  unfold TQMod.put. unfold TQMod.in_dom.
  induction (TQMod.beq_A tid tidR).
  absurd (tid <> tidR); auto.
  unfold WF_currTID in H.
  InvertAll. auto.
  unfold WF_currTID in H.
  InvertAll.
  exists x.
  split; auto.
  subst T'.
  unfold TQMod.put. unfold TQMod.maps_to.
  induction (TQMod.beq_A tid tidR); auto.
  absurd (tid <> tidR); auto.
Qed.


Lemma enter_UNBLK_grt2_hyp2:
forall T T' RdyQ RdyQ' tid tidR tcbR,
  exists_Idle T RdyQ tid
  -> tidR <> tid
  -> WQMod.not_in_dom tidR RdyQ
  -> RdyQ' = addThrd RdyQ tidR
  -> T' = TQMod.put T tidR tcbR
  -> WQMod.not_in_dom tidR RdyQ
  -> exists_Idle T' RdyQ' tid.
Proof.
  intros.
  unfold exists_Idle in *.
  inversion_clear H.
  left. InvertAll.
  exists x. split; auto.
  subst T'.
  unfold TQMod.put. unfold TQMod.maps_to.
  induction (TQMod.beq_A tid tidR); auto.
  absurd (tid <> tidR); auto.

  InvertAll. right.
  exists x. exists x0.
  assert (x <> tidR).
  assert (x = tidR \/ x <> tidR). apply nat_eq_neq_dec.
  inversion H6; auto.
  subst x.
  absurd (~ WQMod.in_dom tidR RdyQ); auto with PMap.
  split.
  subst RdyQ'.
  unfold addThrd.
  unfold WQMod.in_dom.  unfold WQMod.put.
  induction (WQMod.beq_A x tidR); auto.
  contradiction.
  split; auto.
  subst T'.
  unfold TQMod.maps_to. unfold TQMod.put.
  induction (TQMod.beq_A x tidR). contradiction.
  auto.
Qed.


Lemma enter_UNBLK_grt2_hyp3:
forall tidR T T' B B' RdyQ RdyQ' BlkQ BlkQ' w tcbR,
  WF_TWQ T B RdyQ
  -> TQMod.in_dom tidR T
  -> T' = TQMod.put T tidR tcbR
  -> B' = WQSetMod.put B w BlkQ'
  -> BlkQ' = WQMod.remove BlkQ tidR
  -> RdyQ' = addThrd RdyQ tidR
  -> B w = Some BlkQ
  -> WQMod.in_dom tidR BlkQ
  -> WF_TWQ T' B' RdyQ'.
Proof.
  unfold WF_TWQ.
  intros.
  InvertAll.
  split.
  unfold WF_TQ.
  unfold WF_TQ in H7.
  intros.
  rewrite H1 in H8.
  apply H7.
  unfold TQMod.in_dom in H8.
  unfold TQMod.put in H8.
  induction (TQMod.beq_A tid tidR).
  subst tid. auto. auto.
  split.
  unfold WF_WQSet2.
  subst B'.
  unfold WQSetMod.maps_to.
  unfold WQSetMod.put.
  intros.
  induction (WQSetMod.beq_A l w).
  inversion H2. subst Q.
  induction ( WQSetMod.beq_A l' w).
  rewrite <- a0 in a. contradiction.
  subst BlkQ'.
  unfold WF_WQSet2 in H.
  assert (WQMod.disjoint BlkQ Q').
  apply H with w l'; auto.
  unfold WQMod.disjoint in H3.
  unfold WQMod.disjoint.
  intro x.
  assert (WQMod.not_in_dom x BlkQ \/ WQMod.not_in_dom x Q').
  auto.
  inversion_clear H11.
  left.
  unfold WQMod.not_in_dom.
  unfold WQMod.remove.
  induction (WQMod.beq_A x tidR). auto.
  auto. right. auto.

  induction ( WQSetMod.beq_A l' w).
  inversion H8. subst Q'. clear H8.
  subst BlkQ'.
  unfold WF_WQSet2 in H.
  assert (WQMod.disjoint Q BlkQ).
  apply H with l w; auto.
  unfold WQMod.disjoint in H3.
  unfold WQMod.disjoint.
  intro x.
  assert (WQMod.not_in_dom x Q \/ WQMod.not_in_dom x BlkQ).
  auto.
  inversion_clear H8.
  left. auto.
  right.
  unfold WQMod.not_in_dom.
  unfold WQMod.remove.
  induction (WQMod.beq_A x tidR). auto.
  auto.
  unfold WF_WQSet2 in H.
  apply H with l l'; auto.

  unfold disj_RQ_WQ in *.
  intros.
  rewrite H2 in H8.
  unfold WQSetMod.maps_to in H8.  unfold WQSetMod.put in H8.
  induction (WQSetMod.beq_A l w).
  inversion H8. subst Q. clear H8.
  assert (WQMod.disjoint BlkQ RdyQ).
  apply H9 with w; auto.
  unfold WQMod.disjoint.
  unfold WQMod.not_in_dom.
  intro x.
  rewrite H3. rewrite H4.
  unfold addThrd.
  unfold WQMod.disjoint in H8.
  assert (WQMod.not_in_dom x BlkQ \/ WQMod.not_in_dom x RdyQ).
  auto.
  inversion_clear H10.
  left.
  unfold WQMod.remove.
  induction (WQMod.beq_A x tidR); auto.
  unfold WQMod.put.
  induction (WQMod.beq_A x tidR).
  left.
  unfold WQMod.remove.
  induction (WQMod.beq_A x tidR); auto.
  contradiction. right. auto.
  unfold WQMod.disjoint.
  intro.
  rewrite H4.
  unfold addThrd.
  unfold WQMod.not_in_dom. unfold WQMod.put.
  induction (WQMod.beq_A x tidR).
  left.

  unfold WF_WQSet2 in H.
  assert (WQMod.disjoint Q BlkQ).
  apply H with l w; auto.
  unfold WQMod.disjoint in H10.
  assert (WQMod.not_in_dom x Q \/ WQMod.not_in_dom x BlkQ).
  auto. inversion_clear H11; auto.
  subst x.
  absurd (~ WQMod.in_dom tidR BlkQ); auto with PMap.
  assert (WQMod.disjoint Q RdyQ). apply H9 with l; auto.
  unfold WQMod.disjoint in H10.
  assert (WQMod.not_in_dom x Q \/ WQMod.not_in_dom x RdyQ); auto.
Qed.


Lemma updateB_presv_WQSpec:
forall B B' w BlkQ' Delta,
  B' = WQSetMod.put B w BlkQ'
  -> WQSetMod.in_dom w B
  -> WF_WQ_WQSpec B Delta
  -> WF_WQ_WQSpec B' Delta.
Proof.
  unfold WF_WQ_WQSpec.
  intros.
  rewrite H.
  unfold WQSetMod.in_dom.  unfold WQSetMod.put.
  induction (WQSetMod.beq_A l w).
  exists BlkQ'; auto.
  assert (WQSetMod.in_dom l B).
  apply H1 with wqa. auto.
  auto.
Qed.

Lemma inB_indom_TM:
forall Psi'0 T B TM w BlkQ tidR,
  WF_WQSet Psi'0 T B TM
  -> B w = Some BlkQ
  -> WQMod.in_dom tidR BlkQ
  -> ThrdMemMod.in_dom tidR TM.
Proof.
  intros.
  unfold WF_WQSet in H.
  assert (exists tcb : Thread,
            exists M : Mem,
              TQMod.maps_to T tidR tcb /\
              ThrdMemMod.maps_to TM tidR M /\ WF_Wait_Thrd Psi'0 M tcb w).
  apply H with BlkQ; auto.
  InvertAll.
  unfold ThrdMemMod.in_dom.
  exists x0. auto.
Qed.



Lemma enter_UNBLK_grt2_hyp4:
forall M12 cthM2 cthM12 M0'2 cthM11'2 M'1 cthM121 cthM122 M122 rthM,
  MemMod.disjUnion cthM121 cthM122 cthM12
  -> MemMod.disjUnion M122 rthM M12
  -> disjUnionList (M12 :: cthM2 :: cthM12 :: M0'2 :: cthM11'2 :: nil) M'1
  -> exists cthM', exists M12',
            MemMod.disjUnion M12' cthM' M'1 
            /\ disjUnionList (M122 :: cthM121 :: rthM :: nil) M12'
            /\ disjUnionList (cthM2 :: cthM122 :: M0'2 :: cthM11'2 :: nil) cthM'.
Proof.
  intros.
  assert (DUL2mstar (M12 :: cthM2 :: cthM12 :: M0'2 :: cthM11'2 :: nil) M'1).
  apply DUL2mstar_left.
  auto.
  unfold DUL2mstar in H2.
  assert ((eqM M122 *^* eqM rthM *^*
           (eqM cthM2 *^*
             (eqM cthM12 *^* (eqM M0'2 *^* (eqM cthM11'2 *^* Aemp))))) M'1).
  unfold mstar at 1 in H2.
  InvertAll.
  exists x. exists x0.
  split_l. split_r. unfold eqM in H2. subst x.
  exists M122. exists rthM. unfold eqM. auto. clear H2.
  sep on.
  sep lift 3 in H3.
  assert ((eqM cthM121 *^* eqM cthM122 *^* 
           (eqM M0'2 *^* (eqM cthM11'2 *^* Aemp) *^*
            (eqM M122 *^* eqM rthM *^* eqM cthM2))) M'1).
  sep off.
  unfold mstar at 1 in H3. InvertAll.
  exists x. exists x0. split_l. split_r.
  unfold eqM in H3. subst x.
  exists cthM121. exists cthM122. unfold eqM. auto.
  clear H3.
  sep on.
  assert (M'1 ||- (eqM M122 *^* (eqM cthM121 *^* (eqM rthM *^* Aemp))) 
                  *^* (eqM cthM2 *^* (eqM cthM122 *^* (eqM M0'2 *^* (eqM cthM11'2 *^* Aemp))))).
  sep matchH H2 with 9.
  clear H2.
  sep off. unfold mstar at 1 in H3.
  InvertAll.
  exists x0. exists x.
  split_l.
  split.
  cut (DUL2mstar (M122 :: cthM121 :: rthM :: nil) x).
  intro. apply DUL2mstar_right; auto.
  unfold DUL2mstar. auto.
  cut (DUL2mstar (cthM2 :: cthM122 :: M0'2 :: cthM11'2 :: nil) x0).
  intro. apply DUL2mstar_right; auto.
  unfold DUL2mstar. auto.
Qed.




Lemma enter_UNBLK_grt2_hyp5:
forall TM' M'1 M12' cthM' tid tidR M122 cthM121 rthM TM,
  MemSplit (ThrdMemMod.remove (ThrdMemMod.remove TM tid) tidR) M122
  -> MemMod.disjUnion M12' cthM' M'1
  -> disjUnionList (M122 :: cthM121 :: rthM :: nil) M12'
  -> TM' = ThrdMemMod.put (ThrdMemMod.put TM tidR (MemMod.merge cthM121 rthM))
           tid cthM'
  -> tidR <> tid
  -> MemSplit TM' M'1.
Proof.
  intros.
  rewrite H2.
  apply enter_SW_grt_hyp6 with M12'; auto.
  assert (ThrdMemMod.remove (ThrdMemMod.put TM tidR (MemMod.merge cthM121 rthM)) tid
           = ThrdMemMod.put (ThrdMemMod.remove TM tid) tidR (MemMod.merge cthM121 rthM)).
  auto with PMap.
  rewrite H4. clear H4.
  apply enter_SW_grt_hyp6 with M122; auto.
  unfold disjUnionList in H1.
  inversion_clear H1.
  inversion_clear H4.
  InvertAll.
  subst x1. assert (rthM = x0). auto with PMap.
  subst x0.
  inversion_clear H6.
  subst x.
  auto.
Qed.


Lemma enter_UNBLK_grt2_hyp8:
forall cthM121 rthM spR pcR M122 M12',
  disjUnionList (M122 :: cthM121 :: rthM :: nil) M12'
  -> (spR |~>w pcR) rthM
  -> (spR |~>w pcR) (MemMod.merge cthM121 rthM).
Proof.
  intros.
  assert (MemMod.disjUnion cthM121 rthM (MemMod.merge cthM121 rthM)).
  unfold disjUnionList in H.
  InvertAll.
  subst x1. assert (rthM = x0). auto with PMap.
  subst x0.
  split; auto.
  inversion_clear H3; auto.
  apply leadsToW_Mono with rthM cthM121; auto.
  auto with PMap.
Qed.


Lemma mstar_mwand_elim:
forall m1 m2 p M R fl isr M',
  ((m1 *^* m2) -* p) (M, R, fl, isr)
  -> m2 M'
  -> MemMod.disjoint M' M
  -> (m1 -* p) (MemMod.merge M' M, R, fl, isr).
Proof.
  unfold pm_mwand.
  intros.
  UnfoldStateAll.
  assert ((eqM M0 *^* eqM M' *^* eqM M) M'0).
  sep on.
  sep lift 0.
  exists M0. exists (MemMod.merge M' M).
  unfold eqM at 1. split_l. split_l.
  exists M'. exists M. unfold eqM. split; auto.
  split; auto.
  apply H with (MemMod.merge M0 M'); auto.
  unfold mstar in H4.
  unfold eqM in H4.
  InvertAll.
  subst x0 x1 x2.
  inversion_clear H6.
  subst x. auto.
  exists M0. exists M'. split; auto.
  unfold mstar in H4. unfold eqM in H4.
  InvertAll.
  subst x0 x1 x2.
  inversion_clear H6.
  split; auto.
Qed.

Lemma enter_UNBLK_grt2_hyp9:
forall M122 cthM121 rthM M12',
  disjUnionList (M122 :: cthM121 :: rthM :: nil) M12'
  -> MemMod.disjoint cthM121 rthM.
Proof.
  intros.
  unfold disjUnionList in H.
  InvertAll.
  subst x1.
  assert (rthM = x0). auto with PMap.
  subst x0.
  inversion H2; auto.
Qed.


Lemma enter_UNBLK_grt2_hypa:
forall T T' B B' TM TM' Psi'0 tid tidR tcbR M1 M2 w BlkQ BlkQ',
  T' = TQMod.put T tidR tcbR
  -> TM' = ThrdMemMod.put (ThrdMemMod.put TM tidR M1)
           tid M2
  -> B' = WQSetMod.put B w BlkQ'
  -> BlkQ' = WQMod.remove BlkQ tidR
  -> B w = Some BlkQ
  -> unblocked tid B
  -> WF_WQSet Psi'0 T B TM
  -> WQMod.in_dom tidR BlkQ
  -> WF_WQSet2 B
  -> WF_WQSet Psi'0 T' B' TM'.
Proof.
  unfold WF_WQSet.
  intros.
  rewrite H1 in H8.
  unfold WQSetMod.maps_to in H8. unfold WQSetMod.put in H8.
  induction (WQSetMod.beq_A l w).
  inversion H8. subst Q.
  subst l. clear H8.
  assert (tid0 = tidR \/ tid0 <> tidR).
  apply nat_eq_neq_dec.
  inversion H8.
  subst tid0. clear H8.
  rewrite H2 in H9.
  unfold WQMod.in_dom in H9. unfold WQMod.remove in H9.
  induction (WQMod.beq_A tidR tidR).
  inversion H9. inversion H8.
  absurd (tidR = tidR); auto. clear H8.

  assert (exists tcb : Thread,
           exists M : Mem,
           TQMod.maps_to T tid0 tcb /\
           ThrdMemMod.maps_to TM tid0 M /\ WF_Wait_Thrd Psi'0 M tcb w).
  apply H5 with BlkQ; auto.
  rewrite H2 in H9.
  unfold WQMod.in_dom in H9. unfold WQMod.remove in H9.
  induction (WQMod.beq_A tid0 tidR).
  contradiction. auto.
  InvertAll.
  exists x. exists x0.
  split.
  rewrite H.
  unfold TQMod.maps_to. unfold TQMod.put.
  induction (TQMod.beq_A tid0 tidR); auto.
  contradiction.

  split.
  subst TM'.
  unfold ThrdMemMod.put. unfold ThrdMemMod.maps_to.
  induction (ThrdMemMod.beq_A tid0 tid).
  unfold unblocked in H4.
  assert (WQMod.not_in_dom tid BlkQ).
  apply H4 with w; auto.
  rewrite <- a in H0.
  assert (WQMod.not_in_dom tid0 BlkQ').
  subst BlkQ'.
  unfold WQMod.remove. unfold WQMod.not_in_dom.
  induction (WQMod.beq_A tid0 tidR); auto.
  absurd (~ WQMod.in_dom tid0 BlkQ'); auto with PMap.
  induction (ThrdMemMod.beq_A tid0 tidR); auto.
  contradiction. auto.

  assert (exists tcb : Thread,
         exists M : Mem,
           TQMod.maps_to T tid0 tcb /\
           ThrdMemMod.maps_to TM tid0 M /\ WF_Wait_Thrd Psi'0 M tcb l).
  apply H5 with Q; auto.
  InvertAll.
  exists x. exists x0.
  assert (tid0 <> tidR).
  assert (tid0 = tidR \/ tid0 <> tidR).
  apply nat_eq_neq_dec.
  inversion_clear H12; auto.
  unfold WF_WQSet2 in H7.
  assert (WQMod.disjoint BlkQ Q).
  apply H7 with w l; auto.
  assert (WQMod.not_in_dom tid0 BlkQ).
  unfold WQMod.disjoint in H12.
  assert (WQMod.not_in_dom tid0 BlkQ \/ WQMod.not_in_dom tid0 Q). auto.
  inversion_clear H15; auto.
  absurd (~  WQMod.in_dom tid0 Q); auto with PMap.
  subst tid0.
  absurd (~ WQMod.in_dom tidR BlkQ); auto with PMap.
  split.
  subst T'.
  unfold TQMod.maps_to. unfold TQMod.put.
  induction (TQMod.beq_A tid0 tidR); auto.
  contradiction.
  split.
  subst TM'.
  unfold ThrdMemMod.maps_to.
  unfold ThrdMemMod.put.
  induction (ThrdMemMod.beq_A tid0 tid).
  subst tid0.
  unfold unblocked in H4.
  assert (WQMod.not_in_dom tid Q).
  apply H4 with l; auto.
  absurd (~ WQMod.in_dom tid Q); auto with PMap.
  induction (ThrdMemMod.beq_A tid0 tidR); auto.
  contradiction. auto.
Qed.


Lemma enter_UNBLK_grt2_hypb:
forall T T' tid tidR tcbR B B' w BlkQ BlkQ' M22,
  (embed_curr_TCB T tid *^* embed_WQSet T (WQSetMod.remove B w)) M22
  -> T' = TQMod.put T tidR tcbR
  -> tidR <> tid
  -> B w = Some BlkQ
  -> WQMod.in_dom tidR BlkQ
  -> B' = WQSetMod.put B w BlkQ'
  -> WF_WQSet2 B
  -> (embed_curr_TCB T' tid *^* embed_WQSet T' (WQSetMod.remove B' w)) M22.
Proof.
  intros.
  unfold mstar in H.
  InvertAll.
  exists x. exists x0.
  split_l.
  split.
  subst T'.
  unfold embed_curr_TCB.
  unfold embed_curr_TCB in H.
  InvertAll.
  exists x1.
  split; auto.
  unfold TQMod.maps_to. unfold TQMod.put.
  induction (TQMod.beq_A tid tidR); auto.
  absurd (tid <> tidR); auto.
  assert (WQSetMod.remove B' w = WQSetMod.remove B w).
  subst B'.
  auto with PMap.
  rewrite H7.
  rewrite H0.
  apply ext_unblockedT_presv_WQSet; auto.
  unfold unblocked.
  intros.
  unfold WF_WQSet2 in H5.
  assert (w <> w0).
  unfold WQSetMod.maps_to in H9. unfold WQSetMod.remove in H9.
  induction (WQSetMod.beq_A w0 w); auto.
  assert (WQMod.disjoint Q BlkQ).
  apply H5 with w0 w; auto.
  unfold WQSetMod.maps_to in H9. unfold WQSetMod.remove in H9.
  induction (WQSetMod.beq_A w0 w); auto. inversion H9.
  unfold WQMod.disjoint in H11.
  assert (WQMod.not_in_dom tidR Q \/ WQMod.not_in_dom tidR BlkQ).
  auto.  inversion_clear H12; auto.
  absurd (~ WQMod.in_dom tidR BlkQ); auto with PMap.
Qed.
  


Lemma enter_UNBLK_grt2_auxb0_hyp01:
forall cthM1 cthM11 cthM12 cthM11' sp w pc junk1 cthM121 cthM122 cthM11'1 cthM11'2 cthM11'1_1 cthM11'1_2,
  MemMod.disjUnion cthM11 cthM12 cthM1
  -> cthM11' = MemUpdW cthM11 sp (pc + 1) junk1
  -> (sp + 2 |->w w *^* stkSpacePrec (sp + 2) (_UNBLK_Stk + 2)) cthM11
  -> MemMod.disjUnion cthM121 cthM122 cthM12
  -> MemMod.disjUnion cthM11'1 cthM11'2 cthM11'
  -> MemMod.disjUnion cthM11'1_1 cthM11'1_2 cthM11'1
  -> MemMod.disjoint cthM121 cthM11'1_1.
Proof.
  intros.
  assert (MemMod.disjUnion cthM11' cthM12 (MemUpdW cthM1 sp (pc+1) junk1)).
  rewrite H0.
  unfold MemUpdW.
  assert (MemMod.in_dom sp cthM11 /\ MemMod.in_dom (sp+1) cthM11).
  sep on.
  sep lift 1 in H1.
  sep stksplit in H1.
  sep lift 1 in H1.
  apply ptoxW_indom1 with (sp + 2 |->w w *^* stkSpacePrec sp _UNBLK_Stk); auto.
  apply MemMod.update_presv_disjUnion_left.
  apply MemMod.update_presv_disjUnion_left; auto.
  inversion_clear H5; auto.
  inversion_clear H5.
  unfold MemMod.in_dom. unfold MemMod.put.
  induction (MemMod.beq_A (sp + 1) sp); auto.
  exists (pc+1). auto.
  set (cthM1' := (MemUpdW cthM1 sp (pc + 1) junk1)) in *.
  assert (((eqM cthM11'1_1 *^* eqM cthM11'1_2) *^* eqM cthM11'2 *^* (eqM cthM121 *^* eqM cthM122)) cthM1').
  exists cthM11'. exists cthM12.
  split_l. split.
  exists cthM11'1. exists cthM11'2.
  split_l.
  split.
  exists cthM11'1_1. exists cthM11'1_2. unfold eqM. auto.
  unfold eqM. auto.
  exists cthM121. exists cthM122. unfold eqM. auto.
  sep on.
  assert (cthM1' ||- (eqM cthM121 *^* eqM cthM11'1_1 *^* (eqM cthM11'1_2 *^* eqM cthM11'2 *^* eqM cthM122))).
  sep matchH H6 with 5.
  sep off.
  unfold mstar at 1 in H7.
  InvertAll.
  unfold mstar in H7. InvertAll.
  unfold eqM in H7, H12.
  subst x1 x2. inversion H9; auto.
Qed.



Lemma enter_UNBLK_grt2_auxb0_hyp02:
forall cthM1 cthM11 cthM12 cthM11' sp w pc junk1 cthM121 cthM122 cthM11'1 cthM11'2 cthM11'1_1 cthM11'1_2,
  MemMod.disjUnion cthM11 cthM12 cthM1
  -> cthM11' = MemUpdW cthM11 sp (pc + 1) junk1
  -> (sp + 2 |->w w *^* stkSpacePrec (sp + 2) (_UNBLK_Stk + 2)) cthM11
  -> MemMod.disjUnion cthM121 cthM122 cthM12
  -> MemMod.disjUnion cthM11'1 cthM11'2 cthM11'
  -> MemMod.disjUnion cthM11'1_1 cthM11'1_2 cthM11'1
  -> MemMod.disjUnion (MemMod.merge cthM121 cthM11'1_1)
     (MemMod.merge (MemMod.merge cthM11'1_2 cthM11'2) cthM122)
     (MemMod.merge cthM12 cthM11').
Proof.
  intros.
  assert (MemMod.disjUnion cthM11' cthM12 (MemUpdW cthM1 sp (pc+1) junk1)).
  rewrite H0.
  unfold MemUpdW.
  assert (MemMod.in_dom sp cthM11 /\ MemMod.in_dom (sp+1) cthM11).
  sep on.
  sep lift 1 in H1.
  sep stksplit in H1.
  sep lift 1 in H1.
  apply ptoxW_indom1 with (sp + 2 |->w w *^* stkSpacePrec sp _UNBLK_Stk); auto.
  inversion_clear H5.
  apply MemMod.update_presv_disjUnion_left.
  apply MemMod.update_presv_disjUnion_left; auto.
  unfold MemMod.in_dom. unfold MemMod.put.
  induction (MemMod.beq_A (sp + 1) sp); auto.
  exists (pc+1). auto.
  set (cthM1' := (MemUpdW cthM1 sp (pc + 1) junk1)) in *.
  assert (((eqM cthM11'1_1 *^* eqM cthM11'1_2) *^* eqM cthM11'2 *^* (eqM cthM121 *^* eqM cthM122)) cthM1').
  exists cthM11'. exists cthM12.
  split_l. split.
  exists cthM11'1. exists cthM11'2.
  split_l.
  split.
  exists cthM11'1_1. exists cthM11'1_2. unfold eqM. auto.
  unfold eqM. auto.
  exists cthM121. exists cthM122. unfold eqM. auto.
  sep on.
  assert (cthM1' ||- (eqM cthM121 *^* eqM cthM11'1_1 *^* (eqM cthM11'1_2 *^* eqM cthM11'2 *^* eqM cthM122))).
  sep matchH H6 with 5. clear H6.
  sep off.
  assert (MemMod.disjUnion cthM12 cthM11' cthM1'). auto with PMap.
  inversion_clear H6.
  rewrite <- H9.
  unfold mstar at 1 in H7.
  InvertAll.
  unfold mstar in H7. unfold eqM in H7.
  InvertAll.
  subst x1 x2.
  inversion_clear H10. subst x.
  unfold mstar at 1 in H11.
  InvertAll.
  unfold eqM in H13. subst x1.
  unfold mstar in H11.
  unfold eqM in H11.
  InvertAll.
  subst x1 x2.
  inversion_clear H12. subst x.
  inversion_clear H10.
  subst x0. auto.
Qed.


Lemma enter_UNBLK_grt2_auxb0:
forall cthM1 R fl isr M cthM11 cthM12 S' pc pc' M' R' fl' isr' S0 (wqa : MPred) w M0'2 cthM11'2 cthM2 cthM' cthM121 cthM122
       junk1 junk2 cthM11' cthM11'1,
  MemMod.disjUnion cthM11 cthM12 cthM1
  -> (R SP |->w w *^* stkSpacePrec (R SP) (_UNBLK_Stk + 2)) cthM11
  -> cthM11' = MemUpdW cthM11 (R' SP) (pc + 1) junk1
  -> (R' SP + 2 |-> w *^* R' SP + 3 |-> junk2 *^* stkSpacePrec (R' SP) 10) M0'2
  -> (R' SP + 2 |-> w *^* R' SP + 3 |-> junk2 *^*
         stkSpacePrec (R' SP) _UNBLK_Stk) cthM11'1
  -> (R' SP |-> pc + 1 *^* R' SP + 1 |-> junk1) cthM11'2
  -> MemMod.disjUnion cthM11'1 cthM11'2 cthM11'
  -> nextS pc (comml (_call _UNBLK)) (M, R, fl, isr) (M', R', fl', isr')
  -> nextS pc' (comms _retw) S0 S'
  -> S_RF S0 SP = R' SP
  -> isr = S_ISR S'
  -> fl _IF = S_Flag S' _IF
  -> disjUnionList (cthM2 :: cthM122 :: M0'2 :: cthM11'2 :: nil) cthM'
  -> MemMod.disjUnion cthM121 cthM122 cthM12
  -> wqa cthM121
  -> S_RF S0 AX <> 0
  -> wrap_fun (gunblk wqa) pc _UNBLK pc' (cthM1, R, fl, isr)
      (S_Mem_Upd S' (MemMod.merge cthM122 (MemMod.merge M0'2 cthM11'2))).
Proof.
  intros.
  induction S0 as (((Mx, Rx), flx), isrx).
  induction S' as (((My, Ry), fly), isry).
  unfold wrap_fun. UnfoldStateAll.
  unfold conjG.
  exists (MemMod.merge cthM122 (MemMod.merge M0'2 cthM11'2), Rx, flx, isrx).
  split.
  exists (MemMod.merge cthM12 (MemUpdW cthM11 (R SP - 2) (pc+1) junk1), R', fl', isr').
  split.
  apply step_call_hyp_2 with M M' _UNBLK_Stk w; auto.
  unfold gunblk. exists (R' SP). simplS. split.
  assert (R SP - 2 = R' SP).
  apply enter_SW_grt_auxb0_hyp0 with pc M fl isr M' fl' isr' _UNBLK; auto.
  rewrite H15.
  rewrite <- H1.
  assert (S_RF (My, Ry, fly, isry) SP = S_RF (Mx, Rx, flx, isrx) SP + 2).
  apply enter_SW_grt_auxb0_hyp1 with pc'; auto.
  unfold S_RF in H16.
  rewrite H8 in H16.

  unfold prepost.
  simplS. 
  assert (R SP >= 2).
  apply ge_trans with (_UNBLK_Stk + 2); auto with arith.
  sep on.
  sep lift 1 in H0.
  apply stkSpacePrec_sp_ge_sz1 with cthM11 (R SP |->w w); auto.
  assert (R SP = R' SP + 2).
  rewrite <- H15. assert (R SP - 2 + 2 = R SP). auto. auto.

  sep on.
  sep lift 2 in H2.
  sep lift 2 in H3.
  destruct H3 as [cthM11'1_1 [cthM11'1_2 [Hsplit_11'1 [HM11'1_1 HM11'1_2]]]].
  destruct H2 as [M0'2_1 [M0'2_2 [Hsplit0'2 [HM0'2_1 HM0'2_2]]]].
  exists (MemMod.merge cthM121 cthM11'1_1).
  exists (MemMod.merge (MemMod.merge cthM11'1_2 cthM11'2) cthM122).
  exists M0'2_1.
  split.
  exists cthM121. exists cthM11'1_1.
  split. split; auto.

  apply enter_UNBLK_grt2_auxb0_hyp01 
    with cthM1 cthM11 cthM12 cthM11' (R' SP) w pc junk1 
                cthM122 cthM11'1 cthM11'2 cthM11'1_2; auto.
  rewrite H18 in H0; auto.
  split; auto. unfold choice. right. split; auto.
  split_l.

  rewrite H18 in H0.
  split.

  apply enter_UNBLK_grt2_auxb0_hyp02 with cthM1 cthM11 (R' SP) w pc junk1 cthM11'1; auto.

  assert (cthM11'1_2 = M0'2_2).
  assert (R' SP + 2 + 1 = R' SP + 3). omega.
  apply enter_SW_grt_auxb0_hyp26 with (R' SP + 2) w junk2; auto.
  rewrite H2. auto. rewrite H2. auto. rewrite H2.
  apply enter_UNBLK_grt_auxb0_hyp22 with cthM' cthM2; auto.

  destruct (nextS_inversion H6).
  InvertAll.
  subst fl' isr'.
  destruct (nextS_inversion H7).
  InvertAll. subst flx isrx.
  auto.

  destruct (nextS_inversion H7). InvertAll.
  subst flx isrx.
  apply nextS_retw.
  exists (pc+1).
  unfold popW. split; auto.
  rewrite H8.
  apply enter_UNBLK_grt_auxb0_hyp23 with cthM' cthM2 junk1; auto.
Qed.


Lemma enter_UNBLK_grt2_aux7:
forall (p p' : Pred) cthM1 R fl isr M cthM11 cthM12 S' pc pc' M' R' fl' isr' S0 wqa w M0'2 cthM11'2 cthM2 cthM' cthM121 cthM122 
                    junk1 junk2 cthM11' cthM11'1,
  p (cthM1, R, fl, isr)
  -> (forall pc' : nat, pimp (wrap_fun (gunblk wqa) pc _UNBLK pc' (cthM1, R, fl, isr)) p')
  -> MemMod.disjUnion cthM11 cthM12 cthM1 
  -> (R SP |->w w *^* stkSpacePrec (R SP) (_UNBLK_Stk + 2)) cthM11
  -> cthM11' = MemUpdW cthM11 (R' SP) (pc + 1) junk1
  -> (R' SP + 2 |-> w *^* R' SP + 3 |-> junk2 *^* stkSpacePrec (R' SP) 10) M0'2
  -> (R' SP + 2 |-> w *^* R' SP + 3 |-> junk2 *^*
         stkSpacePrec (R' SP) _UNBLK_Stk) cthM11'1
  -> (R' SP |-> pc + 1 *^* R' SP + 1 |-> junk1) cthM11'2
  -> MemMod.disjUnion cthM11'1 cthM11'2 cthM11'
  -> nextS pc (comml (_call _UNBLK)) (M, R, fl, isr) (M', R', fl', isr')
  -> nextS pc' (comms _retw) S0 S'
  -> S_RF S0 SP = R' SP
  -> isr = S_ISR S'
  -> fl _IF = S_Flag S' _IF
  -> disjUnionList (cthM2 :: cthM122 :: M0'2 :: cthM11'2 :: nil) cthM'
  -> MemMod.disjUnion cthM121 cthM122 cthM12
  -> wqa cthM121
  -> S_RF S0 AX <> 0
  -> p' (S_Mem_Upd S' (MemMod.merge cthM122 (MemMod.merge M0'2 cthM11'2))).
Proof.
  intros.
  unfold pimp in H0.
  apply H0 with pc'.
  apply enter_UNBLK_grt2_auxb0 with M cthM11 cthM12 M' R' fl' isr' S0 w cthM2 cthM' cthM121 junk1 junk2 cthM11' cthM11'1; auto.
Qed.


(* >>>>>>>>>>>>>>> *)
Lemma enter_UNBLK_grt:
forall D psi pc theta D' Psi Psi' Psi'0 S S',
  NH_WFcmd psi pc theta (comml (_call _UNBLK))
  -> Minimal_D D
  -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
  -> lift_psi psi rho_ready Psi
  -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
  -> subsetCT Psi' Psi'0
  -> Itp_AIM rho_ready theta Psi'0 S
  -> nextS pc (comml (_call _UNBLK)) S S'
  -> WFST D rho_prim 0 Psi'0 (scap_lift_g UNBLK_grt S').
Proof.
  intros.
  unfold WFST.
  intros.
  unfold scap_lift_g in H7.
  InvertAll.
  induction S' as (((M', R'), fl'), isr').
  induction S as (((M, R), fl), isr).
  unfold S_ISR at 1 in H7.
  unfold S_Flag at 1 in H10.
  induction theta.
  rename n into tag.
  unfold Itp_AIM in H5.
  unfold pmstar in H5.
  InvertAll.
  rename x into T. rename x0 into B. rename x1 into RdyQ. rename x2 into tid.
  rename x3 into M1. rename x4 into M2.

  assert (exists TM, exists cthM, exists M12, exists cthM1, exists cthM2,
        MemMod.disjUnion M12 cthM M1
        /\ MemMod.disjUnion cthM1 cthM2 cthM
        /\ WF_currTID tid T B RdyQ rho_ready
        /\ WF_TWQ T B RdyQ
        /\ exists_Idle T RdyQ tid
        /\ WF_WQ_WQSpec B Delta
        /\ MemSplit (ThrdMemMod.remove TM tid) M12
        /\ ThrdMemMod.maps_to TM tid cthM
        /\ WF_Curr_Thrd Psi'0 (S_Mem_Upd (M1, R, fl, isr) cthM) rho_ready (Ast p g tag) T tid
        /\ WF_ReadyQ Psi'0 T RdyQ TM
        /\ WF_WQSet Psi'0 T B TM
        /\ p (cthM1, R, fl, isr)
        /\ NH_Inv fl cthM2
        /\ (tag = 1 -> NH_WFST rho_ready Psi'0 (NH_ItpG g (cthM, R, fl, isr)))).
  apply decompose_World; auto.
  InvertAll. rename x into TM. rename x0 into cthM. rename x1 into M12. rename x2 into cthM1. rename x3 into cthM2.
  assert (WQMod.not_in_dom tid RdyQ). unfold WF_currTID in H14. InvertAll. auto.
  unfold NH_WFcmd in H.
  inversion_clear H. clear H28.
  induction H27.
  inversion_clear H. clear H27.
  assert (exists w, exists wqa, enable_unblk w wqa (cthM1, R, fl, isr)).
  apply NH_Call_unblk_enable with psi pc p g tag; auto.
  InvertAll. rename x into w. rename x0 into wqa.
  unfold NH_Call_Prim in H28.
  inversion_clear H28.
  inversion H27.
  absurd (_UNBLK <> _SWITCH); auto.
  absurd (_UNBLK <> _BLOCK); auto.
  clear H28. inversion H30.
  rewrite <- H33 in H32.
  rewrite <- H34 in H32.
  rewrite <- H35 in H31.
  clear H30 H33 H34 H35 tag0 p0 g0.
  clear H27.

  assert (exists BlkQ, 
          B w = Some BlkQ
          /\ Delta w = SomeT wqa 
          /\ (pmstar pTrue (mstar (ptoW (S_RF (cthM1, R, fl, isr) SP) w)
                                  (mstar wqa
                                         (stkSpacePrec (S_RF (cthM1, R, fl, isr) SP) (_UNBLK_Stk + 2))
                                  )
                           )
              ) (cthM1, R, fl, isr)).
  apply WF_AIM_World_unblk2 with Psi'0 (M1, R, fl, isr) T RdyQ tid p g tag; auto.
  InvertAll. rename x into BlkQ.
  unfold S_RF in H33.

  assert (exists cthM11, exists cthM12,
            MemMod.disjUnion cthM11 cthM12 cthM1
            /\ (R SP |->w w *^* stkSpacePrec (R SP) (_UNBLK_Stk + 2)) cthM11
            /\ (wqa *^* mTrue) cthM12).
  apply enter_UNBLK_grt_hyp0 with R fl isr; auto.
  InvertAll.
  rename x into cthM11. rename x0 into cthM12.

  assert (disjUnionList (M2 :: M12 :: cthM2 :: cthM12 :: cthM11 :: nil) M).
  apply mem_list_compose_a with cthM1 cthM M1; auto.
  assert 
    (exists junk,
        ((R' SP |-> (pc+1)) *^* (R' SP + 1 |-> junk) *^* (R' SP + 2 |->w w) *^* (stkSpacePrec (R' SP) _UNBLK_Stk)) (MemUpdW cthM11 (R' SP) (pc + 1) junk)
        /\ disjUnionList (M2 :: M12 :: cthM2 :: cthM12 :: (MemUpdW cthM11 (R' SP) (pc + 1) junk) :: nil) M').
  apply call_fun_frame2 with M R fl fl' isr isr' _UNBLK; auto.
  InvertAll. rename x into junk1.
  set (cthM11' := MemUpdW cthM11 (R' SP) (pc + 1) junk1) in *.
  assert
    (exists cthM11'1, exists cthM11'2,
      MemMod.disjUnion cthM11'1 cthM11'2 cthM11'
      /\ ((R' SP + 2 |->w w) *^* (stkSpacePrec (R' SP) _UNBLK_Stk)) cthM11'1
      /\ (R' SP |-> (pc + 1) *^* R' SP + 1 |-> junk1) cthM11'2).
  apply enter_UNBLK_grt_hyp1; auto.
  InvertAll. rename x into cthM11'1. rename x0 into cthM11'2.
  assert (((embed_RdyQ T RdyQ *^* (ptoTQ w (filterTQ T BlkQ))) 
            *^* (embed_curr_TCB T tid *^* embed_WQSet T (WQSetMod.remove B w))) M2).
  apply split_blk_Q_2; auto.
  unfold mstar at 1 in H41.
  InvertAll. rename x into M21. rename x0 into M22.
  assert (exists M1', exists M2', 
           MemMod.disjUnion cthM11'1 M21 M1' 
           /\ disjUnionList (M22 :: M12 :: cthM2 :: cthM12 :: cthM11'2 :: nil) M2'
           /\ MemMod.disjUnion M1' M2' M').
  apply mem_list_compose_d with M2 cthM11'; auto.
  InvertAll.  rename x into M1'. rename x0 into M2'.

  assert (exists tcb, T tid = Some tcb).
  unfold WF_currTID in H14.
  InvertAll.
  exists x. auto. 
  inversion_clear H47. rename x into tcb.
  unfold UNBLK_grt in H8.
  sep on.
  assert (exists junk2, cthM11'1 ||- R' SP + 2 |-> w *^* R' SP + 2 + 1 |-> junk2 *^* stkSpacePrec (R' SP) _UNBLK_Stk).
  apply ptoW_elim1; auto.
  inversion_clear H47. rename x into junk2. sep off.
  clear H38.
  rename H50 into H38.
  assert (R' SP + 2 + 1 = R' SP + 3). omega.
  rewrite H47 in H38. clear H47.
  assert (exists M0' : MemMod.Map,
         MemMod.disjUnion M0' M2' (S_Mem S0) /\
         UNBLK_Post_aux T RdyQ BlkQ (R' SP) (S_RF (M', R', fl', isr') BP) w junk2
           (S_Mem_Upd S0 M0')).
  apply H8 with M1'; auto.
  unfold UNBLK_Prd_aux.
  split. assert (S_Mem (S_Mem_Upd (M', R', fl', isr') M1') = M1'). auto.
  rewrite H47.
  apply mem_compose_c_unblk with cthM11'1 M21; auto.

  split. 
  apply WF_AIM_World_RdyQinT with Psi'0 (S_Mem_Upd (M, R, fl, isr) M1) B tid rho_ready (Ast p g tag).
  auto.
  split.
  apply WF_AIM_World_BlkQinT with Psi'0 (S_Mem_Upd (M, R, fl, isr) M1) B RdyQ tid rho_ready (Ast p g tag) w; auto.
  split; auto.
  split.
  apply WF_AIM_World_RdyNotBlk with Psi'0 (M1, R, fl, isr) T B tid (Ast p g tag) w rho_ready; auto.
  apply WF_AIM_World_WFTQ with Psi'0 (S_Mem_Upd (M, R, fl, isr) M1) B RdyQ tid rho_ready (Ast p g tag); auto.
  clear H8.
  InvertAll.
  rename x into M0'.
  assert ((fl _IF = false) /\ (isr = false)).
  assert (S_Flag (cthM1, R, fl, isr) _IF = false).
  apply H29; auto. split. unfold S_Flag in H8. auto.
  unfold WF_Curr_Thrd in H20.
  assert (beq_nat rho_ready rho_ih = false). simpl. auto.
  rewrite H51 in H20. unfold itp_NH_theta in H20.
  InvertAll. UnfoldStateAll. auto.
  inversion_clear H8.


  exists (pc + 1).
  assert (S_RF S0 SP = R' SP).
  unfold UNBLK_Post_aux in H50. InvertAll. auto.
  split.
  rewrite <- H8 in H42.
  apply enter_UNBLK_grt_hyp3 with junk1 cthM11'2 M22 M12 cthM2 cthM12 M2' M0'; auto.

  exists (packS0 rho_ready (Ast p' g' tag)).
  split.
  apply psi_sw_lookupNPC with psi Psi Psi'; auto.

  intros.
  unfold Itp0.
  assert (D rho_ready = SomeT1 _ (Itp_AIM rho_ready)).
  unfold Minimal_D in H0. InvertAll. auto.
  rewrite H54. clear H54.
  unfold Itp_AIM.
  unfold UNBLK_Post_aux in H50.
  InvertAll.
  clear H54.
  inversion_clear H56. (*two cases:*)

  (* case 1: blkQ is empty *)
  InvertAll.
  exists T. exists B. exists RdyQ. exists tid.
  unfold UNBLK_MPost in H54. unfold UNBLK_MPre in H54.
  assert (S_Mem (S_Mem_Upd S0 M0') = M0'). auto.
  rewrite H56 in H54. clear H56.
  assert (exists M0'1, exists M0'2, 
            MemMod.disjUnion M0'1 M0'2 M0'
            /\ (embed_RdyQ T RdyQ *^* ptoTQ w (filterTQ T BlkQ)) M0'1
            /\ (R' SP + 2 |-> w *^* R' SP + 3 |-> junk2 *^* stkSpacePrec (R' SP) 10) M0'2).
  apply (enter_UNBLK_grt_hyp4 (embed_RdyQ T RdyQ) (ptoTQ w (filterTQ T BlkQ)) (R' SP + 2 |-> w) 
               (R' SP + 3 |-> junk2) (stkSpacePrec (R' SP) 10)); auto.
  InvertAll. rename x into M0'1. rename x0 into M0'2.
  assert (exists M'1, exists M'2,
         MemMod.disjUnion M0'1 M22 M'2
         /\ disjUnionList (M12 :: cthM2 :: cthM12 :: M0'2 :: cthM11'2 :: nil) M'1
         /\ MemMod.disjUnion M'1 M'2 (S_Mem S')).
  apply enter_UNBLK_grt_hyp5 with S0 M0' M2' pc'; auto.
  InvertAll. rename x into M'1. rename x0 into M'2.
  unfold pmstar.
  exists M'1. exists M'2. split; auto.
  split.
  unfold WF_AIM_World.
  split; auto.
  split; auto.
  split; auto.
  split; auto.
  assert (exists cthM', MemMod.disjUnion M12 cthM' M'1 
                        /\ disjUnionList (cthM2 :: cthM12 :: M0'2 :: cthM11'2 :: nil) cthM').
  apply enter_UNBLK_grt_aux5; auto.
  InvertAll.
  rename x into cthM'.
  exists (ThrdMemMod.put TM tid cthM'). exists cthM'.
  split.
  assert (S_Mem (S_Mem_Upd S' M'1) = M'1). auto.
  rewrite H64.
  apply enter_SW_grt_hyp6 with M12; auto.
  split. auto with PMap.

  split.
  assert (S_ISR (M, R, fl, isr) = S_ISR S' /\ S_Flag (M, R, fl, isr) _IF = S_Flag S' _IF).
  apply enter_SW_grt_hyp9 with pc pc' _UNBLK (M', R', fl', isr') S0; auto.
  unfold S_ISR at 1 in H64. unfold S_Flag at 1 in H64.
  inversion_clear H64.
  unfold WF_Curr_Thrd.
  assert (beq_nat rho_ready rho_ih = false).
  apply neq_beq_false. discriminate.
  rewrite H64. clear H64.
  unfold itp_NH_theta.
  split.
  rewrite H52 in H66. rewrite H66. auto.
  split.
  unfold NH_pinvstar.
  unfold pmstar.
  exists (MemMod.merge cthM12 (MemMod.merge M0'2 cthM11'2)).
  exists cthM2. 
  split. assert (MemMod.disjUnion (MemMod.merge cthM12 (MemMod.merge M0'2 cthM11'2)) cthM2 cthM').
  apply enter_UNBLK_grt_hyp6; auto.
  UnfoldStateAll. auto.
  split; auto.

  assert (S_Mem_Upd (S_Mem_Upd (S_Mem_Upd S' M'1) cthM')
        (MemMod.merge cthM12 (MemMod.merge M0'2 cthM11'2)) = S_Mem_Upd S' (MemMod.merge cthM12 (MemMod.merge M0'2 cthM11'2))).
  auto. rewrite H64. clear H64.

  apply enter_UNBLK_grt_aux7 with p cthM1 R fl isr M cthM11 pc pc' M' R' fl' isr' S0 wqa w cthM2 cthM' cthM11' cthM11'1 junk1 junk2; auto.
  assert (exists w : nat,
          exists wqa : WQassert,
            enable_unblk w wqa (cthM1, R, fl, isr) /\
            (forall pc' : nat,
             pimp (wrap_fun (gunblk wqa) pc _UNBLK pc' (cthM1, R, fl, isr)) p') /\
            (forall pc' : nat,
             pimp (wrap_fun (gunblk wqa) pc _UNBLK pc' (cthM1, R, fl, isr) |> g') (g (cthM1, R, fl, isr)))).
  apply H32; auto. InvertAll. 
  assert (w = x /\ wqa = x0). apply enable_unblk_unique_args with (cthM1, R, fl, isr); auto.
  inversion_clear H69. rewrite H72. auto.
  assert (S_Flag (S_Mem_Upd (S_Mem_Upd S' M'1) cthM') = S_Flag S'). auto.
  rewrite H64. unfold NH_Inv. rewrite <- H67. 
  unfold NH_Inv in H24. auto.

  (* WFST *)
  intro Tag1.
  assert (NH_WFST rho_ready Psi'0 (NH_ItpG g (cthM, R, fl, isr))).
  apply H26. auto.
  apply NH_WFST_strengthen with (cthM, R, fl, isr) g; auto.
  unfold pimp.
  intros.
  unfold NH_ItpG.
  assert ((S_Mem_Upd (S_Mem_Upd S' M'1) cthM') = (S_Mem_Upd S' cthM')).
  auto. rewrite H69 in H68. clear H69.

  unfold NH_ItpG in H68.
  InvertAll.
  rename x into cthM'1. rename x0 into cthM'2. rename x1 into SM1. rename x2 into SM2.
  exists cthM1. exists cthM2. exists SM1. exists SM2.
  split. unfold S_Mem. auto.
  split. auto. split. auto. split. auto.
  assert (exists w : nat,
          exists wqa : WQassert,
            enable_unblk w wqa (cthM1, R, fl, isr) /\
            (forall pc' : nat,
             pimp (wrap_fun (gunblk wqa) pc _UNBLK pc' (cthM1, R, fl, isr)) p') /\
            (forall pc' : nat,
             pimp (wrap_fun (gunblk wqa) pc _UNBLK pc' (cthM1, R, fl, isr) |> g') (g (cthM1, R, fl, isr)))).
  apply H32; auto. InvertAll.
  assert (w = x /\ wqa = x0). apply enable_unblk_unique_args with (cthM1, R, fl, isr); auto.
  inversion_clear H75. rewrite <- H78 in H76. clear H77 H78 H72 H74 x x0.
  
  unfold pimp in H76.
  apply H76 with pc'. clear H76.
  unfold linkPG.
  exists (S_Mem_Upd (S_Mem_Upd S' cthM') cthM'1).
  split; auto.
  assert ((cthM2 = cthM'2) /\ (cthM'1 = MemMod.merge cthM12 (MemMod.merge M0'2 cthM11'2))).
  apply enter_UNBLK_grt_hyp8 with S' cthM' fl; auto.
  inversion_clear H72. rewrite H75.
  assert (S_Mem_Upd (S_Mem_Upd S' cthM') (MemMod.merge cthM12 (MemMod.merge M0'2 cthM11'2))
            = S_Mem_Upd S' (MemMod.merge cthM12 (MemMod.merge M0'2 cthM11'2))). auto.
  rewrite H72. clear H72.
  apply enter_UNBLK_grt_auxb0 with M cthM11 M' R' fl' isr' S0 w cthM2 cthM' cthM11' cthM11'1 junk1 junk2; auto.
  split.

  apply enter_SW_grt_hypf with T tcb; auto.
  apply TQMod.extensional_eqMap.
  intros. unfold TQMod.put.
  induction (TQMod.beq_A x tid). rewrite a. auto. auto.
  apply enter_SW_grt_hypg with T tcb; auto.
  unfold WF_currTID in H14. InvertAll. auto.
  apply TQMod.extensional_eqMap.
  intros. unfold TQMod.put.
  induction (TQMod.beq_A x tid). rewrite a. auto. auto.
  apply enter_UNBLK_grt_hypa with M0'1 M22 w BlkQ; auto.

  (* case 2: blkQ is not empty *)
  InvertAll. rename x into tidR. rename x0 into tcbR'.
  assert (exists tcbR, TQMod.maps_to T tidR tcbR /\ WQMod.not_in_dom tidR RdyQ /\ tidR <> tid).
  apply enter_UNBLK_grt2_hyp0 with BlkQ B w Psi'0 TM rho_ready; auto.
  InvertAll.
  induction x as ((spR, statusR), pcR).
  assert (tcbR' = (spR, statusR, pcR)).
  unfold TQMod.maps_to in H57. rewrite H54 in H57.
  inversion H57. auto.
  rewrite H60 in H56. unfold TCB_SP, TCB_PC in H56.
  clear H60 H54 tcbR'.
  exists (TQMod.put T tidR (spR, READY, pcR)).
  set (T' := TQMod.put T tidR (spR, READY, pcR)).
  exists (WQSetMod.put B w (WQMod.remove BlkQ tidR)). 
  set (BlkQ' := WQMod.remove BlkQ tidR).
  set (B' := WQSetMod.put B w BlkQ').
  exists (addThrd RdyQ tidR). set (RdyQ' := addThrd RdyQ tidR).
  exists tid.
  unfold UNBLK_MPost in H56. unfold UNBLK_MPre in H56.
  assert (S_Mem (S_Mem_Upd S0 M0') = M0'). auto.
  rewrite H54 in H56. clear H54.

  assert (exists M0'1, exists M0'2, 
            MemMod.disjUnion M0'1 M0'2 M0'
            /\ (embed_RdyQ T' RdyQ' *^* ptoTQ w (filterTQ T' BlkQ')) M0'1
            /\ (R' SP + 2 |-> w *^* R' SP + 3 |-> junk2 *^* stkSpacePrec (R' SP) 10) M0'2).
  apply (enter_UNBLK_grt_hyp4 (embed_RdyQ T' RdyQ') (ptoTQ w (filterTQ T' BlkQ')) 
           (R' SP + 2 |-> w) (R' SP + 3 |-> junk2) (stkSpacePrec (R' SP) 10)); auto.
  InvertAll. rename x into M0'1. rename x0 into M0'2.
  assert (exists M'1, exists M'2,
         MemMod.disjUnion M0'1 M22 M'2
         /\ disjUnionList (M12 :: cthM2 :: cthM12 :: M0'2 :: cthM11'2 :: nil) M'1
         /\ MemMod.disjUnion M'1 M'2 (S_Mem S')).
  apply enter_UNBLK_grt_hyp5 with S0 M0' M2' pc'; auto.
  InvertAll. rename x into M'1. rename x0 into M'2.
  unfold pmstar.
  exists M'1. exists M'2. split; auto.
  split.
  unfold WF_AIM_World.
  (* WF_currTID *)
  split.
  apply enter_UNBLK_grt2_hyp1 with tidR T B RdyQ BlkQ BlkQ' w (spR, READY, pcR); auto.
  (* exists_Idle *)
  split.
  apply enter_UNBLK_grt2_hyp2 with T RdyQ tidR (spR, READY, pcR); auto.
  (* WF_TWQ *)
  split.
  apply enter_UNBLK_grt2_hyp3 with tidR T B RdyQ BlkQ BlkQ' w (spR, READY, pcR); auto.
    unfold TQMod.in_dom. exists (spR, statusR, pcR). auto.
  (* WF_WQ WQSpec *)
  split.
  apply updateB_presv_WQSpec with B w BlkQ'; auto.
  unfold WQSetMod.in_dom. exists BlkQ. auto.

  (* rest *)
  unfold mstar in H36. InvertAll.
  rename x into cthM121. rename x0 into cthM122.
  assert (ThrdMemMod.in_dom tidR TM).
  apply inB_indom_TM with Psi'0 T B w BlkQ; auto.
  unfold ThrdMemMod.in_dom in H67.
  inversion_clear H67.
  rename x into rthM.
  assert (exists M122,
            MemMod.disjUnion M122 rthM M12
            /\ MemSplit (ThrdMemMod.remove (ThrdMemMod.remove TM tid) tidR) M122).
  apply memsplit_thrd_remove; auto.
    unfold ThrdMemMod.maps_to, ThrdMemMod.remove.
    induction (ThrdMemMod.beq_A tidR tid). contradiction. auto.
  InvertAll.
  rename x into M122.

  assert (exists cthM', exists M12',
            MemMod.disjUnion M12' cthM' M'1 
            /\ disjUnionList (M122 :: cthM121 :: rthM :: nil) M12'
            /\ disjUnionList (cthM2 :: cthM122 :: M0'2 :: cthM11'2 :: nil) cthM').
  apply enter_UNBLK_grt2_hyp4 with M12 cthM12; auto.
  InvertAll.
  rename x into cthM'. rename x0 into M12'.
  exists (ThrdMemMod.put (ThrdMemMod.put TM tidR (MemMod.merge cthM121 rthM)) tid cthM').
  set (TM' := ThrdMemMod.put (ThrdMemMod.put TM tidR (MemMod.merge cthM121 rthM)) tid cthM').
  exists cthM'.
  split.
  assert (S_Mem (S_Mem_Upd S' M'1) = M'1). auto.
  rewrite H73. clear H73.
  apply enter_UNBLK_grt2_hyp5 with M12' cthM' tid tidR M122 cthM121 rthM TM; auto.

  split. unfold TM'.
  auto with PMap.

  (* WF_Curr_Thrd *)
  split.
  assert (S_ISR (M, R, fl, isr) = S_ISR S' /\ S_Flag (M, R, fl, isr) _IF = S_Flag S' _IF).
  apply enter_SW_grt_hyp9 with pc pc' _UNBLK (M', R', fl', isr') S0; auto.
  unfold S_ISR at 1 in H73. unfold S_Flag at 1 in H73.
  inversion_clear H73.
  unfold WF_Curr_Thrd.
  assert (beq_nat rho_ready rho_ih = false). simpl. auto.
  rewrite H73. clear H73.
  unfold itp_NH_theta.
  split.
  rewrite H52 in H75. rewrite H75. auto.
  split.
  unfold NH_pinvstar.
  unfold pmstar.
  exists (MemMod.merge cthM122 (MemMod.merge M0'2 cthM11'2)).
  exists cthM2. 
  split. assert (MemMod.disjUnion (MemMod.merge cthM122 (MemMod.merge M0'2 cthM11'2)) cthM2 cthM').
  apply enter_UNBLK_grt_hyp6; auto.
  UnfoldStateAll. auto.
  split; auto.

  assert (S_Mem_Upd (S_Mem_Upd (S_Mem_Upd S' M'1) cthM')
        (MemMod.merge cthM122 (MemMod.merge M0'2 cthM11'2)) = S_Mem_Upd S' (MemMod.merge cthM122 (MemMod.merge M0'2 cthM11'2))).
  auto. rewrite H73. clear H73.

  apply enter_UNBLK_grt2_aux7 
    with p cthM1 R fl isr M cthM11 cthM12 pc pc' M' R' fl' isr' S0 wqa w cthM2 cthM' cthM121 junk1 junk2 cthM11' cthM11'1; auto.
  assert (exists w : nat,
          exists wqa : WQassert,
            enable_unblk w wqa (cthM1, R, fl, isr) /\
            (forall pc' : nat,
             pimp (wrap_fun (gunblk wqa) pc _UNBLK pc' (cthM1, R, fl, isr)) p') /\
            (forall pc' : nat,
             pimp (wrap_fun (gunblk wqa) pc _UNBLK pc' (cthM1, R, fl, isr) |> g') (g (cthM1, R, fl, isr)))).
  apply H32; auto. InvertAll. 
  assert (w = x /\ wqa = x0). apply enable_unblk_unique_args with (cthM1, R, fl, isr); auto.
  inversion_clear H78. rewrite H81. auto.

  assert (S_Flag (S_Mem_Upd (S_Mem_Upd S' M'1) cthM') = S_Flag S'). auto.
  rewrite H73. unfold NH_Inv. rewrite <- H76. 
  unfold NH_Inv in H24. auto.

  (* WFST *)
  intro Tag1.
  assert (NH_WFST rho_ready Psi'0 (NH_ItpG g (cthM, R, fl, isr))).
  apply H26. auto.
  apply NH_WFST_strengthen with (cthM, R, fl, isr) g; auto.
  unfold pimp.
  intros.
  unfold NH_ItpG.
  assert ((S_Mem_Upd (S_Mem_Upd S' M'1) cthM') = (S_Mem_Upd S' cthM')).
  auto. rewrite H78 in H77. clear H78.

  unfold NH_ItpG in H77.
  InvertAll.
  rename x into cthM'1. rename x0 into cthM'2. rename x1 into SM1. rename x2 into SM2.
  exists cthM1. exists cthM2. exists SM1. exists SM2.
  split. unfold S_Mem. auto.
  split. auto. split. auto. split. auto.
  assert (exists w : nat,
          exists wqa : WQassert,
            enable_unblk w wqa (cthM1, R, fl, isr) /\
            (forall pc' : nat,
             pimp (wrap_fun (gunblk wqa) pc _UNBLK pc' (cthM1, R, fl, isr)) p') /\
            (forall pc' : nat,
             pimp (wrap_fun (gunblk wqa) pc _UNBLK pc' (cthM1, R, fl, isr) |> g') (g (cthM1, R, fl, isr)))).
  apply H32; auto. InvertAll.
  assert (w = x /\ wqa = x0). apply enable_unblk_unique_args with (cthM1, R, fl, isr); auto.
  inversion_clear H84. rewrite <- H87 in H85. clear H87 H86 H81 H83 x x0.

  unfold pimp in H85.
  apply H85 with pc'.
  unfold conjG.
  exists (S_Mem_Upd (S_Mem_Upd S' cthM') cthM'1).
  split; auto.
  assert ((cthM2 = cthM'2) /\ (cthM'1 = MemMod.merge cthM122 (MemMod.merge M0'2 cthM11'2))).
  apply enter_UNBLK_grt_hyp8 with S' cthM' fl; auto.
  inversion_clear H81. rewrite H84.
  assert (S_Mem_Upd (S_Mem_Upd S' cthM') (MemMod.merge cthM122 (MemMod.merge M0'2 cthM11'2))
            = S_Mem_Upd S' (MemMod.merge cthM122 (MemMod.merge M0'2 cthM11'2))). auto.
  rewrite H81. clear H81.
  apply enter_UNBLK_grt2_auxb0 with M cthM11 cthM12 M' R' fl' isr' S0 w cthM2 cthM' cthM121 junk1 junk2 cthM11' cthM11'1; auto.

  (* WF_ReadyQ *)
  split.
  unfold WF_ReadyQ.
  intros.
  assert (tid0 = tidR \/ tid0 <> tidR).
  apply nat_eq_neq_dec.
  inversion_clear H75.

  (* tid0 = tidR *)
  rewrite H76. clear H73 H76 tid0.
  exists (spR, READY, pcR).
  split.
    unfold T'. auto.
  split.
  left. auto.

  exists (MemMod.merge cthM121 rthM).
  split.
    unfold TM'.
    unfold ThrdMemMod.put, ThrdMemMod.maps_to.
    induction (ThrdMemMod.beq_A tidR tid).
        absurd (tidR <> tid); auto.
        induction (ThrdMemMod.beq_A tidR tidR).
          auto.
          absurd (tidR = tidR); auto.

  assert (WF_Wait_Thrd Psi'0 rthM (spR, statusR, pcR) w).
    unfold WF_WQSet in H22.
    assert (exists tcb : Thread,
            exists M : Mem,
              TQMod.maps_to T tidR tcb /\
              ThrdMemMod.maps_to TM tidR M /\ WF_Wait_Thrd Psi'0 M tcb w).
    apply H22 with BlkQ; auto.
    InvertAll.
    unfold TQMod.maps_to in H75, H57.
    rewrite H75 in H57. inversion H57.
    rewrite H78 in H77.
    unfold ThrdMemMod.maps_to in H73.
    rewrite H69 in H73.
    inversion H73.
    auto.

  split.
  unfold WF_Wait_Thrd in H73.
  InvertAll. unfold TCB_SP. unfold TCB_PC.
  apply enter_UNBLK_grt2_hyp8 with M122 M12'; auto.
  left.
  unfold WF_Rdy_Thrd.
  exists rho_ready.
  unfold WF_Wait_Thrd in H73.
  InvertAll.
  rename x into wqa'.
  rename x0 into pw. rename x1 into gw. rename x2 into tagw.
  rewrite H28 in H77. inversion H77.
  rewrite <- H80 in H79. clear H80 H77 wqa'.
  exists pw. exists gw. exists tagw.
  split.
    left. auto.
  split. unfold rho_tcb_consistent.
    split. intro. auto.
           intro. inversion H76.
  split. apply enter_UNBLK_grt2_hyp8 with M122 M12'; auto.
  split; auto.
  intros.
  assert ((INV0 *^* INV1 *^* wqa -*
         itp_NH_theta rho_ready (Ast pw gw tagw) Psi'0) 
          (rthM, R0, fl0, false)).
  apply H79; auto.
  apply (mstar_mwand_elim (INV0 *^* INV1) wqa (itp_NH_theta rho_ready (Ast pw gw tagw) Psi'0)); auto.
  apply enter_UNBLK_grt2_hyp9 with M122 M12'; auto.
  
  (* tid0 <> tidR *)
  unfold WF_ReadyQ in H21.
  assert (exists tcb : Thread,
          TQMod.maps_to T tid0 tcb /\
          (TCB_Status tcb = READY \/ TCB_Status tcb = IDLE) /\
          (exists M : Mem,
             ThrdMemMod.maps_to TM tid0 M /\
             (TCB_SP tcb |~>w TCB_PC tcb) M /\
             (WF_Rdy_Thrd Psi'0 M tcb \/ WF_Rdy_Thrd_IH Psi'0 M tcb))).
  apply H21; auto.
    unfold RdyQ' in H73. unfold addThrd in H73.
    unfold WQMod.in_dom, WQMod.put in H73.
    induction (WQMod.beq_A tid0 tidR). contradiction.
    auto.
  InvertAll.
  exists x.
    split.
    unfold T'.
    unfold TQMod.maps_to, TQMod.put.
    induction (TQMod.beq_A tid0 tidR). contradiction.
    auto.
    split; auto.
    exists x0.
    split; auto.
    unfold TM'.
    unfold ThrdMemMod.maps_to, ThrdMemMod.put.
      induction (ThrdMemMod.beq_A tid0 tid).
      assert (WQMod.not_in_dom tid RdyQ').
        unfold RdyQ'.
        unfold addThrd.
        unfold WQMod.not_in_dom, WQMod.put.
        induction (WQMod.beq_A tid tidR).
          absurd (tid <> tidR); auto.
          unfold WQMod.not_in_dom in H25.
          auto.
      rewrite a in H73.
      absurd (~ WQMod.in_dom tid RdyQ'); auto with PMap.
    induction (ThrdMemMod.beq_A tid0 tidR). contradiction.
    auto.

  (* WF_WQSet *)
  apply enter_UNBLK_grt2_hypa with T B TM tid tidR (spR, READY, pcR) (MemMod.merge cthM121 rthM) cthM' w BlkQ BlkQ';
  auto.
  unfold WF_currTID in H14. InvertAll. auto.
  unfold WF_TWQ in H15. InvertAll; auto.
  apply enter_UNBLK_grt_hypa with M0'1 M22 w BlkQ'; auto.
  apply enter_UNBLK_grt2_hypb with T tidR (spR, READY, pcR) B BlkQ BlkQ'; auto.
  unfold WF_TWQ in H15. InvertAll; auto.
  unfold B'. auto with PMap.

  (* done *)
  inversion H.
  absurd (prim_entry _UNBLK); auto.
Qed.


Lemma enter_UNBLK:
forall D psi pc theta D' Psi Psi' Psi'0 a S S',
  NH_WFcmd psi pc theta (comml (_call _UNBLK))
  -> Minimal_D D
  -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
  -> lift_psi psi rho_ready Psi
  -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
  -> a = Itp_AIM rho_ready theta
  -> subsetCT Psi' Psi'0
  -> a Psi'0 S
  -> nextS pc (comml (_call _UNBLK)) S S'
  -> Itp0 D' UNBLK_Spec0 Psi'0 S'.
Proof.
  intros.
  unfold Itp0.
  unfold UNBLK_Spec0.
  rewrite H1.
  simpl.
  rewrite H4 in H6.
  clear H4 a.
  split.
  apply enter_UNBLK_pre with psi pc theta Psi'0 S; auto.
  split.
  apply enter_UNBLK_disable with psi pc theta Psi'0 S; auto.
  exists 0.
  apply enter_UNBLK_grt with psi pc theta D' Psi Psi' S; auto.
Qed.

Lemma NH_UNBLK_Presv:
  forall D psi pc theta D' Psi Psi' Psi'0 a S S' f',
    NH_WFcmd psi pc theta (comml (_call _UNBLK))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_ready Psi
    -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
    -> a = Itp_AIM rho_ready theta
    -> subsetCT Psi' Psi'0
    -> a Psi'0 S
    -> nextS pc (comml (_call _UNBLK)) S S'
    -> nextPC (comml (_call _UNBLK)) (S_Mem S) (S_RF S) (S_Flag S) pc f'
    -> exists theta0 : Spec0, lookupCT Psi'0 f' theta0 /\ Itp0 D' theta0 Psi'0 S'.
Proof.
  intros.
  assert (lookupCT Psi'0 _UNBLK UNBLK_Spec0).
  apply subsetCT_presv_lookupCT with Psi'; auto.
  rewrite H3.
  unfold ext_IH_Spec, ext_Prim_Specs. 
  apply lookupCT_subsetCT with (consCT _UNBLK UNBLK_Spec0 Psi). 
  apply lookupCT_consCT.
  apply Psi_extend_subsetCT.
  apply Psi_extend2_subsetCT.
  apply Psi_extend_subsetCT.
  apply consCT_subsetCT.
  inversion H8. clear H14 H13 H12 H10 H11 fl R M f pc0.
  exists UNBLK_Spec0.
  split; auto.
  apply enter_UNBLK with D psi pc theta Psi Psi' a S; auto.
Qed.


Lemma call_unblk_disable_itr:
forall psi Psi'0 pc theta S,
  NH_WFcmd psi pc theta (comml (_call _UNBLK)) 
  -> Itp_AIM rho_ready theta Psi'0 S
  -> ~ itr_enabled S.
Proof.
  intros.
  unfold NH_WFcmd in H.
  induction theta.
  InvertAll.
  unfold NH_WFcmd_aux in H1.
  inversion_clear H1.
  unfold NH_Call_Prim in H2.
  InvertAll.
  assert (monotonic_p (fun S => S_Flag S _IF = false)). auto.
  assert (forall S, imprec p S -> (fun S => S_Flag S _IF = false) S).
  apply (imprec_presv_imp_monotonic H4 H5); auto.
  assert (imprec p S).
  apply imprec_presv_imprec with (NH_pinvstar p); auto.
  apply Itp_AIM_pinv with g n Psi'0 rho_ready; auto.
  assert ((fun S0 : State => S_Flag S0 _IF = false) S).
  apply (H6 S H7). intro.
  unfold itr_enabled in H9.
  assert (S_Flag S _IF = false). auto.
  inversion H9. rewrite H11 in H10. inversion H10.
  inversion H2. absurd (prim_entry _UNBLK); auto.
Qed.

Lemma NH_UNBLK_Itp :
  forall D psi pc theta D' Psi Psi',
    NH_WFcmd psi pc theta (comml (_call _UNBLK))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_ready Psi
    -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
    (* -> a = Itp_AIM rho_ready theta *)
    -> WFcmd D' Psi' (packS0 rho_ready theta) pc (comml (_call _UNBLK)).
Proof.
  intros.
  unfold WFcmd.
  split.
  unfold WFcmd_aux.
  intros.
  assert (Itp_AIM rho_ready theta Psi'0 S).
  unfold Itp0 in H4.
  rewrite H1 in H4.
  unfold ext_dict in H4.
  assert (beq_nat rho_ready rho_prim = false). simpl. auto.
  rewrite H6 in H4.
  unfold Minimal_D in H0.
  InvertAll.
  rewrite H7 in H4. auto.
  split.
  apply NH_UNBLK_progress with psi theta Psi'0 (Itp_AIM rho_ready theta); auto.
  split.
  intros.
  apply NH_UNBLK_Presv with D psi pc theta Psi Psi' (Itp_AIM rho_ready theta) S; auto.
  left.
  apply call_unblk_disable_itr with psi Psi'0 pc theta; auto.
  rewrite H1. 
  unfold Itp0.
  unfold ext_dict.
  assert (beq_nat rho_ready rho_prim = false).
  simpl. auto.
  rewrite H4.
  unfold Minimal_D in H0. InvertAll.
  rewrite H5.
  apply Itp_AIM_Mono_CT.
Qed.
