(* This is the SCAP which assumes interrupt is disabled *)
Add LoadPath "..".
Add LoadPath "../CAP0".

Require Import Coqlib.
Require Import Map.
Require Import MultiMapT.
Require Import Arith.
Require Import Machine.
Require Import Assertion.
Require Import Operational_Semantics.
Require Import OpSem_Lemmas.

Require Import CAP0.
Require Import Link.

Module SCAPSpec <: ONE_TYPE.
  Definition B := Spec.
End SCAPSpec.

(* psi : PSpec *)
Module SPSpec := MultiMap SCAPSpec.
Definition SCAP_PSpec := SPSpec.map.

Definition SCAP_emptCT := SPSpec.emp_map.
Definition SCAP_consCT : flab -> Spec -> SCAP_PSpec -> SCAP_PSpec :=
  fun f theta psi => SPSpec.merge (SPSpec.sig_map f theta) psi.

Definition SCAP_lookupCT : SCAP_PSpec -> flab -> Spec -> Prop :=
  fun psi f s =>   SPSpec.lookup psi f s.

Definition Ret_Addr (ra : PC) (S : State) : Prop :=
  forall pc, nextPC (comms _retw) (S_Mem S) (S_RF S) (S_Flag S) pc ra.

Definition NextS_Ret (S S' : State) : Prop :=
  forall pc, nextS pc (comms _retw) S S'.

Definition fun_frame : Grt :=
  fun (S S' : State) => 
    (S_RF S SP = S_RF S' SP) /\ exists ra, Ret_Addr ra S /\ Ret_Addr ra S'. 

Definition sp_call_pred (p : Pred) (pc f: flab) : Pred :=
  p |> (nextS pc (comml (_call f))).


Definition wrap_fun (g : Grt) (pc f pc': flab) : Grt :=
  (nextS pc (comml (_call f))) oo g oo (nextS pc' (comms _retw)).

Definition ZF_st : Pred :=
  fun (S : State) =>  S_Flag S _ZF = true.

Definition ZF_cl : Pred :=
  fun (S : State) => S_Flag S _ZF = false.

Definition conjP (p p' : Pred) : Pred :=
  fun (S : State) => p S /\ p' S.

Definition SCAP_WFcmd (psi : SCAP_PSpec) (pc : flab) (theta : Spec) (c : comm) : Prop :=
  match theta with
    Ast p g tag =>
    match c with
      comms si =>
      match si with
        | _iret => False
        | _retw => forall S, p S -> g S S
        | _jmp f => exists p', exists g',
                      SCAP_lookupCT psi f (Ast p' g' tag) 
                      /\ pimp p p'
                      /\ gimp (p >o g') g
      end
    | comml i  =>
      match i with
        | _cli => False
        | _sti => False
        | _outb_eoi => False
        | _call f => exists p', exists p'', exists g', exists g'',
                       SCAP_lookupCT psi f (Ast p' g' tag)
                       /\ SCAP_lookupCT psi (pc+1) (Ast p'' g'' tag)
                       /\ enable p (nextS pc (comml (_call f)))
                       /\ gimp ((sp_call_pred p pc f) >o g') fun_frame
                       /\ pimp (p |> (nextS pc (comml (_call f)))) p'
                       /\ (forall pc', pimp (p |> (wrap_fun g' pc f pc')) p'')
                       /\ (forall pc', gimp (p >o (wrap_fun g' pc f pc') oo g'') g)
        | _jne f => exists p', exists p'', exists g', exists g'',
                     SCAP_lookupCT psi f (Ast p' g' tag)
                     /\ SCAP_lookupCT psi (pc+1) (Ast p'' g'' tag)
                     /\ pimp (conjP p ZF_cl) p'
                     /\ gimp (p >o (ZF_cl >o g')) g
                     /\ pimp (conjP p ZF_st) p''
                     /\ gimp (p >o (ZF_st >o g'')) g
        | _pushfw => False
        | _popfw => False
        | _ => exists p', exists g',
                 SCAP_lookupCT psi (pc+1) (Ast p' g' tag)
                 /\ enable p (nextS pc c)
                 /\ pimp (p |> (nextS pc c)) p'
                 /\ gimp (p >o (nextS pc c) oo g') g
      end
    end
  end.

Inductive SCAP_WFiseq : SCAP_PSpec -> flab -> Spec -> iseq -> Prop :=
    scap_isq_seq :
        forall psi pc p p' g g' i isq tag,
          SCAP_WFcmd (SCAP_consCT (pc+1) (Ast p' g' tag) psi) pc (Ast p g tag) (comml i)
          -> SCAP_WFiseq psi (pc+1) (Ast p' g' tag) isq
          -> SCAP_WFiseq psi pc (Ast p g tag) (seq i isq)
  | scap_isq_end :
        forall psi pc p g si tag,
          SCAP_WFcmd psi pc (Ast p g tag) (comms si)
          -> SCAP_WFiseq psi pc (Ast p g tag) (sseq si).

Definition SCAP_WFcdhp (psi : SCAP_PSpec) (C: CodeHeap) (psi' : SCAP_PSpec): Prop :=
  forall f theta,
    SCAP_lookupCT psi' f theta -> exists isq, lookupC C f isq /\ SCAP_WFiseq psi f theta isq.

(* scap_lift_g is necessary to link with code handling interrupts *)
(* changed by Xinyu on Dec. 15, 2007 *)
Definition scap_lift_g (g : Grt) :=
  fun (S S' : State) =>
    g S S' /\ (S_ISR S = S_ISR S') /\ (S_Flag S _IF = S_Flag S' _IF).

Fixpoint WFST (D : LDict) (rho : LangID) (n : nat) (Psi : PSpec) (p : Pred) {struct n} : Prop :=
  match n with
    0 => forall S, p S -> exists ra, 
                            Ret_Addr ra S 
                            /\ exists theta0 : Spec0, 
                                 lookupCT Psi ra theta0 
                                 /\ forall pc' S', nextS pc' (comms _retw) S S'
                                                     -> Itp0 D theta0 Psi S'
  | S n' =>
         forall S, p S -> exists ra,
                            Ret_Addr ra S
                            /\ exists p', exists g', exists tag,
                                 lookupCT Psi ra (packS0 rho (Ast p' g' tag))
                                 /\ forall pc' S', nextS pc' (comms _retw) S S' 
                                                    -> p' S' /\ WFST D rho n' Psi ((scap_lift_g g') S')
  end.

Definition Generic_Itp (D : LDict) (rho : LangID) : ITP :=
  fun (theta : Spec) (Psi : PSpec) (S : State) =>
    match theta with
      Ast p g tag => p S /\ S_Flag S _IF = false /\ exists n, WFST D rho n Psi ((scap_lift_g g) S)
    end.

Definition lift_psi (psi : SCAP_PSpec) (rho : LangID) (Psi : PSpec) : Prop :=
  forall f theta, SCAP_lookupCT psi f theta -> lookupCT Psi f (packS0 rho theta).

Definition sink_psi (psi : SCAP_PSpec) (rho : LangID) (Psi : PSpec) : Prop :=
  forall f theta0, lookupCT Psi f theta0 
    -> exists theta : Spec,
      theta0 = (packS0 rho theta) /\ SCAP_lookupCT psi f theta.

Lemma enable_elim :
  forall S p g,
    enable p g 
    -> p S
    -> exists S', g S S'.
Proof.
  intros S p g Hen Hp.
  unfold enable in Hen.
  auto.
Qed.
Implicit Arguments enable_elim [S p g].


Lemma linkPG_intro : forall (p : Pred) (g : Grt) (S S' : State),
  p S -> g S S' -> linkPG p g S'.
Proof.
  intros.
  unfold linkPG.
  exists S; split; trivial.
Qed.
Implicit Arguments linkPG_intro [p g S S'].

Lemma conjG_intro : forall (g g' : Grt) S'' S S',
  g S S'' -> g' S'' S'
  -> conjG g g' S S'.
Proof.
  unfold conjG.
  intros.
  exists S''; split; trivial.
Qed.
Implicit Arguments conjG_intro [g g' S'' S S'].

Lemma conjP_intro : forall (p p' : Pred) S,
  p S -> p' S -> conjP p p' S.
Proof. intros; split; trivial. Qed.

Lemma conjPG_intro : forall (p : Pred) (g :Grt) S S', 
  p S -> g S S' -> conjPG p g S S'.
Proof.
  intros p g S S' Hp Hg; unfold conjPG; split; trivial. 
Qed.
Implicit Arguments conjPG_intro [p g S S'].

Lemma ra_after_call : forall pc ra f S S',
  Ret_Addr ra S'
  -> nextS pc (comml (_call f)) S S'
  -> ra = pc + 1.
Proof.
  intros pc ra f S S'.
  intros Hra HnextS.
  inversion HnextS.
  subst pc0 f0.
  substH Hra with (Hra (pc+1)).
  inversion Hra.
  subst M0 R0 fl0 pc0 pc'.
  subst S S' sp.
  simpl in * .
  inversion H1.
  destruct H2.
  subst R'.
  unfold MemUpd16 in H2.
  destruct H2.
  subst M'.
  unfold leadsToW in H0.
  destruct H0.
  destruct H0.
  destruct H0.
  destruct H2.
  clear Hra.
  clear HnextS.
  clear H1 H3.
  unfold ptoW in H2.
  unfold mstar, pto,ptox,MemMod.disjUnion, MemMod.put in * .
  destruct H2 as [M1 [M2 [H3 [H4 H5]]]].
  destruct H3.
  unfold MemMod.sgltonM, MemMod.maps_to in H4.
  destruct H4.
  subst x0.
  destruct H0.
  unfold RF_update in * .
  unfold beq_reg in * .
  assert (Heq := MemMod.eq_get (R SP - 2) H2).
  simpl in Heq.
  unfold MemMod.beq_A in Heq.
  unfold MemSpec.beq_Domain in Heq.
  destruct (eq_nat_dec (R SP - 2) (R SP - 2 + 1)).
    rewrite plus_comm in e.
    simpl in e.
    destruct (n_Sn (R SP -2) e).
  destruct (eq_nat_dec (R SP - 2) (R SP - 2)).
  unfold MemMod.merge in Heq.
  rewrite H3 in Heq.
  injection Heq;intro;subst; trivial.
  destruct (n0 (refl_equal (R SP - 2))).
Qed.
Implicit Arguments ra_after_call [pc ra f].

Lemma lift_psi_cons : forall psi Psi rho f spec,
  lift_psi psi rho Psi
  -> lift_psi (SCAP_consCT f spec psi) rho (consCT f (packS0 rho spec) Psi).
Proof.
  unfold lift_psi; intros.
  inversion H0.
  injection H1; intros; subst; clear H1.
  compute.
  left; trivial.
  unfold lookupCT, consCT.
  apply OPSpec.lookup_merge_intro.
  right.
  apply H; trivial.
Qed.
Implicit Arguments lift_psi_cons [psi Psi rho f spec].

Lemma psi_cons_a_lookup_a:
forall psi f theta,
  SCAP_lookupCT (SCAP_consCT f theta psi) f theta.
Proof.
  intros.
  unfold SCAP_lookupCT.
  unfold SCAP_consCT.
  unfold SPSpec.lookup.
  simpl.
  left. auto.
Qed.
Hint Resolve psi_cons_a_lookup_a.

Lemma psi_cons_b_lookup_a:
forall psi f theta f' theta',
  SCAP_lookupCT psi f theta
  -> SCAP_lookupCT (SCAP_consCT f' theta' psi) f theta.
Proof.
  unfold SCAP_lookupCT.
  unfold SCAP_consCT.
  unfold SPSpec.lookup.
  intros.
  simpl.
  right; trivial.
Qed.
Hint Resolve psi_cons_b_lookup_a.
