(* This is the SCAP for non-handler code *)
Add LoadPath "..".
Add LoadPath "../CAP0".

Require Import Coqlib.
Require Import Machine.
Require Import Assertion.
Require Import Operational_Semantics.
Require Import SCAP.
Require Import AIM.
Require Import SCAP_READY.

Set Implicit Arguments.
Unset Strict Implicit.

Definition IDLE_Call_Prim (psi : SCAP_PSpec) (pc f: flab) (theta : Spec) : Prop :=
  (NH_Call_Prim psi pc f theta) /\ (f = _SWITCH).


Definition IDLE_WFcmd_aux (psi : SCAP_PSpec) (pc : flab) (theta : Spec) (c : comm) : Prop :=
  match theta with
    Ast p g tag =>
    match c with
      comms si => NH_WFcmd_aux psi pc theta c
    | comml i  =>
      match i with
        _call f => (prim_entry f /\ IDLE_Call_Prim psi pc f theta) 
                     \/ (~ (prim_entry f) /\ NH_Normal_Call psi pc f theta)
      | _  => NH_WFcmd_aux psi pc theta c
      end
    end
  end.

Definition IDLE_WFcmd (psi : SCAP_PSpec) (pc : flab) (theta : Spec) (c : comm) : Prop :=
  match theta with
    Ast p g tag =>
      IDLE_WFcmd_aux psi pc theta c /\ (Prd_Itr_Stk p /\ Grt_Itr_Stk g)
  end.

Inductive IDLE_WFiseq : SCAP_PSpec -> flab -> Spec -> iseq -> Prop :=
    idle_isq_seq :
        forall psi pc p p' g g' i isq tag,
          IDLE_WFcmd (SCAP_consCT (pc+1) (Ast p' g' tag) psi) pc (Ast p g tag) (comml i)
          -> IDLE_WFiseq psi (pc+1) (Ast p' g' tag) isq
          -> IDLE_WFiseq psi pc (Ast p g tag) (seq i isq)
  | idle_isq_end :
        forall psi pc p g si tag,
          IDLE_WFcmd psi pc (Ast p g tag) (comms si)
          -> IDLE_WFiseq psi pc (Ast p g tag) (sseq si).

Definition IDLE_WFcdhp (psi : SCAP_PSpec) (C: CodeHeap) (psi' : SCAP_PSpec): Prop :=
  forall f theta,
    SCAP_lookupCT psi' f theta -> exists isq, lookupC C f isq /\ IDLE_WFiseq psi f theta isq.


