Add LoadPath "..".
Add LoadPath "../CAP0".

Require Import Arith.
Require Import Coqlib.
Require Import Map.
Require Import Machine.
Require Import Assertion.
Require Import Assertion_Lemmas.
Require Import Operational_Semantics.
Require Import OpSem_Lemmas.

Require Import CAP0.
Require Import SCAP.
Require Import AIM.
Require Import QLib.
Require Import SCAP_READY.
Require Import SCAP_IDLE.

Require Import EqNat.

Require Import Linking.
Require Import Linking_Lemmas.
Require Import Linking_Lemmas2.
Require Import SCAP_NH_Lemmas.
Require Import NH_SW.
Require Import READY_BLK.
Require Import READY_UNBLK.

Lemma SCAP_IDLE_in_CAP0_call :
  forall D psi pc theta D' Psi Psi' f,
    IDLE_WFcmd psi pc theta (comml (_call f))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_idle Psi
    -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
    -> WFcmd D' Psi' (packS0 rho_idle theta) pc (comml (_call f)).
Proof.
  intros D psi pc [p g tag]  D' Psi Psi' f.
  intros Hi HD HD' Hlift HPsi'.
  destruct ((fun x => Hi) 0) as [Hi' Hpgik].
  unfold NH_WFcmd_aux in Hi'.
  destruct Hi' as [[Hf Hi'] | [Hf Hi']].
  (* prim_entry f *)
  unfold IDLE_Call_Prim in Hi'.
  destruct Hi' as [H1 H2].
  inversion H1; subst f.
    (* f = _SWITCH *)
    eapply NH_SW_Itp; eauto.
    unfold NH_WFcmd, NH_WFcmd_aux; auto.
  (* ~ prim_entry f *)
  clear Hi.
  rename Hi' into Hi.
  subst D' Psi'.
  unfold IDLE_WFcmd in Hi.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_idle; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  (* interrupt *)
  split.
    destruct (itr_enabled_dec (M, R, fl, isr)) as [Hitr | Hitr];
        [right | left; intro H; destruct (Hitr H)].
    destruct Hitr as [Hfl Hisr].
    assert (Hih : IH_Specified Psi').
      unfold IH_Specified.
      apply lookupCT_subsetCT with (ext_IH_Spec (ext_Prim_Specs Psi)); trivial.
      unfold ext_IH_Spec.
      apply lookupCT_consCT; trivial.
    unfold WF_AIM_World in Hwf.
    destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
    simplS in H7.
    rename H7 into Hwfcth.
    unfold WF_Curr_Thrd in Hwfcth.
    simpl in Hwfcth.
    assert (Hflx : S_Flag (Mcth, R, fl, isr) _IF = true).
      simplS in * ; trivial.
    destruct (NH_Itr_prog (pc:=pc) Hpgik Hflx Hwfcth) as [Scth' HnextScth'].
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (itr_nextS_Mono2 H5 HJM H6 HnextScth') as [M' HnextS].
    exists (M', R', fl', isr'); split_l.
    clear Mcth' R' fl' isr' HnextScth' M' HnextS.
    intros [[[M' R'] fl'] isr'] Hitr.
    exists IH_Spec0.
    split.
      apply Hsub.
      unfold ext_IH_Spec.
      unfold OPSpec.lookup.
      simpl.
      left; trivial.
    simpl.
    rewrite minimalD_elim_ih; trivial.
    unfold Itp_AIM.
    exists T; exists B; exists RdyQ; exists tid.
    assert (Hstk : stkSpace (R SP) (IH_Stk + 6) Mcth).
      destruct Hpgik.
      eapply Prd_Itr_Stk_elim; unfold itr_enabled; eauto.
    destruct (itr_nextS_DMono2 H6 H5 HJM Hstk Hitr)
      as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM' [HnextS HTM']]]]]]].
    unfold pmstar.
    unfold WF_AIM_World.
    exists Mth'; exists Mq; split_l; split_r.
    split.
      eapply WF_CurrTID_nh_ih; eauto.
    split_l; split_l; split_l.
    exists TM'; exists Mcth'; split_l.
    split_l.
    rsplit.
      split.
        apply WF_ReadyQ_Mono_CT_consCT; trivial.
        apply (WF_ReadyQ_untouch H8 HTM' H1).
      apply WF_WQSet_Mono_CT_consCT; trivial.
      apply (WF_WQSet_untouch H9 HTM' H1).
    unfold WF_currTID in H1.
    destruct H1 as [_ [_ [_ [tcb [Hmto [Htcb Hrtc]]]]]].
    clear B RdyQ H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear TM TM' HsplitTM'.
    unfold WF_Curr_Thrd.
    exists tcb; exists (TCB_Status tcb); split_l; split_l.
    eapply NH_Itr_prev; eauto.
  (* ~ prim_entry f *)
  simpl in Hi.
  destruct Hi as [p' [p'' [g' [g'' [Hlookup' [Hlookup'' Hi]]]]]].
  destruct Hi as [Hcall [Hfun [Hpp' [Hpg'p'' Hg''g]]]].
  unfold WF_AIM_World in Hwf.
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct (NH_call_prog Hcall Hwfcth) as [Scth' HnextScth].
  split.
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (nextS_Mono2 H5 HJM H6 HnextScth) as [M' HnextS].
    exists (M', R', fl', isr'); trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  subst R' fl' isr'.
  assert (HnextPC' := nextPC_call pc f M R fl).
  assert (Heq := nextPC_injection HnextPC HnextPC').
  subst pc'.
  clear HnextPC'.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  clear Scth' HnextScth.
  assert (Hlk' : lookupCT Psi' f (packS0 rho_idle (Ast p' g' 1))).
    apply Hsub.
    unfold ext_IH_Spec, ext_Prim_Specs.
    apply lookupCT_subsetCT with Psi.
    apply Hlift; trivial.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply subsetCT_refl.
  assert (Hlk'' : lookupCT Psi' (pc+1) (packS0 rho_idle (Ast p'' g'' tag))).
    apply Hsub.
    unfold ext_IH_Spec, ext_Prim_Specs.
    apply lookupCT_subsetCT with Psi.
    apply Hlift; trivial.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply subsetCT_refl.
  exists (packS0 rho_idle (Ast p' g' 1)).
  split_l.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_idle; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  simplS.
  destruct HM.
  simplS in * .
  exists Mth'; exists Mq; split_l; split_r.
  split_l.
  split_l.
  split_l.
  split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  simplS.
  split.
    clear Hf H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC Hpgik HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd.
    simpl.
    eapply NH_nonprim_call_prev; eauto.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IDLE_in_CAP0_cli :
  forall D psi pc theta D' Psi Psi',
    IDLE_WFcmd psi pc theta (comml (_cli))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_idle Psi
    -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
    -> WFcmd D' Psi' (packS0 rho_idle theta) pc (comml (_cli)).
Proof.
  intros D psi pc [p g tag]  D' Psi Psi'.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold IDLE_WFcmd in Hi.
  destruct Hi as [Hi Hpgik].
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_idle; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  (* interrupt *)
  split.
    destruct (itr_enabled_dec (M, R, fl, isr)) as [Hitr | Hitr];
        [right | left; intro H; destruct (Hitr H)].
    destruct Hitr as [Hfl Hisr].
    assert (Hih : IH_Specified Psi').
      unfold IH_Specified.
      apply lookupCT_subsetCT with (ext_IH_Spec (ext_Prim_Specs Psi)); trivial.
      unfold ext_IH_Spec.
      apply lookupCT_consCT; trivial.
    unfold WF_AIM_World in Hwf.
    destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
    simplS in H7.
    rename H7 into Hwfcth.
    unfold WF_Curr_Thrd in Hwfcth.
    simpl in Hwfcth.
    assert (Hflx : S_Flag (Mcth, R, fl, isr) _IF = true).
      simplS in * ; trivial.
    destruct (NH_Itr_prog (pc:=pc) Hpgik Hflx Hwfcth) as [Scth' HnextScth'].
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (itr_nextS_Mono2 H5 HJM H6 HnextScth') as [M' HnextS].
    exists (M', R', fl', isr'); split_l.
    clear Mcth' R' fl' isr' HnextScth' M' HnextS.
    intros [[[M' R'] fl'] isr'] Hitr.
    exists IH_Spec0.
    split.
      apply Hsub.
      unfold ext_IH_Spec.
      unfold OPSpec.lookup.
      simpl.
      left; trivial.
    simpl.
    rewrite minimalD_elim_ih; trivial.
    unfold Itp_AIM.
    exists T; exists B; exists RdyQ; exists tid.
    assert (Hstk : stkSpace (R SP) (IH_Stk + 6) Mcth).
      destruct Hpgik.
      eapply Prd_Itr_Stk_elim; unfold itr_enabled; eauto.
    destruct (itr_nextS_DMono2 H6 H5 HJM Hstk Hitr)
      as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS HTM']]]]]]].
    unfold pmstar.
    unfold WF_AIM_World.
    exists Mth'; exists Mq; split_l; split_r.
    split. 
      eapply WF_CurrTID_nh_ih; eauto.
    split_l; split_l; split_l.
    exists TM'; exists Mcth'; split_l.
    split_l.
    rsplit.
      split.
        apply WF_ReadyQ_Mono_CT_consCT; trivial.
        apply (WF_ReadyQ_untouch H8 HTM' H1).
      apply WF_WQSet_Mono_CT_consCT; trivial.
      apply (WF_WQSet_untouch H9 HTM' H1).
    unfold WF_currTID in H1.
    destruct H1 as [_ [_ [_ [tcb [Hmto [Htcb Hrtc]]]]]].
    clear B RdyQ H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear TM TM' HsplitTM'.
    unfold WF_Curr_Thrd.
    exists tcb; exists (TCB_Status tcb); split_l; split_l.
    eapply NH_Itr_prev; eauto.
  assert (HnextS := nextS_cli pc M R fl isr).
  assert (HnextPC := nextPC_cli pc M R fl).
  simpl in Hi.
  destruct Hi as [p' [g' [Hlk [Hpp' Hg'g]]]].
  split.
    exists (M, R, Clear_Flag fl _IF, isr); trivial.
  intros f' S' HnextS' HnextPC'.
  destruct S' as [[[M' R'] fl'] isr'].
  assert (Heq := nextS_injection HnextS HnextS').
  simpl in Heq.
  destruct Heq as [Heq1 Heq2]; destruct Heq1; destruct Heq2. 
  assert (Heq := nextPC_injection HnextPC HnextPC'). 
  subst M' R' fl' isr' f'.
  clear HnextS' HnextPC'.
  assert (Hlk' : lookupCT Psi' (pc+1) (packS0 rho_idle (Ast p' g' tag))).
    apply Hsub.
    unfold ext_IH_Spec, ext_Prim_Specs.
    apply lookupCT_subsetCT with Psi.
    apply Hlift; trivial.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply subsetCT_refl.
  exists (packS0 rho_idle (Ast p' g' tag)).
  split_l.
  simpl.
  rewrite minimalD_elim_idle; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World in * .
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  exists Mth; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM; exists Mcth.
  split_l.
  split_l.
  rsplit.
    split_lr.
  simplS; simplS in H7.
  unfold WF_Curr_Thrd in * .
  simpl in * .
  eapply NH_cli_prev; eauto.
Qed.

Lemma SCAP_IDLE_in_CAP0_sti :
  forall D psi pc theta D' Psi Psi',
    IDLE_WFcmd psi pc theta (comml (_sti))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_idle Psi
    -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
    -> WFcmd D' Psi' (packS0 rho_idle theta) pc (comml (_sti)).
Proof.
  intros D psi pc [p g tag]  D' Psi Psi'.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold IDLE_WFcmd in Hi.
  destruct Hi as [Hi Hpgik].
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_idle; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  (* interrupt *)
  split.
    destruct (itr_enabled_dec (M, R, fl, isr)) as [Hitr | Hitr];
        [right | left; intro H; destruct (Hitr H)].
    destruct Hitr as [Hfl Hisr].
    assert (Hih : IH_Specified Psi').
      unfold IH_Specified.
      apply lookupCT_subsetCT with (ext_IH_Spec (ext_Prim_Specs Psi)); trivial.
      unfold ext_IH_Spec.
      apply lookupCT_consCT; trivial.
    unfold WF_AIM_World in Hwf.
    destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
    simplS in H7.
    rename H7 into Hwfcth.
    unfold WF_Curr_Thrd in Hwfcth.
    simpl in Hwfcth.
    assert (Hflx : S_Flag (Mcth, R, fl, isr) _IF = true).
      simplS in * ; trivial.
    destruct (NH_Itr_prog (pc:=pc) Hpgik Hflx Hwfcth) as [Scth' HnextScth'].
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (itr_nextS_Mono2 H5 HJM H6 HnextScth') as [M' HnextS].
    exists (M', R', fl', isr'); split_l.
    clear Mcth' R' fl' isr' HnextScth' M' HnextS.
    intros [[[M' R'] fl'] isr'] Hitr.
    exists IH_Spec0.
    split.
      apply Hsub.
      unfold ext_IH_Spec.
      unfold OPSpec.lookup.
      simpl.
      left; trivial.
    simpl.
    rewrite minimalD_elim_ih; trivial.
    unfold Itp_AIM.
    exists T; exists B; exists RdyQ; exists tid.
    assert (Hstk : stkSpace (R SP) (IH_Stk + 6) Mcth).
      destruct Hpgik.
      eapply Prd_Itr_Stk_elim; unfold itr_enabled; eauto.
    destruct (itr_nextS_DMono2 H6 H5 HJM Hstk Hitr)
      as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS HTM']]]]]]].
    unfold pmstar.
    unfold WF_AIM_World.
    exists Mth'; exists Mq; split_l; split_r.
    split. 
      eapply WF_CurrTID_nh_ih; eauto.
    split_l; split_l; split_l.
    exists TM'; exists Mcth'; split_l.
    split_l.
    rsplit.
      split.
        apply WF_ReadyQ_Mono_CT_consCT; trivial.
        apply (WF_ReadyQ_untouch H8 HTM' H1).
      apply WF_WQSet_Mono_CT_consCT; trivial.
      apply (WF_WQSet_untouch H9 HTM' H1).
    unfold WF_currTID in H1.
    destruct H1 as [_ [_ [_ [tcb [Hmto [Htcb Hrtc]]]]]].
    clear B RdyQ H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear TM TM' HsplitTM'.
    unfold WF_Curr_Thrd.
    exists tcb; exists (TCB_Status tcb); split_l; split_l.
    eapply NH_Itr_prev; eauto.
  assert (HnextS := nextS_sti pc M R fl isr).
  assert (HnextPC := nextPC_sti pc M R fl).
  simpl in Hi.
  destruct Hi as [p' [g' [Hlookup [Hsti [Hpp' Hg'g]]]]].
  split.
    exists (M, R, Set_Flag fl _IF, isr); trivial.
  intros f' S' HnextS' HnextPC'.
  destruct S' as [[[M' R'] fl'] isr'].
  assert (Heq := nextS_injection HnextS HnextS').
  simpl in Heq.
  destruct Heq as [Heq1 Heq2]; destruct Heq1; destruct Heq2.
  assert (Heq := nextPC_injection HnextPC HnextPC'). 
  subst M' R' fl' isr' f'.
  clear HnextS' HnextPC'.
  assert (Hlk' : lookupCT Psi' (pc+1) (packS0 rho_idle (Ast p' g' tag))).
    apply Hsub.
    unfold ext_IH_Spec, ext_Prim_Specs.
    apply lookupCT_subsetCT with Psi.
    apply Hlift; trivial.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply subsetCT_refl.
  exists (packS0 rho_idle (Ast p' g' tag)).
  split_l.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_idle; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World in * .
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  exists Mth;exists Mq.
  split_l.
  split_r.
  split_l.
  split_l.
  split_l.
  split_l.
  exists TM. exists Mcth.
  split_l.
  split_l.
  rsplit.
    split_lr.
  simplS; simplS in H7.
  unfold WF_Curr_Thrd in * .
  simpl in * .
  eapply NH_sti_prev; eauto.
Qed.

Lemma SCAP_IDLE_in_CAP0_jne :
  forall D psi pc theta D' Psi Psi' f,
    IDLE_WFcmd psi pc theta (comml (_jne f))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_idle Psi
    -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
    -> WFcmd D' Psi' (packS0 rho_idle theta) pc (comml (_jne f)).
Proof.
  intros D psi pc [p g tag]  D' Psi Psi' f.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold IDLE_WFcmd in Hi.
  destruct Hi as [Hi Hpgik].
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_idle; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  (* interrupt *)
  split.
    destruct (itr_enabled_dec (M, R, fl, isr)) as [Hitr | Hitr];
        [right | left; intro H; destruct (Hitr H)].
    destruct Hitr as [Hfl Hisr].
    assert (Hih : IH_Specified Psi').
      unfold IH_Specified.
      apply lookupCT_subsetCT with (ext_IH_Spec (ext_Prim_Specs Psi)); trivial.
      unfold ext_IH_Spec.
      apply lookupCT_consCT; trivial.
    unfold WF_AIM_World in Hwf.
    destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
    simplS in H7.
    rename H7 into Hwfcth.
    unfold WF_Curr_Thrd in Hwfcth.
    simpl in Hwfcth.
    assert (Hflx : S_Flag (Mcth, R, fl, isr) _IF = true).
      simplS in * ; trivial.
    destruct (NH_Itr_prog (pc:=pc) Hpgik Hflx Hwfcth) as [Scth' HnextScth'].
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (itr_nextS_Mono2 H5 HJM H6 HnextScth') as [M' HnextS].
    exists (M', R', fl', isr'); split_l.
    clear Mcth' R' fl' isr' HnextScth' M' HnextS.
    intros [[[M' R'] fl'] isr'] Hitr.
    exists IH_Spec0.
    split.
      apply Hsub.
      unfold ext_IH_Spec.
      unfold OPSpec.lookup.
      simpl.
      left; trivial.
    simpl.
    rewrite minimalD_elim_ih; trivial.
    unfold Itp_AIM.
    exists T; exists B; exists RdyQ; exists tid.
    assert (Hstk : stkSpace (R SP) (IH_Stk + 6) Mcth).
      destruct Hpgik.
      eapply Prd_Itr_Stk_elim; unfold itr_enabled; eauto.
    destruct (itr_nextS_DMono2 H6 H5 HJM Hstk Hitr)
      as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS HTM']]]]]]].
    unfold pmstar.
    unfold WF_AIM_World.
    exists Mth'; exists Mq; split_l; split_r.
    split. 
      eapply WF_CurrTID_nh_ih; eauto.
    split_l; split_l; split_l.
    exists TM'; exists Mcth'; split_l.
    split_l.
    rsplit.
      split.
        apply WF_ReadyQ_Mono_CT_consCT; trivial.
        apply (WF_ReadyQ_untouch H8 HTM' H1).
      apply WF_WQSet_Mono_CT_consCT; trivial.
      apply (WF_WQSet_untouch H9 HTM' H1).
    unfold WF_currTID in H1.
    destruct H1 as [_ [_ [_ [tcb [Hmto [Htcb Hrtc]]]]]].
    clear B RdyQ H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear TM TM' HsplitTM'.
    unfold WF_Curr_Thrd.
    exists tcb; exists (TCB_Status tcb); split_l; split_l.
    eapply NH_Itr_prev; eauto.
  simpl in Hi.
  destruct Hi as [p' [p'' [g' [g'' [Hlookup' [Hlookup'' Hi]]]]]].
  destruct Hi as [Hpp' [Hg'g [Hpp'' Hg''g]]].
  unfold WF_AIM_World in Hwf.  
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct (NH_jne_prog (f:=f) (pc:=pc) Hwfcth) as [Scth' HnextScth].
  split.
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (nextS_Mono2 H5 HJM H6 HnextScth) as [M' HnextS].
    exists (M', R', fl', isr'); trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr']]].
  subst M' R' fl' isr'.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  assert (Hlk' : lookupCT Psi' f (packS0 rho_idle (Ast p' g' tag))).
    apply Hsub.
    unfold ext_IH_Spec, ext_Prim_Specs.
    apply lookupCT_subsetCT with Psi.
    apply Hlift; trivial.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply subsetCT_refl.
  assert (Hlk'' : lookupCT Psi' (pc+1) (packS0 rho_idle (Ast p'' g'' tag))).
    apply Hsub.
    unfold ext_IH_Spec, ext_Prim_Specs.
    apply lookupCT_subsetCT with Psi.
    apply Hlift; trivial.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply subsetCT_refl.
  inversion HnextPC.
  (* non-jmp *)
  subst f0 M0 R0 fl0 pc0 pc'.
  simplS in H0.
  rename H0 into Hflzf.
  exists (packS0 rho_idle (Ast p'' g'' tag)).
  split_l.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_idle; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  exists Mth'; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  split.
    clear B RdyQ H1 H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC Hpgik HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd.
    simpl; simplS in * .
    eapply NH_jne_seq_prev; eauto.
    simplS; trivial.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
  (* jmp *)
  subst f0 M0 R0 fl0 pc0 pc'.
  simplS in H0.
  rename H0 into Hflzf.
  exists (packS0 rho_idle (Ast p' g' tag)).
  split_l.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_idle; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  exists Mth'; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  split.
    clear B RdyQ H1 H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC Hpgik HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd. 
    simpl; simplS.
    eapply NH_jne_jmp_prev; eauto.
    simplS; trivial.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IDLE_in_CAP0_movwp :
  forall D psi pc theta D' Psi Psi' o r,
    IDLE_WFcmd psi pc theta (comml (_movwp o r))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_idle Psi
    -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
    -> WFcmd D' Psi' (packS0 rho_idle theta) pc (comml (_movwp o r)).
Proof.
  intros D psi pc [p g tag]  D' Psi Psi' o r.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold IDLE_WFcmd in Hi.
  destruct Hi as [Hi Hpgik].
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_idle; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  (* interrupt *)
  split.
    destruct (itr_enabled_dec (M, R, fl, isr)) as [Hitr | Hitr];
        [right | left; intro H; destruct (Hitr H)].
    destruct Hitr as [Hfl Hisr].
    assert (Hih : IH_Specified Psi').
      unfold IH_Specified.
      apply lookupCT_subsetCT with (ext_IH_Spec (ext_Prim_Specs Psi)); trivial.
      unfold ext_IH_Spec.
      apply lookupCT_consCT; trivial.
    unfold WF_AIM_World in Hwf.
    destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
    simplS in H7.
    rename H7 into Hwfcth.
    unfold WF_Curr_Thrd in Hwfcth.
    simpl in Hwfcth.
    assert (Hflx : S_Flag (Mcth, R, fl, isr) _IF = true).
      simplS in * ; trivial.
    destruct (NH_Itr_prog (pc:=pc) Hpgik Hflx Hwfcth) as [Scth' HnextScth'].
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (itr_nextS_Mono2 H5 HJM H6 HnextScth') as [M' HnextS].
    exists (M', R', fl', isr'); split_l.
    clear Mcth' R' fl' isr' HnextScth' M' HnextS.
    intros [[[M' R'] fl'] isr'] Hitr.
    exists IH_Spec0.
    split.
      apply Hsub.
      unfold ext_IH_Spec.
      unfold OPSpec.lookup.
      simpl.
      left; trivial.
    simpl.
    rewrite minimalD_elim_ih; trivial.
    unfold Itp_AIM.
    exists T; exists B; exists RdyQ; exists tid.
    assert (Hstk : stkSpace (R SP) (IH_Stk + 6) Mcth).
      destruct Hpgik.
      eapply Prd_Itr_Stk_elim; unfold itr_enabled; eauto.
    destruct (itr_nextS_DMono2 H6 H5 HJM Hstk Hitr)
      as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS HTM']]]]]]].
    unfold pmstar.
    unfold WF_AIM_World.
    exists Mth'; exists Mq; split_l; split_r.
    split. 
      eapply WF_CurrTID_nh_ih; eauto.
    split_l; split_l; split_l.
    exists TM'; exists Mcth'; split_l.
    split_l.
    rsplit.
      split.
        apply WF_ReadyQ_Mono_CT_consCT; trivial.
        apply (WF_ReadyQ_untouch H8 HTM' H1).
      apply WF_WQSet_Mono_CT_consCT; trivial.
      apply (WF_WQSet_untouch H9 HTM' H1).
    unfold WF_currTID in H1.
    destruct H1 as [_ [_ [_ [tcb [Hmto [Htcb Hrtc]]]]]].
    clear B RdyQ H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear TM TM' HsplitTM'.
    unfold WF_Curr_Thrd.
    exists tcb; exists (TCB_Status tcb); split_l; split_l.
    eapply NH_Itr_prev; eauto.
  simpl in Hi.
  destruct Hi as [p' [g' [Hlookup' [Hmovwp Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  unfold WF_AIM_World in Hwf.
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct (NH_movwp_prog Hmovwp Hwfcth) as [Scth' HnextScth].
  split.
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (nextS_Mono2 H5 HJM H6 HnextScth) as [M' HnextS].
    exists (M', R', fl', isr'); trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  rewrite (nextPC_inversion HnextPC).
  subst R' fl' isr'.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  clear Scth' HnextScth.
  assert (Hlk' : lookupCT Psi' (pc+1) (packS0 rho_idle (Ast p' g' tag))).
    apply Hsub.
    unfold ext_IH_Spec, ext_Prim_Specs.
    apply lookupCT_subsetCT with Psi.
    apply Hlift; trivial.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply subsetCT_refl.
  exists (packS0 rho_idle (Ast p' g' tag)).
  split_l.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_idle; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  simplS.
  destruct HM.
  simplS in * .
  exists Mth'; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  simplS.
  split.
    clear B RdyQ H1 H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC Hpgik HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd.
    simpl.
    eapply NH_movwp_prev; eauto.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IDLE_in_CAP0_movwld :
  forall D psi pc theta D' Psi Psi' a r,
    IDLE_WFcmd psi pc theta (comml (_movwld a r))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_idle Psi
    -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
    -> WFcmd D' Psi' (packS0 rho_idle theta) pc (comml (_movwld a r)).
Proof.
  intros D psi pc [p g tag]  D' Psi Psi' a r.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold IDLE_WFcmd in Hi.
  destruct Hi as [Hi Hpgik].
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_idle; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  (* interrupt *)
  split.
    destruct (itr_enabled_dec (M, R, fl, isr)) as [Hitr | Hitr];
        [right | left; intro H; destruct (Hitr H)].
    destruct Hitr as [Hfl Hisr].
    assert (Hih : IH_Specified Psi').
      unfold IH_Specified.
      apply lookupCT_subsetCT with (ext_IH_Spec (ext_Prim_Specs Psi)); trivial.
      unfold ext_IH_Spec.
      apply lookupCT_consCT; trivial.
    unfold WF_AIM_World in Hwf.
    destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
    simplS in H7.
    rename H7 into Hwfcth.
    unfold WF_Curr_Thrd in Hwfcth.
    simpl in Hwfcth.
    assert (Hflx : S_Flag (Mcth, R, fl, isr) _IF = true).
      simplS in * ; trivial.
    destruct (NH_Itr_prog (pc:=pc) Hpgik Hflx Hwfcth) as [Scth' HnextScth'].
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (itr_nextS_Mono2 H5 HJM H6 HnextScth') as [M' HnextS].
    exists (M', R', fl', isr'); split_l.
    clear Mcth' R' fl' isr' HnextScth' M' HnextS.
    intros [[[M' R'] fl'] isr'] Hitr.
    exists IH_Spec0.
    split.
      apply Hsub.
      unfold ext_IH_Spec.
      unfold OPSpec.lookup.
      simpl.
      left; trivial.
    simpl.
    rewrite minimalD_elim_ih; trivial.
    unfold Itp_AIM.
    exists T; exists B; exists RdyQ; exists tid.
    assert (Hstk : stkSpace (R SP) (IH_Stk + 6) Mcth).
      destruct Hpgik.
      eapply Prd_Itr_Stk_elim; unfold itr_enabled; eauto.
    destruct (itr_nextS_DMono2 H6 H5 HJM Hstk Hitr)
      as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS HTM']]]]]]].
    unfold pmstar.
    unfold WF_AIM_World.
    exists Mth'; exists Mq; split_l; split_r.
    split. 
      eapply WF_CurrTID_nh_ih; eauto.
    split_l; split_l; split_l.
    exists TM'; exists Mcth'; split_l.
    split_l.
    rsplit.
      split.
        apply WF_ReadyQ_Mono_CT_consCT; trivial.
        apply (WF_ReadyQ_untouch H8 HTM' H1).
      apply WF_WQSet_Mono_CT_consCT; trivial.
      apply (WF_WQSet_untouch H9 HTM' H1).
    unfold WF_currTID in H1.
    destruct H1 as [_ [_ [_ [tcb [Hmto [Htcb Hrtc]]]]]].
    clear B RdyQ H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear TM TM' HsplitTM'.
    unfold WF_Curr_Thrd.
    exists tcb; exists (TCB_Status tcb); split_l; split_l.
    eapply NH_Itr_prev; eauto.
  simpl in Hi.
  destruct Hi as [p' [g' [Hlookup' [Hmovwp Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  unfold WF_AIM_World in Hwf.
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct (NH_movwld_prog Hmovwp Hwfcth) as [Scth' HnextScth].
  split.
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (nextS_Mono2 H5 HJM H6 HnextScth) as [M' HnextS].
    exists (M', R', fl', isr'); trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  rewrite (nextPC_inversion HnextPC).
  destruct HR as [x [Hlds HR']].
  subst R' fl' isr'.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  clear Scth' HnextScth.
  assert (Hlk' : lookupCT Psi' (pc+1) (packS0 rho_idle (Ast p' g' tag))).
    apply Hsub.
    unfold ext_IH_Spec, ext_Prim_Specs.
    apply lookupCT_subsetCT with Psi.
    apply Hlift; trivial.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply subsetCT_refl.
  exists (packS0 rho_idle (Ast p' g' tag)).
  split_l.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_idle; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  simplS.
  destruct HM.
  simplS in * .
  exists Mth'; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  simplS.
  split.
    clear B RdyQ H1 H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC Hpgik HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd.
    simpl.
    eapply NH_movwld_prev; eauto.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IDLE_in_CAP0_movwst :
  forall D psi pc theta D' Psi Psi' o a,
    IDLE_WFcmd psi pc theta (comml (_movwst o a))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_idle Psi
    -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
    -> WFcmd D' Psi' (packS0 rho_idle theta) pc (comml (_movwst o a)).
Proof.
  intros D psi pc [p g tag]  D' Psi Psi' o a.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold IDLE_WFcmd in Hi.
  destruct Hi as [Hi Hpgik].
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_idle; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  (* interrupt *)
  split.
    destruct (itr_enabled_dec (M, R, fl, isr)) as [Hitr | Hitr];
        [right | left; intro H; destruct (Hitr H)].
    destruct Hitr as [Hfl Hisr].
    assert (Hih : IH_Specified Psi').
      unfold IH_Specified.
      apply lookupCT_subsetCT with (ext_IH_Spec (ext_Prim_Specs Psi)); trivial.
      unfold ext_IH_Spec.
      apply lookupCT_consCT; trivial.
    unfold WF_AIM_World in Hwf.
    destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
    simplS in H7.
    rename H7 into Hwfcth.
    unfold WF_Curr_Thrd in Hwfcth.
    simpl in Hwfcth.
    assert (Hflx : S_Flag (Mcth, R, fl, isr) _IF = true).
      simplS in * ; trivial.
    destruct (NH_Itr_prog (pc:=pc) Hpgik Hflx Hwfcth) as [Scth' HnextScth'].
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (itr_nextS_Mono2 H5 HJM H6 HnextScth') as [M' HnextS].
    exists (M', R', fl', isr'); split_l.
    clear Mcth' R' fl' isr' HnextScth' M' HnextS.
    intros [[[M' R'] fl'] isr'] Hitr.
    exists IH_Spec0.
    split.
      apply Hsub.
      unfold ext_IH_Spec.
      unfold OPSpec.lookup.
      simpl.
      left; trivial.
    simpl.
    rewrite minimalD_elim_ih; trivial.
    unfold Itp_AIM.
    exists T; exists B; exists RdyQ; exists tid.
    assert (Hstk : stkSpace (R SP) (IH_Stk + 6) Mcth).
      destruct Hpgik.
      eapply Prd_Itr_Stk_elim; unfold itr_enabled; eauto.
    destruct (itr_nextS_DMono2 H6 H5 HJM Hstk Hitr)
      as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS HTM']]]]]]].
    unfold pmstar.
    unfold WF_AIM_World.
    exists Mth'; exists Mq; split_l; split_r.
    split. 
      eapply WF_CurrTID_nh_ih; eauto.
    split_l; split_l; split_l.
    exists TM'; exists Mcth'; split_l.
    split_l.
    rsplit.
      split.
        apply WF_ReadyQ_Mono_CT_consCT; trivial.
        apply (WF_ReadyQ_untouch H8 HTM' H1).
      apply WF_WQSet_Mono_CT_consCT; trivial.
      apply (WF_WQSet_untouch H9 HTM' H1).
    unfold WF_currTID in H1.
    destruct H1 as [_ [_ [_ [tcb [Hmto [Htcb Hrtc]]]]]].
    clear B RdyQ H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear TM TM' HsplitTM'.
    unfold WF_Curr_Thrd.
    exists tcb; exists (TCB_Status tcb); split_l; split_l.
    eapply NH_Itr_prev; eauto.
  simpl in Hi.
  destruct Hi as [p' [g' [Hlookup' [Hmovwst Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  unfold WF_AIM_World in Hwf.
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct (NH_movwst_prog Hmovwst Hwfcth) as [Scth' HnextScth].
  split.
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (nextS_Mono2 H5 HJM H6 HnextScth) as [M' HnextS].
    exists (M', R', fl', isr'); trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  rewrite (nextPC_inversion HnextPC).
  subst R' fl' isr'.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  clear Scth' HnextScth.
  assert (Hlk' : lookupCT Psi' (pc+1) (packS0 rho_idle (Ast p' g' tag))).
    apply Hsub.
    unfold ext_IH_Spec, ext_Prim_Specs.
    apply lookupCT_subsetCT with Psi.
    apply Hlift; trivial.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply subsetCT_refl.
  exists (packS0 rho_idle (Ast p' g' tag)).
  split_l.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_idle; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  simplS.
  destruct HM.
  simplS in * .
  exists Mth'; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  simplS.
  split.
    clear B RdyQ H1 H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC Hpgik HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd.
    simpl.
    eapply NH_movwst_prev; eauto.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IDLE_in_CAP0_pushwp :
  forall D psi pc theta D' Psi Psi' o,
    IDLE_WFcmd psi pc theta (comml (_pushwp o))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_idle Psi
    -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
    -> WFcmd D' Psi' (packS0 rho_idle theta) pc (comml (_pushwp o)).
Proof.
  intros D psi pc [p g tag]  D' Psi Psi' o.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold IDLE_WFcmd in Hi.
  destruct Hi as [Hi Hpgik].
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_idle; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  (* interrupt *)
  split.
    destruct (itr_enabled_dec (M, R, fl, isr)) as [Hitr | Hitr];
        [right | left; intro H; destruct (Hitr H)].
    destruct Hitr as [Hfl Hisr].
    assert (Hih : IH_Specified Psi').
      unfold IH_Specified.
      apply lookupCT_subsetCT with (ext_IH_Spec (ext_Prim_Specs Psi)); trivial.
      unfold ext_IH_Spec.
      apply lookupCT_consCT; trivial.
    unfold WF_AIM_World in Hwf.
    destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
    simplS in H7.
    rename H7 into Hwfcth.
    unfold WF_Curr_Thrd in Hwfcth.
    simpl in Hwfcth.
    assert (Hflx : S_Flag (Mcth, R, fl, isr) _IF = true).
      simplS in * ; trivial.
    destruct (NH_Itr_prog (pc:=pc) Hpgik Hflx Hwfcth) as [Scth' HnextScth'].
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (itr_nextS_Mono2 H5 HJM H6 HnextScth') as [M' HnextS].
    exists (M', R', fl', isr'); split_l.
    clear Mcth' R' fl' isr' HnextScth' M' HnextS.
    intros [[[M' R'] fl'] isr'] Hitr.
    exists IH_Spec0.
    split.
      apply Hsub.
      unfold ext_IH_Spec.
      unfold OPSpec.lookup.
      simpl.
      left; trivial.
    simpl.
    rewrite minimalD_elim_ih; trivial.
    unfold Itp_AIM.
    exists T; exists B; exists RdyQ; exists tid.
    assert (Hstk : stkSpace (R SP) (IH_Stk + 6) Mcth).
      destruct Hpgik.
      eapply Prd_Itr_Stk_elim; unfold itr_enabled; eauto.
    destruct (itr_nextS_DMono2 H6 H5 HJM Hstk Hitr)
      as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS HTM']]]]]]].
    unfold pmstar.
    unfold WF_AIM_World.
    exists Mth'; exists Mq; split_l; split_r.
    split. 
      eapply WF_CurrTID_nh_ih; eauto.
    split_l; split_l; split_l.
    exists TM'; exists Mcth'; split_l.
    split_l.
    rsplit.
      split.
        apply WF_ReadyQ_Mono_CT_consCT; trivial.
        apply (WF_ReadyQ_untouch H8 HTM' H1).
      apply WF_WQSet_Mono_CT_consCT; trivial.
      apply (WF_WQSet_untouch H9 HTM' H1).
    unfold WF_currTID in H1.
    destruct H1 as [_ [_ [_ [tcb [Hmto [Htcb Hrtc]]]]]].
    clear B RdyQ H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear TM TM' HsplitTM'.
    unfold WF_Curr_Thrd.
    exists tcb; exists (TCB_Status tcb); split_l; split_l.
    eapply NH_Itr_prev; eauto.
  simpl in Hi.
  destruct Hi as [p' [g' [Hlookup' [Hen Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  unfold WF_AIM_World in Hwf.
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct (NH_pushwp_prog Hen Hwfcth) as [Scth' HnextScth].
  split.
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (nextS_Mono2 H5 HJM H6 HnextScth) as [M' HnextS].
    exists (M', R', fl', isr'); trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  rewrite (nextPC_inversion HnextPC).
  subst R' fl' isr'.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  clear Scth' HnextScth.
  assert (Hlk' : lookupCT Psi' (pc+1) (packS0 rho_idle (Ast p' g' tag))).
    apply Hsub.
    unfold ext_IH_Spec, ext_Prim_Specs.
    apply lookupCT_subsetCT with Psi.
    apply Hlift; trivial.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply subsetCT_refl.
  exists (packS0 rho_idle (Ast p' g' tag)).
  split_l.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_idle; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  simplS.
  destruct HM.
  simplS in * .
  exists Mth'; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  simplS.
  split.
    clear B RdyQ H1 H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC Hpgik HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd.
    simpl.
    eapply NH_pushwp_prev; eauto.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IDLE_in_CAP0_pushwm :
  forall D psi pc theta D' Psi Psi' a,
    IDLE_WFcmd psi pc theta (comml (_pushwm a))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_idle Psi
    -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
    -> WFcmd D' Psi' (packS0 rho_idle theta) pc (comml (_pushwm a)).
Proof.
  intros D psi pc [p g tag]  D' Psi Psi' a.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold IDLE_WFcmd in Hi.
  destruct Hi as [Hi Hpgik].
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_idle; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  (* interrupt *)
  split.
    destruct (itr_enabled_dec (M, R, fl, isr)) as [Hitr | Hitr];
        [right | left; intro H; destruct (Hitr H)].
    destruct Hitr as [Hfl Hisr].
    assert (Hih : IH_Specified Psi').
      unfold IH_Specified.
      apply lookupCT_subsetCT with (ext_IH_Spec (ext_Prim_Specs Psi)); trivial.
      unfold ext_IH_Spec.
      apply lookupCT_consCT; trivial.
    unfold WF_AIM_World in Hwf.
    destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
    simplS in H7.
    rename H7 into Hwfcth.
    unfold WF_Curr_Thrd in Hwfcth.
    simpl in Hwfcth.
    assert (Hflx : S_Flag (Mcth, R, fl, isr) _IF = true).
      simplS in * ; trivial.
    destruct (NH_Itr_prog (pc:=pc) Hpgik Hflx Hwfcth) as [Scth' HnextScth'].
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (itr_nextS_Mono2 H5 HJM H6 HnextScth') as [M' HnextS].
    exists (M', R', fl', isr'); split_l.
    clear Mcth' R' fl' isr' HnextScth' M' HnextS.
    intros [[[M' R'] fl'] isr'] Hitr.
    exists IH_Spec0.
    split.
      apply Hsub.
      unfold ext_IH_Spec.
      unfold OPSpec.lookup.
      simpl.
      left; trivial.
    simpl.
    rewrite minimalD_elim_ih; trivial.
    unfold Itp_AIM.
    exists T; exists B; exists RdyQ; exists tid.
    assert (Hstk : stkSpace (R SP) (IH_Stk + 6) Mcth).
      destruct Hpgik.
      eapply Prd_Itr_Stk_elim; unfold itr_enabled; eauto.
    destruct (itr_nextS_DMono2 H6 H5 HJM Hstk Hitr)
      as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS HTM']]]]]]].
    unfold pmstar.
    unfold WF_AIM_World.
    exists Mth'; exists Mq; split_l; split_r.
    split. 
      eapply WF_CurrTID_nh_ih; eauto.
    split_l; split_l; split_l.
    exists TM'; exists Mcth'; split_l.
    split_l.
    rsplit.
      split.
        apply WF_ReadyQ_Mono_CT_consCT; trivial.
        apply (WF_ReadyQ_untouch H8 HTM' H1).
      apply WF_WQSet_Mono_CT_consCT; trivial.
      apply (WF_WQSet_untouch H9 HTM' H1).
    unfold WF_currTID in H1.
    destruct H1 as [_ [_ [_ [tcb [Hmto [Htcb Hrtc]]]]]].
    clear B RdyQ H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear TM TM' HsplitTM'.
    unfold WF_Curr_Thrd.
    exists tcb; exists (TCB_Status tcb); split_l; split_l.
    eapply NH_Itr_prev; eauto.
  destruct Hi as [p' [g' [Hlookup' [Hen Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  unfold WF_AIM_World in Hwf.
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct (NH_pushwm_prog Hen Hwfcth) as [Scth' HnextScth].
  split.
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (nextS_Mono2 H5 HJM H6 HnextScth) as [M' HnextS].
    exists (M', R', fl', isr'); trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  rewrite (nextPC_inversion HnextPC).
  subst R' fl' isr'.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  clear Scth' HnextScth.
  assert (Hlk' : lookupCT Psi' (pc+1) (packS0 rho_idle (Ast p' g' tag))).
    apply Hsub.
    unfold ext_IH_Spec, ext_Prim_Specs.
    apply lookupCT_subsetCT with Psi.
    apply Hlift; trivial.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply subsetCT_refl.
  exists (packS0 rho_idle (Ast p' g' tag)).
  split_l.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_idle; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  simplS.
  destruct HM.
  simplS in * .
  exists Mth'; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  simplS.
  split.
    clear B RdyQ H1 H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC Hpgik HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd.
    simpl.
    eapply NH_pushwm_prev; eauto.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IDLE_in_CAP0_pushfw :
  forall D psi pc theta D' Psi Psi',
    IDLE_WFcmd psi pc theta (comml (_pushfw))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_idle Psi
    -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
    -> WFcmd D' Psi' (packS0 rho_idle theta) pc (comml (_pushfw)).
Proof.
  intros D psi pc [p g tag]  D' Psi Psi'.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold IDLE_WFcmd in Hi.
  destruct Hi as [Hi Hpgik].
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_idle; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  (* interrupt *)
  split.
    destruct (itr_enabled_dec (M, R, fl, isr)) as [Hitr | Hitr];
        [right | left; intro H; destruct (Hitr H)].
    destruct Hitr as [Hfl Hisr].
    assert (Hih : IH_Specified Psi').
      unfold IH_Specified.
      apply lookupCT_subsetCT with (ext_IH_Spec (ext_Prim_Specs Psi)); trivial.
      unfold ext_IH_Spec.
      apply lookupCT_consCT; trivial.
    unfold WF_AIM_World in Hwf.
    destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
    simplS in H7.
    rename H7 into Hwfcth.
    unfold WF_Curr_Thrd in Hwfcth.
    simpl in Hwfcth.
    assert (Hflx : S_Flag (Mcth, R, fl, isr) _IF = true).
      simplS in * ; trivial.
    destruct (NH_Itr_prog (pc:=pc) Hpgik Hflx Hwfcth) as [Scth' HnextScth'].
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (itr_nextS_Mono2 H5 HJM H6 HnextScth') as [M' HnextS].
    exists (M', R', fl', isr'); split_l.
    clear Mcth' R' fl' isr' HnextScth' M' HnextS.
    intros [[[M' R'] fl'] isr'] Hitr.
    exists IH_Spec0.
    split.
      apply Hsub.
      unfold ext_IH_Spec.
      unfold OPSpec.lookup.
      simpl.
      left; trivial.
    simpl.
    rewrite minimalD_elim_ih; trivial.
    unfold Itp_AIM.
    exists T; exists B; exists RdyQ; exists tid.
    assert (Hstk : stkSpace (R SP) (IH_Stk + 6) Mcth).
      destruct Hpgik.
      eapply Prd_Itr_Stk_elim; unfold itr_enabled; eauto.
    destruct (itr_nextS_DMono2 H6 H5 HJM Hstk Hitr)
      as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS HTM']]]]]]].
    unfold pmstar.
    unfold WF_AIM_World.
    exists Mth'; exists Mq; split_l; split_r.
    split. 
      eapply WF_CurrTID_nh_ih; eauto.
    split_l; split_l; split_l.
    exists TM'; exists Mcth'; split_l.
    split_l.
    rsplit.
      split.
        apply WF_ReadyQ_Mono_CT_consCT; trivial.
        apply (WF_ReadyQ_untouch H8 HTM' H1).
      apply WF_WQSet_Mono_CT_consCT; trivial.
      apply (WF_WQSet_untouch H9 HTM' H1).
    unfold WF_currTID in H1.
    destruct H1 as [_ [_ [_ [tcb [Hmto [Htcb Hrtc]]]]]].
    clear B RdyQ H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear TM TM' HsplitTM'.
    unfold WF_Curr_Thrd.
    exists tcb; exists (TCB_Status tcb); split_l; split_l.
    eapply NH_Itr_prev; eauto.
  simpl in Hi.
  destruct Hi as [p' [g' [Hlookup' [Hen Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  unfold WF_AIM_World in Hwf.
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct (NH_pushfw_prog Hen Hwfcth) as [Scth' HnextScth].
  split.
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (nextS_Mono2 H5 HJM H6 HnextScth) as [M' HnextS].
    exists (M', R', fl', isr'); trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  rewrite (nextPC_inversion HnextPC).
  subst R' fl' isr'.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  clear Scth' HnextScth.
  assert (Hlk' : lookupCT Psi' (pc+1) (packS0 rho_idle (Ast p' g' tag))).
    apply Hsub.
    unfold ext_IH_Spec, ext_Prim_Specs.
    apply lookupCT_subsetCT with Psi.
    apply Hlift; trivial.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply subsetCT_refl.
  exists (packS0 rho_idle (Ast p' g' tag)).
  split_l.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_idle; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  simplS.
  destruct HM.
  simplS in * .
  exists Mth'; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  simplS.
  split.
    clear B RdyQ H1 H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC Hpgik HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd.
    simpl.
    eapply NH_pushfw_prev; eauto.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IDLE_in_CAP0_popw :
  forall D psi pc theta D' Psi Psi' r,
    IDLE_WFcmd psi pc theta (comml (_popw r))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_idle Psi
    -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
    -> WFcmd D' Psi' (packS0 rho_idle theta) pc (comml (_popw r)).
Proof.
  intros D psi pc [p g tag]  D' Psi Psi' r.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold IDLE_WFcmd in Hi.
  destruct Hi as [Hi Hpgik].
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_idle; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  (* interrupt *)
  split.
    destruct (itr_enabled_dec (M, R, fl, isr)) as [Hitr | Hitr];
        [right | left; intro H; destruct (Hitr H)].
    destruct Hitr as [Hfl Hisr].
    assert (Hih : IH_Specified Psi').
      unfold IH_Specified.
      apply lookupCT_subsetCT with (ext_IH_Spec (ext_Prim_Specs Psi)); trivial.
      unfold ext_IH_Spec.
      apply lookupCT_consCT; trivial.
    unfold WF_AIM_World in Hwf.
    destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
    simplS in H7.
    rename H7 into Hwfcth.
    unfold WF_Curr_Thrd in Hwfcth.
    simpl in Hwfcth.
    assert (Hflx : S_Flag (Mcth, R, fl, isr) _IF = true).
      simplS in * ; trivial.
    destruct (NH_Itr_prog (pc:=pc) Hpgik Hflx Hwfcth) as [Scth' HnextScth'].
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (itr_nextS_Mono2 H5 HJM H6 HnextScth') as [M' HnextS].
    exists (M', R', fl', isr'); split_l.
    clear Mcth' R' fl' isr' HnextScth' M' HnextS.
    intros [[[M' R'] fl'] isr'] Hitr.
    exists IH_Spec0.
    split.
      apply Hsub.
      unfold ext_IH_Spec.
      unfold OPSpec.lookup.
      simpl.
      left; trivial.
    simpl.
    rewrite minimalD_elim_ih; trivial.
    unfold Itp_AIM.
    exists T; exists B; exists RdyQ; exists tid.
    assert (Hstk : stkSpace (R SP) (IH_Stk + 6) Mcth).
      destruct Hpgik.
      eapply Prd_Itr_Stk_elim; unfold itr_enabled; eauto.
    destruct (itr_nextS_DMono2 H6 H5 HJM Hstk Hitr)
      as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS HTM']]]]]]].
    unfold pmstar.
    unfold WF_AIM_World.
    exists Mth'; exists Mq; split_l; split_r.
    split. 
      eapply WF_CurrTID_nh_ih; eauto.
    split_l; split_l; split_l.
    exists TM'; exists Mcth'; split_l.
    split_l.
    rsplit.
      split.
        apply WF_ReadyQ_Mono_CT_consCT; trivial.
        apply (WF_ReadyQ_untouch H8 HTM' H1).
      apply WF_WQSet_Mono_CT_consCT; trivial.
      apply (WF_WQSet_untouch H9 HTM' H1).
    unfold WF_currTID in H1.
    destruct H1 as [_ [_ [_ [tcb [Hmto [Htcb Hrtc]]]]]].
    clear B RdyQ H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear TM TM' HsplitTM'.
    unfold WF_Curr_Thrd.
    exists tcb; exists (TCB_Status tcb); split_l; split_l.
    eapply NH_Itr_prev; eauto.
  simpl in Hi.
  destruct Hi as [p' [g' [Hlookup' [Hen Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  unfold WF_AIM_World in Hwf.
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct (NH_popw_prog Hen Hwfcth) as [Scth' HnextScth].
  split.
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (nextS_Mono2 H5 HJM H6 HnextScth) as [M' HnextS].
    exists (M', R', fl', isr'); trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  rewrite (nextPC_inversion HnextPC).
  destruct HR as [x [Hlds HR']].
  subst R' fl' isr'.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  clear Scth' HnextScth.
  assert (Hlk' : lookupCT Psi' (pc+1) (packS0 rho_idle (Ast p' g' tag))).
    apply Hsub.
    unfold ext_IH_Spec, ext_Prim_Specs.
    apply lookupCT_subsetCT with Psi.
    apply Hlift; trivial.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply subsetCT_refl.
  exists (packS0 rho_idle (Ast p' g' tag)).
  split_l.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_idle; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  simplS.
  destruct HM.
  simplS in * .
  exists Mth'; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  simplS.
  split.
    clear B RdyQ H1 H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC Hpgik HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd.
    simpl.
    eapply NH_popw_prev; eauto.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IDLE_in_CAP0_popfw :
  forall D psi pc theta D' Psi Psi',
    IDLE_WFcmd psi pc theta (comml (_popfw))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_idle Psi
    -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
    -> WFcmd D' Psi' (packS0 rho_idle theta) pc (comml (_popfw)).
Proof.
  intros D psi pc [p g tag]  D' Psi Psi'.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold IDLE_WFcmd in Hi.
  destruct Hi as [Hi Hpgik].
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_idle; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  (* interrupt *)
  split.
    destruct (itr_enabled_dec (M, R, fl, isr)) as [Hitr | Hitr];
        [right | left; intro H; destruct (Hitr H)].
    destruct Hitr as [Hfl Hisr].
    assert (Hih : IH_Specified Psi').
      unfold IH_Specified.
      apply lookupCT_subsetCT with (ext_IH_Spec (ext_Prim_Specs Psi)); trivial.
      unfold ext_IH_Spec.
      apply lookupCT_consCT; trivial.
    unfold WF_AIM_World in Hwf.
    destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
    simplS in H7.
    rename H7 into Hwfcth.
    unfold WF_Curr_Thrd in Hwfcth.
    simpl in Hwfcth.
    assert (Hflx : S_Flag (Mcth, R, fl, isr) _IF = true).
      simplS in * ; trivial.
    destruct (NH_Itr_prog (pc:=pc) Hpgik Hflx Hwfcth) as [Scth' HnextScth'].
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (itr_nextS_Mono2 H5 HJM H6 HnextScth') as [M' HnextS].
    exists (M', R', fl', isr'); split_l.
    clear Mcth' R' fl' isr' HnextScth' M' HnextS.
    intros [[[M' R'] fl'] isr'] Hitr.
    exists IH_Spec0.
    split.
      apply Hsub.
      unfold ext_IH_Spec.
      unfold OPSpec.lookup.
      simpl.
      left; trivial.
    simpl.
    rewrite minimalD_elim_ih; trivial.
    unfold Itp_AIM.
    exists T; exists B; exists RdyQ; exists tid.
    assert (Hstk : stkSpace (R SP) (IH_Stk + 6) Mcth).
      destruct Hpgik.
      eapply Prd_Itr_Stk_elim; unfold itr_enabled; eauto.
    destruct (itr_nextS_DMono2 H6 H5 HJM Hstk Hitr)
      as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS HTM']]]]]]].
    unfold pmstar.
    unfold WF_AIM_World.
    exists Mth'; exists Mq; split_l; split_r.
    split. 
      eapply WF_CurrTID_nh_ih; eauto.
    split_l; split_l; split_l.
    exists TM'; exists Mcth'; split_l.
    split_l.
    rsplit.
      split.
        apply WF_ReadyQ_Mono_CT_consCT; trivial.
        apply (WF_ReadyQ_untouch H8 HTM' H1).
      apply WF_WQSet_Mono_CT_consCT; trivial.
      apply (WF_WQSet_untouch H9 HTM' H1).
    unfold WF_currTID in H1.
    destruct H1 as [_ [_ [_ [tcb [Hmto [Htcb Hrtc]]]]]].
    clear B RdyQ H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear TM TM' HsplitTM'.
    unfold WF_Curr_Thrd.
    exists tcb; exists (TCB_Status tcb); split_l; split_l.
    eapply NH_Itr_prev; eauto.
  simpl in Hi.
  destruct Hi as [p' [g' [Hlookup' [Hen Hi]]]].
  destruct Hi as [Hsti [Hstipp' [Hstig'g [Hclipp' Hclig'g]]]].
  unfold WF_AIM_World in Hwf.
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct (NH_popfw_prog Hen Hwfcth) as [Scth' HnextScth].
  split.
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (nextS_Mono2 H5 HJM H6 HnextScth) as [M' HnextS].
    exists (M', R', fl', isr'); trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  rewrite (nextPC_inversion HnextPC).
  destruct Hfl as [flnum [Hlds Hfl']].
  subst R' fl' isr'.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  clear Scth' HnextScth.
  assert (Hlk' : lookupCT Psi' (pc+1) (packS0 rho_idle (Ast p' g' tag))).
    apply Hsub.
    unfold ext_IH_Spec, ext_Prim_Specs.
    apply lookupCT_subsetCT with Psi.
    apply Hlift; trivial.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply subsetCT_refl.
  exists (packS0 rho_idle (Ast p' g' tag)).
  split_l.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_idle; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  simplS.
  destruct HM.
  simplS in * .
  exists Mth'; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  simplS.
  split.
    clear B RdyQ H1 H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC Hpgik HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd.
    simpl.
    eapply NH_popfw_prev; eauto.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IDLE_in_CAP0_cmpw :
  forall D psi pc theta D' Psi Psi' o r,
    IDLE_WFcmd psi pc theta (comml (_cmpw o r))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_idle Psi
    -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
    -> WFcmd D' Psi' (packS0 rho_idle theta) pc (comml (_cmpw o r)).
Proof.
  intros D psi pc [p g tag]  D' Psi Psi' o r.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold IDLE_WFcmd in Hi.
  destruct Hi as [Hi Hpgik].
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_idle; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  (* interrupt *)
  split.
    destruct (itr_enabled_dec (M, R, fl, isr)) as [Hitr | Hitr];
        [right | left; intro H; destruct (Hitr H)].
    destruct Hitr as [Hfl Hisr].
    assert (Hih : IH_Specified Psi').
      unfold IH_Specified.
      apply lookupCT_subsetCT with (ext_IH_Spec (ext_Prim_Specs Psi)); trivial.
      unfold ext_IH_Spec.
      apply lookupCT_consCT; trivial.
    unfold WF_AIM_World in Hwf.
    destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
    simplS in H7.
    rename H7 into Hwfcth.
    unfold WF_Curr_Thrd in Hwfcth.
    simpl in Hwfcth.
    assert (Hflx : S_Flag (Mcth, R, fl, isr) _IF = true).
      simplS in * ; trivial.
    destruct (NH_Itr_prog (pc:=pc) Hpgik Hflx Hwfcth) as [Scth' HnextScth'].
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (itr_nextS_Mono2 H5 HJM H6 HnextScth') as [M' HnextS].
    exists (M', R', fl', isr'); split_l.
    clear Mcth' R' fl' isr' HnextScth' M' HnextS.
    intros [[[M' R'] fl'] isr'] Hitr.
    exists IH_Spec0.
    split.
      apply Hsub.
      unfold ext_IH_Spec.
      unfold OPSpec.lookup.
      simpl.
      left; trivial.
    simpl.
    rewrite minimalD_elim_ih; trivial.
    unfold Itp_AIM.
    exists T; exists B; exists RdyQ; exists tid.
    assert (Hstk : stkSpace (R SP) (IH_Stk + 6) Mcth).
      destruct Hpgik.
      eapply Prd_Itr_Stk_elim; unfold itr_enabled; eauto.
    destruct (itr_nextS_DMono2 H6 H5 HJM Hstk Hitr)
      as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS HTM']]]]]]].
    unfold pmstar.
    unfold WF_AIM_World.
    exists Mth'; exists Mq; split_l; split_r.
    split. 
      eapply WF_CurrTID_nh_ih; eauto.
    split_l; split_l; split_l.
    exists TM'; exists Mcth'; split_l.
    split_l.
    rsplit.
      split.
        apply WF_ReadyQ_Mono_CT_consCT; trivial.
        apply (WF_ReadyQ_untouch H8 HTM' H1).
      apply WF_WQSet_Mono_CT_consCT; trivial.
      apply (WF_WQSet_untouch H9 HTM' H1).
    unfold WF_currTID in H1.
    destruct H1 as [_ [_ [_ [tcb [Hmto [Htcb Hrtc]]]]]].
    clear B RdyQ H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear TM TM' HsplitTM'.
    unfold WF_Curr_Thrd.
    exists tcb; exists (TCB_Status tcb); split_l; split_l.
    eapply NH_Itr_prev; eauto.
  simpl in Hi.
  destruct Hi as [p' [g' [Hlookup' [Hen Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  unfold WF_AIM_World in Hwf.
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct (NH_cmpw_prog (o:=o) (r:=r) (pc:=pc) Hwfcth) as [Scth' HnextScth].
  split.
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (nextS_Mono2 H5 HJM H6 HnextScth) as [M' HnextS].
    exists (M', R', fl', isr'); trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  rewrite (nextPC_inversion HnextPC).
  subst R' isr'.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  clear Scth' HnextScth.
  assert (Hlk' : lookupCT Psi' (pc+1) (packS0 rho_idle (Ast p' g' tag))).
    apply Hsub.
    unfold ext_IH_Spec, ext_Prim_Specs.
    apply lookupCT_subsetCT with Psi.
    apply Hlift; trivial.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply subsetCT_refl.
  exists (packS0 rho_idle (Ast p' g' tag)).
  split_l.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_idle; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  simplS.
  destruct HM.
  simplS in * .
  exists Mth'; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  simplS.
  split.
    clear B RdyQ H1 H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC Hpgik HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd.
    simpl.
    eapply NH_cmpw_prev; eauto.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IDLE_in_CAP0_addw :
  forall D psi pc theta D' Psi Psi' o r,
    IDLE_WFcmd psi pc theta (comml (_addw o r))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_idle Psi
    -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
    -> WFcmd D' Psi' (packS0 rho_idle theta) pc (comml (_addw o r)).
Proof.
  intros D psi pc [p g tag]  D' Psi Psi' o r.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold IDLE_WFcmd in Hi.
  destruct Hi as [Hi Hpgik].
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_idle; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  (* interrupt *)
  split.
    destruct (itr_enabled_dec (M, R, fl, isr)) as [Hitr | Hitr];
        [right | left; intro H; destruct (Hitr H)].
    destruct Hitr as [Hfl Hisr].
    assert (Hih : IH_Specified Psi').
      unfold IH_Specified.
      apply lookupCT_subsetCT with (ext_IH_Spec (ext_Prim_Specs Psi)); trivial.
      unfold ext_IH_Spec.
      apply lookupCT_consCT; trivial.
    unfold WF_AIM_World in Hwf.
    destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
    simplS in H7.
    rename H7 into Hwfcth.
    unfold WF_Curr_Thrd in Hwfcth.
    simpl in Hwfcth.
    assert (Hflx : S_Flag (Mcth, R, fl, isr) _IF = true).
      simplS in * ; trivial.
    destruct (NH_Itr_prog (pc:=pc) Hpgik Hflx Hwfcth) as [Scth' HnextScth'].
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (itr_nextS_Mono2 H5 HJM H6 HnextScth') as [M' HnextS].
    exists (M', R', fl', isr'); split_l.
    clear Mcth' R' fl' isr' HnextScth' M' HnextS.
    intros [[[M' R'] fl'] isr'] Hitr.
    exists IH_Spec0.
    split.
      apply Hsub.
      unfold ext_IH_Spec.
      unfold OPSpec.lookup.
      simpl.
      left; trivial.
    simpl.
    rewrite minimalD_elim_ih; trivial.
    unfold Itp_AIM.
    exists T; exists B; exists RdyQ; exists tid.
    assert (Hstk : stkSpace (R SP) (IH_Stk + 6) Mcth).
      destruct Hpgik.
      eapply Prd_Itr_Stk_elim; unfold itr_enabled; eauto.
    destruct (itr_nextS_DMono2 H6 H5 HJM Hstk Hitr)
      as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS HTM']]]]]]].
    unfold pmstar.
    unfold WF_AIM_World.
    exists Mth'; exists Mq; split_l; split_r.
    split. 
      eapply WF_CurrTID_nh_ih; eauto.
    split_l; split_l; split_l.
    exists TM'; exists Mcth'; split_l.
    split_l.
    rsplit.
      split.
        apply WF_ReadyQ_Mono_CT_consCT; trivial.
        apply (WF_ReadyQ_untouch H8 HTM' H1).
      apply WF_WQSet_Mono_CT_consCT; trivial.
      apply (WF_WQSet_untouch H9 HTM' H1).
    unfold WF_currTID in H1.
    destruct H1 as [_ [_ [_ [tcb [Hmto [Htcb Hrtc]]]]]].
    clear B RdyQ H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear TM TM' HsplitTM'.
    unfold WF_Curr_Thrd.
    exists tcb; exists (TCB_Status tcb); split_l; split_l.
    eapply NH_Itr_prev; eauto.
  simpl in Hi.
  destruct Hi as [p' [g' [Hlookup' [Hen Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  unfold WF_AIM_World in Hwf.
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct (NH_addw_prog Hen Hwfcth) as [Scth' HnextScth].
  split.
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (nextS_Mono2 H5 HJM H6 HnextScth) as [M' HnextS].
    exists (M', R', fl', isr'); trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  rewrite (nextPC_inversion HnextPC).
  subst R' isr'.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  clear Scth' HnextScth.
  assert (Hlk' : lookupCT Psi' (pc+1) (packS0 rho_idle (Ast p' g' tag))).
    apply Hsub.
    unfold ext_IH_Spec, ext_Prim_Specs.
    apply lookupCT_subsetCT with Psi.
    apply Hlift; trivial.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply subsetCT_refl.
  exists (packS0 rho_idle (Ast p' g' tag)).
  split_l.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_idle; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  simplS.
  destruct HM.
  simplS in * .
  exists Mth'; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  simplS.
  split.
    clear B RdyQ H1 H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC Hpgik HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd.
    simpl.
    eapply NH_addw_prev; eauto.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IDLE_in_CAP0_subw :
  forall D psi pc theta D' Psi Psi' o r,
    IDLE_WFcmd psi pc theta (comml (_subw o r))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_idle Psi
    -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
    -> WFcmd D' Psi' (packS0 rho_idle theta) pc (comml (_subw o r)).
Proof.
  intros D psi pc [p g tag]  D' Psi Psi' o r.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold IDLE_WFcmd in Hi.
  destruct Hi as [Hi Hpgik].
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_idle; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  (* interrupt *)
  split.
    destruct (itr_enabled_dec (M, R, fl, isr)) as [Hitr | Hitr];
        [right | left; intro H; destruct (Hitr H)].
    destruct Hitr as [Hfl Hisr].
    assert (Hih : IH_Specified Psi').
      unfold IH_Specified.
      apply lookupCT_subsetCT with (ext_IH_Spec (ext_Prim_Specs Psi)); trivial.
      unfold ext_IH_Spec.
      apply lookupCT_consCT; trivial.
    unfold WF_AIM_World in Hwf.
    destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
    simplS in H7.
    rename H7 into Hwfcth.
    unfold WF_Curr_Thrd in Hwfcth.
    simpl in Hwfcth.
    assert (Hflx : S_Flag (Mcth, R, fl, isr) _IF = true).
      simplS in * ; trivial.
    destruct (NH_Itr_prog (pc:=pc) Hpgik Hflx Hwfcth) as [Scth' HnextScth'].
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (itr_nextS_Mono2 H5 HJM H6 HnextScth') as [M' HnextS].
    exists (M', R', fl', isr'); split_l.
    clear Mcth' R' fl' isr' HnextScth' M' HnextS.
    intros [[[M' R'] fl'] isr'] Hitr.
    exists IH_Spec0.
    split.
      apply Hsub.
      unfold ext_IH_Spec.
      unfold OPSpec.lookup.
      simpl.
      left; trivial.
    simpl.
    rewrite minimalD_elim_ih; trivial.
    unfold Itp_AIM.
    exists T; exists B; exists RdyQ; exists tid.
    assert (Hstk : stkSpace (R SP) (IH_Stk + 6) Mcth).
      destruct Hpgik.
      eapply Prd_Itr_Stk_elim; unfold itr_enabled; eauto.
    destruct (itr_nextS_DMono2 H6 H5 HJM Hstk Hitr)
      as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS HTM']]]]]]].
    unfold pmstar.
    unfold WF_AIM_World.
    exists Mth'; exists Mq; split_l; split_r.
    split. 
      eapply WF_CurrTID_nh_ih; eauto.
    split_l; split_l; split_l.
    exists TM'; exists Mcth'; split_l.
    split_l.
    rsplit.
      split.
        apply WF_ReadyQ_Mono_CT_consCT; trivial.
        apply (WF_ReadyQ_untouch H8 HTM' H1).
      apply WF_WQSet_Mono_CT_consCT; trivial.
      apply (WF_WQSet_untouch H9 HTM' H1).
    unfold WF_currTID in H1.
    destruct H1 as [_ [_ [_ [tcb [Hmto [Htcb Hrtc]]]]]].
    clear B RdyQ H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear TM TM' HsplitTM'.
    unfold WF_Curr_Thrd.
    exists tcb; exists (TCB_Status tcb); split_l; split_l.
    eapply NH_Itr_prev; eauto.
  simpl in Hi.
  destruct Hi as [p' [g' [Hlookup' [Hen Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  unfold WF_AIM_World in Hwf.
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct (NH_subw_prog Hen Hwfcth) as [Scth' HnextScth].
  split.
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (nextS_Mono2 H5 HJM H6 HnextScth) as [M' HnextS].
    exists (M', R', fl', isr'); trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  rewrite (nextPC_inversion HnextPC).
  subst R' isr'.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  clear Scth' HnextScth.
  assert (Hlk' : lookupCT Psi' (pc+1) (packS0 rho_idle (Ast p' g' tag))).
    apply Hsub.
    unfold ext_IH_Spec, ext_Prim_Specs.
    apply lookupCT_subsetCT with Psi.
    apply Hlift; trivial.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply subsetCT_refl.
  exists (packS0 rho_idle (Ast p' g' tag)).
  split_l.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_idle; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  simplS.
  destruct HM.
  simplS in * .
  exists Mth'; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  simplS.
  split.
    clear B RdyQ H1 H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC Hpgik HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd.
    simpl.
    eapply NH_subw_prev; eauto.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IDLE_in_CAP0_jmp :
  forall D psi pc theta D' Psi Psi' f,
    IDLE_WFcmd psi pc theta (comms (_jmp f))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_idle Psi
    -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
    -> WFcmd D' Psi' (packS0 rho_idle theta) pc (comms (_jmp f)).
Proof.
  intros D psi pc [p g tag]  D' Psi Psi' f.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold IDLE_WFcmd in Hi.
  destruct Hi as [Hi Hpgik].
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_idle; trivial. 
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  (* interrupt *)
  split.
    destruct (itr_enabled_dec (M, R, fl, isr)) as [Hitr | Hitr];
        [right | left; intro H; destruct (Hitr H)].
    destruct Hitr as [Hfl Hisr].
    assert (Hih : IH_Specified Psi').
      unfold IH_Specified.
      apply lookupCT_subsetCT with (ext_IH_Spec (ext_Prim_Specs Psi)); trivial.
      unfold ext_IH_Spec.
      apply lookupCT_consCT; trivial.
    unfold WF_AIM_World in Hwf.
    destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
    simplS in H7.
    rename H7 into Hwfcth.
    unfold WF_Curr_Thrd in Hwfcth.
    simpl in Hwfcth.
    assert (Hflx : S_Flag (Mcth, R, fl, isr) _IF = true).
      simplS in * ; trivial.
    destruct (NH_Itr_prog (pc:=pc) Hpgik Hflx Hwfcth) as [Scth' HnextScth'].
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (itr_nextS_Mono2 H5 HJM H6 HnextScth') as [M' HnextS].
    exists (M', R', fl', isr'); split_l.
    clear Mcth' R' fl' isr' HnextScth' M' HnextS.
    intros [[[M' R'] fl'] isr'] Hitr.
    exists IH_Spec0.
    split.
      apply Hsub.
      unfold ext_IH_Spec.
      unfold OPSpec.lookup.
      simpl.
      left; trivial.
    simpl.
    rewrite minimalD_elim_ih; trivial.
    unfold Itp_AIM.
    exists T; exists B; exists RdyQ; exists tid.
    assert (Hstk : stkSpace (R SP) (IH_Stk + 6) Mcth).
      destruct Hpgik.
      eapply Prd_Itr_Stk_elim; unfold itr_enabled; eauto.
    destruct (itr_nextS_DMono2 H6 H5 HJM Hstk Hitr)
      as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS HTM']]]]]]].
    unfold pmstar.
    unfold WF_AIM_World.
    exists Mth'; exists Mq; split_l; split_r.
    split. 
      eapply WF_CurrTID_nh_ih; eauto.
    split_l; split_l; split_l.
    exists TM'; exists Mcth'; split_l.
    split_l.
    rsplit.
      split.
        apply WF_ReadyQ_Mono_CT_consCT; trivial.
        apply (WF_ReadyQ_untouch H8 HTM' H1).
      apply WF_WQSet_Mono_CT_consCT; trivial.
      apply (WF_WQSet_untouch H9 HTM' H1).
    unfold WF_currTID in H1.
    destruct H1 as [_ [_ [_ [tcb [Hmto [Htcb Hrtc]]]]]].
    clear B RdyQ H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear TM TM' HsplitTM'.
    unfold WF_Curr_Thrd.
    exists tcb; exists (TCB_Status tcb); split_l; split_l.
    eapply NH_Itr_prev; eauto.
  destruct Hi as [p' [g' [Hlk' [Hpp' Hg'g]]]].
  unfold WF_AIM_World in Hwf.  
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  split.
    exists (M, R, fl, isr); constructor; trivial.
  
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  rewrite (nextPC_inversion HnextPC).
  subst R' fl' isr'.
  assert (HnextScth : nextS pc (comms (_jmp f)) (Mcth, R, fl, isr) (Mcth, R, fl, isr)).
    constructor; trivial.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  clear HnextScth.
  assert (Hlk'' : lookupCT Psi' f (packS0 rho_idle (Ast p' g' tag))).
    apply Hsub.
    unfold ext_IH_Spec, ext_Prim_Specs.
    apply lookupCT_subsetCT with Psi.
    apply Hlift; trivial.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply subsetCT_refl.
  exists (packS0 rho_idle (Ast p' g' tag)).
  split_l.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_idle; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  simplS.
  destruct HM.
  simplS in * .
  exists Mth'; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  simplS.
  split.
    clear B RdyQ H1 H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC Hpgik HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd.
    simpl.
    eapply NH_jmp_prev; eauto.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IDLE_in_CAP0_retw :
  forall D psi pc theta D' Psi Psi',
    IDLE_WFcmd psi pc theta (comms (_retw))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_idle Psi
    -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
    -> WFcmd D' Psi' (packS0 rho_idle theta) pc (comms (_retw)).
Proof.
  intros D psi pc [p g tag]  D' Psi Psi'.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold IDLE_WFcmd in Hi.
  destruct Hi as [Hi Hpgik].
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_idle; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  (* interrupt *)
  split.
    destruct (itr_enabled_dec (M, R, fl, isr)) as [Hitr | Hitr];
        [right | left; intro H; destruct (Hitr H)].
    destruct Hitr as [Hfl Hisr].
    assert (Hih : IH_Specified Psi').
      unfold IH_Specified.
      apply lookupCT_subsetCT with (ext_IH_Spec (ext_Prim_Specs Psi)); trivial.
      unfold ext_IH_Spec.
      apply lookupCT_consCT; trivial.
    unfold WF_AIM_World in Hwf.
    destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
    simplS in H7.
    rename H7 into Hwfcth.
    unfold WF_Curr_Thrd in Hwfcth.
    simpl in Hwfcth.
    assert (Hflx : S_Flag (Mcth, R, fl, isr) _IF = true).
      simplS in * ; trivial.
    destruct (NH_Itr_prog (pc:=pc) Hpgik Hflx Hwfcth) as [Scth' HnextScth'].
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (itr_nextS_Mono2 H5 HJM H6 HnextScth') as [M' HnextS].
    exists (M', R', fl', isr'); split_l.
    clear Mcth' R' fl' isr' HnextScth' M' HnextS.
    intros [[[M' R'] fl'] isr'] Hitr.
    exists IH_Spec0.
    split.
      apply Hsub.
      unfold ext_IH_Spec.
      unfold OPSpec.lookup.
      simpl.
      left; trivial.
    simpl.
    rewrite minimalD_elim_ih; trivial.
    unfold Itp_AIM.
    exists T; exists B; exists RdyQ; exists tid.
    assert (Hstk : stkSpace (R SP) (IH_Stk + 6) Mcth).
      destruct Hpgik.
      eapply Prd_Itr_Stk_elim; unfold itr_enabled; eauto.
    destruct (itr_nextS_DMono2 H6 H5 HJM Hstk Hitr)
      as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS HTM']]]]]]].
    unfold pmstar.
    unfold WF_AIM_World.
    exists Mth'; exists Mq; split_l; split_r.
    split. 
      eapply WF_CurrTID_nh_ih; eauto.
    split_l; split_l; split_l.
    exists TM'; exists Mcth'; split_l.
    split_l.
    rsplit.
      split.
        apply WF_ReadyQ_Mono_CT_consCT; trivial.
        apply (WF_ReadyQ_untouch H8 HTM' H1).
      apply WF_WQSet_Mono_CT_consCT; trivial.
      apply (WF_WQSet_untouch H9 HTM' H1).
    unfold WF_currTID in H1.
    destruct H1 as [_ [_ [_ [tcb [Hmto [Htcb Hrtc]]]]]].
    clear B RdyQ H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear TM TM' HsplitTM'.
    unfold WF_Curr_Thrd.
    exists tcb; exists (TCB_Status tcb); split_l; split_l.
    eapply NH_Itr_prev; eauto.
  simpl in Hi.
  destruct Hi as [Htag Hp].
  unfold WF_AIM_World in Hwf.
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct (NH_retw_prog (pc:=pc) Htag Hp Hwfcth) as [Scth' HnextScth].
  split.
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (nextS_Mono2 H5 HJM H6 HnextScth) as [M' HnextS].
    exists (M', R', fl', isr'); trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R']fl']isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr']]].
  destruct HR as [pc'x [Hlds HR]].
  subst M' R' fl' isr'.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  clear Scth' HnextScth.
  assert (pc' = pc'x).
    eapply leadsToW_unique; eauto.
    simplS in * ; subst; trivial.
    inversion_clear HnextPC.
    subst; trivial.
  subst pc'x.
  assert (HnextPC' : nextPC (comms _retw) Mcth R fl pc pc').
    inversion_clear HnextPC.
    destruct (nextS_inversion HnextS') as [_ [HR [_ _]]].
    destruct HR as [pc'x [Hlds' HR']].
    assert (Heq : pc' = pc'x).
      eapply leadsToW_unique; eauto.
      destruct (MemSplit_Inv H5 H6) as [Moth [Hx _]].
      simplS in * .
      apply (leadsToW_Mono (addr := sp)(x := pc'x) HJM); eauto.
      eapply leadsToW_Mono; eauto.
      subst; trivial.
    subst pc'x.
    apply nextPC_retw with (R SP); trivial.
  destruct (NH_retw_prev Htag HnextPC' HnextS' Hp Hwfcth) as [theta' [Hlk' Hwf]].
  exists (packS0 rho_idle theta'); split_l.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_idle; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  exists Mth'; exists Mq; split_l.
  split_r.
  unfold WF_AIM_World.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  split.
    clear B RdyQ HtoMcth' HTM' H1 H2 H3 H4 H5 H6 H8 H9 HMq.
    clear Mq Mth TM HnextS HJM HJM'.
    unfold WF_Curr_Thrd.
    simpl; simplS.
    trivial.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IDLE_in_CAP0_cmd:
  forall D psi pc theta c D' Psi Psi',
    IDLE_WFcmd psi pc theta c
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_idle Psi
    -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
    -> WFcmd D' Psi' (packS0 rho_idle theta) pc c.
Proof.
  intros D psi pc [p g tag] c D' Psi Psi'.
  intros Hi HD HD' Hlift HPsi'.
  induction c as [ i | s]; [ induction i | induction s].
  (* 19 goals remain, i = cli *)
  eapply SCAP_IDLE_in_CAP0_cli; eauto.
  (* 18 goals remain, i = sti *)
  eapply SCAP_IDLE_in_CAP0_sti; eauto.
  (* 17 goals remain, i = call f *)
  eapply SCAP_IDLE_in_CAP0_call; eauto.
  (* 16 goals remain, i = jne f *)
  eapply SCAP_IDLE_in_CAP0_jne; eauto.
  (* 15 goals remain, i = movwp o r *)
  eapply SCAP_IDLE_in_CAP0_movwp; eauto.
  (* 14 goals remain, i = movwld a r *)
  eapply SCAP_IDLE_in_CAP0_movwld; eauto.
  (* 13 goals remain, i = movwst o a *)
  eapply SCAP_IDLE_in_CAP0_movwst; eauto.
  (* 12 goals remain, i = pushwp o *)
  eapply SCAP_IDLE_in_CAP0_pushwp; eauto.
  (* 11 goals remain, i = pushwm a *)
  eapply SCAP_IDLE_in_CAP0_pushwm; eauto.
  (* 10 goals remain, i = pushfw *)
  eapply SCAP_IDLE_in_CAP0_pushfw; eauto.
  (* 9 goals remain, i = popw r *)
  eapply SCAP_IDLE_in_CAP0_popw; eauto.
  (* 8 goals remain, i = popfw *)
  eapply SCAP_IDLE_in_CAP0_popfw; eauto.
  (* 7 goals remain, i = cmpw o r *)
  eapply SCAP_IDLE_in_CAP0_cmpw; eauto.
  (* 6 goals remain, i = addw o r *)
  eapply SCAP_IDLE_in_CAP0_addw; eauto.
  (* 5 goals remain, i = subw o r *)
  eapply SCAP_IDLE_in_CAP0_subw; eauto.
  (* 4 goals remain, i = outb eoi *)
  destruct Hi as [Hi _]; destruct Hi.
  (* 3 goals remain, i = jmp f *)
  eapply SCAP_IDLE_in_CAP0_jmp; eauto.
  (* 2 goals remain, i = retw *)
  eapply SCAP_IDLE_in_CAP0_retw; eauto.
  (* 1 goals remain, i = iret *)
  simpl in Hi.
  destruct Hi as [Hi _ ]; destruct Hi.
Qed.
Implicit Arguments SCAP_IDLE_in_CAP0_cmd [D psi pc theta c D' Psi Psi'].

Lemma SCAP_IDLE_in_CAP0_iseq :
  forall D psi pc theta isq D' Psi Psi',
    IDLE_WFiseq psi pc theta isq
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_idle Psi
    -> Psi' = ext_IH_Spec (ext_Prim_Specs Psi)
    -> WFiseq D' Psi' (packS0 rho_idle theta) pc isq.
Proof.
  intros D psi pc theta isq D' Psi Psi'.
  intros Hscap HD HD' Hlift HPsi'.
  gen_clear theta pc Hscap HPsi'.
  induction isq; intros theta pc Hscap HPsi'.
    inversion Hscap. subst psi0 pc0 theta i0 isq0.
    eapply wfis_seq with (packS0 rho_idle (Ast p' g' tag)).
    assert (Hlift' := lift_psi_cons (f := pc+1) (spec := Ast p' g' tag) Hlift).
    apply WFcmd_Mono_CT with (ext_IH_Spec (ext_Prim_Specs 
        (consCT (pc + 1) (packS0 rho_idle (Ast p' g' tag)) Psi))).
    subst Psi'.
    unfold ext_IH_Spec, ext_Prim_Specs.
    apply Psi_extend3_subsetCT; auto.
    apply lookupCT_lookupCT_consCT.
    apply lookupCT_consCT.
    apply Psi_extend3_subsetCT; auto.
    apply lookupCT_lookupCT_consCT.
    apply lookupCT_lookupCT_consCT.
    apply lookupCT_consCT.
    apply Psi_extend3_subsetCT; auto.
    apply lookupCT_lookupCT_consCT.
    apply lookupCT_lookupCT_consCT.
    apply lookupCT_lookupCT_consCT.
    apply lookupCT_consCT.
    apply Psi_extend3_subsetCT; auto.
    apply lookupCT_lookupCT_consCT.
    apply lookupCT_lookupCT_consCT.
    apply lookupCT_lookupCT_consCT.
    apply lookupCT_lookupCT_consCT.
    apply lookupCT_consCT.
    apply Psi_extend2_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply Psi_extend_subsetCT.
    apply consCT_subsetCT.
    assert (HPsi'2 := refl_equal (ext_IH_Spec (ext_Prim_Specs 
        (consCT (pc + 1) (packS0 rho_idle (Ast p' g' tag)) Psi)))).
    apply (SCAP_IDLE_in_CAP0_cmd H4 HD HD' Hlift' HPsi'2).
    apply IHisq; trivial.
  inversion Hscap. subst psi0 pc0 theta si.
  eapply wfis_end.
  apply (SCAP_IDLE_in_CAP0_cmd H3 HD HD' Hlift HPsi').
Qed.

Lemma SCAP_IDLE_in_CAP0_cdhp :
  forall D psi C psi' D' Psi Psi' Psi'',
    IDLE_WFcdhp psi C psi'
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_idle Psi
    -> lift_psi psi' rho_idle Psi'
    -> sink_psi psi' rho_idle Psi'
    -> Psi'' = ext_IH_Spec (ext_Prim_Specs Psi)
    -> WFcdhp D' Psi'' C Psi'.
Proof.
  intros D psi C psi' D' Psi Psi' Psi''.
  intros H HD HD' H0 H1 H2 HPsi''.
  intros f theta0 Hf.
  substH H2 with (H2 f theta0 Hf).
  destruct H2 as [theta [Htheta0 Hfx]].
  unfold SCAP_WFcdhp in H.
  substH H with (H f theta Hfx).
  destruct H as [isq [Hisq H]].
  exists isq; split_l.
  destruct theta0.
  injection Htheta0; intros; subst n s.
  eapply SCAP_IDLE_in_CAP0_iseq; eauto.
Qed.

Lemma SCAP_IDLE_in_SCAP_RDY_cmd : forall psi pc theta c,
  IDLE_WFcmd psi pc theta c
  -> NH_WFcmd psi pc theta c.
Proof.
  intros psi pc [p g tag] c.
  induction c as [ i | s]; [ induction i | induction s];
    unfold IDLE_WFcmd, NH_WFcmd, IDLE_WFcmd_aux, NH_WFcmd_aux, IDLE_Call_Prim;
      auto.
  intuition.
Qed.

Lemma SCAP_IDLE_in_SCAP_RDY_iseq : forall psi pc theta isq,
  IDLE_WFiseq psi pc theta isq
  -> NH_WFiseq psi pc theta isq.
Proof.
  intros psi pc theta isq.
  gen_clear pc theta.
  induction isq; intros pc theta Hidle.
    inversion Hidle. subst psi0 pc0 theta i0 isq0.
    eapply nh_isq_seq with p' g'; trivial.
    apply SCAP_IDLE_in_SCAP_RDY_cmd; trivial.
    apply IHisq; trivial.
  inversion Hidle. subst psi0 pc0 theta si.
  eapply nh_isq_end.
  apply SCAP_IDLE_in_SCAP_RDY_cmd; trivial.
Qed.

Lemma SCAP_IDLE_in_SCAP_RDY_cdhp :
  forall psi C psi',
    IDLE_WFcdhp psi C psi'
    -> NH_WFcdhp psi C psi'.
Proof.
  intros psi C psi' Hidle.
  intros f theta Hf.
  substH Hidle with (Hidle f theta Hf).
  destruct Hidle as [isq [Hisq Hidle]].
  exists isq; split_l.
  apply SCAP_IDLE_in_SCAP_RDY_iseq; trivial.
Qed.
