(* This SCAP is used for interrupt hander *)

Add LoadPath "..".
Add LoadPath "../CAP0".

Require Import Coqlib.
Require Import Map.
Require Import MultiMapT.
Require Import Arith.
Require Import Machine.
Require Import Assertion.
Require Import Operational_Semantics.
Require Import CAP0.
Require Import SCAP.
Require Import Bool.
Require Import AIM.
Require Import SCAP.
Require Import SCAP_READY.

Set Implicit Arguments.
Unset Strict Implicit.


(* This is a simplified version of AIM-3 in that         *)
(* we do not allow the handler to enable interrupt again *)

(* trashes the stack space _SW_Stk *)
Definition ih_gy : Grt :=
  fun (S S' : State) => 
    exists sp,
      prepost (mstar INV0 (stkSpacePrec sp _SW_Stk)) (mstar INV0 (stkSpacePrec sp _SW_Stk)) S S'
        /\ S_Flag S _IF = S_Flag S' _IF
        /\ S_ISR S = S_ISR S'
        /\ S_RF S SP = sp
        /\ S_RF S SP = S_RF S' SP.

Definition ih_enable_sw (p : Pred) : Prop :=
    forall S, p S -> (pmstar pTrue (mstar INV0 (stkSpacePrec (S_RF S SP) (_SW_Stk + 2)))) S
                     /\ (S_ISR S = false).

Definition IH_Call_SW (psi : SCAP_PSpec) (pc f : flab) (theta : Spec) : Prop :=
  match theta with
    Ast p g tag =>
        (f = _SWITCH)
        /\ exists p', exists g',
              SCAP_lookupCT psi (pc+1) (Ast p' g' tag)
              /\ ih_enable_sw p
              /\ (forall pc', pimp (p |> (wrap_fun ih_gy pc f pc')) p')
              /\ (forall pc', gimp (p >o (wrap_fun ih_gy pc f pc') oo g') g)
  end.

Definition IH_WFcmd (psi : SCAP_PSpec) (pc : flab) (theta : Spec) (c : comm) : Prop :=
  match theta with
    Ast p g tag =>
    match c with
      comms si =>
      match si with
        _iret => (tag = 0) /\ (forall S, p S -> g S S)
      | _retw => (tag = 1) /\ (forall S, p S -> g S S)
      | _     => SCAP_WFcmd psi pc theta c
      end
      | comml i  =>
      match i with
        | _cli => False
        | _sti => False
        | _outb_eoi => exists p', exists g',
                       SCAP_lookupCT psi (pc+1) (Ast p' g' tag)
                       /\ enable p (nextS pc c)
                       /\ pimp (p |> (nextS pc c)) p'
                       /\ gimp (p >o (nextS pc c) oo g') g
        | _call f => (prim_entry f /\ IH_Call_SW psi pc f theta)
                     \/ (~ (prim_entry f) /\ NH_Normal_Call psi pc f theta)
        | _pushfw => False
        | _popfw => False
        | _ => SCAP_WFcmd psi pc theta c
      end
    end
  end.

Inductive IH_WFiseq : SCAP_PSpec -> flab -> Spec -> iseq -> Prop :=
  | ih_isq_seq :
        forall psi pc p p' g g' i isq tag,
          IH_WFcmd (SCAP_consCT (pc+1) (Ast p' g' tag) psi) pc (Ast p g tag) (comml i)
          -> IH_WFiseq psi (pc+1) (Ast p' g' tag) isq
          -> IH_WFiseq psi pc (Ast p g tag) (seq i isq)
  | ih_isq_end :
        forall psi pc p g si tag,
          IH_WFcmd psi pc (Ast p g tag) (comms si)
          -> IH_WFiseq psi pc (Ast p g tag) (sseq si).

Definition IH_WFcdhp (psi : SCAP_PSpec) (C: CodeHeap) (psi' : SCAP_PSpec): Prop :=
  forall f theta,
    SCAP_lookupCT psi' f theta -> exists isq, lookupC C f isq /\ IH_WFiseq psi f theta isq.

