Add LoadPath "..".
Add LoadPath "../CAP0".

Require Import Arith.
Require Import Coqlib.
Require Import Map.
Require Import Machine.
Require Import Assertion.
Require Import Assertion_Lemmas.
Require Import Assertion_Lemmas5.
Require Import Operational_Semantics.
Require Import OpSem_Lemmas.

Require Import CAP0.
Require Import SCAP.
Require Import AIM.
Require Import QLib.
Require Import SCAP_READY.
Require Import SCAP_Itr.

Require Import EqNat.

Require Import Linking.
Require Import Linking_Lemmas2.

Lemma IH_iret_prog : forall status Psi pc S0
  (p : Pred) (g : Grt) tag,
  tag = 0
  -> (forall S, p S -> g S S)
  -> itp_IH_theta (Ast p g tag) status Psi S0
  -> exists S', nextS pc (comms _iret) S0 S'.
Proof.
  intros status Psi pc [[[M R] fl] isr] p g tag Htag Hpg H.
  unfold itp_IH_theta in H.
  subst tag.
  destruct H as [Hisrf [Hp | [Htag _]]]; [idtac | discriminate Htag].
  unfold lift_IH_PG, pmstar in Hp.
  destruct Hp as [_ [Hp Hwfst]].
  destruct Hp as [Mih [Mnh [HJM [HMih HMnh]]]].
  simplS in * .
  assert (Hg : IH_ItpG g (M, R, fl, isr) (M, R, fl, isr)).
    simplS.
    unfold IH_ItpG.
    exists Mih; exists Mnh.
    exists Mih; exists Mnh.
    split_l; split_l; split_l; split_l.
    apply Hpg; trivial.
  substH Hwfst with (Hwfst _ Hg).
  destruct Hwfst as [S' [HnextS [ra [Hra [rho [theta [Hrho [Hlkra Hwfst]]]]]]]].
  assert (Heq : leadsToW (R SP) ra M).
    unfold Ret_Addr in Hra.
    simpl in Hra.
    substH Hra with (Hra pc).
    inversion_clear Hra.
    subst sp.
    trivial.
  destruct S' as [[[M' R'] fl'] isr'].
  exists (M', R', fl', isr').
  apply HnextS.
Qed.
Implicit Arguments IH_iret_prog [status Psi pc S0 p g tag].

Lemma IH_iret_prev : forall status Psi pc M R fl isr pc' S'
  (p : Pred) (g : Grt) tag,
  tag = 0
  -> nextPC (comms _iret) M R fl pc pc'
  -> nextS pc (comms _iret) (M, R, fl, isr) S'
  -> (forall S, p S -> g S S)
  -> itp_IH_theta (Ast p g tag) status Psi (M,R,fl,isr)
  -> exists theta', exists rho',
    (rho' = rho_ready \/ rho' = rho_idle)
    /\ rho_tcb_consistent2 status rho'
    /\ lookupCT Psi pc' (packS0 rho' theta')
    /\ itp_NH_theta rho' theta' Psi S'.
Proof.
  intros status Psi pc M R fl isr pc' [[[M' R'] fl'] isr'] p g tag. 
  intros Htag HnextPC HnextS Hpg H.
  subst tag.
  destruct H as [Hisrf [Hp | [Htag _]]]; [idtac | discriminate Htag].
  destruct Hp as [_ [Hp Hwfst]].
  destruct Hp as [Mih [Mnh [HJM [HMih HMnh]]]].
  simplS in * .
  assert (Hg : IH_ItpG g (M, R, fl, isr) (M, R, fl, isr)).
    simplS.
    unfold IH_ItpG.
    exists Mih; exists Mnh.
    exists Mih; exists Mnh.
    split_l; split_l; split_l; split_l.
    apply Hpg; trivial.
  substH Hwfst with (Hwfst _ Hg).
  destruct (nextS_inversion HnextS) as [HM' [HR' [Hfl' Hisr']]].
  destruct HM' as [HM' Hlds].
  destruct Hfl' as [fln [Hfl' Hlds']].
  subst M' R' fl' isr'.
  inversion_clear HnextPC.
  subst sp.
  destruct Hwfst as [S' [HnextS' [ra [Hra [rho [theta [Hrho [Htcb [Hlkra Hwfst]]]]]]]]].
  substH Hwfst with (Hwfst pc _ HnextS).
  assert (Hlds2 : leadsToW (R SP) ra M).
    unfold Ret_Addr in Hra.
    simpl in Hra.
    substH Hra with (Hra pc).
    inversion_clear Hra.
    subst sp.
    trivial.
  assert (Heq : ra = pc').
    eapply leadsToW_unique; eauto.
  subst pc'.
  exists theta; exists rho; auto.
Qed.
Implicit Arguments IH_iret_prev [status Psi pc M R fl isr pc' S' p g tag].

Lemma IH_call_prog : forall status Psi f pc S0
  (p : Pred) (g : Grt) tag,
  enable p (nextS pc (comml (_call f)))
  -> itp_IH_theta (Ast p g tag) status Psi S0
  -> exists S', nextS pc (comml (_call f)) S0 S'.
Proof.
  intros status Psi f pc [[[M R] fl] isr] p g tag.
  intros Hen H.
  unfold itp_IH_theta in * .
  destruct H as [Hiff [Htop | Hntop]].
    unfold lift_IH_PG in Htop.
    destruct Htop as [Htag [Hp Hg]].
    destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
    simplS in * .
    destruct (Hen _ HMnh) as [S' HnextSnh]. 
    destruct S' as [[[Mnh' R'] fl'] isr'].
    destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
    exists (M', R', fl', isr'); trivial.
  destruct Hntop as [Htag [Hp Hg]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  simplS in * .
  destruct (Hen _ HMnh) as [S' HnextSnh]. 
  destruct S' as [[[Mnh' R'] fl'] isr'].
  destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
  exists (M', R', fl', isr'); trivial.
Qed.
Implicit Arguments IH_call_prog [status Psi f pc S0 p g tag].

Lemma IH_nonprim_call_prev : forall status Psi f pc S0 S'
  (p p' p'': Pred) (g g' g'': Grt) tag,
  lookupCT Psi f (packS0 rho_ih (Ast p' g' 1))
  -> lookupCT Psi (pc+1) (packS0 rho_ih (Ast p'' g'' tag))
  -> nextS pc (comml (_call f)) S0 S'
  -> enable p (nextS pc (comml (_call f)))
  -> gimp (sp_call_pred p pc f >o g') fun_frame
  -> pimp (p |> nextS pc (comml (_call f))) p'
  -> (forall pc' : nat, pimp (p |> wrap_fun g' pc f pc') p'')
  -> (forall pc' : nat, gimp (p >o wrap_fun g' pc f pc' oo g'') g)
  -> itp_IH_theta (Ast p g tag) status Psi S0
  -> itp_IH_theta (Ast p' g' 1) status Psi S'.
Proof.
  intros status Psi f pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p' p'' g g' g'' tag.
  intros Hlk' Hlk'' HnextS Hen Hfun Hpp' Hg'p'' Hg''g H.
  unfold itp_IH_theta in * .
  destruct H as [Hiff [Htop | Hntop]].
    unfold lift_IH_PG in Htop.
    destruct Htop as [Htag [Hp Hg]].
    destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
    destruct (nextS_inversion HnextS) as [_ [HR [Hfl Hisr]]].
    subst R' fl' isr'.
    unfold enable in Hen.
    destruct (Hen _ HMnh) as [Sx HnextSSx].
    destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
    clear Sx HnextSSx Hen.
    simplS in * .
    split_l.
    assert (HS' : (p |> nextS pc (comml (_call f))) (Mnh', RF_update R SP (R SP - 2), fl, isr)).
      apply linkPG_intro with (Mnh,R,fl,isr); trivial.
    right.
    split_l.
    split.
      exists Mnh'; exists Mih; split_l.
      split_r.
      apply Hpp'; trivial.
    set (S := (M, R, fl, isr)) in * .
    set (S' := (M', RF_update R SP (R SP - 2), fl, isr)) in * .
    set (Snh := (Mnh, R, fl, isr)) in * .
    set (Snh' := (Mnh', RF_update R SP (R SP - 2), fl, isr)) in * .
    exists O.
    simpl.
    intros S'' Hg'.
    destruct S'' as [[[M'' R''] fl''] isr''].
    unfold NH_ItpG in Hg'.
    destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
    unfold S' in Hg'.
    simplS in Hg'.
    destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih Prec1) as [Heq1 Heq2].
    subst Mx My.
    clear HJMxy HMy.
    unfold wrap_fun in Hg'p''.
    simplS in * .
    set (S'' := (M'', R'', fl'', isr'')) in * .
    set (Snh'' := (Mnh'', R'', fl'', isr'')).
    assert (Hfunframe : fun_frame Snh' Snh'').
      apply Hfun.
      unfold sp_call_pred.
      apply conjPG_intro; trivial.
    unfold fun_frame in Hfunframe.
    destruct Hfunframe as [_ [ra [Hra' Hra'']]].
    assert (Hrax := ra_after_call _ _ Hra' HnextSnh').
    subst ra.
    exists (pc + 1).
    unfold S''.
    split.
      unfold S''.
      apply (Ret_Addr_Mono HJM'' Hra'').
    clear Hra'.
    exists (Ast p'' g'' tag).
    split_l.
    intros pc' S''' HnextS''.
    substH Hg'p'' with (Hg'p'' pc').
    assert (HnextSnh''Sx : nextS pc' (comms _retw) (Mnh'', R'', fl'', isr'') (Mnh'', RF_update R'' SP (R'' SP + 2), fl'', isr'')).
      unfold Ret_Addr in Hra''.
      substH Hra'' with (Hra'' pc').
      inversion_clear Hra''.
      subst Snh'' sp; simplS in * .
      apply nextS_retw.
      exists (pc+1).
      unfold popW.
      split_lr.
    destruct S''' as [[[M''' R'''] fl'''] isr'''].
    destruct (nextS_DMono HJM'' HnextSnh''Sx HnextS'') as [Mnh''' [HJM''' HnextSnh'']].
    destruct (nextS_inversion HnextSnh'') as [HM [_ [Hfl Hisr]]].
    subst Mnh''' fl''' isr'''.
    destruct (nextS_inversion HnextS'') as [HM _].
    subst M'''.
    split_l.
    split.
      unfold NH_pinvstar, pmstar.
      exists Mnh''; exists Mih''; split_l.
      split_r by (apply HMih'').
      apply Hg'p''.
      apply linkPG_intro with Snh; trivial.
      apply conjG_intro with Snh''; trivial.
      apply conjG_intro with Snh'; trivial.
    intros Sx H.
    destruct H as [Mnh''0 [Mih''0 [Mnhx [Mihx [HJM''0 [HJMx [HMih''0 [HMihx Hg'']]]]]]]].
    destruct (prec_disjUnion_inv HJM'' HJM''0 HMih'' HMih''0 Prec1) as [Heq1 Heq2].
    subst Mnh''0 Mih''0.
    clear HJM''0 HMih''0.
    assert (IH_ItpG g S Sx).
      unfold IH_ItpG.
      exists Mnh; exists Mih; exists Mnhx; exists Mihx.
      split_l; split_l; split_l; split_l.
      apply (Hg''g pc').
      apply conjG_intro with (Mnh'', R''', fl'', isr''); trivial.
      unfold wrap_fun.
      apply conjPG_intro; auto.
      apply conjG_intro with Snh''; subst S; trivial.
      apply conjG_intro with Snh'; trivial.
    apply (Hg _ H). 
  destruct Hntop as [Htag [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [_ [HR [Hfl Hisr]]].
  subst R' fl' isr'.
  unfold enable in Hen.
  destruct (Hen _ HMnh) as [Sx HnextSSx].
  destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
  clear Sx HnextSSx Hen.
  simplS in * .
  split_l.
  assert (HS' : (p |> nextS pc (comml (_call f))) (Mnh', RF_update R SP (R SP - 2), fl, isr)).
    apply linkPG_intro with (Mnh,R,fl,isr); trivial.
  right.
  split_l.
  split.
    exists Mnh'; exists Mih; split_l.
    split_r.
    apply Hpp'; trivial.
  set (S := (M, R, fl, isr)) in * .
  set (S' := (M', RF_update R SP (R SP - 2), fl, isr)) in * .
  set (Snh := (Mnh, R, fl, isr)) in * .
  set (Snh' := (Mnh', RF_update R SP (R SP - 2), fl, isr)) in * . 
  destruct Hwfst as [n Hwfst].
  exists (1+n).
  simpl.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold NH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  unfold S' in Hg'.
  simplS in Hg'.
  destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih Prec1) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  unfold wrap_fun in Hg'p''.
  simplS in * .
  set (S'' := (M'', R'', fl'', isr'')) in * .
  set (Snh'' := (Mnh'', R'', fl'', isr'')).
  assert (Hfunframe : fun_frame Snh' Snh'').
    apply Hfun.
    unfold sp_call_pred.
    apply conjPG_intro; trivial.
  unfold fun_frame in Hfunframe.
  destruct Hfunframe as [_ [ra [Hra' Hra'']]].
  assert (Hrax := ra_after_call _ _ Hra' HnextSnh').
  subst ra.
  exists (pc + 1).
  unfold S''.
  split.
    unfold S''.
    apply (Ret_Addr_Mono HJM'' Hra'').
  clear Hra'.
  exists p''; exists g''; exists tag.
  split_l.
  split_l.
  intros pc' S''' HnextS''.
  substH Hg'p'' with (Hg'p'' pc').
  assert (HnextSnh''Sx : nextS pc' (comms _retw) (Mnh'', R'', fl'', isr'') (Mnh'', RF_update R'' SP (R'' SP + 2), fl'', isr'')).
    unfold Ret_Addr in Hra''.
    substH Hra'' with (Hra'' pc').
    inversion_clear Hra''.
    subst Snh'' sp; simplS in * .
    apply nextS_retw.
    exists (pc+1).
    unfold popW.
    split_lr.
  destruct S''' as [[[M''' R'''] fl'''] isr'''].
  destruct (nextS_DMono HJM'' HnextSnh''Sx HnextS'') as [Mnh''' [HJM''' HnextSnh'']].
  destruct (nextS_inversion HnextSnh'') as [HM [_ [Hfl Hisr]]].
  subst Mnh''' fl''' isr'''.
  destruct (nextS_inversion HnextS'') as [HM _].
  subst M'''.
  split.
    unfold NH_pinvstar, pmstar.
    exists Mnh''; exists Mih''; split_l.
    split_r by (apply HMih'').
    apply Hg'p''.
    apply linkPG_intro with Snh; trivial.
    apply conjG_intro with Snh''; trivial.
    apply conjG_intro with Snh'; trivial.
  apply (IH_WFST_strengthen (S1 := (M'', R''', fl'', isr'')) (g':= g'') Hwfst).
  substH Hg''g with (Hg''g pc' Snh).
  unfold NH_ItpG.
  intros Sx H.
  destruct H as [Mnh''0 [Mih''0 [Mnhx [Mihx [HJM''0 [HJMx [HMih''0 [HMihx Hg'']]]]]]]].
  destruct Sx as [[[Mx Rx] flx] isrx].
  simplS in * .
  destruct (prec_disjUnion_inv HJM'' HJM''0 HMih'' HMih''0 Prec1) as [Heq1 Heq2].
  subst Mnh''0 Mih''0.
  clear HJM''0 HMih''0.
  exists Mnh; exists Mih; exists Mnhx; exists Mihx.
  split_l; split_l; split_l; split_l.
  apply Hg''g.
  apply conjG_intro with (Mnh'', R''', fl'', isr''); trivial.
  unfold wrap_fun.
  apply conjPG_intro; auto.
  apply conjG_intro with Snh''; subst S; trivial.
  apply conjG_intro with Snh'; trivial.
Qed.
Implicit Arguments IH_nonprim_call_prev [status Psi f pc S0 S' p p' p'' g g' g'' tag].

Lemma IH_jne_prog : forall status Psi f pc S0
  (p : Pred) (g : Grt) tag,
  itp_IH_theta (Ast p g tag) status Psi S0
  -> exists S', nextS pc (comml (_jne f)) S0 S'.
Proof.
  intros status Psi f pc [[[M R] fl] isr] p g tag.
  intros H.
  exists (M, R, fl, isr).
  constructor.
Qed.
Implicit Arguments IH_jne_prog [status Psi f pc S0 p g tag].

Lemma IH_jne_jmp_prev : forall status Psi f pc S0 S'
  (p p': Pred) (g g': Grt) tag,
  lookupCT Psi f (packS0 rho_ih (Ast p' g' tag))
  -> S_Flag S0 _ZF = false
  -> nextS pc (comml (_jne f)) S0 S'
  -> pimp (conjP p ZF_cl) p'
  -> gimp (p >o (ZF_cl >o g')) g
  -> itp_IH_theta (Ast p g tag) status Psi S0
  -> itp_IH_theta (Ast p' g' tag) status Psi S'.
Proof.
  intros status Psi f pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p' g g' tag.
  intros Hlk' Hflzf HnextS Hpp' Hg'g H.
  unfold itp_NH_theta in * .
  destruct H as [Hiff [Htop | Hntop]].
    unfold lift_IH_PG in Htop.
    destruct Htop as [Htag [Hp Hg]].
    destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
    destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
    subst M' R' fl' isr'.
    assert (HS' : (p |> nextS pc (comml (_jne f))) (Mnh, R, fl, isr)).
      apply linkPG_intro with (Mnh,R,fl,isr); trivial.
      constructor.
    split_l.
    left.
    split_l.
    split.
      exists Mnh; exists Mih.
      split_l.
      split_r.
      simplS.
      apply Hpp'; trivial.
      split; trivial.
    unfold pimp, NH_ItpG.
    intros S'' Hg'.
    destruct S'' as [[[M'' R''] fl''] isr''].
    unfold IH_ItpG in Hg'.
    destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
    simplS in Hg'.
    destruct (prec_disjUnion_inv HJMxy HJM HMy HMih Prec1) as [Heq1 Heq2].
    subst Mx My.
    clear HJMxy HMy.
    simplS.
    assert (IH_ItpG g (M, R, fl, isr) (M'', R'', fl'', isr'')).
      unfold IH_ItpG.
      exists Mnh; exists Mih.
      exists Mnh''; exists Mih''.
      split_l; split_l; split_l; split_l.
      apply Hg'g.
      apply conjPG_intro; trivial.
      apply conjPG_intro; trivial.
    apply Hg; trivial.
  destruct Hntop as [Htag [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  subst M' R' fl' isr'.
  assert (HS' : (p |> nextS pc (comml (_jne f))) (Mnh, R, fl, isr)).
    apply linkPG_intro with (Mnh,R,fl,isr); trivial.
    constructor.
  split_l.
  right.
  split_l.
  split.
    exists Mnh; exists Mih.
    split_l.
    split_r.
    simplS.
    apply Hpp'; trivial.
    split; trivial.
  destruct Hwfst as [n Hwfst].
  exists n.
  apply IH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, NH_ItpG.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold NH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  simplS in Hg'.
  destruct (prec_disjUnion_inv HJMxy HJM HMy HMih Prec1) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  simplS.
  exists Mnh; exists Mih.
  exists Mnh''; exists Mih''.
  split_l; split_l; split_l; split_l.
  apply Hg'g.
  apply conjPG_intro; trivial.
  apply conjPG_intro; trivial.
Qed.
Implicit Arguments IH_jne_jmp_prev [status Psi f pc S0 S' p p' g g' tag].

Lemma IH_jne_seq_prev : forall status Psi f pc S0 S'
  (p p'': Pred) (g g'': Grt) tag,
  lookupCT Psi (pc+1) (packS0 rho_ih (Ast p'' g'' tag))
  -> S_Flag S0 _ZF = true
  -> nextS pc (comml (_jne f)) S0 S'
  -> pimp (conjP p ZF_st) p''
  -> gimp (p >o (ZF_st >o g'')) g
  -> itp_IH_theta (Ast p g tag) status Psi S0
  -> itp_IH_theta (Ast p'' g'' tag) status Psi S'.
Proof.
  intros status Psi f pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p'' g g'' tag.
  intros Hlk'' Hflzf HnextS Hpp' Hg'g H.
  unfold itp_IH_theta in * .
  destruct H as [Hiff [Htop | Hntop]].
    unfold lift_IH_PG in Htop.
    destruct Htop as [Htag [Hp Hg]].
    destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
    destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
    subst M' R' fl' isr'.
    assert (HS' : (p |> nextS pc (comml (_jne f))) (Mnh, R, fl, isr)).
      apply linkPG_intro with (Mnh,R,fl,isr); trivial.
      constructor.
    split_l.
    left.
    split_l.
    split.
      exists Mnh; exists Mih.
      split_l.
      split_r.
      simplS.
      apply Hpp'; trivial.
      split; trivial.
    unfold pimp, NH_ItpG.
    intros S'' Hg'.
    destruct S'' as [[[M'' R''] fl''] isr''].
    unfold IH_ItpG in Hg'.
    destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
    simplS in Hg'.
    destruct (prec_disjUnion_inv HJMxy HJM HMy HMih Prec1) as [Heq1 Heq2].
    subst Mx My.
    clear HJMxy HMy.
    simplS.
    assert (IH_ItpG g (M, R, fl, isr) (M'', R'', fl'', isr'')).
      unfold IH_ItpG.
      exists Mnh; exists Mih.
      exists Mnh''; exists Mih''.
      split_l; split_l; split_l; split_l.
      apply Hg'g.
      apply conjPG_intro; trivial.
      apply conjPG_intro; trivial.
    apply Hg; trivial.
  destruct Hntop as [Htag [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  subst M' R' fl' isr'.
  assert (HS' : (p |> nextS pc (comml (_jne f))) (Mnh, R, fl, isr)).
    apply linkPG_intro with (Mnh,R,fl,isr); trivial.
    constructor.
  split_l.
  right.
  split_l.
  split.
    exists Mnh; exists Mih.
    split_l.
    split_r.
    simplS.
    apply Hpp'; trivial.
    split; trivial.
  destruct Hwfst as [n Hwfst].
  exists n.
  apply IH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, NH_ItpG.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold NH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  simplS in Hg'.
  destruct (prec_disjUnion_inv HJMxy HJM HMy HMih Prec1) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  simplS.
  exists Mnh; exists Mih.
  exists Mnh''; exists Mih''.
  split_l; split_l; split_l; split_l.
  apply Hg'g.
  apply conjPG_intro; trivial.
  apply conjPG_intro; trivial.
Qed.
Implicit Arguments IH_jne_seq_prev [status Psi f pc S0 S' p p'' g g'' tag].

Lemma IH_movwp_prog : forall status Psi o r pc S0
  (p : Pred) (g : Grt) tag,
  enable p (nextS pc (comml (_movwp o r)))
  -> itp_IH_theta (Ast p g tag) status Psi S0
  -> exists S', nextS pc (comml (_movwp o r)) S0 S'.
Proof.
  intros status Psi o r pc [[[M R] fl] isr] p g tag.
  intros Hen H.
  unfold itp_IH_theta in * .
  destruct H as [Hiff [Htop | Hntop]].
    unfold lift_IH_PG in Htop.
    destruct Htop as [Htag [Hp Hg]].
    destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
    simplS in * .
    destruct (Hen _ HMnh) as [S' HnextSnh]. 
    destruct S' as [[[Mnh' R'] fl'] isr'].
    destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
    exists (M', R', fl', isr'); trivial.
  destruct Hntop as [Htag [Hp Hg]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  simplS in * .
  destruct (Hen _ HMnh) as [S' HnextSnh]. 
  destruct S' as [[[Mnh' R'] fl'] isr'].
  destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
  exists (M', R', fl', isr'); trivial.
Qed.
Implicit Arguments IH_movwp_prog [status Psi o r pc S0 p g tag].

Lemma IH_movwp_prev : forall status Psi o r pc S0 S'
  (p p' : Pred) (g g' : Grt) tag,
  lookupCT Psi (pc+1) (packS0 rho_ih (Ast p' g' tag))
  -> nextS pc (comml (_movwp o r)) S0 S'
  -> enable p (nextS pc (comml (_movwp o r)))
  -> pimp (p |> nextS pc (comml (_movwp o r))) p'
  -> gimp (p >o nextS pc (comml (_movwp o r)) oo g') g
  -> itp_IH_theta (Ast p g tag) status Psi S0
  -> itp_IH_theta (Ast p' g' tag) status Psi S'.
Proof.
  intros status Psi o r pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p' g g' tag.
  intros Hlk' HnextS Hen Hpp' Hg'g H.
  unfold itp_IH_theta in * .
  destruct H as [Hiff [Htop | Hntop]].
    unfold lift_IH_PG in Htop.
    destruct Htop as [Htag [Hp Hg]].
    destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
    destruct (nextS_inversion HnextS) as [_ [HR [Hfl Hisr]]].
    subst fl' isr'.
    unfold enable in Hen.
    destruct (Hen _ HMnh) as [Sx HnextSSx].
    destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
    clear Sx HnextSSx Hen.
    simplS in * .
    split_l.
    assert (HS' : (p |> nextS pc (comml (_movwp o r))) (Mnh', R', fl, isr)).
      apply linkPG_intro with (Mnh,R,fl,isr); trivial.
    left.
    split_l.
    split.
      exists Mnh'; exists Mih; split_l.
      split_r.
      apply Hpp'; trivial.
    intros S'' Hg'.
    destruct S'' as [[[M'' R''] fl''] isr''].
    unfold NH_ItpG in Hg'.
    destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
    simplS in * .
    destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih Prec1) as [Heq1 Heq2].
    subst Mx My.
    clear HJMxy HMy.
    simplS.
    apply Hg.
    unfold IH_ItpG.
    exists Mnh; exists Mih.
    exists Mnh''; exists Mih''.
    split_l; split_l; split_l; split_l.
    apply Hg'g.
    simplS.
    apply conjG_intro with (Mnh', R', fl, isr); trivial.
    apply conjPG_intro; trivial.
  destruct Hntop as [Htag [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [_ [HR [Hfl Hisr]]].
  subst fl' isr'.
  unfold enable in Hen.
  destruct (Hen _ HMnh) as [Sx HnextSSx].
  destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
  clear Sx HnextSSx Hen.
  simplS in * .
  split_l.
  assert (HS' : (p |> nextS pc (comml (_movwp o r))) (Mnh', R', fl, isr)).
    apply linkPG_intro with (Mnh,R,fl,isr); trivial.
  right.
  split_l.
  split.
    exists Mnh'; exists Mih; split_l.
    split_r.
    apply Hpp'; trivial.
  destruct Hwfst as [n Hwfst].
  exists n.
  apply IH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, NH_ItpG.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold NH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  simplS in * .
  destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih Prec1) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  simplS.
  exists Mnh; exists Mih.
  exists Mnh''; exists Mih''.
  split_l; split_l; split_l; split_l.
  apply Hg'g.
  apply conjG_intro with (Mnh', R', fl, isr); trivial.
  apply conjPG_intro; trivial.
Qed.
Implicit Arguments IH_movwp_prev [status Psi o r pc S0 S' p p' g g' tag].

Lemma IH_movwld_prog : forall status Psi a r pc S0
  (p : Pred) (g : Grt) tag,
  enable p (nextS pc (comml (_movwld a r)))
  -> itp_IH_theta (Ast p g tag) status Psi S0
  -> exists S', nextS pc (comml (_movwld a r)) S0 S'.
Proof.
  intros status Psi a r pc [[[M R] fl] isr] p g tag.
  intros Hen H.
  unfold itp_IH_theta in * .
  destruct H as [Hiff [Htop | Hntop]].
    unfold lift_IH_PG in Htop.
    destruct Htop as [Htag [Hp Hg]].
    destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
    simplS in * .
    destruct (Hen _ HMnh) as [S' HnextSnh]. 
    destruct S' as [[[Mnh' R'] fl'] isr'].
    destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
    exists (M', R', fl', isr'); trivial.
  destruct Hntop as [Htag [Hp Hg]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  simplS in * .
  destruct (Hen _ HMnh) as [S' HnextSnh]. 
  destruct S' as [[[Mnh' R'] fl'] isr'].
  destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
  exists (M', R', fl', isr'); trivial.
Qed.
Implicit Arguments IH_movwld_prog [status Psi a r pc S0 p g tag].

Lemma IH_movwld_prev : forall status Psi a r pc S0 S'
  (p p' : Pred) (g g' : Grt) tag,
  lookupCT Psi (pc+1) (packS0 rho_ih (Ast p' g' tag))
  -> nextS pc (comml (_movwld a r)) S0 S'
  -> enable p (nextS pc (comml (_movwld a r)))
  -> pimp (p |> nextS pc (comml (_movwld a r))) p'
  -> gimp (p >o nextS pc (comml (_movwld a r)) oo g') g
  -> itp_IH_theta (Ast p g tag) status Psi S0
  -> itp_IH_theta (Ast p' g' tag) status Psi S'.
Proof.
  intros status Psi a r pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p' g g' tag.
  intros Hlk' HnextS Hen Hpp' Hg'g H.
  destruct H as [Hiff [Htop | Hntop]].
    unfold lift_IH_PG in Htop.
    destruct Htop as [Htag [Hp Hg]].
    destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
    destruct (nextS_inversion HnextS) as [_ [HR [Hfl Hisr]]].
    subst fl' isr'.
    unfold enable in Hen.
    destruct (Hen _ HMnh) as [Sx HnextSSx].
    destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
    clear Sx HnextSSx Hen.
    simplS in * .
    split_l.
    assert (HS' : (p |> nextS pc (comml (_movwld a r))) (Mnh', R', fl, isr)).
      apply linkPG_intro with (Mnh,R,fl,isr); trivial.
    left.
    split_l.
    split.
      exists Mnh'; exists Mih; split_l.
      split_r.
      apply Hpp'; trivial.
    intros S'' Hg'.
    destruct S'' as [[[M'' R''] fl''] isr''].
    unfold NH_ItpG in Hg'.
    destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
    simplS in * .
    destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih Prec1) as [Heq1 Heq2].
    subst Mx My.
    clear HJMxy HMy.
    simplS.
    apply Hg.
    unfold IH_ItpG.
    exists Mnh; exists Mih.
    exists Mnh''; exists Mih''.
    split_l; split_l; split_l; split_l.
    simplS.
    apply Hg'g.
    apply conjG_intro with (Mnh', R', fl, isr); trivial.
    apply conjPG_intro; trivial.
  destruct Hntop as [Htag [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [_ [HR [Hfl Hisr]]].
  subst fl' isr'.
  unfold enable in Hen.
  destruct (Hen _ HMnh) as [Sx HnextSSx].
  destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
  clear Sx HnextSSx Hen.
  simplS in * .
  split_l.
  assert (HS' : (p |> nextS pc (comml (_movwld a r))) (Mnh', R', fl, isr)).
    apply linkPG_intro with (Mnh,R,fl,isr); trivial.
  right.
  split_l.
  split.
    exists Mnh'; exists Mih; split_l.
    split_r.
    apply Hpp'; trivial.
  destruct Hwfst as [n Hwfst].
  exists n.
  apply IH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, NH_ItpG.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold NH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  simplS in * .
  destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih Prec1) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  simplS.
  exists Mnh; exists Mih.
  exists Mnh''; exists Mih''.
  split_l; split_l; split_l; split_l.
  apply Hg'g.
  apply conjG_intro with (Mnh', R', fl, isr); trivial.
  apply conjPG_intro; trivial.
Qed.
Implicit Arguments IH_movwld_prev [status Psi a r pc S0 S' p p' g g' tag].

Lemma IH_movwst_prog : forall status Psi a r pc S0
  (p : Pred) (g : Grt) tag,
  enable p (nextS pc (comml (_movwst a r)))
  -> itp_IH_theta (Ast p g tag) status Psi S0
  -> exists S', nextS pc (comml (_movwst a r)) S0 S'.
Proof.
  intros status Psi a r pc [[[M R] fl] isr] p g tag.
  intros Hen H.
  unfold itp_IH_theta in * .
  destruct H as [Hiff [Htop | Hntop]].
    unfold lift_IH_PG in Htop.
    destruct Htop as [Htag [Hp Hg]].
    destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
    simplS in * .
    destruct (Hen _ HMnh) as [S' HnextSnh]. 
    destruct S' as [[[Mnh' R'] fl'] isr'].
    destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
    exists (M', R', fl', isr'); trivial.
  destruct Hntop as [Htag [Hp Hg]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  simplS in * .
  destruct (Hen _ HMnh) as [S' HnextSnh]. 
  destruct S' as [[[Mnh' R'] fl'] isr'].
  destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
  exists (M', R', fl', isr'); trivial.
Qed.
Implicit Arguments IH_movwst_prog [status Psi a r pc S0 p g tag].

Lemma IH_movwst_prev : forall status Psi a r pc S0 S'
  (p p' : Pred) (g g' : Grt) tag,
  lookupCT Psi (pc+1) (packS0 rho_ih (Ast p' g' tag))
  -> nextS pc (comml (_movwst a r)) S0 S'
  -> enable p (nextS pc (comml (_movwst a r)))
  -> pimp (p |> nextS pc (comml (_movwst a r))) p'
  -> gimp (p >o nextS pc (comml (_movwst a r)) oo g') g
  -> itp_IH_theta (Ast p g tag) status Psi S0
  -> itp_IH_theta (Ast p' g' tag) status Psi S'.
Proof.
  intros status Psi a r pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p' g g' tag.
  intros Hlk' HnextS Hen Hpp' Hg'g H.
  unfold itp_IH_theta in * .
  destruct H as [Hiff [Htop | Hntop]].
    unfold lift_IH_PG in Htop.
    destruct Htop as [Htag [Hp Hg]].
    destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
    destruct (nextS_inversion HnextS) as [_ [HR [Hfl Hisr]]].
    subst fl' isr'.
    unfold enable in Hen.
    destruct (Hen _ HMnh) as [Sx HnextSSx].
    destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
    clear Sx HnextSSx Hen.
    simplS in * .
    split_l.
    assert (HS' : (p |> nextS pc (comml (_movwst a r))) (Mnh', R', fl, isr)).
      apply linkPG_intro with (Mnh,R,fl,isr); trivial.
    left.
    split_l.
    split.
      exists Mnh'; exists Mih; split_l.
      split_r.
      apply Hpp'; trivial.
    intros S'' Hg'.
    destruct S'' as [[[M'' R''] fl''] isr''].
    unfold NH_ItpG in Hg'.
    destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
    simplS in * .
    destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih Prec1) as [Heq1 Heq2].
    subst Mx My.
    clear HJMxy HMy.
    simplS.
    apply Hg.
    unfold IH_ItpG.
    exists Mnh; exists Mih.
    exists Mnh''; exists Mih''.
    split_l; split_l; split_l; split_l.
    apply Hg'g.
    simplS.
    apply conjG_intro with (Mnh', R', fl, isr); trivial.
    apply conjPG_intro; trivial.
  destruct Hntop as [Htag [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [_ [HR [Hfl Hisr]]].
  subst fl' isr'.
  unfold enable in Hen.
  destruct (Hen _ HMnh) as [Sx HnextSSx].
  destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
  clear Sx HnextSSx Hen.
  simplS in * .
  split_l.
  assert (HS' : (p |> nextS pc (comml (_movwst a r))) (Mnh', R', fl, isr)).
    apply linkPG_intro with (Mnh,R,fl,isr); trivial.
  right.
  split_l.
  split.
    exists Mnh'; exists Mih; split_l.
    split_r.
    apply Hpp'; trivial.
  destruct Hwfst as [n Hwfst].
  exists n.
  apply IH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, NH_ItpG.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold NH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  simplS in * .
  destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih Prec1) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  simplS.
  exists Mnh; exists Mih.
  exists Mnh''; exists Mih''.
  split_l; split_l; split_l; split_l.
  apply Hg'g.
  apply conjG_intro with (Mnh', R', fl, isr); trivial.
  apply conjPG_intro; trivial.
Qed.
Implicit Arguments IH_movwst_prev [status Psi a r pc S0 S' p p' g g' tag].

Lemma IH_pushwp_prog : forall status Psi o pc S0
  (p : Pred) (g : Grt) tag,
  enable p (nextS pc (comml (_pushwp o)))
  -> itp_IH_theta (Ast p g tag) status Psi S0
  -> exists S', nextS pc (comml (_pushwp o)) S0 S'.
Proof.
  intros status Psi o pc [[[M R] fl] isr] p g tag.
  intros Hen H.
  unfold itp_IH_theta in * .
  destruct H as [Hiff [Htop | Hntop]].
    unfold lift_IH_PG in Htop.
    destruct Htop as [Htag [Hp Hg]].
    destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
    simplS in * .
    destruct (Hen _ HMnh) as [S' HnextSnh]. 
    destruct S' as [[[Mnh' R'] fl'] isr'].
    destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
    exists (M', R', fl', isr'); trivial.
  destruct Hntop as [Htag [Hp Hg]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  simplS in * .
  destruct (Hen _ HMnh) as [S' HnextSnh]. 
  destruct S' as [[[Mnh' R'] fl'] isr'].
  destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
  exists (M', R', fl', isr'); trivial.
Qed.
Implicit Arguments IH_pushwp_prog [status Psi o pc S0 p g tag].

Lemma IH_pushwp_prev : forall status Psi o pc S0 S'
  (p p' : Pred) (g g' : Grt) tag,
  lookupCT Psi (pc+1) (packS0 rho_ih (Ast p' g' tag))
  -> nextS pc (comml (_pushwp o)) S0 S'
  -> enable p (nextS pc (comml (_pushwp o)))
  -> pimp (p |> nextS pc (comml (_pushwp o))) p'
  -> gimp (p >o nextS pc (comml (_pushwp o)) oo g') g
  -> itp_IH_theta (Ast p g tag) status Psi S0
  -> itp_IH_theta (Ast p' g' tag) status Psi S'.
Proof.
  intros status Psi o pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p' g g' tag.
  intros Hlk' HnextS Hen Hpp' Hg'g H.
  unfold itp_IH_theta in * .
  destruct H as [Hiff [Htop | Hntop]].
    unfold lift_IH_PG in Htop.
    destruct Htop as [Htag [Hp Hg]].
    destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
    destruct (nextS_inversion HnextS) as [_ [HR [Hfl Hisr]]].
    subst fl' isr'.
    unfold enable in Hen.
    destruct (Hen _ HMnh) as [Sx HnextSSx].
    destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
    clear Sx HnextSSx Hen.
    simplS in * .
    split_l.
    assert (HS' : (p |> nextS pc (comml (_pushwp o))) (Mnh', R', fl, isr)).
      apply linkPG_intro with (Mnh,R,fl,isr); trivial.
    left.
    split_l.
    split.
      exists Mnh'; exists Mih; split_l.
      split_r.
      apply Hpp'; trivial.
    intros S'' Hg'.
    destruct S'' as [[[M'' R''] fl''] isr''].
    unfold NH_ItpG in Hg'.
    destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
    simplS in * .
    destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih Prec1) as [Heq1 Heq2].
    subst Mx My.
    clear HJMxy HMy.
    simplS.
    apply Hg.
    unfold IH_ItpG.
    exists Mnh; exists Mih.
    exists Mnh''; exists Mih''.
    split_l; split_l; split_l; split_l.
    apply Hg'g.
    simplS.
    apply conjG_intro with (Mnh', R', fl, isr); trivial.
    apply conjPG_intro; trivial.
  destruct Hntop as [Htag [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [_ [HR [Hfl Hisr]]].
  subst fl' isr'.
  unfold enable in Hen.
  destruct (Hen _ HMnh) as [Sx HnextSSx].
  destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
  clear Sx HnextSSx Hen.
  simplS in * .
  split_l.
  assert (HS' : (p |> nextS pc (comml (_pushwp o))) (Mnh', R', fl, isr)).
    apply linkPG_intro with (Mnh,R,fl,isr); trivial.
  right.
  split_l.
  split.
    exists Mnh'; exists Mih; split_l.
    split_r.
    apply Hpp'; trivial.
  destruct Hwfst as [n Hwfst].
  exists n.
  apply IH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, NH_ItpG.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold NH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  simplS in * .
  destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih Prec1) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  simplS.
  exists Mnh; exists Mih.
  exists Mnh''; exists Mih''.
  split_l; split_l; split_l; split_l.
  apply Hg'g.
  apply conjG_intro with (Mnh', R', fl, isr); trivial.
  apply conjPG_intro; trivial.
Qed.
Implicit Arguments IH_pushwp_prev [status Psi o pc S0 S' p p' g g' tag].

Lemma IH_pushwm_prog : forall status Psi a pc S0
  (p : Pred) (g : Grt) tag,
  enable p (nextS pc (comml (_pushwm a)))
  -> itp_IH_theta (Ast p g tag) status Psi S0
  -> exists S', nextS pc (comml (_pushwm a)) S0 S'.
Proof.
  intros status Psi a pc [[[M R] fl] isr] p g tag.
  intros Hen H.
  unfold itp_IH_theta in * .
  destruct H as [Hiff [Htop | Hntop]].
    unfold lift_IH_PG in Htop.
    destruct Htop as [Htag [Hp Hg]].
    destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
    simplS in * .
    destruct (Hen _ HMnh) as [S' HnextSnh]. 
    destruct S' as [[[Mnh' R'] fl'] isr'].
    destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
    exists (M', R', fl', isr'); trivial.
  destruct Hntop as [Htag [Hp Hg]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  simplS in * .
  destruct (Hen _ HMnh) as [S' HnextSnh]. 
  destruct S' as [[[Mnh' R'] fl'] isr'].
  destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
  exists (M', R', fl', isr'); trivial.
Qed.
Implicit Arguments IH_pushwm_prog [status Psi a pc S0 p g tag].

Lemma IH_pushwm_prev : forall status Psi a pc S0 S'
  (p p' : Pred) (g g' : Grt) tag,
  lookupCT Psi (pc+1) (packS0 rho_ih (Ast p' g' tag))
  -> nextS pc (comml (_pushwm a)) S0 S'
  -> enable p (nextS pc (comml (_pushwm a)))
  -> pimp (p |> nextS pc (comml (_pushwm a))) p'
  -> gimp (p >o nextS pc (comml (_pushwm a)) oo g') g
  -> itp_IH_theta (Ast p g tag) status Psi S0
  -> itp_IH_theta (Ast p' g' tag) status Psi S'.
Proof.
  intros status Psi a pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p' g g' tag.
  intros Hlk' HnextS Hen Hpp' Hg'g H.
  unfold itp_IH_theta in * .
  destruct H as [Hiff [Htop | Hntop]].
    unfold lift_IH_PG in Htop.
    destruct Htop as [Htag [Hp Hg]].
    destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
    destruct (nextS_inversion HnextS) as [_ [HR [Hfl Hisr]]].
    subst fl' isr'.
    unfold enable in Hen.
    destruct (Hen _ HMnh) as [Sx HnextSSx].
    destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
    clear Sx HnextSSx Hen.
    simplS in * .
    split_l.
    assert (HS' : (p |> nextS pc (comml (_pushwm a))) (Mnh', R', fl, isr)).
      apply linkPG_intro with (Mnh,R,fl,isr); trivial.
    left.
    split_l.
    split.
      exists Mnh'; exists Mih; split_l.
      split_r.
      apply Hpp'; trivial.
    intros S'' Hg'.
    destruct S'' as [[[M'' R''] fl''] isr''].
    unfold NH_ItpG in Hg'.
    destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
    simplS in * .
    destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih Prec1) as [Heq1 Heq2].
    subst Mx My.
    clear HJMxy HMy.
    simplS.
    apply Hg.
    unfold IH_ItpG.
    exists Mnh; exists Mih.
    exists Mnh''; exists Mih''.
    split_l; split_l; split_l; split_l.
    apply Hg'g.
    simplS.
    apply conjG_intro with (Mnh', R', fl, isr); trivial.
    apply conjPG_intro; trivial.
  destruct Hntop as [Htag [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [_ [HR [Hfl Hisr]]].
  subst fl' isr'.
  unfold enable in Hen.
  destruct (Hen _ HMnh) as [Sx HnextSSx].
  destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
  clear Sx HnextSSx Hen.
  simplS in * .
  split_l.
  assert (HS' : (p |> nextS pc (comml (_pushwm a))) (Mnh', R', fl, isr)).
    apply linkPG_intro with (Mnh,R,fl,isr); trivial.
  right.
  split_l.
  split.
    exists Mnh'; exists Mih; split_l.
    split_r.
    apply Hpp'; trivial.
  destruct Hwfst as [n Hwfst].
  exists n.
  apply IH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, IH_ItpG.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold IH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  simplS in * .
  destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih Prec1) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  simplS.
  exists Mnh; exists Mih.
  exists Mnh''; exists Mih''.
  split_l; split_l; split_l; split_l.
  apply Hg'g.
  apply conjG_intro with (Mnh', R', fl, isr); trivial.
  apply conjPG_intro; trivial.
Qed.
Implicit Arguments IH_pushwm_prev [status Psi a pc S0 S' p p' g g' tag].

Lemma IH_popw_prog : forall status Psi r pc S0
  (p : Pred) (g : Grt) tag,
  enable p (nextS pc (comml (_popw r)))
  -> itp_IH_theta (Ast p g tag) status Psi S0
  -> exists S', nextS pc (comml (_popw r)) S0 S'.
Proof.
  intros status Psi r pc [[[M R] fl] isr] p g tag.
  intros Hen H.
  unfold itp_IH_theta in * .
  destruct H as [Hiff [Htop | Hntop]].
    unfold lift_IH_PG in Htop.
    destruct Htop as [Htag [Hp Hg]].
    destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
    simplS in * .
    destruct (Hen _ HMnh) as [S' HnextSnh]. 
    destruct S' as [[[Mnh' R'] fl'] isr'].
    destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
    exists (M', R', fl', isr'); trivial.
  destruct Hntop as [Htag [Hp Hg]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  simplS in * .
  destruct (Hen _ HMnh) as [S' HnextSnh]. 
  destruct S' as [[[Mnh' R'] fl'] isr'].
  destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
  exists (M', R', fl', isr'); trivial.
Qed.
Implicit Arguments IH_popw_prog [status Psi r pc S0 p g tag].

Lemma IH_popw_prev : forall status Psi r pc S0 S'
  (p p' : Pred) (g g' : Grt) tag,
  lookupCT Psi (pc+1) (packS0 rho_ih (Ast p' g' tag))
  -> nextS pc (comml (_popw r)) S0 S'
  -> enable p (nextS pc (comml (_popw r)))
  -> pimp (p |> nextS pc (comml (_popw r))) p'
  -> gimp (p >o nextS pc (comml (_popw r)) oo g') g
  -> itp_IH_theta (Ast p g tag) status Psi S0
  -> itp_IH_theta (Ast p' g' tag) status Psi S'.
Proof.
  intros status Psi r pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p' g g' tag.
  intros Hlk' HnextS Hen Hpp' Hg'g H.
  unfold itp_IH_theta in * .
  destruct H as [Hiff [Htop | Hntop]].
    unfold lift_IH_PG in Htop.
    destruct Htop as [Htag [Hp Hg]].
    destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
    destruct (nextS_inversion HnextS) as [_ [HR [Hfl Hisr]]].
    subst fl' isr'.
    unfold enable in Hen.
    destruct (Hen _ HMnh) as [Sx HnextSSx].
    destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
    clear Sx HnextSSx Hen.
    simplS in * .
    split_l.
    assert (HS' : (p |> nextS pc (comml (_popw r))) (Mnh', R', fl, isr)).
      apply linkPG_intro with (Mnh,R,fl,isr); trivial.
    left.
    split_l.
    split.
      exists Mnh'; exists Mih; split_l.
      split_r.
      apply Hpp'; trivial.
    intros S'' Hg'.
    destruct S'' as [[[M'' R''] fl''] isr''].
    unfold IH_ItpG in Hg'.
    destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
    simplS in * .
    destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih Prec1) as [Heq1 Heq2].
    subst Mx My.
    clear HJMxy HMy.
    simplS.
    apply Hg.
    unfold IH_ItpG.
    exists Mnh; exists Mih.
    exists Mnh''; exists Mih''.
    split_l; split_l; split_l; split_l.
    apply Hg'g.
    apply conjG_intro with (Mnh', R', fl, isr); trivial.
    apply conjPG_intro; trivial.
  destruct Hntop as [Htag [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [_ [HR [Hfl Hisr]]].
  subst fl' isr'.
  unfold enable in Hen.
  destruct (Hen _ HMnh) as [Sx HnextSSx].
  destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
  clear Sx HnextSSx Hen.
  simplS in * .
  split_l.
  assert (HS' : (p |> nextS pc (comml (_popw r))) (Mnh', R', fl, isr)).
    apply linkPG_intro with (Mnh,R,fl,isr); trivial.
  right.
  split_l.
  split.
    exists Mnh'; exists Mih; split_l.
    split_r.
    apply Hpp'; trivial.
  destruct Hwfst as [n Hwfst].
  exists n.
  apply IH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, IH_ItpG.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold IH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  simplS in * .
  destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih Prec1) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  simplS.
  exists Mnh; exists Mih.
  exists Mnh''; exists Mih''.
  split_l; split_l; split_l; split_l.
  apply Hg'g.
  apply conjG_intro with (Mnh', R', fl, isr); trivial.
  apply conjPG_intro; trivial.
Qed.
Implicit Arguments IH_popw_prev [status Psi r pc S0 S' p p' g g' tag].

Lemma IH_cmpw_prog : forall status Psi o r pc S0
  (p : Pred) (g : Grt) tag,
  itp_IH_theta (Ast p g tag) status Psi S0
  -> exists S', nextS pc (comml (_cmpw o r)) S0 S'.
Proof.
  intros status Psi o r pc [[[M R] fl] isr] p g tag.
  intros H.
  destruct (eq_nat_dec (opr_eval R o) (R r)).
  exists (M, R, Set_Flag fl _ZF, isr).
  constructor.
  left; split_lr.
  exists (M, R, Clear_Flag fl _ZF, isr).
  constructor.
  right; split_lr.
Qed.
Implicit Arguments IH_cmpw_prog [status Psi o r pc S0 p g tag].

Lemma IH_cmpw_prev : forall status Psi o r pc S0 S'
  p p' g g' tag,
  lookupCT Psi (pc+1) (packS0 rho_ih (Ast p' g' tag))
  -> nextS pc (comml (_cmpw o r)) S0 S'
  -> enable p (nextS pc (comml (_cmpw o r)))
  -> pimp (p |> nextS pc (comml (_cmpw o r))) p'
  -> gimp (p >o nextS pc (comml (_cmpw o r)) oo g') g
  -> itp_IH_theta (Ast p g tag) status Psi S0
  -> itp_IH_theta (Ast p' g' tag) status Psi S'.
Proof.
  intros status Psi o r pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p' g g' tag.
  intros Hlk' HnextS Hen Hpp' Hg'g H.
  unfold itp_IH_theta in * .
  destruct H as [Hiff [Htop | Hntop]].
    unfold lift_IH_PG in Htop.
    destruct Htop as [Htag [Hp Hg]].
    destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
    destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
    subst M' R' isr'.
    unfold enable in Hen.
    destruct (Hen _ HMnh) as [Sx HnextSSx].
    destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
    clear Sx HnextSSx Hen.
    simplS in * .
    assert (HS' : (p |> nextS pc (comml (_cmpw o r))) (Mnh', R, fl', isr)).
      apply linkPG_intro with (Mnh,R,fl,isr); trivial.
    split.
      destruct Hfl as [[H1 H2] | [H1 H2]]; subst fl'; simpl; auto.
    left.
    split_l.
    split.
      exists Mnh'; exists Mih; split_l.
      split_r.
      apply Hpp'; trivial.
    intros S'' Hg'.
    destruct S'' as [[[M'' R''] fl''] isr''].
    unfold NH_ItpG in Hg'.
    destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
    simplS in * .
    destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih Prec1) as [Heq1 Heq2].
    subst Mx My.
    clear HJMxy HMy.
    simplS.
    apply Hg.
    unfold IH_ItpG.
    exists Mnh; exists Mih.
    exists Mnh''; exists Mih''.
    split_l; split_l; split_l; split_l.
    apply Hg'g.
    simplS.
    apply conjG_intro with (Mnh', R, fl', isr); trivial.
    apply conjPG_intro; trivial.
  destruct Hntop as [Htag [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  subst M' R' isr'.
  unfold enable in Hen.
  destruct (Hen _ HMnh) as [Sx HnextSSx].
  destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
  clear Sx HnextSSx Hen.
  simplS in * .
  split.
    destruct Hfl as [[H1 H2] | [H1 H2]]; subst fl'; simpl; auto.
  right.
  split_l.
  assert (HS' : (p |> nextS pc (comml (_cmpw o r))) (Mnh', R, fl', isr)).
    apply linkPG_intro with (Mnh,R,fl,isr); trivial.
  split.
    exists Mnh'; exists Mih; split_l.
    split_r.
    apply Hpp'; trivial.
  destruct Hwfst as [n Hwfst].
  exists n.
  apply IH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, IH_ItpG.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold IH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  simplS in * .
  destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih Prec1) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  simplS.
  exists Mnh; exists Mih.
  exists Mnh''; exists Mih''.
  split_l; split_l; split_l; split_l.
  apply Hg'g.
  apply conjG_intro with (Mnh', R, fl', isr); trivial.
  apply conjPG_intro; trivial.
Qed.
Implicit Arguments IH_cmpw_prev [status Psi o r pc S0 S' p p' g g' tag].

Lemma IH_addw_prog : forall status Psi o r pc S0
  (p : Pred) (g : Grt) tag,
  enable p (nextS pc (comml (_addw o r)))
  -> itp_IH_theta (Ast p g tag) status Psi S0
  -> exists S', nextS pc (comml (_addw o r)) S0 S'.
Proof.
  intros status Psi o r pc [[[M R] fl] isr] p g tag.
  intros Hen H.
  unfold itp_IH_theta in * .
  destruct H as [Hiff [Htop | Hntop]].
    unfold lift_IH_PG in Htop.
    destruct Htop as [Htag [Hp Hg]].
    destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
    simplS in * .
    destruct (Hen _ HMnh) as [S' HnextSnh]. 
    destruct S' as [[[Mnh' R'] fl'] isr'].
    destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
    exists (M', R', fl', isr'); trivial.
  destruct Hntop as [Htag [Hp Hg]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  simplS in * .
  destruct (Hen _ HMnh) as [S' HnextSnh]. 
  destruct S' as [[[Mnh' R'] fl'] isr'].
  destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
  exists (M', R', fl', isr'); trivial.
Qed.
Implicit Arguments IH_addw_prog [status Psi o r pc S0 p g tag].

Lemma IH_addw_prev : forall status Psi o r pc S0 S' p p' g g' tag,
  lookupCT Psi (pc+1) (packS0 rho_ih (Ast p' g' tag))
  -> nextS pc (comml (_addw o r)) S0 S'
  -> enable p (nextS pc (comml (_addw o r)))
  -> pimp (p |> nextS pc (comml (_addw o r))) p'
  -> gimp (p >o nextS pc (comml (_addw o r)) oo g') g
  -> itp_IH_theta (Ast p g tag) status Psi S0
  -> itp_IH_theta (Ast p' g' tag) status Psi S'.
Proof.
  intros status Psi o r pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p' g g' tag.
  intros Hlk' HnextS Hen Hpp' Hg'g H.
  unfold itp_IH_theta in * .
  destruct H as [Hiff [Htop | Hntop]].
    unfold lift_IH_PG in Htop.
    destruct Htop as [Htag [Hp Hg]].
    destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
    destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
    subst M' isr'.
    unfold enable in Hen.
    destruct (Hen _ HMnh) as [Sx HnextSSx].
    destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
    clear Sx HnextSSx Hen.
    simplS in * .
    split.
      unfold trash_fl in Hfl.
      destruct Hfl as [H1 | H1]; subst fl'; simpl; auto.
    left.
    assert (HS' : (p |> nextS pc (comml (_addw o r))) (Mnh', R', fl', isr)).
      apply linkPG_intro with (Mnh,R,fl,isr); trivial.
    split_l.
    split.
      exists Mnh'; exists Mih; split_l.
      split_r.
      apply Hpp'; trivial.
    intros S'' Hg'.
    destruct S'' as [[[M'' R''] fl''] isr''].
    unfold NH_ItpG in Hg'.
    destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
    simplS in * .
    destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih Prec1) as [Heq1 Heq2].
    subst Mx My.
    clear HJMxy HMy.
    simplS.
    apply Hg.
    unfold IH_ItpG.
    exists Mnh; exists Mih.
    exists Mnh''; exists Mih''.
    split_l; split_l; split_l; split_l.
    apply Hg'g.
    simplS.
    apply conjG_intro with (Mnh', R', fl', isr); trivial.
    apply conjPG_intro; trivial.
  destruct Hntop as [Htag [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  subst M' isr'.
  unfold enable in Hen.
  destruct (Hen _ HMnh) as [Sx HnextSSx].
  destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
  clear Sx HnextSSx Hen.
  simplS in * .
  split.
    unfold trash_fl in Hfl.
    destruct Hfl as [H1 | H1]; subst fl'; simpl; auto.
  right.
  split_l.
  assert (HS' : (p |> nextS pc (comml (_addw o r))) (Mnh', R', fl', isr)).
    apply linkPG_intro with (Mnh,R,fl,isr); trivial.
  split.
    exists Mnh'; exists Mih; split_l.
    split_r.
    apply Hpp'; trivial.
  destruct Hwfst as [n Hwfst].
  exists n.
  apply IH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, IH_ItpG.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold IH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  simplS in * .
  destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih Prec1) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  simplS.
  exists Mnh; exists Mih.
  exists Mnh''; exists Mih''.
  split_l; split_l; split_l; split_l.
  apply Hg'g.
  apply conjG_intro with (Mnh', R', fl', isr); trivial.
  apply conjPG_intro; trivial.
Qed.
Implicit Arguments IH_addw_prev [status Psi o r pc S0 S' p p' g g' tag].

Lemma IH_subw_prog : forall status Psi o r pc S0
  (p : Pred) (g : Grt) tag,
  enable p (nextS pc (comml (_subw o r)))
  -> itp_IH_theta (Ast p g tag) status Psi S0
  -> exists S', nextS pc (comml (_subw o r)) S0 S'.
Proof.
  intros status Psi o r pc [[[M R] fl] isr] p g tag.
  intros Hen H.
  unfold itp_IH_theta in * .
  destruct H as [Hiff [Htop | Hntop]].
    unfold lift_IH_PG in Htop.
    destruct Htop as [Htag [Hp Hg]].
    destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
    simplS in * .
    destruct (Hen _ HMnh) as [S' HnextSnh]. 
    destruct S' as [[[Mnh' R'] fl'] isr'].
    destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
    exists (M', R', fl', isr'); trivial.
  destruct Hntop as [Htag [Hp Hg]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  simplS in * .
  destruct (Hen _ HMnh) as [S' HnextSnh]. 
  destruct S' as [[[Mnh' R'] fl'] isr'].
  destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
  exists (M', R', fl', isr'); trivial.
Qed.
Implicit Arguments IH_subw_prog [status Psi o r pc S0 p g tag].

Lemma IH_subw_prev : forall status Psi o r pc S0 S' p p' g g' tag,
  lookupCT Psi (pc+1) (packS0 rho_ih (Ast p' g' tag))
  -> nextS pc (comml (_subw o r)) S0 S'
  -> enable p (nextS pc (comml (_subw o r)))
  -> pimp (p |> nextS pc (comml (_subw o r))) p'
  -> gimp (p >o nextS pc (comml (_subw o r)) oo g') g
  -> itp_IH_theta (Ast p g tag) status Psi S0
  -> itp_IH_theta (Ast p' g' tag) status Psi S'.
Proof.
  intros status Psi o r pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p' g g' tag.
  intros Hlk' HnextS Hen Hpp' Hg'g H.
  unfold itp_IH_theta in * .
  destruct H as [Hiff [Htop | Hntop]].
    unfold lift_IH_PG in Htop.
    destruct Htop as [Htag [Hp Hg]].
    destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
    destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
    subst M' isr'.
    unfold enable in Hen.
    destruct (Hen _ HMnh) as [Sx HnextSSx].
    destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
    clear Sx HnextSSx Hen.
    simplS in * .
    split.
      unfold trash_fl in Hfl.
      destruct Hfl as [H1 | H1]; subst fl'; simpl; auto.
    left.
    assert (HS' : (p |> nextS pc (comml (_subw o r))) (Mnh', R', fl', isr)).
      apply linkPG_intro with (Mnh,R,fl,isr); trivial.
    split_l.
    split.
      exists Mnh'; exists Mih; split_l.
      split_r.
      apply Hpp'; trivial.
    intros S'' Hg'.
    destruct S'' as [[[M'' R''] fl''] isr''].
    unfold NH_ItpG in Hg'.
    destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
    simplS in * .
    destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih Prec1) as [Heq1 Heq2].
    subst Mx My.
    clear HJMxy HMy.
    simplS.
    apply Hg.
    unfold IH_ItpG.
    exists Mnh; exists Mih.
    exists Mnh''; exists Mih''.
    split_l; split_l; split_l; split_l.
    apply Hg'g.
    simplS.
    apply conjG_intro with (Mnh', R', fl', isr); trivial.
    apply conjPG_intro; trivial.
  destruct Hntop as [Htag [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  subst M' isr'.
  unfold enable in Hen.
  destruct (Hen _ HMnh) as [Sx HnextSSx].
  destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
  clear Sx HnextSSx Hen.
  simplS in * .
  split.
    unfold trash_fl in Hfl.
    destruct Hfl as [H1 | H1]; subst fl'; simpl; auto.
  right.
  split_l.
  assert (HS' : (p |> nextS pc (comml (_subw o r))) (Mnh', R', fl', isr)).
    apply linkPG_intro with (Mnh,R,fl,isr); trivial.
  split.
    exists Mnh'; exists Mih; split_l.
    split_r.
    apply Hpp'; trivial.
  destruct Hwfst as [n Hwfst].
  exists n.
  apply IH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, IH_ItpG.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold IH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  simplS in * .
  destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih Prec1) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  simplS.
  exists Mnh; exists Mih.
  exists Mnh''; exists Mih''.
  split_l; split_l; split_l; split_l.
  apply Hg'g.
  apply conjG_intro with (Mnh', R', fl', isr); trivial.
  apply conjPG_intro; trivial.
Qed.
Implicit Arguments IH_subw_prev [status Psi o r pc S0 S' p p' g g' tag].

Lemma IH_eoi_prog : forall status Psi pc S0
  (p : Pred) (g : Grt) tag,
  enable p (nextS pc (comml (_outb_eoi)))
  -> itp_IH_theta (Ast p g tag) status Psi S0
  -> exists S', nextS pc (comml _outb_eoi) S0 S'.
Proof.
  intros status Psi pc [[[M R] fl] isr] p g tag.
  intros Hen H.
  unfold itp_IH_theta in * .
  destruct H as [Hiff [Htop | Hntop]].
    unfold lift_IH_PG in Htop.
    destruct Htop as [Htag [Hp Hg]].
    destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
    simplS in * .
    destruct (Hen _ HMnh) as [S' HnextSnh]. 
    destruct S' as [[[Mnh' R'] fl'] isr'].
    destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
    exists (M', R', fl', isr'); trivial.
  destruct Hntop as [Htag [Hp Hg]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  simplS in * .
  destruct (Hen _ HMnh) as [S' HnextSnh]. 
  destruct S' as [[[Mnh' R'] fl'] isr'].
  destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
  exists (M', R', fl', isr'); trivial.
Qed.
Implicit Arguments IH_eoi_prog [status Psi pc S0 p g tag].

Lemma IH_eoi_prev : forall status Psi pc S0 S' p p' g g' tag,
  lookupCT Psi (pc+1) (packS0 rho_ih (Ast p' g' tag))
  -> nextS pc (comml _outb_eoi) S0 S'
  -> enable p (nextS pc (comml _outb_eoi))
  -> pimp (p |> nextS pc (comml _outb_eoi)) p'
  -> gimp (p >o nextS pc (comml _outb_eoi) oo g') g
  -> itp_IH_theta (Ast p g tag) status Psi S0
  -> itp_IH_theta (Ast p' g' tag) status Psi S'.
Proof.
  intros status Psi pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p' g g' tag.
  intros Hlk' HnextS Hen Hpp' Hg'g H.
  unfold itp_IH_theta in * .
  destruct H as [Hiff [Htop | Hntop]].
    unfold lift_IH_PG in Htop.
    destruct Htop as [Htag [Hp Hg]].
    destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
    destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
    destruct HR as [HR HR'].
    subst M' R' isr' fl'.
    unfold enable in Hen.
    destruct (Hen _ HMnh) as [Sx HnextSSx].
    destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
    clear Sx HnextSSx Hen.
    simplS in * .
    split_l.
    left.
    assert (HS' : (p |> nextS pc (comml _outb_eoi)) (Mnh', R, fl, false)).
      apply linkPG_intro with (Mnh,R,fl,isr); trivial.
    split_l.
    split.
      exists Mnh'; exists Mih; split_l.
      split_r.
      apply Hpp'; trivial.
    intros S'' Hg'.
    destruct S'' as [[[M'' R''] fl''] isr''].
    unfold NH_ItpG in Hg'.
    destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
    simplS in * .
    destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih Prec1) as [Heq1 Heq2].
    subst Mx My.
    clear HJMxy HMy.
    simplS.
    apply Hg.
    unfold IH_ItpG.
    exists Mnh; exists Mih.
    exists Mnh''; exists Mih''.
    split_l; split_l; split_l; split_l.
    apply Hg'g.
    simplS.
    apply conjG_intro with (Mnh', R, fl, false); trivial.
    apply conjPG_intro; trivial.
  destruct Hntop as [Htag [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  destruct HR as [HR HR'].
  subst M' R' isr' fl'.
  unfold enable in Hen.
  destruct (Hen _ HMnh) as [Sx HnextSSx].
  destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
  clear Sx HnextSSx Hen.
  simplS in * .
  split_l.
  right.
  split_l.
  assert (HS' : (p |> nextS pc (comml _outb_eoi)) (Mnh', R, fl, false)).
    apply linkPG_intro with (Mnh,R,fl,isr); trivial.
  split.
    exists Mnh'; exists Mih; split_l.
    split_r.
    apply Hpp'; trivial.
  destruct Hwfst as [n Hwfst].
  exists n.
  apply IH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, IH_ItpG.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold IH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  simplS in * .
  destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih Prec1) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  simplS.
  exists Mnh; exists Mih.
  exists Mnh''; exists Mih''.
  split_l; split_l; split_l; split_l.
  apply Hg'g.
  apply conjG_intro with (Mnh', R, fl, false); trivial.
  apply conjPG_intro; trivial.
Qed.
Implicit Arguments IH_eoi_prev [status Psi pc S0 S' p p' g g' tag].

Lemma IH_jmp_prev : forall status Psi f pc S0 S' p p' g g' tag,
  lookupCT Psi f (packS0 rho_ih (Ast p' g' tag))
  -> nextS pc (comms (_jmp f)) S0 S'
  -> pimp p p'
  -> gimp (p >o g') g
  -> itp_IH_theta (Ast p g tag) status Psi S0
  -> itp_IH_theta (Ast p' g' tag) status Psi S'.
Proof.
  intros status Psi f pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p' g g' tag.
  intros Hlk' HnextS Hpp' Hg'g H.
  unfold itp_IH_theta in * .
  destruct H as [Hiff [Htop | Hntop]].
    unfold lift_IH_PG in Htop.
    destruct Htop as [Htag [Hp Hg]].
    destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
    destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
    subst M' R' fl' isr'.
    split_l.
    left.
    split_l.
    split.
      exists Mnh; exists Mih; split_l.
      split_r.
      apply Hpp'; trivial.
    intros S'' Hg'.
    destruct S'' as [[[M'' R''] fl''] isr''].
    unfold NH_ItpG in Hg'.
    destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
    simplS in * .
    destruct (prec_disjUnion_inv HJMxy HJM HMy HMih Prec1) as [Heq1 Heq2].
    subst Mx My.
    clear HJMxy HMy.
    simplS.
    apply Hg.
    unfold IH_ItpG.
    exists Mnh; exists Mih.
    exists Mnh''; exists Mih''.
    split_l; split_l; split_l; split_l.
    apply Hg'g.
    apply conjPG_intro; trivial.
  destruct Hntop as [Htag [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  subst M' R' fl' isr'.
  split_l.
  right.
  split_l.
  split.
    exists Mnh; exists Mih; split_l.
    split_r.
    apply Hpp'; trivial.
  destruct Hwfst as [n Hwfst].
  exists n.
  apply IH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, IH_ItpG.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold IH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  simplS in * .
  destruct (prec_disjUnion_inv HJMxy HJM HMy HMih Prec1) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  simplS.
  exists Mnh; exists Mih.
  exists Mnh''; exists Mih''.
  split_l; split_l; split_l; split_l.
  apply Hg'g.
  apply conjPG_intro; trivial.
Qed.
Implicit Arguments IH_jmp_prev [status Psi f pc S0 S' p p' g g' tag].

Lemma IH_retw_prog : forall status Psi pc S0
  (p : Pred) (g : Grt) tag,
  tag = 1
  -> (forall S, p S -> g S S)
  -> itp_IH_theta (Ast p g tag) status Psi S0
  -> exists S', nextS pc (comms (_retw)) S0 S'.
Proof.
  intros status Psi pc [[[M R] fl] isr] p g tag Htag Hpg H.
  unfold itp_IH_theta in * .
  destruct H as [Hiff [Htop | Hntop]].
    unfold lift_IH_PG in Htop.
    destruct Htop as [Htag' [Hp Hg]].
    subst tag; discriminate.
  destruct Hntop as [_ [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  simplS in * .
  assert (Hg : IH_ItpG g (M, R, fl, isr) (M, R, fl, isr)).
    simplS.
    exists Mnh; exists Mih.
    exists Mnh; exists Mih.
    split_l; split_l; split_l; split_l.
    apply Hpg; trivial.
  destruct Hwfst as [n Hwfst].
  destruct n as [ | n].
    simpl in Hwfst.
    substH Hwfst with (Hwfst (M, R, fl, isr) Hg).
    destruct Hwfst as [ra [Hra [theta' [Hlk' Htheta']]]].
    assert (Heq : leadsToW (R SP) ra M).
      unfold Ret_Addr in Hra.
      simpl in Hra.
      substH Hra with (Hra pc).
      inversion_clear Hra.
      subst sp.
      trivial.
    exists (M, RF_update R SP (R SP + 2), fl, isr).
    apply nextS_retw.
    unfold popW.
    exists ra; split_lr.
  simpl in Hwfst.
  substH Hwfst with (Hwfst _ Hg).
  destruct Hwfst as [ra [Hra [p' [g' [tag' [Hlk Hwfst]]]]]].
  assert (Heq : leadsToW (R SP) ra M).
    unfold Ret_Addr in Hra.
    simpl in Hra.
    substH Hra with (Hra pc).
    inversion_clear Hra.
    subst sp.
    trivial.
  exists (M, RF_update R SP (R SP + 2), fl, isr).
  apply nextS_retw.
  unfold popW.
  exists ra; split_lr.
Qed.
Implicit Arguments IH_retw_prog [status Psi pc S0 p g tag].

Lemma IH_retw_prev : forall status Psi pc M R fl isr pc' S'
  (p : Pred) (g : Grt) tag,
  tag = 1
  -> nextPC (comms _retw) M R fl pc pc'
  -> nextS pc (comms _retw) (M, R, fl, isr) S'
  -> (forall S, p S -> g S S)
  -> itp_IH_theta (Ast p g tag) status Psi (M,R,fl,isr)
  -> exists theta', 
    lookupCT Psi pc' (packS0 rho_ih theta')
    /\ itp_IH_theta theta' status Psi S'.
Proof.
  intros status Psi pc M R fl isr pc' [[[M' R'] fl'] isr'] p g tag. 
  intros Htag HnextPC HnextS Hpg H.
  destruct H as [Hiff [Htop | Hntop]].
    unfold lift_IH_PG in Htop.
    destruct Htop as [Htag' [Hp Hg]].
    subst tag; discriminate.
  destruct Hntop as [_ [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  simplS in * .
  destruct (nextS_inversion HnextS) as [HM' [HR' [Hfl' Hisr']]].
  subst M' fl' isr'.
  inversion_clear HnextPC.
  subst sp.
  unfold IH_WFST in Hwfst.
  assert (Hg : IH_ItpG g (M, R, fl, isr) (M, R, fl, isr)).
    simplS.
    exists Mnh; exists Mih.
    exists Mnh; exists Mih.
    split_l; split_l; split_l; split_l.
    apply Hpg; trivial.
  destruct Hwfst as [n Hwfst].
  destruct n as [ | n].
    simpl in Hwfst.
    substH Hwfst with (Hwfst (M, R, fl, isr) Hg).
    destruct Hwfst as [ra [Hra [theta' [Hlk' Htheta']]]].
    assert (Heq : pc' = ra).
      unfold Ret_Addr in Hra.
      simpl in Hra.
      substH Hra with (Hra pc).
      inversion_clear Hra.
      subst sp.
      eapply leadsToW_unique; eauto.
    subst pc'.
    exists theta'; split_l.
    split_l.
    simpl.
    destruct theta' as [p' g' tag'].
    left.
    apply Htheta' with pc; trivial.
  fold IH_WFST in Hwfst.
  substH Hwfst with (Hwfst _ Hg).
  destruct Hwfst as [ra [Hra [p' [g' [tag' [Hlk [Htag' Hwfst]]]]]]].
  assert (Heq : pc' = ra).
    unfold Ret_Addr in Hra.
    simpl in Hra.
    substH Hra with (Hra pc).
    inversion_clear Hra.
    subst sp.
    eapply leadsToW_unique; eauto.
  subst pc'.
  exists (Ast p' g' tag'); split_l.
  split_l.
  right.
  split_l.
  substH Hwfst with (Hwfst _ _ HnextS).
  destruct Hwfst as [Hp' Hwfst].
  destruct Hp' as [Mnh' [Mih' [HJM' Hp']]].
  simplS in * .
  split.
    exists Mnh'; exists Mih'; split_lr.
  exists n; trivial.
Qed.
Implicit Arguments IH_retw_prev [status Psi pc M R fl isr pc' S' p g tag].

