Add LoadPath "..".
Add LoadPath "../CAP0".

Require Import Arith.
Require Import Coqlib.
Require Import Map.
Require Import Machine.
Require Import Assertion.
Require Import Assertion_Lemmas.
Require Import Operational_Semantics.
Require Import OpSem_Lemmas.

Require Import CAP0.
Require Import SCAP.
Require Import AIM.
Require Import QLib.
Require Import SCAP_READY.
Require Import SCAP_Itr.

Require Import EqNat.

Require Import Linking.
Require Import Linking_Lemmas.
Require Import Linking_Lemmas2.
Require Import IH_SW.
Require Import SCAP_Itr_Lemmas.

Lemma IH_lookupCT_SCAP_CAP0 : forall Psi Psi' psi f p g tag,
  SCAP_lookupCT psi f (Ast p g tag)
  -> lift_psi psi rho_ih Psi
  -> subsetCT (ext_Prim_Specs Psi) Psi'
  -> lookupCT Psi' f (packS0 rho_ih (Ast p g tag)).
Proof.
  intros Psi Psi' psi f p g tag Hpsi Hlift Hsub.
  apply Hsub.
  unfold ext_IH_Spec, ext_Prim_Specs.
  apply lookupCT_subsetCT with Psi.
  apply Hlift; trivial.
  apply Psi_extend_subsetCT.
  apply Psi_extend_subsetCT.
  apply Psi_extend_subsetCT.
  apply subsetCT_refl.
Qed.
Hint Local Resolve IH_lookupCT_SCAP_CAP0.

Lemma IH_WFworld_IF_disable : forall Psi S T B RdyQ tid theta,
  WF_AIM_World Psi S T B RdyQ tid rho_ih theta
  -> S_Flag S _IF = false.
Proof.
  intros.
  unfold WF_AIM_World in H.
  destruct H as [H0 [H1 [H2 [H3 [TM [Mcth [Hsplit [Hmto [Hcth Hq]]]]]]]]].
  unfold WF_Curr_Thrd in Hcth.
  simpl in Hcth.
  unfold itp_IH_theta in Hcth.
  destruct S as [[[M R]fl]isr].
  simplS in Hcth. 
  destruct Hcth as [tcb [status Hcth]].
  InvertAll.
  simplS; trivial.
Qed.
Implicit Arguments IH_WFworld_IF_disable [Psi S T B RdyQ tid theta].

Lemma SCAP_IH_in_CAP0_call :
  forall D psi pc theta f D' Psi Psi',
    IH_WFcmd psi pc theta (comml (_call f))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_ih Psi
    -> Psi' = ext_Prim_Specs Psi
    -> WFcmd D' Psi' (packS0 rho_ih theta) pc (comml (_call f)).
Proof.
  intros D psi pc [p g tag] f D' Psi Psi'.
  intros Hi HD HD' Hlift HPsi'.
  destruct ((fun x => Hi) 0) as [[Hf Hi'] | [Hf Hi']].
  (* prim_entry f *)
  unfold IH_Call_SW in Hi'.
  destruct Hi' as [Hfeq Hi']; subst f.
  (* f = _SWITCH *)
  eapply IH_SW_Itp; eauto.
  (* ~ prim_entry f *)
  clear Hi; rename Hi' into Hi.
  unfold NH_Normal_Call in Hi.
  subst D' Psi'.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_ih; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  split.
    left.
    intros [Hx Hy].
    assert (H := IH_WFworld_IF_disable Hwf).
    simplS in * .
    rewrite Hx in H; discriminate.
  simpl in Hi.
  destruct Hi as [p' [p'' [g' [g'' [Hslk' [Hslk'' Hi]]]]]].
  destruct Hi as [Hcall [Hfun [Hpp' [Hpg'p'' Hg''g]]]].
  unfold WF_AIM_World in Hwf.
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct Hwfcth as [tcb [status [Hmto [Hst Hwfcth]]]].
  destruct (IH_call_prog Hcall Hwfcth) as [Scth' HnextScth].
  split.
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (nextS_Mono2 H5 HJM H6 HnextScth) as [M' HnextS].
    exists (M', R', fl', isr'); trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  subst R' fl' isr'.
  assert (HnextPC' := nextPC_call pc f M R fl).
  assert (Heq := nextPC_injection HnextPC HnextPC').
  subst pc'.
  clear HnextPC'.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  clear Scth' HnextScth.
  exists (packS0 rho_ih (Ast p' g' 1)).
  split_l by eauto.
  unfold ext_dict.
  simpl.
  unfold Itp_AIM.
  rewrite minimalD_elim_ih; trivial.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  simplS.
  destruct HM.
  simplS in * .
  exists Mth'; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  simplS.
  split.
    clear Hf B RdyQ H1 H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd.
    simpl.
    exists tcb; exists status; split_l; split_l.
    eapply IH_nonprim_call_prev; eauto.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IH_in_CAP0_jne :
  forall D psi pc theta f D' Psi Psi',
    IH_WFcmd psi pc theta (comml (_jne f))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_ih Psi
    -> Psi' = ext_Prim_Specs Psi
    -> WFcmd D' Psi' (packS0 rho_ih theta) pc (comml (_jne f)).
Proof.
  intros D psi pc [p g tag] f D' Psi Psi'.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_ih; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  split.
    left.
    intros [Hx Hy].
    assert (H := IH_WFworld_IF_disable Hwf).
    simplS in * .
    rewrite Hx in H; discriminate.
  simpl in Hi.
  destruct Hi as [p' [p'' [g' [g'' [Hlookup' [Hlookup'' Hi]]]]]].
  destruct Hi as [Hpp' [Hg'g [Hpp'' Hg''g]]].
  unfold WF_AIM_World in Hwf.  
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct Hwfcth as [tcb [status [Hmto [Hst Hwfcth]]]].
  destruct (IH_jne_prog (f:=f) (pc:=pc) Hwfcth) as [Scth' HnextScth].
  split.
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (nextS_Mono2 H5 HJM H6 HnextScth) as [M' HnextS].
    exists (M', R', fl', isr'); trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr']]].
  subst M' R' fl' isr'.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  inversion HnextPC.
  (* non-jmp *)
  subst f0 M0 R0 fl0 pc0 pc'.
  simplS in H0.
  rename H0 into Hflzf.
  exists (packS0 rho_ih (Ast p'' g'' tag)).
  split_l by eauto.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_ih; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  exists Mth'; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  split.
    clear B RdyQ H1 H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd.
    simpl; simplS in * .
    exists tcb; exists status; split_l; split_l.
    eapply IH_jne_seq_prev; eauto.
    simplS; trivial.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
  (* jmp *)
  subst f0 M0 R0 fl0 pc0 pc'.
  simplS in H0.
  rename H0 into Hflzf.
  exists (packS0 rho_ih (Ast p' g' tag)).
  split_l by eauto.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_ih; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  exists Mth'; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  split.
    clear B RdyQ H1 H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd. 
    exists tcb; exists status; split_l; split_l.
    simpl; simplS.
    eapply IH_jne_jmp_prev; eauto.
    simplS; trivial.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IH_in_CAP0_movwp :
  forall D psi pc theta D' Psi Psi' o r,
    IH_WFcmd psi pc theta (comml (_movwp o r))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_ih Psi
    -> Psi' = ext_Prim_Specs Psi
    -> WFcmd D' Psi' (packS0 rho_ih theta) pc (comml (_movwp o r)).
Proof.
  intros D psi pc [p g tag] D' Psi Psi' o r.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_ih; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  split.
    left.
    intros [Hx Hy].
    assert (H := IH_WFworld_IF_disable Hwf).
    simplS in * .
    rewrite Hx in H; discriminate.
  destruct Hi as [p' [g' [Hlookup' [Hmovwp Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  unfold WF_AIM_World in Hwf.
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct Hwfcth as [tcb [status [Hmto [Hst Hwfcth]]]].
  destruct (IH_movwp_prog Hmovwp Hwfcth) as [Scth' HnextScth].
  split.
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (nextS_Mono2 H5 HJM H6 HnextScth) as [M' HnextS].
    exists (M', R', fl', isr'); trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  rewrite (nextPC_inversion HnextPC).
  subst R' fl' isr'.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  clear Scth' HnextScth.
  exists (packS0 rho_ih (Ast p' g' tag)).
  split_l by eauto.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_ih; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  simplS.
  destruct HM.
  simplS in * .
  exists Mth'; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  simplS.
  split.
    clear B RdyQ H1 H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd.
    exists tcb; exists status; split_l; split_l.
    simpl.
    eapply IH_movwp_prev; eauto.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IH_in_CAP0_movwld :
  forall D psi pc theta D' Psi Psi' a r,
    IH_WFcmd psi pc theta (comml (_movwld a r))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_ih Psi
    -> Psi' = ext_Prim_Specs Psi
    -> WFcmd D' Psi' (packS0 rho_ih theta) pc (comml (_movwld a r)).
Proof.
  intros D psi pc [p g tag] D' Psi Psi' a r.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_ih; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  split.
    left.
    intros [Hx Hy].
    assert (H := IH_WFworld_IF_disable Hwf).
    simplS in * .
    rewrite Hx in H; discriminate.
  simpl in Hi.
  destruct Hi as [p' [g' [Hlookup' [Hmovwp Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  unfold WF_AIM_World in Hwf.
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct Hwfcth as [tcb [status [Hmto [Hst Hwfcth]]]].
  destruct (IH_movwld_prog Hmovwp Hwfcth) as [Scth' HnextScth].
  split.
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (nextS_Mono2 H5 HJM H6 HnextScth) as [M' HnextS].
    exists (M', R', fl', isr'); trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  rewrite (nextPC_inversion HnextPC).
  destruct HR as [x [Hlds HR']].
  subst R' fl' isr'.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  clear Scth' HnextScth.
  exists (packS0 rho_ih (Ast p' g' tag)).
  split_l by eauto.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_ih; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  simplS.
  destruct HM.
  simplS in * .
  exists Mth'; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  simplS.
  split.
    clear B RdyQ H1 H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd.
    simpl.
    exists tcb; exists status; split_l; split_l.
    eapply IH_movwld_prev; eauto.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed. 

Lemma SCAP_IH_in_CAP0_movwst :
  forall D psi pc theta D' Psi Psi' o a,
    IH_WFcmd psi pc theta (comml (_movwst o a))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_ih Psi
    -> Psi' = ext_Prim_Specs Psi
    -> WFcmd D' Psi' (packS0 rho_ih theta) pc (comml (_movwst o a)).
Proof.
  intros D psi pc [p g tag] D' Psi Psi' a r.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_ih; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  split.
    left.
    intros [Hx Hy].
    assert (H := IH_WFworld_IF_disable Hwf).
    simplS in * .
    rewrite Hx in H; discriminate.
  simpl in Hi.
  destruct Hi as [p' [g' [Hlookup' [Hmovwst Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  unfold WF_AIM_World in Hwf.
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct Hwfcth as [tcb [status [Hmto [Hst Hwfcth]]]].
  destruct (IH_movwst_prog Hmovwst Hwfcth) as [Scth' HnextScth].
  split.
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (nextS_Mono2 H5 HJM H6 HnextScth) as [M' HnextS].
    exists (M', R', fl', isr'); trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  rewrite (nextPC_inversion HnextPC).
  subst R' fl' isr'.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  clear Scth' HnextScth.
  exists (packS0 rho_ih (Ast p' g' tag)).
  split_l by eauto.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_ih; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  simplS.
  destruct HM.
  simplS in * .
  exists Mth'; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  simplS.
  split.
    clear B RdyQ H1 H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd.
    simpl.
    exists tcb; exists status; split_l; split_l.
    eapply IH_movwst_prev; eauto.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IH_in_CAP0_pushwp :
  forall D psi pc theta D' Psi Psi' o,
    IH_WFcmd psi pc theta (comml (_pushwp o))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_ih Psi
    -> Psi' = ext_Prim_Specs Psi
    -> WFcmd D' Psi' (packS0 rho_ih theta) pc (comml (_pushwp o)).
Proof.
  intros D psi pc [p g tag] D' Psi Psi' o.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_ih; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  split.
    left.
    intros [Hx Hy].
    assert (H := IH_WFworld_IF_disable Hwf).
    simplS in * .
    rewrite Hx in H; discriminate.
  simpl in Hi.
  destruct Hi as [p' [g' [Hlookup' [Hen Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  unfold WF_AIM_World in Hwf.
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct Hwfcth as [tcb [status [Hmto [Hst Hwfcth]]]].
  destruct (IH_pushwp_prog Hen Hwfcth) as [Scth' HnextScth].
  split.
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (nextS_Mono2 H5 HJM H6 HnextScth) as [M' HnextS].
    exists (M', R', fl', isr'); trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  rewrite (nextPC_inversion HnextPC).
  subst R' fl' isr'.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  clear Scth' HnextScth.
  exists (packS0 rho_ih (Ast p' g' tag)).
  split_l by eauto.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_ih; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  simplS.
  destruct HM.
  simplS in * .
  exists Mth'; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  simplS.
  split.
    clear B RdyQ H1 H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd.
    exists tcb; exists status; split_l; split_l.
    simpl.
    eapply IH_pushwp_prev; eauto.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IH_in_CAP0_pushwm :
  forall D psi pc theta D' Psi Psi' a,
    IH_WFcmd psi pc theta (comml (_pushwm a))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_ih Psi
    -> Psi' = ext_Prim_Specs Psi
    -> WFcmd D' Psi' (packS0 rho_ih theta) pc (comml (_pushwm a)).
Proof.
  intros D psi pc [p g tag] D' Psi Psi' a.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_ih; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  split.
    left.
    intros [Hx Hy].
    assert (H := IH_WFworld_IF_disable Hwf).
    simplS in * .
    rewrite Hx in H; discriminate.
  simpl in Hi.
  destruct Hi as [p' [g' [Hlookup' [Hen Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  unfold WF_AIM_World in Hwf.
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct Hwfcth as [tcb [status [Hmto [Hst Hwfcth]]]].
  destruct (IH_pushwm_prog Hen Hwfcth) as [Scth' HnextScth].
  split.
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (nextS_Mono2 H5 HJM H6 HnextScth) as [M' HnextS].
    exists (M', R', fl', isr'); trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  rewrite (nextPC_inversion HnextPC).
  subst R' fl' isr'.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  clear Scth' HnextScth.
  exists (packS0 rho_ih (Ast p' g' tag)).
  split_l by eauto.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_ih; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  simplS.
  destruct HM.
  simplS in * .
  exists Mth'; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  simplS.
  split.
    clear B RdyQ H1 H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd.
    exists tcb; exists status; split_l; split_l.
    simpl.
    eapply IH_pushwm_prev; eauto.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IH_in_CAP0_popw :
  forall D psi pc theta D' Psi Psi' r,
    IH_WFcmd psi pc theta (comml (_popw r))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_ih Psi
    -> Psi' = ext_Prim_Specs Psi
    -> WFcmd D' Psi' (packS0 rho_ih theta) pc (comml (_popw r)).
Proof.
  intros D psi pc [p g tag] D' Psi Psi' r.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_ih; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  split.
    left.
    intros [Hx Hy].
    assert (H := IH_WFworld_IF_disable Hwf).
    simplS in * .
    rewrite Hx in H; discriminate.
  simpl in Hi.
  destruct Hi as [p' [g' [Hlookup' [Hen Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  unfold WF_AIM_World in Hwf.
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct Hwfcth as [tcb [status [Hmto [Hst Hwfcth]]]].
  destruct (IH_popw_prog Hen Hwfcth) as [Scth' HnextScth].
  split.
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (nextS_Mono2 H5 HJM H6 HnextScth) as [M' HnextS].
    exists (M', R', fl', isr'); trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  rewrite (nextPC_inversion HnextPC).
  destruct HR as [x [Hlds HR']].
  subst R' fl' isr'.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  clear Scth' HnextScth.
  exists (packS0 rho_ih (Ast p' g' tag)).
  split_l by eauto.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_ih; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  simplS.
  destruct HM.
  simplS in * .
  exists Mth'; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  simplS.
  split.
    clear B RdyQ H1 H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd.
    exists tcb; exists status; split_l; split_l.
    simpl.
    eapply IH_popw_prev; eauto.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IH_in_CAP0_cmpw :
  forall D psi pc theta D' Psi Psi' o r,
    IH_WFcmd psi pc theta (comml (_cmpw o r))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_ih Psi
    -> Psi' = ext_Prim_Specs Psi
    -> WFcmd D' Psi' (packS0 rho_ih theta) pc (comml (_cmpw o r)).
Proof.
  intros D psi pc [p g tag] D' Psi Psi' o r.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_ih; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  split.
    left.
    intros [Hx Hy].
    assert (H := IH_WFworld_IF_disable Hwf).
    simplS in * .
    rewrite Hx in H; discriminate.
  simpl in Hi.
  destruct Hi as [p' [g' [Hlookup' [Hen Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  unfold WF_AIM_World in Hwf.
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct Hwfcth as [tcb [status [Hmto [Hst Hwfcth]]]].
  destruct (IH_cmpw_prog (o:=o) (r:=r) (pc:=pc) Hwfcth) as [Scth' HnextScth].
  split.
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (nextS_Mono2 H5 HJM H6 HnextScth) as [M' HnextS].
    exists (M', R', fl', isr'); trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  rewrite (nextPC_inversion HnextPC).
  subst R' isr'.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  clear Scth' HnextScth.
  exists (packS0 rho_ih (Ast p' g' tag)).
  split_l by eauto.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_ih; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  simplS.
  destruct HM.
  simplS in * .
  exists Mth'; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  simplS.
  split.
    clear B RdyQ H1 H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd.
    exists tcb; exists status; split_l; split_l.
    simpl.
    eapply IH_cmpw_prev; eauto.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IH_in_CAP0_addw :
  forall D psi pc theta D' Psi Psi' o r,
    IH_WFcmd psi pc theta (comml (_addw o r))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_ih Psi
    -> Psi' = ext_Prim_Specs Psi
    -> WFcmd D' Psi' (packS0 rho_ih theta) pc (comml (_addw o r)).
Proof.
  intros D psi pc [p g tag] D' Psi Psi' o r.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_ih; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  split.
    left.
    intros [Hx Hy].
    assert (H := IH_WFworld_IF_disable Hwf).
    simplS in * .
    rewrite Hx in H; discriminate.
  simpl in Hi.
  destruct Hi as [p' [g' [Hlookup' [Hen Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  unfold WF_AIM_World in Hwf.
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct Hwfcth as [tcb [status [Hmto [Hst Hwfcth]]]].
  destruct (IH_addw_prog Hen Hwfcth) as [Scth' HnextScth].
  split.
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (nextS_Mono2 H5 HJM H6 HnextScth) as [M' HnextS].
    exists (M', R', fl', isr'); trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  rewrite (nextPC_inversion HnextPC).
  subst R' isr'.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  clear Scth' HnextScth.
  exists (packS0 rho_ih (Ast p' g' tag)).
  split_l by eauto.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_ih; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  simplS.
  destruct HM.
  simplS in * .
  exists Mth'; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  simplS.
  split.
    clear B RdyQ H1 H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd.
    simpl.
    exists tcb; exists status; split_l; split_l.
    eapply IH_addw_prev; eauto.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IH_in_CAP0_subw :
  forall D psi pc theta D' Psi Psi' o r,
    IH_WFcmd psi pc theta (comml (_subw o r))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim) 
    -> lift_psi psi rho_ih Psi
    -> Psi' = ext_Prim_Specs Psi
    -> WFcmd D' Psi' (packS0 rho_ih theta) pc (comml (_subw o r)).
Proof.
  intros D psi pc [p g tag] D' Psi Psi' o r.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_ih; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  split.
    left.
    intros [Hx Hy].
    assert (H := IH_WFworld_IF_disable Hwf).
    simplS in * .
    rewrite Hx in H; discriminate.
  simpl in Hi.
  destruct Hi as [p' [g' [Hlookup' [Hen Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  unfold WF_AIM_World in Hwf.
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct Hwfcth as [tcb [status [Hmto [Hst Hwfcth]]]].
  destruct (IH_subw_prog Hen Hwfcth) as [Scth' HnextScth].
  split.
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (nextS_Mono2 H5 HJM H6 HnextScth) as [M' HnextS].
    exists (M', R', fl', isr'); trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  rewrite (nextPC_inversion HnextPC).
  subst R' isr'.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  clear Scth' HnextScth.
  exists (packS0 rho_ih (Ast p' g' tag)).
  split_l by eauto.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_ih; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  simplS.
  destruct HM.
  simplS in * .
  exists Mth'; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  simplS.
  split.
    clear B RdyQ H1 H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd.
    simpl.
    exists tcb; exists status; split_l; split_l.
    eapply IH_subw_prev; eauto.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IH_in_CAP0_eoi :
  forall D psi pc theta D' Psi Psi',
    IH_WFcmd psi pc theta (comml (_outb_eoi))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_ih Psi
    -> Psi' = ext_Prim_Specs Psi
    -> WFcmd D' Psi' (packS0 rho_ih theta) pc (comml (_outb_eoi)).
Proof.
  intros D psi pc [p g tag] D' Psi Psi'.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_ih; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  split.
    left.
    intros [Hx Hy].
    assert (H := IH_WFworld_IF_disable Hwf).
    simplS in * .
    rewrite Hx in H; discriminate.
  simpl in Hi.
  destruct Hi as [p' [g' [Hlookup' [Hen Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  unfold WF_AIM_World in Hwf.
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct Hwfcth as [tcb [status [Hmto [Hst Hwfcth]]]].
  destruct (IH_eoi_prog Hen Hwfcth) as [Scth' HnextScth].
  split.
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (nextS_Mono2 H5 HJM H6 HnextScth) as [M' HnextS].
    exists (M', R', fl', isr'); trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  rewrite (nextPC_inversion HnextPC).
  destruct HR as [HR HR'].
  subst M' R' fl' isr'.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  clear Scth' HnextScth.
  exists (packS0 rho_ih (Ast p' g' tag)).
  split_l by eauto.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_ih; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  simplS.
  simplS in * .
  exists Mth'; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  simplS.
  split.
    clear B RdyQ H1 H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd.
    simpl.
    exists tcb; exists status; split_l; split_l.
    eapply IH_eoi_prev; eauto.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IH_in_CAP0_jmp :
  forall D psi pc theta D' Psi Psi' f,
    IH_WFcmd psi pc theta (comms (_jmp f))
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_ih Psi
    -> Psi' = ext_Prim_Specs Psi
    -> WFcmd D' Psi' (packS0 rho_ih theta) pc (comms (_jmp f)).
Proof.
  intros D psi pc [p g tag] D' Psi Psi' f.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_ih; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  split.
    left.
    intros [Hx Hy].
    assert (H := IH_WFworld_IF_disable Hwf).
    simplS in * .
    rewrite Hx in H; discriminate.
  destruct Hi as [p' [g' [Hlk' [Hpp' Hg'g]]]].
  unfold WF_AIM_World in Hwf.  
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct Hwfcth as [tcb [status [Hmto [Hst Hwfcth]]]].
  split.
    exists (M, R, fl, isr); constructor; trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  rewrite (nextPC_inversion HnextPC).
  subst R' fl' isr'.
  assert (HnextScth : nextS pc (comms (_jmp f)) (Mcth, R, fl, isr) (Mcth, R, fl, isr)).
    constructor; trivial.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  clear HnextScth.
  exists (packS0 rho_ih (Ast p' g' tag)).
  split_l by eauto.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_ih; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  unfold pmstar.
  unfold WF_AIM_World.
  simplS.
  destruct HM.
  simplS in * .
  exists Mth'; exists Mq; split_l; split_r.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  simplS.
  split.
    clear B RdyQ H1 H2 H3 H4 H5 H6 H8 H9 HtoMcth' HTM' HMq.
    clear Mq Mth TM TM' HnextPC HnextS HJM HJM' HsplitTM'.
    unfold WF_Curr_Thrd.
    simpl.
    exists tcb; exists status; split_l; split_l.
    eapply IH_jmp_prev; eauto.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IH_in_CAP0_retw :
  forall D psi pc theta D' Psi Psi',
    IH_WFcmd psi pc theta (comms _retw)
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_ih Psi
    -> Psi' = ext_Prim_Specs Psi
    -> WFcmd D' Psi' (packS0 rho_ih theta) pc (comms _retw).
Proof.
  intros D psi pc [p g tag] D' Psi Psi'.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_ih; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  split.
    left.
    intros [Hx Hy].
    assert (H := IH_WFworld_IF_disable Hwf).
    simplS in * .
    rewrite Hx in H; discriminate.
  simpl in Hi.
  destruct Hi as [Htag Hp].
  unfold WF_AIM_World in Hwf.
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct Hwfcth as [tcb [status [Hmto [Hst Hwfcth]]]].
  destruct (IH_retw_prog (pc:=pc) Htag Hp Hwfcth) as [Scth' HnextScth].
  split.
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (nextS_Mono2 H5 HJM H6 HnextScth) as [M' HnextS].
    exists (M', R', fl', isr'); trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R']fl']isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr']]].
  destruct HR as [pc'x [Hlds HR]].
  subst M' R' fl' isr'.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  clear Scth' HnextScth.
  assert (pc' = pc'x).
    eapply leadsToW_unique; eauto.
    simplS in * ; subst; trivial.
    inversion_clear HnextPC.
    subst; trivial.
  subst pc'x.
  assert (HnextPC' : nextPC (comms _retw) Mcth R fl pc pc').
    inversion_clear HnextPC.
    destruct (nextS_inversion HnextS') as [_ [HR [_ _]]].
    destruct HR as [pc'x [Hlds' HR']].
    assert (Heq : pc' = pc'x).
      eapply leadsToW_unique; eauto.
      destruct (MemSplit_Inv H5 H6) as [Moth [Hx _]].
      simplS in * .
      apply (leadsToW_Mono (addr := sp)(x := pc'x) HJM); eauto.
      eapply leadsToW_Mono; eauto.
      subst; trivial.
    subst pc'x.
    apply nextPC_retw with (R SP); trivial.
  destruct (IH_retw_prev Htag HnextPC' HnextS' Hp Hwfcth) as [theta' [Hlk' Hwf]].
  exists (packS0 rho_ih theta'); split_l.
  unfold ext_dict.
  simpl.
  rewrite minimalD_elim_ih; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  exists Mth'; exists Mq; split_l.
  split_r.
  unfold WF_AIM_World.
  split_l; split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  split.
    clear B RdyQ HtoMcth' HTM' H1 H2 H3 H4 H5 H6 H8 H9 HMq.
    clear Mq Mth TM HnextS HJM HJM'.
    unfold WF_Curr_Thrd.
    exists tcb; exists status; split_l; split_l.
    simpl; simplS.
    trivial.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IH_in_CAP0_iret :
  forall D psi pc theta D' Psi Psi',
    IH_WFcmd psi pc theta (comms _iret)
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_ih Psi
    -> Psi' = ext_Prim_Specs Psi
    -> WFcmd D' Psi' (packS0 rho_ih theta) pc (comms _iret).
Proof.
  intros D psi pc [p g tag] D' Psi Psi'.
  intros Hi HD HD' Hlift HPsi'.
  subst D' Psi'.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl.
  rewrite minimalD_elim_ih; trivial.
  split_r with Itp_AIM_Mono_CT.
  intros Psi' [[[M R] fl] isr] HItp Hsub.
  unfold Itp_AIM in HItp.
  destruct HItp as [T [B [RdyQ [tid HItp]]]].
  unfold pmstar in HItp.
  destruct HItp as [Mth [Mq [HJM [Hwf HMq]]]].
  apply and_sym_rr.
  split.
    left.
    intros [Hx Hy].
    assert (H := IH_WFworld_IF_disable Hwf).
    simplS in * .
    rewrite Hx in H; discriminate.
  simpl in Hi.
  destruct Hi as [Htag Hp].
  unfold WF_AIM_World in Hwf.
  destruct Hwf as [H1 [H2 [H3 [H4 [TM [Mcth [H5 [H6 [H7 [H8 H9]]]]]]]]]].
  simplS in H7.
  rename H7 into Hwfcth.
  unfold WF_Curr_Thrd in Hwfcth.
  simpl in Hwfcth.
  destruct Hwfcth as [tcb [status [Hmto [Hst Hwfcth]]]].
  destruct (IH_iret_prog (pc:=pc) Htag Hp Hwfcth) as [Scth' HnextScth].
  split.
    destruct Scth' as [[[Mcth' R'] fl'] isr'].
    destruct (nextS_Mono2 H5 HJM H6 HnextScth) as [M' HnextS].
    exists (M', R', fl', isr'); trivial.
  intros pc' S' HnextS HnextPC.
  destruct S' as [[[M' R']fl']isr'].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr']]].
  destruct Hfl as [flnum [Hfl' Hlds]].
  destruct HM as [HM' Hlds'].
  subst M' R' fl' isr'.
  destruct (nextS_DMono2 H6 H5 HJM HnextScth HnextS)
    as [Mcth' [Mth' [TM' [HtoMcth' [HsplitTM' [HJM'  [HnextS' HTM']]]]]]].
  clear Scth' HnextScth.
  assert (HnextPC' : nextPC (comms _iret) Mcth R fl pc pc').
    inversion_clear HnextPC.
    destruct (nextS_inversion HnextS') as [HM [HR [_ _]]].
    destruct HM as [HM' HldsB].
    apply nextPC_iret with (R SP); trivial.
    simplS in * .
    subst sp.
    destruct (MemSplit_Inv H5 H6) as [Moth [Hx _]].
    assert (Hy : MemMod.disjUnion Mcth (MemMod.merge Moth Mq) M).
      eapply disjUnion_assoc_lr; eauto.
    eapply leadsToxW_leadsToW_DMono; eauto.
    apply leadsToBlk_leadsToxW with 6; auto with arith.
  destruct (IH_iret_prev Htag HnextPC' HnextS' Hp Hwfcth) 
    as [theta' [rho' [Hrho' [Hrtc [Hlk' Hwf]]]]].
  exists (packS0 rho' theta'); split_l.
  unfold ext_dict.
  destruct Hrho'; subst rho'.
    simpl.
    rewrite minimalD_elim_Rdy; trivial.
    unfold Itp_AIM.
    exists T; exists B; exists RdyQ; exists tid.
    exists Mth'; exists Mq; split_l.
    split_r.
    unfold WF_AIM_World.
    split.
      eapply WF_CurrTID_ih_nh; eauto.
    split_l; split_l; split_l.
    exists TM'; exists Mcth'; split_l.
    split_l.
    split.
    clear B RdyQ HtoMcth' HTM' H1 H2 H3 H4 H5 H6 H8 H9 HMq.
    clear Mq Mth TM HnextS HJM HJM'.
    unfold WF_Curr_Thrd.
    simpl; simplS.
    trivial.
    split.
      apply (WF_ReadyQ_untouch H8 HTM' H1).
    apply (WF_WQSet_untouch H9 HTM' H1).
  simpl.
  rewrite minimalD_elim_idle; trivial.
  unfold Itp_AIM.
  exists T; exists B; exists RdyQ; exists tid.
  exists Mth'; exists Mq; split_l.
  split_r.
  unfold WF_AIM_World.
  split. 
    eapply WF_CurrTID_ih_nh; eauto.
  split_l; split_l; split_l.
  exists TM'; exists Mcth'; split_l.
  split_l.
  split.
  clear B RdyQ HtoMcth' HTM' H1 H2 H3 H4 H5 H6 H8 H9 HMq.
  clear Mq Mth TM HnextS HJM HJM'.
  unfold WF_Curr_Thrd.
  simpl; simplS.
  trivial.
  split.
    apply (WF_ReadyQ_untouch H8 HTM' H1).
  apply (WF_WQSet_untouch H9 HTM' H1).
Qed.

Lemma SCAP_IH_in_CAP0_cmd:
  forall D psi pc theta c D' Psi Psi',
    IH_WFcmd psi pc theta c
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_ih Psi
    -> Psi' = ext_Prim_Specs Psi
    -> WFcmd D' Psi' (packS0 rho_ih theta) pc c.
Proof.
  intros D psi pc [p g tag] c D' Psi Psi'.
  intros Hi HD HD' Hlift HPsi'.
  induction c as [ i | s]; [ induction i | induction s].
  (* 19 goals remain, i = cli *)
  destruct Hi.
  (* 18 goals remain, i = sti *)
  destruct Hi.
  (* 17 goals remain, i = call f *)
  eapply SCAP_IH_in_CAP0_call; eauto.
  (* 16 goals remain, i = jne f *)
  eapply SCAP_IH_in_CAP0_jne; eauto.
  (* 15 goals remain, i = movwp o r *)
  eapply SCAP_IH_in_CAP0_movwp; eauto.
  (* 14 goals remain, i = movwld a r *)
  eapply SCAP_IH_in_CAP0_movwld; eauto.
  (* 13 goals remain, i = movwst o a *)
  eapply SCAP_IH_in_CAP0_movwst; eauto.
  (* 12 goals remain, i = pushwp o *)
  eapply SCAP_IH_in_CAP0_pushwp; eauto.
  (* 11 goals remain, i = pushwm a *)
  eapply SCAP_IH_in_CAP0_pushwm; eauto.
  (* 10 goals remain, i = pushfw *)
  destruct Hi.
  (* 9 goals remain, i = popw r *)
  eapply SCAP_IH_in_CAP0_popw; eauto.
  (* 8 goals remain, i = popfw *)
  destruct Hi.
  (* 7 goals remain, i = cmpw o r *)
  eapply SCAP_IH_in_CAP0_cmpw; eauto.
  (* 6 goals remain, i = addw o r *)
  eapply SCAP_IH_in_CAP0_addw; eauto.
  (* 5 goals remain, i = subw o r *)
  eapply SCAP_IH_in_CAP0_subw; eauto.
  (* 4 goals remain, i = outb eoi *)
  eapply SCAP_IH_in_CAP0_eoi; eauto.
  (* 3 goals remain, i = jmp f *)
  eapply SCAP_IH_in_CAP0_jmp; eauto.
  (* 2 goals remain, i = retw *)
  eapply SCAP_IH_in_CAP0_retw; eauto.
  (* 1 goals remain, i = iret *)
  eapply SCAP_IH_in_CAP0_iret; eauto.
Qed.
Implicit Arguments SCAP_IH_in_CAP0_cmd [D psi pc theta c D' Psi Psi'].

Lemma SCAP_IH_in_CAP0_iseq:
  forall D psi pc theta isq D' Psi Psi',
    IH_WFiseq psi pc theta isq
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_ih Psi
    -> Psi' = ext_Prim_Specs Psi
    -> WFiseq D' Psi' (packS0 rho_ih theta) pc isq.
Proof.
  intros D psi pc theta isq D' Psi Psi'.
  intros Hscap HD HD' Hlift HPsi'.
  gen_clear theta pc Hscap HPsi'.
  induction isq; intros theta pc Hscap HPsi'.
    inversion Hscap. subst psi0 pc0 theta i0 isq0.
    eapply wfis_seq with (packS0 rho_ih (Ast p' g' tag)).
    assert (Hlift' := lift_psi_cons (f := pc+1) (spec := Ast p' g' tag) Hlift).
    apply WFcmd_Mono_CT with (ext_Prim_Specs 
        (consCT (pc + 1) (packS0 rho_ih (Ast p' g' tag)) Psi)).
    subst Psi'.
    unfold ext_Prim_Specs.
    apply Psi_extend3_subsetCT; auto.
    apply lookupCT_lookupCT_consCT.
    apply lookupCT_consCT.
    apply Psi_extend3_subsetCT; auto.
    apply lookupCT_lookupCT_consCT.
    apply lookupCT_lookupCT_consCT.
    apply lookupCT_consCT.
    apply Psi_extend3_subsetCT; auto.
    apply lookupCT_lookupCT_consCT.
    apply lookupCT_lookupCT_consCT.
    apply lookupCT_lookupCT_consCT.
    apply lookupCT_consCT.
    assert (HPsi'2 := refl_equal (ext_Prim_Specs 
        (consCT (pc + 1) (packS0 rho_ih (Ast p' g' tag)) Psi))).
    apply (SCAP_IH_in_CAP0_cmd H4 HD HD' Hlift' HPsi'2).
    apply IHisq; trivial.
  inversion Hscap. subst psi0 pc0 theta si.
  eapply wfis_end.
  apply (SCAP_IH_in_CAP0_cmd H3 HD HD' Hlift HPsi').
Qed.

Lemma SCAP_IH_in_CAP0_cdhp :
  forall D psi C psi' D' Psi Psi' Psi'',
    IH_WFcdhp psi C psi'
    -> Minimal_D D
    -> D' = ext_dict D rho_prim (Generic_Itp D rho_prim)
    -> lift_psi psi rho_ih Psi
    -> lift_psi psi' rho_ih Psi'
    -> sink_psi psi' rho_ih Psi'
    -> Psi'' = ext_Prim_Specs Psi
    -> WFcdhp D' Psi'' C Psi'.
Proof.
  intros D psi C psi' D' Psi Psi' Psi''. 
  intros H HD HD' H0 H1 H2 HPsi''.
  intros f theta0 Hf.
  substH H2 with (H2 f theta0 Hf).
  destruct H2 as [theta [Htheta0 Hfx]].
  unfold SCAP_WFcdhp in H.
  substH H with (H f theta Hfx).
  destruct H as [isq [Hisq H]].
  exists isq; split_l.
  destruct theta0.
  injection Htheta0; intros; subst n s.
  eapply SCAP_IH_in_CAP0_iseq; eauto.
Qed.
