Add LoadPath ".." .
Add LoadPath "../CAP0" .

Require Import Arith.
Require Omega.
Require Import Coqlib.
Require Import Map.
Require Import Machine.
Require Import Assertion.

Require Import Assertion_Lemmas.
Require Import Assertion_Lemmas5.
Require Import Operational_Semantics.
Require Import OpSem_Lemmas.
Require Import Mem_Lemmas.

Require Import CAP0.
Require Import SCAP.
Require Import AIM.
Require Import QLib.
Require Import SCAP_READY.
Require Import SCAP_IDLE.
Require Import SCAP_Itr.

Require Import EqNat.

Require Import Linking.
Require Import Linking_Lemmas2.

Lemma NH_Itr_prog : forall rho Psi pc M R fl isr p g tag,
  (Prd_Itr_Stk p /\ Grt_Itr_Stk g)
  -> fl _IF = true
  -> itp_NH_theta rho (Ast p g tag) Psi (M, R, fl, isr)
  -> exists S', itr_nextS pc (M, R, fl, isr) S'.
Proof.
  intros rho Psi pc M R fl isr p g tag.
  intros HpgStk Hfl H.
  unfold itp_NH_theta in H.
  destruct H as [Hisr [Hp Hwfst]].
  unfold NH_pinvstar, pmstar in Hp.
  simpl; simplS in * .
  destruct Hp as [Muk [Mis [HJM [Hp HMis]]]].
  destruct HpgStk as [HpStk HgStk].
  unfold Prd_Itr_Stk in HpStk.
  substH HpStk with (HpStk _ Hp). 
  unfold itr_enabled in HpStk; simplS in HpStk. 
  substH HpStk with (HpStk (conj Hfl Hisr)).
  destruct HpStk as [Mu [Mk [HJMuk [HMk HpStk]]]].
  simplS in * .
  set (M'' := MemMod.put (MemMod.put M (R SP - 6) pc) (R SP - 5) 0).
  set (M''' := MemMod.put (MemMod.put M'' (R SP - 4) 0) (R SP - 3) 0).
  set (M' := MemMod.put (MemMod.put M''' (R SP - 2) (FR2nat fl)) (R SP - 1) 0).
  exists (M', RF_update R SP (R SP - 6), Clear_Flag fl _IF, true).
  exists (R SP).
  exists (FR2nat fl).
  exists 0.
  exists M''; exists M'''.
  split_l.
  split. 
    destruct (stkSpacePrec_split HMk) as [Mik [Mjk [HJMk [HMik HMjk]]]].
    apply stkSpace_Mono with Muk Mis; trivial.
    apply stkSpace_Mono_r with Mu Mk; trivial.
    apply stkSpace_Mono with Mik Mjk; trivial.
    apply stkSpacePrec_stkSpace; trivial.
  split_l; split_l; split_l;split_l.
  assert (Hsp : R SP >= 6).
    unfold stkSpacePrec in HMk.
    destruct HMk as [H _ ].
    omega.
  split.
    unfold MemUpd16.
    exists 0.
    replace (R SP - 6 + 1) with (R SP - 5).
    trivial.
    clear rho Psi Hwfst HgStk Hfl Hisr. 
    subst M'' M' M'''.
    clear pc M fl isr tag p g Muk Mis HJM Hp HMis Mu Mk HJMuk HMk HpStk.
    omega.
  split.
    unfold MemUpd16.
    exists 0.
    replace (R SP - 4 + 1) with (R SP - 3).
    trivial.
    clear rho Psi Hwfst HgStk Hfl Hisr. 
    subst M'' M' M'''.
    clear pc M fl isr tag p g Muk Mis HJM Hp HMis Mu Mk HJMuk HMk HpStk.
    omega.
  unfold MemUpd16.
  exists 0.
  replace (R SP - 2 + 1) with (R SP - 1).
  trivial.
  clear rho Psi Hwfst HgStk Hfl Hisr. 
  subst M'' M' M'''.
  clear pc M fl isr tag p g Muk Mis HJM Hp HMis Mu Mk HJMuk HMk HpStk.
  omega.
Qed.
Implicit Arguments NH_Itr_prog [rho Psi pc M R fl isr p g tag].

Lemma NH_Itr_prev : forall rho status Psi pc M R fl isr S' p g tag,
  rho = rho_ready \/ rho = rho_idle
  -> rho_tcb_consistent2 status rho
  -> (Prd_Itr_Stk p /\ Grt_Itr_Stk g)
  -> fl _IF = true
  -> itr_nextS pc (M, R, fl, isr) S'
  -> itp_NH_theta rho (Ast p g tag) Psi (M, R, fl, isr)
  -> itp_IH_theta (Ast IHpred IHgrt 0) status (consCT pc (packS0 rho (Ast p g tag)) Psi) S'.
Proof.  
  intros rho status Psi pc M R fl isr [[[M' R'] fl'] isr'] p g tag.
  intros Hrho Htcs HpgStk Hfl HnextS H.
  unfold itr_nextS in HnextS.
  destruct HnextS as [sp [flnum [cs [M'' [M''' [Hsp [_ [HR' [Hfl' Hitr]]]]]]]]].
  destruct Hitr as [Hisr' [Hflnum HM']].
  simplS in * .
  unfold itp_IH_theta.
  unfold itp_NH_theta in H.
  destruct H as [Hisr [Hp Hwfst]].
  unfold NH_pinvstar, pmstar in Hp.
  simpl; simplS in * .
  destruct Hp as [Muk [Mis [HJM [Hp HMis]]]].
  destruct HpgStk as [HpStk HgStk].
  unfold Prd_Itr_Stk in HpStk.
  substH HpStk with (HpStk _ Hp). 
  unfold itr_enabled in HpStk; simplS in HpStk. 
  substH HpStk with (HpStk (conj Hfl Hisr)).
  destruct HpStk as [Mu [Mk [HJMuk [HMk HpStk]]]].
  simplS in * .
  subst sp isr R' fl' isr' flnum.
  split_l.
  left.
  split_l.
  destruct (stkSpacePrec_split HMk) as [Mik [Mjk [HJMk [HMik HMjk]]]].
  unfold NH_Inv, pmstar in HMis.
  rewrite Hfl in HMis.
  destruct HMis as [Mih [Msh [HJMis [HMih HMsh]]]].
  assert (H : exists Mik',
    MemMod.disjUnion Mik' (MemMod.merge (MemMod.merge Mu Mjk) Mis) M'
    /\ exists M1'1, exists M1'2, 
      MemUpd16 Mik (R SP - 6) pc M1'1
      /\ MemUpd16 M1'1 (R SP - 4) cs M1'2
      /\ MemUpd16 M1'2 (R SP - 2) (FR2nat fl) Mik').
    assert (H : leadsToxW (R SP - 6)  Mik).
      apply stkSpacePrec_leadsToxW with (R SP) 6; trivial; try omega.
    assert (H1 : MemMod.disjUnion Mik (MemMod.merge (MemMod.merge Mu Mjk) Mis) M).
      disjUnion shift_l.
      merge commut Mu.
      merge assoc Mik.
      merge rewrite HJMk.
      merge commut Mk.
      merge rewrite HJMuk.
      trivial.
      disjoint des.
      disjoint lift_l.
    destruct HM' as [HM'1 [HM'2 HM'3]].
    rename M'' into M'1.
    rename M''' into M'2.
    destruct (leadsToxW_MemUpd16_DMono H1 H HM'1) as [Mik'1 [HJ1 Hupd1]].
    destruct ((fun x => HMik) 0) as [Hsp _ ].
    assert (H2 : leadsToxW (R SP - 4) Mik'1).
      apply leadsToxW_MemUpd16_leadsToxW_nt with (R SP - 6) pc Mik; trivial. 
      apply stkSpacePrec_leadsToxW with (R SP) 6; trivial; try omega.
      left; omega.
    destruct (leadsToxW_MemUpd16_DMono HJ1 H2 HM'2) as [Mik'2 [HJ2 HMik'2]].
    assert (H3 : leadsToxW (R SP - 2) Mik'2).
      apply leadsToxW_MemUpd16_leadsToxW_nt with (R SP - 4) cs Mik'1; trivial.
      apply leadsToxW_MemUpd16_leadsToxW_nt with (R SP - 6) pc Mik; 
        trivial; try omega.
      apply stkSpacePrec_leadsToxW with (R SP) 6; trivial; try omega.
      left; omega.
    destruct (leadsToxW_MemUpd16_DMono HJ2 H3 HM'3) as [Mik' [HJ HMik']].
    exists Mik'; split_l.
    exists Mik'1; exists Mik'2; auto.
  destruct H as [Mik' [HJM' HMik'upd]].
  clear HM' M'' M'''.
  destruct ((fun x => HMik) 0) as [Hsp _ ].
  destruct HMik'upd as [Mik'1 [Mik'2 [Hupd1 [Hupd2 Hupd3]]]].
  assert (HMik'1 : stkSpacePrec (R SP) 6 Mik'1).
    apply stkSpacePrec_MemUpd16_stkSpacePrec with Mik (R SP - 6) pc;
      trivial; try omega.
  assert (HMik'2 : stkSpacePrec (R SP) 6 Mik'2).
    apply stkSpacePrec_MemUpd16_stkSpacePrec with Mik'1 (R SP - 4) cs;
      trivial; try omega.
  assert (HMik' : stkSpacePrec (R SP) 6 Mik').
    apply stkSpacePrec_MemUpd16_stkSpacePrec with Mik'2 (R SP - 2) (FR2nat fl);
      trivial; try omega.
  assert (HMik'eqd : MemMod.eqDom Mik Mik').
    assert (HMik'eqd1 : MemMod.eqDom Mik Mik'1).
      apply stkSpacePrec_MemUpd16_eqDom with (R SP) 6 (R SP - 6) pc;
        trivial; try omega.
    assert (HMik'eqd2 : MemMod.eqDom Mik'1 Mik'2).
      apply stkSpacePrec_MemUpd16_eqDom with (R SP) 6 (R SP - 4) cs;
        trivial; try omega.
    assert (HMik'eqd3 : MemMod.eqDom Mik'2 Mik').
      apply stkSpacePrec_MemUpd16_eqDom with (R SP) 6 (R SP - 2) (FR2nat fl);
        trivial; try omega.
    apply MemMod.eqDom_trans with Mik'1; trivial.
    apply MemMod.eqDom_trans with Mik'2; trivial.
  split.
    unfold IHpred, pmstar.
    simplS.
    exists (MemMod.merge Mu (MemMod.merge Mih (MemMod.merge Mik' Mjk))). 
    exists Msh.
    split.
      merge commut Mih.
      merge assoc Mu.
      disjUnion_shift_r.
      merge_commut Mu.
      merge_assoc Mik'.
      merge_commut Mjk.
      disjUnion_shift_r.
      merge rewrite HJMis.
      trivial.
      disjoint des.
      disjoint lift_lr.
      disjoint lift_lr.
      disjoint lift_llr.
      disjoint lift_llr.
      disjoint des.
      disjoint rep_r Mik; disjoint lift_r.
      disjoint lift_r.
      disjoint des.
      disjoint rep_r Mik; disjoint lift_lrr.
      disjoint lift_lrr.
   split_r.
    rsplit.
      auto.
    exists Mu; exists (MemMod.merge Mih (MemMod.merge Mik' Mjk)).
    split.
      split_r.
      disjoint des.
      disjoint lift_lr.
      disjoint rep_r Mik; disjoint lift_r. 
      disjoint lift_r.
    split.
      unfold pTrue; trivial.
    exists Mih; exists (MemMod.merge Mik' Mjk).
    split.
      split_r.
      disjoint des.
      disjoint rep_r Mik.
      disjoint lift_lrr.
      disjoint lift_lrr.
    split_l.
    unfold RF_update, beq_reg.
    apply stkSpacePrec_downup.
    apply (stkSpacePrec_merge HMik' HMjk).
  unfold IH_ItpG.
  intros [[[M'' R''] fl''] isr''].
  simplS.
  intros [Mx [My [M1'' [Msh'' H]]]].
  destruct H as [HJM'_2 [HJM'' [HMy [HMsh'' Hgih]]]].
  assert (HJM'_3 : MemMod.disjUnion 
    (MemMod.merge Mik' (MemMod.merge (MemMod.merge Mu Mjk) Mih)) Msh M').
    disjUnion_shift_r.
    merge assoc Msh.
    merge rewrite HJMis. 
    trivial.
    disjoint des.
    disjoint lift_r.
    disjoint lift_llr.
    disjoint lift_llr.
  destruct (prec_disjUnion_inv HJM'_3 HJM'_2  HMsh HMy Prec1) as [Heq1 Heq2].
  subst Mx My.
  clear HJM'_2 HJM'_3.
  unfold IHgrt in Hgih.
  unfold prepost in Hgih.
  simplS in Hgih.
  destruct Hgih as [Hgih [HRr Hisrr]].
  subst R'' isr''.
  unfold RF_update, beq_reg in Hgih.
  destruct Hgih as [Mx [My [Mjk''ih'' [HMx [HMih''jk'' [HJMin HJM1'']]]]]].
  assert (H : MemMod.disjUnion (MemMod.merge Mih Mjk) (MemMod.merge Mu Mik')
    (MemMod.merge Mik' (MemMod.merge (MemMod.merge Mu Mjk) Mih))).
    merge assoc Mih.
    rewrite (@merge_sym Mih Mjk).
    merge assoc Mik'.
    rewrite (@merge_sym Mik' Mu).
    rewrite (@merge_sym (MemMod.merge Mu Mik')  (MemMod.merge Mjk Mih)).
    split_r.
    disjoint des.
    disjoint lift_llr.
    disjoint rep_r Mik.
    disjoint lift_lr.
    disjoint rep_r Mik.
    disjoint lift_lrr.
    disjoint des.
    disjoint lift_r.
    disjoint lift_lr.
    disjoint rep_l Mik.
    disjoint rep_l Mik; disjoint lift_llr.
    disjoint rep_l Mik; disjoint lift_l.
    disjoint lift_lrr.
  assert (H0 : mstar INV0 (stkSpacePrec (R SP - 6) IH_Stk) (MemMod.merge Mih Mjk)).
    exists Mih; exists Mjk.
    split_r by auto.
    split_r.
    disjoint lift_lrr.
  destruct (prec_disjUnion_inv_left HJMin H  HMx H0) as [Heq1 Heq2].
    apply (@mstar_presv_prec _ _ Prec0 (@precise_stkSpacePrec (R SP - 6) IH_Stk)).
  subst Mx My.
  clear HJMin H H0 HMx HMy.
  exists (M'', R, fl, false).
  assert (HJM1''2 : MemMod.disjUnion Mik' (MemMod.merge Mjk''ih'' Mu) M1'').
    apply disjUnion_sym.
    disjUnion shift_r.
    trivial.
    disjoint rep_r Mik.
    disjoint lift_r.
  split.
    intros pc'.
    apply nextS_iret with (FR2nat fl) (R SP - 6); trivial.
    apply leadsToBlk_Mono with M1'' Msh''; trivial.
    apply leadsToBlk_Mono with  Mik' (MemMod.merge Mjk''ih'' Mu); trivial.
    apply stkSpacePrec_leadsToBlk; trivial.
    apply leadsToW_Mono with M1'' Msh''; trivial.
    apply leadsToW_Mono with  Mik' (MemMod.merge Mjk''ih'' Mu); trivial.
    arith_rep (R SP - 6 + 4) with (R SP - 2).
    apply leadsToxW_MemUpd16_leadsToW with Mik'2; trivial.
    apply leadsToxW_MemUpd16_leadsToxW_nt with (R SP - 4) cs Mik'1; trivial.
    apply leadsToxW_MemUpd16_leadsToxW_nt with (R SP - 6) pc Mik; trivial.
    eapply stkSpacePrec_leadsToxW; eauto.
    omega.
    omega.
    left; omega.
    left; omega.
    rewrite (FR_nat_iso_left fl); trivial.
    arith_rep (R SP - 6 + 6) with (R SP).
    rewrite RF_update_RF_update.
    rewrite RF_update_prev; trivial.
  exists pc.
  split.
    unfold Ret_Addr.
    simplS.
    intros pcx.
    apply nextPC_retw with (sp:=(R SP -6)); trivial.
    apply leadsToW_Mono with M1'' Msh''; trivial.
    apply leadsToW_Mono with  Mik' (MemMod.merge Mjk''ih'' Mu); trivial.
    apply leadsToW_MemUpd16_leadsToW_nt with (R SP - 2) (FR2nat fl) Mik'2; trivial.
    apply leadsToW_MemUpd16_leadsToW_nt with (R SP - 4) cs Mik'1; trivial.
    apply leadsToxW_MemUpd16_leadsToW with Mik; trivial.
    eapply stkSpacePrec_leadsToxW; eauto.
    omega.
    omega.
    right; omega.
    right; omega.
  exists rho.
  exists (Ast p g tag).
  split_l.
  split_l.
  split.
    apply lookupCT_consCT; trivial.
  intros pc' [[[M''' R'''] fl'''] isr'''] HnextS''.
  unfold itp_NH_theta.
  destruct (nextS_inversion HnextS'') as [HM''' [HR''' [Hfl''' Hisr''']]].
  destruct HM''' as [HM''' HldsM'''].
  destruct Hfl''' as [flnum [Hfl''' HldsM'''x]].
  assert (fl''' = fl).
    rewrite RF_update_RF in HldsM'''x.
    assert (leadsToW (R SP - 2) (FR2nat fl) M'').
    apply leadsToW_Mono with M1'' Msh''; trivial.  
    apply leadsToW_Mono with Mik' (MemMod.merge Mjk''ih'' Mu); trivial.
    apply leadsToxW_MemUpd16_leadsToW with Mik'2; trivial.
    apply leadsToxW_MemUpd16_leadsToxW_nt with (R SP - 4) cs Mik'1; trivial.
    apply leadsToxW_MemUpd16_leadsToxW_nt with (R SP - 6) pc Mik; trivial.
    eapply stkSpacePrec_leadsToxW; eauto.
    omega.
    omega.
    left; omega.
    left; omega.
    arith_rep (R SP -6 + 4) with (R SP -2) in HldsM'''x.
    assert (Heq := leadsToW_unique HldsM'''x H).
    assert (Heq' : nat2FR flnum = nat2FR (FR2nat fl)). 
      subst flnum; trivial. 
    subst fl'''.
    rewrite Heq'.
    rewrite (FR_nat_iso_left fl); trivial.
  subst M''' R''' fl''' isr'''.
  simplS.
  split_l.
  destruct HMih''jk'' as [Mih'' [Mjk'' [HJMih''jk'' [HMih'' HMjk'']]]].
  assert (HMjk'eqd : MemMod.eqDom Mjk Mjk'').
    eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
  assert (HMk'' : stkSpacePrec (R SP) (IH_Stk + 6) (MemMod.merge Mik' Mjk'')).
    apply stkSpacePrec_merge; trivial.
  split.
    unfold NH_pinvstar, NH_Inv, pmstar.
    simplS.
    rewrite H.
    rewrite Hfl.
    exists (MemMod.merge Mu (MemMod.merge Mik' Mjk'')).
    exists (MemMod.merge Mih'' Msh'').
    split.
    disjUnion shift_l.
    merge assoc Mu.
    merge assoc Mih''.
    merge commut Mjk''.
    merge rewrite HJMih''jk''.
    merge commut Mjk''ih''.
    merge rewrite HJM1''.
    trivial.
    disjoint des.
    disjoint lift_r.
    disjoint lift_l Mik'.
    apply disjUnion_sym.
    split_r.
    disjoint rep_r Mik.
    disjoint lift_r.
    disjoint lift_r.
    disjoint lift_l Mu.
    apply disjUnion_sym.
    split_r.
    disjoint rep_l Mik.
    disjoint lift_l.
    merge commut Mu.
    disjoint trivial.
    disjoint rep_l Mik.
    disjoint lift_l.
    rewrite RF_update_RF_update.
    rewrite RF_update_RF.
    arith_rep (R SP - 6 + 6) with (R SP).
    rewrite RF_update_prev.
    split.
      apply HpStk; trivial.
    exists Mih''; exists Msh''.
    split; auto.
    split_r. 
    disjoint lift_l. 
    disjoint lift_l.
  intros Htag.
  substH Hwfst with (Hwfst Htag); clear Htag.
  assert (Hsub : subsetCT Psi (consCT pc (packS0 rho (Ast p g tag)) Psi)).
    apply consCT_subsetCT; trivial.
  apply NH_WFST_Mono_CT with Psi; trivial.
  apply NH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, false));
    trivial.
  clear Hwfst.
  intros [[[M'''' R''''] fl''''] isr''''] [Mx1 [Mx2 [M1'''' [M2'''' Hg]]]].
  simplS; simplS in Hg.
  destruct Hg as [Hg1 [Hg2 [Hg3 [Hg4 Hg5]]]].
  unfold NH_Inv in Hg3.
  rewrite H in Hg3.
  rewrite Hfl in Hg3.
  assert (HJM''_2 : MemMod.disjUnion (MemMod.merge Mu (MemMod.merge Mik' Mjk'')) 
    (MemMod.merge Mih'' Msh'') M'').
    disjUnion shift_l.
    merge assoc Mu.
    merge assoc Mih''.
    merge commut Mjk''.
    merge rewrite HJMih''jk''.
    merge commut Mjk''ih''.
    merge rewrite HJM1''.
    trivial.
    disjoint des.
    disjoint lift_r.
    disjoint lift_l Mik'.
    apply disjUnion_sym.
    split_r.
    disjoint rep_r Mik.
    disjoint lift_r.

    disjoint lift_l Mu.
    apply disjUnion_sym.
    split_r.
    disjoint rep_l Mik.
    disjoint lift_l.

    disjoint lift_r.
    merge commut Mu.
    disjoint trivial.
    disjoint rep_l Mik.
    disjoint lift_l.
  assert (H0 : mstar INV0 INV1 (MemMod.merge Mih'' Msh'')).
    exists Mih''; exists Msh''.
    split; auto.
    split_r. 
    disjoint lift_l. 
    disjoint lift_l.
  destruct (prec_disjUnion_inv Hg1 HJM''_2 Hg3 H0 Prec01) as [Heq1 Heq2].
  subst Mx1 Mx2.
  clear H0 Hg3 Hg1.
  unfold NH_ItpG.
  simplS.
  exists (MemMod.merge Mu (MemMod.merge Mik Mjk)). 
  exists (MemMod.merge Mih Msh).
  exists M1''''; exists M2''''.
  split.
    merge rewrite HJMis.
    merge rewrite HJMk.
    merge rewrite HJMuk.
    trivial.
  split_l.
  split.
    unfold NH_Inv.
    rewrite Hfl.
    exists Mih; exists Msh.
    split_r by auto.
    split_r.
    disjoint trivial.
  split_l.
  rewrite RF_update_RF_update in Hg5.
  rewrite RF_update_RF in Hg5. 
  arith_rep (R SP - 6 + 6) with (R SP) in Hg5.
  rewrite RF_update_prev in Hg5.
  unfold Grt_Itr_Stk in HgStk.
  substH HgStk with (HgStk _ _ Hg5).
  unfold itr_enabled in HgStk.
  simplS in HgStk.
  rewrite H in HgStk.
  rewrite Hfl in HgStk.
  substH HgStk with (HgStk (conj (refl_equal _) (refl_equal _))).
  substH HgStk with (HgStk Mu (MemMod.merge Mik' Mjk'')).
  substH HgStk with (HgStk (MemMod.merge Mik Mjk)).
  apply HgStk; trivial.
    split_r.
    disjoint des.
    disjoint rep_r Mik.
    disjoint lift_r.
    disjoint rep_r Mjk.
    disjoint lift_r.
  apply stkSpacePrec_merge; trivial.
Qed.
Implicit Arguments NH_Itr_prev [rho status Psi pc M R fl isr S' p g tag].

Lemma NH_cli_prev : forall rho Psi M R fl isr pc p g tag p' g',
  lookupCT Psi (pc+1) (packS0 rho (Ast p' g' tag))
  -> pimp (p |> gcli) p'
  -> gimp (p >o gcli oo g') g
  -> itp_NH_theta rho (Ast p g tag) Psi (M, R, fl, isr)
  -> itp_NH_theta rho (Ast p' g' tag) Psi (M, R, Clear_Flag fl _IF, isr).
Proof.
  unfold itp_NH_theta.
  intros rho Psi M R fl isr pc p g tag p' g' Hlookup Hpp' Hg'g Hwf.
  simplS in * .
  destruct Hwf as [Hisr [Hpinv Hwfst]].
  split_l.
  split.
    unfold NH_pinvstar in * .
    unfold NH_Inv in * .
    simpl.
    unfold pmstar in Hpinv |- * .
    destruct Hpinv as [M1 [M2 [HJM [HM1 HM2]]]].
    simplS in HM2.
    exists M; exists MemMod.emp.
    split.
      simplS.
      auto with PMap.
    simplS.
    rsplit.
      unfold Aemp; trivial.
    apply Hpp'.
    simplS in HM1.
    apply linkPG_intro with (M1, R, fl, isr); trivial.
    unfold gcli.
    rsplit.              
      split_l.
      split_lr.
    unfold prepost, choice, itr_enabled.
    simplS.
    destruct (fl _IF).
    (* case : fl _IF = true *)
      exists MemMod.emp.
      exists M1.
      exists M2.
      split.
        unfold Aemp; trivial.
      rsplit.
        split; auto with PMap.
      left.
      split_r.
      split_lr.
    (* case : fl _IF = false *)
    exists MemMod.emp.
    exists M.
    exists MemMod.emp.
    unfold Aemp in HM2.
    subst M2.
    split.
      unfold Aemp; trivial.
    rsplit.
      simplS in HJM.
      assert (M1 = M).
        auto with PMap.
      subst M1.
      split; auto with PMap.
    right.
    split.
      intros [Hx Hy]; discriminate Hx.
    unfold Aemp; trivial.
  intro Htag.
  substH Hwfst with (Hwfst Htag).
  apply NH_WFST_strengthen with (g := g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, NH_ItpG, NH_Inv.
  intros [[[Mx Rx] flx] isrx] [Mx1 [Mx2 [Mx1' [Mx2' Hg]]]].
  simplS; simplS in Hg.
  simpl in Hg.
  unfold NH_pinvstar, NH_Inv in Hpinv.
  simplS in Hpinv.
  unfold pmstar in Hpinv.
  destruct Hpinv as [M1 [M2 [HJM [HM1 HM2]]]].
  simplS in HM2.
  exists M1; exists M2.
  exists Mx1'; exists Mx2'.
  destruct Hg as [Hg1 [Hg2 [Hg3 [Hg4 Hg5]]]].
  split_l.
  split_l.
  split_l.
  split_l.
  assert (Heq : Mx1 = M).
    unfold Aemp in Hg3; subst Mx2.
    auto with PMap.
  subst Mx1.
  clear Mx2 Hg1 Hg3.
  apply Hg'g.
  apply conjG_intro with (M, R, Clear_Flag fl _IF, isr); trivial.
  apply conjPG_intro; trivial.
  unfold gcli, prepost, choice, itr_enabled.
  rsplit.
    simplS.
  unfold Clear_Flag; simpl.
    split_l; split_lr.
  exists MemMod.emp; exists M1; exists M2.
  split.
    unfold Aemp; trivial.
  rsplit.
    simplS.
    simplS in HJM.
    split; auto with PMap.
  simplS; simplS in Hisr.    
  destruct (fl _IF); [left | right]; split_r. 
    split_lr.
  intros [Hx Hy]; discriminate Hx.
Qed.
Implicit Arguments NH_cli_prev [rho Psi M R fl isr pc p g tag p' g'].

Lemma NH_sti_prev : forall rho Psi M R fl isr pc p g tag p' g',
  lookupCT Psi (pc+1) (packS0 rho (Ast p' g' tag))
  -> enable_sti p
  -> pimp (p |> gsti) p'
  -> gimp (p >o gsti oo g') g
  -> itp_NH_theta rho (Ast p g tag) Psi (M, R, fl, isr)
  -> itp_NH_theta rho (Ast p' g' tag) Psi (M, R, Set_Flag fl _IF, isr).
Proof.
  intros rho Psi M R fl isr pc p g tag p' g' Hlookup Hsti Hpp' Hg'g Hwf.
  unfold itp_NH_theta in * .
  destruct Hwf as [Hisr [Hpinv Hwfst]].
  split_l.
  split.
    unfold NH_pinvstar in * .
    unfold NH_Inv in * .
    simpl.
    unfold pmstar in Hpinv |- * .
    destruct Hpinv as [M1 [M2 [HJM [HM1 HM2]]]].
    simplS in HM2.
    generalize (refl_equal (fl _IF)); pattern fl at 1; destruct (fl _IF); intro Hfl.
    (* case : fl _IF = true *)
      exists M1; exists M2.
      split_l.
      rsplit.
        unfold Aemp; trivial.
      apply Hpp'.
      simplS in HM1.
      apply linkPG_intro with (M1, R, fl, isr); trivial.
      unfold gsti.
      rsplit.
        split_l.
        split_lr.
      unfold prepost, choice, itr_enabled.
      simplS.
      rewrite Hfl.
      exists MemMod.emp; exists M1; exists MemMod.emp.
      rsplit.
        split.
          unfold Aemp; trivial.
        split; auto with PMap.
      left.
      split.
        split_lr.
      unfold Aemp; trivial.
    (* case : fl _IF = false *)
    unfold enable_sti in Hsti.
    substH Hsti with (Hsti _ HM1).
    unfold pmstar, choice, itr_enabled in Hsti.
    simplS in Hsti.
    destruct Hsti as [M11 [M12 [HJM1 [HM11 Hsti]]]].
    exists M11; exists M12.
    assert (Heq : M1 = M).
      unfold Aemp in HM2; subst M2.
      simplS in HJM.
      auto with PMap.
    subst M1.
    split_l.
    rewrite Hfl in Hsti.
    destruct Hsti as [Hsti | Hsti].
      destruct Hsti as [[Hx _] _]; discriminate Hx.
    destruct Hsti as [_ HM12].
    split_r.
    apply Hpp'.
    apply linkPG_intro with (M, R, fl, isr); trivial.
    unfold gsti.
    rsplit.
      split_l.
      split_lr.
    unfold prepost, choice, itr_enabled.
    simplS.
    rewrite Hfl.
    exists M12; exists M11; exists MemMod.emp.
    rsplit.
      split.
        unfold Aemp; trivial.
      split; auto with PMap.
    right.
    split_r.
    intros [Hx Hy]; discriminate Hx.
  intro Htag.
  substH Hwfst with (Hwfst Htag).
  apply NH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, NH_ItpG, NH_Inv.
  intros [[[Mx Rx] flx] isrx] [Mx1 [Mx2 [Mx1' [Mx2' Hg]]]].
  simplS; simplS in Hg.
  simpl in Hg.
  unfold NH_pinvstar, NH_Inv in Hpinv.
  simplS in Hpinv.
  unfold pmstar in Hpinv.
  destruct Hpinv as [M1 [M2 [HJM [HM1 HM2]]]].
  simplS in HM2.
  exists M1; exists M2.
  exists Mx1'; exists Mx2'.
  destruct Hg as [Hg1 [Hg2 [Hg3 [Hg4 Hg5]]]].
  split_l.
  split_l.
  split_l.
  split_l.
  apply Hg'g.
  apply conjG_intro with (Mx1, R, Set_Flag fl _IF, isr); trivial.
  apply conjPG_intro; trivial.
  unfold gsti, prepost, choice, itr_enabled.
  simplS.
  rsplit.
    unfold Set_Flag; simpl.
    split_l; split_lr.
  destruct (fl _IF).
    assert (Hprec : precise (mstar INV0 INV1)).
      apply Prec01.
    destruct (prec_disjUnion_inv HJM Hg1 HM2 Hg3 Hprec) as [Heq1 Heq2].
    subst Mx1 Mx2.
    exists MemMod.emp; exists M1; exists MemMod.emp.
    rsplit.
      split.
        unfold Aemp; trivial.
      split; auto with PMap.
    left.
    rsplit.
      unfold Aemp; trivial.
    split_lr.
  assert (Heq : M1 = M).
    simplS in HJM.
    unfold Aemp in HM2; subst M2.
    auto with PMap.
  subst M1.
  exists Mx2; exists Mx1; exists MemMod.emp.
  rsplit.
    split. 
      unfold Aemp; trivial.
    split; auto with PMap.
  right.
  split_r.
  intros [Hx Hy]; discriminate Hx.
Qed.
Implicit Arguments NH_sti_prev [rho Psi M R fl isr pc p g tag p' g'].

Lemma NH_call_prog : forall rho Psi f pc S0
  (p : Pred) (g : Grt) tag,
  enable p (nextS pc (comml (_call f)))
  -> itp_NH_theta rho (Ast p g tag) Psi S0
  -> exists S', nextS pc (comml (_call f)) S0 S'.
Proof.
  intros rho Psi f pc [[[M R] fl] isr] p g tag.
  intros Hen H.
  unfold itp_NH_theta, NH_pinvstar, pmstar in * .
  destruct H as [Hisrf [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  simplS in * .
  destruct (Hen _ HMnh) as [S' HnextSnh]. 
  destruct S' as [[[Mnh' R'] fl'] isr'].
  destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
  exists (M', R', fl', isr'); trivial.
Qed.
Implicit Arguments NH_call_prog [rho Psi f pc S0 p g tag].

Lemma NH_nonprim_call_prev : forall rho Psi f pc S0 S'
  (p p' p'': Pred) (g g' g'': Grt) tag,
  lookupCT Psi f (packS0 rho (Ast p' g' 1))
  -> lookupCT Psi (pc+1) (packS0 rho (Ast p'' g'' tag))
  -> nextS pc (comml (_call f)) S0 S'
  -> enable p (nextS pc (comml (_call f)))
  -> gimp (sp_call_pred p pc f >o g') fun_frame
  -> pimp (p |> nextS pc (comml (_call f))) p'
  -> (forall pc' : nat, pimp (p |> wrap_fun g' pc f pc') p'')
  -> (forall pc' : nat, gimp (p >o wrap_fun g' pc f pc' oo g'') g)
  -> itp_NH_theta rho (Ast p g tag) Psi S0
  -> itp_NH_theta rho (Ast p' g' 1) Psi S'.
Proof.
  intros rho Psi f pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p' p'' g g' g'' tag.
  intros Hlk' Hlk'' HnextS Hen Hfun Hpp' Hg'p'' Hg''g H.
  unfold itp_NH_theta, NH_pinvstar, pmstar in * .
  destruct H as [Hisrf [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [_ [HR [Hfl Hisr]]].
  subst R' fl' isr'.
  unfold enable in Hen.
  destruct (Hen _ HMnh) as [Sx HnextSSx].
  destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
  clear Sx HnextSSx Hen.
  simplS in * .
  split_l.
  assert (HS' : (p |> nextS pc (comml (_call f))) (Mnh', RF_update R SP (R SP - 2), fl, isr)).
    apply linkPG_intro with (Mnh,R,fl,isr); trivial.
  split.
    exists Mnh'; exists Mih; split_l.
    split_r.
    apply Hpp'; trivial.
  set (S := (M, R, fl, isr)) in * .
  set (S' := (M', RF_update R SP (R SP - 2), fl, isr)) in * .
  set (Snh := (Mnh, R, fl, isr)) in * .
  set (Snh' := (Mnh', RF_update R SP (R SP - 2), fl, isr)) in * .
  intros Htag; clear Htag.
  destruct (eq_nat_dec tag 1) as [Htag | Htag].
  (* tag = 1 *)
  substH Hwfst with (Hwfst Htag); clear Htag.
  destruct Hwfst as [n Hwfst].
  exists (1+n).
  simpl.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold NH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  unfold S' in Hg'.
  simplS in Hg'.
  destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih (NHInv_prec fl)) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  unfold wrap_fun in Hg'p''.
  simplS in * .
  set (S'' := (M'', R'', fl'', isr'')) in * .
  set (Snh'' := (Mnh'', R'', fl'', isr'')).
  assert (Hfunframe : fun_frame Snh' Snh'').
    apply Hfun.
    unfold sp_call_pred.
    apply conjPG_intro; trivial.
  unfold fun_frame in Hfunframe.
  destruct Hfunframe as [_ [ra [Hra' Hra'']]].
  assert (Hrax := ra_after_call _ _ Hra' HnextSnh').
  subst ra.
  exists (pc + 1).
  unfold S''.
  split.
    unfold S''.
    apply (Ret_Addr_Mono HJM'' Hra'').
  clear Hra'.
  exists p''; exists g''; exists tag.
  split_l.
  intros pc' S''' HnextS''.
  substH Hg'p'' with (Hg'p'' pc').
  assert (HnextSnh''Sx : nextS pc' (comms _retw) (Mnh'', R'', fl'', isr'') (Mnh'', RF_update R'' SP (R'' SP + 2), fl'', isr'')).
    unfold Ret_Addr in Hra''.
    substH Hra'' with (Hra'' pc').
    inversion_clear Hra''.
    subst Snh'' sp; simplS in * .
    apply nextS_retw.
    exists (pc+1).
    unfold popW.
    split_lr.
  destruct S''' as [[[M''' R'''] fl'''] isr'''].
  destruct (nextS_DMono HJM'' HnextSnh''Sx HnextS'') as [Mnh''' [HJM''' HnextSnh'']].
  destruct (nextS_inversion HnextSnh'') as [HM [_ [Hfl Hisr]]].
  subst Mnh''' fl''' isr'''.
  destruct (nextS_inversion HnextS'') as [HM _].
  subst M'''.
  split.
    unfold NH_pinvstar, pmstar.
    exists Mnh''; exists Mih''; split_l.
    split_r by (apply HMih'').
    apply Hg'p''.
    apply linkPG_intro with Snh; trivial.
    apply conjG_intro with Snh''; trivial.
    apply conjG_intro with Snh'; trivial.
  apply (NH_WFST_aux_strengthen (S1 := (M'', R''', fl'', isr''))(g':= g'') Hwfst).
  substH Hg''g with (Hg''g pc' Snh).
  unfold NH_ItpG.
  intros Sx H.
  destruct H as [Mnh''0 [Mih''0 [Mnhx [Mihx [HJM''0 [HJMx [HMih''0 [HMihx Hg'']]]]]]]].
  destruct (prec_disjUnion_inv HJM'' HJM''0 HMih'' HMih''0 (NHInv_prec fl'')) as [Heq1 Heq2].
  subst Mnh''0 Mih''0.
  clear HJM''0 HMih''0.
  exists Mnh; exists Mih; exists Mnhx; exists Mihx.
  split_l; split_l; split_l; split_l.
  apply Hg''g.
  apply conjG_intro with (Mnh'', R''', fl'', isr''); trivial.
  unfold wrap_fun.
  apply conjPG_intro; auto.
  apply conjG_intro with Snh''; subst S; trivial.
  apply conjG_intro with Snh'; trivial.
  (* tag <> 1 *)
  exists 0.
  simpl.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold NH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  unfold S' in Hg'.
  simplS in Hg'.
  destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih (NHInv_prec fl)) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  unfold wrap_fun in Hg'p''.
  simplS in * .
  set (S'' := (M'', R'', fl'', isr'')) in * .
  set (Snh'' := (Mnh'', R'', fl'', isr'')).
  assert (Hfunframe : fun_frame Snh' Snh'').
    apply Hfun.
    unfold sp_call_pred.
    apply conjPG_intro; trivial.
  unfold fun_frame in Hfunframe.
  destruct Hfunframe as [_ [ra [Hra' Hra'']]].
  assert (Hrax := ra_after_call _ _ Hra' HnextSnh').
  subst ra.
  exists (pc + 1).
  unfold S''.
  split.
    unfold S''.
    apply (Ret_Addr_Mono HJM'' Hra'').
  clear Hra'.
  exists p''; exists g''; exists tag.
  split_l.
  split_l.
  intros pc' S''' HnextS''.
  substH Hg'p'' with (Hg'p'' pc').
  assert (HnextSnh''Sx : nextS pc' (comms _retw) (Mnh'', R'', fl'', isr'') (Mnh'', RF_update R'' SP (R'' SP + 2), fl'', isr'')).
    unfold Ret_Addr in Hra''.
    substH Hra'' with (Hra'' pc').
    inversion_clear Hra''.
    subst Snh'' sp; simplS in * .
    apply nextS_retw.
    exists (pc+1).
    unfold popW.
    split_lr.
  destruct S''' as [[[M''' R'''] fl'''] isr'''].
  destruct (nextS_DMono HJM'' HnextSnh''Sx HnextS'') as [Mnh''' [HJM''' HnextSnh'']].
  destruct (nextS_inversion HnextSnh'') as [HM [_ [Hfl Hisr]]].
  subst Mnh''' fl''' isr'''.
  destruct (nextS_inversion HnextS'') as [HM _].
  subst M'''.
  unfold NH_pinvstar, pmstar.
  exists Mnh''; exists Mih''; split_l.
  split_r by (apply HMih'').
  apply Hg'p''.
  apply linkPG_intro with Snh; trivial.
  apply conjG_intro with Snh''; trivial.
  apply conjG_intro with Snh'; trivial.
Qed.
Implicit Arguments NH_nonprim_call_prev [rho Psi f pc S0 S' p p' p'' g g' g'' tag].

Lemma NH_jne_prog : forall rho Psi f pc S0
  (p : Pred) (g : Grt) tag,
  itp_NH_theta rho (Ast p g tag) Psi S0
  -> exists S', nextS pc (comml (_jne f)) S0 S'.
Proof.
  intros rho Psi f pc [[[M R] fl] isr] p g tag.
  intros H.
  exists (M, R, fl, isr).
  constructor.
Qed.
Implicit Arguments NH_jne_prog [rho Psi f pc S0 p g tag].

Lemma NH_jne_jmp_prev : forall rho Psi f pc S0 S'
  (p p': Pred) (g g': Grt) tag,
  lookupCT Psi f (packS0 rho (Ast p' g' tag))
  -> S_Flag S0 _ZF = false
  -> nextS pc (comml (_jne f)) S0 S'
  -> pimp (conjP p ZF_cl) p'
  -> gimp (p >o (ZF_cl >o g')) g
  -> itp_NH_theta rho (Ast p g tag) Psi S0
  -> itp_NH_theta rho (Ast p' g' tag) Psi S'.
Proof.
  intros rho Psi f pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p' g g' tag.
  intros Hlk' Hflzf HnextS Hpp' Hg'g H.
  unfold itp_NH_theta, NH_pinvstar, pmstar in * .
  destruct H as [Hisrf [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  subst M' R' fl' isr'.
  assert (HS' : (p |> nextS pc (comml (_jne f))) (Mnh, R, fl, isr)).
    apply linkPG_intro with (Mnh,R,fl,isr); trivial.
    constructor.
  split_l.
  split.
    exists Mnh; exists Mih.
    split_l.
    split_r.
    simplS.
    apply Hpp'; trivial.
    split; trivial.
  intros Htag. 
  substH Hwfst with (Hwfst Htag); clear Htag.
  apply NH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, NH_ItpG.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold NH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  simplS in Hg'.
  destruct (prec_disjUnion_inv HJMxy HJM HMy HMih (NHInv_prec fl)) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  simplS.
  exists Mnh; exists Mih.
  exists Mnh''; exists Mih''.
  split_l; split_l; split_l; split_l.
  apply Hg'g.
  apply conjPG_intro; trivial.
  apply conjPG_intro; trivial.
Qed.
Implicit Arguments NH_jne_jmp_prev [rho Psi f pc S0 S' p p' g g' tag].

Lemma NH_jne_seq_prev : forall rho Psi f pc S0 S'
  (p p'': Pred) (g g'': Grt) tag,
  lookupCT Psi (pc+1) (packS0 rho (Ast p'' g'' tag))
  -> S_Flag S0 _ZF = true
  -> nextS pc (comml (_jne f)) S0 S'
  -> pimp (conjP p ZF_st) p''
  -> gimp (p >o (ZF_st >o g'')) g
  -> itp_NH_theta rho (Ast p g tag) Psi S0
  -> itp_NH_theta rho (Ast p'' g'' tag) Psi S'.
Proof.
  intros rho Psi f pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p'' g g'' tag.
  intros Hlk'' Hflzf HnextS Hpp' Hg'g H.
  unfold itp_NH_theta, NH_pinvstar, pmstar in * .
  destruct H as [Hisrf [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  subst M' R' fl' isr'.
  assert (HS' : (p |> nextS pc (comml (_jne f))) (Mnh, R, fl, isr)).
    apply linkPG_intro with (Mnh,R,fl,isr); trivial.
    constructor.
  split_l.
  split.
    exists Mnh; exists Mih.
    split_l.
    split_r.
    simplS.
    apply Hpp'; trivial.
    split; trivial.
  intros Htag. 
  substH Hwfst with (Hwfst Htag); clear Htag.
  apply NH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, NH_ItpG.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold NH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  simplS in Hg'.
  destruct (prec_disjUnion_inv HJMxy HJM HMy HMih (NHInv_prec fl)) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  simplS.
  exists Mnh; exists Mih.
  exists Mnh''; exists Mih''.
  split_l; split_l; split_l; split_l.
  apply Hg'g.
  apply conjPG_intro; trivial.
  apply conjPG_intro; trivial.
Qed.
Implicit Arguments NH_jne_seq_prev [rho Psi f pc S0 S' p p'' g g'' tag].

Lemma NH_movwp_prog : forall rho Psi o r pc S0
  (p : Pred) (g : Grt) tag,
  enable p (nextS pc (comml (_movwp o r)))
  -> itp_NH_theta rho (Ast p g tag) Psi S0
  -> exists S', nextS pc (comml (_movwp o r)) S0 S'.
Proof.
  intros rho Psi o r pc [[[M R] fl] isr] p g tag.
  intros Hen H.
  unfold itp_NH_theta, NH_pinvstar, pmstar in * .
  destruct H as [Hisrf [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  simplS in * .
  destruct (Hen _ HMnh) as [S' HnextSnh]. 
  destruct S' as [[[Mnh' R'] fl'] isr'].
  destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
  exists (M', R', fl', isr'); trivial.
Qed.
Implicit Arguments NH_movwp_prog [rho Psi o r pc S0 p g tag].

Lemma NH_movwp_prev : forall rho Psi o r pc S0 S'
  (p p' : Pred) (g g' : Grt) tag,
  lookupCT Psi (pc+1) (packS0 rho (Ast p' g' tag))
  -> nextS pc (comml (_movwp o r)) S0 S'
  -> enable p (nextS pc (comml (_movwp o r)))
  -> pimp (p |> nextS pc (comml (_movwp o r))) p'
  -> gimp (p >o nextS pc (comml (_movwp o r)) oo g') g
  -> itp_NH_theta rho (Ast p g tag) Psi S0
  -> itp_NH_theta rho (Ast p' g' tag) Psi S'.
Proof.
  intros rho Psi o r pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p' g g' tag.
  intros Hlk' HnextS Hen Hpp' Hg'g H.
  unfold itp_NH_theta, NH_pinvstar, pmstar in * .
  destruct H as [Hisrf [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [_ [HR [Hfl Hisr]]].
  subst fl' isr'.
  unfold enable in Hen.
  destruct (Hen _ HMnh) as [Sx HnextSSx].
  destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
  clear Sx HnextSSx Hen.
  simplS in * .
  split_l.
  assert (HS' : (p |> nextS pc (comml (_movwp o r))) (Mnh', R', fl, isr)).
    apply linkPG_intro with (Mnh,R,fl,isr); trivial.
  split.
    exists Mnh'; exists Mih; split_l.
    split_r.
    apply Hpp'; trivial.
  intros Htag.
  substH Hwfst with (Hwfst Htag); clear Htag.
  apply NH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, NH_ItpG.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold NH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  simplS in * .
  destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih (NHInv_prec fl)) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  simplS.
  exists Mnh; exists Mih.
  exists Mnh''; exists Mih''.
  split_l; split_l; split_l; split_l.
  apply Hg'g.
  apply conjG_intro with (Mnh', R', fl, isr); trivial.
  apply conjPG_intro; trivial.
Qed.
Implicit Arguments NH_movwp_prev [rho Psi o r pc S0 S' p p' g g' tag].

Lemma NH_movwld_prog : forall rho Psi a r pc S0
  (p : Pred) (g : Grt) tag,
  enable p (nextS pc (comml (_movwld a r)))
  -> itp_NH_theta rho (Ast p g tag) Psi S0
  -> exists S', nextS pc (comml (_movwld a r)) S0 S'.
Proof.
  intros rho Psi a r pc [[[M R] fl] isr] p g tag.
  intros Hen H.
  unfold itp_NH_theta, NH_pinvstar, pmstar in * .
  destruct H as [Hisrf [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  simplS in * .
  destruct (Hen _ HMnh) as [S' HnextSnh]. 
  destruct S' as [[[Mnh' R'] fl'] isr'].
  destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
  exists (M', R', fl', isr'); trivial.
Qed.
Implicit Arguments NH_movwld_prog [rho Psi a r pc S0 p g tag].

Lemma NH_movwld_prev : forall rho Psi a r pc S0 S'
  (p p' : Pred) (g g' : Grt) tag,
  lookupCT Psi (pc+1) (packS0 rho (Ast p' g' tag))
  -> nextS pc (comml (_movwld a r)) S0 S'
  -> enable p (nextS pc (comml (_movwld a r)))
  -> pimp (p |> nextS pc (comml (_movwld a r))) p'
  -> gimp (p >o nextS pc (comml (_movwld a r)) oo g') g
  -> itp_NH_theta rho (Ast p g tag) Psi S0
  -> itp_NH_theta rho (Ast p' g' tag) Psi S'.
Proof.
  intros rho Psi a r pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p' g g' tag.
  intros Hlk' HnextS Hen Hpp' Hg'g H.
  unfold itp_NH_theta, NH_pinvstar, pmstar in * .
  destruct H as [Hisrf [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [_ [HR [Hfl Hisr]]].
  subst fl' isr'.
  unfold enable in Hen.
  destruct (Hen _ HMnh) as [Sx HnextSSx].
  destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
  clear Sx HnextSSx Hen.
  simplS in * .
  split_l.
  assert (HS' : (p |> nextS pc (comml (_movwld a r))) (Mnh', R', fl, isr)).
    apply linkPG_intro with (Mnh,R,fl,isr); trivial.
  split.
    exists Mnh'; exists Mih; split_l.
    split_r.
    apply Hpp'; trivial.
  intros Htag.
  substH Hwfst with (Hwfst Htag); clear Htag.
  apply NH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, NH_ItpG.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold NH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  simplS in * .
  destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih (NHInv_prec fl)) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  simplS.
  exists Mnh; exists Mih.
  exists Mnh''; exists Mih''.
  split_l; split_l; split_l; split_l.
  apply Hg'g.
  apply conjG_intro with (Mnh', R', fl, isr); trivial.
  apply conjPG_intro; trivial.
Qed.
Implicit Arguments NH_movwld_prev [rho Psi a r pc S0 S' p p' g g' tag].

Lemma NH_movwst_prog : forall rho Psi a r pc S0
  (p : Pred) (g : Grt) tag,
  enable p (nextS pc (comml (_movwst a r)))
  -> itp_NH_theta rho (Ast p g tag) Psi S0
  -> exists S', nextS pc (comml (_movwst a r)) S0 S'.
Proof.
  intros rho Psi a r pc [[[M R] fl] isr] p g tag.
  intros Hen H.
  unfold itp_NH_theta, NH_pinvstar, pmstar in * .
  destruct H as [Hisrf [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  simplS in * .
  destruct (Hen _ HMnh) as [S' HnextSnh]. 
  destruct S' as [[[Mnh' R'] fl'] isr'].
  destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
  exists (M', R', fl', isr'); trivial.
Qed.
Implicit Arguments NH_movwst_prog [rho Psi a r pc S0 p g tag].

Lemma NH_movwst_prev : forall rho Psi a r pc S0 S'
  (p p' : Pred) (g g' : Grt) tag,
  lookupCT Psi (pc+1) (packS0 rho (Ast p' g' tag))
  -> nextS pc (comml (_movwst a r)) S0 S'
  -> enable p (nextS pc (comml (_movwst a r)))
  -> pimp (p |> nextS pc (comml (_movwst a r))) p'
  -> gimp (p >o nextS pc (comml (_movwst a r)) oo g') g
  -> itp_NH_theta rho (Ast p g tag) Psi S0
  -> itp_NH_theta rho (Ast p' g' tag) Psi S'.
Proof.
  intros rho Psi a r pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p' g g' tag.
  intros Hlk' HnextS Hen Hpp' Hg'g H.
  unfold itp_NH_theta, NH_pinvstar, pmstar in * .
  destruct H as [Hisrf [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [_ [HR [Hfl Hisr]]].
  subst fl' isr'.
  unfold enable in Hen.
  destruct (Hen _ HMnh) as [Sx HnextSSx].
  destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
  clear Sx HnextSSx Hen.
  simplS in * .
  split_l.
  assert (HS' : (p |> nextS pc (comml (_movwst a r))) (Mnh', R', fl, isr)).
    apply linkPG_intro with (Mnh,R,fl,isr); trivial.
  split.
    exists Mnh'; exists Mih; split_l.
    split_r.
    apply Hpp'; trivial.
  intros Htag.
  substH Hwfst with (Hwfst Htag); clear Htag.
  apply NH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, NH_ItpG.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold NH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  simplS in * .
  destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih (NHInv_prec fl)) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  simplS.
  exists Mnh; exists Mih.
  exists Mnh''; exists Mih''.
  split_l; split_l; split_l; split_l.
  apply Hg'g.
  apply conjG_intro with (Mnh', R', fl, isr); trivial.
  apply conjPG_intro; trivial.
Qed.
Implicit Arguments NH_movwst_prev [rho Psi a r pc S0 S' p p' g g' tag].

Lemma NH_pushwp_prog : forall rho Psi o pc S0
  (p : Pred) (g : Grt) tag,
  enable p (nextS pc (comml (_pushwp o)))
  -> itp_NH_theta rho (Ast p g tag) Psi S0
  -> exists S', nextS pc (comml (_pushwp o)) S0 S'.
Proof.
  intros rho Psi o pc [[[M R] fl] isr] p g tag.
  intros Hen H.
  unfold itp_NH_theta, NH_pinvstar, pmstar in * .
  destruct H as [Hisrf [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  simplS in * .
  destruct (Hen _ HMnh) as [S' HnextSnh]. 
  destruct S' as [[[Mnh' R'] fl'] isr'].
  destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
  exists (M', R', fl', isr'); trivial.
Qed.
Implicit Arguments NH_pushwp_prog [rho Psi o pc S0 p g tag].

Lemma NH_pushwp_prev : forall rho Psi o pc S0 S'
  (p p' : Pred) (g g' : Grt) tag,
  lookupCT Psi (pc+1) (packS0 rho (Ast p' g' tag))
  -> nextS pc (comml (_pushwp o)) S0 S'
  -> enable p (nextS pc (comml (_pushwp o)))
  -> pimp (p |> nextS pc (comml (_pushwp o))) p'
  -> gimp (p >o nextS pc (comml (_pushwp o)) oo g') g
  -> itp_NH_theta rho (Ast p g tag) Psi S0
  -> itp_NH_theta rho (Ast p' g' tag) Psi S'.
Proof.
  intros rho Psi o pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p' g g' tag.
  intros Hlk' HnextS Hen Hpp' Hg'g H.
  unfold itp_NH_theta, NH_pinvstar, pmstar in * .
  destruct H as [Hisrf [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [_ [HR [Hfl Hisr]]].
  subst fl' isr'.
  unfold enable in Hen.
  destruct (Hen _ HMnh) as [Sx HnextSSx].
  destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
  clear Sx HnextSSx Hen.
  simplS in * .
  split_l.
  assert (HS' : (p |> nextS pc (comml (_pushwp o))) (Mnh', R', fl, isr)).
    apply linkPG_intro with (Mnh,R,fl,isr); trivial.
  split.
    exists Mnh'; exists Mih; split_l.
    split_r.
    apply Hpp'; trivial.
  intros Htag.
  substH Hwfst with (Hwfst Htag); clear Htag.
  apply NH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, NH_ItpG.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold NH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  simplS in * .
  destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih (NHInv_prec fl)) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  simplS.
  exists Mnh; exists Mih.
  exists Mnh''; exists Mih''.
  split_l; split_l; split_l; split_l.
  apply Hg'g.
  apply conjG_intro with (Mnh', R', fl, isr); trivial.
  apply conjPG_intro; trivial.
Qed.
Implicit Arguments NH_pushwp_prev [rho Psi o pc S0 S' p p' g g' tag].

Lemma NH_pushwm_prog : forall rho Psi a pc S0
  (p : Pred) (g : Grt) tag,
  enable p (nextS pc (comml (_pushwm a)))
  -> itp_NH_theta rho (Ast p g tag) Psi S0
  -> exists S', nextS pc (comml (_pushwm a)) S0 S'.
Proof.
  intros rho Psi a pc [[[M R] fl] isr] p g tag.
  intros Hen H.
  unfold itp_NH_theta, NH_pinvstar, pmstar in * .
  destruct H as [Hisrf [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  simplS in * .
  destruct (Hen _ HMnh) as [S' HnextSnh]. 
  destruct S' as [[[Mnh' R'] fl'] isr'].
  destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
  exists (M', R', fl', isr'); trivial.
Qed.
Implicit Arguments NH_pushwm_prog [rho Psi a pc S0 p g tag].

Lemma NH_pushwm_prev : forall rho Psi a pc S0 S'
  (p p' : Pred) (g g' : Grt) tag,
  lookupCT Psi (pc+1) (packS0 rho (Ast p' g' tag))
  -> nextS pc (comml (_pushwm a)) S0 S'
  -> enable p (nextS pc (comml (_pushwm a)))
  -> pimp (p |> nextS pc (comml (_pushwm a))) p'
  -> gimp (p >o nextS pc (comml (_pushwm a)) oo g') g
  -> itp_NH_theta rho (Ast p g tag) Psi S0
  -> itp_NH_theta rho (Ast p' g' tag) Psi S'.
Proof.
  intros rho Psi a pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p' g g' tag.
  intros Hlk' HnextS Hen Hpp' Hg'g H.
  unfold itp_NH_theta, NH_pinvstar, pmstar in * .
  destruct H as [Hisrf [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [_ [HR [Hfl Hisr]]].
  subst fl' isr'.
  unfold enable in Hen.
  destruct (Hen _ HMnh) as [Sx HnextSSx].
  destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
  clear Sx HnextSSx Hen.
  simplS in * .
  split_l.
  assert (HS' : (p |> nextS pc (comml (_pushwm a))) (Mnh', R', fl, isr)).
    apply linkPG_intro with (Mnh,R,fl,isr); trivial.
  split.
    exists Mnh'; exists Mih; split_l.
    split_r.
    apply Hpp'; trivial.
  intros Htag.
  substH Hwfst with (Hwfst Htag); clear Htag.
  apply NH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, NH_ItpG.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold NH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  simplS in * .
  destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih (NHInv_prec fl)) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  simplS.
  exists Mnh; exists Mih.
  exists Mnh''; exists Mih''.
  split_l; split_l; split_l; split_l.
  apply Hg'g.
  apply conjG_intro with (Mnh', R', fl, isr); trivial.
  apply conjPG_intro; trivial.
Qed.
Implicit Arguments NH_pushwm_prev [rho Psi a pc S0 S' p p' g g' tag].

Lemma NH_pushfw_prog : forall rho Psi  pc S0
  (p : Pred) (g : Grt) tag,
  enable p (nextS pc (comml (_pushfw )))
  -> itp_NH_theta rho (Ast p g tag) Psi S0
  -> exists S', nextS pc (comml (_pushfw )) S0 S'.
Proof.
  intros rho Psi  pc [[[M R] fl] isr] p g tag.
  intros Hen H.
  unfold itp_NH_theta, NH_pinvstar, pmstar in * .
  destruct H as [Hisrf [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  simplS in * .
  destruct (Hen _ HMnh) as [S' HnextSnh]. 
  destruct S' as [[[Mnh' R'] fl'] isr'].
  destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
  exists (M', R', fl', isr'); trivial.
Qed.
Implicit Arguments NH_pushfw_prog [rho Psi pc S0 p g tag].

Lemma NH_pushfw_prev : forall rho Psi  pc S0 S'
  (p p' : Pred) (g g' : Grt) tag,
  lookupCT Psi (pc+1) (packS0 rho (Ast p' g' tag))
  -> nextS pc (comml (_pushfw )) S0 S'
  -> enable p (nextS pc (comml (_pushfw )))
  -> pimp (p |> nextS pc (comml (_pushfw ))) p'
  -> gimp (p >o nextS pc (comml (_pushfw )) oo g') g
  -> itp_NH_theta rho (Ast p g tag) Psi S0
  -> itp_NH_theta rho (Ast p' g' tag) Psi S'.
Proof.
  intros rho Psi  pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p' g g' tag.
  intros Hlk' HnextS Hen Hpp' Hg'g H.
  unfold itp_NH_theta, NH_pinvstar, pmstar in * .
  destruct H as [Hisrf [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [_ [HR [Hfl Hisr]]].
  subst fl' isr'.
  unfold enable in Hen.
  destruct (Hen _ HMnh) as [Sx HnextSSx].
  destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
  clear Sx HnextSSx Hen.
  simplS in * .
  split_l.
  assert (HS' : (p |> nextS pc (comml (_pushfw ))) (Mnh', R', fl, isr)).
    apply linkPG_intro with (Mnh,R,fl,isr); trivial.
  split.
    exists Mnh'; exists Mih; split_l.
    split_r.
    apply Hpp'; trivial.
  intros Htag.
  substH Hwfst with (Hwfst Htag); clear Htag.
  apply NH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, NH_ItpG.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold NH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  simplS in * .
  destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih (NHInv_prec fl)) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  simplS.
  exists Mnh; exists Mih.
  exists Mnh''; exists Mih''.
  split_l; split_l; split_l; split_l.
  apply Hg'g.
  apply conjG_intro with (Mnh', R', fl, isr); trivial.
  apply conjPG_intro; trivial.
Qed.
Implicit Arguments NH_pushfw_prev [rho Psi pc S0 S' p p' g g' tag].

Lemma NH_popw_prog : forall rho Psi r pc S0
  (p : Pred) (g : Grt) tag,
  enable p (nextS pc (comml (_popw r)))
  -> itp_NH_theta rho (Ast p g tag) Psi S0
  -> exists S', nextS pc (comml (_popw r)) S0 S'.
Proof.
  intros rho Psi r pc [[[M R] fl] isr] p g tag.
  intros Hen H.
  unfold itp_NH_theta, NH_pinvstar, pmstar in * .
  destruct H as [Hisrf [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  simplS in * .
  destruct (Hen _ HMnh) as [S' HnextSnh]. 
  destruct S' as [[[Mnh' R'] fl'] isr'].
  destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
  exists (M', R', fl', isr'); trivial.
Qed.
Implicit Arguments NH_popw_prog [rho Psi r pc S0 p g tag].

Lemma NH_popw_prev : forall rho Psi r pc S0 S'
  (p p' : Pred) (g g' : Grt) tag,
  lookupCT Psi (pc+1) (packS0 rho (Ast p' g' tag))
  -> nextS pc (comml (_popw r)) S0 S'
  -> enable p (nextS pc (comml (_popw r)))
  -> pimp (p |> nextS pc (comml (_popw r))) p'
  -> gimp (p >o nextS pc (comml (_popw r)) oo g') g
  -> itp_NH_theta rho (Ast p g tag) Psi S0
  -> itp_NH_theta rho (Ast p' g' tag) Psi S'.
Proof.
  intros rho Psi r pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p' g g' tag.
  intros Hlk' HnextS Hen Hpp' Hg'g H.
  unfold itp_NH_theta, NH_pinvstar, pmstar in * .
  destruct H as [Hisrf [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [_ [HR [Hfl Hisr]]].
  subst fl' isr'.
  unfold enable in Hen.
  destruct (Hen _ HMnh) as [Sx HnextSSx].
  destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
  clear Sx HnextSSx Hen.
  simplS in * .
  split_l.
  assert (HS' : (p |> nextS pc (comml (_popw r))) (Mnh', R', fl, isr)).
    apply linkPG_intro with (Mnh,R,fl,isr); trivial.
  split.
    exists Mnh'; exists Mih; split_l.
    split_r.
    apply Hpp'; trivial.
  intros Htag.
  substH Hwfst with (Hwfst Htag); clear Htag.
  apply NH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, NH_ItpG.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold NH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  simplS in * .
  destruct (prec_disjUnion_inv HJMxy HJM' HMy HMih (NHInv_prec fl)) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  simplS.
  exists Mnh; exists Mih.
  exists Mnh''; exists Mih''.
  split_l; split_l; split_l; split_l.
  apply Hg'g.
  apply conjG_intro with (Mnh', R', fl, isr); trivial.
  apply conjPG_intro; trivial.
Qed.
Implicit Arguments NH_popw_prev [rho Psi r pc S0 S' p p' g g' tag].

Lemma NH_popfw_prog : forall rho Psi pc S0
  (p : Pred) (g : Grt) tag,
  enable p (nextS pc (comml (_popfw)))
  -> itp_NH_theta rho (Ast p g tag) Psi S0
  -> exists S', nextS pc (comml (_popfw)) S0 S'.
Proof.
  intros rho Psi pc [[[M R] fl] isr] p g tag.
  intros Hen H.
  unfold itp_NH_theta, NH_pinvstar, pmstar in * .
  destruct H as [Hisrf [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  simplS in * .
  destruct (Hen _ HMnh) as [S' HnextSnh]. 
  destruct S' as [[[Mnh' R'] fl'] isr'].
  destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
  exists (M', R', fl', isr'); trivial.
Qed.
Implicit Arguments NH_popfw_prog [rho Psi pc S0 p g tag].

Lemma NH_popfw_prev : forall rho Psi pc S0 S'
  (p p' : Pred) (g g' : Grt) tag,
  lookupCT Psi (pc+1) (packS0 rho (Ast p' g' tag))
  -> nextS pc (comml (_popfw)) S0 S'
  -> enable p (nextS pc (comml (_popfw)))
  -> enable_popfw_sti (conjP p stk_IF_st)
  -> pimp ((conjP p stk_IF_st) |> gpopfw_sti) p'
  -> gimp ((conjP p stk_IF_st) >o gpopfw_sti oo g') g
  -> pimp ((conjP p stk_IF_cl) |> gpopfw_cli) p'
  -> gimp ((conjP p stk_IF_cl) >o gpopfw_cli oo g') g
  -> itp_NH_theta rho (Ast p g tag) Psi S0
  -> itp_NH_theta rho (Ast p' g' tag) Psi S'.
Proof.
  intros rho Psi pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p' g g' tag.
  intros Hlk' HnextS Hen Hsti Hstipp' Hstig'g Hclipp' Hclig'g H.
  destruct H as [Hisrf [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [HM' [HR' [Hfl' Hisr']]].
  destruct Hfl' as [flnum [Hlds Hfl']].
  subst M' fl' isr'.
  unfold enable in Hen.
  destruct (Hen _ HMnh) as [Sx HnextSSx].
  destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
  destruct (nextS_inversion HnextSnh') as [HMnh' _].
  subst Mnh'; clear Sx HnextSSx Hen HJM'.
  simplS in * .
  split_l.
  assert (HS' : (p |> nextS pc (comml (_popfw))) (Mnh, R', nat2FR flnum, isr)).
    apply linkPG_intro with (Mnh,R,fl,isr); trivial.
  unfold NH_pinvstar, pmstar.
  simplS.
  generalize (refl_equal (nat2FR flnum _IF)); pattern flnum at 1; destruct (nat2FR flnum _IF); intro Hfln;
    (generalize (refl_equal (fl _IF)); pattern fl at 1; destruct (fl _IF); intro Hfl).
  (* 1 -> 1 *)
  unfold NH_Inv in * .
  rewrite Hfl in HMih.
  rewrite Hfln.
  assert (Hgpopfw_sti : gpopfw_sti (Mnh, R, fl, isr)
     (Mnh, R', nat2FR flnum, isr)).
    unfold gpopfw_sti.
    simpl.
    unfold prepost, choice, itr_enabled.
    simplS.
    rewrite Hfl.
    rewrite Hfln.
    rsplit.
      auto.
    exists MemMod.emp; exists Mnh; exists MemMod.emp.
    rsplit.
      unfold Aemp; auto with PMap.
    left; unfold Aemp; auto.
  split.
    exists Mnh; exists Mih.
    split_l.
    split_r.
    apply Hstipp'; trivial.
    apply linkPG_intro with (Mnh, R, fl, isr); trivial.
    split; auto.
    destruct (nextS_inversion HnextSnh') as [_ [_ [Hflx _]]].
    destruct Hflx as [x [Hx Hy]].
    exists x; simplS.
    split_l.
    rewrite <- Hy; trivial.
  intros Htag.
  substH Hwfst with (Hwfst Htag); clear Htag.
  apply NH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, NH_ItpG.
  intros [[[Mx Rx] flx] isrx] [Mx1 [Mx2 [Mx1' [Mx2' Hg]]]].
  simplS; simplS in Hg.
  destruct Hg as [Hg1 [Hg2 [Hg3 [Hg4 Hg5]]]].
  unfold NH_Inv in Hg3.
  rewrite Hfln in Hg3.
  destruct (prec_disjUnion_inv HJM Hg1 HMih Hg3 (Prec01)) as [Heq1 Heq2].
  subst Mx1 Mx2.
  exists Mnh; exists Mih.
  exists Mx1'; exists Mx2'.
  split_l.
  split_l.
  unfold NH_Inv; rewrite Hfl.
  split_l.
  split_l.
  apply Hstig'g.
  apply conjG_intro with (Mnh, R', nat2FR flnum, isr); trivial.
  apply conjPG_intro; trivial.
  split; auto.
  destruct (nextS_inversion HnextSnh') as [_ [_ [Hflx _]]].
  destruct Hflx as [x [Hx Hy]].
  exists x; simplS.
  split_l.
  rewrite <- Hy; trivial.
  (* 0 -> 1 *)
  unfold NH_Inv, Aemp in HMih. 
  rewrite Hfl in HMih. 
  subst Mih.
  assert (Mnh = M).
    auto with PMap.
  subst Mnh.
  clear HJM.
  unfold enable_popfw_sti, pmstar, pTrue, choice in Hsti.
  unfold conjP, stk_IF_st in Hsti.
  simplS in Hsti.
  assert (Hflnex : exists flnum : nat,
            leadsToW (R SP) flnum M /\
            nat2FR flnum _IF = true).
    destruct (nextS_inversion HnextSnh') as [_ [_ [Hflx _]]].
    destruct Hflx as [flnumx [Hldsx Heq]].
    exists flnumx; split; auto.
    rewrite <- Heq; trivial.
  substH Hsti with (Hsti _ _ _ _ (conj HMnh Hflnex)).
  destruct Hsti as [Mnh [Mih [HJM [Hlds' Hsti]]]].
  rewrite Hfl in Hsti.
  destruct Hsti as [ [H1 H2] | [H1 H2]]; [discriminate H1 | clear H1].
  assert (Hgpopfw_sti : gpopfw_sti (M, R, fl, isr)
    (Mnh, R', nat2FR flnum, isr)).
    unfold gpopfw_sti.
    unfold prepost, choice, itr_enabled.
    simpl.
    rewrite Hfl.
    rewrite Hfln.
    rsplit.
      auto.
    exists Mih; exists Mnh; exists MemMod.emp.
    rsplit.
      unfold Aemp; auto with PMap.
    right; split_r.
    intros [Hx Hy]; discriminate Hx.
  split.
    exists Mnh; exists Mih.
    split_l.
    unfold NH_Inv. rewrite Hfln.
    split_r.
    apply Hstipp'; trivial.
    apply linkPG_intro with (M, R, fl, isr); trivial.
    split; auto.
  intros Htag.
  substH Hwfst with (Hwfst Htag); clear Htag.
  apply NH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, NH_ItpG.
  intros [[[Mx Rx] flx] isrx] [Mx1 [Mx2 [Mx1' [Mx2' Hg]]]].
  simplS; simplS in Hg.
  destruct Hg as [Hg1 [Hg2 [Hg3 [Hg4 Hg5]]]].
  unfold NH_Inv in Hg3.
  rewrite Hfln in Hg3.
  destruct (prec_disjUnion_inv HJM Hg1 H2 Hg3 (Prec01)) as [Heq1 Heq2].
  subst Mx1 Mx2.
  exists M; exists MemMod.emp.
  exists Mx1'; exists Mx2'.
  split_l by auto with PMap.
  split_l.
  unfold NH_Inv; rewrite Hfl.
  split.
    unfold Aemp; trivial.
  split_l.
  apply Hstig'g.
  apply conjG_intro with (Mnh, R', nat2FR flnum, isr); trivial.
  apply conjPG_intro; trivial.
  split; auto.
  (* 1 -> 0 *)
  unfold NH_Inv in * .
  rewrite Hfl in HMih.
  rewrite Hfln.
  assert (Hgpopfw_cli : gpopfw_cli (Mnh, R, fl, isr)
     (M, R', nat2FR flnum, isr)).
    unfold gpopfw_cli.
    simplS.
    rsplit.              
      auto.
    unfold prepost, choice, itr_enabled.
    simplS.
    rewrite Hfl.
    exists MemMod.emp; exists Mnh; exists Mih.
    unfold Aemp.
    split_l.
    rsplit.
      auto with PMap.
    left; auto.
  split.
    exists M; exists MemMod.emp.
    split.
      auto with PMap.
    rsplit.
      unfold Aemp; trivial.
    apply Hclipp'.
    apply linkPG_intro with (Mnh, R, fl, isr); trivial.
    split; auto.
    destruct (nextS_inversion HnextSnh') as [_ [_ [Hflx _]]].
    destruct Hflx as [x [Hx Hy]].
    exists x; simplS.
    split_l.
    rewrite <- Hy; trivial.
  intros Htag.
  substH Hwfst with (Hwfst Htag); clear Htag.
  apply NH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, NH_ItpG.
  intros [[[Mx Rx] flx] isrx] [Mx1 [Mx2 [Mx1' [Mx2' Hg]]]].
  simplS; simplS in Hg.
  destruct Hg as [Hg1 [Hg2 [Hg3 [Hg4 Hg5]]]].
  unfold NH_Inv in Hg3.
  rewrite Hfln in Hg3.
  assert (Mx2 = MemMod.emp).
    unfold Aemp; trivial.
  subst Mx2.
  assert (Mx1 = M).
    auto with PMap.
  subst Mx1.
  exists Mnh; exists Mih.
  exists Mx1'; exists Mx2'.
  split.
    auto with PMap.
  split_l.
  split.
    unfold NH_Inv; rewrite Hfl; trivial.
  split_l.
  apply Hclig'g.
  apply conjG_intro with (M, R', nat2FR flnum, isr); trivial.
  apply conjPG_intro; trivial.
  split; auto.
  destruct (nextS_inversion HnextSnh') as [_ [_ [Hflx _]]].
  destruct Hflx as [x [Hx Hy]].
  exists x; simplS.
  split_l.
  rewrite <- Hy; trivial.
  (* 0 -> 0 *)
  unfold NH_Inv in * .
  rewrite Hfl in HMih.
  rewrite Hfln.
  assert (Mih = MemMod.emp).
    unfold Aemp in HMih; trivial.
  subst Mih.
  assert (Mnh = M).
    auto with PMap.
  subst Mnh.
  assert (Hgpopfw_cli : gpopfw_cli (M, R, fl, isr)
     (M, R', nat2FR flnum, isr)).
    unfold gpopfw_cli.
    simplS.
    rsplit.
      auto.
    unfold prepost, choice, itr_enabled.
    unfold Aemp.
    simplS.
    rewrite Hfl.
    exists MemMod.emp; exists M; exists MemMod.emp.
    split_l.
    rsplit.
      auto with PMap.
    right; split_r.
    intros [Hx Hy]; discriminate Hx.
  split.
    exists M; exists MemMod.emp.
    split.
      auto with PMap.
    rsplit.
      unfold Aemp; trivial.
    apply Hclipp'; trivial.
    apply linkPG_intro with (M, R, fl, isr); trivial.
    split; auto.
    destruct (nextS_inversion HnextSnh') as [_ [_ [Hflx _]]].
    destruct Hflx as [x [Hx Hy]].
    exists x; simplS.
    split_l.
    rewrite <- Hy; trivial.
  intros Htag.
  substH Hwfst with (Hwfst Htag); clear Htag.
  apply NH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, NH_ItpG.
  intros [[[Mx Rx] flx] isrx] [Mx1 [Mx2 [Mx1' [Mx2' Hg]]]].
  simplS; simplS in Hg.
  destruct Hg as [Hg1 [Hg2 [Hg3 [Hg4 Hg5]]]].
  unfold NH_Inv in Hg3.
  rewrite Hfln in Hg3.
  assert (Mx2 = MemMod.emp).
    unfold Aemp; trivial.
  subst Mx2.
  assert (Mx1 = M).
    auto with PMap.
  subst Mx1.
  exists M; exists MemMod.emp.
  exists Mx1'; exists Mx2'.
  split.
    auto with PMap.
  split_l.
  split.
    unfold NH_Inv; rewrite Hfl; trivial.
  split_l.
  apply Hclig'g.
  apply conjG_intro with (M, R', nat2FR flnum, isr); trivial.
  apply conjPG_intro; trivial.
  split; auto.
  exists flnum; auto.
Qed.
Implicit Arguments NH_popfw_prev [rho Psi pc S0 S' p p' g g' tag].

Lemma NH_cmpw_prog : forall rho Psi o r pc S0
  (p : Pred) (g : Grt) tag,
  itp_NH_theta rho (Ast p g tag) Psi S0
  -> exists S', nextS pc (comml (_cmpw o r)) S0 S'.
Proof.
  intros rho Psi o r pc [[[M R] fl] isr] p g tag.
  intros H.
  destruct (eq_nat_dec (opr_eval R o) (R r)).
  exists (M, R, Set_Flag fl _ZF, isr).
  constructor.
  left; split_lr.
  exists (M, R, Clear_Flag fl _ZF, isr).
  constructor.
  right; split_lr.
Qed.
Implicit Arguments NH_cmpw_prog [rho Psi o r pc S0 p g tag].

Lemma NH_cmpw_prev : forall rho Psi o r pc S0 S' 
  p p' g g' tag,
  lookupCT Psi (pc+1) (packS0 rho (Ast p' g' tag))
  -> nextS pc (comml (_cmpw o r)) S0 S'
  -> enable p (nextS pc (comml (_cmpw o r)))
  -> pimp (p |> nextS pc (comml (_cmpw o r))) p'
  -> gimp (p >o nextS pc (comml (_cmpw o r)) oo g') g
  -> itp_NH_theta rho (Ast p g tag) Psi S0
  -> itp_NH_theta rho (Ast p' g' tag) Psi S'.
Proof.
  intros rho Psi o r pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p' g g' tag.
  intros Hlk' HnextS Hen Hpp' Hg'g H.
  unfold itp_NH_theta, NH_pinvstar, pmstar in * .
  destruct H as [Hisrf [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  subst M' R' isr'.
  unfold enable in Hen.
  destruct (Hen _ HMnh) as [Sx HnextSSx].
  destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
  clear Sx HnextSSx Hen.
  simplS in * .
  split_l.
  assert (HS' : (p |> nextS pc (comml (_cmpw o r))) (Mnh', R, fl', isr)).
    apply linkPG_intro with (Mnh,R,fl,isr); trivial.
  assert (HInv' : NH_Inv fl' Mih).
    unfold NH_Inv.
    destruct Hfl.
      destruct H; subst fl'.
      unfold Set_Flag.
      simpl.
      trivial.
    destruct H; subst fl'.
    unfold Clear_Flag.
    simpl.
    trivial.
  split.
    exists Mnh'; exists Mih; split_l.
    split_r.
    apply Hpp'; trivial.
  intros Htag.
  substH Hwfst with (Hwfst Htag); clear Htag.
  apply NH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, NH_ItpG.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold NH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  simplS in * .
  destruct (prec_disjUnion_inv HJMxy HJM' HMy HInv' (NHInv_prec fl')) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  simplS.
  exists Mnh; exists Mih.
  exists Mnh''; exists Mih''.
  split_l; split_l; split_l; split_l.
  apply Hg'g.
  apply conjG_intro with (Mnh', R, fl', isr); trivial.
  apply conjPG_intro; trivial.
Qed.
Implicit Arguments NH_cmpw_prev [rho Psi o r pc S0 S' p p' g g' tag].

Lemma NH_addw_prog : forall rho Psi o r pc S0
  (p : Pred) (g : Grt) tag,
  enable p (nextS pc (comml (_addw o r)))
  -> itp_NH_theta rho (Ast p g tag) Psi S0
  -> exists S', nextS pc (comml (_addw o r)) S0 S'.
Proof.
  intros rho Psi o r pc [[[M R] fl] isr] p g tag.
  intros Hen H.
  unfold itp_NH_theta, NH_pinvstar, pmstar in * .
  destruct H as [Hisrf [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  simplS in * .
  destruct (Hen _ HMnh) as [S' HnextSnh]. 
  destruct S' as [[[Mnh' R'] fl'] isr'].
  destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
  exists (M', R', fl', isr'); trivial.
Qed.
Implicit Arguments NH_addw_prog [rho Psi o r pc S0 p g tag].

Lemma NH_addw_prev : forall rho Psi o r pc S0 S' p p' g g' tag,
  lookupCT Psi (pc+1) (packS0 rho (Ast p' g' tag))
  -> nextS pc (comml (_addw o r)) S0 S'
  -> enable p (nextS pc (comml (_addw o r)))
  -> pimp (p |> nextS pc (comml (_addw o r))) p'
  -> gimp (p >o nextS pc (comml (_addw o r)) oo g') g
  -> itp_NH_theta rho (Ast p g tag) Psi S0
  -> itp_NH_theta rho (Ast p' g' tag) Psi S'.
Proof.
  intros rho Psi o r pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p' g g' tag.
  intros Hlk' HnextS Hen Hpp' Hg'g H.
  unfold itp_NH_theta, NH_pinvstar, pmstar in * .
  destruct H as [Hisrf [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  subst M' isr'.
  unfold enable in Hen.
  destruct (Hen _ HMnh) as [Sx HnextSSx].
  destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
  clear Sx HnextSSx Hen.
  simplS in * .
  split_l.
  assert (HInv' : NH_Inv fl' Mih).
    unfold NH_Inv.
    destruct Hfl; subst fl'.
      unfold Set_Flag.
      simpl.
      trivial.
    unfold Clear_Flag.
    simpl.
    trivial.
  assert (HS' : (p |> nextS pc (comml (_addw o r))) (Mnh', R', fl', isr)).
    apply linkPG_intro with (Mnh,R,fl,isr); trivial.
  unfold NH_pinvstar, pmstar.
  split.
    exists Mnh'; exists Mih; split_l.
    split_r.
    apply Hpp'; trivial.
  intros Htag.
  substH Hwfst with (Hwfst Htag); clear Htag.
  apply NH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, NH_ItpG.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold NH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  simplS in * .
  destruct (prec_disjUnion_inv HJMxy HJM' HMy HInv' (NHInv_prec fl')) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  simplS.
  exists Mnh; exists Mih.
  exists Mnh''; exists Mih''.
  split_l; split_l; split_l; split_l.
  apply Hg'g.
  apply conjG_intro with (Mnh', R', fl', isr); trivial.
  apply conjPG_intro; trivial.
Qed.
Implicit Arguments NH_addw_prev [rho Psi o r pc S0 S' p p' g g' tag].

Lemma NH_subw_prog : forall rho Psi o r pc S0
  (p : Pred) (g : Grt) tag,
  enable p (nextS pc (comml (_subw o r)))
  -> itp_NH_theta rho (Ast p g tag) Psi S0
  -> exists S', nextS pc (comml (_subw o r)) S0 S'.
Proof.
  intros rho Psi o r pc [[[M R] fl] isr] p g tag.
  intros Hen H.
  unfold itp_NH_theta, NH_pinvstar, pmstar in * .
  destruct H as [Hisrf [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  simplS in * .
  destruct (Hen _ HMnh) as [S' HnextSnh]. 
  destruct S' as [[[Mnh' R'] fl'] isr'].
  destruct (nextS_Mono HJM HnextSnh) as [M' [HJM' HnextS]].
  exists (M', R', fl', isr'); trivial.
Qed.
Implicit Arguments NH_subw_prog [rho Psi o r pc S0 p g tag].

Lemma NH_subw_prev : forall rho Psi o r pc S0 S' p p' g g' tag,
  lookupCT Psi (pc+1) (packS0 rho (Ast p' g' tag))
  -> nextS pc (comml (_subw o r)) S0 S'
  -> enable p (nextS pc (comml (_subw o r)))
  -> pimp (p |> nextS pc (comml (_subw o r))) p'
  -> gimp (p >o nextS pc (comml (_subw o r)) oo g') g
  -> itp_NH_theta rho (Ast p g tag) Psi S0
  -> itp_NH_theta rho (Ast p' g' tag) Psi S'.
Proof.
  intros rho Psi o r pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p' g g' tag.
  intros Hlk' HnextS Hen Hpp' Hg'g H.
  unfold itp_NH_theta, NH_pinvstar, pmstar in * .
  destruct H as [Hisrf [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  subst M' isr'.
  unfold enable in Hen.
  destruct (Hen _ HMnh) as [Sx HnextSSx].
  destruct (nextS_DMono HJM HnextSSx HnextS) as [Mnh' [HJM' HnextSnh']].
  clear Sx HnextSSx Hen.
  simplS in * .
  split_l.
  assert (HInv' : NH_Inv fl' Mih).
    unfold NH_Inv.
    destruct Hfl; subst fl'.
      unfold Set_Flag.
      simpl.
      trivial.
    unfold Clear_Flag.
    simpl.
    trivial.
  assert (HS' : (p |> nextS pc (comml (_subw o r))) (Mnh', R', fl', isr)).
    apply linkPG_intro with (Mnh,R,fl,isr); trivial.
  unfold NH_pinvstar, pmstar.
  split.
    exists Mnh'; exists Mih; split_l.
    split_r.
    apply Hpp'; trivial.
  intros Htag.
  substH Hwfst with (Hwfst Htag); clear Htag.
  apply NH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, NH_ItpG.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold NH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  simplS in * .
  destruct (prec_disjUnion_inv HJMxy HJM' HMy HInv' (NHInv_prec fl')) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  simplS.
  exists Mnh; exists Mih.
  exists Mnh''; exists Mih''.
  split_l; split_l; split_l; split_l.
  apply Hg'g.
  apply conjG_intro with (Mnh', R', fl', isr); trivial.
  apply conjPG_intro; trivial.
Qed.
Implicit Arguments NH_subw_prev [rho Psi o r pc S0 S' p p' g g' tag].

Lemma NH_jmp_prev : forall rho Psi f pc S0 S' p p' g g' tag,
  lookupCT Psi f (packS0 rho (Ast p' g' tag))
  -> nextS pc (comms (_jmp f)) S0 S'
  -> pimp p p'
  -> gimp (p >o g') g
  -> itp_NH_theta rho (Ast p g tag) Psi S0
  -> itp_NH_theta rho (Ast p' g' tag) Psi S'.
Proof.
  intros rho Psi f pc [[[M R] fl] isr] [[[M' R'] fl'] isr'] p p' g g' tag.
  intros Hlk' HnextS Hpp' Hg'g H.
  unfold itp_NH_theta, NH_pinvstar, pmstar in * .
  destruct H as [Hisrf [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  destruct (nextS_inversion HnextS) as [HM [HR [Hfl Hisr]]].
  subst M' R' fl' isr'.
  split_l.
  split.
    exists Mnh; exists Mih; split_l.
    split_r.
    apply Hpp'; trivial.
  intros Htag.
  substH Hwfst with (Hwfst Htag); clear Htag.
  apply NH_WFST_strengthen with (g:=g) (S0:= (M, R, fl, isr)); trivial.
  unfold pimp, NH_ItpG.
  intros S'' Hg'.
  destruct S'' as [[[M'' R''] fl''] isr''].
  unfold NH_ItpG in Hg'.
  destruct Hg' as [Mx [My [Mnh'' [Mih'' [HJMxy [HJM'' [HMy [HMih'' Hg']]]]]]]].
  simplS in * .
  destruct (prec_disjUnion_inv HJMxy HJM HMy HMih (NHInv_prec fl)) as [Heq1 Heq2].
  subst Mx My.
  clear HJMxy HMy.
  simplS.
  exists Mnh; exists Mih.
  exists Mnh''; exists Mih''.
  split_l; split_l; split_l; split_l.
  apply Hg'g.
  apply conjPG_intro; trivial.
Qed.
Implicit Arguments NH_jmp_prev [rho Psi f pc S0 S' p p' g g' tag].

Lemma NH_retw_prog : forall rho Psi pc S0
  (p : Pred) (g : Grt) tag,
  tag = 1
  -> (forall S, p S -> g S S)
  -> itp_NH_theta rho (Ast p g tag) Psi S0
  -> exists S', nextS pc (comms (_retw)) S0 S'.
Proof.
  intros rho Psi pc [[[M R] fl] isr] p g tag Htag Hpg H.
  unfold itp_NH_theta, NH_pinvstar, pmstar in * .
  destruct H as [Hisrf [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  simplS in * .
  substH Hwfst with (Hwfst Htag).
  unfold NH_WFST in Hwfst.
  assert (Hg : NH_ItpG g (M, R, fl, isr) (M, R, fl, isr)).
    simplS.
    exists Mnh; exists Mih.
    exists Mnh; exists Mih.
    split_l; split_l; split_l; split_l.
    apply Hpg; trivial.
  destruct Hwfst as [n Hwfst].
  destruct n as [ | n].
    simpl in Hwfst.
    substH Hwfst with (Hwfst (M, R, fl, isr) Hg).
    destruct Hwfst as [ra [Hra [p' [g' [tag' [Hlookup [Htag' Hp']]]]]]].
    assert (Heq : leadsToW (R SP) ra M).
      unfold Ret_Addr in Hra.
      simpl in Hra.
      substH Hra with (Hra pc).
      inversion_clear Hra.
      subst sp.
      trivial.
    exists (M, RF_update R SP (R SP + 2), fl, isr).
    apply nextS_retw.
    unfold popW.
    exists ra; split_lr.
  simpl in Hwfst.
  substH Hwfst with (Hwfst _ Hg).
  destruct Hwfst as [ra [Hra [p' [g' [tag' [Hlk Hwfst]]]]]].
  assert (Heq : leadsToW (R SP) ra M).
    unfold Ret_Addr in Hra.
    simpl in Hra.
    substH Hra with (Hra pc).
    inversion_clear Hra.
    subst sp.
    trivial.
  exists (M, RF_update R SP (R SP + 2), fl, isr).
  apply nextS_retw.
  unfold popW.
  exists ra; split_lr.
Qed.
Implicit Arguments NH_retw_prog [rho Psi pc S0 p g tag].

Lemma NH_retw_prev : forall rho Psi pc M R fl isr pc' S'
  (p : Pred) (g : Grt) tag,
  tag = 1
  -> nextPC (comms _retw) M R fl pc pc'
  -> nextS pc (comms _retw) (M, R, fl, isr) S'
  -> (forall S, p S -> g S S)
  -> itp_NH_theta rho (Ast p g tag) Psi (M,R,fl,isr)
  -> exists theta', 
    lookupCT Psi pc' (packS0 rho theta')
    /\ itp_NH_theta rho theta' Psi S'.
Proof.
  intros rho Psi pc M R fl isr pc' [[[M' R'] fl'] isr'] p g tag. 
  intros Htag HnextPC HnextS Hpg H.
  unfold itp_NH_theta, NH_pinvstar, pmstar in * .
  destruct H as [Hisrf [Hp Hwfst]].
  destruct Hp as [Mnh [Mih [HJM [HMnh HMih]]]].
  simplS in * .
  substH Hwfst with (Hwfst Htag).
  destruct (nextS_inversion HnextS) as [HM' [HR' [Hfl' Hisr']]].
  subst M' fl' isr'.
  inversion_clear HnextPC.
  subst sp.
  unfold NH_WFST in Hwfst.
  assert (Hg : NH_ItpG g (M, R, fl, isr) (M, R, fl, isr)).
    simplS.
    exists Mnh; exists Mih.
    exists Mnh; exists Mih.
    split_l; split_l; split_l; split_l.
    apply Hpg; trivial.
  destruct Hwfst as [n Hwfst].
  destruct n as [ | n].
    simpl in Hwfst.
    substH Hwfst with (Hwfst (M, R, fl, isr) Hg).
    destruct Hwfst as [ra [Hra [p' [g' [tag' [Hlookup [Htag' Hp']]]]]]].
    assert (Heq : pc' = ra).
      unfold Ret_Addr in Hra.
      simpl in Hra.
      substH Hra with (Hra pc).
      inversion_clear Hra.
      subst sp.
      eapply leadsToW_unique; eauto.
    subst pc'.
    exists (Ast p' g' tag'); split_l.
    split_l.
    simpl.
    substH Hp' with (Hp' _ _ HnextS).
    unfold NH_pinvstar, mstar in Hp'.
    destruct Hp' as [Mnh' [Mih' [HJM' Hp']]].
    simplS in * .
    split.
      exists Mnh'; exists Mih'; split_lr.
    intros Htag'x.
    destruct (Htag' Htag'x).
  simpl in Hwfst.
  substH Hwfst with (Hwfst _ Hg).
  destruct Hwfst as [ra [Hra [p' [g' [tag' [Hlk Hwfst]]]]]].
  assert (Heq : pc' = ra).
    unfold Ret_Addr in Hra.
    simpl in Hra.
    substH Hra with (Hra pc).
    inversion_clear Hra.
    subst sp.
    eapply leadsToW_unique; eauto.
  subst pc'.
  exists (Ast p' g' tag'); split_l.
  split_l.
  substH Hwfst with (Hwfst _ _ HnextS).
  destruct Hwfst as [Hp' Hwfst].
  unfold NH_pinvstar, mstar in Hp'.
  destruct Hp' as [Mnh' [Mih' [HJM' Hp']]].
  simplS in * .
  split.
    exists Mnh'; exists Mih'; split_lr.
  intros Htag'.
  exists n; trivial.
Qed.
Implicit Arguments NH_retw_prev [rho Psi pc M R fl isr pc' S' p g tag].
