(* This is the SCAP for non-handler code *)
Add LoadPath "..".
Add LoadPath "../CAP0".

Require Import Coqlib.
Require Import Map.
Require Import MultiMapT.
Require Import Arith.
Require Import Machine.
Require Import Assertion.
Require Import Operational_Semantics.
Require Import CAP0.
Require Import SCAP.
Require Import AIM.
Require Import Bool.

Set Implicit Arguments.
Unset Strict Implicit.

Definition Prd_Itr_Stk (p : Pred) : Prop :=
  forall S, p S -> itr_enabled S
    -> exists M1, exists M2, MemMod.disjUnion M1 M2 (S_Mem S)
      /\ stkSpacePrec (S_RF S SP) (IH_Stk + 6) M2
      /\ (forall M2', stkSpacePrec (S_RF S SP) (IH_Stk + 6) M2'
        -> p (S_Mem_Upd S (MemMod.merge M1 M2')) ).

(* Definition Grt_Itr_Stk (g : Grt) : Prop := *)
(*   forall S S',  g S S' -> itr_enabled S *)
(*     -> exists M1, exists M2, MemMod.disjUnion M1 M2 (S_Mem S) *)
(*       /\ stkSpacePrec (S_RF S SP) (IH_Stk + 6) M2 *)
(*       /\ (forall M2', stkSpacePrec (S_RF S SP) (IH_Stk + 6) M2' *)
(*         -> g (S_Mem_Upd S (MemMod.merge M1 M2')) S'). *)

(* This definition seems to be too strong *)
(* should be the following: *)
Definition Grt_Itr_Stk (g : Grt) : Prop :=
  forall S S',  
    g S S' 
    -> itr_enabled S
    -> forall M1 M2 M2',
         MemMod.disjUnion M1 M2 (S_Mem S)
         -> stkSpacePrec (S_RF S SP) (IH_Stk + 6) M2
         -> stkSpacePrec (S_RF S SP) (IH_Stk + 6) M2'
         -> g (S_Mem_Upd S (MemMod.merge M1 M2')) S'.

(* the SCAP for non-handler & READY thread *)


(* trashes the stack space _SW_Stk *)
Definition gy : Grt :=
  fun (S S' : State) => 
    exists sp,
      prepost (mstar (mstar INV0 INV1) (stkSpacePrec sp _SW_Stk)) (mstar (mstar INV0 INV1) (stkSpacePrec sp _SW_Stk)) S S' 
        /\ S_Flag S _IF = S_Flag S' _IF
        /\ S_ISR S = S_ISR S'
        /\ S_RF S SP = sp
        /\ S_RF S SP = S_RF S' SP.


(* trashes the stack space _BLK_Stk *)
Definition gblk (m : MPred) : Grt :=
  fun (S S' : State) =>
    exists sp,
      prepost (mstar (mstar INV0 INV1) (stkSpacePrec sp _BLK_Stk)) 
              (mstar (mstar (mstar INV0 INV1) m) (stkSpacePrec sp _BLK_Stk)) 
      S S'
       /\ S_Flag S _IF = S_Flag S' _IF
       /\ S_ISR S = S_ISR S'
       /\ S_RF S SP = sp
       /\ S_RF S SP = S_RF S' SP.


Definition enable_sw (p : Pred) : Prop :=
    forall S, p S -> (pmstar pTrue (mstar (mstar INV0 INV1) (stkSpacePrec (S_RF S SP) (_SW_Stk + 2)))) S.


(*
(* we need to ensure that the current thread is not IDLE *)
Definition enable_blk (p : Pred) (wqa : WQassert): Prop :=
  disj_PM p wqa  (* probably not necessary, double check here *)
  /\ forall S, p S ->
        exists w, (Delta w = SomeT wqa) 
                  /\ (pmstar pTrue (mstar (ptoW (S_RF S SP) w)
                                           (mstar (mstar INV0 INV1) 
                                                  (stkSpacePrec (S_RF S SP) (_BLK_Stk + 2))
                                           )
                                   )
                     ) S.
*)


Definition enable_blk (w : nat) (wqa : WQassert) (S : State) : Prop :=
   (Delta w = SomeT wqa) 
   /\ (pTrue *<* ((S_RF S SP |->w w) *^* ((INV0 *^* INV1) *^* (stkSpacePrec (S_RF S SP) (_BLK_Stk + 2))))) S.


(* trash all registers *)
Definition gunblk (m : MPred) : Grt :=
  fun (S S' : State) =>
    exists sp,
      prepost (mstar (choice (S_RF S' AX = 0) Aemp m) (stkSpacePrec sp _UNBLK_Stk)) (stkSpacePrec sp _UNBLK_Stk) S S'
      /\ S_Flag S _IF = S_Flag S' _IF
      /\ S_ISR S = S_ISR S'
      /\ S_RF S SP = sp
      /\ S_RF S SP = S_RF S' SP.

(*
Definition enable_unblk (p : Pred) (wqa : WQassert) : Prop :=
    forall S, p S ->
              exists w, (Delta w = SomeT wqa)
                        /\ (pmstar pTrue (mstar (ptoW (S_RF S SP) w)
                                                (mstar wqa 
                                                       (stkSpacePrec (S_RF S SP) (_UNBLK_Stk + 2))
                                                )
                                         )
                           ) S.
*)


Definition enable_unblk (w : nat) (wqa : WQassert) (S : State) : Prop :=
   (Delta w = SomeT wqa) 
   /\ (pTrue *<* ((S_RF S SP |->w w) *^* (wqa *^* (stkSpacePrec (S_RF S SP) (_UNBLK_Stk + 2))))) S.


Inductive NH_Call_Prim_aux (psi : SCAP_PSpec) (pc f: flab) (theta : Spec) : Prop :=
    Call_SW : 
       forall p p' g g' tag, (f = _SWITCH) 
                         -> theta = (Ast p g tag) 
                         -> SCAP_lookupCT psi (pc+1) (Ast p' g' tag)
                         -> enable_sw p
                         -> (forall pc', pimp (p |> (wrap_fun gy pc f pc')) p')
                         -> (forall pc', gimp (p >o (wrap_fun gy pc f pc') oo g') g)
                         -> NH_Call_Prim_aux psi pc f theta
  | Call_BLK :
       forall p p' g g' tag, 
                         (f = _BLOCK)
                         -> theta = (Ast p g tag) 
                         -> SCAP_lookupCT psi (pc+1) (Ast p' g' tag)
                         -> (forall S, p S -> exists w, exists wqa,
                               (enable_blk w wqa S)
                               /\ (forall pc', pimp ((wrap_fun (gblk wqa) pc f pc') S) p')
                               /\ (forall pc', pimp (((wrap_fun (gblk wqa) pc f pc') S) |> g') (g S))
                            )
                         -> NH_Call_Prim_aux psi pc f theta
  | Call_UNBLK :
       forall p p' g g' tag,
                         (f = _UNBLK)
                         -> theta = (Ast p g tag)
                         -> SCAP_lookupCT psi (pc+1) (Ast p' g' tag)
                         -> (forall S, p S -> exists w, exists wqa,
                               (enable_unblk w wqa S)
                               /\ (forall pc', pimp ((wrap_fun (gunblk wqa) pc f pc') S) p')
                               /\ (forall pc', pimp (((wrap_fun (gunblk wqa) pc f pc') S) |> g') (g S))
                            )
                         -> NH_Call_Prim_aux psi pc f theta.
(*
  | Call_BLK :
       forall p p' g g' wqa tag, 
                         (f = _BLOCK)
                         -> theta = (Ast p g tag) 
                         -> enable_blk p wqa
                         -> (forall pc', pimp (p |> (wrap_fun (gblk wqa) pc f pc')) p')
                         -> (forall pc', gimp (p >o (wrap_fun (gblk wqa) pc f pc') oo g') g)
                         -> NH_Call_Prim_aux psi pc f theta
                         -> SCAP_lookupCT psi (pc+1) (Ast p' g' tag)
  | Call_UNBLK :
       forall p p' g g' wqa tag, 
                         (f = _UNBLK) 
                         -> theta = (Ast p g tag)
                         -> SCAP_lookupCT psi (pc+1) (Ast p' g' tag)
                         -> enable_unblk p wqa
                         -> (forall pc', pimp (p |> (wrap_fun (gunblk wqa) pc f pc')) p')
                         -> (forall pc', gimp (p >o (wrap_fun (gunblk wqa) pc f pc') oo g') g)
                         -> NH_Call_Prim_aux psi pc f theta.
*)

Definition NH_Call_Prim (psi : SCAP_PSpec) (pc f: flab) (theta : Spec) : Prop :=
  match theta with
    Ast p g _ => (NH_Call_Prim_aux psi pc f theta) /\ (forall S, p S -> S_Flag S _IF = false)
  end.

Definition gcli : Grt :=
  fun (S S' : State) =>
    prepost Aemp (choice (itr_enabled S) (mstar INV0 INV1) Aemp) S S'
    /\ (S_Flag S' _IF = false) /\ (S_ISR S = S_ISR S') /\ (S_RF S = S_RF S').

(* note that this definition is for non-handler code *)
(* we always assume that ISR = false                 *)
Definition gsti : Grt :=
  fun (S S' : State) =>
    prepost (choice (itr_enabled S) Aemp (mstar INV0 INV1)) Aemp S S'
    /\ (S_Flag S' _IF = true) /\ (S_ISR S = S_ISR S') /\ (S_RF S = S_RF S').

(* note that this definition is for non-handler code *)
(* we always assume that ISR = false                 *)
Definition enable_sti (p : Pred) : Prop :=
  forall S,
    p S -> (pmstar pTrue (choice (itr_enabled S) Aemp (mstar INV0 INV1))) S.


(* popfw *)

Definition enable_popfw_sti (p : Pred) : Prop :=
  forall M R fl isr,
    p (M,R,fl,isr) -> (mstar (leadsToxW (R SP)) (choice (fl _IF=true) Aemp (mstar INV0 INV1))) M.

Definition stk_IF_st : Pred :=
  fun (S : State) =>  exists flnum, 
    leadsToW ((S_RF S) SP) flnum (S_Mem S) /\ nat2FR flnum _IF = true.

Definition stk_IF_cl : Pred :=
  fun (S : State) =>  exists flnum, 
    leadsToW ((S_RF S) SP) flnum (S_Mem S) /\ nat2FR flnum _IF = false.

Definition gpopfw_cli : Grt :=
  fun (S S' : State) =>
    prepost Aemp (choice (itr_enabled S) (mstar INV0 INV1) Aemp) S S'
    /\ (S_Flag S' _IF = false) /\ (S_ISR S = S_ISR S') 
    /\ (S_RF S' = RF_update (S_RF S) SP ((S_RF S) SP +2)).

Definition gpopfw_sti : Grt :=
  fun (S S' : State) =>
    prepost (choice (itr_enabled S) Aemp (mstar INV0 INV1)) Aemp S S'
    /\ (S_Flag S' _IF = true) /\ (S_ISR S = S_ISR S') 
    /\ (S_RF S' = RF_update (S_RF S) SP ((S_RF S) SP +2)).

Definition NH_Normal_Call (psi : SCAP_PSpec) (pc f : flab) (theta : Spec) :=
  match theta with
    Ast p g tag =>
        exists p', exists p'', exists g', exists g'',
                     SCAP_lookupCT psi f (Ast p' g' 1)
                     /\ SCAP_lookupCT psi (pc+1) (Ast p'' g'' tag)
                     /\ enable p (nextS pc (comml (_call f)))
                     /\ gimp ((sp_call_pred p pc f) >o g') fun_frame
                     /\ pimp (p |> (nextS pc (comml (_call f)))) p'
                     /\ (forall pc', pimp (p |> (wrap_fun g' pc f pc')) p'')
                     /\ (forall pc', gimp (p >o (wrap_fun g' pc f pc') oo g'') g)
  end.

Definition NH_WFcmd_aux (psi : SCAP_PSpec) (pc : flab) (theta : Spec) (c : comm) : Prop :=
  match theta with
    Ast p g tag =>
    match c with
      comms si =>
      match si with
        _iret => False
      | _retw => (tag = 1) /\ (forall S, p S -> g S S)     (* tag = 1 means this is not topmost fun *)
      | _jmp f => exists p', exists g',
                    SCAP_lookupCT psi f (Ast p' g' tag)
                    /\ pimp p p'
                    /\ gimp (p >o g') g
      end
    | comml i  =>
      match i with
        _call f => (prim_entry f /\ NH_Call_Prim psi pc f theta)
                     \/ (~ (prim_entry f) /\ NH_Normal_Call psi pc f theta)
      | _cli => exists p', exists g',
                  SCAP_lookupCT psi (pc+1) (Ast p' g' tag)
                  /\ pimp (p |> gcli) p'
                  /\ gimp (p >o gcli oo g') g
      | _sti => exists p', exists g',
                  SCAP_lookupCT psi (pc+1) (Ast p' g' tag)
                  /\ enable_sti p
                  /\ pimp (p |> gsti) p'
                  /\ gimp (p >o gsti oo g') g
      | _outb_eoi => False
      | _pushfw => exists p', exists g',
                 SCAP_lookupCT psi (pc+1) (Ast p' g' tag)
                 /\ enable p (nextS pc c)
                 /\ pimp (p |> (nextS pc c)) p'
                 /\ gimp (p >o (nextS pc c) oo g') g
      | _popfw => exists p', exists g',
                 SCAP_lookupCT psi (pc+1) (Ast p' g' tag)
                 /\ enable p (nextS pc c)
                 /\ enable_popfw_sti (conjP p stk_IF_st)
                 /\ pimp ((conjP p stk_IF_st) |> gpopfw_sti) p'
                 /\ gimp ((conjP p stk_IF_st) >o gpopfw_sti oo g') g
                 /\ pimp ((conjP p stk_IF_cl) |> gpopfw_cli) p'
                 /\ gimp ((conjP p stk_IF_cl) >o gpopfw_cli oo g') g
      | _  => SCAP_WFcmd psi pc theta c
      end
    end
  end.

Definition NH_WFcmd (psi : SCAP_PSpec) (pc : flab) (theta : Spec) (c : comm) : Prop :=
  match theta with
    Ast p g tag =>
        NH_WFcmd_aux psi pc theta c /\ (Prd_Itr_Stk p /\ Grt_Itr_Stk g)
  end.

(* Definition NH_WFcmd (psi : SCAP_PSpec) (pc : flab) (theta : Spec) (c : comm) : Prop := *)
(*   match theta with *)
(*     Ast p g tag => *)
(*         NH_WFcmd_aux psi pc theta c /\ (Prd_Itr_Stk p /\ Grt_Itr_Stk g) *)
(*   end. *)

Inductive NH_WFiseq : SCAP_PSpec -> flab -> Spec -> iseq -> Prop :=
    nh_isq_seq :
        forall psi pc p p' g g' i isq tag,
          NH_WFcmd (SCAP_consCT (pc+1) (Ast p' g' tag) psi) pc (Ast p g tag) (comml i)
          -> NH_WFiseq psi (pc+1) (Ast p' g' tag) isq
          -> NH_WFiseq psi pc (Ast p g tag) (seq i isq)
  | nh_isq_end :
        forall psi pc p g si tag,
          NH_WFcmd psi pc (Ast p g tag) (comms si)
          -> NH_WFiseq psi pc (Ast p g tag) (sseq si).

Definition NH_WFcdhp (psi : SCAP_PSpec) (C: CodeHeap) (psi' : SCAP_PSpec): Prop :=
  forall f theta,
    SCAP_lookupCT psi' f theta -> exists isq, lookupC C f isq /\ NH_WFiseq psi f theta isq.

Lemma Prd_Itr_Stk_equal : forall (p p' : Pred),
  (forall S, p' S -> p S)
  -> (forall S, p S -> p' S)
  -> Prd_Itr_Stk p
  -> Prd_Itr_Stk p'.
Proof.
  unfold Prd_Itr_Stk.
  intros p p' Hp'p Hpp' Hp.
  intros S HS HIS.
  substH Hp with (Hp S (Hp'p _ HS) HIS).
  destruct Hp as [M1 [M2 [HJM [HM2 Hp]]]].
  exists M1; exists M2; split_l; split_l.
  intros.
  apply Hpp'.
  auto.
Qed.

Lemma Grt_Itr_Stk_equal : forall (g g' : Grt),
  (forall S S', g S S' -> g' S S')
  -> (forall S S', g' S S' -> g S S')
  -> Grt_Itr_Stk g
  -> Grt_Itr_Stk g'.
Proof.
  unfold Grt_Itr_Stk.
  intros g g' Hgg' Hg'g Hg.
  intros S S' HSS' HIS.
  intros M1 M2 M2' HJM HM2 HM2'.
  apply Hgg'.
  eapply Hg; eauto.
Qed.
