(* This is the SCAP which assumes interrupt is disabled *)
Add LoadPath "..".
Add LoadPath "../CAP0".

Require Import Coqlib.
Require Import Map.
Require Import MultiMapT.
Require Import Arith.
Require Import Machine.
Require Import Assertion.
Require Import Operational_Semantics.
Require Import OpSem_Lemmas.

Require Import CAP0.
Require Import Link.

Require Import SCAP.

Lemma WFST_strengthen :
  forall D rho n Psi S0 g S1 g',
    WFST D rho n Psi (scap_lift_g g S0)
    -> pimp (scap_lift_g g' S1) (scap_lift_g g S0)
    -> WFST D rho n Psi (scap_lift_g g' S1).
Proof.
  destruct n; simpl in * ; intros.
    apply H.
    apply H0; trivial.
  apply H.
  apply H0; trivial.
Qed.
Implicit Arguments WFST_strengthen [D rho n Psi S0 g S1 g'].

Lemma SCAP_Mono_CT : 
  forall D rho p g tag, 
    Mono_CT_Itp D
    -> Mono_CT_assert (Generic_Itp D rho (Ast p g tag)).
Proof.
  intros D rho p g tag HItp.
  unfold Mono_CT_assert, Generic_Itp.
  intros Psi S [Hp [Hif Hwfst]] Psi' Hsub.
  split; trivial.
  split; trivial.
  destruct Hwfst as [n Hwfst].
  exists n.
  clear Hif.
  gen_clear p g S Hp Hwfst.
  induction n; intros p g S Hp Hwfst.
    unfold WFST in Hwfst |- * .
    intros S' Hg. 
    substH Hwfst with (Hwfst S' Hg).
    destruct Hwfst as [ra [Hra [theta [Hthetain Hwfst]]]].
    exists ra; split; trivial.
    exists theta; split; trivial.
    apply Hsub; trivial.
    intros pc' S'' HnextS.
    substH Hwfst with (Hwfst pc' S'' HnextS).
    substH HItp with (HItp theta).
    unfold Mono_CT_assert in HItp.
    eapply HItp; eauto.
  simpl in * .
  intros.
  substH Hwfst with (Hwfst S0 H).
  destruct Hwfst as [ra [Hra [p' [g' [tag' [Hrax Hwfst]]]]]].
  exists ra; split_l.
  exists p'; exists g'; exists tag'.
  split.
    apply Hsub; trivial.
  intros pc' S' HnextS.
  substH Hwfst with (Hwfst pc' S' HnextS).
  destruct Hwfst.
  split_l.
  apply IHn with p'; trivial.
Qed.

Lemma WFST_nextS : forall D rho n Psi p pc M R fl isr,
  WFST D rho n Psi p
  -> p (M, R, fl, isr)
  -> nextS pc (comms _retw) (M, R, fl, isr) (M, RF_update R SP (R SP + 2), fl, isr).
Proof.
  intros.
  apply nextS_retw.
  destruct n; simpl in * .
    substH H with (H (M, R, fl, isr) H0); clear H0.
  destruct H as [ra [Hra H]].
  unfold Ret_Addr in Hra.
  substH Hra with (Hra (R SP)).
  inversion_clear Hra.
  unfold popW.
  subst sp.
  simpl in H1.
  exists ra.
  split_lr.
    substH H with (H (M, R, fl, isr) H0); clear H0.
  destruct H as [ra [Hra H]].
  unfold Ret_Addr in Hra.
  substH Hra with (Hra (R SP)).
  inversion_clear Hra.
  unfold popW.
  subst sp.
  simpl in H1.
  exists ra.
  split_lr.
Qed.
Implicit Arguments WFST_nextS [D rho n Psi p pc M R fl isr].

Lemma SCAP_IF_disable : forall D rho Psi theta S0,
  Generic_Itp D rho theta Psi S0
  -> S_Flag S0 _IF = false.
Proof.
  intros.
  destruct theta as [p g tag].
  simpl in H.
  destruct H as [_ [H _]]; trivial.
Qed.
Implicit Arguments SCAP_IF_disable [D rho Psi theta S0].

Lemma SCAP_in_CAP0_call : forall D rho psi pc theta D' Psi f,
  SCAP_WFcmd psi pc theta (comml (_call f))
  -> Mono_CT_Itp D
  -> D rho = NoneT1 ITP
  -> D' = ext_dict D rho (Generic_Itp D rho)
  -> lift_psi psi rho Psi
  -> WFcmd D' Psi (packS0 rho theta) pc (comml (_call f)).
Proof.
  intros D rho psi pc [p g tag] D' Psi f.
  simpl.
  intros Hi HDmono HD HD' Hlift.
  subst D'.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl; rewrite <- EqNat.beq_nat_refl.
  split_r with (SCAP_Mono_CT D rho p g tag HDmono).
  intros Psi' S HItp Hsub.
  destruct Hi as [p' [p'' [g' [g'' Hi]]]].
  destruct Hi as [Hf [Hf'' [Hen [Hra [Hpp' [Hg'p'' Hg''g]]]]]].
  destruct S as [[[M R] fl] isr].
  rrsplit.
    left.
    unfold not, itr_enabled.
    intros [Hx _]; rewrite (SCAP_IF_disable HItp) in Hx; discriminate Hx.
  simpl in HItp.
  destruct HItp as [Hp [Hfl Hwfst]].
  split.
    destruct (Hen _ Hp) as [S' HnextS].
    exists S'; trivial.
  intros f' [[[M' R'] fl'] isr'] HnextS HnextPC.
  destruct (nextS_inversion HnextS) as [HM' [HR' [Hfl' Hisr']]].
  assert (Heq :=nextPC_inversion HnextPC).
  simpl in Heq.
  subst fl' isr' f'.
  assert (Hlk' : lookupCT Psi' f (packS0 rho (Ast p' g' tag))).
    apply Hsub.
    apply Hlift.
    trivial.
  assert (Hlk'' : lookupCT Psi' (pc+1) (packS0 rho (Ast p'' g'' tag))).
    apply Hsub.
    apply Hlift.
    trivial.
  exists (packS0 rho (Ast p' g' tag)).
  split_l.
  unfold ext_dict.
  simpl.
  rewrite <- EqNat.beq_nat_refl.
  unfold Generic_Itp; simpl.
  assert (HS' : (p |> nextS pc (comml (_call f))) (M', R', fl, isr)).
    apply linkPG_intro with (M,R,fl,isr); trivial.
  split.
    apply Hpp'; trivial.
  split; trivial.
  destruct Hwfst as [n Hwfst].
  exists (1+n).
  simpl.
  intros S'' Hg'S'S''.
  unfold wrap_fun in Hg'p''.
  destruct Hg'S'S'' as [Hg' [Hgisr Hgfl]].
  assert (Hxx : fun_frame (M',R',fl,isr) S'').
    apply Hra.
    unfold sp_call_pred.
    apply conjPG_intro; trivial.
  unfold fun_frame in Hxx.
  destruct Hxx as [_ [ra [Hra' Hra'']]].
  assert (Hrax := ra_after_call _ _ Hra' HnextS).
  subst ra.
  exists (pc + 1).
  split; trivial.
  clear Hra' Hra''.
  exists p''; exists g''; exists tag.
  split_l.
  intros pc' S''' HnextS''.
  substH Hg'p'' with (Hg'p'' pc').
  split.
    apply Hg'p''.
    apply linkPG_intro with (M,R,fl,isr); trivial.
    apply conjG_intro with S''; trivial.
    apply conjG_intro with (M',R',fl,isr); trivial.
  apply (WFST_strengthen (S1 := S''') (g' :=g'') Hwfst).
  substH Hg''g with (Hg''g pc' (M,R,fl,isr)).
  intros S'''' [H [H1 H2]].
  split.
  apply Hg''g.
  apply conjG_intro with S'''; trivial.
  unfold wrap_fun.
  apply conjPG_intro; auto.
  apply conjG_intro with S''; trivial.
  apply conjG_intro with (M', R', fl, isr); trivial.
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct S''' as [[[M''' R'''] fl'''] isr'''].
  destruct S'''' as [[[M'''' R''''] fl''''] isr''''].
  simplS in * .
  destruct (nextS_inversion HnextS'') as [_ [_ [Hfl'' Hisr'']]].
  subst isr'' fl''' isr'''.
  split_l.
  rewrite Hgfl; rewrite H2; trivial.
Qed.

Lemma SCAP_in_CAP0_jne : forall D rho psi pc theta D' Psi f,
  SCAP_WFcmd psi pc theta (comml (_jne f))
  -> Mono_CT_Itp D
  -> D rho = NoneT1 ITP
  -> D' = ext_dict D rho (Generic_Itp D rho)
  -> lift_psi psi rho Psi
  -> WFcmd D' Psi (packS0 rho theta) pc (comml (_jne f)).
Proof.
  intros D rho psi pc [p g tag] D' Psi f.
  simpl.
  intros Hi HDmono HD HD' Hlift.
  subst D'.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl; rewrite <- EqNat.beq_nat_refl.
  split_r with (SCAP_Mono_CT D rho p g tag HDmono).
  destruct Hi as [p' [p'' [g' [g'' Hi]]]].
  destruct Hi as [Hf [Hf'' [Hpp' [Hg'g [Hpp'' Hg''g]]]]].
  intros Psi' S HItp Hsub.
  rrsplit.
    left.
    unfold not, itr_enabled.
    intros [Hx _]; rewrite (SCAP_IF_disable HItp) in Hx; discriminate Hx.
  destruct S as [[[M R] fl] isr].
  simpl in HItp.
  destruct HItp as [Hp [Hfl Hwfst]].
  split.
    assert (HnextS := nextS_jne pc f M R fl isr).
    exists (M,R,fl,isr); trivial.
  intros f' [[[M' R'] fl'] isr'] HnextS HnextPC.
  destruct (nextS_inversion HnextS) as [HM' [HR' [Hfl' Hisr']]].
  subst M' R' fl' isr'.
  simpl in HnextPC.
  destruct (nextPC_inversion HnextPC) as [Hpc1 Hpc2].
  generalize (refl_equal (fl _ZF)); pattern fl at 1; destruct (fl _ZF); intro Hzf.
    substH Hpc1 with (Hpc1 (refl_equal _)).
    subst f'.
    exists (packS0 rho (Ast p'' g'' tag)).
    split.
      apply Hsub.
      apply Hlift; trivial.
    unfold ext_dict; simpl.
    rewrite <- EqNat.beq_nat_refl.
    unfold Generic_Itp; simpl.
    assert (HS' : (p |> nextS pc (comml (_jne f))) (M, R, fl, isr)).
      apply linkPG_intro with (M,R,fl,isr); trivial.
    split.
      apply Hpp''; trivial.
      apply conjP_intro; auto.
      unfold ZF_st.
      split_l.
    destruct Hwfst as [n Hwfst].
    exists n.
    apply (WFST_strengthen Hwfst (S1:=(M,R,fl,isr))(g':=g'')).
    intros S'' Hg''.
    destruct Hg'' as [Hg'' [H1 H2]].
    split.
      apply Hg''g.
      apply conjPG_intro; trivial.
      apply conjPG_intro; trivial.
    split_lr.
  substH Hpc2 with (Hpc2 (refl_equal _)).
  subst f'.
  exists (packS0 rho (Ast p' g' tag)).
  split.
    apply Hsub.
    apply Hlift; trivial.
  unfold ext_dict; simpl.
  rewrite <- EqNat.beq_nat_refl.
  unfold Generic_Itp; simpl.
  assert (HS' : (p |> nextS pc (comml (_jne f))) (M, R, fl, isr)).
    apply linkPG_intro with (M,R,fl,isr); trivial.
  split.
    apply Hpp'; trivial.
    apply conjP_intro; auto.
    unfold ZF_cl; trivial.
    split; trivial.
  destruct Hwfst as [n Hwfst].
  exists n.
  apply (WFST_strengthen Hwfst (S1:=(M,R,fl,isr))(g':=g')).
    intros S' Hg'.
    destruct Hg' as [Hg' [H1 H2]].
    split.
      apply Hg'g.
      apply conjPG_intro; trivial.
      apply conjPG_intro; trivial.
    split_lr.
Qed.

Lemma SCAP_in_CAP0_movwp : forall D rho psi pc theta D' Psi o r,
  SCAP_WFcmd psi pc theta (comml (_movwp o r))
  -> Mono_CT_Itp D
  -> D rho = NoneT1 ITP
  -> D' = ext_dict D rho (Generic_Itp D rho)
  -> lift_psi psi rho Psi
  -> WFcmd D' Psi (packS0 rho theta) pc (comml (_movwp o r)).
Proof.
  intros D rho psi pc [p g tag] D' Psi o r.
  simpl.
  intros Hi HDmono HD HD' Hlift.
  subst D'.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl; rewrite <- EqNat.beq_nat_refl.
  split_r with (SCAP_Mono_CT D rho p g tag HDmono).
  destruct Hi as [p' [g' [Hslk' [Hen Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  intros Psi' S HItp Hsub.
  rrsplit.
    left.
    unfold not, itr_enabled.
    intros [Hx _]; rewrite (SCAP_IF_disable HItp) in Hx; discriminate Hx.
  destruct S as [[[M R] fl] isr].
  simpl in HItp.
  destruct HItp as [Hp [Hfl Hwfst]].
  split.
    destruct (enable_elim Hen Hp) as [S' HnextS].
    exists S'; trivial.
  intros f' [[[M' R'] fl'] isr'] HnextS HnextPC.
  destruct (nextS_inversion HnextS) as [HM' [HR' [Hfl' Hisr']]].
  assert (Hpc' := nextPC_inversion HnextPC).
  simpl in Hpc'.
  subst M' R' fl' isr' f'.
  exists (packS0 rho (Ast p' g' tag)).
  split.
    apply Hsub.
    apply Hlift; trivial.
  unfold ext_dict.
  simpl.
  rewrite <- EqNat.beq_nat_refl.
  unfold Generic_Itp; simpl.
  assert (HS' : (p |> nextS pc (comml (_movwp o r))) (M, RF_update R r (opr_eval R o), fl, isr)).
    apply linkPG_intro with (M,R,fl,isr); trivial.
  split.
    apply Hpp'; trivial.
  split_l.  
  destruct Hwfst as [n Hwfst].
  exists n.
  apply (WFST_strengthen Hwfst (S1:=(M,RF_update R r (opr_eval R o),fl,isr)) (g':=g')).
    intros S' Hg'.
    destruct Hg' as [Hg' [H1 H2]].
  split.
    apply Hg'g.
    apply conjG_intro with (M, RF_update R r (opr_eval R o), fl, isr); trivial.
    apply conjPG_intro; trivial.
  split_lr.
Qed.

Lemma SCAP_in_CAP0_movwld : forall D rho psi pc theta D' Psi a r,
  SCAP_WFcmd psi pc theta (comml (_movwld a r))
  -> Mono_CT_Itp D
  -> D rho = NoneT1 ITP
  -> D' = ext_dict D rho (Generic_Itp D rho)
  -> lift_psi psi rho Psi
  -> WFcmd D' Psi (packS0 rho theta) pc (comml (_movwld a r)).
Proof.
  intros D rho psi pc [p g tag] D' Psi a r.
  simpl.
  intros Hi HDmono HD HD' Hlift.
  subst D'.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl; rewrite <- EqNat.beq_nat_refl.
  split_r with (SCAP_Mono_CT D rho p g tag HDmono).
  destruct Hi as [p' [g' [Hslk' [Hen Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  intros Psi' S HItp Hsub.
  rrsplit.
    left.
    unfold not, itr_enabled.
    intros [Hx _]; rewrite (SCAP_IF_disable HItp) in Hx; discriminate Hx.
  destruct S as [[[M R] fl] isr].
  simpl in HItp.
  destruct HItp as [Hp [Hfl Hwfst]].
  split.
    destruct (enable_elim Hen Hp) as [S' HnextS].
    exists S'; trivial.
  intros f' [[[M' R'] fl'] isr'] HnextS HnextPC.
  destruct (nextS_inversion HnextS) as [HM' [HR' [Hfl' Hisr']]].
  assert (Hpc' := nextPC_inversion HnextPC).
  simpl in Hpc'.
  subst M' fl' isr' f'.
  exists (packS0 rho (Ast p' g' tag)).
  split.
    apply Hsub.
    apply Hlift; trivial.
  unfold ext_dict.
  simpl.
  rewrite <- EqNat.beq_nat_refl.
  unfold Generic_Itp; simpl.
  assert (HS' : (p |> nextS pc (comml (_movwld a r))) (M, R', fl, isr)).
    apply linkPG_intro with (M,R,fl,isr); trivial.
  split.
    apply Hpp'; trivial.
  split; trivial.
  destruct Hwfst as [n Hwfst].
  exists n.
  apply (WFST_strengthen Hwfst (S1:=(M,R',fl,isr)) (g':=g')).
  intros S' [Hg' [H1 H2]].
  split.
    apply Hg'g.
    apply conjG_intro with (M, R', fl, isr); trivial.
    apply conjPG_intro; trivial.
  split_lr.
Qed.

Lemma SCAP_in_CAP0_movwst : forall D rho psi pc theta D' Psi o a,
  SCAP_WFcmd psi pc theta (comml (_movwst o a))
  -> Mono_CT_Itp D
  -> D rho = NoneT1 ITP
  -> D' = ext_dict D rho (Generic_Itp D rho)
  -> lift_psi psi rho Psi
  -> WFcmd D' Psi (packS0 rho theta) pc (comml (_movwst o a)).
Proof.
  intros D rho psi pc [p g tag] D' Psi o a.
  simpl.
  intros Hi HDmono HD HD' Hlift.
  subst D'.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl; rewrite <- EqNat.beq_nat_refl.
  split_r with (SCAP_Mono_CT D rho p g tag HDmono).
  destruct Hi as [p' [g' [Hslk' [Hen Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  intros Psi' S HItp Hsub.
  rrsplit.
    left.
    unfold not, itr_enabled.
    intros [Hx _]; rewrite (SCAP_IF_disable HItp) in Hx; discriminate Hx.
  destruct S as [[[M R] fl] isr].
  simpl in HItp.
  destruct HItp as [Hp [Hfl Hwfst]].
  split.
    destruct (enable_elim Hen Hp) as [S' HnextS].
    exists S'; trivial.
  intros f' [[[M' R'] fl'] isr'] HnextS HnextPC.
  destruct (nextS_inversion HnextS) as [_ [_ [Hfl' Hisr']]].
  assert (Hpc' := nextPC_inversion HnextPC).
  simpl in Hpc'.
  subst fl' isr' f'.
  exists (packS0 rho (Ast p' g' tag)).
  split.
    apply Hsub.
    apply Hlift; trivial.
  unfold ext_dict.
  simpl.
  rewrite <- EqNat.beq_nat_refl.
  unfold Generic_Itp; simpl.
  assert (HS' : (p |> nextS pc (comml (_movwst o a))) (M', R', fl, isr)).
    apply linkPG_intro with (M,R,fl,isr); trivial.
  split.
    apply Hpp'; trivial.
  split; trivial.
  destruct Hwfst as [n Hwfst].
  exists n.
  apply (WFST_strengthen Hwfst (S1:=(M',R',fl,isr)) (g':=g')).
  intros S' [Hg' [H1 H2]].
  split.
    apply Hg'g.
    apply conjG_intro with (M', R', fl, isr); trivial.
    apply conjPG_intro; trivial.
  split_lr.
Qed.

Lemma SCAP_in_CAP0_pushwp : forall D rho psi pc theta D' Psi o,
  SCAP_WFcmd psi pc theta (comml (_pushwp o))
  -> Mono_CT_Itp D
  -> D rho = NoneT1 ITP
  -> D' = ext_dict D rho (Generic_Itp D rho)
  -> lift_psi psi rho Psi
  -> WFcmd D' Psi (packS0 rho theta) pc (comml (_pushwp o)).
Proof.
  intros D rho psi pc [p g tag] D' Psi o.
  simpl.
  intros Hi HDmono HD HD' Hlift.
  subst D'.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl; rewrite <- EqNat.beq_nat_refl.
  split_r with (SCAP_Mono_CT D rho p g tag HDmono).
  destruct Hi as [p' [g' [Hslk' [Hen Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  intros Psi' S HItp Hsub.
  rrsplit.
    left.
    unfold not, itr_enabled.
    intros [Hx _]; rewrite (SCAP_IF_disable HItp) in Hx; discriminate Hx.
  destruct S as [[[M R] fl] isr].
  simpl in HItp.
  destruct HItp as [Hp [Hfl Hwfst]].
  split.
    destruct (enable_elim Hen Hp) as [S' HnextS].
    exists S'; trivial.
  intros f' [[[M' R'] fl'] isr'] HnextS HnextPC.
  destruct (nextS_inversion HnextS) as [_ [_ [Hfl' Hisr']]].
  assert (Hpc' := nextPC_inversion HnextPC).
  simpl in Hpc'.
  subst fl' isr' f'.
  exists (packS0 rho (Ast p' g' tag)).
  split.
    apply Hsub.
    apply Hlift; trivial.
  unfold ext_dict.
  simpl.
  rewrite <- EqNat.beq_nat_refl.
  unfold Generic_Itp; simpl.
  assert (HS' : (p |> nextS pc (comml (_pushwp o))) (M', R', fl, isr)).
    apply linkPG_intro with (M,R,fl,isr); trivial.
  split.
    apply Hpp'; trivial.
  split; trivial.
  destruct Hwfst as [n Hwfst].
  exists n.
  apply (WFST_strengthen Hwfst (S1:=(M',R',fl,isr)) (g':=g')).
  intros S' [Hg' [H1 H2]].
  split.
    apply Hg'g.
    apply conjG_intro with (M', R', fl, isr); trivial.
    apply conjPG_intro; trivial.
  split_lr.
Qed.

Lemma SCAP_in_CAP0_pushwm : forall D rho psi pc theta D' Psi a,
  SCAP_WFcmd psi pc theta (comml (_pushwm a))
  -> Mono_CT_Itp D
  -> D rho = NoneT1 ITP
  -> D' = ext_dict D rho (Generic_Itp D rho)
  -> lift_psi psi rho Psi
  -> WFcmd D' Psi (packS0 rho theta) pc (comml (_pushwm a)).
Proof.
  intros D rho psi pc [p g tag] D' Psi a.
  simpl.
  intros Hi HDmono HD HD' Hlift.
  subst D'.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl; rewrite <- EqNat.beq_nat_refl.
  split_r with (SCAP_Mono_CT D rho p g tag HDmono).
  destruct Hi as [p' [g' [Hslk' [Hen Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  intros Psi' S HItp Hsub.
  rrsplit.
    left.
    unfold not, itr_enabled.
    intros [Hx _]; rewrite (SCAP_IF_disable HItp) in Hx; discriminate Hx.
  destruct S as [[[M R] fl] isr].
  simpl in HItp.
  destruct HItp as [Hp [Hfl Hwfst]].
  split.
    destruct (enable_elim Hen Hp) as [S' HnextS].
    exists S'; trivial.
  intros f' [[[M' R'] fl'] isr'] HnextS HnextPC.
  destruct (nextS_inversion HnextS) as [_ [_ [Hfl' Hisr']]].
  assert (Hpc' := nextPC_inversion HnextPC).
  simpl in Hpc'.
  subst fl' isr' f'.
  exists (packS0 rho (Ast p' g' tag)).
  split.
    apply Hsub.
    apply Hlift; trivial.
  unfold ext_dict.
  simpl.
  rewrite <- EqNat.beq_nat_refl.
  unfold Generic_Itp; simpl.
  assert (HS' : (p |> nextS pc (comml (_pushwm a))) (M', R', fl, isr)).
    apply linkPG_intro with (M,R,fl,isr); trivial.
  split.
    apply Hpp'; trivial.
  split_l.
  destruct Hwfst as [n Hwfst].
  exists n.
  apply (WFST_strengthen Hwfst (S1:=(M',R',fl,isr)) (g':=g')).
  intros S' [Hg' [H1 H2]].
  split.
    apply Hg'g.
    apply conjG_intro with (M', R', fl, isr); trivial.
    apply conjPG_intro; trivial.
  split_lr.
Qed.

Lemma SCAP_in_CAP0_popw : forall D rho psi pc theta D' Psi r,
  SCAP_WFcmd psi pc theta (comml (_popw r))
  -> Mono_CT_Itp D
  -> D rho = NoneT1 ITP
  -> D' = ext_dict D rho (Generic_Itp D rho)
  -> lift_psi psi rho Psi
  -> WFcmd D' Psi (packS0 rho theta) pc (comml (_popw r)).
Proof.
  intros D rho psi pc [p g tag] D' Psi r.
  simpl.
  intros Hi HDmono HD HD' Hlift.
  subst D'.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl; rewrite <- EqNat.beq_nat_refl.
  split_r with (SCAP_Mono_CT D rho p g tag HDmono).
  destruct Hi as [p' [g' [Hslk' [Hen Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  intros Psi' S HItp Hsub.
  rrsplit.
    left.
    unfold not, itr_enabled.
    intros [Hx _]; rewrite (SCAP_IF_disable HItp) in Hx; discriminate Hx.
  destruct S as [[[M R] fl] isr].
  simpl in HItp.
  destruct HItp as [Hp [Hfl Hwfst]].
  split.
    destruct (enable_elim Hen Hp) as [S' HnextS].
    exists S'; trivial.
  intros f' [[[M' R'] fl'] isr'] HnextS HnextPC.
  destruct (nextS_inversion HnextS) as [_ [_ [Hfl' Hisr']]].
  assert (Hpc' := nextPC_inversion HnextPC).
  simpl in Hpc'.
  subst fl' isr' f'.
  exists (packS0 rho (Ast p' g' tag)).
  split.
    apply Hsub.
    apply Hlift; trivial.
  unfold ext_dict.
  simpl.
  rewrite <- EqNat.beq_nat_refl.
  unfold Generic_Itp; simpl.
  assert (HS' : (p |> nextS pc (comml (_popw r))) (M', R', fl, isr)).
    apply linkPG_intro with (M,R,fl,isr); trivial.
  split.
    apply Hpp'; trivial.
  split_l.
  destruct Hwfst as [n Hwfst].
  exists n.
  apply (WFST_strengthen Hwfst (S1:=(M',R',fl,isr)) (g':=g')).
  intros S' [Hg' [H1 H2]].
  split.
    apply Hg'g.
    apply conjG_intro with (M', R', fl, isr); trivial.
    apply conjPG_intro; trivial.
  split_lr.
Qed.

Lemma SCAP_in_CAP0_cmpw : forall D rho psi pc theta D' Psi o r,
  SCAP_WFcmd psi pc theta (comml (_cmpw o r))
  -> Mono_CT_Itp D
  -> D rho = NoneT1 ITP
  -> D' = ext_dict D rho (Generic_Itp D rho)
  -> lift_psi psi rho Psi
  -> WFcmd D' Psi (packS0 rho theta) pc (comml (_cmpw o r)).
Proof.
  intros D rho psi pc [p g tag] D' Psi o r.
  simpl.
  intros Hi HDmono HD HD' Hlift.
  subst D'.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl; rewrite <- EqNat.beq_nat_refl.
  split_r with (SCAP_Mono_CT D rho p g tag HDmono).
  destruct Hi as [p' [g' [Hslk' [Hen Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  intros Psi' S HItp Hsub.
  rrsplit.
    left.
    unfold not, itr_enabled.
    intros [Hx _]; rewrite (SCAP_IF_disable HItp) in Hx; discriminate Hx.
  destruct S as [[[M R] fl] isr].
  simpl in HItp.
  destruct HItp as [Hp [Hfl Hwfst]].
  split.
    destruct (enable_elim Hen Hp) as [S' HnextS].
    exists S'; trivial.
  intros f' [[[M' R'] fl'] isr'] HnextS HnextPC.
  destruct (nextS_inversion HnextS) as [_ [_ [Hfl' Hisr']]].
  assert (Hpc' := nextPC_inversion HnextPC).
  simpl in Hpc'.
  subst isr' f'.
  exists (packS0 rho (Ast p' g' tag)).
  split.
    apply Hsub.
    apply Hlift; trivial.
  unfold ext_dict.
  simpl.
  rewrite <- EqNat.beq_nat_refl.
  unfold Generic_Itp; simpl.
  assert (HS' : (p |> nextS pc (comml (_cmpw o r))) (M', R', fl', isr)).
    apply linkPG_intro with (M,R,fl,isr); trivial.
  split.
    apply Hpp'; trivial.
  split.
    destruct Hfl' as [[Hr Hfl'] | [Hr Hfl']]; subst fl'; trivial.
  destruct Hwfst as [n Hwfst].
  exists n.
  apply (WFST_strengthen Hwfst (S1:=(M',R',fl',isr)) (g':=g')).
  intros S' [Hg' [H1 H2]].
  split.
    apply Hg'g.
    apply conjG_intro with (M', R', fl', isr); trivial.
    apply conjPG_intro; trivial.
  split_l.
  destruct Hfl' as [[Hr Hfl'] | [Hr Hfl']]; subst fl'; trivial.
Qed.

Lemma SCAP_in_CAP0_addw : forall D rho psi pc theta D' Psi o r,
  SCAP_WFcmd psi pc theta (comml (_addw o r))
  -> Mono_CT_Itp D
  -> D rho = NoneT1 ITP
  -> D' = ext_dict D rho (Generic_Itp D rho)
  -> lift_psi psi rho Psi
  -> WFcmd D' Psi (packS0 rho theta) pc (comml (_addw o r)).
Proof.
  intros D rho psi pc [p g tag] D' Psi o r.
  simpl.
  intros Hi HDmono HD HD' Hlift.
  subst D'.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl; rewrite <- EqNat.beq_nat_refl.
  split_r with (SCAP_Mono_CT D rho p g tag HDmono).
  destruct Hi as [p' [g' [Hslk' [Hen Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  intros Psi' S HItp Hsub.
  rrsplit.
    left.
    unfold not, itr_enabled.
    intros [Hx _]; rewrite (SCAP_IF_disable HItp) in Hx; discriminate Hx.
  destruct S as [[[M R] fl] isr].
  simpl in HItp.
  destruct HItp as [Hp [Hfl Hwfst]].
  split.
    destruct (enable_elim Hen Hp) as [S' HnextS].
    exists S'; trivial.
  intros f' [[[M' R'] fl'] isr'] HnextS HnextPC.
  destruct (nextS_inversion HnextS) as [_ [_ [Hfl' Hisr']]].
  assert (Hpc' := nextPC_inversion HnextPC).
  simpl in Hpc'.
  subst isr' f'.
  exists (packS0 rho (Ast p' g' tag)).
  split.
    apply Hsub.
    apply Hlift; trivial.
  unfold ext_dict.
  simpl.
  rewrite <- EqNat.beq_nat_refl.
  unfold Generic_Itp; simpl.
  assert (HS' : (p |> nextS pc (comml (_addw o r))) (M', R', fl', isr)).
    apply linkPG_intro with (M,R,fl,isr); trivial.
  split.
    apply Hpp'; trivial.
  split.
    destruct Hfl'; subst fl'; trivial.
  destruct Hwfst as [n Hwfst].
  exists n.
  apply (WFST_strengthen Hwfst (S1:=(M',R',fl',isr)) (g':=g')).
  intros S' [Hg' [H1 H2]].
  split.
    apply Hg'g.
    apply conjG_intro with (M', R', fl', isr); trivial.
    apply conjPG_intro; trivial.
  split_l.
  destruct Hfl'; subst fl'; trivial.
Qed.

Lemma SCAP_in_CAP0_subw : forall D rho psi pc theta D' Psi o r,
  SCAP_WFcmd psi pc theta (comml (_subw o r))
  -> Mono_CT_Itp D
  -> D rho = NoneT1 ITP
  -> D' = ext_dict D rho (Generic_Itp D rho)
  -> lift_psi psi rho Psi
  -> WFcmd D' Psi (packS0 rho theta) pc (comml (_subw o r)).
Proof.
  intros D rho psi pc [p g tag] D' Psi o r.
  simpl.
  intros Hi HDmono HD HD' Hlift.
  subst D'.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl; rewrite <- EqNat.beq_nat_refl.
  split_r with (SCAP_Mono_CT D rho p g tag HDmono).
  destruct Hi as [p' [g' [Hslk' [Hen Hi]]]].
  destruct Hi as [Hpp' Hg'g].
  intros Psi' S HItp Hsub.
  rrsplit.
    left.
    unfold not, itr_enabled.
    intros [Hx _]; rewrite (SCAP_IF_disable HItp) in Hx; discriminate Hx.
  destruct S as [[[M R] fl] isr].
  simpl in HItp.
  destruct HItp as [Hp [Hfl Hwfst]].
  split.
    destruct (enable_elim Hen Hp) as [S' HnextS].
    exists S'; trivial.
  intros f' [[[M' R'] fl'] isr'] HnextS HnextPC.
  destruct (nextS_inversion HnextS) as [_ [_ [Hfl' Hisr']]].
  assert (Hpc' := nextPC_inversion HnextPC).
  simpl in Hpc'.
  subst isr' f'.
  exists (packS0 rho (Ast p' g' tag)).
  split.
    apply Hsub.
    apply Hlift; trivial.
  unfold ext_dict.
  simpl.
  rewrite <- EqNat.beq_nat_refl.
  unfold Generic_Itp; simpl.
  assert (HS' : (p |> nextS pc (comml (_subw o r))) (M', R', fl', isr)).
    apply linkPG_intro with (M,R,fl,isr); trivial.
  split.
    apply Hpp'; trivial.
  split.
    destruct Hfl'; subst fl'; trivial.
  destruct Hwfst as [n Hwfst].
  exists n.
  apply (WFST_strengthen Hwfst (S1:=(M',R',fl',isr)) (g':=g')).
  intros S' [Hg' [H1 H2]].
  split.
    apply Hg'g.
    apply conjG_intro with (M', R', fl', isr); trivial.
    apply conjPG_intro; trivial.
  split_l.
  destruct Hfl'; subst fl'; trivial.
Qed.

Lemma SCAP_in_CAP0_jmp : forall D rho psi pc theta D' Psi f,
  SCAP_WFcmd psi pc theta (comms (_jmp f))
  -> Mono_CT_Itp D
  -> D rho = NoneT1 ITP
  -> D' = ext_dict D rho (Generic_Itp D rho)
  -> lift_psi psi rho Psi
  -> WFcmd D' Psi (packS0 rho theta) pc (comms (_jmp f)).
Proof.
  intros D rho psi pc [p g tag] D' Psi f.
  simpl.
  intros Hi HDmono HD HD' Hlift.
  subst D'.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl; rewrite <- EqNat.beq_nat_refl.
  split_r with (SCAP_Mono_CT D rho p g tag HDmono).
  destruct Hi as [p' [g' [Hslk' [Hpp' Hg'g]]]].
  intros Psi' S HItp Hsub.
  rrsplit.
    left.
    unfold not, itr_enabled.
    intros [Hx _]; rewrite (SCAP_IF_disable HItp) in Hx; discriminate Hx.
  destruct S as [[[M R] fl] isr].
  simpl in HItp.
  destruct HItp as [Hp [Hfl Hwfst]].
  split.
    exists (M,R,fl,isr); constructor.
  intros f' [[[M' R'] fl'] isr'] HnextS HnextPC.
  destruct (nextS_inversion HnextS) as [HM' [HR' [Hfl' Hisr']]].
  assert (Hpc' := nextPC_inversion HnextPC).
  simpl in Hpc'.
  subst M' R' fl' isr' f'.
  exists (packS0 rho (Ast p' g' tag)).
  split.
    apply Hsub.
    apply Hlift; trivial.
  unfold ext_dict.
  simpl.
  rewrite <- EqNat.beq_nat_refl.
  unfold Generic_Itp; simpl.
  assert (HS' : (p |> nextS pc (comms (_jmp f))) (M, R, fl, isr)).
    apply linkPG_intro with (M,R,fl,isr); trivial.
  split.
    apply Hpp'; trivial.
  split_l.
  destruct Hwfst as [n Hwfst].
  exists n.
  apply (WFST_strengthen Hwfst (S1:=(M,R,fl,isr)) (g':=g')).
  intros S' [Hg' [H1 H2]].
  split.
    apply Hg'g.
    apply conjPG_intro; trivial.
  split_lr.
Qed.

(* unfinished *)

Lemma SCAP_in_CAP0_retw : forall D rho psi pc theta D' Psi,
  SCAP_WFcmd psi pc theta (comms (_retw))
  -> Mono_CT_Itp D
  -> D rho = NoneT1 ITP
  -> D' = ext_dict D rho (Generic_Itp D rho)
  -> lift_psi psi rho Psi
  -> WFcmd D' Psi (packS0 rho theta) pc (comms (_retw)).
Proof.
  intros D rho psi pc [p g tag] D' Psi.
  simpl.
  intros Hi HDmono HD HD' Hlift.
  subst D'.
  unfold WFcmd, WFcmd_aux.
  unfold ext_dict; simpl; rewrite <- EqNat.beq_nat_refl.
  split_r with (SCAP_Mono_CT D rho p g tag HDmono).
  intros Psi' S HItp Hsub.
  rrsplit.
    left.
    unfold not, itr_enabled.
    intros [Hx _]; rewrite (SCAP_IF_disable HItp) in Hx; discriminate Hx.
  destruct S as [[[M R] fl] isr].
  simpl in HItp.
  destruct HItp as [Hp [Hfl Hwfst]].
  assert (Hg : scap_lift_g g (M, R, fl, isr) (M, R, fl, isr)).
    split.
      apply Hi; trivial.
    split_lr.
  split.
    destruct Hwfst as [n Hwfst].
    destruct n.
      simpl in Hwfst.
      substH Hwfst with (Hwfst (M, R, fl, isr) Hg).
      destruct Hwfst as [ra [Hra Hwfst]].
      unfold Ret_Addr in Hra.
      substH Hra with (Hra (R SP)).
      exists (M, RF_update R SP (R SP + 2), fl, isr).
      apply nextS_retw.
      inversion_clear Hra.
      unfold popW.
      subst sp.
      simplS in * .
      exists ra.
      split_lr.
    substH Hwfst with (Hwfst (M, R, fl, isr) Hg).
    destruct Hwfst as [ra [Hra Hwfst]].
    unfold Ret_Addr in Hra.
    substH Hra with (Hra (R SP)).
    exists (M, RF_update R SP (R SP + 2), fl, isr).
    apply nextS_retw.
    inversion_clear Hra.
    unfold popW.
    subst sp.
    simplS in * .
    exists ra.
    split_lr.
  intros f' [[[M' R'] fl'] isr'] HnextS HnextPC.
  destruct (nextS_inversion HnextS) as [HM' [HR' [Hfl' Hisr']]].
  subst fl' isr'.
  destruct Hwfst as [n Hwfst].
  destruct n as [ | n].
    simpl in Hwfst.
    substH Hwfst with (Hwfst (M,R,fl,isr) Hg).
    destruct Hwfst as [ra [Hra [theta0 [Hlk Hwfst]]]].
    substH Hwfst with (Hwfst pc (M',R',fl,isr) HnextS).
    substH Hra with (Hra pc).
    assert (Heq := nextPC_injection HnextPC Hra).
    subst f'.
    clear Hra.
    exists theta0; split_l.
    apply WFstate_dict_enlarge with D; trivial.
    apply (ext_dict_subsetD _ _ (Generic_Itp D rho) HD).
  simpl in Hwfst.
  substH Hwfst with (Hwfst (M,R,fl,isr) Hg).
  destruct Hwfst as [ra [Hra [p' [g' [tag' [Hrain Hwfst]]]]]].
  substH Hra with (Hra pc).
  assert (Heq := nextPC_injection HnextPC Hra).
  subst f'.
  clear Hra.
  substH Hwfst with (Hwfst pc (M',R',fl,isr) HnextS).
  destruct Hwfst as [Hp'S' Hwfst].
  exists (packS0 rho (Ast p' g' tag')).
  split_l.
  unfold ext_dict; simpl.
  rewrite <- EqNat.beq_nat_refl.
  unfold Generic_Itp; simpl.
  split_l.
  split_l.
  exists n; trivial.
Qed.

Lemma SCAP_in_CAP0_cmd_aux:
  forall D rho psi pc theta c D' Psi,
    SCAP_WFcmd psi pc theta c
    -> Mono_CT_Itp D
    -> D rho = NoneT1 ITP
    -> D' = ext_dict D rho (Generic_Itp D rho)
    -> lift_psi psi rho Psi
    -> WFcmd D' Psi (packS0 rho theta) pc c.
Proof.
  intros D rho psi pc [p g tag] c D' Psi.
  intros Hi HDmono HD HD' Hlift.
  induction c as [ i | s]; [ induction i | induction s].
  (* 19 goals remain, i = cli *)
  destruct Hi.
  (* 18 goals remain, i = sti *)
  destruct Hi.
  (* 17 goals remain, i = call f *)
  eapply SCAP_in_CAP0_call; eauto.
  (* 16 goals remain, i = jne f *)
  eapply SCAP_in_CAP0_jne; eauto.
  (* 15 goals remain, i = movwp o r *)
  eapply SCAP_in_CAP0_movwp; eauto.
  (* 14 goals remain, i = movwld a r *)
  eapply SCAP_in_CAP0_movwld; eauto.
  (* 13 goals remain, i = movwst o a *)
  eapply SCAP_in_CAP0_movwst; eauto.
  (* 12 goals remain, i = pushwp o *)
  eapply SCAP_in_CAP0_pushwp; eauto.
  (* 11 goals remain, i = pushwm a *)
  eapply SCAP_in_CAP0_pushwm; eauto.
  (* 10 goals remain, i = pushfw *)
  destruct Hi.
  (* 9 goals remain, i = popw r *)
  eapply SCAP_in_CAP0_popw; eauto.
  (* 8 goals remain, i = popfw *)
  destruct Hi.
  (* 7 goals remain, i = cmpw o r *)
  eapply SCAP_in_CAP0_cmpw; eauto.
  (* 6 goals remain, i = addw o r *)
  eapply SCAP_in_CAP0_addw; eauto.
  (* 5 goals remain, i = subw o r *)
  eapply SCAP_in_CAP0_subw; eauto.
  (* 4 goals remain, i = outb eoi *)
  destruct Hi.
  (* 3 goals remain, i = jmp f *)
  eapply SCAP_in_CAP0_jmp; eauto.
  (* 2 goals remain, i = retw *)
  eapply SCAP_in_CAP0_retw; eauto.
  (* 1 goals remain, i = iret *)
  destruct Hi.
Qed.
Implicit Arguments SCAP_in_CAP0_cmd_aux [D rho psi pc theta c D' Psi].

Lemma SCAP_in_CAP0_iseq:
  forall D rho psi pc theta isq D' Psi,
    SCAP_WFiseq psi pc theta isq
    -> Mono_CT_Itp D
    -> D rho = NoneT1 ITP
    -> D' = ext_dict D rho (Generic_Itp D rho)
    -> lift_psi psi rho Psi
    -> WFiseq D' Psi (packS0 rho theta) pc isq.
Proof.
  intros D rho psi pc theta isq D' Psi Hscap HDmono HD HD' Hlift.
  gen_clear theta pc Hscap.
  induction isq; intros theta pc Hscap.
    inversion Hscap. subst psi0 pc0 theta i0 isq0.
    eapply wfis_seq with (packS0 rho (Ast p' g' tag)).
    assert (Hlift' := lift_psi_cons (f := pc+1) (spec := Ast p' g' tag) Hlift).
    apply (SCAP_in_CAP0_cmd_aux H4 HDmono HD HD' Hlift').
    apply IHisq; trivial.
  inversion Hscap. subst psi0 pc0 theta si.
  eapply wfis_end.
  apply (SCAP_in_CAP0_cmd_aux H3 HDmono HD HD' Hlift).
Qed.

Lemma SCAP_in_CAP0_cdhp:
  forall D rho psi C psi' D' Psi Psi',
    SCAP_WFcdhp psi C psi'
    -> Mono_CT_Itp D
    -> D rho = NoneT1 ITP
    -> D' = ext_dict D rho (Generic_Itp D rho)
    -> lift_psi psi rho Psi
    -> lift_psi psi' rho Psi'
    -> sink_psi psi' rho Psi'
    -> WFcdhp D' Psi C Psi'.
Proof.
  intros D rho psi C psi' D' Psi Psi' H HDmono H0 H1 H2 H3 H4; 
    unfold SCAP_WFcdhp, WFcdhp; intros f theta0 Hf.
  substH H4 with (H4 f theta0 Hf).
  destruct H4 as [theta [Htheta0 Hfx]].
  unfold SCAP_WFcdhp in H.
  substH H with (H f theta Hfx).
  destruct H as [isq [Hisq H]].
  exists isq; split_l.
  subst theta0.
  eapply SCAP_in_CAP0_iseq; eauto.
Qed.
