(***************************************************************)
(* Coq code for AIM                                            *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

Add LoadPath ".." .
Add LoadPath "../CAP0" .
Add LoadPath "../SCAP" .
Add LoadPath "../miniOS" .

Require Import Machine.
Require Import Operational_Semantics.
Require Import AIM.

Require Import Kernel.

Definition L_AVAIL   := 0.
Definition L_WAITING := 2.

Notation Local " i ; I " := (seq i I)
  (at level 60, right associativity).

Notation Local " i ;; " := (sseq i)
  (at level 60).

(*************************************
 * Source code for lock release
 *************************************

# lock_release(l)
# precondition: interrupt is enabled
lock_release:
        pushw   %bp
        movw    %sp, %bp
        pushw   4(%bp)
        cli
        call    l_release
        sti
        addw    $2, %sp
        popw    %bp
        retw

l_release:
        pushw   %bp
        movw    %sp, %bp
        movw    4(%bp), %bx             # bx: lock pointer
        pushw   L_WAITING(%bx)
        call    unblock
        addw    $2, %sp
        cmpw    $0, %ax
        jne     l_rel_done
        movw    %sp, %bp
        movw    4(%bp), %bx
        movw    $1, L_AVAIL(%bx)
l_rel_done:
        popw    %bp
        retw


 ************************************)

Parameters lock_release l_release : Label.

Axiom l_release_not_prim : ~ prim_entry l_release.

Definition l_rel_done : Label := l_release + 10.

Definition I_l_rel_done : iseq :=
(* <l_rel_done + 0> : *) _popw BP;
(* <l_rel_done + 1> : *) _retw;; .

Definition I_l_release : iseq :=
(* <l_release + 0> : *) _pushwp (Opr_reg BP);
(* <l_release + 1> : *) _movwp  (Opr_reg SP) BP;
(* <l_release + 2> : *) _movwld (A_reg 4 BP) BX;
(* <l_release + 3> : *) _pushwm (A_reg L_WAITING BX);
(* <l_release + 4> : *) _call   _UNBLK;
(* <l_release + 5> : *) _addw   (Opr_imm 2) SP;
(* <l_release + 6> : *) _cmpw   (Opr_imm 0) AX;
(* <l_release + 7> : *) _jne    l_rel_done;
(* <l_release + 8> : *) _movwp  (Opr_reg SP) BP;
(* <l_release + 9> : *) _movwld (A_reg 4 BP) BX;
(* <l_release + a> : *) _movwst (Opr_imm 1) (A_reg L_AVAIL BX);
(* <l_release + b> : *) I_l_rel_done .


Definition I_lock_release: iseq :=
(* <lock_release + 0> : *) _pushwp (Opr_reg BP);
(* <lock_release + 1> : *) _movwp  (Opr_reg SP) BP;
(* <lock_release + 2> : *) _pushwm (A_reg 4 BP);
(* <lock_release + 3> : *) _cli;
(* <lock_release + 4> : *) _call   l_release;
(* <lock_release + 5> : *) _sti;
(* <lock_release + 6> : *) _addw   (Opr_imm 2) SP;
(* <lock_release + 7> : *) _popw   BP;
(* <lock_release + 8> : *) _retw;; .





(*************************************
 * Source code for lock acquire
 *************************************

lock_acquire:
        pushw   %bp
        movw    %sp, %bp
        pushw   4(%bp)
        cli
        call    l_acquire
        sti
        addw    $2, %sp
        popw    %bp
        retw

# precondition: interrupt is disabled
l_acquire:
        pushw   %bp
        movw    %sp, %bp
        movw    4(%bp), %bx             # lock pointer
        movw    L_AVAIL(%bx), %dx
        cmpw    $0, %dx
        jne     l_setacq
        pushw   L_WAITING(%bx)
        call    sleep
        addw    $2, %sp
        popw    %bp
        retw

l_setacq:
        movw    $0, L_AVAIL(%bx)
        popw    %bp
        retw

*************************************)

Parameters l_setacq l_acquire lock_acquire: Label.

Axiom l_acquire_not_prim : ~ prim_entry l_acquire.

Definition I_l_acquire:=
(* <l_acquire + 0> : *)  _pushwp (Opr_reg BP);
(* <l_acquire + 1> : *)  _movwp  (Opr_reg SP) BP;
(* <l_acquire + 2> : *)  _movwld (A_reg 4 BP) BX;
(* <l_acquire + 3> : *)  _movwld (A_reg L_AVAIL BX) DX;
(* <l_acquire + 4> : *)  _cmpw   (Opr_imm 0) DX;
(* <l_acquire + 5> : *)  _jne    l_setacq;
(* <l_acquire + 6> : *)  _pushwm (A_reg L_WAITING BX);
(* <l_acquire + 7> : *)  _call   L_sleep;
(* <l_acquire + 8> : *)  _addw   (Opr_imm 2) SP;
(* <l_acquire + 9> : *)  _popw   BP;
(* <l_acquire + a> : *)  _retw;; .


Definition I_l_setacq:=
(* <l_setacq + 0> : *)  _movwst (Opr_imm 0) (A_reg L_AVAIL BX);
(* <l_setacq + 1> : *)  _popw   BP;
(* <l_setacq + 2> : *)  _retw;; .


Definition I_lock_acquire :=
(* <lock_acquire + 0> : *)  _pushwp (Opr_reg BP);
(* <lock_acquire + 1> : *)  _movwp  (Opr_reg SP) BP;
(* <lock_acquire + 2> : *)  _pushwm (A_reg 4 BP);
(* <lock_acquire + 3> : *)  _cli;
(* <lock_acquire + 4> : *)  _call   l_acquire;
(* <lock_acquire + 5> : *)  _sti;
(* <lock_acquire + 6> : *)  _addw   (Opr_imm 2) SP;
(* <lock_acquire + 7> : *)  _popw   BP;
(* <lock_acquire + 8> : *)  _retw;; .





(*******************************************
 * Source code for cond var : Mesa style
 *******************************************

        .set    C_WAITING, 0

# cond_wait(l, cv)
# precondition: interrupt is enabled
cond_wait:
        pushw   %bp
        movw    %sp, %bp
        pushw   6(%bp)                  # lock pointer
        cli
        call    l_release
        movw    4(%bp), %bx
        pushw   C_WAITING(%bx)
        call    sleep
        sti
        addw    $2, %sp
        cli
        call    l_acquire
        sti
        addw    $2, %sp
        popw    %bp
        retw

# mesa style implementation
# cond_signal(cv)
# precondition: interrupt is enabled
cond_signal:
        pushw   %bp
        movw    %sp, %bp
        movw    4(%bp), %bx
        cli
        pushw   C_WAITING(%bx)
        call    unblock
        sti
        addw    $2, %sp
        popw    %bp
        retw

********************************************)

Definition C_WAITING := 0.

Parameters cond_wait cond_signal : Label.

Definition I_cond_wait :=
(* <cond_wait + 0> : *) _pushwp (Opr_reg BP);
(* <cond_wait + 1> : *) _movwp  (Opr_reg SP) BP;
(* <cond_wait + 2> : *) _pushwm (A_reg 6 BP);
(* <cond_wait + 3> : *) _cli;
(* <cond_wait + 4> : *) _call   l_release;
(* <cond_wait + 5> : *) _movwld (A_reg 4 BP) BX;
(* <cond_wait + 6> : *) _pushwm (A_reg C_WAITING BX);
(* <cond_wait + 7> : *) _call   L_sleep;
(* <cond_wait + 8> : *) _sti;
(* <cond_wait + 9> : *) _addw   (Opr_imm 2) SP;
(* <cond_wait + a> : *) _cli;
(* <cond_wait + b> : *) _call   l_acquire;
(* <cond_wait + c> : *) _sti;
(* <cond_wait + d> : *) _addw   (Opr_imm 2) SP;
(* <cond_wait + e> : *) _popw   BP;
(* <cond_wait + f> : *) _retw;; .

Definition I_cond_signal :=
(* <cond_signal + 0> : *) _pushwp (Opr_reg BP);
(* <cond_signal + 1> : *) _movwp  (Opr_reg SP) BP;
(* <cond_signal + 2> : *) _movwld (A_reg 4 BP) BX;
(* <cond_signal + 3> : *) _cli;
(* <cond_signal + 4> : *) _pushwm (A_reg C_WAITING BX);
(* <cond_signal + 5> : *) _call   _UNBLK;
(* <cond_signal + 6> : *) _sti;
(* <cond_signal + 7> : *) _addw   (Opr_imm 2) SP;
(* <cond_signal + 8> : *) _popw   BP;
(* <cond_signal + 9> : *) _retw;; .
