(***************************************************************)
(* Coq code for AIM                                            *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

Add LoadPath "..".
Add LoadPath "../CAP0".
Add LoadPath "../SCAP".
Add LoadPath "../miniOS".

Require Import Coqlib.

Require Import List.
Require Import Arith.
Require Import Machine.
Require Import Assertion.
Require Import Assertion_Lemmas.
Require Import Operational_Semantics.

Require Import AIM.
Require Import QLib.
Require Import Assertion_Lemmas4.
Require Import SCAP.
Require Import SCAP_READY.
Require Import Linking.
Require Import Assembly.
Require Import Omega.
Require Import Arith_Lemmas.

Require Import Kernel.
Require Import Classical.

Lemma Memp_dec:
  forall M, M = MemMod.emp \/ M <> MemMod.emp.
Proof.
  intro.
  apply classic.
Qed.


Lemma not_Memp_indom:
  forall M, M <> MemMod.emp -> exists l, MemMod.in_dom l M.
Proof.
  intros.
  assert ((forall l, MemMod.not_in_dom l M) \/ ~ (forall l, MemMod.not_in_dom l M)).
  apply classic.
  inversion H0.
  assert (M = MemMod.emp).
  apply MemMod.extensional_eqMap.
  intro.
  unfold MemMod.emp.
  assert (MemMod.not_in_dom x M).
  apply H1. unfold MemMod.not_in_dom in H2. auto.
  contradiction.
  assert (exists l, ~ MemMod.not_in_dom l M).
  apply (not_all_ex_not MemMod.A (fun l => MemMod.not_in_dom l M)). auto.
  inversion H2.
  exists x.
  auto with PMap.
Qed.

Lemma disj_eq_emp:
forall M1 M2, MemMod.disjoint M1 M2 -> M1 = M2 -> M1 = MemMod.emp.
Proof.
  intros.
  assert (M1 = MemMod.emp \/ M1 <> MemMod.emp). apply Memp_dec.
  inversion_clear H1; auto.
  assert (exists l, MemMod.in_dom l M1).
  apply not_Memp_indom. auto.
  unfold MemMod.disjoint in H.
  inversion H1.
  assert (MemMod.not_in_dom x M1 \/ MemMod.not_in_dom x M2).
  apply H.
  inversion_clear H4.
  absurd (~ MemMod.in_dom x M1); auto with PMap.
  rewrite <- H0 in H5.
  absurd (~ MemMod.in_dom x M1); auto with PMap.
Qed.


(* Note here we define Gamma as a total mapping  *)
(* if a lock l is not specified in Gamma, we let *)
(* Gamma l = Aemp                                *)
Parameter Gamma : Label -> MPred.

Parameter precGamma : forall l, precise (Gamma l).

Definition Lock_INV_H_aux (l : Label) (b w : int) : MPred :=
  fun (M : Mem) => 
    ((l + L_AVAIL |->w b) *^* (l + L_WAITING |->w w) *^* (choice (b=0) Aemp (Gamma l))) M
      /\ Delta w = SomeT (Gamma l).

(* Hoare style invariant assotiated with lock l *)
Definition Lock_INV_H (l :Label) : MPred := 
  fun (M : Mem) =>
    exists b, exists w, 
    ((l + L_AVAIL |->w b) *^* (l + L_WAITING |->w w) *^* (choice (b=0) Aemp (Gamma l))) M
      /\ Delta w = SomeT (Gamma l).

Definition Lock_INV_H_0 (l : Label) : MPred :=
  fun (M : Mem) =>
    exists w, 
    ((l + L_AVAIL |->w 0) *^* (l + L_WAITING |->w w)) M
      /\ Delta w = SomeT (Gamma l).

Definition Lock_INV_H_1 (l :Label) : MPred := 
  fun (M : Mem) =>
    exists w, 
    ((l + L_AVAIL |->w 1) *^* (l + L_WAITING |->w w) *^* (Gamma l)) M
      /\ Delta w = SomeT (Gamma l).

Definition LockSet := list Label.

(* Hoare style, Mesa style and spin locks *)
Parameters lockSet_H lockSet_M lockSet_Spin : LockSet.

Fixpoint WF_lockSet_H (ls : LockSet) {struct ls} : MPred :=
  match ls with
    nil => Aemp
  | l :: ls' => (Lock_INV_H l) *^* WF_lockSet_H ls'
  end.


(* set of condition variables *)
Definition CVSet := list Label.

(* Mesa-style condition variables *)
Parameter cvSet_M : CVSet.

Definition CV_INV_M (cv : Label) : MPred :=
  fun (M : Mem) =>
    exists w, (cv + C_WAITING |->w w) M /\ Delta w = SomeT Aemp.

Fixpoint removeFirstL (l : Label) (ls : LockSet) {struct ls} : LockSet :=
  match ls with
    nil => nil
  | l' :: ls' => if (eq_nat_dec l l') then ls' else l' :: (removeFirstL l ls')
  end.

  
Lemma WF_LockSet_decomp:
  forall l ls,
    In l ls -> leqvM (WF_lockSet_H ls) ((Lock_INV_H l) *^* (WF_lockSet_H (removeFirstL l ls))).
Proof.
  intros.
  induction ls.
  unfold In in H.
  absurd False; auto.
  unfold In in H.
  inversion_clear H.
  subst a. clear IHls.
  unfold leqvM.
  intros.
  split.
  intro WFLset.
  unfold WF_lockSet_H in WFLset.
  fold WF_lockSet_H in WFLset.
  unfold removeFirstL.
  induction (eq_nat_dec l l). auto.
  absurd (l = l); auto.
  intro.
  unfold removeFirstL in H.
  induction (eq_nat_dec l l).
  unfold WF_lockSet_H. fold WF_lockSet_H. auto.
  absurd (l = l); auto.
  assert (leqvM (WF_lockSet_H ls)
           (Lock_INV_H l *^* WF_lockSet_H (removeFirstL l ls))).
  apply IHls; auto.
  unfold leqvM in *.
  unfold removeFirstL.
  induction (eq_nat_dec l a).
  rewrite <- a0.
  intro h.
  unfold WF_lockSet_H at 1. fold WF_lockSet_H. split; auto.
  fold removeFirstL.
  unfold WF_lockSet_H at 2. fold WF_lockSet_H.
  intro.
  unfold WF_lockSet_H at 1. fold WF_lockSet_H.
  split.
  intro.
  sep on.
  sep lift 1.
  destruct H1 as [M1 [M2 [HMsplit [HM1 HM2]]]].
  exists M1. exists M2. split_l.
  split_l.
  assert (WF_lockSet_H ls M2 <->
      (Lock_INV_H l *^* WF_lockSet_H (removeFirstL l ls)) M2).
  apply H.
  inversion_clear H1.
  assert ((Lock_INV_H l *^* WF_lockSet_H (removeFirstL l ls)) M2).
  apply H2; auto.
  sep on.
  sep matchH H1 with 2.
  intro.
  sep on.
  sep lift 1 in H1.
  destruct H1 as [M1 [M2 [HMsplit [HM1 HM2]]]].
  exists M1. exists M2. split_l.
  split_l.
  assert (WF_lockSet_H ls M2 <->
      (Lock_INV_H l *^* WF_lockSet_H (removeFirstL l ls)) M2).
  apply H.
  inversion_clear H1.
  apply H3.
  sep on.
  sep matchH HM2 with 2.
Qed.

Fixpoint WF_cvSet_M (cvs : CVSet) {struct cvs} : MPred :=
  match cvs with
    nil => Aemp
  | cv :: cvs' => (CV_INV_M cv) *^* WF_cvSet_M cvs'
  end.

Parameter lock_inv : 
  exists m, leqvM INV1 ((WF_lockSet_H lockSet_H) *^* (WF_cvSet_M cvSet_M) *^* m).


Fixpoint removeFirstCV (l : Label) (cvs : CVSet) {struct cvs} : LockSet :=
  match cvs with
    nil => nil
  | l' :: cvs' => if (eq_nat_dec l l') then cvs' else l' :: (removeFirstCV l cvs')
  end.

  
Lemma WF_CVSet_decomp:
  forall cv cvs,
    In cv cvs -> leqvM (WF_cvSet_M cvs) ((CV_INV_M cv) *^* (WF_cvSet_M (removeFirstCV cv cvs))).
Proof.
  intros.
  induction cvs.
  unfold In in H.
  absurd False; auto.
  unfold In in H.
  inversion_clear H.
  subst a. clear IHcvs.
  unfold leqvM.
  intro.
  split.
  intro WFCVset.
  unfold WF_cvSet_M in WFCVset.
  fold WF_cvSet_M in WFCVset.
  unfold removeFirstCV.
  induction (eq_nat_dec cv cv). auto.
  absurd (cv = cv); auto.
  intro.
  unfold removeFirstCV in H.
  induction (eq_nat_dec cv cv).
  unfold WF_cvSet_M. fold WF_cvSet_M. auto.
  absurd (cv = cv); auto.
  assert (leqvM (WF_cvSet_M cvs)
           (CV_INV_M cv *^* WF_cvSet_M (removeFirstCV cv cvs))).
  apply IHcvs; auto.
  unfold leqvM in *.
  unfold removeFirstCV.
  induction (eq_nat_dec cv a).
  rewrite <- a0.
  intro h.
  unfold WF_cvSet_M at 1. fold WF_cvSet_M. split; auto.
  fold removeFirstCV.
  unfold WF_cvSet_M. fold WF_cvSet_M.
  intro.
  split.
  intro.
  sep on.
  sep lift 1.
  destruct H1 as [M1 [M2 [HMsplit [HM1 HM2]]]].
  exists M1. exists M2. split_l.
  split_l.
  assert (WF_cvSet_M cvs M2 <->
      (CV_INV_M cv *^* WF_cvSet_M (removeFirstCV cv cvs)) M2).
  apply H.
  inversion_clear H1.
  assert ((CV_INV_M cv *^* WF_cvSet_M (removeFirstCV cv cvs)) M2).
  apply H2; auto.
  sep on.
  sep matchH H1 with 2.
  intro.
  sep on.
  sep lift 1 in H1.
  destruct H1 as [M1 [M2 [HMsplit [HM1 HM2]]]].
  exists M1. exists M2. split_l.
  split_l.
  assert (WF_cvSet_M cvs M2 <->
      (CV_INV_M cv *^* WF_cvSet_M (removeFirstCV cv cvs)) M2).
  apply H.
  inversion_clear H1.
  apply H3.
  sep on.
  sep matchH HM2 with 2.
Qed.

Lemma decomp_ex_Lock_INV_H:
forall M l m,
  M ||- Lock_INV_H l *^* m
    -> exists b, exists w, 
          (M ||- (l + L_AVAIL |->w b) *^* (l + L_WAITING |->w w) *^* (choice (b=0) Aemp (Gamma l)) *^* m)
          /\ Delta w = SomeT (Gamma l).
Proof.
  intros.
  unfold mjudge in *.
  unfold mstar in H.
  InvertAll.
  unfold Lock_INV_H in H.
  InvertAll.
  exists x1. exists x2.
  split; auto.
  unfold mstar at 1.
  exists x. exists x0. auto.
Qed.

Lemma decomp_ex_Lock_INV_H1:
forall M l m,
    (exists b, exists w, 
          (M ||- (l + L_AVAIL |->w b) *^* (l + L_WAITING |->w w) *^* (choice (b=0) Aemp (Gamma l)) *^* m)
          /\ Delta w = SomeT (Gamma l))
    -> M ||- Lock_INV_H l *^* m.
Proof.
  intros.
  unfold mjudge in *.
  unfold mstar at 1 in H.
  InvertAll.
  unfold mstar. exists x1. exists x2.
  split_l. split; auto.
  unfold Lock_INV_H.
  exists x. exists x0.
  split; auto.
Qed.



Lemma INV1_cvInv:
  forall M cv,
    (INV1 *^* mTrue) M 
    -> In cv cvSet_M 
    -> (CV_INV_M cv *^* mTrue) M.
Proof.
  intros.
  assert (exists m, leqvM INV1 ((WF_lockSet_H lockSet_H) *^* (WF_cvSet_M cvSet_M) *^* m)).
  apply lock_inv.
  destruct H1 as [mP Heqv].
  sep on.
  sep leqv_subst in H with (WF_lockSet_H lockSet_H *^* WF_cvSet_M cvSet_M *^* mP).
  assert (leqvM (WF_cvSet_M cvSet_M) ((CV_INV_M cv) *^* (WF_cvSet_M (removeFirstCV cv cvSet_M)))).
  apply WF_CVSet_decomp; auto.
  sep on.
  sep lift 1 in H.
  sep leqv_subst in H with (CV_INV_M cv *^* WF_cvSet_M (removeFirstCV cv cvSet_M)).
  sep on. sep lift 0 in H.
  sep matchH H with 2.
Qed.


Lemma INV1_decom_l:
  forall l, In l lockSet_H -> exists m, leqvM INV1 (Lock_INV_H l *^* m).
Proof.
  intros.
  assert (exists m, leqvM INV1 ((WF_lockSet_H lockSet_H) *^* (WF_cvSet_M cvSet_M) *^* m)).
  apply lock_inv.
  destruct H0 as [mP Heqv].
  assert (leqvM (WF_lockSet_H lockSet_H) ((Lock_INV_H l) *^* (WF_lockSet_H (removeFirstL l lockSet_H)))).
  apply WF_LockSet_decomp; auto.
  exists (WF_lockSet_H (removeFirstL l lockSet_H) *^* WF_cvSet_M cvSet_M *^* mP).
  unfold leqvM.
  intro M.
  unfold leqvM in Heqv.
  assert (INV1 M <-> (WF_lockSet_H lockSet_H *^* WF_cvSet_M cvSet_M *^* mP) M).
  apply Heqv. inversion_clear H1.
  split.
  intro.
  assert ((WF_lockSet_H lockSet_H *^* WF_cvSet_M cvSet_M *^* mP) M).
  apply H2; auto.
  sep on.
  sep lift 0 in H4.
  sep leqv_subst in H4 with (Lock_INV_H l *^* WF_lockSet_H (removeFirstL l lockSet_H)).
  sep on.
  sep matchH H4 with 4.
  intro.
  apply H3.
  sep on.
  sep lift 0.
  sep leqv_subst with (Lock_INV_H l *^* WF_lockSet_H (removeFirstL l lockSet_H)).
  sep on.
  sep matchH H1 with 4.
Qed.


Lemma INV1_lockInv1:
forall M l,
  INV1 M -> In l lockSet_H -> (Lock_INV_H l *^* mTrue) M.
Proof.
  intros.
  assert (exists m, leqvM INV1 (Lock_INV_H l *^* m)).
  apply INV1_decom_l; auto.
  destruct H1 as [mp Heqv].
  assert ((Lock_INV_H l *^* mp) M).
  apply leqvM_app_subst with INV1; auto. clear H.
  sep on.
  sep matchH H1 with 2.
Qed.


Lemma psi_cons_a_lookup_a:
forall psi f theta,
  SCAP_lookupCT (SCAP_consCT f theta psi) f theta.
Proof.
  intros.
  unfold SCAP_lookupCT.
  unfold SCAP_consCT.
  unfold SPSpec.lookup.
  simpl.
  left. auto.
Qed.
Hint Resolve psi_cons_a_lookup_a.

Lemma forall_enable_exists_l_release:
  forall (P : nat -> nat -> nat -> nat -> Pred) (g : Grt),
    (forall a b c d, enable (P a b c d) g) -> 
        enable (fun S => exists a, exists b, exists c, exists d, P a b c d S) g.
Proof.
  intros.
  unfold enable. unfold enable in H.
  intros. InvertAll.
  apply H with x x0 x1 x2; auto.
Qed.

Lemma forall_pimp_exists_l_release:
  forall (P Q : nat -> nat -> nat -> nat -> Pred) (g : Grt) (p p' : Pred),
    (forall a b c d, exists a', exists b', exists c', exists d',
      pimp ((P a b c d) |> g) (Q a' b' c' d'))
    -> frame g
    -> p = (fun S => exists a, exists b, exists c, exists d, P a b c d S)
    -> p' = (fun S => exists a, exists b, exists c, exists d, Q a b c d S)
    -> enable p g
    -> pimp ((fun S=> imprec p S) |> g) (fun S => imprec p' S).
Proof.
  intros.
  apply frame_rule_1; auto.
  unfold pimp. intros.
  unfold linkPG in H4. rewrite H1 in H4.
  InvertAll. rename x into S0.
  assert (exists a' : nat,
        exists b' : nat,
          exists c' : nat,
            exists d' : nat, pimp (P x0 x1 x2 x3 |> g) (Q a' b' c' d')).
  apply H. clear H.
  InvertAll.
  rewrite H2. exists x. exists x4. exists x5. exists x6.
  unfold pimp in H5. apply H5. unfold linkPG.
  exists S0. split; auto.
Qed.


Lemma Lock_INV_convert_a:
  forall l m M w,
    M ||- Lock_INV_H l *^* Gamma l *^* m
      -> (MemUpdW M (l + L_AVAIL) 1 w) ||- Lock_INV_H l *^* m.
Proof.
  intros.
  sep lift 0 in H.
  assert (exists b, exists w, 
          (M ||- (l + L_AVAIL |->w b) *^* (l + L_WAITING |->w w) *^* (choice (b=0) Aemp (Gamma l)) *^* (Gamma l *^* m))
          /\ Delta w = SomeT (Gamma l)).
  apply decomp_ex_Lock_INV_H; auto.
  destruct H0 as [b [w0 [Hmem Hdelta]]].
  cut (exists b, exists w0,
          (MemUpdW M (l + L_AVAIL) 1 w ||- (l + L_AVAIL |->w b) *^* (l + L_WAITING |->w w0) *^* (choice (b=0) Aemp (Gamma l)) *^* m)
          /\ Delta w0 = SomeT (Gamma l)).
  intro.
  apply decomp_ex_Lock_INV_H1; auto.
  exists 1. exists w0.
  split; auto.
  sep lift 0.
  sep upd.
  sep lift 0 in Hmem.
  sep match_left Hmem; auto. sep trivial.
  sep match_left Hmem; auto.
  sep lift 1.
  sep lift 2 in Hmem.
  sep match_left Hmem; auto.
  unfold choice in *.
  destruct Hmem as [M01 [M02 [HM0split [HM01 HM02]]]].
  sep off.
  right. split; auto.
  assert (b = 0 \/ b <> 0).
  apply nat_eq_neq_dec.
  inversion_clear H0.
  subst b.
  inversion_clear HM01.
  inversion_clear H0.
  unfold Aemp in H2.
  subst M01.
  assert (M02 = M0). auto with PMap.
  subst M02. auto.
  inversion_clear H0. absurd (0 = 0); auto.
  inversion_clear HM01.
  inversion H0. contradiction.
  inversion_clear H0.
  assert (M01 = M02).
  assert (precise (Gamma l)).
  apply precGamma.
  unfold precise in H0.
  apply H0 with M0; auto.
  apply MemMod.disjUnion_subseteq_left with M02; auto.
  apply MemMod.disjUnion_subseteq_right with M01; auto.
  assert (MemMod.disjoint M01 M02). inversion_clear HM0split; auto.
  assert (M01 = MemMod.emp).
  apply disj_eq_emp with M02; auto.
  rewrite H5 in HM0split.
  assert (M02 = M0). auto with PMap.
  rewrite <- H6. auto.
Qed.


Lemma gimp_l_release:
  forall (f f' : nat -> nat) (P P' Q : nat -> nat -> nat -> nat -> Pred) (g g' g'': Grt) (p : Pred),
    (forall a b c d,
      pimp ((P a b c d) |> g) (P' (f a) b c d))
    -> frame g
    -> p = (fun S => exists a, exists b, exists c, exists d, P a b c d S)
    -> enable p g
    -> g' = (fun S S' => 
                forall sp l pc bp,
                  leadstoPQ (P' sp l pc bp) (Q (f' sp) l pc bp) S S')
    -> gimp (fun S S' => 
                forall sp l pc bp,
                  leadstoPQ (P sp l pc bp) (Q (f' (f sp)) l pc bp) S S') g''
    -> gimp ((fun S : State => imprec p S) >o g oo g') g''.
Proof.
  intros f f' P P' Q g g' g'' p.
  set (g''' := 
             fun S S' => 
                forall sp l pc bp,
                  leadstoPQ (P sp l pc bp) (Q (f' (f sp)) l pc bp) S S').
  intros.
  apply gimp_transt with g'''; auto.
  unfold gimp.
  unfold conjG, conjPG.
  intros.
  InvertAll. rename x into S0.
  unfold g'''.
  intros.
  rewrite H3 in H7. clear H3.
  assert (leadstoPQ (P' (f sp) l pc bp) (Q (f' (f sp) ) l pc bp) S0 S').
  apply H7. clear H7.
  assert (gimp (g oo (leadstoPQ (P' (f sp) l pc bp) (Q (f' (f sp)) l pc bp)))
                     (leadstoPQ (P sp l pc bp) (Q (f' (f sp)) l pc bp))).
  apply leadstoPQ_sameQ_reverseP; auto.
  rewrite H1 in H2.
  unfold enable. unfold enable in H2.
  intros. apply H2. exists sp. exists l. exists pc. exists bp. auto.
  unfold gimp in H5. apply H5.
  unfold conjG. exists S0. split; auto.
Qed.

Lemma sep_pure:
  forall m p m' M, ((fun M => m M /\ p) *^* m') M <-> (m *^* m') M /\ p.
Proof.
  split.
  intros.
  unfold mstar in H.
  InvertAll. split; auto.
  unfold mstar. exists x. exists x0. auto.
  intro. unfold mstar in H.
  InvertAll.
  unfold mstar. exists x. exists x0. auto.
Qed.


Lemma sep_pure_a:
  forall m p m' M, ((fun M => m M /\ p) *^* m') M -> (m *^* m') M /\ p.
Proof.
  intros.
  assert (((fun M => m M /\ p) *^* m') M <-> (m *^* m') M /\ p).
  apply sep_pure.
  inversion_clear H0. apply H1; auto.
Qed.

Lemma sep_pure_b:
  forall m p m' M,  (m *^* m') M /\ p -> ((fun M => m M /\ p) *^* m') M.
Proof.
  intros.
  assert (((fun M => m M /\ p) *^* m') M <-> (m *^* m') M /\ p).
  apply sep_pure.
  inversion_clear H0. apply H2; auto.
Qed.

Lemma INV1_lock_wid:
forall M l,
  In l lockSet_H -> INV1 M -> (l + L_WAITING |->w_ *^* mTrue) M.
Proof.
  intros.
  assert ((Lock_INV_H l *^* mTrue) M).
  apply INV1_lockInv1; auto.
  unfold Lock_INV_H in H1.
  assert (exists b, exists w, (l + L_AVAIL |->w b *^* l + L_WAITING |->w w *^* choice (b = 0) Aemp (Gamma l) *^* mTrue) M).
  unfold mstar at 1 in H1.
  InvertAll.
  exists x1. exists x2. unfold mstar at 1.
  exists x. exists x0. auto.
  InvertAll.
  clear H1.
  sep on.
  sep lift 1 in H2. 
  sep matchH H2 with 4.
Qed.


Lemma IF_false_Prd_Itr_Stk:
  forall (p : Pred), (forall S, p S -> S_Flag S _IF = false) -> Prd_Itr_Stk p.
Proof.
    intros.
    unfold Prd_Itr_Stk.
    intros.
    unfold itr_enabled in H1.
    inversion_clear H1.
    assert (S_Flag S _IF = false). apply H; auto.
    rewrite H1 in H2. inversion H2.
Qed.

Lemma leadstoPP_ID:
forall p S, leadstoPQ p p S S.
Proof.
  unfold leadstoPQ.
  induction S as (((M, R), fl), isr).
  intros.
  exists M1. split; auto.
Qed.

Parameter L_sleep_neq_sw: L_sleep <> _SWITCH.
Parameter L_sleep_neq_blk: L_sleep <> _BLOCK.
Parameter L_sleep_neq_unblk: L_sleep <> _UNBLK.
Hint Resolve L_sleep_neq_sw L_sleep_neq_blk L_sleep_neq_unblk.



Lemma ptoW_ptoxW_decide:
forall M l v A B,
  M ||- l |->w v *^* A
    -> M ||- l |->w_ *^* B
      -> M ||- l |->w v *^* B.
Proof.
  intros.
  assert (exists v', M ||- l |->w v' *^* B).
  apply ptoxW_ptoW1; auto.
  inversion H1.
  assert (x = v).
  apply ptoW_ptoW_eq1 with M l B A; auto.
  subst x. auto.
Qed.


Lemma leadsToW_ptoxW_unique:
forall M l v A,
  M ||- l |->w_ *^* A -> M ||- l |~>w v -> M ||- l |->w v *^* A.
Proof.
  intros.
  assert (M ||- l |->w v *^* mTrue).
  apply leadsToW_ptoW. auto.
  apply ptoW_ptoxW_decide with mTrue; auto.
Qed.

Definition leadstoPQ2 (P Q : MPred -> Pred) : Grt :=
  fun (S S' : State) =>
    forall mP, P mP S -> Q mP S'.




Lemma stkUpd_presv_stkSpace:
forall M sp x x' sz,
  sz >= 2 ->
  M ||- stkSpacePrec sp sz
    -> MemUpdW M (sp - 2) x x' ||- stkSpacePrec sp sz.
Proof.
  intros.
  assert (sp >= sz).
  apply stkSpacePrec_sp_ge_sz with M; auto.
  assert (sp >= 2). apply ge_trans with sz; auto.
  sep stksplit.
  sep stksplit in H0.
  sep lift 1.
  sep upd2.
  sep lift 1 in H0. auto.
Qed.

Lemma stkUpd_presv_stkSpace1:
forall M sp x x' sz B,
  sz >= 2 ->
  M ||- stkSpacePrec sp sz *^* B
    -> MemUpdW M (sp - 2) x x' ||- stkSpacePrec sp sz *^* B.
Proof.
  intros.
  assert (sp >= sz).
  apply stkSpacePrec_sp_ge_sz1 with M B; auto.
  assert (sp >= 2). apply ge_trans with sz; auto.
  sep stksplit.
  sep stksplit in H0.
  sep lift 1.
  sep upd2.
  sep lift 1 in H0. auto.
Qed.

Ltac sep_stk_update :=
  match goal with
    | |- MemUpdW ?M (?l - 2) ?v ?junk ||- stkSpacePrec ?l ?sz *^* ?B => 
      apply stkUpd_presv_stkSpace1; [(auto with arith || omega || (unfold sz; omega) || idtac) | idtac]
    | |- MemUpdW ?M (?l - 2) ?v ?junk ||- stkSpacePrec ?l ?sz
      => apply stkUpd_presv_stkSpace; [(auto with arith || omega || (unfold sz; omega) || idtac) | idtac]
    | _ => fail 1 "Error: format"
  end.

Tactic Notation "sep" "stk" "upd" := sep_stk_update.

Ltac sep_blur :=
  match goal with
    | H : ?M ||- ?l |->w_ *^* ?B
      |- ?M ||- ?l |~>w_ => (apply ptoxW_leadsToxW1 with B; auto) || idtac
    | H : ?M ||- ?l |->w ?v *^* ?B
      |- ?M ||- ?l |~>w ?v => (apply ptoW_leadsToW1 with B; auto) || idtac
    | H : ?M ||- ?l |->w ?v *^* ?B
      |- ?M ||- ?l |~>w_ => (apply ptoW_leadsToxW1 with B; auto) || idtac
    | _ => fail 1 "format error; tactic cannot apply"
  end.




Ltac sep_mem_dec H H' :=
  match type of H with
    | ?M ||- (?l |->w ?v) *^* ?A
      => match type of H' with
           | M ||- (l |->w_) *^* ?B
             => substH H' with (ptoW_ptoxW_decide M l v A B H H')
         end
  end.

Tactic Notation "sep" "mem_dec" hyp (H) hyp (H') := sep_mem_dec H H'.


Ltac mexists5 x x0 x1 x2 x3 :=
  exists x; exists x0; exists x1; exists x2; exists x3.



(*
(* Lemmas below are proved but not used *)

Lemma ptoW_indom:
forall l v (M: Mem),
  (l |->w v) M -> MemMod.in_dom l M.
Proof.
  unfold MemMod.in_dom.
  intros.
  unfold ptoW in H.
  unfold mstar in H.
  InvertAll.
  unfold pto in H.
  unfold MemMod.sgltonM in H.
  inversion_clear H.
  inversion_clear H0.
  exists v.
  subst M.
  unfold MemMod.merge.
  unfold MemMod.maps_to in H1.
  rewrite H1; auto.
Qed.

Lemma ptoW_indom1:
forall l v m (M: Mem),
  (l |->w v *^* m) M -> MemMod.in_dom l M.
Proof.
  intros.
  unfold mstar in H.
  InvertAll.
  assert (MemMod.in_dom l x).
  apply ptoW_indom with v. auto.
  inversion_clear H0.
  rewrite H4. auto with PMap.
Qed.

Lemma ptoW_linearity:
forall l v v' M,
  ~ (l |->w v *^* l |->w v') M.
Proof.
  intros.
  intro.
  unfold mstar in H.
  InvertAll.
  inversion_clear H0.
  unfold MemMod.disjoint in H1.
  assert (MemMod.not_in_dom l x \/ MemMod.not_in_dom l x0).
  apply H1. clear H1.
  inversion_clear H0.
  assert (MemMod.in_dom l x).
  apply ptoW_indom with v. auto.
  absurd (MemMod.in_dom l x); auto with PMap.
  assert (MemMod.in_dom l x0).
  apply ptoW_indom with v'; auto.
  absurd (MemMod.in_dom l x0); auto with PMap.
Qed.

Lemma ptoW_linearity1:
forall l v v' m M,
  ~ (l |->w v *^* l |->w v' *^* m) M.
Proof.
  intros.
  intro.
  unfold mstar at 1 in H.enbedBlkQ_inWQSet
  InvertAll.
  absurd (~ (l |->w v *^* l |->w v') x); auto.
  apply ptoW_linearity.
Qed.


Lemma Lock_INV_linearity:
forall l M,
  ~ (Lock_INV_H l *^* Lock_INV_H l) M.
Proof.
  intros.
  intro.
  sep on.
  assert (exists b, exists w, 
              (M ||- (l + L_AVAIL |->w b) *^* (l + L_WAITING |->w w) *^* (choice (b=0) Aemp (Gamma l)) *^* Lock_INV_H l)
              /\ Delta w = SomeT (Gamma l)).
  apply decomp_ex_Lock_INV_H. auto.
  InvertAll.
  sep lift 3 in H1.
  assert (exists b, exists w, 
              (M ||- (l + L_AVAIL |->w b) *^* (l + L_WAITING |->w w) *^* (choice (b=0) Aemp (Gamma l)) *^* 
                (l + L_AVAIL |->w x *^* l + L_WAITING |->w x0 *^*
                  choice (x = 0) Aemp (Gamma l))
              )
              /\ Delta w = SomeT (Gamma l)).
  apply decomp_ex_Lock_INV_H. auto.
  InvertAll.
  assert (M ||- l + L_AVAIL |->w x1 *^* l + L_AVAIL |->w x *^* mTrue).
  sep matchH H3 with 3.
  sep off.
  assert (~ (l + L_AVAIL |->w x1 *^* l + L_AVAIL |->w x *^* mTrue) M).
  apply ptoW_linearity1. contradiction.
Qed.

Lemma Lock_INV_linearity1:
forall l m M,
  ~ (Lock_INV_H l *^* Lock_INV_H l *^* m) M.
Proof.
  intros. intro.
  unfold mstar at 1 in H.
  InvertAll.
  assert (~ (Lock_INV_H l *^* Lock_INV_H l) x).
  apply Lock_INV_linearity.
  contradiction.
Qed.

Lemma WF_LockSet_unique:
forall l ls M,
  WF_lockSet_H (l :: ls) M -> ~ In l ls.
Proof.
  intros l ls.
  induction ls.
  intros. intro.
  unfold In. auto.
  intros.
  intro.
  assert (a = l \/ In l ls).
  apply in_inv. auto.
  inversion_clear H1.
  subst a.
  unfold WF_lockSet_H in H.
  fold WF_lockSet_H in H.
  sep on.
  assert (M ||- Lock_INV_H l *^* Lock_INV_H l *^* WF_lockSet_H ls).
  sep matchH H with 3.
  clear H.
  sep off.
  assert (~ (Lock_INV_H l *^* Lock_INV_H l *^* WF_lockSet_H ls) M).
  apply Lock_INV_linearity1.
  contradiction.
  assert (~ In l ls).
  unfold WF_lockSet_H in H.
  fold WF_lockSet_H in H.
  sep on.
  sep lift 1 in H.
  destruct H as [M1 [M2 [HMsplit [HM1 HM2]]]].
  apply IHls with M2.
  unfold WF_lockSet_H. fold WF_lockSet_H.
  sep on.
  sep matchH HM2 with 2.
  contradiction.
Qed.

Lemma remove_notin_presv:
forall (l : nat) ls,
  ~ In l ls -> remove eq_nat_dec l ls = ls.
Proof.
  intros l ls.
  induction ls.
  unfold remove. auto.
  intros.
  unfold remove.
  induction (eq_nat_dec l a).
  subst l.
  assert (In a (a :: ls)).
  unfold In. left. auto.
  contradiction.
  cut (remove eq_nat_dec l ls = ls).
  intro.
  unfold remove in H0.
  rewrite H0. auto.
  apply IHls.
  intro.
  assert (In l (a :: ls)).
  unfold In. right.
  auto. contradiction.
Qed.
*)
