(***************************************************************)
(* Coq code for AIM                                            *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

Add LoadPath "..".
Add LoadPath "../CAP0".
Add LoadPath "../SCAP".
Add LoadPath "../miniOS".
Add LoadPath "../examples".

Require Import Coqlib.

Require Import List.
Require Import Arith.
Require Import Machine.
Require Import Mem_Lemmas.
Require Import Assertion.
Require Import Assertion_Lemmas.
Require Import Assertion_Lemmas4.
Require Import Assertion_Lemmas5.
Require Import Operational_Semantics.
Require Import OpSem_Lemmas.
Require Import OpSem_Lemmas2.
Require Import Reg.

Require Import AIM.
Require Import AIM_Lemmas.
Require Import QLib.

Require Import SCAP.
Require Import SCAP_READY.
Require Import Linking.
Require Import Linking_Lemmas2.
Require Import Omega.
Require Import Assembly.
Require Import SyncUtil.

Notation _STK_SIZE_cond_signal := (4 + 2 + _UNBLK_Stk + IH_Stk) (only parsing).

Definition psi_cond_signal := SCAP_emptCT.

Definition cond_signal_MPre (sp cv pc : Label) (M : Mem) : Prop :=
    ((sp |->w pc) *^* (sp + 2 |->w cv) *^* stkSpacePrec sp _STK_SIZE_cond_signal) M /\ In cv cvSet_M.

Definition cond_signal_Prd_aux (sp cv pc bp : Label) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
      R SP = sp
      /\ R BP = bp
      /\ (cond_signal_MPre sp cv pc) M
      /\ fl _IF = true
      /\ isr = false
  end.

Definition cond_signal_Prd (S : State) : Prop :=
  imprec (fun S => exists sp, exists cv, exists pc, exists bp, cond_signal_Prd_aux sp cv pc bp S) S.


Definition cond_signal_MPost := cond_signal_MPre.

Definition cond_signal_Post_aux := cond_signal_Prd_aux.

Definition cond_signal_grt (S S' : State) : Prop :=
  forall sp cv pc bp,
    leadstoPQ (cond_signal_Prd_aux sp cv pc bp) (cond_signal_Post_aux sp cv pc bp) S S'.

(*
Definition cond_signal_psi :=
   SCAP_consCT l_acquire (Ast l_acquire_Prd l_acquire_grt 1)  
  (SCAP_consCT l_release (Ast l_release_Prd l_release_grt 1)
  (SCAP_consCT L_sleep   (Ast L_sleep_0_p   L_sleep_0_g   1)
                         SCAP_emptCT
  )).
*)

Definition Ast_cond_signal_pre (cv bp sp pc : Label) (mP : MPred)
  (S : State) : Prop :=
  (stkSpacePrec sp _STK_SIZE_cond_signal
    *^* (sp |->w pc)
    *^* (sp + 2 |->w cv)
    *^* mP) (S_Mem S)
  /\ S_RF S BP = bp
  /\ S_RF S SP = sp
  /\ sp >= _STK_SIZE_cond_signal
  /\ In cv cvSet_M
  /\ S_Flag S _IF = true
  /\ S_ISR S = false.

Definition Ast_cond_signal_post := Ast_cond_signal_pre.

Definition p_cond_signal_0 (S : State) : Prop :=
  exists cv, exists bp, exists sp, exists pc,
    Ast_cond_signal_pre cv bp sp pc mTrue S.

Definition g_cond_signal_0 (S S' : State) : Prop := forall mP,
  forall cv bp sp pc,
    Ast_cond_signal_pre cv bp sp pc mP S
    -> Ast_cond_signal_post cv bp sp pc mP S'.

(* <cond_signal + 0> : _pushwp (Opr_reg BP); *)

Definition Ast_cond_signal_1 (cv bp sp pc : Label) (mP : MPred)
  (S : State) : Prop :=
  (stkSpacePrec (sp-2) (_STK_SIZE_cond_signal - 2)
    *^* (sp - 2 |->w bp)
    *^* (sp |->w pc)
    *^* (sp + 2 |->w cv)
    *^* mP) (S_Mem S)
  /\ S_RF S BP = bp
  /\ S_RF S SP = sp - 2
  /\ sp >= _STK_SIZE_cond_signal
  /\ In cv cvSet_M
  /\ S_Flag S _IF = true
  /\ S_ISR S = false.

Definition p_cond_signal_1 (S : State) : Prop :=
  exists cv, exists bp, exists sp, exists pc,
    Ast_cond_signal_1 cv bp sp pc mTrue S.

Definition g_cond_signal_1 (S S' : State) : Prop := forall mP,
  forall cv bp sp pc,
    Ast_cond_signal_1 cv bp sp pc mP S
    -> Ast_cond_signal_post cv bp sp pc mP S'.

(* <cond_signal + 1> :  _movwp  (Opr_reg SP) BP; *)
  
Definition Ast_cond_signal_2 (cv bp sp pc : Label) (mP : MPred)
  (S : State) : Prop :=
  (stkSpacePrec (sp-2) (_STK_SIZE_cond_signal - 2)
    *^* (sp - 2 |->w bp)
    *^* (sp |->w pc)
    *^* (sp + 2 |->w cv)
    *^* mP) (S_Mem S)
  /\ S_RF S BP = sp - 2
  /\ S_RF S SP = sp - 2
  /\ sp >= _STK_SIZE_cond_signal
  /\ In cv cvSet_M
  /\ S_Flag S _IF = true
  /\ S_ISR S = false.

Definition p_cond_signal_2 (S : State) : Prop :=
  exists cv, exists bp, exists sp, exists pc,
    Ast_cond_signal_2 cv bp sp pc mTrue S.

Definition g_cond_signal_2 (S S' : State) : Prop := forall mP,
  forall cv bp sp pc,
    Ast_cond_signal_2 cv bp sp pc mP S
    -> Ast_cond_signal_post cv bp sp pc mP S'.

(* <cond_signal + 2> : _movwld (A_reg 4 BP) BX; *)

Definition Ast_cond_signal_3 (cv bp sp pc : Label) (mP : MPred)
  (S : State) : Prop :=
  (stkSpacePrec (sp-2) (_STK_SIZE_cond_signal - 2)
    *^* (sp - 2 |->w bp)
    *^* (sp |->w pc)
    *^* (sp + 2 |->w cv)
    *^* mP) (S_Mem S)
  /\ S_RF S BX = cv
  /\ S_RF S SP = sp - 2
  /\ sp >= _STK_SIZE_cond_signal
  /\ In cv cvSet_M
  /\ S_Flag S _IF = true
  /\ S_ISR S = false.

Definition p_cond_signal_3 (S : State) : Prop :=
  exists cv, exists bp, exists sp, exists pc,
    Ast_cond_signal_3 cv bp sp pc mTrue S.

Definition g_cond_signal_3 (S S' : State) : Prop := forall mP,
  forall cv bp sp pc,
    Ast_cond_signal_3 cv bp sp pc mP S
    -> Ast_cond_signal_post cv bp sp pc mP S'.

(* <cond_signal + 3> :  _cli; *)

Definition Ast_cond_signal_4 (cv bp sp pc : Label) (mP : MPred)
  (S : State) : Prop :=
  ((INV0 *^* INV1)
    *^* (stkSpacePrec (sp-2) (_STK_SIZE_cond_signal - 2)
    *^* (sp - 2 |->w bp)
    *^* (sp |->w pc)
    *^* (sp + 2 |->w cv)
    *^* mP)) (S_Mem S)
  /\ S_RF S BX = cv
  /\ S_RF S SP = sp - 2
  /\ sp >= _STK_SIZE_cond_signal
  /\ In cv cvSet_M
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.

Definition p_cond_signal_4 (S : State) : Prop :=
  exists cv, exists bp, exists sp, exists pc,
    Ast_cond_signal_4 cv bp sp pc mTrue S.

Definition g_cond_signal_4 (S S' : State) : Prop := forall mP,
  forall cv bp sp pc,
    Ast_cond_signal_4 cv bp sp pc mP S
    -> Ast_cond_signal_post cv bp sp pc mP S'.

(* <cond_signal + 4> :  _pushwm (A_reg C_WAITING BX); *)

Definition Ast_cond_signal_5 (cv bp sp pc : Label) (mP : MPred)
  (S : State) : Prop :=
  ((INV0 *^* INV1)
    *^* stkSpacePrec (sp-2-2) (_STK_SIZE_cond_signal - 2 - 2 )
    *^* (sp - 2 - 2 |->w_)
    *^* (sp - 2 |->w bp)
    *^* (sp |->w pc)
    *^* (sp + 2 |->w cv)
    *^* mP) (S_Mem S)
  /\ (exists w, (sp - 2 - 2 |~>w w) (S_Mem S)
    /\ Delta w = SomeT Aemp)
  /\ S_RF S SP = sp - 2 - 2
  /\ sp >= _STK_SIZE_cond_signal
  /\ In cv cvSet_M
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.

Definition p_cond_signal_5 (S : State) : Prop :=
  exists cv, exists bp, exists sp, exists pc,
    Ast_cond_signal_5 cv bp sp pc mTrue S.

Definition g_cond_signal_5 (S S' : State) : Prop := forall mP,
  forall cv bp sp pc,
    Ast_cond_signal_5 cv bp sp pc mP S
    -> Ast_cond_signal_post cv bp sp pc mP S'.

(* <cond_signal + 5> :  _call   _UNBLK; *)

Definition Ast_cond_signal_6 (cv bp sp pc : Label) (mP : MPred)
  (S : State) : Prop :=
  ((INV0 *^* INV1)
    *^* (stkSpacePrec (sp-2-2) (_STK_SIZE_cond_signal - 2 - 2 )
    *^* (sp - 2 - 2 |->w_)
    *^* (sp - 2 |->w bp)
    *^* (sp |->w pc)
    *^* (sp + 2 |->w cv)
    *^* mP)) (S_Mem S)
  /\ S_RF S SP = sp - 2 - 2
  /\ sp >= _STK_SIZE_cond_signal
  /\ In cv cvSet_M
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.

Definition p_cond_signal_6 (S : State) : Prop :=
  exists cv, exists bp, exists sp, exists pc,
    Ast_cond_signal_6 cv bp sp pc mTrue S.

Definition g_cond_signal_6 (S S' : State) : Prop := forall mP,
  forall cv bp sp pc,
    Ast_cond_signal_6 cv bp sp pc mP S
    -> Ast_cond_signal_post cv bp sp pc mP S'.

(* <cond_signal + 6> :  _sti; *)

Definition Ast_cond_signal_7 (cv bp sp pc : Label) (mP : MPred)
  (S : State) : Prop :=
  (stkSpacePrec (sp-2-2) (_STK_SIZE_cond_signal - 2 - 2 )
    *^* (sp - 2 - 2 |->w_)
    *^* (sp - 2 |->w bp)
    *^* (sp |->w pc)
    *^* (sp + 2 |->w cv)
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 2 - 2
  /\ sp >= _STK_SIZE_cond_signal
  /\ In cv cvSet_M
  /\ S_Flag S _IF = true
  /\ S_ISR S = false.

Definition p_cond_signal_7 (S : State) : Prop :=
  exists cv, exists bp, exists sp, exists pc,
    Ast_cond_signal_7 cv bp sp pc mTrue S.

Definition g_cond_signal_7 (S S' : State) : Prop := forall mP,
  forall cv bp sp pc,
    Ast_cond_signal_7 cv bp sp pc mP S
    -> Ast_cond_signal_post cv bp sp pc mP S'.

(* <cond_signal + 7> :  _addw   (Opr_imm 2) SP; *)

Definition Ast_cond_signal_8 (cv bp sp pc : Label) (mP : MPred)
  (S : State) : Prop :=
  (stkSpacePrec (sp-2) (_STK_SIZE_cond_signal - 2)
    *^* (sp - 2 |->w bp)
    *^* (sp |->w pc)
    *^* (sp + 2 |->w cv)
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 2
  /\ sp >= _STK_SIZE_cond_signal
  /\ In cv cvSet_M
  /\ S_Flag S _IF = true
  /\ S_ISR S = false.

Definition p_cond_signal_8 (S : State) : Prop :=
  exists cv, exists bp, exists sp, exists pc,
    Ast_cond_signal_8 cv bp sp pc mTrue S.

Definition g_cond_signal_8 (S S' : State) : Prop := forall mP,
  forall cv bp sp pc,
    Ast_cond_signal_8 cv bp sp pc mP S
    -> Ast_cond_signal_post cv bp sp pc mP S'.

(* <cond_signal + 8> : _popw   BP; *)

Definition Ast_cond_signal_9 (cv bp sp pc : Label) (mP : MPred)
  (S : State) : Prop :=
  (stkSpacePrec sp (_STK_SIZE_cond_signal)
    *^* (sp |->w pc)
    *^* (sp + 2 |->w cv)
    *^* mP) (S_Mem S)
  /\ S_RF S BP = bp
  /\ S_RF S SP = sp
  /\ sp >= _STK_SIZE_cond_signal
  /\ In cv cvSet_M
  /\ S_Flag S _IF = true
  /\ S_ISR S = false.

Definition p_cond_signal_9 (S : State) : Prop :=
  exists cv, exists bp, exists sp, exists pc,
    Ast_cond_signal_9 cv bp sp pc mTrue S.

Definition g_cond_signal_9 (S S' : State) : Prop := forall mP,
  forall cv bp sp pc,
    Ast_cond_signal_9 cv bp sp pc mP S
    -> Ast_cond_signal_post cv bp sp pc mP S'.

Lemma p_cond_signal_0_sound : forall S,
  cond_signal_Prd S -> p_cond_signal_0 S.
Proof.
  unfold cond_signal_Prd, p_cond_signal_0.
  intros S HS.
  unfold imprec, pmstar in HS.
  destruct HS as [M1 [M2 [HJM [HM1 HM2]]]].
  destruct S as [[[M R] fl] isr].
  destruct HM1 as [sp [cv [pc [bp HS]]]].
  unfold cond_signal_Prd_aux in HS.
  unfold Ast_cond_signal_pre.
  simplS in * .
  unfold cond_signal_MPre in HS.
  exists cv; exists bp; exists sp; exists pc.
  destruct HS as [HSP [HBP [HM1 [Hif Hisr]]]].
  destruct HM1 as [HM1 Hcv].
  rsplit.
  repeat split_l; trivial.
    sep on.
    sep lift 2 in HM1.
    assert (HM1x : M1 ||- stkSpace sp (4 + 2 + _UNBLK_Stk + IH_Stk)).
      sep trivial.
    destruct HM1x; trivial.
  exists M1; exists M2; split_l.
  split_r.
  sep on.
  sep matchH HM1 with 2.
Qed.

Lemma g_cond_signal_0_sound : forall S S',
  g_cond_signal_0 S S' -> cond_signal_grt S S'.
Proof.
  unfold g_cond_signal_0, cond_signal_grt.
  unfold cond_signal_Prd_aux, cond_signal_Post_aux.
  unfold leadstoPQ.
  intros S S' HSS'.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  intros sp cv pc bp M1 M2 HJM HS.
  assert (H : Ast_cond_signal_pre cv bp sp pc (mEq M2) (M, R, fl, isr)).
    unfold cond_signal_MPre in HS.
    destruct HS as [HSP [HBP [[HM1 Hcv] [Hif Hisr]]]].
    unfold Ast_cond_signal_pre.
    rsplit.
    simplS in * .
    repeat split_l; trivial.
      sep on.
      sep lift 2 in HM1.
      assert (HM1x : M1 ||- stkSpace sp (4 + 2 + _UNBLK_Stk + IH_Stk)).
        sep trivial.
      destruct HM1x; trivial.
    exists M1; exists M2; split_l.
    split_r.
    sep on.
    sep matchH HM1 with 2.
  unfold mEq; trivial.
  assert (H0 := HSS' _ _ _ _ _ H).
  unfold Ast_cond_signal_post, Ast_cond_signal_pre in H0.
  simplS in * .
  destruct H0 as [HM' [HBP' [HSP' [Hsp [Hcv [Hif' Hisr']]]]]].
  destruct HM' as [M1' [M2' [HJM' [HM1' HM2']]]].
  exists M1'.
  unfold cond_signal_Prd_aux.
  unfold cond_signal_MPre.
  unfold mEq in HM2'.
  subst M2'.
  repeat split_l; trivial.
  sep on.
  sep matchH HM1' with 3.
Qed.

Lemma g_cond_signal_0_complete : forall S S',
  cond_signal_grt S S' -> g_cond_signal_0 S S'.
Proof.
  unfold g_cond_signal_0, cond_signal_grt.
  unfold cond_signal_Prd_aux, cond_signal_Post_aux.
  unfold leadstoPQ.
  intros S S' HSS'.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  unfold Ast_cond_signal_post, Ast_cond_signal_pre.
  intros mP cv bp sp pc HS.
  simplS in * .
  destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hif Hisr]]]]]].
  destruct HM as [M1 [M2 [HJM [HM1 HM2]]]].
  substH HSS' with (HSS' sp cv pc bp M1 M2 HJM).
  assert (R SP = sp /\
         R BP = bp /\
         cond_signal_MPre sp cv pc M1 /\ fl _IF = true /\ isr = false).
    split_l.
    split_l.
    split_r by auto.
    unfold cond_signal_MPre.
    split_r.
    sep on.
    sep matchH HM1 with 3.
  substH HSS' with (HSS' H); clear H.
  destruct HSS' as [M1' [HJM' HM1']].
  unfold cond_signal_Prd_aux in HM1'.
  destruct HM1' as [HSP' [HBP' [HM1' [Hif' Hisr']]]].
  rsplit.
  repeat split_l; trivial.
  exists M1'; exists M2; split_l.
  split_r.
  unfold cond_signal_MPre in HM1'.
  destruct HM1' as [HM1' _].
  sep on.
  sep matchH HM1' with 3.
Qed.

Lemma wff_cmd_cond_signal1:
   NH_WFcmd
     (SCAP_consCT (cond_signal + 1) (Ast p_cond_signal_1 g_cond_signal_1 1)
        psi_cond_signal) cond_signal (Ast cond_signal_Prd cond_signal_grt 1)
     (comml (_pushwp (Opr_reg BP))).
Proof.
  assert (Hpp' : forall cv bp sp pc mP S S',
    Ast_cond_signal_pre cv bp sp pc mP S
    -> nextS (cond_signal) (comml (_pushwp (Opr_reg BP))) S S'
    -> Ast_cond_signal_1 cv bp sp pc mP S').
    intros cv bp sp pc mP S S' HS HSS'.
    unfold Ast_cond_signal_pre, Ast_cond_signal_1 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspgt [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simplS in * .
    destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hif Hisr]]]]]].
    split.
      subst M' R'.
      sep on.
      rewrite HBP.
      rewrite HSP.
      sep lift 1.
      sep upd.
      sep lift 0 in HM.
      sep stksplit in HM.
      sep matchH HM with 5.
    subst M' R' fl' isr'.
    reg simpl.
    rewrite HSP.
    repeat split_l; trivial.
  simpl.
  split.
    exists p_cond_signal_1; exists g_cond_signal_1.
  split_l (* lookup *).
  split.
  (* enable *)
    apply enable_pushwp with (_STK_SIZE_cond_signal).
    intros S HS.
    destruct S as [[[M R] fl] isr].
    substH HS with (p_cond_signal_0_sound _ HS).
    unfold p_cond_signal_0 in HS.
    destruct HS as [cv [bp [sp [pc HS]]]].
    unfold Ast_cond_signal_pre in HS.
    destruct HS as [HM [HBP [HSP _]]].
    simplS in * .
    rewrite HSP.
    sep on.
    sep lift 0 in HM.
    sep trivial.
    omega.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold p_cond_signal_1, p_cond_signal_2.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HSS'].
    substH HS with (p_cond_signal_0_sound _ HS).
    destruct HS as [cv [bp [sp [pc HS]]]].
    exists cv; exists bp; exists sp; exists pc.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HxS HSS'] HS'S''].
  apply g_cond_signal_0_sound.
  unfold g_cond_signal_0, g_cond_signal_1 in * .
  intros mP cv bp sp pc HS.
  apply HS'S''.
  apply Hpp' with S; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS HIS.
    unfold cond_signal_Prd in HS.
    destruct HS as [M1 [M2 [HJM [HM1 HM2]]]].
    destruct HM1 as [sp [cv [pc [bp HM1]]]].
    unfold cond_signal_Prd_aux in HM1.
    destruct HM1 as [HSP [HBP [HM1 [Hif Hisr]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in HIS.
    simplS in * .
    rewrite HSP.
    unfold cond_signal_MPre in HM1.
    destruct HM1 as [HM1 Hcv].
    sep on.
    sep lift 2 in HM1.
    sep stksplit (IH_Stk + 6) in HM1.
    sep lift 0 in HM1.
    destruct HM1 as [M11 [M12 [HJM1 [HM11 HM12]]]].
    exists (MemMod.merge M12 M2); exists M11.
    split.
      split.
        disjoint des.
        disjoint lift_r.
      destruct ((fun x => HJM) 0); destruct ((fun x => HJM1) 0).
      subst M1 M.
      merge commut (MemMod.merge M12 M2).
      merge assoc M11.
      trivial.
      disjoint des.
      disjoint lift_r.
    split_l.
    intros M11' HM11'.
    unfold cond_signal_Prd.
    exists (MemMod.merge M11' M12); exists M2.
    simplS.
    assert (HeqD : MemMod.eqDom M11 M11').
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    split.
      split.
      disjoint des.
      disjoint rep_l M11.
      disjoint lift_l.
      disjoint lift_l.
      merge_commut M2.
      merge assoc M2.
      merge commut M2.
      merge commut M12.
      trivial.
      disjoint trivial.
      disjoint rep_r M11.
      disjoint des.
      disjoint lift_r.
      disjoint rep_r M11.
      disjoint lift_r.
    split_r.
    exists sp; exists cv; exists pc; exists bp.
    unfold cond_signal_Prd_aux.
    split_l.
    split_l.
    unfold cond_signal_MPre.
    split_r.
      repeat split_l; trivial.
    sep on.
    sep lift 2.
    assert (H : stkSpace sp (IH_Stk+6) M11).
      sep on.
      sep trivial.
    destruct H.
    sep stksplit (IH_Stk+6).
    sep lift 1.
    exists M11'; exists M12.
    split.
      split_r.
      disjoint rep_l M11.
    split_l.
    sep on.
    sep matchH HM12 with 3.
  (* Grt_Itr_Stk *)
  apply (Grt_Itr_Stk_equal g_cond_signal_0_sound g_cond_signal_0_complete).
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold g_cond_signal_0 in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP cv bp sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  apply HSS'.
  clear HSS'.
  unfold Ast_cond_signal_pre in * .
  simplS in * .
  destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hif Hisr]]]]]].
  rsplit.
    repeat split_l; trivial.
  sep on.
  sep lift 0 in HM.
  sep stksplit (IH_Stk + 6) in HM.
  sep lift 0 in HM.
  destruct HM as [M2x [M1x [HJM12 [HM2x HM1x]]]].
  sep lift 0.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  assert (M2x = M2' /\ M1x = M1).
    sep off.
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    rewrite HSP in HM2'.
    assert (Hx := prec_disjUnion_inv H H0 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  rewrite HSP in HM2.
  split_l.
  sep on.
  sep matchH HM1x with 3.
Qed.

Lemma wff_cmd_cond_signal2:
   NH_WFcmd
     (SCAP_consCT (cond_signal + 1 + 1)
        (Ast p_cond_signal_2 g_cond_signal_2 1) psi_cond_signal)
     (cond_signal + 1) (Ast p_cond_signal_1 g_cond_signal_1 1)
     (comml (_movwp (Opr_reg SP) BP)).
Proof.
  assert (Hpp' : forall cv bp sp pc mP S S',
    Ast_cond_signal_1 cv bp sp pc mP S
    -> nextS (cond_signal + 1) (comml (_movwp  (Opr_reg SP) BP)) S S'
    -> Ast_cond_signal_2 cv bp sp pc mP S').
    intros cv bp sp pc mP S S' HS HSS'.
    unfold Ast_cond_signal_1, Ast_cond_signal_2 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    simplS in * .
    destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hif Hisr]]]]]].
    subst M' R' fl' isr'.
    reg simpl.
    repeat split_l; trivial.
  simpl.
  split.
    exists p_cond_signal_2; exists g_cond_signal_2.
  split_l.
  split.
    (* enable movw *)
    apply trivial_enable_movwp.
  split. (* p ==> p' *)
    unfold p_cond_signal_1, p_cond_signal_2.
    intros S' H.
    unfold linkPG in H.
    destruct H as [S H].
    destruct H as [H HSS'].
    destruct H as [cv [bp [sp [pc HS]]]].
    exists cv; exists bp; exists sp; exists pc.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold conjPG, conjG in H.
  destruct H as [S' [[_ HSS'] HS'S'']].
  unfold g_cond_signal_1, g_cond_signal_2 in * .
  intros mP cv bp sp pc HS.
  eapply HS'S''; eauto.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold p_cond_signal_1 in HS.
    destruct HS as [cv [bp [sp [pc HS]]]].
    unfold Ast_cond_signal_1 in HS.
    destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hif Hisr]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    rewrite HSP.
    sep on.
    sep lift 0 in HM.
    sep stksplit (IH_Stk + 6) in HM.
    sep lift 0 in HM.
    destruct HM as [M1 [M2 [HJM [HM1 HM2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    split_l.
    intros M1' HM1'.
    unfold p_cond_signal_1.
    exists cv; exists bp; exists sp; exists pc.
    unfold Ast_cond_signal_1.
    simplS in * .
    rsplit.
      repeat split_l; trivial.
    sep on.
    sep lift 0.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      sep off.
      split.
        disjoint rep_l M1.
        eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    split_l.
    sep on.
    sep matchH HM2 with 4.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold g_cond_signal_1 in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP cv bp sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  apply HSS'.
  clear HSS'.
  unfold Ast_cond_signal_1 in * .
  simplS in * .
  destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hif Hisr]]]]]].
  rsplit.
    repeat split_l; trivial.
  sep on.
  sep lift 0 in HM.
  sep stksplit (IH_Stk + 6) in HM.
  sep lift 0 in HM.
  destruct HM as [M2x [M1x [HJM12 [HM2x HM1x]]]].
  sep lift 0.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  assert (M2x = M2' /\ M1x = M1).
    sep off.
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    rewrite HSP in HM2'.
    assert (Hx := prec_disjUnion_inv H H0 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  rewrite HSP in HM2.
  split_l.
  sep on.
  sep matchH HM1x with 3.
Qed.

Lemma wff_cmd_cond_signal3:
   NH_WFcmd
     (SCAP_consCT (cond_signal + 1 + 1 + 1)
        (Ast p_cond_signal_3 g_cond_signal_3 1) psi_cond_signal)
     (cond_signal + 1 + 1) (Ast p_cond_signal_2 g_cond_signal_2 1)
     (comml (_movwld (A_reg 4 BP) BX)).
Proof.
  assert (Hpp' : forall cv bp sp pc mP S S',
    Ast_cond_signal_2 cv bp sp pc mP S
    -> nextS (cond_signal + 1 + 1) (comml (_movwld (A_reg 4 BP) BX)) S S'
    -> Ast_cond_signal_3 cv bp sp pc mP S').
    intros cv bp sp pc mP S S' HS HSS'.
    unfold Ast_cond_signal_2, Ast_cond_signal_3 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HR' as [cvx [Hlds HR']].
    simplS in * .
    destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hif Hisr]]]]]].
    assert (Heq : cvx = cv).
      unfold addr_eval in Hlds.
      rewrite HBP in Hlds.
      arith_rep (sp - 2 + 4) with (sp + 2) in Hlds.
      sep on.
      sep lift 3 in HM.
      sep inj.
      auto.
    subst cvx M' R' fl' isr'.
    reg simpl.
    repeat split_l; trivial.
  simpl.
  split.
    exists p_cond_signal_3; exists g_cond_signal_3.
  split_l.
  split.
  (* enable *)
    apply enable_movwld.
    intros S HS.
    destruct S as [[[M R] fl] isr].
    unfold p_cond_signal_2 in HS.
    destruct HS as [cv [bp [sp [pc HS]]]].
    unfold Ast_cond_signal_2 in HS.
    destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hif Hisr]]]]]].
    simplS in * .
    unfold addr_eval.
    rewrite HBP.
    arith_rep (sp - 2 + 4) with (sp + 2).
    sep on.
    sep lift 3 in HM.
    sep trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold p_cond_signal_2, p_cond_signal_3.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HSS'].
    destruct HS as [cv [bp [sp [pc HS]]]]. 
    exists cv; exists bp; exists sp; exists pc.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HxS HSS'] HS'S''].
  unfold g_cond_signal_2, g_cond_signal_3 in * .
  intros mP cv bp sp pc HS.
  eapply HS'S''; eauto.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold p_cond_signal_2 in HS.
    destruct HS as [cv [bp [sp [pc HS]]]].
    unfold Ast_cond_signal_2 in HS.
    destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hif Hisr]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    rewrite HSP.
    sep on.
    sep lift 0 in HM.
    sep stksplit (IH_Stk + 6) in HM.
    sep lift 0 in HM.
    destruct HM as [M1 [M2 [HJM [HM1 HM2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    split_l.
    intros M1' HM1'.
    unfold p_cond_signal_2.
    exists cv; exists bp; exists sp; exists pc.
    unfold Ast_cond_signal_2.
    simplS in * .
    rsplit.
      repeat split_l; trivial.
    sep on.
    sep lift 0.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      sep off.
      split.
        disjoint rep_l M1.
        eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    split_l.
    sep on.
    sep matchH HM2 with 4.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold g_cond_signal_2 in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP cv bp sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  apply HSS'.
  clear HSS'.
  unfold Ast_cond_signal_2 in * .
  simplS in * .
  destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hif Hisr]]]]]].
  rsplit.
    repeat split_l; trivial.
  sep on.
  sep lift 0 in HM.
  sep stksplit (IH_Stk + 6) in HM.
  sep lift 0 in HM.
  destruct HM as [M2x [M1x [HJM12 [HM2x HM1x]]]].
  sep lift 0.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  assert (M2x = M2' /\ M1x = M1).
    sep off.
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    rewrite HSP in HM2'.
    assert (Hx := prec_disjUnion_inv H H0 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  rewrite HSP in HM2.
  split_l.
  sep on.
  sep matchH HM1x with 3.
Qed.

Lemma wff_cmd_cond_signal4:
   NH_WFcmd
     (SCAP_consCT (cond_signal + 1 + 1 + 1 + 1)
        (Ast p_cond_signal_4 g_cond_signal_4 1) psi_cond_signal)
     (cond_signal + 1 + 1 + 1) (Ast p_cond_signal_3 g_cond_signal_3 1)
     (comml _cli).
Proof.
  assert (Hpp': forall cv bp sp pc mP S S', 
    Ast_cond_signal_3 cv bp sp pc mP S
    -> gcli S S'
    -> Ast_cond_signal_4 cv bp sp pc mP S').
    intros cv bp sp pc mP S S' HS HSS'.
    unfold linkPG, conjPG, conjG.
    unfold Ast_cond_signal_3, gcli, Ast_cond_signal_4 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct HS as [HM [HBX [HSP [Hsp [Hcv [Hif Hisr]]]]]].
    destruct HSS' as [Hshift [Hif' [Hisr' HR']]].
    assert (Hshift' : prepost Aemp (INV0 *^* INV1) (M, R, fl, isr) (M', R', fl', isr')).
      apply (prepost_weaken_post (mQ' := INV0 *^* INV1) Hshift).
      intros Mx H.
      apply (choice_elim_left H); unfold itr_enabled; auto.
    split.
      sep on.
      assert (HMx : S_Mem (M, R, fl, isr) ||- 
        (Aemp) *^* (stkSpacePrec (sp - 2) (4 + 2 + _UNBLK_Stk + IH_Stk - 2) *^*
           sp - 2 |->w bp *^* sp |->w pc *^* sp + 2 |->w cv *^* mP)).
        sep matchH HM with 4.
      substH HM with (prepost_elim Hshift' precise_emp HMx).
      sep on.
      sep matchH HM with 5.
    simplS in * .
    subst R' isr'.
    repeat split_l; trivial.
  simpl.
  split.
  exists p_cond_signal_4; exists g_cond_signal_4.
  split_l. (* lookup *)
  split.
  (* p ==> p' *)
  unfold p_cond_signal_3, p_cond_signal_4.
  intros S' H.
  unfold linkPG in H.
  destruct H as [S [HS HSS']].
  destruct HS as [cv [bp [sp [pc HS]]]].
  exists cv; exists bp; exists sp; exists pc.
  apply Hpp' with S; trivial.
  (* g' ==> g *)
  unfold g_cond_signal_3, g_cond_signal_4.
  unfold linkPG, conjPG, conjG.
  intros S S'' H.
  destruct H as [S' H].
  destruct H as [[HxS HSS'] HS'S''].
  intros mP cv bp sp pc HS.
  apply HS'S''.
  apply Hpp' with S; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold p_cond_signal_3 in HS.
    destruct HS as [cv [bp [sp [pc HS]]]].
    unfold Ast_cond_signal_3 in HS.
    destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hif Hisr]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    rewrite HSP.
    sep on.
    sep lift 0 in HM.
    sep stksplit (IH_Stk + 6) in HM.
    sep lift 0 in HM.
    destruct HM as [M1 [M2 [HJM [HM1 HM2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    split_l.
    intros M1' HM1'.
    unfold p_cond_signal_3.
    exists cv; exists bp; exists sp; exists pc.
    unfold Ast_cond_signal_3.
    simplS in * .
    rsplit.
      repeat split_l; trivial.
    sep on.
    sep lift 0.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      sep off.
      split.
        disjoint rep_l M1.
        eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    split_l.
    sep on.
    sep matchH HM2 with 4.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold g_cond_signal_3 in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP cv bp sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  apply HSS'.
  clear HSS'.
  unfold Ast_cond_signal_3 in * .
  simplS in * .
  destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hif Hisr]]]]]].
  rsplit.
    repeat split_l; trivial.
  sep on.
  sep lift 0 in HM.
  sep stksplit (IH_Stk + 6) in HM.
  sep lift 0 in HM.
  destruct HM as [M2x [M1x [HJM12 [HM2x HM1x]]]].
  sep lift 0.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  assert (M2x = M2' /\ M1x = M1).
    sep off.
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    rewrite HSP in HM2'.
    assert (Hx := prec_disjUnion_inv H H0 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  rewrite HSP in HM2.
  split_l.
  sep on.
  sep matchH HM1x with 3.
Qed.

Lemma wff_cmd_cond_signal5:
   NH_WFcmd
     (SCAP_consCT (cond_signal + 1 + 1 + 1 + 1 + 1)
        (Ast p_cond_signal_5 g_cond_signal_5 1) psi_cond_signal)
     (cond_signal + 1 + 1 + 1 + 1) (Ast p_cond_signal_4 g_cond_signal_4 1)
     (comml (_pushwm (A_reg C_WAITING BX))).
Proof.
  assert (Hpp' : forall cv bp sp pc mP S S',
    Ast_cond_signal_4 cv bp sp pc mP S
    -> nextS (cond_signal+1+1+1+1) (comml (_pushwm (A_reg C_WAITING BX))) S S'
    -> Ast_cond_signal_5 cv bp sp pc mP S').
    intros cv bp sp pc mP S S' HS HSS'.
    unfold Ast_cond_signal_4, Ast_cond_signal_5 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [wx [Hlds [Hspgt [HldsB HM']]]].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM', Hlds.
    simplS in * .
    destruct HS as [HM [HBX [HSP [Hsp [Hcv [Hif Hisr]]]]]].
    assert (Hw : exists w,
      (cv + C_WAITING |~>w w) M /\ Delta w = SomeT Aemp).
      assert ((CV_INV_M cv *^* mTrue) M).
      apply INV1_cvInv; trivial.
      sep on.
      sep matchH HM with 5.
      destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
      unfold CV_INV_M in HM1.
      destruct HM1 as [w [HM1 Hw]].
      exists w; split_r.
      unfold leadsToW.
      unfold imprecM.
      exists M1; exists M2; auto.
    destruct Hw as [w [Hw1 Hw2]].
    assert (Heq : w = wx).
      rewrite HBX in Hlds.
      sep on. 
      sep inj.
      auto.
    subst wx.
    subst M' R' fl' isr'.
    rewrite HSP.
    split.
      sep on.
      sep lift 3.
      sep exists_left w.
      sep upd.
      sep lift 2 in HM.
      sep stksplit in HM.
      sep matchH HM with 6.
    reg simpl.
    repeat split_l; trivial.
    exists w; split_r.
    sep on.
    sep upd.
    sep lift 2 in HM.
    sep stksplit in HM.
    sep lift 1 in HM.
    sep trivial.
  simpl.
  split.
  exists p_cond_signal_5; exists g_cond_signal_5.
  split_l (* lookup *).
  split.
  (* enable *)
    apply enable_pushwm with (_STK_SIZE_cond_signal - 2); try omega.
    intros S HS.
    destruct S as [[[M R] fl] isr].
    unfold p_cond_signal_4 in HS.
    destruct HS as [cv [bp [sp [pc HS]]]].
    unfold Ast_cond_signal_4 in HS.
    simplS in * .
    destruct HS as [HM [HBX [HSP [Hsp [Hcv [Hif Hisr]]]]]].
    rewrite HSP.
    unfold addr_eval.
    rewrite HBX.
    assert (Hw : exists w,
      (cv + C_WAITING |~>w w) M /\ Delta w = SomeT Aemp).
      assert ((CV_INV_M cv *^* mTrue) M).
        apply INV1_cvInv; trivial.
      sep on.
      sep matchH HM with 5.
      destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
      unfold CV_INV_M in HM1.
      destruct HM1 as [w [HM1 Hw]].
      exists w; split_r.
      unfold leadsToW.
      unfold imprecM.
      exists M1; exists M2; auto.
    destruct Hw as [w [Hw1 Hw2]].
    split.
      sep on.
      sep lift 2 in HM.
      sep trivial.
      sep on.
    sep trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold p_cond_signal_4, p_cond_signal_5.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HSS'].
    destruct HS as [cv [bp [sp [pc HS]]]].
    exists cv; exists bp; exists sp; exists pc.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HxS HSS'] HS'S''].
  unfold g_cond_signal_4, g_cond_signal_5 in * .
  intros mP cv bp sp pc HS.
  apply HS'S''.
  apply Hpp' with S; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold p_cond_signal_4 in HS.
    destruct HS as [cv [bp [sp [pc HS]]]].
    unfold Ast_cond_signal_4 in HS.
    destruct HS as [HM [HBX [HSP [Hsp [Hcv [Hif Hisr]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold g_cond_signal_5 in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP cv bp sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_cond_signal_4 in HS.
  simplS in * .
  destruct HS as [HM [HBX [HSP [Hsp [Hcv [Hif Hisr]]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma wff_cmd_cond_signal6:
   NH_WFcmd
     (SCAP_consCT (cond_signal + 1 + 1 + 1 + 1 + 1 + 1)
        (Ast p_cond_signal_6 g_cond_signal_6 1) psi_cond_signal)
     (cond_signal + 1 + 1 + 1 + 1 + 1)
     (Ast p_cond_signal_5 g_cond_signal_5 1) (comml (_call _UNBLK)).
Proof.
  assert (Hpp''' : forall cv bp sp pc mP S S' S'' S''' pc',
    Ast_cond_signal_5 cv bp sp pc mP S
    -> nextS (cond_signal + 1 + 1 + 1 + 1 + 1) (comml (_call _UNBLK)) S S'
    -> (gunblk Aemp) S' S''
    -> nextS pc' (comms _retw) S'' S'''
    -> Ast_cond_signal_6 cv bp sp pc mP S''').
    intros cv bp sp pc mP.
    intros S S' S'' S''' pc'.
    intros HS HSS' HS'S'' HS''S'''.
    unfold Ast_cond_signal_5, Ast_cond_signal_6 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct S'' as [[[M'' R''] fl''] isr''].
    destruct S''' as [[[M''' R'''] fl'''] isr'''].
    simplS in * .
    destruct HS as [HM [Hw [HSP [Hsp [Hcv [Hif Hisr]]]]]].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspg' [Hlds Hupd]].
    unfold gunblk in HS'S''.
    destruct HS'S'' as [sp' [HM'' [Hif'' [Hisr'' HSP'']]]].
    destruct (nextS_inversion HS''S''') as [HM''' [HR''' [Hfl''' Hisr''']]].
    destruct HR''' as [pc''' [Hlds''' HR''']].
    simplS in * .
    subst M''' R''' fl''' isr'''.
    subst isr'' R' fl' isr'.
    clear HSS' HS''S'''.
    reg simpl.
    destruct HSP'' as [Hsp' HSP''].
    subst sp'.
    rewrite HSP in HSP''.
    rewrite HSP in HM''.
    destruct Hw as [w [Hldsw Hw]].
    split.
      sep on.
      substH Hupd with (MemUpd16_MemUpdW Hupd).
      destruct Hupd as [junk HM'].
      rewrite HSP in HM'.
      assert (HM''x : prepost
        (stkSpacePrec (sp - 2 - 2 - 2) 10)
        (stkSpacePrec (sp - 2 - 2 - 2) 10)
        (M', RF_update R SP (sp - 2 - 2 - 2), fl, isr)
        (M'', R'', fl'', isr)).
        eapply prepost_weaken_pre; eauto.
        intros Mx HMx.
        destruct HMx as [Mx1 [Mx2 [HJMx [HMx1 HMx2]]]].
        assert (Mx1 = MemMod.emp).
          destruct (eq_nat_dec (R'' AX) 0).
            rewrite e in HMx1.
            substH HMx1 with (choice_elim_left HMx1 (refl_equal _)).
            unfold Aemp in HMx1; trivial.
          substH HMx1 with (choice_elim_right HMx1 n).
          unfold Aemp in HMx1; trivial.
        assert (Mx2 = Mx).
          subst Mx1.
          auto with PMap.
        subst Mx1 Mx2.
        trivial.
      clear HM''; rename HM''x into HM''.
      sep lift 2.
      sep lift 2 in HM.
      sep stksplit in HM.
      clear HSP HSP'' Hspg'.
      sep stksplit.
      sep stksplit 10.
      sep stksplit 10 in HM.
      assert (HM'x : M' ||-
          stkSpacePrec (sp - 2 - 2 - 2) 10 *^*
          stkSpacePrec (sp - 2 - 2 - 2 - 10)
          (4 + 2 + 10 + IH_Stk - 2 - 2 - 2 - 10) *^*
          (sp - 2 - 2 - 2 |->w (cond_signal + 1 + 1 + 1 + 1 + 1 + 1)) *^*
          (sp - 2 - 2 |->w_) *^*
          (sp - 2 |->w bp) *^*
          (sp |->w pc) *^*
          (sp + 2 |->w cv) *^* mP *^*
          (INV0 *^* INV1)).
        subst M'.
        sep lift 2.
        sep upd.
        sep matchH HM with 7.
      replace M'' with (S_Mem (M'', R'', fl'', isr)); trivial.
      replace M' with (S_Mem (M', RF_update R SP (sp - 2 - 2 - 2), fl, isr)) in HM'x; trivial.
      sep lift 0 in HM'x.
      sep lift 1.
      sep off.
      substH HM'' with (prepost_elim HM'' (@precise_stkSpacePrec _ _) HM'x).
      simplS in * .
      sep on.
      sep matchH HM'' with 10.
    rewrite <- HSP''.
    split.
      generalize Hsp; clearall.
      omega.
    rewrite <- Hif''.
    auto.
  simpl.
  split.
    left.
    split.
      unfold prim_entry.
      right; right; trivial.
    split.
      constructor 3 with p_cond_signal_5 p_cond_signal_6 
        g_cond_signal_5 g_cond_signal_6 1; trivial.
    (* enable *)
    unfold enable_unblk.
    intros S HS.
    unfold p_cond_signal_5 in HS.
    destruct HS as [cv [bp [sp [pc HS]]]].
    assert (Hw : exists w, (sp - 2 - 2 |~>w w) (S_Mem S)
      /\ Delta w = SomeT Aemp).
      destruct HS as [HM [Hw [HSP [Hsp [Hcv [Hif Hisr]]]]]].
      destruct Hw as [w [Hldsw Hw]].
      exists w; auto.
    destruct Hw as [w [Hldsw Hw]].
    exists w; exists Aemp.
    split.
      split_l.
      destruct HS as [HM [_ [HSP [Hsp [Hcv [Hif Hisr]]]]]].
      unfold pTrue, pmstar.
      destruct S as [[[M R] fl] isr].
      simplS in * .
      rewrite HSP.
      sep on.
      assert (HMx : M ||- ((INV0 *^* INV1)
        *^* stkSpacePrec (sp-2-2-12) IH_Stk
        *^* (sp - 2 |->w bp)
        *^* (sp |->w pc)
        *^* (sp + 2 |->w cv)
        *^* mTrue)
        *^* (sp - 2 - 2 |->w w *^* (Aemp *^* stkSpacePrec (sp - 2 - 2) (_UNBLK_Stk + 2)))).
        sep lift 3 in HM.
        sep elim_left HM.
        sep inj.
        intro Heq; subst v.
        sep matchH HM with 8.
        generalize HM; clearall_arith; intros.
        clear HSP.
        sep stksplit (10+2) in HM.
        arith_rep (4 + 2 + 10 + IH_Stk - 2 - 2 - (10+2)) with 
          IH_Stk in HM.
      clear HM.
      destruct HMx as [M1 [M2 [HJM [HM1 HM2]]]].
      exists M1; exists M2; split_l.
      auto.
  split.
    (* p ==> p' *)
    intros pc' S'''.
    unfold p_cond_signal_5, p_cond_signal_6.
    unfold wrap_fun.
    unfold linkPG, wrap_fun, conjG.
    intros H.
    destruct H as [S'' [H HS''S''']].
    destruct H as [S' [HSS' HS'S'']].
    exists cv; exists bp; exists sp; exists pc.
    eapply Hpp'''; eauto.
    (* g''' ==> g *)
    intros pc' S''''.
    unfold g_cond_signal_5, g_cond_signal_6.
    unfold wrap_fun.
    intros HSS''''.
    unfold conjG at 1 in HSS'''' .
    destruct HSS'''' as [S''' [HSS''' H]].
    destruct HSS''' as [S'' [HSS'' HS''S''']].
    destruct HSS'' as [S' [HSS' HS'S'']].
    clear cv bp sp pc Hldsw HS.
    intros mP cv bp sp pc HS.
    apply H; trivial.
    eapply Hpp'''; eauto.
    (* p ==> if=false *)
    intros S HS.
    destruct HS as [cv [bp [sp [pc HS]]]].
    unfold Ast_cond_signal_5 in HS.
    InvertAll; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold p_cond_signal_5 in HS.
    destruct HS as [cv [ bp [sp [pc HS]]]].
    unfold Ast_cond_signal_5 in HS.
    destruct HS as [HM [_ [HSP [Hsp [Hcv [Hif Hisr]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold g_cond_signal_5 in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP cv bp sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_cond_signal_5 in HS.
  simplS in * .
  destruct HS as [HM [_ [HSP [Hsp [Hcv [Hif Hisr]]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma wff_cmd_cond_signal7:
   NH_WFcmd
     (SCAP_consCT (cond_signal + 1 + 1 + 1 + 1 + 1 + 1 + 1)
        (Ast p_cond_signal_7 g_cond_signal_7 1) psi_cond_signal)
     (cond_signal + 1 + 1 + 1 + 1 + 1 + 1)
     (Ast p_cond_signal_6 g_cond_signal_6 1) (comml _sti).
Proof.
  assert (Hpp': forall cv bp sp pc mP S S', 
    Ast_cond_signal_6 cv bp sp pc mP S
    -> gsti S S' 
    -> Ast_cond_signal_7 cv bp sp pc mP S').
    intros cv bp sp pc mP S S' HS HSS'.
    unfold linkPG, conjPG, conjG.
    unfold Ast_cond_signal_6, gsti, Ast_cond_signal_7 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct HS as [HM [HSP [Hsp [Hcv [Hif Hisr]]]]].
    destruct HSS' as [Hshift [Hif' [Hisr' HR']]].
    assert (Hshift' : prepost (INV0 *^* INV1) Aemp
      (M, R, fl, isr) (M', R', fl', isr')).
      apply (prepost_weaken_pre (mP':=INV0 *^* INV1) Hshift).
      intros Mx HMx.
      assert (Hitr : ~ itr_enabled (M, R, fl, isr)).
        unfold itr_enabled.
        intros [H1 H2].
        simplS in * .
        rewrite Hif in H1; discriminate H1.
      eapply choice_elim_right; eauto.
    clear Hshift; rename Hshift' into Hshift.
    split.
      sep on.
      substH HM with (prepost_elim Hshift Prec01 HM).
      sep on.
      simplS in * .
      sep matchH HM with 6.
      sep trivial.
    simplS in * .
    subst R' isr'.
    repeat split_l; trivial.
  simpl.
  split.
  exists p_cond_signal_7; exists g_cond_signal_7.
  split_l.
  split.
  (* enable *)
     unfold enable_sti.
     intros S HS.
     unfold p_cond_signal_6 in HS.
     unfold pTrue, pmstar, choice.
     destruct HS as [cv [bp [sp [pc HS]]]].
     unfold Ast_cond_signal_6 in HS.
     destruct HS as [HM [HSP [Hsp [Hcv [Hif Hisr]]]]].
     destruct S as [[[M R] fl] isr].
     simplS in * .
     destruct HM as [M2 [M1 [HJM [HM2 HM1]]]].
     exists M1; exists M2.
     split.
       auto with PMap.
     split_l.
     right.
     split_r.
     intros [H1 H2].
     simplS in * .
     rewrite Hif in H1; discriminate H1.
  split.
  (* p ==> p' *)
  unfold p_cond_signal_6, p_cond_signal_7.
  intros S' H.
  unfold linkPG in H.
  destruct H as [S [HS HSS']].
  destruct HS as [cv [bp [sp [pc HS]]]].
  exists cv; exists bp; exists sp; exists pc.
  apply Hpp' with S; trivial.
  (* g' ==> g *)
  unfold g_cond_signal_6, g_cond_signal_7.
  unfold linkPG, conjPG, conjG.
  intros S S'' H.
  destruct H as [S' H].
  destruct H as [[HxS HSS'] HS'S''].
  intros mP cv bp sp pc HS.
  apply HS'S''.
  apply Hpp' with S; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold p_cond_signal_6 in HS.
    destruct HS as [cv [ bp [sp [pc HS]]]].
    unfold Ast_cond_signal_6 in HS.
    destruct HS as [HM [HSP [Hsp [Hcv [Hif Hisr]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold g_cond_signal_6 in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP cv bp sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_cond_signal_6 in HS.
  simplS in * .
  destruct HS as [HM [HSP [Hsp [Hcv [Hif Hisr]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma wff_cmd_cond_signal8:
   NH_WFcmd
     (SCAP_consCT (cond_signal + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
        (Ast p_cond_signal_8 g_cond_signal_8 1) psi_cond_signal)
     (cond_signal + 1 + 1 + 1 + 1 + 1 + 1 + 1)
     (Ast p_cond_signal_7 g_cond_signal_7 1) (comml (_addw (Opr_imm 2) SP)).
Proof.
  assert (Hpp' : forall cv bp sp pc mP S S',
    Ast_cond_signal_7 cv bp sp pc mP S
    -> nextS (cond_signal + 1 + 1 + 1 + 1 + 1 + 1 + 1) (comml (_addw (Opr_imm 2) SP)) S S'
    -> Ast_cond_signal_8 cv bp sp pc mP S').
    intros cv bp sp pc mP S S' HS HSS'.
    unfold Ast_cond_signal_7, Ast_cond_signal_8 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    simplS in * .
    destruct HS as [HM [HSP [Hsp [Hcv [Hif Hisr]]]]].
    subst M' R' isr'.
    split.
      sep on.
      sep matchH HM with 5.
      sep stksplit. 
      sep matchH HM with 2.
    reg simpl.
    unfold opr_eval.
    rewrite HSP.
    split_l by omega.
    split_l.
    split_l.
    split_r.
    unfold trash_fl in Hfl'.
    destruct Hfl'; subst fl'; auto.
  split.
  simpl.
  exists p_cond_signal_8; exists g_cond_signal_8.
  split_l.
  split.
    (* enable addw *)
    apply trivial_enable_addw.
  split. (* p ==> p' *)
    unfold p_cond_signal_7, p_cond_signal_8.
    intros S' H.
    unfold linkPG in H.
    destruct H as [S H].
    simpl in H.
    destruct H as [H HnextS].
    destruct H as [cv [bp [sp [pc H]]]].
    exists cv; exists bp; exists sp; exists pc.
    apply Hpp' with S; trivial.
  (* gimp *)
  intros S S'' H.
  unfold conjPG, conjG in H.
  destruct H as [S' [[HSx HnextS] HS'S'']].
  unfold g_cond_signal_7, g_cond_signal_8 in * .
  intros mP cv bp sp pc HS.
  eapply HS'S''; eauto.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold p_cond_signal_7 in HS.
    destruct HS as [cv [bp [sp [pc HS]]]].
    unfold Ast_cond_signal_7 in HS.
    destruct HS as [HM [HSP [Hsp [Hcv [Hif Hisr]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    rewrite HSP.
    sep on.
    sep lift 0 in HM.
    sep stksplit (IH_Stk + 6) in HM.
    sep lift 0 in HM.
    destruct HM as [M1 [M2 [HJM [HM1 HM2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    split_l.
    intros M1' HM1'.
    unfold p_cond_signal_7.
    exists cv; exists bp; exists sp; exists pc.
    unfold Ast_cond_signal_7.
    simplS in * .
    rsplit.
      repeat split_l; trivial.
    sep on.
    sep lift 0.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      sep off.
      split.
        disjoint rep_l M1.
        eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    split_l.
    sep on.
    sep matchH HM2 with 4.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold g_cond_signal_7 in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP cv bp sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  apply HSS'.
  clear HSS'.
  unfold Ast_cond_signal_7 in * .
  simplS in * .
  destruct HS as [HM [HSP [Hsp [Hcv [Hif Hisr]]]]].
  rsplit.
    repeat split_l; trivial.
  sep on.
  sep lift 0 in HM.
  sep stksplit (IH_Stk + 6) in HM.
  sep lift 0 in HM.
  destruct HM as [M2x [M1x [HJM12 [HM2x HM1x]]]].
  sep lift 0.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  assert (M2x = M2' /\ M1x = M1).
    sep off.
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    rewrite HSP in HM2'.
    assert (Hx := prec_disjUnion_inv H H0 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  rewrite HSP in HM2.
  split_l.
  sep on.
  sep matchH HM1x with 3.
Qed.

Lemma wff_cmd_cond_signal9:
   NH_WFcmd
     (SCAP_consCT (cond_signal + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
        (Ast p_cond_signal_9 g_cond_signal_9 1) psi_cond_signal)
     (cond_signal + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
     (Ast p_cond_signal_8 g_cond_signal_8 1) (comml (_popw BP)).
Proof.
  assert (Hpp' : forall cv bp sp pc mP S S',
    Ast_cond_signal_8 cv bp sp pc mP S
    -> nextS (cond_signal + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1) (comml (_popw BP)) S S'
    -> Ast_cond_signal_9 cv bp sp pc mP S').
    intros cv bp sp pc mP S S' HS HSS'.
    unfold Ast_cond_signal_8, Ast_cond_signal_9 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HR' as [bpx [Hlds HR']].
    simplS in * .
    destruct HS as [HM [HSP [Hsp [Hcv [Hif Hisr]]]]].
    assert (Heq : bp = bpx).
      sep on.
      sep lift 1 in HM.
      rewrite HSP in Hlds.
      sep inj.
      auto.
    subst bpx.
    subst M' fl' R' isr'.
    split.
      sep on.
      sep matchH HM with 5.
      sep stksplit.
      sep matchH HM with 2.
    rewrite HSP.
    reg simpl.
    split_l.
    split_l by omega.
    auto.
  simpl.
  split.
  exists p_cond_signal_9; exists g_cond_signal_9.
  split_l.
  split.
  (* enable *)
    apply enable_popw.
    intros S HS.
    destruct S as [[[M R] fl] isr].
    unfold p_cond_signal_8 in HS.
    destruct HS as [cv [bp [sp [pc HS]]]].
    unfold Ast_cond_signal_8 in HS.
    destruct HS as [HM [HSP [Hsp [Hcv [Hif Hisr]]]]].
    simplS in * .
    rewrite HSP.
    sep on.
    sep lift 1 in HM.
    sep trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold p_cond_signal_8, p_cond_signal_9.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [cv [bp [sp [pc HS]]]].
    exists cv; exists bp; exists sp; exists pc.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HxS HSS'] HS'S''].
  unfold g_cond_signal_8, g_cond_signal_9 in * .
  intros mP cv bp sp pc HS.
  apply HS'S''.
  apply Hpp' with S; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold p_cond_signal_8 in HS.
    destruct HS as [cv [bp [sp [pc HS]]]].
    unfold Ast_cond_signal_8 in HS.
    destruct HS as [HM [HSP [Hsp [Hcv [Hif Hisr]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    rewrite HSP.
    sep on.
    sep lift 0 in HM.
    sep stksplit (IH_Stk + 6) in HM.
    sep lift 0 in HM.
    destruct HM as [M1 [M2 [HJM [HM1 HM2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    split_l.
    intros M1' HM1'.
    unfold p_cond_signal_8.
    exists cv; exists bp; exists sp; exists pc.
    unfold Ast_cond_signal_8.
    simplS in * .
    rsplit.
      repeat split_l; trivial.
    sep on.
    sep lift 0.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      sep off.
      split.
        disjoint rep_l M1.
        eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    split_l.
    sep on.
    sep matchH HM2 with 4.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold g_cond_signal_8 in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP cv bp sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  apply HSS'.
  clear HSS'.
  unfold Ast_cond_signal_8 in * .
  simplS in * .
  destruct HS as [HM [HSP [Hsp [Hcv [Hif Hisr]]]]].
  rsplit.
    repeat split_l; trivial.
  sep on.
  sep lift 0 in HM.
  sep stksplit (IH_Stk + 6) in HM.
  sep lift 0 in HM.
  destruct HM as [M2x [M1x [HJM12 [HM2x HM1x]]]].
  sep lift 0.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  assert (M2x = M2' /\ M1x = M1).
    sep off.
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    rewrite HSP in HM2'.
    assert (Hx := prec_disjUnion_inv H H0 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  rewrite HSP in HM2.
  split_l.
  sep on.
  sep matchH HM1x with 3.
Qed.

Lemma wff_cmd_cond_signal10:
 NH_WFiseq psi_cond_signal (cond_signal + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
   (Ast p_cond_signal_9 g_cond_signal_9 1) (sseq _retw).
Proof.
  constructor.
  simpl.
  split.
  split_l.
  intros S HS.
  unfold g_cond_signal_9, p_cond_signal_9 in * .
  unfold Ast_cond_signal_9, Ast_cond_signal_post, Ast_cond_signal_pre.
  intros mP cv bp sp pc H.
  InvertAll.
  repeat split_l; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold p_cond_signal_9 in HS.
    destruct HS as [cv [bp [sp [pc HS]]]].
    unfold Ast_cond_signal_9 in HS.
    destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hif Hisr]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    rewrite HSP.
    sep on.
    sep lift 0 in HM.
    sep stksplit (IH_Stk + 6) in HM.
    sep lift 0 in HM.
    destruct HM as [M1 [M2 [HJM [HM1 HM2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    split_l.
    intros M1' HM1'.
    unfold p_cond_signal_9.
    exists cv; exists bp; exists sp; exists pc.
    unfold Ast_cond_signal_9.
    simplS in * .
    rsplit.
      repeat split_l; trivial.
    sep on.
    sep lift 0.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      sep off.
      split.
        disjoint rep_l M1.
        eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    split_l.
    sep on.
    sep matchH HM2 with 4.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold g_cond_signal_9 in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP cv bp sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  apply HSS'.
  clear HSS'.
  unfold Ast_cond_signal_9 in * .
  simplS in * .
  destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hif Hisr]]]]]].
  rsplit.
    repeat split_l; trivial.
  sep on.
  sep lift 0 in HM.
  sep stksplit (IH_Stk + 6) in HM.
  sep lift 0 in HM.
  destruct HM as [M2x [M1x [HJM12 [HM2x HM1x]]]].
  sep lift 0.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  assert (M2x = M2' /\ M1x = M1).
    sep off.
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    rewrite HSP in HM2'.
    assert (Hx := prec_disjUnion_inv H H0 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  rewrite HSP in HM2.
  split_l.
  sep on.
  sep matchH HM1x with 3.
Qed.

Lemma wff_cond_signal:
  NH_WFiseq psi_cond_signal cond_signal (Ast cond_signal_Prd cond_signal_grt 1) I_cond_signal.
Proof.
  unfold I_cond_signal.
  apply nh_isq_seq with p_cond_signal_1 g_cond_signal_1.
  apply wff_cmd_cond_signal1.
  apply nh_isq_seq with p_cond_signal_2 g_cond_signal_2.
  apply wff_cmd_cond_signal2.
  apply nh_isq_seq with p_cond_signal_3 g_cond_signal_3.
  apply wff_cmd_cond_signal3.
  apply nh_isq_seq with p_cond_signal_4 g_cond_signal_4.
  apply wff_cmd_cond_signal4.
  apply nh_isq_seq with p_cond_signal_5 g_cond_signal_5.
  apply wff_cmd_cond_signal5.
  apply nh_isq_seq with p_cond_signal_6 g_cond_signal_6.
  apply wff_cmd_cond_signal6.
  apply nh_isq_seq with p_cond_signal_7 g_cond_signal_7.
  apply wff_cmd_cond_signal7.
  apply nh_isq_seq with p_cond_signal_8 g_cond_signal_8.
  apply wff_cmd_cond_signal8.
  apply nh_isq_seq with p_cond_signal_9 g_cond_signal_9.
  apply wff_cmd_cond_signal9.
  apply wff_cmd_cond_signal10.
Qed.
