(***************************************************************)
(* Coq code for AIM                                            *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

Add LoadPath ".." .
Add LoadPath "../CAP0" .
Add LoadPath "../SCAP" .
Add LoadPath "../miniOS" .
Add LoadPath "../examples" .


Require Import Coqlib.

Require Import List.
Require Import Arith.
Require Import Machine.
Require Import Reg.
Require Import Mem_Lemmas.
Require Import Assertion.
Require Import Assertion_Lemmas.
Require Import Assertion_Lemmas3.
Require Import Assertion_Lemmas4.
Require Import Assertion_Lemmas5.
Require Import Operational_Semantics.
Require Import OpSem_Lemmas.
Require Import OpSem_Lemmas2.

Require Import AIM.
Require Import AIM_Lemmas.
Require Import QLib.

Require Import SCAP.
Require Import SCAP_READY.
Require Import Linking.
Require Import Omega.
Require Import Assembly.
Require Import SyncUtil.
Require Import Kernel.
Require Import SleepSpec.
Require Import lock_release.
Require Import lock_acquire.

(* the stack space required by cond_wait *)
(* assuming:
   _STK_SIZE_l_acquire > 2 + IH_Stk
   _STK_SIZE_l_acquire > stk_l_release
   _STK_SIZE_l_acquire > 2 + _STK_SIZE_sleep *)
Notation _STK_SIZE_cond_wait := (40 + 6 + IH_Stk) (only parsing).

(* Definition _STK_SIZE_cond_wait := 6 + _STK_SIZE_l_acquire. *)

Definition psi_cond_wait :=
   SCAP_consCT l_acquire (Ast l_acquire_Prd l_acquire_grt 1)  
  (SCAP_consCT l_release (Ast l_release_Prd l_release_grt 1)
  (SCAP_consCT L_sleep   (Ast p_sleep g_sleep 1)
                         SCAP_emptCT
  )).

Definition Ast_l_release_pre (lk bp sp pc : Label) (mP : MPred) (S : State) : Prop :=
  ((INV0 *^* INV1) 
    *^* (stkSpacePrec sp stk_l_release)
    *^* (sp |->w pc)
    *^* (sp + 2 |->w lk)
    *^* (Gamma lk)
    *^* mP) (S_Mem S)
  /\ S_RF S BP = bp
  /\ S_RF S SP = sp
  /\ In lk lockSet_H
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.
  
Definition Ast_l_release_post (lk bp sp pc : Label) (mP : MPred) (S : State) : Prop :=
  ((INV0 *^* INV1) 
    *^* (stkSpacePrec sp stk_l_release)
    *^* (sp |->w pc)
    *^* (sp + 2 |->w lk)
    *^* mP) (S_Mem S)
  /\ S_RF S BP = bp
  /\ S_RF S SP = sp
  /\ In lk lockSet_H
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.

Definition p_l_release (S : State) : Prop :=
  exists lk, exists bp, exists sp, exists pc,
      Ast_l_release_pre lk bp sp pc mTrue S.

Definition g_l_release (S S' : State) : Prop := forall mP,
  forall lk bp sp pc,
    Ast_l_release_pre lk bp sp pc mP S
    -> Ast_l_release_post lk bp sp pc mP S'.

Lemma p_l_release_sound : forall S,
  p_l_release S -> l_release_Prd S.
Proof.
  intros [[[M R] fl] isr].
  unfold p_l_release, l_release_Prd.
  unfold Ast_l_release_pre, l_release_Prd_aux.
  intros H.
  destruct H as [lk [bp [sp [pc H]]]].
  exists sp; exists lk; exists pc; exists bp; exists mTrue.
  destruct H as [HM [HBP [HSP [Hlk [Hif Hisr]]]]].
  simplS in * .
  unfold l_release_MPre.
  split_l; split_l.
  split.
  sep on.
  sep matchH HM with 7.
  auto.
Qed.

Lemma g_l_release_sound : forall S S',
  l_release_grt S S' -> g_l_release S S'.
Proof.
  intros [[[M R] fl] isr] [[[M' R'] fl'] isr'].
  unfold l_release_grt, g_l_release.
  unfold leadstoPQ2, l_release_Prd_aux, l_release_Post_aux_0.
  unfold Ast_l_release_pre, Ast_l_release_post.
  unfold l_release_MPre, l_release_MPost.
  intros.
  simplS in * .
  InvertAll.
  assert (R SP = sp /\
      R BP = bp /\
      (sp |->w pc *^* sp + 2 |->w lk *^* stkSpacePrec sp stk_l_release *^*
       INV1 *^* INV0 *^* Gamma lk *^* mP) M /\
      In lk lockSet_H /\ fl _IF = false /\ isr = false).
  split_l; split_l.
  split.
    sep on.
    sep matchH H1 with 7.
  auto.
  substH H with (H _ _ _ _ _ H5).
  clear H5.
  InvertAll.
  split.
  destruct H7 as [M1' [M2' [HJM' [HM1' HM2']]]].
  destruct HM1' as [HM1' Hlk].
  exists M1'; exists M2'; split_l.
  split_r.
  sep on.
  sep matchH HM1' with 7.
  auto.
Qed.

Definition Ast_l_acquire_pre (lk bp sp pc : Label) (mP : MPred) (S : State) : Prop :=
  ((INV0 *^* INV1) 
    *^* (stkSpacePrec sp _STK_SIZE_l_acquire)
    *^* (sp |->w pc)
    *^* (sp + 2 |->w lk)
    *^* mP) (S_Mem S)
  /\ S_RF S BP = bp
  /\ S_RF S SP = sp
  /\ In lk lockSet_H
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.
  
Definition Ast_l_acquire_post (lk bp sp pc : Label) (mP : MPred) (S : State) : Prop :=
  ((INV0 *^* INV1) 
    *^* (stkSpacePrec sp _STK_SIZE_l_acquire)
    *^* (sp |->w pc)
    *^* (sp + 2 |->w lk)
    *^* (Gamma lk)
    *^* mP) (S_Mem S)
  /\ S_RF S BP = bp
  /\ S_RF S SP = sp
  /\ In lk lockSet_H
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.

Definition p_l_acquire (S : State) : Prop :=
  exists lk, exists bp, exists sp, exists pc,
      Ast_l_acquire_pre lk bp sp pc mTrue S.

Definition g_l_acquire (S S' : State) : Prop := forall mP,
  forall lk bp sp pc,
    Ast_l_acquire_pre lk bp sp pc mP S
    -> Ast_l_acquire_post lk bp sp pc mP S'.

Lemma p_l_acquire_sound : forall S,
  p_l_acquire S -> l_acquire_Prd S.
Proof.
  intros [[[M R] fl] isr].
  unfold p_l_acquire, l_acquire_Prd.
  unfold Ast_l_acquire_pre, l_acquire_Prd_aux.
  intros H.
  destruct H as [lk [bp [sp [pc H]]]].
  exists sp; exists lk; exists pc; exists bp; exists mTrue.
  destruct H as [HM [HBP [HSP [Hlk [Hif Hisr]]]]].
  simplS in * .
  unfold l_acquire_MPre.
  split_l; split_l.
  split.
  sep on.
  sep matchH HM with 7.
  auto.
Qed.

Lemma g_l_acquire_sound : forall S S',
  l_acquire_grt S S' -> g_l_acquire S S'.
Proof.
  intros [[[M R] fl] isr] [[[M' R'] fl'] isr'].
  unfold l_acquire_grt, g_l_acquire.
  unfold leadstoPQ2, l_acquire_Prd_aux, l_acquire_Post_aux.
  unfold Ast_l_acquire_pre, Ast_l_acquire_post.
  unfold l_acquire_MPre, l_acquire_MPost.
  intros.
  simplS in * .
  InvertAll.
  assert (R SP = sp /\
      R BP = bp /\
      (sp |->w pc *^* sp + 2 |->w lk *^* stkSpacePrec sp _STK_SIZE_l_acquire *^*
       INV1 *^* INV0 *^* mP) M /\
      In lk lockSet_H /\ fl _IF = false /\ isr = false).
  split_l; split_l.
  split.
    sep on.
    sep matchH H1 with 7.
  auto.
  substH H with (H _ _ _ _ _ H5).
  clear H5.
  InvertAll.
  split.
  unfold l_acquire_MPre in H7.
  sep on.
  sep matchH H7 with 7.
  auto.
Qed.

Definition cond_wait_MPre (sp cv l pc : Label) (M : Mem) : Prop :=
    ((sp |->w pc) *^* (sp + 2 |->w cv) *^* (sp + 4 |->w l) *^* stkSpacePrec sp _STK_SIZE_cond_wait
       *^* (Gamma l)) M /\ In l lockSet_H /\ In cv cvSet_M.

Definition cond_wait_Prd_aux (sp cv l pc bp : Label) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
      R SP = sp
      /\ R BP = bp
      /\ (cond_wait_MPre sp cv l pc) M
      /\ fl _IF = true
      /\ isr = false
  end.

Definition cond_wait_Prd (S : State) : Prop :=
  imprec (fun S => exists sp, exists cv, exists l, exists pc, exists bp, cond_wait_Prd_aux sp cv l pc bp S) S.

Definition cond_wait_MPost := cond_wait_MPre.

Definition cond_wait_Post_aux := cond_wait_Prd_aux.

Definition cond_wait_grt (S S' : State) : Prop :=
  forall sp cv l pc bp,
    leadstoPQ (cond_wait_Prd_aux sp cv l pc bp) (cond_wait_Post_aux sp cv l pc bp) S S'.

Definition Ast_cond_wait_pre (lk cv bp sp pc : Label) (mP : MPred)
  (S : State) : Prop :=
  (stkSpacePrec sp _STK_SIZE_cond_wait
    *^* (sp |->w pc)
    *^* (sp + 2 |->w cv)
    *^* (sp + 4 |->w lk)
    *^* (Gamma lk)
    *^* mP) (S_Mem S)
  /\ S_RF S BP = bp
  /\ S_RF S SP = sp
  /\ sp >= _STK_SIZE_cond_wait
  /\ In cv cvSet_M
  /\ In lk lockSet_H
  /\ S_Flag S _IF = true
  /\ S_ISR S = false.

Definition Ast_cond_wait_post := Ast_cond_wait_pre.

Definition p_cond_wait_0 (S : State) : Prop :=
  exists lk, exists cv, exists bp, exists sp, exists pc,
    Ast_cond_wait_pre lk cv bp sp pc mTrue S.

Definition g_cond_wait_0 (S S' : State) : Prop := forall mP,
  forall lk cv bp sp pc,
    Ast_cond_wait_pre lk cv bp sp pc mP S
    -> Ast_cond_wait_post lk cv bp sp pc mP S'.

(* <cond_wait + 0> : _pushwp (Opr_reg BP); *)

Definition Ast_cond_wait_1 (lk cv bp sp pc : Label) (mP : MPred)
  (S : State) : Prop :=
  (stkSpacePrec (sp - 2) (_STK_SIZE_cond_wait - 2)
    *^* (sp - 2 |->w bp)
    *^* (sp |->w pc)
    *^* (sp + 2 |->w cv)
    *^* (sp + 4 |->w lk)
    *^* (Gamma lk)
    *^* mP) (S_Mem S)
  /\ S_RF S BP = bp
  /\ S_RF S SP = sp - 2
  /\ sp >= _STK_SIZE_cond_wait
  /\ In cv cvSet_M
  /\ In lk lockSet_H
  /\ S_Flag S _IF = true
  /\ S_ISR S = false.

Definition p_cond_wait_1 (S : State) : Prop :=
  exists lk, exists cv, exists bp, exists sp, exists pc,
    Ast_cond_wait_1 lk cv bp sp pc mTrue S.

Definition g_cond_wait_1 (S S' : State) : Prop := forall mP,
  forall lk cv bp sp pc,
    Ast_cond_wait_1 lk cv bp sp pc mP S
    -> Ast_cond_wait_post lk cv bp sp pc mP S'.

(* <cond_wait + 1> : _movwp  (Opr_reg SP) BP; *)

Definition Ast_cond_wait_2 (lk cv bp sp pc : Label) (mP : MPred)
  (S : State) : Prop :=
  (stkSpacePrec (sp - 2) (_STK_SIZE_cond_wait - 2)
    *^* (sp - 2 |->w bp)
    *^* (sp |->w pc)
    *^* (sp + 2 |->w cv)
    *^* (sp + 4 |->w lk)
    *^* (Gamma lk)
    *^* mP) (S_Mem S)
  /\ S_RF S BP = sp - 2
  /\ S_RF S SP = sp - 2
  /\ sp >= _STK_SIZE_cond_wait
  /\ In cv cvSet_M
  /\ In lk lockSet_H
  /\ S_Flag S _IF = true
  /\ S_ISR S = false.

Definition p_cond_wait_2 (S : State) : Prop :=
  exists lk, exists cv, exists bp, exists sp, exists pc,
    Ast_cond_wait_2 lk cv bp sp pc mTrue S.

Definition g_cond_wait_2 (S S' : State) : Prop := forall mP,
  forall lk cv bp sp pc,
    Ast_cond_wait_2 lk cv bp sp pc mP S
    -> Ast_cond_wait_post lk cv bp sp pc mP S'.

(* <cond_wait + 2> : _pushwm (A_reg 6 BP); *)

Definition Ast_cond_wait_3 (lk cv bp sp pc : Label) (mP : MPred)
  (S : State) : Prop :=
  (stkSpacePrec (sp - 4) (_STK_SIZE_cond_wait - 4)
    *^* (sp - 4 |->w lk)
    *^* (sp - 2 |->w bp)
    *^* (sp |->w pc)
    *^* (sp + 2 |->w cv)
    *^* (sp + 4 |->w lk)
    *^* (Gamma lk)
    *^* mP) (S_Mem S)
  /\ S_RF S BP = sp - 2
  /\ S_RF S SP = sp - 4
  /\ sp >= _STK_SIZE_cond_wait
  /\ In cv cvSet_M
  /\ In lk lockSet_H
  /\ S_Flag S _IF = true
  /\ S_ISR S = false.

Definition p_cond_wait_3 (S : State) : Prop :=
  exists lk, exists cv, exists bp, exists sp, exists pc,
    Ast_cond_wait_3 lk cv bp sp pc mTrue S.

Definition g_cond_wait_3 (S S' : State) : Prop := forall mP,
  forall lk cv bp sp pc,
    Ast_cond_wait_3 lk cv bp sp pc mP S
    -> Ast_cond_wait_post lk cv bp sp pc mP S'.

(* <cond_wait + 3> :  _cli; *)

Definition Ast_cond_wait_4 (lk cv bp sp pc : Label) (mP : MPred)
  (S : State) : Prop :=
  ((INV0 *^* INV1)
    *^* (stkSpacePrec (sp - 4) (_STK_SIZE_cond_wait - 4))
    *^* (sp - 4 |->w lk)
    *^* (sp - 2 |->w bp)
    *^* (sp |->w pc)
    *^* (sp + 2 |->w cv)
    *^* (sp + 4 |->w lk)
    *^* (Gamma lk)
    *^* mP) (S_Mem S)
  /\ S_RF S BP = sp - 2
  /\ S_RF S SP = sp - 4
  /\ sp >= _STK_SIZE_cond_wait
  /\ In cv cvSet_M
  /\ In lk lockSet_H
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.

Definition p_cond_wait_4 (S : State) : Prop :=
  exists lk, exists cv, exists bp, exists sp, exists pc,
    Ast_cond_wait_4 lk cv bp sp pc mTrue S.

Definition g_cond_wait_4 (S S' : State) : Prop := forall mP,
  forall lk cv bp sp pc,
    Ast_cond_wait_4 lk cv bp sp pc mP S
    -> Ast_cond_wait_post lk cv bp sp pc mP S'.

(* <cond_wait + 4> :  _call   l_release; *)

Definition Ast_cond_wait_5 (lk cv bp sp pc : Label) (mP : MPred)
  (S : State) : Prop :=
  ((INV0 *^* INV1)
    *^* (stkSpacePrec (sp - 4) (_STK_SIZE_cond_wait - 4))
    *^* (sp - 4 |->w lk)
    *^* (sp - 2 |->w bp)
    *^* (sp |->w pc)
    *^* (sp + 2 |->w cv)
    *^* (sp + 4 |->w lk)
    *^* mP) (S_Mem S)
  /\ S_RF S BP = sp - 2
  /\ S_RF S SP = sp - 4
  /\ sp >= _STK_SIZE_cond_wait
  /\ In cv cvSet_M
  /\ In lk lockSet_H
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.

Definition p_cond_wait_5 (S : State) : Prop :=
  exists lk, exists cv, exists bp, exists sp, exists pc,
    Ast_cond_wait_5 lk cv bp sp pc mTrue S.

Definition g_cond_wait_5 (S S' : State) : Prop := forall mP,
  forall lk cv bp sp pc,
    Ast_cond_wait_5 lk cv bp sp pc mP S
    -> Ast_cond_wait_post lk cv bp sp pc mP S'.

(* <cond_wait + 5> : _movwld (A_reg 4 BP) BX; *)

Definition Ast_cond_wait_6 (lk cv bp sp pc : Label) (mP : MPred)
  (S : State) : Prop :=
  ((INV0 *^* INV1)
    *^* (stkSpacePrec (sp - 4) (_STK_SIZE_cond_wait - 4))
    *^* (sp - 4 |->w lk)
    *^* (sp - 2 |->w bp)
    *^* (sp |->w pc)
    *^* (sp + 2 |->w cv)
    *^* (sp + 4 |->w lk)
    *^* mP) (S_Mem S)
  /\ S_RF S BX = cv
  /\ S_RF S SP = sp - 4
  /\ sp >= _STK_SIZE_cond_wait
  /\ In cv cvSet_M
  /\ In lk lockSet_H
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.

Definition p_cond_wait_6 (S : State) : Prop :=
  exists lk, exists cv, exists bp, exists sp, exists pc,
    Ast_cond_wait_6 lk cv bp sp pc mTrue S.

Definition g_cond_wait_6 (S S' : State) : Prop := forall mP,
  forall lk cv bp sp pc,
    Ast_cond_wait_6 lk cv bp sp pc mP S
    -> Ast_cond_wait_post lk cv bp sp pc mP S'.

(* <cond_wait + 6> : _pushwm (A_reg C_WAITING BX); *)

Definition Ast_cond_wait_7 (lk cv bp sp pc : Label) (mP : MPred)
  (S : State) : Prop :=
  ((INV0 *^* INV1)
    *^* (stkSpacePrec (sp - 6) (_STK_SIZE_cond_wait - 6))
    *^* (sp - 6 |->w_)
    *^* (sp - 4 |->w lk)
    *^* (sp - 2 |->w bp)
    *^* (sp |->w pc)
    *^* (sp + 2 |->w cv)
    *^* (sp + 4 |->w lk)
    *^* mP) (S_Mem S)
  /\ (exists w, (sp - 6 |~>w w) (S_Mem S) 
    /\ Delta w = SomeT Aemp)
  /\ S_RF S SP = sp - 6
  /\ sp >= _STK_SIZE_cond_wait
  /\ In cv cvSet_M
  /\ In lk lockSet_H
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.

Definition p_cond_wait_7 (S : State) : Prop :=
  exists lk, exists cv, exists bp, exists sp, exists pc,
    Ast_cond_wait_7 lk cv bp sp pc mTrue S.

Definition g_cond_wait_7 (S S' : State) : Prop := forall mP,
  forall lk cv bp sp pc,
    Ast_cond_wait_7 lk cv bp sp pc mP S
    -> Ast_cond_wait_post lk cv bp sp pc mP S'.

(* <cond_wait + 7> : _call   L_sleep; *)

Definition Ast_cond_wait_8 (lk cv bp sp pc : Label) (mP : MPred)
  (S : State) : Prop :=
  ((INV0 *^* INV1)
    *^* (stkSpacePrec (sp - 6) (_STK_SIZE_cond_wait - 6)
    *^* (sp - 6 |->w_)
    *^* (sp - 4 |->w lk)
    *^* (sp - 2 |->w bp)
    *^* (sp |->w pc)
    *^* (sp + 2 |->w cv)
    *^* (sp + 4 |->w lk)
    *^* mP)) (S_Mem S)
  /\ S_RF S SP = sp - 6
  /\ sp >= _STK_SIZE_cond_wait
  /\ In cv cvSet_M
  /\ In lk lockSet_H
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.

Definition p_cond_wait_8 (S : State) : Prop :=
  exists lk, exists cv, exists bp, exists sp, exists pc,
    Ast_cond_wait_8 lk cv bp sp pc mTrue S.

Definition g_cond_wait_8 (S S' : State) : Prop := forall mP,
  forall lk cv bp sp pc,
    Ast_cond_wait_8 lk cv bp sp pc mP S
    -> Ast_cond_wait_post lk cv bp sp pc mP S'.

(* <cond_wait + 8> :  _sti; *)

Definition Ast_cond_wait_9 (lk cv bp sp pc : Label) (mP : MPred)
  (S : State) : Prop :=
  (stkSpacePrec (sp - 6) (_STK_SIZE_cond_wait - 6)
    *^* (sp - 6 |->w_)
    *^* (sp - 4 |->w lk)
    *^* (sp - 2 |->w bp)
    *^* (sp |->w pc)
    *^* (sp + 2 |->w cv)
    *^* (sp + 4 |->w lk)
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 6
  /\ sp >= _STK_SIZE_cond_wait
  /\ In cv cvSet_M
  /\ In lk lockSet_H
  /\ S_Flag S _IF = true
  /\ S_ISR S = false.

Definition p_cond_wait_9 (S : State) : Prop :=
  exists lk, exists cv, exists bp, exists sp, exists pc,
    Ast_cond_wait_9 lk cv bp sp pc mTrue S.

Definition g_cond_wait_9 (S S' : State) : Prop := forall mP,
  forall lk cv bp sp pc,
    Ast_cond_wait_9 lk cv bp sp pc mP S
    -> Ast_cond_wait_post lk cv bp sp pc mP S'.

(* <cond_wait + 9> :  _addw   (Opr_imm 2) SP; *)

Definition Ast_cond_wait_a (lk cv bp sp pc : Label) (mP : MPred)
  (S : State) : Prop :=
  (stkSpacePrec (sp - 4) (_STK_SIZE_cond_wait - 4)
    *^* (sp - 4 |->w lk)
    *^* (sp - 2 |->w bp)
    *^* (sp |->w pc)
    *^* (sp + 2 |->w cv)
    *^* (sp + 4 |->w lk)
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 4
  /\ sp >= _STK_SIZE_cond_wait
  /\ In cv cvSet_M
  /\ In lk lockSet_H
  /\ S_Flag S _IF = true
  /\ S_ISR S = false.

Definition p_cond_wait_a (S : State) : Prop :=
  exists lk, exists cv, exists bp, exists sp, exists pc,
    Ast_cond_wait_a lk cv bp sp pc mTrue S.

Definition g_cond_wait_a (S S' : State) : Prop := forall mP,
  forall lk cv bp sp pc,
    Ast_cond_wait_a lk cv bp sp pc mP S
    -> Ast_cond_wait_post lk cv bp sp pc mP S'.

(* <cond_wait + a> :  _cli; *)

Definition Ast_cond_wait_b (lk cv bp sp pc : Label) (mP : MPred)
  (S : State) : Prop :=
  ((INV0 *^* INV1)
    *^* (stkSpacePrec (sp - 4) (_STK_SIZE_cond_wait - 4))
    *^* (sp - 4 |->w lk)
    *^* (sp - 2 |->w bp)
    *^* (sp |->w pc)
    *^* (sp + 2 |->w cv)
    *^* (sp + 4 |->w lk)
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 4
  /\ sp >= _STK_SIZE_cond_wait
  /\ In cv cvSet_M
  /\ In lk lockSet_H
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.

Definition p_cond_wait_b (S : State) : Prop :=
  exists lk, exists cv, exists bp, exists sp, exists pc,
    Ast_cond_wait_b lk cv bp sp pc mTrue S.

Definition g_cond_wait_b (S S' : State) : Prop := forall mP,
  forall lk cv bp sp pc,
    Ast_cond_wait_b lk cv bp sp pc mP S
    -> Ast_cond_wait_post lk cv bp sp pc mP S'.

(* <cond_wait + b> :  _call   l_acquire; *)

Definition Ast_cond_wait_c (lk cv bp sp pc : Label) (mP : MPred)
  (S : State) : Prop :=
  ((INV0 *^* INV1)
    *^* (stkSpacePrec (sp - 4) (_STK_SIZE_cond_wait - 4)
    *^* (sp - 4 |->w lk)
    *^* (sp - 2 |->w bp)
    *^* (sp |->w pc)
    *^* (sp + 2 |->w cv)
    *^* (sp + 4 |->w lk)
    *^* (Gamma lk)
    *^* mP)) (S_Mem S)
  /\ S_RF S SP = sp - 4
  /\ sp >= _STK_SIZE_cond_wait
  /\ In cv cvSet_M
  /\ In lk lockSet_H
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.

Definition p_cond_wait_c (S : State) : Prop :=
  exists lk, exists cv, exists bp, exists sp, exists pc,
    Ast_cond_wait_c lk cv bp sp pc mTrue S.

Definition g_cond_wait_c (S S' : State) : Prop := forall mP,
  forall lk cv bp sp pc,
    Ast_cond_wait_c lk cv bp sp pc mP S
    -> Ast_cond_wait_post lk cv bp sp pc mP S'.

(* <cond_wait + c> :  _sti; *)

Definition Ast_cond_wait_d (lk cv bp sp pc : Label) (mP : MPred)
  (S : State) : Prop :=
  (stkSpacePrec (sp - 4) (_STK_SIZE_cond_wait - 4)
    *^* (sp - 4 |->w lk)
    *^* (sp - 2 |->w bp)
    *^* (sp |->w pc)
    *^* (sp + 2 |->w cv)
    *^* (sp + 4 |->w lk)
    *^* (Gamma lk)
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 4
  /\ sp >= _STK_SIZE_cond_wait
  /\ In cv cvSet_M
  /\ In lk lockSet_H
  /\ S_Flag S _IF = true
  /\ S_ISR S = false.

Definition p_cond_wait_d (S : State) : Prop :=
  exists lk, exists cv, exists bp, exists sp, exists pc,
    Ast_cond_wait_d lk cv bp sp pc mTrue S.

Definition g_cond_wait_d (S S' : State) : Prop := forall mP,
  forall lk cv bp sp pc,
    Ast_cond_wait_d lk cv bp sp pc mP S
    -> Ast_cond_wait_post lk cv bp sp pc mP S'.

(* <cond_wait + d> :  _addw   (Opr_imm 2) SP; *)

Definition Ast_cond_wait_e (lk cv bp sp pc : Label) (mP : MPred)
  (S : State) : Prop :=
  (stkSpacePrec (sp - 2) (_STK_SIZE_cond_wait - 2)
    *^* (sp - 2 |->w bp)
    *^* (sp |->w pc)
    *^* (sp + 2 |->w cv)
    *^* (sp + 4 |->w lk)
    *^* (Gamma lk)
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 2
  /\ sp >= _STK_SIZE_cond_wait
  /\ In cv cvSet_M
  /\ In lk lockSet_H
  /\ S_Flag S _IF = true
  /\ S_ISR S = false.

Definition p_cond_wait_e (S : State) : Prop :=
  exists lk, exists cv, exists bp, exists sp, exists pc,
    Ast_cond_wait_e lk cv bp sp pc mTrue S.

Definition g_cond_wait_e (S S' : State) : Prop := forall mP,
  forall lk cv bp sp pc,
    Ast_cond_wait_e lk cv bp sp pc mP S
    -> Ast_cond_wait_post lk cv bp sp pc mP S'.

(* <cond_wait + e> :  _popw   BP; *)

Definition Ast_cond_wait_f (lk cv bp sp pc : Label) (mP : MPred)
  (S : State) : Prop :=
  (stkSpacePrec sp (_STK_SIZE_cond_wait)
    *^* (sp |->w pc)
    *^* (sp + 2 |->w cv)
    *^* (sp + 4 |->w lk)
    *^* (Gamma lk)
    *^* mP) (S_Mem S)
  /\ S_RF S BP = bp
  /\ S_RF S SP = sp
  /\ sp >= _STK_SIZE_cond_wait
  /\ In cv cvSet_M
  /\ In lk lockSet_H
  /\ S_Flag S _IF = true
  /\ S_ISR S = false.

Definition p_cond_wait_f (S : State) : Prop :=
  exists lk, exists cv, exists bp, exists sp, exists pc,
    Ast_cond_wait_f lk cv bp sp pc mTrue S.

Definition g_cond_wait_f (S S' : State) : Prop := forall mP,
  forall lk cv bp sp pc,
    Ast_cond_wait_f lk cv bp sp pc mP S
    -> Ast_cond_wait_post lk cv bp sp pc mP S'.

(* <cond_wait + f> :  _retw;;  *)

Definition p_cond_wait := p_cond_wait_0.

Definition g_cond_wait := g_cond_wait_0.

Lemma wff_cmd_cond_wait1:
   NH_WFcmd
     (SCAP_consCT (cond_wait + 1) (Ast p_cond_wait_1 g_cond_wait_1 1)
        psi_cond_wait) cond_wait (Ast p_cond_wait g_cond_wait 1)
     (comml (_pushwp (Opr_reg BP))).
Proof.
  unfold p_cond_wait, g_cond_wait.
  assert (Hpp' : forall lk cv bp sp pc mP S S',
    Ast_cond_wait_pre lk cv bp sp pc mP S
    -> nextS (cond_wait) (comml (_pushwp (Opr_reg BP))) S S'
    -> Ast_cond_wait_1 lk cv bp sp pc mP S').
    intros lk cv bp sp pc mP S S' HS HSS'.
    unfold Ast_cond_wait_pre, Ast_cond_wait_1 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspgt [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simplS in * .
    destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
    split.
      subst M' R'.
      sep on.
      rewrite HBP.
      rewrite HSP.
      sep lift 1.
      sep upd.
      sep lift 0 in HM.
      sep stksplit in HM.
      sep matchH HM with 5.
    subst M' R' fl' isr'.
    reg simpl.
    rewrite HSP.
    repeat split_l; trivial.
  simpl.
  split.
    exists p_cond_wait_1; exists g_cond_wait_1.
  split_l (* lookup *).
  split.
  (* enable *)
    apply enable_pushwp with (_STK_SIZE_cond_wait).
    intros S HS.
    destruct S as [[[M R] fl] isr].
    unfold p_cond_wait_0 in HS.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    unfold Ast_cond_wait_pre in HS.
    destruct HS as [HM [HBP [HSP _]]].
    simplS in * .
    rewrite HSP.
    sep on.
    sep lift 0 in HM.
    sep trivial.
    omega.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold p_cond_wait_1, p_cond_wait_2.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HSS'].
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    exists lk; exists cv; exists bp; exists sp; exists pc.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HxS HSS'] HS'S''].
  unfold g_cond_wait_0, g_cond_wait_1 in * .
  intros mP lk cv bp sp pc HS.
  apply HS'S''.
  apply Hpp' with S; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold p_cond_wait_1 in HS.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    unfold Ast_cond_wait_1 in HS.
    destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    rewrite HSP.
    sep on.
    sep lift 0 in HM.
    sep stksplit (IH_Stk + 6) in HM.
    sep lift 0 in HM.
    destruct HM as [M1 [M2 [HJM [HM1 HM2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    split_l.
    intros M1' HM1'.
    unfold p_cond_wait_0.
    exists lk; exists cv; exists bp; exists sp; exists pc.
    unfold Ast_cond_wait_pre.
    simplS in * .
    rsplit.
      repeat split_l; trivial.
    sep on.
    sep lift 0.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      sep off.
      split.
        disjoint rep_l M1.
        eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    split_l.
    sep on.
    sep matchH HM2 with 4.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold g_cond_wait_0 in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP lk cv bp sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  apply HSS'.
  clear HSS'.
  unfold Ast_cond_wait_pre in * .
  simplS in * .
  destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
  rsplit.
    repeat split_l; trivial.
  sep on.
  sep lift 0 in HM.
  sep stksplit (IH_Stk + 6) in HM.
  sep lift 0 in HM.
  destruct HM as [M2x [M1x [HJM12 [HM2x HM1x]]]].
  sep lift 0.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  assert (M2x = M2' /\ M1x = M1).
    sep off.
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    rewrite HSP in HM2'.
    assert (Hx := prec_disjUnion_inv H H0 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  rewrite HSP in HM2.
  split_l.
  sep on.
  sep matchH HM1x with 3.
Qed.

Lemma wff_cmd_cond_wait2:
   NH_WFcmd
     (SCAP_consCT (cond_wait + 1 + 1) (Ast p_cond_wait_2 g_cond_wait_2 1)
        psi_cond_wait) (cond_wait + 1) (Ast p_cond_wait_1 g_cond_wait_1 1)
     (comml (_movwp (Opr_reg SP) BP)).
Proof.
  assert (Hpp' : forall lk cv bp sp pc mP S S',
    Ast_cond_wait_1 lk cv bp sp pc mP S
    -> nextS (cond_wait + 1) (comml (_movwp  (Opr_reg SP) BP)) S S'
    -> Ast_cond_wait_2 lk cv bp sp pc mP S').
    intros lk cv bp sp pc mP S S' HS HSS'.
    unfold Ast_cond_wait_1, Ast_cond_wait_2 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    simplS in * .
    destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
    subst M' R' fl' isr'.
    reg simpl.
    repeat split_l; trivial.
  simpl.
  split.
    exists p_cond_wait_2; exists g_cond_wait_2.
  split_l.
  split.
    (* enable movw *)
    apply trivial_enable_movwp.
  split. (* p ==> p' *)
    unfold p_cond_wait_1, p_cond_wait_2.
    intros S' H.
    unfold linkPG in H.
    destruct H as [S H].
    destruct H as [H HSS'].
    destruct H as [lk [cv [bp [sp [pc HS]]]]].
    exists lk; exists cv; exists bp; exists sp; exists pc.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold conjPG, conjG in H.
  destruct H as [S' [[_ HSS'] HS'S'']].
  unfold g_cond_wait_1, g_cond_wait_2 in * .
  intros mP lk cv bp sp pc HS.
  eapply HS'S''; eauto.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold p_cond_wait_1 in HS.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    unfold Ast_cond_wait_1 in HS.
    destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    rewrite HSP.
    sep on.
    sep lift 0 in HM.
    sep stksplit (IH_Stk + 6) in HM.
    sep lift 0 in HM.
    destruct HM as [M1 [M2 [HJM [HM1 HM2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    split_l.
    intros M1' HM1'.
    unfold p_cond_wait_1.
    exists lk; exists cv; exists bp; exists sp; exists pc.
    unfold Ast_cond_wait_1.
    simplS in * .
    rsplit.
      repeat split_l; trivial.
    sep on.
    sep lift 0.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      sep off.
      split.
        disjoint rep_l M1.
        eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    split_l.
    sep on.
    sep matchH HM2 with 4.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold g_cond_wait_1 in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP lk cv bp sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  apply HSS'.
  clear HSS'.
  unfold Ast_cond_wait_1 in * .
  simplS in * .
  destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
  rsplit.
    repeat split_l; trivial.
  sep on.
  sep lift 0 in HM.
  sep stksplit (IH_Stk + 6) in HM.
  sep lift 0 in HM.
  destruct HM as [M2x [M1x [HJM12 [HM2x HM1x]]]].
  sep lift 0.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  assert (M2x = M2' /\ M1x = M1).
    sep off.
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    rewrite HSP in HM2'.
    assert (Hx := prec_disjUnion_inv H H0 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  rewrite HSP in HM2.
  split_l.
  sep on.
  sep matchH HM1x with 3.
Qed.

Lemma wff_cmd_cond_wait3:
   NH_WFcmd
     (SCAP_consCT (cond_wait + 1 + 1 + 1)
        (Ast p_cond_wait_3 g_cond_wait_3 1) psi_cond_wait)
     (cond_wait + 1 + 1) (Ast p_cond_wait_2 g_cond_wait_2 1)
     (comml (_pushwm (A_reg 6 BP))).
Proof.
  assert (Hpp' : forall lk cv bp sp pc mP S S',
    Ast_cond_wait_2 lk cv bp sp pc mP S
    -> nextS (cond_wait+1+1) (comml (_pushwm (A_reg 6 BP))) S S'
    -> Ast_cond_wait_3 lk cv bp sp pc mP S').
    intros lk cv bp sp pc mP S S' HS HSS'.
    unfold Ast_cond_wait_2, Ast_cond_wait_3 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [wx [Hlds [Hspgt [HldsB HM']]]].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM', Hlds.
    simplS in * .
    destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
    assert (Heq : lk = wx).
      rewrite HBP in Hlds.
      arith_rep (sp - 2 + 6) with (sp + 4) in Hlds.
      sep on.
      sep lift 4 in HM.
      sep inj.
      auto.
    subst wx.
    subst M' R' fl' isr'.
    rewrite HSP.
    arith_rep (sp - 2 - 2) with (sp - 4).
    split.
      sep on.
      sep lift 1.
      sep upd.
      sep lift 0 in HM.
      sep stksplit in HM.
      arith_rep (sp - 2 - 2) with (sp - 4) in HM.
      arith_rep (_STK_SIZE_cond_wait - 2 - 2) with (_STK_SIZE_cond_wait - 4) in HM.
      sep matchH HM with 6.
    reg simpl.
    repeat split_l; trivial.
  simpl.
  split.
  exists p_cond_wait_3; exists g_cond_wait_3.
  split_l (* lookup *).
  split.
  (* enable *)
    apply enable_pushwm with (_STK_SIZE_cond_wait - 2); try omega.
    intros S HS.
    destruct S as [[[M R] fl] isr].
    unfold p_cond_wait_2 in HS.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    unfold Ast_cond_wait_2 in HS.
    simplS in * .
    destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
    rewrite HSP.
    unfold addr_eval.
    rewrite HBP.
    split.
      sep on.
      sep lift 0 in HM.
      sep trivial.
    sep on.
    arith_rep (sp - 2 + 6) with (sp + 4).
    sep lift 4 in HM.
    sep trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold p_cond_wait_2, p_cond_wait_3.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HSS'].
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    exists lk; exists cv; exists bp; exists sp; exists pc.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HxS HSS'] HS'S''].
  unfold g_cond_wait_2, g_cond_wait_3 in * .
  intros mP lk cv bp sp pc HS.
  apply HS'S''.
  apply Hpp' with S; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold p_cond_wait_2 in HS.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    unfold Ast_cond_wait_2 in HS.
    destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    rewrite HSP.
    sep on.
    sep lift 0 in HM.
    sep stksplit (IH_Stk + 6) in HM.
    sep lift 0 in HM.
    destruct HM as [M1 [M2 [HJM [HM1 HM2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    split_l.
    intros M1' HM1'.
    unfold p_cond_wait_2.
    exists lk; exists cv; exists bp; exists sp; exists pc.
    unfold Ast_cond_wait_2.
    simplS in * .
    rsplit.
      repeat split_l; trivial.
    sep on.
    sep lift 0.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      sep off.
      split.
        disjoint rep_l M1.
        eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    split_l.
    sep on.
    sep matchH HM2 with 4.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold g_cond_wait_2 in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP lk cv bp sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  apply HSS'.
  clear HSS'.
  unfold Ast_cond_wait_2 in * .
  simplS in * .
  destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
  rsplit.
    repeat split_l; trivial.
  sep on.
  sep lift 0 in HM.
  sep stksplit (IH_Stk + 6) in HM.
  sep lift 0 in HM.
  destruct HM as [M2x [M1x [HJM12 [HM2x HM1x]]]].
  sep lift 0.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  assert (M2x = M2' /\ M1x = M1).
    sep off.
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    rewrite HSP in HM2'.
    assert (Hx := prec_disjUnion_inv H H0 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  rewrite HSP in HM2.
  split_l.
  sep on.
  sep matchH HM1x with 3.
Qed.

Lemma wff_cmd_cond_wait4:
   NH_WFcmd
     (SCAP_consCT (cond_wait + 1 + 1 + 1 + 1)
        (Ast p_cond_wait_4 g_cond_wait_4 1) psi_cond_wait)
     (cond_wait + 1 + 1 + 1) (Ast p_cond_wait_3 g_cond_wait_3 1)
     (comml _cli).
Proof.
  assert (Hpp': forall lk cv bp sp pc mP S S', 
    Ast_cond_wait_3 lk cv bp sp pc mP S
    -> gcli S S'
    -> Ast_cond_wait_4 lk cv bp sp pc mP S').
    intros lk cv bp sp pc mP S S' HS HSS'.
    unfold linkPG, conjPG, conjG.
    unfold Ast_cond_wait_3, gcli, Ast_cond_wait_4 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct HS as [HM [HBX [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
    destruct HSS' as [Hshift [Hif' [Hisr' HR']]].
    assert (Hshift' : prepost Aemp (INV0 *^* INV1) (M, R, fl, isr) (M', R', fl', isr')).
      apply (prepost_weaken_post (mQ' := INV0 *^* INV1) Hshift).
      intros Mx H.
      apply (choice_elim_left H); unfold itr_enabled; auto.
    split.
      sep on.
      assert (HMx : S_Mem (M, R, fl, isr) ||- 
        (Aemp) *^* (stkSpacePrec (sp - 4) (40 + 6 + IH_Stk - 4) *^* sp - 4 |->w lk *^*
           sp - 2 |->w bp *^* sp |->w pc *^* sp + 2 |->w cv *^*
           sp + 4 |->w lk *^* Gamma lk *^* mP)).
        sep matchH HM with 4.
      substH HM with (prepost_elim Hshift' precise_emp HMx).
      sep on.
      sep matchH HM with 5.
    simplS in * .
    subst R' isr'.
    repeat split_l; trivial.
  simpl.
  split.
  exists p_cond_wait_4; exists g_cond_wait_4.
  split_l. (* lookup *)
  split.
  (* p ==> p' *)
  unfold p_cond_wait_3, p_cond_wait_4.
  intros S' H.
  unfold linkPG in H.
  destruct H as [S [HS HSS']].
  destruct HS as [lk [cv [bp [sp [pc HS]]]]].
  exists lk; exists cv; exists bp; exists sp; exists pc.
  apply Hpp' with S; trivial.
  (* g' ==> g *)
  unfold g_cond_wait_3, g_cond_wait_4.
  unfold linkPG, conjPG, conjG.
  intros S S'' H.
  destruct H as [S' H].
  destruct H as [[HxS HSS'] HS'S''].
  intros mP lk cv bp sp pc HS.
  apply HS'S''.
  apply Hpp' with S; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold p_cond_wait_3 in HS.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    unfold Ast_cond_wait_3 in HS.
    destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    rewrite HSP.
    sep on.
    sep lift 0 in HM.
    sep stksplit (IH_Stk + 6) in HM.
    sep lift 0 in HM.
    destruct HM as [M1 [M2 [HJM [HM1 HM2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    split_l.
    intros M1' HM1'.
    unfold p_cond_wait_3.
    exists lk; exists cv; exists bp; exists sp; exists pc.
    unfold Ast_cond_wait_3.
    simplS in * .
    rsplit.
      repeat split_l; trivial.
    sep on.
    sep lift 0.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      sep off.
      split.
        disjoint rep_l M1.
        eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    split_l.
    sep on.
    sep matchH HM2 with 4.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold g_cond_wait_3 in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP lk cv bp sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  apply HSS'.
  clear HSS'.
  unfold Ast_cond_wait_3 in * .
  simplS in * .
  destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
  rsplit.
    repeat split_l; trivial.
  sep on.
  sep lift 0 in HM.
  sep stksplit (IH_Stk + 6) in HM.
  sep lift 0 in HM.
  destruct HM as [M2x [M1x [HJM12 [HM2x HM1x]]]].
  sep lift 0.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  assert (M2x = M2' /\ M1x = M1).
    sep off.
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    rewrite HSP in HM2'.
    assert (Hx := prec_disjUnion_inv H H0 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  rewrite HSP in HM2.
  split_l.
  sep on.
  sep matchH HM1x with 3.
Qed.


Lemma wff_cmd_cond_wait5:
   NH_WFcmd
     (SCAP_consCT (cond_wait + 1 + 1 + 1 + 1 + 1)
        (Ast p_cond_wait_5 g_cond_wait_5 1) psi_cond_wait)
     (cond_wait + 1 + 1 + 1 + 1) (Ast p_cond_wait_4 g_cond_wait_4 1)
     (comml (_call l_release)).
Proof.
  assert(Hpp' : forall lk cv bp sp pc mQ mP S S',
    Ast_cond_wait_4 lk cv bp sp pc mP S
    -> nextS (cond_wait + 1 + 1 + 1 + 1) (comml (_call l_release)) S S'
    -> mQ = (stkSpacePrec (sp - 6 - stk_l_release) (_STK_SIZE_cond_wait - 6 - stk_l_release)
            *^* (sp - 2 |->w bp)
            *^* (sp |->w pc)
            *^* (sp + 2 |->w cv)
            *^* (sp + 4 |->w lk))
    -> Ast_l_release_pre lk (sp - 2) (sp - 6) (cond_wait + 1 + 1 + 1 + 1 + 1)
      (mQ *^* mP) S').
    intros lk cv bp sp pc mQ mP S S'.
    intros HS HSS' HmQ.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    unfold Ast_cond_wait_4 in HS.
    destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspx [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    subst M' R' fl' isr'.
    unfold Ast_l_release_pre.
    simplS in * .
    reg simpl.
    split.
      sep on.
      rewrite HSP.
      assert (H1 : sp - 4 - 2 = sp - 6).
        generalize Hsp; clearall.
        omega.
      rewrite H1.
      subst mQ.
      sep lift 3.
      sep_update.
      sep lift 2 in HM.
      sep stksplit in HM.
      rewrite H1 in HM.
      assert (H2 : _STK_SIZE_cond_wait - 4 - 2 = _STK_SIZE_cond_wait - 6).
        clearall; omega.
      rewrite H2 in HM.
      assert (H3 :  _STK_SIZE_cond_wait - 6 >= stk_l_release).
        unfold stk_l_release.
        clearall; omega.
      assert (H4 :  sp - 6 >= stk_l_release).
        unfold stk_l_release.
        generalize Hsp; clearall; omega.
      sep stksplit stk_l_release in HM.
      assert (H5 :  sp - 6 + 2 = sp - 4).
        generalize Hsp; clearall; omega.
      rewrite H5.
      sep matchH HM with 11.
    split_l.
    split.
      rewrite HSP; omega.
    auto.
  assert (Hpp''' : forall lk cv bp sp pc pc' mP S S' S'' S''',
    Ast_cond_wait_4 lk cv bp sp pc mP S
    -> nextS (cond_wait + 1 + 1 + 1 + 1) (comml (_call l_release)) S S'
    -> l_release_grt S' S''
    -> nextS pc' (comms (_retw)) S'' S'''
    -> Ast_cond_wait_5 lk cv bp sp pc mP S''').
    intros lk cv bp sp pc pc' mP.
    intros S S' S'' S''' HS HSS' HS'S'' HS''S'''.
    substH HS'S'' with (g_l_release_sound _ _ HS'S'').
    unfold g_l_release in HS'S''.
    set (mQ := (stkSpacePrec (sp - 6 - stk_l_release) (_STK_SIZE_cond_wait - 6 - stk_l_release)
            *^* (sp - 2 |->w bp)
            *^* (sp |->w pc)
            *^* (sp + 2 |->w cv)
            *^* (sp + 4 |->w lk))).
    assert (HS' : Ast_l_release_pre lk (sp - 2) (sp - 6) (cond_wait + 1 + 1 + 1 + 1 + 1) (mQ *^* mP) S').
      apply Hpp' with cv bp pc S; trivial.
    assert (HS'' := HS'S'' _ _ _ _ _ HS').
    subst mQ.
    clear HS'S''.
    unfold Ast_l_release_post in HS''.
    unfold Ast_cond_wait_5.
    unfold Ast_cond_wait_4 in HS.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct S'' as [[[M'' R''] fl''] isr''].
    destruct S''' as [[[M''' R'''] fl'''] isr'''].
    destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
    simplS in * .
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct (nextS_inversion HS''S''') as [HM''' [HR''' [Hfl''' Hisr''']]].
    destruct HR''' as [pcx [Hlds HR''']].
    destruct HS'' as [HM'' [HBP'' [HSP'' [Hlkx [Hif'' Hisr'']]]]].
    assert (Hpcx : pcx = cond_wait + 1 + 1 + 1 + 1 + 1).
      subst R' M''' R'''.
      subst fl' isr' fl''' isr'''.
      rewrite HSP'' in Hlds; clear HSP''.
      sep on.
      sep lift 3 in HM''.
      sep_inj.
      auto.
    subst pcx; clear Hlds.
    subst M''' R''' fl''' isr'''.
    reg simpl.
    rewrite HBP''.
    rewrite HSP''.
    assert (H0 : sp - 6 + 2 = sp - 4).
      generalize Hsp; clearall.
      omega.
    rewrite H0.
    rewrite H0 in HM''.
    split.
      sep on.
      sep matchH HM'' with 9.
      assert (H3 :  _STK_SIZE_cond_wait - 4 >= 2).
        clearall; omega.
      assert (H4 :  sp - 4 >= 2).
        generalize Hsp; clearall; omega.
      sep stksplit.
      assert (H1 : sp - 4 - 2 = sp - 6).
        generalize Hsp; clearall; omega.
      rewrite H1.
      assert (H2 : _STK_SIZE_cond_wait - 4 - 2 = _STK_SIZE_cond_wait - 6).
        clearall; omega.
      rewrite H2.
      clear H3 H4.
      assert (H3 :  _STK_SIZE_cond_wait - 6 >= stk_l_release).
        unfold stk_l_release.
        clearall; omega.
      assert (H4 :  sp - 6 >= stk_l_release).
        unfold stk_l_release.
        generalize Hsp; clearall; omega.
      sep stksplit stk_l_release.
      sep matchH HM'' with 3.
    repeat split_l; trivial.
  simpl.
  split.
  right.
  split.
    apply l_release_not_primentry.
  exists l_release_Prd; exists p_cond_wait_5.
  exists l_release_grt; exists g_cond_wait_5.
  split.
    unfold psi_cond_wait; auto.
  split_l.
  split.
    clear Hpp' Hpp'''.
    unfold enable, p_cond_wait_4.
    intros S HS.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    unfold Ast_cond_wait_4 in HS.
    destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
    (* enable call *)
    destruct S as [[[M R] fl] isr].
    simplS in * .
    exists (MemUpdW M (sp - 4 - 2) (cond_wait + 1 + 1 + 1 + 1 + 1) 0, RF_update R SP (sp - 4 - 2), fl, isr).
    apply nextS_call.
    unfold pushW.
    split.
      rewrite HSP.
      sep on.
      sep lift 2 in HM.
      sep stksplit 2 in HM.
      sep lift 0 in HM.
      sep trivial.
    split.
      rewrite HSP.
      apply MemUpdW_MemUpd16 with 0; trivial.
    rewrite HSP; trivial.
  split.
    (* g' => fun_frame *)
    unfold sp_call_pred, fun_frame.
    unfold linkPG, conjPG.
    intros S' S'' HS'S''.
    destruct HS'S'' as [H HS'S''].
    destruct H as [S [HS HSS']].
    substH HS'S'' with (g_l_release_sound _ _ HS'S'').
    unfold g_l_release in HS'S''.
    unfold p_cond_wait_4 in HS.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    set (mQ := (stkSpacePrec (sp - 6 - stk_l_release) (_STK_SIZE_cond_wait - 6 - stk_l_release)
            *^* (sp - 2 |->w bp)
            *^* (sp |->w pc)
            *^* (sp + 2 |->w cv)
            *^* (sp + 4 |->w lk))).
    assert (HS' : Ast_l_release_pre lk (sp - 2) (sp - 6) (cond_wait + 1 + 1 + 1 + 1 + 1) (mQ *^* mTrue) S').
      apply Hpp' with cv bp pc S; trivial.
    clear Hpp' HS HSS' Hpp'''.
    substH HS'S'' with (HS'S'' _ _ _ _ _ HS').
    rename HS'S'' into HS''.
    destruct S' as [[[M' R'] fl'] isr'].
    destruct S'' as [[[M'' R''] fl''] isr''].
    simplS in * .
    unfold Ast_l_release_pre, Ast_l_release_post in * .
    destruct HS' as [HM' [HBP' [HSP' [Hlk' [Hif' Hisr']]]]].
    destruct HS'' as [HM'' [HBP'' [HSP'' [Hlk'' [Hif'' Hisr'']]]]].
    simplS in * .
    split.
      rewrite HSP'; rewrite HSP''; trivial.
    unfold Ret_Addr.
    exists (cond_wait + 1 + 1 + 1 + 1 + 1).
    simplS.
    split.
      intros pc''.
      apply nextPC_retw with (sp - 6); trivial.
      clear HM''.
      sep on.
      subst mQ.
      sep lift 3 in HM'.
      sep trivial.
    intros pc''.
    apply nextPC_retw with (sp - 6); trivial.
    clear HM'.
    sep on.
    subst mQ.
    sep lift 3 in HM''.
    sep trivial.
  split.
    (* p => p' *)
    intros S' H.
    unfold conjPG, conjG, linkPG in H.
    destruct H as [S [HS HnextSS']].
    apply p_l_release_sound.
    unfold p_cond_wait_4 in HS.
    unfold p_l_release.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    exists lk; exists (sp - 2); exists (sp - 6); exists (cond_wait + 1 + 1 + 1 + 1 + 1).
    set (mQ := (stkSpacePrec (sp - 6 - stk_l_release) (_STK_SIZE_cond_wait - 6 - stk_l_release)
            *^* (sp - 2 |->w bp)
            *^* (sp |->w pc)
            *^* (sp + 2 |->w cv)
            *^* (sp + 4 |->w lk))).
    assert (HS' : Ast_l_release_pre lk (sp - 2) (sp - 6) (cond_wait + 1 + 1 + 1 + 1 + 1) (mQ *^* mTrue) S').
      apply Hpp' with cv bp pc S; trivial.
    clear Hpp''' HS HnextSS'.
    unfold Ast_l_release_pre in * .
    destruct HS' as [HM' [HBP' [HSP' [Hlk' [Hif' Hisr']]]]].
    split.
      sep on.
      sep matchH HM' with 10.
    auto.
  split.
    (* p => p''' *)
    intros pc' S'''.
    unfold linkPG, wrap_fun, conjG.
    intros H.
    destruct H as [S H].
    destruct H as [HS H].
    destruct H as [S'' [H HS''S''']].
    destruct H as [S' [HSS' HS'S'']].
    unfold p_cond_wait_4, p_cond_wait_5 in * .
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    exists lk; exists cv; exists bp; exists sp; exists pc.
    eapply Hpp'''; eauto.
  (* g''' => g *)
  intros pc' S S'''' H.
  unfold conjPG, wrap_fun, conjG, linkPG in H.
  destruct H as [S''' [HS HS'''S'''']].
  destruct HS as [HSx [S'' [Hx HnextS'']]].
  destruct Hx as [S' [HSS' HS'S'']].
  unfold g_cond_wait_4, g_cond_wait_5 in * .
  intros mP lk cv bp sp pc HS.
  eapply HS'''S''''; eauto.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold p_cond_wait_4 in HS.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    unfold Ast_cond_wait_4 in HS.
    destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold g_cond_wait_4 in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP lk cv bp sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_cond_wait_4 in HS.
  simplS in * .
  destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma wff_cmd_cond_wait6:
   NH_WFcmd
     (SCAP_consCT (cond_wait + 1 + 1 + 1 + 1 + 1 + 1)
        (Ast p_cond_wait_6 g_cond_wait_6 1) psi_cond_wait)
     (cond_wait + 1 + 1 + 1 + 1 + 1) (Ast p_cond_wait_5 g_cond_wait_5 1)
     (comml (_movwld (A_reg 4 BP) BX)).
Proof.
  assert (Hpp' : forall lk cv bp sp pc mP S S',
    Ast_cond_wait_5 lk cv bp sp pc mP S
    -> nextS (cond_wait + 1 + 1 + 1 + 1 + 1) (comml (_movwld (A_reg 4 BP) BX)) S S'
    -> Ast_cond_wait_6 lk cv bp sp pc mP S').
    intros lk cv bp sp pc mP S S' HS HSS'.
    unfold Ast_cond_wait_5, Ast_cond_wait_6 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HR' as [cvx [Hlds HR']].
    simplS in * .
    destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
    assert (Heq : cvx = cv).
      unfold addr_eval in Hlds.
      rewrite HBP in Hlds.
      arith_rep (sp - 2 + 4) with (sp + 2) in Hlds.
      sep on.
      sep lift 6 in HM.
      sep inj.
      auto.
    subst cvx M' R' fl' isr'.
    reg simpl.
    repeat split_l; trivial.
  simpl.
  split.
    exists p_cond_wait_6; exists g_cond_wait_6.
  split_l.
  split.
  (* enable *)
    apply enable_movwld.
    intros S HS.
    destruct S as [[[M R] fl] isr].
    unfold p_cond_wait_5 in HS.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    unfold Ast_cond_wait_5 in HS.
    destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
    simplS in * .
    unfold addr_eval.
    rewrite HBP.
    arith_rep (sp - 2 + 4) with (sp + 2).
    sep on.
    sep lift 6 in HM.
    sep trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold p_cond_wait_5, p_cond_wait_6.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HSS'].
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    exists lk; exists cv; exists bp; exists sp; exists pc.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HxS HSS'] HS'S''].
  unfold g_cond_wait_5, g_cond_wait_6 in * .
  intros mP lk cv bp sp pc HS.
  eapply HS'S''; eauto.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold p_cond_wait_5 in HS.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    unfold Ast_cond_wait_5 in HS.
    destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold g_cond_wait_5 in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP lk cv bp sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_cond_wait_5 in HS.
  simplS in * .
  destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma wff_cmd_cond_wait7:
   NH_WFcmd
     (SCAP_consCT (cond_wait + 1 + 1 + 1 + 1 + 1 + 1 + 1)
        (Ast p_cond_wait_7 g_cond_wait_7 1) psi_cond_wait)
     (cond_wait + 1 + 1 + 1 + 1 + 1 + 1)
     (Ast p_cond_wait_6 g_cond_wait_6 1)
     (comml (_pushwm (A_reg C_WAITING BX))).
Proof.
  assert (Hpp' : forall lk cv bp sp pc mP S S',
    Ast_cond_wait_6 lk cv bp sp pc mP S
    -> nextS (cond_wait+1+1+1+1+1+1) (comml (_pushwm (A_reg C_WAITING BX))) S S'
    -> Ast_cond_wait_7 lk cv bp sp pc mP S').
    intros lk cv bp sp pc mP S S' HS HSS'.
    unfold Ast_cond_wait_6, Ast_cond_wait_7 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [wx [Hlds [Hspgt [HldsB HM']]]].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM', Hlds.
    simplS in * .
    destruct HS as [HM [HBX [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
    assert (Hw : exists w,
      (cv + C_WAITING |~>w w) M /\ Delta w = SomeT Aemp).
      assert ((CV_INV_M cv *^* mTrue) M).
      apply INV1_cvInv; trivial.
      sep on.
      sep matchH HM with 5.
      destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
      unfold CV_INV_M in HM1.
      destruct HM1 as [w [HM1 Hw]].
      exists w; split_r.
      unfold leadsToW.
      unfold imprecM.
      exists M1; exists M2; auto.
    destruct Hw as [w [Hw1 Hw2]].
    assert (Heq : w = wx).
      rewrite HBX in Hlds.
      sep on. 
      sep inj.
      auto.
    subst wx.
    subst M' R' fl' isr'.
    rewrite HSP.
    split.
      sep on.
      sep lift 3.
      sep exists_left w.
      arith_rep (sp - 6) with (sp - 4 - 2).
      sep upd.
      sep lift 2 in HM.
      sep stksplit in HM.
      arith_rep (_STK_SIZE_cond_wait - 6) with (_STK_SIZE_cond_wait - 4 - 2).
      sep matchH HM with 6.
    reg simpl.
    repeat split_l; trivial.
    exists w; split_r.
    sep on.
    arith_rep (sp - 6) with (sp - 4 - 2).
    sep upd.
    sep lift 2 in HM.
    sep stksplit in HM.
    sep lift 1 in HM.
    sep trivial.
    omega.
  simpl.
  split.
  exists p_cond_wait_7; exists g_cond_wait_7.
  split_l (* lookup *).
  split.
  (* enable *)
    apply enable_pushwm with (_STK_SIZE_cond_wait - 4); try omega.
    intros S HS.
    destruct S as [[[M R] fl] isr].
    unfold p_cond_wait_6 in HS.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    unfold Ast_cond_wait_6 in HS.
    simplS in * .
    destruct HS as [HM [HBX [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
    rewrite HSP.
    unfold addr_eval.
    rewrite HBX.
    assert (Hw : exists w,
      (cv + C_WAITING |~>w w) M /\ Delta w = SomeT Aemp).
      assert ((CV_INV_M cv *^* mTrue) M).
        apply INV1_cvInv; trivial.
      sep on.
      sep matchH HM with 5.
      destruct H as [M1 [M2 [HJM [HM1 HM2]]]].
      unfold CV_INV_M in HM1.
      destruct HM1 as [w [HM1 Hw]].
      exists w; split_r.
      unfold leadsToW.
      unfold imprecM.
      exists M1; exists M2; auto.
    destruct Hw as [w [Hw1 Hw2]].
    split.
      sep on.
      sep lift 2 in HM.
      sep trivial.
      sep on.
    sep trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold p_cond_wait_6, p_cond_wait_7.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HSS'].
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    exists lk; exists cv; exists bp; exists sp; exists pc.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HxS HSS'] HS'S''].
  unfold g_cond_wait_6, g_cond_wait_7 in * .
  intros mP lk cv bp sp pc HS.
  apply HS'S''.
  apply Hpp' with S; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold p_cond_wait_6 in HS.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    unfold Ast_cond_wait_6 in HS.
    destruct HS as [HM [HBX [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold g_cond_wait_6 in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP lk cv bp sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_cond_wait_6 in HS.
  simplS in * .
  destruct HS as [HM [HBX [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma wff_cmd_cond_wait8:
   NH_WFcmd
     (SCAP_consCT (cond_wait + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
        (Ast p_cond_wait_8 g_cond_wait_8 1) psi_cond_wait)
     (cond_wait + 1 + 1 + 1 + 1 + 1 + 1 + 1)
     (Ast p_cond_wait_7 g_cond_wait_7 1) (comml (_call L_sleep)).
Proof.
  assert(Hpp' : forall lk cv bp sp pc mQ mP S S',
    Ast_cond_wait_7 lk cv bp sp pc mP S
    -> nextS (cond_wait + 1 + 1 + 1 + 1 + 1 + 1 + 1) (comml (_call L_sleep)) S S'
    -> mQ = (stkSpacePrec (sp - 8 - _STK_SIZE_sleep) (_STK_SIZE_cond_wait - 8 - _STK_SIZE_sleep)
            *^* (sp - 4 |->w lk)
            *^* (sp - 2 |->w bp)
            *^* (sp |->w pc)
            *^* (sp + 2 |->w cv)
            *^* (sp + 4 |->w lk))
    -> exists w, Ast_sleep_0 w Aemp (sp - 8) (cond_wait +1 +1 +1 +1 +1 +1 +1 +1)
      (mQ *^* mP) S'
      /\ (sp - 6 |~>w w) (S_Mem S)).
    intros lk cv bp sp pc mQ mP S S'.
    intros HS HSS' HmQ.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    unfold Ast_cond_wait_7 in HS.
    destruct HS as [HM [Hw [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspx [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    subst M' R' fl' isr'.
    destruct Hw as [w [Hldsw Hw]].
    exists w; split_r.
    unfold Ast_sleep_0.
    simplS in * .
    reg simpl.
    split.
      sep on.
      rewrite HSP.
      assert (H1 : sp - 6 - 2 = sp - 8).
        generalize Hsp; clearall.
        omega.
      rewrite H1.
      subst mQ.
      sep lift 3.
      sep_update.
      sep lift 2 in HM.
      sep stksplit in HM.
      rewrite H1 in HM.
      assert (H2 : _STK_SIZE_cond_wait - 6 - 2 = _STK_SIZE_cond_wait - 8).
        clearall.
        omega.
      rewrite H2 in HM.
      assert (H3 : 40 + 6 + IH_Stk - 8 >= _STK_SIZE_sleep).
        clearall.
        simpl.
        omega.
      sep stksplit _STK_SIZE_sleep in HM.
      assert (H4 : sp - 8 + 2 = sp - 6).
        generalize Hsp; clearall; omega.
      rewrite H4.
      sep lift 3 in HM.
      sep elim_left HM.
      sep inj.
      intro; subst v.
      sep matchH HM with 11.
    rewrite HSP.
    split.
      generalize Hsp; clearall; omega.
    split.
      generalize Hsp; clearall; omega.
    auto.
  assert (Hpp''' : forall lk cv bp sp pc pc' mP S S' S'' S''',
    Ast_cond_wait_7 lk cv bp sp pc mP S
    -> nextS (cond_wait +1 +1 +1 +1 +1 +1 +1) (comml (_call L_sleep)) S S'
    -> g_sleep S' S''
    -> nextS pc' (comms (_retw)) S'' S'''
    -> Ast_cond_wait_8 lk cv bp sp pc mP S''').
    intros lk cv bp sp pc pc' mP.
    intros S S' S'' S''' HS HSS' HS'S'' HS''S'''.
    unfold g_sleep in HS'S''.
    set (mQ := (stkSpacePrec (sp - 8 - _STK_SIZE_sleep) (_STK_SIZE_cond_wait - 8 - _STK_SIZE_sleep)
            *^* (sp - 4 |->w lk)
            *^* (sp - 2 |->w bp)
            *^* (sp |->w pc)
            *^* (sp + 2 |->w cv)
            *^* (sp + 4 |->w lk))).
    assert (HS' : exists w, 
      Ast_sleep_0 w Aemp (sp - 8) (cond_wait +1 +1 +1 +1 +1 +1 +1 +1) (mQ *^* mP) S'
      /\ (sp - 6 |~>w w) (S_Mem S)).
      eapply Hpp'; eauto.
    destruct HS' as [w [HS' Hldsw]].
    assert (HS'' := HS'S'' _ _ _ _ _ HS').
    clear HS'S''.
    destruct HS'' as [HS'' HS'S''].
    subst mQ.
    unfold Ast_cond_wait_7, Ast_cond_wait_8 in * .
    destruct HS as [HM [Hw [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
    destruct Hw as [wx [Hldswx Hw]].
    assert (wx = w).
      sep on.
      sep inj.
      auto.
    subst wx; clear Hldswx.
    unfold Ast_sleep_post in HS''.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct S'' as [[[M'' R''] fl''] isr''].
    destruct S''' as [[[M''' R'''] fl'''] isr'''].
    simplS in * .
    clear Hpp'.
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct (nextS_inversion HS''S''') as [HM''' [HR''' [Hfl''' Hisr''']]].
    destruct HR''' as [pcx [Hlds HR''']].
    destruct HS'' as [HM'' [Hif'' Hisr'']].
    assert (Hpcx : pcx = cond_wait +1 +1 +1 +1 +1 +1 +1 +1).
      subst R' R'' M''' R'''.
      subst fl' isr' fl''' isr'''.
      sep on.
      sep lift 4 in HM''.
      reg simpl.
      rewrite HSP in Hlds.
      assert (H1 : sp - 6 - 2 = sp - 8).
        generalize Hsp; clearall; omega.
      rewrite H1 in Hlds.
      sep_inj.
      auto.
    subst pcx; clear Hlds.
    subst R' R'' M''' R''' fl''' isr'''.
    reg simpl.
    assert (H1 : sp - 8 + 2 = sp - 6).
      generalize Hsp; clearall; omega.
      rewrite H1 in HM''.
    split.
      sep on.
      sep matchH HM'' with 9.
      sep stksplit.
      assert (H2 : sp - 6 - 2 = sp - 8).
        generalize Hsp; clearall; omega.
      rewrite H2.
      assert (H3 : _STK_SIZE_cond_wait - 6 - 2 = _STK_SIZE_cond_wait - 8).
        clearall; omega.
      rewrite H3.
      assert (H4 : _STK_SIZE_cond_wait - 8 >= _STK_SIZE_sleep).
        clearall; omega.
      assert (H5 : sp - 8 >= _STK_SIZE_sleep).
       generalize Hsp; clearall; omega.
      sep stksplit (_STK_SIZE_sleep).
      sep matchH HM'' with 3.
    repeat split_l; trivial.
    rewrite HSP.
    generalize Hsp; clearall; omega.
  simpl.
  split.
  right.
  split.
    apply sleep_not_prim.
  exists p_sleep; exists p_cond_wait_8.
  exists g_sleep; exists g_cond_wait_8.
  split.
    unfold psi_cond_wait; auto.
  split_l.
  split.
    clear Hpp' Hpp'''.
    unfold enable, p_cond_wait_7.
    intros S HS.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    unfold Ast_cond_wait_7 in HS.
    destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
    (* enable call *)
    destruct S as [[[M R] fl] isr].
    simplS in * .
    exists (MemUpdW M (sp - 6 - 2) (cond_wait +1 +1 +1 +1 +1 +1 +1 +1) 0, RF_update R SP (sp - 6 - 2), fl, isr).
    apply nextS_call.
    unfold pushW.
    split.
      rewrite HSP.
      sep on.
      sep lift 2 in HM.
      sep stksplit 2 in HM.
      sep lift 0 in HM.
      sep trivial.
    split.
      rewrite HSP.
      apply MemUpdW_MemUpd16 with 0; trivial.
    rewrite HSP; trivial.
  split.
    (* g' => fun_frame *)
    unfold sp_call_pred, fun_frame.
    unfold linkPG, conjPG.
    intros S' S'' HS'S''.
    destruct HS'S'' as [H HS'S''].
    destruct H as [S [HS HSS']].
    unfold g_sleep in HS'S''.
    unfold p_cond_wait_7 in HS.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    set (mQ := (stkSpacePrec (sp - 8 - _STK_SIZE_sleep) (_STK_SIZE_cond_wait - 8 - _STK_SIZE_sleep)
            *^* (sp - 4 |->w lk)
            *^* (sp - 2 |->w bp)
            *^* (sp |->w pc)
            *^* (sp + 2 |->w cv)
            *^* (sp + 4 |->w lk))).
    assert (HS' : exists w, 
      Ast_sleep_0 w Aemp (sp - 8) (cond_wait +1 +1 +1 +1 +1 +1 +1 +1) (mQ *^* mTrue) S'
      /\ (sp - 6 |~>w w) (S_Mem S)).
      eapply Hpp'; eauto.
    destruct HS' as [w [HS' Hldsw]].
    clear Hpp' HS HSS' Hpp'''.
    substH HS'S'' with (HS'S'' _ _ _ _ _ HS').
    rename HS'S'' into HS''.
    destruct S' as [[[M' R'] fl'] isr'].
    destruct S'' as [[[M'' R''] fl''] isr''].
    simplS in * .
    unfold Ast_sleep_0, Ast_sleep_post in * .
    destruct HS' as [HM' [HSP' [Hsp' [Hif' [Hisr' Hw]]]]].
    destruct HS'' as [[HM'' [Hif'' Hisr'']] HR''].
    simplS in * .
    subst R'' isr' isr''.
    split_l.
    unfold Ret_Addr.
    exists (cond_wait +1 +1 +1 +1 +1 +1 +1 +1).
    simplS.
    split.
      intros pc''.
      apply nextPC_retw with (sp - 8); trivial.
      clear HM''.
      sep on.
      subst mQ.
      sep lift 3 in HM'.
      sep trivial.
    intros pc''.
    apply nextPC_retw with (sp - 8); trivial.
    clear HM'.
    sep on.
    subst mQ.
    sep lift 4 in HM''.
    sep trivial.
  split.
    (* p => p' *)
    intros S' H.
    unfold conjPG, conjG, linkPG in H.
    destruct H as [S [HS HnextSS']].
    unfold p_cond_wait_7 in HS.
    unfold p_sleep.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    set (mQ := (stkSpacePrec (sp - 8 - _STK_SIZE_sleep) (_STK_SIZE_cond_wait - 8 - _STK_SIZE_sleep)
            *^* (sp - 4 |->w lk)
            *^* (sp - 2 |->w bp)
            *^* (sp |->w pc)
            *^* (sp + 2 |->w cv)
            *^* (sp + 4 |->w lk))).
    assert (HS' : exists w, 
      Ast_sleep_0 w Aemp (sp - 8) (cond_wait +1 +1 +1 +1 +1 +1 +1 +1) (mQ *^* mTrue) S'
      /\ (sp - 6 |~>w w) (S_Mem S)).
      eapply Hpp'; eauto.
    destruct HS' as [w [HS' Hldsw]].
    exists w; exists  Aemp; exists (sp - 8); exists (cond_wait +1 +1 +1 +1 +1 +1 +1 +1).
    clear Hpp''' HS HnextSS'.
    unfold Ast_sleep_0 in * .
    destruct HS' as [HM' [HSP' [Hsp' [Hif' [Hisr' Hw]]]]].
    split.
      sep on.
      sep matchH HM' with 10.
    auto.
  split.
    (* p => p''' *)
    intros pc' S'''.
    unfold linkPG, wrap_fun, conjG.
    intros H.
    destruct H as [S H].
    destruct H as [HS H].
    destruct H as [S'' [H HS''S''']].
    destruct H as [S' [HSS' HS'S'']].
    unfold p_cond_wait_7, p_cond_wait_8 in * .
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    exists lk; exists cv; exists bp; exists sp; exists pc.
    eapply Hpp'''; eauto.
  (* g''' => g *)
  intros pc' S S'''' H.
  unfold conjPG, wrap_fun, conjG, linkPG in H.
  destruct H as [S''' [HS HS'''S'''']].
  destruct HS as [HSx [S'' [Hx HnextS'']]].
  destruct Hx as [S' [HSS' HS'S'']].
  unfold g_cond_wait_7, g_cond_wait_8 in * .
  intros mP lk cv bp sp pc HS.
  eapply HS'''S''''; eauto.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold p_cond_wait_7 in HS.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    unfold Ast_cond_wait_7 in HS.
    destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold g_cond_wait_7 in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP lk cv bp sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_cond_wait_7 in HS.
  simplS in * .
  destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma wff_cmd_cond_wait9:
   NH_WFcmd
     (SCAP_consCT (cond_wait +1 +1 +1 +1 +1 +1 +1 +1 +1)
        (Ast p_cond_wait_9 g_cond_wait_9 1) psi_cond_wait)
     (cond_wait +1 +1 +1 +1 +1 +1 +1 +1)
     (Ast p_cond_wait_8 g_cond_wait_8 1) (comml _sti).
Proof.
  assert (Hpp': forall lk cv bp sp pc mP S S', 
    Ast_cond_wait_8 lk cv bp sp pc mP S
    -> gsti S S' 
    -> Ast_cond_wait_9 lk cv bp sp pc mP S').
    intros lk cv bp sp pc mP S S' HS HSS'.
    unfold linkPG, conjPG, conjG.
    unfold Ast_cond_wait_8, gsti, Ast_cond_wait_9 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct HS as [HM [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]].
    destruct HSS' as [Hshift [Hif' [Hisr' HR']]].
    assert (Hshift' : prepost (INV0 *^* INV1) Aemp
      (M, R, fl, isr) (M', R', fl', isr')).
      apply (prepost_weaken_pre (mP':=INV0 *^* INV1) Hshift).
      intros Mx HMx.
      assert (Hitr : ~ itr_enabled (M, R, fl, isr)).
        unfold itr_enabled.
        intros [H1 H2].
        simplS in * .
        rewrite Hif in H1; discriminate H1.
      eapply choice_elim_right; eauto.
    clear Hshift; rename Hshift' into Hshift.
    split.
      sep on.
      substH HM with (prepost_elim Hshift Prec01 HM).
      sep on.
      simplS in * .
      sep matchH HM with 6.
      sep trivial.
    simplS in * .
    subst R' isr'.
    repeat split_l; trivial.
  simpl.
  split.
  exists p_cond_wait_9; exists g_cond_wait_9.
  split_l.
  split.
  (* enable *)
     unfold enable_sti.
     intros S HS.
     unfold p_cond_wait_8 in HS.
     unfold pTrue, pmstar, choice.
     destruct HS as [lk [cv [bp [sp [pc HS]]]]].
     unfold Ast_cond_wait_8 in HS.
     destruct HS as [HM [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]].
     destruct S as [[[M R] fl] isr].
     simplS in * .
     destruct HM as [M2 [M1 [HJM [HM2 HM1]]]].
     exists M1; exists M2.
     split.
       auto with PMap.
     split_l.
     right.
     split_r.
     intros [H1 H2].
     simplS in * .
     rewrite Hif in H1; discriminate H1.
  split.
  (* p ==> p' *)
  unfold p_cond_wait_8, p_cond_wait_9.
  intros S' H.
  unfold linkPG in H.
  destruct H as [S [HS HSS']].
  destruct HS as [lk [cv [bp [sp [pc HS]]]]].
  exists lk; exists cv; exists bp; exists sp; exists pc.
  apply Hpp' with S; trivial.
  (* g' ==> g *)
  unfold g_cond_wait_8, g_cond_wait_9.
  unfold linkPG, conjPG, conjG.
  intros S S'' H.
  destruct H as [S' H].
  destruct H as [[HxS HSS'] HS'S''].
  intros mP lk cv bp sp pc HS.
  apply HS'S''.
  apply Hpp' with S; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold p_cond_wait_8 in HS.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    unfold Ast_cond_wait_8 in HS.
    destruct HS as [HM [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold g_cond_wait_8 in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP lk cv bp sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_cond_wait_8 in HS.
  simplS in * .
  destruct HS as [HM [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma wff_cmd_cond_waita:
   NH_WFcmd
     (SCAP_consCT (cond_wait +1 +1 +1 +1 +1 +1 +1 +1 +1 +1)
        (Ast p_cond_wait_a g_cond_wait_a 1) psi_cond_wait)
     (cond_wait +1 +1 +1 +1 +1 +1 +1 +1 +1)
     (Ast p_cond_wait_9 g_cond_wait_9 1) (comml (_addw (Opr_imm 2) SP)).
Proof.
  assert (Hpp' : forall lk cv bp sp pc mP S S',
    Ast_cond_wait_9 lk cv bp sp pc mP S
    -> nextS (cond_wait +1 +1 +1 +1 +1 +1 +1 +1 +1) (comml (_addw (Opr_imm 2) SP)) S S'
    -> Ast_cond_wait_a lk cv bp sp pc mP S').
    intros lk cv bp sp pc mP S S' HS HSS'.
    unfold Ast_cond_wait_9, Ast_cond_wait_a in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    simplS in * .
    destruct HS as [HM [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]].
    subst M' R' isr'.
    split.
      sep on.
      sep matchH HM with 5.
      sep stksplit.
      assert (H1 : sp - 4 - 2 = sp - 6).
        generalize Hsp; clearall; omega.
      rewrite H1.
      assert (H2 : _STK_SIZE_cond_wait - 4 - 2 = _STK_SIZE_cond_wait - 6).
        clearall; omega.
      rewrite H2.
      sep matchH HM with 2.
    reg simpl.
    unfold opr_eval.
    rewrite HSP.
    split_l by (generalize Hsp; clearall; omega).
    repeat split_l; trivial.
    unfold trash_fl in Hfl'.
    destruct Hfl'; subst fl'; auto.
  split.
  simpl.
  exists p_cond_wait_a; exists g_cond_wait_a.
  split_l.
  split.
    (* enable addw *)
    apply trivial_enable_addw.
  split. (* p ==> p' *)
    unfold p_cond_wait_9, p_cond_wait_a.
    intros S' H.
    unfold linkPG in H.
    destruct H as [S H].
    simpl in H.
    destruct H as [H HnextS].
    destruct H as [lk [cv [bp [sp [pc H]]]]].
    exists lk; exists cv; exists bp; exists sp; exists pc.
    apply Hpp' with S; trivial.
  (* gimp *)
  intros S S'' H.
  unfold conjPG, conjG in H.
  destruct H as [S' [[HSx HnextS] HS'S'']].
  unfold g_cond_wait_9, g_cond_wait_a in * .
  intros mP lk cv bp sp pc HS.
  eapply HS'S''; eauto.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold p_cond_wait_9 in HS.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    unfold Ast_cond_wait_9 in HS.
    destruct HS as [HM [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    rewrite HSP.
    sep on.
    sep lift 0 in HM.
    sep stksplit (IH_Stk + 6) in HM.
    sep lift 0 in HM.
    destruct HM as [M1 [M2 [HJM [HM1 HM2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    split_l.
    intros M1' HM1'.
    unfold p_cond_wait_9.
    exists lk; exists cv; exists bp; exists sp; exists pc.
    unfold Ast_cond_wait_9.
    simplS in * .
    rsplit.
      repeat split_l; trivial.
    sep on.
    sep lift 0.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      sep off.
      split.
        disjoint rep_l M1.
        eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    split_l.
    sep on.
    sep matchH HM2 with 4.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold g_cond_wait_9 in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP lk cv bp sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  apply HSS'.
  clear HSS'.
  unfold Ast_cond_wait_9 in * .
  simplS in * .
  destruct HS as [HM [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]].
  rsplit.
    repeat split_l; trivial.
  sep on.
  sep lift 0 in HM.
  sep stksplit (IH_Stk + 6) in HM.
  sep lift 0 in HM.
  destruct HM as [M2x [M1x [HJM12 [HM2x HM1x]]]].
  sep lift 0.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  assert (M2x = M2' /\ M1x = M1).
    sep off.
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    rewrite HSP in HM2'.
    assert (Hx := prec_disjUnion_inv H H0 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  rewrite HSP in HM2.
  split_l.
  sep on.
  sep matchH HM1x with 3.
Qed.

Lemma wff_cmd_cond_waitb:
   NH_WFcmd
     (SCAP_consCT (cond_wait +1 +1 +1 +1 +1 +1 +1 +1 +1 +1 +1)
        (Ast p_cond_wait_b g_cond_wait_b 1) psi_cond_wait)
     (cond_wait +1 +1 +1 +1 +1 +1 +1 +1 +1 +1)
     (Ast p_cond_wait_a g_cond_wait_a 1) (comml _cli).
Proof.
  assert (Hpp': forall lk cv bp sp pc mP S S', 
    Ast_cond_wait_a lk cv bp sp pc mP S
    -> gcli S S'
    -> Ast_cond_wait_b lk cv bp sp pc mP S').
    intros lk cv bp sp pc mP S S' HS HSS'.
    unfold linkPG, conjPG, conjG.
    unfold Ast_cond_wait_a, gcli, Ast_cond_wait_b in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct HS as [HM [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]].
    destruct HSS' as [Hshift [Hif' [Hisr' HR']]].
    assert (Hshift' : prepost Aemp (INV0 *^* INV1) (M, R, fl, isr) (M', R', fl', isr')).
      apply (prepost_weaken_post (mQ' := INV0 *^* INV1) Hshift).
      intros Mx H.
      apply (choice_elim_left H); unfold itr_enabled; auto.
    split.
      sep on.
      assert (HMx : S_Mem (M, R, fl, isr) ||- 
        (Aemp) *^* (stkSpacePrec (sp - 4) (40 + 6 + IH_Stk - 4) *^* sp - 4 |->w lk *^*
           sp - 2 |->w bp *^* sp |->w pc *^* sp + 2 |->w cv *^*
           sp + 4 |->w lk *^* mP)).
        sep matchH HM with 4.
      substH HM with (prepost_elim Hshift' precise_emp HMx).
      sep on.
      sep matchH HM with 5.
    simplS in * .
    subst R' isr'.
    repeat split_l; trivial.
  simpl.
  split.
  exists p_cond_wait_b; exists g_cond_wait_b.
  split_l. (* lookup *)
  split.
  (* p ==> p' *)
  unfold p_cond_wait_a, p_cond_wait_b.
  intros S' H.
  unfold linkPG in H.
  destruct H as [S [HS HSS']].
  destruct HS as [lk [cv [bp [sp [pc HS]]]]].
  exists lk; exists cv; exists bp; exists sp; exists pc.
  apply Hpp' with S; trivial.
  (* g' ==> g *)
  unfold g_cond_wait_a, g_cond_wait_b.
  unfold linkPG, conjPG, conjG.
  intros S S'' H.
  destruct H as [S' H].
  destruct H as [[HxS HSS'] HS'S''].
  intros mP lk cv bp sp pc HS.
  apply HS'S''.
  apply Hpp' with S; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold p_cond_wait_a in HS.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    unfold Ast_cond_wait_a in HS.
    destruct HS as [HM [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    rewrite HSP.
    sep on.
    sep lift 0 in HM.
    sep stksplit (IH_Stk + 6) in HM.
    sep lift 0 in HM.
    destruct HM as [M1 [M2 [HJM [HM1 HM2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    split_l.
    intros M1' HM1'.
    unfold p_cond_wait_a.
    exists lk; exists cv; exists bp; exists sp; exists pc.
    unfold Ast_cond_wait_a.
    simplS in * .
    rsplit.
      repeat split_l; trivial.
    sep on.
    sep lift 0.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      sep off.
      split.
        disjoint rep_l M1.
        eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    split_l.
    sep on.
    sep matchH HM2 with 4.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold g_cond_wait_a in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP lk cv bp sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  apply HSS'.
  clear HSS'.
  unfold Ast_cond_wait_a in * .
  simplS in * .
  destruct HS as [HM [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]].
  rsplit.
    repeat split_l; trivial.
  sep on.
  sep lift 0 in HM.
  sep stksplit (IH_Stk + 6) in HM.
  sep lift 0 in HM.
  destruct HM as [M2x [M1x [HJM12 [HM2x HM1x]]]].
  sep lift 0.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  assert (M2x = M2' /\ M1x = M1).
    sep off.
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    rewrite HSP in HM2'.
    assert (Hx := prec_disjUnion_inv H H0 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  rewrite HSP in HM2.
  split_l.
  sep on.
  sep matchH HM1x with 3.
Qed.

Lemma wff_cmd_cond_waitc:
   NH_WFcmd
     (SCAP_consCT (cond_wait +1 +1 +1 +1 +1 +1 +1 +1 +1 +1 +1 +1)
        (Ast p_cond_wait_c g_cond_wait_c 1) psi_cond_wait)
     (cond_wait +1 +1 +1 +1 +1 +1 +1 +1 +1 +1 +1)
     (Ast p_cond_wait_b g_cond_wait_b 1) (comml (_call l_acquire)).
Proof.
  set (cond_wait_b := cond_wait +1 +1 +1 +1 +1 +1 +1 +1 +1 +1 +1).
  assert(Hpp' : forall lk cv bp sp pc mQ mP S S',
    Ast_cond_wait_b lk cv bp sp pc mP S
    -> nextS (cond_wait_b) (comml (_call l_acquire)) S S'
    -> mQ = (stkSpacePrec (sp - 6 - _STK_SIZE_l_acquire) (_STK_SIZE_cond_wait - 6 - _STK_SIZE_l_acquire)
            *^* (sp - 2 |->w bp)
            *^* (sp |->w pc)
            *^* (sp + 2 |->w cv)
            *^* (sp + 4 |->w lk))
    -> Ast_l_acquire_pre lk (S_RF S BP) (sp - 6) (cond_wait_b + 1)
      (mQ *^* mP) S').
    intros lk cv bp sp pc mQ mP S S'.
    intros HS HSS' HmQ.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    unfold Ast_cond_wait_b in HS.
    destruct HS as [HM [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspx [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    subst M' R' fl' isr'.
    unfold Ast_l_acquire_pre.
    simplS in * .
    reg simpl.
    split.
      sep on.
      rewrite HSP.
      assert (H1 : sp - 4 - 2 = sp - 6).
        generalize Hsp; clearall; omega.
      rewrite H1.
      subst mQ.
      sep lift 3.
      sep_update.
      sep lift 2 in HM.
      sep stksplit in HM.
      rewrite H1 in HM.
      assert (H2 : _STK_SIZE_cond_wait - 4 - 2 = _STK_SIZE_cond_wait - 6).
        clearall; omega.
      rewrite H2 in HM.
      assert (H3 : _STK_SIZE_cond_wait - 6 >= _STK_SIZE_l_acquire).
        unfold _STK_SIZE_l_acquire.
        clearall; omega.
      assert (H4 : sp - 6 >= _STK_SIZE_l_acquire).
        unfold _STK_SIZE_l_acquire.
        generalize Hsp; clearall; omega.
      sep stksplit _STK_SIZE_l_acquire in HM.
      assert (H5 : sp - 6 + 2 = sp - 4).
        generalize Hsp; clearall; omega.
      rewrite H5.
      sep matchH HM with 11.
    split_l.
    split.
      rewrite HSP; omega.
    auto.
  assert (Hpp''' : forall lk cv bp sp pc pc' mP S S' S'' S''',
    Ast_cond_wait_b lk cv bp sp pc mP S
    -> nextS (cond_wait_b) (comml (_call l_acquire)) S S'
    -> l_acquire_grt S' S''
    -> nextS pc' (comms (_retw)) S'' S'''
    -> Ast_cond_wait_c lk cv bp sp pc mP S''').
    intros lk cv bp sp pc pc' mP.
    intros S S' S'' S''' HS HSS' HS'S'' HS''S'''.
    substH HS'S'' with (g_l_acquire_sound _ _ HS'S'').
    unfold g_l_acquire in HS'S''.
    set (mQ := (stkSpacePrec (sp - 6 - _STK_SIZE_l_acquire) (_STK_SIZE_cond_wait - 6 - _STK_SIZE_l_acquire)
            *^* (sp - 2 |->w bp)
            *^* (sp |->w pc)
            *^* (sp + 2 |->w cv)
            *^* (sp + 4 |->w lk))).
    assert (HS' : Ast_l_acquire_pre lk (S_RF S BP) (sp - 6) (cond_wait_b + 1) (mQ *^* mP) S').
      eapply Hpp'; eauto.
    assert (HS'' := HS'S'' _ _ _ _ _ HS').
    subst mQ.
    clear HS'S''.
    unfold Ast_l_acquire_post in HS''.
    unfold Ast_cond_wait_c.
    unfold Ast_cond_wait_b in HS.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct S'' as [[[M'' R''] fl''] isr''].
    destruct S''' as [[[M''' R'''] fl'''] isr'''].
    destruct HS as [HM [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]].
    simplS in * .
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct (nextS_inversion HS''S''') as [HM''' [HR''' [Hfl''' Hisr''']]].
    destruct HR''' as [pcx [Hlds HR''']].
    destruct HS'' as [HM'' [HBP'' [HSP'' [Hlkx [Hif'' Hisr'']]]]].
    assert (Hpcx : pcx = cond_wait_b + 1).
      subst R' M''' R'''.
      subst fl' isr' fl''' isr'''.
      rewrite HSP'' in Hlds; clear HSP''.
      sep on.
      sep lift 3 in HM''.
      sep_inj.
      auto.
    subst pcx; clear Hlds.
    subst M''' R''' fl''' isr'''.
    reg simpl.
    rewrite HSP''.
    assert (H5 : sp - 6 + 2 = sp - 4).
      generalize Hsp; clearall; omega.
    rewrite H5 in HM''.
    split.
      sep on.
      sep matchH HM'' with 11.
      sep stksplit.
      assert (H1 : sp - 4 - 2 = sp - 6).
        generalize Hsp; clearall; omega.
      rewrite H1.
      assert (H2 : _STK_SIZE_cond_wait - 4 - 2 = _STK_SIZE_cond_wait - 6).
        clearall; omega.
      rewrite H2.
      assert (H3 : _STK_SIZE_cond_wait - 6 >= _STK_SIZE_l_acquire).
        unfold _STK_SIZE_l_acquire.
        clearall; omega.
      assert (H4 : sp - 6 >= _STK_SIZE_l_acquire).
        unfold _STK_SIZE_l_acquire.
        generalize Hsp; clearall; omega.
      sep stksplit _STK_SIZE_l_acquire.
      sep matchH HM'' with 3.
    repeat split_l; trivial.
  simpl.
  split.
  right.
  split.
    apply l_acquire_not_prim.
  exists l_acquire_Prd; exists p_cond_wait_c.
  exists l_acquire_grt; exists g_cond_wait_c.
  split.    
    unfold psi_cond_wait; auto.
  split_l.
  split.
    clear Hpp' Hpp'''.
    unfold enable, p_cond_wait_b.
    intros S HS.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    unfold Ast_cond_wait_b in HS.
    destruct HS as [HM [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]].
    (* enable call *)
    destruct S as [[[M R] fl] isr].
    simplS in * .
    exists (MemUpdW M (sp - 4 - 2) (cond_wait_b+1) 0, RF_update R SP (sp - 4 - 2), fl, isr).
    apply nextS_call.
    unfold pushW.
    split.
      rewrite HSP.
      sep on.
      sep lift 2 in HM.
      sep stksplit 2 in HM.
      sep lift 0 in HM.
      sep trivial.
    split.
      rewrite HSP.
      apply MemUpdW_MemUpd16 with 0; trivial.
    rewrite HSP; trivial.
  split.
    (* g' => fun_frame *)
    unfold sp_call_pred, fun_frame.
    unfold linkPG, conjPG.
    intros S' S'' HS'S''.
    destruct HS'S'' as [H HS'S''].
    destruct H as [S [HS HSS']].
    substH HS'S'' with (g_l_acquire_sound _ _ HS'S'').
    unfold g_l_acquire in HS'S''.
    unfold p_cond_wait_b in HS.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    set (mQ := (stkSpacePrec (sp - 6 - _STK_SIZE_l_acquire) (_STK_SIZE_cond_wait - 6 - _STK_SIZE_l_acquire)
            *^* (sp - 2 |->w bp)
            *^* (sp |->w pc)
            *^* (sp + 2 |->w cv)
            *^* (sp + 4 |->w lk))).
    assert (HS' : Ast_l_acquire_pre lk (S_RF S BP) (sp - 6) (cond_wait_b+1) (mQ *^* mTrue) S').
      eapply Hpp'; eauto.
    clear Hpp' HS HSS' Hpp'''.
    substH HS'S'' with (HS'S'' _ _ _ _ _ HS').
    rename HS'S'' into HS''.
    destruct S' as [[[M' R'] fl'] isr'].
    destruct S'' as [[[M'' R''] fl''] isr''].
    simplS in * .
    unfold Ast_l_acquire_pre, Ast_l_acquire_post in * .
    destruct HS' as [HM' [HBP' [HSP' [Hlk' [Hif' Hisr']]]]].
    destruct HS'' as [HM'' [HBP'' [HSP'' [Hlk'' [Hif'' Hisr'']]]]].
    simplS in * .
    split.
      rewrite HSP'; rewrite HSP''; trivial.
    unfold Ret_Addr.
    exists (cond_wait_b + 1).
    simplS.
    split.
      intros pc''.
      apply nextPC_retw with (sp - 6); trivial.
      clear HM''.
      sep on.
      subst mQ.
      sep lift 3 in HM'.
      sep trivial.
    intros pc''.
    apply nextPC_retw with (sp - 6); trivial.
    clear HM'.
    sep on.
    subst mQ.
    sep lift 3 in HM''.
    sep trivial.
  split.
    (* p => p' *)
    intros S' H.
    unfold conjPG, conjG, linkPG in H.
    destruct H as [S [HS HnextSS']].
    apply p_l_acquire_sound.
    unfold p_cond_wait_b in HS.
    unfold p_l_acquire.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    exists lk; exists (S_RF S BP); exists (sp - 6); exists (cond_wait_b + 1).
    set (mQ := (stkSpacePrec (sp - 6 - _STK_SIZE_l_acquire) (_STK_SIZE_cond_wait - 6 - _STK_SIZE_l_acquire)
            *^* (sp - 2 |->w bp)
            *^* (sp |->w pc)
            *^* (sp + 2 |->w cv)
            *^* (sp + 4 |->w lk))).
    assert (HS' : Ast_l_acquire_pre lk (S_RF S BP) (sp - 6) (cond_wait_b + 1) (mQ *^* mTrue) S').
      eapply Hpp'; eauto.
    clear Hpp''' HS HnextSS'.
    unfold Ast_l_acquire_pre in * .
    destruct HS' as [HM' [HBP' [HSP' [Hlk' [Hif' Hisr']]]]].
    split.
      sep on.
      sep matchH HM' with 10.
    auto.
  split.
    (* p => p''' *)
    intros pc' S'''.
    unfold linkPG, wrap_fun, conjG.
    intros H.
    destruct H as [S H].
    destruct H as [HS H].
    destruct H as [S'' [H HS''S''']].
    destruct H as [S' [HSS' HS'S'']].
    unfold p_cond_wait_b, p_cond_wait_c in * .
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    exists lk; exists cv; exists bp; exists sp; exists pc.
    eapply Hpp'''; eauto.
  (* g''' => g *)
  intros pc' S S'''' H.
  unfold conjPG, wrap_fun, conjG, linkPG in H.
  destruct H as [S''' [HS HS'''S'''']].
  destruct HS as [HSx [S'' [Hx HnextS'']]].
  destruct Hx as [S' [HSS' HS'S'']].
  unfold g_cond_wait_b, g_cond_wait_c in * .
  intros mP lk cv bp sp pc HS.
  eapply HS'''S''''; eauto.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold p_cond_wait_b in HS.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    unfold Ast_cond_wait_b in HS.
    destruct HS as [HM [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold g_cond_wait_b in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP lk cv bp sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_cond_wait_b in HS.
  simplS in * .
  destruct HS as [HM [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma wff_cmd_cond_waitd:
   NH_WFcmd
     (SCAP_consCT
        (cond_wait + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
        (Ast p_cond_wait_d g_cond_wait_d 1) psi_cond_wait)
     (cond_wait + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
     (Ast p_cond_wait_c g_cond_wait_c 1) (comml _sti).
Proof.
  set (cond_wait_c := cond_wait + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1).
  assert (Hpp': forall lk cv bp sp pc mP S S', 
    Ast_cond_wait_c lk cv bp sp pc mP S
    -> gsti S S' 
    -> Ast_cond_wait_d lk cv bp sp pc mP S').
    intros lk cv bp sp pc mP S S' HS HSS'.
    unfold linkPG, conjPG, conjG.
    unfold Ast_cond_wait_c, gsti, Ast_cond_wait_d in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct HS as [HM [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]].
    destruct HSS' as [Hshift [Hif' [Hisr' HR']]].
    assert (Hshift' : prepost (INV0 *^* INV1) Aemp
      (M, R, fl, isr) (M', R', fl', isr')).
      apply (prepost_weaken_pre (mP':=INV0 *^* INV1) Hshift).
      intros Mx HMx.
      assert (Hitr : ~ itr_enabled (M, R, fl, isr)).
        unfold itr_enabled.
        intros [H1 H2].
        simplS in * .
        rewrite Hif in H1; discriminate H1.
      eapply choice_elim_right; eauto.
    clear Hshift; rename Hshift' into Hshift.
    split.
      sep on.
      substH HM with (prepost_elim Hshift Prec01 HM).
      sep on.
      simplS in * .
      sep matchH HM with 6.
      sep trivial.
    simplS in * .
    subst R' isr'.
    repeat split_l; trivial.
  simpl.
  split.
  exists p_cond_wait_d; exists g_cond_wait_d.
  split_l.
  split.
  (* enable *)
     unfold enable_sti.
     intros S HS.
     unfold p_cond_wait_c in HS.
     unfold pTrue, pmstar, choice.
     destruct HS as [lk [cv [bp [sp [pc HS]]]]].
     unfold Ast_cond_wait_c in HS.
     destruct HS as [HM [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]].
     destruct S as [[[M R] fl] isr].
     simplS in * .
     destruct HM as [M2 [M1 [HJM [HM2 HM1]]]].
     exists M1; exists M2.
     split.
       auto with PMap.
     split_l.
     right.
     split_r.
     intros [H1 H2].
     simplS in * .
     rewrite Hif in H1; discriminate H1.
  split.
  (* p ==> p' *)
  unfold p_cond_wait_c, p_cond_wait_d.
  intros S' H.
  unfold linkPG in H.
  destruct H as [S [HS HSS']].
  destruct HS as [lk [cv [bp [sp [pc HS]]]]].
  exists lk; exists cv; exists bp; exists sp; exists pc.
  apply Hpp' with S; trivial.
  (* g' ==> g *)
  unfold g_cond_wait_c, g_cond_wait_d.
  unfold linkPG, conjPG, conjG.
  intros S S'' H.
  destruct H as [S' H].
  destruct H as [[HxS HSS'] HS'S''].
  intros mP lk cv bp sp pc HS.
  apply HS'S''.
  apply Hpp' with S; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold p_cond_wait_c in HS.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    unfold Ast_cond_wait_c in HS.
    destruct HS as [HM [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold g_cond_wait_c in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP lk cv bp sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_cond_wait_c in HS.
  simplS in * .
  destruct HS as [HM [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma wff_cmd_cond_waite:
   NH_WFcmd
     (SCAP_consCT
        (cond_wait + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
        (Ast p_cond_wait_e g_cond_wait_e 1) psi_cond_wait)
     (cond_wait + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
     (Ast p_cond_wait_d g_cond_wait_d 1) (comml (_addw (Opr_imm 2) SP)).
Proof.
  set (cond_wait_d := cond_wait + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1).
  assert (Hpp' : forall lk cv bp sp pc mP S S',
    Ast_cond_wait_d lk cv bp sp pc mP S
    -> nextS (cond_wait_d) (comml (_addw (Opr_imm 2) SP)) S S'
    -> Ast_cond_wait_e lk cv bp sp pc mP S').
    intros lk cv bp sp pc mP S S' HS HSS'.
    unfold Ast_cond_wait_d, Ast_cond_wait_e in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    simplS in * .
    destruct HS as [HM [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]].
    subst M' R' isr'.
    split.
      sep on.
      sep matchH HM with 5.
      sep stksplit.
      assert (H1 : sp - 2 - 2 = sp - 4).
        generalize Hsp; clearall; omega.
      rewrite H1.
      assert (H2 : _STK_SIZE_cond_wait - 2 - 2 = _STK_SIZE_cond_wait - 4).
        clearall; omega.
      rewrite H2.
      sep matchH HM with 2.
    reg simpl.
    unfold opr_eval.
    rewrite HSP.
    split_l by (generalize Hsp; clearall; omega).
    repeat split_l; trivial.
    unfold trash_fl in Hfl'.
    destruct Hfl'; subst fl'; auto.
  split.
  simpl.
  exists p_cond_wait_e; exists g_cond_wait_e.
  split_l.
  split.
    (* enable addw *)
    apply trivial_enable_addw.
  split. (* p ==> p' *)
    unfold p_cond_wait_d, p_cond_wait_e.
    intros S' H.
    unfold linkPG in H.
    destruct H as [S H].
    simpl in H.
    destruct H as [H HnextS].
    destruct H as [lk [cv [bp [sp [pc H]]]]].
    exists lk; exists cv; exists bp; exists sp; exists pc.
    apply Hpp' with S; trivial.
  (* gimp *)
  intros S S'' H.
  unfold conjPG, conjG in H.
  destruct H as [S' [[HSx HnextS] HS'S'']].
  unfold g_cond_wait_d, g_cond_wait_e in * .
  intros mP lk cv bp sp pc HS.
  eapply HS'S''; eauto.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold p_cond_wait_d in HS.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    unfold Ast_cond_wait_d in HS.
    destruct HS as [HM [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    rewrite HSP.
    sep on.
    sep lift 0 in HM.
    sep stksplit (IH_Stk + 6) in HM.
    sep lift 0 in HM.
    destruct HM as [M1 [M2 [HJM [HM1 HM2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    split_l.
    intros M1' HM1'.
    unfold p_cond_wait_d.
    exists lk; exists cv; exists bp; exists sp; exists pc.
    unfold Ast_cond_wait_d.
    simplS in * .
    rsplit.
      repeat split_l; trivial.
    sep on.
    sep lift 0.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      sep off.
      split.
        disjoint rep_l M1.
        eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    split_l.
    sep on.
    sep matchH HM2 with 4.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold g_cond_wait_d in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP lk cv bp sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  apply HSS'.
  clear HSS'.
  unfold Ast_cond_wait_d in * .
  simplS in * .
  destruct HS as [HM [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]].
  rsplit.
    repeat split_l; trivial.
  sep on.
  sep lift 0 in HM.
  sep stksplit (IH_Stk + 6) in HM.
  sep lift 0 in HM.
  destruct HM as [M2x [M1x [HJM12 [HM2x HM1x]]]].
  sep lift 0.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  assert (M2x = M2' /\ M1x = M1).
    sep off.
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    rewrite HSP in HM2'.
    assert (Hx := prec_disjUnion_inv H H0 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  rewrite HSP in HM2.
  split_l.
  sep on.
  sep matchH HM1x with 3.
Qed.

Lemma wff_cmd_cond_waitf:
   NH_WFcmd
     (SCAP_consCT
        (cond_wait + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1) 
        (Ast p_cond_wait_f g_cond_wait_f 1) psi_cond_wait)
     (cond_wait + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
     (Ast p_cond_wait_e g_cond_wait_e 1) (comml (_popw BP)).
Proof.
  set (cond_wait_e := cond_wait + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1).
  assert (Hpp' : forall lk cv bp sp pc mP S S',
    Ast_cond_wait_e lk cv bp sp pc mP S
    -> nextS cond_wait_e (comml (_popw BP)) S S'
    -> Ast_cond_wait_f lk cv bp sp pc mP S').
    intros lk cv bp sp pc mP S S' HS HSS'.
    unfold Ast_cond_wait_e, Ast_cond_wait_f in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HR' as [bpx [Hlds HR']].
    simplS in * .
    destruct HS as [HM [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]].
    assert (Heq : bp = bpx).
      sep on.
      sep lift 1 in HM.
      rewrite HSP in Hlds.
      sep inj.
      auto.
    subst bpx.
    subst M' fl' R' isr'.
    split.
      sep on.
      sep matchH HM with 5.
      sep stksplit.
      sep matchH HM with 2.
    rewrite HSP.
    reg simpl.
    split_l.
    split_l by omega.
    auto.
  simpl.
  split.
  exists p_cond_wait_f; exists g_cond_wait_f.
  split_l.
  split.
  (* enable *)
    apply enable_popw.
    intros S HS.
    destruct S as [[[M R] fl] isr].
    unfold p_cond_wait_e in HS.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    unfold Ast_cond_wait_e in HS.
    destruct HS as [HM [HSP [Hsp [Hcv [Hif Hisr]]]]].
    simplS in * .
    rewrite HSP.
    sep on.
    sep lift 1 in HM.
    sep trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold p_cond_wait_e, p_cond_wait_f.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    exists lk; exists cv; exists bp; exists sp; exists pc.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HxS HSS'] HS'S''].
  unfold g_cond_wait_e, g_cond_wait_f in * .
  intros mP lk cv bp sp pc HS.
  apply HS'S''.
  apply Hpp' with S; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold p_cond_wait_e in HS.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    unfold Ast_cond_wait_e in HS.
    destruct HS as [HM [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    rewrite HSP.
    sep on.
    sep lift 0 in HM.
    sep stksplit (IH_Stk + 6) in HM.
    sep lift 0 in HM.
    destruct HM as [M1 [M2 [HJM [HM1 HM2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    split_l.
    intros M1' HM1'.
    unfold p_cond_wait_e.
    exists lk; exists cv; exists bp; exists sp; exists pc.
    unfold Ast_cond_wait_e.
    simplS in * .
    rsplit.
      repeat split_l; trivial.
    sep on.
    sep lift 0.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      sep off.
      split.
        disjoint rep_l M1.
        eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    split_l.
    sep on.
    sep matchH HM2 with 4.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold g_cond_wait_8 in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP lk cv bp sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  apply HSS'.
  clear HSS'.
  unfold Ast_cond_wait_e in * .
  simplS in * .
  destruct HS as [HM [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]].
  rsplit.
    repeat split_l; trivial.
  sep on.
  sep lift 0 in HM.
  sep stksplit (IH_Stk + 6) in HM.
  sep lift 0 in HM.
  destruct HM as [M2x [M1x [HJM12 [HM2x HM1x]]]].
  sep lift 0.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  assert (M2x = M2' /\ M1x = M1).
    sep off.
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    rewrite HSP in HM2'.
    assert (Hx := prec_disjUnion_inv H H0 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  rewrite HSP in HM2.
  split_l.
  sep on.
  sep matchH HM1x with 3.
Qed.

Lemma wff_cmd_cond_wait10:
 NH_WFiseq psi_cond_wait
   (cond_wait + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
   (Ast p_cond_wait_f g_cond_wait_f 1) (sseq _retw).
Proof.
  constructor.
  simpl.
  split.
  split_l.
  intros S HS.
  unfold g_cond_wait_f, p_cond_wait_f in * .
  unfold Ast_cond_wait_f, Ast_cond_wait_post, Ast_cond_wait_f.
  intros mP lk cv bp sp pc H.
  InvertAll.
  repeat split_l; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold p_cond_wait_f in HS.
    destruct HS as [lk [cv [bp [sp [pc HS]]]]].
    unfold Ast_cond_wait_f in HS.
    destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    rewrite HSP.
    sep on.
    sep lift 0 in HM.
    sep stksplit (IH_Stk + 6) in HM.
    sep lift 0 in HM.
    destruct HM as [M1 [M2 [HJM [HM1 HM2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    split_l.
    intros M1' HM1'.
    unfold p_cond_wait_f.
    exists lk; exists cv; exists bp; exists sp; exists pc.
    unfold Ast_cond_wait_f.
    simplS in * .
    rsplit.
      repeat split_l; trivial.
    sep on.
    sep lift 0.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      sep off.
      split.
        disjoint rep_l M1.
        eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    split_l.
    sep on.
    sep matchH HM2 with 4.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold g_cond_wait_f in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP lk cv bp sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  apply HSS'.
  clear HSS'.
  unfold Ast_cond_wait_f in * .
  simplS in * .
  destruct HS as [HM [HBP [HSP [Hsp [Hcv [Hlk [Hif Hisr]]]]]]].
  rsplit.
    repeat split_l; trivial.
  sep on.
  sep lift 0 in HM.
  sep stksplit (IH_Stk + 6) in HM.
  sep lift 0 in HM.
  destruct HM as [M2x [M1x [HJM12 [HM2x HM1x]]]].
  sep lift 0.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  assert (M2x = M2' /\ M1x = M1).
    sep off.
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    rewrite HSP in HM2'.
    assert (Hx := prec_disjUnion_inv H H0 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  rewrite HSP in HM2.
  split_l.
  sep on.
  sep matchH HM1x with 3.
Qed.

Lemma wff_cond_wait:
  NH_WFiseq psi_cond_wait cond_wait (Ast p_cond_wait g_cond_wait 1) I_cond_wait.
Proof.
  unfold I_cond_wait.
  apply nh_isq_seq with p_cond_wait_1 g_cond_wait_1.
  apply wff_cmd_cond_wait1.
  apply nh_isq_seq with p_cond_wait_2 g_cond_wait_2.
  apply wff_cmd_cond_wait2.
  apply nh_isq_seq with p_cond_wait_3 g_cond_wait_3.
  apply wff_cmd_cond_wait3.
  apply nh_isq_seq with p_cond_wait_4 g_cond_wait_4.
  apply wff_cmd_cond_wait4.
  apply nh_isq_seq with p_cond_wait_5 g_cond_wait_5.
  apply wff_cmd_cond_wait5.
  apply nh_isq_seq with p_cond_wait_6 g_cond_wait_6.
  apply wff_cmd_cond_wait6.
  apply nh_isq_seq with p_cond_wait_7 g_cond_wait_7.
  apply wff_cmd_cond_wait7.
  apply nh_isq_seq with p_cond_wait_8 g_cond_wait_8.
  apply wff_cmd_cond_wait8.
  apply nh_isq_seq with p_cond_wait_9 g_cond_wait_9.
  apply wff_cmd_cond_wait9.
  apply nh_isq_seq with p_cond_wait_a g_cond_wait_a.
  apply wff_cmd_cond_waita.
  apply nh_isq_seq with p_cond_wait_b g_cond_wait_b.
  apply wff_cmd_cond_waitb.
  apply nh_isq_seq with p_cond_wait_c g_cond_wait_c.
  apply wff_cmd_cond_waitc.
  apply nh_isq_seq with p_cond_wait_d g_cond_wait_d.
  apply wff_cmd_cond_waitd.
  apply nh_isq_seq with p_cond_wait_e g_cond_wait_e.
  apply wff_cmd_cond_waite.
  apply nh_isq_seq with p_cond_wait_f g_cond_wait_f.
  apply wff_cmd_cond_waitf.
  apply wff_cmd_cond_wait10.
Qed.
