(***************************************************************)
(* Coq code for AIM                                            *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

Add LoadPath "..".
Add LoadPath "../CAP0".
Add LoadPath "../SCAP".
Add LoadPath "../miniOS".

Require Import Coqlib.

Require Import List.
Require Import Arith.
Require Import Machine.
Require Import Assertion.
Require Import Assertion_Lemmas.
Require Import Assertion_Lemmas4.
Require Import Assertion_Lemmas5.
Require Import Operational_Semantics.
Require Import OpSem_Lemmas.
Require Import OpSem_Lemmas2.
Require Import Reg.
Require Import Mem_Lemmas.

Require Import AIM.
Require Import QLib.

Require Import SCAP.
Require Import SCAP_READY.
Require Import Linking.
Require Import Linking_Lemmas.
Require Import Omega.
Require Import Assembly.
Require Import SyncUtil.
Require Import Kernel.
Require Import SleepSpec.

(* the stack space required by l_release *)
Definition _STK_SIZE_l_acquire := 6 + _STK_SIZE_sleep.

Definition l_acquire_MPre (sp l pc : Label) : MPred :=
    (sp |->w pc) *^* (sp + 2 |->w l) *^* stkSpacePrec sp _STK_SIZE_l_acquire
       *^* INV1 *^* INV0.

Definition l_acquire_Prd_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
      R SP = sp
      /\ R BP = bp
      /\ (l_acquire_MPre sp l pc *^* mP) M
      /\ In l lockSet_H
      /\ fl _IF = false
      /\ isr = false
  end.

Definition l_acquire_Prd (S : State) : Prop :=
  exists sp, exists l, exists pc, exists bp, exists mP, l_acquire_Prd_aux sp l pc bp mP S.


Definition l_acquire_MPost (sp l pc: Label) : MPred :=
    (l_acquire_MPre sp l pc) *^* (Gamma l).


Definition l_acquire_Post_aux (sp l pc bp : Label) (mP :MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
      R SP = sp
      /\ R BP = bp
      /\ (l_acquire_MPost sp l pc *^* mP) M
      /\ In l lockSet_H
      /\ fl _IF = false
      /\ isr = false
  end.

Definition l_acquire_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (l_acquire_Prd_aux sp l pc bp) (l_acquire_Post_aux sp l pc bp) S S'.


Definition l_setacq_Prd_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R SP = sp
        /\ R BX = l
        /\ ((sp |->w bp) *^* (sp + 2 |->w pc) *^* (sp + 4 |->w l) 
              *^* stkSpacePrec sp (_STK_SIZE_l_acquire - 2)
              *^* INV1 *^* INV0 *^* mP) M
        /\ (exists avail, (l + L_AVAIL |~>w avail) M  /\ R DX = avail /\ avail <> 0)
        /\ In l lockSet_H
        /\ fl _IF = false
        /\ isr = false
  end.


Definition l_setacq_Prd (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, l_setacq_Prd_aux sp l pc bp mP S.

Definition l_setacq_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (l_setacq_Prd_aux sp l pc bp) (l_acquire_Post_aux (sp +2) l pc bp) S S'.



Definition l_acquire_psi := SCAP_consCT L_sleep  (Ast L_sleep_0_p L_sleep_0_g 1) 
                           (SCAP_consCT l_setacq (Ast l_setacq_Prd l_setacq_grt 1) 
                                        SCAP_emptCT).

Lemma L_sleep_in_psi:
   SCAP_lookupCT l_acquire_psi L_sleep (Ast L_sleep_0_p L_sleep_0_g 1).
Proof.
  unfold SCAP_lookupCT.
  unfold l_acquire_psi.
  unfold SCAP_consCT at 1.
  unfold SPSpec.lookup.
  simpl. left. auto.
Qed.


Lemma l_setacq_in_psi:
   SCAP_lookupCT l_acquire_psi l_setacq (Ast l_setacq_Prd l_setacq_grt 1).
Proof.
  unfold SCAP_lookupCT.
  unfold l_acquire_psi.
  unfold SCAP_consCT at 1.
  unfold SPSpec.lookup.
  simpl. right. left. auto.
Qed.


(* spec for the next instruction *)

Definition l_acquire1_MPre (sp l pc bp : Label) :=
    (sp |->w bp) *^* (sp + 2 |->w pc) *^* (sp + 4 |->w l) *^* stkSpacePrec sp (_STK_SIZE_l_acquire - 2)
       *^* INV1 *^* INV0.

Definition l_acquire1_Prd_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R SP = sp
        /\ R BP = bp
        /\ (l_acquire1_MPre (R SP) l pc (R BP) *^* mP) M
        /\ In l lockSet_H
        /\ fl _IF = false
        /\ isr = false
  end.

Definition l_acquire1_Prd (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, l_acquire1_Prd_aux sp l pc bp mP S.


Definition l_acquire1_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (l_acquire1_Prd_aux sp l pc bp) (l_acquire_Post_aux (sp +2) l pc bp) S S'.

Lemma l_acquire_Prd_IF_false:
   forall S : State, l_acquire_Prd S -> S_Flag S _IF = false.
Proof.
  unfold l_acquire_Prd.
  intros.
  induction S as (((M, R), fl), isr).
  unfold l_acquire_Prd_aux in H.
  InvertAll.
  UnfoldStateAll. auto.
Qed.


Lemma l_acquire_Grt_Itr_Stk: Grt_Itr_Stk l_acquire_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  unfold l_acquire_grt.
  unfold leadstoPQ2. intros.
  unfold l_acquire_Prd_aux in H4.
  induction S as (((M, R), fl), isr).
  UnfoldStateAll. InvertAll.
  unfold itr_enabled in H0.
  inversion_clear H0.
  UnfoldStateAll.
  rewrite H9 in H8. inversion H8.
Qed.

Lemma _STI_SIZE_acq_ge2:    _STK_SIZE_l_acquire >= 2.
Proof.
  unfold _STK_SIZE_l_acquire. omega.
Qed.
Hint Resolve _STI_SIZE_acq_ge2.


Lemma l_acquire_Prd_aux_stkSpace:
  forall sp l pc bp mP S,
   l_acquire_Prd_aux sp l pc bp mP S -> stkSpace (S_RF S SP) _STK_SIZE_l_acquire (S_Mem S).
Proof.
  intros.
  unfold l_acquire_Prd_aux in H.
  induction S as (((M, R), fl), isr).
  unfold l_acquire_MPre in H.
  InvertAll.
  sep on.
  sep lift 2 in H1.
  UnfoldStateAll.
  rewrite H0.
  sep trivial.
Qed.



Lemma l_acquire_step1:
forall f sp l pc bp mP S S',
  l_acquire_Prd_aux sp l pc bp mP S  
  -> nextS f (comml (_pushwp (Opr_reg BP))) S S'
  -> l_acquire1_Prd_aux (sp - 2) l pc bp mP S'.
Proof.
  intros.
  rename H0 into Htrans.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion Htrans).
  destruct H0 as [Hspgt2 [Hmemblk HM']].
  destruct H1 as [HR' [Hfl' Hisr']].
  subst R' fl' isr'. clear Htrans.
  unfold opr_eval in HM'.
  unfold l_acquire_Prd_aux in H.
  destruct H as [HRsp [HRbp [Hmem [Hlock [Hfl Hisr]]]]].
  unfold l_acquire_MPre in Hmem.
  unfold l_acquire1_Prd_aux.
  reg simpl.
  rewrite HRsp.
  split_l; split_l.
  split; auto.
  unfold l_acquire1_MPre.
  rewrite HRsp in Hspgt2.
  assert (sp - 2 + 2 = sp). auto.
  rewrite H. clear H.
  assert (sp - 2 + 4 = sp + 2). auto.
  rewrite H. clear H.
  substH HM' with (MemUpd16_MemUpdW HM').
  destruct HM' as [junk HM'].
  subst M'.
  sep on.
  rewrite HRsp. rewrite HRbp.
  sep lift 0.
  sep upd2.
  sep lift 2 in Hmem.
  sep stksplit in Hmem.
  sep matchH Hmem with 7.
Qed.


Lemma wff_cmd_l_acquire1:
   NH_WFcmd
     (SCAP_consCT (l_acquire + 1) (Ast l_acquire1_Prd l_acquire1_grt 1)
        l_acquire_psi) l_acquire (Ast l_acquire_Prd l_acquire_grt 1)
     (comml (_pushwp (Opr_reg BP))).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists l_acquire1_Prd. exists l_acquire1_grt.
  split. auto.

  assert (forall sp l pc bp mP, enable (l_acquire_Prd_aux sp l pc bp mP) (nextS l_acquire (comml (_pushwp (Opr_reg BP))))).
  intros.
  apply enable_pushwp with _STK_SIZE_l_acquire; auto.
  apply l_acquire_Prd_aux_stkSpace.
  auto.

  split.
  unfold enable in *.
  intros.
  unfold l_acquire_Prd in H0.
  InvertAll.
  apply H with x x0 x1 x2 x3; auto.
  split.
  unfold pimp.
  unfold linkPG.
  intros.
  InvertAll.
  unfold l_acquire_Prd in H0.
  InvertAll.
  unfold l_acquire1_Prd.
  mexists5 (x0 - 2) x1 x2 x3 x4.

  apply l_acquire_step1 with l_acquire x; auto.
  unfold gimp.
  unfold conjPG. unfold conjG.
  intros.
  InvertAll.
  unfold l_acquire_grt.
  unfold leadstoPQ2.
  intros.
  unfold l_acquire1_grt in H2.
  unfold leadstoPQ2 in H2.
  assert (l_acquire_Post_aux (sp - 2 + 2) l pc bp mP S').
  apply H2.
  apply l_acquire_step1 with l_acquire S; auto.
  induction S as (((M, R), fl), isr).
  induction x as (((M', R'), fl'), isr').
  destruct (nextS_inversion H3).
  InvertAll.
  unfold l_acquire_Prd_aux in H0.
  InvertAll.
  rewrite H10 in H8.
  assert (sp - 2 + 2 = sp). auto.
  rewrite H15 in H4. auto.
  split. 
  apply IF_false_Prd_Itr_Stk.
  apply l_acquire_Prd_IF_false.
  apply l_acquire_Grt_Itr_Stk.
Qed.

Definition l_acquire2_Prd_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R BP = sp
        /\ l_acquire1_Prd_aux sp l pc bp mP (M, RF_update R BP bp, fl, isr)
  end.

Definition l_acquire2_Prd (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, l_acquire2_Prd_aux sp l pc bp mP S.

Definition l_acquire2_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (l_acquire2_Prd_aux sp l pc bp) (l_acquire_Post_aux (sp +2) l pc bp) S S'.


Lemma l_acquire1_Prd_IF_false:
 forall S : State, l_acquire1_Prd S -> S_Flag S _IF = false.
Proof.
    unfold l_acquire1_Prd.
    intros.
    induction S as (((M, R), fl), isr).
    unfold l_acquire1_Prd_aux in H.
    UnfoldStateAll. InvertAll. auto.
Qed.

Lemma l_acquire1_Grt_Itr_Stk: Grt_Itr_Stk l_acquire1_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  unfold l_acquire1_grt.
  unfold leadstoPQ2. intros.
  unfold l_acquire1_Prd_aux in H4.
  induction S as (((M, R), fl), isr).
  UnfoldStateAll. InvertAll.
  unfold itr_enabled in H0.
  inversion_clear H0.
  UnfoldStateAll.
  rewrite H9 in H8. inversion H8.
Qed.

Lemma l_acquire_step2:
forall f sp l pc bp mP S S',
  l_acquire1_Prd_aux sp l pc bp mP S
  -> nextS f (comml (_movwp (Opr_reg SP) BP)) S S'
  -> l_acquire2_Prd_aux sp l pc bp mP S'.
Proof.
  intros.
  rename H0 into Htrans.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion Htrans).
  subst M'.
  destruct H1 as [HR' [Hfl' Hisr']].
  unfold opr_eval in HR'.
  subst R' fl' isr'. clear Htrans.
  unfold l_acquire1_Prd_aux in H.
  destruct H as [HRsp [HRbp [Hmem [Hlock [Hfl Hisr]]]]].
  unfold l_acquire2_Prd_aux.
  reg simpl.
  rewrite HRsp.
  split_l.
  unfold l_acquire1_Prd_aux.
  split_l; split_l.
  split; auto.
  reg simpl. rewrite HRbp in Hmem. auto.
Qed.


Lemma wff_cmd_l_acquire2:
   NH_WFcmd
     (SCAP_consCT (l_acquire + 1 + 1) (Ast l_acquire2_Prd l_acquire2_grt 1)
        l_acquire_psi) (l_acquire + 1) (Ast l_acquire1_Prd l_acquire1_grt 1)
     (comml (_movwp (Opr_reg SP) BP)).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists l_acquire2_Prd. exists l_acquire2_grt.
  split. auto.
  split; auto.

  split.
  unfold pimp.
  unfold linkPG.
  intros.
  InvertAll.
  unfold l_acquire1_Prd in H.
  InvertAll.
  unfold l_acquire2_Prd.
  mexists5 x0 x1 x2 x3 x4.

  apply l_acquire_step2 with (l_acquire + 1) x; auto.

  unfold gimp.
  unfold conjPG. unfold conjG.
  intros.
  InvertAll.
  unfold l_acquire1_grt.
  unfold leadstoPQ2.
  intros.
  unfold l_acquire2_grt in H1.
  unfold leadstoPQ2 in H1.
  apply H1.
  apply l_acquire_step2 with (l_acquire + 1) S; auto.


  split. 
    apply IF_false_Prd_Itr_Stk.
    apply l_acquire1_Prd_IF_false.
    apply l_acquire1_Grt_Itr_Stk.
Qed.


(* _movwld (A_reg 4 BP) BX *)
Definition l_acquire3_Prd_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R BX = l
        /\ exists bx, l_acquire2_Prd_aux sp l pc bp mP (M, RF_update R BX bx, fl, isr)
  end.

Definition l_acquire3_Prd (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, l_acquire3_Prd_aux sp l pc bp mP S.

Definition l_acquire3_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (l_acquire3_Prd_aux sp l pc bp) (l_acquire_Post_aux (sp +2) l pc bp) S S'.

Lemma l_acquire2_Prd_IF_false:
 forall S : State, l_acquire2_Prd S -> S_Flag S _IF = false.
Proof.
    unfold l_acquire2_Prd.
    intros.
    induction S as (((M, R), fl), isr).
    unfold l_acquire2_Prd_aux in H.
    unfold l_acquire1_Prd_aux in H.
    UnfoldStateAll. InvertAll.
    auto.
Qed.

Lemma l_acquire2_Grt_Itr_Stk: Grt_Itr_Stk l_acquire2_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  unfold l_acquire2_grt.
  unfold leadstoPQ2. intros.
  unfold l_acquire2_Prd_aux in H4.
  induction S as (((M, R), fl), isr).
  UnfoldStateAll.
  unfold l_acquire1_Prd_aux in H4.
  InvertAll.
  unfold itr_enabled in H0.
  inversion_clear H0.
  UnfoldStateAll.
  rewrite H10 in H9. inversion H9.
Qed.

Lemma STK_SZ_acq_N2_ge2: _STK_SIZE_l_acquire - 2 >= 2.
Proof.
  unfold _STK_SIZE_l_acquire.
  omega.
Qed.
Hint Resolve STK_SZ_acq_N2_ge2.

Lemma l_acquire2_Prd_1:
   forall sp l pc bp mP S,
   l_acquire2_Prd_aux sp l pc bp mP S ->
   ((addr_eval (S_RF S) (A_reg 4 BP)) |~>w l) (S_Mem S).
Proof.
  intros.
  unfold l_acquire2_Prd_aux in H.
  induction S as (((M, R), fl), isr).
  unfold l_acquire1_Prd_aux in H.
  unfold l_acquire1_MPre in H.
  InvertAll.
  reg simpl.
  unfold addr_eval.
  unfold S_RF. unfold S_Mem.
  sep on.
  sep lift 2 in H2.
  rewrite H0. rewrite H in H2.
  sep trivial.
Qed.

Lemma l_acquire_step3:
forall f sp l pc bp mP S S',
  l_acquire2_Prd_aux sp l pc bp mP S
  -> nextS f (comml (_movwld (A_reg 4 BP) BX)) S S'
  -> l_acquire3_Prd_aux sp l pc bp mP S'.
Proof.
  intros.
  rename H0 into Htrans.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion Htrans).
  subst M'.
  destruct H1 as [[x [Hmemx HR']] [Hfl' Hisr']].
  assert (x = l).
  assert ((addr_eval (S_RF (M, R, fl, isr)) (A_reg 4 BP) |~>w l) (S_Mem (M, R, fl, isr))).
  apply l_acquire2_Prd_1 with sp pc bp mP; auto.
  UnfoldStateAll.
  apply leadsToW_unique with (addr_eval R (A_reg 4 BP)) M; auto.
  subst x.
  subst R' fl' isr'.
  unfold l_acquire3_Prd_aux.
  split. reg simpl. auto.
  exists (R BX).
  assert (RF_update (RF_update R BX l) BX (R BX) = R).
  assert (RF_update (RF_update R BX l) BX (R BX) = RF_update R BX (R BX)).
  apply RF_update_RF_update.
  rewrite H0.
  apply RF_update_prev.
  rewrite H0.
  auto.
Qed.


Lemma wff_cmd_l_acquire3:
   NH_WFcmd
     (SCAP_consCT (l_acquire + 1 + 1 + 1)
        (Ast l_acquire3_Prd l_acquire3_grt 1) l_acquire_psi)
     (l_acquire + 1 + 1) (Ast l_acquire2_Prd l_acquire2_grt 1)
     (comml (_movwld (A_reg 4 BP) BX)).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists l_acquire3_Prd. exists l_acquire3_grt.
  split. auto.

  (* enable *)
  generalize (l_acquire + 1 + 1). intro f.
  assert (forall sp l pc bp mP, enable (l_acquire2_Prd_aux sp l pc bp mP) 
                                    (nextS f (comml (_movwld (A_reg 4 BP) BX)))).
  intros.
  apply enable_movwld.
  intros.
  assert (((addr_eval (S_RF S) (A_reg 4 BP)) |~>w l) (S_Mem S)).
  apply l_acquire2_Prd_1 with sp pc bp mP; auto.
  apply leadsToW_leadsToxW with l; auto.

  split.
  unfold enable in *.
  intros.
  unfold l_acquire2_Prd in H0.
  InvertAll.
  apply H with x x0 x1 x2 x3; auto.
  split.
  unfold pimp.
  unfold linkPG.
  intros.
  InvertAll.
  unfold l_acquire2_Prd in H0.
  InvertAll.
  unfold l_acquire3_Prd.
  exists x0. exists x1. exists x2. exists x3. exists x4.

  apply l_acquire_step3 with f x; auto.

  unfold gimp.
  unfold conjPG. unfold conjG.
  intros.
  InvertAll.
  unfold l_acquire2_grt.
  unfold leadstoPQ2.
  intros.
  unfold l_acquire3_grt in H2.
  unfold leadstoPQ2 in H2.
  apply H2.
  apply l_acquire_step3 with f S; auto.

  split. 
    apply IF_false_Prd_Itr_Stk.
      apply l_acquire2_Prd_IF_false.
    apply l_acquire2_Grt_Itr_Stk.
Qed.


Definition l_acquire4_Prd_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R SP = sp
        /\ R BX = l
        /\ (l_acquire1_MPre sp l pc bp *^* mP) M
        /\ (exists avail, (l + L_AVAIL |~>w avail) M  /\ R DX = avail)
        /\ In l lockSet_H
        /\ fl _IF = false
        /\ isr = false
  end.

Definition l_acquire4_Prd (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, l_acquire4_Prd_aux sp l pc bp mP S.

Definition l_acquire4_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (l_acquire4_Prd_aux sp l pc bp) (l_acquire_Post_aux (sp +2) l pc bp) S S'.

Lemma l_acquire3_Prd_1:
   forall sp l pc bp mP S,
   l_acquire3_Prd_aux sp l pc bp mP S ->
   ((addr_eval (S_RF S) (A_reg L_AVAIL BX)) |~>w_) (S_Mem S).
Proof.
  intros.
  unfold l_acquire3_Prd_aux in H.
  induction S as (((M, R), fl), isr).
  unfold l_acquire2_Prd_aux in H.
  unfold l_acquire1_Prd_aux in H.
  unfold l_acquire1_MPre in H.
  InvertAll.
  reg simpl.
  unfold addr_eval.
  unfold S_RF. unfold S_Mem.
  sep on.
  sep lift 4 in H3.
  assert (forall M, INV1 M -> (Lock_INV_H l *^* mTrue) M).
  intros. apply INV1_lockInv1; auto.
  sep weaker_subst in H3 with (Lock_INV_H l *^* mTrue).
  sep on. sep lift 0 in H3.
  assert (exists b, exists w, 
          (M ||- (l + L_AVAIL |->w b) *^* (l + L_WAITING |->w w) *^* (choice (b=0) Aemp (Gamma l)) *^* 
           (mTrue *^*
            (INV0 *^* mP *^*
             (R SP |->w bp *^* R SP + 2 |->w pc *^* R SP + 4 |->w l *^*
              stkSpacePrec (R SP) (_STK_SIZE_l_acquire - 2))))
          )
          /\ Delta w = SomeT (Gamma l)).
  apply decomp_ex_Lock_INV_H; auto.
  InvertAll.
  rewrite H0. sep lift 0 in H10.
  sep trivial.
Qed.

Lemma l_acquire_step4:
   forall f sp l pc bp mP S S',
     l_acquire3_Prd_aux sp l pc bp mP S
     -> nextS f (comml (_movwld (A_reg L_AVAIL BX) DX)) S S'
     -> l_acquire4_Prd_aux sp l pc bp mP S'.
Proof.
  intros.
  rename H0 into Htrans.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion Htrans).
  subst M'.
  destruct H1 as [[x [Hmemx HR']] [Hfl' Hisr']].
  unfold l_acquire4_Prd_aux.
  subst R' fl' isr'.
  reg simpl.
  unfold l_acquire3_Prd_aux in H.
  destruct H as [HRbx [bx0 Hpre]].
  unfold l_acquire2_Prd_aux in Hpre.
  reg simpl.
  unfold l_acquire1_Prd_aux in Hpre.
  reg simpl.
  destruct Hpre as [HRbp [HRsp [Htmp [Hmem [Hlock [Hfl Hisr]]]]]].
  split_l. split_l.
  split. rewrite HRsp in Hmem. auto.
  split; auto.
  exists x.
  split; auto.
  unfold addr_eval in Hmemx.
  rewrite HRbx in Hmemx. auto.
Qed.


Lemma l_acquire3_Prd_IF_false:
   forall S : State, l_acquire3_Prd S -> S_Flag S _IF = false.
Proof.
    unfold l_acquire3_Prd.
    intros.
    induction S as (((M, R), fl), isr).
    unfold l_acquire3_Prd_aux in H.
    unfold l_acquire2_Prd_aux in H.
    unfold l_acquire1_Prd_aux in H.
    InvertAll. auto.
Qed.

Lemma l_acquire3_Grt_Itr_Stk: Grt_Itr_Stk l_acquire3_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  unfold l_acquire3_grt.
  unfold leadstoPQ2. intros.
  unfold l_acquire3_Prd_aux in H4.
  induction S as (((M, R), fl), isr).
  UnfoldStateAll.
  unfold l_acquire2_Prd_aux in H4.
  unfold l_acquire1_Prd_aux in H4.
  InvertAll.
  unfold itr_enabled in H0.
  inversion_clear H0.
  UnfoldStateAll.
  rewrite H11 in H10. inversion H10.
Qed.

Lemma wff_cmd_l_acquire4:
   NH_WFcmd
     (SCAP_consCT (l_acquire + 1 + 1 + 1 + 1)
        (Ast l_acquire4_Prd l_acquire4_grt 1) l_acquire_psi)
     (l_acquire + 1 + 1 + 1) (Ast l_acquire3_Prd l_acquire3_grt 1)
     (comml (_movwld (A_reg L_AVAIL BX) DX)).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists l_acquire4_Prd. exists l_acquire4_grt.
  split. auto.

  (* enable *)
  generalize (l_acquire + 1 + 1 + 1). intro f.
  assert (forall sp l pc bp mP, enable (l_acquire3_Prd_aux sp l pc bp mP) 
                                    (nextS f (comml (_movwld (A_reg L_AVAIL BX) DX)))).
  intros.
  apply enable_movwld.
  intros.
  apply l_acquire3_Prd_1 with sp l pc bp mP; auto.

  split.
  unfold enable in *.
  intros.
  unfold l_acquire3_Prd in H0.
  InvertAll.
  apply H with x x0 x1 x2 x3; auto.
  split.
  unfold pimp.
  unfold linkPG.
  intros.
  InvertAll.
  unfold l_acquire3_Prd in H0.
  InvertAll.
  unfold l_acquire4_Prd.
  exists x0. exists x1. exists x2. exists x3. exists x4.
  apply l_acquire_step4 with f x; auto.
  unfold gimp.
  unfold conjPG. unfold conjG.
  intros.
  InvertAll.
  unfold l_acquire3_grt.
  unfold leadstoPQ2.
  intros.
  unfold l_acquire4_grt in H2.
  unfold leadstoPQ2 in H2.
  apply H2.
  apply l_acquire_step4 with f S; auto.

  split. 
    apply IF_false_Prd_Itr_Stk.
      apply l_acquire3_Prd_IF_false.
    apply l_acquire3_Grt_Itr_Stk.
Qed.

Definition l_acquire5_Prd_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        exists fl', l_acquire4_Prd_aux sp l pc bp mP (M, R, fl', isr) /\
                    ((R DX = 0) /\ (fl = Set_Flag fl' _ZF)
                         \/ (R DX <> 0) /\ (fl = Clear_Flag fl' _ZF))
  end.

Definition l_acquire5_Prd (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, l_acquire5_Prd_aux sp l pc bp mP S.

Definition l_acquire5_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (l_acquire5_Prd_aux sp l pc bp) (l_acquire_Post_aux (sp +2) l pc bp) S S'.


Lemma l_acquire_step5:
   forall f sp l pc bp mP S S',
     l_acquire4_Prd_aux sp l pc bp mP S
     -> nextS f (comml (_cmpw (Opr_imm 0) DX)) S S'
     -> l_acquire5_Prd_aux sp l pc bp mP S'.
Proof.
  intros.
  rename H into Hpre. rename H0 into Htrans.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion Htrans).
  InvertAll.
  subst M'. subst R' isr'.
  unfold l_acquire5_Prd_aux.
  exists fl.
  unfold opr_eval in H0.
  split; auto.
  inversion_clear H0.
  left. 
  inversion H. split; auto.
  inversion H. right. split; auto.
Qed.

Lemma l_acquire4_Prd_IF_false:
   forall S : State, l_acquire4_Prd S -> S_Flag S _IF = false.
Proof.
   intros. unfold l_acquire4_Prd in H.
   InvertAll.
   unfold l_acquire4_Prd_aux in H0.
   induction S as (((M, R), fl), isr). UnfoldStateAll.
   InvertAll. auto.
Qed.

Lemma l_acquire4_Grt_Itr_Stk: Grt_Itr_Stk l_acquire4_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  unfold l_acquire4_grt.
  unfold leadstoPQ2.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  UnfoldStateAll.
  intros.
  unfold l_acquire4_Prd_aux in H4.
  InvertAll.
  unfold itr_enabled in H0.
  UnfoldStateAll.
  inversion_clear H0.
  rewrite H10 in H9. inversion H9.
Qed.


Lemma wff_cmd_l_acquire5:
   NH_WFcmd
     (SCAP_consCT (l_acquire + 1 + 1 + 1 + 1 + 1)
        (Ast l_acquire5_Prd l_acquire5_grt 1) l_acquire_psi)
     (l_acquire + 1 + 1 + 1 + 1) (Ast l_acquire4_Prd l_acquire4_grt 1)
     (comml (_cmpw (Opr_imm 0) DX)).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists l_acquire5_Prd. exists l_acquire5_grt.
  split. auto.

  generalize (l_acquire + 1 + 1 + 1 + 1). intro f.
  split. apply trivial_enable_cmpw.

  split.
  unfold pimp. unfold linkPG.
  intros S1 Htmp.
  destruct Htmp as [S0 [Hpre Htrans]].
  unfold l_acquire5_Prd.
  unfold l_acquire4_Prd in Hpre.
  destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
  exists sp. exists l. exists pc. exists bp. exists mP.
  apply l_acquire_step5 with f S0. auto. auto.
  unfold gimp.
  unfold conjG. unfold conjPG.
  intros.
  destruct H as [S0 [Htmp Hgrt5]].
  destruct Htmp as [Hpre Htrans].
  unfold l_acquire4_grt.
  unfold leadstoPQ2.
  intros.
  unfold l_acquire5_grt in Hgrt5.
  unfold leadstoPQ2 in Hgrt5.
  apply Hgrt5.
  apply l_acquire_step5 with f S; auto.

  split.
  apply IF_false_Prd_Itr_Stk.
  apply l_acquire4_Prd_IF_false.
  apply l_acquire4_Grt_Itr_Stk.
Qed.


Definition l_acquire6_Prd_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        (l_acquire4_Prd_aux sp l pc bp mP (M, R, fl, isr)) /\ R DX = 0
  end.

Definition l_acquire6_Prd (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, l_acquire6_Prd_aux sp l pc bp mP S.

Definition l_acquire6_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (l_acquire6_Prd_aux sp l pc bp) (l_acquire_Post_aux (sp +2) l pc bp) S S'.


Lemma l_acquire5_Prd_IF_false:
   forall S : State, l_acquire5_Prd S -> S_Flag S _IF = false.
Proof.
  unfold l_acquire5_Prd.
  unfold l_acquire5_Prd_aux.
  unfold l_acquire4_Prd_aux.
  intros. induction S as (((M, R), fl), isr).
  UnfoldStateAll. InvertAll.
  inversion_clear H1. InvertAll.
  rewrite H9. unfold Set_Flag. simpl. auto.
  InvertAll. rewrite H9. unfold Clear_Flag. simpl. auto.
Qed.

Lemma l_acquire5_Grt_Itr_Stk: Grt_Itr_Stk l_acquire5_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  unfold l_acquire5_grt.
  unfold leadstoPQ2.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  UnfoldStateAll.
  intros.
  unfold l_acquire5_Prd_aux in H4.
  unfold l_acquire4_Prd_aux in H4.
  InvertAll.
  unfold itr_enabled in H0.
  UnfoldStateAll.
  inversion_clear H0.
  inversion_clear H6.
  inversion_clear H0.
  rewrite H15 in H11. unfold Set_Flag in H11.
  simpl in H11. rewrite H11 in H10. inversion H10.
  inversion_clear H0.
  rewrite H15 in H11. unfold Clear_Flag in H11.
  simpl in H11. rewrite H11 in H10. inversion H10.
Qed.

Lemma l_acquire_step6_a:
forall sp l pc bp mP S,
  l_acquire5_Prd_aux sp l pc bp mP S
  -> ZF_cl S
  -> l_setacq_Prd_aux sp l pc bp mP S.
Proof.
  intros.
  unfold l_acquire5_Prd_aux in H.
  induction S as (((M, R), fl), isr).
  destruct H as [fl' [Hpre4 Hcases]].
  unfold ZF_cl in H0.
  unfold S_Flag in H0.
  inversion_clear Hcases.
  inversion_clear H.
  rewrite H2 in H0.
  unfold Set_Flag in H0.
  simpl in H0.
  inversion H0.
  inversion_clear H.
  rename H1 into Hdx0. rename H0 into Hzffalse.
  rename H2 into Hfl.
  unfold l_acquire4_Prd_aux in Hpre4.
  destruct Hpre4 as [Hsp [Hbx [Hmem1 [Hmem2 [Hlock [Hfl' Hisr]]]]]].
  unfold l_setacq_Prd_aux.
  split_l. split_l.
  split.
  unfold l_acquire1_MPre in Hmem1.
  auto.
  rewrite Hfl.
  unfold Clear_Flag. simpl.
  split; auto.
  destruct Hmem2 as [avail [Hinmem Hinreg]].
  exists avail. split_l. split_l.
  rewrite <- Hinreg. auto.
Qed.

Lemma l_acquire_step6_b:
forall sp l pc bp mP S,
  l_acquire5_Prd_aux sp l pc bp mP S 
  -> ZF_st S 
  -> l_acquire6_Prd_aux sp l pc bp mP S.
Proof.
  intros.
  unfold l_acquire5_Prd_aux in H.
  induction S as (((M, R), fl), isr).
  destruct H as [fl' [Hpre6 Hcases]].
  unfold ZF_st in H0.
  unfold S_Flag in H0.
  inversion_clear Hcases.
  inversion_clear H.
  rename H1 into Hax0. rename H0 into Hzftrue.
  rename H2 into Hfl.
  unfold l_acquire6_Prd_aux. split; auto.
  unfold l_acquire4_Prd_aux in *.
  InvertAll.
  split_l. split_l. split_l. rewrite Hfl.
  unfold Set_Flag. simpl. split; auto.
  exists x. split; auto.
  inversion_clear H.
  rewrite H2 in H0.
  unfold Clear_Flag in H0. simpl in H0.
  inversion H0.
Qed.

Lemma wff_cmd_l_acquire6:
   NH_WFcmd
     (SCAP_consCT (l_acquire + 1 + 1 + 1 + 1 + 1 + 1)
        (Ast l_acquire6_Prd l_acquire6_grt 1) l_acquire_psi)
     (l_acquire + 1 + 1 + 1 + 1 + 1) (Ast l_acquire5_Prd l_acquire5_grt 1)
     (comml (_jne l_setacq)).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists l_setacq_Prd. exists l_acquire6_Prd. 
  exists l_setacq_grt. exists l_acquire6_grt.
  split. unfold SCAP_lookupCT. unfold SCAP_consCT.
  unfold SPSpec.lookup. unfold ListT.In. right.
  assert (SCAP_lookupCT l_acquire_psi l_setacq (Ast l_setacq_Prd l_setacq_grt 1)).
  apply l_setacq_in_psi. auto.
  split; auto.

  split.
  unfold conjP.
  unfold pimp.
  intros.
  destruct H as [Hpre Hzf].
  unfold l_acquire5_Prd in Hpre.
  destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
  unfold l_setacq_Prd.
  exists sp. exists l. exists pc. exists bp. exists mP.

  apply l_acquire_step6_a; auto.
  split.
  unfold gimp.
  unfold conjPG.
  intros.
  destruct H as [Hpre [Hzf Hgrtdone]].
  unfold l_acquire5_grt.
  unfold leadstoPQ2.
  intros.
  unfold l_setacq_grt in Hgrtdone.
  unfold leadstoPQ2 in Hgrtdone.
  apply Hgrtdone.
  apply l_acquire_step6_a; auto.
  unfold conjP.
  split.
  unfold pimp.
  intros.
  destruct H as [Hpre Hzf].
  unfold l_acquire5_Prd in Hpre.
  destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
  unfold l_acquire6_Prd.
  exists sp. exists l. exists pc. exists bp. exists mP.


  apply l_acquire_step6_b; auto.
  unfold gimp.
  unfold conjPG.
  intros.
  destruct H as [Hpre [Hzf Hgrtdone]].
  unfold l_acquire5_grt.
  unfold leadstoPQ2.
  intros.
  unfold l_acquire6_grt in Hgrtdone.
  unfold leadstoPQ2 in Hgrtdone.
  apply Hgrtdone.
  apply l_acquire_step6_b; auto.


  split.
  apply IF_false_Prd_Itr_Stk.
  apply l_acquire5_Prd_IF_false.
  apply l_acquire5_Grt_Itr_Stk.
Qed.



Definition l_acquire7_MPre (sp l pc bp wid : Label) :=
  (sp |->w wid) *^* (sp +2 |->w bp) *^* (sp + 4|->w pc) *^* (sp + 6 |->w l) 
  *^* stkSpacePrec sp (_STK_SIZE_l_acquire - 4)
       *^* INV1 *^* INV0.

Definition l_acquire7_Prd_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R SP = sp
        /\ (exists w, (l_acquire7_MPre sp l pc bp w *^* mP) M 
             /\ ((sp |->w w) *^* (l + L_WAITING |->w w) *^* (l + L_AVAIL |->w 0) *^* mTrue) M)
        /\ In l lockSet_H
        /\ fl _IF = false
        /\ isr = false
  end.

Definition l_acquire7_Prd (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, l_acquire7_Prd_aux sp l pc bp mP S.

Definition l_acquire7_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (l_acquire7_Prd_aux sp l pc bp) (l_acquire_Post_aux (sp +4) l pc bp) S S'.


Lemma l_acquire6_Prd_IF_false:
   forall S : State, l_acquire6_Prd S -> S_Flag S _IF = false.
Proof.
  unfold l_acquire6_Prd. 
  unfold l_acquire6_Prd_aux. intros.
  induction S as (((M, R), fl), isr).
  UnfoldStateAll. InvertAll.
  unfold l_acquire4_Prd_aux in H. InvertAll. auto.
Qed.


Lemma l_acquire6_Grt_Itr_Stk: Grt_Itr_Stk l_acquire6_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  unfold l_acquire6_grt.
  unfold leadstoPQ2.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  UnfoldStateAll.
  intros.
  unfold l_acquire6_Prd_aux in H4.
  unfold l_acquire4_Prd_aux in H4.
  InvertAll.
  unfold itr_enabled in H0.
  UnfoldStateAll.
  inversion_clear H0.
  rewrite H11 in H10. inversion H10.
Qed.

Lemma l_acquire7_Prd_IF_false:
   forall S : State, l_acquire7_Prd S -> S_Flag S _IF = false.
Proof.
  unfold l_acquire7_Prd. 
  unfold l_acquire7_Prd_aux. intros.
  induction S as (((M, R), fl), isr).
  UnfoldStateAll. InvertAll. auto.
Qed.


Lemma l_acquire7_Grt_Itr_Stk: Grt_Itr_Stk l_acquire7_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  unfold l_acquire7_grt.
  unfold leadstoPQ2.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  UnfoldStateAll.
  intros.
  unfold l_acquire7_Prd_aux in H4.
  InvertAll.
  unfold itr_enabled in H0.
  UnfoldStateAll.
  inversion_clear H0.
  rewrite H8 in H7. inversion H7.
Qed.



Lemma l_acquire6_Prd_enable:
   forall f sp l pc bp mP,
   enable (l_acquire6_Prd_aux sp l pc bp mP)
     (nextS f (comml (_pushwm (A_reg L_WAITING BX)))).
Proof.
  unfold enable.
  intros.
  unfold l_acquire6_Prd_aux in H.
  induction S as (((M, R), fl), isr).
  destruct H as [Hpre HRdx0].
  unfold l_acquire4_Prd_aux in Hpre.
  destruct Hpre as [HRsp [HRbx [Hmem1 [Hmem2 [Hlock [Hfl Hisr]]]]]].
  unfold l_acquire1_MPre in Hmem1.
  sep on.
  sep lift 4 in Hmem1.
  assert (forall M, INV1 M -> (Lock_INV_H l *^* mTrue) M).
  intros.
  apply INV1_lockInv1; auto.
  sep weaker_subst in Hmem1 with (Lock_INV_H l *^* mTrue).
  sep on. 
  sep lift 0 in Hmem1.
  assert (exists b, exists w, 
          (M ||- (l + L_AVAIL |->w b) *^* (l + L_WAITING |->w w) *^* (choice (b=0) Aemp (Gamma l)) *^* 
              (mTrue *^*
               (INV0 *^* mP *^*
                (sp |->w bp *^* sp + 2 |->w pc *^* sp + 4 |->w l *^*
                 stkSpacePrec sp (_STK_SIZE_l_acquire - 2))))
           )
          /\ Delta w = SomeT (Gamma l)).
  apply decomp_ex_Lock_INV_H; auto.
  InvertAll.
  exists (MemUpdW M (R SP - 2) x0 0, RF_update R SP (R SP - 2), fl, isr).
  apply nextS_pushwm with (addr_eval R (A_reg L_WAITING BX)) x0. auto.
  unfold addr_eval.
  rewrite HRbx.
  sep on.
  sep lift 1 in H2. sep trivial.
  unfold pushW.
  split.
  rewrite HRsp.
  sep on.
  sep lift 9 in H2.
  sep stksplit 2 in H2. 
  sep lift 0 in H2. sep trivial.
  split; auto.
  apply MemUpdW_MemUpd16 with 0; auto.
Qed.

Lemma l_acquire_step7:
forall f sp l pc bp mP S S',
  l_acquire6_Prd_aux sp l pc bp mP S  
  -> nextS f (comml (_pushwm (A_reg L_WAITING BX))) S S'
  -> l_acquire7_Prd_aux (sp - 2) l pc bp mP S'.
Proof.
  intros.
  rename H0 into Htrans.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion Htrans).
  destruct H0 as [w [Hsrc [Hspgt2 [Hmemblk HM']]]].
  destruct H1 as [HR' [Hfl' Hisr']].
  subst R' fl' isr'. clear Htrans.
  unfold l_acquire7_Prd_aux.
  unfold l_acquire6_Prd_aux in H.
  unfold l_acquire4_Prd_aux in H.
  destruct H as [[HRsp [HRbx [Hmem1 [Hmem2 [Hlock [Hfl Hisr]]]]]] HRdx0].
  reg simpl.
  split. rewrite HRsp. auto.
  split; auto.
  exists w.
  split.
  unfold l_acquire7_MPre.
  substH HM' with (MemUpd16_MemUpdW HM').
  destruct HM' as [junk HM'].
  subst M'. rewrite HRsp.
  sep on. sep lift 0.
  sep upd.
  unfold l_acquire1_MPre in Hmem1.
  sep lift 3 in Hmem1.
  sep stksplit in Hmem1.
  assert (_STK_SIZE_l_acquire - 2 - 2 = _STK_SIZE_l_acquire - 4); auto.
  rewrite H in Hmem1.
  rewrite HRsp in Hspgt2.
  assert (sp - 2 + 2 = sp); auto.
  rewrite H0. clear H0.
  assert (sp - 2 + 4 = sp + 2); auto.
  rewrite H0; clear H0.
  assert (sp - 2 + 6 = sp + 4). 
  auto. rewrite H0; clear H0.
  sep matchH Hmem1 with 8.
  substH HM' with (MemUpd16_MemUpdW HM').  
  destruct HM' as [junk HM'].
  subst M'. rewrite HRsp.
  sep on. sep lift 0.
  sep upd.
  unfold l_acquire1_MPre in Hmem1.
  sep lift 4 in Hmem1.
  assert (forall M, INV1 M -> (Lock_INV_H l *^* mTrue) M).
  intros.
  apply INV1_lockInv1; auto.
  sep weaker_subst in Hmem1 with (Lock_INV_H l *^* mTrue).
  sep on. 
  sep lift 0 in Hmem1.
  assert (exists b, exists w, 
          (M ||- (l + L_AVAIL |->w b) *^* (l + L_WAITING |->w w) *^* (choice (b=0) Aemp (Gamma l)) *^* 
              (mTrue *^*
               (INV0 *^* mP *^*
                (sp |->w bp *^* sp + 2 |->w pc *^* sp + 4 |->w l *^*
                 stkSpacePrec sp (_STK_SIZE_l_acquire - 2))))
           )
          /\ Delta w = SomeT (Gamma l)).
  apply decomp_ex_Lock_INV_H; auto.
  InvertAll.
  assert (x = x1).
  sep lift 0 in H2.
  apply  ptoW_leadsToW_eq1 with M (l + L_AVAIL) 
           (l + L_WAITING |->w x0 *^*
            (choice (x = 0) Aemp (Gamma l) *^*
             (mTrue *^*
              (INV0 *^* mP *^*
               (sp |->w bp *^* sp + 2 |->w pc *^* sp + 4 |->w l *^*
                stkSpacePrec sp (_STK_SIZE_l_acquire - 2)))))); auto.
  rewrite <- H5 in H1.
  rewrite HRdx0 in H1. rewrite H1 in H2.
  unfold addr_eval in Hsrc.
  rewrite HRbx in Hsrc.
  assert (x0 = w).
  sep lift 1 in H2.
  apply  ptoW_leadsToW_eq1 with M (l + L_WAITING) 
           (choice (0 = 0) Aemp (Gamma l) *^*
            (mTrue *^*
             (INV0 *^* mP *^*
              (sp |->w bp *^* sp + 2 |->w pc *^* sp + 4 |->w l *^*
               stkSpacePrec sp (_STK_SIZE_l_acquire - 2)))) *^*
            l + L_AVAIL |->w 0); auto.
  rewrite <- H6.
  sep lift 9 in H2.
  sep stksplit in H2.
  sep lift 1 in H2.
  sep match_left H2. auto.
  sep lift 1 in H2. sep match_left H2; auto.
  sep lift 8 in H2. sep match_left H2; auto.
Qed.


Lemma wff_cmd_l_acquire7:
   NH_WFcmd
     (SCAP_consCT (l_acquire + 1 + 1 + 1 + 1 + 1 + 1 + 1)
        (Ast l_acquire7_Prd l_acquire7_grt 1) l_acquire_psi)
     (l_acquire + 1 + 1 + 1 + 1 + 1 + 1) (Ast l_acquire6_Prd l_acquire6_grt 1)
     (comml (_pushwm (A_reg L_WAITING BX))).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists l_acquire7_Prd. exists l_acquire7_grt.
  split. auto.

  generalize (l_acquire + 1 + 1 + 1 + 1 + 1 + 1). intro f.

  assert (forall sp l pc bp mP, enable (l_acquire6_Prd_aux sp l pc bp mP) 
                                    (nextS f (comml (_pushwm (A_reg L_WAITING BX))))).

  apply l_acquire6_Prd_enable.
  split.
  unfold enable in *.
  intros.
  unfold l_acquire6_Prd in H0.
  InvertAll.
  apply H with x x0 x1 x2 x3; auto.

  split.
  unfold pimp.
  unfold linkPG.
  intros.
  InvertAll.
  unfold l_acquire6_Prd in H0.
  InvertAll.
  unfold l_acquire7_Prd.
  mexists5 (x0 - 2) x1 x2 x3 x4.


  apply l_acquire_step7 with f x; auto.

  unfold gimp.
  unfold conjPG. unfold conjG.
  intros.
  InvertAll.
  unfold l_acquire6_grt.
  unfold leadstoPQ2.
  intros.
  unfold l_acquire7_grt in H2.
  unfold leadstoPQ2 in H2.
  assert (l_acquire_Post_aux (sp - 2 + 4) l pc bp mP S').
  apply H2.
  apply l_acquire_step7 with f S; auto.
  induction S as (((M, R), fl), isr).
  induction x as (((M', R'), fl'), isr').
  destruct (nextS_inversion H3).
  InvertAll.
  unfold l_acquire6_Prd_aux in H0.
  unfold l_acquire4_Prd_aux in H0.
  InvertAll.
  reg simpl.
  rewrite <- H0.
  rewrite <- H0 in H4.
  assert (R SP - 2 + 4 = R SP + 2). auto.
  rewrite H18 in H4. auto.

  split.
  apply IF_false_Prd_Itr_Stk.
  apply l_acquire6_Prd_IF_false.
  apply l_acquire6_Grt_Itr_Stk.
Qed.


Definition l_acquire8_MPre (sp l pc bp : Label) :=
    (sp + 2 |->w bp) *^* (sp + 4 |->w pc) *^* (sp + 6 |->w l) *^* stkSpacePrec (sp + 2) (_STK_SIZE_l_acquire - 2)
       *^* INV1 *^* INV0 *^* Gamma l.

Definition l_acquire8_Prd_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R SP = sp
        /\ (l_acquire8_MPre (R SP) l pc bp *^* mP) M
        /\ In l lockSet_H
        /\ fl _IF = false
        /\ isr = false
  end.

Definition l_acquire8_Prd (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, l_acquire8_Prd_aux sp l pc bp mP S.


Definition l_acquire8_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (l_acquire8_Prd_aux sp l pc bp) (l_acquire_Post_aux (sp +4) l pc bp) S S'.


Lemma _STK_SZ_acq_N4_ge2:
  _STK_SIZE_l_acquire - 4 >= 2.
Proof.
  unfold _STK_SIZE_l_acquire.
  omega.
Qed.
Hint Resolve _STK_SZ_acq_N4_ge2.

Lemma _STK_SZ_acq_N4_N2_eqsleep:
  _STK_SIZE_l_acquire - 4 - 2 = _STK_SIZE_sleep.
Proof.
  unfold _STK_SIZE_l_acquire. omega.
Qed.
Hint Resolve _STK_SZ_acq_N4_N2_eqsleep.

Definition L_sleep_closure_acq (sp l pc bp : int) (mP : MPred) :=
  sp + 2 |->w bp *^* sp + 4 |->w pc *^*
                 sp + 6 |->w l *^* mP.

Lemma l_acquire_step8_call2:
  forall f sp l pc bp mP S S', 
    l_acquire7_Prd_aux sp l pc bp mP S
    -> nextS f (comml (_call L_sleep)) S S'
    -> exists w, Ast_sleep_0 w (Gamma l) (sp - 2) (f + 1) 
                    (L_sleep_closure_acq sp l pc bp mP) S'.
Proof.
  intros.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion H0).
  destruct H1 as [Hspgt2 [Hpcspace HM']].
  destruct H2 as [HR' [Hfl' Hisr']].
  substH HM' with (MemUpd16_MemUpdW HM').
  destruct HM' as [junk HM'].
  subst M' R' fl' isr'. clear H0.
  unfold l_acquire7_Prd_aux in H.
  destruct H as [HRsp [Hmem [Hlock [Hfl Hisr]]]].
  destruct Hmem as [w [Hmem1 Hmem2]].
  rewrite HRsp.
  exists w.
  unfold Ast_sleep_0.
  UnfoldStateAll. reg simpl.
  split.
  unfold l_acquire7_MPre in Hmem1.
  rewrite HRsp in Hspgt2.
  sep on.
  sep lift 3. sep upd.
  assert (sp - 2 + 2 = sp). auto.
  rewrite H. clear H.
  sep lift 4 in Hmem1.
  sep stksplit in Hmem1.
  assert (_STK_SIZE_l_acquire - 4 - 2 = _STK_SIZE_sleep); auto.
  rewrite H in Hmem1.
  unfold L_sleep_closure_acq.
  sep matchH Hmem1 with 9.
  split_l.
  split.
  unfold l_acquire7_MPre in Hmem1.
  sep on. sep lift 4 in Hmem1.
  sep stksplit in Hmem1.
  assert (sp - 2 >= _STK_SIZE_l_acquire - 4 - 2).
  apply stkSpacePrec_sp_ge_sz1 with M 
              (sp - 2 |->w_ *^*
               (INV1 *^* (INV0 *^* mP) *^*
                (sp |->w w *^* sp + 2 |->w bp *^* sp + 4 |->w pc *^*
                 sp + 6 |->w l))); auto.
  assert (_STK_SIZE_l_acquire - 4 - 2 = _STK_SIZE_sleep). auto.
  rewrite H0 in H. auto.
  split_l. split_l.
  unfold l_acquire7_MPre in Hmem1.
  sep on.
  sep lift 5 in Hmem1.
  assert (forall M, INV1 M -> (Lock_INV_H l *^* mTrue) M).
  intros.
  apply INV1_lockInv1; auto.
  sep weaker_subst in Hmem1 with (Lock_INV_H l *^* mTrue).
  sep on.
  sep lift 0 in Hmem1.
  assert (exists b, exists w0, 
          (M ||- (l + L_AVAIL |->w b) *^* (l + L_WAITING |->w w0) *^* (choice (b=0) Aemp (Gamma l)) *^* 
              (mTrue *^*
               (INV0 *^* mP *^*
                (sp |->w w *^* sp + 2 |->w bp *^* sp + 4 |->w pc *^*
                 sp + 6 |->w l *^* stkSpacePrec sp (_STK_SIZE_l_acquire - 4))))
           )
          /\ Delta w0 = SomeT (Gamma l)).
  apply decomp_ex_Lock_INV_H; auto.
  destruct H1 as [b [w0 [Hmem' Hdelta]]].
  sep lift 1 in Hmem'.
  sep lift 1 in Hmem2.
  assert (w = w0).
  eapply ptoW_ptoW_eq1; eauto.
  rewrite H1. auto.
Qed.


Lemma l_acquire7_enable:
   forall f sp l pc bp mP,
   enable (l_acquire7_Prd_aux sp l pc bp mP)
     (nextS f (comml (_call L_sleep))).
Proof.
  intros.
  unfold enable.
  intros.
  unfold l_acquire7_Prd_aux in H.
  induction S as (((M, R), fl), isr).
  exists (MemUpdW M (R SP - 2) (f+1) 0, RF_update R SP (R SP - 2), fl, isr).
  apply nextS_call. 
  unfold pushW.
  destruct H as [HRsp [Hmem [Hlock [Hfl Hisr]]]].
  split.
  destruct Hmem as [w [Hmem1 Hmem2]].
  unfold l_acquire7_MPre in Hmem1.
  sep on.
  sep lift 4 in Hmem1.
  sep stksplit 2 in Hmem1.
  rewrite HRsp. sep lift 0 in Hmem1.
  sep trivial.
  split; auto.
  apply  MemUpdW_MemUpd16 with 0; auto.
Qed.

Lemma wff_cmd_l_acquire8:
   NH_WFcmd
     (SCAP_consCT (l_acquire + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
        (Ast l_acquire8_Prd l_acquire8_grt 1) l_acquire_psi)
     (l_acquire + 1 + 1 + 1 + 1 + 1 + 1 + 1) (Ast l_acquire7_Prd l_acquire7_grt 1)
     (comml (_call L_sleep)).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  right.

  split.
  intro. unfold prim_entry in H.
  inversion_clear H. absurd (L_sleep <> _SWITCH); auto.
  inversion_clear H0. absurd (L_sleep <> _BLOCK); auto.
  absurd (L_sleep <> _UNBLK); auto.

  unfold NH_Normal_Call.
  exists L_sleep_0_p. exists l_acquire8_Prd.
  exists L_sleep_0_g. exists l_acquire8_grt.
  split. 
  unfold SCAP_lookupCT. unfold SCAP_consCT.
  unfold SPSpec.lookup. unfold ListT.In. right.
  assert (SCAP_lookupCT l_acquire_psi L_sleep (Ast L_sleep_0_p L_sleep_0_g 1)).
  apply L_sleep_in_psi. auto.
  split. auto.

  generalize (l_acquire + 1 + 1 + 1 + 1 + 1 + 1 + 1). intro f.


  split. 
  unfold enable. unfold l_acquire7_Prd.
  assert (forall f sp l pc bp mP,
   enable (l_acquire7_Prd_aux sp l pc bp mP)
     (nextS f (comml (_call L_sleep)))).
  apply l_acquire7_enable.
  intros.
  InvertAll. unfold enable in H.
  apply H with x x0 x1 x2 x3; auto.

Lemma l_acquire_step8_call1: 
  forall f, gimp (sp_call_pred l_acquire7_Prd f L_sleep >o L_sleep_0_g) fun_frame.
Proof.
  unfold gimp.
  unfold sp_call_pred. unfold conjPG. unfold linkPG.
  intros.
  destruct H as [[S0 [Hpre Htrans]] Hgrt].
  unfold l_acquire7_Prd in Hpre.
  destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
  assert (exists w, Ast_sleep_0 w (Gamma l) (sp - 2) (f + 1) (L_sleep_closure_acq sp l pc bp mP) S).
  apply l_acquire_step8_call2 with S0; auto.
  rename H into HFpre.
  destruct HFpre as [w HFpre].
  unfold L_sleep_0_g in Hgrt.
  assert (Ast_sleep_post w (Gamma l) (sp-2) (f+1) (L_sleep_closure_acq sp l pc bp mP) S' /\ S_RF S' = S_RF S).
  apply Hgrt; auto.
  clear Hgrt.
  unfold fun_frame.
  destruct H as [HPost HRF'].
  split. rewrite HRF'. auto.
  exists (f + 1).
  unfold Ret_Addr.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  UnfoldStateAll.
  split.
  intro.
  apply nextPC_retw with (sp - 2).
  unfold Ast_sleep_0 in HFpre. UnfoldStateAll.
  InvertAll. auto.
  unfold Ast_sleep_0 in HFpre. UnfoldStateAll.
  InvertAll.
  sep on. sep lift 3 in H.
  sep trivial.
  intro.
  unfold Ast_sleep_post in HPost.
  UnfoldStateAll.
  InvertAll.
  unfold Ast_sleep_0 in HFpre.
  UnfoldStateAll. InvertAll.
  apply nextPC_retw with (sp - 2); auto.
  rewrite HRF'; auto.
  sep on.
  sep lift 4 in H.
  sep trivial.
Qed.
  split.
  apply l_acquire_step8_call1.
  split.
  unfold pimp. unfold linkPG. 
  intros.
  destruct H as [S0 [Hpre Htrans]].
  unfold l_acquire7_Prd in Hpre.
  destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
  assert (exists w, Ast_sleep_0 w (Gamma l) (sp - 2) (f + 1) (L_sleep_closure_acq sp l pc bp mP) S).
  apply l_acquire_step8_call2 with S0; auto.
  destruct H as [w HFpre].
  unfold L_sleep_0_p. exists w. exists (Gamma l). exists (sp - 2).
  exists (f + 1).
  induction S as (((M, R), fl), isr).
  unfold Ast_sleep_0 in *. UnfoldStateAll.
  InvertAll.
  split; auto.
  sep on.
  sep matchH H with 5.
Lemma l_acquire_step8_call3:
forall f pc' sp l pc bp mP S S',
  l_acquire7_Prd_aux sp l pc bp mP S
  -> wrap_fun L_sleep_0_g f L_sleep pc' S S'
  -> l_acquire8_Prd_aux sp l pc bp mP S'.
Proof.
  intros.
  rename H into Hpre. rename H0 into Htrans.
  unfold wrap_fun in Htrans.
  unfold conjG in Htrans.
  destruct Htrans as [S2 [[S1 [Hcall Hgrt]] Hretw]].
  assert (sp >= 2).
  unfold l_acquire7_Prd_aux in Hpre.
  induction S as (((M, R), fl), isr).
  induction S1 as (((M', R'), fl'), isr').
  InvertAll.
  destruct (nextS_inversion Hcall).
  InvertAll.
  rewrite H in H8; auto.
  rename H into Hspge2.
  assert (exists w, Ast_sleep_0 w (Gamma l) (sp - 2) (f + 1) (L_sleep_closure_acq sp l pc bp mP) S1).
  apply l_acquire_step8_call2 with S; auto.
  destruct H as [w HFpre].
  unfold L_sleep_0_g in Hgrt.
  assert (Ast_sleep_post w (Gamma l) (sp-2) (f+1) (L_sleep_closure_acq sp l pc bp mP) S2 /\ S_RF S2 = S_RF S1).
  apply Hgrt; auto.
  clear Hgrt Hcall.
  destruct H as [Hpre' HR].
  induction S2 as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion Hretw).
  subst M'.
  destruct H0 as [[pc'' [HMpc HR']] [Hfl' Hisr']].
  subst fl' isr' R'. clear Hretw. clear HMpc.
  unfold Ast_sleep_post in Hpre'.
  UnfoldStateAll.
  destruct Hpre' as [Hmem [Hfl Hisr]].
  unfold l_acquire8_Prd_aux.
  reg simpl.
  assert (R SP + 2 = sp).
  unfold Ast_sleep_0 in HFpre.
  induction S1 as (((M2, R2), fl2), isr2).
  rewrite HR.
  UnfoldStateAll.
  InvertAll.
  rewrite H1. auto.
  split_l. split. rewrite H. clear H.
  unfold l_acquire8_MPre.
  unfold L_sleep_closure_acq in Hmem.
  assert (sp - 2 + 2 = sp). auto.
  rewrite H in Hmem.
  sep on.
  sep lift 3.
  sep stksplit.
  assert (sp + 2 - 2 = sp). auto.
  rewrite H0.
  sep stksplit.
  assert (_STK_SIZE_l_acquire - 2 - 2 - 2 = 18 + 2 + _BLK_Stk). auto with arith.
  rewrite H1.
  sep matchH Hmem with 10.
  split; auto.
  unfold l_acquire7_Prd_aux in Hpre.
  induction S as (((Mx, Rx), flx), isrx).
  InvertAll; auto.
Qed.
  split.
  unfold linkPG. unfold pimp.
  intros.
  InvertAll.
  unfold l_acquire7_Prd in H.
  InvertAll.
  unfold l_acquire8_Prd.
  mexists5 x0 x1 x2 x3 x4.
  apply l_acquire_step8_call3 with f pc' x; auto.

  intros.
  unfold gimp.
  unfold conjPG. unfold conjG.
  intros.
  destruct H as [S'' [[Hpre Hfun] Htrans5]].
  unfold l_acquire7_grt.
  unfold leadstoPQ2.
  intros.
  unfold l_acquire8_grt in Htrans5.
  unfold leadstoPQ2 in Htrans5.
  apply Htrans5.
  apply l_acquire_step8_call3 with f pc' S; auto.


  split.
  apply IF_false_Prd_Itr_Stk.
  apply l_acquire7_Prd_IF_false.
  apply l_acquire7_Grt_Itr_Stk.
Qed.


(* _addw (Opr_imm 2) SP *)
Definition l_acquire9_Prd_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R SP = sp
        /\ sp >= 2
        /\ l_acquire8_Prd_aux (sp - 2) l pc bp mP (M, RF_update R SP (sp - 2), fl, isr)
  end.

Definition l_acquire9_Prd (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, l_acquire9_Prd_aux sp l pc bp mP S.

Definition l_acquire9_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (l_acquire9_Prd_aux sp l pc bp) (l_acquire_Post_aux (sp +2) l pc bp) S S'.


Lemma l_acquire_step9:
   forall f sp l pc bp mP S S',
     l_acquire8_Prd_aux sp l pc bp mP S 
     -> nextS f (comml (_addw (Opr_imm 2) SP)) S S'
     -> l_acquire9_Prd_aux (sp + 2) l pc bp mP S'.
Proof.
  intros.
  rename H0 into Htrans.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion Htrans).
  subst M'.
  destruct H1 as [HR' [Hfl' Hisr']].
  unfold opr_eval in HR'.
  subst R' isr'.
  clear Htrans.
  unfold l_acquire8_Prd_aux in H.
  destruct H as [HRsp [Hmem [Hlock [Hfl Hisr]]]].
  unfold l_acquire9_Prd_aux.
  reg simpl.
  rewrite HRsp.
  split; auto with arith.
  split; auto with arith.
  assert (sp + 2 - 2 = sp). auto.
  rewrite H.
  rewrite <- HRsp.
  assert (RF_update R SP (R SP) = R).
  apply RF_update_prev.
  rewrite H0.
  unfold l_acquire8_Prd_aux.
  split_l. split; auto.
  split_l. split_r.
  unfold trash_fl in Hfl'.
  inversion Hfl'.
  subst fl'. unfold Set_Flag. simpl. auto.
  subst fl'. unfold Clear_Flag. simpl. auto.
Qed.

Lemma l_acquire8_Prd_IF_false:
   forall S : State, l_acquire8_Prd S -> S_Flag S _IF = false.
Proof.
  unfold l_acquire8_Prd.
  unfold l_acquire8_Prd_aux. intros.
  induction S as (((M, R), fl), isr).
  UnfoldStateAll. InvertAll.
  auto.
Qed.

Lemma l_acquire8_Grt_Itr_Stk: Grt_Itr_Stk l_acquire8_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  unfold l_acquire8_grt.
  unfold leadstoPQ2.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  UnfoldStateAll.
  intros.
  unfold l_acquire8_Prd_aux in H4.
  InvertAll.
  unfold itr_enabled in H0.
  UnfoldStateAll.
  inversion_clear H0.
  rewrite H8 in H7. inversion H7.
Qed.

Lemma wff_cmd_l_acquire9:
   NH_WFcmd
     (SCAP_consCT (l_acquire + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
        (Ast l_acquire9_Prd l_acquire9_grt 1) l_acquire_psi)
     (l_acquire + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1) (Ast l_acquire8_Prd l_acquire8_grt 1)
     (comml (_addw (Opr_imm 2) SP)).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists l_acquire9_Prd. exists l_acquire9_grt.
  split. auto.

  generalize (l_acquire + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1).
  intro f.
  split. apply trivial_enable_addw.

  split.
  unfold pimp. unfold linkPG.
  intros. 
  destruct H as [S0 [Hpre Htrans]].
  unfold l_acquire8_Prd in Hpre.
  destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
  unfold l_acquire9_Prd.
  mexists5 (sp + 2) l pc bp mP.
  apply l_acquire_step9 with f S0; auto.

  unfold gimp. unfold conjG. unfold conjPG.
  intros.
  destruct H as [S'' [[Hpre Htrans] Hgrt7]].
  unfold l_acquire8_grt.
  unfold l_acquire9_grt in Hgrt7.
  unfold leadstoPQ2 in *.
  intros.
  assert (sp + 2 + 2 = sp + 4). auto.
  rewrite <- H0.
  apply Hgrt7.
  apply l_acquire_step9 with f S; auto.

  split.
  apply IF_false_Prd_Itr_Stk.
  apply l_acquire8_Prd_IF_false.
  apply l_acquire8_Grt_Itr_Stk.
Qed.



Lemma l_acquire9_Prd_IF_false:
   forall S : State, l_acquire9_Prd S -> S_Flag S _IF = false.
Proof.
  unfold l_acquire9_Prd.
  unfold l_acquire9_Prd_aux. unfold l_acquire8_Prd_aux.
  intros. induction S as (((M, R), fl), isr).
  UnfoldStateAll. InvertAll. auto.
Qed.

Lemma l_acquire9_Grt_Itr_Stk: Grt_Itr_Stk l_acquire9_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  unfold l_acquire9_grt.
  unfold leadstoPQ2.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  UnfoldStateAll.
  intros.
  unfold l_acquire9_Prd_aux in H4.
  unfold l_acquire8_Prd_aux in H4.
  InvertAll.
  unfold itr_enabled in H0.
  UnfoldStateAll.
  inversion_clear H0.
  rewrite H10 in H9. inversion H9.
Qed.

Definition l_acquire_Post (S : State) : Prop :=
  exists sp, exists l, exists pc, exists bp, exists mP, l_acquire_Post_aux sp l pc bp mP S.

Definition l_acquire_last_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (l_acquire_Post_aux sp l pc bp) (l_acquire_Post_aux sp l pc bp) S S'.


Lemma wff_cmd_l_acquirea:
   NH_WFcmd
     (SCAP_consCT (l_acquire + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
        (Ast l_acquire_Post l_acquire_last_grt 1) l_acquire_psi)
     (l_acquire + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1) (Ast l_acquire9_Prd l_acquire9_grt 1)
     (comml (_popw BP)).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists l_acquire_Post. exists l_acquire_last_grt.
  split. auto.

  generalize (l_acquire + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1). intro f.
  assert (forall sp l pc bp mP, enable (l_acquire9_Prd_aux sp l pc bp mP) (nextS f (comml (_popw BP)))).
  intros.
  apply enable_popw.
  intros. assert ((S_RF S SP |~>w bp) (S_Mem S)).
Lemma l_acquire9_Prd_1:
  forall sp l pc bp mP S, l_acquire9_Prd_aux sp l pc bp mP S -> (S_RF S SP |~>w bp) (S_Mem S).
Proof.
  intros.
  unfold l_acquire9_Prd_aux in H.
  induction S as (((M, R), fl), isr).
  InvertAll.
  UnfoldStateAll.
  unfold l_acquire8_Prd_aux in H2.
  unfold l_acquire8_MPre in H2. reg simpl.
  InvertAll.
  sep on.
  assert (sp - 2 + 2 = sp). auto.
  rewrite H5 in H2.
  sep lift 0 in H2. rewrite H0.
  sep trivial.
Qed.
  apply l_acquire9_Prd_1 with sp l pc mP. auto.
  apply leadsToW_leadsToxW with bp; auto.
  split.
  unfold enable in *.
  intros.
  unfold l_acquire9_Prd in H0.
  InvertAll.
  apply H with x x0 x1 x2 x3; auto.

Lemma l_acquire_stepa:
forall f sp l pc bp mP S S',
  l_acquire9_Prd_aux sp l pc bp mP S  
  -> nextS f (comml (_popw BP)) S S'
  -> l_acquire_Post_aux (sp + 2) l pc bp mP S'.
Proof.
  intros.
  rename H0 into Htrans.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion Htrans).
  subst M'.
  destruct H1 as [[x [Hmemx HR']] [Hfl' Hisr']].
  assert (x = bp).
  assert ((S_RF (M, R, fl, isr) SP |~>w bp) (S_Mem (M, R, fl, isr))).
  apply l_acquire9_Prd_1 with sp l pc mP; auto.
  UnfoldStateAll.
  apply leadsToW_unique with (R SP) M; auto.
  subst x.
  subst R' fl' isr'.
  unfold l_acquire_Post_aux.
  reg simpl.
  clear Htrans.
  unfold l_acquire9_Prd_aux in H.
  InvertAll.
  split. rewrite H0. auto.
  split_l.
  unfold l_acquire8_Prd_aux in H2.
  reg simpl.
  InvertAll.
  split; auto.
  unfold l_acquire8_MPre in H2.
  unfold l_acquire_MPost.
  unfold l_acquire_MPre.
  sep on.
  sep lift 2.
  sep stksplit.
  assert (sp + 2 - 2 = sp). auto.
  rewrite H5.
  assert (sp + 2 + 2 = sp + 4). auto.
  rewrite H7.
  assert (sp - 2 + 2 = sp); auto.
  rewrite H8 in H2.
  assert (sp - 2 + 4 = sp + 2). auto.
  rewrite H9 in H2.
  assert (sp - 2 + 6 = sp + 4). auto.
  rewrite H10 in H2.
  sep matchH H2 with 8.
Qed.
  
  split.
  unfold pimp.
  unfold linkPG.
  intros.
  InvertAll.
  unfold l_acquire9_Prd in H0.
  InvertAll.
  unfold l_acquire_Post.
  exists (x0 + 2). exists x1. exists x2. exists x3. exists x4.
  apply l_acquire_stepa with f x; auto.

  unfold gimp.
  unfold conjPG. unfold conjG.
  intros.
  InvertAll.
  unfold l_acquire9_grt.
  unfold leadstoPQ2.
  intros.
  unfold l_acquire_last_grt in H2.
  unfold leadstoPQ2 in H2.
  apply H2.
  apply l_acquire_stepa with f S; auto.

  split.
  apply IF_false_Prd_Itr_Stk.
  apply l_acquire9_Prd_IF_false.
  apply l_acquire9_Grt_Itr_Stk.
Qed.


Lemma wff_cmd_l_acquireb:
   NH_WFiseq l_acquire_psi (l_acquire + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
     (Ast l_acquire_Post l_acquire_last_grt 1) (sseq _retw).
Proof.
  apply  nh_isq_end.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  split; auto.
  intros.
  unfold l_acquire_last_grt.
  intros.
  unfold leadstoPQ2. auto.
  split.
  apply IF_false_Prd_Itr_Stk.
Lemma l_acquire_Post_IF_false:
   forall S : State, l_acquire_Post S -> S_Flag S _IF = false.
Proof.
  unfold l_acquire_Post.
  unfold l_acquire_Post_aux. intros.
  induction S as (((M, R), fl), isr). UnfoldStateAll.
  InvertAll. auto.
Qed.

Lemma l_acquire_last_Grt_Itr_Stk : Grt_Itr_Stk l_acquire_last_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  unfold l_acquire_last_grt.
  unfold leadstoPQ2.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  UnfoldStateAll.
  intros.
  unfold l_acquire_Post_aux in H4.
  InvertAll.
  unfold itr_enabled in H0.
  UnfoldStateAll.
  inversion_clear H0.
  rewrite H9 in H8. inversion H8.
Qed.

  apply l_acquire_Post_IF_false.
  apply l_acquire_last_Grt_Itr_Stk.
Qed.


Lemma wff_l_acquire:
  NH_WFiseq l_acquire_psi l_acquire (Ast l_acquire_Prd l_acquire_grt 1) I_l_acquire.
Proof.
  unfold I_l_acquire.
  apply nh_isq_seq with l_acquire1_Prd l_acquire1_grt.
  apply wff_cmd_l_acquire1.
  apply nh_isq_seq with l_acquire2_Prd l_acquire2_grt.
  apply wff_cmd_l_acquire2.
  apply nh_isq_seq with l_acquire3_Prd l_acquire3_grt.
  apply wff_cmd_l_acquire3.
  apply nh_isq_seq with l_acquire4_Prd l_acquire4_grt.
  apply wff_cmd_l_acquire4.
  apply nh_isq_seq with l_acquire5_Prd l_acquire5_grt.
  apply wff_cmd_l_acquire5.
  apply nh_isq_seq with l_acquire6_Prd l_acquire6_grt.
  apply wff_cmd_l_acquire6.
  apply nh_isq_seq with l_acquire7_Prd l_acquire7_grt.
  apply wff_cmd_l_acquire7.
  apply nh_isq_seq with l_acquire8_Prd l_acquire8_grt.
  apply wff_cmd_l_acquire8.
  apply nh_isq_seq with l_acquire9_Prd l_acquire9_grt.
  apply wff_cmd_l_acquire9.
  apply nh_isq_seq with l_acquire_Post l_acquire_last_grt.
  apply wff_cmd_l_acquirea.
  apply wff_cmd_l_acquireb.
Qed.


Definition l_setacq1_MPre (sp l pc bp : Label) :=
    (sp |->w bp) *^* (sp + 2 |->w pc) *^* (sp + 4 |->w l) *^* stkSpacePrec sp (_STK_SIZE_l_acquire - 2)
       *^* INV1 *^* INV0 *^* Gamma l.

Definition l_setacq1_Prd_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R SP = sp
        /\ (l_setacq1_MPre sp l pc bp *^* mP) M
        /\ In l lockSet_H
        /\ fl _IF = false
        /\ isr = false
  end.

Definition l_setacq1_Prd (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, l_setacq1_Prd_aux sp l pc bp mP S.

Definition l_setacq1_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (l_setacq1_Prd_aux sp l pc bp) (l_acquire_Post_aux (sp +2) l pc bp) S S'.

Definition l_setacq_psi :=  SCAP_emptCT.

Lemma wff_cmd_l_setacq1:
   NH_WFcmd
     (SCAP_consCT (l_setacq + 1)
        (Ast l_setacq1_Prd l_setacq1_grt 1) l_setacq_psi)
     l_setacq
     (Ast l_setacq_Prd l_setacq_grt 1)
     (comml (_movwst (Opr_imm 0) (A_reg L_AVAIL BX))).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists l_setacq1_Prd. exists l_setacq1_grt.
  split. auto.

  generalize l_setacq. intro f.
  assert (forall sp l pc bp mP, enable (l_setacq_Prd_aux sp l pc bp mP) (nextS f (comml (_movwst (Opr_imm 0) (A_reg L_AVAIL BX))))).
  intros.
  apply enable_movwst.
  intros.

Lemma l_setacq_Prd_1:
forall sp l pc bp mP S,
   l_setacq_Prd_aux sp l pc bp mP S ->
   (addr_eval (S_RF S) (A_reg L_AVAIL BX) |~>w_) (S_Mem S).
Proof.
  intros.
  unfold addr_eval.
  unfold l_setacq_Prd_aux in H.
  induction S as (((M, R), fl), isr).
  InvertAll.
  UnfoldStateAll.
  rewrite H. reg simpl.
  sep on.
  sep lift 4 in H1.
  assert (forall M, INV1 M -> (Lock_INV_H l *^* mTrue) M).
  intros.
  apply INV1_lockInv1; auto.
  sep weaker_subst in H1 with (Lock_INV_H l *^* mTrue).
  sep on.
  sep lift 0 in H1.
  assert (exists b, exists w,
    (M ||- (l + L_AVAIL |->w b) *^* (l + L_WAITING |->w w) *^* (choice (b=0) Aemp (Gamma l)) *^*
           (mTrue *^*
            (INV0 *^* mP *^*
             (sp |->w bp *^* sp + 2 |->w pc *^* sp + 4 |->w l *^*
              stkSpacePrec sp (_STK_SIZE_l_acquire - 2))))
    )           
    /\ Delta w = SomeT (Gamma l)).
  apply decomp_ex_Lock_INV_H; auto.
  clear H1. InvertAll.
  sep lift 0 in H1.
  sep trivial.
Qed.


  apply l_setacq_Prd_1 with sp l pc bp mP. auto.
  split.
  unfold enable in *.
  intros.
  unfold l_setacq_Prd in H0.
  InvertAll.
  apply H with x x0 x1 x2 x3; auto.
  split.
  unfold pimp. unfold linkPG.
  intros.
  InvertAll.
  unfold l_setacq1_Prd.
  unfold l_setacq_Prd in H0.
  InvertAll.
  exists x0. exists x1. exists x2. exists x3. exists x4.

Lemma l_setacq_step1:
forall f sp l pc bp mP S S',
  l_setacq_Prd_aux sp l pc bp mP S
  -> nextS f (comml (_movwst (Opr_imm 0) (A_reg L_AVAIL BX))) S S'
  -> l_setacq1_Prd_aux sp l pc bp mP S'.
Proof.
  intros.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion H0).
  clear H0.
  InvertAll.
  subst R' fl' isr'.
  rename H3 into Htargetx.
  rename H5 into Hnewmem.
  unfold addr_eval in *.
  unfold opr_eval in *.
  substH Hnewmem with (MemUpd16_MemUpdW Hnewmem).
  destruct Hnewmem as [junk HM'].
  unfold l_setacq1_Prd_aux.
  unfold l_setacq_Prd_aux in H.
  destruct H as [HRsp [HRbx [Hmem1 [Hmem2 [Hlock [Hfl Hisr]]]]]].
  split_l.
  split; auto.
  unfold l_setacq1_MPre.
  sep on.
  assert (exists m, leqvM INV1 (Lock_INV_H l *^* m)).
  apply INV1_decom_l; auto.
  destruct H as [mP' Hleqv].
  sep lift 4.
  sep leqv_subst with (Lock_INV_H l *^* mP').
  sep lift 4 in Hmem1.
  sep leqv_subst in Hmem1 with (Lock_INV_H l *^* mP').
  sep on.

  sep lift 0.
  apply decomp_ex_Lock_INV_H1.
  exists 0.
  sep lift 0 in Hmem1.
  assert (exists b, exists w, 
          (M ||- (l + L_AVAIL |->w b) *^* (l + L_WAITING |->w w) *^* (choice (b=0) Aemp (Gamma l)) *^*
              (mP' *^*
               (INV0 *^* mP *^*
                (sp |->w bp *^* sp + 2 |->w pc *^* sp + 4 |->w l *^*
                 stkSpacePrec sp (_STK_SIZE_l_acquire - 2))))
          )
          /\ Delta w = SomeT (Gamma l)).
  apply decomp_ex_Lock_INV_H. auto.
  clear Hmem1. destruct H0 as [b0 [w0 [Hmem1 Hdelta]]].
  exists w0.
  split; auto.
  subst M'. rewrite HRbx.
  sep lift 0.
  sep upd.
  destruct Hmem2 as [avail [HMavail [HRdx Havailne0]]].
  assert (b0 = avail).
  sep lift 0 in Hmem1.
  apply ptoW_leadsToW_eq1 with M (l + L_AVAIL) 
              (l + L_WAITING |->w w0 *^*
               (choice (b0 = 0) Aemp (Gamma l) *^*
                (mP' *^*
                 (INV0 *^* mP *^*
                  (sp |->w bp *^* sp + 2 |->w pc *^* sp + 4 |->w l *^*
                   stkSpacePrec sp (_STK_SIZE_l_acquire - 2)))))); auto.
  rewrite H0 in Hmem1. clear H0.
  sep matchH Hmem1 with 11.
  unfold choice in *.
  inversion_clear Hmem1.
  inversion_clear H0. contradiction.
  inversion_clear H0. exists MemMod.emp. exists h.
  split; auto with PMap.
  split; auto.
  left.
  unfold Aemp.
  auto.
Qed.

  apply l_setacq_step1 with f x; auto.

  unfold gimp.
  unfold conjPG. unfold conjG.
  intros.
  InvertAll.
  unfold l_setacq_grt.
  unfold leadstoPQ2.
  intros.
  unfold l_setacq1_grt in H2.
  unfold leadstoPQ2 in H2.
  apply H2.
  apply l_setacq_step1 with f S; auto.

  split.
  apply IF_false_Prd_Itr_Stk.
Lemma l_setacq_Prd_IF_false:
   forall S : State, l_setacq_Prd S -> S_Flag S _IF = false.
Proof.
  unfold l_setacq_Prd.
  unfold l_setacq_Prd_aux. intros.
  induction S as (((M, R), fl), isr). UnfoldStateAll.
  InvertAll. auto.
Qed.

Lemma l_setacq_Grt_Itr_Stk : Grt_Itr_Stk l_setacq_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  unfold l_setacq_grt.
  unfold leadstoPQ2.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  UnfoldStateAll.
  intros.
  unfold l_setacq_Prd_aux in H4.
  InvertAll.
  unfold itr_enabled in H0.
  UnfoldStateAll.
  inversion_clear H0.
  rewrite H12 in H9. inversion H9.
Qed.

  apply l_setacq_Prd_IF_false.
  apply l_setacq_Grt_Itr_Stk.
Qed.


Lemma l_setacq1_Prd_IF_false:
   forall S : State, l_setacq1_Prd S -> S_Flag S _IF = false.
Proof.
  unfold l_setacq1_Prd.
  unfold l_setacq1_Prd_aux. intros.
  induction S as (((M, R), fl), isr). UnfoldStateAll.
  InvertAll. auto.
Qed.

Lemma l_setacq1_Grt_Itr_Stk : Grt_Itr_Stk l_setacq1_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  unfold l_setacq1_grt.
  unfold leadstoPQ2.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  UnfoldStateAll.
  intros.
  unfold l_setacq1_Prd_aux in H4.
  InvertAll.
  unfold itr_enabled in H0.
  UnfoldStateAll.
  inversion_clear H0.
  rewrite H8 in H7. inversion H7.
Qed.



Lemma wff_cmd_l_setacq2:
   NH_WFcmd
     (SCAP_consCT (l_setacq + 1 + 1)
        (Ast l_acquire_Post l_acquire_last_grt 1) l_setacq_psi)
     (l_setacq + 1)
     (Ast l_setacq1_Prd l_setacq1_grt 1) (comml (_popw BP)).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists l_acquire_Post. exists l_acquire_last_grt.
  split. auto.

  generalize (l_setacq + 1). intro f.
  assert (forall sp l pc bp mP, enable (l_setacq1_Prd_aux sp l pc bp mP) (nextS f (comml (_popw BP)))).
  intros.
  apply enable_popw.
  intros. assert ((S_RF S SP |~>w bp) (S_Mem S)).
Lemma l_setacq1_Prd_1:
  forall sp l pc bp mP S, l_setacq1_Prd_aux sp l pc bp mP S -> (S_RF S SP |~>w bp) (S_Mem S).
Proof.
  intros.
  unfold l_setacq1_Prd_aux in H.
  induction S as (((M, R), fl), isr).
  InvertAll.
  UnfoldStateAll.
  unfold l_setacq1_MPre in H.
  sep on.
  rewrite H0.
  sep lift 0 in H.
  sep trivial.
Qed.
  apply l_setacq1_Prd_1 with sp l pc mP. auto.
  apply leadsToW_leadsToxW with bp; auto.
  split.
  unfold enable in *.
  intros.
  unfold l_setacq1_Prd in H0.
  InvertAll.
  apply H with x x0 x1 x2 x3; auto.

Lemma l_setacq_step2:
forall f sp l pc bp mP S S',
  l_setacq1_Prd_aux sp l pc bp mP S  
  -> nextS f (comml (_popw BP)) S S'
  -> l_acquire_Post_aux (sp + 2) l pc bp mP S'.
Proof.
  intros.
  rename H0 into Htrans.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion Htrans).
  subst M'.
  destruct H1 as [[x [Hmemx HR']] [Hfl' Hisr']].
  assert (x = bp).
  assert ((S_RF (M, R, fl, isr) SP |~>w bp) (S_Mem (M, R, fl, isr))).
  apply l_setacq1_Prd_1 with sp l pc mP; auto.
  UnfoldStateAll.
  apply leadsToW_unique with (R SP) M; auto.
  subst x.
  subst R' fl' isr'.
  unfold l_acquire_Post_aux.
  reg simpl.
  clear Htrans.
  unfold l_setacq1_Prd_aux in H.
  InvertAll.
  split. rewrite H0. auto.
  split_l.
  split; auto.
  unfold l_setacq1_MPre in H.
  unfold l_acquire_MPost.
  unfold l_acquire_MPre.
  sep on.
  sep lift 2.
  sep stksplit.
  assert (sp + 2 - 2 = sp). auto.
  rewrite H3.
  assert (sp + 2 + 2 = sp + 4). auto.
  rewrite H5.
  sep matchH H with 8.
Qed.
  
  split.
  unfold pimp.
  unfold linkPG.
  intros.
  InvertAll.
  unfold l_setacq1_Prd in H0.
  InvertAll.
  unfold l_acquire_Post.
  exists (x0 + 2). exists x1. exists x2. exists x3. exists x4.
  apply l_setacq_step2 with f x; auto.

  unfold gimp.
  unfold conjPG. unfold conjG.
  intros.
  InvertAll.
  unfold l_setacq1_grt.
  unfold leadstoPQ2.
  intros.
  unfold l_acquire_last_grt in H2.
  unfold leadstoPQ2 in H2.
  apply H2.
  apply l_setacq_step2 with f S; auto.

  split.
  apply IF_false_Prd_Itr_Stk.
  apply l_setacq1_Prd_IF_false.
  apply l_setacq1_Grt_Itr_Stk.
Qed.


Lemma wff_cmd_l_setacq3:
   NH_WFiseq l_setacq_psi (l_setacq + 1 + 1)
     (Ast l_acquire_Post l_acquire_last_grt 1) (sseq _retw).
Proof.
  apply  nh_isq_end.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  split; auto.
  intros.
  unfold l_acquire_last_grt.
  intros.
  unfold leadstoPQ2. auto.
  split.
  apply IF_false_Prd_Itr_Stk.
  apply l_acquire_Post_IF_false.
  apply l_acquire_last_Grt_Itr_Stk.
Qed.


Lemma wff_l_block:
  NH_WFiseq l_setacq_psi l_setacq (Ast l_setacq_Prd l_setacq_grt 1) I_l_setacq.
Proof.
  unfold I_l_setacq.
  apply nh_isq_seq with l_setacq1_Prd l_setacq1_grt.
  apply wff_cmd_l_setacq1.
  apply nh_isq_seq with l_acquire_Post l_acquire_last_grt.
  apply wff_cmd_l_setacq2.
  apply wff_cmd_l_setacq3.
Qed.



(* the stack space required by l_release *)
Definition _STK_SIZE_lock_acquire := 6 + IH_Stk + _STK_SIZE_l_acquire.

Definition lock_acquire_psi := SCAP_consCT l_acquire (Ast l_acquire_Prd l_acquire_grt 1) SCAP_emptCT.

Definition lock_acquire_MPre (sp l pc: Label) : MPred :=
  (sp |->w pc) *^* (sp + 2 |->w l) *^* stkSpacePrec sp _STK_SIZE_lock_acquire.

Definition lock_acquire_Prd_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
      R SP = sp
      /\ R BP = bp
      /\ (lock_acquire_MPre sp l pc *^* mP) M
      /\ In l lockSet_H
      /\ fl _IF = true
      /\ isr = false
  end.

Definition lock_acquire_Prd (S : State) : Prop :=
  exists sp, exists l, exists pc, exists bp, exists mP, lock_acquire_Prd_aux sp l pc bp mP S.


Definition lock_acquire_MPost (sp l pc : Label) : MPred :=
    (lock_acquire_MPre sp l pc) *^* (Gamma l).

Definition lock_acquire_Post_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
      R SP = sp
      /\ R BP = bp
      /\ (lock_acquire_MPost sp l pc *^* mP) M
      /\ In l lockSet_H
      /\ fl _IF = true
      /\ isr = false
  end.

Definition lock_acquire_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (lock_acquire_Prd_aux sp l pc bp) (lock_acquire_Post_aux sp l pc bp) S S'.

Definition lock_acquire1_MPre (sp l pc bp : Label) :=
    (sp |->w bp) *^* (sp + 2 |->w pc) *^* (sp + 4 |->w l) *^* stkSpacePrec sp (_STK_SIZE_lock_acquire - 2).

Definition lock_acquire1_Prd_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R SP = sp
        /\ R BP = bp
        /\ (lock_acquire1_MPre sp l pc bp *^* mP) M
        /\ In l lockSet_H 
        /\ fl _IF = true
        /\ isr = false
  end.

Definition lock_acquire1_Prd (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, lock_acquire1_Prd_aux sp l pc bp mP S.

Definition lock_acquire1_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (lock_acquire1_Prd_aux sp l pc bp) (lock_acquire_Post_aux (sp +2) l pc bp) S S'.


Definition lock_acquire2_Prd_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R BP = sp
        /\ lock_acquire1_Prd_aux sp l pc bp mP (M, RF_update R BP bp, fl, isr)
  end.


Definition lock_acquire2_Prd (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, lock_acquire2_Prd_aux sp l pc bp mP S.

Definition lock_acquire2_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (lock_acquire2_Prd_aux sp l pc bp) (lock_acquire_Post_aux (sp +2) l pc bp) S S'.


Definition lock_acquire3_MPre (sp l pc bp : Label) :=
    (sp |->w l) *^* (sp + 2 |->w bp) *^* (sp + 4 |->w pc) *^* (sp + 6 |->w l) *^* stkSpacePrec sp (_STK_SIZE_lock_acquire - 4).

Definition lock_acquire3_Prd_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R SP = sp
        /\ (lock_acquire3_MPre sp l pc bp *^* mP) M
        /\ In l lockSet_H 
        /\ fl _IF = true
        /\ isr = false
  end.

Definition lock_acquire3_Prd (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, lock_acquire3_Prd_aux sp l pc bp mP S.


Definition lock_acquire3_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (lock_acquire3_Prd_aux sp l pc bp) (lock_acquire_Post_aux (sp + 4) l pc bp) S S'.


Definition lock_acquire4_MPre (sp l pc bp : Label) :=
    ((lock_acquire3_MPre sp l pc bp) *^* (INV1 *^* INV0)).



Definition lock_acquire4_Prd_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R SP = sp
        /\ (lock_acquire4_MPre (R SP) l pc bp *^* mP) M
        /\ In l lockSet_H 
        /\ fl _IF = false
        /\ isr = false
  end.

Definition lock_acquire4_Prd (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, lock_acquire4_Prd_aux sp l pc bp mP S.


Definition lock_acquire4_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (lock_acquire4_Prd_aux sp l pc bp) (lock_acquire_Post_aux (sp + 4) l pc bp) S S'.


Definition lock_acquire6_MPre (sp l pc bp : Label) :=
  (sp |->w_) *^* (sp + 2 |->w bp) *^* (sp + 4 |->w pc) *^* (sp + 6 |->w l) *^* stkSpacePrec sp (_STK_SIZE_lock_acquire - 4) *^* Gamma l.


Definition lock_acquire5_MPre (sp l pc bp : Label) :=
  (lock_acquire6_MPre sp l pc bp) *^* (INV1 *^* INV0).

Definition lock_acquire5_Prd_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R SP = sp
        /\ (lock_acquire5_MPre sp l pc bp *^* mP) M
        /\ In l lockSet_H
        /\ fl _IF = false
        /\ isr = false
  end.

Definition lock_acquire5_Prd (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, lock_acquire5_Prd_aux sp l pc bp mP S.


Definition lock_acquire5_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (lock_acquire5_Prd_aux sp l pc bp) (lock_acquire_Post_aux (sp + 4) l pc bp) S S'.


Definition lock_acquire6_Prd_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R SP = sp
        /\ (lock_acquire6_MPre sp l pc bp *^* mP) M
        /\ In l lockSet_H
        /\ fl _IF = true
        /\ isr = false
  end.

Definition lock_acquire6_Prd (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, lock_acquire6_Prd_aux sp l pc bp mP S.


Definition lock_acquire6_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (lock_acquire6_Prd_aux sp l pc bp) (lock_acquire_Post_aux (sp + 4) l pc bp) S S'.

(* _addw (Opr_imm 2) SP *)
Definition lock_acquire7_Prd_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R SP = sp
        /\ sp >= 2
        /\ lock_acquire6_Prd_aux (sp - 2) l pc bp mP (M, RF_update R SP (sp - 2), fl, isr)
  end.

Definition lock_acquire7_Prd (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, lock_acquire7_Prd_aux sp l pc bp mP S.


Definition lock_acquire7_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (lock_acquire7_Prd_aux sp l pc bp) (lock_acquire_Post_aux (sp +2) l pc bp) S S'.

Definition lock_acquire_Post (S : State) : Prop :=
  exists sp, exists l, exists pc, exists bp, exists mP, lock_acquire_Post_aux sp l pc bp mP S.


Definition lock_acquire_last_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (lock_acquire_Post_aux sp l pc bp) (lock_acquire_Post_aux sp l pc bp) S S'.


Lemma wff_cmd_lock_acquire1:
   NH_WFcmd
     (SCAP_consCT (lock_acquire + 1)
        (Ast lock_acquire1_Prd lock_acquire1_grt 1) lock_acquire_psi)
     lock_acquire (Ast lock_acquire_Prd lock_acquire_grt 1)
     (comml (_pushwp (Opr_reg BP))).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists lock_acquire1_Prd. exists lock_acquire1_grt.
  split. auto.

  assert (forall sp l pc bp mP, enable (lock_acquire_Prd_aux sp l pc bp mP) (nextS lock_acquire (comml (_pushwp (Opr_reg BP))))).
  intros.
  apply enable_pushwp with _STK_SIZE_lock_acquire; auto.
Lemma lock_acquire_Prd_aux_stkSpace:
   forall sp l pc bp mP S,
   lock_acquire_Prd_aux sp l pc bp mP S ->
   stkSpace (S_RF S SP) _STK_SIZE_lock_acquire (S_Mem S).
Proof.
  intros.
  unfold lock_acquire_Prd_aux in H.
  induction S as (((M, R), fl), isr).
  unfold lock_acquire_MPre in H.
  InvertAll.
  sep on.
  sep lift 2 in H1.
  UnfoldStateAll.
  rewrite H0.
  sep trivial.
Qed.

  apply lock_acquire_Prd_aux_stkSpace.
Lemma STK_SZ_acquire:
   _STK_SIZE_lock_acquire >= 2.
Proof.
  unfold _STK_SIZE_lock_acquire.
  auto with arith.
Qed.
Hint Resolve STK_SZ_acquire.
  auto.

  split.
  unfold enable in *.
  intros.
  unfold lock_acquire_Prd in H0.
  InvertAll.
  apply H with x x0 x1 x2 x3; auto.
  split.
  unfold pimp.
  unfold linkPG.
  intros.
  InvertAll.
  unfold lock_acquire_Prd in H0.
  InvertAll.
  unfold lock_acquire1_Prd.
  mexists5 (x0 - 2) x1 x2 x3 x4.
Lemma lock_acquire_step1:
forall f sp l pc bp mP S S',
  lock_acquire_Prd_aux sp l pc bp mP S  
  -> nextS f (comml (_pushwp (Opr_reg BP))) S S'
  -> lock_acquire1_Prd_aux (sp - 2) l pc bp mP S'.
Proof.
  intros.
  rename H0 into Htrans.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion Htrans).
  destruct H0 as [Hspgt2 [Hmemblk HM']].
  destruct H1 as [HR' [Hfl' Hisr']].
  subst R' fl' isr'. clear Htrans.
  unfold opr_eval in HM'.
  unfold lock_acquire_Prd_aux in H.
  destruct H as [HRsp [HRbp [Hmem [Hlock [Hfl Hisr]]]]].
  unfold lock_acquire_MPre in Hmem.
  unfold lock_acquire1_Prd_aux.
  reg simpl.
  rewrite HRsp.
  split_l; split_l.
  split; auto.
  unfold lock_acquire1_MPre.
  rewrite HRsp in Hspgt2.
  assert (sp - 2 + 2 = sp). auto.
  rewrite H. clear H.
  assert (sp - 2 + 4 = sp + 2). auto.
  rewrite H. clear H.
  substH HM' with (MemUpd16_MemUpdW HM').
  destruct HM' as [junk HM'].
  subst M'.
  sep on.
  rewrite HRsp. rewrite HRbp.
  sep lift 0.
  sep upd2.
  sep lift 2 in Hmem.
  sep stksplit in Hmem.
  sep matchH Hmem with 4.
Qed.


  apply lock_acquire_step1 with lock_acquire x; auto.

  unfold gimp.
  unfold conjPG. unfold conjG.
  intros.
  InvertAll.
  unfold lock_acquire_grt.
  unfold leadstoPQ2.
  intros.
  unfold lock_acquire1_grt in H2.
  unfold leadstoPQ2 in H2.
  assert (lock_acquire_Post_aux (sp - 2 + 2) l pc bp mP S').
  apply H2.
  apply lock_acquire_step1 with lock_acquire S; auto.
  induction S as (((M, R), fl), isr).
  induction x as (((M', R'), fl'), isr').
  destruct (nextS_inversion H3).
  InvertAll.
  unfold lock_acquire_Prd_aux in H0.
  InvertAll.
  rewrite H10 in H8.
  assert (sp - 2 + 2 = sp). auto.
  rewrite H15 in H4. auto.

Lemma lock_acquire_stk_geIH: _STK_SIZE_lock_acquire >= IH_Stk + 6.
Proof.
  unfold _STK_SIZE_lock_acquire. omega.
Qed.
Hint Resolve lock_acquire_stk_geIH.

Lemma lock_acquire_Prd_Itr_Stk: Prd_Itr_Stk lock_acquire_Prd.
Proof.
    unfold Prd_Itr_Stk.
    intros S Hpre Henable.
    unfold lock_acquire_Prd in Hpre.
    destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
    unfold lock_acquire_Prd_aux in Hpre.
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in Henable.
    simplS in * .
    destruct Hpre as [HRsp [HRbp [Hmem [Hlock [Hfl Hisr]]]]].
    rewrite HRsp.
    unfold lock_acquire_MPre in Hmem.
    sep on.
    sep lift 2 in Hmem.
    sep stksplit (IH_Stk + 6) in Hmem.
    sep lift 0 in Hmem.
    destruct Hmem as [M1 [M2 [Hsplit [Hmem1 Hmem2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    split_l.
    intros M1' Hmem1'.
    unfold lock_acquire_Prd.
    mexists5 sp l pc bp mP.
    unfold lock_acquire_Prd_aux.
    unfold lock_acquire_MPre.
    split_l. split_l.
    split; auto.
    sep on.
    sep lift 2.
    assert (sp >= IH_Stk + 6).
    apply stkSpacePrec_sp_ge_sz with M1; auto.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      split.
        disjoint rep_l M1.
        apply stkSpacePrec_stkSpacePrec_eqDom with sp (IH_Stk + 6); auto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      apply stkSpacePrec_stkSpacePrec_eqDom with sp (IH_Stk + 6); auto.
    split_l.
    sep on.
    sep matchH Hmem2 with 4.
Qed.

Lemma lock_acquire_Grt_Itr_Stk: Grt_Itr_Stk lock_acquire_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros S S' Hgrt Henable.
  unfold lock_acquire_grt in * .
  intros M1 M2 M2' Hsplit HM2 HM2'.
  unfold leadstoPQ2 in *.
  intros sp l pc bp mP Hpre.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in Henable.
  apply Hgrt.
  clear Hgrt.
  unfold lock_acquire_Prd_aux in * .
  simplS in * .
  destruct Hpre as [HRsp [HRbp [Hmem [Hlock [Hfl Hisr]]]]].
  split_l. split_l.
  split; auto.
  unfold lock_acquire_MPre in *.
  sep on.
  sep lift 2 in Hmem.
  sep stksplit (IH_Stk + 6) in Hmem.
  sep lift 0 in Hmem.
  destruct Hmem as [M2x [M1x [Hsplit' [HM2x HM1x]]]].
  sep lift 2.
  assert (sp >= IH_Stk + 6).
  apply stkSpacePrec_sp_ge_sz with M2x; auto.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  assert (M2x = M2' /\ M1x = M1).
    sep off.
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      apply stkSpacePrec_stkSpacePrec_eqDom with (R SP) (IH_Stk + 6); auto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    rewrite HRsp in HM2'.
    assert (Hx := prec_disjUnion_inv H0 H1 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H0; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  rewrite HRsp in HM2.
  split_l.
  sep on.
  sep matchH HM1x with 4.
Qed.


  split.
  apply lock_acquire_Prd_Itr_Stk.
  apply lock_acquire_Grt_Itr_Stk.
Qed.


Lemma wff_cmd_lock_acquire2:
   NH_WFcmd
     (SCAP_consCT (lock_acquire + 1 + 1)
        (Ast lock_acquire2_Prd lock_acquire2_grt 1) lock_acquire_psi)
     (lock_acquire + 1) (Ast lock_acquire1_Prd lock_acquire1_grt 1)
     (comml (_movwp (Opr_reg SP) BP)).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists lock_acquire2_Prd. exists lock_acquire2_grt.
  split. auto.
  split; auto.

  split.
  unfold pimp.
  unfold linkPG.
  intros.
  InvertAll.
  unfold lock_acquire1_Prd in H.
  InvertAll.
  unfold lock_acquire2_Prd.
  mexists5 x0 x1 x2 x3 x4.
Lemma lock_acquire_step2:
forall f sp l pc bp mP S S',
  lock_acquire1_Prd_aux sp l pc bp mP S
  -> nextS f (comml (_movwp (Opr_reg SP) BP)) S S'
  -> lock_acquire2_Prd_aux sp l pc bp mP S'.
Proof.
  intros.
  rename H0 into Htrans.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion Htrans).
  subst M'.
  destruct H1 as [HR' [Hfl' Hisr']].
  unfold opr_eval in HR'.
  subst R' fl' isr'. clear Htrans.
  unfold lock_acquire1_Prd_aux in H.
  destruct H as [HRsp [HRbp [Hmem [Hlock [Hfl Hisr]]]]].
  unfold lock_acquire2_Prd_aux.
  reg simpl.
  rewrite HRsp.
  split_l.
  rewrite <- HRbp.
  assert (RF_update R BP (R BP) = R).
  apply RF_update_prev.
  rewrite H. clear H.
  rewrite HRbp.
  unfold lock_acquire1_Prd_aux.
  split_l; split_l.
  split; auto.
Qed.


  apply lock_acquire_step2 with (lock_acquire + 1) x; auto.

  unfold gimp.
  unfold conjPG. unfold conjG.
  intros.
  InvertAll.
  unfold lock_acquire1_grt.
  unfold leadstoPQ2.
  intros.
  unfold lock_acquire2_grt in H1.
  unfold leadstoPQ2 in H1.
  apply H1.
  apply lock_acquire_step2 with (lock_acquire + 1) S; auto.

Lemma stk_lock_acquire_N2_geIH: _STK_SIZE_lock_acquire - 2 >= IH_Stk + 6.
Proof.
  unfold _STK_SIZE_lock_acquire. unfold _STK_SIZE_l_acquire. omega.
Qed.
Hint Resolve stk_lock_acquire_N2_geIH.

Lemma lock_acquire1_Prd_Itr_Stk: Prd_Itr_Stk lock_acquire1_Prd.
Proof.
    unfold Prd_Itr_Stk.
    intros S Hpre Henable.
    unfold lock_acquire1_Prd in Hpre.
    destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
    unfold lock_acquire1_Prd_aux in Hpre.
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in Henable.
    simplS in * .
    destruct Hpre as [HRsp [HRbp [Hmem [Hlock [Hfl Hisr]]]]].
    rewrite HRsp.
    unfold lock_acquire1_MPre in Hmem.
    sep on.
    sep lift 3 in Hmem.
    sep stksplit (IH_Stk + 6) in Hmem.
    sep lift 0 in Hmem.
    destruct Hmem as [M1 [M2 [Hsplit [Hmem1 Hmem2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    split_l.
    intros M1' Hmem1'.
    unfold lock_acquire1_Prd.
    mexists5 sp l pc bp mP.
    unfold lock_acquire1_Prd_aux.
    unfold lock_acquire1_MPre.
    split_l. split_l.
    split; auto.
    sep on.
    sep lift 3.
    assert (sp >= IH_Stk + 6).
    apply stkSpacePrec_sp_ge_sz with M1; auto.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      split.
        disjoint rep_l M1.
        apply stkSpacePrec_stkSpacePrec_eqDom with sp (IH_Stk + 6); auto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      apply stkSpacePrec_stkSpacePrec_eqDom with sp (IH_Stk + 6); auto.
    split_l.
    sep on.
    sep matchH Hmem2 with 5.
Qed.

Lemma lock_acquire1_Grt_Itr_Stk: Grt_Itr_Stk lock_acquire1_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros S S' Hgrt Henable.
  unfold lock_acquire1_grt in * .
  intros M1 M2 M2' Hsplit HM2 HM2'.
  unfold leadstoPQ2 in *.
  intros sp l pc bp mP Hpre.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in Henable.
  apply Hgrt.
  clear Hgrt.
  unfold lock_acquire1_Prd_aux in * .
  simplS in * .
  destruct Hpre as [HRsp [HRbp [Hmem [Hlock [Hfl Hisr]]]]].
  split_l. split_l.
  split; auto.
  unfold lock_acquire1_MPre in *.
  sep on.
  sep lift 3 in Hmem.
  sep stksplit (IH_Stk + 6) in Hmem.
  sep lift 0 in Hmem.
  destruct Hmem as [M2x [M1x [Hsplit' [HM2x HM1x]]]].
  sep lift 3.
  assert (sp >= IH_Stk + 6).
  apply stkSpacePrec_sp_ge_sz with M2x; auto.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  assert (M2x = M2' /\ M1x = M1).
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      apply stkSpacePrec_stkSpacePrec_eqDom with (R SP) (IH_Stk + 6); auto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    rewrite HRsp in HM2'.
    assert (Hx := prec_disjUnion_inv H0 H1 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H0; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  rewrite HRsp in HM2.
  split_l.
  sep on.
  sep matchH HM1x with 5.
Qed.

  split.
    apply lock_acquire1_Prd_Itr_Stk.
    apply lock_acquire1_Grt_Itr_Stk.
Qed.


Lemma wff_cmd_lock_acquire3:
   NH_WFcmd
     (SCAP_consCT (lock_acquire + 1 + 1 + 1)
        (Ast lock_acquire3_Prd lock_acquire3_grt 1) lock_acquire_psi)
     (lock_acquire + 1 + 1) (Ast lock_acquire2_Prd lock_acquire2_grt 1)
     (comml (_pushwm (A_reg 4 BP))).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists lock_acquire3_Prd. exists lock_acquire3_grt.
  split. auto.

  generalize (lock_acquire + 1 + 1). intro f.

  split.
  assert (forall f sp l pc bp mP,
   enable (lock_acquire2_Prd_aux sp l pc bp mP)
     (nextS f (comml (_pushwm (A_reg 4 BP))))).
Lemma lock_acquire2_Prd_enable:
   forall f sp l pc bp mP,
   enable (lock_acquire2_Prd_aux sp l pc bp mP)
     (nextS f (comml (_pushwm (A_reg 4 BP)))).
Proof.
  unfold enable.
  intros.
  assert (((addr_eval (S_RF S) (A_reg 4 BP)) |~>w l) (S_Mem S)).
Lemma lock_acquire2_Prd_1:
   forall sp l pc bp mP S,
   lock_acquire2_Prd_aux sp l pc bp mP S ->
   ((addr_eval (S_RF S) (A_reg 4 BP)) |~>w l) (S_Mem S).
Proof.
  intros.
  induction S as (((M, R), fl), isr).
  unfold lock_acquire2_Prd_aux in H.
  unfold S_RF. unfold S_Mem.
  unfold addr_eval.
  inversion_clear H. rewrite H0.
  inversion_clear H1.
  unfold lock_acquire1_MPre in H2.
  reg simpl.
  InvertAll.
  sep on.
  sep lift 2 in H2.
  sep trivial.
Qed.
  apply lock_acquire2_Prd_1 with sp pc bp mP; auto.
  induction S as (((M, R), fl), isr).
  exists (MemUpdW M (R SP - 2) l 0, RF_update R SP (R SP - 2), fl, isr).
  apply nextS_pushwm with (addr_eval R (A_reg 4 BP)) l. auto.
  auto.
  unfold pushW.
  unfold lock_acquire2_Prd_aux in H.
  destruct H as [HRbp Hpre].
  unfold lock_acquire1_Prd_aux in Hpre.
  unfold lock_acquire1_MPre in Hpre.
  reg simpl.
  destruct Hpre as [HRsp [Hjunk [Hmem [Hlock [Hfl Hisr]]]]].
  split.
  sep on.
  sep lift 3 in Hmem.
  sep stksplit 2 in Hmem.
  sep lift 0 in Hmem.
  rewrite HRsp.
  sep trivial.
  split; auto.
  apply  MemUpdW_MemUpd16 with 0; auto.
Qed.

  apply lock_acquire2_Prd_enable.
  unfold enable in *.
  intros.
  unfold lock_acquire2_Prd in H0.
  InvertAll.
  apply H with x x0 x1 x2 x3; auto.

  split.
  unfold pimp.
  unfold linkPG.
  intros.
  InvertAll.
  unfold lock_acquire2_Prd in H.
  InvertAll.
  unfold lock_acquire3_Prd.
  mexists5 (x0 - 2) x1 x2 x3 x4.
Lemma lock_acquire_step3:
forall f sp l pc bp mP S S',
  lock_acquire2_Prd_aux sp l pc bp mP S  
  -> nextS f (comml (_pushwm (A_reg 4 BP))) S S'
  -> lock_acquire3_Prd_aux (sp - 2) l pc bp mP S'.
Proof.
  intros.
  rename H0 into Htrans.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion Htrans).
  destruct H0 as [w [Hsrc [Hspgt2 [Hmemblk HM']]]].
  destruct H1 as [HR' [Hfl' Hisr']].
  subst R' fl' isr'. clear Htrans.
  assert ((addr_eval (S_RF (M, R, fl, isr)) (A_reg 4 BP) |~>w l) (S_Mem (M, R, fl, isr))).
  apply lock_acquire2_Prd_1 with sp pc bp mP; auto.
  UnfoldStateAll.
  assert (w = l).
  apply leadsToW_leadsToW_eq with M (addr_eval R (A_reg 4 BP)); auto.
  subst w. clear H0.
  unfold addr_eval in Hsrc.
  unfold lock_acquire2_Prd_aux in H.
  destruct H as [HRbp Hpre].
  unfold lock_acquire1_Prd_aux in Hpre.
  unfold lock_acquire1_MPre in Hpre.
  reg simpl.
  destruct Hpre as [HRsp [Htmp [Hmem [Hlock [Hfl Hisr]]]]].
  unfold lock_acquire3_Prd_aux.
  reg simpl.
  split. rewrite HRsp. auto.
  split; auto.
  unfold lock_acquire3_MPre.
  substH HM' with (MemUpd16_MemUpdW HM').
  destruct HM' as [junk HM'].
  subst M'.
  sep on.
  sep lift 0.
  rewrite HRsp.
  sep upd2.
  sep lift 3 in Hmem.
  sep stksplit in Hmem.
  sep lift 1 in Hmem.
  rewrite HRsp in Hspgt2.
  assert (sp - 2 + 2 = sp). auto.
  rewrite H. clear H.
  assert (sp - 2 + 4 = sp + 2). auto.
  rewrite H. clear H.
  assert (sp - 2 + 6 = sp + 4). auto.
  rewrite H. clear H.
  assert (_STK_SIZE_lock_acquire - 2 - 2 = _STK_SIZE_lock_acquire - 4). auto with arith.
  rewrite H in Hmem.
  sep matchH Hmem with 6.
Qed.

  apply lock_acquire_step3 with f x; auto.

  unfold gimp.
  unfold conjPG. unfold conjG.
  intros.
  InvertAll.
  unfold lock_acquire2_grt.
  unfold leadstoPQ2.
  intros.
  unfold lock_acquire3_grt in H1.
  unfold leadstoPQ2 in H1.
  assert (lock_acquire_Post_aux (sp - 2 + 4) l pc bp mP S').
  apply H1.
  apply lock_acquire_step3 with f S; auto.
  induction S as (((M, R), fl), isr).
  induction x as (((M', R'), fl'), isr').
  destruct (nextS_inversion H2).
  InvertAll.
  unfold lock_acquire2_Prd_aux in H.
  unfold lock_acquire1_Prd_aux in H.
  InvertAll.
  reg simpl.
  rewrite H in H7.
  assert (sp - 2 + 4 = sp + 2); auto.
  rewrite H16 in H3. auto.

Lemma lock_acquire2_Prd_Itr_Stk: Prd_Itr_Stk lock_acquire2_Prd.
Proof.
    unfold Prd_Itr_Stk.
    intros S Hpre Henable.
    unfold lock_acquire2_Prd in Hpre.
    destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
    unfold lock_acquire2_Prd_aux in Hpre.
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in Henable.
    simplS in * .
    destruct Hpre as [HRbp Hpre].
    unfold lock_acquire1_Prd_aux in Hpre.
    reg simpl.
    destruct Hpre as [HRsp [Hbp0 [Hmem [Hlock [Hfl Hisr]]]]].
    unfold lock_acquire1_MPre in Hmem.
    sep on.
    sep lift 3 in Hmem.
    sep stksplit (IH_Stk + 6) in Hmem.
    sep lift 0 in Hmem.
    destruct Hmem as [M1 [M2 [Hsplit [Hmem1 Hmem2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    rewrite HRsp.
    split_l.
    intros M1' Hmem1'.
    unfold lock_acquire2_Prd.
    mexists5 sp l pc bp mP.
    unfold lock_acquire2_Prd_aux.
    split_l.
    unfold lock_acquire1_Prd_aux.
    unfold lock_acquire1_MPre.
    reg simpl.
    split_l. split_l.
    split; auto.
    sep on.
    sep lift 3.
    assert (sp >= IH_Stk + 6).
    apply stkSpacePrec_sp_ge_sz with M1; auto.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      split.
        disjoint rep_l M1.
        apply stkSpacePrec_stkSpacePrec_eqDom with sp (IH_Stk + 6); auto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      apply stkSpacePrec_stkSpacePrec_eqDom with sp (IH_Stk + 6); auto.
    split_l.
    sep on.
    sep matchH Hmem2 with 5.
Qed.

Lemma lock_acquire2_Grt_Itr_Stk: Grt_Itr_Stk lock_acquire2_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros S S' Hgrt Henable.
  unfold lock_acquire2_grt in * .
  intros M1 M2 M2' Hsplit HM2 HM2'.
  unfold leadstoPQ2 in *.
  intros sp l pc bp mP Hpre.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in Henable.
  apply Hgrt.
  clear Hgrt.
  unfold lock_acquire2_Prd_aux in * .
  simplS in * .
  destruct Hpre as [HRbp Hpre].
  unfold lock_acquire1_Prd_aux in *.
  reg simpl.
  destruct Hpre as [HRsp [HRbp0 [Hmem [Hlock [Hfl Hisr]]]]].
  split_l. split_l. split_l. 
  split; auto.
  unfold lock_acquire1_MPre in *.
  sep on.
  sep lift 3 in Hmem.
  sep stksplit (IH_Stk + 6) in Hmem.
  sep lift 0 in Hmem.
  destruct Hmem as [M2x [M1x [Hsplit' [HM2x HM1x]]]].
  sep lift 3.
  assert (sp >= IH_Stk + 6).
  apply stkSpacePrec_sp_ge_sz with M2x; auto.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  rewrite HRsp in HM2. rewrite HRsp in HM2'.
  assert (M2x = M2' /\ M1x = M1).
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      apply stkSpacePrec_stkSpacePrec_eqDom with sp (IH_Stk + 6); auto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    assert (Hx := prec_disjUnion_inv H0 H1 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H0; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  split_l.
  sep on.
  sep matchH HM1x with 5.
Qed.

  split.
  apply lock_acquire2_Prd_Itr_Stk.
  apply lock_acquire2_Grt_Itr_Stk.
Qed.



Lemma wff_cmd_lock_acquire4:
   NH_WFcmd
     (SCAP_consCT (lock_acquire + 1 + 1 + 1 + 1)
        (Ast lock_acquire4_Prd lock_acquire4_grt 1) lock_acquire_psi)
     (lock_acquire + 1 + 1 + 1) (Ast lock_acquire3_Prd lock_acquire3_grt 1)
     (comml _cli).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  exists lock_acquire4_Prd. exists lock_acquire4_grt.
  split. auto.
  split.
  unfold pimp.
  unfold linkPG.
  intros.
  InvertAll.
  unfold lock_acquire3_Prd in H.
  InvertAll.
  unfold lock_acquire4_Prd.
  mexists5 x0 x1 x2 x3 x4.
Lemma lock_acquire_step4:
forall sp l pc bp mP S S',
  lock_acquire3_Prd_aux sp l pc bp mP S
  -> gcli S S'
  -> lock_acquire4_Prd_aux sp l pc bp mP S'.
Proof.
  intros.
  rename H0 into Htrans.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  unfold gcli in Htrans.
  UnfoldStateAll.
  destruct Htrans as [Hpp [Hfl' [Hisr' HR']]].
  subst R' isr'.
  unfold prepost in Hpp.
  destruct Hpp as [M1 [M2 [M1' [HM1 [HM1' [HsplitM HsplitM']]]]]].
  UnfoldStateAll.
  unfold lock_acquire4_Prd_aux.
  unfold lock_acquire3_Prd_aux in H.
  destruct H as [HRsp [Hmem [Hlock [Hfl Hisr]]]].
  split_l. split; auto.
  unfold lock_acquire4_MPre.
  assert ((INV1 *^* INV0 *^* (lock_acquire3_MPre (R SP) l pc bp *^* mP)) M').
  exists M1'. exists M2.
  split_l.
  unfold Aemp in HM1.
  rewrite HM1 in HsplitM.
  assert (M2 = M); auto with PMap.
  subst M2.
  split; auto.
  unfold choice in HM1'.
  inversion_clear HM1'.
  inversion H; auto.
  inversion H. assert (itr_enabled (M, R, fl, isr)).
  unfold itr_enabled. UnfoldStateAll. auto.
  contradiction.
  sep on. rewrite HRsp. auto.
  sep on.
  sep matchH H with 4.
Qed.

  apply lock_acquire_step4 with x; auto.

  unfold gimp.
  unfold conjPG. unfold conjG.
  intros.
  InvertAll.
  unfold lock_acquire3_grt.
  unfold leadstoPQ2.
  intros.
  unfold lock_acquire4_grt in H1.
  unfold leadstoPQ2 in H1.
  apply H1.
  apply lock_acquire_step4 with S; auto.

Lemma lock_acquire3_Prd_Itr_Stk: Prd_Itr_Stk lock_acquire3_Prd.
Proof.
    unfold Prd_Itr_Stk.
    intros S Hpre Henable.
    unfold lock_acquire3_Prd in Hpre.
    destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
    unfold lock_acquire3_Prd_aux in Hpre.
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in Henable.
    simplS in * .
    destruct Hpre as [HRsp [Hmem [Hlock [Hfl Hisr]]]].
    unfold lock_acquire3_MPre in Hmem.
    sep on.
    sep lift 4 in Hmem.
Lemma stk_lock_acquire_N4_geIH:
  _STK_SIZE_lock_acquire - 4 >= IH_Stk + 6.
Proof.
  unfold _STK_SIZE_lock_acquire.
  unfold _STK_SIZE_l_acquire. omega.
Qed.
Hint Resolve stk_lock_acquire_N4_geIH.
    sep stksplit (IH_Stk + 6) in Hmem.
    sep lift 0 in Hmem.
    destruct Hmem as [M1 [M2 [Hsplit [Hmem1 Hmem2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    rewrite HRsp.
    split_l.
    intros M1' Hmem1'.
    unfold lock_acquire3_Prd.
    mexists5 sp l pc bp mP.
    unfold lock_acquire3_Prd_aux.
    split_l.
    split; auto.
    unfold lock_acquire3_MPre.
    sep on.
    sep lift 4.
    assert (sp >= IH_Stk + 6).
    apply stkSpacePrec_sp_ge_sz with M1; auto.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      split.
        disjoint rep_l M1.
        apply stkSpacePrec_stkSpacePrec_eqDom with sp (IH_Stk + 6); auto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      apply stkSpacePrec_stkSpacePrec_eqDom with sp (IH_Stk + 6); auto.
    split_l.
    sep on.
    sep matchH Hmem2 with 6.
Qed.


Lemma lock_acquire3_Grt_Itr_Stk : Grt_Itr_Stk lock_acquire3_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros S S' Hgrt Henable.
  unfold lock_acquire3_grt in * .
  intros M1 M2 M2' Hsplit HM2 HM2'.
  unfold leadstoPQ2 in *.
  intros sp l pc bp mP Hpre.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in Henable.
  apply Hgrt.
  clear Hgrt.
  unfold lock_acquire3_Prd_aux in * .
  simplS in * .
  destruct Hpre as [HRsp [Hmem [Hlock [Hfl Hisr]]]].
  split_l. 
  unfold lock_acquire3_MPre in *.
  split; auto.
  sep on.
  sep lift 4 in Hmem.
  sep stksplit (IH_Stk + 6) in Hmem.
  sep lift 0 in Hmem.
  destruct Hmem as [M2x [M1x [Hsplit' [HM2x HM1x]]]].
  sep lift 4.
  assert (sp >= IH_Stk + 6).
  apply stkSpacePrec_sp_ge_sz with M2x; auto.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  rewrite HRsp in HM2, HM2'.
  assert (M2x = M2' /\ M1x = M1).
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      apply stkSpacePrec_stkSpacePrec_eqDom with sp (IH_Stk + 6); auto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    assert (Hx := prec_disjUnion_inv H0 H1 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H0; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  split_l.
  sep on.
  sep matchH HM1x with 6.
Qed.

  split.
  apply lock_acquire3_Prd_Itr_Stk.
  apply lock_acquire3_Grt_Itr_Stk.
Qed.

Parameter l_acquire_not_primentry:
  ~ prim_entry l_acquire.

Lemma lock_acquire4_enable:
   forall f sp l pc bp mP,
   enable (lock_acquire4_Prd_aux sp l pc bp mP)
     (nextS f (comml (_call l_acquire))).
Proof.
  intros.
  unfold enable.
  intros.
  unfold lock_acquire4_Prd_aux in H.
  induction S as (((M, R), fl), isr).
  exists (MemUpdW M (R SP - 2) (f+1) 0, RF_update R SP (R SP - 2), fl, isr).
  apply nextS_call. 
  unfold pushW.
  destruct H as [HRsp [Hmem [Hlock [Hfl Hisr]]]].
  split.
  unfold lock_acquire4_MPre in Hmem.
  unfold lock_acquire3_MPre in Hmem.
  sep on.
  sep lift 4 in Hmem.
  sep stksplit 2 in Hmem.
  sep lift 0 in Hmem.
  sep trivial.
  split; auto.
  apply  MemUpdW_MemUpd16 with 0; auto.
Qed.

Definition l_acquire_closure_step5 (sp l pc bp : int) (mP: MPred) : MPred :=
  (sp + 2 |->w bp *^* sp + 4 |->w pc *^* sp + 6 |->w l) *^* stkSpacePrec (sp - 2 - _STK_SIZE_l_acquire) IH_Stk *^* mP.

Lemma stk_lock_acquire_call1:
_STK_SIZE_lock_acquire - 4 - 2 - _STK_SIZE_l_acquire = IH_Stk.
Proof.
  unfold _STK_SIZE_lock_acquire. omega.
Qed.
Hint Resolve stk_lock_acquire_call1.

Lemma stk_lock_acquire_call2:
_STK_SIZE_lock_acquire - 4 - 2 >= _STK_SIZE_l_acquire.
Proof.
  unfold _STK_SIZE_lock_acquire. omega.
Qed.
Hint Resolve stk_lock_acquire_call2.

Lemma lock_acquire_step5_call3:
  forall f sp l pc bp mP S S', 
    lock_acquire4_Prd_aux sp l pc bp mP S
    -> nextS f (comml (_call l_acquire)) S S'
    -> l_acquire_Prd_aux (sp - 2) l (f + 1) (S_RF S' BP) (l_acquire_closure_step5 sp l pc bp mP) S'.
Proof.
  intros.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion H0).
  destruct H1 as [Hspgt2 [Hpcspace HM']].
  destruct H2 as [HR' [Hfl' Hisr']].
  substH HM' with (MemUpd16_MemUpdW HM').
  destruct HM' as [junk HM'].
  subst M' R' fl' isr'. clear H0.
  UnfoldStateAll.
  reg simpl.
  unfold lock_acquire4_Prd_aux in H.
  destruct H as [HRsp [Hmem [Hlock [Hfl Hisr]]]].
  rewrite HRsp in Hmem.
  unfold lock_acquire4_MPre in Hmem.
  unfold lock_acquire3_MPre in Hmem.
  unfold l_acquire_Prd_aux.
  reg simpl.
  rewrite HRsp. split_l. split_l.
  split; auto.
  unfold l_acquire_MPre.
  unfold l_acquire_closure_step5.
  sep on.
  sep lift 0.
  sep upd2.
  sep lift 4 in Hmem.
  sep stksplit in Hmem.
  sep stksplit _STK_SIZE_l_acquire in Hmem.
  assert (_STK_SIZE_lock_acquire - 4 - 2 - _STK_SIZE_l_acquire = IH_Stk).
  auto.
  rewrite H in Hmem.
  rewrite HRsp in Hspgt2.
  assert (sp - 2 + 2 = sp). auto.
  rewrite H0.
  sep matchH Hmem with 10.
Qed.

Lemma lock_acquire_step5_call2: 
  forall f, gimp (sp_call_pred lock_acquire4_Prd f l_acquire >o l_acquire_grt) fun_frame.
Proof.
  unfold gimp.
  unfold sp_call_pred. unfold conjPG. unfold linkPG.
  intros.
  destruct H as [[S0 [Hpre Htrans]] Hgrt].
  unfold lock_acquire4_Prd in Hpre.
  destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
  assert (l_acquire_Prd_aux (sp - 2) l (f + 1) (S_RF S BP) (l_acquire_closure_step5 sp l pc bp mP) S).
  apply lock_acquire_step5_call3 with S0; auto.
  rename H into HFpre.
  unfold l_acquire_grt in Hgrt.
  unfold leadstoPQ2 in Hgrt.
  assert (l_acquire_Post_aux (sp - 2) l (f+1) (S_RF S BP) (l_acquire_closure_step5 sp l pc bp mP) S').
  apply Hgrt; auto.
  clear Hgrt.
  unfold fun_frame.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  unfold l_acquire_Prd_aux in HFpre.
  unfold l_acquire_Post_aux in H.
  UnfoldStateAll.
  destruct HFpre as [HRsp [HRbp [Hmem [Hlock [Hfl Hisr]]]]].
  destruct H as [HRsp' [HRbp' [Hmem' [Hfl' Hisr']]]].
  split.
  rewrite HRsp. rewrite HRsp'. auto.
  exists (f + 1).
  unfold Ret_Addr.
  UnfoldStateAll.
  split.
  intro.
  apply nextPC_retw with (sp - 2). auto.
  unfold l_acquire_MPre in Hmem.
  sep on.
  sep lift 0 in Hmem.
  sep trivial.
  intro.
  apply nextPC_retw with (sp - 2); auto.
  unfold l_acquire_MPost in Hmem'.
  unfold l_acquire_MPre in Hmem'.
  sep on.
  sep lift 0 in Hmem'.
  sep trivial.
Qed.

Lemma lock_acquire_step5_call4_aux:
forall f f' sp l pc bp mP S S',
  lock_acquire4_Prd_aux sp l pc bp mP S
  -> wrap_fun l_acquire_grt f l_acquire f' S S'
  -> lock_acquire5_Prd_aux sp l pc bp mP S'.
Proof.
  intros.
  rename H into Hpre.
  unfold wrap_fun in H0.
  unfold conjG in H0.
  destruct H0 as [S2 [[S1 [Hcall Hgrt]] Hretw]].
  assert (sp >= 2).
  unfold lock_acquire4_Prd_aux in Hpre.
  induction S as (((M, R), fl), isr).
  induction S1 as (((M', R'), fl'), isr').
  InvertAll.
  destruct (nextS_inversion Hcall).
  InvertAll.
  rewrite H in H7; auto.
  rename H into Hspge2.
  assert (l_acquire_Prd_aux (sp - 2) l (f + 1) (S_RF S1 BP) (l_acquire_closure_step5 sp l pc bp mP) S1).
  apply lock_acquire_step5_call3 with S; auto.
  clear Hpre.
  rename H into HFpre.
  unfold l_acquire_grt in Hgrt.
  unfold leadstoPQ2 in Hgrt.
  assert (l_acquire_Post_aux (sp - 2) l (f+1) (S_RF S1 BP) (l_acquire_closure_step5 sp l pc bp mP) S2).
  apply Hgrt; auto.
  clear HFpre Hgrt Hcall.
  rename H into Hpre.
  induction S2 as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion Hretw).
  subst M'.
  destruct H0 as [[pc'' [HMpc HR']] [Hfl' Hisr']].
  subst fl' isr' R'. clear Hretw. clear HMpc.
  unfold l_acquire_Post_aux in Hpre.
  destruct Hpre as [HRsp [HRbp [Hmem [Hlock [Hfl Hisr]]]]].
  unfold lock_acquire5_Prd_aux.
  reg simpl.
  unfold l_acquire_MPost in Hmem.
  unfold l_acquire_MPre in Hmem.
  assert (R SP + 2 = sp).
  rewrite HRsp. auto.
  split_l.
  split; auto.
  unfold lock_acquire5_MPre.
  unfold lock_acquire6_MPre.
  unfold l_acquire_closure_step5 in Hmem.
  sep on.
  sep lift 4.
  sep stksplit.
  sep lift 2 in Hmem.
  assert (sp - 2 >= _STK_SIZE_l_acquire).
  eapply stkSpacePrec_sp_ge_sz1; eauto.
  assert (_STK_SIZE_lock_acquire - 4 - 2 >= _STK_SIZE_l_acquire).
  auto.
  sep stksplit _STK_SIZE_l_acquire.
  assert (_STK_SIZE_lock_acquire - 4 - 2 - _STK_SIZE_l_acquire = IH_Stk).
  auto.
  rewrite H2.
  assert (sp - 2 + 2 = sp). auto.
  rewrite H3 in Hmem.
  sep matchH Hmem with 10.
Qed.

Lemma lock_acquire_step5_call4:
forall f pc' : nat,
    pimp (lock_acquire4_Prd |> wrap_fun l_acquire_grt f l_acquire pc')
      lock_acquire5_Prd.
Proof.
  unfold pimp. unfold linkPG.
  intros.
  destruct H as [S0 [Hpre Htrans]].
  unfold lock_acquire4_Prd in Hpre.
  destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
  unfold lock_acquire5_Prd.
  mexists5 sp l pc bp mP.
  apply lock_acquire_step5_call4_aux with f pc' S0; auto.
Qed.


Lemma lock_acquire_step5_call5:
forall f pc' : nat,
    gimp
      (lock_acquire4_Prd >o wrap_fun l_acquire_grt f l_acquire pc'
       oo lock_acquire5_grt) lock_acquire4_grt.
Proof.
  intros.
  unfold gimp.
  unfold conjPG. unfold conjG.
  intros.
  destruct H as [S'' [[Hpre Hfun] Htrans5]].
  unfold lock_acquire4_grt.
  unfold leadstoPQ2.
  intros.
  unfold lock_acquire5_grt in Htrans5.
  unfold leadstoPQ2 in Htrans5.
  apply Htrans5.
  apply lock_acquire_step5_call4_aux with f pc' S; auto.
Qed.

Lemma lock_acquire4_Prd_IF_false:
   forall S : State, lock_acquire4_Prd S -> S_Flag S _IF = false.
Proof.
  intros.
  unfold lock_acquire4_Prd in H.
  unfold lock_acquire4_Prd_aux in H.
  induction S as (((M, R), fl), isr).
  InvertAll.
  unfold S_Flag. auto.
Qed.

Lemma lock_acquire4_Grt_Itr_Stk:
   Grt_Itr_Stk lock_acquire4_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  induction S as (((M, R), fl), isr).
  unfold lock_acquire4_grt. unfold leadstoPQ2.
  intros.
  unfold lock_acquire4_Prd_aux in H4.
  UnfoldStateAll.
  InvertAll.
  unfold itr_enabled in H0.
  inversion_clear H0.
  UnfoldStateAll.
  rewrite H8 in H7. inversion H7.
Qed.


Lemma wff_cmd_lock_acquire5:
   NH_WFcmd
     (SCAP_consCT (lock_acquire + 1 + 1 + 1 + 1 + 1)
        (Ast lock_acquire5_Prd lock_acquire5_grt 1) lock_acquire_psi)
     (lock_acquire + 1 + 1 + 1 + 1)
     (Ast lock_acquire4_Prd lock_acquire4_grt 1) (comml (_call l_acquire)).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux. right.
  split. apply l_acquire_not_primentry.
  unfold NH_Normal_Call.
  exists l_acquire_Prd. exists lock_acquire5_Prd.
  exists l_acquire_grt. exists lock_acquire5_grt.

  split. unfold SCAP_lookupCT. unfold SCAP_consCT.

  unfold SPSpec.lookup. unfold ListT.In. right.
  assert (SCAP_lookupCT lock_acquire_psi l_acquire (Ast l_acquire_Prd l_acquire_grt 1)).
  unfold lock_acquire_psi.
  auto. auto. split; auto.
  generalize (lock_acquire + 1 + 1 + 1 + 1).
  intro f.

  split.
  assert (forall sp l pc bp mP, enable (lock_acquire4_Prd_aux sp l pc bp mP) (nextS f (comml (_call l_acquire)))).
  apply lock_acquire4_enable.
  unfold enable in *.
  intros.
  unfold lock_acquire4_Prd in H0.
  InvertAll.
  apply H with x x0 x1 x2 x3; auto.
  split. apply lock_acquire_step5_call2.
  split. unfold pimp. unfold linkPG.
  intros. InvertAll.
  unfold lock_acquire4_Prd in H.
  InvertAll.
  unfold l_acquire_Prd.
  mexists5 (x0 - 2) x1 (f+1) (S_RF S BP) (l_acquire_closure_step5 x0 x1 x2 x3 x4).
  apply lock_acquire_step5_call3 with x; auto.
  split. intro.
  apply lock_acquire_step5_call4.
  intro. apply lock_acquire_step5_call5.

  split. 
    apply IF_false_Prd_Itr_Stk.
    apply lock_acquire4_Prd_IF_false.
    apply lock_acquire4_Grt_Itr_Stk.
Qed.



Lemma lock_acquire_step6_enable:
  enable_sti lock_acquire5_Prd.
Proof.
  unfold lock_acquire5_Prd. unfold enable_sti.
  unfold lock_acquire5_Prd_aux.
  intros.
  induction S as (((M, R), fl), isr).
  destruct H as [sp [l [pc [bp [m [HRsp [Hmem [Hlock [Hfl Hisr]]]]]]]]].
  unfold lock_acquire5_MPre in Hmem.
  assert ((INV1 *^* INV0 *^* (lock_acquire6_MPre sp l pc bp *^* m)) M).
  sep on.
  sep matchH Hmem with 4.
  clear Hmem. rename H into Hmem.
  destruct Hmem as [M1 [M2 [HMsplit [Hinv Hrest]]]].
  exists M2. exists M1.
  UnfoldStateAll.
  split; auto with PMap.
  split. unfold pTrue. auto.
  unfold choice. right.
  split; auto.
  intro. unfold itr_enabled in H.
  UnfoldStateAll. inversion H.
  rewrite H0 in Hfl. inversion Hfl.
Qed.


Lemma lock_acquire_step6_Prd_aux:
forall sp l pc bp mP S S',
  lock_acquire5_Prd_aux sp l pc bp mP S
  -> gsti S S'
  -> lock_acquire6_Prd_aux sp l pc bp mP S'.
Proof.
  unfold lock_acquire5_Prd_aux.
  intros.
  induction S as (((M, R), fl), isr).
  destruct H as [HRsp [Hmem [Hlock [Hfl Hisr]]]].
  rename H0 into Htrans.
  induction S' as (((M', R'), fl'), isr').
  unfold gsti in Htrans.
  UnfoldStateAll.
  destruct Htrans as [Hpp [Hfl' [Hisr' HR']]].
  subst R' isr'.
  unfold prepost in Hpp.
  destruct Hpp as [M1 [M2 [M1' [HM1 [Hm1' [HsplitM HsplitM']]]]]].
  UnfoldStateAll.
  assert (M2 = M').
  unfold Aemp in Hm1'.
  rewrite Hm1' in HsplitM'. auto with PMap.
  subst M'. clear HsplitM' Hm1'  M1'.
  unfold choice in HM1.
  inversion_clear HM1.
  inversion H. unfold itr_enabled in H0.
  inversion_clear H0. unfold S_Flag in H2.
  rewrite H2 in Hfl. inversion Hfl.
  inversion_clear H. clear H0. rename H1 into Hinv.
  unfold lock_acquire5_MPre in Hmem.
  assert ((INV1 *^* INV0 *^* (lock_acquire6_MPre sp l pc bp *^* mP)) M).
  sep on. sep matchH Hmem with 4.
  clear Hmem. rename H into Hmem.
  destruct Hmem as [M1x [M2x [HMsplit' [Hinv' Hrest]]]].
  assert (M1 = M1x).
  assert (precise (INV1 *^* INV0)).
  apply mstar_presv_prec; auto.
  apply Prec1. apply Prec0.
  unfold precise in H.
  apply H with M; auto.
  apply MemMod.disjUnion_subseteq_left with M2; auto.
  apply MemMod.disjUnion_subseteq_left with M2x; auto.
  assert (M2 = M2x). apply MemMod.disjUnion_unique with M1 M1x M; auto.
  subst M2.
  unfold lock_acquire6_Prd_aux.
  split_l. split; auto.
Qed.

Lemma lock_acquire5_Prd_IF_false:
   forall S : State, lock_acquire5_Prd S -> S_Flag S _IF = false.
Proof.
  unfold lock_acquire5_Prd.
  intros.
  destruct H as [sp [l [oc [bp [mP Hpre]]]]].
  unfold lock_acquire5_Prd_aux in Hpre.
  induction S as (((M, R), fl), isr).
  InvertAll. auto.
Qed.

Lemma lock_acquire5_Grt_Itr_Stk:
   Grt_Itr_Stk lock_acquire5_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  induction S as (((M, R), fl), isr).
  unfold lock_acquire5_grt. unfold leadstoPQ2.
  intros.
  unfold lock_acquire5_Prd_aux in H4.
  UnfoldStateAll.
  InvertAll.
  unfold itr_enabled in H0.
  inversion_clear H0.
  UnfoldStateAll.
  rewrite H8 in H7. inversion H7.
Qed.



Lemma wff_cmd_lock_acquire6:
   NH_WFcmd
     (SCAP_consCT (lock_acquire + 1 + 1 + 1 + 1 + 1 + 1)
        (Ast lock_acquire6_Prd lock_acquire6_grt 1) lock_acquire_psi)
     (lock_acquire + 1 + 1 + 1 + 1 + 1)
     (Ast lock_acquire5_Prd lock_acquire5_grt 1) (comml _sti).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  exists lock_acquire6_Prd. exists lock_acquire6_grt.
  split; auto.

  split. apply lock_acquire_step6_enable.
  split.
  unfold pimp. unfold linkPG.
  intros. 
  destruct H as [S0 [Hpre Htrans]].
  unfold lock_acquire5_Prd in Hpre.
  destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
  unfold lock_acquire6_Prd.
  mexists5 sp l pc bp mP.
  apply lock_acquire_step6_Prd_aux with S0; auto.

  unfold gimp. unfold conjG. unfold conjPG.
  intros.
  destruct H as [S'' [[Hpre Htrans] Hgrt6]].
  unfold lock_acquire5_grt.
  unfold lock_acquire6_grt in Hgrt6.
  unfold leadstoPQ2 in *.
  intros.
  apply Hgrt6.
  apply lock_acquire_step6_Prd_aux with S; auto.
  split. 
    apply IF_false_Prd_Itr_Stk.
    apply lock_acquire5_Prd_IF_false.
    apply lock_acquire5_Grt_Itr_Stk.
Qed.

Lemma lock_acquire_step7:
forall f sp l pc bp mP S S',
  lock_acquire6_Prd_aux sp l pc bp mP S  
  -> nextS f (comml (_addw (Opr_imm 2) SP)) S S'
  -> lock_acquire7_Prd_aux (sp + 2) l pc bp mP S'.
Proof.
  intros.
  rename H0 into Htrans.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion Htrans).
  subst M'.
  destruct H1 as [HR' [Hfl' Hisr']].
  unfold opr_eval in HR'.
  subst R' isr'.
  clear Htrans.
  unfold lock_acquire6_Prd_aux in H.
  destruct H as [HRsp [Hmem [Hlock [Hfl Hisr]]]].
  unfold lock_acquire7_Prd_aux.
  reg simpl.
  rewrite HRsp.
  split; auto with arith.
  split; auto with arith.
  assert (sp + 2 - 2 = sp). auto.
  rewrite H.
  unfold lock_acquire6_Prd_aux.
  reg simpl.
  split; auto.
  split; auto. split; auto. split_r.
  unfold trash_fl in Hfl'.
  inversion Hfl'.
  subst fl'. unfold Set_Flag. simpl. auto.
  subst fl'. unfold Clear_Flag. simpl. auto.
Qed.


Lemma lock_acquire6_Prd_Itr_Stk : Prd_Itr_Stk lock_acquire6_Prd.
Proof.
    unfold Prd_Itr_Stk.
    intros S Hpre Henable.
    unfold lock_acquire6_Prd in Hpre.
    destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
    unfold lock_acquire6_Prd_aux in Hpre.
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in Henable.
    simplS in * .
    destruct Hpre as [HRsp [Hmem [Hlock [Hfl Hisr]]]].
    unfold lock_acquire6_MPre in Hmem.
    sep on.
    sep lift 4 in Hmem.
    sep stksplit (IH_Stk + 6) in Hmem.
    sep lift 0 in Hmem.
    destruct Hmem as [M1 [M2 [Hsplit [Hmem1 Hmem2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    rewrite HRsp.
    split_l.
    intros M1' Hmem1'.
    unfold lock_acquire6_Prd.
    mexists5 sp l pc bp mP.
    unfold lock_acquire6_Prd_aux.
    unfold lock_acquire6_MPre.
    split_l.
    split; auto.
    sep on.
    sep lift 4.
    assert (sp >= IH_Stk + 6).
    apply stkSpacePrec_sp_ge_sz with M1; auto.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      split.
        disjoint rep_l M1.
        apply stkSpacePrec_stkSpacePrec_eqDom with sp (IH_Stk + 6); auto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      apply stkSpacePrec_stkSpacePrec_eqDom with sp (IH_Stk + 6); auto.
    split_l.
    sep on.
    sep matchH Hmem2 with 7.
Qed.

Lemma lock_acquire6_Grt_Itr_Stk : Grt_Itr_Stk lock_acquire6_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros S S' Hgrt Henable.
  unfold lock_acquire6_grt in * .
  intros M1 M2 M2' Hsplit HM2 HM2'.
  unfold leadstoPQ2 in *.
  intros sp l pc bp mP Hpre.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in Henable.
  apply Hgrt.
  clear Hgrt.
  unfold lock_acquire6_Prd_aux in * .
  simplS in * .
  destruct Hpre as [HRsp [Hmem [Hlock [Hfl Hisr]]]].
  split_l.
  split; auto.
  unfold lock_acquire6_MPre in *.
  sep on.
  sep lift 4 in Hmem.
  sep stksplit (IH_Stk + 6) in Hmem.
  sep lift 0 in Hmem.
  destruct Hmem as [M2x [M1x [Hsplit' [HM2x HM1x]]]].
  sep lift 4.
  assert (sp >= IH_Stk + 6).
  apply stkSpacePrec_sp_ge_sz with M2x; auto.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  rewrite HRsp in HM2, HM2'.
  assert (M2x = M2' /\ M1x = M1).
    sep off.
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      apply stkSpacePrec_stkSpacePrec_eqDom with sp (IH_Stk + 6); auto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    assert (Hx := prec_disjUnion_inv H0 H1 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H0; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  split_l.
  sep on.
  sep matchH HM1x with 7.
Qed.

Lemma wff_cmd_lock_acquire7:
   NH_WFcmd
     (SCAP_consCT (lock_acquire + 1 + 1 + 1 + 1 + 1 + 1 + 1)
        (Ast lock_acquire7_Prd lock_acquire7_grt 1) lock_acquire_psi)
     (lock_acquire + 1 + 1 + 1 + 1 + 1 + 1)
     (Ast lock_acquire6_Prd lock_acquire6_grt 1)
     (comml (_addw (Opr_imm 2) SP)).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists lock_acquire7_Prd. exists lock_acquire7_grt.
  split. auto.

  generalize (lock_acquire + 1 + 1 + 1 + 1 + 1 + 1).
  intro f.
  split.
  apply trivial_enable_addw.


  split.
  unfold pimp. unfold linkPG.
  intros. 
  destruct H as [S0 [Hpre Htrans]].
  unfold lock_acquire6_Prd in Hpre.
  destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
  unfold lock_acquire7_Prd.
  mexists5 (sp + 2) l pc bp mP.
  apply lock_acquire_step7 with f S0; auto.

  unfold gimp. unfold conjG. unfold conjPG.
  intros.
  destruct H as [S'' [[Hpre Htrans] Hgrt7]].
  unfold lock_acquire6_grt.
  unfold lock_acquire7_grt in Hgrt7.
  unfold leadstoPQ2 in *.
  intros.
  assert (sp + 2 + 2 = sp + 4). auto.
  rewrite <- H0.
  apply Hgrt7.
  apply lock_acquire_step7 with f S; auto.


  split.
  apply lock_acquire6_Prd_Itr_Stk.
  apply lock_acquire6_Grt_Itr_Stk.
Qed.



Lemma lock_acquire7_Prd_Itr_Stk : Prd_Itr_Stk lock_acquire7_Prd.
Proof.
    unfold Prd_Itr_Stk.
    intros S Hpre Henable.
    unfold lock_acquire7_Prd in Hpre.
    destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
    unfold lock_acquire7_Prd_aux in Hpre.
    destruct S as [[[M R] fl] isr].
    unfold lock_acquire6_Prd_aux in Hpre.
    unfold itr_enabled in Henable.
    simplS in * .
    reg simpl.
    destruct Hpre as [HRsp [Hspgt2 [Htmp [Hmem [Hlock [Hfl Hisr]]]]]].
    unfold lock_acquire6_MPre in Hmem.
    sep on.
    sep lift 4 in Hmem.
    assert (_STK_SIZE_lock_acquire - 4 >= IH_Stk + 6 - 2).
    unfold _STK_SIZE_lock_acquire. unfold _STK_SIZE_l_acquire. omega.
    sep stksplit (IH_Stk + 6 - 2) in Hmem.
    sep lift 0 in Hmem.
    assert (M ||- stkSpacePrec (sp - 2) (IH_Stk + 6 - 2) *^* sp - 2 |->w_ *^* 
                  (stkSpacePrec (sp - 2 - (IH_Stk + 6 - 2))
                    (_STK_SIZE_lock_acquire - 4 - (IH_Stk + 6 - 2)) *^*
                    (Gamma l *^* mP *^*
                      (sp - 2 + 2 |->w bp *^* sp - 2 + 4 |->w pc *^*
                        sp - 2 + 6 |->w l)))).
    sep matchH Hmem with 8.
    clear Hmem. rename H0 into Hmem.
    destruct Hmem as [M1 [M2 [Hsplit [Hmem1 Hmem2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    split.
    rewrite HRsp.
    sep on.
    sep stksplit. auto.
    intros M1' Hmem1'.
    unfold lock_acquire7_Prd.
    mexists5 sp l pc bp mP.
    unfold lock_acquire7_Prd_aux.
    unfold lock_acquire6_Prd_aux.
    unfold lock_acquire6_MPre.
    reg simpl.
    split_l.
    split; auto. split_l. split; auto.
    sep on.
    sep lift 4.
    assert (sp - 2 >= IH_Stk + 6 - 2).
    apply stkSpacePrec_sp_ge_sz1 with M1 (sp - 2 |->w_); auto.
    sep stksplit (IH_Stk + 6 - 2).
    sep lift 1.
    assert (MemMod.merge M2 M1'
           ||- stkSpacePrec (sp - 2) (IH_Stk + 6 - 2) *^* sp - 2 |->w_ *^* 
             (Gamma l *^* mP *^*
               (sp - 2 + 2 |->w bp *^* sp - 2 + 4 |->w pc *^*
                 sp - 2 + 6 |->w l) *^*
               stkSpacePrec (sp - 2 - (IH_Stk + 6 - 2))
               (_STK_SIZE_lock_acquire - 4 - (IH_Stk + 6 - 2)))).
    exists M1'; exists M2.
    assert (stkSpacePrec sp (IH_Stk + 6) M1).
    sep on.
    sep stksplit. auto.
    split.
      split.
        disjoint rep_l M1.
        apply stkSpacePrec_stkSpacePrec_eqDom with sp (IH_Stk + 6); auto.
        rewrite <- HRsp. auto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      apply stkSpacePrec_stkSpacePrec_eqDom with sp (IH_Stk + 6); auto.
      rewrite <- HRsp. auto.
    split. 
    sep on. assert (M1' ||- stkSpacePrec (R SP) (IH_Stk + 6)). auto.
    rewrite HRsp in H2. 
    sep stksplit in H2. auto.
    sep on.
    sep matchH Hmem2 with 6. sep matchH H1 with 8.
Qed.

Lemma lock_acquire7_Grt_Itr_Stk : Grt_Itr_Stk lock_acquire7_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros S S' Hgrt Henable.
  unfold lock_acquire7_grt in * .
  intros M1 M2 M2' Hsplit HM2 HM2'.
  unfold leadstoPQ2 in *.
  intros sp l pc bp mP Hpre.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in Henable.
  apply Hgrt.
  clear Hgrt.
  unfold lock_acquire7_Prd_aux in * .
  unfold lock_acquire6_Prd_aux in * .
  simplS in * .
  reg simpl.
  destruct Hpre as [HRsp [Hspgt2 [Htmp [Hmem [Hlock [Hfl Hisr]]]]]].
  split_l. split_l. split_l.
  split; auto.
  unfold lock_acquire6_MPre in *.
  sep on.
  sep lift 4 in Hmem.
  assert (_STK_SIZE_lock_acquire - 4 >= IH_Stk + 6 - 2).
  unfold _STK_SIZE_lock_acquire. unfold _STK_SIZE_l_acquire. omega.
  sep stksplit (IH_Stk + 6 - 2) in Hmem.
  sep lift 0 in Hmem.
  assert (MemMod.merge M1 M2' ||- stkSpacePrec (sp - 2) (IH_Stk + 6 - 2) *^* sp - 2 |->w_ *^* 
    (stkSpacePrec (sp - 2 - (IH_Stk + 6 - 2))
      (_STK_SIZE_lock_acquire - 4 - (IH_Stk + 6 - 2)) *^*
      (Gamma l *^* mP *^*
        (sp - 2 + 2 |->w bp *^* sp - 2 + 4 |->w pc *^*
          sp - 2 + 6 |->w l)))).
  sep matchH Hmem with 8.
  clear Hmem. rename H0 into Hmem.
  destruct Hmem as [M2x [M1x [Hsplit' [HM2x HM1x]]]].
  sep lift 4.
  assert (sp - 2 >= IH_Stk + 6 - 2).
  apply stkSpacePrec_sp_ge_sz1 with M2x (sp - 2 |->w_); auto.
  sep stksplit (IH_Stk + 6 - 2).
  sep lift 1.
  assert (M
          ||- stkSpacePrec (sp - 2) (IH_Stk + 6 - 2) *^* sp - 2 |->w_ *^* 
            (Gamma l *^* mP *^*
              (sp - 2 + 2 |->w bp *^* sp - 2 + 4 |->w pc *^*
                sp - 2 + 6 |->w l) *^*
              stkSpacePrec (sp - 2 - (IH_Stk + 6 - 2))
              (_STK_SIZE_lock_acquire - 4 - (IH_Stk + 6 - 2)))).
  rewrite HRsp in HM2, HM2'.
  assert (M2x = M2' /\ M1x = M1).
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      apply stkSpacePrec_stkSpacePrec_eqDom with sp (IH_Stk + 6); auto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    assert (M2x ||- stkSpacePrec sp (IH_Stk + 6)).
    sep stksplit. sep on. auto.
    assert (Hx := prec_disjUnion_inv H1 H2 HM2' H3 (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H1; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  split.
  assert (M2 ||- stkSpacePrec sp (IH_Stk + 6)). auto.
  sep on.
  sep stksplit in H1. auto.
  sep on.
  sep matchH HM1x with 6.
  sep matchH H1 with 8.
Qed.


Lemma wff_cmd_lock_acquire8:
   NH_WFcmd
     (SCAP_consCT (lock_acquire + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
        (Ast lock_acquire_Post lock_acquire_last_grt 1) lock_acquire_psi)
     (lock_acquire + 1 + 1 + 1 + 1 + 1 + 1 + 1)
     (Ast lock_acquire7_Prd lock_acquire7_grt 1) (comml (_popw BP)).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists lock_acquire_Post. exists lock_acquire_last_grt.
  split. auto.

  generalize (lock_acquire + 1 + 1 + 1 + 1 + 1 + 1 + 1). intro f.
  assert (forall sp l pc bp mP, enable (lock_acquire7_Prd_aux sp l pc bp mP) (nextS f (comml (_popw BP)))).
  intros.
  apply enable_popw.
  intros. assert ((S_RF S SP |~>w bp) (S_Mem S)).
Lemma lock_acquire7_Prd_1:
  forall sp l pc bp mP S, lock_acquire7_Prd_aux sp l pc bp mP S -> (S_RF S SP |~>w bp) (S_Mem S).
Proof.
  intros.
  unfold lock_acquire7_Prd_aux in H.
  induction S as (((M, R), fl), isr).
  InvertAll.
  UnfoldStateAll.
  unfold lock_acquire6_Prd_aux in H2.
  unfold lock_acquire6_MPre in H2. reg simpl.
  InvertAll.
  sep on.
  assert (sp - 2 + 2 = sp). auto.
  rewrite H5 in H2.
  sep lift 1 in H2. rewrite H0.
  sep trivial.
Qed.

  apply lock_acquire7_Prd_1 with sp l pc mP. auto.
  apply leadsToW_leadsToxW with bp; auto.
  split.
  unfold enable in *.
  intros.
  unfold lock_acquire7_Prd in H0.
  InvertAll.
  apply H with x x0 x1 x2 x3; auto.

Lemma lock_acquire_step8:
forall f sp l pc bp mP S S',
  lock_acquire7_Prd_aux sp l pc bp mP S  
  -> nextS f (comml (_popw BP)) S S'
  -> lock_acquire_Post_aux (sp + 2) l pc bp mP S'.
Proof.
  intros.
  rename H0 into Htrans.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion Htrans).
  subst M'.
  destruct H1 as [[x [Hmemx HR']] [Hfl' Hisr']].
  assert (x = bp).
  assert ((S_RF (M, R, fl, isr) SP |~>w bp) (S_Mem (M, R, fl, isr))).
  apply lock_acquire7_Prd_1 with sp l pc mP; auto.
  UnfoldStateAll.
  apply leadsToW_unique with (R SP) M; auto.
  subst x.
  subst R' fl' isr'.
  unfold lock_acquire_Post_aux.
  reg simpl.
  clear Htrans.
  unfold lock_acquire7_Prd_aux in H.
  InvertAll.
  split. rewrite H0. auto.
  split_l.
  unfold lock_acquire6_Prd_aux in H2.
  reg simpl.
  InvertAll.
  split; auto.
  unfold lock_acquire6_MPre in H2.
  unfold lock_acquire_MPost.
  unfold lock_acquire_MPre.
  sep on.
  sep lift 2.
  sep stksplit.
  assert (sp + 2 - 2 = sp). auto.
  rewrite H5.
  sep stksplit.
  assert (sp - 2 + 2 = sp). auto.
  rewrite H7 in H2.
  assert (_STK_SIZE_lock_acquire - 2 - 2 = _STK_SIZE_lock_acquire - 4).
  auto. rewrite H8.
  assert (sp + 2 + 2 = sp + 4). auto.
  rewrite H9.
  assert (sp - 2 + 4 = sp + 2). auto.
  rewrite H10 in H2.
  assert (sp - 2 + 6 = sp + 4). auto.
  rewrite H11 in H2.
  sep matchH H2 with 7.
Qed.

  split.
  unfold pimp.
  unfold linkPG.
  intros.
  InvertAll.
  unfold lock_acquire7_Prd in H0.
  InvertAll.
  unfold lock_acquire_Post.
  exists (x0 + 2). exists x1. exists x2. exists x3. exists x4.
  apply lock_acquire_step8 with f x; auto.

  unfold gimp.
  unfold conjPG. unfold conjG.
  intros.
  InvertAll.
  unfold lock_acquire7_grt.
  unfold leadstoPQ2.
  intros.
  unfold lock_acquire_last_grt in H2.
  unfold leadstoPQ2 in H2.
  apply H2.
  apply lock_acquire_step8 with f S; auto.


  split.
  apply lock_acquire7_Prd_Itr_Stk.
  apply lock_acquire7_Grt_Itr_Stk.
Qed.


Lemma lock_acquirePost_Prd_Itr_Stk : Prd_Itr_Stk lock_acquire_Post.
Proof.
    unfold Prd_Itr_Stk.
    intros S Hpre Henable.
    unfold lock_acquire_Post in Hpre.
    destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
    unfold lock_acquire_Post_aux in Hpre.
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in Henable.
    simplS in * .
    destruct Hpre as [HRsp [HRbp [Hmem [Hlock [Hfl Hisr]]]]].
    unfold lock_acquire_MPost in Hmem.
    unfold lock_acquire_MPre in Hmem.
    sep on.
    sep lift 2 in Hmem.
    sep stksplit (IH_Stk + 6) in Hmem.
    sep lift 0 in Hmem.
    destruct Hmem as [M1 [M2 [Hsplit [Hmem1 Hmem2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    rewrite HRsp.
    split_l.
    intros M1' Hmem1'.
    unfold lock_acquire_Post.
    mexists5 sp l pc bp mP.
    unfold lock_acquire_Post_aux.
    unfold lock_acquire_MPost.
    unfold lock_acquire_MPre.
    split_l. split_l.
    split; auto.
    sep on.
    sep lift 2.
    assert (sp >= IH_Stk + 6).
    apply stkSpacePrec_sp_ge_sz with M1; auto.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      split.
        disjoint rep_l M1.
        apply stkSpacePrec_stkSpacePrec_eqDom with sp (IH_Stk + 6); auto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      apply stkSpacePrec_stkSpacePrec_eqDom with sp (IH_Stk + 6); auto.
    split_l.
    sep on.
    sep matchH Hmem2 with 5.
Qed.

Lemma lock_acquirePost_Grt_Itr_Stk : Grt_Itr_Stk lock_acquire_last_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros S S' Hgrt Henable.
  unfold lock_acquire_last_grt in * .
  intros M1 M2 M2' Hsplit HM2 HM2'.
  unfold leadstoPQ2 in *.
  intros sp l pc bp mP Hpre.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in Henable.
  apply Hgrt.
  clear Hgrt.
  unfold lock_acquire_Post_aux in * .
  simplS in * .
  destruct Hpre as [HRsp [HRbp [Hmem [Hlock [Hfl Hisr]]]]].
  split_l. split_l.
  split; auto.
  unfold lock_acquire_MPost in *.
  unfold lock_acquire_MPre in *.
  sep on.
  sep lift 2 in Hmem.
  sep stksplit (IH_Stk + 6) in Hmem.
  sep lift 0 in Hmem.
  destruct Hmem as [M2x [M1x [Hsplit' [HM2x HM1x]]]].
  sep lift 2.
  assert (sp >= IH_Stk + 6).
  apply stkSpacePrec_sp_ge_sz with M2x; auto.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  rewrite HRsp in HM2, HM2'.
  assert (M2x = M2' /\ M1x = M1).
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      apply stkSpacePrec_stkSpacePrec_eqDom with sp (IH_Stk + 6); auto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    assert (Hx := prec_disjUnion_inv H0 H1 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H0; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  split_l.
  sep on.
  sep matchH HM1x with 5.
Qed.

Lemma wff_cmd_lock_acquire9:
 NH_WFiseq lock_acquire_psi (lock_acquire + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
   (Ast lock_acquire_Post lock_acquire_last_grt 1) 
   (sseq _retw).
Proof.
  apply  nh_isq_end.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  split; auto.
  intros.
  unfold lock_acquire_last_grt.
  intros.
  unfold leadstoPQ2. auto.
  split.
  apply lock_acquirePost_Prd_Itr_Stk.
  apply lock_acquirePost_Grt_Itr_Stk.
Qed.


Lemma wff_lock_acquire:
  NH_WFiseq lock_acquire_psi lock_acquire (Ast lock_acquire_Prd lock_acquire_grt 1) I_lock_acquire.
Proof.
  unfold I_lock_acquire.
  apply nh_isq_seq with lock_acquire1_Prd lock_acquire1_grt.
  apply wff_cmd_lock_acquire1.
  apply nh_isq_seq with lock_acquire2_Prd lock_acquire2_grt.
  apply wff_cmd_lock_acquire2.
  apply nh_isq_seq with lock_acquire3_Prd lock_acquire3_grt.
  apply wff_cmd_lock_acquire3.
  apply nh_isq_seq with lock_acquire4_Prd lock_acquire4_grt.
  apply wff_cmd_lock_acquire4.
  apply nh_isq_seq with lock_acquire5_Prd lock_acquire5_grt.
  apply wff_cmd_lock_acquire5.
  apply nh_isq_seq with lock_acquire6_Prd lock_acquire6_grt.
  apply wff_cmd_lock_acquire6.
  apply nh_isq_seq with lock_acquire7_Prd lock_acquire7_grt.
  apply wff_cmd_lock_acquire7.
  apply nh_isq_seq with lock_acquire_Post lock_acquire_last_grt.
  apply wff_cmd_lock_acquire8.
  apply wff_cmd_lock_acquire9.
Qed.
