(***************************************************************)
(* Coq code for AIM                                            *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

Add LoadPath "..".
Add LoadPath "../CAP0".
Add LoadPath "../SCAP".

Require Import Coqlib.

Require Import Arith.
Require Import Machine.
Require Import Assertion.
Require Import Operational_Semantics.
Require Import OpSem_Lemmas.
Require Import Reg.
Require Import Assertion_Lemmas3.
Require Import Assertion_Lemmas4.

Require Import AIM.
Require Import SCAP.
Require Import QLib.
Require Import Linking.
Require Import Linking_Lemmas3.

Require Import Kernel.
Require Import BlockSpec.

(********   Block  *************************** 
# precondition requires that interrupt be disabled
# block(q)
1158 <block>:
        movw    (curr_thrd), %bx	#curr_thrd = 2254
        movw    $BLOCKED, TCB_STATUS(%bx)
1160    movw    %sp, TCB_STACK(%bx)
        movw    %sp, %bp
        movw    2(%bp), %ax
        pushw   %ax			#BlockQ
        pushw   %bx
1165    call    enQueue	(1190)
        pushw   $readyQ			#readyQ = 2250
        call    deQueue (1212)
# note that we can always find a thread in the readyQ
# because our program logic guarantees the idle thread
# will never be blocked. So the idle thread must always
# be in the ready queue at this moment
        addw    $6, %sp
        movw    %ax, %bx
1170    movw    %ax, (curr_thrd)
        movw    TCB_STACK(%bx), %sp
1172    retw
************************************************)

(*  <block+0> : movw (curr_thrd), %bx  #curr_thrd = 2254 *)

Lemma SCAP_WFcmd_block_0_movwld :
  SCAP_WFcmd (SCAP_consCT (L_block+1) (Ast p_block_1 g_block_1 sctag) psi_block)
    (L_block) (Ast p_block g_block sctag) 
    (comml (_movwld (A_imm curr_thrd) BX)).
Proof.
  assert (Hpp' : forall T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S S',
    Ast_block_pre T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S
    -> nextS (L_block) (comml (_movwld (A_imm curr_thrd) BX)) S S'
    -> Ast_block_1 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S').
    intros T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S S' HS HSS'.
    unfold Ast_block_pre, Ast_block_1 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HR' as [tidx [Hlds HR']].
    simplS in * .
    destruct HS as [HM [Hsp [Hspg [HRT [HtR [Hex [Htcb [Hmto HT]]]]]]]].
    assert (Heq : tidx = tid).
      unfold addr_eval in Hlds.
      sep on.
      sep lift 0 in HM.
      sep inj.
      auto.
    subst tidx M' R' fl' isr'.
    reg simpl.
    repeat split_l.
    trivial.
  simpl.
  exists p_block_1; exists g_block_1.
  split_l.
  split.
  (* enable *)
    unfold enable.
    intros S HS.
    destruct S as [[[M R] fl] isr].
    unfold p_block in HS.
    destruct HS as [T [RdyQ [BlkQ [l_ptr [tid [tcb [sp [pc [jk1 [jk2 HS]]]]]]]]]].
    unfold Ast_block_pre in HS.
    destruct HS as [HM [Hsp [Hspg [HRT [HtR [Hex [Htcb Hmto]]]]]]].
    simplS in * .
    exists (M, RF_update R BX tid, fl, isr).
    apply nextS_movwld with curr_thrd tid; trivial.
    sep on.
    sep lift 0 in HM.
    sep trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold p_block, p_block_1.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [T [RdyQ [BlkQ [l_ptr [tid [tcb [sp [pc [jk1 [jk2 HS]]]]]]]]]]. 
    exists T; exists RdyQ; exists BlkQ; exists l_ptr.
    exists tid; exists tcb; exists sp; exists pc; exists jk1; exists jk2.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold g_block, g_block_1 in * .
  intros mP T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 HS.
  eapply HgS'S''; eauto.
Qed.

(* <block + 1> :  movw $BLOCKED, TCB_STATUS(%bx) *)

Lemma SCAP_WFcmd_block_1_movwst :
  SCAP_WFcmd (SCAP_consCT (L_block+1+1) (Ast p_block_2 g_block_2 sctag) psi_block)
    (L_block+1) (Ast p_block_1 g_block_1 sctag) 
    (comml (_movwst (Opr_imm 2) (A_reg TCB_STATUS BX))).
Proof.
  assert (Hpp' : forall T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S S',
    Ast_block_1 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S
    -> nextS (L_block+1) (comml (_movwst (Opr_imm 2) (A_reg TCB_STATUS BX))) S S'
    -> Ast_block_2 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S').
    intros T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S S' HS HSS'.
    unfold Ast_block_1, Ast_block_2 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hlds HM'].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simpl in Hlds.
    simplS in * .
    destruct HS as [HM [Hbx [Hsp [Hspg [HRT [HtR [Hex [Htcb [Hmto HT]]]]]]]]].
    subst M' R' fl' isr'.
    split.
      sep on.
      unfold TCB_STATUS.
      unfold embed_TCB in * .
      sep lift 3 in HM.
      sep lift 3.
      apply ptoTCB_intro1 with 2.
        substH HM with (ptoTCB_elim1 HM).
        destruct HM as [tst [HM Htst]].
        sep lift 1 in HM.
        sep lift 1.
        arith_rep (tid + 4 + 2) with (tid + 6).
        arith_rep (tid + 4 + 2) with (tid + 6) in HM.
        rewrite Hbx.
        sep upd.
        destruct tcb as [[spx statusx] pcx].
        unfold TCB_SP, TCB_Status_Upd in * .
        sep matchH HM with 8.
      destruct tcb as [[spx statusx] pcx].
      unfold TCB_SP, TCB_Status, TCB_Status_Upd in * .
      simpl; trivial.
    repeat split_l; trivial.
  simpl.
  exists p_block_2; exists g_block_2.
  split_l.
  split.
    (* enable movw *)
    unfold enable, p_block_1.
    intros S H.
    destruct H as [T [RdyQ [BlkQ [l_ptr [tid [tcb [sp [pc [jk1 [jk2 HS]]]]]]]]]].
    unfold Ast_block_1 in HS.
    destruct HS as [HM [Hbx [Hsp [Hspg [HRT [HtR [Hex [Htcb [Hmto HT]]]]]]]]].
    destruct S as [[[M R] fl] isr].
    simplS in * .
    set (M' := MemUpdW M (tid + 6) 2 0).
    exists (M', R, fl, isr).
    apply nextS_movwst with (tid + 6) 2; auto.
      unfold addr_eval, TCB_STATUS.
      rewrite Hbx; trivial.
      unfold embed_TCB in HM.
      sep on.
      sep lift 3 in HM.
      substH HM with (ptoTCB_elim1 HM).
      destruct HM as [tst [HM Htst]].
      sep lift 1 in HM.
      arith_rep (tid + 4 + 2) with (tid + 6) in HM.
      sep trivial.
    apply MemUpdW_MemUpd16 with 0; trivial.
  split. (* p ==> p' *)
    unfold p_block_1, p_block_2.
    intros S' H.
    unfold linkPG in H.
    destruct H as [S H].
    destruct H as [H HnextS].
    destruct H as [T [RdyQ [BlkQ [l_ptr [tid [tcb [sp [pc [jk1 [jk2 HS]]]]]]]]]].
    exists T; exists RdyQ; exists BlkQ; exists l_ptr. 
    exists tid; exists tcb; exists sp; exists pc; exists jk1; exists jk2.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold conjPG, conjG in H.
  destruct H as [S' [[_ HnextS] HSS']].
  unfold g_block_1, g_block_2 in * .
  intros mP T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 HS.
  eapply HSS'; eauto.
Qed.

(*  <block+2> : movw  %sp, TCB_STACK(%bx) *)

Lemma SCAP_WFcmd_block_2_movwst :
  SCAP_WFcmd (SCAP_consCT (L_block+2+1) (Ast p_block_3 g_block_3 sctag) psi_block)
    (L_block+2) (Ast p_block_2 g_block_2 sctag)
    (comml (_movwst (Opr_reg SP) (A_reg TCB_STACK BX))).
Proof.
  assert (Hpp' : forall T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S S',
    Ast_block_2 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S
    -> nextS (L_block+2) (comml (_movwst (Opr_reg SP) (A_reg TCB_STACK BX))) S S'
    -> Ast_block_3 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S').
    intros T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S S' HS HSS'.
    unfold Ast_block_2, Ast_block_3 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hlds HM'].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simpl in Hlds.
    simplS in * .
    destruct HS as [HM [Hbx [Hsp [Hspg [HRT [HtR [Hex [Htcb [Hmto HT]]]]]]]]].
    subst M' R' fl' isr'.
    destruct tcb as [[spx statusx] pcx].
    unfold TCB_Status, TCB_Status_Upd in * .
    split.
      unfold embed_TCB in * .
      sep on.
      unfold TCB_STACK.
      sep lift 3 in HM.
      sep lift 3.
      apply ptoTCB_intro1 with 2.
        substH HM with (ptoTCB_elim1 HM).
        destruct HM as [tst [HM Htst]].
        sep lift 0 in HM.
        sep lift 0.
        unfold TCB_SP in * .
        rewrite Hbx.
        rewrite Hsp.
        sep upd.
        assert (Htst' : tst = 2).
          destruct tst; simpl in Htst.
          discriminate Htst.
          destruct tst; simpl in Htst.
          discriminate Htst.
          destruct tst; simpl in Htst.
          trivial.
          discriminate Htst.
        subst tst.
        sep matchH HM with 8.
      simpl; trivial.
    repeat split_l; trivial.
  simpl.
  exists p_block_3; exists g_block_3.
  split_l.
  split.
    (* enable movw *)
    unfold enable, p_block_2.
    intros S H.
    destruct H as [T [RdyQ [BlkQ [l_ptr [tid [tcb [sp [pc [jk1 [jk2 HS]]]]]]]]]].
    unfold Ast_block_2 in HS.
    destruct HS as [HM [Hbx [Hsp [Hspg [HRT [HtR [Hex [Htcb [Hmto HT]]]]]]]]].
    destruct S as [[[M R] fl] isr].
    simplS in * .
    set (M' := MemUpdW M (tid + 4) sp 0).
    exists (M', R, fl, isr).
    apply nextS_movwst with (tid + 4) sp; auto.
      unfold addr_eval, TCB_STACK.
      rewrite Hbx; trivial.
      unfold embed_TCB in * .
      sep on.
      sep lift 3 in HM.
      substH HM with (ptoTCB_elim1 HM).
      destruct HM as [tst [HM Htst]].
      sep lift 0 in HM.
      sep trivial.
    apply MemUpdW_MemUpd16 with 0; trivial.
  split. (* p ==> p' *)
    unfold p_block_2, p_block_3.
    intros S' H.
    unfold linkPG in H.
    destruct H as [S H].
    destruct H as [H HnextS].
    destruct H as [T [RdyQ [BlkQ [l_ptr [tid [tcb [sp [pc [jk1 [jk2 HS]]]]]]]]]]. 
    exists T; exists RdyQ; exists BlkQ; exists l_ptr. 
    exists tid; exists tcb; exists sp; exists pc; exists jk1; exists jk2.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold conjPG, conjG in H.
  destruct H as [S' [[_ HnextS] HSS']].
  unfold g_block_2, g_block_3 in * .
  intros mP T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 HS.
  eapply HSS'; eauto.
Qed.

(*  <block+3> :  movw    %sp, %bp *)

Lemma SCAP_WFcmd_block_3_movwp :
  SCAP_WFcmd (SCAP_consCT (L_block+3+1) (Ast p_block_4 g_block_4 sctag) psi_block)
    (L_block+3) (Ast p_block_3 g_block_3 sctag)
    (comml (_movwp  (Opr_reg SP) BP)).
Proof.
  assert (Hpp' : forall T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S S',
    Ast_block_3 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S
    -> nextS (L_block+3) (comml (_movwp  (Opr_reg SP) BP)) S S'
    -> Ast_block_4 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S').
    intros T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S S' HS HSS'.
    unfold Ast_block_3, Ast_block_4 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    simplS in * .
    destruct HS as [HM [Hbx [Hsp [Hspg [HRT [HtR [Hex [Htcb [Hmto HT]]]]]]]]].
    subst M' R' fl' isr'.
    reg simpl.
    repeat split_l; trivial.
  simpl.
  exists p_block_4; exists g_block_4.
  split_l.
  split.
    (* enable movw *)
    unfold enable, p_block_3.
    intros S H.
    destruct H as [T [RdyQ [BlkQ [l_ptr [tid [tcb [sp [pc [jk1 [jk2 HS]]]]]]]]]].
    unfold Ast_block_3 in HS.
    destruct HS as [HM [Hbx [Hsp [Hspg [HRT [HtR [Hex [Htcb [Hmto HT]]]]]]]]].
    destruct S as [[[M R] fl] isr].
    simplS in * .
    set (R' := RF_update R BP sp).
    exists (M, R', fl, isr).
    apply nextS_movwp with sp; auto.
  split. (* p ==> p' *)
    unfold p_block_3, p_block_4.
    intros S' H.
    unfold linkPG in H.
    destruct H as [S H].
    destruct H as [H HnextS].
    destruct H as [T [RdyQ [BlkQ [l_ptr [tid [tcb [sp [pc [jk1 [jk2 HS]]]]]]]]]].
    exists T; exists RdyQ; exists BlkQ; exists l_ptr; exists tid.
    exists tcb; exists sp; exists pc; exists jk1; exists jk2.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold conjPG, conjG in H.
  destruct H as [S' [[_ HnextS] HS'S'']].
  unfold g_block_3, g_block_4 in * .
  intros mP T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 HS.
  eapply HS'S''; eauto.
Qed.

(*  <block+4> :  movw    2(%bp), %ax *)

Lemma SCAP_WFcmd_block_4_movwld :
  SCAP_WFcmd (SCAP_consCT (L_block+4+1) (Ast p_block_5 g_block_5 sctag) psi_block)
    (L_block+4) (Ast p_block_4 g_block_4 sctag)
    (comml (_movwld (A_reg 2 BP) AX)).
Proof.
  assert (Hpp' : forall T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S S',
    Ast_block_4 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S
    -> nextS (L_block+4) (comml (_movwld (A_reg 2 BP) AX)) S S'
    -> Ast_block_5 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S').
    intros T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S S' HS HSS'.
    unfold Ast_block_4, Ast_block_5 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HR' as [l_ptrx [Hlds HR']].
    simplS in * .
    destruct HS as [HM [Hbx [Hsp [Hbp [Hspg [HRT [HtR [Hex [Htcb [Hmto HT]]]]]]]]]].
    assert (Heq : l_ptrx = l_ptr).
      unfold addr_eval in Hlds.
      rewrite Hbp in Hlds.
      sep on.
      sep lift 7 in HM.
      sep inj.
      auto.
    subst l_ptrx M' R' fl' isr'.
    reg simpl.
    repeat split_l; trivial.
  simpl.
  exists p_block_5; exists g_block_5.
  split_l.
  split.
  (* enable *)
    unfold enable.
    intros S HS.
    destruct S as [[[M R] fl] isr].
    unfold p_block_4 in HS.
    destruct HS as [T [RdyQ [BlkQ [ltr [tid [tcb [sp [pc [jk1 [jk2 HS]]]]]]]]]].
    unfold Ast_block_4 in HS.
    destruct HS as [HM [Hbx [Hsp [Hbp [Hspg [HRT [HtR [Hex [Htcb [Hmto HT]]]]]]]]]].
    simplS in * .
    exists (M, RF_update R AX ltr, fl, isr).
    apply nextS_movwld with (sp+2) ltr; auto.
      unfold addr_eval; rewrite Hbp; trivial.
    sep on.
    sep lift 7 in HM.
    apply ptoW_leadsToW1 with mTrue.
    sep subst_left with (sp + 2 |-> ltr *^* sp + 2 + 1 |->_).
    arith_rep (sp + 2 + 1) with (sp + 3).
    sep lift 0.
    sep match_left HM.
    sep lift 0 in HM.
    sep match_left HM.
    sep trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold p_block_4, p_block_5.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [T [RdyQ [BlkQ [ltr [tid [tcb [sp [pc [jk1 [jk2 HS]]]]]]]]]].
    exists T; exists RdyQ; exists BlkQ; exists ltr.
    exists tid; exists tcb; exists sp; exists pc; exists jk1; exists jk2.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold p_block_4, g_block_5 in * .
  intros mP T RdyQ BlkQ ltr tid tcb sp pc jk1 jk2 HS.
  eapply HgS'S''; eauto.
Qed.

(*  <block+5> : pushw   %ax  #BlockQ *)

Lemma SCAP_WFcmd_block_5_pushwp :
  SCAP_WFcmd (SCAP_consCT (L_block+5+1) (Ast p_block_6 g_block_6 sctag) psi_block)
    (L_block+5) (Ast p_block_5 g_block_5 sctag) 
    (comml (_pushwp (Opr_reg AX))).
Proof.
  assert (Hpp' : forall T RdyQ BlkQ ltr tid tcb sp pc jk1 jk2 mP S S',
    Ast_block_5 T RdyQ BlkQ ltr tid tcb sp pc jk1 jk2 mP S
    -> nextS (L_block+5) (comml (_pushwp (Opr_reg AX))) S S'
    -> Ast_block_6 T RdyQ BlkQ ltr tid tcb sp pc jk1 jk2 mP S').
    intros T RdyQ BlkQ ltr tid tcb sp pc jk1 jk2 mP S S' HS HSS'.
    unfold Ast_block_5, Ast_block_6 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspgt [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simplS in * .
    destruct HS as [HM [Hax [Hbx [Hsp [Hspg [HRT [HtR [Hex [Htcb [Hmto HT]]]]]]]]]].
    split.
      subst M' R' fl' isr'.
      sep on.
      sep lift 5.
      rewrite Hsp.
      rewrite Hax.
      sep upd.
      sep matchH HM with 7.
      sep stksplit in HM.
      sep trivial.
    subst R' fl' isr'; reg simpl.
    rewrite Hsp; rewrite Hbx.
    repeat split_l; trivial.
  simpl.
  exists p_block_6; exists g_block_6.
  split_l.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold p_block_5 in HpS.
    destruct HpS as [T [RdyQ [BlkQ [ltr [tid [tcb [sp [pc [jk1 [jk2 HS]]]]]]]]]].
    exists (MemUpdW M (sp-2) (R AX) 0, RF_update R SP (sp - 2), fl, isr).
    apply nextS_pushwp with (R AX); trivial.
    unfold pushW.
    destruct HS as [HM [Hax [Hbx [Hsp [Hspg [HRT [HtR [Hex [Htcb [Hmto HT]]]]]]]]]].
    simplS in * .
    rewrite Hsp; rewrite Hax.
    split.
      apply stkSpace_weaken with (_BLK_Stk);
        [idtac | omega].
      sep on.
      sep lift 4 in HM.
      sep trivial.
    split.
      apply MemUpdW_MemUpd16 with 0; trivial.
    trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold p_block_5, g_block_6.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [T [RdyQ [BlkQ [ltr [tid [tcb [sp [pc [jk1 [jk2 HS]]]]]]]]]].
    exists T; exists RdyQ; exists BlkQ; exists ltr; exists tid; 
      exists tcb; exists sp; exists pc; exists jk1; exists jk2.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold p_block_5, g_block_6 in * .
  intros mP T RdyQ BlkQ ltr tid tcb sp pc jk1 jk2 HS.
  eapply HgS'S''; eauto.
Qed.

(*  <block+6> :  pushw   %bx *)

Lemma SCAP_WFcmd_block_6_pushwp :
  SCAP_WFcmd (SCAP_consCT (L_block+6+1) (Ast p_block_7 g_block_7 sctag) psi_block)
    (L_block+6) (Ast p_block_6 g_block_6 sctag) 
    (comml (_pushwp (Opr_reg BX))).
Proof.
  assert (Hpp' : forall T RdyQ BlkQ ltr tid tcb sp pc jk1 jk2 mP S S',
    Ast_block_6 T RdyQ BlkQ ltr tid tcb sp pc jk1 jk2 mP S
    -> nextS (L_block+6) (comml (_pushwp (Opr_reg BX))) S S'
    -> Ast_block_7 T RdyQ BlkQ ltr tid tcb sp pc jk1 jk2 mP S').
    intros T RdyQ BlkQ ltr tid tcb sp pc jk1 jk2 mP S S' HS HSS'.
    unfold Ast_block_6, Ast_block_7 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspgt [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simplS in * .
    destruct HS as [HM [Hbx [Hsp [Hspg [HRT [HtR [Hex [Htcb [Hmto HT]]]]]]]]].
    split.
      subst M' R' fl' isr'.
      sep on.
      sep lift 5.
      rewrite Hsp.
      rewrite Hbx.
      arith_rep (sp - 2 - 2) with (sp - 4).
      sep upd.
      sep matchH HM with 10.
      sep stksplit in HM.
      arith_rep (sp - 2 - 2) with (sp - 4) in HM.
    subst R' fl' isr'; reg simpl.
    rewrite Hsp; rewrite Hbx.
    split_l.
    split_l by omega.
    repeat split_l; trivial.
  simpl.
  exists p_block_7; exists g_block_7.
  split_l.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold p_block_6 in HpS.
    destruct HpS as [T [RdyQ [BlkQ [ltr [tid [tcb [sp [pc [jk1 [jk2 HS]]]]]]]]]].
    exists (MemUpdW M (sp-4) (R BX) 0, RF_update R SP (sp - 4), fl, isr).
    apply nextS_pushwp with (R BX); trivial.
    unfold pushW.
    destruct HS as [HM [Hbx [Hsp [Hspg [HRT [HtR [Hex [Htcb [Hmto HT]]]]]]]]].
    simplS in * .
    rewrite Hsp; rewrite Hbx.
    split.
      apply stkSpace_weaken with (_BLK_Stk - 2);
        [idtac | omega].
      sep on.
      sep lift 4 in HM.
      sep trivial.
    split.
      apply MemUpdW_MemUpd16 with 0; trivial.
    arith_rep (sp - 2 - 2) with (sp - 4).
    arith_rep (sp - 2 - 2) with (sp - 4).
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold p_block_6, p_block_7.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [T [RdyQ [BlkQ [ltr [tid [tcb [sp [pc [jk1 [jk2 HS]]]]]]]]]].
    exists T; exists RdyQ; exists BlkQ; exists ltr; exists tid; 
      exists tcb; exists sp; exists pc; exists jk1; exists jk2.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold g_block_6, g_block_7 in * .
  intros mP T RdyQ BlkQ ltr tid tcb sp pc jk1 jk2 HS.
  eapply HgS'S''; eauto.
Qed.

(*  <block + 7> :  call    enQueue	(1190)  *)

Lemma SCAP_WFcmd_block_7_call :
  SCAP_WFcmd (SCAP_consCT (L_block+7+1) (Ast p_block_8 g_block_8 sctag) psi_block)
    (L_block+7) (Ast p_block_7 g_block_7 sctag) 
    (comml (_call    L_enQueue)).
Proof.
  assert(Hpp' : forall T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mQ mP S S',
    Ast_block_7 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S
    -> nextS (L_block+7) (comml (_call L_enQueue)) S S'
    -> mQ = (stkSpacePrec (sp - 8) 2
        *^* (curr_thrd |->w tid)
        *^* (embed_RdyQ T RdyQ)
        *^* (sp - 6 |->w (L_block+7+1))
        *^* (sp     |-> pc) 
        *^* (sp + 1 |-> jk1) 
        *^* (sp + 2 |-> l_ptr) 
        *^* (sp + 3 |-> jk2))
    -> Ast_enQueue_pre T BlkQ l_ptr tid (sp, BLOCKED, pc) (S_RF S' BP)
    (sp - 6) (mQ *^* mP) S').
    intros T RdyQ BlkQ ltr tid tcb sp pc jk1 jk2 mQ mP S S'.
    intros HS HSS' HmQ.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    unfold Ast_block_7 in HS.
    destruct HS as [HM [Hbx [Hsp [Hspg [HRT [HtR [Hex [Htcb [Hmto HT]]]]]]]]].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspx [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    subst M' R' fl' isr'.
    unfold Ast_enQueue_pre.
    simplS in * .
    reg simpl.
    split.
      sep on.
      rewrite Hsp.
      arith_rep (sp - 4 - 2) with (sp - 6).
      subst mQ.
      sep lift 8.
      sep upd.
      sep lift 4 in HM.
      sep stksplit in HM.
      arith_rep (sp - 4 - 2) with (sp - 6) in HM.
      sep stksplit 2 in HM.
      arith_rep (sp - 6 - 2) with (sp - 8) in HM.
      arith_rep (10 - 4 - 2 - 2) with 2 in HM.
      unfold embed_BlkQ in HM.
      arith_rep (sp - 6 + 2) with (sp - 4).
      arith_rep (sp - 6 + 4) with (sp - 2).
      sep matchH HM with 11.
    split.
      assert (tid <> 0).
        apply HT.
        exists tcb; trivial.
        clear HM Hbx Hsp Hspg Hex Htcb Hmto Hlds HSS'.
        induction tid.
          destruct (H (refl_equal _)).
        clear IHtid H.
        omega.
    rewrite Hsp; omega.
  assert (Hpp''' : forall T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 pc' mP S S' S'' S''',
    Ast_block_7 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S
    -> nextS (L_block+7) (comml (_call L_enQueue)) S S'
    -> g_enQueue S' S''
    -> nextS pc' (comms (_retw)) S'' S'''
    -> exists T', exists BlkQ',
      Ast_block_8 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 T' BlkQ' mP S''').
    intros T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 pc' mP.
    intros S S' S'' S''' HS HSS' HS'S'' HS''S'''.
    exists (TQMod.put T tid (sp, BLOCKED, pc)).
    exists (addThrd BlkQ tid).
    unfold g_enQueue in HS'S''.
    set (mQ := (stkSpacePrec (sp - 8) 2 
         *^* (curr_thrd |->w tid)
         *^* (embed_RdyQ T RdyQ)
         *^* (sp - 6 |->w (L_block+7+1))
        *^* (sp     |-> pc) 
        *^* (sp + 1 |-> jk1) 
        *^* (sp + 2 |-> l_ptr) 
        *^* (sp + 3 |-> jk2))).
    assert (HS' : Ast_enQueue_pre T BlkQ l_ptr tid (sp, BLOCKED, pc) (S_RF S' BP) (sp-6) (mQ *^* mP) S').
      apply Hpp' with RdyQ tcb jk1 jk2 S; trivial.
    assert (HgS'' := HS'S'' _ _ _ _ _ _ _ _ HS').
    destruct HgS'' as [T' [BlkQ' [HS'' [HT' HBlkQ']]]].
    subst mQ.
    clear HS'S''.
    unfold Ast_enQueue_post in HS''.
    unfold Ast_block_8.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct S'' as [[[M'' R''] fl''] isr''].
    destruct S''' as [[[M''' R'''] fl'''] isr'''].
    destruct HS as [HM [Hbx [Hsp [Hspg HS]]]]. 
    destruct HS as [HRT [HtR [Hex [Htcb [Hmto HT]]]]].
    simplS in * .
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct (nextS_inversion HS''S''') as [HM''' [HR''' [Hfl''' Hisr''']]].
    destruct HR''' as [pcx [Hlds HR''']].
    destruct HS'' as [HM'' [HBP'' HSP'']].
    assert (Hpcx : pcx = L_block + 7 + 1).
      subst R' M''' R'''.
      clear Hpp' HT HRT HtR Hex Htcb Hmto.
      clear Hbx Hsp Hspg HM HS' HS''S''' HM'.
      clear HT' HBlkQ' HSS'.
      subst fl' isr' fl''' isr'''.
      rewrite HSP'' in Hlds; clear HSP''.
      sep on.
      sep lift 7 in HM''.
      sep_inj.
      auto.
    subst pcx; clear Hlds.
    subst M''' R''' fl''' isr'''.
    reg simpl.
    rewrite HSP''.
    arith_rep (sp - 6 + 2) with (sp - 4) in HM''.
    arith_rep (sp - 6 + 4) with (sp - 2) in HM''.
    arith_rep (sp - 6 + 2) with (sp - 4).
    split.
      unfold embed_BlkQ.
      rewrite <- HT'.
      rewrite <- HBlkQ'.
      unfold embed_RdyQ in * .
      rewrite (filterTQ_prev _ _ _ _ _ HRT HtR HT').
      sep on.
      sep matchH HM'' with 9.
      sep stksplit.
      arith_rep (sp - 4 - 2) with (sp - 6).
      sep stksplit 2.
      arith_rep (10-4-2-2) with 2.
      arith_rep (sp-6-2) with (sp-8).
      sep matchH HM'' with 3.
    repeat split_l; trivial.
  simpl.
  exists p_enQueue; exists p_block_8.
  exists g_enQueue; exists g_block_8.
  split.    
    unfold SCAP_lookupCT, SCAP_consCT.
    apply SPSpec.lookup_merge_intro_r.
    apply L_enQueue_in_psi_block.
  split_l.
  split.
    clear Hpp' Hpp'''.
    unfold enable, p_block_7.
    intros S HS.
    destruct HS as [T [RdyQ [BlkQ [l_ptr [tid [tcb [sp [pc [jk1 [jk2 HS]]]]]]]]]].
    unfold Ast_block_7 in HS.
    destruct HS as [HM [Hbx [Hsp [Hspg [HRT [HtR [Hex [Htcb [Hmto HT]]]]]]]]].
    (* enable call *)
    destruct S as [[[M R] fl] isr].
    simplS in * .
    exists (MemUpdW M (sp - 4 - 2) (L_block+7+1) 0, RF_update R SP (sp - 4 - 2), fl, isr).
    apply nextS_call.
    unfold pushW.
    split.
      apply stkSpacePrec_stkSpace1 with mTrue.
      rewrite Hsp.
      sep on.
      sep lift 4 in HM.
      sep stksplit 2 in HM.
      sep matchH HM with 10.
      split.
      rewrite Hsp.
      apply MemUpdW_MemUpd16 with 0; trivial.
    rewrite Hsp; trivial.
  split.
    (* g' => fun_frame *)
    unfold sp_call_pred, fun_frame.
    unfold linkPG, conjPG.
    intros S' S'' HS'S''.
    destruct HS'S'' as [H HS'S''].
    destruct H as [S [HS HSS']].
    unfold p_enQueue in HS'S''.
    destruct HS as [T [RdyQ [BlkQ [l_ptr [tid [tcb [sp [pc [jk1 [jk2 HS]]]]]]]]]].
    set (mQ := (stkSpacePrec (sp - 8) 2 
         *^* (curr_thrd |->w tid)
         *^* (embed_RdyQ T RdyQ)
         *^* (sp - 6 |->w (L_block+7+1))
        *^* (sp     |-> pc) 
        *^* (sp + 1 |-> jk1) 
        *^* (sp + 2 |-> l_ptr) 
        *^* (sp + 3 |-> jk2))).
    assert (HS' : Ast_enQueue_pre T BlkQ l_ptr tid (sp, BLOCKED, pc) (S_RF S' BP)
      (sp - 6) (mQ *^* mTrue) S').
      eapply Hpp'; eauto.
    clear Hpp' HS HSS' Hpp'''.
    substH HS'S'' with (HS'S'' _ _ _ _ _ _ _ _ HS').
    destruct HS'S'' as [T' [BlkQ' HS'']].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct S'' as [[[M'' R''] fl''] isr''].
    simplS in * .
    destruct HS'' as [HS'' _].
    unfold Ast_enQueue_pre, Ast_enQueue_post in * .
    destruct HS' as [HM' [Htid [HBP' HSP']]].
    destruct HS'' as [HM'' [HBP'' HSP'']].
    simplS in * .
    split.
      rewrite HSP'; rewrite HSP''; trivial.
    unfold Ret_Addr.
    exists (L_block + 7 + 1).
    simplS.
    split.
      intros pc''.
      apply nextPC_retw with (sp - 6); trivial.
      clear HM''.
      sep on.
      subst mQ.
      sep lift 8 in HM'.
      sep trivial.
    intros pc''.
    apply nextPC_retw with (sp - 6); trivial.
    clear HM'.
    sep on.
    subst mQ.
    sep lift 7 in HM''.
    sep trivial.
  split.
    (* p => p' *)
    intros S' H.
    unfold conjPG, conjG, linkPG in H.
    destruct H as [S [HS HnextSS']].
    unfold p_block_7 in HS.
    unfold p_enQueue.
    destruct HS as [T [RdyQ [BlkQ [l_ptr [tid [tcb [sp [pc [jk1 [jk2 HS]]]]]]]]]].
    exists T; exists BlkQ; exists l_ptr; exists tid; exists (sp, BLOCKED, pc).
    exists (S_RF S' BP); exists (sp - 6).
    set (mQ := (stkSpacePrec (sp - 8) 2 
         *^* (curr_thrd |->w tid)
         *^* (embed_RdyQ T RdyQ)
         *^* (sp - 6 |->w (L_block+7+1))
         *^* (sp     |-> pc) 
         *^* (sp + 1 |-> jk1) 
         *^* (sp + 2 |-> l_ptr) 
         *^* (sp + 3 |-> jk2))).
    assert (H : Ast_enQueue_pre T BlkQ l_ptr tid (sp, BLOCKED, pc) (S_RF S' BP)
      (sp - 6) (mQ *^* mTrue) S'). 
      eapply Hpp'; eauto.
    clear Hpp''' HS HnextSS'.
    unfold Ast_enQueue_pre in * .
    destruct H as [HM' [Htid [HBP' HSP']]].
    split.
      sep on.
      sep matchH HM' with 10.
    auto.
  split.
    (* p => p''' *)
    intros pc' S'''.
    unfold linkPG, wrap_fun, conjG.
    intros H.
    destruct H as [S H].
    destruct H as [HS H].
    destruct H as [S'' [H HS''S''']].
    destruct H as [S' [HSS' HS'S'']].
    unfold p_block_7, g_block_8 in * .
    destruct HS as [T [RdyQ [BlkQ [l_ptr [tid [tcb [sp [pc [jk1 [jk2 HS]]]]]]]]]].
    assert (H : exists T', exists BlkQ',
      Ast_block_8 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 T' BlkQ' mTrue S''').
      eapply Hpp'''; eauto.
    exists T; exists RdyQ; exists BlkQ; exists l_ptr.
    exists tid; exists tcb; exists sp; exists pc; exists jk1; exists jk2.
    apply H.
  (* g''' => g *)
  intros pc' S S'''' H.
  unfold conjPG, wrap_fun, conjG, linkPG in H.
  destruct H as [S''' [HS HS'''S'''']].
  destruct HS as [HSx [S'' [Hx HnextS'']]].
  destruct Hx as [S' [HSS' HS'S'']].
  unfold p_block_7, g_block_8 in * .
  intros mP T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 HS.
  assert (H : exists T', exists BlkQ',
    Ast_block_8 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 T' BlkQ' mP S''').
    eapply Hpp'''; eauto.
  destruct H as [T' [BlkQ' HS']].
  eapply HS'''S''''; eauto.
Qed.

Lemma SCAP_WFcmd_block_8_pushwp :
  SCAP_WFcmd (SCAP_consCT (L_block+8+1) (Ast p_block_9 g_block_9 sctag) psi_block)
    (L_block+8) (Ast p_block_8 g_block_8 sctag) 
    (comml (_pushwp (Opr_imm ready_Q))).
Proof.
  assert (Hpp' : forall T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 T' BlkQ' mP S S',
    Ast_block_8 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 T' BlkQ' mP S
    -> nextS (L_block+8) (comml (_pushwp (Opr_imm ready_Q))) S S'
    -> Ast_block_9 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 T' BlkQ' mP S').
    intros T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 T' BlkQ' mP S S' HS HSS'.
    unfold Ast_block_8, Ast_block_9 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspgt [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simplS in * .
    destruct HS as [HM [Hsp [Hspg [HRT [HtR [Hex [Htcb [Hmto [HT' HBlkQ']]]]]]]]].
    split.
      subst M' R' fl' isr'.
      sep on.
      sep lift 4.
      rewrite Hsp.
      arith_rep (sp - 4 - 2) with (sp - 6).
      sep upd.
      sep matchH HM with 13.
      sep stksplit in HM.
      arith_rep (sp - 4 - 2) with (sp - 6) in HM.
    subst R' fl' isr'; reg simpl.
    rewrite Hsp.
    split_l by omega.
    repeat split_l. 
    trivial.
  simpl.
  exists p_block_9; exists g_block_9.
  split_l.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold p_block_8 in HpS.
    destruct HpS as [T [RdyQ [BlkQ [l_ptr [tid [tcb [sp [pc [jk1 [jk2 [T' [BlkQ' HS]]]]]]]]]]]].
    exists (MemUpdW M (sp-6) ready_Q 0, RF_update R SP (sp - 6), fl, isr).
    apply nextS_pushwp with ready_Q; trivial.
    unfold pushW.
    destruct HS as [HM [Hsp [Hspg [HRT [HtR [Hex [Htcb [Hmto [HT' HBlkQ']]]]]]]]].
    simplS in * .
    rewrite Hsp.
    split.
      apply stkSpace_weaken with (_BLK_Stk - 4);
        [idtac | omega].
      sep on.
      sep lift 3 in HM.
      sep trivial.
    arith_rep (sp - 4 - 2) with (sp - 6).
    split.
      apply MemUpdW_MemUpd16 with 0; trivial.
    trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold p_block_8, p_block_9.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [T [RdyQ [BlkQ [l_ptr [tid [tcb [sp [pc [jk1 [jk2 [T' [BlkQ' HS]]]]]]]]]]]].
    exists T; exists RdyQ; exists BlkQ; exists l_ptr; exists tid; exists tcb;
      exists sp; exists pc; exists jk1; exists jk2; exists T'; exists BlkQ'.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold p_block_8, g_block_9 in * .
  intros mP T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 T' BlkQ' HS.
  eapply HgS'S''; eauto.
Qed.

Lemma SCAP_WFcmd_block_9_call :
  SCAP_WFcmd (SCAP_consCT (L_block+9+1) (Ast p_block_10 g_block_10 sctag) psi_block)
    (L_block+9) (Ast p_block_9 g_block_9 sctag) 
    (comml (_call L_deQueue)).
Proof.
  assert(Hpp' : forall T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 
    T' BlkQ' mQ mP S S',
    Ast_block_9 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 T' BlkQ' mP S
    -> nextS (L_block+9) (comml (_call L_deQueue)) S S'
    -> mQ = ((curr_thrd |->w tid)
        *^* (embed_BlkQ l_ptr T' BlkQ')
        *^* (sp - 8 |->w (L_block+9+1))
        *^* (sp - 4 |->w tid)
        *^* (sp - 2 |->w l_ptr)
        *^* (sp     |-> pc) 
        *^* (sp + 1 |-> jk1) 
        *^* (sp + 2 |-> l_ptr) 
        *^* (sp + 3 |-> jk2))
    -> Ast_deQueue_pre T' RdyQ ready_Q (S_RF S' BP) (sp - 8)
    (mQ *^* mP) S').
    intros T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 T' BlkQ' mQ mP S S'.
    intros HS HSS' HmQ.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    unfold Ast_block_9 in HS.
    destruct HS as [HM [Hsp [Hspg [HRT [HtR [Hex [Htcb [Hmto [HT HBlkQ']]]]]]]]].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspx [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    subst M' R' fl' isr'.
    unfold Ast_deQueue_pre.
    simplS in * .
    reg simpl.
    split.
      sep on.
      rewrite Hsp.
      arith_rep (sp - 6 - 2) with (sp - 8).
      subst mQ.
      sep lift 5.
      sep_update.
      sep lift 3 in HM.
      sep stksplit in HM.
      arith_rep (sp - 6 - 2) with (sp - 8) in HM.
      arith_rep (10 - 6 - 2) with 2 in HM.
      unfold embed_RdyQ in HM.
      arith_rep (sp - 8 + 2) with (sp - 6).
      sep matchH HM with 13.
    split_l.
    rewrite Hsp; omega.
  assert (Hpp''' : forall T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2
    T' BlkQ' pc' mP S S' S'' S''',
    Ast_block_9 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 T' BlkQ' mP S
    -> nextS (L_block+9) (comml (_call L_deQueue)) S S'
    -> g_deQueue S' S''
    -> nextS pc' (comms (_retw)) S'' S'''
    -> exists RdyQ', exists tid', exists tcb',
      Ast_block_10 T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 
        T' RdyQ' BlkQ' tid' tcb' mP S''').
    intros T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 T' BlkQ' pc' mP.
    intros S S' S'' S''' HS HSS' HS'S'' HS''S'''.
    unfold p_deQueue in HS'S''.
    set (mQ := (curr_thrd |->w tid)
        *^* (embed_BlkQ l_ptr T' BlkQ')
        *^* (sp - 8 |->w (L_block+9+1))
        *^* (sp - 4 |->w tid)
        *^* (sp - 2 |->w l_ptr)
        *^* (sp     |-> pc) 
        *^* (sp + 1 |-> jk1) 
        *^* (sp + 2 |-> l_ptr) 
        *^* (sp + 3 |-> jk2)).
    assert (HS' : Ast_deQueue_pre T' RdyQ ready_Q (S_RF S' BP) (sp-8)
      (mQ *^* mP) S').
      eapply Hpp'; eauto.
    clear Hpp'.
    substH HS'S'' with (HS'S'' _ _ _ _ _ _ HS') into HS''.
    unfold Ast_block_9 in HS.
    destruct HS as [_ [Hsp [Hspg [HRT [HtR [Hex [Htcb [Hmto [HT' HBlkQ']]]]]]]]].
    destruct HS'' as [HS'' | HS''].
      destruct HS'' as [Hempty [HS'' Hbp]].
      rewrite (filterTQ_prev _ _ _ _ _ HRT HtR HT') in Hempty.
      destruct (filterTQ_not_empty _ _ _ _ Hex Htcb Hmto Hempty).
    destruct HS'' as [RdyQ' [tid' [tcb' [HS'' [HRdyQ' Hnin]]]]].
    exists RdyQ'; exists tid'; exists tcb'.
    subst mQ.
    unfold Ast_deQueue_post_b in HS''.
    unfold Ast_block_10.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct S'' as [[[M'' R''] fl''] isr''].
    destruct S''' as [[[M''' R'''] fl'''] isr'''].
    simplS in * .
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct (nextS_inversion HS''S''') as [HM''' [HR''' [Hfl''' Hisr''']]].
    destruct HR''' as [pcx [Hlds HR''']].
    destruct HS'' as [HM'' [HAX'' [HBP'' [HSP'' Hmto']]]].
    assert (Hpcx : pcx = L_block + 9 + 1).
      subst R' M''' R'''.
      clear HT' HRT HtR Hex Htcb Hmto.
      clear Hsp Hspg HS' HS''S''' HM'.
      clear HBlkQ' HSS'.
      subst fl' isr' fl''' isr'''.
      rewrite HSP'' in Hlds; clear HSP''.
      sep on.
      sep lift 6 in HM''.
      sep_inj.
      auto.
    subst pcx; clear Hlds.
    subst M''' R''' fl''' isr'''.
    reg simpl.
    rewrite HAX''.
    rewrite HSP''.
    arith_rep (sp - 8 + 2) with (sp - 6) in HM''.
    arith_rep (sp - 8 + 2) with (sp - 6).
    split.
      unfold embed_BlkQ in * .
      unfold embed_RdyQ.
      sep on.
      sep matchH HM'' with 11.
      sep stksplit.
      arith_rep (sp - 6 - 2) with (sp - 8).
      arith_rep (10 - 6 - 2) with 2.
      sep matchH HM'' with 2.
    repeat split_l; trivial.
  simpl.
  exists p_deQueue; exists p_block_10.
  exists g_deQueue; exists g_block_10.
  split.
    unfold SCAP_lookupCT, SCAP_consCT.
    apply SPSpec.lookup_merge_intro_r.
    apply L_deQueue_in_psi_block.
  split_l.
  split.
    (* enable call *)
    clear Hpp' Hpp'''.
    unfold enable, p_block_9.
    intros S HS.
    destruct HS as [T [RdyQ [BlkQ [l_ptr [tid [tcb [sp [pc [jk1 [jk2 HS]]]]]]]]]].
    destruct HS as [T' [BlkQ' HS]].
    unfold Ast_block_9 in HS.
    destruct HS as [HM [Hsp [Hspg _]]].
    destruct S as [[[M R] fl] isr].
    simplS in * .
    exists (MemUpdW M (sp - 6 - 2) (L_block+9+1) 0, RF_update R SP (sp - 6 - 2), fl, isr).
    apply nextS_call.
    unfold pushW.
    split.
      apply stkSpace_weaken with (10 - 6); [idtac | omega].
      apply stkSpacePrec_stkSpace1 with mTrue.
      rewrite Hsp.
      sep on.
      sep matchH HM with 10.
    split.
      rewrite Hsp.
      apply MemUpdW_MemUpd16 with 0; trivial.
    rewrite Hsp; trivial.
  split.
    (* g' => fun_frame *)
    unfold sp_call_pred, fun_frame.
    unfold linkPG, conjPG.
    intros S' S'' HS'S''.
    destruct HS'S'' as [H HS'S''].
    destruct H as [S [HS HSS']].
    unfold p_deQueue in HS'S''.
    destruct HS as [T [RdyQ [BlkQ [l_ptr [tid [tcb [sp [pc [jk1 [jk2 HS]]]]]]]]]].
    destruct HS as [T' [BlkQ' HS]].
    set (mQ := (curr_thrd |->w tid)
        *^* (embed_BlkQ l_ptr T' BlkQ')
        *^* (sp - 8 |->w (L_block+9+1))
        *^* (sp - 4 |->w tid)
        *^* (sp - 2 |->w l_ptr)
        *^* (sp     |-> pc) 
        *^* (sp + 1 |-> jk1) 
        *^* (sp + 2 |-> l_ptr) 
        *^* (sp + 3 |-> jk2)).
    assert (HS' : Ast_deQueue_pre T' RdyQ ready_Q (S_RF S' BP)
      (sp - 8) (mQ *^* mTrue) S').
      eapply Hpp'; eauto.
    clear Hpp' HSS' Hpp'''.
    substH HS'S'' with (HS'S'' _ _ _ _ _ _ HS') into HS''.
    destruct HS as [_ [Hsp [Hspg [HRT [HtR [Hex [Htcb [Hmto [HT' HBlkQ']]]]]]]]].
    destruct HS'' as [HS'' | HS''].
      destruct HS'' as [Hempty [HS'' Hbp]].
      rewrite (filterTQ_prev _ _ _ _ _ HRT HtR HT') in Hempty.
      destruct (filterTQ_not_empty _ _ _ _ Hex Htcb Hmto Hempty).
    destruct HS'' as [RdyQ' [tid' [tcb' [HS'' [HRdyQ' Hnin']]]]].
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct S'' as [[[M'' R''] fl''] isr''].
    simplS in * .
    unfold Ast_deQueue_pre, Ast_deQueue_post_b in * .
    destruct HS' as [HM' [HBP' HSP']].
    destruct HS'' as [HM'' [HAX'' [HBP'' [HSP'' Hmto']]]].
    simplS in * .
    split.
      rewrite HSP'; rewrite HSP''; trivial.
    unfold Ret_Addr.
    exists (L_block + 9 + 1).
    simplS.
    split.
      intros pc''.
      apply nextPC_retw with (sp - 8); trivial.
      clear HM''.
      sep on.
      subst mQ.
      sep lift 5 in HM'.
      sep trivial.
    intros pc''.
    apply nextPC_retw with (sp - 8); trivial.
    clear HM'.
    sep on.
    subst mQ.
    sep lift 6 in HM''.
    sep trivial.
  split.
    (* p => p' *)
    intros S' H.
    unfold conjPG, conjG, linkPG in H.
    destruct H as [S [HS HnextSS']].
    unfold p_block_9 in HS.
    unfold p_deQueue.
    destruct HS as [T [RdyQ [BlkQ [l_ptr [tid [tcb [sp [pc [jk1 [jk2 HS]]]]]]]]]].
    destruct HS as [T' [BlkQ' HS]].
    exists T'; exists RdyQ; exists ready_Q; exists (S_RF S' BP); 
      exists (sp - 8).
    set (mQ := (curr_thrd |->w tid)
        *^* (embed_BlkQ l_ptr T' BlkQ')
        *^* (sp - 8 |->w (L_block+9+1))
        *^* (sp - 4 |->w tid)
        *^* (sp - 2 |->w l_ptr)
        *^* (sp     |-> pc) 
        *^* (sp + 1 |-> jk1) 
        *^* (sp + 2 |-> l_ptr) 
        *^* (sp + 3 |-> jk2)).
    assert (H : Ast_deQueue_pre T' RdyQ ready_Q (S_RF S' BP) 
           (sp - 8) (mQ *^* mTrue) S').
      eapply Hpp'; eauto.
    clear Hpp''' Hpp' HS HnextSS'.
    unfold Ast_deQueue_pre in * .
    destruct H as [HM' H].
    split_r.
    sep on.
    sep matchH HM' with 10.
  split.
    (* p => p''' *)
    intros pc' S'''.
    unfold linkPG, wrap_fun, conjG.
    intros H.
    destruct H as [S H].
    destruct H as [HS H].
    destruct H as [S'' [H HS''S''']].
    destruct H as [S' [HSS' HS'S'']].
    unfold p_block_9, p_block_10 in * .
    destruct HS as [T [RdyQ [BlkQ [l_ptr [tid [tcb [sp [pc [jk1 [jk2 HS]]]]]]]]]].
    destruct HS as [T' [BlkQ' HS]].
    assert (H : exists RdyQ', exists tid', exists tcb',
      Ast_block_10 T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ'
        tid' tcb' mTrue S''').
      eapply Hpp'''; eauto.
    destruct H as [RdyQ' [tid' [tcb' HS''']]].
    exists T; exists RdyQ; exists BlkQ; exists l_ptr.
    exists tid; exists sp; exists pc; exists jk1; exists jk2;
    exists T'; exists RdyQ'.
    exists BlkQ'; exists tid'; exists tcb'.
    trivial.
  (* g'' => g *)
  intros pc' S S'''' H.
  unfold conjPG, wrap_fun, conjG, linkPG in H.
  destruct H as [S''' [HS HS'''S'''']].
  destruct HS as [HSx [S'' [H HS'']]].
  destruct H as [S' [HSS' HS'S'']].
  unfold p_block_9, p_block_10 in * .
  intros mP T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 T' BlkQ' HS.
  assert (H : exists RdyQ', exists tid', exists tcb',
    Ast_block_10 T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ'
    tid' tcb' mP S''').
    eapply Hpp'''; eauto.
  destruct H as [RdyQ' [tid' [tcb' HS''']]].
  apply HS'''S'''' with T' RdyQ' BlkQ' tid' tcb'; trivial.
Qed.

(*  <block+10>:   addw    $6, %sp *)

Lemma SCAP_WFcmd_block_10_addw :
  SCAP_WFcmd (SCAP_consCT (L_block+10+1) (Ast p_block_11 g_block_11 sctag) psi_block)
    (L_block+10) (Ast p_block_10 g_block_10 sctag) 
    (comml (_addw   (Opr_imm 6 ) SP)).
Proof.
  assert (Hpp' : forall T RdyQ BlkQ l_ptr tid sp pc jk1 jk2
    T' RdyQ' BlkQ' tid' tcb' mP S S',
    Ast_block_10 T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb' mP S
    -> nextS (L_block+10) (comml (_addw   (Opr_imm 6 ) SP)) S S'
    -> Ast_block_11 T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb' mP S').
    intros T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb' mP S S' HS HSS'.
    unfold Ast_block_10, Ast_block_11 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    simplS in * .
    destruct HS as [HM [Hax [Hsp [Hspg [HT' [HBlkQ' [HRdyQ [Hmto Hnin]]]]]]]].
    subst M' R' isr'.
    split.
      sep on.
      sep matchH HM with 10.
      sep stksplit. 
      sep stksplit. 
      sep stksplit.
      arith_rep (sp - 2 - 2) with (sp - 4).
      arith_rep (sp - 4 - 2) with (sp - 6).
      arith_rep (10 - 2 - 2 - 2) with (10 - 6).
      sep matchH HM with 4.
    reg simpl.
    unfold opr_eval.
    split_l.
    split_l by omega.
    repeat split_l; trivial.
  simpl.
  exists p_block_11; exists g_block_11.
  split_l.
  split.
    (* enable addw *)
    unfold enable, p_block_10.
    intros S H.
    destruct H as [T [RdyQ [BlkQ [l_ptr [tid [sp [pc [jk1 [jk2 H]]]]]]]]].
    destruct H as [T' [RdyQ' [BlkQ' [tid' [tcb' HS]]]]].
    destruct HS as [HM [Hax [Hsp [Hspg [HT' [HBlkQ' [HRdyQ [Hmto Hnin]]]]]]]].
    destruct S as [[[M R] fl] isr].
    simplS in * .
    set (R' := RF_update R SP (6 + R SP)).
    exists (M, R', Clear_Flag fl _ZF, isr).
    apply nextS_addw with 6 (R SP); auto.
    unfold trash_fl.
    right; trivial.
  split. (* p ==> p' *)
    unfold p_block_10, g_block_11.
    intros S' H.
    unfold linkPG in H.
    destruct H as [S H].
    simpl in H.
    destruct H as [H HnextS].
    destruct H as [T [RdyQ [BlkQ [l_ptr [tid [sp [pc [jk1 [jk2 H]]]]]]]]].
    destruct H as [T' [RdyQ' [BlkQ' [tid' [tcb' HS]]]]].
    exists T; exists RdyQ; exists BlkQ; exists l_ptr.
    exists tid; exists sp; exists pc; exists jk1; exists jk2.
    exists T'; exists RdyQ'; exists BlkQ'; exists tid'; exists tcb'.
    apply Hpp' with S; trivial.
  (* gimp *)
  intros S S'' H.
  unfold conjPG, conjG in H.
  destruct H as [S' [[HSx HnextS] HS'S'']].
  unfold p_block_10, p_block_11 in * .
  intros mP T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb' HS.
  eapply HS'S''; eauto.
Qed.

Lemma SCAP_WFcmd_block_11_movw :
  SCAP_WFcmd (SCAP_consCT (L_block+11+1) (Ast p_block_12 g_block_12 sctag) psi_block)
    (L_block+11) (Ast p_block_11 g_block_11 sctag) 
    (comml (_movwp  (Opr_reg AX) BX)).
Proof.
  assert (Hpp' : forall T RdyQ BlkQ l_ptr tid sp pc jk1 jk2
    T' RdyQ' BlkQ' tid' tcb' mP S S',
    Ast_block_11 T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb' mP S
    -> nextS (L_block+11) (comml (_movwp  (Opr_reg AX) BX)) S S'
    -> Ast_block_12 T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb' mP S').
    intros T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb' mP S S' HS HSS'.
    unfold Ast_block_11, Ast_block_12 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    simplS in * .
    destruct HS as [HM [Hax [Hsp [Hspg [HT' [HBlkQ' [HRdyQ [Hmto Hnin]]]]]]]].
    subst M' R' fl' isr'.
    split_l.
    reg simpl.
    repeat split_l; trivial.
  simpl.
  exists p_block_12; exists g_block_12.
  split_l.
  split.
    (* enable movw *)
    unfold enable, p_block_11.
    intros S H.
    destruct H as [T [RdyQ [BlkQ [l_ptr [tid [sp [pc [jk1 [jk2 H]]]]]]]]].
    destruct H as [T' [RdyQ' [BlkQ' [tid' [tcb' HS]]]]].
    unfold Ast_block_11 in HS.
    destruct HS as [HM [Hax [Hsp [Hspg [HT' [HBlkQ' [HRdyQ' [Hmto Hnin]]]]]]]].
    destruct S as [[[M R] fl] isr].
    simplS in * .
    set (R' := RF_update R BX tid').
    exists (M, R', fl, isr).
    apply nextS_movwp with tid'; auto.
  split. (* p ==> p' *)
    unfold p_block_11, p_block_12.
    intros S' H.
    unfold linkPG in H.
    destruct H as [S H].
    destruct H as [H HnextS].
    destruct H as [T [RdyQ [BlkQ [l_ptr [tid [sp [pc [jk1 [jk2 H]]]]]]]]].
    destruct H as [T' [RdyQ' [BlkQ' [tid' [tcb' HS]]]]].
    exists T; exists RdyQ; exists BlkQ; exists l_ptr; exists tid.
    exists sp; exists pc; exists jk1; exists jk2; exists T'; exists RdyQ'.
    exists BlkQ'; exists tid'; exists tcb'. 
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold conjPG, conjG in H.
  destruct H as [S' [[_ HnextS] HS'S'']].
  unfold p_block_11, p_block_12 in * .
  intros mP T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb' HS.
  eapply HS'S''; eauto.
Qed.

Lemma SCAP_WFcmd_block_12_movwst :
  SCAP_WFcmd (SCAP_consCT (L_block+12+1) (Ast p_block_13 g_block_13 sctag) psi_block)
    (L_block+12) (Ast p_block_12 g_block_12 sctag) 
    (comml (_movwst (Opr_reg AX) (A_imm curr_thrd))).
Proof.
  assert (Hpp' : forall T RdyQ BlkQ l_ptr tid sp pc jk1 jk2
    T' RdyQ' BlkQ' tid' tcb' mP S S',
    Ast_block_12 T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb' mP S
    -> nextS (L_block+12) (comml (_movwst (Opr_reg AX) (A_imm curr_thrd))) S S'
    -> Ast_block_13 T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb' mP S').
    intros T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb' mP S S' HS HSS'.
    unfold Ast_block_12, Ast_block_13 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hlds HM'].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simpl in Hlds.
    simplS in * .
    destruct HS as [HM [Hax [Hbx [Hsp [Hspg [HT' [HBlkQ' [HRdyQ' [Htid'tcb' Htid'RdyQ']]]]]]]]].
    subst M' R' fl' isr'.
    split.
      sep on.
      sep lift 0 in HM.
      sep lift 0.
      rewrite Hax.
      sep upd.
      sep matchH HM with 6.
    repeat split_l; trivial.
  simpl.
  exists p_block_13; exists g_block_13.
  split_l.
  split.
    (* enable movw *)
    unfold enable, p_block_12.
    intros S H.
    destruct H as [T [RdyQ [BlkQ [l_ptr [tid [sp [pc [jk1 [jk2 HS]]]]]]]]].
    destruct HS as [T' [RdyQ' [BlkQ' [tid' [tcb' HS]]]]].
    destruct HS as [HM [Hax [Hbx [Hsp [Hspg [HT' [HBlkQ' [HRdyQ' [Htid'tcb' Htid'RdyQ']]]]]]]]].
    destruct S as [[[M R] fl] isr].
    simplS in * .
    set (M' := MemUpdW M curr_thrd tid' 0).
    exists (M', R, fl, isr).
    apply nextS_movwst with curr_thrd tid'; auto.
      sep on.
      sep lift 0 in HM.
      sep trivial.
    apply MemUpdW_MemUpd16 with 0; trivial.
  split. (* pimp *)
    unfold p_block_12, p_block_13.
    intros S' H.
    unfold linkPG in H.
    destruct H as [S H].
    destruct H as [H HnextS].
    destruct H as [T [RdyQ [BlkQ [l_ptr [tid [sp [pc [jk1 [jk2 HS]]]]]]]]].
    destruct HS as [T' [RdyQ' [BlkQ' [tid' [tcb' HS]]]]].
    exists T; exists RdyQ; exists BlkQ; exists l_ptr.
    exists tid; exists sp; exists pc; exists jk1; exists jk2.
    exists T'; exists RdyQ'; exists BlkQ'.
    exists tid'; exists tcb'.
    apply Hpp' with S; trivial.
  (* gimp *)
  intros S S'' H.
  unfold conjPG, conjG in H.
  destruct H as [S' [[_ HnextS] HSS']].
  unfold p_block_12, p_block_13 in * .
  intros mP T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb' HS.
  eapply HSS'; eauto.
Qed.

Lemma SCAP_WFcmd_block_13_movwld : 
  SCAP_WFcmd (SCAP_consCT (L_block+13+1) (Ast p_block_14 g_block_14 sctag) psi_block)
    (L_block+13) (Ast p_block_13 g_block_13 sctag) 
    (comml (_movwld (A_reg TCB_STACK BX) SP)).
Proof.
  simpl.
  exists p_block_14; exists g_block_14.
  split_l.
  assert (Hpp' : forall T RdyQ BlkQ l_ptr tid sp pc jk1 jk2
    T' RdyQ' BlkQ' tid' tcb' mP S S',
    Ast_block_13 T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb' mP S
    -> nextS (L_block+13) (comml (_movwld (A_reg TCB_STACK BX) SP)) S S'
    -> Ast_block_14 T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb' (TCB_SP tcb') mP S').
    intros T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb' mP S S' HS HSS'.
    unfold Ast_block_13, Ast_block_14 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HR' as [sp' [Hlds HR']].
    simplS in * .
    destruct HS as [HM [Hbx [Hsp [Hspg [HT' [HBlkQ' [HRdyQ' [Htid'tcb' Htid'RdyQ']]]]]]]].
    assert (Heq : sp' = TCB_SP tcb').
      unfold addr_eval in Hlds.
      rewrite Hbx in Hlds.
      unfold TCB_STACK in Hlds.
      destruct tcb' as [[spx' stx'] pcx'].
      unfold embed_TCB in HM.
      clear Hbx Hsp Hspg HT' HBlkQ' HRdyQ' Htid'tcb' Htid'RdyQ'.
      clear HSS'.
      subst M'.
      sep on.
      sep lift 3 in HM.
      substH HM with (ptoTCB_elim1 HM).
      destruct HM as [tst [HM Htst]].
      sep lift 0 in HM.
      sep inj.
      auto.
    subst M' R' fl' isr'.
    split_l.
    split.
      reg simpl.
      trivial.
    repeat split_l; trivial.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold p_block_13 in HpS.
    destruct HpS as [T [RdyQ [BlkQ [l_ptr [tid [sp [pc [jk1 [jk2 HS]]]]]]]]].
    destruct HS as [T' [RdyQ' [BlkQ' [tid' [tcb' HS]]]]].
    unfold Ast_block_13 in HS.
    destruct HS as [HM [HBX [HSP [Hspg [HT' [HBlkQ' [HRdyQ' [Hmto Hnin]]]]]]]].
    simplS in * .
    exists (M, RF_update R SP (TCB_SP tcb'), fl, isr).
    apply nextS_movwld with (tid' + TCB_STACK) (TCB_SP tcb'); trivial.
      simpl.
      subst; trivial.
    unfold TCB_STACK.
    unfold embed_TCB in HM.
    sep on.
    sep lift 3 in HM.
    substH HM with (ptoTCB_elim1 HM).
    destruct HM as [tst [HM Htst]].
    sep lift 0 in HM.
    sep trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold p_block_13, p_block_14.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [T [RdyQ [BlkQ [l_ptr [tid [sp [pc [jk1 [jk2 HS]]]]]]]]]. 
    destruct HS as [T' [RdyQ' [BlkQ' [tid' [tcb' HS]]]]].
    exists T; exists RdyQ; exists BlkQ; exists l_ptr.
    exists tid; exists sp; exists pc; exists jk1; exists jk2.
    exists T'; exists RdyQ'; exists BlkQ'.
    exists tid'; exists tcb'; exists (TCB_SP tcb').
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold p_block_13, p_block_14 in * .
  intros mP T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb' HS.
  eapply HgS'S''; eauto.
Qed.

Lemma SCAP_WFcmd_block_14_retw : 
  SCAP_WFcmd psi_block (L_block+14) 
    (Ast p_block_14 g_block_14 sctag) (comms _retw).
Proof.
  unfold p_block_14, g_block_14.
  simpl.
  intros S H.
  intros.
  unfold Ast_block_14 in H0.
  InvertAll.
  exists T'; exists RdyQ'; exists BlkQ'. 
  exists tid'; exists sp'; exists tcb'.
  repeat split_l.
  trivial.
Qed.

Lemma SCAP_WFiseq_block :
  SCAP_WFiseq psi_block L_block spec_block I_block.
Proof.
unfold I_block, spec_block.
apply scap_isq_seq with p_block_1 g_block_1.
apply SCAP_WFcmd_block_0_movwld.
apply scap_isq_seq with p_block_2 g_block_2.
apply SCAP_WFcmd_block_1_movwst.
assert (H : L_block+1+1 = L_block+2); [omega | rewrite H; clear H].
apply scap_isq_seq with p_block_3 g_block_3.
apply SCAP_WFcmd_block_2_movwst.
assert (H : L_block+2+1 = L_block+3); [omega | rewrite H; clear H].
apply scap_isq_seq with p_block_4 g_block_4.
apply SCAP_WFcmd_block_3_movwp.
assert (H : L_block+3+1 = L_block+4); [omega | rewrite H; clear H].
apply scap_isq_seq with p_block_5 g_block_5.
apply SCAP_WFcmd_block_4_movwld.
assert (H : L_block+4+1 = L_block+5); [omega | rewrite H; clear H].
apply scap_isq_seq with p_block_6 g_block_6.
apply SCAP_WFcmd_block_5_pushwp.
assert (H : L_block+5+1 = L_block+6); [omega | rewrite H; clear H].
apply scap_isq_seq with p_block_7 g_block_7.
apply SCAP_WFcmd_block_6_pushwp.
assert (H : L_block+6+1 = L_block+7); [omega | rewrite H; clear H].
apply scap_isq_seq with p_block_8 g_block_8.
apply SCAP_WFcmd_block_7_call.
assert (H : L_block+7+1 = L_block+8); [omega | rewrite H; clear H].
apply scap_isq_seq with p_block_9 g_block_9.
apply SCAP_WFcmd_block_8_pushwp.
assert (H : L_block+8+1 = L_block+9); [omega | rewrite H; clear H].
apply scap_isq_seq with p_block_10 g_block_10.
apply SCAP_WFcmd_block_9_call.
assert (H : L_block+9+1 = L_block+10); [omega | rewrite H; clear H].
apply scap_isq_seq with p_block_11 g_block_11.
apply SCAP_WFcmd_block_10_addw.
assert (H : L_block+10+1 = L_block+11); [omega | rewrite H; clear H].
apply scap_isq_seq with p_block_12 g_block_12.
apply SCAP_WFcmd_block_11_movw.
assert (H : L_block+11+1 = L_block+12); [omega | rewrite H; clear H].
apply scap_isq_seq with p_block_13 g_block_13.
apply SCAP_WFcmd_block_12_movwst.
assert (H : L_block+12+1 = L_block+13); [omega | rewrite H; clear H].
apply scap_isq_seq with p_block_14 g_block_14.
apply SCAP_WFcmd_block_13_movwld.
assert (H : L_block+13+1 = L_block+14); [omega | rewrite H; clear H].
apply scap_isq_end.
apply SCAP_WFcmd_block_14_retw.
Qed.
