(***************************************************************)
(* Coq code for AIM                                            *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

Add LoadPath "..".
Add LoadPath "../CAP0".
Add LoadPath "../SCAP".

Require Import Coqlib.

Require Import Arith.
Require Import Machine.
Require Import Operational_Semantics.
Require Import Assertion.
Require Import Assertion_Lemmas4.

Require Import AIM.
Require Import SCAP.
Require Import QLib.
Require Import Linking.
Require Import Linking_Lemmas3.

Require Import Kernel.

Require Export EnQueueSpec.
Require Export DeQueueSpec.

Definition Ast_block_pre (T : TQ) (RdyQ BlkQ : WQ) (l_ptr : Label)
  (tid : TID) (tcb : Thread) (sp pc jk1 jk2 : nat) (mP : MPred) (S : State) : Prop :=
  ((curr_thrd |->w tid)
    *^* (embed_TCB tid tcb)
    *^* (embed_RdyQ T RdyQ) 
    *^* (embed_BlkQ l_ptr T BlkQ) 
    *^* (stkSpacePrec sp _BLK_Stk) 
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk1) 
    *^* (sp + 2 |-> l_ptr) 
    *^* (sp + 3 |-> jk2) 
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp
  /\ sp >= _BLK_Stk
  /\ subsetQ_TQ RdyQ T
  /\ WQMod.not_in_dom tid RdyQ
  /\ exists_Idle T RdyQ tid
  /\ TCB_Status tcb = READY
  /\ TQMod.maps_to T tid tcb
  /\ WF_TQ T.

Definition Ast_block_post (T' : TQ) (RdyQ' BlkQ' : WQ) (l_ptr : Label)
  (tid' : TID) (tcb' : Thread) (sp' sp pc jk1 jk2 : nat) (mP : MPred) (S' : State) : Prop :=
  ((curr_thrd |->w tid')
    *^* (embed_TCB tid' tcb')
    *^* (embed_RdyQ T' RdyQ') 
    *^* (embed_BlkQ l_ptr T' BlkQ') 
    *^* (stkSpacePrec sp _BLK_Stk) 
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk1) 
    *^* (sp + 2 |-> l_ptr) 
    *^* (sp + 3 |-> jk2) 
    *^* mP) (S_Mem S')
  /\ S_RF S' SP = sp'
  /\ TQMod.maps_to T' tid' tcb'
  /\ TCB_SP tcb' = sp'
  /\ WQMod.not_in_dom tid' RdyQ'.

Definition p_block (S : State) : Prop :=
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists tid, exists tcb, exists sp, exists pc,
    exists jk1, exists jk2,
      Ast_block_pre T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mTrue S.

Definition g_block (S S' : State) : Prop := forall mP,
  forall T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2,
    Ast_block_pre T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S
    -> exists T', exists RdyQ', exists BlkQ', exists tid', exists sp', exists tcb',
      T' = TQMod.put T tid (sp, BLOCKED, pc)
      /\ BlkQ' = addThrd BlkQ tid
      /\ RdyQ = addThrd RdyQ' tid'
      /\ Ast_block_post T' RdyQ' BlkQ' l_ptr tid' tcb' sp' sp pc jk1 jk2 mP S'.

(* 1158 <block>:   movw   (curr_thrd), %bx  # curr_thrd = 2254 *)

Definition Ast_block_1 (T : TQ) (RdyQ BlkQ : WQ) (l_ptr : Label)
  (tid : TID) (tcb : Thread) (sp pc jk1 jk2 : nat) (mP : MPred) (S : State) : Prop :=
  ((curr_thrd |->w tid)
    *^* (embed_TCB tid tcb)
    *^* (embed_RdyQ T RdyQ)
    *^* (embed_BlkQ l_ptr T BlkQ)
    *^* (stkSpacePrec sp _BLK_Stk)
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk1) 
    *^* (sp + 2 |-> l_ptr) 
    *^* (sp + 3 |-> jk2) 
    *^* mP) (S_Mem S)
  /\ S_RF S BX = tid
  /\ S_RF S SP = sp
  /\ sp >= _BLK_Stk
  /\ subsetQ_TQ RdyQ T
  /\ WQMod.not_in_dom tid RdyQ
  /\ exists_Idle T RdyQ tid
  /\ TCB_Status tcb = READY
  /\ TQMod.maps_to T tid tcb
  /\ WF_TQ T.

Definition p_block_1 (S : State) : Prop := 
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists tid, exists tcb, exists sp, exists pc, exists jk1, exists jk2,
      Ast_block_1 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mTrue S.

Definition g_block_1 (S S' : State) := forall mP,
  forall T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2,
    Ast_block_1 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S
    -> exists T', exists RdyQ', exists BlkQ', exists tid', exists sp', exists tcb',
      T' = TQMod.put T tid (sp, BLOCKED, pc)
      /\ BlkQ' = addThrd BlkQ tid
      /\ RdyQ = addThrd RdyQ' tid'
      /\ Ast_block_post T' RdyQ' BlkQ' l_ptr tid' tcb' sp' sp pc jk1 jk2 mP S'.

(* 1159 <block+1>: movw   $BLOCKED, TCB_STATUS(%bx)  *)

Definition Ast_block_2 (T : TQ) (RdyQ BlkQ : WQ) (l_ptr : Label)
  (tid : TID) (tcb : Thread) (sp pc jk1 jk2 : nat) (mP : MPred) (S : State) : Prop :=
  ((curr_thrd |->w tid)
    *^* (embed_TCB tid (TCB_Status_Upd tcb BLOCKED))
    *^* (embed_RdyQ T RdyQ) 
    *^* (embed_BlkQ l_ptr T BlkQ) 
    *^* (stkSpacePrec sp _BLK_Stk) 
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk1) 
    *^* (sp + 2 |-> l_ptr) 
    *^* (sp + 3 |-> jk2) 
    *^* mP) (S_Mem S)
  /\ S_RF S BX = tid
  /\ S_RF S SP = sp
  /\ sp >= _BLK_Stk
  /\ subsetQ_TQ RdyQ T
  /\ WQMod.not_in_dom tid RdyQ
  /\ exists_Idle T RdyQ tid
  /\ TCB_Status tcb = READY
  /\ TQMod.maps_to T tid tcb
  /\ WF_TQ T.

Definition p_block_2 (S : State) : Prop := 
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists tid, exists tcb, exists sp, exists pc, exists jk1, exists jk2,
      Ast_block_2 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mTrue S.

Definition g_block_2 (S S' : State) := forall mP,
  forall T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2,
    Ast_block_2 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S
    -> exists T', exists RdyQ', exists BlkQ', exists tid', exists sp', exists tcb',
      T' = TQMod.put T tid (sp, BLOCKED, pc)
      /\ BlkQ' = addThrd BlkQ tid
      /\ RdyQ = addThrd RdyQ' tid'
      /\ Ast_block_post T' RdyQ' BlkQ' l_ptr tid' tcb' sp' sp pc jk1 jk2 mP S'.

(* 1160 <block+2>:  movw    %sp, TCB_STACK(%bx) *)

Definition Ast_block_3 (T : TQ) (RdyQ BlkQ : WQ) (l_ptr : Label)
  (tid : TID) (tcb : Thread) (sp pc jk1 jk2 : nat) (mP : MPred) (S : State) : Prop :=
  ((curr_thrd |->w tid)
    *^* (embed_TCB tid (sp, BLOCKED, pc))
    *^* (embed_RdyQ T RdyQ) 
    *^* (embed_BlkQ l_ptr T BlkQ) 
    *^* (stkSpacePrec sp _BLK_Stk) 
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk1) 
    *^* (sp + 2 |-> l_ptr) 
    *^* (sp + 3 |-> jk2) 
    *^* mP) (S_Mem S)
  /\ S_RF S BX = tid
  /\ S_RF S SP = sp
  /\ sp >= _BLK_Stk
  /\ subsetQ_TQ RdyQ T
  /\ WQMod.not_in_dom tid RdyQ
  /\ exists_Idle T RdyQ tid
  /\ TCB_Status tcb = READY
  /\ TQMod.maps_to T tid tcb
  /\ WF_TQ T.

Definition p_block_3 (S : State) : Prop := 
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists tid, exists tcb, exists sp, exists pc, exists jk1, exists jk2,
      Ast_block_3 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mTrue S.

Definition g_block_3 (S S' : State) := forall mP,
  forall T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2,
    Ast_block_3 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S
    -> exists T', exists RdyQ', exists BlkQ', exists tid', exists sp', exists tcb',
      T' = TQMod.put T tid (sp, BLOCKED, pc)
      /\ BlkQ' = addThrd BlkQ tid
      /\ RdyQ = addThrd RdyQ' tid'
      /\ Ast_block_post T' RdyQ' BlkQ' l_ptr tid' tcb' sp' sp pc jk1 jk2 mP S'.

(* 1161 <block+3>:  movw   %sp, %bp *)

Definition Ast_block_4 (T : TQ) (RdyQ BlkQ : WQ) (l_ptr : Label)
  (tid : TID) (tcb : Thread) (sp pc jk1 jk2 : nat) (mP : MPred) (S : State) : Prop :=
  ((curr_thrd |->w tid)
    *^* (embed_TCB tid (sp, BLOCKED, pc))
    *^* (embed_RdyQ T RdyQ) 
    *^* (embed_BlkQ l_ptr T BlkQ) 
    *^* (stkSpacePrec sp _BLK_Stk) 
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk1) 
    *^* (sp + 2 |-> l_ptr) 
    *^* (sp + 3 |-> jk2) 
    *^* mP) (S_Mem S)
  /\ S_RF S BX = tid
  /\ S_RF S SP = sp
  /\ S_RF S BP = sp
  /\ sp >= _BLK_Stk
  /\ subsetQ_TQ RdyQ T
  /\ WQMod.not_in_dom tid RdyQ
  /\ exists_Idle T RdyQ tid
  /\ TCB_Status tcb = READY
  /\ TQMod.maps_to T tid tcb
  /\ WF_TQ T.

Definition p_block_4 (S : State) : Prop := 
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists tid, exists tcb, exists sp, exists pc, exists jk1, exists jk2,
      Ast_block_4 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mTrue S.

Definition g_block_4 (S S' : State) := forall mP,
  forall T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2,
    Ast_block_4 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S
    -> exists T', exists RdyQ', exists BlkQ', exists tid', exists sp', exists tcb',
      T' = TQMod.put T tid (sp, BLOCKED, pc)
      /\ BlkQ' = addThrd BlkQ tid
      /\ RdyQ = addThrd RdyQ' tid'
      /\ Ast_block_post T' RdyQ' BlkQ' l_ptr tid' tcb' sp' sp pc jk1 jk2 mP S'.

(* 1162 <block+4>:  movw   2(%bp), %ax *)

Definition Ast_block_5 (T : TQ) (RdyQ BlkQ : WQ) (l_ptr : Label)
  (tid : TID) (tcb : Thread) (sp pc jk1 jk2 : nat) (mP : MPred) (S : State) : Prop :=
  ((curr_thrd |->w tid)
    *^* (embed_TCB tid (sp, BLOCKED, pc))
    *^* (embed_RdyQ T RdyQ) 
    *^* (embed_BlkQ l_ptr T BlkQ) 
    *^* (stkSpacePrec sp _BLK_Stk) 
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk1) 
    *^* (sp + 2 |-> l_ptr) 
    *^* (sp + 3 |-> jk2) 
    *^* mP) (S_Mem S)
  /\ S_RF S AX = l_ptr
  /\ S_RF S BX = tid
  /\ S_RF S SP = sp
  /\ sp >= _BLK_Stk
  /\ subsetQ_TQ RdyQ T
  /\ WQMod.not_in_dom tid RdyQ
  /\ exists_Idle T RdyQ tid
  /\ TCB_Status tcb = READY
  /\ TQMod.maps_to T tid tcb
  /\ WF_TQ T.

Definition p_block_5 (S : State) : Prop := 
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists tid, exists tcb, exists sp, exists pc, exists jk1, exists jk2,
      Ast_block_5 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mTrue S.

Definition g_block_5 (S S' : State) := forall mP,
  forall T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 ,
    Ast_block_5 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S
    -> exists T', exists RdyQ', exists BlkQ', exists tid', exists sp', exists tcb',
      T' = TQMod.put T tid (sp, BLOCKED, pc)
      /\ BlkQ' = addThrd BlkQ tid
      /\ RdyQ = addThrd RdyQ' tid'
      /\ Ast_block_post T' RdyQ' BlkQ' l_ptr tid' tcb' sp' sp pc jk1 jk2 mP S'.

(* 1163 <block+5>:  pushw   %ax                 #BlockQ *)

Definition Ast_block_6 (T : TQ) (RdyQ BlkQ : WQ) (l_ptr : Label)
  (tid : TID) (tcb : Thread) (sp pc jk1 jk2 : nat) (mP : MPred) (S : State) : Prop :=
  ((curr_thrd |->w tid)
    *^* (embed_TCB tid (sp, BLOCKED, pc))
    *^* (embed_RdyQ T RdyQ) 
    *^* (embed_BlkQ l_ptr T BlkQ) 
    *^* (stkSpacePrec (sp - 2) (_BLK_Stk - 2)) 
    *^* (sp - 2 |->w l_ptr)
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk1) 
    *^* (sp + 2 |-> l_ptr) 
    *^* (sp + 3 |-> jk2) 
    *^* mP) (S_Mem S)
  /\ S_RF S BX = tid
  /\ S_RF S SP = sp - 2
  /\ sp >= _BLK_Stk
  /\ subsetQ_TQ RdyQ T
  /\ WQMod.not_in_dom tid RdyQ
  /\ exists_Idle T RdyQ tid
  /\ TCB_Status tcb = READY
  /\ TQMod.maps_to T tid tcb
  /\ WF_TQ T.

Definition p_block_6 (S : State) : Prop := 
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists tid, exists tcb, exists sp, exists pc, exists jk1, exists jk2,
      Ast_block_6 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mTrue S.

Definition g_block_6 (S S' : State) := forall mP,
  forall T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2,
    Ast_block_6 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S
    -> exists T', exists RdyQ', exists BlkQ', exists tid', exists sp', exists tcb',
      T' = TQMod.put T tid (sp, BLOCKED, pc)
      /\ BlkQ' = addThrd BlkQ tid
      /\ RdyQ = addThrd RdyQ' tid'
      /\ Ast_block_post T' RdyQ' BlkQ' l_ptr tid' tcb' sp' sp pc jk1 jk2 mP S'.

(* 1164 <block+6>:  pushw   %bx        *)

Definition Ast_block_7 (T : TQ) (RdyQ BlkQ : WQ) (l_ptr : Label)
  (tid : TID) (tcb : Thread) (sp pc jk1 jk2 : nat) (mP : MPred) (S : State) : Prop :=
  ((curr_thrd |->w tid)
    *^* (embed_TCB tid (sp, BLOCKED, pc))
    *^* (embed_RdyQ T RdyQ) 
    *^* (embed_BlkQ l_ptr T BlkQ) 
    *^* (stkSpacePrec (sp - 4) (_BLK_Stk - 4)) 
    *^* (sp - 4 |->w tid)
    *^* (sp - 2 |->w l_ptr)
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk1) 
    *^* (sp + 2 |-> l_ptr) 
    *^* (sp + 3 |-> jk2) 
    *^* mP) (S_Mem S)
  /\ S_RF S BX = tid
  /\ S_RF S SP = sp - 4
  /\ sp >= _BLK_Stk
  /\ subsetQ_TQ RdyQ T
  /\ WQMod.not_in_dom tid RdyQ
  /\ exists_Idle T RdyQ tid
  /\ TCB_Status tcb = READY
  /\ TQMod.maps_to T tid tcb
  /\ WF_TQ T.

Definition p_block_7 (S : State) : Prop := 
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists tid, exists tcb, exists sp, exists pc, exists jk1, exists jk2,
      Ast_block_7 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mTrue S.

Definition g_block_7 (S S' : State) := forall mP,
  forall T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2,
    Ast_block_7 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 mP S
    -> exists T', exists RdyQ', exists BlkQ', exists tid', exists sp', exists tcb',
      T' = TQMod.put T tid (sp, BLOCKED, pc)
      /\ BlkQ' = addThrd BlkQ tid
      /\ RdyQ = addThrd RdyQ' tid'
      /\ Ast_block_post T' RdyQ' BlkQ' l_ptr tid' tcb' sp' sp pc jk1 jk2 mP S'.

(* 1165 <block+7>:  call    L_enQueue	*)

Definition Ast_block_8 (T : TQ) (RdyQ BlkQ : WQ) (l_ptr : Label)
  (tid : TID) (tcb : Thread) (sp pc jk1 jk2 : nat) 
  (T' : TQ) (BlkQ' : WQ) (mP : MPred) (S : State) : Prop :=
  ((curr_thrd |->w tid)
    *^* (embed_RdyQ T' RdyQ) 
    *^* (embed_BlkQ l_ptr T' BlkQ') 
    *^* (stkSpacePrec (sp - 4) (_BLK_Stk - 4)) 
    *^* (sp - 4 |->w tid)
    *^* (sp - 2 |->w l_ptr)
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk1) 
    *^* (sp + 2 |-> l_ptr) 
    *^* (sp + 3 |-> jk2) 
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 4
  /\ sp >= _BLK_Stk
  /\ subsetQ_TQ RdyQ T
  /\ WQMod.not_in_dom tid RdyQ
  /\ exists_Idle T RdyQ tid
  /\ TCB_Status tcb = READY
  /\ TQMod.maps_to T tid tcb
  /\ T' = TQMod.put T tid (sp, BLOCKED, pc)
  /\ BlkQ' = addThrd BlkQ tid.

Definition p_block_8 (S : State) : Prop := 
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists tid, exists tcb, exists sp, exists pc,  exists jk1, exists jk2,
      exists T', exists BlkQ',
        Ast_block_8 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 T' BlkQ' mTrue S.

Definition g_block_8 (S S' : State) := forall mP,
  forall T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 T' BlkQ',
    Ast_block_8 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 T' BlkQ' mP S
    -> exists T', exists RdyQ', exists BlkQ', exists tid', exists sp', exists tcb',
      T' = TQMod.put T tid (sp, BLOCKED, pc)
      /\ BlkQ' = addThrd BlkQ tid
      /\ RdyQ = addThrd RdyQ' tid'
      /\ Ast_block_post T' RdyQ' BlkQ' l_ptr tid' tcb' sp' sp pc jk1 jk2 mP S'.

(* 1166 <block+8>:  pushw   $readyQ             #readyQ = 2250 *)

Definition Ast_block_9 (T : TQ) (RdyQ BlkQ : WQ) (l_ptr : Label)
  (tid : TID) (tcb : Thread) (sp pc jk1 jk2 : nat)
  (T' : TQ) (BlkQ' : WQ) (mP : MPred) (S : State) : Prop :=
  ((curr_thrd |->w tid)
    *^* (embed_RdyQ T' RdyQ) 
    *^* (embed_BlkQ l_ptr T' BlkQ') 
    *^* (stkSpacePrec (sp - 6) (_BLK_Stk - 6)) 
    *^* (sp - 6 |->w ready_Q)
    *^* (sp - 4 |->w tid)
    *^* (sp - 2 |->w l_ptr)
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk1) 
    *^* (sp + 2 |-> l_ptr) 
    *^* (sp + 3 |-> jk2) 
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 6
  /\ sp >= _BLK_Stk
  /\ subsetQ_TQ RdyQ T
  /\ WQMod.not_in_dom tid RdyQ
  /\ exists_Idle T RdyQ tid
  /\ TCB_Status tcb = READY
  /\ TQMod.maps_to T tid tcb
  /\ T' = TQMod.put T tid (sp, BLOCKED, pc)
  /\ BlkQ' = addThrd BlkQ tid.

Definition p_block_9 (S : State) : Prop := 
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists tid, exists tcb, exists sp, exists pc, exists jk1, exists jk2,
      exists T', exists BlkQ',
        Ast_block_9 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 T' BlkQ' mTrue S.

Definition g_block_9 (S S' : State) := forall mP,
  forall T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 T' BlkQ',
    Ast_block_9 T RdyQ BlkQ l_ptr tid tcb sp pc jk1 jk2 T' BlkQ' mP S
    -> exists T', exists RdyQ', exists BlkQ', exists tid', exists sp', exists tcb',
      T' = TQMod.put T tid (sp, BLOCKED, pc)
      /\ BlkQ' = addThrd BlkQ tid
      /\ RdyQ = addThrd RdyQ' tid'
      /\ Ast_block_post T' RdyQ' BlkQ' l_ptr tid' tcb' sp' sp pc jk1 jk2 mP S'.

(* 1167 <block+9>:  call    L_deQueue   *)

Definition Ast_block_10 (T : TQ) (RdyQ BlkQ : WQ) (l_ptr : Label)
  (tid : TID) (sp pc jk1 jk2 : nat) 
  (T' : TQ) (RdyQ' BlkQ' : WQ) (tid' : TID) (tcb' : Thread)
  (mP : MPred) (S : State) : Prop :=
  ((curr_thrd |->w tid)
    *^* (embed_TCB tid' tcb')
    *^* (embed_RdyQ T' RdyQ')
    *^* (embed_BlkQ l_ptr T' BlkQ')
    *^* (stkSpacePrec (sp - 6) (_BLK_Stk - 6))
    *^* (sp - 6 |->w ready_Q)
    *^* (sp - 4 |->w tid)
    *^* (sp - 2 |->w l_ptr)
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk1) 
    *^* (sp + 2 |-> l_ptr) 
    *^* (sp + 3 |-> jk2) 
    *^* mP) (S_Mem S)
  /\ S_RF S AX = tid'
  /\ S_RF S SP = sp - 6
  /\ sp >= _BLK_Stk
  /\ T' = TQMod.put T tid (sp, BLOCKED, pc)
  /\ BlkQ' = addThrd BlkQ tid
  /\ RdyQ = addThrd RdyQ' tid'
  /\ TQMod.maps_to T' tid' tcb'
  /\ WQMod.not_in_dom tid' RdyQ'.

Definition p_block_10 (S : State) : Prop := 
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists tid, exists sp, exists pc, exists jk1, exists jk2,
      exists T', exists RdyQ', exists BlkQ', 
        exists tid', exists tcb',
   Ast_block_10 T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb' mTrue S.

Definition g_block_10 (S S' : State) := forall mP,
  forall T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb',
   Ast_block_10 T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb' mP S
    -> exists T', exists RdyQ', exists BlkQ', exists tid', exists sp', exists tcb',
      T' = TQMod.put T tid (sp, BLOCKED, pc)
      /\ BlkQ' = addThrd BlkQ tid
      /\ RdyQ = addThrd RdyQ' tid'
      /\ Ast_block_post T' RdyQ' BlkQ' l_ptr tid' tcb' sp' sp pc jk1 jk2 mP S'.

(* 1168 <block+10>: addw    $6, %sp     *)

Definition Ast_block_11 (T : TQ) (RdyQ BlkQ : WQ) (l_ptr : Label)
  (tid : TID) (sp pc jk1 jk2 : nat) 
  (T' : TQ) (RdyQ' BlkQ' : WQ) (tid' : TID) (tcb' : Thread)
  (mP : MPred) (S : State) : Prop :=
  ((curr_thrd |->w tid)
    *^* (embed_TCB tid' tcb')
    *^* (embed_RdyQ T' RdyQ')
    *^* (embed_BlkQ l_ptr T' BlkQ')
    *^* (stkSpacePrec sp _BLK_Stk)
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk1) 
    *^* (sp + 2 |-> l_ptr) 
    *^* (sp + 3 |-> jk2) 
    *^* mP) (S_Mem S)
  /\ S_RF S AX = tid'
  /\ S_RF S SP = sp
  /\ sp >= _BLK_Stk
  /\ T' = TQMod.put T tid (sp, BLOCKED, pc)
  /\ BlkQ' = addThrd BlkQ tid
  /\ RdyQ = addThrd RdyQ' tid'
  /\ TQMod.maps_to T' tid' tcb'
  /\ WQMod.not_in_dom tid' RdyQ'.

Definition p_block_11 (S : State) : Prop := 
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists tid, exists sp, exists pc, exists jk1, exists jk2,
      exists T', exists RdyQ', exists BlkQ', 
        exists tid', exists tcb',
   Ast_block_11 T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb' mTrue S.

Definition g_block_11 (S S' : State) := forall mP,
  forall T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb',
   Ast_block_11 T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb' mP S
    -> exists T', exists RdyQ', exists BlkQ', exists tid', exists sp', exists tcb',
      T' = TQMod.put T tid (sp, BLOCKED, pc)
      /\ BlkQ' = addThrd BlkQ tid
      /\ RdyQ = addThrd RdyQ' tid'
      /\ Ast_block_post T' RdyQ' BlkQ' l_ptr tid' tcb' sp' sp pc jk1 jk2 mP S'.

(* 1169 <block+11>: movw    %ax, %bx    *)

Definition Ast_block_12 (T : TQ) (RdyQ BlkQ : WQ) (l_ptr : Label)
  (tid : TID) (sp pc jk1 jk2 : nat) 
  (T' : TQ) (RdyQ' BlkQ' : WQ) (tid' : TID) (tcb' : Thread)
  (mP : MPred) (S : State) : Prop :=
  ((curr_thrd |->w tid)
    *^* (embed_TCB tid' tcb')
    *^* (embed_RdyQ T' RdyQ')
    *^* (embed_BlkQ l_ptr T' BlkQ')
    *^* (stkSpacePrec sp _BLK_Stk)
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk1) 
    *^* (sp + 2 |-> l_ptr) 
    *^* (sp + 3 |-> jk2) 
    *^* mP) (S_Mem S)
  /\ S_RF S AX = tid'
  /\ S_RF S BX = tid'
  /\ S_RF S SP = sp
  /\ sp >= _BLK_Stk
  /\ T' = TQMod.put T tid (sp, BLOCKED, pc)
  /\ BlkQ' = addThrd BlkQ tid
  /\ RdyQ = addThrd RdyQ' tid'
  /\ TQMod.maps_to T' tid' tcb'
  /\ WQMod.not_in_dom tid' RdyQ'.

Definition p_block_12 (S : State) : Prop := 
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists tid, exists sp, exists pc, exists jk1, exists jk2,
      exists T', exists RdyQ', exists BlkQ', 
        exists tid', exists tcb',
   Ast_block_12 T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb' mTrue S.

Definition g_block_12 (S S' : State) := forall mP,
  forall T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb',
   Ast_block_12 T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb' mP S
    -> exists T', exists RdyQ', exists BlkQ', exists tid', exists sp', exists tcb',
      T' = TQMod.put T tid (sp, BLOCKED, pc)
      /\ BlkQ' = addThrd BlkQ tid
      /\ RdyQ = addThrd RdyQ' tid'
      /\ Ast_block_post T' RdyQ' BlkQ' l_ptr tid' tcb' sp' sp pc jk1 jk2 mP S'.

(* 1170 <block+12>: movw    %ax, (curr_thrd) *)

Definition Ast_block_13 (T : TQ) (RdyQ BlkQ : WQ) (l_ptr : Label)
  (tid : TID) (sp pc jk1 jk2 : nat) 
  (T' : TQ) (RdyQ' BlkQ' : WQ) (tid' : TID) (tcb' : Thread)
  (mP : MPred) (S : State) : Prop :=
  ((curr_thrd |->w tid')
    *^* (embed_TCB tid' tcb')
    *^* (embed_RdyQ T' RdyQ')
    *^* (embed_BlkQ l_ptr T' BlkQ')
    *^* (stkSpacePrec sp _BLK_Stk)
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk1) 
    *^* (sp + 2 |-> l_ptr) 
    *^* (sp + 3 |-> jk2) 
    *^* mP) (S_Mem S)
  /\ S_RF S BX = tid'
  /\ S_RF S SP = sp
  /\ sp >= _BLK_Stk
  /\ T' = TQMod.put T tid (sp, BLOCKED, pc)
  /\ BlkQ' = addThrd BlkQ tid
  /\ RdyQ = addThrd RdyQ' tid'
  /\ TQMod.maps_to T' tid' tcb'
  /\ WQMod.not_in_dom tid' RdyQ'.

Definition p_block_13 (S : State) : Prop := 
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists tid, exists sp, exists pc, exists jk1, exists jk2,
      exists T', exists RdyQ', exists BlkQ', 
        exists tid', exists tcb',
   Ast_block_13 T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb' mTrue S.

Definition g_block_13 (S S' : State) := forall mP,
  forall T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb',
   Ast_block_13 T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb' mP S
    -> exists T', exists RdyQ', exists BlkQ', exists tid', exists sp', exists tcb',
      T' = TQMod.put T tid (sp, BLOCKED, pc)
      /\ BlkQ' = addThrd BlkQ tid
      /\ RdyQ = addThrd RdyQ' tid'
      /\ Ast_block_post T' RdyQ' BlkQ' l_ptr tid' tcb' sp' sp pc jk1 jk2 mP S'.

(* 1171 <block+13>: movw    TCB_STACK(%bx), %sp *)

Definition Ast_block_14 (T : TQ) (RdyQ BlkQ : WQ) (l_ptr : Label)
  (tid : TID) (sp pc jk1 jk2 : nat) 
  (T' : TQ) (RdyQ' BlkQ' : WQ) (tid' : TID) (tcb' : Thread) (sp' : nat)
  (mP : MPred) (S : State) : Prop :=
  ((curr_thrd |->w tid')
    *^* (embed_TCB tid' tcb')
    *^* (embed_RdyQ T' RdyQ')
    *^* (embed_BlkQ l_ptr T' BlkQ')
    *^* (stkSpacePrec sp _BLK_Stk)
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk1) 
    *^* (sp + 2 |-> l_ptr) 
    *^* (sp + 3 |-> jk2) 
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp'
  /\ sp >= _BLK_Stk
  /\ T' = TQMod.put T tid (sp, BLOCKED, pc)
  /\ BlkQ' = addThrd BlkQ tid
  /\ RdyQ = addThrd RdyQ' tid'
  /\ TCB_SP tcb' = sp'
  /\ TQMod.maps_to T' tid' tcb'
  /\ WQMod.not_in_dom tid' RdyQ'.

Definition p_block_14 (S : State) : Prop := 
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists tid, exists sp, exists pc, exists jk1, exists jk2,
      exists T', exists RdyQ', exists BlkQ', 
        exists tid', exists tcb', exists sp',
   Ast_block_14 T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb' sp' mTrue S.

Definition g_block_14 (S S' : State) := forall mP,
  forall T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb' sp',
   Ast_block_14 T RdyQ BlkQ l_ptr tid sp pc jk1 jk2 T' RdyQ' BlkQ' tid' tcb' sp' mP S
    -> exists T', exists RdyQ', exists BlkQ', exists tid', exists sp', exists tcb',
      T' = TQMod.put T tid (sp, BLOCKED, pc)
      /\ BlkQ' = addThrd BlkQ tid
      /\ RdyQ = addThrd RdyQ' tid'
      /\ Ast_block_post T' RdyQ' BlkQ' l_ptr tid' tcb' sp' sp pc jk1 jk2 mP S'.

(* 1172 <block+14>: retw                *)

Lemma p_block_sound : forall S,
  Linking.BLK_Prd S -> p_block S.
Proof.
  intros S HS.
  unfold p_block, BLK_Prd in * .
  destruct HS as [T [RdyQ [BlkQ [tid [sp [l_ptr [tcb [pc HS]]]]]]]].
  destruct HS as [HM [Hex [Hmto [HRT [HBT [Htcb [HSP [HtR [HtB [HRB HT]]]]]]]]]].
  destruct HM as [jk1 [jk2 HM]].
  exists T; exists RdyQ; exists BlkQ; exists l_ptr. 
  exists tid; exists tcb; exists sp; exists pc; exists jk1; exists jk2.
  destruct S as [[[M R] fl] isr].
  unfold Ast_block_pre, BLK_MPre in * .
  simplS in * .
  split.
    sep on.
    sep lift 0 in HM.
    substH HM with (embed_curr_TCB_elim1 _ _ _ _ HM).
    destruct HM as [tcbx [Hmtox HM]].
    assert (tcbx = tcb).
      unfold TQMod.maps_to in * .
      rewrite Hmto in Hmtox.
      injection Hmtox; auto.
    subst tcbx.
    unfold embed_TCB.
    unfold embed_BlkQ.
    sep matchH HM with 12.
  split_l.
  split.
    sep on.
    sep lift 7 in HM.
    assert (HMx : M ||- stkSpace sp 10).
      sep trivial.
    destruct HMx; trivial.
  repeat split_l; trivial.
Qed.

Lemma g_block_sound : forall S S',
  g_block S S' -> Linking.BLK_grt S S'.
Proof.
  unfold g_block, BLK_grt.
  intros S S' HSS'.
  intros M0 M1 HJM T T' RdyQ BlkQ BlkQ' tid sp BlkQ_ptr tcb pc jk1 jk2.
  unfold BLK_Prd_aux.
  intros H.
  destruct H as [HMPre [Hex [Hmto [HRT [HBT [Htcb [HSP [HtR [HtB [HRB HT]]]]]]]]]].
  intros HT' HBlkQ'.
  substH HSS' with (HSS' (mEq M1) T RdyQ BlkQ BlkQ_ptr tid tcb sp pc).
  assert (Ast_block_pre T RdyQ BlkQ BlkQ_ptr tid tcb sp pc jk1 jk2 (mEq M1) S).
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    unfold BLK_MPre in HMPre.
    simplS in * .
    split.
      simplS.
      exists M0; exists M1; split_l.
      rsplit.
        unfold mEq; trivial.
      sep on.
      sep lift 0 in HMPre.
      substH HMPre with (embed_curr_TCB_elim1 _ _ _ _ HMPre).
      destruct HMPre as [tcbx [Hmtox HM]].
      assert (tcbx = tcb).
        unfold TQMod.maps_to in * .
        rewrite Hmto in Hmtox.
        injection Hmtox; auto.
      subst tcbx.
      unfold embed_TCB.
      unfold embed_BlkQ.
      sep matchH HM with 12.
    simplS.
    split_l.
    split.
      sep on.
      sep lift 7 in HMPre.
      assert (HMx : M0 ||- stkSpace sp 10).
      sep trivial.
      destruct HMx; trivial.
    repeat split_l; trivial.
  substH HSS' with (HSS' jk1 jk2 H).
  destruct HSS' as [Tx' [RdyQx' [BlkQx' [tid' [sp' [tcb' Hx]]]]]].
  destruct Hx as [HTx' [HBlkQx' [HRdyQ HS']]].
  unfold Ast_block_post in HS'.
  destruct HS' as [HM' [Hsp' [Hmto' [Htcb' Hnin']]]].
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  simplS in * .
  destruct HM' as [M2' [M1' [HJM' [HM2' HM1']]]].
  unfold mEq in HM1'; subst M1'.
  exists M2'; exists RdyQx'; exists tid'; exists sp'; exists tcb'.
  split_l.
  split_l.
  unfold BLK_Post_aux.
  unfold BLK_MPost.
  simplS.
  assert (Tx' = T').
    subst T' Tx'; trivial.
  clear HTx'.
  subst Tx'.
  split_l.
  split_l.
  split_l.
  split_l.
  sep on.
  sep lift 0.
  apply embed_curr_TCB_intro1 with tcb'; trivial.
  unfold embed_BlkQ in HM2'.
  unfold embed_TCB in HM2'.
  sep matchH HM2' with 8.
  rewrite <- HBlkQ' in HBlkQx'.
  subst BlkQx'.
  trivial.
Qed.

Definition spec_block := Ast p_block g_block sctag.

Definition psi_block  :=
  (SCAP_consCT L_enQueue (Ast p_enQueue g_enQueue sctag)
  (SCAP_consCT L_deQueue (Ast p_deQueue g_deQueue sctag)
    SCAP_emptCT)).

Lemma L_deQueue_in_psi_block : 
  SCAP_lookupCT psi_block L_deQueue (Ast p_deQueue g_deQueue sctag).
Proof.
  unfold SCAP_lookupCT, psi_block.
  unfold SCAP_consCT.
  apply SPSpec.lookup_merge_intro_r.
  simpl.
  SPSpec.lookup_solve.
Qed.

Lemma L_enQueue_in_psi_block : 
  SCAP_lookupCT psi_block L_enQueue (Ast p_enQueue g_enQueue sctag).
Proof.
  unfold SCAP_lookupCT, psi_block.
  unfold SCAP_consCT.
  simpl.
  SPSpec.lookup_solve.
Qed.
