(***************************************************************)
(* Coq code for AIM                                            *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(* by wangwei *)
Add LoadPath "..".

Add LoadPath "../CAP0".
Add LoadPath "../SCAP".

Require Import Coqlib.

Require Import Arith.
Require Import Machine.
Require Import Assertion.
Require Import Operational_Semantics.
Require Import OpSem_Lemmas.

Require Import AIM.
Require Import SCAP.
Require Import QLib.
Require Import Linking.
Require Import Linking_Lemmas3.
Require Import QLemmas.

Require Import Reg.
Require Import Assertion_Lemmas3.
Require Import Assertion_Lemmas4.

Require Import Kernel.

Require Import DeQueueSpec.

(***************************************)

(*  <DeQueue + 0> : pushw %bp *)
Lemma SCAP_WFcmd_deQueue_0_pushw :
  SCAP_WFcmd (SCAP_consCT (L_deQueue + 1) (Ast L_deQueue_1_p L_deQueue_1_g sctag) psi_deQueue_Block0)
    (L_deQueue) (Ast p_deQueue g_deQueue sctag) 
    (comml (_pushwp (Opr_reg BP))).
Proof.
  assert (Hpp': forall T Q ptr bp sp mP S S',
    Ast_deQueue_pre T Q ptr bp sp mP S
    -> nextS (L_deQueue) (comml (_pushwp (Opr_reg BP))) S S'
    -> Ast_deQueue_1 T Q ptr bp sp  mP S').
    unfold Ast_deQueue_pre,Ast_deQueue_1.
    intros T Q ptr bp sp mP S S' Hp HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [[HSS1 [HSS2 [jk HSS3]]] [HSS4 [HSS5 HSS6]]].
    destruct Hp as [Hp1 [Hp2 Hp3]].
    subst M' R' fl' isr'.
    rewrite Hp2; rewrite Hp3.
    intuition.
    sep on.
    sep lift 1 in Hp1.
    sep lift 1.
    assert (MemUpdW M (sp - 2) bp jk ||- 
      sp - 2 |->w bp *^* (sp + 2 |->w ptr *^* mP *^* ptoTQ ptr (filterTQ T Q))).
      sep upd.
      sep matchH Hp1 with 10.
      sep trivial.
    unfold MemUpdW in H.
    assumption.
  unfold SCAP_WFcmd.
  exists L_deQueue_1_p;exists L_deQueue_1_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [ptr [bp [sp HS]]]]].
  destruct HS as [HS1 [HS2 HS3]].
  exists (MemUpdW M (sp-2) bp 0, RF_update R SP (sp - 2), fl, isr).
  simplS in * .
  apply nextS_pushwp with bp.
  unfold opr_eval.
  rewrite HS2;reflexivity.
  unfold pushW.
  intuition.
  rewrite HS3.
    sep on.
    sep lift 1 in HS1.
    sep trivial.
  exists 0.
  unfold MemUpdW.
  rewrite HS3;reflexivity.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold p_deQueue, L_deQueue_1_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [ptr [bp [sp H]]]]].
  exists T; exists Q; exists ptr; exists bp; exists sp.
  apply Hpp' with S;assumption.
  (* g' ==> g *)
  unfold g_deQueue, L_deQueue_1_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S''.
  apply Hpp' with S;assumption.
Qed.

(* <DeQueue + 1> : movw %sp, %bp *)
Lemma SCAP_WFcmd_deQueue_1_movwp :
  SCAP_WFcmd (SCAP_consCT (L_deQueue + 1 + 1) (Ast L_deQueue_2_p L_deQueue_2_g sctag) psi_deQueue_Block0)
    (L_deQueue + 1) (Ast L_deQueue_1_p L_deQueue_1_g sctag) 
    (comml (_movwp (Opr_reg SP) BP)).
Proof.
  assert (Hpp': forall T Q ptr bp sp mP S S',
    Ast_deQueue_1 T Q ptr bp sp mP S
    -> nextS (L_deQueue + 1) (comml (_movwp (Opr_reg SP) BP)) S S'
    -> Ast_deQueue_2 T Q ptr bp sp  mP S').
    unfold Ast_deQueue_1,Ast_deQueue_2.
    intros T Q ptr bp sp mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [HSS1 [HSS2 [HSS3 HSS4]]].
    destruct HS as [HS1 [HS2 HS3]].
    subst M' R' fl' isr'.
    rewrite HS2.
    intuition.
  unfold SCAP_WFcmd.
  exists L_deQueue_2_p;exists L_deQueue_2_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [ptr [bp [sp HS]]]]].
  destruct HS as [HS1 [HS2 HS3]].
  exists (M, RF_update R BP (sp - 2), fl, isr).
  simplS in * .
  apply nextS_movwp with (sp - 2);trivial.
  unfold opr_eval.
  rewrite HS2;reflexivity.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_deQueue_1_p, L_deQueue_2_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [ptr [bp [sp H]]]]].
  exists T; exists Q; exists ptr; exists bp; exists sp.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_deQueue_1_g, L_deQueue_2_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S''.
  apply Hpp' with S;assumption.
Qed.

(* <DeQueue + 2> : movw 4(%bp), %bx *)
Lemma SCAP_WFcmd_deQueue_2_movwld :
  SCAP_WFcmd (SCAP_consCT (L_deQueue + 2 + 1) (Ast L_deQueue_3_p L_deQueue_3_g sctag) psi_deQueue_Block0)
    (L_deQueue + 2) (Ast L_deQueue_2_p L_deQueue_2_g sctag) 
    (comml (_movwld (A_reg 4 BP) BX)).
Proof.
  assert (Hpp': forall T Q ptr bp sp mP S S',
    Ast_deQueue_2 T Q ptr bp sp mP S
    -> nextS (L_deQueue + 2) (comml (_movwld (A_reg 4 BP) BX)) S S'
    -> exists head, exists tail, exists prev,
      Ast_deQueue_3 T Q ptr bp sp head tail prev mP S').
    unfold Ast_deQueue_2,Ast_deQueue_3.
    intros T Q ptr bp sp mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [HSS1 [[x [HSS2 HSS3]] [HSS4 HSS5]]].
    destruct HS as [HS1 [HS2 [HS3 HS4]]].
    subst M' R' fl' isr'.
    sep on.
    sep lift 0 in HS1.
    destruct (ptoTQ_elim1 HS1) as [head [tail [prev HS]]].
    clear HS1.
    exists head; exists tail; exists prev.
    assert (E: x = ptr).
      rewrite HS3 in HSS2.
      sep on.
      sep lift 4 in HS.
      arith_rep (sp-2+4) with (sp+2) in HSS2.
      sep inj;auto.
    intuition.
  unfold SCAP_WFcmd.
  exists L_deQueue_3_p;exists L_deQueue_3_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [ptr [bp [sp HS]]]]].
  destruct HS as [HS1 [HS2 [HS3 HS4]]].
  exists (M, RF_update R BX ptr, fl, isr).
  simplS in * .
  apply nextS_movwld with (sp+2) ptr;trivial.
  unfold addr_eval.
  rewrite HS3;omega.
    sep on.
    sep lift 2 in HS1.
    sep trivial.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_deQueue_2_p, L_deQueue_3_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [ptr [bp [sp H]]]]].
  exists T; exists Q; exists ptr; exists bp; exists sp.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_deQueue_2_g, L_deQueue_3_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  assert (exists head : nat,
           exists tail : nat,
             exists prev : nat,
               Ast_deQueue_3 T Q ptr bp sp head tail prev mP S').
  apply Hpp' with S;assumption.
  destruct H0 as [head [tail [prev H0]]].
  apply HgS'S'' with head tail prev;assumption.
Qed.

(* <DeQueue + 3> : movw Q_HEAD(%bx), %dx *)
Lemma SCAP_WFcmd_deQueue_3_movwld :
  SCAP_WFcmd (SCAP_consCT (L_deQueue + 3 + 1) (Ast L_deQueue_4_p L_deQueue_4_g sctag) psi_deQueue_Block0)
    (L_deQueue + 3) (Ast L_deQueue_3_p L_deQueue_3_g sctag) 
    (comml (_movwld (A_reg Q_HEAD BX) DX)).
Proof.
  assert (Hpp': forall T Q ptr bp sp head tail prev mP S S',
    Ast_deQueue_3 T Q ptr bp sp head tail prev mP S
    -> nextS (L_deQueue + 3) (comml (_movwld (A_reg Q_HEAD BX) DX)) S S'
    -> Ast_deQueue_4 T Q ptr bp sp head tail prev mP S').
    unfold Ast_deQueue_3, Ast_deQueue_4.
    intros T Q ptr bp sp head tail prev mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [HSS1 [[x [HSS2 HSS3]] [HSS4 HSS5]]].
    destruct HS as [HS1 [HS2 [HS3 [HS4 HS5]]]].
    subst M' R' fl' isr'.
    assert (E: x = head).
      rewrite HS4 in HSS2.
      arith_rep (ptr + 0) with ptr in HSS2.
      sep on.
      sep lift 0 in HS1.
      sep inj;auto.
    intuition.
  unfold SCAP_WFcmd.
  exists L_deQueue_4_p; exists L_deQueue_4_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [ptr [bp [sp [head [tail [prev HS]]]]]]]].
  destruct HS as [HS1 [HS2 [HS3 [HS4 HS5]]]].
  exists (M, RF_update R DX head, fl, isr).
  simplS in * .
  apply nextS_movwld with ptr head;trivial.
  unfold addr_eval.
  rewrite HS4;omega.
    sep on.
    sep lift 0 in HS1.
    sep trivial.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_deQueue_3_p, L_deQueue_4_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [ptr [bp [sp [head [tail [prev H]]]]]]]].
  exists T; exists Q; exists ptr; exists bp; exists sp.
  exists head; exists tail; exists prev.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_deQueue_3_g, L_deQueue_4_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S'' with head tail prev.
  apply Hpp' with S;assumption.
Qed.

Lemma filterTQ_add_sub_equal:
  forall T Q head  tcb TQ',
  filterTQ T Q =
  (fun l' : nat => if eq_nat_dec head l' then Some tcb else TQ' l')  
  -> TQ' head = None
  -> TQMod.maps_to T head tcb /\
  filterTQ T (fun l : nat => if eq_nat_dec head l then None else Q l) = TQ'/\
  Q = addThrd (fun l : nat => if eq_nat_dec head l then None else Q l) head.
Proof.
  intros.
  split.
  assert (filterTQ T Q head =  (fun l' : nat => if eq_nat_dec head l' then Some tcb else TQ' l') head).
  rewrite H.
  reflexivity.
  unfold filterTQ in H1.
  destruct (Q head); destruct (eq_nat_dec head head).
  unfold TQMod.maps_to.
  assumption.
  elim n; reflexivity.
  inversion H1.
  elim n; reflexivity.
  assert (P :forall y,(filterTQ T Q) y =
      (fun l' : nat => if eq_nat_dec head l' then Some tcb else TQ' l') y).
  rewrite H.
  reflexivity.
  clear H.
  split.
  apply TQMod.extensional_eqMap.
  intro x.
  substH P with (P x).
  unfold filterTQ in * .
  destruct (eq_nat_dec head x).
  subst head.
  rewrite H0.
  reflexivity.
  rewrite <- P.
  reflexivity.
  unfold addThrd.
  unfold WQMod.put.
  apply WQMod.extensional_eqMap.
  intro x.
  substH P with (P x).
  simpl in P.
  unfold WQMod.beq_A.
  unfold WQSpec.beq_Domain.
  unfold filterTQ in P.
  destruct (eq_nat_dec x head);destruct (eq_nat_dec head x); destruct (Q x).
  induction b; reflexivity.
  inversion P.
  induction b; reflexivity.
  rewrite e in n;elim n;reflexivity.
  rewrite e in n;elim n;reflexivity.
  rewrite e in n;elim n;reflexivity.
  reflexivity.
  reflexivity.
Qed.

Lemma Queue_T_split : 
  forall head tail prev T Q M ,
    M ||- Queue Thread ptoTCB head tail prev 0 (filterTQ T Q)
    -> head > 0
    -> exists Q', exists tcb, exists next, exists next', exists tail', 
      M ||- (ptoNode Thread ptoTCB head prev next tcb *^*
        Queue Thread ptoTCB next' tail' head 0 (filterTQ T Q'))
      /\ Q = addThrd Q' head /\  WQMod.not_in_dom head Q'/\
      TQMod.maps_to T head tcb
      /\ (tail' = tail /\ tail' > 0 /\ next' = next \/ 
     tail' = 0/\ tail = head /\ next' = 0 /\ next = 0).
Proof.
  intros head tail prev T Q M HQ Hhead.
  substH HQ with (Queue_unemp_get HQ Hhead).
  destruct HQ as [TQ' [tcb [next' [next'' [tail' [HM [HQTQ' [HTQ' Htail']]]]]]]].
  exists (fun l : nat => if eq_nat_dec head l then None else Q l); exists tcb; exists next'; exists next''; exists tail'.
  destruct (filterTQ_add_sub_equal T Q head  tcb TQ' HQTQ' HTQ') as [Pmap [PTQ' PQQ']].
  rewrite PTQ'.
  unfold WQMod.not_in_dom.
  destruct (eq_nat_dec head head).
  intuition.
  elim n; reflexivity.
Qed.
Implicit Arguments Queue_T_split [head tail prev T Q M].

Lemma Queue_T_split_1 : 
  forall head tail prev T Q M A,
    M ||- (Queue Thread ptoTCB head tail prev 0 (filterTQ T Q))*^* A
    -> head > 0
    -> exists Q', exists tcb, exists next, exists next', exists tail', 
      M ||- (ptoNode Thread ptoTCB head prev next tcb *^*
        Queue Thread ptoTCB next' tail' head 0 (filterTQ T Q'))*^* A
      /\ Q = addThrd Q' head /\  WQMod.not_in_dom head Q'/\
      TQMod.maps_to T head tcb
      /\ (tail' = tail /\ tail' > 0 /\ next' = next \/ 
     tail' = 0/\ tail = head /\ next' = 0 /\ next = 0).
Proof.
  intros.
  destruct H  as [M1 [M2 [P1 [P2 P3]]]].
  substH P2 with (Queue_T_split P2).
  substH P2 with (P2 H0).
  InvertAll.
  exists x; exists x0; exists x1; exists x2; exists x3.
  split.
  exists M1;exists M2;intuition.
  intuition.
Qed.
Implicit Arguments Queue_T_split_1 [head tail prev T Q M A]. 

(* <DeQueue + 4> : cmpw $0, %dx *)
Lemma SCAP_WFcmd_deQueue_4_cmpw :
  SCAP_WFcmd (SCAP_consCT (L_deQueue + 4 + 1) (Ast L_deQueue_5_p L_deQueue_5_g sctag) psi_deQueue_Block0)
    (L_deQueue + 4) (Ast L_deQueue_4_p L_deQueue_4_g sctag) 
    (comml (_cmpw (Opr_imm 0) DX)).
Proof.
  assert (Hpp': forall T Q ptr bp sp head tail prev mP S S',
    Ast_deQueue_4 T Q ptr bp sp head tail prev mP S
    -> nextS (L_deQueue + 4)  (comml (_cmpw (Opr_imm 0) DX)) S S'
    -> (Ast_deQueue_5_emp T Q ptr bp sp prev mP S' \/
    exists Q', exists tcb, exists next, exists next', exists tail', 
      Ast_deQueue_5_get T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S')).
    unfold Ast_deQueue_4, Ast_deQueue_5_emp, Ast_deQueue_5_get.
    intros T Q ptr bp sp head tail prev mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [HSS1 [HSS2 [HSS3 HSS4]]].
    destruct HS as [HS1 [HS2 [HS3 [HS4 [HS5 HS6]]]]].
    destruct HSS3 as [HSS3|HSS3];destruct HSS3 as [HSS31 HSS32].
    subst M' R' fl' isr'.
    left.
    rewrite <- HSS31 in HS5.
    subst head.
    intuition.
    sep on.
    sep lift 2 in HS1.
    assert (Tail0 := Queue_head0_tail0_1 HS1).
    subst tail.
    sep matchH HS1 with 10.
    right.
    sep on.
    sep lift 2 in HS1.
    assert (headn0 : head > 0).
      rewrite HS5 in HSS31.
      omega.
    destruct (Queue_T_split_1 HS1 headn0) as [Q' [tcb [next [next' [tail' [P1 [P2 [P3 [P4 P5]]]]]]]]].
    exists Q'; exists tcb; exists next; exists next'; exists tail'.
    subst M' R' fl' isr'.
    split.
    sep on.
    sep matchH P1 with 10.
    intuition.
  unfold SCAP_WFcmd.
  exists L_deQueue_5_p; exists L_deQueue_5_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [ptr [bp [sp [head [tail [prev HS]]]]]]]].
  destruct HS as [HS1 [HS2 [HS3 [HS4 [HS5 HS6]]]]].
  destruct (eq_nat_dec 0 head).
    exists (M, R,Set_Flag fl _ZF, isr).
    simplS in * .
    apply nextS_cmpw;trivial.
    left.
    intuition.
    simpl.
    rewrite <- e in HS5;auto.
    exists (M, R,Clear_Flag fl _ZF, isr).
    simplS in * .
    apply nextS_cmpw;trivial.
    right.
    intuition.
    simpl in H.
    apply n;rewrite <- H in HS5;auto.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_deQueue_4_p, L_deQueue_5_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [ptr [bp [sp [head [tail [prev H]]]]]]]].
  exists T; exists Q; exists ptr; exists bp; exists sp.
  exists head; exists tail; exists prev.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_deQueue_4_g, L_deQueue_5_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S'' with head tail prev.
  apply Hpp' with S;assumption.
Qed.

(* <DeQueue + 5> : jne deq_a *)
Lemma SCAP_WFcmd_deQueue_5_jne :
  SCAP_WFcmd (SCAP_consCT (L_deQueue + 5 + 1) (Ast L_deQueue_6_p L_deQueue_6_g sctag) psi_deQueue_Block0)
    (L_deQueue + 5) (Ast L_deQueue_5_p L_deQueue_5_g sctag) 
    (comml (_jne L_deQ_a)).
Proof.
  unfold SCAP_WFcmd.
  exists L_deQueue_9_p; exists L_deQueue_6_p.
  exists L_deQueue_9_g; exists L_deQueue_6_g.
  split.
  apply psi_cons_b_lookup_a.
  unfold psi_deQueue_Block0.
  apply psi_cons_a_lookup_a.
  split_l.
  assert (Hget: forall T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S,
    Ast_deQueue_5_get T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S
    -> Ast_deQueue_9 T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S).
    unfold Ast_deQueue_5_get, Ast_deQueue_9.
  intros.
  InvertAll.
  intuition.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_deQueue_5_p, L_deQueue_9_p.
  intros S' H.
  unfold conjP in H.
  destruct H as [H1 H2].
  destruct H1 as [T [Q [ptr [bp [sp [head [tail [prev H]]]]]]]].
  destruct H as [P|P].
  destruct P as [P1 [P2 [P3 [P4 [P5 [P6 P7]]]]]].
  rewrite  H2 in P6.
  inversion P6.
  destruct P as [Q' [tcb [next [next' [tail' P]]]]].
  exists T; exists Q; exists Q'; exists ptr;exists tcb; exists bp.
  exists sp;exists head; exists tail; exists prev; exists next.
  exists next'; exists tail'.
  apply Hget;assumption.
  split.
  (* g' ==> g *)
  unfold L_deQueue_5_g, L_deQueue_9_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [P1 [P2 P3]].
  intros T Q ptr bp sp head tail prev mP H.
  destruct H as [P|P].
  destruct P as [P01 [P02 [P03 [P04 [P05 [P06 P07]]]]]].
  rewrite  P2 in P06.
  inversion P06.
  destruct P as [Q' [tcb [next [next' [tail' P]]]]].
  apply P3 with Q' tcb head tail prev next next' tail'.
  apply Hget;assumption.
 assert (Hemp: forall T Q ptr bp sp prev mP S,
    Ast_deQueue_5_emp T Q ptr bp sp prev mP S
    -> Ast_deQueue_6 T Q ptr bp sp prev mP S).
    unfold Ast_deQueue_5_emp, Ast_deQueue_6.
  intros.
  InvertAll.
  intuition.
  split.
  (* p ==> p'' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_deQueue_5_p, L_deQueue_6_p.
  intros S' H.
  unfold conjP in H.
  destruct H as [H1 H2].
  destruct H1 as [T [Q [ptr [bp [sp [head [tail [prev H]]]]]]]].
  destruct H as [P|P].
  exists T; exists Q; exists ptr; exists bp; exists sp;exists prev.
  apply Hemp;assumption.
  destruct P as [Q' [tcb [next [next' [tail' [P1 [P2 [P3 [P4 [P5 [P6 P7]]]]]]]]]]].
  rewrite  H2 in P6.
  inversion P6.
  (* g'' ==> g *)
  unfold L_deQueue_5_g, L_deQueue_6_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [P1 [P2 P3]].
  intros T Q ptr bp sp head tail prev mP H.
  destruct H as [P|P].
  apply P3 with prev.
  apply Hemp;assumption.
  destruct P as [Q' [tcb [next [next' [tail' [P01 [P02 [P03 [P04 [P05 [P06 P07]]]]]]]]]]].
  rewrite  P2 in P06.
  inversion P06.
Qed.

(* <DeQueue + 6> : movw $0, %ax *)
Lemma SCAP_WFcmd_deQueue_6_movwp :
  SCAP_WFcmd (SCAP_consCT (L_deQueue + 6 + 1) (Ast L_deQueue_7_p L_deQueue_7_g sctag) psi_deQueue_Block0)
    (L_deQueue + 6) (Ast L_deQueue_6_p L_deQueue_6_g sctag) 
    (comml (_movwp (Opr_imm 0) AX)).
Proof.
  assert (Hpp': forall T Q ptr bp sp prev mP S S',
    Ast_deQueue_6 T Q ptr bp sp prev mP S
    -> nextS (L_deQueue + 6) (comml (_movwp (Opr_imm 0) AX)) S S'
    -> Ast_deQueue_7 T Q ptr bp sp prev mP S').
    unfold Ast_deQueue_6, Ast_deQueue_7.
    intros T Q ptr bp sp prev mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [HSS1 [HSS2 [HSS3 HSS4]]].
    destruct HS as [HS1 [HS2 HS3]].
    subst M' R' fl' isr'.
    intuition.
  unfold SCAP_WFcmd.
  exists L_deQueue_7_p; exists L_deQueue_7_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [ptr [bp [sp [prev HS]]]]]].
  destruct HS as [HS1 [HS2 HS3]].
  exists (M, RF_update R AX 0, fl, isr).
  simplS in * .
  apply nextS_movwp with 0;trivial.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_deQueue_6_p, L_deQueue_7_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [ptr [bp [sp [prev H]]]]]].
  exists T; exists Q; exists ptr; exists bp; exists sp; exists prev.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_deQueue_6_g, L_deQueue_7_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S'' with prev.
  apply Hpp' with S;assumption.
Qed.

(* <DeQueue + 7> : popw %bp *)
Lemma SCAP_WFcmd_deQueue_7_popw :
  SCAP_WFcmd (SCAP_consCT (L_deQueue + 7 + 1) (Ast L_deQueue_8_p L_deQueue_8_g sctag) psi_deQueue_Block0)
    (L_deQueue + 7) (Ast L_deQueue_7_p L_deQueue_7_g sctag) 
    (comml (_popw BP)).
Proof.
  assert (Hpp': forall T Q ptr bp sp prev mP S S',
    Ast_deQueue_7 T Q ptr bp sp prev mP S
    -> nextS (L_deQueue + 7) (comml (_popw BP)) S S'
    -> Ast_deQueue_8 T Q ptr bp sp prev mP S').
    unfold Ast_deQueue_7, Ast_deQueue_8.
    intros T Q ptr bp sp prev mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [HSS1 [[x [HSS2 HSS3]] [HSS4 HSS5]]].
    destruct HS as [HS1 [HS2 [HS3 HS4]]].
    subst M' R' fl' isr'.
    assert (E : x = bp ).
      rewrite HS3 in HSS2.
      sep on.
      sep lift 3 in HS1.
      sep inj;auto.
    subst x.
    rewrite HS3.
    arith_rep (sp-2+2) with sp.
    intuition.
  sep on.
  sep matchH HS1 with 10.
  arith_rep sp with ((sp-2)+2).
  assert (h ||- ptoxW (sp-2)).
  sep trivial.
  sep trivial.
  unfold SCAP_WFcmd.
  exists L_deQueue_8_p; exists L_deQueue_8_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [ptr [bp [sp [prev HS]]]]]].
  destruct HS as [HS1 [HS2 [HS3 HS4]]].
  exists (M,  RF_update (RF_update R SP sp) BP bp, fl, isr).
  simplS in * .
  apply nextS_popw with (RF_update R SP sp) bp;trivial.
  unfold popW.
  rewrite HS3;intuition.
  sep on.
  sep lift 3 in HS1;sep trivial.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_deQueue_7_p, L_deQueue_8_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [ptr [bp [sp [prev H]]]]]].
  exists T; exists Q; exists ptr; exists bp; exists sp; exists prev.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_deQueue_7_g, L_deQueue_8_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S'' with prev.
  apply Hpp' with S;assumption.
Qed.

(* <DeQueue + 8> : retw *)
Lemma SCAP_WFcmd_deQueue_8_retw : 
  SCAP_WFcmd psi_deQueue_Block0 (L_deQueue+8) 
    (Ast L_deQueue_8_p L_deQueue_8_g sctag) (comms _retw).
Proof.
  unfold L_deQueue_8_g.
  simpl.
  intros S H.
  intros.
  unfold Ast_deQueue_8 in H0.
  InvertAll.
  left.
  sep on.
  split.
  sep lift 2 in H1.
  destruct (Queue_ptr0_emp_1 H1).
  rewrite H6.
  trivial.
  unfold Ast_deQueue_post_a.
  intuition.
  sep on.
  sep lift 0.
  apply ptoTQ_intro1 with 0 0 prev.
  sep matchH H1 with 10.
Qed.

(* <DeQueue + 9> : movw %bx, %cx *)
Lemma SCAP_WFcmd_deQueue_9_movwp :
  SCAP_WFcmd (SCAP_consCT (L_deQueue + 9 + 1) (Ast L_deQueue_10_p L_deQueue_10_g sctag) psi_deQueue_Block1)
    (L_deQueue + 9) (Ast L_deQueue_9_p L_deQueue_9_g sctag) 
    (comml (_movwp (Opr_reg BX) CX)).
Proof.
  assert (Hpp': forall T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S S',
    Ast_deQueue_9 T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S
    -> nextS (L_deQueue + 9) (comml (_movwp (Opr_reg BX) CX)) S S'
    -> Ast_deQueue_10 T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S').
    unfold Ast_deQueue_9, Ast_deQueue_10.
    intros T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [HSS1 [HSS2 [HSS3 HSS4]]].
    destruct HS as [HS1 [HS2 [HS3 [HS4 [HS5 [HS6 [HS7 [HS8 [HS9 [HS10 HS11]]]]]]]]]].
    subst M' R' fl' isr'.
    intuition.
  unfold SCAP_WFcmd.
  exists L_deQueue_10_p; exists L_deQueue_10_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [Q' [ptr [tcb [bp [sp [head [tail [prev [next [next' [tail' HS]]]]]]]]]]]]].
  destruct HS as [HS1 [HS2 [HS3 [HS4 [ HS5 [HS6 [HS7 [HS8 [HS9 [HS10 HS11]]]]]]]]]].
  exists (M, RF_update R CX ptr, fl, isr).
  simplS in * .
  apply nextS_movwp with ptr;trivial.
  unfold opr_eval.
  rewrite HS4;reflexivity.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_deQueue_9_p, L_deQueue_10_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [Q' [ptr [tcb [bp [sp [head [tail [prev [next [next' [tail' H]]]]]]]]]]]]].
  exists T; exists Q; exists Q'; exists ptr; exists tcb; exists bp.
  exists sp; exists head; exists tail; exists prev; exists next.
  exists next'; exists tail'.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_deQueue_9_g, L_deQueue_10_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S'' with Q' tcb head tail prev next next' tail'.
  apply Hpp' with S;assumption.
Qed.

(* <DeQueue + 10> : movw %dx, %bx *)
Lemma SCAP_WFcmd_deQueue_10_movwp :
  SCAP_WFcmd (SCAP_consCT (L_deQueue + 10 + 1) (Ast L_deQueue_11_p L_deQueue_11_g sctag) psi_deQueue_Block1)
    (L_deQueue + 10) (Ast L_deQueue_10_p L_deQueue_10_g sctag) 
    (comml (_movwp (Opr_reg DX) BX)).
Proof.
  assert (Hpp': forall T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S S',
    Ast_deQueue_10 T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S
    -> nextS (L_deQueue + 10) (comml (_movwp (Opr_reg DX) BX)) S S'
    -> Ast_deQueue_11 T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S').
    unfold Ast_deQueue_10, Ast_deQueue_11.
    intros T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [HSS1 [HSS2 [HSS3 HSS4]]].
    destruct HS as [HS1 [HS2 [HS3 [HS4 [HS5 [HS6 [HS7 [HS8 [HS9 [HS10 HS11]]]]]]]]]].
    subst M' R' fl' isr'.
    intuition.
  unfold SCAP_WFcmd.
  exists L_deQueue_11_p; exists L_deQueue_11_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [Q' [ptr [tcb [bp [sp [head [tail [prev [next [next' [tail' HS]]]]]]]]]]]]].
  destruct HS as [HS1 [HS2 [HS3 [HS4 [ HS5 [HS6 [HS7 [HS8 [HS9 [HS10 HS11]]]]]]]]]].
  exists (M, RF_update R BX head, fl, isr).
  simplS in * .
  apply nextS_movwp with head;trivial.
  unfold opr_eval.
  rewrite HS5;reflexivity.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_deQueue_10_p, L_deQueue_11_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [Q' [ptr [tcb [bp [sp [head [tail [prev [next [next' [tail' H]]]]]]]]]]]]].
  exists T; exists Q; exists Q'; exists ptr; exists tcb; exists bp.
  exists sp; exists head; exists tail; exists prev; exists next.
  exists next'; exists tail'.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_deQueue_10_g, L_deQueue_11_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S'' with Q' tcb head tail prev next next' tail'.
  apply Hpp' with S;assumption.
Qed.

(* <DeQueue + 11> : movw ND_NEXT(%bx), %dx *)
Lemma SCAP_WFcmd_deQueue_11_movwld :
  SCAP_WFcmd (SCAP_consCT (L_deQueue + 11 + 1) (Ast L_deQueue_12_p L_deQueue_12_g sctag) psi_deQueue_Block1)
    (L_deQueue + 11) (Ast L_deQueue_11_p L_deQueue_11_g sctag) 
    (comml (_movwld (A_reg ND_NEXT BX) DX)).
Proof.
  assert (Hpp': forall T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S S',
    Ast_deQueue_11 T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S
    -> nextS (L_deQueue + 11) (comml (_movwld (A_reg ND_NEXT BX) DX)) S S'
    -> Ast_deQueue_12 T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S').
    unfold Ast_deQueue_11, Ast_deQueue_12.
    intros T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [HSS1 [[x [HSS2 HSS3]] [HSS4 HSS5]]].
    destruct HS as [HS1 [HS2 [HS3 [HS4 [HS5 [HS6 [HS7 [HS8 [HS9 [HS10 HS11]]]]]]]]]].
    subst M' R' fl' isr'.
    unfold ptoNode in HS1.
    assert (E : x = next).
      sep on.
      sep lift 3 in HS1.
      rewrite HS2 in HSS2.
      sep inj;auto.
    subst x.
    split.
    sep on.
    sep matchH HS1 with 10.
    intuition.
  unfold SCAP_WFcmd.
  exists L_deQueue_12_p; exists L_deQueue_12_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [Q' [ptr [tcb [bp [sp [head [tail [prev [next [next' [tail' HS]]]]]]]]]]]]].
  destruct HS as [HS1 [HS2 [HS3 [HS4 [ HS5 [HS6 [HS7 [HS8 [HS9 [HS10 HS11]]]]]]]]]].
  exists (M, RF_update R DX next, fl, isr).
  simplS in * .
  apply nextS_movwld with (head+2) next;trivial.
  unfold addr_eval.
  rewrite HS2;reflexivity.
  sep on.
  unfold ptoNode in HS1.
  sep lift 3 in HS1.
  sep trivial.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_deQueue_11_p, L_deQueue_12_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [Q' [ptr [tcb [bp [sp [head [tail [prev [next [next' [tail' H]]]]]]]]]]]]].
  exists T; exists Q; exists Q'; exists ptr; exists tcb; exists bp.
  exists sp; exists head; exists tail; exists prev; exists next.
  exists next'; exists tail'.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_deQueue_11_g, L_deQueue_12_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S'' with Q' tcb head tail prev next next' tail'.
  apply Hpp' with S;assumption.
Qed.

(* <DeQueue + 12> : movw %bx, %ax *)
Lemma SCAP_WFcmd_deQueue_12_movwp :
  SCAP_WFcmd (SCAP_consCT (L_deQueue + 12 + 1) (Ast L_deQueue_13_p L_deQueue_13_g sctag) psi_deQueue_Block1)
    (L_deQueue + 12) (Ast L_deQueue_12_p L_deQueue_12_g sctag) 
    (comml (_movwp (Opr_reg BX) AX)).
Proof.
  assert (Hpp': forall T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S S',
    Ast_deQueue_12 T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S
    -> nextS (L_deQueue + 12) (comml (_movwp (Opr_reg BX) AX)) S S'
    -> Ast_deQueue_13 T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S').
    unfold Ast_deQueue_12, Ast_deQueue_13.
    intros T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [HSS1 [HSS2 [HSS3 HSS4]]].
    destruct HS as [HS1 [HS2 [HS3 [HS4 [HS5 [HS6 [HS7 [HS8 [HS9 [HS10 HS11]]]]]]]]]].
    subst M' R' fl' isr'.
    intuition.
  unfold SCAP_WFcmd.
  exists L_deQueue_13_p; exists L_deQueue_13_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [Q' [ptr [tcb [bp [sp [head [tail [prev [next [next' [tail' HS]]]]]]]]]]]]].
  destruct HS as [HS1 [HS2 [HS3 [HS4 [ HS5 [HS6 [HS7 [HS8 [HS9 [HS10 HS11]]]]]]]]]].
  exists (M, RF_update R AX head, fl, isr).
  simplS in * .
  apply nextS_movwp with head;trivial.
  unfold opr_eval.
  rewrite HS3;reflexivity.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_deQueue_12_p, L_deQueue_13_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [Q' [ptr [tcb [bp [sp [head [tail [prev [next [next' [tail' H]]]]]]]]]]]]].
  exists T; exists Q; exists Q'; exists ptr; exists tcb; exists bp.
  exists sp; exists head; exists tail; exists prev; exists next.
  exists next'; exists tail'.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_deQueue_12_g, L_deQueue_13_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S'' with Q' tcb head tail prev next next' tail'.
  apply Hpp' with S;assumption.
Qed.

(* <DeQueue + 13> : movw %cx, %bx *)
Lemma SCAP_WFcmd_deQueue_13_movwp :
  SCAP_WFcmd (SCAP_consCT (L_deQueue + 13 + 1) (Ast L_deQueue_14_p L_deQueue_14_g sctag) psi_deQueue_Block1)
    (L_deQueue + 13) (Ast L_deQueue_13_p L_deQueue_13_g sctag) 
    (comml (_movwp (Opr_reg CX) BX)).
Proof.
  assert (Hpp': forall T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S S',
    Ast_deQueue_13 T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S
    -> nextS (L_deQueue + 13) (comml (_movwp (Opr_reg CX) BX)) S S'
    -> Ast_deQueue_14 T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S').
    unfold Ast_deQueue_13, Ast_deQueue_14.
    intros T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [HSS1 [HSS2 [HSS3 HSS4]]].
    destruct HS as [HS1 [HS2 [HS3 [HS4 [HS5 [HS6 [HS7 [HS8 [HS9 [HS10 [HS11 HS12]]]]]]]]]]].
    subst M' R' fl' isr'.
    intuition.
  unfold SCAP_WFcmd.
  exists L_deQueue_14_p; exists L_deQueue_14_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [Q' [ptr [tcb [bp [sp [head [tail [prev [next [next' [tail' HS]]]]]]]]]]]]].
  destruct HS as [HS1 [HS2 [HS3 [HS4 [ HS5 [HS6 [HS7 [HS8 [HS9 [HS10 [HS11 HS12]]]]]]]]]]].
  exists (M, RF_update R BX ptr, fl, isr).
  simplS in * .
  apply nextS_movwp with ptr;trivial.
  unfold opr_eval.
  rewrite HS4;reflexivity.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_deQueue_13_p, L_deQueue_14_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [Q' [ptr [tcb [bp [sp [head [tail [prev [next [next' [tail' H]]]]]]]]]]]]].
  exists T; exists Q; exists Q'; exists ptr; exists tcb; exists bp.
  exists sp; exists head; exists tail; exists prev; exists next.
  exists next'; exists tail'.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_deQueue_13_g, L_deQueue_14_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S'' with Q' tcb head tail prev next next' tail'.
  apply Hpp' with S;assumption.
Qed.

(* <DeQueue + 14> : cmpw $0, %dx *)
Lemma SCAP_WFcmd_deQueue_14_cmpw :
  SCAP_WFcmd (SCAP_consCT (L_deQueue + 14 + 1) (Ast L_deQueue_15_p L_deQueue_15_g sctag) psi_deQueue_Block1)
    (L_deQueue + 14) (Ast L_deQueue_14_p L_deQueue_14_g sctag) 
    (comml (_cmpw (Opr_imm 0) DX)).
Proof.
  assert (Hpp': forall T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S S',
    Ast_deQueue_14 T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S
    -> nextS (L_deQueue + 14)  (comml (_cmpw (Opr_imm 0) DX)) S S'
    -> (Ast_deQueue_15_emp T Q Q' ptr tcb bp sp head tail prev mP S' \/
    exists next'', 
      Ast_deQueue_15_get T Q Q' ptr tcb bp sp head tail prev next'' mP S')).
    unfold Ast_deQueue_14, Ast_deQueue_15_emp, Ast_deQueue_15_get.
    intros T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [HSS1 [HSS2 [HSS3 HSS4]]].
    destruct HS as [HS1 [HS2 [HS3 [HS4 [HS5 [HS6 [HS7 [HS8 [HS9 [HS10 [HS11 HS12]]]]]]]]]]].
    destruct HSS3 as [HSS3|HSS3];destruct HSS3 as [HSS31 HSS32].
    subst M' R' fl' isr'.
    left.
    rewrite <- HSS31 in HS4.
    subst next.
    destruct HS12 as [[P1 [P2 P3]] |[P1 [P2 [P3 P4]]]].
    subst tail' next'.
    sep on.
    sep lift 5 in HS1.
    assert (P := Queue_head0_tail0_1 HS1).
    rewrite P in P2;inversion P2.
    subst tail' tail next'.
    intuition.
    right.
    destruct HS12 as [[P1 [P2 P3]] |[P1 [P2 [P3 P4]]]].
    exists next.
    subst M' R' fl' isr' next' tail'.
    intuition.
    rewrite HS4 in HSS31;rewrite P4 in HSS31; elim HSS31;reflexivity.
  unfold SCAP_WFcmd.
  exists L_deQueue_15_p; exists L_deQueue_15_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [Q' [ptr [tcb [bp [sp [head [tail [prev [next [next' [tail' HS]]]]]]]]]]]]].
  destruct HS as [HS1 [HS2 [HS3 [HS4 [HS5 [HS6 [HS7 [HS8 [HS9 [HS10 [HS11 HS12]]]]]]]]]]].
  destruct (eq_nat_dec 0 next).
    exists (M, R,Set_Flag fl _ZF, isr).
    simplS in * .
    apply nextS_cmpw;trivial.
    left.
    split.
    simpl.
    rewrite <- e in HS4;auto.
    auto.
    exists (M, R,Clear_Flag fl _ZF, isr).
    simplS in * .
    apply nextS_cmpw;trivial.
    right.
    intuition.
    simpl in H.
    apply n;rewrite <- H1 in HS4;auto.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_deQueue_14_p, L_deQueue_15_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [Q' [ptr [tcb [bp [sp [head [tail [prev [next [next' [tail' H]]]]]]]]]]]]].
  exists T; exists Q; exists Q'; exists ptr; exists tcb; exists bp; exists sp.
  exists head; exists tail; exists prev.
  apply Hpp' with next next' tail' S; assumption.
  (* g' ==> g *)
  unfold L_deQueue_14_g, L_deQueue_15_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S'' with Q' tcb head tail prev.
  apply Hpp' with next next' tail' S;assumption.
Qed.

(* <DeQueue + 15> : jne deq_b *)
Lemma SCAP_WFcmd_deQueue_15_jne :
  SCAP_WFcmd (SCAP_consCT (L_deQueue + 15 + 1) (Ast L_deQueue_16_p L_deQueue_16_g sctag) psi_deQueue_Block1)
    (L_deQueue + 15) (Ast L_deQueue_15_p L_deQueue_15_g sctag) 
    (comml (_jne L_deQ_b)).
Proof.
  unfold SCAP_WFcmd.
  exists L_deQueue_17_p; exists L_deQueue_16_p.
  exists L_deQueue_17_g; exists L_deQueue_16_g.
  split.
  apply psi_cons_b_lookup_a.
  unfold psi_deQueue_Block1.
  apply psi_cons_a_lookup_a.
  split_l.
  assert (Hget: forall T Q Q' ptr tcb bp sp head tail prev next mP S,
    Ast_deQueue_15_get T Q Q' ptr tcb bp sp head tail prev next mP S
    -> Ast_deQueue_17 T Q Q' ptr tcb bp sp head prev next tail mP S).
    unfold Ast_deQueue_15_get, Ast_deQueue_17.
  intros.
  InvertAll.
  intuition.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_deQueue_15_p, L_deQueue_17_p.
  intros S' H.
  unfold conjP in H.
  destruct H as [H1 H2].
  destruct H1 as [T [Q [Q' [ptr [tcb [bp [sp [head [tail [prev H]]]]]]]]]].
  destruct H as [P|P].
  destruct P as [P1 [P2 [P3 [P4 [P5 [P6 [P7 P8]]]]]]].
  rewrite  H2 in P7.
  inversion P7.
  destruct P as [next P].
  exists T; exists Q; exists Q'; exists ptr;exists tcb; exists bp.
  exists sp;exists head; exists prev; exists next; exists tail.
  apply Hget;assumption.
  split.
  (* g' ==> g *)
  unfold L_deQueue_15_g, L_deQueue_17_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [P1 [P2 P3]].
  intros T Q Q' ptr tcb bp sp head tail prev  mP H.
  destruct H as [P|P].
  destruct P as [P01 [P02 [P03 [P04 [P05 [P06 [P07 P08]]]]]]].
  rewrite  P2 in P07.
  inversion P07.
  destruct P as [next P].
  apply P3 with Q' tcb head prev next tail.
  apply Hget;assumption.
 assert (Hemp: forall T Q Q' ptr tcb bp sp head tail prev mP S,
    Ast_deQueue_15_emp T Q Q' ptr tcb bp sp head tail prev mP S
    -> Ast_deQueue_16 T Q Q' ptr tcb bp sp head tail prev mP S).
    unfold Ast_deQueue_15_emp, Ast_deQueue_16.
  intros.
  InvertAll.
  intuition.
  split.
  (* p ==> p'' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_deQueue_15_p, L_deQueue_16_p.
  intros S' H.
  unfold conjP in H.
  destruct H as [H1 H2].
  destruct H1 as [T [Q [Q' [ptr [tcb [bp [sp [head [tail [prev H]]]]]]]]]].
  destruct H as [P|P].
  exists T; exists Q; exists Q'; exists ptr; exists tcb; exists bp; exists sp.
  exists head; exists tail; exists prev.
  apply Hemp;assumption.
  destruct P as [next [P1 [P2 [P3 [P4 [P5 [P6 [P7 P8]]]]]]]].
  rewrite  H2 in P7.
  inversion P7.
  (* g'' ==> g *)
  unfold L_deQueue_15_g, L_deQueue_16_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [P1 [P2 P3]].
  intros T Q Q' ptr tcb bp sp head tail prev mP H.
  destruct H as [P|P].
  apply P3 with Q' tcb head tail prev.
  apply Hemp;assumption.
  destruct P as [next [P01 [P02 [P03 [P04 [P05 [P06 [P07 P08]]]]]]]].
  rewrite  P2 in P07.
  inversion P07.
Qed.

(* <DeQueue + 16> : movw %dx, Q_TAIL(%bx) *)
Lemma SCAP_WFcmd_deQueue_16_movwst :
  SCAP_WFcmd (SCAP_consCT (L_deQueue + 16 + 1) (Ast L_deQueue_17_p L_deQueue_17_g sctag) psi_deQueue_Block1)
    (L_deQueue + 16) (Ast L_deQueue_16_p L_deQueue_16_g sctag) 
    (comml (_movwst (Opr_reg DX) (A_reg Q_TAIL BX))).
Proof.
  assert (Hpp': forall T Q Q' ptr tcb bp sp head tail prev mP S S',
    Ast_deQueue_16 T Q Q' ptr tcb bp sp head tail prev  mP S
    -> nextS (L_deQueue + 16) (comml (_movwst (Opr_reg DX)(A_reg Q_TAIL BX))) S S'
    -> Ast_deQueue_17 T Q Q' ptr tcb bp sp head prev 0 0 mP S').
    unfold Ast_deQueue_16, Ast_deQueue_17.
    intros T Q Q' ptr tcb bp sp head tail prev mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [[HSS1 [jk HSS2]] [HSS3 [HSS4 HSS5]]].
    destruct HS as [HS1 [HS2 [HS3 [HS4 [HS5 [HS6 [HS7 [HS8 [HS9 HS10]]]]]]]]].
    subst M' R' fl' isr'.
    intuition.
    sep on.
    assert (MemUpdW M (ptr + 2) 0 jk
   ||- ptr |->w head *^* ptr + 2 |->w 0 *^* head |->w prev *^*
       head + 2 |->w 0 *^* ptoTCB (head + 4) tcb *^*
       Queue Thread ptoTCB 0 0 head 0 (filterTQ T Q') *^* 
       sp - 2 |->w bp *^* sp + 2 |->w ptr *^* mP).
    sep lift 1.
    sep upd.
    sep matchH HS1 with 10.
    unfold MemUpdW in H.
    rewrite HS2; rewrite HS4.
    assumption.
  unfold SCAP_WFcmd.
  exists L_deQueue_17_p; exists L_deQueue_17_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [Q' [ptr [tcb [bp [sp [head [tail [prev  HS]]]]]]]]]].
  destruct HS as [HS1 [HS2 [HS3 [HS4 [ HS5 [HS6 [HS7 [HS8 [HS9 HS10]]]]]]]]].
  exists (MemUpdW M (ptr+2) 0 0, R, fl, isr).
  simplS in * .  
  apply nextS_movwst with (ptr+2) 0;trivial.
  unfold addr_eval.
  rewrite HS2;reflexivity.
  simpl;auto.
  sep on.
  sep lift 1 in HS1.
  sep trivial.
  apply MemUpdW_MemUpd16 with 0; trivial.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_deQueue_16_p, L_deQueue_17_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [Q' [ptr [tcb [bp [sp [head [tail [prev H]]]]]]]]]].
  exists T; exists Q; exists Q'; exists ptr; exists tcb; exists bp.
  exists sp; exists head; exists prev; exists 0;exists 0.
  apply Hpp' with tail S; assumption.
  (* g' ==> g *)
  unfold L_deQueue_16_g, L_deQueue_17_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S'' with Q' tcb head prev 0 0.
  apply Hpp' with tail S;assumption.
Qed.

(* <DeQueue + 17> : movw %dx, Q_HEAD(%bx) *)
Lemma SCAP_WFcmd_deQueue_17_movwst :
  SCAP_WFcmd (SCAP_consCT (L_deQueue + 17 + 1) (Ast L_deQueue_18_p L_deQueue_18_g sctag) psi_deQueue_Block1)
    (L_deQueue + 17) (Ast L_deQueue_17_p L_deQueue_17_g sctag) 
    (comml (_movwst (Opr_reg DX) (A_reg Q_HEAD BX))).
Proof.
  assert (Hpp': forall T Q Q' ptr tcb bp sp head prev head' tail' mP S S',
    Ast_deQueue_17 T Q Q' ptr tcb bp sp head prev head' tail'  mP S
    -> nextS (L_deQueue + 17) (comml (_movwst (Opr_reg DX)(A_reg Q_HEAD BX))) S S'
    -> Ast_deQueue_18 T Q Q' ptr tcb bp sp head prev head' tail' mP S').
    unfold Ast_deQueue_17, Ast_deQueue_18.
    intros T Q Q' ptr tcb bp sp head prev head' tail' mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [[HSS1 [jk HSS2]] [HSS3 [HSS4 HSS5]]].
    destruct HS as [HS1 [HS2 [HS3 [HS4 [HS5 [HS6 [HS7 [HS8 [HS9 HS10]]]]]]]]].
    subst M' R' fl' isr'.
    intuition.
    sep on.
    assert (MemUpdW M ptr head' jk
   ||- ptr |->w head' *^* ptr + 2 |->w tail' *^* head |->w prev *^*
       head + 2 |->w head' *^* ptoTCB (head + 4) tcb *^*
       Queue Thread ptoTCB head' tail' head 0 (filterTQ T Q') *^* 
       sp - 2 |->w bp *^* sp + 2 |->w ptr *^* mP).
    sep lift 0.
    sep upd.
    sep matchH HS1 with 10.
    unfold MemUpdW in H.
    rewrite HS2; rewrite HS4.
    arith_rep (ptr+0) with ptr.
  unfold SCAP_WFcmd.
  exists L_deQueue_18_p; exists L_deQueue_18_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [Q' [ptr [tcb [bp [sp [head [prev [head' [tail'  HS]]]]]]]]]]].
  destruct HS as [HS1 [HS2 [HS3 [HS4 [ HS5 [HS6 [HS7 [HS8 [HS9 HS10]]]]]]]]].
  exists (MemUpdW M ptr head'  0, R, fl, isr).
  simplS in * .  
  apply nextS_movwst with ptr head';trivial.
  unfold addr_eval.
  rewrite HS2;auto.
  simpl;auto.
  sep on.
  sep lift 0 in HS1.
  sep trivial.
  apply MemUpdW_MemUpd16 with 0; trivial.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_deQueue_17_p, L_deQueue_18_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [Q' [ptr [tcb [bp [sp [head [prev [head' [tail' H]]]]]]]]]]].
  exists T; exists Q; exists Q'; exists ptr; exists tcb; exists bp.
  exists sp; exists head; exists prev; exists head';exists tail'.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_deQueue_17_g, L_deQueue_18_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S'' with Q' tcb head prev head' tail'.
  apply Hpp' with S;assumption.
Qed.

(* <DeQueue + 18> : popw %bp *)
Lemma SCAP_WFcmd_deQueue_18_popw :
  SCAP_WFcmd (SCAP_consCT (L_deQueue + 18 + 1) (Ast L_deQueue_19_p L_deQueue_19_g sctag) psi_deQueue_Block1)
    (L_deQueue + 18) (Ast L_deQueue_18_p L_deQueue_18_g sctag) 
    (comml (_popw BP)).
Proof.
  assert (Hpp': forall T Q Q' ptr tcb bp sp head prev head' tail' mP S S',
    Ast_deQueue_18 T Q Q' ptr tcb bp sp head prev head' tail' mP S
    -> nextS (L_deQueue + 18) (comml (_popw BP)) S S'
    -> Ast_deQueue_19 T Q Q' ptr tcb bp sp head prev head' tail' mP S').
    unfold Ast_deQueue_18, Ast_deQueue_19.
    intros T Q Q' ptr tcb bp sp head prev head' tail' mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [HSS1 [[x [HSS2 HSS3]] [HSS4 HSS5]]].
    destruct HS as [HS1 [HS2 [HS3 [HS4 [HS5 [HS6 [HS7 [HS8 HS9]]]]]]]].
    subst M' R' fl' isr'.
    assert (E : x = bp ).
      rewrite HS4 in HSS2.
      sep on.
      sep lift 6 in HS1.
      sep inj;auto.
    subst x.
    rewrite HS4.
    arith_rep (sp-2+2) with sp.
    intuition.
  sep on.
  sep matchH HS1 with 10.
  arith_rep sp with ((sp-2)+2).
  assert (h ||- ptoxW (sp-2)).
  sep trivial.
  sep trivial.
  unfold SCAP_WFcmd.
  exists L_deQueue_19_p; exists L_deQueue_19_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [Q' [ptr [tcb [bp [sp [head [prev [head' [tail' HS]]]]]]]]]]].
  destruct HS as [HS1 [HS2 [HS3 [HS4 [HS5 [HS6 [HS7 [HS8 HS9]]]]]]]].
  exists (M,  RF_update (RF_update R SP sp) BP bp, fl, isr).
  simplS in * .
  apply nextS_popw with (RF_update R SP sp) bp;trivial.
  unfold popW.
  rewrite HS4;intuition.
  sep on.
  sep lift 6 in HS1;sep trivial.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_deQueue_18_p, L_deQueue_19_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as  [T [Q [Q' [ptr [tcb [bp [sp [head [prev [head' [tail' HS]]]]]]]]]]].
  exists T; exists Q; exists Q'; exists ptr; exists tcb; exists bp.
  exists sp; exists head; exists prev; exists head'; exists tail'.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_deQueue_18_g, L_deQueue_19_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S'' with Q' tcb head prev head' tail'.
  apply Hpp' with S;assumption.
Qed.

(* <DeQueue + 19> : retw *)
Lemma SCAP_WFcmd_deQueue_19_retw : 
  SCAP_WFcmd psi_deQueue_Block1 (L_deQueue+19) 
    (Ast L_deQueue_19_p L_deQueue_19_g sctag) (comms _retw).
Proof.
  unfold L_deQueue_19_g.
  simpl.
  intros S H.
  intros.
  unfold Ast_deQueue_19 in H0.
  InvertAll.
  right.
  exists Q'; exists head; exists tcb.
  repeat split_l.
  unfold embed_TCB.
  sep on.
  sep lift 3.
  apply ptoTQ_intro1 with head' tail' head.
  sep matchH H1 with 10.
  assumption.
  trivial.
Qed.

Lemma SCAP_WFiseq_deQueue_Block0 :
  SCAP_WFiseq psi_deQueue_Block0 L_deQueue spec_deQueue I_deQueue.
Proof.
unfold I_deQueue, spec_deQueue.
apply scap_isq_seq with L_deQueue_1_p L_deQueue_1_g.
apply SCAP_WFcmd_deQueue_0_pushw.
apply scap_isq_seq with L_deQueue_2_p L_deQueue_2_g.
apply SCAP_WFcmd_deQueue_1_movwp.
assert (H : L_deQueue+1+1 = L_deQueue+2); [omega | rewrite H; clear H].
apply scap_isq_seq with L_deQueue_3_p L_deQueue_3_g.
apply SCAP_WFcmd_deQueue_2_movwld.
assert (H : L_deQueue+2+1 = L_deQueue+3); [omega | rewrite H; clear H].
apply scap_isq_seq with L_deQueue_4_p L_deQueue_4_g.
apply SCAP_WFcmd_deQueue_3_movwld.
assert (H : L_deQueue+3+1 = L_deQueue+4); [omega | rewrite H; clear H].
apply scap_isq_seq with L_deQueue_5_p L_deQueue_5_g.
apply SCAP_WFcmd_deQueue_4_cmpw.
assert (H : L_deQueue+4+1 = L_deQueue+5); [omega | rewrite H; clear H].
apply scap_isq_seq with L_deQueue_6_p L_deQueue_6_g.
apply SCAP_WFcmd_deQueue_5_jne.
assert (H : L_deQueue+5+1 = L_deQueue+6); [omega | rewrite H; clear H].
apply scap_isq_seq with L_deQueue_7_p L_deQueue_7_g.
apply SCAP_WFcmd_deQueue_6_movwp.
assert (H : L_deQueue+6+1 = L_deQueue+7); [omega | rewrite H; clear H].
apply scap_isq_seq with L_deQueue_8_p L_deQueue_8_g.
apply SCAP_WFcmd_deQueue_7_popw.
assert (H : L_deQueue+7+1 = L_deQueue+8); [omega | rewrite H; clear H].
apply scap_isq_end.
apply SCAP_WFcmd_deQueue_8_retw.
Qed.

Lemma SCAP_WFiseq_deQueue_Block1 :
  SCAP_WFiseq psi_deQueue_Block1 L_deQ_a (Ast L_deQueue_9_p L_deQueue_9_g sctag) I_deQ_a.
Proof.
unfold I_deQ_a, L_deQ_a.
apply scap_isq_seq with L_deQueue_10_p L_deQueue_10_g.
apply SCAP_WFcmd_deQueue_9_movwp.
apply scap_isq_seq with L_deQueue_11_p L_deQueue_11_g.
arith_rep (L_deQueue + 9 + 1) with (L_deQueue + 10).
apply SCAP_WFcmd_deQueue_10_movwp.
apply scap_isq_seq with L_deQueue_12_p L_deQueue_12_g.
arith_rep (L_deQueue + 9 + 1 + 1) with (L_deQueue + 11).
apply SCAP_WFcmd_deQueue_11_movwld.
apply scap_isq_seq with L_deQueue_13_p L_deQueue_13_g.
arith_rep (L_deQueue + 9 + 1 + 1 + 1) with (L_deQueue + 12).
apply SCAP_WFcmd_deQueue_12_movwp.
apply scap_isq_seq with L_deQueue_14_p L_deQueue_14_g.
arith_rep (L_deQueue + 9 + 1 + 1+ 1 + 1) with (L_deQueue + 13).
apply SCAP_WFcmd_deQueue_13_movwp.
apply scap_isq_seq with L_deQueue_15_p L_deQueue_15_g.
arith_rep (L_deQueue + 9 + 1 + 1+ 1 + 1 +1) with (L_deQueue + 14).
apply SCAP_WFcmd_deQueue_14_cmpw.
apply scap_isq_seq with L_deQueue_16_p L_deQueue_16_g.
arith_rep (L_deQueue + 9 + 1 + 1+ 1 + 1 +1+1) with (L_deQueue + 15).
apply SCAP_WFcmd_deQueue_15_jne.
apply scap_isq_seq with L_deQueue_17_p L_deQueue_17_g.
arith_rep (L_deQueue + 9 + 1 + 1+ 1 + 1 +1+1+1) with (L_deQueue + 16).
apply SCAP_WFcmd_deQueue_16_movwst.
apply scap_isq_seq with L_deQueue_18_p L_deQueue_18_g.
arith_rep (L_deQueue + 9 + 1 + 1+ 1 + 1 +1+1+1+1) with (L_deQueue + 17).
apply SCAP_WFcmd_deQueue_17_movwst.
apply scap_isq_seq with L_deQueue_19_p L_deQueue_19_g.
arith_rep (L_deQueue + 9 + 1 + 1+ 1 + 1 +1+1+1+1+1) with (L_deQueue + 18).
apply SCAP_WFcmd_deQueue_18_popw.
arith_rep (L_deQueue + 9 + 1 + 1+ 1 + 1 +1+1+1+1+1+1) with (L_deQueue + 19).
apply scap_isq_end.
apply SCAP_WFcmd_deQueue_19_retw.
Qed.
