(***************************************************************)
(* Coq code for AIM                                            *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(* by wangwei *)

Add LoadPath "..".
Add LoadPath "../CAP0".
Add LoadPath "../SCAP".

Require Import Coqlib.

Require Import Arith.
Require Import Machine.
Require Import Assertion.
Require Import Operational_Semantics.

Require Import Kernel.

Require Import AIM.
Require Import SCAP.
Require Import QLib.
Require Import Linking.

Definition Ast_deQueue_pre (T : TQ) (Q : WQ) (ptr : Label)
  (bp sp : nat) (mP : MPred) (S : State): Prop :=
    (ptoTQ ptr (filterTQ T Q)
      *^* stkSpacePrec sp 2 
      *^* sp + 2|->w ptr
      *^* mP) (S_Mem S)
    /\ S_RF S BP = bp
    /\ S_RF S SP = sp.

Definition p_deQueue (S : State) : Prop :=
  exists T, exists Q, exists ptr,
    exists bp, exists sp, 
      Ast_deQueue_pre T Q ptr bp sp mTrue S.

Definition Ast_deQueue_post_a (T : TQ) (Q : WQ) (ptr : Label)
  (bp sp : nat) (mP : MPred) (S' : State): Prop :=
  (ptoTQ ptr (filterTQ T Q)
    *^* stkSpacePrec sp 2 
    *^* sp + 2|->w ptr
    *^* mP) (S_Mem S')
  /\ S_RF S' AX = 0
  /\ S_RF S' BP = bp
  /\ S_RF S' SP = sp.

(* exit of Block3 , get a node *)
Definition Ast_deQueue_post_b (T : TQ) (Q' : WQ) (ptr : Label)
  (tid' : TID) (tcb' : Thread)
  (bp sp : nat) (mP : MPred) (S' : State): Prop :=
  (embed_TCB tid' tcb' 
    *^* ptoTQ ptr (filterTQ T Q')
    *^* stkSpacePrec sp 2 
    *^* sp + 2 |->w ptr
    *^* mP) (S_Mem S')
  /\ S_RF S' AX = tid'
  /\ S_RF S' BP = bp
  /\ S_RF S' SP = sp
  /\ TQMod.maps_to T tid' tcb'.

Definition g_deQueue (S S' : State) : Prop := forall mP,
  forall T Q ptr bp sp,
    Ast_deQueue_pre T Q ptr bp sp mP S
    -> (forall l, (filterTQ T Q) l = None)
      /\ Ast_deQueue_post_a T Q ptr bp sp mP S'
    \/ exists Q', exists tid', exists tcb',
      Ast_deQueue_post_b T Q' ptr tid' tcb' bp sp mP S'
      /\ Q = addThrd Q' tid'
      /\ WQMod.not_in_dom tid' Q'.

(* <deQueue> : Block0 *)
(* <deQueue> : pushw %bp *)
Definition Ast_deQueue_1 (T : TQ) (Q : WQ) (ptr : Label)
  (bp sp : nat) (mP : MPred) (S : State) :Prop :=
  (ptoTQ ptr (filterTQ T Q)
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w ptr
      *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 2
  /\ sp >= 2.

Definition L_deQueue_1_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists bp, exists sp,
    Ast_deQueue_1 T Q ptr bp sp mTrue S.

Definition L_deQueue_1_g (S S' : State) : Prop :=
  forall T Q ptr bp sp mP,
    Ast_deQueue_1 T Q ptr bp sp mP S
    -> (forall l, (filterTQ T Q) l = None)
    /\ Ast_deQueue_post_a T Q ptr bp sp mP S'
    \/ exists Q', exists tid', exists tcb',
      Ast_deQueue_post_b T Q' ptr tid' tcb' bp sp mP S'
      /\ Q = addThrd Q' tid'
      /\ WQMod.not_in_dom tid' Q'.

(* <deQueue> : movw %sp, %bp *)
Definition Ast_deQueue_2 (T : TQ) (Q : WQ) (ptr : Label)
  (bp sp : nat) (mP : MPred) (S : State) :Prop :=
  (ptoTQ ptr (filterTQ T Q)
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w ptr
      *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 2
  /\ S_RF S BP = sp - 2
  /\ sp >= 2.

Definition L_deQueue_2_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists bp, exists sp,
    Ast_deQueue_2 T Q ptr bp sp mTrue S.

Definition L_deQueue_2_g (S S' : State) : Prop :=
  forall T Q ptr bp sp mP,
    Ast_deQueue_2 T Q ptr bp sp mP S
    -> (forall l, (filterTQ T Q) l = None)
    /\ Ast_deQueue_post_a T Q ptr bp sp mP S'
    \/ exists Q', exists tid', exists tcb',
      Ast_deQueue_post_b T Q' ptr tid' tcb' bp sp mP S'
      /\ Q = addThrd Q' tid'
      /\ WQMod.not_in_dom tid' Q'.

(* <deQueue> : movw 4(%bp), %bx *)
Definition Ast_deQueue_3 (T : TQ) (Q : WQ) (ptr : Label)
  (bp sp head tail prev: nat) (mP : MPred) (S : State) :Prop :=
  (ptr |->w head
    *^* ptr + 2 |->w tail
    *^* (Queue Thread ptoTCB head tail prev 0 (filterTQ T Q))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w ptr
      *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 2
  /\ S_RF S BP = sp - 2
  /\ S_RF S BX = ptr
  /\ sp >= 2.

Definition L_deQueue_3_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists bp, exists sp,
    exists head, exists tail, exists prev,
      Ast_deQueue_3 T Q ptr bp sp head tail prev mTrue S.

Definition L_deQueue_3_g (S S' : State) : Prop :=
  forall T Q ptr bp sp head tail prev mP,
    Ast_deQueue_3 T Q ptr bp sp head tail prev mP S
    -> (forall l, (filterTQ T Q) l = None)
    /\ Ast_deQueue_post_a T Q ptr bp sp mP S'
    \/ exists Q', exists tid', exists tcb',
      Ast_deQueue_post_b T Q' ptr tid' tcb' bp sp mP S'
      /\ Q = addThrd Q' tid'
      /\ WQMod.not_in_dom tid' Q'.

(* <deQueue> : movw Q_HEAD(%bx), %dx *)
Definition Ast_deQueue_4 (T : TQ) (Q : WQ) (ptr : Label)
  (bp sp head tail prev: nat) (mP : MPred) (S : State) :Prop :=
  (ptr |->w head
    *^* ptr + 2 |->w tail
    *^* (Queue Thread ptoTCB head tail prev 0 (filterTQ T Q))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w ptr
      *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 2
  /\ S_RF S BP = sp - 2
  /\ S_RF S BX = ptr
  /\ S_RF S DX = head
  /\ sp >= 2.

Definition L_deQueue_4_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists bp, exists sp,
    exists head, exists tail, exists prev,
      Ast_deQueue_4 T Q ptr bp sp head tail prev mTrue S.

Definition L_deQueue_4_g (S S' : State) : Prop :=
  forall T Q ptr bp sp head tail prev mP,
    Ast_deQueue_4 T Q ptr bp sp head tail prev mP S
    -> (forall l, (filterTQ T Q) l = None)
    /\ Ast_deQueue_post_a T Q ptr bp sp mP S'
    \/ exists Q', exists tid', exists tcb',
      Ast_deQueue_post_b T Q' ptr tid' tcb' bp sp mP S'
      /\ Q = addThrd Q' tid'
      /\ WQMod.not_in_dom tid' Q'.

(* <deQueue> : cmpw $0, %dx *)
Definition Ast_deQueue_5_emp (T : TQ) (Q : WQ) (ptr : Label)
  (bp sp prev: nat) (mP : MPred) (S : State) :Prop :=
  (ptr |->w 0
    *^* ptr + 2 |->w 0
    *^* (Queue Thread ptoTCB 0 0 prev 0 (filterTQ T Q))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w ptr
      *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 2
  /\ S_RF S BP = sp - 2
  /\ S_RF S BX = ptr
  /\ S_RF S DX = 0
  /\ S_Flag S _ZF = true
  /\ sp >= 2.

Definition Ast_deQueue_5_get (T : TQ) (Q Q' : WQ) (ptr : Label) (tcb : Thread)
  (bp sp head tail prev next next' tail': nat) (mP : MPred) (S : State) :Prop :=
  (ptr |->w head
    *^* ptr + 2 |->w tail
    *^* (ptoNode Thread ptoTCB head prev next tcb)
    *^* (Queue Thread ptoTCB next' tail' head 0 (filterTQ T Q'))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w ptr
      *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 2
  /\ S_RF S BP = sp - 2
  /\ S_RF S BX = ptr
  /\ S_RF S DX = head
  /\ S_Flag S _ZF = false
  /\ sp >= 2
  /\ head > 0
  /\ Q = addThrd Q' head
  /\ WQMod.not_in_dom head Q'
  /\ TQMod.maps_to T head tcb
  /\ (tail' = tail /\ tail' > 0 /\ next' = next \/ 
     tail' = 0/\ tail = head /\ next' = 0 /\ next = 0).

Definition L_deQueue_5_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists bp, exists sp,
    exists head, exists tail, exists prev,
      Ast_deQueue_5_emp T Q ptr bp sp prev mTrue S \/
  exists Q', exists tcb, exists next, exists next', exists tail',
      Ast_deQueue_5_get T Q Q' ptr tcb bp sp head tail prev next next' tail' mTrue S.

Definition L_deQueue_5_g (S S' : State) : Prop :=
  forall T Q ptr bp sp head tail prev mP,
    (Ast_deQueue_5_emp T Q ptr bp sp prev mP S \/
      exists Q', exists tcb, exists next, exists next', exists tail',
      Ast_deQueue_5_get T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S)
    -> (forall l, (filterTQ T Q) l = None)
    /\ Ast_deQueue_post_a T Q ptr bp sp mP S'
    \/ exists Q', exists tid', exists tcb',
      Ast_deQueue_post_b T Q' ptr tid' tcb' bp sp mP S'
      /\ Q = addThrd Q' tid'
      /\ WQMod.not_in_dom tid' Q'.

(* <deQueue> : jne deq_a *)
Definition Ast_deQueue_6 (T : TQ) (Q : WQ) (ptr : Label)
  (bp sp prev: nat) (mP : MPred) (S : State) :Prop :=
  (ptr |->w 0
    *^* ptr + 2 |->w 0
    *^* (Queue Thread ptoTCB 0 0 prev 0 (filterTQ T Q))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w ptr
      *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 2
  /\ sp >= 2.

Definition L_deQueue_6_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists bp, exists sp,
    exists prev,
      Ast_deQueue_6 T Q ptr bp sp prev mTrue S.

Definition L_deQueue_6_g (S S' : State) : Prop :=
  forall T Q ptr bp sp prev mP,
    Ast_deQueue_6 T Q ptr bp sp prev mP S
    -> (forall l, (filterTQ T Q) l = None)
    /\ Ast_deQueue_post_a T Q ptr bp sp mP S'
    \/ exists Q', exists tid', exists tcb',
      Ast_deQueue_post_b T Q' ptr tid' tcb' bp sp mP S'
      /\ Q = addThrd Q' tid'
      /\ WQMod.not_in_dom tid' Q'.

(* <deQueue> : movw de $0, %ax *)
Definition Ast_deQueue_7 (T : TQ) (Q : WQ) (ptr : Label)
  (bp sp prev: nat) (mP : MPred) (S : State) :Prop :=
  (ptr |->w 0
    *^* ptr + 2 |->w 0
    *^* (Queue Thread ptoTCB 0 0 prev 0 (filterTQ T Q))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w ptr
      *^* mP) (S_Mem S)
  /\ S_RF S AX = 0
  /\ S_RF S SP = sp - 2
  /\ sp >= 2.

Definition L_deQueue_7_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists bp, exists sp,
    exists prev,
      Ast_deQueue_7 T Q ptr bp sp prev mTrue S.

Definition L_deQueue_7_g (S S' : State) : Prop :=
  forall T Q ptr bp sp prev mP,
    Ast_deQueue_7 T Q ptr bp sp prev mP S
    -> (forall l, (filterTQ T Q) l = None)
    /\ Ast_deQueue_post_a T Q ptr bp sp mP S'
    \/ exists Q', exists tid', exists tcb',
      Ast_deQueue_post_b T Q' ptr tid' tcb' bp sp mP S'
      /\ Q = addThrd Q' tid'
      /\ WQMod.not_in_dom tid' Q'.

(* <deQueue> : popw %bp *)
Definition Ast_deQueue_8 (T : TQ) (Q : WQ) (ptr : Label)
  (bp sp prev: nat) (mP : MPred) (S : State) :Prop :=
  (ptr |->w 0
    *^* ptr + 2 |->w 0
    *^* (Queue Thread ptoTCB 0 0 prev 0 (filterTQ T Q))
    *^* stkSpacePrec sp 2
    *^* sp + 2 |->w ptr
      *^* mP) (S_Mem S)
  /\ S_RF S SP = sp
  /\ S_RF S BP = bp
  /\ S_RF S AX = 0
  /\ sp >= 2.

Definition L_deQueue_8_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists bp, exists sp,
    exists prev,
      Ast_deQueue_8 T Q ptr bp sp prev mTrue S.

Definition L_deQueue_8_g (S S' : State) : Prop :=
  forall T Q ptr bp sp prev mP,
    Ast_deQueue_8 T Q ptr bp sp prev mP S
    -> (forall l, (filterTQ T Q) l = None)
    /\ Ast_deQueue_post_a T Q ptr bp sp mP S'
    \/ exists Q', exists tid', exists tcb',
      Ast_deQueue_post_b T Q' ptr tid' tcb' bp sp mP S'
      /\ Q = addThrd Q' tid'
      /\ WQMod.not_in_dom tid' Q'.
(* <deQueue> : retw *)

(* <Dequeue> : Block1 *)
Definition Ast_deQueue_9 (T : TQ) (Q Q' : WQ) (ptr : Label) (tcb : Thread)
  (bp sp head tail prev next next' tail': nat) (mP : MPred) (S : State) :Prop :=
  (ptr |->w head
    *^* ptr + 2 |->w tail
    *^* (ptoNode Thread ptoTCB head prev next tcb)
    *^* (Queue Thread ptoTCB next' tail' head 0 (filterTQ T Q'))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w ptr
      *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 2
  /\ S_RF S BP = sp - 2
  /\ S_RF S BX = ptr
  /\ S_RF S DX = head
  /\ sp >= 2
  /\ head > 0
  /\ Q = addThrd Q' head
  /\ WQMod.not_in_dom head Q'
  /\ TQMod.maps_to T head tcb
  /\ (tail' = tail /\ tail' > 0 /\ next' = next \/ 
     tail' = 0/\ tail = head /\ next' = 0 /\ next = 0).

Definition L_deQueue_9_p (S : State) : Prop :=
  exists T, exists Q, exists Q', exists ptr, exists tcb, 
    exists bp, exists sp,exists head, exists tail, exists prev,
      exists next, exists next', exists tail',
      Ast_deQueue_9 T Q Q' ptr tcb bp sp head tail prev next next' tail' mTrue S.

Definition L_deQueue_9_g (S S' : State) : Prop :=
  forall T Q Q' ptr tcb bp sp head tail prev next next' tail' mP,
      Ast_deQueue_9 T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S
    -> (forall l, (filterTQ T Q) l = None)
    /\ Ast_deQueue_post_a T Q ptr bp sp mP S'
    \/ exists Q', exists tid', exists tcb',
      Ast_deQueue_post_b T Q' ptr tid' tcb' bp sp mP S'
      /\ Q = addThrd Q' tid'
      /\ WQMod.not_in_dom tid' Q'.

(* <Dequeue> :  movw    %bx, %cx *)
Definition Ast_deQueue_10 (T : TQ) (Q Q' : WQ) (ptr : Label) (tcb : Thread)
  (bp sp head tail prev next next' tail': nat) (mP : MPred) (S : State) :Prop :=
  (ptr |->w head
    *^* ptr + 2 |->w tail
    *^* (ptoNode Thread ptoTCB head prev next tcb)
    *^* (Queue Thread ptoTCB next' tail' head 0 (filterTQ T Q'))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w ptr
      *^* mP) (S_Mem S)
  /\ S_RF S CX = ptr
  /\ S_RF S SP = sp - 2
  /\ S_RF S BP = sp - 2
  /\ S_RF S DX = head
  /\ sp >= 2
  /\ head > 0
  /\ Q = addThrd Q' head
  /\ WQMod.not_in_dom head Q'
  /\ TQMod.maps_to T head tcb
  /\ (tail' = tail /\ tail' > 0 /\ next' = next \/ 
     tail' = 0/\ tail = head /\ next' = 0 /\ next = 0).

Definition L_deQueue_10_p (S : State) : Prop :=
  exists T, exists Q, exists Q', exists ptr, exists tcb, 
    exists bp, exists sp,exists head, exists tail, exists prev,
      exists next, exists next', exists tail',
      Ast_deQueue_10 T Q Q' ptr tcb bp sp head tail prev next next' tail' mTrue S.

Definition L_deQueue_10_g (S S' : State) : Prop :=
  forall T Q Q' ptr tcb bp sp head tail prev next next' tail' mP,
      Ast_deQueue_10 T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S
    -> (forall l, (filterTQ T Q) l = None)
    /\ Ast_deQueue_post_a T Q ptr bp sp mP S'
    \/ exists Q', exists tid', exists tcb',
      Ast_deQueue_post_b T Q' ptr tid' tcb' bp sp mP S'
      /\ Q = addThrd Q' tid'
      /\ WQMod.not_in_dom tid' Q'.

(* <Dequeue> :  movw    %dx, %bx *)
Definition Ast_deQueue_11 (T : TQ) (Q Q' : WQ) (ptr : Label) (tcb : Thread)
  (bp sp head tail prev next next' tail': nat) (mP : MPred) (S : State) :Prop :=
  (ptr |->w head
    *^* ptr + 2 |->w tail
    *^* (ptoNode Thread ptoTCB head prev next tcb)
    *^* (Queue Thread ptoTCB next' tail' head 0 (filterTQ T Q'))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w ptr
      *^* mP) (S_Mem S)
  /\ S_RF S BX = head
  /\ S_RF S CX = ptr
  /\ S_RF S SP = sp - 2
  /\ S_RF S BP = sp - 2
  /\ sp >= 2
  /\ head > 0
  /\ Q = addThrd Q' head
  /\ WQMod.not_in_dom head Q'
  /\ TQMod.maps_to T head tcb
  /\ (tail' = tail /\ tail' > 0 /\ next' = next \/ 
     tail' = 0/\ tail = head /\ next' = 0 /\ next = 0).

Definition L_deQueue_11_p (S : State) : Prop :=
  exists T, exists Q, exists Q', exists ptr, exists tcb, 
    exists bp, exists sp,exists head, exists tail, exists prev,
      exists next, exists next', exists tail',
      Ast_deQueue_11 T Q Q' ptr tcb bp sp head tail prev next next' tail' mTrue S.

Definition L_deQueue_11_g (S S' : State) : Prop :=
  forall T Q Q' ptr tcb bp sp head tail prev next next' tail' mP,
      Ast_deQueue_11 T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S
    -> (forall l, (filterTQ T Q) l = None)
    /\ Ast_deQueue_post_a T Q ptr bp sp mP S'
    \/ exists Q', exists tid', exists tcb',
      Ast_deQueue_post_b T Q' ptr tid' tcb' bp sp mP S'
      /\ Q = addThrd Q' tid'
      /\ WQMod.not_in_dom tid' Q'.

(* <Dequeue> : movw ND_NEXT(%bx), %dx *)
Definition Ast_deQueue_12 (T : TQ) (Q Q' : WQ) (ptr : Label) (tcb : Thread)
  (bp sp head tail prev next next' tail': nat) (mP : MPred) (S : State) :Prop :=
  (ptr |->w head
    *^* ptr + 2 |->w tail
    *^* head |->w prev 
    *^* head + 2 |->w next
    *^* ptoTCB (head + 4) tcb 
    *^* (Queue Thread ptoTCB next' tail' head 0 (filterTQ T Q'))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w ptr
      *^* mP) (S_Mem S)
  /\ S_RF S DX = next
  /\ S_RF S BX = head
  /\ S_RF S CX = ptr
  /\ S_RF S SP = sp - 2
  /\ S_RF S BP = sp - 2
  /\ sp >= 2
  /\ head > 0
  /\ Q = addThrd Q' head
  /\ WQMod.not_in_dom head Q'
  /\ TQMod.maps_to T head tcb
  /\ (tail' = tail /\ tail' > 0 /\ next' = next \/ 
     tail' = 0/\ tail = head /\ next' = 0 /\ next = 0).

Definition L_deQueue_12_p (S : State) : Prop :=
  exists T, exists Q, exists Q', exists ptr, exists tcb, 
    exists bp, exists sp,exists head, exists tail, exists prev,
      exists next, exists next', exists tail',
      Ast_deQueue_12 T Q Q'  ptr tcb bp sp head tail prev next next' tail' mTrue S.

Definition L_deQueue_12_g (S S' : State) : Prop :=
  forall T Q Q' ptr tcb bp sp head tail prev next next' tail' mP,
      Ast_deQueue_12 T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S
    -> (forall l, (filterTQ T Q) l = None)
    /\ Ast_deQueue_post_a T Q ptr bp sp mP S'
    \/ exists Q', exists tid', exists tcb',
      Ast_deQueue_post_b T Q' ptr tid' tcb' bp sp mP S'
      /\ Q = addThrd Q' tid'
      /\ WQMod.not_in_dom tid' Q'.

(* <Dequeue> :  movw %bx, %ax *)
Definition Ast_deQueue_13 (T : TQ) (Q Q' : WQ) (ptr : Label) (tcb : Thread)
  (bp sp head tail prev next next' tail': nat) (mP : MPred) (S : State) :Prop :=
  (ptr |->w head
    *^* ptr + 2 |->w tail
    *^* head |->w prev 
    *^* head + 2 |->w next
    *^* ptoTCB (head + 4) tcb 
    *^* (Queue Thread ptoTCB next' tail' head 0 (filterTQ T Q'))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w ptr
      *^* mP) (S_Mem S)
  /\ S_RF S AX = head
  /\ S_RF S DX = next
  /\ S_RF S CX = ptr
  /\ S_RF S SP = sp - 2
  /\ S_RF S BP = sp - 2
  /\ sp >= 2
  /\ head > 0
  /\ Q = addThrd Q' head
  /\ WQMod.not_in_dom head Q'
  /\ TQMod.maps_to T head tcb
  /\ (tail' = tail /\ tail' > 0 /\ next' = next \/ 
     tail' = 0/\ tail = head /\ next' = 0 /\ next = 0).

Definition L_deQueue_13_p (S : State) : Prop :=
  exists T, exists Q, exists Q', exists ptr, exists tcb, 
    exists bp, exists sp,exists head, exists tail, exists prev,
      exists next, exists next', exists tail',
      Ast_deQueue_13 T Q Q' ptr tcb bp sp head tail prev next next' tail' mTrue S.

Definition L_deQueue_13_g (S S' : State) : Prop :=
  forall T Q Q' ptr tcb bp sp head tail prev next next' tail' mP,
      Ast_deQueue_13 T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S
    -> (forall l, (filterTQ T Q) l = None)
    /\ Ast_deQueue_post_a T Q ptr bp sp mP S'
    \/ exists Q', exists tid', exists tcb',
      Ast_deQueue_post_b T Q' ptr tid' tcb' bp sp mP S'
      /\ Q = addThrd Q' tid'
      /\ WQMod.not_in_dom tid' Q'.

(* <Dequeue> :  movw %cx, %bx *)
Definition Ast_deQueue_14 (T : TQ) (Q Q' : WQ) (ptr : Label) (tcb : Thread)
  (bp sp head tail prev next next' tail': nat) (mP : MPred) (S : State) :Prop :=
  (ptr |->w head
    *^* ptr + 2 |->w tail
    *^* head |->w prev 
    *^* head + 2 |->w next
    *^* ptoTCB (head + 4) tcb 
    *^* (Queue Thread ptoTCB next' tail' head 0 (filterTQ T Q'))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w ptr
      *^* mP) (S_Mem S)
  /\ S_RF S BX = ptr
  /\ S_RF S AX = head
  /\ S_RF S DX = next
  /\ S_RF S SP = sp - 2
  /\ S_RF S BP = sp - 2
  /\ sp >= 2
  /\ head > 0
  /\ Q = addThrd Q' head
  /\ WQMod.not_in_dom head Q'
  /\ TQMod.maps_to T head tcb
  /\ (tail' = tail /\ tail' > 0 /\ next' = next \/ 
     tail' = 0/\ tail = head /\ next' = 0 /\ next = 0).

Definition L_deQueue_14_p (S : State) : Prop :=
  exists T, exists Q, exists Q', exists ptr, exists tcb, 
    exists bp, exists sp,exists head, exists tail, exists prev,
      exists next, exists next', exists tail',
      Ast_deQueue_14 T Q Q' ptr tcb bp sp head tail prev next next' tail' mTrue S.

Definition L_deQueue_14_g (S S' : State) : Prop :=
  forall T Q Q' ptr tcb bp sp head tail prev next next' tail' mP,
      Ast_deQueue_14 T Q Q' ptr tcb bp sp head tail prev next next' tail' mP S
    -> (forall l, (filterTQ T Q) l = None)
    /\ Ast_deQueue_post_a T Q ptr bp sp mP S'
    \/ exists Q', exists tid', exists tcb',
      Ast_deQueue_post_b T Q' ptr tid' tcb' bp sp mP S'
      /\ Q = addThrd Q' tid'
      /\ WQMod.not_in_dom tid' Q'.

(* <Dequeue> : cmpw $0, %dx *)
Definition Ast_deQueue_15_emp (T : TQ) (Q Q' : WQ) (ptr : Label) (tcb : Thread)
  (bp sp head tail prev: nat) (mP : MPred) (S : State) :Prop :=
  (ptr |->w head
    *^* ptr + 2 |->w tail
    *^* head |->w prev 
    *^* head + 2 |->w 0
    *^* ptoTCB (head + 4) tcb 
    *^* (Queue Thread ptoTCB 0 0 head 0 (filterTQ T Q'))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w ptr
      *^* mP) (S_Mem S)
  /\ S_RF S BX = ptr
  /\ S_RF S AX = head
  /\ S_RF S DX = 0
  /\ S_RF S SP = sp - 2
  /\ S_RF S BP = sp - 2
  /\ S_Flag S _ZF = true
  /\ sp >= 2
  /\ head > 0
  /\ Q = addThrd Q' head
  /\ WQMod.not_in_dom head Q'
  /\ TQMod.maps_to T head tcb.

Definition Ast_deQueue_15_get (T : TQ) (Q Q' : WQ) (ptr : Label) (tcb : Thread)
  (bp sp head tail prev next: nat) (mP : MPred) (S : State) :Prop :=
  (ptr |->w head
    *^* ptr + 2 |->w tail
    *^* head |->w prev 
    *^* head + 2 |->w next
    *^* ptoTCB (head + 4) tcb 
    *^* (Queue Thread ptoTCB next tail head 0 (filterTQ T Q'))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w ptr
      *^* mP) (S_Mem S)
  /\ S_RF S BX = ptr
  /\ S_RF S AX = head
  /\ S_RF S DX = next
  /\ S_RF S SP = sp - 2
  /\ S_RF S BP = sp - 2
  /\ S_Flag S _ZF = false
  /\ sp >= 2
  /\ head > 0
  /\ next <> 0
  /\ Q = addThrd Q' head
  /\ WQMod.not_in_dom head Q'
  /\ TQMod.maps_to T head tcb.

Definition L_deQueue_15_p (S : State) : Prop :=
  exists T, exists Q, exists Q', exists ptr, exists tcb, 
    exists bp, exists sp,exists head, exists tail, exists prev,
      Ast_deQueue_15_emp T Q Q' ptr tcb bp sp head tail prev mTrue S\/ 
      exists next,
      Ast_deQueue_15_get T Q Q' ptr tcb bp sp head tail prev next mTrue S.

Definition L_deQueue_15_g (S S' : State) : Prop :=
  forall T Q Q' ptr tcb bp sp head tail prev mP,
    (Ast_deQueue_15_emp T Q Q' ptr tcb bp sp head tail prev mP S \/
      exists next,
        Ast_deQueue_15_get T Q Q' ptr tcb bp sp head tail prev next mP S)
    -> (forall l, (filterTQ T Q) l = None)
    /\ Ast_deQueue_post_a T Q ptr bp sp mP S'
    \/ exists Q', exists tid', exists tcb',
      Ast_deQueue_post_b T Q' ptr tid' tcb' bp sp mP S'
      /\ Q = addThrd Q' tid'
      /\ WQMod.not_in_dom tid' Q'.

(* <deQueue> : jne deq_b *)
Definition Ast_deQueue_16 (T : TQ) (Q Q' : WQ) (ptr : Label) (tcb : Thread)
  (bp sp head tail prev: nat) (mP : MPred) (S : State) :Prop :=
  (ptr |->w head
    *^* ptr + 2 |->w tail
    *^* head |->w prev 
    *^* head + 2 |->w 0
    *^* ptoTCB (head + 4) tcb 
    *^* (Queue Thread ptoTCB 0 0 head 0 (filterTQ T Q'))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w ptr
      *^* mP) (S_Mem S)
  /\ S_RF S BX = ptr
  /\ S_RF S AX = head
  /\ S_RF S DX = 0
  /\ S_RF S SP = sp - 2
  /\ S_RF S BP = sp - 2
  /\ sp >= 2
  /\ head > 0
  /\ Q = addThrd Q' head
  /\ WQMod.not_in_dom head Q'
  /\ TQMod.maps_to T head tcb.

Definition L_deQueue_16_p (S : State) : Prop :=
  exists T, exists Q, exists Q', exists ptr, exists tcb, 
    exists bp, exists sp,exists head, exists tail, exists prev,
      Ast_deQueue_16 T Q Q' ptr tcb bp sp head tail prev mTrue S.

Definition L_deQueue_16_g (S S' : State) : Prop :=
  forall T Q Q' ptr tcb bp sp head tail prev mP,
      Ast_deQueue_16 T Q Q' ptr tcb bp sp head tail prev mP S
    -> (forall l, (filterTQ T Q) l = None)
    /\ Ast_deQueue_post_a T Q ptr bp sp mP S'
    \/ exists Q', exists tid', exists tcb',
      Ast_deQueue_post_b T Q' ptr tid' tcb' bp sp mP S'
      /\ Q = addThrd Q' tid'
      /\ WQMod.not_in_dom tid' Q'.

(* <deQueue> : movw %dx, Q_TAIL(%bx) *)

(* <deQueue> : Block3 *)
Definition Ast_deQueue_17 (T : TQ) (Q Q' : WQ) (ptr : Label) (tcb : Thread)
  (bp sp head prev head' tail': nat) (mP : MPred) (S : State) :Prop :=
  (ptr |->w head
    *^* ptr + 2 |->w tail'
    *^* head |->w prev 
    *^* head + 2 |->w head'
    *^* ptoTCB (head + 4) tcb 
    *^* (Queue Thread ptoTCB head' tail' head 0 (filterTQ T Q'))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w ptr
      *^* mP) (S_Mem S)
  /\ S_RF S BX = ptr
  /\ S_RF S AX = head
  /\ S_RF S DX = head'
  /\ S_RF S SP = sp - 2
  /\ S_RF S BP = sp - 2
  /\ sp >= 2
  /\ head > 0
  /\ Q = addThrd Q' head
  /\ WQMod.not_in_dom head Q'
  /\ TQMod.maps_to T head tcb.

Definition L_deQueue_17_p (S : State) : Prop :=
  exists T, exists Q, exists Q', exists ptr, exists tcb, 
    exists bp, exists sp,exists head, exists prev, exists head',
      exists tail',
        Ast_deQueue_17 T Q Q' ptr tcb bp sp head prev head' tail' mTrue S.

Definition L_deQueue_17_g (S S' : State) : Prop :=
  forall T Q Q' ptr tcb bp sp head prev head' tail' mP,
      Ast_deQueue_17 T Q Q' ptr tcb bp sp head prev head' tail' mP S
    -> (forall l, (filterTQ T Q) l = None)
    /\ Ast_deQueue_post_a T Q ptr bp sp mP S'
    \/ exists Q', exists tid', exists tcb',
      Ast_deQueue_post_b T Q' ptr tid' tcb' bp sp mP S'
      /\ Q = addThrd Q' tid'
      /\ WQMod.not_in_dom tid' Q'.

(* <deQueue> : movw %dx, Q_HEAD(%bx) *)
Definition Ast_deQueue_18 (T : TQ) (Q Q' : WQ) (ptr : Label) (tcb : Thread)
  (bp sp head prev head' tail': nat) (mP : MPred) (S : State) :Prop :=
  (ptr |->w head'
    *^* ptr + 2 |->w tail'
    *^* head |->w prev 
    *^* head + 2 |->w head'
    *^* ptoTCB (head + 4) tcb 
    *^* (Queue Thread ptoTCB head' tail' head 0 (filterTQ T Q'))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w ptr
      *^* mP) (S_Mem S)
  /\ S_RF S BX = ptr
  /\ S_RF S AX = head
  /\ S_RF S SP = sp - 2
  /\ S_RF S BP = sp - 2
  /\ sp >= 2
  /\ head > 0
  /\ Q = addThrd Q' head
  /\ WQMod.not_in_dom head Q'
  /\ TQMod.maps_to T head tcb.

Definition L_deQueue_18_p (S : State) : Prop :=
  exists T, exists Q, exists Q', exists ptr, exists tcb, 
    exists bp, exists sp,exists head, exists prev, exists head',
      exists tail',
        Ast_deQueue_18 T Q Q' ptr tcb bp sp head prev head' tail' mTrue S.

Definition L_deQueue_18_g (S S' : State) : Prop :=
  forall T Q Q' ptr tcb bp sp head prev head' tail' mP,
      Ast_deQueue_18 T Q Q' ptr tcb bp sp head prev head' tail' mP S
    -> (forall l, (filterTQ T Q) l = None)
    /\ Ast_deQueue_post_a T Q ptr bp sp mP S'
    \/ exists Q', exists tid', exists tcb',
      Ast_deQueue_post_b T Q' ptr tid' tcb' bp sp mP S'
      /\ Q = addThrd Q' tid'
      /\ WQMod.not_in_dom tid' Q'.

(* <deQueue> : popw %bp *)
Definition Ast_deQueue_19 (T : TQ) (Q Q' : WQ) (ptr : Label) (tcb : Thread)
  (bp sp head prev head' tail': nat) (mP : MPred) (S : State) :Prop :=
  (ptr |->w head'
    *^* ptr + 2 |->w tail'
    *^* head |->w prev 
    *^* head + 2 |->w head'
    *^* ptoTCB (head + 4) tcb 
    *^* (Queue Thread ptoTCB head' tail' head 0 (filterTQ T Q'))
    *^* stkSpacePrec sp 2 
    *^* sp + 2 |->w ptr
      *^* mP) (S_Mem S)
  /\ S_RF S BP = bp
  /\ S_RF S BX = ptr
  /\ S_RF S AX = head
  /\ S_RF S SP = sp
  /\ sp >= 2
  /\ head > 0
  /\ Q = addThrd Q' head
  /\ WQMod.not_in_dom head Q'
  /\ TQMod.maps_to T head tcb.

Definition L_deQueue_19_p (S : State) : Prop :=
  exists T, exists Q, exists Q', exists ptr, exists tcb, 
    exists bp, exists sp,exists head, exists prev, exists head',
      exists tail',
        Ast_deQueue_19 T Q Q' ptr tcb bp sp head prev head' tail' mTrue S.

Definition L_deQueue_19_g (S S' : State) : Prop :=
  forall T Q Q' ptr tcb bp sp head prev head' tail' mP,
      Ast_deQueue_19 T Q Q' ptr tcb bp sp head prev head' tail' mP S
    -> (forall l, (filterTQ T Q) l = None)
    /\ Ast_deQueue_post_a T Q ptr bp sp mP S'
    \/ exists Q', exists tid', exists tcb',
      Ast_deQueue_post_b T Q' ptr tid' tcb' bp sp mP S'
      /\ Q = addThrd Q' tid'
      /\ WQMod.not_in_dom tid' Q'.

(* <deQueue> : retw *)


Definition spec_deQueue := Ast p_deQueue g_deQueue sctag.

Definition psi_deQueue_Block0  :=
  (SCAP_consCT L_deQ_a (Ast L_deQueue_9_p L_deQueue_9_g sctag)
    SCAP_emptCT).

Definition psi_deQueue_Block1  :=
  (SCAP_consCT L_deQ_b (Ast L_deQueue_17_p L_deQueue_17_g sctag)
    SCAP_emptCT).

Definition psi_deQueue_Block2  := 
  SCAP_emptCT.
