(***************************************************************)
(* Coq code for AIM                                            *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(* by wangwei *)
Add LoadPath "..".

Add LoadPath "../CAP0".
Add LoadPath "../SCAP".

Require Import Coqlib.

Require Import Arith.
Require Import Machine.
Require Import Reg.
Require Import Assertion.
Require Import Assertion_Lemmas3.
Require Import Assertion_Lemmas4.
Require Import Operational_Semantics.
Require Import OpSem_Lemmas.

Require Import AIM.
Require Import SCAP.
Require Import QLib.
Require Import Linking.
Require Import Linking_Lemmas3.
Require Import QLemmas.

Require Import Kernel.

Require Import EnQueueSpec.

(** Functioon: EnQueue, add a node into Queue **)
(** Two blocks: EnQueue, EnQ_NotNull             **)
(***********************************************)
(********Source Code *************************** 
1190 <EnQueue>:
1190:   pushw   %bp
        movw    %sp, %bp
        movw    4(%bp), %ax             # ax: pointer to node
        movw    6(%bp), %bx             # bx: pointer to Q
        movw    Q_TAIL(%bx), %dx
1195:   movw    %bx, %cx                # save bx
        movw    %ax, %bx
        movw    %dx, ND_PREV(%bx)
        movw    $0, ND_NEXT(%bx)
        cmpw    $0, %dx
1200:   jne     EnQ_NotNull
        movw    %cx, %bx                # bx: pointers to Q
        movw    %ax, Q_HEAD(%bx)        # head = newtcb
        movw    %ax, Q_TAIL(%bx)
        popw    %bp
        retw
EnQ_NotNull :
1206:   movw    %dx, %bx
        movw    %ax, ND_NEXT(%bx)
        movw    %cx, %bx                # bx: pointers to Q
        movw    %ax, Q_TAIL(%bx)
1210:   popw    %bp
        retw
************************************************)

(***************************************)

(*  <EnQueue + 0> : pushw %bp *)
Lemma SCAP_WFcmd_enQueue_0_pushw :
  SCAP_WFcmd (SCAP_consCT (L_enQueue + 1) (Ast L_enQueue_1_p L_enQueue_1_g sctag) psi_enQueue_Block0)
    (L_enQueue) (Ast p_enQueue g_enQueue sctag) 
    (comml (_pushwp (Opr_reg BP))).
Proof.
  assert (Hpp': forall T Q ptr tid tcb bp sp mP S S',
    Ast_enQueue_pre T Q ptr tid tcb bp sp mP S
    -> nextS (L_enQueue) (comml (_pushwp (Opr_reg BP))) S S'
    -> Ast_enQueue_1 T Q ptr tid tcb bp sp mP S').
    unfold Ast_enQueue_pre,Ast_enQueue_1.
    intros T Q ptr tid tcb bp sp mP S S' Hp HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [[HSS1 [HSS2 [jk HSS3]]] [HSS4 [HSS5 HSS6]]].
    destruct Hp as [Hp1 [Hp2 [Hp3 Hp4]]].
    subst M' R' fl' isr'.
    rewrite Hp3; rewrite Hp4.
    intuition.
    sep on.
    sep lift 2 in Hp1.
    sep lift 2.
    assert (MemUpdW M (sp - 2) bp jk ||- 
      sp - 2 |->w bp *^*
       (sp + 2 |->w tid *^* (sp + 4 |->w ptr *^* mP) *^*
        (embed_TCB tid tcb *^* ptoTQ ptr (filterTQ T Q)))).
      sep upd.
      sep matchH Hp1 with 10.
      sep trivial.
    unfold MemUpdW in H.
    assumption.
  unfold SCAP_WFcmd.
  exists L_enQueue_1_p;exists L_enQueue_1_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [ptr [tid [tcb [bp [sp HS]]]]]]].
  destruct HS as [HS1 [HS2 [HS3 HS4]]].
  exists (MemUpdW M (sp-2) bp 0, RF_update R SP (sp - 2), fl, isr).
  simplS in * .
  apply nextS_pushwp with bp.
  unfold opr_eval.
  rewrite HS3;reflexivity.
  unfold pushW.
  intuition.
  rewrite HS4.
    sep on.
    sep lift 2 in HS1.
    sep trivial.
  exists 0.
  unfold MemUpdW.
  rewrite HS4;reflexivity.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold p_enQueue, L_enQueue_1_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [ptr [tid [tcb [bp [sp H]]]]]]].
  exists T; exists Q; exists ptr; exists tid; exists tcb .
  exists bp; exists sp.
  apply Hpp' with S;assumption.
  (* g' ==> g *)
  unfold g_enQueue, L_enQueue_1_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S''.
  apply Hpp' with S;assumption.
Qed.

(* <DeQueue + 1> : movw %sp, %bp *)
Lemma SCAP_WFcmd_enQueue_1_movwp :
  SCAP_WFcmd (SCAP_consCT (L_enQueue + 1 + 1) (Ast L_enQueue_2_p L_enQueue_2_g sctag) psi_enQueue_Block0)
    (L_enQueue + 1) (Ast L_enQueue_1_p L_enQueue_1_g sctag) 
    (comml (_movwp (Opr_reg SP) BP)).
Proof.
  assert (Hpp': forall T Q ptr tid tcb bp sp mP S S',
    Ast_enQueue_1 T Q ptr tid tcb bp sp mP S
    -> nextS (L_enQueue + 1) (comml (_movwp (Opr_reg SP) BP)) S S'
    -> Ast_enQueue_2 T Q ptr tid tcb bp sp mP S').
    unfold Ast_enQueue_1,Ast_enQueue_2.
    intros T Q ptr tid tcb bp sp mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [HSS1 [HSS2 [HSS3 HSS4]]].
    destruct HS as [HS1 [HS2 [HS3 [HS4 HS5]]]].
    subst M' R' fl' isr'.
    rewrite HS5.
    intuition.
  unfold SCAP_WFcmd.
  exists L_enQueue_2_p;exists L_enQueue_2_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [ptr [tid [tcb [bp [sp HS]]]]]]].
  destruct HS as [HS1 [HS2 [HS3 [HS4 HS5]]]].
  exists (M, RF_update R BP (sp - 2), fl, isr).
  simplS in * .
  apply nextS_movwp with (sp - 2);trivial.
  unfold opr_eval.
  rewrite HS5;reflexivity.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_enQueue_1_p, L_enQueue_2_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [ptr [tid [tcb [bp [sp H]]]]]]].
  exists T; exists Q; exists ptr; exists tid; exists tcb; 
    exists bp; exists sp.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_enQueue_1_g, L_enQueue_2_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S''.
  apply Hpp' with S;assumption.
Qed.

(* <EnQueue + 2> : movw 4(%bp), %ax *)
Lemma SCAP_WFcmd_enQueue_2_movwld :
  SCAP_WFcmd (SCAP_consCT (L_enQueue + 2 + 1) (Ast L_enQueue_3_p L_enQueue_3_g sctag) psi_enQueue_Block0)
    (L_enQueue + 2) (Ast L_enQueue_2_p L_enQueue_2_g sctag) 
    (comml (_movwld (A_reg 4 BP) AX)).
Proof.
  assert (Hpp': forall T Q ptr tid tcb bp sp mP S S',
    Ast_enQueue_2 T Q ptr tid tcb bp sp mP S
    -> nextS (L_enQueue + 2) (comml (_movwld (A_reg 4 BP) AX)) S S'
    -> Ast_enQueue_3 T Q ptr tid tcb bp sp mP S').
    unfold Ast_enQueue_2,Ast_enQueue_3.
    intros T Q ptr tid tcb bp sp mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [HSS1 [[x [HSS2 HSS3]] [HSS4 HSS5]]].
    destruct HS as [HS1 [HS2 [HS3 [HS4 HS5]]]].
    subst M' R' fl' isr'.
    assert (E: x = tid).
      rewrite HS4 in HSS2.
      sep on.
      sep lift 3 in HS1.
      arith_rep (sp-2+4) with (sp+2) in HSS2.
      sep inj; auto.
    intuition.
  unfold SCAP_WFcmd.
  exists L_enQueue_3_p;exists L_enQueue_3_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [ptr [tid [tcb [bp [sp HS]]]]]]].
  destruct HS as [HS1 [HS2 [HS3 [HS4 HS5]]]].
  exists (M, RF_update R AX tid, fl, isr).
  simplS in * .
  apply nextS_movwld with (sp+2) tid;trivial.
  unfold addr_eval.
  rewrite HS4;omega.
    sep on.
    sep lift 3 in HS1.
    sep trivial.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_enQueue_2_p, L_enQueue_3_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [ptr [tid [tcb [bp [sp H]]]]]]].
  exists T; exists Q; exists ptr; exists tid; exists tcb.
  exists bp; exists sp.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_enQueue_2_g, L_enQueue_3_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S''.
  apply Hpp' with S; assumption.
Qed.

(* Lemma ptoTQ_elim :    *)
(*    forall Q ptr M, *)
(*      M ||- (ptoTQ ptr Q) *)
(*      -> exists head, exists tail, exists prev, *)
(*        M ||- (ptr |->w head *^* ptr + 2 |->w tail *^*  *)
(*          Queue Thread ptoTCB head tail prev 0 Q). *)
(* Proof. *)
(*   intros. *)
(*   sep off. *)
(*   unfold ptoTQ , ptoQ2 in H. *)
(*   destruct H as [head [tail [prev H]]]. *)
(*   exists head; exists tail; exists prev. *)
(*   assumption. *)
(* Qed. *)
(* Implicit Arguments ptoTQ_elim [Q ptr M]. *)

(* Lemma ptoTQ_elim1 :    *)
(*    forall Q ptr M A, *)
(*      M ||- (ptoTQ ptr Q *^* A) *)
(*      -> exists head, exists tail, exists prev, *)
(*        M ||- (ptr |->w head *^* ptr + 2 |->w tail *^*  *)
(*          Queue Thread ptoTCB head tail prev 0 Q *^* A). *)
(* Proof. *)
(*   intros. *)
(*   destruct H as [M1 [M2 [Hdis [HQ HA]]]]. *)
(*   destruct (ptoTQ_elim HQ) as [head [tail [prev H]]]. *)
(*   exists head; exists tail; exists prev. *)
(*   exists M1;exists M2. *)
(*   intuition. *)
(* Qed. *)
(* Implicit Arguments ptoTQ_elim1 [Q ptr M A]. *)

(* <EnQueue + 3> : movw 6(%bp), %bx *)
Lemma SCAP_WFcmd_enQueue_3_movwld :
  SCAP_WFcmd (SCAP_consCT (L_enQueue + 3 + 1) (Ast L_enQueue_4_p L_enQueue_4_g sctag) psi_enQueue_Block0)
    (L_enQueue + 3) (Ast L_enQueue_3_p L_enQueue_3_g sctag) 
    (comml (_movwld (A_reg 6 BP) BX)).
Proof.
  assert (Hpp': forall T Q ptr tid tcb bp sp mP S S',
    Ast_enQueue_3 T Q ptr tid tcb bp sp mP S
    -> nextS (L_enQueue + 3) (comml (_movwld (A_reg 6 BP) BX)) S S'
    -> exists head, exists tail, exists prev,
      Ast_enQueue_4 T Q ptr tid tcb bp sp head tail prev mP S').
    unfold Ast_enQueue_4, Ast_enQueue_4.
    intros T Q ptr tid tcb bp sp mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [HSS1 [[x [HSS2 HSS3]] [HSS4 HSS5]]].
    destruct HS as [HS1 [HS2 [HS3 [HS4 [HS5 HS6]]]]].
    simplS in * .
    subst M' R' fl' isr'.
    sep on.
    sep lift 1 in HS1.
    destruct (ptoTQ_elim1 HS1) as [head [tail [prev HS]]].
    clear HS1.
    exists head; exists tail; exists prev.
    assert (E: x = ptr).
      rewrite HS4 in HSS2.
      sep lift 5 in HS.
      arith_rep (sp-2+6) with (sp+4) in HSS2.
      sep inj;auto.
    intuition.
  sep on.
  sep matchH HS with 10.
  unfold SCAP_WFcmd.
  exists L_enQueue_4_p;exists L_enQueue_4_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [ptr [tid [tcb [bp [sp HS]]]]]]].
  destruct HS as [HS1 [HS2 [HS3 [HS4 [HS5 HS6]]]]].
  exists (M, RF_update R BX ptr, fl, isr).
  simplS in * .
  apply nextS_movwld with (sp+4) ptr;trivial.
  unfold addr_eval.
  rewrite HS4;omega.
    sep on.
    sep lift 4 in HS1.
    sep trivial.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_enQueue_3_p, L_enQueue_4_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [ptr [tid [tcb [bp [sp H]]]]]]].
  exists T; exists Q; exists ptr; exists tid; exists tcb.
  exists bp; exists sp.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_enQueue_3_g, L_enQueue_4_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  assert (exists head : nat,
           exists tail : nat,
             exists prev : nat,
               Ast_enQueue_4 T Q ptr tid tcb bp sp head tail prev mP S').
  apply Hpp' with S;assumption.
  destruct H0 as [head [tail [prev H0]]].
  apply HgS'S'' with head tail prev;assumption.
Qed.

(* <EnQueue + 4> : movw Q_TAIL(%bx), %dx *)
Lemma SCAP_WFcmd_enQueue_4_movwld :
  SCAP_WFcmd (SCAP_consCT (L_enQueue + 4 + 1) (Ast L_enQueue_5_p L_enQueue_5_g sctag) psi_enQueue_Block0)
    (L_enQueue + 4) (Ast L_enQueue_4_p L_enQueue_4_g sctag) 
    (comml (_movwld (A_reg Q_TAIL BX) DX)).
Proof.
  assert (Hpp': forall T Q ptr tid tcb bp sp head tail prev mP S S',
    Ast_enQueue_4 T Q ptr tid tcb bp sp head tail prev mP S
    -> nextS (L_enQueue + 4) (comml (_movwld (A_reg Q_TAIL BX) DX)) S S'
    -> Ast_enQueue_5 T Q ptr tid tcb bp sp head tail prev mP S').
    unfold Ast_enQueue_4, Ast_enQueue_5.
    intros T Q ptr tid tcb bp sp head tail prev mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [HSS1 [[x [HSS2 HSS3]] [HSS4 HSS5]]].
    destruct HS as [HS1 [HS2 [HS3 [HS4 [HS5 [HS6 HS7]]]]]].
    subst M' R' fl' isr'.
    assert (E: x = tail).
      rewrite HS7 in HSS2.
      sep on.
      sep lift 2 in HS1.
      sep inj;auto.
    intuition.
  unfold SCAP_WFcmd.
  exists L_enQueue_5_p; exists L_enQueue_5_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [ptr [tid [tcb [bp [sp [head [tail [prev HS]]]]]]]]]].
  destruct HS as [HS1 [HS2 [HS3 [HS4 [HS5 [HS6 HS7]]]]]].
  exists (M, RF_update R DX tail, fl, isr).
  simplS in * .
  apply nextS_movwld with (ptr + 2) tail;trivial.
  unfold addr_eval.
  rewrite HS7;reflexivity.
    sep on.
    sep lift 2 in HS1.
    sep trivial.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_enQueue_4_p, L_enQueue_5_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [ptr [tid [tcb [bp [sp [head [tail [prev H]]]]]]]]]].
  exists T; exists Q; exists ptr; exists tid; exists tcb.
  exists bp; exists sp; exists head; exists tail; exists prev.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_enQueue_4_g, L_enQueue_5_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S'' with head tail prev.
  apply Hpp' with S;assumption.
Qed.

(* <EnQueue + 5> : movw %bx, %cx *)
Lemma SCAP_WFcmd_enQueue_5_movwp :
  SCAP_WFcmd (SCAP_consCT (L_enQueue + 5 + 1) (Ast L_enQueue_6_p L_enQueue_6_g sctag) psi_enQueue_Block0)
    (L_enQueue + 5) (Ast L_enQueue_5_p L_enQueue_5_g sctag) 
    (comml (_movwp (Opr_reg BX) CX)).
Proof.
  assert (Hpp': forall T Q ptr tid tcb bp sp head tail prev mP S S',
    Ast_enQueue_5 T Q ptr tid tcb bp sp head tail prev mP S
    -> nextS (L_enQueue + 5) (comml (_movwp (Opr_reg BX) CX)) S S'
    -> Ast_enQueue_6 T Q ptr tid tcb bp sp head tail prev mP S').
    unfold Ast_enQueue_5,Ast_enQueue_6.
    intros T Q ptr tid tcb bp sp head tail prev mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [HSS1 [HSS2 [HSS3 HSS4]]].
    destruct HS as [HM [Htid [Hsp [HBP [HSP [HAX [HBX HDX]]]]]]].
    subst M' R' fl' isr'.
    rewrite HBX.
    intuition.
  unfold SCAP_WFcmd.
  exists L_enQueue_6_p;exists L_enQueue_6_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [ptr [tid [tcb [bp [sp [head [tail [prev HS]]]]]]]]]].
  destruct HS as [HM [Htid [Hsp [HBP [HSP [HAX [HBX HDX]]]]]]].
  exists (M, RF_update R CX ptr, fl, isr).
  simplS in * .
  apply nextS_movwp with ptr; trivial.
  unfold opr_eval.
  rewrite HBX;reflexivity.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_enQueue_5_p, L_enQueue_6_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [ptr [tid [tcb [bp [sp [head [tail [prev H]]]]]]]]]].
  exists T; exists Q; exists ptr; exists tid; exists tcb; 
    exists bp; exists sp; exists head; exists tail; exists prev.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_enQueue_5_g, L_enQueue_6_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S'' with head tail prev.
  apply Hpp' with S;assumption.
Qed.

(* <EnQueue + 6> : movw %ax, %bx *)
Lemma SCAP_WFcmd_enQueue_6_movwp :
  SCAP_WFcmd (SCAP_consCT (L_enQueue + 6 + 1) (Ast L_enQueue_7_p L_enQueue_7_g sctag) psi_enQueue_Block0)
    (L_enQueue + 6) (Ast L_enQueue_6_p L_enQueue_6_g sctag) 
    (comml (_movwp (Opr_reg AX) BX)).
Proof.
  assert (Hpp': forall T Q ptr tid tcb bp sp head tail prev mP S S',
    Ast_enQueue_6 T Q ptr tid tcb bp sp head tail prev mP S
    -> nextS (L_enQueue + 6) (comml (_movwp (Opr_reg AX) BX)) S S'
    -> Ast_enQueue_7 T Q ptr tid tcb bp sp head tail prev mP S').
    unfold Ast_enQueue_6,Ast_enQueue_7.
    intros T Q ptr tid tcb bp sp head tail prev mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [HSS1 [HSS2 [HSS3 HSS4]]].
    destruct HS as [HM [Htid [Hsp [HBP [HSP [HAX [HDX HCX]]]]]]].
    subst M' R' fl' isr'.
    rewrite HAX.
    unfold embed_TCB in HM.
    intuition.
  unfold SCAP_WFcmd.
  exists L_enQueue_7_p;exists L_enQueue_7_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [ptr [tid [tcb [bp [sp [head [tail [prev HS]]]]]]]]]].
  destruct HS as [HM [Htid [Hsp [HBP [HSP [HAX [HDX HCX]]]]]]].
  exists (M, RF_update R BX tid, fl, isr).
  simplS in * .
  apply nextS_movwp with tid; trivial.
  unfold opr_eval.
  rewrite HAX;reflexivity.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_enQueue_6_p, L_enQueue_7_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [ptr [tid [tcb [bp [sp [head [tail [prev H]]]]]]]]]].
  exists T; exists Q; exists ptr; exists tid; exists tcb; 
    exists bp; exists sp; exists head; exists tail; exists prev.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_enQueue_6_g, L_enQueue_7_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S'' with head tail prev.
  apply Hpp' with S;assumption.
Qed.

(* <EnQueue + 7> : movw %dx, ND_PREV(%bx) *)
Lemma SCAP_WFcmd_enQueue_7_movwst :
  SCAP_WFcmd (SCAP_consCT (L_enQueue + 7 + 1) (Ast L_enQueue_8_p L_enQueue_8_g sctag) psi_enQueue_Block0)
    (L_enQueue + 7) (Ast L_enQueue_7_p L_enQueue_7_g sctag) 
    (comml (_movwst (Opr_reg DX) (A_reg ND_PREV BX))).
Proof.
  assert (Hpp': forall T Q ptr tid tcb bp sp head tail prev mP S S',
    Ast_enQueue_7 T Q ptr tid tcb bp sp head tail prev  mP S
    -> nextS (L_enQueue + 7) (comml (_movwst (Opr_reg DX)(A_reg ND_PREV BX))) S S'
    -> Ast_enQueue_8 T Q ptr tid tcb bp sp head tail prev mP S').
    unfold Ast_enQueue_7, Ast_enQueue_8.
    intros T Q ptr tid tcb bp sp head tail prev mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [[HSS1 [jk HSS2]] [HSS3 [HSS4 HSS5]]].
    destruct HS as [HM [Htid [Hsp [HBP [HSP [HAX [HDX [HCX HBX]]]]]]]].
    subst M' R' fl' isr'.
    intuition.
    sep on.
    assert (MemUpdW M tid tail jk
   ||- tid |->w tail *^* tid + 2 |->w_ *^* ptoTCB (tid + 4) tcb *^*
       ptr |->w head *^* ptr + 2 |->w tail *^*
       Queue Thread ptoTCB head tail prev 0 (filterTQ T Q) *^*
       sp - 2 |->w bp *^* sp + 2 |->w tid *^* sp + 4 |->w ptr *^* mP).
    sep lift 0.

    sep upd.
    sep matchH HM with 10.
    unfold MemUpdW in H.
    rewrite HBX; rewrite HDX.
    replace (tid+0) with tid.
    assumption.
    auto.
  unfold SCAP_WFcmd.
  exists L_enQueue_8_p; exists L_enQueue_8_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [ptr [tid [tcb [bp [sp [head [tail [prev  HS]]]]]]]]]].
  destruct HS as [HM [Htid [Hsp [HBP [HSP [HAX [HDX [HCX HBX]]]]]]]].
  exists (MemUpdW M tid tail 0, R, fl, isr).
  simplS in * .  
  apply nextS_movwst with tid tail;trivial.
  unfold addr_eval.
  rewrite HBX;auto.
  simpl;auto.
  sep on.
  sep lift 0 in HM.
  sep trivial.
  apply MemUpdW_MemUpd16 with 0; trivial.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_enQueue_7_p, L_enQueue_8_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [ptr [tid [tcb [bp [sp [head [tail [prev H]]]]]]]]]].
  exists T; exists Q; exists ptr; exists tid; exists tcb; exists bp.
  exists sp; exists head; exists tail; exists prev.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_enQueue_7_g, L_enQueue_8_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S'' with head tail prev .
  apply Hpp' with S;assumption.
Qed.

(* <EnQueue + 8> : movw $0, ND_NEXT(%bx) *)
Lemma SCAP_WFcmd_enQueue_8_movwst :
  SCAP_WFcmd (SCAP_consCT (L_enQueue + 8 + 1) (Ast L_enQueue_9_p L_enQueue_9_g sctag) psi_enQueue_Block0)
    (L_enQueue + 8) (Ast L_enQueue_8_p L_enQueue_8_g sctag) 
    (comml (_movwst (Opr_imm 0) (A_reg ND_NEXT BX))).
Proof.
  assert (Hpp': forall T Q ptr tid tcb bp sp head tail prev mP S S',
    Ast_enQueue_8 T Q ptr tid tcb bp sp head tail prev  mP S
    -> nextS (L_enQueue + 8) (comml (_movwst (Opr_imm 0)(A_reg ND_NEXT BX))) S S'
    -> Ast_enQueue_9 T Q ptr tid tcb bp sp head tail prev mP S').
    unfold Ast_enQueue_8, Ast_enQueue_9.
    intros T Q ptr tid tcb bp sp head tail prev mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [[HSS1 [jk HSS2]] [HSS3 [HSS4 HSS5]]].
    destruct HS as [HM [Htid [Hsp [HBP [HSP [HAX [HDX [HCX HBX]]]]]]]].
    subst M' R' fl' isr'.
    intuition.
    sep on.
    assert (MemUpdW M (tid + 2) 0 jk
   ||- tid |->w tail *^* tid + 2 |->w 0 *^* ptoTCB (tid + 4) tcb *^*
       ptr |->w head *^* ptr + 2 |->w tail *^*
       Queue Thread ptoTCB head tail prev 0 (filterTQ T Q) *^*
       sp - 2 |->w bp *^* sp + 2 |->w tid *^* sp + 4 |->w ptr *^* mP).
    sep lift 1.
    sep upd.
    sep matchH HM with 10.
    unfold MemUpdW in H.
    rewrite HBX; assumption.
  unfold SCAP_WFcmd.
  exists L_enQueue_9_p; exists L_enQueue_9_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [ptr [tid [tcb [bp [sp [head [tail [prev  HS]]]]]]]]]].
  destruct HS as [HM [Htid [Hsp [HBP [HSP [HAX [HDX [HCX HBX]]]]]]]].
  exists (MemUpdW M (tid + 2) 0 0, R, fl, isr).
  simplS in * .  
  apply nextS_movwst with (tid + 2) 0;trivial.
  unfold addr_eval.
  rewrite HBX;auto.
  sep on.
  sep lift 1 in HM.
  sep trivial.
  apply MemUpdW_MemUpd16 with 0; trivial.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [ptr [tid [tcb [bp [sp [head [tail [prev H]]]]]]]]]].
  exists T; exists Q; exists ptr; exists tid; exists tcb; exists bp.
  exists sp; exists head; exists tail; exists prev.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_enQueue_8_g, L_enQueue_9_g in *.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S'' with head tail prev .
  apply Hpp' with S;assumption.
Qed.

(* <EnQueue + 9> : cmpw $0, %dx *)
Lemma SCAP_WFcmd_enQueue_9_cmpw :
  SCAP_WFcmd (SCAP_consCT (L_enQueue + 9 + 1) (Ast L_enQueue_10_p L_enQueue_10_g sctag) psi_enQueue_Block0)
    (L_enQueue + 9) (Ast L_enQueue_9_p L_enQueue_9_g sctag) 
    (comml (_cmpw (Opr_imm 0) DX)).
Proof.
  assert (Hpp': forall T Q ptr tid tcb bp sp head tail prev mP S S',
    Ast_enQueue_9 T Q ptr tid tcb bp sp head tail prev mP S
    -> nextS (L_enQueue + 9)  (comml (_cmpw (Opr_imm 0) DX)) S S'
    -> (Ast_enQueue_10_emp T Q ptr tid tcb bp sp prev mP S' \/ 
      Ast_enQueue_10_notemp T Q ptr tid tcb bp sp head tail prev mP S')).
    unfold Ast_enQueue_9, Ast_enQueue_10_emp, Ast_enQueue_10_notemp.
    intros T Q ptr tid tcb bp sp head tail prev mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [HSS1 [HSS2 [HSS3 HSS4]]].
    destruct HS as [HM [Htid [Hsp [HBP [HSP [HAX [HDX [HCX HBX]]]]]]]].
    destruct HSS3 as [HSS3|HSS3];destruct HSS3 as [HSS31 HSS32].
    subst M' R' fl' isr'.
    left.
    rewrite <- HSS31 in HDX.
    subst tail.
    intuition.
    sep on.
    sep lift 5 in HM.
    assert (Head0 := Queue_tail0_head0_1 HM).
    subst head.
    sep matchH HM with 10.
    right.
    assert (tailn0 : tail > 0).
      rewrite HDX in HSS31.
      omega.
    subst M' R' fl' isr'.
    intuition.
  unfold SCAP_WFcmd.
  exists L_enQueue_10_p; exists L_enQueue_10_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [ptr [tid [tcb [bp [sp [head [tail [prev HS]]]]]]]]]].
  destruct HS as [HM [Htid [Hsp [HBP [HSP [HAX [HDX [HCX HBX]]]]]]]].
  destruct (eq_nat_dec 0 tail).
    exists (M, R,Set_Flag fl _ZF, isr).
    simplS in * .
    apply nextS_cmpw;trivial.
    left.
    intuition.
    simpl.
    rewrite <- e in HDX;auto.
    exists (M, R,Clear_Flag fl _ZF, isr).
    simplS in * .
    apply nextS_cmpw;trivial.
    right.
    intuition.
    simpl in H.
    apply n;rewrite <- H in HDX;auto.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_enQueue_9_p, L_enQueue_10_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [ptr [tid [tcb [bp [sp [head [tail [prev H]]]]]]]]]].
  exists T; exists Q; exists ptr; exists tid; exists tcb.
  exists bp; exists sp; exists head; exists tail; exists prev.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_enQueue_9_g, L_enQueue_10_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S'' with head tail prev.
  apply Hpp' with S;assumption.
Qed.

(* <EnQueue + 10> : jne EnQ_NotNull *)
Lemma SCAP_WFcmd_enQueue_10_jne :
  SCAP_WFcmd (SCAP_consCT (L_enQueue + 10 + 1) (Ast L_enQueue_11_p L_enQueue_11_g sctag) psi_enQueue_Block0)
    (L_enQueue + 10) (Ast L_enQueue_10_p L_enQueue_10_g sctag) 
    (comml (_jne L_enQ_NotNull)).
Proof.
  unfold SCAP_WFcmd.
  exists L_enQueue_16_p; exists L_enQueue_11_p.
  exists L_enQueue_16_g; exists L_enQueue_11_g.
  split.
  apply psi_cons_b_lookup_a.
  unfold psi_enQueue_Block0.
  apply psi_cons_a_lookup_a.
  split_l.
  assert (Hget: forall T Q ptr tid tcb bp sp head tail prev mP S,
    Ast_enQueue_10_notemp T Q ptr tid tcb bp sp head tail prev mP S
    -> Ast_enQueue_16 T Q ptr tid tcb bp sp head tail prev mP S).
    unfold Ast_enQueue_10_notemp, Ast_enQueue_16.
  intros.
  InvertAll.
  intuition.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_enQueue_10_p, L_enQueue_16_p.
  intros S' H.
  unfold conjP in H.
  destruct H as [H1 H2].
  destruct H1 as [T [Q [ptr [tid [tcb [bp [sp [head [tail [prev H]]]]]]]]]].
  destruct H as [P|P].
  destruct P as [HM [Htid [Hsp [HBP [HSP [HAX [HCX HZF]]]]]]].
  rewrite  H2 in HZF.
  inversion HZF.
  exists T; exists Q; exists ptr;exists tid; exists tcb; exists bp.
  exists sp;exists head; exists tail; exists prev.
  apply Hget;assumption.
  split.
  (* g' ==> g *)
  unfold L_enQueue_10_g, L_enQueue_16_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [P1 [P2 P3]].
  intros T Q ptr tid tcb bp sp head tail prev mP H.
  destruct H as [P|P].
  destruct P as [HM [Htid [Hsp [HBP [HSP [HAX [HCX HZF]]]]]]].
  rewrite  P2 in HZF.
  inversion HZF.
  apply P3 with head tail prev.
  apply Hget;assumption.
 assert (Hemp: forall T Q ptr tid tcb bp sp prev mP S,
    Ast_enQueue_10_emp T Q ptr tid tcb bp sp prev mP S
    -> Ast_enQueue_11 T Q ptr tid tcb bp sp prev mP S).
    unfold Ast_enQueue_10_emp, Ast_enQueue_11.
  intros.
  InvertAll.
  intuition.
  split.
  (* p ==> p'' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_enQueue_10_p, L_enQueue_11_p.
  intros S' H.
  unfold conjP in H.
  destruct H as [H1 H2].
  destruct H1 as [T [Q [ptr [tid [tcb [bp [sp [head [tail [prev H]]]]]]]]]].
  destruct H as [P|P].
  exists T; exists Q; exists ptr; exists tid; exists tcb .
  exists bp; exists sp;exists prev.
  apply Hemp;assumption.
  destruct P as [HM [Htid [Hsp [Htail [HBP [HSP [HAX [HDX [HCX [HBX HZF]]]]]]]]]].
  rewrite  H2 in HZF.
  inversion HZF.
  (* g'' ==> g *)
  unfold L_enQueue_10_g, L_enQueue_11_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [P1 [P2 P3]].
  intros T Q ptr tid tcb bp sp head tail prev mP H.
  destruct H as [P|P].
  apply P3 with prev.
  apply Hemp;assumption.
  destruct P as [HM [Htid [Hsp [Htail [HBP [HSP [HAX [HDX [HCX [HBX HZF]]]]]]]]]].
  rewrite  P2 in HZF.
  inversion HZF.
Qed.

(* <EnQueue + 11> : movw %cx, %bx *)
Lemma SCAP_WFcmd_enQueue_11_movwp :
  SCAP_WFcmd (SCAP_consCT (L_enQueue + 11 + 1) (Ast L_enQueue_12_p L_enQueue_12_g sctag) psi_enQueue_Block0)
    (L_enQueue + 11) (Ast L_enQueue_11_p L_enQueue_11_g sctag) 
    (comml (_movwp (Opr_reg CX) BX)).
Proof.
  assert (Hpp': forall T Q ptr tid tcb bp sp prev mP S S',
    Ast_enQueue_11 T Q ptr tid tcb bp sp prev mP S
    -> nextS (L_enQueue + 11) (comml (_movwp (Opr_reg CX) BX)) S S'
    -> Ast_enQueue_12 T Q ptr tid tcb bp sp prev mP S').
    unfold Ast_enQueue_11, Ast_enQueue_12.
    intros T Q ptr tid tcb bp sp prev mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [HSS1 [HSS2 [HSS3 HSS4]]].
    destruct HS as [HM [Htid [Hsp [HBP [HSP [HAX HCX]]]]]].
    subst M' R' fl' isr'.
    rewrite HCX.
    intuition.
  unfold SCAP_WFcmd.
  exists L_enQueue_12_p;exists L_enQueue_12_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [ptr [tid [tcb [bp [sp [prev HS]]]]]]]].
  destruct HS as [HM [Htid [Hsp [HBP [HSP [HAX  HCX]]]]]].
  exists (M, RF_update R BX ptr, fl, isr).
  simplS in * .
  apply nextS_movwp with ptr; trivial.
  unfold opr_eval.
  rewrite HCX;reflexivity.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_enQueue_11_p, L_enQueue_12_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [ptr [tid [tcb [bp [sp [prev H]]]]]]]].
  exists T; exists Q; exists ptr; exists tid; exists tcb; 
    exists bp; exists sp; exists prev.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_enQueue_11_g, L_enQueue_12_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S'' with prev.
  apply Hpp' with S;assumption.
Qed.

(* <EnQueue + 12> : movw %ax, Q_HEAD(%bx) *)
Lemma SCAP_WFcmd_enQueue_12_movwst :
  SCAP_WFcmd (SCAP_consCT (L_enQueue + 12 + 1) (Ast L_enQueue_13_p L_enQueue_13_g sctag) psi_enQueue_Block0)
    (L_enQueue + 12) (Ast L_enQueue_12_p L_enQueue_12_g sctag) 
    (comml (_movwst (Opr_reg AX) (A_reg Q_HEAD BX))).
Proof.
  assert (Hpp': forall T Q ptr tid tcb bp sp prev mP S S',
    Ast_enQueue_12 T Q ptr tid tcb bp sp prev  mP S
    -> nextS (L_enQueue + 12) (comml (_movwst (Opr_reg AX)(A_reg Q_HEAD BX))) S S'
    -> Ast_enQueue_13 T Q ptr tid tcb bp sp prev mP S').
    unfold Ast_enQueue_12, Ast_enQueue_13.
    intros T Q ptr tid tcb bp sp prev mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [[HSS1 [jk HSS2]] [HSS3 [HSS4 HSS5]]].
    destruct HS as [HM [Htid [Hsp [HBP [HSP [HAX HBX]]]]]].
    subst M' R' fl' isr'.
    intuition.
    sep on.
    assert (MemUpdW M ptr tid jk
   ||- tid |->w 0 *^* tid + 2 |->w 0 *^* ptoTCB (tid + 4) tcb *^*
       ptr |->w tid *^* ptr + 2 |->w 0 *^*
       Queue Thread ptoTCB 0 0 prev 0 (filterTQ T Q) *^* 
       sp - 2 |->w bp *^* sp + 2 |->w tid *^* sp + 4 |->w ptr *^* mP).
    sep lift 3.
    sep upd.
    sep matchH HM with 10.
    unfold MemUpdW in H.
    rewrite HBX; rewrite HAX. 
    arith_rep (ptr+0) with ptr.
  unfold SCAP_WFcmd.
  exists L_enQueue_13_p; exists L_enQueue_13_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [ptr [tid [tcb [bp [sp [prev  HS]]]]]]]].
  destruct HS as [HM [Htid [Hsp [HBP [HSP [HAX HBX]]]]]].
  exists (MemUpdW M ptr tid 0, R, fl, isr).
  simplS in * .  
  apply nextS_movwst with ptr tid;trivial.
  unfold addr_eval.
  rewrite HBX;auto.
  unfold opr_eval.
  rewrite HAX;auto.
  sep on.
  sep lift 3 in HM.
  sep trivial.
  apply MemUpdW_MemUpd16 with 0; trivial.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [ptr [tid [tcb [bp [sp [prev H]]]]]]]].
  exists T; exists Q; exists ptr; exists tid; exists tcb; exists bp.
  exists sp; exists prev.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_enQueue_12_g, L_enQueue_13_g in *.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S'' with prev .
  apply Hpp' with S;assumption.
Qed.

(* <EnQueue + 13> : movw %ax, Q_TAIL(%bx) *)
Lemma SCAP_WFcmd_enQueue_13_movwst :
  SCAP_WFcmd (SCAP_consCT (L_enQueue + 13 + 1) (Ast L_enQueue_14_p L_enQueue_14_g sctag) psi_enQueue_Block0)
    (L_enQueue + 13) (Ast L_enQueue_13_p L_enQueue_13_g sctag) 
    (comml (_movwst (Opr_reg AX) (A_reg Q_TAIL BX))).
Proof.
  assert (Hpp': forall T Q ptr tid tcb bp sp prev mP S S',
    Ast_enQueue_13 T Q ptr tid tcb bp sp prev mP S
    -> nextS (L_enQueue + 13) (comml (_movwst (Opr_reg AX)(A_reg Q_TAIL BX))) S S'
    -> Ast_enQueue_14 T Q ptr tid tcb bp sp prev mP S').
    unfold Ast_enQueue_13, Ast_enQueue_14.
    intros T Q ptr tid tcb bp sp prev mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [[HSS1 [jk HSS2]] [HSS3 [HSS4 HSS5]]].
    destruct HS as [HM [Htid [Hsp [HBP [HSP [HAX HBX]]]]]].
    subst M' R' fl' isr'.
    intuition.
    sep on.
    assert (MemUpdW M (ptr + 2) tid jk
   ||- tid |->w 0 *^* tid + 2 |->w 0 *^* ptoTCB (tid + 4) tcb *^*
       ptr |->w tid *^* ptr + 2 |->w tid *^*
       Queue Thread ptoTCB 0 0 prev 0 (filterTQ T Q) *^* 
       sp - 2 |->w bp *^* sp + 2 |->w tid *^* sp + 4 |->w ptr *^* mP).
    sep lift 4.
    sep upd.
    sep matchH HM with 10.
    unfold MemUpdW in H.
    rewrite HBX; rewrite HAX. 
    assumption.
  unfold SCAP_WFcmd.
  exists L_enQueue_14_p; exists L_enQueue_14_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [ptr [tid [tcb [bp [sp [prev  HS]]]]]]]].
  destruct HS as [HM [Htid [Hsp [HBP [HSP [HAX HBX]]]]]].
  exists (MemUpdW M (ptr + 2) tid 0, R, fl, isr).
  simplS in * .  
  apply nextS_movwst with (ptr + 2) tid;trivial.
  unfold addr_eval.
  rewrite HBX;auto.
  unfold opr_eval.
  rewrite HAX;auto.
  sep on.
  sep lift 4 in HM.
  sep trivial.
  apply MemUpdW_MemUpd16 with 0; trivial.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [ptr [tid [tcb [bp [sp [prev H]]]]]]]].
  exists T; exists Q; exists ptr; exists tid; exists tcb; exists bp.
  exists sp; exists prev.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_enQueue_13_g, L_enQueue_14_g in *.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S'' with prev .
  apply Hpp' with S;assumption.
Qed.

(* <DeQueue + 14> : popw %bp *)
Lemma SCAP_WFcmd_enQueue_14_popw :
  SCAP_WFcmd (SCAP_consCT (L_enQueue + 14 + 1) (Ast L_enQueue_15_p L_enQueue_15_g sctag) psi_enQueue_Block0)
    (L_enQueue + 14) (Ast L_enQueue_14_p L_enQueue_14_g sctag) 
    (comml (_popw BP)).
Proof.
  assert (Hpp': forall T Q ptr tid tcb bp sp prev mP S S',
    Ast_enQueue_14 T Q ptr tid tcb bp sp prev mP S
    -> nextS (L_enQueue + 14) (comml (_popw BP)) S S'
    -> Ast_enQueue_15 T Q ptr tid tcb bp sp prev mP S').
    unfold Ast_enQueue_14, Ast_enQueue_15.
    intros T Q ptr tid tcb bp sp prev mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [HSS1 [[x [HSS2 HSS3]] [HSS4 HSS5]]].
    destruct HS as  [HM [Htid [Hsp [HBP HSP]]]].
    subst M' R' fl' isr'.
    assert (E : x = bp ).
      rewrite HSP in HSS2.
      sep on.
      sep lift 6 in HM.
      sep inj;auto.
    subst x.
    rewrite HSP.
    arith_rep (sp-2+2) with sp.
    intuition.
  sep on.
  sep matchH HM with 10.
  assert (h ||- ptoxW (sp-2)).
  sep trivial.
  arith_rep sp with ((sp -2) + 2).
  sep trivial.
  unfold SCAP_WFcmd.
  exists L_enQueue_15_p; exists L_enQueue_15_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [ptr [tid [tcb [bp [sp [prev HS]]]]]]]].
  destruct HS as [HM [Htid [Hsp [HBP HSP]]]].
  exists (M,  RF_update (RF_update R SP sp) BP bp, fl, isr).
  simplS in * .
  apply nextS_popw with (RF_update R SP sp) bp;trivial.
  unfold popW.
  rewrite HSP;intuition.
  sep on.
  sep lift 6 in HM;sep trivial.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_enQueue_14_p, L_enQueue_15_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [ptr [tid [tcb [bp [sp [prev H]]]]]]]].
  exists T; exists Q; exists ptr; exists tid; exists tcb.
  exists bp; exists sp; exists prev.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_enQueue_14_g, L_enQueue_15_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S'' with prev.
  apply Hpp' with S;assumption.
Qed.

(* <EnQueue + 15> : retw *)
Lemma SCAP_WFcmd_enQueue_15_retw : 
  SCAP_WFcmd psi_enQueue_Block0 (L_enQueue+15) 
    (Ast L_enQueue_15_p L_enQueue_15_g sctag) (comms _retw).
Proof.
  unfold L_enQueue_15_g.
  simpl.
  intros S H.
  intros.
  unfold Ast_enQueue_15 in H0.
  InvertAll.
  exists (TQMod.put T tid tcb); exists (addThrd Q tid).
  unfold Ast_enQueue_post.
  intuition.
  sep on.
  sep lift 0.
  apply ptoTQ_intro1 with tid tid 0.
  sep lift 5 in H1.
  destruct (Queue_ptr0_emp_1 H1).
  sep lift 2.
  assert ((filterTQ (TQMod.put T tid tcb) (addThrd Q tid)) = 
 (fun l' => if eq_nat_dec tid l' then Some tcb else None)).
  apply TQMod.extensional_eqMap .
  intro x.
  assert (filterTQ T Q x = None).
  rewrite H6;reflexivity;clear H6.
  unfold filterTQ in * .
  unfold addThrd.
  unfold WQMod.put, TQMod.put, WQMod.beq_A, TQMod.beq_A, WQSpec.beq_Domain, TQSpec.beq_Domain.
  destruct (eq_nat_dec x tid); destruct (eq_nat_dec tid x); auto;(subst tid;elim n;auto).
  rewrite H7.
  apply (Queue_cons_fst_intro_1 Thread ptoTCB tid 0 0 tcb (S_Mem S)).
  unfold ptoNode.
  sep matchH H4 with 10.
  assumption.
Qed.

(* <EnQueue + 16> : movw %dx, %bx *)
Lemma SCAP_WFcmd_enQueue_16_movwp :
  SCAP_WFcmd (SCAP_consCT (L_enQueue + 16 + 1) (Ast L_enQueue_17_p L_enQueue_17_g sctag) psi_enQueue_Block1)
    (L_enQueue + 16) (Ast L_enQueue_16_p L_enQueue_16_g sctag) 
    (comml (_movwp (Opr_reg DX) BX)).
Proof.
  assert (Hpp': forall T Q ptr tid tcb bp sp head tail prev mP S S',
    Ast_enQueue_16 T Q ptr tid tcb bp sp head tail prev mP S
    -> nextS (L_enQueue + 16) (comml (_movwp (Opr_reg DX) BX)) S S'
    -> Ast_enQueue_17 T Q ptr tid tcb bp sp head tail prev mP S').
    unfold Ast_enQueue_16,Ast_enQueue_17.
    intros T Q ptr tid tcb bp sp head tail prev mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [HSS1 [HSS2 [HSS3 HSS4]]].
    destruct HS as [HM [Htid [Hsp [Htail [HBP [HSP [HAX [HDX HCX]]]]]]]].
    subst M' R' fl' isr'.
    rewrite HDX.
    intuition.
  unfold SCAP_WFcmd.
  exists L_enQueue_17_p;exists L_enQueue_17_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [ptr [tid [tcb [bp [sp [head [tail [prev HS]]]]]]]]]].
  destruct HS as [HM [Htid [Hsp [Htail [HBP [HSP [HAX [HDX HCX]]]]]]]].
  exists (M, RF_update R BX tail, fl, isr).
  simplS in * .
  apply nextS_movwp with tail; trivial.
  unfold opr_eval.
  rewrite HDX;reflexivity.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_enQueue_16_p, L_enQueue_17_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [ptr [tid [tcb [bp [sp [head [tail [prev H]]]]]]]]]].
  exists T; exists Q; exists ptr; exists tid; exists tcb; 
    exists bp; exists sp; exists head; exists tail; exists prev.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_enQueue_16_g, L_enQueue_17_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S'' with head tail prev.
  apply Hpp' with S;assumption.
Qed.

(* <EnQueue + 17> : movw %ax, ND_NEXT(%bx) *)
Lemma SCAP_WFcmd_enQueue_17_movwst :
  SCAP_WFcmd (SCAP_consCT (L_enQueue + 17 + 1) (Ast L_enQueue_18_p L_enQueue_18_g sctag) psi_enQueue_Block1)
    (L_enQueue + 17) (Ast L_enQueue_17_p L_enQueue_17_g sctag) 
    (comml (_movwst (Opr_reg AX) (A_reg ND_NEXT BX))).
Proof.
  assert (Hpp': forall T Q ptr tid tcb bp sp head tail prev mP S S',
    Ast_enQueue_17 T Q ptr tid tcb bp sp head tail prev  mP S
    -> nextS (L_enQueue + 17) (comml (_movwst (Opr_reg AX)(A_reg ND_NEXT BX))) S S'
    -> Ast_enQueue_18 T Q ptr tid tcb bp sp head tail prev mP S').
    unfold Ast_enQueue_17, Ast_enQueue_18.
    intros T Q ptr tid tcb bp sp head tail prev mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [[HSS1 [jk HSS2]] [HSS3 [HSS4 HSS5]]].
    destruct HS as [HM [Htid [Hsp [Htail [HBP [HSP [HAX [HCX HBX]]]]]]]].
    subst M' R' fl' isr'.
    intuition.
    sep on.
    sep lift 5 in HM.
    rewrite HAX; rewrite HBX.
    substH HM with (Queue_upd_next_1 tid jk HM Htail ).
    sep matchH HM with 10.
  unfold SCAP_WFcmd.
  exists L_enQueue_18_p; exists L_enQueue_18_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [ptr [tid [tcb [bp [sp [head [tail [prev  HS]]]]]]]]]].
  destruct HS as [HM [Htid [Hsp [Htail [HBP [HSP [HAX [HCX HBX]]]]]]]].
  exists (MemUpdW M (tail + 2) tid 0, R, fl, isr).
  simplS in * .  
  apply nextS_movwst with (tail+2) tid;trivial.
  unfold addr_eval.
  rewrite HBX;auto.
  simpl;auto.
  sep on.
  sep lift 5 in HM.
  substH HM with (Queue_unemp_split_1 HM Htail).
  destruct HM as [Q' [tcb' [prev' [head' [tail' [HM P]]]]]].
  unfold ptoNode in HM.
  sep lift 1 in HM.
  sep trivial.
  apply MemUpdW_MemUpd16 with 0; trivial.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_enQueue_17_p, L_enQueue_18_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [ptr [tid [tcb [bp [sp [head [tail [prev H]]]]]]]]]].
  exists T; exists Q; exists ptr; exists tid; exists tcb; exists bp.
  exists sp; exists head; exists tail; exists prev.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_enQueue_17_g, L_enQueue_18_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S'' with head tail prev .
  apply Hpp' with S;assumption.
Qed.

(* <EnQueue + 18> : movw %cx, %bx *)
Lemma SCAP_WFcmd_enQueue_18_movwp :
  SCAP_WFcmd (SCAP_consCT (L_enQueue + 18 + 1) (Ast L_enQueue_19_p L_enQueue_19_g sctag) psi_enQueue_Block1)
    (L_enQueue + 18) (Ast L_enQueue_18_p L_enQueue_18_g sctag) 
    (comml (_movwp (Opr_reg CX) BX)).
Proof.
  assert (Hpp': forall T Q ptr tid tcb bp sp head tail prev mP S S',
    Ast_enQueue_18 T Q ptr tid tcb bp sp head tail prev mP S
    -> nextS (L_enQueue + 18) (comml (_movwp (Opr_reg CX) BX)) S S'
    -> Ast_enQueue_19 T Q ptr tid tcb bp sp head tail prev mP S').
    unfold Ast_enQueue_18,Ast_enQueue_19.
    intros T Q ptr tid tcb bp sp head tail prev mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [HSS1 [HSS2 [HSS3 HSS4]]].
    destruct HS as [HM [Htid [Hsp [Htail [HBP [HSP [HAX HCX]]]]]]].
    subst M' R' fl' isr'.
    rewrite HCX.
    intuition.
  unfold SCAP_WFcmd.
  exists L_enQueue_19_p;exists L_enQueue_19_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [ptr [tid [tcb [bp [sp [head [tail [prev HS]]]]]]]]]].
  destruct HS as [HM [Htid [Hsp [Htail [HBP [HSP [HAX HCX]]]]]]].
  exists (M, RF_update R BX ptr, fl, isr).
  simplS in * .
  apply nextS_movwp with ptr; trivial.
  unfold opr_eval.
  rewrite HCX;reflexivity.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_enQueue_18_p, L_enQueue_19_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [ptr [tid [tcb [bp [sp [head [tail [prev H]]]]]]]]]].
  exists T; exists Q; exists ptr; exists tid; exists tcb; 
    exists bp; exists sp; exists head; exists tail; exists prev.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_enQueue_18_g, L_enQueue_19_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S'' with head tail prev.
  apply Hpp' with S;assumption.
Qed.

(* <EnQueue + 19> : movw %ax, Q_TAIL(%bx) *)
Lemma SCAP_WFcmd_enQueue_19_movwst :
  SCAP_WFcmd (SCAP_consCT (L_enQueue + 19 + 1) (Ast L_enQueue_20_p L_enQueue_20_g sctag) psi_enQueue_Block1)
    (L_enQueue + 19) (Ast L_enQueue_19_p L_enQueue_19_g sctag) 
    (comml (_movwst (Opr_reg AX) (A_reg Q_TAIL BX))).
Proof.
  assert (Hpp': forall T Q ptr tid tcb bp sp head tail prev mP S S',
    Ast_enQueue_19 T Q ptr tid tcb bp sp head tail prev  mP S
    -> nextS (L_enQueue + 19) (comml (_movwst (Opr_reg AX)(A_reg Q_TAIL BX))) S S'
    -> Ast_enQueue_20 T Q ptr tid tcb bp sp head tail prev mP S').
    unfold Ast_enQueue_19, Ast_enQueue_20.
    intros T Q ptr tid tcb bp sp head tail prev mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [[HSS1 [jk HSS2]] [HSS3 [HSS4 HSS5]]].
    destruct HS as [HM [Htid [Hsp [Htail [HBP [HSP [HAX HBX]]]]]]].
    subst M' R' fl' isr'.
    intuition.
    sep on.
    assert (MemUpdW M (ptr + 2) tid jk
   ||- tid |->w tail *^* tid + 2 |->w 0 *^* ptoTCB (tid + 4) tcb *^*
       ptr |->w head *^* ptr + 2 |->w tid *^*
       Queue Thread ptoTCB head tail prev tid (filterTQ T Q) *^*
       sp - 2 |->w bp *^* sp + 2 |->w tid *^* sp + 4 |->w ptr *^* mP).
    sep lift 4.
    sep upd.
    sep matchH HM with 10.
    rewrite HAX; rewrite HBX.
    unfold MemUpdW in H.
    assumption.
  unfold SCAP_WFcmd.
  exists L_enQueue_20_p; exists L_enQueue_20_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [ptr [tid [tcb [bp [sp [head [tail [prev  HS]]]]]]]]]].
  destruct HS as [HM [Htid [Hsp [Htail [HBP [HSP [HAX HBX]]]]]]].
  exists (MemUpdW M (ptr + 2) tid 0, R, fl, isr).
  simplS in * .  
  apply nextS_movwst with (ptr+2) tid;trivial.
  unfold addr_eval.
  rewrite HBX;auto.
  simpl;auto.
  sep on.
  sep lift 4 in HM.
  sep trivial.
  apply MemUpdW_MemUpd16 with 0; trivial.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_enQueue_19_p, L_enQueue_20_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [ptr [tid [tcb [bp [sp [head [tail [prev H]]]]]]]]]].
  exists T; exists Q; exists ptr; exists tid; exists tcb; exists bp.
  exists sp; exists head; exists tail; exists prev.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_enQueue_19_g, L_enQueue_20_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S'' with head tail prev .
  apply Hpp' with S;assumption.
Qed.

(* <DeQueue + 20> : popw %bp *)
Lemma SCAP_WFcmd_enQueue_20_popw :
  SCAP_WFcmd (SCAP_consCT (L_enQueue + 20 + 1) (Ast L_enQueue_21_p L_enQueue_21_g sctag) psi_enQueue_Block1)
    (L_enQueue + 20) (Ast L_enQueue_20_p L_enQueue_20_g sctag) 
    (comml (_popw BP)).
Proof.
  assert (Hpp': forall T Q ptr tid tcb bp sp head tail prev mP S S',
    Ast_enQueue_20 T Q ptr tid tcb bp sp head tail prev mP S
    -> nextS (L_enQueue + 20) (comml (_popw BP)) S S'
    -> Ast_enQueue_21 T Q ptr tid tcb bp sp head tail prev mP S').
    unfold Ast_enQueue_20, Ast_enQueue_21.
    intros T Q ptr tid tcb bp sp head tail prev mP S S' HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [HSS1 [[x [HSS2 HSS3]] [HSS4 HSS5]]].
    destruct HS as  [HM [Htid [Hsp [Htail [HBP HSP]]]]].
    subst M' R' fl' isr'.
    assert (E : x = bp ).
      rewrite HSP in HSS2.
      sep on.
      sep lift 6 in HM.
      sep inj;auto.
    subst x.
    rewrite HSP.
    arith_rep (sp-2+2) with sp.
    intuition.
  sep on.
  sep matchH HM with 10.
  assert (h ||- ptoxW (sp-2)).
  sep trivial.
  arith_rep sp with ((sp -2) + 2).
  sep trivial.
  unfold SCAP_WFcmd.
  exists L_enQueue_21_p; exists L_enQueue_21_g.
  split_l.
  split.
  (* enable *)
  unfold enable.
  intros S PS.
  destruct S as [[[M R] fl] isr].
  destruct PS as [T [Q [ptr [tid [tcb [bp [sp [head [tail [prev HS]]]]]]]]]].
  destruct HS as [HM [Htid [Hsp [Htail [HBP HSP]]]]].
  exists (M,  RF_update (RF_update R SP sp) BP bp, fl, isr).
  simplS in * .
  apply nextS_popw with (RF_update R SP sp) bp;trivial.
  unfold popW.
  rewrite HSP;intuition.
  sep on.
  sep lift 6 in HM;sep trivial.
  split.
  (* p ==> p' *)
  unfold pimp.
  unfold linkPG, conjPG, conjG.
  unfold L_enQueue_20_p, L_enQueue_21_p.
  intros S' H.
  destruct H as [S [H1 H2]].
  destruct H1 as [T [Q [ptr [tid [tcb [bp [sp [head [tail [prev H]]]]]]]]]].
  exists T; exists Q; exists ptr; exists tid; exists tcb.
  exists bp; exists sp; exists head; exists tail; exists prev.
  apply Hpp' with S; assumption.
  (* g' ==> g *)
  unfold L_enQueue_20_g, L_enQueue_21_g.
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[Hps HnextSS'] HgS'S''].
  intros.
  apply HgS'S'' with head tail prev.
  apply Hpp' with S;assumption.
Qed.

(* <EnQueue + 21> : retw *)
Lemma SCAP_WFcmd_enQueue_21_retw : 
  SCAP_WFcmd psi_enQueue_Block1 (L_enQueue+21) 
    (Ast L_enQueue_21_p L_enQueue_21_g sctag) (comms _retw).
Proof.
  unfold L_enQueue_21_g.
  simpl.
  intros S H.
  intros.
  unfold Ast_enQueue_21 in H0.
  InvertAll.
  exists (TQMod.put T tid tcb); exists (addThrd Q tid).
  unfold Ast_enQueue_post.
  intuition.
  sep on.
  sep lift 0.
  apply ptoTQ_intro1 with head tid prev.
  sep lift 2.
  assert ((filterTQ (TQMod.put T tid tcb) (addThrd Q tid)) = 
    (fun l' => if eq_nat_dec tid l' then Some tcb else (filterTQ T Q) l')).
  apply TQMod.extensional_eqMap.
  intro x.
  unfold filterTQ, addThrd.
  unfold WQMod.put, TQMod.put.
  unfold WQMod.beq_A, TQMod.beq_A, WQSpec.beq_Domain, TQSpec.beq_Domain.
  destruct (eq_nat_dec x tid); destruct (eq_nat_dec tid x); try reflexivity;
    rewrite e in n; elim n; reflexivity.
  rewrite H5.
  apply (Queue_cons_more_intro_1 Thread ptoTCB head tail prev (filterTQ T Q) (S_Mem S) tid 0 tcb).
  unfold ptoNode.
  sep matchH H1 with 10.
  assumption.
  assumption.
Qed.

Lemma SCAP_WFiseq_enQueue_Block0 :
  SCAP_WFiseq psi_enQueue_Block0 L_enQueue spec_enQueue I_enQueue.
Proof.
unfold I_enQueue, spec_enQueue.
apply scap_isq_seq with L_enQueue_1_p L_enQueue_1_g.
apply SCAP_WFcmd_enQueue_0_pushw.
apply scap_isq_seq with L_enQueue_2_p L_enQueue_2_g.
apply SCAP_WFcmd_enQueue_1_movwp.
assert (H : L_enQueue+1+1 = L_enQueue+2); [omega | rewrite H; clear H].
apply scap_isq_seq with L_enQueue_3_p L_enQueue_3_g.
apply SCAP_WFcmd_enQueue_2_movwld.
assert (H : L_enQueue+2+1 = L_enQueue+3); [omega | rewrite H; clear H].
apply scap_isq_seq with L_enQueue_4_p L_enQueue_4_g.
apply SCAP_WFcmd_enQueue_3_movwld.
assert (H : L_enQueue+3+1 = L_enQueue+4); [omega | rewrite H; clear H].
apply scap_isq_seq with L_enQueue_5_p L_enQueue_5_g.
apply SCAP_WFcmd_enQueue_4_movwld.
assert (H : L_enQueue+4+1 = L_enQueue+5); [omega | rewrite H; clear H].
apply scap_isq_seq with L_enQueue_6_p L_enQueue_6_g.
apply SCAP_WFcmd_enQueue_5_movwp.
assert (H : L_enQueue+5+1 = L_enQueue+6); [omega | rewrite H; clear H].
apply scap_isq_seq with L_enQueue_7_p L_enQueue_7_g.
apply SCAP_WFcmd_enQueue_6_movwp.
assert (H : L_enQueue+6+1 = L_enQueue+7); [omega | rewrite H; clear H].
apply scap_isq_seq with L_enQueue_8_p L_enQueue_8_g.
apply SCAP_WFcmd_enQueue_7_movwst.
assert (H : L_enQueue+7+1 = L_enQueue+8); [omega | rewrite H; clear H].
apply scap_isq_seq with L_enQueue_9_p L_enQueue_9_g.
apply SCAP_WFcmd_enQueue_8_movwst.
assert (H : L_enQueue+8+1 = L_enQueue+9); [omega | rewrite H; clear H].
apply scap_isq_seq with L_enQueue_10_p L_enQueue_10_g.
apply SCAP_WFcmd_enQueue_9_cmpw.
assert (H : L_enQueue+9+1 = L_enQueue+10); [omega | rewrite H; clear H].
apply scap_isq_seq with L_enQueue_11_p L_enQueue_11_g.
apply SCAP_WFcmd_enQueue_10_jne.
assert (H : L_enQueue+10+1 = L_enQueue+11); [omega | rewrite H; clear H].
apply scap_isq_seq with L_enQueue_12_p L_enQueue_12_g.
apply SCAP_WFcmd_enQueue_11_movwp.
assert (H : L_enQueue+11+1 = L_enQueue+12); [omega | rewrite H; clear H].
apply scap_isq_seq with L_enQueue_13_p L_enQueue_13_g.
apply SCAP_WFcmd_enQueue_12_movwst.
assert (H : L_enQueue+12+1 = L_enQueue+13); [omega | rewrite H; clear H].
apply scap_isq_seq with L_enQueue_14_p L_enQueue_14_g.
apply SCAP_WFcmd_enQueue_13_movwst.
assert (H : L_enQueue+13+1 = L_enQueue+14); [omega | rewrite H; clear H].
apply scap_isq_seq with L_enQueue_15_p L_enQueue_15_g.
apply SCAP_WFcmd_enQueue_14_popw.
assert (H : L_enQueue+14+1 = L_enQueue+15); [omega | rewrite H; clear H].
apply scap_isq_end.
apply SCAP_WFcmd_enQueue_15_retw.
Qed.

Lemma SCAP_WFiseq_enQueue_Block1 :
  SCAP_WFiseq psi_enQueue_Block1 L_enQ_NotNull (Ast L_enQueue_16_p L_enQueue_16_g sctag) I_enQ_NotNull.
Proof.
unfold I_enQ_NotNull, L_enQ_NotNull.
apply scap_isq_seq with L_enQueue_17_p L_enQueue_17_g.
apply SCAP_WFcmd_enQueue_16_movwp.
apply scap_isq_seq with L_enQueue_18_p L_enQueue_18_g.
arith_rep (L_enQueue + 16 + 1) with (L_enQueue + 17).
apply SCAP_WFcmd_enQueue_17_movwst.
apply scap_isq_seq with L_enQueue_19_p L_enQueue_19_g.
arith_rep (L_enQueue + 16 + 1 + 1) with (L_enQueue + 18).
apply SCAP_WFcmd_enQueue_18_movwp.
apply scap_isq_seq with L_enQueue_20_p L_enQueue_20_g.
arith_rep (L_enQueue + 16 + 1 + 1 + 1) with (L_enQueue + 19).
apply SCAP_WFcmd_enQueue_19_movwst.
apply scap_isq_seq with L_enQueue_21_p L_enQueue_21_g.
arith_rep (L_enQueue + 16 + 1 + 1+ 1 + 1) with (L_enQueue + 20).
apply SCAP_WFcmd_enQueue_20_popw.
arith_rep (L_enQueue + 16 + 1 + 1+ 1 + 1) with (L_enQueue + 20).
apply scap_isq_end.
apply SCAP_WFcmd_enQueue_21_retw.
Qed.
