(***************************************************************)
(* Coq code for AIM                                            *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(* by wangwei *)

Add LoadPath "..".
Add LoadPath "../CAP0".
Add LoadPath "../SCAP".

Require Import Coqlib.

Require Import Arith.
Require Import Machine.
Require Import Assertion.
Require Import Operational_Semantics.

Require Import Kernel.

Require Import AIM.
Require Import SCAP.
Require Import QLib.
Require Import Linking.

Definition Ast_enQueue_pre (T : TQ) (Q : WQ)
  (ptr : Label) (tid : TID) (tcb : Thread)
  (bp sp : nat) (mP : MPred) (S : State) : Prop :=
  (embed_TCB tid tcb
    *^* ptoTQ ptr (filterTQ T Q)
    *^* stkSpacePrec sp 2 
    *^* sp + 2 |->w tid
    *^* sp + 4 |->w ptr
    *^* mP) (S_Mem S)
  /\ tid > 0
  /\ S_RF S BP = bp
  /\ S_RF S SP = sp.

Definition p_enQueue (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists tid, exists tcb,
    exists bp, exists sp,
      Ast_enQueue_pre T Q ptr tid tcb bp sp mTrue S.

Definition Ast_enQueue_post (T' : TQ) (Q' : WQ)
  (ptr : Label) (tid : TID) (tcb : Thread)
  (bp sp : nat) (mP : MPred) (S' : State) : Prop :=
  (ptoTQ ptr (filterTQ T' Q')
    *^* stkSpacePrec sp 2 
    *^* sp + 2 |->w tid
    *^* sp + 4 |->w ptr
    *^* mP) (S_Mem S')
  /\ S_RF S' BP = bp
  /\ S_RF S' SP = sp.

Definition g_enQueue (S S' : State) : Prop :=
  forall T Q ptr tid tcb bp sp mP,
    Ast_enQueue_pre T Q ptr tid tcb bp sp mP S
    -> exists T', exists Q',
      Ast_enQueue_post T' Q' ptr tid tcb bp sp mP S'
      /\ T' = TQMod.put T tid tcb
      /\ Q' = addThrd Q tid.

(* <enQueue> : pushw %bp *)
Definition Ast_enQueue_1  (T : TQ) (Q : WQ)
  (ptr : Label) (tid : TID) (tcb : Thread)
  (bp sp : nat) (mP : MPred) (S : State) : Prop :=
  (embed_TCB tid tcb
    *^* ptoTQ ptr (filterTQ T Q)
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w tid
    *^* sp + 4 |->w ptr
    *^* mP) (S_Mem S)
  /\ tid > 0
  /\ sp >=2 
  /\ S_RF S BP = bp
  /\ S_RF S SP = sp - 2.

Definition L_enQueue_1_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists tid, exists tcb,
    exists bp, exists sp,  
    Ast_enQueue_1 T Q ptr tid tcb bp sp mTrue S.

Definition L_enQueue_1_g (S S' : State) : Prop :=
  forall T Q ptr tid tcb bp sp mP,
    Ast_enQueue_1 T Q ptr tid tcb bp sp mP S 
    -> exists T', exists Q',
      Ast_enQueue_post T' Q' ptr tid tcb bp sp mP S'
      /\ T' = TQMod.put T tid tcb
      /\ Q' = addThrd Q tid.

(* <enQueue> : movw %sp, %bp *)
Definition Ast_enQueue_2  (T : TQ) (Q : WQ)
  (ptr : Label) (tid : TID) (tcb : Thread)
  (bp sp : nat) (mP : MPred) (S : State) : Prop :=
  (embed_TCB tid tcb
    *^* ptoTQ ptr (filterTQ T Q)
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w tid
    *^* sp + 4 |->w ptr
    *^* mP) (S_Mem S)
  /\ tid > 0
  /\ sp >= 2
  /\ S_RF S BP = sp - 2
  /\ S_RF S SP = sp - 2 .

Definition L_enQueue_2_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists tid, exists tcb,
    exists bp, exists sp,  
    Ast_enQueue_2 T Q ptr tid tcb bp sp mTrue S.

Definition L_enQueue_2_g (S S' : State) : Prop :=
  forall T Q ptr tid tcb bp sp mP,
    Ast_enQueue_2 T Q ptr tid tcb bp sp mP S 
    -> exists T', exists Q',
      Ast_enQueue_post T' Q' ptr tid tcb bp sp mP S'
      /\ T' = TQMod.put T tid tcb
      /\ Q' = addThrd Q tid.

(* <enQueue> : movw    4(%bp), %ax *)
Definition Ast_enQueue_3  (T : TQ) (Q : WQ)
  (ptr : Label) (tid : TID) (tcb : Thread)
  (bp sp : nat) (mP : MPred) (S : State) : Prop :=
  (embed_TCB tid tcb
    *^* ptoTQ ptr (filterTQ T Q)
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w tid
    *^* sp + 4 |->w ptr
    *^* mP) (S_Mem S)
  /\ tid > 0
  /\ sp >= 2 
  /\ S_RF S BP = sp - 2
  /\ S_RF S SP = sp - 2 
  /\ S_RF S AX = tid.

Definition L_enQueue_3_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists tid, exists tcb,
    exists bp, exists sp,  
    Ast_enQueue_3 T Q ptr tid tcb bp sp mTrue S.

Definition L_enQueue_3_g (S S' : State) : Prop :=
  forall T Q ptr tid tcb bp sp mP,
    Ast_enQueue_3 T Q ptr tid tcb bp sp mP S 
    -> exists T', exists Q',
      Ast_enQueue_post T' Q' ptr tid tcb bp sp mP S'
      /\ T' = TQMod.put T tid tcb
      /\ Q' = addThrd Q tid.

(* <enQueue> : movw    6(%bp), %bx *)
Definition Ast_enQueue_4  (T : TQ) (Q : WQ)
  (ptr : Label) (tid : TID) (tcb : Thread)
  (bp sp head tail prev : nat) (mP : MPred) (S : State) : Prop :=
  (embed_TCB tid tcb
    *^* ptr |->w head
    *^* ptr + 2 |->w tail
    *^* (Queue Thread ptoTCB head tail prev 0 (filterTQ T Q))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w tid
    *^* sp + 4 |->w ptr
    *^* mP) (S_Mem S)
  /\ tid > 0
  /\ sp >= 2 
  /\ S_RF S BP = sp - 2
  /\ S_RF S SP = sp - 2 
  /\ S_RF S AX = tid
  /\ S_RF S BX = ptr.

Definition L_enQueue_4_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists tid, exists tcb,
    exists bp, exists sp,   
      exists head, exists tail, exists prev,
    Ast_enQueue_4 T Q ptr tid tcb bp sp head tail prev mTrue S.

Definition L_enQueue_4_g (S S' : State) : Prop :=
  forall T Q ptr tid tcb bp sp head tail prev mP,
    Ast_enQueue_4 T Q ptr tid tcb bp sp head tail prev mP S 
    -> exists T', exists Q',
      Ast_enQueue_post T' Q' ptr tid tcb bp sp mP S'
      /\ T' = TQMod.put T tid tcb
      /\ Q' = addThrd Q tid.

(* <enQueue> :  movw    Q_TAIL(%bx), %dx *)
Definition Ast_enQueue_5  (T : TQ) (Q : WQ)
  (ptr : Label) (tid : TID) (tcb : Thread)
  (bp sp head tail prev : nat) (mP : MPred) (S : State) : Prop :=
  (embed_TCB tid tcb
    *^* ptr |->w head
    *^* ptr + 2 |->w tail
    *^* (Queue Thread ptoTCB head tail prev 0 (filterTQ T Q))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w tid
    *^* sp + 4 |->w ptr
    *^* mP) (S_Mem S)
  /\ tid > 0
  /\ sp >= 2 
  /\ S_RF S BP = sp - 2
  /\ S_RF S SP = sp - 2 
  /\ S_RF S AX = tid
  /\ S_RF S BX = ptr
  /\ S_RF S DX = tail.

Definition L_enQueue_5_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists tid, exists tcb,
    exists bp, exists sp,   
      exists head, exists tail, exists prev,
    Ast_enQueue_5 T Q ptr tid tcb bp sp head tail prev mTrue S.

Definition L_enQueue_5_g (S S' : State) : Prop :=
  forall T Q ptr tid tcb bp sp head tail prev mP,
    Ast_enQueue_5 T Q ptr tid tcb bp sp head tail prev mP S 
    -> exists T', exists Q',
      Ast_enQueue_post T' Q' ptr tid tcb bp sp mP S'
      /\ T' = TQMod.put T tid tcb
      /\ Q' = addThrd Q tid.

(* <enQueue> :   movw    %bx, %cx *)
Definition Ast_enQueue_6  (T : TQ) (Q : WQ)
  (ptr : Label) (tid : TID) (tcb : Thread)
  (bp sp head tail prev : nat) (mP : MPred) (S : State) : Prop :=
  (embed_TCB tid tcb
    *^* ptr |->w head
    *^* ptr + 2 |->w tail
    *^* (Queue Thread ptoTCB head tail prev 0 (filterTQ T Q))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w tid
    *^* sp + 4 |->w ptr
    *^* mP) (S_Mem S)
  /\ tid > 0
  /\ sp >= 2 
  /\ S_RF S BP = sp - 2
  /\ S_RF S SP = sp - 2 
  /\ S_RF S AX = tid
  /\ S_RF S DX = tail
  /\ S_RF S CX = ptr.

Definition L_enQueue_6_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists tid, exists tcb,
    exists bp, exists sp,   
      exists head, exists tail, exists prev,
    Ast_enQueue_6 T Q ptr tid tcb bp sp head tail prev mTrue S.

Definition L_enQueue_6_g (S S' : State) : Prop :=
  forall T Q ptr tid tcb bp sp head tail prev mP,
    Ast_enQueue_6 T Q ptr tid tcb bp sp head tail prev mP S 
    -> exists T', exists Q',
      Ast_enQueue_post T' Q' ptr tid tcb bp sp mP S'
      /\ T' = TQMod.put T tid tcb
      /\ Q' = addThrd Q tid.

(* <enQueue> :   movw    %ax, %bx *)
Definition Ast_enQueue_7  (T : TQ) (Q : WQ)
  (ptr : Label) (tid : TID) (tcb : Thread)
  (bp sp head tail prev: nat) (mP : MPred) (S : State) : Prop :=
  (tid |->w_
    *^* tid+2 |->w_
    *^* ptoTCB (tid+4) tcb
    *^* ptr |->w head
    *^* ptr + 2 |->w tail
    *^* (Queue Thread ptoTCB head tail prev 0 (filterTQ T Q))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w tid
    *^* sp + 4 |->w ptr
    *^* mP) (S_Mem S)
  /\ tid > 0
  /\ sp >= 2 
  /\ S_RF S BP = sp - 2
  /\ S_RF S SP = sp - 2 
  /\ S_RF S AX = tid
  /\ S_RF S DX = tail
  /\ S_RF S CX = ptr
  /\ S_RF S BX = tid.

Definition L_enQueue_7_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists tid, exists tcb,
    exists bp, exists sp,    
      exists head, exists tail, exists prev, 
    Ast_enQueue_7 T Q ptr tid tcb bp sp head tail prev mTrue S.

Definition L_enQueue_7_g (S S' : State) : Prop :=
  forall T Q ptr tid tcb bp sp head tail prev mP,
    Ast_enQueue_7 T Q ptr tid tcb bp sp head tail prev mP S 
    -> exists T', exists Q',
      Ast_enQueue_post T' Q' ptr tid tcb bp sp mP S'
      /\ T' = TQMod.put T tid tcb
      /\ Q' = addThrd Q tid.

(* <enQueue> :   movw    %dx, ND_PREV(%bx) *)
Definition Ast_enQueue_8  (T : TQ) (Q : WQ)
  (ptr : Label) (tid : TID) (tcb : Thread)
  (bp sp head tail prev: nat) (mP : MPred) (S : State) : Prop :=
  (tid |->w tail 
    *^* tid + 2 |->w_ 
    *^* ptoTCB (tid+4) tcb
    *^* ptr |->w head
    *^* ptr + 2 |->w tail
    *^* (Queue Thread ptoTCB head tail prev 0 (filterTQ T Q))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w tid
    *^* sp + 4 |->w ptr
    *^* mP) (S_Mem S)
  /\ tid > 0
  /\ sp >= 2 
  /\ S_RF S BP = sp - 2
  /\ S_RF S SP = sp - 2 
  /\ S_RF S AX = tid
  /\ S_RF S DX = tail
  /\ S_RF S CX = ptr
  /\ S_RF S BX = tid.

Definition L_enQueue_8_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists tid, exists tcb,
    exists bp, exists sp,    
      exists head, exists tail, exists prev,
    Ast_enQueue_8 T Q ptr tid tcb bp sp head tail prev mTrue S.

Definition L_enQueue_8_g (S S' : State) : Prop :=
  forall T Q ptr tid tcb bp sp head tail prev mP,
    Ast_enQueue_8 T Q ptr tid tcb bp sp head tail prev mP S 
    -> exists T', exists Q',
      Ast_enQueue_post T' Q' ptr tid tcb bp sp mP S'
      /\ T' = TQMod.put T tid tcb
      /\ Q' = addThrd Q tid.

(* <enQueue> :   movw    $0, ND_NEXT(%bx) *)
Definition Ast_enQueue_9  (T : TQ) (Q : WQ)
  (ptr : Label) (tid : TID) (tcb : Thread)
  (bp sp head tail prev : nat) (mP : MPred) (S : State) : Prop :=
  (tid |->w tail 
    *^* tid + 2 |->w 0 
    *^* ptoTCB (tid+4) tcb
    *^* ptr |->w head
    *^* ptr + 2 |->w tail
    *^* (Queue Thread ptoTCB head tail prev 0 (filterTQ T Q))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w tid
    *^* sp + 4 |->w ptr
    *^* mP) (S_Mem S)
  /\ tid > 0
  /\ sp >= 2 
  /\ S_RF S BP = sp - 2
  /\ S_RF S SP = sp - 2 
  /\ S_RF S AX = tid
  /\ S_RF S DX = tail
  /\ S_RF S CX = ptr
  /\ S_RF S BX = tid.

Definition L_enQueue_9_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists tid, exists tcb,
    exists bp, exists sp,    
      exists head, exists tail, exists prev,
    Ast_enQueue_9 T Q ptr tid tcb bp sp head tail prev mTrue S.

Definition L_enQueue_9_g (S S' : State) : Prop :=
  forall T Q ptr tid tcb bp sp head tail prev mP,
    Ast_enQueue_9 T Q ptr tid tcb bp sp head tail prev mP S 
    -> exists T', exists Q',
      Ast_enQueue_post T' Q' ptr tid tcb bp sp mP S'
      /\ T' = TQMod.put T tid tcb
      /\ Q' = addThrd Q tid.

(* <enQueue> :  cmpw    $0, %dx *)
Definition Ast_enQueue_10_emp  (T : TQ) (Q : WQ)
  (ptr : Label) (tid : TID) (tcb : Thread)
  (bp sp prev : nat) (mP : MPred) (S : State) : Prop :=
  (tid |->w 0 
    *^* tid + 2 |->w 0 
    *^* ptoTCB (tid+4) tcb
    *^* ptr |->w 0
    *^* ptr + 2 |->w 0
    *^* (Queue Thread ptoTCB 0 0 prev 0 (filterTQ T Q))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w tid
    *^* sp + 4 |->w ptr
    *^* mP) (S_Mem S)
  /\ tid > 0
  /\ sp >= 2 
  /\ S_RF S BP = sp - 2
  /\ S_RF S SP = sp - 2 
  /\ S_RF S AX = tid
  /\ S_RF S CX = ptr
  /\ S_Flag S _ZF = true.

Definition Ast_enQueue_10_notemp  (T : TQ) (Q : WQ)
  (ptr : Label) (tid : TID) (tcb : Thread)
  (bp sp head tail prev : nat) (mP : MPred) (S : State) : Prop :=
  (tid |->w tail 
    *^* tid + 2 |->w 0 
    *^* ptoTCB (tid+4) tcb
    *^* ptr |->w head
    *^* ptr + 2 |->w tail
    *^* (Queue Thread ptoTCB head tail prev 0 (filterTQ T Q))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w tid
    *^* sp + 4 |->w ptr
    *^* mP) (S_Mem S)
  /\ tid > 0
  /\ sp >= 2 
  /\ tail > 0
  /\ S_RF S BP = sp - 2
  /\ S_RF S SP = sp - 2 
  /\ S_RF S AX = tid
  /\ S_RF S DX = tail
  /\ S_RF S CX = ptr
  /\ S_RF S BX = tid
  /\ S_Flag S _ZF = false.

Definition L_enQueue_10_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists tid, exists tcb,
    exists bp, exists sp,    
      exists head, exists tail, exists prev,
    Ast_enQueue_10_emp T Q ptr tid tcb bp sp prev mTrue S \/
    Ast_enQueue_10_notemp T Q ptr tid tcb bp sp head tail prev mTrue S.

Definition L_enQueue_10_g (S S' : State) : Prop :=
  forall T Q ptr tid tcb bp sp head tail prev mP,
    Ast_enQueue_10_emp T Q ptr tid tcb bp sp prev mP S \/
    Ast_enQueue_10_notemp T Q ptr tid tcb bp sp head tail prev mP S
    -> exists T', exists Q',
      Ast_enQueue_post T' Q' ptr tid tcb bp sp mP S'
      /\ T' = TQMod.put T tid tcb
      /\ Q' = addThrd Q tid.

(* <enQueue> : jne L_enQ_NotNull *)
Definition Ast_enQueue_11  (T : TQ) (Q : WQ)
  (ptr : Label) (tid : TID) (tcb : Thread)
  (bp sp prev : nat) (mP : MPred) (S : State) : Prop :=
  (tid |->w 0 
    *^* tid + 2 |->w 0 
    *^* ptoTCB (tid+4) tcb
    *^* ptr |->w 0
    *^* ptr + 2 |->w 0
    *^* (Queue Thread ptoTCB 0 0 prev 0 (filterTQ T Q))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w tid
    *^* sp + 4 |->w ptr
    *^* mP) (S_Mem S)
  /\ tid > 0
  /\ sp >= 2 
  /\ S_RF S BP = sp - 2
  /\ S_RF S SP = sp - 2 
  /\ S_RF S AX = tid
  /\ S_RF S CX = ptr.

Definition L_enQueue_11_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists tid, exists tcb,
    exists bp, exists sp,   exists prev,
    Ast_enQueue_11 T Q ptr tid tcb bp sp prev mTrue S.

Definition L_enQueue_11_g (S S' : State) : Prop :=
  forall T Q ptr tid tcb bp sp prev mP,
    Ast_enQueue_11 T Q ptr tid tcb bp sp prev mP S
    -> exists T', exists Q',
      Ast_enQueue_post T' Q' ptr tid tcb bp sp mP S'
      /\ T' = TQMod.put T tid tcb
      /\ Q' = addThrd Q tid.

(* <enQueue> : movw    %cx, %bx *)
Definition Ast_enQueue_12  (T : TQ) (Q : WQ)
  (ptr : Label) (tid : TID) (tcb : Thread)
  (bp sp prev : nat) (mP : MPred) (S : State) : Prop :=
  (tid |->w 0 
    *^* tid + 2 |->w 0 
    *^* ptoTCB (tid+4) tcb
    *^* ptr |->w 0
    *^* ptr + 2 |->w 0
    *^* (Queue Thread ptoTCB 0 0 prev 0 (filterTQ T Q))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w tid
    *^* sp + 4 |->w ptr
    *^* mP) (S_Mem S)
  /\ tid > 0
  /\ sp >= 2 
  /\ S_RF S BP = sp - 2
  /\ S_RF S SP = sp - 2 
  /\ S_RF S AX = tid
  /\ S_RF S BX = ptr.

Definition L_enQueue_12_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists tid, exists tcb,
    exists bp, exists sp,   exists prev,
    Ast_enQueue_12 T Q ptr tid tcb bp sp prev mTrue S.

Definition L_enQueue_12_g (S S' : State) : Prop :=
  forall T Q ptr tid tcb bp sp prev mP,
    Ast_enQueue_12 T Q ptr tid tcb bp sp prev mP S
    -> exists T', exists Q',
      Ast_enQueue_post T' Q' ptr tid tcb bp sp mP S'
      /\ T' = TQMod.put T tid tcb
      /\ Q' = addThrd Q tid.

(* <enQueue> : movw    %ax, Q_HEAD(%bx) *)
Definition Ast_enQueue_13  (T : TQ) (Q : WQ)
  (ptr : Label) (tid : TID) (tcb : Thread)
  (bp sp prev : nat) (mP : MPred) (S : State) : Prop :=
  (tid |->w 0 
    *^* tid + 2 |->w 0 
    *^* ptoTCB (tid+4) tcb
    *^* ptr |->w tid
    *^* ptr + 2 |->w 0
    *^* (Queue Thread ptoTCB 0 0 prev 0 (filterTQ T Q))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w tid
    *^* sp + 4 |->w ptr
    *^* mP) (S_Mem S)
  /\ tid > 0
  /\ sp >= 2 
  /\ S_RF S BP = sp - 2
  /\ S_RF S SP = sp - 2 
  /\ S_RF S AX = tid
  /\ S_RF S BX = ptr.

Definition L_enQueue_13_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists tid, exists tcb,
    exists bp, exists sp,   exists prev,
    Ast_enQueue_13 T Q ptr tid tcb bp sp prev mTrue S.

Definition L_enQueue_13_g (S S' : State) : Prop :=
  forall T Q ptr tid tcb bp sp prev mP,
    Ast_enQueue_13 T Q ptr tid tcb bp sp prev mP S
    -> exists T', exists Q',
      Ast_enQueue_post T' Q' ptr tid tcb bp sp mP S'
      /\ T' = TQMod.put T tid tcb
      /\ Q' = addThrd Q tid.

(* <enQueue> : movw    %ax, Q_TAIL(%bx) *)
Definition Ast_enQueue_14  (T : TQ) (Q : WQ)
  (ptr : Label) (tid : TID) (tcb : Thread)
  (bp sp prev : nat) (mP : MPred) (S : State) : Prop :=
  (tid |->w 0 
    *^* tid + 2 |->w 0 
    *^* ptoTCB (tid+4) tcb
    *^* ptr |->w tid
    *^* ptr + 2 |->w tid
    *^* (Queue Thread ptoTCB 0 0 prev 0 (filterTQ T Q))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w tid
    *^* sp + 4 |->w ptr
    *^* mP) (S_Mem S)
  /\ tid > 0
  /\ sp >= 2 
  /\ S_RF S BP = sp - 2
  /\ S_RF S SP = sp - 2.

Definition L_enQueue_14_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists tid, exists tcb,
    exists bp, exists sp,   exists prev,
    Ast_enQueue_14 T Q ptr tid tcb bp sp prev mTrue S.

Definition L_enQueue_14_g (S S' : State) : Prop :=
  forall T Q ptr tid tcb bp sp prev mP,
    Ast_enQueue_14 T Q ptr tid tcb bp sp prev mP S
    -> exists T', exists Q',
      Ast_enQueue_post T' Q' ptr tid tcb bp sp mP S'
      /\ T' = TQMod.put T tid tcb
      /\ Q' = addThrd Q tid.

(* <enQueue> :  popw    %bp *)
Definition Ast_enQueue_15  (T : TQ) (Q : WQ)
  (ptr : Label) (tid : TID) (tcb : Thread)
  (bp sp prev : nat) (mP : MPred) (S : State) : Prop :=
  (tid |->w 0 
    *^* tid + 2 |->w 0 
    *^* ptoTCB (tid+4) tcb
    *^* ptr |->w tid
    *^* ptr + 2 |->w tid
    *^* (Queue Thread ptoTCB 0 0 prev 0 (filterTQ T Q))
    *^* stkSpacePrec sp 2
    *^* sp + 2 |->w tid
    *^* sp + 4 |->w ptr
    *^* mP) (S_Mem S)
  /\ tid > 0
  /\ sp >= 2 
  /\ S_RF S BP = bp
  /\ S_RF S SP = sp.

Definition L_enQueue_15_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists tid, exists tcb,
    exists bp, exists sp,   exists prev,
    Ast_enQueue_15 T Q ptr tid tcb bp sp prev mTrue S.

Definition L_enQueue_15_g (S S' : State) : Prop :=
  forall T Q ptr tid tcb bp sp prev mP,
    Ast_enQueue_15 T Q ptr tid tcb bp sp prev mP S
    -> exists T', exists Q',
      Ast_enQueue_post T' Q' ptr tid tcb bp sp mP S'
      /\ T' = TQMod.put T tid tcb
      /\ Q' = addThrd Q tid.

(* <enQueue> : retw *)

(* <enQueue> : enQ_NotNull *)
Definition Ast_enQueue_16  (T : TQ) (Q : WQ)
  (ptr : Label) (tid : TID) (tcb : Thread)
  (bp sp head tail prev : nat) (mP : MPred) (S : State) : Prop :=
  (tid |->w tail 
    *^* tid + 2 |->w 0 
    *^* ptoTCB (tid+4) tcb
    *^* ptr |->w head
    *^* ptr + 2 |->w tail
    *^* (Queue Thread ptoTCB head tail prev 0 (filterTQ T Q))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w tid
    *^* sp + 4 |->w ptr
    *^* mP) (S_Mem S)
  /\ tid > 0
  /\ sp >= 2 
  /\ tail > 0
  /\ S_RF S BP = sp - 2
  /\ S_RF S SP = sp - 2 
  /\ S_RF S AX = tid
  /\ S_RF S DX = tail
  /\ S_RF S CX = ptr.

Definition L_enQueue_16_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists tid, exists tcb,
    exists bp, exists sp,    
      exists head, exists tail, exists prev,
    Ast_enQueue_16 T Q ptr tid tcb bp sp head tail prev mTrue S.

Definition L_enQueue_16_g (S S' : State) : Prop :=
  forall T Q ptr tid tcb bp sp head tail prev mP,
    Ast_enQueue_16 T Q ptr tid tcb bp sp head tail prev mP S
    -> exists T', exists Q',
      Ast_enQueue_post T' Q' ptr tid tcb bp sp mP S'
      /\ T' = TQMod.put T tid tcb
      /\ Q' = addThrd Q tid.

(* <enQueue> : movw    %dx, %bx *)
Definition Ast_enQueue_17  (T : TQ) (Q : WQ)
  (ptr : Label) (tid : TID) (tcb : Thread)
  (bp sp head tail prev : nat) (mP : MPred) (S : State) : Prop :=
  (tid |->w tail 
    *^* tid + 2 |->w 0 
    *^* ptoTCB (tid+4) tcb
    *^* ptr |->w head
    *^* ptr + 2 |->w tail
    *^* (Queue Thread ptoTCB head tail prev 0 (filterTQ T Q))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w tid
    *^* sp + 4 |->w ptr
    *^* mP) (S_Mem S)
  /\ tid > 0
  /\ sp >= 2 
  /\ tail > 0
  /\ S_RF S BP = sp - 2
  /\ S_RF S SP = sp - 2 
  /\ S_RF S AX = tid
  /\ S_RF S CX = ptr
  /\ S_RF S BX = tail.

Definition L_enQueue_17_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists tid, exists tcb,
    exists bp, exists sp,    
      exists head, exists tail, exists prev,
    Ast_enQueue_17 T Q ptr tid tcb bp sp head tail prev mTrue S.

Definition L_enQueue_17_g (S S' : State) : Prop :=
  forall T Q ptr tid tcb bp sp head tail prev mP,
    Ast_enQueue_17 T Q ptr tid tcb bp sp head tail prev mP S
    -> exists T', exists Q',
      Ast_enQueue_post T' Q' ptr tid tcb bp sp mP S'
      /\ T' = TQMod.put T tid tcb
      /\ Q' = addThrd Q tid.

(* <enQueue> : movw    %ax, ND_NEXT(%bx) *)
Definition Ast_enQueue_18  (T : TQ) (Q : WQ)
  (ptr : Label) (tid : TID) (tcb : Thread)
  (bp sp head tail prev : nat) (mP : MPred) (S : State) : Prop :=
  (tid |->w tail 
    *^* tid + 2 |->w 0 
    *^* ptoTCB (tid+4) tcb
    *^* ptr |->w head
    *^* ptr + 2 |->w tail
    *^* (Queue Thread ptoTCB head tail prev tid (filterTQ T Q))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w tid
    *^* sp + 4 |->w ptr
    *^* mP) (S_Mem S)
  /\ tid > 0
  /\ sp >= 2 
  /\ tail > 0
  /\ S_RF S BP = sp - 2
  /\ S_RF S SP = sp - 2 
  /\ S_RF S AX = tid
  /\ S_RF S CX = ptr.

Definition L_enQueue_18_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists tid, exists tcb,
    exists bp, exists sp,    
      exists head, exists tail, exists prev,
    Ast_enQueue_18 T Q ptr tid tcb bp sp head tail prev mTrue S.

Definition L_enQueue_18_g (S S' : State) : Prop :=
  forall T Q ptr tid tcb bp sp head tail prev mP,
    Ast_enQueue_18 T Q ptr tid tcb bp sp head tail prev mP S
    -> exists T', exists Q',
      Ast_enQueue_post T' Q' ptr tid tcb bp sp mP S'
      /\ T' = TQMod.put T tid tcb
      /\ Q' = addThrd Q tid.

(* <enQueue> :  movw    %cx, %bx *)
Definition Ast_enQueue_19  (T : TQ) (Q : WQ)
  (ptr : Label) (tid : TID) (tcb : Thread)
  (bp sp head tail prev : nat) (mP : MPred) (S : State) : Prop :=
  (tid |->w tail 
    *^* tid + 2 |->w 0
    *^* ptoTCB (tid+4) tcb
    *^* ptr |->w head
    *^* ptr + 2 |->w tail
    *^* (Queue Thread ptoTCB head tail prev tid (filterTQ T Q))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w tid
    *^* sp + 4 |->w ptr
    *^* mP) (S_Mem S)
  /\ tid > 0
  /\ sp >= 2 
  /\ tail > 0
  /\ S_RF S BP = sp - 2
  /\ S_RF S SP = sp - 2 
  /\ S_RF S AX = tid
  /\ S_RF S BX = ptr.

Definition L_enQueue_19_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists tid, exists tcb,
    exists bp, exists sp,    
      exists head, exists tail, exists prev,
    Ast_enQueue_19 T Q ptr tid tcb bp sp head tail prev mTrue S.

Definition L_enQueue_19_g (S S' : State) : Prop :=
  forall T Q ptr tid tcb bp sp head tail prev mP,
    Ast_enQueue_19 T Q ptr tid tcb bp sp head tail prev mP S
    -> exists T', exists Q',
      Ast_enQueue_post T' Q' ptr tid tcb bp sp mP S'
      /\ T' = TQMod.put T tid tcb
      /\ Q' = addThrd Q tid.

(* <enQueue> :  movw    %ax, Q_TAIL(%bx) *)
Definition Ast_enQueue_20  (T : TQ) (Q : WQ)
  (ptr : Label) (tid : TID) (tcb : Thread)
  (bp sp head tail prev : nat) (mP : MPred) (S : State) : Prop :=
  (tid |->w tail 
    *^* tid + 2 |->w 0 
    *^* ptoTCB (tid+4) tcb
    *^* ptr |->w head
    *^* ptr + 2 |->w tid
    *^* (Queue Thread ptoTCB head tail prev tid (filterTQ T Q))
    *^* sp - 2 |->w bp
    *^* sp + 2 |->w tid
    *^* sp + 4 |->w ptr
    *^* mP) (S_Mem S)
  /\ tid > 0
  /\ sp >= 2 
  /\ tail > 0
  /\ S_RF S BP = sp - 2
  /\ S_RF S SP = sp - 2.

Definition L_enQueue_20_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists tid, exists tcb,
    exists bp, exists sp,    
      exists head, exists tail, exists prev,
    Ast_enQueue_20 T Q ptr tid tcb bp sp head tail prev mTrue S.

Definition L_enQueue_20_g (S S' : State) : Prop :=
  forall T Q ptr tid tcb bp sp head tail prev mP,
    Ast_enQueue_20 T Q ptr tid tcb bp sp head tail prev mP S
    -> exists T', exists Q',
      Ast_enQueue_post T' Q' ptr tid tcb bp sp mP S'
      /\ T' = TQMod.put T tid tcb
      /\ Q' = addThrd Q tid.

(* <enQueue> : popw    %bp *)
Definition Ast_enQueue_21  (T : TQ) (Q : WQ)
  (ptr : Label) (tid : TID) (tcb : Thread)
  (bp sp head tail prev : nat) (mP : MPred) (S : State) : Prop :=
  (tid |->w tail 
    *^* tid + 2 |->w 0
    *^* ptoTCB (tid+4) tcb
    *^* ptr |->w head
    *^* ptr + 2 |->w tid
    *^* (Queue Thread ptoTCB head tail prev tid (filterTQ T Q))
    *^* stkSpacePrec sp 2 
    *^* sp + 2 |->w tid
    *^* sp + 4 |->w ptr
    *^* mP) (S_Mem S)
  /\ tid > 0
  /\ sp >= 2 
  /\ tail > 0
  /\ S_RF S BP = bp
  /\ S_RF S SP = sp.

Definition L_enQueue_21_p (S : State) : Prop :=
  exists T, exists Q, exists ptr, exists tid, exists tcb,
    exists bp, exists sp,    
      exists head, exists tail, exists prev,
    Ast_enQueue_21 T Q ptr tid tcb bp sp head tail prev mTrue S.

Definition L_enQueue_21_g (S S' : State) : Prop :=
  forall T Q ptr tid tcb bp sp head tail prev mP,
    Ast_enQueue_21 T Q ptr tid tcb bp sp head tail prev mP S
    -> exists T', exists Q',
      Ast_enQueue_post T' Q' ptr tid tcb bp sp mP S'
      /\ T' = TQMod.put T tid tcb
      /\ Q' = addThrd Q tid.

(* <enQueue> : retw *)

Definition spec_enQueue := Ast p_enQueue g_enQueue sctag.

Definition psi_enQueue_Block0  :=
  (SCAP_consCT L_enQ_NotNull (Ast L_enQueue_16_p L_enQueue_16_g sctag)
    SCAP_emptCT).

Definition psi_enQueue_Block1  := 
  SCAP_emptCT.
