(***************************************************************)
(* Coq code for AIM                                            *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

Add LoadPath ".." .
Add LoadPath "../CAP0" .
Add LoadPath "../SCAP" .

Require Import Machine.
Require Import Operational_Semantics.
Require Import AIM.

Notation Local " i ; I " := (seq i I)
  (at level 60, right associativity).

Notation Local " i ;; " := (sseq i)
  (at level 60).

(** Thread Control Block constructure **)
Definition TCB_STATUS : nat := 6.
Definition TCB_STACK : nat := 4.
Notation TCB_PREV := 0.
Notation TCB_NEXT := 2.

(** Queue and node fields **)
Notation Q_HEAD  := 0.
Notation Q_TAIL  := 2.
Notation ND_PREV := 0.
Notation ND_NEXT := 2.

(* Notation _STK_SIZE_unblock := 10 (only parsing). *)
(* Notation _STK_SIZE_swith := 8 (only parsing). *)
(* Notation _STK_SIZE_block := 10 (only parsing). *)

(** Functioon: scheduler, system thread scheduler**)
(********Source Code *************************** 

# precondition requires that interrupt be disabled
1146 <scheduler>:
        movw    (curr_thrd), %bx
        movw    %sp, TCB_STACK(%bx)
        pushw   $readyQ
        pushw   %bx
1150:   call    enQueue (L_enQueue)
1151:   addw    $2, %sp
1152:   call    deQueue (L_deQueue)
1153:   addw    $2, %sp
        movw    %ax, %bx
        movw    %ax, (curr_thrd)
        movw    TCB_STACK(%bx), %sp
1157:   retw

# precondition requires that interrupt be disabled
# block(q)
1158 <block>:
        movw    (curr_thrd), %bx	#curr_thrd = 2254
        movw    $BLOCKED, TCB_STATUS(%bx)
1160    movw    %sp, TCB_STACK(%bx)
        movw    %sp, %bp
        movw    2(%bp), %ax
        pushw   %ax			#BlockQ
        pushw   %bx
1165    call    enQueue	(1190)
        pushw   $readyQ			#readyQ = 2250
        call    deQueue (1212)
# note that we can always find a thread in the readyQ
# because our program logic guarantees the idle thread
# will never be blocked. So the idle thread must always
# be in the ready queue at this moment
        addw    $6, %sp
        movw    %ax, %bx
1170    movw    %ax, (curr_thrd)
        movw    TCB_STACK(%bx), %sp
1172    retw


# unblock(q)
1173<unblock>:
        pushw   %bp
        movw    %sp, %bp
1175:   movw    4(%bp), %bx             # the pointer to the block queue
        pushw   %bx
        call    deQueue                 # not change bp
        addw    $2, %sp
        cmpw    $0, %ax
1180:   jne     wakeup
        popw    %bp
        retw                            # block_done

1183:wakeup:
        movw    %ax, %bx
        movw    $READY, TCB_STATUS(%bx)
1185:   pushw   $readyQ
        pushw   %bx
        call    enQueue
        popw    %ax
        addw    $2, %sp
1190:   popw    %bp
        retw

1190 <enQueue>:
        pushw   %bp
        movw    %sp, %bp
        movw    4(%bp), %ax             # ax: pointer to node
        movw    6(%bp), %bx             # bx: pointer to Q
        movw    Q_TAIL(%bx), %dx
1195:   movw    %bx, %cx                # save bx
        movw    %ax, %bx
        movw    %dx, ND_PREV(%bx)
        movw    $0, ND_NEXT(%bx)
        cmpw    $0, %dx
1200:   jne     not_null
        movw    %cx, %bx                # bx: pointers to Q
        movw    %ax, Q_HEAD(%bx)        # head = newtcb
        movw    %ax, Q_TAIL(%bx)
        popw    %bp
        retw
1206: <not_null>:
        movw    %dx, %bx
        movw    %ax, ND_NEXT(%bx)
        movw    %cx, %bx                # bx: pointers to Q
        movw    %ax, Q_TAIL(%bx)
1210:   popw    %bp
        retw

1212 <deQueue>:                         # Block0
        pushw   %bp
        movw    %sp, %bp
        movw    4(%bp), %bx             # bx: pointer to Q
1215:   movw    Q_HEAD(%bx), %dx        # dx: head
        cmpw    $0, %dx
        jne     deq_a
        movw    $0, %ax                 # return value
        popw    %bp
1220:   retw
1221: <deq_a>:                                  # Block1
        movw    %bx, %cx                # save bx
        movw    %dx, %bx
        movw    ND_NEXT(%bx), %dx
        movw    %bx, %ax                # return value
1225:   movw    %cx, %bx                # bx: pointer to Q
        cmpw    $0, %dx
        jne     deq_b
        movw    %dx, Q_TAIL(%bx)
1229: <deq_b>:                          # Block2:
        movw    %dx, Q_HEAD(%bx)
        popw    %bp
        retw

************************************************)

(** Instruction Label **)

Parameter L_enQueue: Label.
Definition L_enQ_NotNull := L_enQueue + 16.

Parameter L_deQueue: Label.
Definition L_deQ_a := L_deQueue + 9.
Definition L_deQ_b := L_deQueue + 17.

Definition L_schedule: Label := _SWITCH.

Definition L_block : Label := _BLOCK.

Definition L_unblock  : Label := _UNBLK.

Definition L_wakeup := L_unblock +1+1+1+1 +1+1+1+1 +1+1.

(** Instruction sequence **)

Definition I_enQueue  := 
  (seq (_pushwp (Opr_reg BP))
  (seq (_movwp  (Opr_reg SP) BP)
  (seq (_movwld (A_reg 4 BP) AX)
  (seq (_movwld (A_reg 6 BP) BX)
  (seq (_movwld (A_reg Q_TAIL BX) DX)
  (seq (_movwp  (Opr_reg BX) CX)
  (seq (_movwp  (Opr_reg AX) BX)
  (seq (_movwst (Opr_reg DX) (A_reg ND_PREV BX))
  (seq (_movwst (Opr_imm 0 ) (A_reg ND_NEXT BX))
  (seq (_cmpw   (Opr_imm 0 ) DX)
  (seq (_jne    L_enQ_NotNull)
  (seq (_movwp  (Opr_reg CX) BX)
  (seq (_movwst (Opr_reg AX) (A_reg Q_HEAD BX))
  (seq (_movwst (Opr_reg AX) (A_reg Q_TAIL BX))
  (seq (_popw   BP)
  (sseq (_retw))))))))))))))))).

Definition I_enQ_NotNull :=
  (seq (_movwp  (Opr_reg DX) BX)
  (seq (_movwst (Opr_reg AX) (A_reg ND_NEXT BX))
  (seq (_movwp  (Opr_reg CX) BX)
  (seq (_movwst (Opr_reg AX) (A_reg Q_TAIL BX))
  (seq (_popw   BP)
  (sseq (_retw))))))).

(*********************************************)

Definition I_deQueue  := 
  (seq (_pushwp (Opr_reg BP))
  (seq (_movwp  (Opr_reg SP) BP)
  (seq (_movwld (A_reg 4 BP) BX)
  (seq (_movwld (A_reg Q_HEAD BX) DX)
  (seq (_cmpw   (Opr_imm 0 ) DX)
  (seq (_jne    L_deQ_a)
  (seq (_movwp  (Opr_imm 0 ) AX)
  (seq (_popw   BP)
  (sseq (_retw)))))))))).

Definition I_deQ_a  := 
  (seq (_movwp  (Opr_reg BX) CX)
  (seq (_movwp  (Opr_reg DX) BX)
  (seq (_movwld (A_reg ND_NEXT BX) DX)
  (seq (_movwp  (Opr_reg BX) AX)
  (seq (_movwp  (Opr_reg CX) BX)
  (seq (_cmpw   (Opr_imm 0 ) DX)
  (seq (_jne    L_deQ_b)
  (seq (_movwst (Opr_reg DX) (A_reg Q_TAIL BX))
  (seq (_movwst (Opr_reg DX) (A_reg Q_HEAD BX))
  (seq (_popw   BP)
  (sseq (_retw)))))))))))).

Definition I_deQ_b  := 
  (seq (_movwst (Opr_reg DX) (A_reg Q_HEAD BX))
  (seq (_popw   BP)
  (sseq (_retw)))).

(*********************************************)

Definition I_schedule:= 
  (seq (_movwld (A_imm curr_thrd) BX)
  (seq (_movwst (Opr_reg SP) (A_reg TCB_STACK BX))
  (seq (_pushwp (Opr_imm ready_Q))
  (seq (_pushwp (Opr_reg BX))
  (seq (_call    L_enQueue)
  (seq (_addw   (Opr_imm 2 ) SP)
  (seq (_call    L_deQueue)
  (seq (_addw   (Opr_imm 2 ) SP)
  (seq (_movwp  (Opr_reg AX) BX)
  (seq (_movwst (Opr_reg AX) (A_imm curr_thrd))
  (seq (_movwld (A_reg TCB_STACK BX) SP)
  (sseq (_retw))))))))))))).

(**********************************************)

Definition I_block:= 
(* 00 *)  (seq (_movwld (A_imm curr_thrd) BX)
(* 01 *)  (seq (_movwst (Opr_imm 2) (A_reg TCB_STATUS BX))
(* 02 *)  (seq (_movwst (Opr_reg SP) (A_reg TCB_STACK BX))
(* 03 *)  (seq (_movwp  (Opr_reg SP) BP)
(* 04 *)  (seq (_movwld (A_reg 2 BP) AX)
(* 05 *)  (seq (_pushwp (Opr_reg AX))
(* 06 *)  (seq (_pushwp (Opr_reg BX))
(* 07 *)  (seq (_call    L_enQueue)
(* 08 *)  (seq (_pushwp (Opr_imm ready_Q))
(* 09 *)  (seq (_call    L_deQueue)
(* 10 *)  (seq (_addw   (Opr_imm 6 ) SP)
(* 11 *)  (seq (_movwp  (Opr_reg AX) BX)        
(* 12 *)  (seq (_movwst (Opr_reg AX) (A_imm curr_thrd))
(* 13 *)  (seq (_movwld (A_reg TCB_STACK BX) SP)
(* 14 *)  (sseq (_retw)))))))))))))))).

(**********************************************)

Definition I_unblock  :=
  (seq (_pushwp (Opr_reg BP))
  (seq (_movwp  (Opr_reg SP) BP)
  (seq (_movwld (A_reg 4 BP) BX)
  (seq (_pushwp (Opr_reg BX))
  (seq (_call    L_deQueue)
  (seq (_addw   (Opr_imm 2 ) SP)
  (seq (_cmpw   (Opr_imm 0 ) AX)
  (seq (_jne    L_wakeup)
  (seq (_popw   BP)
  (sseq (_retw))))))))))).

Definition I_wakeup  :=
  (seq (_movwp  (Opr_reg AX) BX)
  (seq (_movwst (Opr_imm 1) (A_reg TCB_STATUS BX))
  (seq (_pushwp (Opr_imm ready_Q))
  (seq (_pushwp (Opr_reg BX))
  (seq (_call    L_enQueue)
  (seq (_popw   AX)
  (seq (_addw   (Opr_imm 2 ) SP)
  (seq (_popw   BP)
  (sseq (_retw)))))))))).

(**********************************************)

(********Source Code *************************** 

# precondition: interrupt is disabled
# sleep(q)
sleep:
        pushw   %ax
        pushw   %bx
        pushw   %cx
        pushw   %dx
        pushw   %di
        pushw   %si
        pushw   %bp
        pushfw
        movw    %sp, %bp
        pushw   18(%bp)
        call    block
        addw    $2, %sp
        popfw
        popw    %bp
        popw    %si
        popw    %di
        popw    %dx
        popw    %cx
        popw    %bx
        popw    %ax
        retw

yield:
        cli
        pushw   %ax
        pushw   %bx
        pushw   %cx
        pushw   %dx
        pushw   %di
        pushw   %si
        pushw   %bp
        pushfw
        call    scheduler
        popfw
        popw    %bp
        popw    %si
        popw    %di
        popw    %dx
        popw    %cx
        popw    %bx
        popw    %ax
        sti
        retw

*********************************************)

(** Instruction Label **)
Parameter L_sleep : Label.

Parameter L_yield : Label.

Parameter sleep_not_prim : ~ prim_entry L_sleep.

Notation _STK_SIZE_yield:= (16 + 2 + 8 + IH_Stk) (only parsing).

Notation _STK_SIZE_sleep := (18 + 2 + _BLK_Stk) (only parsing).

Definition I_yield := 
  _cli;
  _pushwp (Opr_reg AX);
  _pushwp (Opr_reg BX);
  _pushwp (Opr_reg CX);
  _pushwp (Opr_reg DX);
  _pushwp (Opr_reg DI);
  _pushwp (Opr_reg SI);
  _pushwp (Opr_reg BP);
  _pushfw;
  _call _SWITCH;
  _popfw;
  _popw           (BP);
  _popw           (SI);
  _popw           (DI);
  _popw           (DX);
  _popw           (CX);
  _popw           (BX);
  _popw           (AX);
  _sti;
  _retw;; .

Definition I_sleep := 
(* <sleep + 0> : *)  _pushwp (Opr_reg AX);
(* <sleep + 1> : *)  _pushwp (Opr_reg BX);
(* <sleep + 2> : *)  _pushwp (Opr_reg CX);
(* <sleep + 3> : *)  _pushwp (Opr_reg DX);
(* <sleep + 4> : *)  _pushwp (Opr_reg DI);
(* <sleep + 5> : *)  _pushwp (Opr_reg SI);
(* <sleep + 6> : *)  _pushwp (Opr_reg BP);
(* <sleep + 7> : *)  _pushfw;
(* <sleep + 8> : *)  _movwp (Opr_reg SP) BP;
(* <sleep + 9> : *)  _pushwm (A_reg 18 BP);
(* <sleep + 10>: *)  _call _BLOCK;
(* <sleep + 11>: *)  _addw   (Opr_imm 2) SP;
(* <sleep + 12>: *)  _popfw;
(* <sleep + 13>: *)  _popw           (BP);
(* <sleep + 14>: *)  _popw           (SI);
(* <sleep + 15>: *)  _popw           (DI);
(* <sleep + 16>: *)  _popw           (DX);
(* <sleep + 17>: *)  _popw           (CX);
(* <sleep + 18>: *)  _popw           (BX);
(* <sleep + 19>: *)  _popw           (AX);
(* <sleep + 20>: *)  _retw;; .
