(***************************************************************)
(* Coq code for AIM                                            *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

Add LoadPath "..".
Add LoadPath "../CAP0".
Add LoadPath "../SCAP".

Require Import Coqlib.

Require Import Arith.
Require Import Machine.
Require Import Assertion.
Require Import Assertion_Lemmas.
Require Import Operational_Semantics.
Require Import OpSem_Lemmas.

Require Import AIM.
Require Import AIM_Lemmas.
Require Import SCAP.
Require Import SCAP_READY.
Require Import QLib.
Require Import Linking.
Require Import Assertion_Lemmas4.
Require Import Assertion_Lemmas5.
Require Import Reg.

Require Import Kernel.

Require Import SleepSpec.

(********Source Code *************************** 

# precondition: interrupt is disabled
# sleep(q)
sleep:
        pushw   %ax
        pushw   %bx
        pushw   %cx
        pushw   %dx
        pushw   %di
        pushw   %si
        pushw   %bp
        pushfw
        movw    %sp, %bp
        pushw   18(%bp)
        call    block
        addw    $2, %sp
        popfw
        popw    %bp
        popw    %si
        popw    %di
        popw    %dx
        popw    %cx
        popw    %bx
        popw    %ax
        retw

************************************************)

Lemma NH_WFcmd_sleep_0_pushwp :
  NH_WFcmd (SCAP_consCT (L_sleep+1) (Ast L_sleep_1_p L_sleep_1_g tag_sleep) psi_sleep)
    (L_sleep) (Ast L_sleep_0_p L_sleep_0_g tag_sleep) 
    (comml (_pushwp (Opr_reg AX))).
Proof.
  simpl.
  split.
  exists L_sleep_1_p; exists L_sleep_1_g.
  split_l.
  assert (Hpp' : forall l_ptr wqa sp pc mP S S',
    Ast_sleep_0 l_ptr wqa sp pc mP S
    -> nextS (L_sleep) (comml (_pushwp (Opr_reg AX))) S S'
    -> Ast_sleep_1 l_ptr wqa sp pc (S_RF S AX) mP S').
    intros l_ptr wqa sp pc mP S S' HS HSS'.
    unfold Ast_sleep_0, Ast_sleep_1 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspgt [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simplS in * .
    destruct HS as [HM [Hsp [Hspgt' [Hif Hisr]]]].
    split.
      subst M' R' sp.
      sep on.
      sep lift 3.
      reg simpl.
      sep upd.
      sep lift 2 in HM.
      sep stksplit in HM.
      sep matchH HM with 6.
    subst R' fl' isr' sp; reg simpl.
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_sleep_0_p in HpS.
    destruct HpS as [l_ptr [wqa [sp [pc HS]]]].
    exists (MemUpdW M (sp-2) (R AX) 0, RF_update R SP (sp - 2), fl, isr).
    apply nextS_pushwp with  (R AX); trivial.
    unfold pushW.
    destruct HS as [HM [Hsp [Hspgt [Hif Hisr]]]].
    simplS in * .
    split.
      apply stkSpace_weaken with _STK_SIZE_sleep;
        [idtac | omega].
      sep on.
      sep lift 2 in HM.
      rewrite Hsp.
      sep trivial.
    split.
      subst sp; eapply MemUpdW_MemUpd16; eauto.
    subst sp; trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_sleep_0_p, L_sleep_1_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [l_ptr [wqa [sp [pc HS]]]].
    exists l_ptr; exists wqa; exists sp; exists pc; exists (S_RF S AX).
    apply Hpp'; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_sleep_0_g, L_sleep_1_g in * .
  intros mP l_ptr wqa sp pc HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [_ [Hsp _]].
  simplS in * .
  subst sp R' fl'.
  reg simpl.
  apply (extensionality reg nat R'' R).
  induction x; auto.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_sleep_0_p in HS.
    destruct HS as [l_ptr [wqa [sp [pc HS]]]].
    unfold Ast_sleep_0 in HS.
    destruct HS as [HM [HSP [Hsp [HIF [Hisr Hwqa]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite HIF in IHif.
    discriminate.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_sleep_0_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP l_ptr wqa sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_sleep_0 in HS.
  simplS in * .
  destruct HS as [HM [HSP [Hsp [Hif [Hisr Hwqa]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate.
Qed.

Lemma NH_WFcmd_sleep_1_pushwp :
  NH_WFcmd (SCAP_consCT 
    (L_sleep+1+1)
      (Ast L_sleep_2_p L_sleep_2_g tag_sleep) psi_sleep)
    (L_sleep+1) 
      (Ast L_sleep_1_p L_sleep_1_g tag_sleep) 
    (comml (_pushwp (Opr_reg BX))).
Proof.
  simpl.
  split.
  exists L_sleep_2_p; exists L_sleep_2_g.
  split_l.
  assert (Hpp' : forall l_ptr wqa sp pc ax mP S S',
    Ast_sleep_1 l_ptr wqa sp pc ax mP S
    -> nextS (L_sleep+1) (comml (_pushwp (Opr_reg BX))) S S'
    -> Ast_sleep_2 l_ptr wqa sp pc ax (S_RF S BX) mP S').
    intros l_ptr wqa sp pc ax mP S S' HS HSS'.
    unfold Ast_sleep_1, Ast_sleep_2 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspgt [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simplS in * .
    destruct HS as [HM [Hax [Hsp [Hspgt' [Hif Hisr]]]]].
    split.
      subst M' R'.
      rewrite Hsp.
      sep on.
      sep lift 3.
      arith_rep (sp - 2 - 2) with (sp - 4).
      reg simpl.
      sep upd.
      sep lift 2 in HM.
      sep stksplit in HM.
      clear Hax Hsp Hspgt Hlds.
      arith_rep (sp - 2 - 2) with (sp - 4) in HM.
      arith_rep (18 + 2 + _BLK_Stk - 2 - 2) with
        (18 + 2 + _BLK_Stk - 4) in HM.
      sep matchH HM with 4.
    subst R' fl' isr'; reg simpl.
    rewrite Hsp.
    split_l.
    split_l by omega.
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HS.
    destruct S as [[[M R] fl] isr].
    unfold L_sleep_1_p in HS.
    destruct HS as [l_ptr [wqa [sp [pc [ax HS]]]]].
    exists (MemUpdW M (sp-4) (R BX) 0, RF_update R SP (sp - 4), fl, isr).
    apply nextS_pushwp with  (R BX); trivial.
    unfold pushW.
    destruct HS as [HM [Hax [Hsp [Hspgt [Hif Hisr]]]]].
    simplS in * .
    split.
      apply stkSpace_weaken with (_STK_SIZE_sleep - 2);
        [idtac | omega].
      sep on.
      sep lift 2 in HM.
      rewrite Hsp.
      sep trivial.
    split.
      rewrite Hsp. 
      arith_rep (sp - 2 - 2) with (sp - 4).
      eapply MemUpdW_MemUpd16; eauto.
    rewrite Hsp. 
    arith_rep (sp - 2 - 2) with (sp - 4).
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_sleep_1_p, L_sleep_2_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [l_ptr [wqa [sp [pc [ax HS]]]]].
    exists l_ptr; exists wqa; exists sp; exists pc; exists ax; exists (S_RF S BX).
    apply Hpp'; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_sleep_1_g, L_sleep_2_g in * .
  intros mP l_ptr wqa sp pc ax HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [_ [Hax [Hsp _]]].
  simplS in * .
  subst ax sp R' fl'.
  reg simpl.
  repeat split_l.
  trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_sleep_1_p in HS.
    destruct HS as [l_ptr [wqa [sp [pc [ax HS]]]]].
    unfold Ast_sleep_1 in HS.
    destruct HS as [HM [HAX [HSP [Hsp [Hif [Hisr Hwqa]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_sleep_1_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP l_ptr wqa sp pc ax HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_sleep_1 in HS.
  simplS in * .
  destruct HS as [HM [HAX [HSP [Hsp [Hif [Hisr Hwqa]]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma NH_WFcmd_sleep_2_pushwp :
  NH_WFcmd (SCAP_consCT 
    (L_sleep+2+1)
      (Ast L_sleep_3_p L_sleep_3_g tag_sleep) psi_sleep)
    (L_sleep+2) 
      (Ast L_sleep_2_p L_sleep_2_g tag_sleep) 
    (comml (_pushwp (Opr_reg CX))).
Proof.
  simpl.
  split.
  exists L_sleep_3_p; exists L_sleep_3_g.
  split_l.
  assert (Hpp' : forall l_ptr wqa sp pc ax bx mP S S',
    Ast_sleep_2 l_ptr wqa sp pc ax bx mP S
    -> nextS (L_sleep+2) (comml (_pushwp (Opr_reg CX))) S S'
    -> Ast_sleep_3 l_ptr wqa sp pc ax bx (S_RF S CX) mP S').
    intros l_ptr wqa sp pc ax bx mP S S' HS HSS'.
    unfold Ast_sleep_2, Ast_sleep_3 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspgt [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simplS in * .
    destruct HS as [HM [Hbx [Hsp [Hspgt' [Hif Hisr]]]]].
    split.
      subst M' R'.
      rewrite Hsp.
      sep on.
      sep lift 3.
      arith_rep (sp - 4 - 2) with (sp - 6).
      reg simpl.
      sep upd.
      sep lift 2 in HM.
      sep stksplit in HM.
      arith_rep (sp - 4 - 2) with (sp - 6) in HM.
      arith_rep (_STK_SIZE_sleep - 4 - 2) with (_STK_SIZE_sleep - 6) in HM.
      sep matchH HM with 5.
    subst R' fl' isr'; reg simpl.
    rewrite Hsp.
    arith_rep (sp - 4 - 2) with (sp - 6).
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_sleep_2_p in HpS.
    destruct HpS as [l_ptr [wqa [sp [pc [ax [bx HS]]]]]].
    exists (MemUpdW M (sp-6) (R CX) 0, RF_update R SP (sp - 6), fl, isr).
    apply nextS_pushwp with  (R CX); trivial.
    unfold pushW.
    destruct HS as [HM [Hbx [Hsp [Hspgt [Hif Hisr]]]]].
    simplS in * .
    split.
      apply stkSpace_weaken with (_STK_SIZE_sleep - 4);
        [idtac | omega].
      sep on.
      sep lift 2 in HM.
      rewrite Hsp.
      sep trivial.
    split.
      rewrite Hsp. 
      arith_rep (sp - 4 - 2) with (sp - 6).
      eapply MemUpdW_MemUpd16; eauto.
    rewrite Hsp. 
    arith_rep (sp - 4 - 2) with (sp - 6).
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_sleep_2_p, L_sleep_3_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx HS]]]]]].
    exists l_ptr; exists wqa; exists sp; exists pc; exists ax; exists bx;
      exists (S_RF S CX).
    apply Hpp'; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_sleep_2_g, L_sleep_3_g in * .
  intros mP l_ptr wqa sp pc ax bx HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [_ [Hbx [Hsp _]]].
  simplS in * .
  subst bx sp R' fl'.
  reg simpl.
  repeat split_l.
  trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_sleep_2_p in HS.
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx HS]]]]]].
    unfold Ast_sleep_2 in HS.
    destruct HS as [HM [HBX [HSP [Hsp [Hif [Hisr Hwqa]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_sleep_2_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP l_ptr wqa sp pc ax bx HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_sleep_2 in HS.
  simplS in * .
  destruct HS as [HM [HBX [HSP [Hsp [Hif [Hisr Hwqa]]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma NH_WFcmd_sleep_3_pushwp :
  NH_WFcmd (SCAP_consCT 
    (L_sleep+3+1)
      (Ast L_sleep_4_p L_sleep_4_g tag_sleep) psi_sleep)
    (L_sleep+3) 
      (Ast L_sleep_3_p L_sleep_3_g tag_sleep) 
    (comml (_pushwp (Opr_reg DX))).
Proof.
  simpl.
  split.
  exists L_sleep_4_p; exists L_sleep_4_g.
  split_l.
  assert (Hpp' : forall l_ptr wqa sp pc ax bx cx mP S S',
    Ast_sleep_3 l_ptr wqa sp pc ax bx cx mP S
    -> nextS (L_sleep+3) (comml (_pushwp (Opr_reg DX))) S S'
    -> Ast_sleep_4 l_ptr wqa sp pc ax bx cx (S_RF S DX) mP S').
    intros l_ptr wqa sp pc ax bx cx mP S S' HS HSS'.
    unfold Ast_sleep_2, Ast_sleep_3 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspgt [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simplS in * .
    destruct HS as [HM [Hcx [Hsp [Hspgt' [Hif Hisr]]]]].
    split.
      subst M' R'.
      rewrite Hsp.
      sep on.
      sep lift 3.
      arith_rep (sp - 6 - 2) with (sp - 8).
      reg simpl.
      simplS in * .
      sep upd.
      sep lift 2 in HM.
      sep stksplit in HM.
      arith_rep (sp - 6 - 2) with (sp - 8) in HM.
      arith_rep (_STK_SIZE_sleep - 6 - 2) with (_STK_SIZE_sleep - 8) in HM.
      sep matchH HM with 5.
    subst R' fl' isr'; reg simpl.
    rewrite Hsp.
    arith_rep (sp - 6 - 2) with (sp - 8).
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_sleep_4_p in HpS.
    destruct HpS as [l_ptr [wqa [sp [pc [ax [bx [cx HS]]]]]]].
    exists (MemUpdW M (sp-8) (R DX) 0, RF_update R SP (sp - 8), fl, isr).
    apply nextS_pushwp with  (R DX); trivial.
    unfold pushW.
    destruct HS as [HM [Hcx [Hsp [Hspgt [Hif Hisr]]]]].
    simplS in * .
    split.
      apply stkSpace_weaken with (_STK_SIZE_sleep - 6);
        [idtac | omega].
      sep on.
      sep lift 2 in HM.
      rewrite Hsp.
      sep trivial.
    split.
      rewrite Hsp. 
      arith_rep (sp - 6 - 2) with (sp - 8).
      eapply MemUpdW_MemUpd16; eauto.
    rewrite Hsp.
    arith_rep (sp - 6 - 2) with (sp - 8).
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_sleep_3_p, L_sleep_4_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx HS]]]]]]].
    exists l_ptr; exists wqa; exists sp; exists pc; exists ax; exists bx;
      exists cx; exists (S_RF S DX).
    apply Hpp'; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_sleep_3_g, L_sleep_4_g in * .
  intros mP l_ptr wqa sp pc ax bx cx HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [_ [Hcx [Hsp _]]].
  simplS in * .
  subst cx sp R' fl'.
  reg simpl.
  repeat split_l.
  trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_sleep_3_p in HS.
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx HS]]]]]]].
    unfold Ast_sleep_3 in HS.
    destruct HS as [HM [HCX [HSP [Hsp [Hif [Hisr Hwqa]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_sleep_3_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP l_ptr wqa sp pc ax bx cx HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_sleep_3 in HS.
  simplS in * .
  destruct HS as [HM [HCX [HSP [Hsp [Hif [Hisr Hwqa]]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma NH_WFcmd_sleep_4_pushwp :
  NH_WFcmd (SCAP_consCT 
    (L_sleep+4+1)
      (Ast L_sleep_5_p L_sleep_5_g tag_sleep) psi_sleep)
    (L_sleep+4) 
      (Ast L_sleep_4_p L_sleep_4_g tag_sleep) 
    (comml (_pushwp (Opr_reg DI))).
Proof.
  simpl.
  split.
  exists L_sleep_5_p; exists L_sleep_5_g.
  split_l.
  assert (Hpp' : forall l_ptr wqa sp pc ax bx cx dx mP S S',
    Ast_sleep_4 l_ptr wqa sp pc ax bx cx dx mP S
    -> nextS (L_sleep+4) (comml (_pushwp (Opr_reg DI))) S S'
    -> Ast_sleep_5 l_ptr wqa sp pc ax bx cx dx (S_RF S DI) mP S').
    intros l_ptr wqa sp pc ax bx cx dx mP S S' HS HSS'.
    unfold Ast_sleep_3, Ast_sleep_4 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspgt [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simplS in * .
    destruct HS as [HM [Hdx [Hsp [Hspgt' [Hif Hisr]]]]].
    split.
      subst M' R'.
      rewrite Hsp.
      sep on.
      sep lift 3.
      arith_rep (sp - 8 - 2) with (sp - 10).
      reg simpl.
      simplS in * .
      sep upd.
      sep lift 2 in HM.
      sep stksplit in HM.
      arith_rep (sp - 8 - 2) with (sp - 10) in HM.
      arith_rep (_STK_SIZE_sleep - 8 - 2) with (_STK_SIZE_sleep - 10) in HM.
      sep matchH HM with 8.
    subst R' fl' isr'; reg simpl.
    rewrite Hsp.
    arith_rep (sp - 8 - 2) with (sp - 10).
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_sleep_5_p in HpS.
    destruct HpS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx HS]]]]]]]].
    exists (MemUpdW M (sp-10) (R DI) 0, RF_update R SP (sp - 10), fl, isr).
    apply nextS_pushwp with  (R DI); trivial.
    unfold pushW.
    destruct HS as [HM [Hdx [Hsp [Hspgt [Hif Hisr]]]]].
    simplS in * .
    split.
      apply stkSpace_weaken with (_STK_SIZE_sleep - 8);
        [idtac | omega].
      sep on.
      sep lift 2 in HM.
      rewrite Hsp.
      sep trivial.
    split.
      rewrite Hsp. 
      arith_rep (sp - 8 - 2) with (sp - 10).
      eapply MemUpdW_MemUpd16; eauto.
    rewrite Hsp.
    arith_rep (sp - 8 - 2) with (sp - 10).
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_sleep_4_p, L_sleep_5_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx HS]]]]]]]].
    exists l_ptr; exists wqa; exists sp; exists pc; exists ax; exists bx;
      exists cx; exists dx; exists (S_RF S DI).
    apply Hpp'; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_sleep_4_g, L_sleep_5_g in * .
  intros mP l_ptr wqa sp pc ax bx cx dx HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [_ [Hdx [Hsp _]]].
  simplS in * .
  subst dx sp R' fl'.
  reg simpl.
  repeat split_l.
  trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_sleep_4_p in HS.
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx HS]]]]]]]].
    unfold Ast_sleep_4 in HS.
    destruct HS as [HM [HDX [HSP [Hsp [Hif [Hisr Hwqa]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_sleep_4_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP l_ptr wqa sp pc ax bx cx dx HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_sleep_4 in HS.
  simplS in * .
  destruct HS as [HM [HDX [HSP [Hsp [Hif [Hisr Hwqa]]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma NH_WFcmd_sleep_5_pushwp :
  NH_WFcmd (SCAP_consCT 
    (L_sleep+5+1)
      (Ast L_sleep_6_p L_sleep_6_g tag_sleep) psi_sleep)
    (L_sleep+5) 
      (Ast L_sleep_5_p L_sleep_5_g tag_sleep) 
    (comml (_pushwp (Opr_reg SI))).
Proof.
  simpl.
  split.
  exists L_sleep_6_p; exists L_sleep_6_g.
  split_l.
  assert (Hpp' : forall l_ptr wqa sp pc ax bx cx dx di mP S S',
    Ast_sleep_5 l_ptr wqa sp pc ax bx cx dx di mP S
    -> nextS (L_sleep+5) (comml (_pushwp (Opr_reg SI))) S S'
    -> Ast_sleep_6 l_ptr wqa sp pc ax bx cx dx di (S_RF S SI) mP S').
    intros l_ptr wqa sp pc ax bx cx dx di mP S S' HS HSS'.
    unfold Ast_sleep_4, Ast_sleep_5 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspgt [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simplS in * .
    destruct HS as [HM [Hdi [Hsp [Hspgt' [Hif Hisr]]]]].
    split.
      subst M' R'.
      rewrite Hsp.
      sep on.
      sep lift 3.
      arith_rep (sp - 10 - 2) with (sp - 12).
      reg simpl.
      simplS in * .
      sep upd.
      sep lift 2 in HM.
      sep stksplit in HM.
      arith_rep (sp - 10 - 2) with (sp - 12) in HM.
      arith_rep (_STK_SIZE_sleep - 10 - 2) with (_STK_SIZE_sleep - 12) in HM.
      sep matchH HM with 9.
    subst R' fl' isr'; reg simpl.
    rewrite Hsp.
    arith_rep (sp - 10 - 2) with (sp - 12).
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_sleep_6_p in HpS.
    destruct HpS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di HS]]]]]]]]].
    exists (MemUpdW M (sp-12) (R SI) 0, RF_update R SP (sp - 12), fl, isr).
    apply nextS_pushwp with  (R SI); trivial.
    unfold pushW.
    destruct HS as [HM [Hdi [Hsp [Hspgt [Hif Hisr]]]]].
    simplS in * .
    split.
      apply stkSpace_weaken with (_STK_SIZE_sleep - 10);
        [idtac | omega].
      sep on.
      sep lift 2 in HM.
      rewrite Hsp.
      sep trivial.
    split.
      rewrite Hsp. 
      arith_rep (sp - 10 - 2) with (sp - 12).
      eapply MemUpdW_MemUpd16; eauto.
    rewrite Hsp.
    arith_rep (sp - 10 - 2) with (sp - 12).
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_sleep_5_p, L_sleep_6_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di HS]]]]]]]]].
    exists l_ptr; exists wqa; exists sp; exists pc; exists ax; exists bx;
      exists cx; exists dx; exists di; exists (S_RF S SI).
    apply Hpp'; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_sleep_5_g, L_sleep_6_g in * .
  intros mP l_ptr wqa sp pc ax bx cx dx di HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ _ _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [_ [Hdi [Hsp _]]].
  simplS in * .
  subst di sp R' fl'.
  reg simpl.
  repeat split_l.
  trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_sleep_5_p in HS.
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di HS]]]]]]]]].
    unfold Ast_sleep_5 in HS.
    destruct HS as [HM [HDI [HSP [Hsp [Hif [Hisr Hwqa]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_sleep_5_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP l_ptr wqa sp pc ax bx cx dx di HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_sleep_5 in HS.
  simplS in * .
  destruct HS as [HM [HDI [HSP [Hsp [Hif [Hisr Hwqa]]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma NH_WFcmd_sleep_6_pushwp :
  NH_WFcmd (SCAP_consCT 
    (L_sleep+6+1)
      (Ast L_sleep_7_p L_sleep_7_g tag_sleep) psi_sleep)
    (L_sleep+6) 
      (Ast L_sleep_6_p L_sleep_6_g tag_sleep) 
    (comml (_pushwp (Opr_reg BP))).
Proof.
  simpl.
  split.
  exists L_sleep_7_p; exists L_sleep_7_g.
  split_l.
  assert (Hpp' : forall l_ptr wqa sp pc ax bx cx dx di si mP S S',
    Ast_sleep_6 l_ptr wqa sp pc ax bx cx dx di si mP S
    -> nextS (L_sleep + 6) (comml (_pushwp (Opr_reg BP))) S S'
    -> Ast_sleep_7 l_ptr wqa sp pc ax bx cx dx di si (S_RF S BP) mP S').
    intros l_ptr wqa sp pc ax bx cx dx di si mP S S' HS HSS'.
    unfold Ast_sleep_5, Ast_sleep_6 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspgt [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simplS in * .
    destruct HS as [HM [Hsi [Hsp [Hspgt' [Hif Hisr]]]]].
    split.
      subst M' R'.
      rewrite Hsp.
      sep on.
      sep lift 3.
      arith_rep (sp - 12 - 2) with (sp - 14).
      reg simpl.
      simplS in * .
      sep upd.
      sep lift 2 in HM.
      sep stksplit in HM.
      arith_rep (sp - 12 - 2) with (sp - 14) in HM.
      arith_rep (_STK_SIZE_sleep - 12 - 2) with (_STK_SIZE_sleep - 14) in HM.
      sep matchH HM with 10.
    subst R' fl' isr'; reg simpl.
    rewrite Hsp.
    arith_rep (sp - 12 - 2) with (sp - 14).
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_sleep_6_p in HpS.
    destruct HpS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si HS]]]]]]]]]].
    exists (MemUpdW M (sp-14) (R BP) 0, RF_update R SP (sp - 14), fl, isr).
    apply nextS_pushwp with  (R BP); trivial.
    unfold pushW.
    destruct HS as [HM [Hbp [Hsp [Hspgt [Hif Hisr]]]]].
    simplS in * .
    split.
      apply stkSpace_weaken with (_STK_SIZE_sleep - 12);
        [idtac | omega].
      sep on.
      sep lift 2 in HM.
      rewrite Hsp.
      sep trivial.
    split.
      rewrite Hsp.
      arith_rep (sp - 12 - 2) with (sp - 14).
      eapply MemUpdW_MemUpd16; eauto.
    rewrite Hsp.
    arith_rep (sp - 12 - 2) with (sp - 14).
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_sleep_6_p, L_sleep_7_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si HS]]]]]]]]]].
    exists l_ptr; exists wqa; exists sp; exists pc; exists ax; exists bx;
      exists cx; exists dx; exists di; exists si;
        exists (S_RF S BP).
    apply Hpp'; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_sleep_6_g, L_sleep_7_g in * .
  intros mP l_ptr wqa sp pc ax bx cx dx di si HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ _ _ _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [_ [Hsi [Hsp _]]].
  simplS in * .
  subst si sp R' fl'.
  reg simpl.
  repeat split_l.
  trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_sleep_6_p in HS.
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si HS]]]]]]]]]].
    unfold Ast_sleep_6 in HS.
    destruct HS as [HM [HSI [HSP [Hsp [Hif [Hisr Hwqa]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_sleep_6_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP l_ptr wqa sp pc ax bx cx dx di si HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_sleep_6 in HS.
  simplS in * .
  destruct HS as [HM [HSI [HSP [Hsp [Hif [Hisr Hwqa]]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma NH_WFcmd_sleep_7_pushfw :
  NH_WFcmd (SCAP_consCT 
    (L_sleep+7+1)
      (Ast L_sleep_8_p L_sleep_8_g tag_sleep) psi_sleep)
    (L_sleep+7) 
      (Ast L_sleep_7_p L_sleep_7_g tag_sleep) 
    (comml (_pushfw)).
Proof.
  simpl.
  split.
  exists L_sleep_8_p; exists L_sleep_8_g.
  split_l.
  assert (Hpp' : forall l_ptr wqa sp pc ax bx cx dx di si bp mP S S',
    Ast_sleep_7 l_ptr wqa sp pc ax bx cx dx di si bp mP S
    -> nextS (L_sleep+7) (comml (_pushfw)) S S'
    -> Ast_sleep_8 l_ptr wqa sp pc ax bx cx dx di si bp (FR2nat (S_Flag S)) mP S').
    intros l_ptr wqa sp pc ax bx cx dx di si bp mP S S' HS HSS'.
    unfold Ast_sleep_7, Ast_sleep_8 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspgt [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simplS in * .
    destruct HS as [HM [Hbp [Hsp [Hspgt' [Hif Hisr]]]]].
    split.
      subst M' R'.
      sep on.
      sep lift 3.
      rewrite Hsp.
      arith_rep (sp - 14 - 2) with (sp - 16).
      sep upd.
      sep lift 2 in HM.
      sep stksplit in HM.
      arith_rep (sp - 14 - 2) with (sp - 16) in HM.
      arith_rep (_STK_SIZE_sleep - 14 - 2) with (_STK_SIZE_sleep - 16) in HM.
      sep matchH HM with 11.
    subst R' fl' isr' bp; reg simpl.
    rewrite Hsp.
    split_l by omega.
    split_l.
    rewrite FR_nat_iso_left.
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_sleep_7_p in HpS.
    destruct HpS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si [bp HS]]]]]]]]]]].
    exists (MemUpdW M (sp-16) (FR2nat fl) 0, RF_update R SP (sp - 16), fl, isr).
    apply nextS_pushfw; trivial.
    unfold pushW.
    destruct HS as [HM [Hbp [Hsp [Hspgt [Hif Hisr]]]]].
    simplS in * .
    rewrite Hsp.
    arith_rep (sp-14-2) with (sp - 16).
    split.
      apply stkSpace_weaken with (_STK_SIZE_sleep - 14);
        [idtac | omega].
      sep on.
      sep lift 2 in HM.
      sep trivial.
    split.
      apply MemUpdW_MemUpd16 with 0; trivial.
    trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_sleep_7_p, L_sleep_8_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si [bp HS]]]]]]]]]]].
    exists l_ptr; exists wqa; exists sp; exists pc; exists ax; exists bx; exists cx; 
      exists dx; exists di; exists si; exists bp; exists (FR2nat (S_Flag S)).
    apply Hpp'; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_sleep_7_g, L_sleep_8_g in * .
  intros mP l_ptr wqa sp pc ax bx cx dx di si bp HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ _ _ _ _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [_ [Hbp [Hsp _]]].
  simplS in * .
  subst ax bx cx dx di si sp bp R' fl'.
  repeat split_l; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_sleep_7_p in HS.
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si [bp HS]]]]]]]]]]].
    unfold Ast_sleep_7 in HS.
    destruct HS as [HM [HBP [HSP [Hsp [Hif [Hisr Hwqa]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_sleep_7_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP l_ptr wqa sp pc ax bx cx dx di si bp HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_sleep_7 in HS.
  simplS in * .
  destruct HS as [HM [HBP [HSP [Hsp [Hif [Hisr Hwqa]]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

(* <sleep + 8> :  _movwp (Opr_reg SP) BP; *)

Lemma NH_WFcmd_sleep_8_movwp :
  NH_WFcmd (SCAP_consCT 
    (L_sleep+8+1)
      (Ast L_sleep_9_p L_sleep_9_g tag_sleep) psi_sleep)
    (L_sleep+8) 
      (Ast L_sleep_8_p L_sleep_8_g tag_sleep) 
    (comml (_movwp (Opr_reg SP) BP)).
Proof.
  simpl.
  split.
  exists L_sleep_9_p; exists L_sleep_9_g.
  split_l.
  assert (Hpp' : forall l_ptr wqa sp pc ax bx cx dx di si bp flnum mP S S',
    Ast_sleep_8 l_ptr wqa sp pc ax bx cx dx di si bp flnum mP S
    -> nextS (L_sleep+8) (comml (_movwp (Opr_reg SP) BP)) S S'
    -> Ast_sleep_9 l_ptr wqa sp pc ax bx cx dx di si bp flnum mP S').
    intros l_ptr wqa sp pc ax bx cx dx di si bp flnum mP S S' HS HSS'.
    unfold Ast_sleep_8, Ast_sleep_9 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    simpl in HR'.
    simplS in * .
    destruct HS as [HM [Hsp [Hspgt' [Hfl [Hif [Hisr Hwqa]]]]]].
    split.
      subst M' R' fl' isr'.
      sep on.
      sep matchH HM with 10.
    subst fl R' fl' isr'; reg simpl.
    rewrite Hsp.
    repeat split_l; trivial.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_sleep_8_p in HpS.
    destruct HpS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si [bp [flnum HS]]]]]]]]]]]].
    exists (M, RF_update R BP (R SP), fl, isr).
    apply nextS_movwp with (R SP); trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_sleep_8_p, L_sleep_9_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si [bp [flnum HS]]]]]]]]]]]].
    exists l_ptr; exists wqa; exists sp; exists pc; exists ax; exists bx; exists cx; 
      exists dx; exists di; exists si; exists bp; exists flnum.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_sleep_8_g, L_sleep_9_g in * .
  intros mP l_ptr wqa sp pc ax bx cx dx di si bp flnum HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ _ _ _ _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  repeat split_l; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_sleep_8_p in HS.
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si [bp [fln HS]]]]]]]]]]]].
    unfold Ast_sleep_8 in HS.
    destruct HS as [HM [HSP [Hsp [Hfln [Hif [Hisr Hwqa]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_sleep_8_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP l_ptr wqa sp pc ax bx cx dx di si bp fln HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_sleep_8 in HS.
  simplS in * .
  destruct HS as [HM [HSP [Hsp [Hfln [Hif [Hisr Hwqa]]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

(* <sleep + 9> :  _pushwm (A_reg 18 BP); *)

Lemma NH_WFcmd_sleep_9_pushwm :
  NH_WFcmd (SCAP_consCT 
    (L_sleep+9+1)
      (Ast L_sleep_10_p L_sleep_10_g tag_sleep) psi_sleep)
    (L_sleep+9) 
      (Ast L_sleep_9_p L_sleep_9_g tag_sleep) 
    (comml (_pushwm (A_reg 18 BP))).
Proof.
  simpl.
  split.
  exists L_sleep_10_p; exists L_sleep_10_g.
  split_l.
  assert (Hpp' : forall l_ptr wqa sp pc ax bx cx dx di si bp flnum mP S S',
    Ast_sleep_9 l_ptr wqa sp pc ax bx cx dx di si bp flnum mP S
    -> nextS (L_sleep+9) (comml (_pushwm (A_reg 18 BP))) S S'
    -> Ast_sleep_10 l_ptr wqa sp pc ax bx cx dx di si bp flnum mP S').
    intros l_ptr wqa sp pc ax bx cx dx di si bp flnum mP S S' HS HSS'.
    unfold Ast_sleep_9, Ast_sleep_10 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [l_ptrx [Hlds [Hspgt [HldsB HM']]]].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM', Hlds.
    simplS in * .
    destruct HS as [HM [Hbp [Hsp [Hspgt' [Hfl [Hif Hisr]]]]]].
    split.
      subst M' R' fl' isr'.
      assert (Heq : l_ptr = l_ptrx).
        rewrite Hbp in Hlds.
        rewrite Hsp in Hlds.
        arith_rep (sp - 16 + 18) with (sp + 2) in Hlds.
        sep on.
        sep lift 12 in HM.
        sep inj.
        auto.
      subst l_ptrx.
      rewrite Hsp.
      arith_rep (sp - 16 - 2) with (sp - 18).
      sep on.
      sep lift 3.
      sep upd.
      sep lift 2 in HM.
      sep stksplit in HM.
      arith_rep (sp - 16 - 2) with (sp - 18) in HM.
      arith_rep (_STK_SIZE_sleep - 16 - 2) with (_STK_SIZE_sleep - 18) in HM.
      sep matchH HM with 12.
    subst R' fl' isr'; reg simpl.
    rewrite Hsp.
    split_l by omega.
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_sleep_9_p in HpS.
    destruct HpS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si [bp [flnum HS]]]]]]]]]]]].
    destruct HS as [HM [Hbp [Hsp [Hspgt' [Hfl [Hif Hisr]]]]]].
    simplS in * .
    exists (MemUpdW M (sp-18) l_ptr 0, RF_update R SP (sp - 18), fl, isr).
    apply nextS_pushwm with (sp + 2) l_ptr; trivial.
      simpl.
      rewrite Hbp.
      rewrite Hsp.
      omega.
      sep on.
      sep lift 12 in HM.
      sep trivial.
    unfold pushW.
    rewrite Hsp.
    arith_rep (sp-16-2) with (sp - 18).
    split.
      apply stkSpace_weaken with (_STK_SIZE_sleep - 16);
        [idtac | omega].
      sep on.
      sep lift 2 in HM.
      sep trivial.
    split.
      apply MemUpdW_MemUpd16 with 0; trivial.
    trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_sleep_9_p, L_sleep_10_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si [bp [flnum HS]]]]]]]]]]]].
    exists l_ptr; exists wqa; exists sp; exists pc; exists ax; exists bx; exists cx; 
      exists dx; exists di; exists si; exists bp; exists flnum.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_sleep_9_g, L_sleep_10_g in * .
  intros mP l_ptr wqa sp pc ax bx cx dx di si bp flnum HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ _ _ _ _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [_ [Hbp [Hsp _]]].
  simplS in * .
  subst ax bx cx dx di si sp bp R' fl'.
  repeat split_l; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_sleep_9_p in HS.
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si [bp [fln HS]]]]]]]]]]]].
    unfold Ast_sleep_9 in HS.
    destruct HS as [HM [HBP [HSP [Hsp [Hfln [Hif [Hisr Hwqa]]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_sleep_9_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP l_ptr wqa sp pc ax bx cx dx di si bp fln HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_sleep_9 in HS.
  simplS in * .
  destruct HS as [HM [HBP [HSP [Hsp [Hfln [Hif [Hisr Hwqa]]]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

(* <sleep + 10>:  _call L_block; *)

Lemma NH_WFcmd_sleep_10_call :
  NH_WFcmd (SCAP_consCT 
    (L_sleep+10+1)
      (Ast L_sleep_11_p L_sleep_11_g tag_sleep) psi_sleep)
    (L_sleep+10) 
      (Ast L_sleep_10_p L_sleep_10_g tag_sleep) 
    (comml (_call _BLOCK)).
Proof.
  simpl.
  split.
  assert (Hpp''' : forall l_ptr wqa sp pc ax bx cx dx di si bp flnum
    mP S S' S'' S''' pc',
    Ast_sleep_10 l_ptr wqa sp pc ax bx cx dx di si bp flnum mP S
    -> nextS (L_sleep+10) (comml (_call _BLOCK)) S S'
    -> (gblk wqa) S' S''
    -> nextS pc' (comms _retw) S'' S'''
    -> Ast_sleep_11 l_ptr wqa sp pc ax bx cx dx di si bp flnum mP S''').
    intros l_ptr wqa sp pc ax bx cx dx di si bp flnum mP.
    intros S S' S'' S''' pc'.
    intros HS HSS' HS'S'' HS''S'''.
    unfold Ast_sleep_10, Ast_sleep_11 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct S'' as [[[M'' R''] fl''] isr''].
    destruct S''' as [[[M''' R'''] fl'''] isr'''].
    simplS in * .
    destruct HS as [HM [Hsp [Hspg [Hfl [Hif [Hisr Hwqa]]]]]].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspg' [Hlds Hupd]].
    unfold gblk in HS'S''.
    destruct HS'S'' as [sp' [HM'' [Hif'' [Hisr'' HSP]]]].
    destruct (nextS_inversion HS''S''') as [HM''' [HR''' [Hfl''' Hisr''']]].
    destruct HR''' as [pc''' [Hlds''' HR''']].
    simplS in * .
    subst M''' R''' fl''' isr'''.
    subst isr'' R' fl' isr'.
    clear HSS' HS''S'''.
    reg simpl.
    destruct HSP as [Hsp' HSP].
    subst sp'.
    rewrite <- HSP.
    rewrite Hsp.
    rewrite Hsp in HM''.
    split.
      sep on.
      sep lift 3.
      apply stkSpacePrec_mergeW1; [idtac | omega | omega ].
      arith_rep (18 + 2 + _BLK_Stk - 18 - 2) with _BLK_Stk.
      assert (Hx : M''
        ||- (INV0 *^* INV1 *^* wqa *^* stkSpacePrec (sp - 18 - 2) (_BLK_Stk + 2 - 2)) *^*
          (sp - 18 - 2 |->w (L_sleep + 10 + 1) *^*
           sp - 18 |->w l_ptr *^*
           sp - 16 |->w flnum *^*
           sp - 14 |->w bp *^*
           sp - 12 |->w si *^*
           sp - 10 |->w di *^*
          sp - 8 |->w dx *^*
          sp - 6 |->w cx *^*
          sp - 4 |->w bx *^* 
          sp - 2 |->w ax *^* 
          sp |->w pc *^* 
          sp + 2 |->w l_ptr *^* mP)).
        replace M'' with (S_Mem (M'', R'', fl'', false));
          trivial.
        sep off.
        assert (forall sp size, precise (INV0 *^* INV1 *^* stkSpacePrec sp size)).
          intros spx sizex.
          apply mstar_presv_prec.
          apply Prec01.
          apply precise_stkSpacePrec.
        apply prepost_elim with (INV0 *^* INV1 *^* stkSpacePrec (sp - 18 - 2)  (_BLK_Stk + 2 - 2))
          (M', RF_update R SP (sp - 18 - 2), fl, isr);
          auto.
        simplS.
        substH Hupd with (MemUpd16_MemUpdW Hupd).
        destruct Hupd as [x HM'].
        rewrite Hsp in HM'.
        subst M'.
        sep on.
        sep lift 3.
        sep upd.
        arith_rep (18+2+_BLK_Stk-18) with (_BLK_Stk + 2) in HM.
        sep matchH HM with 11.
        sep stksplit in HM.
        sep matchH HM with 2.
      arith_rep (_BLK_Stk + 2 - 2) with _BLK_Stk in Hx.
      sep matchH Hx with 13.
    split_l by omega.
    rewrite <- Hif''.
    auto.
  left.
  split.
    unfold prim_entry.
    right; left; trivial.
  split.
    constructor 2 with L_sleep_10_p L_sleep_11_p 
      L_sleep_10_g L_sleep_11_g tag_sleep; trivial.
    (* enable *)
    unfold enable_blk.
    intros S HS.
    unfold L_sleep_10_p in HS.
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si [bp [flnum HS]]]]]]]]]]]].
    exists l_ptr; exists wqa.
    split.
      destruct HS as [HM [Hsp [Hspg [Hfl [Hif [Hisr Hwqa]]]]]].
      split_l.
      unfold pTrue, pmstar.
      destruct S as [[[M R] fl] isr].
      simplS in * .
      rewrite Hsp.
      assert (HMx : (mTrue *^* (sp - 18 |->w l_ptr *^* 
        (INV0 *^* INV1 *^* stkSpacePrec (sp - 18) (18 + 2 + _BLK_Stk - 18)))) M).
        sep on.
        sep matchH HM with 12.
      arith_rep (18 + 2 + _BLK_Stk - 18) with (_BLK_Stk + 2) in HMx.
  split.
    (* p ==> p' *)
    intros pc' S'''.
    unfold L_sleep_10_p, L_sleep_11_p.
    unfold wrap_fun.
    unfold linkPG, wrap_fun, conjG.
    intros H.
    destruct H as [S'' [H HS''S''']].
    destruct H as [S' [HSS' HS'S'']].
    exists l_ptr; exists wqa; exists sp; exists pc; exists ax; exists bx; 
      exists cx; exists dx; exists di; exists si; exists bp; exists flnum.
    eapply Hpp'''; eauto.
    (* g''' ==> g *)
    intros pc' S''''.
    unfold L_sleep_10_g, L_sleep_11_g.
    unfold wrap_fun.
    intros HSS''''.
    unfold conjG at 1 in HSS'''' .
    destruct HSS'''' as [S''' [HSS''' H]].
    destruct HSS''' as [S'' [HSS'' HS''S''']].
    destruct HSS'' as [S' [HSS' HS'S'']].
    assert (Hwqa : Delta l_ptr = SomeT wqa).
      destruct HS as [HM [Hsp [Hspg [Hfl [Hif [Hisr Hwqa]]]]]].
      trivial.
    assert (Hlds : leadsToW (S_RF S SP) l_ptr (S_Mem S)).
      destruct HS as [HM [Hsp [Hspg [Hfl [Hif [Hisr _]]]]]].
      rewrite Hsp.
      sep on.
      sep lift 3 in HM.
      sep trivial.
    clear sp pc ax bx cx dx di si bp flnum HS.
    intros mP l_ptrx wqax sp pc ax bx cx dx di si bp flnum HS.
    assert (Heq : l_ptrx = l_ptr).
      destruct HS as [HM [Hsp [Hspg [Hfl [Hif [Hisr Hwqax]]]]]].
      rewrite Hsp in Hlds.
      sep on.
      sep lift 3 in HM.
      sep inj.
      auto.
    subst l_ptrx.
    assert (Heq : wqax = wqa).
      destruct HS as [HM [Hsp [Hspg [Hfl [Hif [Hisr Hwqax]]]]]].
      rewrite Hwqax in Hwqa.
      injection Hwqa; trivial.
    subst wqax.
    apply H with flnum.
    eapply Hpp'''; eauto.
    (* p ==> if=false *)
    intros S HS.
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si [bp [flnum HS]]]]]]]]]]]].
    unfold Ast_sleep_10 in HS.
    destruct HS as [HM [Hsp [Hspg [Hif [Hfl [Hisr Hwqa]]]]]].
    trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_sleep_10_p in HS.
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si [bp [fln HS]]]]]]]]]]]].
    unfold Ast_sleep_10 in HS.
    destruct HS as [HM [HSP [Hsp [Hfln [Hif [Hisr Hwqa]]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_sleep_10_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP l_ptr wqa sp pc ax bx cx dx di si bp fln HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_sleep_10 in HS.
  simplS in * .
  destruct HS as [HM [HSP [Hsp [Hfln [Hif [Hisr Hwqa]]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

(* <sleep + 11>:  _addw   (Opr_imm 2) SP; *)

Lemma NH_WFcmd_sleep_11_addw :
  NH_WFcmd (SCAP_consCT 
    (L_sleep+11+1)
      (Ast L_sleep_12_p L_sleep_12_g tag_sleep) psi_sleep)
    (L_sleep+11)
      (Ast L_sleep_11_p L_sleep_11_g tag_sleep) 
    (comml (_addw (Opr_imm 2) SP)).
Proof.
  simpl.
  split.
  exists L_sleep_12_p; exists L_sleep_12_g.
  split_l.
  assert (Hpp' : forall l_ptr wqa sp pc ax bx cx dx di si bp flnum mP S S',
    Ast_sleep_11 l_ptr wqa sp pc ax bx cx dx di si bp flnum mP S
    -> nextS (L_sleep+11) (comml (_addw (Opr_imm 2) SP)) S S'
    -> Ast_sleep_12 l_ptr wqa sp pc ax bx cx dx di si bp flnum mP S').
    intros l_ptr wqa sp pc ax bx cx dx di si bp flnum mP S S' HS HSS'.
    unfold Ast_sleep_11, Ast_sleep_12 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    clear HSS'.
    simplS in * .
    destruct HS as [HM [Hsp [Hspgt' [Hflnum [Hif Hisr]]]]].
    subst M' R' isr'.
    split.
      sep on.
      sep matchH HM with 13.
      sep stksplit.
      arith_rep (sp - 16 - 2) with (sp - 18).
      arith_rep (_STK_SIZE_sleep - 16 - 2) with (_STK_SIZE_sleep - 18).
      sep matchH HM with 2.
    unfold opr_eval.
    rewrite Hsp.
    reg simpl.
    split_l by omega.
    unfold trash_fl in Hfl'.
    destruct Hfl'; subst fl'; auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_sleep_11_p in HpS.
    destruct HpS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si [bp [flnum HS]]]]]]]]]]]].
    destruct HS as [HM [Hsp [Hspgt' [Hflnum [Hif Hisr]]]]].
    simplS in * .
    exists (M, RF_update R SP (sp - 16), Clear_Flag fl _ZF, isr).
    apply nextS_addw with 2 (sp - 18); trivial.
      rewrite Hsp; trivial.
      arith_rep (2 + (sp - 18)) with (sp - 16).
      right; trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_sleep_11_p, L_sleep_12_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si [bp [flnum HS]]]]]]]]]]]].
    exists l_ptr; exists wqa; exists sp; exists pc; exists ax; exists bx; 
      exists cx; exists dx; exists di; exists si; exists bp; exists flnum.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_sleep_11_g, L_sleep_12_g in * .
  intros mP l_ptr wqa sp pc ax bx cx dx di si bp flnum HS.
  apply HgS'S'' with flnum.
  apply Hpp' with S; trivial.
  trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_sleep_11_p in HS.
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si [bp [fln HS]]]]]]]]]]]].
    unfold Ast_sleep_11 in HS.
    destruct HS as [HM [HSP [Hsp [Hfln [Hif Hisr]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_sleep_11_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP l_ptr wqa sp pc ax bx cx dx di si bp fln HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_sleep_11 in HS.
  simplS in * .
  destruct HS as [HM [HSP [Hsp [Hfln [Hif Hisr]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

(* <sleep + 12>:  _popfw; *)

Lemma NH_WFcmd_sleep_12_popfw :
  NH_WFcmd (SCAP_consCT 
    (L_sleep+12+1)
      (Ast L_sleep_13_p L_sleep_13_g tag_sleep) psi_sleep)
    (L_sleep+12)
      (Ast L_sleep_12_p L_sleep_12_g tag_sleep) 
    (comml (_popfw)).
Proof.
  simpl.
  split.
  exists L_sleep_13_p; exists L_sleep_13_g.
  split_l.
  assert (Hpp' : forall l_ptr wqa sp pc ax bx cx dx di si bp flnum mP S S',
    Ast_sleep_12 l_ptr wqa sp pc ax bx cx dx di si bp flnum mP S
    -> stk_IF_cl S
    -> gpopfw_cli S S'
    -> Ast_sleep_13 l_ptr wqa sp pc ax bx cx dx di si bp mP S').
    intros l_ptr wqa sp pc ax bx cx dx di si bp flnum mP S S' HS Hcl HSS'.
    unfold Ast_sleep_12, Ast_sleep_13, stk_IF_cl in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    unfold gpopfw_cli in HSS'.
    simplS in * .
    destruct HSS' as [Hshift [Hif' [Hisr' HR']]].
    assert (HM' : M' = M).
      cut (forall M0 : Mem,
            choice (itr_enabled (M, R, fl, isr)) (INV0 *^* INV1) Aemp M0 ->
            Aemp M0).
      intros Hx.
      substH Hshift with (prepost_weaken_post (mQ':=Aemp) Hshift Hx).
      apply (prepost_elim_emp_M _ _ _ _ _ _ _ _ Hshift).
      intros.
      apply (choice_elim_right H).
      intros [H1 H2].
      simplS in * .
      destruct HS as [HM [Hsp [Hspgt' [Hflnum [Hif Hisr]]]]].
      rewrite Hif in H1; discriminate H1.
    clear Hshift.
    subst M' R' isr'.
    destruct HS as [HM [Hsp [Hspg [Hif [Hfl Hisr]]]]].
    split.
      sep on.
      sep matchH HM with 11.
      sep stksplit.
      arith_rep (sp - 14 - 2) with (sp - 16).
      arith_rep (_STK_SIZE_sleep - 14 - 2) with (_STK_SIZE_sleep - 16).
      sep matchH HM with 2.
    rewrite Hsp.
    reg simpl.
    split_l by omega.
    auto.
  assert (Hpp'_sti : forall l_ptr wqa sp pc ax bx cx dx di si bp flnum mP S S',
    Ast_sleep_12 l_ptr wqa sp pc ax bx cx dx di si bp flnum mP S
    -> stk_IF_st S
    -> gpopfw_sti S S'
    -> Ast_sleep_13 l_ptr wqa sp pc ax bx cx dx di si bp mP S').
    clear Hpp'.
    intros l_ptr wqa sp pc ax bx cx dx di si bp flnum mP S S' HS Hst HSS'.
    unfold stk_IF_st in * .
    destruct Hst as [flnumx [Hlds Hif]].
    unfold Ast_sleep_12 in HS.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct HS as [HM [Hsp [Hspg [Hifx [Hfl Hisr]]]]].
    simplS in * .
    assert (Heq : flnum = flnumx).
      sep on.
      sep lift 4 in HM.
      rewrite Hsp in Hlds.
      sep inj.
      auto.
    subst flnumx.
    rewrite Hifx in Hif; discriminate Hif.
  split.
  (* enable popfw *)
    clear Hpp' Hpp'_sti.
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_sleep_12_p in HpS.
    destruct HpS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si [bp [flnum HS]]]]]]]]]]]].
    exists (M, RF_update R SP (sp - 14),(nat2FR flnum), isr).
    apply nextS_popfw with flnum; trivial.
    unfold popW.
    destruct HS as [HM [Hsp [Hspgt [Hif [Hfl Hisr]]]]].
    simplS in * .
    rewrite Hsp.
    split.
      sep on.
      sep lift 4 in HM.
      sep trivial.
    arith_rep (sp-16+2) with (sp - 14).
  split.
  (* enable sti *)
    unfold enable_popfw_sti, stk_IF_st, conjP.
    intros M R fl isr HS.
    destruct HS as [HS Hflnum].
    unfold L_sleep_10_p in HS.
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si [bp [flnum HS]]]]]]]]]]]].
    destruct Hflnum as [flnumx [Hlds Hif]].
    unfold Ast_sleep_12 in HS. 
    destruct HS as [HM [Hsp [Hspg [Hifx [Hfl Hisr]]]]].
    assert (Heq : flnum = flnumx).
      sep on.
      sep lift 4 in HM.
      rewrite Hsp in Hlds.
      sep inj.
      auto.
    subst flnumx.
    rewrite Hifx in Hif; discriminate Hif.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_sleep_12_p, L_sleep_13_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HSS'].
    destruct HS as [HS Hst].
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si [bp [flnum HS]]]]]]]]]]]].
    exists l_ptr; exists wqa; exists sp; exists pc; exists ax; exists bx; exists cx; 
      exists dx; exists di; exists si; exists bp.
    eapply Hpp'_sti; eauto.
  split.
  (* g' ==> g *)
    intros S S'' H.
    unfold linkPG, conjPG, conjG in H.
    destruct H as [S' H].
    destruct H as [[HpS HnextSS'] HgS'S''].
    destruct HpS as [_ Hst].
    unfold L_sleep_12_g, L_sleep_13_g in * .
    intros mP l_ptr wqa sp pc ax bx cx dx di si bp flnum HS.
    apply HgS'S''.
    apply Hpp'_sti with flnum S; auto.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_sleep_10_p, L_sleep_11_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HSS'].
    destruct HS as [HS Hcl].
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si [bp [flnum HS]]]]]]]]]]]].
    exists l_ptr; exists wqa; exists sp; exists pc; exists ax; exists bx; exists cx; 
      exists dx; exists di; exists si; exists bp.
    eapply Hpp'; eauto.
  (* g' ==> g *)
    intros S S'' H.
    unfold linkPG, conjPG, conjG in H.
    destruct H as [S' H].
    destruct H as [[HpS HnextSS'] HgS'S''].
    destruct HpS as [_ Hst].
    unfold L_sleep_12_g, L_sleep_13_g in * .
    intros mP l_ptr wqa sp pc ax bx cx dx di si bp flnum HS.
    apply HgS'S''.
    apply Hpp' with flnum S; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_sleep_12_p in HS.
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si [bp [fln HS]]]]]]]]]]]].
    unfold Ast_sleep_12 in HS.
    destruct HS as [HM [HSP [Hsp [Hfln [Hif Hisr]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_sleep_12_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP l_ptr wqa sp pc ax bx cx dx di si bp fln HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_sleep_12 in HS.
  simplS in * .
  destruct HS as [HM [HSP [Hsp [Hfln [Hif Hisr]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma NH_WFcmd_sleep_13_popw :
  NH_WFcmd (SCAP_consCT 
    (L_sleep+13+1)
      (Ast L_sleep_14_p L_sleep_14_g tag_sleep) psi_sleep)
    (L_sleep+13)
      (Ast L_sleep_13_p L_sleep_13_g tag_sleep) 
    (comml (_popw BP)).
Proof.
  simpl.
  split.
  exists L_sleep_14_p; exists L_sleep_14_g.
  split_l.
  assert (Hpp' : forall l_ptr wqa sp pc ax bx cx dx di si bp mP S S',
    Ast_sleep_13 l_ptr wqa sp pc ax bx cx dx di si bp mP S
    -> nextS (L_sleep+13) (comml (_popw BP)) S S'
    -> Ast_sleep_14 l_ptr wqa sp pc ax bx cx dx di si mP S').
    intros l_ptr wqa sp pc ax bx cx dx di si bp mP S S' HS HSS'.
    unfold Ast_sleep_13, Ast_sleep_14 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HR' as [bpx [Hlds HR']].
    simplS in * .
    destruct HS as [HM [Hsp [Hspgt' [Hif Hisr]]]].
    assert (Heq : bp = bpx).
      sep on.
      sep lift 4 in HM.
      rewrite Hsp in Hlds.
      sep inj.
      auto.
    subst bpx.
    subst M' fl' R' isr'.
    split.
      sep on.
      sep matchH HM with 12.
      sep stksplit.
      arith_rep (sp - 12 - 2) with (sp-14).
      arith_rep (_STK_SIZE_sleep - 12 - 2) with (_STK_SIZE_sleep-14).
      sep matchH HM with 2.
    rewrite Hsp.
    reg simpl.
    split_l by omega.
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_sleep_13_p in HpS.
    destruct HpS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si [bp HS]]]]]]]]]]].
    exists (M, RF_update (RF_update R SP (sp - 12)) BP bp, fl, isr).
    apply nextS_popw with (RF_update R SP (sp - 12)) bp; trivial.
    unfold popW.
    destruct HS as [HM [Hsp [Hspgt [Hif Hisr]]]].
    simplS in * .
    rewrite Hsp.
    split.
      sep on.
      sep lift 4 in HM.
      sep trivial.
    arith_rep (sp - 14 + 2) with (sp - 12).
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_sleep_13_p, L_sleep_14_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si [bp HS]]]]]]]]]]].
    exists l_ptr; exists wqa; exists sp; exists pc; exists ax; exists bx; exists cx; 
      exists dx; exists di; exists si.
    apply Hpp' with bp S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_sleep_13_g, L_sleep_14_g in * .
  intros mP l_ptr wqa sp pc ax bx cx dx di si bp HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ _ _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [HM [Hsp _]].
  simplS in * .
  destruct HR' as [bpx [Hlds HR']].
  assert (Heq : bp = bpx).
    sep on.
    sep lift 4 in HM.
    rewrite Hsp in Hlds.
    sep inj.
    auto.
  subst bpx ax bx cx dx di si sp R' fl' isr'.
  reg simpl.
  repeat split_l; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_sleep_13_p in HS.
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si [bp HS]]]]]]]]]]].
    unfold Ast_sleep_13 in HS.
    destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_sleep_13_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP l_ptr wqa sp pc ax bx cx dx di si bp HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_sleep_13 in HS.
  simplS in * .
  destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

(* <sleep + 14>:  _popw           (SI); *)

Lemma NH_WFcmd_sleep_14_popw :
  NH_WFcmd (SCAP_consCT 
    (L_sleep+14+1)
      (Ast L_sleep_15_p L_sleep_15_g tag_sleep) psi_sleep)
    (L_sleep+14)
      (Ast L_sleep_14_p L_sleep_14_g tag_sleep) 
    (comml (_popw SI)).
Proof.
  simpl.
  split.
  exists L_sleep_15_p; exists L_sleep_15_g.
  split_l.
  assert (Hpp' : forall l_ptr wqa sp pc ax bx cx dx di si mP S S',
    Ast_sleep_14 l_ptr wqa sp pc ax bx cx dx di si mP S
    -> nextS (L_sleep+14) (comml (_popw SI)) S S'
    -> Ast_sleep_15 l_ptr wqa sp pc ax bx cx dx di mP S').
    intros l_ptr wqa sp pc ax bx cx dx di si mP S S' HS HSS'.
    unfold Ast_sleep_14, Ast_sleep_15 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HR' as [six [Hlds HR']].
    simplS in * .
    destruct HS as [HM [Hsp [Hspgt' [Hif Hisr]]]].
    assert (Heq : si = six).
      sep on.
      sep lift 4 in HM.
      rewrite Hsp in Hlds.
      sep inj.
      auto.
    subst six.
    subst M' fl' R' isr'.
    split.
      sep on.
      sep matchH HM with 10.
      sep stksplit.
      arith_rep (sp - 10 - 2) with (sp-12).
      arith_rep (_STK_SIZE_sleep - 10 - 2) with (_STK_SIZE_sleep-12).
      sep matchH HM with 2.
    rewrite Hsp.
    reg simpl.
    split_l by omega.
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_sleep_14_p in HpS.
    destruct HpS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si HS]]]]]]]]]].
    exists (M, RF_update (RF_update R SP (sp - 10)) SI si, fl, isr).
    apply nextS_popw with (RF_update R SP (sp - 10)) si; trivial.
    unfold popW.
    destruct HS as [HM [Hsp [Hspgt [Hif Hisr]]]].
    simplS in * .
    rewrite Hsp.
    split.
      sep on.
      sep lift 4 in HM.
      sep trivial.
    arith_rep (sp - 12 + 2) with (sp - 10).
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_sleep_12_p, L_sleep_13_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si HS]]]]]]]]]].
    exists l_ptr; exists wqa; exists sp; exists pc; exists ax; exists bx; exists cx; 
      exists dx; exists di.
    apply Hpp' with si S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_sleep_14_g, L_sleep_15_g in * .
  intros mP l_ptr wqa sp pc ax bx cx dx di si HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [HM [Hsp _]].
  simplS in * .
  destruct HR' as [six [Hlds HR']].
  assert (Heq : si = six).
    sep on.
    sep lift 4 in HM.
    rewrite Hsp in Hlds.
    sep inj.
    auto.
  subst six ax bx cx dx di sp R' fl' isr'.
  reg simpl.
  repeat split_l; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_sleep_13_p in HS.
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di [si HS]]]]]]]]]].
    unfold Ast_sleep_13 in HS.
    destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_sleep_13_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP l_ptr wqa sp pc ax bx cx dx di si HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_sleep_13 in HS.
  destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
  destruct IHS as [IHif IHisr].
  simplS in * .
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

(* <sleep + 15>:  _popw  (DI) *)

Lemma NH_WFcmd_sleep_15_popw :
  NH_WFcmd (SCAP_consCT 
    (L_sleep+15+1)
      (Ast L_sleep_16_p L_sleep_16_g tag_sleep) psi_sleep)
    (L_sleep+15)
      (Ast L_sleep_15_p L_sleep_15_g tag_sleep) 
    (comml (_popw DI)).
Proof.
  simpl.
  split.
  exists L_sleep_16_p; exists L_sleep_16_g.
  split_l.
  assert (Hpp' : forall l_ptr wqa sp pc ax bx cx dx di mP S S',
    Ast_sleep_15 l_ptr wqa sp pc ax bx cx dx di mP S
    -> nextS (L_sleep+15) (comml (_popw DI)) S S'
    -> Ast_sleep_16 l_ptr wqa sp pc ax bx cx dx mP S').
    intros l_ptr wqa sp pc ax bx cx dx di mP S S' HS HSS'.
    unfold Ast_sleep_15, Ast_sleep_16 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HR' as [dix [Hlds HR']].
    simplS in * .
    destruct HS as [HM [Hsp [Hspgt' [Hif Hisr]]]].
    assert (Heq : di = dix).
      sep on.
      sep lift 4 in HM.
      rewrite Hsp in Hlds.
      sep inj.
      auto.
    subst dix.
    subst M' fl' R' isr'.
    split.
      sep on.
      sep matchH HM with 9.
      sep stksplit.
      arith_rep (sp - 8 - 2) with (sp-10).
      arith_rep (_STK_SIZE_sleep - 8 - 2) with (_STK_SIZE_sleep-10).
      sep matchH HM with 2.
    rewrite Hsp.
    reg simpl.
    split_l by omega.
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_sleep_15_p in HpS.
    destruct HpS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di HS]]]]]]]]].
    exists (M, RF_update (RF_update R SP (sp - 8)) DI di, fl, isr).
    apply nextS_popw with (RF_update R SP (sp - 8)) di; trivial.
    unfold popW.
    destruct HS as [HM [Hsp [Hspgt [Hif Hisr]]]].
    simplS in * .
    rewrite Hsp.
    split.
      sep on.
      sep lift 4 in HM.
      sep trivial.
    arith_rep (sp - 10 + 2) with (sp - 8).
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_sleep_15_p, L_sleep_16_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di HS]]]]]]]]].
    exists l_ptr; exists wqa; exists sp; exists pc; exists ax; exists bx; exists cx; 
      exists dx.
    apply Hpp' with di S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_sleep_15_g, L_sleep_16_g in * .
  intros mP l_ptr wqa sp pc ax bx cx dx di HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [HM [Hsp _]].
  simplS in * .
  destruct HR' as [dix [Hlds HR']].
  assert (Heq : di = dix).
    sep on.
    sep lift 4 in HM.
    rewrite Hsp in Hlds.
    sep inj.
    auto.
  subst dix ax bx cx dx sp R' fl' isr'.
  reg simpl.
  repeat split_l; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_sleep_15_p in HS.
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx [di HS]]]]]]]]].
    unfold Ast_sleep_15 in HS.
    destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_sleep_15_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP l_ptr wqa sp pc ax bx cx dx di HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_sleep_15 in HS.
  destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
  destruct IHS as [IHif IHisr].
  simplS in * .
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

(* <sleep + 16>:  _popw (DX) *)

Lemma NH_WFcmd_sleep_16_popw :
  NH_WFcmd (SCAP_consCT 
    (L_sleep+16+1)
      (Ast L_sleep_17_p L_sleep_17_g tag_sleep) psi_sleep)
    (L_sleep+16)
      (Ast L_sleep_16_p L_sleep_16_g tag_sleep) 
    (comml (_popw DX)).
Proof.
  simpl.
  split.
  exists L_sleep_17_p; exists L_sleep_17_g.
  split_l.
  assert (Hpp' : forall l_ptr wqa sp pc ax bx cx dx mP S S',
    Ast_sleep_16 l_ptr wqa sp pc ax bx cx dx mP S
    -> nextS (L_sleep+16) (comml (_popw DX)) S S'
    -> Ast_sleep_17 l_ptr wqa sp pc ax bx cx mP S').
    intros l_ptr wqa sp pc ax bx cx dx mP S S' HS HSS'.
    unfold Ast_sleep_16, Ast_sleep_17 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HR' as [dxx [Hlds HR']].
    simplS in * .
    destruct HS as [HM [Hsp [Hspgt' [Hif Hisr]]]].
    assert (Heq : dx = dxx).
      sep on.
      sep lift 4 in HM.
      rewrite Hsp in Hlds.
      sep inj.
      auto.
    subst dxx.
    subst M' fl' R' isr'.
    split.
      sep on.
      sep matchH HM with 8.
      sep stksplit.
      arith_rep (sp - 6 - 2) with (sp-8).
      arith_rep (_STK_SIZE_sleep - 6 - 2) with (_STK_SIZE_sleep-8).
      sep matchH HM with 2.
    rewrite Hsp.
    reg simpl.
    split_l by omega.
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_sleep_16_p in HpS.
    destruct HpS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx HS]]]]]]]].
    exists (M, RF_update (RF_update R SP (sp - 6)) DX dx, fl, isr).
    apply nextS_popw with (RF_update R SP (sp - 6)) dx; trivial.
    unfold popW.
    destruct HS as [HM [Hsp [Hspgt [Hif Hisr]]]].
    simplS in * .
    rewrite Hsp.
    split.
      sep on.
      sep lift 4 in HM.
      sep trivial.
    arith_rep (sp - 8 + 2) with (sp - 6).
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_sleep_16_p, L_sleep_17_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx HS]]]]]]]].
    exists l_ptr; exists wqa; exists sp; exists pc; exists ax; exists bx; exists cx.
    apply Hpp' with dx S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_sleep_16_g, L_sleep_17_g in * .
  intros mP l_ptr wqa sp pc ax bx cx dx HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [HM [Hsp _]].
  simplS in * .
  destruct HR' as [dxx [Hlds HR']].
  assert (Heq : dx = dxx).
    sep on.
    sep lift 4 in HM.
    rewrite Hsp in Hlds.
    sep inj.
    auto.
  subst ax bx cx dxx sp R' fl' isr'.
  reg simpl.
  repeat split_l.
  trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_sleep_16_p in HS.
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx [dx HS]]]]]]]].
    unfold Ast_sleep_16 in HS.
    destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_sleep_16_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP l_ptr wqa sp pc ax bx cx dx HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_sleep_16 in HS.
  destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
  destruct IHS as [IHif IHisr].
  simplS in * .
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

(* <sleep + 17>:   _popw  (CX); *)

Lemma NH_WFcmd_sleep_17_popw :
  NH_WFcmd (SCAP_consCT 
    (L_sleep+17+1)
      (Ast L_sleep_18_p L_sleep_18_g tag_sleep) psi_sleep)
    (L_sleep+17)
      (Ast L_sleep_17_p L_sleep_17_g tag_sleep) 
    (comml (_popw CX)).
Proof.
  simpl.
  split.
  exists L_sleep_18_p; exists L_sleep_18_g.
  split_l.
  assert (Hpp' : forall l_ptr wqa sp pc ax bx cx mP S S',
    Ast_sleep_17 l_ptr wqa sp pc ax bx cx mP S
    -> nextS (L_sleep+17) (comml (_popw CX)) S S'
    -> Ast_sleep_18 l_ptr wqa sp pc ax bx mP S').
    intros l_ptr wqa sp pc ax bx cx mP S S' HS HSS'.
    unfold Ast_sleep_17, Ast_sleep_18 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HR' as [cxx [Hlds HR']].
    simplS in * .
    destruct HS as [HM [Hsp [Hspgt' [Hif Hisr]]]].
    assert (Heq : cx = cxx).
      sep on.
      sep lift 4 in HM.
      rewrite Hsp in Hlds.
      sep inj.
      auto.
    subst cxx.
    subst M' fl' R' isr'.
    split.
      sep on.
      sep matchH HM with 8.
      sep stksplit.
      arith_rep (sp - 4 - 2) with (sp-6).
      arith_rep (_STK_SIZE_sleep - 4 - 2) with (_STK_SIZE_sleep-6).
      sep matchH HM with 2.
    rewrite Hsp.
    reg simpl.
    split_l by omega.
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_sleep_17_p in HpS.
    destruct HpS as [l_ptr [wqa [sp [pc [ax [bx [cx HS]]]]]]].
    exists (M, RF_update (RF_update R SP (sp - 4)) CX cx, fl, isr).
    apply nextS_popw with (RF_update R SP (sp - 4)) cx; trivial.
    unfold popW.
    destruct HS as [HM [Hsp [Hspgt [Hif Hisr]]]].
    simplS in * .
    rewrite Hsp.
    split.
      sep on.
      sep lift 4 in HM.
      sep trivial.
    arith_rep (sp - 6 + 2) with (sp - 4).
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_sleep_17_p, L_sleep_18_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx HS]]]]]]].
    exists l_ptr; exists wqa; exists sp; exists pc; exists ax; exists bx.
    apply Hpp' with cx S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_sleep_17_g, L_sleep_18_g in * .
  intros mP l_ptr wqa sp pc ax bx cx HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [HM [Hsp _]].
  simplS in * .
  destruct HR' as [cxx [Hlds HR']].
  assert (Heq : cx = cxx).
    sep on.
    sep lift 4 in HM.
    rewrite Hsp in Hlds.
    sep inj.
    auto.
  subst ax bx cxx sp R' fl' isr'.
  reg simpl.
  repeat split_l.
  trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_sleep_17_p in HS.
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx [cx HS]]]]]]].
    unfold Ast_sleep_17 in HS.
    destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_sleep_17_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP l_ptr wqa sp pc ax bx cx HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_sleep_17 in HS.
  destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
  destruct IHS as [IHif IHisr].
  simplS in * .
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

(* <sleep + 18>:  _popw  (BX); *)

Lemma NH_WFcmd_sleep_18_popw :
  NH_WFcmd (SCAP_consCT 
    (L_sleep+18+1)
      (Ast L_sleep_19_p L_sleep_19_g tag_sleep) psi_sleep)
    (L_sleep+18)
      (Ast L_sleep_18_p L_sleep_18_g tag_sleep) 
    (comml (_popw BX)).
Proof.
  simpl.
  split.
  exists L_sleep_19_p; exists L_sleep_19_g.
  split_l.
  assert (Hpp' : forall l_ptr wqa sp pc ax bx mP S S',
    Ast_sleep_18 l_ptr wqa sp pc ax bx mP S
    -> nextS (L_sleep+18) (comml (_popw BX)) S S'
    -> Ast_sleep_19 l_ptr wqa sp pc ax mP S').
    intros l_ptr wqa sp pc ax bx mP S S' HS HSS'.
    unfold Ast_sleep_18, Ast_sleep_19 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HR' as [bxx [Hlds HR']].
    simplS in * .
    destruct HS as [HM [Hsp [Hspgt' [Hif Hisr]]]].
    assert (Heq : bx = bxx).
      sep on.
      sep lift 4 in HM.
      rewrite Hsp in Hlds.
      sep inj.
      auto.
    subst bxx.
    subst M' fl' R' isr'.
    split.
      sep on.
      sep matchH HM with 6.
      sep stksplit.
      arith_rep (sp - 2 - 2) with (sp-4).
      arith_rep (_STK_SIZE_sleep - 2 - 2) with (_STK_SIZE_sleep-4).
      sep matchH HM with 2.
    rewrite Hsp.
    reg simpl.
    split_l by omega.
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_sleep_18_p in HpS.
    destruct HpS as [l_ptr [wqa [sp [pc [ax [bx HS]]]]]].
    exists (M, RF_update (RF_update R SP (sp - 2)) BX bx, fl, isr).
    apply nextS_popw with (RF_update R SP (sp - 2)) bx; trivial.
    unfold popW.
    destruct HS as [HM [Hsp [Hspgt [Hif Hisr]]]].
    simplS in * .
    rewrite Hsp.
    split.
      sep on.
      sep lift 4 in HM.
      sep trivial.
    arith_rep (sp - 4 + 2) with (sp - 2).
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_sleep_18_p, L_sleep_19_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx HS]]]]]].
    exists l_ptr; exists wqa; exists sp; exists pc; exists ax.
    apply Hpp' with bx S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_sleep_18_g, L_sleep_19_g in * .
  intros mP l_ptr wqa sp pc ax bx HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [HM [Hsp _]].
  simplS in * .
  destruct HR' as [bxx [Hlds HR']].
  assert (Heq : bx = bxx).
    sep on.
    sep lift 4 in HM.
    rewrite Hsp in Hlds.
    sep inj.
    auto.
  subst ax bxx sp R' fl' isr'.
  reg simpl.
  repeat split_l.
  trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_sleep_18_p in HS.
    destruct HS as [l_ptr [wqa [sp [pc [ax [bx HS]]]]]].
    unfold Ast_sleep_18 in HS.
    destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_sleep_18_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP l_ptr wqa sp pc ax bx HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_sleep_18 in HS.
  destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
  destruct IHS as [IHif IHisr].
  simplS in * .
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

(* <sleep + 19>:  _popw  (AX); *)

Lemma NH_WFcmd_sleep_19_popw :
  NH_WFcmd (SCAP_consCT 
    (L_sleep+19+1)
      (Ast L_sleep_20_p L_sleep_20_g tag_sleep) psi_sleep)
    (L_sleep+19)
      (Ast L_sleep_19_p L_sleep_19_g tag_sleep) 
    (comml (_popw AX)).
Proof.
  simpl.
  split.
  exists L_sleep_20_p; exists L_sleep_20_g.
  split_l.
  assert (Hpp' : forall l_ptr wqa sp pc ax mP S S',
    Ast_sleep_19 l_ptr wqa sp pc ax mP S
    -> nextS (L_sleep+19) (comml (_popw AX)) S S'
    -> Ast_sleep_20 l_ptr wqa sp pc mP S').
    intros l_ptr wqa sp pc ax mP S S' HS HSS'.
    unfold Ast_sleep_19, Ast_sleep_20 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HR' as [axx [Hlds HR']].
    simplS in * .
    destruct HS as [HM [Hsp [Hspgt' [Hif Hisr]]]].
    assert (Heq : ax = axx).
      sep on.
      sep lift 4 in HM.
      rewrite Hsp in Hlds.
      sep inj.
      auto.
    subst axx.
    subst M' fl' R' isr'.
    split.
      sep on.
      sep matchH HM with 5.
      sep stksplit.
      sep matchH HM with 2.
    rewrite Hsp.
    reg simpl.
    split_l by omega.
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_sleep_19_p in HpS.
    destruct HpS as [l_ptr [wqa [sp [pc [ax HS]]]]].
    exists (M, RF_update (RF_update R SP sp) AX ax, fl, isr).
    apply nextS_popw with (RF_update R SP sp) ax; trivial.
    unfold popW.
    destruct HS as [HM [Hsp [Hspgt [Hif Hisr]]]].
    simplS in * .
    rewrite Hsp.
    split.
      sep on.
      sep lift 4 in HM.
      sep trivial.
    arith_rep (sp - 2 + 2) with sp.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_sleep_19_p, L_sleep_20_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [l_ptr [wqa [sp [pc [ax HS]]]]].
    exists l_ptr; exists wqa; exists sp; exists pc.
    apply Hpp' with ax S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_sleep_19_g, L_sleep_20_g in * .
  intros mP l_ptr wqa sp pc ax HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [HM [Hsp HS]].
  simplS in * .
  destruct HR' as [axx [Hlds HR']].
  assert (Heq : ax = axx).
    sep on.
    sep lift 4 in HM.
    rewrite Hsp in Hlds.
    sep inj.
    auto.
  subst axx R' fl' isr'.
  reg simpl.
  repeat split_l.
  trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_sleep_19_p in HS.
    destruct HS as [l_ptr [wqa [sp [pc [ax HS]]]]].
    unfold Ast_sleep_19 in HS.
    destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_sleep_19_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP l_ptr wqa sp pc ax HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_sleep_19 in HS.
  destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
  destruct IHS as [IHif IHisr].
  simplS in * .
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma NH_WFcmd_sleep_20_retw :
  NH_WFcmd psi_sleep
    (L_sleep+20)
      (Ast L_sleep_20_p L_sleep_20_g tag_sleep) 
    (comms (_retw)).
Proof.
  simpl.
  split.
  split_l.
  intros S HS.
  unfold L_sleep_20_g.
  unfold Ast_sleep_20, Ast_sleep_post.
  intros mP l_ptr wqa sp pc H.
  destruct H as [H1 [H2 [H3 H4]]].
  split_l by auto.
  repeat split_l.
  trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_sleep_20_p in HS.
    destruct HS as [l_ptr [wqa [sp [pc HS]]]].
    unfold Ast_sleep_20 in HS.
    destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_sleep_20_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP l_ptr wqa sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_sleep_20 in HS.
  destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
  destruct IHS as [IHif IHisr].
  simplS in * .
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma NH_WFiseq_sleep :
  NH_WFiseq psi_sleep L_sleep spec_sleep I_sleep.
Proof.
unfold I_sleep, spec_sleep.
apply nh_isq_seq with L_sleep_1_p L_sleep_1_g.
apply NH_WFcmd_sleep_0_pushwp.
apply nh_isq_seq with L_sleep_2_p L_sleep_2_g.
apply NH_WFcmd_sleep_1_pushwp.
assert (H : L_sleep+1+1 = L_sleep+2); [omega | rewrite H; clear H].
apply nh_isq_seq with L_sleep_3_p L_sleep_3_g.
apply NH_WFcmd_sleep_2_pushwp.
assert (H : L_sleep+2+1 = L_sleep+3); [omega | rewrite H; clear H].
apply nh_isq_seq with L_sleep_4_p L_sleep_4_g.
apply NH_WFcmd_sleep_3_pushwp.
assert (H : L_sleep+3+1 = L_sleep+4); [omega | rewrite H; clear H].
apply nh_isq_seq with L_sleep_5_p L_sleep_5_g.
apply NH_WFcmd_sleep_4_pushwp.
assert (H : L_sleep+4+1 = L_sleep+5); [omega | rewrite H; clear H].
apply nh_isq_seq with L_sleep_6_p L_sleep_6_g.
apply NH_WFcmd_sleep_5_pushwp.
assert (H : L_sleep+5+1 = L_sleep+6); [omega | rewrite H; clear H].
apply nh_isq_seq with L_sleep_7_p L_sleep_7_g.
apply NH_WFcmd_sleep_6_pushwp.
assert (H : L_sleep+6+1 = L_sleep+7); [omega | rewrite H; clear H].
apply nh_isq_seq with L_sleep_8_p L_sleep_8_g.
apply NH_WFcmd_sleep_7_pushfw.
assert (H : L_sleep+7+1 = L_sleep+8); [omega | rewrite H; clear H].
apply nh_isq_seq with L_sleep_9_p L_sleep_9_g.
apply NH_WFcmd_sleep_8_movwp.
assert (H : L_sleep+8+1 = L_sleep+9); [omega | rewrite H; clear H].
apply nh_isq_seq with L_sleep_10_p L_sleep_10_g.
apply NH_WFcmd_sleep_9_pushwm.
assert (H : L_sleep+9+1 = L_sleep+10); [omega | rewrite H; clear H].
apply nh_isq_seq with L_sleep_11_p L_sleep_11_g.
apply NH_WFcmd_sleep_10_call.
assert (H : L_sleep+10+1 = L_sleep+11); [omega | rewrite H; clear H].
apply nh_isq_seq with L_sleep_12_p L_sleep_12_g.
apply NH_WFcmd_sleep_11_addw.
assert (H : L_sleep+11+1 = L_sleep+12); [omega | rewrite H; clear H].
apply nh_isq_seq with L_sleep_13_p L_sleep_13_g.
apply NH_WFcmd_sleep_12_popfw.
assert (H : L_sleep+12+1 = L_sleep+13); [omega | rewrite H; clear H].
apply nh_isq_seq with L_sleep_14_p L_sleep_14_g.
apply NH_WFcmd_sleep_13_popw.
assert (H : L_sleep+13+1 = L_sleep+14); [omega | rewrite H; clear H].
apply nh_isq_seq with L_sleep_15_p L_sleep_15_g.
apply NH_WFcmd_sleep_14_popw.
assert (H : L_sleep+14+1 = L_sleep+15); [omega | rewrite H; clear H].
apply nh_isq_seq with L_sleep_16_p L_sleep_16_g.
apply NH_WFcmd_sleep_15_popw.
assert (H : L_sleep+15+1 = L_sleep+16); [omega | rewrite H; clear H].
apply nh_isq_seq with L_sleep_17_p L_sleep_17_g.
apply NH_WFcmd_sleep_16_popw.
assert (H : L_sleep+16+1 = L_sleep+17); [omega | rewrite H; clear H].
apply nh_isq_seq with L_sleep_18_p L_sleep_18_g.
apply NH_WFcmd_sleep_17_popw.
assert (H : L_sleep+17+1 = L_sleep+18); [omega | rewrite H; clear H].
apply nh_isq_seq with L_sleep_19_p L_sleep_19_g.
apply NH_WFcmd_sleep_18_popw.
assert (H : L_sleep+18+1 = L_sleep+19); [omega | rewrite H; clear H].
apply nh_isq_seq with L_sleep_20_p L_sleep_20_g.
apply NH_WFcmd_sleep_19_popw.
assert (H : L_sleep+19+1 = L_sleep+20); [omega | rewrite H; clear H].
apply nh_isq_end.
apply NH_WFcmd_sleep_20_retw.
Qed.
