(***************************************************************)
(* Coq code for AIM                                            *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

Add LoadPath ".." .
Add LoadPath "../CAP0" .
Add LoadPath "../SCAP" .

Require Import Coqlib.

Require Import Arith.
Require Import Machine.
Require Import Assertion.
Require Import Operational_Semantics.
Require Import AIM.
Require Import SCAP.
Require Import QLib.
Require Import Linking.

Require Import Kernel.

Definition Ast_sleep_0 (l_ptr : Label) (wqa : WQassert)
  (sp pc : Label) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec sp _STK_SIZE_sleep
    *^* sp |->w pc
    *^* sp+2 |->w l_ptr
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp
  /\ sp >= _STK_SIZE_sleep
  /\ S_Flag S _IF = false
  /\ S_ISR S = false
  /\ Delta l_ptr = SomeT wqa.

Definition Ast_sleep_post (l_ptr : Label) (wqa : WQassert)
  (sp pc : Label) (mP : MPred) (S' : State) : Prop :=
  (INV1 *^* INV0 *^* wqa
    *^* stkSpacePrec sp _STK_SIZE_sleep
    *^* sp |->w pc
    *^* sp+2 |->w l_ptr
    *^* mP) (S_Mem S')
  /\ S_Flag S' _IF = false
  /\ S_ISR S' = false.
  
Definition L_sleep_0_p (S : State) : Prop :=
  exists l_ptr, exists wqa, exists sp, exists pc,
    Ast_sleep_0 l_ptr wqa sp pc mTrue S.

Definition L_sleep_0_g (S S' : State) : Prop := forall mP,
  forall l_ptr wqa sp pc,
    Ast_sleep_0 l_ptr wqa sp pc mP S
    -> Ast_sleep_post l_ptr wqa sp pc mP S'
    /\ S_RF S' = S_RF S.

(* <sleep+0>:   pushw   %ax *)

Definition Ast_sleep_1 (l_ptr : Label) (wqa : WQassert)
  (sp pc ax : Label) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec (sp-2) (_STK_SIZE_sleep-2)
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* sp+2 |->w l_ptr
    *^* mP) (S_Mem S)
  /\ S_RF S AX = ax
  /\ S_RF S SP = sp - 2
  /\ sp >= _STK_SIZE_sleep
  /\ S_Flag S _IF = false
  /\ S_ISR S = false
  /\ Delta l_ptr = SomeT wqa.

Definition L_sleep_1_p (S : State) : Prop :=
  exists l_ptr, exists wqa, exists sp, exists pc, exists ax, 
    Ast_sleep_1 l_ptr wqa sp pc ax mTrue S.

Definition L_sleep_1_g (S S' : State) : Prop := forall mP,
  forall l_ptr wqa sp pc ax,
    Ast_sleep_1 l_ptr wqa sp pc ax mP S
    -> Ast_sleep_post l_ptr wqa sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = S_RF S BX
    /\ S_RF S' CX = S_RF S CX
    /\ S_RF S' DX = S_RF S DX
    /\ S_RF S' DI = S_RF S DI
    /\ S_RF S' SI = S_RF S SI
    /\ S_RF S' BP = S_RF S BP
    /\ S_RF S' SP = sp.

(* <sleep+1>:   pushw   %bx *)

Definition Ast_sleep_2 (l_ptr : Label) (wqa : WQassert)
  (sp pc ax bx : Label) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec (sp-4) (_STK_SIZE_sleep-4)
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* sp+2 |->w l_ptr
    *^* mP) (S_Mem S)
  /\ S_RF S BX = bx
  /\ S_RF S SP = sp - 4
  /\ sp >= _STK_SIZE_sleep
  /\ S_Flag S _IF = false
  /\ S_ISR S = false
  /\ Delta l_ptr = SomeT wqa.

Definition L_sleep_2_p (S : State) : Prop :=
  exists l_ptr, exists wqa, exists sp, exists pc, exists ax, exists bx,
    Ast_sleep_2 l_ptr wqa sp pc ax bx mTrue S.

Definition L_sleep_2_g (S S' : State) : Prop := forall mP,
  forall l_ptr wqa sp pc ax bx,
    Ast_sleep_2 l_ptr wqa sp pc ax bx mP S
    -> Ast_sleep_post l_ptr wqa sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = S_RF S CX
    /\ S_RF S' DX = S_RF S DX
    /\ S_RF S' DI = S_RF S DI
    /\ S_RF S' SI = S_RF S SI
    /\ S_RF S' BP = S_RF S BP
    /\ S_RF S' SP = sp.

(* <sleep+2>:   pushw   %cx *)

Definition Ast_sleep_3 (l_ptr : Label) (wqa : WQassert)
  (sp pc ax bx cx : Label) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec (sp-6) (_STK_SIZE_sleep-6)
    *^* sp-6 |->w cx
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* sp+2 |->w l_ptr
    *^* mP) (S_Mem S)
  /\ S_RF S CX = cx
  /\ S_RF S SP = sp - 6
  /\ sp >= _STK_SIZE_sleep
  /\ S_Flag S _IF = false
  /\ S_ISR S = false
  /\ Delta l_ptr = SomeT wqa.

Definition L_sleep_3_p (S : State) : Prop :=
  exists l_ptr, exists wqa, exists sp, exists pc, exists ax, exists bx, exists cx,
    Ast_sleep_3 l_ptr wqa sp pc ax bx cx mTrue S.

Definition L_sleep_3_g (S S' : State) : Prop := forall mP,
  forall l_ptr wqa sp pc ax bx cx,
    Ast_sleep_3 l_ptr wqa sp pc ax bx cx mP S
    -> Ast_sleep_post l_ptr wqa sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = cx
    /\ S_RF S' DX = S_RF S DX
    /\ S_RF S' DI = S_RF S DI
    /\ S_RF S' SI = S_RF S SI
    /\ S_RF S' BP = S_RF S BP
    /\ S_RF S' SP = sp.

(* <sleep+3>:   pushw   %dx *)

Definition Ast_sleep_4 (l_ptr : Label) (wqa : WQassert) 
  (sp pc ax bx cx dx : Label) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec (sp-8) (_STK_SIZE_sleep-8)
    *^* sp-8 |->w dx
    *^* sp-6 |->w cx
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* sp+2 |->w l_ptr
    *^* mP) (S_Mem S)
  /\ S_RF S DX = dx
  /\ S_RF S SP = sp - 8
  /\ sp >= _STK_SIZE_sleep
  /\ S_Flag S _IF = false
  /\ S_ISR S = false
  /\ Delta l_ptr = SomeT wqa.

Definition L_sleep_4_p (S : State) : Prop :=
  exists l_ptr, exists wqa, exists sp, exists pc, exists ax, exists bx, 
    exists cx, exists dx,
    Ast_sleep_4 l_ptr wqa sp pc ax bx cx dx mTrue S.

Definition L_sleep_4_g (S S' : State) : Prop := forall mP,
  forall l_ptr wqa sp pc ax bx cx dx,
    Ast_sleep_4 l_ptr wqa sp pc ax bx cx dx mP S
    -> Ast_sleep_post l_ptr wqa sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = cx
    /\ S_RF S' DX = dx
    /\ S_RF S' DI = S_RF S DI
    /\ S_RF S' SI = S_RF S SI
    /\ S_RF S' BP = S_RF S BP
    /\ S_RF S' SP = sp.

(* <sleep+4>:   pushw   %di *)

Definition Ast_sleep_5 (l_ptr : Label) (wqa : WQassert)
  (sp pc ax bx cx dx di : Label) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec (sp-10) (_STK_SIZE_sleep-10)
    *^* sp-10 |->w di
    *^* sp-8 |->w dx
    *^* sp-6 |->w cx
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* sp+2 |->w l_ptr
    *^* mP) (S_Mem S)
  /\ S_RF S DI = di
  /\ S_RF S SP = sp - 10
  /\ sp >= _STK_SIZE_sleep
  /\ S_Flag S _IF = false
  /\ S_ISR S = false
  /\ Delta l_ptr = SomeT wqa.

Definition L_sleep_5_p (S : State) : Prop :=
  exists l_ptr, exists wqa, exists sp, exists pc, exists ax, exists bx,
    exists cx, exists dx, exists di,
    Ast_sleep_5 l_ptr wqa sp pc ax bx cx dx di mTrue S.

Definition L_sleep_5_g (S S' : State) : Prop := forall mP,
  forall l_ptr wqa sp pc ax bx cx dx di,
    Ast_sleep_5 l_ptr wqa sp pc ax bx cx dx di mP S
    -> Ast_sleep_post l_ptr wqa sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = cx
    /\ S_RF S' DX = dx
    /\ S_RF S' DI = di
    /\ S_RF S' SI = S_RF S SI
    /\ S_RF S' BP = S_RF S BP
    /\ S_RF S' SP = sp.

(* <sleep+5>:   pushw   %si *)

Definition Ast_sleep_6 (l_ptr : Label) (wqa : WQassert)
  (sp pc ax bx cx dx di si : Label) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec (sp-12) (_STK_SIZE_sleep-12)
    *^* sp-12 |->w si
    *^* sp-10 |->w di
    *^* sp-8 |->w dx
    *^* sp-6 |->w cx
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* sp+2 |->w l_ptr
    *^* mP) (S_Mem S)
  /\ S_RF S SI = si
  /\ S_RF S SP = sp - 12
  /\ sp >= _STK_SIZE_sleep
  /\ S_Flag S _IF = false
  /\ S_ISR S = false
  /\ Delta l_ptr = SomeT wqa.

Definition L_sleep_6_p (S : State) : Prop :=
  exists l_ptr, exists wqa, exists sp, exists pc, exists ax, exists bx,
    exists cx, exists dx, exists di, exists si,
    Ast_sleep_6 l_ptr wqa sp pc ax bx cx dx di si mTrue S.

Definition L_sleep_6_g (S S' : State) : Prop := forall mP,
  forall l_ptr wqa sp pc ax bx cx dx di si,
    Ast_sleep_6 l_ptr wqa sp pc ax bx cx dx di si mP S
    -> Ast_sleep_post l_ptr wqa sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = cx
    /\ S_RF S' DX = dx
    /\ S_RF S' DI = di
    /\ S_RF S' SI = si
    /\ S_RF S' BP = S_RF S BP
    /\ S_RF S' SP = sp.

(* <sleep+6>:   pushw   %bp *)

Definition Ast_sleep_7 (l_ptr : Label) (wqa : WQassert)
  (sp pc ax bx cx dx di si bp : Label) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec (sp-14) (_STK_SIZE_sleep-14)
    *^* sp-14 |->w bp
    *^* sp-12 |->w si
    *^* sp-10 |->w di
    *^* sp-8 |->w dx
    *^* sp-6 |->w cx
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* sp+2 |->w l_ptr
    *^* mP) (S_Mem S)
  /\ S_RF S BP = bp
  /\ S_RF S SP = sp - 14
  /\ sp >= _STK_SIZE_sleep
  /\ S_Flag S _IF = false
  /\ S_ISR S = false
  /\ Delta l_ptr = SomeT wqa.

Definition L_sleep_7_p (S : State) : Prop :=
  exists l_ptr, exists wqa, exists sp, exists pc, exists ax, exists bx,
    exists cx, exists dx, exists di, exists si, exists bp,
    Ast_sleep_7 l_ptr wqa sp pc ax bx cx dx di si bp mTrue S.

Definition L_sleep_7_g (S S' : State) : Prop := forall mP,
  forall l_ptr wqa sp pc ax bx cx dx di si bp,
    Ast_sleep_7 l_ptr wqa sp pc ax bx cx dx di si bp mP S
    -> Ast_sleep_post l_ptr wqa sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = cx
    /\ S_RF S' DX = dx
    /\ S_RF S' DI = di
    /\ S_RF S' SI = si
    /\ S_RF S' BP = bp
    /\ S_RF S' SP = sp.

(* <sleep+7>:   pushfw *)

Definition Ast_sleep_8 (l_ptr : Label) (wqa : WQassert)
  (sp pc ax bx cx dx di si bp flnum : Label) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec (sp-16) (_STK_SIZE_sleep - 16)
    *^* sp-16 |->w flnum
    *^* sp-14 |->w bp
    *^* sp-12 |->w si
    *^* sp-10 |->w di
    *^* sp-8 |->w dx
    *^* sp-6 |->w cx
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* sp+2 |->w l_ptr
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 16
  /\ sp >= _STK_SIZE_sleep
  /\ S_Flag S = nat2FR flnum
  /\ S_Flag S _IF = false
  /\ S_ISR S = false
  /\ Delta l_ptr = SomeT wqa.

Definition L_sleep_8_p (S : State) : Prop :=
  exists l_ptr, exists wqa, exists sp, exists pc, exists ax, exists bx,
    exists cx, exists dx, exists di, exists si, exists bp, exists flnum,
    Ast_sleep_8 l_ptr wqa sp pc ax bx cx dx di si bp flnum mTrue S.

Definition L_sleep_8_g (S S' : State) : Prop := forall mP,
  forall l_ptr wqa sp pc ax bx cx dx di si bp flnum,
    Ast_sleep_8 l_ptr wqa sp pc ax bx cx dx di si bp flnum mP S
    -> Ast_sleep_post l_ptr wqa sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = cx
    /\ S_RF S' DX = dx
    /\ S_RF S' DI = di
    /\ S_RF S' SI = si
    /\ S_RF S' BP = bp
    /\ S_RF S' SP = sp.

(* <sleep+8>:   movw    %sp, %bp *)

Definition Ast_sleep_9 (l_ptr : Label) (wqa : WQassert)
  (sp pc ax bx cx dx di si bp flnum : Label) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec (sp-16) (_STK_SIZE_sleep - 16)
    *^* sp-16 |->w flnum
    *^* sp-14 |->w bp
    *^* sp-12 |->w si
    *^* sp-10 |->w di
    *^* sp-8 |->w dx
    *^* sp-6 |->w cx
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* sp+2 |->w l_ptr
    *^* mP) (S_Mem S)
  /\ S_RF S BP = S_RF S SP
  /\ S_RF S SP = sp - 16
  /\ sp >= _STK_SIZE_sleep
  /\ (nat2FR flnum) _IF = false
  /\ S_Flag S _IF = false
  /\ S_ISR S = false
  /\ Delta l_ptr = SomeT wqa.

Definition L_sleep_9_p (S : State) : Prop :=
  exists l_ptr, exists wqa, exists sp, exists pc, exists ax, exists bx,
    exists cx, exists dx, exists di, exists si, exists bp, exists flnum,
    Ast_sleep_9 l_ptr wqa sp pc ax bx cx dx di si bp flnum mTrue S.

Definition L_sleep_9_g (S S' : State) : Prop := forall mP,
  forall l_ptr wqa sp pc ax bx cx dx di si bp flnum,
    Ast_sleep_9 l_ptr wqa sp pc ax bx cx dx di si bp flnum mP S
    -> Ast_sleep_post l_ptr wqa sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = cx
    /\ S_RF S' DX = dx
    /\ S_RF S' DI = di
    /\ S_RF S' SI = si
    /\ S_RF S' BP = bp
    /\ S_RF S' SP = sp.

(* <sleep+9>:   pushw   18(%bp) *)

Definition Ast_sleep_10 (l_ptr : Label) (wqa : WQassert) 
  (sp pc ax bx cx dx di si bp flnum : Label) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec (sp-18) (_STK_SIZE_sleep - 18)
    *^* sp-18 |->w l_ptr
    *^* sp-16 |->w flnum
    *^* sp-14 |->w bp
    *^* sp-12 |->w si
    *^* sp-10 |->w di
    *^* sp-8 |->w dx
    *^* sp-6 |->w cx
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* sp+2 |->w l_ptr
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 18
  /\ sp >= _STK_SIZE_sleep
  /\ (nat2FR flnum) _IF = false
  /\ S_Flag S _IF = false
  /\ S_ISR S = false
  /\ Delta l_ptr = SomeT wqa.

Definition L_sleep_10_p (S : State) : Prop :=
  exists l_ptr, exists wqa, exists sp, exists pc, exists ax, exists bx,
    exists cx, exists dx, exists di, exists si, exists bp, exists flnum,
    Ast_sleep_10 l_ptr wqa sp pc ax bx cx dx di si bp flnum mTrue S.

Definition L_sleep_10_g (S S' : State) : Prop := forall mP,
  forall l_ptr wqa sp pc ax bx cx dx di si bp flnum,
    Ast_sleep_10 l_ptr wqa sp pc ax bx cx dx di si bp flnum mP S
    -> Ast_sleep_post l_ptr wqa sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = cx
    /\ S_RF S' DX = dx
    /\ S_RF S' DI = di
    /\ S_RF S' SI = si
    /\ S_RF S' BP = bp
    /\ S_RF S' SP = sp.

(* <sleep+10>:  call    block *)

Definition Ast_sleep_11 (l_ptr : Label) (wqa : WQassert) 
  (sp pc ax bx cx dx di si bp flnum : Label) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 *^* wqa
    *^* stkSpacePrec (sp-18) (_STK_SIZE_sleep - 18)
    *^* sp-18 |->w l_ptr
    *^* sp-16 |->w flnum
    *^* sp-14 |->w bp
    *^* sp-12 |->w si
    *^* sp-10 |->w di
    *^* sp-8 |->w dx
    *^* sp-6 |->w cx
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* sp+2 |->w l_ptr
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 18
  /\ sp >= _STK_SIZE_sleep
  /\ (nat2FR flnum) _IF = false
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.

Definition L_sleep_11_p (S : State) : Prop :=
  exists l_ptr, exists wqa, exists sp, exists pc, exists ax, exists bx,
    exists cx, exists dx, exists di, exists si, exists bp, exists flnum,
    Ast_sleep_11 l_ptr wqa sp pc ax bx cx dx di si bp flnum mTrue S.

Definition L_sleep_11_g (S S' : State) : Prop := forall mP,
  forall l_ptr wqa sp pc ax bx cx dx di si bp flnum,
    Ast_sleep_11 l_ptr wqa sp pc ax bx cx dx di si bp flnum mP S
    -> Ast_sleep_post l_ptr wqa sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = cx
    /\ S_RF S' DX = dx
    /\ S_RF S' DI = di
    /\ S_RF S' SI = si
    /\ S_RF S' BP = bp
    /\ S_RF S' SP = sp.

(* <sleep+11>:  addw    $2, %sp *)

Definition Ast_sleep_12 (l_ptr : Label) (wqa : WQassert) 
  (sp pc ax bx cx dx di si bp flnum : Label) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 *^* wqa
    *^* stkSpacePrec (sp-16) (_STK_SIZE_sleep - 16)
    *^* sp-16 |->w flnum
    *^* sp-14 |->w bp
    *^* sp-12 |->w si
    *^* sp-10 |->w di
    *^* sp-8 |->w dx
    *^* sp-6 |->w cx
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* sp+2 |->w l_ptr
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 16
  /\ sp >= _STK_SIZE_sleep
  /\ (nat2FR flnum) _IF = false
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.

Definition L_sleep_12_p (S : State) : Prop :=
  exists l_ptr, exists wqa, exists sp, exists pc, exists ax, exists bx,
    exists cx, exists dx, exists di, exists si, exists bp, exists flnum,
    Ast_sleep_12 l_ptr wqa sp pc ax bx cx dx di si bp flnum mTrue S.

Definition L_sleep_12_g (S S' : State) : Prop := forall mP,
  forall l_ptr wqa sp pc ax bx cx dx di si bp flnum,
    Ast_sleep_12 l_ptr wqa sp pc ax bx cx dx di si bp flnum mP S
    -> Ast_sleep_post l_ptr wqa sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = cx
    /\ S_RF S' DX = dx
    /\ S_RF S' DI = di
    /\ S_RF S' SI = si
    /\ S_RF S' BP = bp
    /\ S_RF S' SP = sp.

(* <sleep+12>:  popfw *)

Definition Ast_sleep_13 (l_ptr : Label) (wqa : WQassert)
  (sp pc ax bx cx dx di si bp : Label) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 *^* wqa
    *^* stkSpacePrec (sp-14) (_STK_SIZE_sleep - 14)
    *^* sp-14 |->w bp
    *^* sp-12 |->w si
    *^* sp-10 |->w di
    *^* sp-8 |->w dx
    *^* sp-6 |->w cx
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* sp+2 |->w l_ptr
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 14
  /\ sp >= _STK_SIZE_sleep
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.

Definition L_sleep_13_p (S : State) : Prop :=
  exists l_ptr, exists wqa, exists sp, exists pc, exists ax, exists bx,
    exists cx, exists dx, exists di, exists si, exists bp,
    Ast_sleep_13 l_ptr wqa sp pc ax bx cx dx di si bp mTrue S.

Definition L_sleep_13_g (S S' : State) : Prop := forall mP,
  forall l_ptr wqa sp pc ax bx cx dx di si bp,
    Ast_sleep_13 l_ptr wqa sp pc ax bx cx dx di si bp mP S
    -> Ast_sleep_post l_ptr wqa sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = cx
    /\ S_RF S' DX = dx
    /\ S_RF S' DI = di
    /\ S_RF S' SI = si
    /\ S_RF S' BP = bp
    /\ S_RF S' SP = sp.

(* <sleep+13>:  popw    %bp *)

Definition Ast_sleep_14 (l_ptr : Label) (wqa : WQassert)
  (sp pc ax bx cx dx di si : Label) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 *^* wqa
    *^* stkSpacePrec (sp-12) (_STK_SIZE_sleep - 12)
    *^* sp-12 |->w si
    *^* sp-10 |->w di
    *^* sp-8 |->w dx
    *^* sp-6 |->w cx
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* sp+2 |->w l_ptr
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 12
  /\ sp >= _STK_SIZE_sleep
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.

Definition L_sleep_14_p (S : State) : Prop :=
  exists l_ptr, exists wqa, exists sp, exists pc, exists ax, exists bx,
    exists cx, exists dx, exists di, exists si,
    Ast_sleep_14 l_ptr wqa sp pc ax bx cx dx di si mTrue S.

Definition L_sleep_14_g (S S' : State) : Prop := forall mP,
  forall l_ptr wqa sp pc ax bx cx dx di si,
    Ast_sleep_14 l_ptr wqa sp pc ax bx cx dx di si mP S
    -> Ast_sleep_post l_ptr wqa sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = cx
    /\ S_RF S' DX = dx
    /\ S_RF S' DI = di
    /\ S_RF S' SI = si
    /\ S_RF S' BP = S_RF S BP
    /\ S_RF S' SP = sp.

(* <sleep+14>:  popw    %si *)

Definition Ast_sleep_15 (l_ptr : Label) (wqa : WQassert) 
  (sp pc ax bx cx dx di : Label) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0  *^* wqa
    *^* stkSpacePrec (sp-10) (_STK_SIZE_sleep - 10)
    *^* sp-10 |->w di
    *^* sp-8 |->w dx
    *^* sp-6 |->w cx
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* sp+2 |->w l_ptr
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 10
  /\ sp >= _STK_SIZE_sleep
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.

Definition L_sleep_15_p (S : State) : Prop :=
  exists l_ptr, exists wqa, exists sp, exists pc, exists ax, exists bx,
    exists cx, exists dx, exists di,
    Ast_sleep_15 l_ptr wqa sp pc ax bx cx dx di mTrue S.

Definition L_sleep_15_g (S S' : State) : Prop := forall mP,
  forall l_ptr wqa sp pc ax bx cx dx di,
    Ast_sleep_15 l_ptr wqa sp pc ax bx cx dx di mP S
    -> Ast_sleep_post l_ptr wqa sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = cx
    /\ S_RF S' DX = dx
    /\ S_RF S' DI = di
    /\ S_RF S' SI = S_RF S SI
    /\ S_RF S' BP = S_RF S BP
    /\ S_RF S' SP = sp.

(* <sleep+15>:  popw    %di *)

Definition Ast_sleep_16 (l_ptr : Label) (wqa : WQassert)
  (sp pc ax bx cx dx : Label) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 *^* wqa
    *^* stkSpacePrec (sp-8) (_STK_SIZE_sleep - 8)
    *^* sp-8 |->w dx
    *^* sp-6 |->w cx
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* sp+2 |->w l_ptr
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 8
  /\ sp >= _STK_SIZE_sleep
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.

Definition L_sleep_16_p (S : State) : Prop :=
  exists l_ptr, exists wqa, exists sp, exists pc, exists ax, exists bx,
    exists cx, exists dx,
    Ast_sleep_16 l_ptr wqa sp pc ax bx cx dx mTrue S.

Definition L_sleep_16_g (S S' : State) : Prop := forall mP,
  forall l_ptr wqa sp pc ax bx cx dx,
    Ast_sleep_16 l_ptr wqa sp pc ax bx cx dx mP S
    -> Ast_sleep_post l_ptr wqa sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = cx
    /\ S_RF S' DX = dx
    /\ S_RF S' DI = S_RF S DI
    /\ S_RF S' SI = S_RF S SI
    /\ S_RF S' BP = S_RF S BP
    /\ S_RF S' SP = sp.

(* <sleep+16>:  popw    %dx *)

Definition Ast_sleep_17 (l_ptr : Label) (wqa : WQassert) 
  (sp pc ax bx cx : Label) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 *^* wqa
    *^* stkSpacePrec (sp-6) (_STK_SIZE_sleep - 6)
    *^* sp-6 |->w cx
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* sp+2 |->w l_ptr
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 6
  /\ sp >= _STK_SIZE_sleep
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.

Definition L_sleep_17_p (S : State) : Prop :=
  exists l_ptr, exists wqa, exists sp, exists pc, exists ax, exists bx,
    exists cx,
    Ast_sleep_17 l_ptr wqa sp pc ax bx cx mTrue S.

Definition L_sleep_17_g (S S' : State) : Prop := forall mP,
  forall l_ptr wqa sp pc ax bx cx,
    Ast_sleep_17 l_ptr wqa sp pc ax bx cx mP S
    -> Ast_sleep_post l_ptr wqa sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = cx
    /\ S_RF S' DX = S_RF S DX
    /\ S_RF S' DI = S_RF S DI
    /\ S_RF S' SI = S_RF S SI
    /\ S_RF S' BP = S_RF S BP
    /\ S_RF S' SP = sp.

(* <sleep+17>:  popw    %cx *)

Definition Ast_sleep_18 (l_ptr : Label) (wqa : WQassert)
  (sp pc ax bx : Label) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 *^* wqa
    *^* stkSpacePrec (sp-4) (_STK_SIZE_sleep - 4)
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* sp+2 |->w l_ptr
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 4
  /\ sp >= _STK_SIZE_sleep
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.

Definition L_sleep_18_p (S : State) : Prop :=
  exists l_ptr, exists wqa, exists sp, exists pc, exists ax, exists bx,
    Ast_sleep_18 l_ptr wqa sp pc ax bx mTrue S.

Definition L_sleep_18_g (S S' : State) : Prop := forall mP,
  forall l_ptr wqa sp pc ax bx,
    Ast_sleep_18 l_ptr wqa sp pc ax bx mP S
    -> Ast_sleep_post l_ptr wqa sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = S_RF S CX
    /\ S_RF S' DX = S_RF S DX
    /\ S_RF S' DI = S_RF S DI
    /\ S_RF S' SI = S_RF S SI
    /\ S_RF S' BP = S_RF S BP
    /\ S_RF S' SP = sp.

(* <sleep+18>:  popw    %bx *)

Definition Ast_sleep_19 (l_ptr : Label) (wqa : WQassert) 
  (sp pc ax : Label) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 *^* wqa
    *^* stkSpacePrec (sp-2) (_STK_SIZE_sleep - 2)
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* sp+2 |->w l_ptr
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 2
  /\ sp >= _STK_SIZE_sleep
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.

Definition L_sleep_19_p (S : State) : Prop :=
  exists l_ptr, exists wqa, exists sp, exists pc, exists ax,
    Ast_sleep_19 l_ptr wqa sp pc ax mTrue S.

Definition L_sleep_19_g (S S' : State) : Prop := forall mP,
  forall l_ptr wqa sp pc ax,
    Ast_sleep_19 l_ptr wqa sp pc ax mP S
    -> Ast_sleep_post l_ptr wqa sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = S_RF S BX
    /\ S_RF S' CX = S_RF S CX
    /\ S_RF S' DX = S_RF S DX
    /\ S_RF S' DI = S_RF S DI
    /\ S_RF S' SI = S_RF S SI
    /\ S_RF S' BP = S_RF S BP
    /\ S_RF S' SP = sp.

(* <sleep+19>:  popw    %ax *)

Definition Ast_sleep_20 (l_ptr : Label) (wqa : WQassert)
  (sp pc : Label) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 *^* wqa
    *^* stkSpacePrec sp _STK_SIZE_sleep
    *^* sp   |->w pc
    *^* sp+2 |->w l_ptr
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp
  /\ sp >= _STK_SIZE_sleep
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.

Definition L_sleep_20_p (S : State) : Prop :=
  exists l_ptr, exists wqa, exists sp, exists pc,
    Ast_sleep_20 l_ptr wqa sp pc mTrue S.

Definition L_sleep_20_g (S S' : State) : Prop := forall mP,
  forall l_ptr wqa sp pc,
    Ast_sleep_20 l_ptr wqa sp pc mP S
    -> Ast_sleep_post l_ptr wqa sp pc mP S'
    /\ S_RF S' AX = S_RF S AX
    /\ S_RF S' BX = S_RF S BX
    /\ S_RF S' CX = S_RF S CX
    /\ S_RF S' DX = S_RF S DX
    /\ S_RF S' DI = S_RF S DI
    /\ S_RF S' SI = S_RF S SI
    /\ S_RF S' BP = S_RF S BP
    /\ S_RF S' SP = sp.

(* <sleep+20>:  retw *)

Definition tag_sleep := 1.

Definition p_sleep := L_sleep_0_p.

Definition g_sleep := L_sleep_0_g.

Definition spec_sleep := Ast p_sleep g_sleep tag_sleep.

Definition psi_sleep  := SCAP_emptCT.
