(***************************************************************)
(* Coq code for AIM                                            *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

Add LoadPath "..".
Add LoadPath "../CAP0".
Add LoadPath "../SCAP".

Require Import Coqlib.

Require Import Arith.
Require Import Machine.
Require Import Assertion.
Require Import Operational_Semantics.
Require Import OpSem_Lemmas.

Require Import AIM.
Require Import SCAP.
Require Import QLib.
Require Import Linking.
Require Import Linking_Lemmas3.

Require Import Reg.
Require Import Assertion_Lemmas3.
Require Import Assertion_Lemmas4.

Require Import Kernel.

Require Import SwitchSpec.

(*  <schedule+0> : movw (curr_thrd), %bx *)

Lemma SCAP_WFcmd_schedule_0_movwld :
  SCAP_WFcmd (SCAP_consCT (L_schedule+1) (Ast L_schedule_1_p L_schedule_1_g sctag) psi_schedule)
    (L_schedule) (Ast L_schedule_pre_p L_schedule_0_g sctag) 
    (comml (_movwld (A_imm curr_thrd) BX)).
Proof.
  unfold SCAP_WFcmd.
  exists L_schedule_1_p;exists L_schedule_1_g.
  split_l.
  split.
  (* enable *)
    unfold enable.
    intros S PS.
    destruct S as [[[M R] fl] isr].
    unfold L_schedule_pre_p in PS.
    destruct PS as [T [RdyQ [tid [tcb [sp [pc [jk HS]]]]]]].
    destruct HS as [HS1 [HS2 [HS3 [HS4 [HS5 [HS6 HS7]]]]]].
    exists (M, RF_update R BX tid, fl, isr).
    apply nextS_movwld with curr_thrd tid.
      simpl;reflexivity.
      sep on.
      simpl in HS1.
      sep lift 0 in HS1.
      sep trivial.
      reflexivity.
  assert (Hpp': forall T RdyQ tid tcb sp pc jk mP S S',
    Ast_schedule_pre T RdyQ tid tcb sp pc jk mP S
    -> nextS (L_schedule) (comml (_movwld (A_imm curr_thrd) BX)) S S'
    -> Ast_schedule_1 T RdyQ tid tcb sp pc jk mP S').
  unfold Ast_schedule_pre,Ast_schedule_1.
  intros T RdyQ tid tcb sp pc jk mP S S' Hp HSS'.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  simplS in * .
  substH HSS' with (nextS_inversion HSS').
  simpl in HSS'.
  destruct HSS' as [HM' [[tidx [Hc HR']] [Hfl' Hisr']]] .
  destruct Hp as [HM [HSP [Hsp8 [Hsub [ Hnot [Hmap HT]]]]]].
  assert (Etid :tidx = tid).
    sep on.
    sep lift 0 in HM.
    sep inj.
    auto.
  subst tidx M' R' fl' isr'.
  intuition.
  split.
  (* p ==> p' *)
    unfold pimp.
    unfold linkPG, conjPG, conjG.
    unfold L_schedule_pre_p,L_schedule_1_p.
    intros S' H.
    destruct H as [S [H1 H2]].
    destruct H1 as [T [RdyQ [tid [tcb [sp [pc [jk H]]]]]]].
    exists T; exists RdyQ; exists tid; exists tcb; exists sp; exists pc;exists jk.
    apply Hpp' with S;assumption.
  (* g' ==> g *)
     unfold L_schedule_0_g, L_schedule_1_g.
     intros S S'' H.
     unfold linkPG, conjPG, conjG in H.
     destruct H as [S' H].
     destruct H as [[Hps HnextSS'] HgS'S''].
     intros mP T RdyQ tid tcb sp pc  jk HS.
     apply HgS'S''.
     apply Hpp' with S;assumption.
Qed.

(* <schedule + 1> :        movw    %sp, TCB_STACK(%bx)  *)
Lemma SCAP_WFcmd_schedule_1_movwst :
  SCAP_WFcmd (SCAP_consCT (L_schedule+1+1) (Ast L_schedule_2_p L_schedule_2_g sctag) psi_schedule)
    (L_schedule+1) (Ast L_schedule_1_p L_schedule_1_g sctag) 
    (comml (_movwst (Opr_reg SP) (A_reg TCB_STACK BX))).
Proof.
  unfold SCAP_WFcmd.
  exists L_schedule_2_p;exists L_schedule_2_g.
  split_l.
  split.
  (* enable *)
    unfold enable.
    intros S PS.
    destruct S as [[[M R] fl] isr].
    unfold L_schedule_1_p in PS.
    destruct PS as [T [RdyQ [tid [tcb [sp [pc [jk HS]]]]]]].
    destruct HS as [HS1 [HS2 [HS3 [HS4 [HS5 [HS6 HS7]]]]]].
    exists (MemUpdW M (tid + 4) sp 0, R, fl, isr).
    simpl in * .
    apply nextS_movwst with (tid+4) sp.
      simpl;rewrite HS2;unfold TCB_STACK;reflexivity.
      simpl;rewrite HS3;reflexivity.
      sep on.
      unfold embed_TCB in HS1.
      sep lift 3 in HS1.
      substH HS1 with (ptoTCB_elim1 HS1).
      destruct HS1 as [tst [HM HST]].
      sep lift 0 in HM.
      sep trivial.
      apply MemUpdW_MemUpd16 with 0;trivial.
  assert (Hpp': forall T RdyQ tid tcb sp pc jk mP S S',
    Ast_schedule_1 T RdyQ tid tcb sp pc jk mP S
    -> nextS (L_schedule + 1) (comml (_movwst (Opr_reg SP) (A_reg TCB_STACK BX))) S S'
    -> Ast_schedule_2 T RdyQ tid tcb sp pc jk mP S').
    unfold Ast_schedule_1, Ast_schedule_2.
    intros T RdyQ tid tcb sp pc jk mP S S' Hp HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    simplS in * .
    substH HSS' with (nextS_inversion HSS').
    simpl in HSS'.
    destruct HSS' as [[HM [x HM']] [HR' [Hfl' Hisr']]] .
    unfold embed_TCB in * .
    destruct Hp as [HM1 [Htid [Hsp [Hsp8 [Hsub [Hnotin [Hmap HT]]]]]]].
    sep on.
    sep lift 3 in HM1.
    subst M' R' fl' isr'.
    unfold TCB_STACK in * .
    intuition.
    sep lift 5.
    substH HM1 with (ptoTCB_elim1 HM1).
    destruct HM1 as [tst [HM1 Htst]].
    apply ptoTCB_intro1 with tst.
    assert ((MemUpdW M (R BX + 4) (R SP) x)
      ||- tid + 4 |->w TCB_SP (sp, TCB_Status tcb, pc) *^* tid + 4 + 2 |->w tst *^*
        (embed_RdyQ T RdyQ *^* (stkSpacePrec sp 8 *^* (sp |-> pc*^*(sp+1|-> jk *^* mP))) *^*
          curr_thrd |->w tid *^* (tid |->w_ *^* tid + 2 |->w_))).
    simpl in * . 
    rewrite Hsp; rewrite Htid.
    sep lift 0.
    sep upd.
    sep matchH HM1 with 10.
  unfold MemUpdW in H.
  assumption.
  simpl.
  assumption.
  split.
  (* p ==> p' *)
    unfold pimp.
    unfold linkPG, conjPG, conjG.
    unfold L_schedule_1_p,L_schedule_2_p.
    intros S' H.
    destruct H as [S [H1 H2]].
    destruct H1 as [T [RdyQ [tid [tcb [sp [pc [jk H]]]]]]].
    exists T; exists RdyQ; exists tid; exists tcb; exists sp; exists pc;exists jk.
    apply Hpp' with S; assumption.
  (* g' ==> g *)
     unfold L_schedule_2_g, L_schedule_1_g.
     intros S S'' H.
     unfold linkPG, conjPG, conjG in H.
     destruct H as [S' H].
     destruct H as [[Hps HnextSS'] HgS'S''].
     intros mP T RdyQ tid tcb sp pc jk HS.
     apply HgS'S''.
     apply Hpp' with S;assumption.
Qed.

(* <schedule+2> : pushw $readyQ *)
Lemma SCAP_WFcmd_schedule_2_pushwp :
  SCAP_WFcmd (SCAP_consCT (L_schedule+2+1) (Ast L_schedule_3_p L_schedule_3_g sctag) psi_schedule)
    (L_schedule+2) (Ast L_schedule_2_p L_schedule_2_g sctag) 
    (comml (_pushwp (Opr_imm ready_Q))).
Proof.
  unfold SCAP_WFcmd.
  exists L_schedule_3_p;exists L_schedule_3_g.
  split_l.
  split.
  (* enable *)
    unfold enable.
    intros S PS.
    destruct S as [[[M R] fl] isr].
    unfold L_schedule_2_p in PS.
    destruct PS as [T [RdyQ [tid [tcb [sp [pc [jk HS]]]]]]].
    destruct HS as [HS1 [HS2 [HS3 [HS4 [HS5 [HS6 [HS7 HS8]]]]]]].
    exists (MemUpdW M (sp-2) ready_Q 0, RF_update R SP (sp - 2), fl, isr).
    simpl in * .
    apply nextS_pushwp with ready_Q.
      simpl;reflexivity.
      unfold pushW.
      rewrite HS3.
      split.
        apply stkSpace_weaken with 8; [idtac | omega].
        sep on.
        sep lift 3 in HS1.
        sep trivial.
      intuition.
      apply MemUpdW_MemUpd16 with 0;trivial.
  (* p nextS -> p' *)
  assert (Hpp': forall T RdyQ tid tcb sp pc jk mP S S',
    Ast_schedule_2 T RdyQ tid tcb sp pc jk mP S
    -> nextS (L_schedule + 2) (comml (_pushwp (Opr_imm ready_Q))) S S'
    -> Ast_schedule_3 T RdyQ tid tcb sp pc jk mP S').
    unfold Ast_schedule_2, Ast_schedule_3.
    intros T RdyQ tid tcb sp pc jk mP S S' HS Hnext.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    substH Hnext with (nextS_inversion Hnext).
    simpl in * .
    destruct Hnext as [[Hnext1 [Hnext2 [x Hnext3]]] [Hnext4 [Hnext5 Hnext6]]].
    destruct HS as [HS1 [HS2 [HS3 [HS4 [HS5 [HS6 [HS7 HS8]]]]]]].
    subst M' R' fl' isr'.
    rewrite HS3.
    intuition.
    sep on.
    sep lift 3 in HS1.
    sep lift 4.
    assert (MemUpdW M (sp - 2) ready_Q x
      ||- sp - 2 |->w ready_Q *^*
        (sp |-> pc*^*(sp+1|->jk *^* mP) *^*
          (curr_thrd |->w tid *^* embed_TCB tid (sp, TCB_Status tcb, pc) *^*
            embed_RdyQ T RdyQ *^* stkSpacePrec (sp - 2) 6))).
      sep upd.
    sep matchH HS1 with 14.
    sep stksplit in HS1.
    arith_rep (8-2) with 6 in HS1.
    assumption.
  split.
  (* p ==> p' *)
    unfold pimp.
    unfold linkPG, conjPG, conjG.
    unfold L_schedule_2_p,L_schedule_3_p.
    intros S' H.
    destruct H as [S [H1 H2]].
    destruct H1 as [T [RdyQ [tid [tcb [sp [pc [jk H]]]]]]].
    exists T; exists RdyQ; exists tid; exists tcb; exists sp; exists pc;exists jk.
    apply Hpp' with S; assumption.
  (* g' ==> g *)
     unfold L_schedule_3_g, L_schedule_2_g.
     intros S S'' H.
     unfold linkPG, conjPG, conjG in H.
     destruct H as [S' H].
     destruct H as [[Hps HnextSS'] HgS'S''].
     intros mP T RdyQ tid tcb sp pc jk HS.
     apply HgS'S''.
     apply Hpp' with S;assumption.
Qed.

(* <schedule> : pushw %bx *)
Lemma SCAP_WFcmd_schedule_3_pushwp :
  SCAP_WFcmd (SCAP_consCT (L_schedule+3+1) (Ast L_schedule_4_p L_schedule_4_g sctag) psi_schedule)
    (L_schedule+3) (Ast L_schedule_3_p L_schedule_3_g sctag) 
    (comml (_pushwp (Opr_reg BX))).
Proof.
   unfold SCAP_WFcmd.
  exists L_schedule_4_p;exists L_schedule_4_g.
  split_l.
  split.
  (* enable *)
    unfold enable.
    intros S PS.
    destruct S as [[[M R] fl] isr].
    unfold L_schedule_3_p in PS.
    destruct PS as [T [RdyQ [tid [tcb [sp [pc [jk HS]]]]]]].
    destruct HS as [HS1 [HS2 [HS3 [HS4 [HS5 [HS6 [HS7 HS8]]]]]]].
    exists (MemUpdW M (sp-4) tid 0, RF_update R SP (sp - 4), fl, isr).
    simpl in * .
    apply nextS_pushwp with tid.
      simpl;rewrite HS2;reflexivity.
      unfold pushW.
      rewrite HS3.
      split.
        apply stkSpace_weaken with 6; [idtac | omega].
        sep on.
        sep lift 3 in HS1.
        sep trivial.
      intuition.
      arith_rep (sp-2-2) with (sp-4).
      apply MemUpdW_MemUpd16 with 0;trivial.
  (* p nextS -> p' *)
  assert (Hpp': forall T RdyQ tid tcb sp pc jk mP S S',
    Ast_schedule_3 T RdyQ tid tcb sp pc jk mP S
    -> nextS (L_schedule + 3) (comml (_pushwp (Opr_reg BX))) S S'
    -> Ast_schedule_4 T RdyQ tid tcb sp pc jk mP S').
    unfold Ast_schedule_4, Ast_schedule_3.
    intros T RdyQ tid tcb sp pc jk mP S S' HS Hnext.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    substH Hnext with (nextS_inversion Hnext).
    simpl in * .
    destruct Hnext as [[Hnext1 [Hnext2 [x Hnext3]]] [Hnext4 [Hnext5 Hnext6]]].
    destruct HS as [HS1 [HS2 [HS3 [HS4 [HS5 [HS6 [HS7 HS8]]]]]]].
    subst M' R' fl' isr'.
    rewrite HS3.
    intuition.
    sep on.
    sep lift 3 in HS1.
    sep lift 4.
    assert (MemUpdW M (sp - 4) tid x
      ||- sp - 4 |->w tid *^*
       (sp - 2 |->w ready_Q *^* (sp |-> pc *^* (sp+1|->jk *^* mP)) *^*
        (curr_thrd |->w tid *^* embed_TCB tid (sp, TCB_Status tcb, pc) *^*
         embed_RdyQ T RdyQ *^* stkSpacePrec (sp - 4) 4))).
      sep upd.
    sep matchH HS1 with 13.
    sep stksplit in HS1.
    arith_rep (sp-2-2) with (sp-4) in HS1.
    rewrite HS2.
    arith_rep (sp-2-2) with (sp-4).
    arith_rep (sp-2-2) with (sp-4).
  split.
  (* p ==> p' *)
    unfold pimp.
    unfold linkPG, conjPG, conjG.
    unfold L_schedule_4_p,L_schedule_3_p.
    intros S' H.
    destruct H as [S [H1 H2]].
    destruct H1 as [T [RdyQ [tid [tcb [sp [pc [jk H]]]]]]].
    exists T; exists RdyQ; exists tid; exists tcb; exists sp; exists pc;exists jk.
    apply Hpp' with S; assumption.
  (* g' ==> g *)
     unfold L_schedule_3_g, L_schedule_4_g.
     intros S S'' H.
     unfold linkPG, conjPG, conjG in H.
     destruct H as [S' H].
     destruct H as [[Hps HnextSS'] HgS'S''].
     intros mP T RdyQ tid tcb sp pc jk HS.
     apply HgS'S''.
     apply Hpp' with S;assumption.
Qed.

Lemma subsetQ_TQ_put_sub:
  forall T RdyQ tid tcb,
    subsetQ_TQ RdyQ T ->
  subsetQ_TQ (addThrd RdyQ tid) (TQMod.put T tid tcb).
Proof.
  intros.
  unfold subsetQ_TQ in * .
  unfold addThrd.
  unfold WQMod.put,TQMod.put.
  intro tid0.
  unfold WQMod.in_dom in * .
  unfold TQMod.in_dom in * .
  unfold WQMod.beq_A,TQMod.beq_A.
  unfold WQSpec.beq_Domain,TQSpec.beq_Domain.
  destruct  (eq_nat_dec tid0 tid).
  intro.
  exists tcb;reflexivity.
  apply H.
Qed.

(* <schedule> : call enQueue (L_enQueue) *)

Lemma SCAP_WFcmd_schedule_4_call :
  SCAP_WFcmd (SCAP_consCT (L_schedule+4+1) (Ast L_schedule_5_p L_schedule_5_g sctag) psi_schedule)
    (L_schedule+4) (Ast L_schedule_4_p L_schedule_4_g sctag) 
    (comml (_call    L_enQueue)).
Proof.
  assert(Hpp' : forall T RdyQ tid tcb sp pc jk mP S S',
    Ast_schedule_4 T RdyQ tid tcb sp pc jk mP S
    -> nextS (L_schedule+4) (comml (_call L_enQueue)) S S'
    -> Ast_enQueue_pre T RdyQ ready_Q tid (sp, TCB_Status tcb, pc) (S_RF S' BP)
    (sp - 6) (curr_thrd |->w tid  *^* sp - 6 |->w (L_schedule + 4 + 1) *^*
      sp |-> pc *^*(sp+1 |-> jk *^* mP)) S').
    intros T RdyQ tid tcb sp pc jk mP S S'.
    intros HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    unfold Ast_schedule_4 in HS.
    destruct HS as [HM [Hbx [Hsp [Hspg [HRT [HtR [Hmto HT]]]]]]].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspx [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    subst M' R' fl' isr'.
    unfold Ast_enQueue_pre.
    simplS in * .
    reg simpl.
    split.
      sep on.
      rewrite Hsp.
      arith_rep (sp - 4 - 2) with (sp - 6).
      sep lift 6.
      sep_update.
      sep lift 3 in HM.
      sep stksplit in HM.
      arith_rep (sp - 4 - 2) with (sp - 6) in HM.
      sep stksplit 2 in HM.
      arith_rep (sp - 6 - 2) with (sp - 8) in HM.
      arith_rep (8 - 4 - 2 - 2) with 0 in HM.
      sep lift 1 in HM.
      substH HM with (stkSpacePrec_unit1 HM).
      unfold embed_RdyQ in HM.
      arith_rep (sp - 6 + 2) with (sp - 4).
      arith_rep (sp - 6 + 4) with (sp - 2).
      sep matchH HM with 11.
    split.
      assert (tid <> 0).
        apply HT.
        exists tcb; trivial.
        clear HM Hbx Hsp Hspg Hmto Hlds HSS'.
        induction tid.
          destruct (H (refl_equal _)).
        clear IHtid H.
        omega.
    rewrite Hsp; omega.
  assert (Hpp''' : forall T RdyQ tid tcb sp pc jk
    pc' mP S S' S'' S''',
    Ast_schedule_4 T RdyQ tid tcb sp pc jk mP S
    -> nextS (L_schedule+4) (comml (_call L_enQueue)) S S'
    -> g_enQueue S' S''
    -> nextS pc' (comms (_retw)) S'' S'''
    -> exists T',
      Ast_schedule_5 T T' RdyQ tid tcb sp pc jk mP S''').
    intros T RdyQ tid tcb sp pc jk pc' mP.
    intros S S' S'' S''' HS HSS' HS'S'' HS''S'''.
    exists (TQMod.put T tid (sp, TCB_Status tcb, pc)).
    unfold g_enQueue in HS'S''.
    assert (HS' : Ast_enQueue_pre T RdyQ ready_Q tid (sp, TCB_Status tcb, pc) (S_RF S' BP) (sp-6) 
      (curr_thrd |->w tid  
        *^* sp - 6 |->w (L_schedule + 4 + 1) 
        *^* sp |-> pc 
        *^*(sp+1 |->jk *^* mP)) S').
      apply Hpp' with S; trivial.
    assert (HgS'' := HS'S'' _ _ _ _ _ _ _ _ HS').
    destruct HgS'' as [T' [RdyQ' [HS'' [HT' HRdyQ']]]].
    clear HS'S''.
    unfold Ast_enQueue_post in HS''.
    unfold Ast_schedule_5.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct S'' as [[[M'' R''] fl''] isr''].
    destruct S''' as [[[M''' R'''] fl'''] isr'''].
    destruct HS as [HM [Hbx [Hsp [Hspg HS]]]]. 
    destruct HS as [HRT [HtR [Hmto HT]]].
    simplS in * .
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct (nextS_inversion HS''S''') as [HM''' [HR''' [Hfl''' Hisr''']]].
    destruct HR''' as [pcx [Hlds HR''']].
    destruct HS'' as [HM'' HR''].
    assert (Hpcx : pcx = L_schedule + 4 + 1).
      subst R' M''' R'''.
      clear Hpp' HT HRT HtR Hmto.
      clear Hbx Hsp Hspg HM HS' HS''S''' HM'.
      clear HT' HRdyQ' HSS'.
      subst fl' isr' fl''' isr'''.
      destruct HR'' as [HBP'' HSP''].
      rewrite HSP'' in Hlds.
      sep on.
      sep lift 5 in HM''.
      sep_inj.
      auto.
    subst pcx; clear Hlds.
    subst M''' R''' fl''' isr'''.
    reg simpl.
    destruct HR'' as [HBP'' HSP''].
    rewrite HSP''.
    arith_rep (sp - 6 + 2) with (sp - 4) in HM''.
    arith_rep (sp - 6 + 4) with (sp - 2) in HM''.
    arith_rep (sp - 6 + 2) with (sp - 4).
    split.
      unfold embed_RdyQ.
      rewrite <- HT'.
      rewrite <- HRdyQ'.
      unfold embed_RdyQ in HM'' .
      sep on.
      sep matchH HM'' with 9.
      sep stksplit.
      arith_rep (sp - 4 - 2) with (sp - 6).
      arith_rep (8-4-2) with 2.
      sep matchH HM'' with 3.
    repeat split_l; trivial.
    apply subsetQ_TQ_put_sub;assumption.
  simpl.
  exists p_enQueue; exists L_schedule_5_p.
  exists g_enQueue; exists L_schedule_5_g.
  split.    
    unfold SCAP_lookupCT, SCAP_consCT.
    apply SPSpec.lookup_merge_intro_r.
    apply L_enQueue_in_psi_schedule.
  split_l.
  split.
    clear Hpp' Hpp'''.
    unfold enable, L_schedule_4_p.
    intros S HS.
    destruct HS as [T [RdyQ [tid [tcb [sp [pc [jk HS]]]]]]].
    unfold Ast_schedule_4 in HS.
    destruct HS as [HM [Hbx [Hsp [Hspg [HRT [HtR[Hmto HT]]]]]]].
    (* enable call *)
    destruct S as [[[M R] fl] isr].
    simplS in * .
    exists (MemUpdW M (sp - 4 - 2) (L_schedule+4+1) 0, RF_update R SP (sp - 4 - 2), fl, isr).
    apply nextS_call.
    unfold pushW.
    split.
      apply stkSpacePrec_stkSpace1 with mTrue.
      rewrite Hsp.
      sep on.
      sep lift 3 in HM.
      sep stksplit 2 in HM.
      sep matchH HM with 10.
      split.
      rewrite Hsp.
      apply MemUpdW_MemUpd16 with 0; trivial.
    rewrite Hsp; trivial.
  split.
    (* g' => fun_frame *)
    unfold sp_call_pred, fun_frame.
    unfold linkPG, conjPG.
    intros S' S'' HS'S''.
    destruct HS'S'' as [H HS'S''].
    destruct H as [S [HS HSS']].
    unfold g_enQueue in HS'S''.
    destruct HS as [T [RdyQ [tid [tcb [sp [pc [jk HS]]]]]]].
    assert (HS' :  Ast_enQueue_pre T RdyQ ready_Q tid (sp, TCB_Status tcb, pc)
      (S_RF S' BP) (sp - 6) 
      (curr_thrd |->w tid  
        *^* sp - 6 |->w (L_schedule + 4 + 1) 
        *^* sp |-> pc 
        *^* (sp+1|-> jk *^* mTrue)) S').
      eapply Hpp'; eauto.
    clear Hpp' HS HSS' Hpp'''.
    substH HS'S'' with (HS'S'' _ _ _ _ _ _ _ _ HS').
    destruct HS'S'' as [T' [BlkQ' HS'']].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct S'' as [[[M'' R''] fl''] isr''].
    simplS in * .
    destruct HS'' as [HS'' _].
    unfold Ast_enQueue_pre, Ast_enQueue_post in * .
    destruct HS' as [HM' [Htid Hsp']].
    destruct HS'' as [HM'' [HBP'' HSP'']].
    simplS in * .
    split.
      destruct Hsp' as [HBP' HSP'].
      rewrite HSP''; rewrite HSP'; trivial.
    unfold Ret_Addr.
    exists (L_schedule + 4 + 1).
    simplS.
    split.
      intros pc''.
      destruct Hsp'.
      apply nextPC_retw with (sp - 6); trivial.
      clear HM''.
      sep on.
      sep lift 6 in HM'.
      sep trivial.
    intros pc''.
    destruct Hsp'.
    apply nextPC_retw with (sp - 6); trivial.
    clear HM'.
    sep on.
    sep lift 5 in HM''.
    sep trivial.
  split.
    (* p => p' *)
    intros S' H.
    unfold conjPG, conjG, linkPG in H.
    destruct H as [S [HS HnextSS']].
    unfold L_schedule_4_p in HS.
    unfold p_enQueue.
    destruct HS as [T [RdyQ [tid [tcb [sp [pc [jk HS]]]]]]].
    exists T; exists RdyQ; exists ready_Q; exists tid; exists (sp, TCB_Status tcb, pc).
    exists (S_RF S' BP); exists (sp - 6).
    assert (H :  Ast_enQueue_pre T RdyQ ready_Q tid (sp, TCB_Status tcb, pc)
      (S_RF S' BP) (sp - 6) (curr_thrd |->w tid  *^* sp - 6 |->w (L_schedule + 4 + 1) *^*
      sp |-> pc *^*(sp+1|->jk*^* mTrue)) S').
      eapply Hpp'; eauto.
    clear Hpp''' HS HnextSS'.
    unfold Ast_enQueue_pre in * .
    destruct H as [HM' [Htid Hsp']].
    split.
      sep on.
      sep matchH HM' with 10.
    auto.
  split.
    (* p => p''' *)
    intros pc' S'''.
    unfold linkPG, wrap_fun, conjG.
    intros H.
    destruct H as [S H].
    destruct H as [HS H].
    destruct H as [S'' [H HS''S''']].
    destruct H as [S' [HSS' HS'S'']].
    unfold L_schedule_4_p, L_schedule_5_p in * .
    destruct HS as [T [RdyQ [tid [tcb [sp [pc[jk  HS]]]]]]].
    assert (H : exists T',
      Ast_schedule_5 T T' RdyQ tid tcb sp pc jk mTrue S''').
      eapply Hpp'''; eauto.
    exists T; exists RdyQ.
    exists tid; exists tcb; exists sp; exists pc ;exists jk.
    apply H.
  (* g''' => g *)
  intros pc' S S'''' H.
  unfold conjPG, wrap_fun, conjG, linkPG in H.
  destruct H as [S''' [HS HS'''S'''']].
  destruct HS as [HSx [S'' [Hx HnextS'']]].
  destruct Hx as [S' [HSS' HS'S'']].
  unfold L_schedule_4_g, L_schedule_5_g in * .
  intros mP T RdyQ tid tcb sp pc jk HS.
  assert (H : exists T',
    Ast_schedule_5 T T' RdyQ tid tcb sp pc jk mP S''').
    eapply Hpp'''; eauto.
  destruct H as [T' HS'].
  eapply HS'''S''''; eauto.
Qed.

(* <schedule> : addw $2, %sp *)
Lemma SCAP_WFcmd_schedule_5_addw:
  SCAP_WFcmd (SCAP_consCT (L_schedule+5+1) (Ast L_schedule_6_p L_schedule_6_g sctag) psi_schedule)
    (L_schedule+5) (Ast L_schedule_5_p L_schedule_5_g sctag) 
    (comml (_addw   (Opr_imm 2 ) SP)).
Proof.
  assert (Hpp' : forall T T' RdyQ tid tcb sp pc jk mP S S',
    Ast_schedule_5 T T' RdyQ tid tcb sp pc jk mP S
    -> nextS (L_schedule+5) (comml (_addw   (Opr_imm 2 ) SP)) S S'
    -> Ast_schedule_6 T T' RdyQ tid tcb sp pc jk mP S').
    intros T T' RdyQ  tid tcb sp pc jk mP S S' HS HSS'.
    unfold Ast_schedule_5, Ast_schedule_6 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    simplS in * .
    destruct HS as [HM [Hsp [Hspg [HT' [HRdyQ [Htcb [Hmto Hnin]]]]]]].
    subst M' R' isr'.
    split.
      sep on.
      sep matchH HM with 10.
      sep stksplit.
      arith_rep (sp - 2 - 2) with (sp - 4).
      arith_rep (8 - 2 - 2) with (8 - 4).
      sep matchH HM with 2.
    reg simpl.
    unfold opr_eval.
    split_l.
    split_l by omega.
    repeat split_l; trivial.
  simpl.
  exists L_schedule_6_p; exists L_schedule_6_g.
  split_l.
  split.
    (* enable addw *)
    unfold enable, L_schedule_5_p.
    intros S H.
    destruct H as [T [RdyQ [tid [tcb [sp [pc [jk H]]]]]]]. 
    destruct H as [T' HS].
    destruct HS as [HM [Hsp [Hspg [HT' [HRdyQ [Htcb [Hmto Hnin]]]]]]].
    destruct S as [[[M R] fl] isr].
    simplS in * .
    set (R' := RF_update R SP (2 + R SP)).
    exists (M, R', Clear_Flag fl _ZF, isr).
    apply nextS_addw with 2 (R SP); auto.
    unfold trash_fl.
    right; trivial.
  split. (* p ==> p' *)
    unfold L_schedule_5_p, L_schedule_5_p.
    intros S' H.
    unfold linkPG in H.
    destruct H as [S H].
    simpl in H.
    destruct H as [H HnextS].
    destruct H as [T [RdyQ [tid [tcb [sp [pc [jk H]]]]]]]. 
    destruct H as [T' HS].
    exists T;  exists RdyQ;
    exists tid; exists tcb; exists sp; exists pc;exists jk;exists T'.
    apply Hpp' with S; trivial.
  (* gimp *)
  intros S S'' H.
  unfold conjPG, conjG in H.
  destruct H as [S' [[HSx HnextS] HS'S'']].
  unfold L_schedule_5_g, L_schedule_6_g in * .
  intros.
  eapply HS'S''; eauto.
Qed.

(* <schedule> : call deQueue (L_deQueue) *)
Lemma SCAP_WFcmd_schedule_6_call :
  SCAP_WFcmd (SCAP_consCT (L_schedule+6+1) (Ast L_schedule_7_p L_schedule_7_g sctag) psi_schedule)
    (L_schedule+6) (Ast L_schedule_6_p L_schedule_6_g sctag) 
    (comml (_call L_deQueue)).
Proof.
  assert(Hpp' : forall T RdyQ tid tcb sp pc jk 
    T' mP S S',
    Ast_schedule_6 T T' RdyQ tid tcb sp pc jk mP S
    -> nextS (L_schedule+6) (comml (_call L_deQueue)) S S'
    -> Ast_deQueue_pre T' (addThrd RdyQ tid) ready_Q (S_RF S BP) (sp - 4)
    ((curr_thrd |->w tid)*^* stkSpacePrec (sp-6) 2
        *^* (sp - 4 |->w (L_schedule+6+1))
        *^* (sp |-> pc)*^* (sp+1|->jk *^* mP)) S').
    intros T RdyQ tid tcb sp pc jk T' mP S S'.
    intros HS HSS'.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    unfold Ast_schedule_6 in HS.
    destruct HS as [HM [HT' [Hsp [Hspg [HRT [HtR [Hmto HT]]]]]]].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspx [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    subst M' R' fl' isr'.
    unfold Ast_deQueue_pre.
    simplS in * .
    reg simpl.
    split.
      sep on.
      rewrite Hsp.
      arith_rep (sp - 2 - 2) with (sp - 4).
      sep lift 5.
      sep_update.
      sep lift 2 in HM.
      sep stksplit in HM.
      arith_rep (sp - 2 - 2) with (sp - 4) in HM.
      arith_rep (8 - 2 - 2) with 4 in HM.
      unfold embed_RdyQ in HM.
      arith_rep (sp - 4 + 2) with (sp - 2).
      sep stksplit 2 in HM.
      arith_rep (sp-4-2) with (sp-6) in HM.
      arith_rep (4-2) with 2 in HM.
      sep matchH HM with 10.
    split_l.
    rewrite Hsp; omega.
  assert (Hpp''' : forall T RdyQ tid tcb sp pc jk
    T' pc' mP S S' S'' S''',
    Ast_schedule_6 T T' RdyQ tid tcb sp pc jk mP S
    -> nextS (L_schedule+6) (comml (_call L_deQueue)) S S'
    -> g_deQueue S' S''
    -> nextS pc' (comms (_retw)) S'' S'''
    -> exists RdyQ', exists tid', exists tcb',
      Ast_schedule_7 T T' RdyQ RdyQ' tid tid' 
      tcb tcb' sp pc jk mP S''').
    intros T RdyQ tid tcb sp pc jk T' pc' mP.
    intros S S' S'' S''' HS HSS' HS'S'' HS''S'''.
    unfold g_deQueue in HS'S''.
    assert (HS' : Ast_deQueue_pre T' (addThrd RdyQ tid) ready_Q (S_RF S BP) (sp-4)
      (((curr_thrd |->w tid)*^* stkSpacePrec (sp-6) 2
        *^* (sp - 4 |->w (L_schedule+6+1))
        *^* (sp |-> pc)) *^*(sp+1|->jk *^* mP)) S').
      eapply Hpp'; eauto.
    clear Hpp'.
    substH HS'S'' with (HS'S'' _ _ _ _ _ _ HS') into HS''.
    unfold Ast_schedule_6 in HS.
    destruct HS as [_ [HT' [Hsp [Hspg [HRT [HtR [Hmto HT]]]]]]].
    destruct HS'' as [HS'' | HS''].
      destruct HS'' as [Hempty [HS'' Hbp]].
      substH Hempty with (Hempty tid).
      unfold filterTQ in Hempty.
      unfold addThrd,WQMod.put in Hempty.
      subst T'.
      unfold TQMod.put,WQMod.beq_A,TQMod.beq_A in Hempty.
      unfold WQSpec.beq_Domain,TQSpec.beq_Domain in Hempty.
      destruct (eq_nat_dec tid tid).
      inversion Hempty.
      elim n;reflexivity.
    destruct HS'' as [RdyQ' [tid' [tcb' [HS'' [HRdyQ' Hnin]]]]].
    exists RdyQ'; exists tid'; exists tcb'.
    unfold Ast_deQueue_post_b in HS''.
    unfold Ast_schedule_7 .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct S'' as [[[M'' R''] fl''] isr''].
    destruct S''' as [[[M''' R'''] fl'''] isr'''].
    simplS in * .
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct (nextS_inversion HS''S''') as [HM''' [HR''' [Hfl''' Hisr''']]].
    destruct HR''' as [pcx [Hlds HR''']].
    destruct HS'' as [HM'' [HAX'' [HBP'' [HSP'' Hmto']]]].
    assert (Hpcx : pcx = L_schedule+6 + 1).
      subst R' M''' R'''.
      clear HT' HRT HtR  Hmto.
      clear Hsp Hspg HS' HS''S''' HM'.
      clear HSS'.
      subst fl' isr' fl''' isr'''.
      rewrite HSP'' in Hlds; clear HSP''.
      sep on.
      sep lift 6 in HM''.
      sep_inj.
      auto.
    subst pcx; clear Hlds.
    subst M''' R''' fl''' isr'''.
    reg simpl.
    rewrite HAX''.
    rewrite HSP''.
    arith_rep (sp - 4 + 2) with (sp - 2) in HM''.
    arith_rep (sp - 4 + 2) with (sp - 2).
    split.
      unfold embed_RdyQ.
      sep on.
      sep matchH HM'' with 11.
      sep stksplit.
      arith_rep (sp - 2 - 2) with (sp - 4).
      arith_rep (8 - 2 - 2) with 4.
      sep stksplit 2.
      arith_rep (sp-4-2) with (sp-6).
      arith_rep (4-2) with 2.
      sep matchH HM'' with 3.
    repeat split_l; trivial.
  simpl.
  exists p_deQueue; exists L_schedule_7_p.
  exists g_deQueue; exists L_schedule_7_g.
  split.
    unfold SCAP_lookupCT, SCAP_consCT.
    apply SPSpec.lookup_merge_intro_r.
    apply L_deQueue_in_psi_schedule.
  split_l.
  split.
    (* enable call *)
    clear Hpp' Hpp'''.
    unfold enable, L_schedule_6_p.
    intros S HS.
    destruct HS as [T [RdyQ [tid [tcb [sp [pc [jk [T' HS]]]]]]]].
    unfold Ast_schedule_6 in HS.
    destruct HS as [HM [HT'[Hsp [Hspg _]]]].
    destruct S as [[[M R] fl] isr].
    simplS in * .
    exists (MemUpdW M (sp - 2 - 2) (L_schedule+6+1) 0, RF_update R SP (sp - 2 - 2), fl, isr).
    apply nextS_call.
    unfold pushW.
    split.
      apply stkSpace_weaken with (8 - 2); [idtac | omega].
      apply stkSpacePrec_stkSpace1 with mTrue.
      rewrite Hsp.
      sep on.
      sep matchH HM with 10.
    split.
      rewrite Hsp.
      apply MemUpdW_MemUpd16 with 0; trivial.
    rewrite Hsp; trivial.
  split.
    (* g' => fun_frame *)
    unfold sp_call_pred, fun_frame.
    unfold linkPG, conjPG.
    intros S' S'' HS'S''.
    destruct HS'S'' as [H HS'S''].
    destruct H as [S [HS HSS']].
    unfold g_deQueue in HS'S''.
    destruct HS as [T [RdyQ [tid [tcb [sp [pc[jk [T' HS]]]]]]]].
    assert (HS' : Ast_deQueue_pre T' (addThrd RdyQ tid) ready_Q (S_RF S BP)
      (sp - 4) (((curr_thrd |->w tid)*^* stkSpacePrec (sp-6) 2
        *^* (sp - 4 |->w (L_schedule+6+1))
        *^* (sp |-> pc))*^*(sp+1|->jk *^* mTrue)) S').
      eapply Hpp'; eauto.
    clear Hpp' HSS' Hpp'''.
    substH HS'S'' with (HS'S'' _ _ _ _ _ _ HS') into HS''.
    destruct HS as [_ [HT' [Hsp [Hspg [HRT [HtR [Hmto HT]]]]]]].
    destruct HS'' as [HS'' | HS''].
      destruct HS'' as [Hempty [HS'' Hbp]].
      substH Hempty with (Hempty tid).
      unfold filterTQ in Hempty.
      unfold addThrd,WQMod.put in Hempty.
      subst T'.
      unfold TQMod.put,WQMod.beq_A,TQMod.beq_A in Hempty.
      unfold WQSpec.beq_Domain,TQSpec.beq_Domain in Hempty.
      destruct (eq_nat_dec tid tid).
      inversion Hempty.
      elim n;reflexivity.
    destruct HS'' as [RdyQ' [tid' [tcb' [HS'' [HRdyQ' Hnin']]]]].
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct S'' as [[[M'' R''] fl''] isr''].
    simplS in * .
    unfold Ast_deQueue_pre, Ast_deQueue_post_b in * .
    destruct HS' as [HM' [HBP' HSP']].
    destruct HS'' as [HM'' [HAX'' [HBP'' [HSP'' Hmto']]]].
    simplS in * .
    split.
      rewrite HSP'; rewrite HSP''; trivial.
    unfold Ret_Addr.
    exists (L_schedule + 6 + 1).
    simplS.
    split.
      intros pc''.
      apply nextPC_retw with (sp - 4); trivial.
      clear HM''.
      sep on.
      sep lift 5 in HM'.
      sep trivial.
    intros pc''.
    apply nextPC_retw with (sp - 4); trivial.
    clear HM'.
    sep on.
    sep lift 6 in HM''.
    sep trivial.
  split.
    (* p => p' *)
    intros S' H.
    unfold conjPG, conjG, linkPG in H.
    destruct H as [S [HS HnextSS']].
    unfold L_schedule_6_p in HS.
    unfold p_deQueue.
    destruct HS as [T [RdyQ [tid [tcb [sp [pc [jk [T' HS]]]]]]]].
    exists T'; exists (addThrd RdyQ tid); exists ready_Q; exists (S_RF S BP); 
      exists (sp - 4).
    assert (H : Ast_deQueue_pre T' (addThrd RdyQ tid) ready_Q (S_RF S BP) 
           (sp - 4) (((curr_thrd |->w tid)*^* stkSpacePrec (sp-6) 2
        *^* (sp - 4 |->w (L_schedule+6+1))
        *^* (sp |-> pc))*^*(sp+1|->jk *^* mTrue)) S').
      eapply Hpp'; eauto.
    clear Hpp''' Hpp' HS HnextSS'.
    unfold Ast_deQueue_pre in * .
    destruct H as [HM' H].
    split_r.
    sep on.
    sep matchH HM' with 10.
  split.
    (* p => p''' *)
    intros pc' S'''.
    unfold linkPG, wrap_fun, conjG.
    intros H.
    destruct H as [S H].
    destruct H as [HS H].
    destruct H as [S'' [H HS''S''']].
    destruct H as [S' [HSS' HS'S'']].
    unfold L_schedule_6_p, L_schedule_7_p in * .
    destruct HS as [T [RdyQ [tid [tcb [sp [pc [jk [T' HS]]]]]]]].
    assert (H : exists RdyQ', exists tid', exists tcb',
      Ast_schedule_7 T T' RdyQ RdyQ' tid tid' tcb tcb' sp pc jk mTrue S''').
      eapply Hpp'''; eauto.
    destruct H as [RdyQ' [tid' [tcb' HS''']]].
    exists T; exists RdyQ; exists tid;exists tcb;
      exists sp; exists pc; exists jk;exists T'; exists RdyQ'.
     exists tid'; exists tcb'.
    trivial.
  (* g'' => g *)
  intros pc' S S'''' H.
  unfold conjPG, wrap_fun, conjG, linkPG in H.
  destruct H as [S''' [HS HS'''S'''']].
  destruct HS as [HSx [S'' [H HS'']]].
  destruct H as [S' [HSS' HS'S'']].
  unfold L_schedule_6_g, L_schedule_7_g in * .
  intros mP T T' RdyQ tid tcb sp pc  jk HS.
  assert (H : exists RdyQ', exists tid', exists tcb',
    Ast_schedule_7 T T' RdyQ RdyQ' tid tid' tcb tcb' sp pc jk mP S''').
    eapply Hpp'''; eauto.
  destruct H as [RdyQ' [tid' [tcb' HS''']]].
  apply HS'''S'''' with T' RdyQ' tid' tcb'; trivial.
Qed.

(* <schedule> : addw $2, %sp *)
Lemma SCAP_WFcmd_schedule_7_addw:
  SCAP_WFcmd (SCAP_consCT (L_schedule+7+1) (Ast L_schedule_8_p L_schedule_8_g sctag) psi_schedule)
    (L_schedule+7) (Ast L_schedule_7_p L_schedule_7_g sctag) 
    (comml (_addw   (Opr_imm 2 ) SP)).
Proof.
  assert (Hpp' : forall T T' RdyQ RdyQ' tid tid' tcb tcb' sp pc jk mP S S',
    Ast_schedule_7 T T' RdyQ RdyQ' tid tid' tcb tcb' sp pc jk mP S
    -> nextS (L_schedule+7) (comml (_addw   (Opr_imm 2 ) SP)) S S'
    -> Ast_schedule_8 T T' RdyQ RdyQ' tid tid' tcb tcb' sp pc jk mP S').
    intros T T' RdyQ RdyQ' tid tid' tcb tcb' sp pc jk mP S S' HS HSS'.
    unfold Ast_schedule_6, Ast_schedule_8 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HS as [HM [Hsp [Hspg [HT' [HRdyQ [Htcb [Hmto [Hnin HT]]]]]]]].
    simplS in * .
    subst M' R' isr'.
    split.
      sep on.
      sep matchH HM with 10.
      sep stksplit.
      sep matchH HM with 2.
    reg simpl.
    unfold opr_eval.
    split_l.
    split_l by omega.
    repeat split_l; trivial.
  simpl.
  exists L_schedule_8_p; exists L_schedule_8_g.
  split_l.
  split.
    (* enable addw *)
    unfold enable, L_schedule_7_p.
    intros S H.
    destruct H as [T [RdyQ [tid [tcb [sp [pc [jk [T' [RdyQ' [tid' [tcb' H]]]]]]]]]]]. 
    destruct H as [HM [Hsp [Hspg [HT' [HRdyQ [Htcb [Hmto [Hnin HT]]]]]]]].
    destruct S as [[[M R] fl] isr].
    simplS in * .
    set (R' := RF_update R SP (2 + R SP)).
    exists (M, R', Clear_Flag fl _ZF, isr).
    apply nextS_addw with 2 (R SP); auto.
    unfold trash_fl.
    right; trivial.
  split. (* p ==> p' *)
    unfold L_schedule_7_p, L_schedule_7_p.
    intros S' H.
    unfold linkPG in H.
    destruct H as [S H].
    simpl in H.
    destruct H as [H HnextS].
    destruct H as [T [RdyQ [tid [tcb [sp [pc [jk H]]]]]]]. 
    destruct H as [T' [RdyQ' [tid' [tcb' HS]]]].
    exists T;  exists RdyQ;exists tid; exists tcb; 
      exists sp; exists pc;exists jk;exists T';exists RdyQ';
        exists tid'; exists tcb'.
    apply Hpp' with S; trivial.
  (* gimp *)
  intros S S'' H.
  unfold conjPG, conjG in H.
  destruct H as [S' [[HSx HnextS] HS'S'']].
  unfold L_schedule_7_g, L_schedule_8_g in * .
  intros.
  eapply HS'S''; eauto.
Qed.

(* <schedule> : movw %ax, %bx *)
Lemma SCAP_WFcmd_schedule_8_movwp :
  SCAP_WFcmd (SCAP_consCT (L_schedule+8+1) (Ast L_schedule_9_p L_schedule_9_g sctag) psi_schedule)
    (L_schedule+8) (Ast L_schedule_8_p L_schedule_8_g sctag) 
    (comml (_movwp  (Opr_reg AX) BX)).
Proof.
  assert (Hpp' : forall T RdyQ tid tcb sp pc jk
    T' RdyQ' tid' tcb' mP S S',
    Ast_schedule_8 T T' RdyQ RdyQ' tid tid' tcb tcb' sp pc jk mP S
    -> nextS (L_schedule + 8) (comml (_movwp  (Opr_reg AX) BX)) S S'
    -> Ast_schedule_9 T T' RdyQ RdyQ' tid tid' tcb tcb' sp pc jk mP S').
    intros T RdyQ tid tcb  sp pc jk T' RdyQ' tid' tcb' mP S S' HS HSS'.
    unfold Ast_schedule_8, Ast_schedule_9 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    simplS in * .
    destruct HS as [HM [Hax [Hsp [Hspg [HT' [HBlkQ' [HRdyQ [Hmto Hnin]]]]]]]].
    subst M' R' fl' isr'.
    split_l.
    reg simpl.
    repeat split_l; trivial.
  simpl.
  exists L_schedule_9_p; exists L_schedule_9_g.
  split_l.
  split.
    (* enable movw *)
    unfold enable, L_schedule_8_p.
    intros S H.
    destruct H as [T [RdyQ [tid [tcb [sp [pc[ jk H]]]]]]].
    destruct H as [T' [RdyQ' [tid' [tcb' HS]]]].
    unfold Ast_schedule_8 in HS.
    destruct HS as [HM [Hax [Hsp [Hspg [HT' [HBlkQ' [HRdyQ' [Hmto Hnin]]]]]]]].
    destruct S as [[[M R] fl] isr].
    simplS in * .
    set (R' := RF_update R BX tid').
    exists (M, R', fl, isr).
    apply nextS_movwp with tid'; auto.
  split. (* p ==> p' *)
    unfold L_schedule_9_p, L_schedule_9_p.
    intros S' H.
    unfold linkPG in H.
    destruct H as [S H].
    destruct H as [H HnextS].
    destruct H as [T [RdyQ [tid [tcb [sp [pc [jk H]]]]]]].
    destruct H as [T' [RdyQ' [tid' [tcb' HS]]]].
    exists T; exists RdyQ; exists tid;exists tcb.
    exists sp; exists pc;exists jk; exists T'; exists RdyQ'.
    exists tid'; exists tcb'. 
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold conjPG, conjG in H.
  destruct H as [S' [[_ HnextS] HS'S'']].
  unfold L_schedule_8_g, L_schedule_9_g in * .
  intros.
  eapply HS'S''; eauto.
Qed.

(* <schedule> : movw %ax, (curr_thrd) *)
Lemma SCAP_WFcmd_schedule_9_movwst :
  SCAP_WFcmd (SCAP_consCT (L_schedule+9+1) (Ast L_schedule_10_p L_schedule_10_g sctag) psi_schedule)
    (L_schedule+9) (Ast L_schedule_9_p L_schedule_9_g sctag) 
    (comml (_movwst (Opr_reg AX) (A_imm curr_thrd))).
Proof.
  assert (Hpp' : forall T RdyQ  tid tcb sp pc jk
    T' RdyQ' tid' tcb' mP S S',
    Ast_schedule_9 T T' RdyQ RdyQ' tid tid' tcb tcb' sp pc jk mP S
    -> nextS (L_schedule+9) (comml (_movwst (Opr_reg AX) (A_imm curr_thrd))) S S'
    -> Ast_schedule_10 T T' RdyQ RdyQ' tid tid' tcb tcb' sp pc jk mP S').
    intros T RdyQ tid tcb sp pc jk T' RdyQ' tid' tcb' mP S S' HS HSS'.
    unfold Ast_schedule_9, Ast_schedule_10 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hlds HM'].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simpl in Hlds.
    simplS in * .
    destruct HS as [HM [Hax [Hbx [Hsp [Hspg [HT' [HBlkQ' [HRdyQ' [Htid'tcb' Htid'RdyQ']]]]]]]]].
    subst M' R' fl' isr'.
    split.
      sep on.
      sep lift 0 in HM.
      sep lift 0.
      rewrite Hax.
      sep upd.
      sep matchH HM with 6.
    repeat split_l; trivial.
  simpl.
  exists L_schedule_10_p; exists L_schedule_10_g.
  split_l.
  split.
    (* enable movw *)
    unfold enable, L_schedule_9_p.
    intros S H.
    destruct H as [T [RdyQ [tid [tcb [sp [pc [jk HS]]]]]]].
    destruct HS as [T' [RdyQ' [tid' [tcb' HS]]]].
    destruct HS as [HM [Hax [Hbx [Hsp [Hspg [HT' [HBlkQ' [HRdyQ' [Htid'tcb' Htid'RdyQ']]]]]]]]].
    destruct S as [[[M R] fl] isr].
    simplS in * .
    set (M' := MemUpdW M curr_thrd tid' 0).
    exists (M', R, fl, isr).
    apply nextS_movwst with curr_thrd tid'; auto.
      sep on.
      sep lift 0 in HM.
      sep trivial.
    apply MemUpdW_MemUpd16 with 0; trivial.
  split. (* pimp *)
    unfold L_schedule_9_p, L_schedule_10_p.
    intros S' H.
    unfold linkPG in H.
    destruct H as [S H].
    destruct H as [H HnextS].
    destruct H as [T [RdyQ [tid [tcb [sp [pc [jk HS]]]]]]].
    destruct HS as [T' [RdyQ' [tid' [tcb' HS]]]].
    exists T; exists RdyQ;exists tid;
    exists tcb; exists sp; exists pc;exists jk.
    exists T'; exists RdyQ'.
    exists tid'; exists tcb'.
    apply Hpp' with S; trivial.
  (* gimp *)
  intros S S'' H.
  unfold conjPG, conjG in H.
  destruct H as [S' [[_ HnextS] HSS']].
  unfold L_schedule_9_g, L_schedule_10_g in * .
  intros.
  eapply HSS'; eauto.
Qed.

Lemma SCAP_WFcmd_schedule_10_movwld : 
  SCAP_WFcmd (SCAP_consCT (L_schedule+10+1) (Ast L_schedule_11_p L_schedule_11_g sctag) psi_schedule)
    (L_schedule+10) (Ast L_schedule_10_p L_schedule_10_g sctag) 
    (comml (_movwld (A_reg TCB_STACK BX) SP)).
Proof.
  simpl.
  exists L_schedule_11_p; exists L_schedule_11_g.
  split_l.
  assert (Hpp' : forall T RdyQ tid tcb sp pc jk
    T' RdyQ' tid' tcb' mP S S',
    Ast_schedule_10 T T' RdyQ RdyQ' tid tid' tcb tcb' sp pc jk  mP S
    -> nextS (L_schedule+10) (comml (_movwld (A_reg TCB_STACK BX) SP)) S S'
    -> Ast_schedule_11 T T' RdyQ RdyQ' tid tid' tcb tcb' sp (TCB_SP tcb') pc jk mP S').
    intros T RdyQ tid tcb sp pc  jk T' RdyQ' tid' tcb' mP S S' HS HSS'.
    unfold Ast_schedule_10, Ast_schedule_11 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HR' as [sp' [Hlds HR']].
    destruct HS as [HM [Hbx [Hsp [Hspg [HT' [HH [HBlkQ' [HRdyQ' [Htid'tcb' Htid'RdyQ']]]]]]]]].
    simplS in * .
    assert (Heq : sp' = TCB_SP tcb').
      unfold addr_eval in Hlds.
      rewrite HT' in Hlds.
      unfold TCB_STACK in Hlds.
      destruct tcb' as [[spx' stx'] pcx'].
      unfold embed_TCB in HM.
      clear Hbx Hsp Hspg HT' HBlkQ' HRdyQ' Htid'tcb' Htid'RdyQ'.
      clear HSS'.
      subst M'.
      sep on.
      sep lift 3 in HM.
      substH HM with (ptoTCB_elim1 HM).
      destruct HM as [tst [HM Htst]].
      sep lift 0 in HM.
      sep inj.
      auto.
    subst M' R' fl' isr'.
    split_l.
    split.
      reg simpl.
      trivial.
    repeat split_l; trivial.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_schedule_10_p in HpS.
    destruct HpS as [T [RdyQ [tid [tcb [sp [pc [jk HS]]]]]]].
    destruct HS as [T' [RdyQ' [tid' [tcb' HS]]]].
    unfold Ast_schedule_10 in HS.
    destruct HS as [HM [HBX [HH [HSP [Hspg [HT' [HBlkQ' [HRdyQ' [Hmto Hnin]]]]]]]]].
    simplS in * .
    exists (M, RF_update R SP (TCB_SP tcb'), fl, isr).
    apply nextS_movwld with (tid' + TCB_STACK) (TCB_SP tcb'); trivial.
      simpl.
      rewrite <- Hspg; trivial.
    unfold TCB_STACK.
    unfold embed_TCB in HM.
    sep on.
    sep lift 3 in HM.
    substH HM with (ptoTCB_elim1 HM).
    destruct HM as [tst [HM Htst]].
    sep lift 0 in HM.
    sep trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_schedule_10_p, L_schedule_11_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [T [RdyQ [tid [tcb [sp [pc [jk HS]]]]]]]. 
    destruct HS as [T' [RdyQ' [tid' [tcb' HS]]]].
    exists T; exists RdyQ;exists tid.
    exists tcb; exists sp; exists pc;exists jk.
    exists T'; exists RdyQ'.
    exists tid'; exists tcb'; exists (TCB_SP tcb').
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_schedule_10_g, L_schedule_11_g in * .
  intros.
  eapply HgS'S''; eauto.
Qed.

Lemma SCAP_WFcmd_schedule_11_retw : 
  SCAP_WFcmd psi_schedule (L_schedule+11) 
    (Ast L_schedule_11_p L_schedule_11_g sctag) (comms _retw).
Proof.
  unfold L_schedule_11_g.
  simpl.
  intros S H.
  intros.
  unfold Ast_schedule_11 in H0.
  InvertAll.
  exists T'; exists RdyQ''. 
  exists tid'; exists sp'; exists tcb'.
  repeat split_l.
  trivial.
Qed.

Lemma SCAP_WFiseq_scheduler :
  SCAP_WFiseq psi_schedule L_schedule spec_schedule I_schedule.
Proof.
unfold I_schedule, spec_schedule.
apply scap_isq_seq with L_schedule_1_p L_schedule_1_g.
apply SCAP_WFcmd_schedule_0_movwld.
apply scap_isq_seq with L_schedule_2_p L_schedule_2_g.
apply SCAP_WFcmd_schedule_1_movwst.
assert (H : L_schedule+1+1 = L_schedule+2); [omega | rewrite H; clear H].
apply scap_isq_seq with L_schedule_3_p L_schedule_3_g.
apply SCAP_WFcmd_schedule_2_pushwp.
assert (H : L_schedule+2+1 = L_schedule+3); [omega | rewrite H; clear H].
apply scap_isq_seq with L_schedule_4_p L_schedule_4_g.
apply SCAP_WFcmd_schedule_3_pushwp.
assert (H : L_schedule+3+1 = L_schedule+4); [omega | rewrite H; clear H].
apply scap_isq_seq with L_schedule_5_p L_schedule_5_g.
apply SCAP_WFcmd_schedule_4_call.
assert (H : L_schedule+4+1 = L_schedule+5); [omega | rewrite H; clear H].
apply scap_isq_seq with L_schedule_6_p L_schedule_6_g.
apply SCAP_WFcmd_schedule_5_addw.
assert (H : L_schedule+5+1 = L_schedule+6); [omega | rewrite H; clear H].
apply scap_isq_seq with L_schedule_7_p L_schedule_7_g.
apply SCAP_WFcmd_schedule_6_call.
assert (H : L_schedule+6+1 = L_schedule+7); [omega | rewrite H; clear H].
apply scap_isq_seq with L_schedule_8_p L_schedule_8_g.
apply SCAP_WFcmd_schedule_7_addw.
assert (H : L_schedule+7+1 = L_schedule+8); [omega | rewrite H; clear H].
apply scap_isq_seq with L_schedule_9_p L_schedule_9_g.
apply SCAP_WFcmd_schedule_8_movwp.
assert (H : L_schedule+8+1 = L_schedule+9); [omega | rewrite H; clear H].
apply scap_isq_seq with L_schedule_10_p L_schedule_10_g.
apply SCAP_WFcmd_schedule_9_movwst.
assert (H : L_schedule+9+1 = L_schedule+10); [omega | rewrite H; clear H].
apply scap_isq_seq with L_schedule_11_p L_schedule_11_g.
apply SCAP_WFcmd_schedule_10_movwld.
assert (H : L_schedule+10+1 = L_schedule+11); [omega | rewrite H; clear H].
apply scap_isq_end.
apply SCAP_WFcmd_schedule_11_retw.
Qed.
