(***************************************************************)
(* Coq code for AIM                                            *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

Add LoadPath "..".
Add LoadPath "../CAP0".
Add LoadPath "../SCAP".

Require Import Coqlib.

Require Import Arith.
Require Import Machine.
Require Import Assertion.
Require Import Assertion_Lemmas4.
Require Import Operational_Semantics.
Require Import AIM.
Require Import SCAP.
Require Import QLib.
Require Import Linking.
Require Import Linking_Lemmas3.

Require Import Kernel.

Require Export EnQueueSpec.
Require Export DeQueueSpec.

Definition Ast_schedule_pre (T : TQ) (RdyQ : WQ)(tid : TID) 
  (tcb : Thread) (sp pc jk : nat) (mP : MPred) (S : State) : Prop :=
  ((curr_thrd |->w tid)
    *^* (embed_TCB tid tcb)
    *^* (embed_RdyQ T RdyQ) 
    *^* (stkSpacePrec sp _SW_Stk) 
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk) 
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp
  /\ sp >= _SW_Stk
  /\ subsetQ_TQ RdyQ T
  /\ WQMod.not_in_dom tid RdyQ
  /\ TQMod.maps_to T tid tcb
  /\ WF_TQ T.

Definition L_schedule_pre_p (S : State) : Prop :=
  exists T, exists RdyQ, exists tid, exists tcb, 
    exists sp, exists pc,exists jk,
      Ast_schedule_pre T RdyQ tid tcb sp pc jk mTrue S.

Definition Ast_schedule_post (T' : TQ) (RdyQ' : WQ) 
  (tid' : TID) (tcb' : Thread) (sp' sp pc jk : nat) (mP : MPred) 
  (S' : State) : Prop :=
  ((curr_thrd |->w tid')
    *^* (embed_TCB tid' tcb')
    *^* (embed_RdyQ T' RdyQ')  
    *^* (stkSpacePrec sp _SW_Stk) 
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk) 
    *^* mP) (S_Mem S')
  /\ S_RF S' SP = sp'
  /\ TQMod.maps_to T' tid' tcb'
  /\ TCB_SP tcb' = sp'
  /\ WQMod.not_in_dom tid' RdyQ'.

Definition L_schedule_0_g (S S' : State) : Prop := forall mP,
  forall T RdyQ tid tcb sp pc jk,
    Ast_schedule_pre T RdyQ tid tcb sp pc jk mP S
    -> exists T', exists RdyQ', exists tid', exists sp', exists tcb',
       T' = TQMod.put T tid (sp, TCB_Status tcb, pc)
      /\ addThrd RdyQ tid = addThrd RdyQ' tid'
      /\ Ast_schedule_post T' RdyQ' tid' tcb' sp' sp pc jk mP S'.


(* <schedule>: movw    (curr_thrd), %bx *)
Definition Ast_schedule_1 (T : TQ) (RdyQ : WQ)(tid : TID) 
  (tcb : Thread) (sp pc jk: nat) (mP : MPred) (S : State) : Prop :=
  ((curr_thrd |->w tid)
    *^* (embed_TCB tid tcb)
    *^* (embed_RdyQ T RdyQ) 
    *^* (stkSpacePrec sp _SW_Stk) 
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk)
    *^* mP) (S_Mem S)
  /\ S_RF S BX = tid
  /\ S_RF S SP = sp
  /\ sp >= _SW_Stk
  /\ subsetQ_TQ RdyQ T
  /\ WQMod.not_in_dom tid RdyQ
  /\ TQMod.maps_to T tid tcb
  /\ WF_TQ T.

Definition L_schedule_1_p (S : State) : Prop :=
  exists T, exists RdyQ, exists tid, exists tcb, 
    exists sp, exists pc,exists jk,
      Ast_schedule_1 T RdyQ tid tcb sp pc jk mTrue S.

Definition L_schedule_1_g (S S' : State) := forall mP,
  forall T RdyQ tid tcb sp pc jk,
    Ast_schedule_1 T RdyQ tid tcb sp pc jk mP S
    -> exists T', exists RdyQ', exists tid', exists sp', exists tcb',
      T' = TQMod.put T tid (sp, TCB_Status tcb , pc)
      /\ addThrd RdyQ tid = addThrd RdyQ' tid'
      /\ Ast_schedule_post T' RdyQ' tid' tcb' sp' sp pc jk mP S'.


(* <schedule>: movw    %sp, TCB_STACK(%bx)*)
Definition Ast_schedule_2 (T : TQ) (RdyQ : WQ)(tid : TID) 
  (tcb : Thread) (sp pc jk: nat) (mP : MPred) (S : State) : Prop :=
  ((curr_thrd |->w tid)
    *^* (embed_TCB tid (sp,TCB_Status tcb, pc))
    *^* (embed_RdyQ T RdyQ) 
    *^* (stkSpacePrec sp _SW_Stk) 
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk)
    *^* mP) (S_Mem S)
  /\ S_RF S BX = tid
  /\ S_RF S SP = sp
  /\ sp >= _SW_Stk
  /\ subsetQ_TQ RdyQ T
  /\ WQMod.not_in_dom tid RdyQ
  /\ TQMod.maps_to T tid tcb
  /\ WF_TQ T.

Definition L_schedule_2_p (S : State) : Prop :=
  exists T, exists RdyQ, exists tid, exists tcb, 
    exists sp, exists pc,exists jk,
      Ast_schedule_2 T RdyQ tid tcb sp pc jk mTrue S.

Definition L_schedule_2_g (S S' : State) := forall mP,
  forall T RdyQ tid tcb sp pc jk,
    Ast_schedule_2 T RdyQ tid tcb sp pc jk mP S
    -> exists T', exists RdyQ', exists tid', exists sp', exists tcb',
      T' = TQMod.put T tid (sp, TCB_Status tcb , pc)
      /\ addThrd RdyQ tid = addThrd RdyQ' tid'
      /\ Ast_schedule_post T' RdyQ' tid' tcb' sp' sp pc jk mP S'.

(* <schedule>: pushw  $readyQ *)
Definition Ast_schedule_3 (T : TQ) (RdyQ : WQ)(tid : TID) 
  (tcb : Thread) (sp pc jk: nat) (mP : MPred) (S : State) : Prop :=
  ((curr_thrd |->w tid)
    *^* (embed_TCB tid (sp,TCB_Status tcb, pc))
    *^* (embed_RdyQ T RdyQ) 
    *^* (stkSpacePrec (sp - 2) (_SW_Stk - 2)) 
    *^* (sp - 2 |->w ready_Q)
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk) 
    *^* mP) (S_Mem S)
  /\ S_RF S BX = tid
  /\ S_RF S SP = sp - 2
  /\ sp >= _SW_Stk
  /\ subsetQ_TQ RdyQ T
  /\ WQMod.not_in_dom tid RdyQ
  /\ TQMod.maps_to T tid tcb
  /\ WF_TQ T.

Definition L_schedule_3_p (S : State) : Prop :=
  exists T, exists RdyQ, exists tid, exists tcb, 
    exists sp, exists pc, exists jk,
      Ast_schedule_3 T RdyQ tid tcb sp pc jk mTrue S.

Definition L_schedule_3_g (S S' : State) := forall mP,
  forall T RdyQ tid tcb sp pc jk,
    Ast_schedule_3 T RdyQ tid tcb sp pc jk mP S
    -> exists T', exists RdyQ', exists tid', exists sp', exists tcb',
      T' = TQMod.put T tid (sp, TCB_Status tcb , pc)
      /\ addThrd RdyQ tid = addThrd RdyQ' tid'
      /\ Ast_schedule_post T' RdyQ' tid' tcb' sp' sp pc jk mP S'.


(* <schedule>: pushw %bx *)

Definition Ast_schedule_4 (T : TQ) (RdyQ : WQ)(tid : TID) 
  (tcb : Thread) (sp pc jk : nat) (mP : MPred) (S : State) : Prop :=
  ((curr_thrd |->w tid)
    *^* (embed_TCB tid (sp,TCB_Status tcb, pc))
    *^* (embed_RdyQ T RdyQ) 
    *^* (stkSpacePrec (sp - 4) (_SW_Stk - 4)) 
    *^* (sp - 4 |->w tid)
    *^* (sp - 2 |->w ready_Q)
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk)
    *^* mP) (S_Mem S)
  /\ S_RF S BX = tid
  /\ S_RF S SP = sp - 4
  /\ sp >= _SW_Stk
  /\ subsetQ_TQ RdyQ T
  /\ WQMod.not_in_dom tid RdyQ
  /\ TQMod.maps_to T tid tcb
  /\ WF_TQ T.

Definition L_schedule_4_p (S : State) : Prop :=
  exists T, exists RdyQ, exists tid, exists tcb, 
    exists sp, exists pc, exists jk,
      Ast_schedule_4 T RdyQ tid tcb sp pc jk mTrue S.

Definition L_schedule_4_g (S S' : State) := forall mP,
  forall T RdyQ tid tcb sp pc jk,
    Ast_schedule_4 T RdyQ tid tcb sp pc jk mP S
    -> exists T', exists RdyQ', exists tid', exists sp', exists tcb',
      T' = TQMod.put T tid (sp, TCB_Status tcb , pc)
      /\ addThrd RdyQ tid = addThrd RdyQ' tid'
      /\ Ast_schedule_post T' RdyQ' tid' tcb' sp' sp pc  jk mP S'.


(* <schedule>: call L_enQueue *)
Definition Ast_schedule_5 (T T' : TQ) (RdyQ : WQ)(tid : TID) 
  (tcb : Thread) (sp pc jk : nat) (mP : MPred) (S : State) : Prop :=
  ((curr_thrd |->w tid)
    *^* (embed_RdyQ T'  (addThrd RdyQ tid)) 
    *^* (stkSpacePrec (sp - 4) (_SW_Stk - 4)) 
    *^* (sp - 4 |->w tid)
    *^* (sp - 2 |->w ready_Q)
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk)
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 4
  /\ sp >= _SW_Stk
  /\ subsetQ_TQ (addThrd RdyQ tid) T'
  /\ WQMod.not_in_dom tid RdyQ
  /\ TQMod.maps_to T tid tcb
  /\ T' = TQMod.put T tid (sp, TCB_Status tcb , pc)
  /\ WF_TQ T.

Definition L_schedule_5_p (S : State) : Prop :=
  exists T, exists RdyQ, exists tid, exists tcb, 
    exists sp, exists pc,exists jk, exists T',
      Ast_schedule_5 T T' RdyQ tid tcb sp pc jk mTrue S.

Definition L_schedule_5_g (S S' : State) := forall mP,
  forall T T' RdyQ tid tcb sp pc jk,
    Ast_schedule_5 T T' RdyQ tid tcb sp pc jk mP S
    -> exists T', exists RdyQ', exists tid', exists sp', exists tcb',
      T' = TQMod.put T tid (sp, TCB_Status tcb , pc)
      /\ addThrd RdyQ tid = addThrd RdyQ' tid'
      /\ Ast_schedule_post T' RdyQ' tid' tcb' sp' sp pc jk mP S'.

(* <schedule> : addw $2, %sp *)
Definition Ast_schedule_6 (T T' : TQ) (RdyQ : WQ)(tid : TID) 
  (tcb : Thread) (sp pc jk : nat) (mP : MPred) (S : State) : Prop :=
  ((curr_thrd |->w tid)
    *^* (embed_RdyQ T' (addThrd RdyQ tid)) 
    *^* (stkSpacePrec (sp - 2) (_SW_Stk - 2)) 
    *^* (sp - 2 |->w ready_Q)
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk)
    *^* mP) (S_Mem S)
  /\ T' = TQMod.put T tid (sp, TCB_Status tcb , pc)
  /\ S_RF S SP = sp - 2
  /\ sp >= _SW_Stk
  /\ subsetQ_TQ (addThrd RdyQ tid) T'
  /\ WQMod.not_in_dom tid RdyQ
  /\ TQMod.maps_to T tid tcb
  /\ WF_TQ T.

Definition L_schedule_6_p (S : State) : Prop :=
  exists T, exists RdyQ, exists tid, exists tcb, 
    exists sp, exists pc,exists jk, exists T', 
      Ast_schedule_6 T T' RdyQ tid tcb sp pc jk mTrue S.

Definition L_schedule_6_g (S S' : State) := forall mP,
  forall T T' RdyQ tid tcb sp pc jk,
    Ast_schedule_6 T T' RdyQ  tid tcb sp pc jk mP S
    -> exists T', exists RdyQ', exists tid', exists sp', exists tcb',
      T' = TQMod.put T tid (sp, TCB_Status tcb , pc)
      /\ addThrd RdyQ tid = addThrd RdyQ' tid'
      /\ Ast_schedule_post T' RdyQ' tid' tcb' sp' sp pc jk mP S'.


(* <schedule> : call L_deQueue *)
Definition Ast_schedule_7 (T T' : TQ) (RdyQ RdyQ' : WQ)(tid tid': TID) 
  (tcb tcb': Thread) (sp pc jk : nat) (mP : MPred) (S : State) : Prop :=
  ((curr_thrd |->w tid)
    *^* (embed_TCB tid' tcb')
    *^* (embed_RdyQ T' RdyQ') 
    *^* (stkSpacePrec (sp - 2) (_SW_Stk - 2)) 
    *^* (sp - 2 |->w ready_Q)
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk)
    *^* mP) (S_Mem S)
  /\ T' = TQMod.put T tid (sp, TCB_Status tcb , pc)
  /\ addThrd RdyQ tid = addThrd RdyQ' tid'
  /\ S_RF S AX = tid'
  /\ S_RF S SP = sp - 2
  /\ TQMod.maps_to T' tid' tcb'
  /\ TQMod.maps_to T tid tcb
  /\ sp >= _SW_Stk
  /\ WQMod.not_in_dom tid' RdyQ'.

Definition L_schedule_7_p (S : State) : Prop :=
  exists T, exists RdyQ, exists tid, exists tcb, 
    exists sp, exists pc,exists jk, exists T', exists RdyQ'',
      exists tid', exists tcb',
      Ast_schedule_7 T T' RdyQ RdyQ'' tid tid' tcb tcb' sp pc jk mTrue S.

Definition L_schedule_7_g (S S' : State) := forall mP,
  forall T T' RdyQ RdyQ'' tid tid' tcb tcb' sp pc jk,
    Ast_schedule_7 T T' RdyQ RdyQ'' tid tid' tcb tcb' sp pc jk mP S
    -> exists T', exists RdyQ', exists tid', exists sp', exists tcb',
      T' = TQMod.put T tid (sp, TCB_Status tcb , pc)
      /\ addThrd RdyQ tid = addThrd RdyQ' tid'
      /\ Ast_schedule_post T' RdyQ' tid' tcb' sp' sp pc jk mP S'.


(* <schedule> : addw $2, %sp*)
Definition Ast_schedule_8 (T T' : TQ) (RdyQ RdyQ' : WQ)(tid tid': TID) 
  (tcb tcb': Thread) (sp pc jk : nat) (mP : MPred) (S : State) : Prop :=
  ((curr_thrd |->w tid)
    *^* (embed_TCB tid' tcb')
    *^* (embed_RdyQ T' RdyQ') 
    *^* (stkSpacePrec sp _SW_Stk) 
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk)
    *^* mP) (S_Mem S)
  /\ S_RF S AX = tid'
  /\ S_RF S SP = sp
  /\ sp >= _SW_Stk
  /\ T' = TQMod.put T tid (sp, TCB_Status tcb , pc)
  /\ addThrd RdyQ tid = addThrd RdyQ' tid'
  /\ TQMod.maps_to T' tid' tcb'
  /\ TQMod.maps_to T tid tcb
  /\ WQMod.not_in_dom tid' RdyQ'.

Definition L_schedule_8_p (S : State) : Prop :=
  exists T, exists RdyQ, exists tid, exists tcb, 
    exists sp, exists pc, exists jk, exists T', exists RdyQ'',
      exists tid', exists tcb',
      Ast_schedule_8 T T' RdyQ RdyQ'' tid tid' tcb tcb' sp pc jk mTrue S.

Definition L_schedule_8_g (S S' : State) := forall mP,
  forall T T' RdyQ RdyQ'' tid tid' tcb tcb' sp pc jk,
    Ast_schedule_8 T T' RdyQ RdyQ'' tid tid' tcb tcb' sp pc jk mP S
    -> exists T', exists RdyQ', exists tid', exists sp', exists tcb',
      T' = TQMod.put T tid (sp, TCB_Status tcb , pc)
      /\ addThrd RdyQ tid = addThrd RdyQ' tid'
      /\ Ast_schedule_post T' RdyQ' tid' tcb' sp' sp pc jk mP S'.

(* <schedule> : movw %ax, %bx *)
Definition Ast_schedule_9 (T T' : TQ) (RdyQ RdyQ' : WQ)(tid tid': TID) 
  (tcb tcb': Thread) (sp pc jk : nat) (mP : MPred) (S : State) : Prop :=
  ((curr_thrd |->w tid)
    *^* (embed_TCB tid' tcb')
    *^* (embed_RdyQ T' RdyQ') 
    *^* (stkSpacePrec sp _SW_Stk) 
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk)
    *^* mP) (S_Mem S)
  /\ S_RF S AX = tid'
  /\ S_RF S BX = tid'
  /\ S_RF S SP = sp
  /\ T' = TQMod.put T tid (sp, TCB_Status tcb , pc)
  /\ addThrd RdyQ tid = addThrd RdyQ' tid'
  /\ TQMod.maps_to T' tid' tcb'
  /\ TQMod.maps_to T tid tcb
  /\ sp >= _SW_Stk
  /\ WQMod.not_in_dom tid' RdyQ'.

Definition L_schedule_9_p (S : State) : Prop :=
  exists T, exists RdyQ, exists tid, exists tcb, 
    exists sp, exists pc ,exists jk, exists T', exists RdyQ'',
      exists tid', exists tcb',
      Ast_schedule_9 T T' RdyQ RdyQ'' tid tid' tcb tcb' sp pc jk mTrue S.

Definition L_schedule_9_g (S S' : State) := forall mP,
  forall T T' RdyQ RdyQ'' tid tid' tcb tcb' sp pc jk,
    Ast_schedule_9 T T' RdyQ RdyQ'' tid tid' tcb tcb' sp pc jk mP S
    -> exists T', exists RdyQ', exists tid', exists sp', exists tcb',
      T' = TQMod.put T tid (sp, TCB_Status tcb , pc)
      /\ addThrd RdyQ tid = addThrd RdyQ' tid'
      /\ Ast_schedule_post T' RdyQ' tid' tcb' sp' sp pc jk mP S'.


(* <schedule> : movw %ax, (curr_thrd) *)
Definition Ast_schedule_10 (T T' : TQ) (RdyQ RdyQ' : WQ)(tid tid': TID) 
  (tcb tcb': Thread) (sp pc jk: nat) (mP : MPred) (S : State) : Prop :=
  ((curr_thrd |->w tid')
    *^* (embed_TCB tid' tcb')
    *^* (embed_RdyQ T' RdyQ') 
    *^* (stkSpacePrec sp _SW_Stk) 
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk)
    *^* mP) (S_Mem S)
  /\ T' = TQMod.put T tid (sp, TCB_Status tcb , pc)
  /\ addThrd RdyQ tid = addThrd RdyQ' tid'
  /\ S_RF S AX = tid'
  /\ S_RF S BX = tid'
  /\ S_RF S SP = sp
  /\ TQMod.maps_to T' tid' tcb'
  /\ sp >= _SW_Stk
  /\ TQMod.maps_to T tid tcb
  /\ WQMod.not_in_dom tid' RdyQ'
(*  /\ subsetQ_TQ RdyQ T
  /\ WQMod.not_in_dom tid RdyQ
  /\ WF_TQ T*).

Definition L_schedule_10_p (S : State) : Prop :=
  exists T, exists RdyQ, exists tid, exists tcb, 
    exists sp, exists pc,exists jk, exists T', exists RdyQ'',
      exists tid', exists tcb',
      Ast_schedule_10 T T' RdyQ RdyQ'' tid tid' tcb tcb' sp pc jk mTrue S.

Definition L_schedule_10_g (S S' : State) := forall mP,
  forall T T' RdyQ RdyQ'' tid tid' tcb tcb' sp pc jk,
    Ast_schedule_10 T T' RdyQ RdyQ'' tid tid' tcb tcb' sp pc jk mP S
    -> exists T', exists RdyQ', exists tid', exists sp', exists tcb',
      T' = TQMod.put T tid (sp, TCB_Status tcb , pc)
      /\ addThrd RdyQ tid = addThrd RdyQ' tid'
      /\ Ast_schedule_post T' RdyQ' tid' tcb' sp' sp pc jk mP S'.


(* <schedule> :  movw    TCB_STACK(%bx), %sp*)
Definition Ast_schedule_11 (T T' : TQ) (RdyQ RdyQ' : WQ)(tid tid': TID) 
  (tcb tcb': Thread) (sp sp' pc jk: nat) (mP : MPred) (S : State) : Prop :=
  ((curr_thrd |->w tid')
    *^* (embed_TCB tid' tcb')
    *^* (embed_RdyQ T' RdyQ') 
    *^* (stkSpacePrec sp _SW_Stk) 
    *^* (sp     |-> pc) 
    *^* (sp + 1 |-> jk)
    *^* mP) (S_Mem S)
  /\ T' = TQMod.put T tid (sp, TCB_Status tcb , pc)
  /\ addThrd RdyQ tid = addThrd RdyQ' tid'
  /\ S_RF S AX = tid'
  /\ S_RF S BX = tid'
  /\ TCB_SP tcb' = sp'
  /\ S_RF S SP = sp'
  /\ TQMod.maps_to T' tid' tcb'
  /\ TQMod.maps_to T tid tcb
  /\ sp >= _SW_Stk
  /\ WQMod.not_in_dom tid' RdyQ'
(*  /\ subsetQ_TQ RdyQ T
  /\ WQMod.not_in_dom tid RdyQ
  /\ TQMod.maps_to T tid tcb
  /\ WF_TQ T*).

Definition L_schedule_11_p (S : State) : Prop :=
  exists T, exists RdyQ, exists tid, exists tcb, 
    exists sp, exists pc,exists jk, exists T', exists RdyQ'',
      exists tid', exists tcb', exists sp',
      Ast_schedule_11 T T' RdyQ RdyQ'' tid tid' tcb tcb' sp sp' pc jk mTrue S.

Definition L_schedule_11_g (S S' : State) := forall mP,
  forall T T' RdyQ RdyQ'' tid tid' tcb tcb' sp sp' pc jk,
    Ast_schedule_11 T T' RdyQ RdyQ'' tid tid' tcb tcb' sp sp' pc jk mP S
    -> exists T', exists RdyQ', exists tid', exists sp', exists tcb',
      T' = TQMod.put T tid (sp, TCB_Status tcb , pc)
      /\ addThrd RdyQ tid = addThrd RdyQ' tid'
      /\ Ast_schedule_post T' RdyQ' tid' tcb' sp' sp pc jk mP S'.


(* <schedule> : retw *)

Lemma L_schedule_pre_p_sound : forall S,
  Linking.SW_Prd S -> L_schedule_pre_p S.
Proof.
  intros S HS.
  unfold L_schedule_pre_p.
  unfold SW_Prd in HS.
  destruct HS as [T [RdyQ [tid [sp [tcb [pc [[jk P1] [P2 [P3 [P4 [P5 P6]]]]]]]]]]].
  unfold Ast_schedule_pre.
  exists T; exists RdyQ; exists tid; exists tcb; exists sp; exists pc;exists jk.
  unfold SW_MPre in P1.
  intuition.
  sep on.
  sep lift 0 in P1.
  substH P1 with (embed_curr_TCB_elim1 (S_Mem S) T tid  (embed_RdyQ T RdyQ *^*
    (sp     |-> pc 
    *^* (sp + 1 |-> jk *^* (stkSpacePrec sp 8 *^* mTrue)))) P1).
  destruct P1 as [tcbx [E P1]].
  assert (tcbx = tcb).
      unfold TQMod.maps_to in * .
      rewrite E in P2.
      inversion P2.
      reflexivity.
  subst tcbx.
  unfold embed_TCB.
  sep matchH P1 with 8.
  sep on.
  sep lift 4 in P1.
  assert (P : (S_Mem S) ||- stkSpace sp 8).
  sep trivial.
  destruct P.
  assumption.
Qed.

Lemma L_schedule_0_g_sound : forall S S',
  L_schedule_0_g S S' -> Linking.SW_grt S S'.
Proof.
  unfold L_schedule_0_g,SW_grt.
  intros S S' HSS'.
  intros M0 M1 HJM T T' RdyQ tid sp tcb pc jk.
  unfold SW_Prd_aux.
  intros H.
  destruct H as [HMPre [Hmto [HRT [HSP [HtR  HT]]]]].
  intros HT'.
  substH HSS' with (HSS' (mEq M1) T RdyQ tid tcb sp pc jk).
  assert (Ast_schedule_pre T RdyQ tid tcb sp pc jk (mEq M1) S).
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    unfold SW_MPre in HMPre.
    simplS in * .
    split.
      simplS.
      exists M0; exists M1; split_l.
      rsplit.
        unfold mEq; trivial.
      sep on.
      sep lift 0 in HMPre.
      substH HMPre with (embed_curr_TCB_elim1 _ _ _ _ HMPre).
      destruct HMPre as [tcbx [Hmtox HM]].
      assert (tcbx = tcb).
        unfold TQMod.maps_to in * .
        rewrite Hmto in Hmtox.
        injection Hmtox; auto.
      subst tcbx.
      unfold embed_TCB.
      sep matchH HM with 10.
    simplS.
    split_l.
    split.
      sep on.
      sep lift 4 in HMPre.
      assert (HMx : M0 ||- stkSpace sp 8).
      sep trivial.
      destruct HMx; trivial.
    repeat split_l; trivial.
  substH HSS' with (HSS' H).
  destruct HSS' as [Tx' [RdyQx' [tid' [sp' [tcb' Hx]]]]].
  destruct Hx as [HTx' [HRdyQ HS']].
  unfold Ast_schedule_post in HS'.
  destruct HS' as [HM' [Hsp' [Hmto' [Htcb' Hnin']]]].
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  simplS in * .
  destruct HM' as [M2' [M1' [HJM' [HM2' HM1']]]].
  unfold mEq in HM1'; subst M1'.
  exists M2'; exists RdyQx'; exists tid'; exists sp'; exists tcb'.
  split_l.
  split_l.
  unfold SW_Post_aux.
  unfold SW_MPost.
  simplS.
  assert (Tx' = T').
    subst T' Tx'; trivial.
  clear HTx'.
  subst Tx'.
  intuition.
  sep on.
  sep lift 0.
  apply embed_curr_TCB_intro1 with tcb'; trivial.
  unfold embed_TCB in HM2'.
  sep matchH HM2' with 8.
Qed.

Definition spec_schedule := Ast L_schedule_pre_p L_schedule_0_g sctag.

Definition psi_schedule  :=
  (SCAP_consCT L_enQueue (Ast p_enQueue g_enQueue sctag)
  (SCAP_consCT L_deQueue (Ast p_deQueue g_deQueue sctag)
    SCAP_emptCT)).

Lemma L_deQueue_in_psi_schedule : 
  SCAP_lookupCT psi_schedule L_deQueue (Ast p_deQueue g_deQueue sctag).
Proof.
  unfold SCAP_lookupCT, psi_schedule.
  unfold SCAP_consCT.
  apply SPSpec.lookup_merge_intro_r.
  simpl.
  SPSpec.lookup_solve.
Qed.

Lemma L_enQueue_in_psi_schedule : 
  SCAP_lookupCT psi_schedule L_enQueue (Ast p_enQueue g_enQueue sctag).
Proof.
  unfold SCAP_lookupCT, psi_schedule.
  unfold SCAP_consCT.
  simpl.
  SPSpec.lookup_solve.
Qed.
