(***************************************************************)
(* Coq code for AIM                                            *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(*******************************************)
(** Prove Timer (timer interrupt handler) with AIM  *******)
(** $Id: Timer.v,v 1.6 2008/04/04 14:36:16 ssg05 Exp $ **)
Add LoadPath "..".
Add LoadPath "../CAP0".
Add LoadPath "../SCAP".

Require Import Coqlib.

Require Import Arith.
Require Import Machine.
Require Import Assertion.
Require Import Operational_Semantics.

Require Import AIM.
Require Import SCAP.
Require Import SCAP_Itr.
Require Import Linking.

Require Import OpSem_Lemmas.
Require Import Assertion_Lemmas.
Require Import Assertion_Lemmas2.
Require Import Assertion_Lemmas4.
Require Import Assertion_Lemmas5.
Require Import Reg.

(** Should be moved to Assertion_Lemmas4.v **)
Notation Local " i ; I " := (seq i I) 
  (at level 60, right associativity).

Notation Local " i ;; " := (sseq i)
  (at level 60).

(** Functioon: timer, timer interrupt handler**)
(********Source Code *************************** 
# timer(q)
1127: timer:
        pushw   %ax
        pushw   %bx
        pushw   %cx
1130:   pushw   %dx
        pushw   %di
        pushw   %si
        pushw   %bp
        movb    $0x20, %al
1135:   movw    $0x20, %dx
        outb    %al, %dx                # send EOI
        call    scheduler               # instruction _SWITCH
        popw    %bp
        popw    %si
1140:   popw    %di
        popw    %dx
        popw    %cx
        popw    %bx
        popw    %ax
1145:   iret
************************************************)
(** Instruction Label **)
Definition I_timer := 
  _pushwp (Opr_reg AX);
  _pushwp (Opr_reg BX);
  _pushwp (Opr_reg CX);
  _pushwp (Opr_reg DX);
  _pushwp (Opr_reg DI);
  _pushwp (Opr_reg SI);
  _pushwp (Opr_reg BP);
  _movwp  (Opr_imm 32) AX;
  _movwp  (Opr_imm 32) DX;
  _outb_eoi;
  _call _SWITCH;
  _popw           (BP);
  _popw           (SI);
  _popw           (DI);
  _popw           (DX);
  _popw           (CX);
  _popw           (BX);
  _popw           (AX);
  _iret;; .

Definition IHpred_alt : Pred :=
  fun (S : State) =>
    (mstar mTrue (mstar INV0 (stkSpacePrec (S_RF S SP) IH_Stk))) (S_Mem S)
      /\ (S_ISR S) = true
      /\ (S_Flag S _IF) = false.

Definition ihtag    := 0.

Notation s_timer    := (Ast IHpred_alt IHgrt ihtag).
Notation Psi_timer  :=  SCAP_emptCT.

Definition timer_post (sp : Label) (S' : State) : Prop :=
  (INV0 *^* stkSpacePrec sp IH_Stk) (S_Mem S') /\ S_ISR S' = false.

Definition ih_gy_alt : Grt :=
  fun (S S' : State) =>
    exists sp, (forall M0 M1, MemMod.disjUnion M0 M1 (S_Mem S)
      -> (INV0 *^* (stkSpacePrec sp _SW_Stk)) M0
      -> exists M0', MemMod.disjUnion M0' M1 (S_Mem S')
        /\ (INV0 *^* (stkSpacePrec sp _SW_Stk)) M0')
        /\ S_Flag S _IF = S_Flag S' _IF
        /\ S_ISR S = S_ISR S'
        /\ S_RF S SP = sp
        /\ S_RF S SP = S_RF S' SP.

Lemma ih_pre_imply:
  forall st, IHpred st -> IHpred_alt st.
Proof.
intros st [pre [inIsr IFtrue]].
split; auto.
assert ((mTrue *^*  (INV0 *^* stkSpacePrec (S_RF st SP + 6) (IH_Stk + 6))) (S_Mem st)).
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 exists M1. exists M2. split_l. repeat split; auto.
sep on.
 sep lift 2 in H.
 assert ( S_Mem st ||- 
   stkSpacePrec (S_RF st SP + 6) 6 *^* stkSpacePrec (S_RF st SP + 6 - 6) (IH_Stk + 6 - 6) *^* (mTrue *^* INV0)).
 apply stkSpacePrec_splith1; auto.
 omega.
 arith_rep (S_RF st SP + 6 - 6) with (S_RF st SP) in H0.
 replace (IH_Stk + 6 - 6) with (IH_Stk) in H0.
 assert ( S_Mem st ||-
   (mTrue *^* stkSpacePrec (S_RF st SP + 6) 6) *^* (INV0 *^* stkSpacePrec (S_RF st SP) IH_Stk)).
  sep matchH H0 with 4.
 sep off.
destruct H1 as [M1 [M2 [djh [mM1 mM2]]]].
 exists M1. exists M2. split_l. repeat split; auto.
simpl; trivial.
Qed.

Lemma ih_gy_derive:
  forall st st', ih_gy st st' -> ih_gy_alt st st'.
Proof.
unfold ih_gy, ih_gy_alt.
intros st st' ih.
destruct ih as [sp [pre [eqFlg [eqisr [eqsp eqsp']]]]].
exists sp.
split; auto.
intros M1 M2 djh mM1.
unfold prepost in pre.
destruct pre as [Ma [Mb [Ma' [mMa [mMa' [djha djh']]]]]].
exists Ma'. split_r.

assert (Hpre: precise (INV0 *^* stkSpacePrec sp _SW_Stk)).
 apply mstar_presv_prec.
  apply Prec0.
  apply precise_stkSpacePrec.
substH Hpre with (Hpre M1 Ma (S_Mem st)).

assert (eqM: M2 = Mb).
 apply MemMod.disjUnion_unique with M1 Ma (S_Mem st); auto.
 apply Hpre; auto.
 apply MemMod.disjUnion_subseteq_left with M2; auto.
 apply MemMod.disjUnion_subseteq_left with Mb; auto.
rewrite eqM; auto.
Qed.
