(***************************************************************)
(* Coq code for AIM                                            *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(*******************************************)
(** Prove Timer (timer interrupt handler) with AIM  *******)
(** $Id: TimerProof.v,v 1.9 2008/04/04 14:36:16 ssg05 Exp $ **)
Add LoadPath "..".
Add LoadPath "../CAP0".
Add LoadPath "../SCAP".

Require Import Coqlib.

Require Import Arith.
Require Import Machine.
Require Import Assertion.
Require Import Operational_Semantics.

Require Import AIM.
Require Import SCAP.
Require Import SCAP_Itr.
Require Import Linking.

Require Import OpSem_Lemmas.
Require Import Assertion_Lemmas.
Require Import Assertion_Lemmas2.
Require Import Assertion_Lemmas4.
Require Import Assertion_Lemmas5.
Require Import Reg.

Require Import Timer.

Definition timer_pushbx_MPre (sp ax : nat) : MPred :=
  fun (M : Mem) => (INV0 *^* stkSpacePrec (sp-2) (IH_Stk-2) *^* sp-2 |->w ax) M.

Definition timer_pushbx_pre (sp ax : nat) (S : State) : Prop :=
  timer_pushbx_MPre sp ax (S_Mem S)
  /\ S_RF S AX = ax
  /\ S_RF S SP = sp - 2
  /\ sp >= IH_Stk /\ S_ISR S = true /\ S_Flag S _IF = false.

Definition p_timer_pushbx (S : State) : Prop :=
  exists sp, exists ax, imprec (timer_pushbx_pre sp ax) S.

Definition timer_pushbx_post (sp ax : nat) (S S':State) (M0':Mem) : Prop :=
  timer_post sp (S_Mem_Upd S' M0') 
	/\ S_RF S' AX = ax  	  /\ S_RF S' BX = S_RF S BX
	/\ S_RF S' CX = S_RF S CX /\ S_RF S' DX = S_RF S DX
	/\ S_RF S' DI = S_RF S DI /\ S_RF S' SI = S_RF S SI
	/\ S_RF S' BP = S_RF S BP /\ S_RF S' SP = sp.

Definition g_timer_pushbx (S S' : State) : Prop := 
  forall M0 M1, MemMod.disjUnion M0 M1 (S_Mem S)
    -> forall sp ax, timer_pushbx_pre sp ax (S_Mem_Upd S M0)
      -> exists M0', MemMod.disjUnion M0' M1 (S_Mem S') 
	/\ timer_pushbx_post sp ax S S' M0'.

Lemma timer_pushax_has_stack:
  forall R h,
    (INV0 *^* stkSpacePrec (R SP) IH_Stk) h
    -> R SP >=  IH_Stk /\ (exists ax, ((R SP-2) |~>w ax) h).
Proof.
intros R h pre.
assert (R SP >= IH_Stk).
 destruct pre as [M1 [M2 [djh [mM1 [spge mM2]]]]]. auto. split_l.
sep on.
 assert (Hx : h ||- stkSpacePrec (R SP - 2) (IH_Stk - 2)
    *^* R SP - 2 |->w_  *^* INV0).
  assocr.
  apply stkSpacePrec_splitW1.
  arith_rep (R SP - 2 + 2) with (R SP).
  replace (IH_Stk - 2 + 2) with (IH_Stk).
  sep matchH pre with 2.
  simpl; trivial.
 sep lift 1 in Hx.
 sep elim_left Hx.
 exists v.
 sep on.
 sep trivial.
Qed.

Lemma timer_pushax_pre2post:
  forall R h h',
    (INV0 *^* stkSpacePrec (R SP) IH_Stk) h
    -> MemUpd16 h (R SP - 2) (R AX) h' -> R SP >= IH_Stk
    ->  timer_pushbx_MPre (R SP) (R AX) h'.
Proof.
intros R h h' pre HM' spge.
substH HM' with (MemUpd16_MemUpdW HM').
destruct HM' as [junk HM'].
unfold timer_pushbx_MPre.
  subst h'.
  sep on.
  assert (Hx : h ||- stkSpacePrec (R SP - 2) (IH_Stk - 2)
    *^* R SP - 2 |->w_  *^* INV0).
  assocr.
  apply stkSpacePrec_splitW1.
  arith_rep (R SP - 2 + 2) with (R SP).
  arith_rep (IH_Stk - 2 + 2) with (IH_Stk).
  sep matchH pre with 2.
  sep lift 2.
  sep upd.
  sep matchH Hx with 5.
Qed.

Lemma IH_WFcmd_1127_pushax:
  IH_WFcmd (SCAP_consCT (h_entry + 1) (Ast p_timer_pushbx g_timer_pushbx ihtag) Psi_timer) 
     h_entry s_timer (comml (_pushwp (Opr_reg AX))).
Proof.
exists p_timer_pushbx. exists g_timer_pushbx.
split. constructor. auto.
split.
 intros st [pre [inIsr IFtrue]].
 destruct pre as [M1 [M2 [djh [pre' mM2]]]].
 induction st as (((h, R), fl), isr); UnfoldStateAll.
 destruct (timer_pushax_has_stack R M2 mM2) as [spge [ax ld2sp]].
 exists ((MemMod.put (MemMod.put h (R SP -2) (opr_eval R (Opr_reg AX))) (R SP-2+1)(opr_eval R (Opr_reg AX))), (RF_update R SP (R SP-2)), fl, isr).
 apply nextS_pushwp with (opr_eval R (Opr_reg AX)); auto.
 unfold pushW. split.
  split_l. 
  omega.
  apply leadsToW_leadsToBlk with ax; auto.
  apply leadsToW_Mono with M2 M1; auto. apply MemMod.disjUnion_commut; auto.
 unfold opr_eval, MemUpd16. split_r.
 exists (opr_eval R (Opr_reg AX)); auto.
split. (** pimp **)
 intros st' [st [pS nS]].
 destruct pS as [pre [inIsr IFtrue]].
 destruct pre as [M1 [M2 [djh [pre' mM2]]]].
 destruct (timer_pushax_has_stack (S_RF st) M2 mM2) as [spge [ax ld2sp]].

 induction st as (((h, R), fl), isr);  UnfoldStateAll.
 induction st' as (((h', R'), fl'), isr').
 destruct (nextS_inversion nS).
 clear nS. 
 unfold opr_eval in H. destruct H as [spgt2' [ldt memUp]]. 
 destruct H0 as [rfUp [eqfl eqisr]].
 exists (R SP). exists (R AX).

 assert (leadsToxW (R SP -2) M2).
 apply leadsToW_leadsToxW with ax; auto.
 destruct (leadsToxW_MemUpd16_DMonoR djh H memUp) as [M2' [djh' upM2]].

 rewrite rfUp. rewrite eqfl. rewrite eqisr.
 exists M2'. exists M1. 
 split. apply MemMod.disjUnion_commut; auto. split_r.
 repeat split; auto.
 UnfoldStateAll. 
 apply timer_pushax_pre2post with M2; auto.
(** gimp **)
 intros st st''  [st' [[pS nS] gS]].
 destruct pS as [pre [inIsr IFtrue]].  (** 1. pS -> nS -> pS', S.H = M1 *^* M2 **)
 destruct pre as [M1 [M2 [djh [pre' mM2]]]].
 destruct (timer_pushax_has_stack (S_RF st) M2 mM2) as [spge [ax ld2sp]].

 induction st as (((h, R), fl), isr);  UnfoldStateAll.
 induction st' as (((h', R'), fl'), isr').
 destruct (nextS_inversion nS).
 clear nS. 
 unfold opr_eval in H. destruct H as [spgt2' [ldt memUp]]. 
 destruct H0 as [rfUp [eqfl eqisr]].

 assert (leadsToxW (R SP -2) M2).
 apply leadsToW_leadsToxW with ax; auto.
 destruct (leadsToxW_MemUpd16_DMonoR djh H memUp) as [M2' [djh' upM2]].

 assert (timer_pushbx_pre (R SP) (R AX) (M2', R', fl', isr')).
  rewrite rfUp. rewrite eqfl. rewrite eqisr.
  repeat split; auto.
  UnfoldStateAll. unfold RF_update; unfold beq_reg.
  apply timer_pushax_pre2post with M2; auto.
 rename H0 into mM2'.
 assert (MemMod.disjUnion M2' M1 h').
   apply MemMod.disjUnion_commut; auto. clear djh'. rename H0 into djh'. 
 
 induction st'' as (((h'', R''), fl''), isr'').
 unfold g_timer_pushbx in gS.       (** 2. pS -> gS -> pS'', S.H = M1 *^* M2 **)
 substH gS with (gS M2' M1 djh' (R SP) (R AX) mM2').
 rewrite rfUp in gS.
 destruct gS as [M2'' [djh'' [[pre'' NinISr] [eqax [eqbx [eqcx [eqdx [eqdi [eqsi [eqbp eqsp]]]]]]]]]].

 unfold IHgrt. unfold prepost. UnfoldStateAll. reg simpl.
 split.
  exists M2. exists M1. exists M2''.
  split_l.
   rewrite eqsp. split_l. split_r. apply MemMod.disjUnion_commut; auto.
   split_r.
   apply (extensionality reg nat R R'').
   induction x; auto.
Qed.

Definition timer_pushcx_MPre (sp ax bx : nat) : MPred :=
  fun (M : Mem) => (INV0 *^* stkSpacePrec (sp-4) (IH_Stk-4) 
    *^* sp-4 |->w bx *^* sp-2 |->w ax) M.

Definition timer_pushcx_pre (sp ax bx : nat) (S : State) : Prop :=
  timer_pushcx_MPre sp ax bx (S_Mem S)
  /\ S_RF S BX = bx
  /\ S_RF S SP = sp - 4
  /\ sp >= IH_Stk /\ S_ISR S = true /\ S_Flag S _IF = false.

Definition p_timer_pushcx (S : State) : Prop :=
  exists sp, exists ax, exists bx, imprec (timer_pushcx_pre sp ax bx) S.

Definition timer_pushcx_post (sp ax bx : nat) (S S':State) (M0':Mem) : Prop :=
  timer_post sp (S_Mem_Upd S' M0') 
	/\ S_RF S' AX = ax  	  /\ S_RF S' BX = bx
	/\ S_RF S' CX = S_RF S CX /\ S_RF S' DX = S_RF S DX
	/\ S_RF S' DI = S_RF S DI /\ S_RF S' SI = S_RF S SI
	/\ S_RF S' BP = S_RF S BP /\ S_RF S' SP = sp.

Definition g_timer_pushcx (S S' : State) : Prop := 
  forall M0 M1, MemMod.disjUnion M0 M1 (S_Mem S)
    -> forall sp ax bx, timer_pushcx_pre sp ax bx (S_Mem_Upd S M0)
      -> exists M0', MemMod.disjUnion M0' M1 (S_Mem S') 
	/\ timer_pushcx_post sp ax bx S S' M0'.

Lemma timer_pushbx_has_stack:
  forall h sp ax, timer_pushbx_MPre sp ax h -> sp >= IH_Stk -> exists bx, ((sp-4) |~>w bx) h.
Proof.
unfold timer_pushbx_MPre; intros h sp ax pre spge.
  sep on.
  assert (Hx : h ||- stkSpacePrec (sp-4) (IH_Stk-4) *^* sp-4 |->w_
    *^* (sp-2 |->w ax  *^* INV0)).
  assocr.
  apply stkSpacePrec_splitW1.
  arith_rep (sp - 4 + 2) with (sp-2).
  arith_rep (IH_Stk - 4 + 2) with (IH_Stk-2).
  sep matchH pre with 2.
  sep lift 1 in Hx.
  sep elim_left Hx.
  exists v.
  sep on.
  sep trivial.
Qed.

Lemma timer_pushbx_pre2post:
  forall sp ax bx h h', timer_pushbx_MPre sp ax h -> MemUpd16 h (sp - 4) bx h'
    -> sp >= IH_Stk -> timer_pushcx_MPre sp ax bx h'. 
Proof.
intros sp ax bx h h' pre HM' spge.
substH HM' with (MemUpd16_MemUpdW HM').
destruct HM' as [junk HM'].
unfold timer_pushcx_MPre, timer_pushbx_MPre in * .
  subst h'.
  sep on.
  sep lift 2.
  reg simpl.
  sep upd.

  assert (Hx : h ||- stkSpacePrec (sp-4) (IH_Stk-4) *^* sp-4 |->w_
    *^* (sp-2 |->w ax  *^* INV0)).
  assocr.
  apply stkSpacePrec_splitW1.
  arith_rep (sp - 4 + 2) with (sp-2).
  arith_rep (IH_Stk - 4 + 2) with (IH_Stk-2).
  sep matchH pre with 2.
  sep matchH Hx with 4.
Qed.

Lemma IH_WFcmd_1128_pushbx:
  IH_WFcmd (SCAP_consCT (h_entry + 1 + 1) (Ast p_timer_pushcx g_timer_pushcx ihtag) Psi_timer) 
     (h_entry + 1) (Ast p_timer_pushbx g_timer_pushbx ihtag) (comml (_pushwp (Opr_reg BX))).
Proof.
exists p_timer_pushcx. exists g_timer_pushcx.
split. constructor. auto.
split.
 intros st [sp [ax pre]].
 induction st as (((h, R), fl), isr).  
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqax [eqsp [spge [inIsr IFtrue]]]]].
 UnfoldStateAll.
 destruct (timer_pushbx_has_stack M1 sp ax mM1 spge) as [bx ld2sp].
 exists ((MemMod.put (MemMod.put h (R SP -2) (opr_eval R (Opr_reg BX))) (R SP -2+1)(opr_eval R (Opr_reg BX))), (RF_update R SP (R SP -2)), fl, isr).
 apply nextS_pushwp with (opr_eval R (Opr_reg BX)); auto.
 unfold pushW. split.  split_l. 
 omega.
  apply leadsToW_leadsToBlk with bx; auto. rewrite eqsp. replace (sp-2-2) with (sp-4).
  apply leadsToW_Mono with M1 M2; auto. omega.
 unfold opr_eval, MemUpd16. split_r.
 exists (opr_eval R (Opr_reg BX)); auto.
split. (** pimp **)
 intros st' [st [pS nS]].
 destruct pS as [sp [ax pre]].
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqax [eqsp [spge [inIsr IFtrue]]]]].
 destruct (timer_pushbx_has_stack M1 sp ax mM1 spge) as [bx ld2sp].

 induction st as (((h, R), fl), isr);  UnfoldStateAll.
 induction st' as (((h', R'), fl'), isr').
 destruct (nextS_inversion nS).
 clear nS. 
 unfold opr_eval in H. destruct H as [spgt2' [ldt memUp]]. 
 destruct H0 as [rfUp [eqfl eqisr]].
 exists sp. exists ax. exists (R BX).

 assert (leadsToxW (R SP -2) M1).
 apply leadsToW_leadsToxW with bx. rewrite eqsp. replace (sp-2-2) with (sp-4); auto. omega.
 destruct (leadsToxW_MemUpd16_DMono djh H memUp) as [M1' [djh' upM1]].

 rewrite rfUp. rewrite eqfl. rewrite eqisr.
 exists M1'. exists M2. split_l. split_r.
 repeat split; auto.
 UnfoldStateAll. 
 apply timer_pushbx_pre2post with M1; auto.  rewrite eqsp in upM1. replace (sp-4) with (sp-2-2); auto. omega.
 UnfoldStateAll. reg simpl. rewrite eqsp. omega.
(** gimp **)
 intros st st''  [st' [[pS nS] gS]].
 induction st as (((h, R), fl), isr).
 induction st' as (((h', R'), fl'), isr').
 induction st'' as (((h'', R''), fl''), isr'').

 destruct (nextS_inversion nS). clear nS.  (** nS **)
 unfold opr_eval in H. destruct H as [spgt2' [ldt memUp]]. 
 destruct H0 as [rfUp [eqfl eqisr]].  
 
 unfold g_timer_pushbx.   (** goal: old gS  **)
 intros M1 M2 djh sp ax [mM1 [eqax [eqsp [spge [inIsr IFtrue]]]]]. UnfoldStateAll. 
 destruct (timer_pushbx_has_stack M1 sp ax mM1 spge) as [bx ld2sp].

 assert (leadsToxW (R SP -2) M1).
 apply leadsToW_leadsToxW with bx. rewrite eqsp. replace (sp-2-2) with (sp-4); auto. omega.
 destruct (leadsToxW_MemUpd16_DMono djh H memUp) as [M1' [djh' upM1]].

 assert (timer_pushcx_pre sp ax (R BX) (M1', R', fl', isr')).
  rewrite rfUp. rewrite eqfl. rewrite eqisr.
  repeat split; auto.
  UnfoldStateAll.
  apply timer_pushbx_pre2post with M1; auto.  rewrite eqsp in upM1. replace (sp-4) with (sp-2-2); auto. omega.
  unfold S_RF. reg simpl. rewrite eqsp. omega.
 rename H0 into mM1'.

 unfold g_timer_pushcx in gS.       (** 2. pS -> gS -> pS'', S.H = M1 *^* M2 **)
 substH gS with (gS M1' M2 djh' sp ax (R BX) mM1').
 rewrite rfUp in gS.
 destruct gS as [M1'' [djh'' [pre'' [eqax'' [eqbx [eqcx [eqdx [eqdi [eqsi [eqbp eqsp'']]]]]]]]]].

 UnfoldStateAll. reg simpl.
 exists M1''. split_l. split_l.
 repeat split; auto.
Qed.

Definition timer_pushdx_MPre (sp ax bx cx: nat) : MPred :=
  fun (M : Mem) => (INV0 *^* stkSpacePrec (sp-6) (IH_Stk-6) 
    *^* sp-6 |->w cx *^* sp-4 |->w bx *^* sp-2 |->w ax) M.

Definition timer_pushdx_pre (sp ax bx cx: nat) (S : State) : Prop :=
  timer_pushdx_MPre sp ax bx cx (S_Mem S)
  /\ S_RF S CX = cx
  /\ S_RF S SP = sp - 6
  /\ sp >= IH_Stk /\ S_ISR S = true /\ S_Flag S _IF = false.

Definition p_timer_pushdx (S : State) : Prop :=
  exists sp, exists ax, exists bx, exists cx, imprec (timer_pushdx_pre sp ax bx cx) S.

Definition timer_pushdx_post (sp ax bx cx: nat) (S S':State) (M0':Mem) : Prop :=
  timer_post sp (S_Mem_Upd S' M0') 
	/\ S_RF S' AX = ax  	  /\ S_RF S' BX = bx
	/\ S_RF S' CX = cx        /\ S_RF S' DX = S_RF S DX
	/\ S_RF S' DI = S_RF S DI /\ S_RF S' SI = S_RF S SI
	/\ S_RF S' BP = S_RF S BP /\ S_RF S' SP = sp.

Definition g_timer_pushdx (S S' : State) : Prop := 
  forall M0 M1, MemMod.disjUnion M0 M1 (S_Mem S)
    -> forall sp ax bx cx, timer_pushdx_pre sp ax bx cx (S_Mem_Upd S M0)
      -> exists M0', MemMod.disjUnion M0' M1 (S_Mem S') 
	/\ timer_pushdx_post sp ax bx cx S S' M0'.

Lemma timer_pushcx_has_stack:
  forall h sp ax bx, timer_pushcx_MPre sp ax bx h -> sp >= IH_Stk -> exists cx, (sp-6 |~>w cx) h.
Proof.
unfold timer_pushcx_MPre; intros h sp ax bx pre spge.
  sep on.
  assert (Hx : h ||- stkSpacePrec (sp-6) (IH_Stk-6) *^* sp-6 |->w_
    *^* (sp-4 |->w bx *^* sp-2 |->w ax  *^* INV0)).
  assocr.
  apply stkSpacePrec_splitW1.
  arith_rep (sp - 6 + 2) with (sp-4).
  arith_rep (IH_Stk - 6 + 2) with (IH_Stk-4).
  sep matchH pre with 3.
  sep lift 1 in Hx.
  sep elim_left Hx.
  exists v.
  sep on.
  sep trivial.
Qed.

Lemma timer_pushcx_pre2post:
  forall sp ax bx cx h h', timer_pushcx_MPre sp ax bx h -> MemUpd16 h (sp-6) cx h'
    -> sp >= IH_Stk -> timer_pushdx_MPre sp ax bx cx h'.
Proof.
intros sp ax bx cx h h' pre HM' spge.
substH HM' with (MemUpd16_MemUpdW HM').
destruct HM' as [junk HM'].
unfold timer_pushcx_MPre, timer_pushdx_MPre in * .
  subst h'.
  sep on.
  sep lift 2.
  reg simpl.
  sep upd.

  assert (Hx : h ||- stkSpacePrec (sp-6) (IH_Stk-6) *^* sp-6 |->w_
    *^* (sp-4 |->w bx *^* sp-2 |->w ax  *^* INV0)).
   assocr.
   apply stkSpacePrec_splitW1.
   arith_rep (sp - 6 + 2) with (sp-4).
   arith_rep (IH_Stk - 6 + 2) with (IH_Stk-4).
   sep matchH pre with 3.
  sep matchH Hx with 5.
Qed.

Lemma IH_WFcmd_1129_pushcx:
  IH_WFcmd (SCAP_consCT (h_entry + 1 + 1 + 1) (Ast p_timer_pushdx g_timer_pushdx ihtag) Psi_timer) 
     (h_entry + 1 + 1) (Ast p_timer_pushcx g_timer_pushcx ihtag) (comml (_pushwp (Opr_reg CX))).
Proof.
exists p_timer_pushdx. exists g_timer_pushdx.
split. constructor. auto.
split.
 intros st [sp [ax [bx pre]]].
 induction st as (((h, R), fl), isr).  
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqbx [eqsp [spge [inIsr IFtrue]]]]].
 UnfoldStateAll.
 destruct (timer_pushcx_has_stack M1 sp ax bx mM1 spge) as [cx ld2sp].
 exists ((MemMod.put (MemMod.put h (R SP -2) (opr_eval R (Opr_reg CX))) (R SP -2+1)(opr_eval R (Opr_reg CX))), (RF_update R SP (R SP -2)), fl, isr).
 apply nextS_pushwp with (opr_eval R (Opr_reg CX)); auto.
 unfold pushW. split.  split_l. 
 omega.
  apply leadsToW_leadsToBlk with cx; auto. rewrite eqsp. replace (sp-4-2) with (sp-6).
  apply leadsToW_Mono with M1 M2; auto. omega.
 unfold opr_eval, MemUpd16. split_r.
 exists (opr_eval R (Opr_reg CX)); auto.
split. (** pimp **)
 intros st' [st [pS nS]].
 destruct pS as [sp [ax [bx pre]]].
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqbx [eqsp [spge [inIsr IFtrue]]]]].
 destruct (timer_pushcx_has_stack M1 sp ax bx mM1 spge) as [cx ld2sp].

 induction st as (((h, R), fl), isr);  UnfoldStateAll.
 induction st' as (((h', R'), fl'), isr').
 destruct (nextS_inversion nS).
 clear nS. 
 unfold opr_eval in H. destruct H as [spgt2' [ldt memUp]]. 
 destruct H0 as [rfUp [eqfl eqisr]].
 exists sp. exists ax. exists bx. exists (R CX).

 assert (leadsToxW (R SP -2) M1).
 apply leadsToW_leadsToxW with cx. rewrite eqsp. replace (sp-4-2) with (sp-6); auto. omega.
 destruct (leadsToxW_MemUpd16_DMono djh H memUp) as [M1' [djh' upM1]].

 rewrite rfUp. rewrite eqfl. rewrite eqisr.
 exists M1'. exists M2. split_l. split_r.
 repeat split; auto.
 UnfoldStateAll. 
 apply timer_pushcx_pre2post with M1; auto.  rewrite eqsp in upM1. replace (sp-6) with (sp-4-2); auto. omega.
 UnfoldStateAll. reg simpl. rewrite eqsp. omega.
(** gimp **)
 intros st st''  [st' [[pS nS] gS]].
 induction st as (((h, R), fl), isr).
 induction st' as (((h', R'), fl'), isr').
 induction st'' as (((h'', R''), fl''), isr'').

 destruct (nextS_inversion nS). clear nS.  (** nS **)
 unfold opr_eval in H. destruct H as [spgt2' [ldt memUp]]. 
 destruct H0 as [rfUp [eqfl eqisr]].  
 
 intros M1 M2 djh sp ax bx [mM1 [eqbx [eqsp [spge [inIsr IFtrue]]]]]. UnfoldStateAll. 
 destruct (timer_pushcx_has_stack M1 sp ax bx mM1 spge) as [cx ld2sp].

 assert (leadsToxW (R SP -2) M1).
 apply leadsToW_leadsToxW with cx. rewrite eqsp. replace (sp-4-2) with (sp-6); auto. omega.
 destruct (leadsToxW_MemUpd16_DMono djh H memUp) as [M1' [djh' upM1]].

 assert (timer_pushdx_pre sp ax bx (R CX) (M1', R', fl', isr')).
  rewrite rfUp. rewrite eqfl. rewrite eqisr.
  repeat split; auto.
  UnfoldStateAll.
  apply timer_pushcx_pre2post with M1; auto.  rewrite eqsp in upM1. replace (sp-6) with (sp-4-2); auto. omega.
  unfold S_RF. reg simpl. rewrite eqsp. omega.
 rename H0 into mM1'.

 substH gS with (gS M1' M2 djh' sp ax bx (R CX) mM1'). (** 2. pS -> gS -> pS'', S.H = M1 *^* M2 **)
 rewrite rfUp in gS.
 destruct gS as [M1'' [djh'' [pre'' [eqax [eqbx'' [eqcx [eqdx [eqdi [eqsi [eqbp eqsp'']]]]]]]]]].

 UnfoldStateAll. reg simpl.
 exists M1''. split_l. split_l.
 repeat split; auto.
Qed.

Definition timer_pushdi_MPre (sp ax bx cx dx: nat) : MPred :=
  fun (M : Mem) => (INV0 *^* stkSpacePrec (sp-8) (IH_Stk-8) 
    *^* sp-8 |->w dx *^* sp-6 |->w cx *^* sp-4 |->w bx *^* sp-2 |->w ax) M.

Definition timer_pushdi_pre (sp ax bx cx dx: nat) (S : State) : Prop :=
  timer_pushdi_MPre sp ax bx cx dx (S_Mem S)
  /\ S_RF S DX = dx
  /\ S_RF S SP = sp - 8
  /\ sp >= IH_Stk /\ S_ISR S = true /\ S_Flag S _IF = false.

Definition p_timer_pushdi (S : State) : Prop :=
  exists sp, exists ax, exists bx, exists cx, exists dx, imprec (timer_pushdi_pre sp ax bx cx dx) S.

Definition timer_pushdi_post (sp ax bx cx dx: nat) (S S':State) (M0':Mem) : Prop :=
  timer_post sp (S_Mem_Upd S' M0') 
	/\ S_RF S' AX = ax  	  /\ S_RF S' BX = bx
	/\ S_RF S' CX = cx        /\ S_RF S' DX = dx
	/\ S_RF S' DI = S_RF S DI /\ S_RF S' SI = S_RF S SI
	/\ S_RF S' BP = S_RF S BP /\ S_RF S' SP = sp.

Definition g_timer_pushdi (S S' : State) : Prop := 
  forall M0 M1, MemMod.disjUnion M0 M1 (S_Mem S)
    -> forall sp ax bx cx dx, timer_pushdi_pre sp ax bx cx dx (S_Mem_Upd S M0)
      -> exists M0', MemMod.disjUnion M0' M1 (S_Mem S') 
	/\ timer_pushdi_post sp ax bx cx dx S S' M0'.

Lemma timer_pushdx_has_stack:
  forall h sp ax bx cx, timer_pushdx_MPre sp ax bx cx h -> sp>=IH_Stk -> exists dx, (sp-8 |~>w dx) h.
Proof.
unfold timer_pushdx_MPre; intros h sp ax bx cx pre spge.
  sep on.
  assert (Hx : h ||- stkSpacePrec (sp-8) (IH_Stk-8) *^* sp-8 |->w_
    *^* (sp-6 |->w cx *^* sp-4 |->w bx *^* sp-2 |->w ax  *^* INV0)).
   assocr.
   apply stkSpacePrec_splitW1.
   arith_rep (sp - 8 + 2) with (sp-6).
   arith_rep (IH_Stk - 8 + 2) with (IH_Stk-6).
   sep matchH pre with 3.
  sep lift 1 in Hx.
  sep elim_left Hx.
  exists v.
  sep on.
  sep trivial.
Qed.

Lemma timer_pushdx_pre2post:
  forall sp ax bx cx dx h h', timer_pushdx_MPre sp ax bx cx h -> MemUpd16 h (sp-8) dx h'
    -> sp >= IH_Stk -> timer_pushdi_MPre sp ax bx cx dx h'.
Proof.
intros sp ax bx cx dx h h' pre HM' spge.
substH HM' with (MemUpd16_MemUpdW HM').
destruct HM' as [junk HM'].
unfold timer_pushdi_MPre, timer_pushdx_MPre in * .
  subst h'.
  sep on.
  sep lift 2.
  reg simpl.
  sep upd.
  assert (Hx : h ||- stkSpacePrec (sp-8) (IH_Stk-8) *^* sp-8 |->w_
    *^* (sp-6 |->w cx *^* sp-4 |->w bx *^* sp-2 |->w ax  *^* INV0)).
   assocr.
   apply stkSpacePrec_splitW1.
   arith_rep (sp - 8 + 2) with (sp-6).
   arith_rep (IH_Stk - 8 + 2) with (IH_Stk-6).
   sep matchH pre with 3.
  sep matchH Hx with 6.
Qed.

Lemma IH_WFcmd_1130_pushdx:
  IH_WFcmd (SCAP_consCT (h_entry + 1 + 1 + 1 + 1) (Ast p_timer_pushdi g_timer_pushdi ihtag) Psi_timer) 
     (h_entry + 1 + 1 + 1) (Ast p_timer_pushdx g_timer_pushdx ihtag) (comml (_pushwp (Opr_reg DX))).
Proof.
exists p_timer_pushdi. exists g_timer_pushdi.
split. constructor. auto.
split.
 intros st [sp [ax [bx [cx pre]]]].
 induction st as (((h, R), fl), isr).  
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqcx [eqsp [spge [inIsr IFtrue]]]]].
 UnfoldStateAll.
 destruct (timer_pushdx_has_stack M1 sp ax bx cx mM1 spge) as [dx ld2sp].
 exists ((MemMod.put (MemMod.put h (R SP -2) (opr_eval R (Opr_reg DX))) (R SP -2+1)(opr_eval R (Opr_reg DX))), (RF_update R SP (R SP -2)), fl, isr).
 apply nextS_pushwp with (opr_eval R (Opr_reg DX)); auto.
 unfold pushW. split.  split_l. 
 omega.
  apply leadsToW_leadsToBlk with dx; auto. rewrite eqsp. replace (sp-6-2) with (sp-8).
  apply leadsToW_Mono with M1 M2; auto. omega.
 unfold opr_eval, MemUpd16. split_r.
 exists (opr_eval R (Opr_reg DX)); auto.
split. (** pimp **)
 intros st' [st [pS nS]].
 induction st as (((h, R), fl), isr).
 induction st' as (((h', R'), fl'), isr').
 destruct (nextS_inversion nS). clear nS. 
 unfold opr_eval in H. destruct H as [spgt2' [ldt memUp]]. 
 destruct H0 as [rfUp [eqfl eqisr]].

 destruct pS as [sp [ax [bx [cx pre]]]].
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqcx [eqsp [spge [inIsr IFtrue]]]]].
 destruct (timer_pushdx_has_stack M1 sp ax bx cx mM1 spge) as [dx ld2sp]. UnfoldStateAll.
 assert (leadsToxW (R SP -2) M1).
 apply leadsToW_leadsToxW with dx. rewrite eqsp. replace (sp-6-2) with (sp-8); auto. omega.
 destruct (leadsToxW_MemUpd16_DMono djh H memUp) as [M1' [djh' upM1]].

 exists sp. exists ax. exists bx. exists cx. exists (R DX).
 rewrite rfUp. rewrite eqfl. rewrite eqisr.
 exists M1'. exists M2. split_l. split_r.
 repeat split; auto.
 unfold S_Mem, S_Mem_Upd. 
 apply timer_pushdx_pre2post with M1; auto.  rewrite eqsp in upM1. replace (sp-8) with (sp-6-2); auto. omega.
 unfold S_Mem_Upd, S_RF. reg simpl. rewrite eqsp. omega.
(** gimp **)
 intros st st''  [st' [[pS nS] gS]].
 induction st as (((h, R), fl), isr).
 induction st' as (((h', R'), fl'), isr').
 induction st'' as (((h'', R''), fl''), isr'').

 destruct (nextS_inversion nS). clear nS.  (** nS **)
 unfold opr_eval in H. destruct H as [spgt2' [ldt memUp]]. 
 destruct H0 as [rfUp [eqfl eqisr]].  
 
 intros M1 M2 djh sp ax bx cx [mM1 [eqcx [eqsp [spge [inIsr IFtrue]]]]]. 
 destruct (timer_pushdx_has_stack M1 sp ax bx cx mM1 spge) as [dx ld2sp]. UnfoldStateAll. 
 assert (leadsToxW (R SP -2) M1).
 apply leadsToW_leadsToxW with dx. rewrite eqsp. replace (sp-6-2) with (sp-8); auto. omega.
 destruct (leadsToxW_MemUpd16_DMono djh H memUp) as [M1' [djh' upM1]].

 assert (timer_pushdi_pre sp ax bx cx (R DX) (M1', R', fl', isr')).
  rewrite rfUp. rewrite eqfl. rewrite eqisr.
  repeat split; auto. unfold S_Mem, S_Mem_Upd.
  apply timer_pushdx_pre2post with M1; auto.  rewrite eqsp in upM1. replace (sp-8) with (sp-6-2); auto. omega.
  unfold S_RF. reg simpl. rewrite eqsp. omega.
 rename H0 into mM1'.

 substH gS with (gS M1' M2 djh' sp ax bx cx (R DX) mM1'). (** 2. pS -> gS -> pS'', S.H = M1 *^* M2 **)
 rewrite rfUp in gS.
 destruct gS as [M1'' [djh'' [pre'' [eqax [eqbx [eqcx'' [eqdx [eqdi [eqsi [eqbp eqsp'']]]]]]]]]].

 UnfoldStateAll. reg simpl.
 exists M1''. split_l. split_l.
 repeat split; auto.
Qed.

Definition timer_pushsi_MPre (sp ax bx cx dx di: nat) : MPred :=
  fun (M : Mem) => (INV0 *^* stkSpacePrec (sp-10) (IH_Stk-10) 
    *^* sp-10 |->w di *^* sp-8 |->w dx *^* sp-6 |->w cx 
      *^* sp-4 |->w bx *^* sp-2 |->w ax) M.

Definition timer_pushsi_pre (sp ax bx cx dx di : nat) (S : State) : Prop :=
  timer_pushsi_MPre sp ax bx cx dx di (S_Mem S)
  /\ S_RF S DI = di
  /\ S_RF S SP = sp - 10
  /\ sp >= IH_Stk /\ S_ISR S = true /\ S_Flag S _IF = false.

Definition p_timer_pushsi (S : State) : Prop :=
  exists sp, exists ax, exists bx, exists cx, exists dx, exists di,
    imprec (timer_pushsi_pre sp ax bx cx dx di) S.

Definition timer_pushsi_post (sp ax bx cx dx di: nat) (S S':State) (M0':Mem) : Prop :=
  timer_post sp (S_Mem_Upd S' M0') 
	/\ S_RF S' AX = ax  	  /\ S_RF S' BX = bx
	/\ S_RF S' CX = cx        /\ S_RF S' DX = dx
	/\ S_RF S' DI = di        /\ S_RF S' SI = S_RF S SI
	/\ S_RF S' BP = S_RF S BP /\ S_RF S' SP = sp.

Definition g_timer_pushsi (S S' : State) : Prop := 
  forall M0 M1, MemMod.disjUnion M0 M1 (S_Mem S)
    -> forall sp ax bx cx dx di, timer_pushsi_pre sp ax bx cx dx di (S_Mem_Upd S M0)
      -> exists M0', MemMod.disjUnion M0' M1 (S_Mem S') 
	/\ timer_pushsi_post sp ax bx cx dx di S S' M0'.

Lemma timer_pushdi_has_stack:
  forall h sp ax bx cx dx, timer_pushdi_MPre sp ax bx cx dx h -> sp>=IH_Stk -> exists di, (sp-10 |~>w di) h.
Proof.
unfold timer_pushdi_MPre; intros h sp ax bx cx dx pre spge.
  sep on.
  assert (Hx : h ||- stkSpacePrec (sp-10) (IH_Stk-10) *^* sp-10 |->w_
    *^* (sp-8 |->w dx *^* sp-6 |->w cx *^* sp-4 |->w bx *^* sp-2 |->w ax  *^* INV0)).
   assocr.
   apply stkSpacePrec_splitW1.
   arith_rep (sp - 10 + 2) with (sp-8).
   arith_rep (IH_Stk - 10 + 2) with (IH_Stk-8).
   sep matchH pre with 3.
  sep lift 1 in Hx.
  sep elim_left Hx.
  exists v.
  sep on.
  sep trivial.
Qed.

Lemma timer_pushdi_pre2post:
  forall sp ax bx cx dx di h h', timer_pushdi_MPre sp ax bx cx dx h -> MemUpd16 h (sp-10) di h'
    ->sp >= IH_Stk -> timer_pushsi_MPre sp ax bx cx dx di h'.
Proof.
intros sp ax bx cx dx di h h' pre HM' spge.
substH HM' with (MemUpd16_MemUpdW HM').
destruct HM' as [junk HM'].
unfold timer_pushdi_MPre, timer_pushsi_MPre in * .
  subst h'.
  sep on.
  sep lift 2.
  reg simpl.
  sep upd.
  assert (Hx : h ||- stkSpacePrec (sp-10) (IH_Stk-10) *^* sp-10 |->w_
    *^* (sp-8 |->w dx *^* sp-6 |->w cx *^* sp-4 |->w bx *^* sp-2 |->w ax  *^* INV0)).
   assocr.
   apply stkSpacePrec_splitW1.
   arith_rep (sp - 10 + 2) with (sp-8).
   arith_rep (IH_Stk - 10 + 2) with (IH_Stk-8).
   sep matchH pre with 3.
  sep matchH Hx with 7.
Qed.

Lemma IH_WFcmd_1131_pushdi:
  IH_WFcmd (SCAP_consCT (h_entry + 1 + 1 + 1 + 1 + 1) (Ast p_timer_pushsi g_timer_pushsi ihtag) Psi_timer) 
     (h_entry + 1 + 1 + 1 + 1) (Ast p_timer_pushdi g_timer_pushdi ihtag) (comml (_pushwp (Opr_reg DI))).
Proof.
exists p_timer_pushsi. exists g_timer_pushsi.
split. constructor. auto.
split.
 intros st [sp [ax [bx [cx [dx pre]]]]].
 induction st as (((h, R), fl), isr).  
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqdx [eqsp [spge [inIsr IFtrue]]]]].
 UnfoldStateAll.
 destruct (timer_pushdi_has_stack M1 sp ax bx cx dx mM1 spge) as [di ld2sp].
 exists ((MemMod.put (MemMod.put h (R SP -2) (opr_eval R (Opr_reg DI))) (R SP -2+1)(opr_eval R (Opr_reg DI))), (RF_update R SP (R SP -2)), fl, isr).
 apply nextS_pushwp with (opr_eval R (Opr_reg DI)); auto.
 unfold pushW. split.  split_l. 
 omega.
  apply leadsToW_leadsToBlk with di; auto. rewrite eqsp. replace (sp-8-2) with (sp-10).
  apply leadsToW_Mono with M1 M2; auto. omega.
 unfold opr_eval, MemUpd16. split_r.
 exists (opr_eval R (Opr_reg DI)); auto.
split. (** pimp **)
 intros st' [st [pS nS]].
 induction st as (((h, R), fl), isr).
 induction st' as (((h', R'), fl'), isr').
 destruct (nextS_inversion nS). clear nS. 
 unfold opr_eval in H. destruct H as [spgt2' [ldt memUp]]. 
 destruct H0 as [rfUp [eqfl eqisr]].

 destruct pS as [sp [ax [bx [cx [dx pre]]]]].
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqdx [eqsp [spge [inIsr IFtrue]]]]].
 destruct (timer_pushdi_has_stack M1 sp ax bx cx dx mM1 spge) as [di ld2sp]. UnfoldStateAll.
 assert (leadsToxW (R SP -2) M1).
 apply leadsToW_leadsToxW with di. rewrite eqsp. replace (sp-8-2) with (sp-10); auto. omega.
 destruct (leadsToxW_MemUpd16_DMono djh H memUp) as [M1' [djh' upM1]].

 exists sp. exists ax. exists bx. exists cx. exists dx. exists (R DI).
 rewrite rfUp. rewrite eqfl. rewrite eqisr.
 exists M1'. exists M2. split_l. split_r.
 repeat split; auto.
 unfold S_Mem, S_Mem_Upd. 
 apply timer_pushdi_pre2post with M1; auto.  rewrite eqsp in upM1. replace (sp-10) with (sp-8-2); auto. omega.
 unfold S_Mem_Upd, S_RF. reg simpl. rewrite eqsp. omega.
(** gimp **)
 intros st st''  [st' [[pS nS] gS]].
 induction st as (((h, R), fl), isr).
 induction st' as (((h', R'), fl'), isr').
 induction st'' as (((h'', R''), fl''), isr'').

 destruct (nextS_inversion nS). clear nS.  (** nS **)
 unfold opr_eval in H. destruct H as [spgt2' [ldt memUp]]. 
 destruct H0 as [rfUp [eqfl eqisr]].  
 
 intros M1 M2 djh sp ax bx cx dx [mM1 [eqdx [eqsp [spge [inIsr IFtrue]]]]]. 
 destruct (timer_pushdi_has_stack M1 sp ax bx cx dx mM1 spge) as [di ld2sp]. UnfoldStateAll. 
 assert (leadsToxW (R SP -2) M1).
 apply leadsToW_leadsToxW with di. rewrite eqsp. replace (sp-8-2) with (sp-10); auto. omega.
 destruct (leadsToxW_MemUpd16_DMono djh H memUp) as [M1' [djh' upM1]].

 assert (timer_pushsi_pre sp ax bx cx dx (R DI) (M1', R', fl', isr')).
  rewrite rfUp. rewrite eqfl. rewrite eqisr.
  repeat split; auto. unfold S_Mem, S_Mem_Upd.
  apply timer_pushdi_pre2post with M1; auto.  rewrite eqsp in upM1. replace (sp-10) with (sp-8-2); auto. omega.
  unfold S_RF. reg simpl. rewrite eqsp. omega.
 rename H0 into mM1'.

 substH gS with (gS M1' M2 djh' sp ax bx cx dx (R DI) mM1'). (** 2. pS -> gS -> pS'', S.H = M1 *^* M2 **)
 rewrite rfUp in gS.
 destruct gS as [M1'' [djh'' [pre'' [eqax [eqbx [eqcx [eqdx'' [eqdi [eqsi [eqbp eqsp'']]]]]]]]]].

 UnfoldStateAll. reg simpl.
 exists M1''. split_l. split_l.
 repeat split; auto.
Qed.

Definition timer_pushbp_MPre (sp ax bx cx dx di si: nat) : MPred :=
  fun (M : Mem) => (INV0 *^* stkSpacePrec (sp-12) (IH_Stk-12) 
    *^* sp-12 |->w si *^* sp-10 |->w di *^* sp-8 |->w dx 
      *^* sp-6  |->w cx *^* sp-4  |->w bx *^* sp-2 |->w ax) M.


Definition timer_pushbp_pre (sp ax bx cx dx di si : nat) (S : State) : Prop :=
  timer_pushbp_MPre sp ax bx cx dx di si (S_Mem S)
  /\ S_RF S SI = si
  /\ S_RF S SP = sp - 12
  /\ sp >= IH_Stk /\ S_ISR S = true /\ S_Flag S _IF = false.

Definition p_timer_pushbp (S : State) : Prop :=
  exists sp, exists ax, exists bx, exists cx, exists dx, exists di, exists si,
    imprec (timer_pushbp_pre sp ax bx cx dx di si) S.

Definition timer_pushbp_post (sp ax bx cx dx di si: nat) (S S':State) (M0':Mem) : Prop :=
  timer_post sp (S_Mem_Upd S' M0') 
	/\ S_RF S' AX = ax  	  /\ S_RF S' BX = bx
	/\ S_RF S' CX = cx        /\ S_RF S' DX = dx
	/\ S_RF S' DI = di        /\ S_RF S' SI = si
	/\ S_RF S' BP = S_RF S BP /\ S_RF S' SP = sp.

Definition g_timer_pushbp (S S' : State) : Prop := 
  forall M0 M1, MemMod.disjUnion M0 M1 (S_Mem S)
    -> forall sp ax bx cx dx di si, timer_pushbp_pre sp ax bx cx dx di si (S_Mem_Upd S M0)
      -> exists M0', MemMod.disjUnion M0' M1 (S_Mem S') 
	/\ timer_pushbp_post sp ax bx cx dx di si S S' M0'.

Lemma timer_pushsi_has_stack:
  forall h sp ax bx cx dx di, timer_pushsi_MPre sp ax bx cx dx di h -> sp>=IH_Stk -> exists si, (sp-12 |~>w si) h.
Proof.
unfold timer_pushsi_MPre; intros h sp ax bx cx dx di pre spge.
  sep on.
  assert (Hx : h ||- stkSpacePrec (sp-12) (IH_Stk-12) *^* sp-12 |->w_
    *^* (sp-10 |->w di *^* sp-8 |->w dx *^* sp-6 |->w cx *^* sp-4 |->w bx *^* sp-2 |->w ax  *^* INV0)).
   assocr.
   apply stkSpacePrec_splitW1.
   arith_rep (sp - 12 + 2) with (sp-10).
   arith_rep (IH_Stk - 12 + 2) with (IH_Stk-10).
   sep matchH pre with 3.
  sep lift 1 in Hx.
  sep elim_left Hx.
  exists v.
  sep on.
  sep trivial.
Qed.

Lemma timer_pushsi_pre2post:
  forall sp ax bx cx dx di si h h', timer_pushsi_MPre sp ax bx cx dx di h
    -> MemUpd16 h (sp-12) si h' ->sp >= IH_Stk -> timer_pushbp_MPre sp ax bx cx dx di si h'.
Proof.
intros sp ax bx cx dx di si h h' pre HM' spge.
substH HM' with (MemUpd16_MemUpdW HM').
destruct HM' as [junk HM'].
unfold timer_pushbp_MPre, timer_pushsi_MPre in * .
  subst h'.
  sep on.
  sep lift 2.
  reg simpl.
  sep upd.
  assert (Hx : h ||- stkSpacePrec (sp-12) (IH_Stk-12) *^* sp-12 |->w_
    *^* (sp-10 |->w di *^* sp-8 |->w dx *^* sp-6 |->w cx *^* sp-4 |->w bx *^* sp-2 |->w ax  *^* INV0)).
   assocr.
   apply stkSpacePrec_splitW1.
   arith_rep (sp - 12 + 2) with (sp-10).
   arith_rep (IH_Stk - 12 + 2) with (IH_Stk-10).
   sep matchH pre with 3.
  sep matchH Hx with 8.
Qed.

Lemma IH_WFcmd_1132_pushsi:
  IH_WFcmd (SCAP_consCT (h_entry + 1 + 1 + 1 + 1 + 1 + 1) (Ast p_timer_pushbp g_timer_pushbp ihtag) Psi_timer) 
     (h_entry + 1 + 1 + 1 + 1 + 1) (Ast p_timer_pushsi g_timer_pushsi ihtag) (comml (_pushwp (Opr_reg SI))).
Proof.
exists p_timer_pushbp. exists g_timer_pushbp.
split. constructor. auto.
split.
 intros st [sp [ax [bx [cx [dx [di pre]]]]]].
 induction st as (((h, R), fl), isr).  
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqdi [eqsp [spge [inIsr IFtrue]]]]].
 UnfoldStateAll.
 destruct (timer_pushsi_has_stack M1 sp ax bx cx dx di mM1 spge) as [si ld2sp].
 exists ((MemMod.put (MemMod.put h (R SP -2) (opr_eval R (Opr_reg SI))) (R SP -2+1)(opr_eval R (Opr_reg SI))), (RF_update R SP (R SP -2)), fl, isr).
 apply nextS_pushwp with (opr_eval R (Opr_reg SI)); auto.
 unfold pushW. split.  split_l. 
 omega.
  apply leadsToW_leadsToBlk with si; auto. rewrite eqsp. replace (sp-10-2) with (sp-12).
  apply leadsToW_Mono with M1 M2; auto. omega.
 unfold opr_eval, MemUpd16. split_r.
 exists (opr_eval R (Opr_reg SI)); auto.
split. (** pimp **)
 intros st' [st [pS nS]].
 induction st as (((h, R), fl), isr).
 induction st' as (((h', R'), fl'), isr').
 destruct (nextS_inversion nS). clear nS. 
 unfold opr_eval in H. destruct H as [spgt2' [ldt memUp]]. 
 destruct H0 as [rfUp [eqfl eqisr]].

 destruct pS as [sp [ax [bx [cx [dx [di pre]]]]]].
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqdi [eqsp [spge [inIsr IFtrue]]]]].
 destruct (timer_pushsi_has_stack M1 sp ax bx cx dx di mM1 spge) as [si ld2sp]. UnfoldStateAll.
 assert (leadsToxW (R SP -2) M1).
 apply leadsToW_leadsToxW with si. rewrite eqsp. replace (sp-10-2) with (sp-12); auto. omega.
 destruct (leadsToxW_MemUpd16_DMono djh H memUp) as [M1' [djh' upM1]].

 exists sp. exists ax. exists bx. exists cx. exists dx. exists di. exists (R SI).
 rewrite rfUp. rewrite eqfl. rewrite eqisr.
 exists M1'. exists M2. split_l. split_r.
 repeat split; auto.
 unfold S_Mem, S_Mem_Upd. 
 apply timer_pushsi_pre2post with M1; auto.  rewrite eqsp in upM1. replace (sp-12) with (sp-10-2); auto. omega.
 unfold S_Mem_Upd, S_RF. reg simpl. rewrite eqsp. omega.
(** gimp **)
 intros st st''  [st' [[pS nS] gS]].
 induction st as (((h, R), fl), isr).
 induction st' as (((h', R'), fl'), isr').
 induction st'' as (((h'', R''), fl''), isr'').

 destruct (nextS_inversion nS). clear nS.  (** nS **)
 unfold opr_eval in H. destruct H as [spgt2' [ldt memUp]]. 
 destruct H0 as [rfUp [eqfl eqisr]].  
 
 intros M1 M2 djh sp ax bx cx dx di [mM1 [eqdi [eqsp [spge [inIsr IFtrue]]]]]. 
 destruct (timer_pushsi_has_stack M1 sp ax bx cx dx di mM1 spge) as [si ld2sp]. UnfoldStateAll. 
 assert (leadsToxW (R SP -2) M1).
 apply leadsToW_leadsToxW with si. rewrite eqsp. replace (sp-10-2) with (sp-12); auto. omega.
 destruct (leadsToxW_MemUpd16_DMono djh H memUp) as [M1' [djh' upM1]].

 assert (timer_pushbp_pre sp ax bx cx dx di (R SI) (M1', R', fl', isr')).
  rewrite rfUp. rewrite eqfl. rewrite eqisr.
  repeat split; auto. unfold S_Mem, S_Mem_Upd.
  apply timer_pushsi_pre2post with M1; auto.  rewrite eqsp in upM1. replace (sp-12) with (sp-10-2); auto. omega.
  unfold S_RF. reg simpl. rewrite eqsp. omega.
 rename H0 into mM1'.

 substH gS with (gS M1' M2 djh' sp ax bx cx dx di (R SI) mM1'). (** 2. pS -> gS -> pS'', S.H = M1 *^* M2 **)
 rewrite rfUp in gS.
 destruct gS as [M1'' [djh'' [pre'' [eqax [eqbx [eqcx [eqdx [eqdi'' [eqsi [eqbp eqsp'']]]]]]]]]].

 UnfoldStateAll. reg simpl.
 exists M1''. split_l. split_l.
 repeat split; auto.
Qed.

Definition timer_movax_MPre (sp ax bx cx dx di si bp: nat) : MPred :=
  fun (M : Mem) => (INV0 *^* stkSpacePrec (sp-14) (IH_Stk-14) 
    *^* sp-14 |->w bp *^* sp-12 |->w si *^* sp-10 |->w di *^* sp-8 |->w dx 
      *^* sp-6  |->w cx *^* sp-4  |->w bx *^* sp-2  |->w ax) M.

Definition timer_movax_pre (sp ax bx cx dx di si bp: nat) (S : State) : Prop :=
  timer_movax_MPre sp ax bx cx dx di si bp (S_Mem S)
  /\ S_RF S BP = bp
  /\ S_RF S SP = sp - 14
  /\ sp >= IH_Stk /\ S_ISR S = true /\ S_Flag S _IF = false.

Definition p_timer_movax (S : State) : Prop :=
  exists sp, exists ax, exists bx, exists cx, exists dx, exists di, exists si, exists bp,
    imprec (timer_movax_pre sp ax bx cx dx di si bp) S.

Definition timer_movax_post (sp ax bx cx dx di si bp: nat) (S':State) (M0':Mem) : Prop :=
  timer_post sp (S_Mem_Upd S' M0') 
	/\ S_RF S' AX = ax  	  /\ S_RF S' BX = bx
	/\ S_RF S' CX = cx        /\ S_RF S' DX = dx
	/\ S_RF S' DI = di        /\ S_RF S' SI = si
	/\ S_RF S' BP = bp        /\ S_RF S' SP = sp.

Definition g_timer_movax (S S' : State) : Prop := 
  forall M0 M1, MemMod.disjUnion M0 M1 (S_Mem S)
    -> forall sp ax bx cx dx di si bp, timer_movax_pre sp ax bx cx dx di si bp (S_Mem_Upd S M0)
      -> exists M0', MemMod.disjUnion M0' M1 (S_Mem S') 
	/\ timer_movax_post sp ax bx cx dx di si bp S' M0'.

Lemma timer_pushbp_has_stack:
  forall h sp ax bx cx dx di si, timer_pushbp_MPre sp ax bx cx dx di si h -> sp>=IH_Stk -> exists bp, (sp-14 |~>w bp) h.
Proof.
unfold timer_pushbp_MPre; intros h sp ax bx cx dx di si pre spge.
  sep on.
  assert (Hx : h ||- stkSpacePrec (sp-14) (IH_Stk-14) *^* sp-14 |->w_
    *^* (sp-12 |->w si *^* sp-10 |->w di *^* sp-8 |->w dx *^* sp-6 |->w cx *^* sp-4 |->w bx *^* sp-2 |->w ax  *^* INV0)).
   assocr.
   apply stkSpacePrec_splitW1.
   arith_rep (sp - 14 + 2) with (sp-12).
   arith_rep (IH_Stk - 14 + 2) with (IH_Stk-12).
   sep matchH pre with 4.
  sep lift 1 in Hx.
  sep elim_left Hx.
  exists v.
  sep on.
  sep trivial.
Qed.

Lemma timer_pushbp_pre2post:
  forall sp ax bx cx dx di si bp h h', timer_pushbp_MPre sp ax bx cx dx di si h
    -> MemUpd16 h (sp-14) bp h' -> sp>=IH_Stk -> timer_movax_MPre sp ax bx cx dx di si bp h'.
Proof.
intros sp ax bx cx dx di si bp h h' pre HM' spge.
substH HM' with (MemUpd16_MemUpdW HM').
destruct HM' as [junk HM'].
unfold timer_pushbp_MPre, timer_movax_MPre in * .
  subst h'.
  sep on.
  sep lift 2.
  reg simpl.
  sep upd.
  assert (Hx : h ||- stkSpacePrec (sp-14) (IH_Stk-14) *^* sp-14 |->w_
    *^* (sp-12 |->w si *^* sp-10 |->w di *^* sp-8 |->w dx *^* sp-6 |->w cx *^* sp-4 |->w bx *^* sp-2 |->w ax  *^* INV0)).
   assocr.
   apply stkSpacePrec_splitW1.
   arith_rep (sp - 14 + 2) with (sp-12).
   arith_rep (IH_Stk - 14 + 2) with (IH_Stk-12).
   sep matchH pre with 4.
  sep matchH Hx with 9.
Qed.

Lemma IH_WFcmd_1133_pushbp:
  IH_WFcmd (SCAP_consCT (h_entry + 1 + 1 + 1 + 1 + 1 + 1 + 1) (Ast p_timer_movax g_timer_movax ihtag) Psi_timer) 
     (h_entry + 1 + 1 + 1 + 1 + 1 + 1) (Ast p_timer_pushbp g_timer_pushbp ihtag) (comml (_pushwp (Opr_reg BP))).
Proof.
exists p_timer_movax. exists g_timer_movax.
split. constructor. auto.
split.
 intros st [sp [ax [bx [cx [dx [di [si pre]]]]]]].
 induction st as (((h, R), fl), isr).  
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqbp [eqsp [spge [inIsr IFtrue]]]]].
 UnfoldStateAll.
 destruct (timer_pushbp_has_stack M1 sp ax bx cx dx di si mM1 spge) as [bp ld2sp].
 exists ((MemMod.put (MemMod.put h (R SP -2) (opr_eval R (Opr_reg BP))) (R SP -2+1)(opr_eval R (Opr_reg BP))), (RF_update R SP (R SP -2)), fl, isr).
 apply nextS_pushwp with (opr_eval R (Opr_reg BP)); auto.
 unfold pushW. split.  split_l. 
 omega.
  apply leadsToW_leadsToBlk with bp; auto. rewrite eqsp. replace (sp-12-2) with (sp-14).
  apply leadsToW_Mono with M1 M2; auto. omega.
 unfold opr_eval, MemUpd16. split_r.
 exists (opr_eval R (Opr_reg BP)); auto.
split. (** pimp **)
 intros st' [st [pS nS]].
 induction st as (((h, R), fl), isr).
 induction st' as (((h', R'), fl'), isr').
 destruct (nextS_inversion nS). clear nS. 
 unfold opr_eval in H. destruct H as [spgt2' [ldt memUp]]. 
 destruct H0 as [rfUp [eqfl eqisr]].

 destruct pS as [sp [ax [bx [cx [dx [di [si pre]]]]]]].
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqsi [eqsp [spge [inIsr IFtrue]]]]].
 destruct (timer_pushbp_has_stack M1 sp ax bx cx dx di si mM1 spge) as [bp ld2sp]. UnfoldStateAll.
 assert (leadsToxW (R SP -2) M1).
 apply leadsToW_leadsToxW with bp. rewrite eqsp. replace (sp-12-2) with (sp-14); auto. omega.
 destruct (leadsToxW_MemUpd16_DMono djh H memUp) as [M1' [djh' upM1]].

 exists sp. exists ax. exists bx. exists cx. exists dx. exists di. exists si. exists (R BP).
 rewrite rfUp. rewrite eqfl. rewrite eqisr.
 exists M1'. exists M2. split_l. split_r.
 repeat split; auto.
 unfold S_Mem, S_Mem_Upd. 
 apply timer_pushbp_pre2post with M1; auto.  rewrite eqsp in upM1. replace (sp-14) with (sp-12-2); auto. omega.
 unfold S_Mem_Upd, S_RF. reg simpl. rewrite eqsp. omega.
(** gimp **)
 intros st st''  [st' [[pS nS] gS]].
 induction st as (((h, R), fl), isr).
 induction st' as (((h', R'), fl'), isr').
 induction st'' as (((h'', R''), fl''), isr'').

 destruct (nextS_inversion nS). clear nS.  (** nS **)
 unfold opr_eval in H. destruct H as [spgt2' [ldt memUp]]. 
 destruct H0 as [rfUp [eqfl eqisr]].  
 
 intros M1 M2 djh sp ax bx cx dx di si [mM1 [eqsi [eqsp [spge [inIsr IFtrue]]]]]. 
 destruct (timer_pushbp_has_stack M1 sp ax bx cx dx di si mM1 spge) as [bp ld2sp]. UnfoldStateAll. 
 assert (leadsToxW (R SP -2) M1).
 apply leadsToW_leadsToxW with bp. rewrite eqsp. replace (sp-12-2) with (sp-14); auto. omega.
 destruct (leadsToxW_MemUpd16_DMono djh H memUp) as [M1' [djh' upM1]].

 assert (timer_movax_pre sp ax bx cx dx di si (R BP) (M1', R', fl', isr')).
  rewrite rfUp. rewrite eqfl. rewrite eqisr.
  repeat split; auto. unfold S_Mem, S_Mem_Upd.
  apply timer_pushbp_pre2post with M1; auto.  rewrite eqsp in upM1. replace (sp-14) with (sp-12-2); auto. omega.
  unfold S_RF. reg simpl. rewrite eqsp. omega.
 rename H0 into mM1'.

 substH gS with (gS M1' M2 djh' sp ax bx cx dx di si (R BP) mM1'). (** 2. pS -> gS -> pS'', S.H = M1 *^* M2 **)
 destruct gS as [M1'' [djh'' [pre'' [eqax [eqbx [eqcx [eqdx [eqdi [eqsi'' [eqbp eqsp'']]]]]]]]]].

 exists M1''. split_l. split_l.
 repeat split; auto.
Qed.

Lemma IH_WFcmd_1134_movax:
  IH_WFcmd (SCAP_consCT (h_entry + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1) (Ast (conjP p_timer_movax (fun st : State => S_RF st AX = 32))
    g_timer_movax ihtag) Psi_timer) (h_entry + 1 + 1 + 1 + 1 + 1 + 1 + 1)
     (Ast p_timer_movax g_timer_movax ihtag) (comml (_movwp (Opr_imm 32) AX)).
Proof.
exists (conjP p_timer_movax (fun st : State => S_RF st AX = 32)).  exists g_timer_movax.
split. constructor. auto.
split.
 intros st pS.
 induction st as (((h, R), fl), isr).
 exists (h, (RF_update R AX (opr_eval R (Opr_imm 32))), fl, isr).
 apply nextS_movwp with (opr_eval R (Opr_imm 32)); auto.
split.
 intros st' [st [pS nS]].
 induction st as (((h, R), fl), isr). 
 induction st' as (((h', R'), fl'), isr').
 destruct (nextS_inversion nS). clear nS.
 destruct H0 as [rfUp [eqfl eqisr]].
 rewrite H. rewrite rfUp. rewrite eqfl. rewrite eqisr.

 destruct pS as [sp [ax [bx [cx [dx [di [si [bp pre]]]]]]]].
 split_r.
 exists sp. exists ax. exists bx. exists cx. exists dx. exists di. exists si. exists bp.
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqbp [eqsp [spge [inIsr IFtrue]]]]].
 exists M1. exists M2. split_l. split_r.
 split; auto.
(** gimp **)
 intros st st''  [st' [[pS nS] gS]].
 induction st as (((h, R), fl), isr). 
 induction st' as (((h', R'), fl'), isr').
 destruct (nextS_inversion nS). clear nS.
 destruct H0 as [rfUp [eqfl eqisr]].
 rewrite H in gS. rewrite rfUp in gS. rewrite eqfl in gS. rewrite eqisr in gS.
 unfold g_timer_movax. intros. 
 apply gS with M0; auto.
Qed.

Lemma IH_WFcmd_1135_movdx:
  IH_WFcmd (SCAP_consCT (h_entry + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1) (Ast (conjP p_timer_movax
    (fun st : State => S_RF st AX = 32 /\ S_RF st DX = 32)) g_timer_movax ihtag) Psi_timer)
     (h_entry + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1) (Ast (conjP p_timer_movax (fun st : State => S_RF st AX = 32))
        g_timer_movax ihtag) (comml (_movwp (Opr_imm 32) DX)).
Proof.
exists (conjP p_timer_movax (fun st : State => S_RF st AX = 32 /\ S_RF st DX =32)).  exists g_timer_movax.
split. constructor. auto.
split.
 intros st pS.
 induction st as (((h, R), fl), isr).
 exists (h, (RF_update R DX (opr_eval R (Opr_imm 32))), fl, isr).
 apply nextS_movwp with (opr_eval R (Opr_imm 32)); auto.
split.
 intros st' [st [pS nS]].
 induction st as (((h, R), fl), isr). 
 induction st' as (((h', R'), fl'), isr').
 destruct (nextS_inversion nS). clear nS.
 destruct H0 as [rfUp [eqfl eqisr]].
 rewrite H. rewrite rfUp. rewrite eqfl. rewrite eqisr.

 destruct pS as [[sp [ax [bx [cx [dx [di [si [bp pre]]]]]]]] eqax].
 split.
 exists sp. exists ax. exists bx. exists cx. exists dx. exists di. exists si. exists bp.
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqbp [eqsp [spge [inIsr IFtrue]]]]].
 exists M1. exists M2. split_l. split_r.
 split; auto. split; auto.
(** gimp **)
 intros st st''  [st' [[pS nS] gS]].
 induction st as (((h, R), fl), isr). 
 induction st' as (((h', R'), fl'), isr').
 destruct (nextS_inversion nS). clear nS.
 destruct H0 as [rfUp [eqfl eqisr]].
 rewrite H in gS. rewrite rfUp in gS. rewrite eqfl in gS. rewrite eqisr in gS.
 unfold g_timer_movax. intros.
 apply gS with M0; auto.
Qed.

Definition timer_scheduler_pre  (sp ax bx cx dx di si bp: nat) (S : State) : Prop :=
  timer_movax_MPre sp ax bx cx dx di si bp (S_Mem S)
  /\ S_RF S SP = sp - 14
  /\ sp >= IH_Stk /\ S_ISR S = false /\ S_Flag S _IF = false.

Definition p_timer_scheduler (S : State) : Prop :=
  exists sp, exists ax, exists bx, exists cx, exists dx, exists di, exists si, exists bp,
    imprec (timer_scheduler_pre sp ax bx cx dx di si bp) S.

Definition g_timer_scheduler (S S' : State) : Prop :=
  forall M0 M1, MemMod.disjUnion M0 M1 (S_Mem S)
    -> forall sp ax bx cx dx di si bp, timer_scheduler_pre sp ax bx cx dx di si bp (S_Mem_Upd S M0)
      -> exists M0', MemMod.disjUnion M0' M1 (S_Mem S')
	/\ timer_movax_post sp ax bx cx dx di si bp S' M0'.

Lemma IH_WFcmd_1136_eoi:
  IH_WFcmd (SCAP_consCT (h_entry + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1) (Ast p_timer_scheduler g_timer_scheduler ihtag) Psi_timer)
    (h_entry + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1) (Ast (conjP p_timer_movax (fun st : State => S_RF st AX = 32 /\ S_RF st DX = 32))
      g_timer_movax ihtag) (comml _outb_eoi).
Proof.
exists p_timer_scheduler. exists g_timer_scheduler.
split. constructor. auto.
split.
 intros st [[sp [ax [bx [cx [dx [di [si [bp pre]]]]]]]] [eqax eqdx]].
 induction st as (((h, R), fl), isr).
 exists (h,R,fl,false).
 apply nextS_EOI; auto.
split. (** pimp **)
 intros st' [st [pS nS]].
 induction st as (((h, R), fl), isr).
 induction st' as (((h', R'), fl'), isr').
 destruct (nextS_inversion nS). clear nS. 
 destruct H0 as [[eqR [eqax0 eqdx0]] [eqfl eqisr]].

 destruct pS as [[sp [ax [bx [cx [dx [di [si [bp pre]]]]]]]] [eqax eqdx]].
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqbp [eqsp [spge [inIsr IFtrue]]]]]. UnfoldStateAll.

 exists sp. exists ax. exists bx. exists cx. exists dx. exists di. exists si. exists bp.
 rewrite H. rewrite eqR. rewrite eqfl. rewrite eqisr.
 exists M1. exists M2. split_l. split_r.
 repeat split; auto.
(** gimp **)
 intros st st''  [st' [[pS nS] gS]]. clear pS.
 induction st as (((h, R), fl), isr).
 induction st' as (((h', R'), fl'), isr').
 induction st'' as (((h'', R''), fl''), isr'').

 destruct (nextS_inversion nS). clear nS.  (** nS **)
 destruct H0 as [[eqR [eqax0 eqdx0]] [eqfl eqisr]].
 rewrite H in gS. rewrite eqR in gS. rewrite eqfl in gS. rewrite eqisr in gS.

 intros M1 M2 djh sp ax bx cx dx di si bp pre.
 assert (pre': timer_scheduler_pre sp ax bx cx dx di si bp (M1, R, fl, false)).
  destruct pre as [mM1 [eqbp [eqsp [spge [inIsr IFtrue]]]]].
  split; auto.

 unfold g_timer_scheduler in gS. UnfoldStateAll. 
 substH gS with (gS M1 M2 djh sp ax bx cx dx di si bp pre'). (** 2. pS -> gS -> pS'', S.H = M1 *^* M2 **)
 destruct gS as [M1'' [djh'' [pre'' [eqax [eqbx [eqcx [eqdx [eqdi [eqsi [eqbp'' eqsp'']]]]]]]]]].

 exists M1''. split_l. split_l.
 repeat split; auto.
Qed.

Definition p_timer_popbp (S : State) : Prop := p_timer_scheduler S.
Definition g_timer_popbp (S S' : State) : Prop := g_timer_scheduler S S'.

Lemma  timer_scheduler_pre2post:
  forall pc' sp ax bx cx dx di si bp S S' S'' S''' M1 M2,
    MemMod.disjUnion M1 M2 (S_Mem S) /\ timer_scheduler_pre sp ax bx cx dx di si bp (S_Mem_Upd S M1) /\ mTrue M2
    -> nextS (h_entry + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1) (comml (_call _SWITCH)) S S'
    -> ih_gy S' S''
    -> nextS pc' (comms _retw) S'' S'''
    -> exists M1', MemMod.disjUnion M1' M2 (S_Mem S''') /\ timer_scheduler_pre sp ax bx cx dx di si bp (S_Mem_Upd S''' M1').
Proof.
intros pc' sp ax bx cx dx di si bp S S' S'' S''' M1 M2 [djh [pre mM2]] nSC ihg nSR.
unfold timer_scheduler_pre in * .
destruct S as [[[M R] fl] isr].
destruct S' as [[[M' R'] fl'] isr'].
destruct S'' as [[[M'' R''] fl''] isr''].
destruct S''' as [[[M''' R'''] fl'''] isr'''].
simplS in * .
destruct pre as [mM1 [eqsp [spge [inIsr IFtrue]]]].

destruct (nextS_inversion nSC) as [HM' [HR' [Hfl' Hisr']]]. clear nSC.
assert (ihga: ih_gy_alt (M', R', fl', isr') (M'', R'', fl'', isr'')).
 apply ih_gy_derive; auto. clear ihg. rename ihga into ihg.
destruct ihg as [sp' [HM'' [Hif'' [Hisr'' HSP]]]]. 
destruct (nextS_inversion nSR) as [HM''' [HR''' [Hfl''' Hisr''']]]. clear nSR.
destruct HR''' as [pc''' [Hlds''' HR''']].
destruct HM' as [Hspg' [Hlds Hupd]].
simplS in * .

subst M''' R''' fl''' isr'''.
subst isr'' R' fl' isr' isr.
reg simpl.
destruct HSP as [Hsp' HSP].
subst sp'.
rewrite <- HSP.
rewrite eqsp.
rewrite eqsp in HM''.

unfold timer_movax_MPre in * .
sep on.
 sep lift 1 in mM1.
 sep stksplit in mM1.
 arith_rep (14+8+2-14-2) with 8 in mM1.
sep off.

rewrite eqsp in Hupd.
assert (leadsToxW (sp-14-2) M1).
sep on.
 sep lift 1 in mM1.
 sep trivial.
destruct (leadsToxW_MemUpd16_DMono djh H Hupd) as [M1' [djh' upM1]].

assert (mM1a: ((INV0 *^* stkSpacePrec (sp - 14 - 2) 8)
  *^* (sp - 14 - 2 |->w_ *^* sp - 14 |->w bp *^* sp - 12 |->w si *^* sp - 10 |->w di 
    *^* sp - 8 |->w dx *^* sp - 6 |->w cx *^* sp - 4 |->w bx *^* sp - 2 |->w ax)) M1).
sep on.
 simpl in mM1. sep matchH mM1 with 13. clear mM1.
destruct mM1a as [M3 [M4 [djM1 [mM3 mM4]]]].

assert (leadsToxW (sp-14-2) M4).
sep on.
 sep lift 0 in mM4. sep trivial.
destruct (leadsToxW_MemUpd16_DMonoR djM1 H0 upM1) as [M4' [djM4' upM4]].
assert (djh'a: MemMod.disjUnion M3 (MemMod.merge M4' M2) M').
  destruct djM4' as [dij eqM4].
  apply MemMod.disjUnion_assoc_left_right; auto.
  rewrite <- eqM4; auto.

substH HM'' with (HM'' M3 (MemMod.merge M4' M2) djh'a mM3).
destruct HM'' as [M3'' [djM'' mM3'']].

exists (MemMod.merge M3'' M4').
split.
 apply MemMod.disjUnion_assoc_right_left; auto. 
 apply MemMod.disj_merge_disj_right with M3; auto. 
 destruct djM4' as [djM4' mergeM4']. 
 rewrite <- mergeM4'. 
 inversion djh'; auto.
split.
 assert ((INV0 *^* stkSpacePrec (sp - 14 - 2) 8
   *^* ((sp - 14 - 2 |->w_ *^* sp - 14 |->w bp *^*
         sp - 12 |->w si *^* sp - 10 |->w di *^* sp - 8 |->w dx *^*
         sp - 6 |->w cx *^* sp - 4 |->w bx *^* sp - 2 |->w ax))) (MemMod.merge M3'' M4')).
 exists M3''. exists M4'.
 split.
  apply MemMod.disj_merge_disjUnion; auto. 
  inversion djM''.
  apply MemMod.disj_x_merge_disj_left with M2; auto.  (** MemMod.disjUnion M3'' M4' (MemMod.merge M3'' M4') **)
 split_l.

assert (mM4': (sp - 14 - 2 |->w (h_entry + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1) *^* sp - 14 |->w bp *^*
    sp - 12 |->w si *^* sp - 10 |->w di *^* sp - 8 |->w dx *^* 
    sp - 6 |->w cx *^* sp - 4 |->w bx *^* sp - 2 |->w ax) M4').
 substH upM4 with (MemUpd16_MemUpdW upM4).
 destruct upM4 as [x upM4].
 rewrite upM4.
 sep on.
 sep lift 0.
 sep upd.
 sep matchH mM4 with 3.
sep on.
sep matchH mM4' with 3.

sep on.
 sep lift 1.
 sep stksplit.
 simpl.
 sep matchH H1 with 4.
repeat split; auto.
 omega.
rewrite <- Hif''; trivial.
Qed.

Lemma IH_WFcmd_1137_scheduler:
  IH_WFcmd
     (SCAP_consCT  (h_entry + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1) (Ast p_timer_popbp g_timer_popbp ihtag) Psi_timer)
      (h_entry + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1) (Ast p_timer_scheduler g_timer_scheduler ihtag) (comml (_call _SWITCH)).
Proof.
left. unfold prim_entry.
split. left. trivial.
split_l.
exists p_timer_popbp. exists g_timer_popbp.
split. constructor; auto. 
split.  (** enable **)
 intros st [sp [ax [bx [cx [dx [di [si [bp pre]]]]]]]].
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqsp [spge [inIsr IFtrue]]]].
 induction st as (((h, R), fl), isr).  UnfoldStateAll.
 split_r.
  unfold timer_movax_MPre in mM1.
  assert (mM1r: ((sp - 14 |->w bp *^* sp - 12 |->w si *^* sp - 10 |->w di *^* sp - 8 |->w dx *^*
         sp - 6 |->w cx *^* sp - 4 |->w bx *^* sp - 2 |->w ax)
         *^* (INV0 *^* stkSpacePrec (sp - 14) (IH_Stk - 14))) M1).
  sep on.
    sep matchH mM1 with 4.
  destruct mM1r as [M3 [M4 [djM1 [mM3 mM4]]]].
  exists (MemMod.merge M2 M3). exists M4.
  split.
   destruct djM1 as [djm1 eqm1].
   apply MemMod.disjUnion_assoc_right_left; auto.
   rewrite <- eqm1. apply MemMod.disjUnion_commut; auto.
   split_l. 
  rewrite eqsp.   
   arith_rep (8+2) with (14+8+2-14).
split. (** pimp **)
 intros pc'.
 intros st' [st [pS nS]].
 destruct nS as [stR [[stC [nSC gSwtch]] nSR]].
 unfold p_timer_popbp.  unfold p_timer_scheduler in * .
 destruct pS as [sp [ax [bx [cx [dx [di [si [bp pre]]]]]]]].
 exists sp; exists ax; exists bx; exists cx; exists dx; exists di; exists si; exists bp.
 destruct pre as [M1 [M2 pre]].
 destruct (timer_scheduler_pre2post pc' sp ax bx cx dx di si bp st stC stR st' M1 M2
   pre nSC gSwtch nSR) as [M1' [dijh' pre']].
 exists M1'. exists M2. split_l. split_l. unfold mTrue; auto.
(** gimp **)
 intros pc' st st''  [st' [[pS nS] gS]].
 destruct nS as [stR [[stC [nSC gSwtch]] nSR]].
 unfold g_timer_popbp in gS. unfold g_timer_scheduler in * .
 intros M1 M2 djh sp ax bx cx dx di si bp pre.

 assert (IpS: MemMod.disjUnion M1 M2 (S_Mem st) /\
        timer_scheduler_pre sp ax bx cx dx di si bp (S_Mem_Upd st M1) /\ mTrue M2).
  split_l. split. trivial. unfold mTrue; trivial.
 destruct (timer_scheduler_pre2post pc' sp ax bx cx dx di si bp st stC stR st' M1 M2 
   IpS nSC gSwtch nSR) as [M1' [dijh' pre']]. 
 apply gS with M1'; auto.
Qed.

Definition timer_popsi_pre (sp ax bx cx dx di si : nat) (S : State) : Prop :=
  timer_pushbp_MPre sp ax bx cx dx di si (S_Mem S)
  /\ S_RF S SP = sp - 12
  /\ sp >= IH_Stk /\ S_ISR S = false /\ S_Flag S _IF = false.

Definition p_timer_popsi (S : State) : Prop :=
  exists sp, exists ax, exists bx, exists cx, exists dx, exists di, exists si,
    imprec (timer_popsi_pre sp ax bx cx dx di si) S.

Definition g_timer_popsi (S S' : State) : Prop := 
  forall M0 M1, MemMod.disjUnion M0 M1 (S_Mem S)
    -> forall sp ax bx cx dx di si, timer_popsi_pre sp ax bx cx dx di si (S_Mem_Upd S M0)
      -> exists M0', MemMod.disjUnion M0' M1 (S_Mem S') 
	/\ timer_pushbp_post sp ax bx cx dx di si S S' M0'.

Lemma timer_popbp_has_item:
  forall sp ax bx cx dx di si bp h, timer_movax_MPre sp ax bx cx dx di si bp h -> sp>= IH_Stk -> (sp-14 |~>w bp) h.
Proof.
unfold timer_movax_MPre; intros sp ax bx cx dx di si bp h pre spge.
assert (Hx : h ||- stkSpacePrec (sp-14) (IH_Stk-14) *^* sp-14 |->w bp
  *^* (sp-12 |->w si *^* sp-10 |->w di *^* sp-8 |->w dx *^* sp-6 |->w cx *^* sp-4 |->w bx *^* sp-2 |->w ax  *^* INV0)). 
  sep on.
  sep matchH pre with 4.
  sep lift 1 in Hx.
  sep on. sep trivial.
Qed.

Lemma timer_popbp_pre2post:
  forall sp ax bx cx dx di si bp h, timer_movax_MPre sp ax bx cx dx di si bp h
    -> sp >= IH_Stk -> timer_pushbp_MPre sp ax bx cx dx di si h.
Proof.
intros sp ax bx cx dx di si bp h pre spge.
unfold timer_pushbp_MPre, timer_movax_MPre in * .
  sep on.
  sep lift 1.
  apply stkSpacePrec_mergeW1.
  arith_rep (sp - 12 - 2) with (sp-14).
  arith_rep (IH_Stk - 12 - 2) with (IH_Stk-14).
  sep matchH pre with 4. omega. omega.
Qed.

Lemma IH_WFcmd_1138_popbp:
  IH_WFcmd (SCAP_consCT (h_entry + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1) (Ast p_timer_popsi g_timer_popsi ihtag) Psi_timer)
  (h_entry + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1) (Ast p_timer_popbp g_timer_popbp ihtag) (comml (_popw BP)).
Proof.
exists p_timer_popsi. exists g_timer_popsi.
split. constructor; auto.
split. (**  enable nextS **)
 intros st [sp [ax [bx [cx [dx [di [si [bp pre]]]]]]]].
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqsp [spge [NinIsr IFtrue]]]].
 induction st as (((h, R), fl), isr).  UnfoldStateAll.
 assert ((sp-14 |~>w bp) M1).
  apply timer_popbp_has_item with ax bx cx dx di si; auto.

 exists (h, (RF_update (RF_update R SP (sp-14+2)) BP bp), fl, isr).
 apply nextS_popw with (RF_update R SP (sp-14+2)) bp; auto.
 unfold popW; rewrite eqsp.
 split; auto. 
 apply leadsToW_Mono with M1 M2; auto.
split. (** pimp **)
 intros st' [st [pS nS]].

 destruct pS as [sp [ax [bx [cx [dx [di [si [bp pre]]]]]]]].
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqsp [spge [NinIsr IFtrue]]]]. UnfoldStateAll.

 induction st as (((h, R), fl), isr).
 induction st' as (((h', R'), fl'), isr').
 destruct (nextS_inversion nS); clear nS.
 rewrite eqsp in H0. 
 destruct H0 as [[bp' [ldt eqr]] [eqfl eqisr]].
 rewrite eqfl. rewrite eqisr.  rewrite eqr. rewrite H.

 exists sp; exists ax; exists bx; exists cx; exists dx; exists di; exists si. 
 exists M1. exists M2. split_l. split_r.
 repeat split; auto. unfold S_Mem_Upd, S_Mem.
 apply timer_popbp_pre2post with bp; auto.

 unfold S_Mem_Upd, S_RF, RF_update; unfold beq_reg. omega.
(**  gimp **)
 intros st st'' [st' [[pS nS] gS]].
 induction st as (((h, R), fl), isr).
 induction st' as (((h', R'), fl'), isr').
 destruct (nextS_inversion nS); clear nS.
 destruct H0 as [[bp' [ldt eqr]] [eqfl eqisr]].
 rewrite eqfl in gS. rewrite eqisr in gS.  rewrite eqr in gS. rewrite H in gS.
 induction st'' as (((h'', R''), fl''), isr'').

 intros M1 M2 djh sp ax bx cx dx di si bp post.
 destruct post as [mM1 [eqsp [spge [NinIsr IFtrue]]]]. UnfoldStateAll.

 assert (eqbp: bp' = bp).
  rewrite eqsp in ldt.
  apply leadsToW_unique with (sp-14) h; auto.
 apply leadsToW_Mono with M1 M2; auto.
  apply timer_popbp_has_item with ax bx cx dx di si; auto.
 rewrite eqbp in gS. 

 assert (pre:  timer_popsi_pre sp ax bx cx dx di si
         (M1, RF_update (RF_update R SP (R SP + 2)) BP bp, fl, isr)).
   repeat split; auto. simplS.
   apply timer_popbp_pre2post with bp; auto.
   unfold S_Mem_Upd, S_RF, RF_update; unfold beq_reg. rewrite eqsp. omega.

 unfold g_timer_popsi in gS; unfold timer_pushbp_post in gS. (** 2. pS -> gS -> pS'', S.H = M1 *^* M2 **)
 UnfoldStateAll.
 substH gS with (gS M1 M2 djh sp ax bx cx dx di si pre).
 destruct gS as [M1'' [djh'' [[pre'' NinISr] [eqax'' [eqbx [eqcx [eqdx'' [eqdi [eqsi [eqbp'' eqsp'']]]]]]]]]].

 exists M1''. split_l.
 repeat split; auto.
Qed.

Definition timer_popdi_pre (sp ax bx cx dx di : nat) (S : State) : Prop :=
  timer_pushsi_MPre sp ax bx cx dx di (S_Mem S)
  /\ S_RF S SP = sp - 10
  /\ sp >= IH_Stk /\ S_ISR S = false /\ S_Flag S _IF = false.

Definition p_timer_popdi (S : State) : Prop :=
  exists sp, exists ax, exists bx, exists cx, exists dx, exists di,
    imprec (timer_popdi_pre sp ax bx cx dx di) S.

Definition g_timer_popdi (S S' : State) : Prop := 
  forall M0 M1, MemMod.disjUnion M0 M1 (S_Mem S)
    -> forall sp ax bx cx dx di, timer_popdi_pre sp ax bx cx dx di (S_Mem_Upd S M0)
      -> exists M0', MemMod.disjUnion M0' M1 (S_Mem S') 
	/\ timer_pushsi_post sp ax bx cx dx di S S' M0'.

Lemma timer_popsi_has_item:
  forall sp ax bx cx dx di si h, timer_pushbp_MPre sp ax bx cx dx di si h -> sp >= IH_Stk -> (sp-12 |~>w si) h.
Proof.
unfold timer_pushbp_MPre; intros sp ax bx cx dx di si h pre spge.
  sep on.
  assert (Hx : h ||- stkSpacePrec (sp-12) (IH_Stk-12) *^* sp-12 |->w si
    *^* (sp-10 |->w di *^* sp-8 |->w dx *^* sp-6 |->w cx *^* sp-4 |->w bx *^* sp-2 |->w ax  *^* INV0)).
   sep matchH pre with 4.
   unfold mjudge.
  sep lift 1 in Hx.
  sep on. sep trivial.
Qed.

Lemma timer_popsi_pre2post:
  forall sp ax bx cx dx di si h, timer_pushbp_MPre sp ax bx cx dx di si h
    -> sp >= IH_Stk -> timer_pushsi_MPre sp ax bx cx dx di h.
Proof.
intros sp ax bx cx dx di si h pre spge.
unfold timer_pushbp_MPre, timer_pushsi_MPre in * .
  sep on.
  sep lift 1.
  apply stkSpacePrec_mergeW1.
  arith_rep (sp - 10 - 2) with (sp-12).
  arith_rep (IH_Stk - 10 - 2) with (IH_Stk-12).
  sep matchH pre with 4. omega. omega.
Qed.

Lemma IH_WFcmd_1139_popsi:
  IH_WFcmd (SCAP_consCT (h_entry + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1) (Ast p_timer_popdi g_timer_popdi ihtag) Psi_timer)
  (h_entry + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1) (Ast p_timer_popsi g_timer_popsi ihtag) (comml (_popw SI)).
Proof.
exists p_timer_popdi. exists g_timer_popdi.
split. constructor; auto.
split. (**  enable nextS **)
 intros st [sp [ax [bx [cx [dx [di [si pre]]]]]]].
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqsp [spge [NinIsr IFtrue]]]].
 induction st as (((h, R), fl), isr).  UnfoldStateAll.
 assert ((sp-12 |~>w si) M1).
  apply timer_popsi_has_item with ax bx cx dx di; auto.

 exists (h, (RF_update (RF_update R SP (sp-12+2)) SI si), fl, isr).
 apply nextS_popw with (RF_update R SP (sp-12+2)) si; auto.
 unfold popW; rewrite eqsp.
 split; auto. 
 apply leadsToW_Mono with M1 M2; auto.
split. (** pimp **)
 intros st' [st [pS nS]].

 destruct pS as [sp [ax [bx [cx [dx [di [si pre]]]]]]].
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqsp [spge [NinIsr IFtrue]]]].  

 induction st as (((h, R), fl), isr). UnfoldStateAll.
 induction st' as (((h', R'), fl'), isr').
 destruct (nextS_inversion nS); clear nS.
 rewrite eqsp in H0. 
 destruct H0 as [[bp' [ldt eqr]] [eqfl eqisr]].
 rewrite eqfl. rewrite eqisr.  rewrite eqr. rewrite H.

 exists sp; exists ax; exists bx; exists cx; exists dx; exists di. 
 exists M1. exists M2. split_l. split_r.
 repeat split; auto. simplS. 
 apply timer_popsi_pre2post with si; auto.

 unfold S_Mem_Upd, S_RF, RF_update; unfold beq_reg. omega.
(**  gimp **)
 intros st st'' [st' [[pS nS] gS]].
 induction st as (((h, R), fl), isr).
 induction st' as (((h', R'), fl'), isr').
 destruct (nextS_inversion nS); clear nS.
 destruct H0 as [[si' [ldt eqr]] [eqfl eqisr]].
 rewrite eqfl in gS. rewrite eqisr in gS.  rewrite eqr in gS. rewrite H in gS.
 induction st'' as (((h'', R''), fl''), isr'').

 intros M1 M2 djh sp ax bx cx dx di si post.
 destruct post as [mM1 [eqsp [spge [NinIsr IFtrue]]]]. UnfoldStateAll.

 assert (eqsi: si' = si).
  rewrite eqsp in ldt.
  apply leadsToW_unique with (sp-12) h; auto.
 apply leadsToW_Mono with M1 M2; auto.
  apply timer_popsi_has_item with ax bx cx dx di; auto.
 rewrite eqsi in gS. 

 assert (pre:  timer_popdi_pre sp ax bx cx dx di
         (M1, RF_update (RF_update R SP (R SP + 2)) SI si, fl, isr)).
   repeat split; auto. simplS.
   apply timer_popsi_pre2post with si; auto.
   unfold S_Mem_Upd, S_RF, RF_update; unfold beq_reg. rewrite eqsp. omega.

 unfold g_timer_popdi in gS; unfold timer_pushsi_post in gS. (** 2. pS -> gS -> pS'', S.H = M1 *^* M2 **)
 UnfoldStateAll.
 substH gS with (gS M1 M2 djh sp ax bx cx dx di pre).
 destruct gS as [M1'' [djh'' [[pre'' NinISr] [eqax [eqbx [eqcx [eqdx [eqdi [eqsi'' [eqbp eqsp'']]]]]]]]]].

 exists M1''. split_l.
 repeat split; auto.
Qed.

Definition timer_popdx_pre (sp ax bx cx dx : nat) (S : State) : Prop :=
  timer_pushdi_MPre sp ax bx cx dx (S_Mem S)
  /\ S_RF S SP = sp - 8
  /\ sp >= IH_Stk /\ S_ISR S = false /\ S_Flag S _IF = false.

Definition p_timer_popdx (S : State) : Prop :=
  exists sp, exists ax, exists bx, exists cx, exists dx, imprec (timer_popdx_pre sp ax bx cx dx) S.

Definition g_timer_popdx (S S' : State) : Prop := 
  forall M0 M1, MemMod.disjUnion M0 M1 (S_Mem S)
    -> forall sp ax bx cx dx, timer_popdx_pre sp ax bx cx dx (S_Mem_Upd S M0)
      -> exists M0', MemMod.disjUnion M0' M1 (S_Mem S') 
	/\ timer_pushdi_post sp ax bx cx dx S S' M0'.

Lemma timer_popdi_has_item:
  forall sp ax bx cx dx di h, timer_pushsi_MPre sp ax bx cx dx di h -> sp >= IH_Stk -> (sp-10 |~>w di) h.
Proof.
unfold timer_pushsi_MPre; intros sp ax bx cx dx di h pre spge.
  sep on.
  assert (Hx : h ||- stkSpacePrec (sp-10) (IH_Stk-10) *^* sp-10 |->w di
    *^* (sp-8 |->w dx *^* sp-6 |->w cx *^* sp-4 |->w bx *^* sp-2 |->w ax  *^* INV0)).
   sep matchH pre with 4.
  sep lift 1 in Hx.
  sep on. sep trivial.
Qed.

Lemma timer_popdi_pre2post:
  forall sp ax bx cx dx di h, timer_pushsi_MPre sp ax bx cx dx di h
    -> sp >= IH_Stk -> timer_pushdi_MPre sp ax bx cx dx h.
Proof.
intros sp ax bx cx dx di h pre spge.
unfold timer_pushdi_MPre, timer_pushsi_MPre in * .
  sep on.
  sep lift 1.
  apply stkSpacePrec_mergeW1.
  arith_rep (sp - 8 - 2) with (sp-10).
  arith_rep (IH_Stk - 8 - 2) with (IH_Stk-10).
  sep matchH pre with 4. omega. omega.
Qed.

Lemma IH_WFcmd_1140_popdi:
  IH_WFcmd (SCAP_consCT (h_entry + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1) (Ast p_timer_popdx g_timer_popdx ihtag) Psi_timer)
  (h_entry + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1) (Ast p_timer_popdi g_timer_popdi ihtag) (comml (_popw DI)).
Proof.
exists p_timer_popdx. exists g_timer_popdx.
split. constructor; auto.
split. (**  enable nextS **)
 intros st [sp [ax [bx [cx [dx [di pre]]]]]].
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqsp [spge [NinIsr IFtrue]]]].
 induction st as (((h, R), fl), isr).  UnfoldStateAll.
 assert ((sp-10 |~>w di) M1).
  apply timer_popdi_has_item with ax bx cx dx; auto.

 exists (h, (RF_update (RF_update R SP (sp-10+2)) DI di), fl, isr).
 apply nextS_popw with (RF_update R SP (sp-10+2)) di; auto.
 unfold popW; rewrite eqsp.
 split; auto. 
 apply leadsToW_Mono with M1 M2; auto.
split. (** pimp **)
 intros st' [st [pS nS]].

 destruct pS as [sp [ax [bx [cx [dx [di pre]]]]]].
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqsp [spge [NinIsr IFtrue]]]].  

 induction st as (((h, R), fl), isr). UnfoldStateAll.
 induction st' as (((h', R'), fl'), isr').
 destruct (nextS_inversion nS); clear nS.
 rewrite eqsp in H0. 
 destruct H0 as [[bp' [ldt eqr]] [eqfl eqisr]].
 rewrite eqfl. rewrite eqisr.  rewrite eqr. rewrite H.

 exists sp; exists ax; exists bx; exists cx; exists dx. 
 exists M1. exists M2. split_l. split_r.
 repeat split; auto. simplS. 
 apply timer_popdi_pre2post with di; auto.

 unfold S_Mem_Upd, S_RF, RF_update; unfold beq_reg. omega.
(**  gimp **)
 intros st st'' [st' [[pS nS] gS]].
 induction st as (((h, R), fl), isr).
 induction st' as (((h', R'), fl'), isr').
 destruct (nextS_inversion nS); clear nS.
 destruct H0 as [[di' [ldt eqr]] [eqfl eqisr]].
 rewrite eqfl in gS. rewrite eqisr in gS.  rewrite eqr in gS. rewrite H in gS.
 induction st'' as (((h'', R''), fl''), isr'').

 intros M1 M2 djh sp ax bx cx dx di post.
 destruct post as [mM1 [eqsp [spge [NinIsr IFtrue]]]]. UnfoldStateAll.

 assert (eqdi: di' = di).
  rewrite eqsp in ldt.
  apply leadsToW_unique with (sp-10) h; auto.
 apply leadsToW_Mono with M1 M2; auto.
  apply timer_popdi_has_item with ax bx cx dx; auto.
 rewrite eqdi in gS. 

 assert (pre:  timer_popdx_pre sp ax bx cx dx
         (M1, RF_update (RF_update R SP (R SP + 2)) DI di, fl, isr)).
   repeat split; auto. simplS.
   apply timer_popdi_pre2post with di; auto.
   unfold S_Mem_Upd, S_RF, RF_update; unfold beq_reg. rewrite eqsp. omega.

 unfold g_timer_popdx in gS; unfold timer_pushdi_post in gS. (** 2. pS -> gS -> pS'', S.H = M1 *^* M2 **)
 UnfoldStateAll.
 substH gS with (gS M1 M2 djh sp ax bx cx dx pre).
 destruct gS as [M1'' [djh'' [[pre'' NinISr] [eqax [eqbx [eqcx [eqdx [eqdi'' [eqsi [eqbp eqsp'']]]]]]]]]].

 exists M1''. split_l.
 repeat split; auto.
Qed.

Definition timer_popcx_pre (sp ax bx cx : nat) (S : State) : Prop :=
  timer_pushdx_MPre sp ax bx cx (S_Mem S)
  /\ S_RF S SP = sp - 6
  /\ sp >= IH_Stk /\ S_ISR S = false /\ S_Flag S _IF = false.

Definition p_timer_popcx (S : State) : Prop :=
  exists sp, exists ax, exists bx, exists cx, imprec (timer_popcx_pre sp ax bx cx) S.

Definition g_timer_popcx (S S' : State) : Prop := 
  forall M0 M1, MemMod.disjUnion M0 M1 (S_Mem S)
    -> forall sp ax bx cx, timer_popcx_pre sp ax bx cx (S_Mem_Upd S M0)
      -> exists M0', MemMod.disjUnion M0' M1 (S_Mem S') 
	/\ timer_pushdx_post sp ax bx cx S S' M0'.

Lemma timer_popdx_has_item:
  forall sp ax bx cx dx h, timer_pushdi_MPre sp ax bx cx dx h -> sp >= IH_Stk -> (sp-8 |~>w dx) h.
Proof.
unfold timer_pushdi_MPre; intros sp ax bx cx dx h pre spge.
  sep on.
  assert (Hx : h ||- stkSpacePrec (sp-8) (IH_Stk-8) *^* sp-8 |->w dx
    *^* (sp-6 |->w cx *^* sp-4 |->w bx *^* sp-2 |->w ax  *^* INV0)).
   sep matchH pre with 4.
  sep lift 1 in Hx.
  sep on. sep trivial.
Qed.

Lemma timer_popdx_pre2post:
  forall sp ax bx cx dx h, timer_pushdi_MPre sp ax bx cx dx h
    -> sp >= IH_Stk -> timer_pushdx_MPre sp ax bx cx h.
Proof.
intros sp ax bx cx dx h pre spge.
unfold timer_pushdi_MPre, timer_pushdx_MPre in * .
  sep on.
  sep lift 1.
  apply stkSpacePrec_mergeW1.
  arith_rep (sp - 6 - 2) with (sp-8).
  arith_rep (IH_Stk - 6 - 2) with (IH_Stk-8).
  sep matchH pre with 4. omega. omega.
Qed.

Lemma IH_WFcmd_1141_popdx:
  IH_WFcmd (SCAP_consCT (h_entry + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1) (Ast p_timer_popcx g_timer_popcx ihtag) Psi_timer)
  (h_entry + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1) (Ast p_timer_popdx g_timer_popdx ihtag) (comml (_popw DX)).
Proof.
exists p_timer_popcx. exists g_timer_popcx.
split. constructor; auto.
split. (**  enable nextS **)
 intros st [sp [ax [bx [cx [dx pre]]]]].
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqsp [spge [NinIsr IFtrue]]]].
 induction st as (((h, R), fl), isr).  UnfoldStateAll.
 assert ((sp-8 |~>w dx) M1).
  apply timer_popdx_has_item with ax bx cx; auto.

 exists (h, (RF_update (RF_update R SP (sp-8+2)) DX dx), fl, isr).
 apply nextS_popw with (RF_update R SP (sp-8+2)) dx; auto.
 unfold popW; rewrite eqsp.
 split; auto. 
 apply leadsToW_Mono with M1 M2; auto.
split. (** pimp **)
 intros st' [st [pS nS]].

 destruct pS as [sp [ax [bx [cx [dx pre]]]]].
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqsp [spge [NinIsr IFtrue]]]].  

 induction st as (((h, R), fl), isr). UnfoldStateAll.
 induction st' as (((h', R'), fl'), isr').
 destruct (nextS_inversion nS); clear nS.
 rewrite eqsp in H0. 
 destruct H0 as [[dx' [ldt eqr]] [eqfl eqisr]].
 rewrite eqfl. rewrite eqisr.  rewrite eqr. rewrite H.

 exists sp; exists ax; exists bx; exists cx. 
 exists M1. exists M2. split_l. split_r.
 repeat split; auto. simplS. 
 apply timer_popdx_pre2post with dx; auto.

 unfold S_Mem_Upd, S_RF, RF_update; unfold beq_reg. omega.
(**  gimp **)
 intros st st'' [st' [[pS nS] gS]].
 induction st as (((h, R), fl), isr).
 induction st' as (((h', R'), fl'), isr').
 destruct (nextS_inversion nS); clear nS.
 destruct H0 as [[dx' [ldt eqr]] [eqfl eqisr]].
 rewrite eqfl in gS. rewrite eqisr in gS.  rewrite eqr in gS. rewrite H in gS.
 induction st'' as (((h'', R''), fl''), isr'').

 intros M1 M2 djh sp ax bx cx dx post.
 destruct post as [mM1 [eqsp [spge [NinIsr IFtrue]]]]. UnfoldStateAll.

 assert (eqdx: dx' = dx).
  rewrite eqsp in ldt.
  apply leadsToW_unique with (sp-8) h; auto.
 apply leadsToW_Mono with M1 M2; auto.
  apply timer_popdx_has_item with ax bx cx; auto.
 rewrite eqdx in gS. 

 assert (pre:  timer_popcx_pre sp ax bx cx
         (M1, RF_update (RF_update R SP (R SP + 2)) DX dx, fl, isr)).
   repeat split; auto. simplS.
   apply timer_popdx_pre2post with dx; auto.
   unfold S_Mem_Upd, S_RF, RF_update; unfold beq_reg. rewrite eqsp. omega.

 unfold g_timer_popcx in gS; unfold timer_pushdx_post in gS. (** 2. pS -> gS -> pS'', S.H = M1 *^* M2 **)
 UnfoldStateAll.
 substH gS with (gS M1 M2 djh sp ax bx cx pre).
 destruct gS as [M1'' [djh'' [[pre'' NinISr] [eqax [eqbx [eqcx [eqdx'' [eqdi [eqsi [eqbp eqsp'']]]]]]]]]].

 exists M1''. split_l.
 repeat split; auto.
Qed.

Definition timer_popbx_pre (sp ax bx : nat) (S : State) : Prop :=
  timer_pushcx_MPre sp ax bx (S_Mem S)
  /\ S_RF S SP = sp - 4
  /\ sp >= IH_Stk /\ S_ISR S = false /\ S_Flag S _IF = false.

Definition p_timer_popbx (S : State) : Prop :=
  exists sp, exists ax, exists bx, imprec (timer_popbx_pre sp ax bx) S.

Definition g_timer_popbx (S S' : State) : Prop := 
  forall M0 M1, MemMod.disjUnion M0 M1 (S_Mem S)
    -> forall sp ax bx, timer_popbx_pre sp ax bx (S_Mem_Upd S M0)
      -> exists M0', MemMod.disjUnion M0' M1 (S_Mem S') 
	/\ timer_pushcx_post sp ax bx S S' M0'.

Lemma timer_popcx_has_item:
  forall sp ax bx cx h, timer_pushdx_MPre sp ax bx cx h -> sp >= IH_Stk -> (sp-6 |~>w cx) h.
Proof.
unfold timer_pushdx_MPre; intros sp ax bx cx h pre spge.
  sep on.
  assert (Hx : h ||- stkSpacePrec (sp-6) (IH_Stk-6) *^* sp-6 |->w cx
    *^* (sp-4 |->w bx *^* sp-2 |->w ax  *^* INV0)).
   sep matchH pre with 3.
  sep lift 1 in Hx.
  sep on. sep trivial.
Qed.

Lemma timer_popcx_pre2post:
  forall sp ax bx cx h, timer_pushdx_MPre sp ax bx cx h
    -> sp >= IH_Stk -> timer_pushcx_MPre sp ax bx h.
Proof.
intros sp ax bx cx h pre spge.
unfold timer_pushcx_MPre, timer_pushdx_MPre in * .
  sep on.
  sep lift 1.
  apply stkSpacePrec_mergeW1.
  arith_rep (sp - 4 - 2) with (sp-6).
  arith_rep (IH_Stk - 4 - 2) with (IH_Stk-6).
  sep matchH pre with 3. omega. omega.
Qed.

Lemma IH_WFcmd_1142_popcx:
  IH_WFcmd (SCAP_consCT (h_entry + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1) (Ast p_timer_popbx g_timer_popbx ihtag) Psi_timer)
  (h_entry + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1) (Ast p_timer_popcx g_timer_popcx ihtag) (comml (_popw CX)).
Proof.
exists p_timer_popbx. exists g_timer_popbx.
split. constructor; auto.
split. (**  enable nextS **)
 intros st [sp [ax [bx [cx pre]]]].
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqsp [spge [NinIsr IFtrue]]]].
 induction st as (((h, R), fl), isr).  UnfoldStateAll.
 assert ((sp-6 |~>w cx) M1).
  apply timer_popcx_has_item with ax bx; auto.

 exists (h, (RF_update (RF_update R SP (sp-6+2)) CX cx), fl, isr).
 apply nextS_popw with (RF_update R SP (sp-6+2)) cx; auto.
 unfold popW; rewrite eqsp.
 split; auto. 
 apply leadsToW_Mono with M1 M2; auto.
split. (** pimp **)
 intros st' [st [pS nS]].

 destruct pS as [sp [ax [bx [cx pre]]]].
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqsp [spge [NinIsr IFtrue]]]].  

 induction st as (((h, R), fl), isr). UnfoldStateAll.
 induction st' as (((h', R'), fl'), isr').
 destruct (nextS_inversion nS); clear nS.
 rewrite eqsp in H0. 
 destruct H0 as [[cx' [ldt eqr]] [eqfl eqisr]].
 rewrite eqfl. rewrite eqisr.  rewrite eqr. rewrite H.

 exists sp; exists ax; exists bx. 
 exists M1. exists M2. split_l. split_r.
 repeat split; auto. simplS. 
 apply timer_popcx_pre2post with cx; auto.

 unfold S_Mem_Upd, S_RF, RF_update; unfold beq_reg. omega.
(**  gimp **)
 intros st st'' [st' [[pS nS] gS]].
 induction st as (((h, R), fl), isr).
 induction st' as (((h', R'), fl'), isr').
 destruct (nextS_inversion nS); clear nS.
 destruct H0 as [[cx' [ldt eqr]] [eqfl eqisr]].
 rewrite eqfl in gS. rewrite eqisr in gS.  rewrite eqr in gS. rewrite H in gS.
 induction st'' as (((h'', R''), fl''), isr'').

 intros M1 M2 djh sp ax bx cx post.
 destruct post as [mM1 [eqsp [spge [NinIsr IFtrue]]]]. UnfoldStateAll.

 assert (eqcx: cx' = cx).
  rewrite eqsp in ldt.
  apply leadsToW_unique with (sp-6) h; auto.
 apply leadsToW_Mono with M1 M2; auto.
  apply timer_popcx_has_item with ax bx; auto.
 rewrite eqcx in gS. 

 assert (pre:  timer_popbx_pre sp ax bx
         (M1, RF_update (RF_update R SP (R SP + 2)) CX cx, fl, isr)).
   repeat split; auto. simplS.
   apply timer_popcx_pre2post with cx; auto.
   unfold S_Mem_Upd, S_RF, RF_update; unfold beq_reg. rewrite eqsp. omega.

 unfold g_timer_popbx in gS; unfold timer_pushcx_post in gS. (** 2. pS -> gS -> pS'', S.H = M1 *^* M2 **)
 UnfoldStateAll.
 substH gS with (gS M1 M2 djh sp ax bx pre).
 destruct gS as [M1'' [djh'' [[pre'' NinISr] [eqax [eqbx [eqcx'' [eqdx [eqdi [eqsi [eqbp eqsp'']]]]]]]]]].

 exists M1''. split_l.
 repeat split; auto.
Qed.

Definition timer_popax_pre (sp ax : nat) (S : State) : Prop :=
  timer_pushbx_MPre sp ax (S_Mem S)
  /\ S_RF S SP = sp - 2
  /\ sp >= IH_Stk /\ S_ISR S = false /\ S_Flag S _IF = false.

Definition p_timer_popax (S : State) : Prop :=
  exists sp, exists ax, imprec (timer_popax_pre sp ax) S.

Definition g_timer_popax (S S' : State) : Prop := 
  forall M0 M1, MemMod.disjUnion M0 M1 (S_Mem S)
    -> forall sp ax, timer_popax_pre sp ax (S_Mem_Upd S M0)
      -> exists M0', MemMod.disjUnion M0' M1 (S_Mem S') 
	/\ timer_pushbx_post sp ax S S' M0'.

Lemma timer_popbx_has_item:
  forall sp ax bx h, timer_pushcx_MPre sp ax bx h -> sp >= IH_Stk -> (sp-4 |~>w bx) h.
Proof.
unfold timer_pushcx_MPre; intros sp ax bx h pre spge.
  sep on.
  assert (Hx : h ||- stkSpacePrec (sp-4) (IH_Stk-4) *^* sp-4 |->w bx
    *^* (sp-2 |->w ax  *^* INV0)).
   sep matchH pre with 3.
  sep lift 1 in Hx.
  sep on. sep trivial.
Qed.

Lemma timer_popbx_pre2post:
  forall sp ax bx h, timer_pushcx_MPre sp ax bx h
    -> sp >= IH_Stk -> timer_pushbx_MPre sp ax h.
Proof.
intros sp ax bx h pre spge.
unfold timer_pushcx_MPre, timer_pushbx_MPre in * .
  sep on.
  sep lift 1.
  apply stkSpacePrec_mergeW1.
  arith_rep (sp - 2 - 2) with (sp-4).
  arith_rep (IH_Stk - 2 - 2) with (IH_Stk-4).
  sep matchH pre with 2. omega. omega.
Qed.

Lemma IH_WFcmd_1143_popbx:
  IH_WFcmd (SCAP_consCT (h_entry + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1) (Ast p_timer_popax g_timer_popax ihtag) Psi_timer)
  (h_entry + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1) (Ast p_timer_popbx g_timer_popbx ihtag) (comml (_popw BX)).
Proof.
exists p_timer_popax. exists g_timer_popax.
split. constructor; auto.
split. (**  enable nextS **)
 intros st [sp [ax [bx pre]]].
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqsp [spge [NinIsr IFtrue]]]].
 induction st as (((h, R), fl), isr).  UnfoldStateAll.
 assert ((sp-4 |~>w bx) M1).
  apply timer_popbx_has_item with ax; auto.

 exists (h, (RF_update (RF_update R SP (sp-4+2)) BX bx), fl, isr).
 apply nextS_popw with (RF_update R SP (sp-4+2)) bx; auto.
 unfold popW; rewrite eqsp.
 split; auto. 
 apply leadsToW_Mono with M1 M2; auto.
split. (** pimp **)
 intros st' [st [pS nS]].

 destruct pS as [sp [ax [bx pre]]].
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqsp [spge [NinIsr IFtrue]]]].  

 induction st as (((h, R), fl), isr). UnfoldStateAll.
 induction st' as (((h', R'), fl'), isr').
 destruct (nextS_inversion nS); clear nS.
 rewrite eqsp in H0. 
 destruct H0 as [[bx' [ldt eqr]] [eqfl eqisr]].
 rewrite eqfl. rewrite eqisr.  rewrite eqr. rewrite H.

 exists sp; exists ax. 
 exists M1. exists M2. split_l. split_r.
 repeat split; auto. simplS. 
 apply timer_popbx_pre2post with bx; auto.

 unfold S_Mem_Upd, S_RF, RF_update; unfold beq_reg. omega.
(**  gimp **)
 intros st st'' [st' [[pS nS] gS]].
 induction st as (((h, R), fl), isr).
 induction st' as (((h', R'), fl'), isr').
 destruct (nextS_inversion nS); clear nS.
 destruct H0 as [[bx' [ldt eqr]] [eqfl eqisr]].
 rewrite eqfl in gS. rewrite eqisr in gS.  rewrite eqr in gS. rewrite H in gS.
 induction st'' as (((h'', R''), fl''), isr'').

 intros M1 M2 djh sp ax bx post.
 destruct post as [mM1 [eqsp [spge [NinIsr IFtrue]]]]. UnfoldStateAll.

 assert (eqbx: bx' = bx).
  rewrite eqsp in ldt.
  apply leadsToW_unique with (sp-4) h; auto.
 apply leadsToW_Mono with M1 M2; auto.
  apply timer_popbx_has_item with ax; auto.
 rewrite eqbx in gS. 

 assert (pre:  timer_popax_pre sp ax (M1, RF_update (RF_update R SP (R SP + 2)) BX bx, fl, isr)).
   repeat split; auto. simplS.
   apply timer_popbx_pre2post with bx; auto.
   unfold S_Mem_Upd, S_RF, RF_update; unfold beq_reg. rewrite eqsp. omega.

 unfold g_timer_popax in gS; unfold timer_pushbx_post in gS. (** 2. pS -> gS -> pS'', S.H = M1 *^* M2 **)
 UnfoldStateAll.
 substH gS with (gS M1 M2 djh sp ax pre).
 destruct gS as [M1'' [djh'' [[pre'' NinISr] [eqax [eqbx'' [eqcx [eqdx [eqdi [eqsi [eqbp eqsp'']]]]]]]]]].

 exists M1''. split_l.
 repeat split; auto.
Qed.

Definition timer_iret_pre (sp : nat) (S : State) : Prop :=
  (INV0 *^* stkSpacePrec sp IH_Stk) (S_Mem S)
  /\ S_RF S SP = sp
  /\ sp >= IH_Stk /\ S_ISR S = false /\ S_Flag S _IF = false.

Definition p_timer_iret (S : State) : Prop := exists sp, imprec (timer_iret_pre sp) S.

Definition g_timer_iret (S S' : State) : Prop := 
  forall M0 M1, MemMod.disjUnion M0 M1 (S_Mem S)
    -> forall sp, timer_iret_pre sp (S_Mem_Upd S M0)
      -> exists M0', MemMod.disjUnion M0' M1 (S_Mem S') 
	/\  timer_post sp (S_Mem_Upd S' M0') /\ S_RF S' = S_RF S.

Lemma timer_popax_has_item:
  forall sp ax h, timer_pushbx_MPre sp ax h -> sp >= IH_Stk -> (sp-2 |~>w ax) h.
Proof.
unfold timer_pushbx_MPre; intros sp ax h pre spge.
  sep on.
  assert (Hx : h ||- stkSpacePrec (sp-2) (IH_Stk-2) *^* sp-2 |->w ax *^* INV0).
   sep matchH pre with 3.
  sep lift 1 in Hx.
  sep on. sep trivial.
Qed.

Lemma timer_popax_pre2post:
  forall sp ax h, timer_pushbx_MPre sp ax h ->  sp >= IH_Stk -> (INV0 *^* stkSpacePrec sp IH_Stk) h.
Proof.
intros sp ax h pre spge.
unfold timer_pushbx_MPre in * .
  sep on.
  sep lift 1.
  apply stkSpacePrec_mergeW1.
  sep matchH pre with 2. omega. omega. 
Qed.

Lemma IH_WFcmd_1144_popax:
  IH_WFcmd (SCAP_consCT (h_entry + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1) (Ast p_timer_iret g_timer_iret ihtag) Psi_timer)
  (h_entry + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1) (Ast p_timer_popax g_timer_popax ihtag) (comml (_popw AX)).
Proof.
exists p_timer_iret. exists g_timer_iret.
split. constructor; auto.
split. (**  enable nextS **)
 intros st [sp [ax pre]].
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqsp [spge [NinIsr IFtrue]]]].
 induction st as (((h, R), fl), isr).  UnfoldStateAll.
 assert ((sp-2 |~>w ax) M1).
  apply timer_popax_has_item; auto.

 exists (h, (RF_update (RF_update R SP (sp-2+2)) AX ax), fl, isr).
 apply nextS_popw with (RF_update R SP (sp-2+2)) ax; auto.
 unfold popW; rewrite eqsp.
 split; auto. 
 apply leadsToW_Mono with M1 M2; auto.
split. (** pimp **)
 intros st' [st [pS nS]].

 destruct pS as [sp [ax pre]].
 destruct pre as [M1 [M2 [djh [pre mM2]]]].
 destruct pre as [mM1 [eqsp [spge [NinIsr IFtrue]]]].  

 induction st as (((h, R), fl), isr). UnfoldStateAll.
 induction st' as (((h', R'), fl'), isr').
 destruct (nextS_inversion nS); clear nS.
 rewrite eqsp in H0. 
 destruct H0 as [[ax' [ldt eqr]] [eqfl eqisr]].
 rewrite eqfl. rewrite eqisr.  rewrite eqr. rewrite H.

 exists sp. exists M1. exists M2. split_l. split_r.
 repeat split; auto. simplS. 
 apply timer_popax_pre2post with ax; auto.

 unfold S_Mem_Upd, S_RF, RF_update; unfold beq_reg. omega.
(**  gimp **)
 intros st st'' [st' [[pS nS] gS]].
 induction st as (((h, R), fl), isr).
 induction st' as (((h', R'), fl'), isr').
 destruct (nextS_inversion nS); clear nS.
 destruct H0 as [[ax' [ldt eqr]] [eqfl eqisr]].
 rewrite eqfl in gS. rewrite eqisr in gS.  rewrite eqr in gS. rewrite H in gS.
 induction st'' as (((h'', R''), fl''), isr'').

 intros M1 M2 djh sp ax post.
 destruct post as [mM1 [eqsp [spge [NinIsr IFtrue]]]]. UnfoldStateAll.

 assert (eqax: ax' = ax).
  rewrite eqsp in ldt.
  apply leadsToW_unique with (sp-2) h; auto.
 apply leadsToW_Mono with M1 M2; auto.
  apply timer_popax_has_item; auto.
 rewrite eqax in gS. 

 assert (pre:  timer_iret_pre sp (M1, RF_update (RF_update R SP (R SP + 2)) AX ax, fl, isr)).
   repeat split; auto. simplS.
   apply timer_popax_pre2post with ax; auto.
   unfold S_Mem_Upd, S_RF, RF_update; unfold beq_reg.  rewrite eqsp. omega.

 unfold g_timer_iret in gS. (** 2. pS -> gS -> pS'', S.H = M1 *^* M2 **)
 UnfoldStateAll.
 substH gS with (gS M1 M2 djh sp pre).
 destruct gS as [M1'' [djh'' [[pre'' NinISr] eqR]]].

 exists M1''. split_l.
 rewrite eqsp in eqR. rewrite eqR.
 
 repeat split; auto.
 simplS; reg simpl. omega.
Qed.

Lemma IH_WFcmd_1145_iret:
 IH_WFcmd Psi_timer (h_entry + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1) 
   (Ast p_timer_iret g_timer_iret ihtag) (comms _iret).
Proof.
  simpl.
  split_l. (** ihtag = zero, see the defination in ./SCAP/AIM.v and ./SCAP/SCAP_Itr.v *)
  intros S HS M1 M2 djh sp [mM1 [eqsp [spge [NinIsr IFtrue]]]].
  unfold p_timer_iret in HS.
  unfold timer_iret_pre in HS.
  exists M1. 
  split_l. 
  repeat split; auto. 
Qed.

(**************************************************************)
(** Well-formed Instruction Sequence                    *******)
Lemma IH_WFiseq_timer :
  IH_WFiseq Psi_timer h_entry s_timer I_timer.
Proof.
unfold I_timer.
apply ih_isq_seq with p_timer_pushbx g_timer_pushbx.
apply IH_WFcmd_1127_pushax.
apply ih_isq_seq with p_timer_pushcx g_timer_pushcx.
apply IH_WFcmd_1128_pushbx.
apply ih_isq_seq with p_timer_pushdx g_timer_pushdx.
apply IH_WFcmd_1129_pushcx.
apply ih_isq_seq with p_timer_pushdi g_timer_pushdi.
apply IH_WFcmd_1130_pushdx.
apply ih_isq_seq with p_timer_pushsi g_timer_pushsi.
apply IH_WFcmd_1131_pushdi.
apply ih_isq_seq with p_timer_pushbp g_timer_pushbp.
apply IH_WFcmd_1132_pushsi.
apply ih_isq_seq with p_timer_movax g_timer_movax.
apply IH_WFcmd_1133_pushbp.

apply ih_isq_seq with (conjP p_timer_movax (fun st : State => S_RF st AX = 32)) g_timer_movax.
apply IH_WFcmd_1134_movax.
apply ih_isq_seq with (conjP p_timer_movax (fun st : State => S_RF st AX = 32 /\ S_RF st DX =32)) g_timer_movax.
apply IH_WFcmd_1135_movdx.
apply ih_isq_seq with p_timer_scheduler g_timer_scheduler.
apply IH_WFcmd_1136_eoi.

apply ih_isq_seq with p_timer_popbp g_timer_popbp.
apply IH_WFcmd_1137_scheduler.
apply ih_isq_seq with p_timer_popsi g_timer_popsi.
apply IH_WFcmd_1138_popbp.
apply ih_isq_seq with p_timer_popdi g_timer_popdi.
apply IH_WFcmd_1139_popsi.
apply ih_isq_seq with p_timer_popdx g_timer_popdx.
apply IH_WFcmd_1140_popdi.
apply ih_isq_seq with p_timer_popcx g_timer_popcx.
apply IH_WFcmd_1141_popdx.
apply ih_isq_seq with p_timer_popbx g_timer_popbx.
apply IH_WFcmd_1142_popcx.
apply ih_isq_seq with p_timer_popax g_timer_popax.
apply IH_WFcmd_1143_popbx.
apply ih_isq_seq with p_timer_iret g_timer_iret.
apply IH_WFcmd_1144_popax.

apply ih_isq_end.
apply IH_WFcmd_1145_iret.
Qed.
