(***************************************************************)
(* Coq code for AIM                                            *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

Add LoadPath "..".
Add LoadPath "../CAP0".
Add LoadPath "../SCAP".

Require Import Coqlib.

Require Import Arith.
Require Import Machine.
Require Import Assertion.
Require Import Operational_Semantics.
Require Import OpSem_Lemmas.
Require Import OpSem_Lemmas2.
Require Import Reg.
Require Import Assertion_Lemmas3.
Require Import Assertion_Lemmas4.

Require Import AIM.
Require Import SCAP.
Require Import QLib.
Require Import Linking.
Require Import Linking_Lemmas3.

Require Import Kernel.

Require Import UnblockSpec.

(* <unblock + 0>:      pushw   %bp *)

Lemma SCAP_WFcmd_0_pushbp:
  SCAP_WFcmd (SCAP_consCT (L_unblock + 1) 
    (Ast p_unblock_1 g_unblock_1 sctag) psi_unblock)
     L_unblock spec_unblock (comml (_pushwp (Opr_reg BP))).
Proof.
  assert (Hpp' : forall T RdyQ BlkQ l_ptr bp sp jk mP S S',
    Ast_unblock_pre T RdyQ BlkQ l_ptr bp sp jk mP S
    -> nextS (L_unblock) (comml (_pushwp (Opr_reg BP))) S S'
    -> Ast_unblock_1 T RdyQ BlkQ l_ptr bp sp jk mP S').
    intros T RdyQ BlkQ l_ptr bp sp jk mP S S' HS HSS'.
    unfold Ast_unblock_pre, Ast_unblock_1 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspgt [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simplS in * .
    destruct HS as [HM [Hsp [Hbp [Hspg [HRT [HBT [HRB HT]]]]]]].
    split.
      subst M' R' fl' isr'.
      sep on.
      sep lift 3.
      rewrite Hsp.
      rewrite Hbp.
      sep upd.
      sep matchH HM with 7.
      sep stksplit in HM.
      sep trivial.
    subst R' fl' isr'; reg simpl.
    rewrite Hsp; rewrite Hbp.
    repeat split_l; trivial.
  simpl.
  exists p_unblock_1; exists g_unblock_1.
  split_l.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold p_unblock in HpS.
    destruct HpS as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk HS]]]]]]].
    exists (MemUpdW M (sp-2) (R BP) 0, RF_update R SP (sp - 2), fl, isr).
    apply nextS_pushwp with (R BP); trivial.
    unfold pushW.
    destruct HS as [HM [Hbp [Hsp [Hspg [HRT [HBT [HRB HT]]]]]]].
    simplS in * .
    rewrite Hsp; rewrite Hbp.
    split.
      apply stkSpace_weaken with (_UNBLK_Stk);
        [idtac | omega].
      sep on.
      sep lift 2 in HM.
      sep trivial.
    split.
      apply MemUpdW_MemUpd16 with 0; trivial.
    trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold p_unblock, p_unblock_1.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk HS]]]]]]].
    exists T; exists RdyQ; exists BlkQ; exists l_ptr.
      exists bp; exists sp; exists jk.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold g_unblock, g_unblock_1 in * .
  intros mP T RdyQ BlkQ l_ptr bp sp jk HS.
  eapply HgS'S''; eauto.
Qed.

(* <unblock+1>:        movw    %sp, %bp *)

Lemma SCAP_WFcmd_1_movwp:
  SCAP_WFcmd (SCAP_consCT (L_unblock +1+1) 
    (Ast p_unblock_2 g_unblock_2 sctag) psi_unblock)
     (L_unblock +1) (Ast p_unblock_1 g_unblock_1 sctag)
     (comml (_movwp  (Opr_reg SP) BP)).
Proof.
  assert (Hpp' : forall T RdyQ BlkQ l_ptr bp sp jk mP S S',
    Ast_unblock_1 T RdyQ BlkQ l_ptr bp sp jk mP S
    -> nextS (L_unblock+1) (comml (_movwp  (Opr_reg SP) BP)) S S'
    -> Ast_unblock_2 T RdyQ BlkQ l_ptr bp sp jk mP S').
    intros T RdyQ BlkQ l_ptr bp sp jk mP S S' HS HSS'.
    unfold Ast_unblock_1, Ast_unblock_2 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    simplS in * .
    destruct HS as [HM [Hsp [Hbp [Hspg [HRT [HBT [HRB HT]]]]]]].
    subst M' R' fl' isr'.
    reg simpl.
    repeat split_l; trivial.
  simpl.
  exists p_unblock_2; exists g_unblock_2.
  split_l.
  split.
    (* enable movw *)
    unfold enable, p_unblock.
    intros S H.
    destruct H as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk HS]]]]]]].
    unfold Ast_unblock_pre in HS.
    destruct HS as [HM [Hsp [Hbp [Hspg [HRT [HBT [HRB HT]]]]]]].
    destruct S as [[[M R] fl] isr].
    simplS in * .
    set (R' := RF_update R BP (sp - 2)).
    exists (M, R', fl, isr).
    apply nextS_movwp with (sp - 2); auto.
  split. (* p ==> p' *)
    unfold p_unblock_1, p_unblock_2.
    intros S' H.
    unfold linkPG in H.
    destruct H as [S H].
    destruct H as [H HnextS].
    destruct H as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk HS]]]]]]].
    exists T; exists RdyQ; exists BlkQ; exists l_ptr.
    exists bp; exists sp; exists jk.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold conjPG, conjG in H.
  destruct H as [S' [[_ HnextS] HS'S'']].
  unfold g_unblock_1, g_unblock_2 in * .
  intros mP T RdyQ BlkQ l_ptr bp sp jk HS.
  eapply HS'S''; eauto.
Qed.

(* <unblock+2> :   movw    4(%bp), %bx             # the pointer to the block queue *)

Lemma SCAP_WFcmd_2_movwld:
  SCAP_WFcmd (SCAP_consCT (L_unblock +1+1+1) 
    (Ast p_unblock_3 g_unblock_3 sctag) psi_unblock)
     (L_unblock +1+1) (Ast p_unblock_2 g_unblock_2 sctag)
     (comml (_movwld (A_reg 4 BP) BX)).
Proof.
  assert (Hpp' : forall T RdyQ BlkQ l_ptr bp sp jk mP S S',
    Ast_unblock_2 T RdyQ BlkQ l_ptr bp sp jk mP S
    -> nextS (L_unblock +1+1) (comml (_movwld (A_reg 4 BP) BX)) S S'
    -> Ast_unblock_3 T RdyQ BlkQ l_ptr bp sp jk mP S').
    intros T RdyQ BlkQ l_ptr bp sp jk mP S S' HS HSS'.
    unfold Ast_unblock_2, Ast_unblock_3 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HR' as [l_ptrx [Hlds HR']].
    simplS in * .
    destruct HS as [HM [Hsp [Hbp [Hspg [HRT [HBT [HRB HT]]]]]]].
    assert (Heq : l_ptrx = l_ptr).
      unfold addr_eval in Hlds.
      rewrite Hbp in Hlds.
      sep on.
      sep lift 4 in HM.
      arith_rep (sp - 2 + 4) with (sp + 2) in Hlds.
      sep inj.
      auto.
    subst l_ptrx M' R' fl' isr'.
    reg simpl.
    repeat split_l; trivial.
  simpl.
  exists p_unblock_3; exists g_unblock_3.
  split_l.
  split.
  (* enable *)
    unfold enable.
    intros S HS.
    destruct S as [[[M R] fl] isr].
    unfold p_unblock_2 in HS.
    destruct HS as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk HS]]]]]]].
    unfold Ast_unblock_2 in HS.
    destruct HS as [HM [Hsp [Hbp [Hspg [HRT [HBT [HRB HT]]]]]]].
    simplS in * .
    exists (M, RF_update R BX l_ptr, fl, isr).
    apply nextS_movwld with (sp+2) l_ptr; auto.
      unfold addr_eval; rewrite Hbp; trivial.
      omega.
    sep on.
    sep lift 4 in HM.
    sep trivial.
    apply ptoW_leadsToW1 with mTrue.
    sep subst_left with (sp + 2 |-> l_ptr *^* sp + 2 + 1 |->_).
    arith_rep (sp + 2 + 1) with (sp + 3).
    sep lift 0.
    sep match_left HM.
    sep lift 0 in HM.
    sep match_left HM.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold p_unblock_2, p_unblock_3.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk HS]]]]]]].
    exists T; exists RdyQ; exists BlkQ; exists l_ptr.
    exists bp; exists sp; exists jk.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold g_unblock_2, g_unblock_3 in * .
  intros mP T RdyQ BlkQ l_ptr bp sp jk HS.
  eapply HgS'S''; eauto.
Qed.

(* <unblock+3> :        pushw   %bx *)

Lemma SCAP_WFcmd_3_pushwp:
  SCAP_WFcmd (SCAP_consCT (L_unblock +1+1+1+1) 
    (Ast p_unblock_4 g_unblock_4 sctag) psi_unblock)
     (L_unblock +1+1+1) (Ast p_unblock_3 g_unblock_3 sctag)
     (comml (_pushwp (Opr_reg BX))).
Proof.
  assert (Hpp' : forall T RdyQ BlkQ l_ptr bp sp jk mP S S',
    Ast_unblock_3 T RdyQ BlkQ l_ptr bp sp jk mP S
    -> nextS (L_unblock +1+1+1) (comml (_pushwp (Opr_reg BX))) S S'
    -> Ast_unblock_4 T RdyQ BlkQ l_ptr bp sp jk mP S').
    intros T RdyQ BlkQ l_ptr bp sp jk mP S S' HS HSS'.
    unfold Ast_unblock_3, Ast_unblock_4 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspgt [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simplS in * .
    destruct HS as [HM [Hbx [Hsp [Hbp [Hspg [HRT [HBT [HRB HT]]]]]]]].
    split.
      subst M' R' fl' isr'.
      sep on.
      sep lift 3.
      rewrite Hsp.
      rewrite Hbx.
      sep upd.
      sep matchH HM with 7.
      sep stksplit in HM.
      sep trivial.
    subst R' fl' isr'; reg simpl.
    rewrite Hsp; rewrite Hbx.
    repeat split_l; trivial.
  simpl.
  exists p_unblock_4; exists g_unblock_4.
  split_l.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold p_unblock_4 in HpS.
    destruct HpS as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk HS]]]]]]].
    exists (MemUpdW M (sp-2-2) (R BX) 0, RF_update R SP (sp - 2-2), fl, isr).
    apply nextS_pushwp with (R BX); trivial.
    unfold pushW.
    unfold Ast_unblock_3 in HS.
    destruct HS as [HM [Hbx [Hsp [Hbp [Hspg [HRT [HBT [HRB HT]]]]]]]].
    simplS in * .
    rewrite Hsp; rewrite Hbx.
    split.
      apply stkSpace_weaken with (_UNBLK_Stk - 2);
        [idtac | omega].
      sep on.
      sep lift 2 in HM.
      sep trivial.
    split.
      apply MemUpdW_MemUpd16 with 0; trivial.
    trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold p_unblock_3, g_unblock_4.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk HS]]]]]]].
    exists T; exists RdyQ; exists BlkQ; exists l_ptr.
    exists bp; exists sp; exists jk.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold g_unblock_3, g_unblock_4 in * .
  intros mP T RdyQ BlkQ l_ptr bp sp jk HS.
  eapply HgS'S''; eauto.
Qed.

(* <unblock+4> :        call    deQueue                 # not change bp *)

Lemma SCAP_WFcmd_4_call:
  SCAP_WFcmd (SCAP_consCT (L_unblock +1+1+1+1+1) 
    (Ast p_unblock_5 g_unblock_5 sctag) psi_unblock)
     (L_unblock +1+1+1+1) (Ast p_unblock_4 g_unblock_4 sctag)
     (comml (_call L_deQueue)).
Proof.
  assert(Hpp' : forall T RdyQ BlkQ l_ptr bp sp jk mQ mP S S',
    Ast_unblock_4 T RdyQ BlkQ l_ptr bp sp jk mP S
    -> nextS (L_unblock +1+1+1+1) (comml (_call L_deQueue)) S S'
    -> mQ = ((embed_RdyQ T RdyQ)
        *^* (stkSpacePrec (sp -2-2-2-2) (_UNBLK_Stk -2-2-2-2))
        *^* (sp -2-2-2 |->w (L_unblock +1+1+1+1+1))
        *^* (sp - 2 |->w bp)
        *^* (sp + 2 |-> l_ptr) 
        *^* (sp + 3 |-> jk))
    -> Ast_deQueue_pre T BlkQ l_ptr (S_RF S' BP) (sp -2-2-2)
    (mQ *^* mP) S').
    intros T RdyQ BlkQ l_ptr bp sp jk mQ mP S S'.
    intros HS HSS' HmQ.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    unfold Ast_unblock_4 in HS.
    destruct HS as [HM [Hbx [Hsp [Hbp [Hspg [HRT [HBT [HRB HT]]]]]]]].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspx [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    subst M' R' fl' isr'.
    unfold Ast_deQueue_pre.
    simplS in * .
    reg simpl.
    split.
      sep on.
      rewrite Hsp.
      subst mQ.
      sep lift 5.
      sep upd.
      sep lift 2 in HM.
      sep stksplit in HM.
      unfold embed_BlkQ in HM.
      sep stksplit 2 in HM.
      assert (Ha1 : sp - 2 - 2 - 2 + 2 = sp - 2 - 2).
        generalize Hspg; clearall.
        omega.
      rewrite Ha1.
      sep matchH HM with 13.
    split_l.
    rewrite Hsp; trivial.
  assert (Hpp''' : forall T RdyQ BlkQ l_ptr bp sp jk
    pc' mP S S' S'' S''',
    Ast_unblock_4 T RdyQ BlkQ l_ptr bp sp jk mP S
    -> nextS (L_unblock +1+1+1+1) (comml (_call L_deQueue)) S S'
    -> g_deQueue S' S''
    -> nextS pc' (comms (_retw)) S'' S'''
    -> Ast_unblock_5_a T RdyQ BlkQ l_ptr bp sp jk mP S'''
    \/ exists BlkQ', exists tid', exists tcb',
      Ast_unblock_5_b T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S''').
    intros T RdyQ BlkQ l_ptr bp sp jk pc' mP.
    intros S S' S'' S''' HS HSS' HS'S'' HS''S'''.
    unfold g_deQueue in HS'S''.
    set (mQ := (embed_RdyQ T RdyQ)
        *^* (stkSpacePrec (sp -2-2-2-2) (_UNBLK_Stk -2-2-2-2))
        *^* (sp -2-2-2 |->w (L_unblock +1+1+1+1+1))
        *^* (sp - 2 |->w bp)
        *^* (sp + 2 |-> l_ptr) 
        *^* (sp + 3 |-> jk)).
    assert (HS' : Ast_deQueue_pre T BlkQ l_ptr (S_RF S' BP) (sp-2-2-2)
      (mQ *^* mP) S').
      eapply Hpp'; eauto.
    clear Hpp'.
    substH HS'S'' with (HS'S'' _ _ _ _ _ _ HS') into HS''.
    unfold Ast_unblock_4 in HS.
    destruct HS as [_ [Hbx [Hsp [Hbp [Hspg [HRT [HBT [HRB HT]]]]]]]].
    destruct HS'' as [HS'' | HS''].
      destruct HS'' as [Hempty HS''].
      left.
      unfold Ast_deQueue_post_a in HS''.
      unfold Ast_unblock_5_a.
      destruct S as [[[M R] fl] isr].
      destruct S' as [[[M' R'] fl'] isr'].
      destruct S'' as [[[M'' R''] fl''] isr''].
      destruct S''' as [[[M''' R'''] fl'''] isr'''].
      simplS in * .
      destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
      destruct (nextS_inversion HS''S''') as [HM''' [HR''' [Hfl''' Hisr''']]].
      destruct HR''' as [pcx [Hlds HR''']].
      destruct HS'' as [HM'' [HAX'' [HBP'' HSP'']]].
      assert (Hpcx : pcx = L_unblock +1+1+1+1+1).
        subst R' M''' R'''.
        subst fl' isr' fl''' isr'''.
        rewrite HSP'' in Hlds; clear HSP''.
        sep on.
        subst mQ.
        sep lift 5 in HM''.
        sep_inj.
        auto.
      subst pcx; clear Hlds.
      subst M''' R''' fl''' isr'''.
      reg simpl.
      rewrite HAX''.
      rewrite HSP''.
      rewrite HBP''.
      subst R'.
      reg simpl.
      assert (Ha1 : sp - 2 - 2 - 2 + 2 = sp - 2 - 2).
        generalize Hspg; clearall.
        omega.
      split.
        subst mQ.
        unfold embed_BlkQ in * .
        clear HRT HRB HBT HT Hbx Hsp Hbp HSS' HS' Hempty.
        clear HAX'' HBP'' HSP'' HM' HS''S'''.
        sep on.
        rewrite Ha1 in HM''.
        sep matchH HM'' with 11.
        clear Ha1.
        sep stksplit.
        sep stksplit 2.
        sep matchH HM'' with 2.
      repeat split_l; trivial.
    destruct HS'' as [BlkQ' [tid' [tcb' [HS'' [HBlkQ' Hnin]]]]].
    right.
    exists BlkQ'; exists tid'; exists tcb'.
    subst mQ.
    unfold Ast_deQueue_post_b in HS''.
    unfold Ast_unblock_5_b.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct S'' as [[[M'' R''] fl''] isr''].
    destruct S''' as [[[M''' R'''] fl'''] isr'''].
    simplS in * .
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct (nextS_inversion HS''S''') as [HM''' [HR''' [Hfl''' Hisr''']]].
    destruct HR''' as [pcx [Hlds HR''']].
    destruct HS'' as [HM'' [HAX'' [HBP'' [HSP'' Hmto']]]].
    assert (Hpcx : pcx = L_unblock +1+1+1+1+1).
      subst R' M''' R'''.
      subst fl' isr' fl''' isr'''.
      rewrite HSP'' in Hlds; clear HSP''.
      sep on.
      sep lift 6 in HM''.
      sep_inj.
      auto.
    subst pcx; clear Hlds.
    subst M''' R''' fl''' isr'''.
    reg simpl.
    rewrite HAX''.
    rewrite HSP''.
    rewrite HBP''.
    subst R'.
    reg simpl.
    clear HAX'' HSP'' HBP''.
    assert (Ha1 : sp - 2 - 2 - 2 + 2 = sp - 2 - 2).
      generalize Hspg; clearall.
      omega.
    rewrite Ha1.
    rewrite Ha1 in HM''.
    clear Ha1.
    split.
      unfold embed_BlkQ in * .
      sep on.
      sep matchH HM'' with 11.
      generalize Hspg HM''; clearall; intros.
      sep stksplit.
      sep stksplit 2.
      sep matchH HM'' with 2.
    repeat split_l; trivial.
    apply HT; exists tcb'; trivial.
    subst BlkQ.
    unfold addThrd.
    auto with PMap.
    subst BlkQ.
    unfold addThrd.
    rewrite WQMod.remove_cancel_put.
    rewrite <- WQMod.remove_nothing.
    trivial.
    trivial.
  simpl.
  exists p_deQueue; exists p_unblock_5.
  exists g_deQueue; exists g_unblock_5.
  split.
    unfold SCAP_lookupCT, SCAP_consCT.
    apply SPSpec.lookup_merge_intro_r.
    apply L_deQueue_in_psi_unblock.
  split_l.
  split.
    (* enable call *)
    clear Hpp' Hpp'''.
    unfold enable, p_unblock_4.
    intros S HS.
    destruct HS as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk HS]]]]]]].
    unfold Ast_unblock_4 in HS.
    destruct HS as [HM [HBX [HSP [HBP [Hsp _]]]]].
    destruct S as [[[M R] fl] isr].
    simplS in * .
    exists (MemUpdW M (sp - 2 - 2 - 2) (L_unblock+1+1+1+1+1) 0, RF_update R SP (sp - 2 - 2 - 2), fl, isr).
    apply nextS_call.
    unfold pushW.
    rewrite HSP.
    split.
      apply stkSpace_weaken with (10 - 2 - 2 ); [idtac | omega].
      sep on.
      sep lift 2 in HM.
      sep trivial.
    split_r.
    apply MemUpdW_MemUpd16 with 0; trivial.
  split.
    (* g' => fun_frame *)
    unfold sp_call_pred, fun_frame.
    unfold linkPG, conjPG.
    intros S' S'' HS'S''.
    destruct HS'S'' as [H HS'S''].
    destruct H as [S [HS HSS']].
    unfold p_deQueue in HS'S''.
    destruct HS as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk HS]]]]]]].
    set (mQ := (embed_RdyQ T RdyQ)
        *^* (stkSpacePrec (sp -2-2-2-2) (_UNBLK_Stk -2-2-2-2))
        *^* (sp -2-2-2 |->w (L_unblock +1+1+1+1+1))
        *^* (sp - 2 |->w bp)
        *^* (sp + 2 |-> l_ptr) 
        *^* (sp + 3 |-> jk)).
    assert (HS' : Ast_deQueue_pre T BlkQ l_ptr (S_RF S' BP) (sp-2-2-2)
      (mQ *^* mTrue) S').
      eapply Hpp'; eauto.
    clear Hpp' HSS' Hpp'''.
    substH HS'S'' with (HS'S'' _ _ _ _ _ _ HS') into HS''.
    unfold Ast_unblock_4 in HS.
    destruct HS as [_ [HBX [HSP [HBP [Hsp [HRT [HBT [HRB HT]]]]]]]].
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct S'' as [[[M'' R''] fl''] isr''].
    simplS in * .
    unfold Ast_deQueue_pre, Ast_deQueue_post_b, Ast_deQueue_post_a in * .
    destruct HS' as [HM' [HBP' HSP']].
    simplS in * .
    split.
      rewrite HSP'.
      destruct HS'' as [HS'' | HS''].
        destruct HS'' as [_ HS''].
        destruct HS'' as [_ [_ [HBP'' HSP'']]].
        rewrite HSP''; trivial.
      destruct HS'' as [Q' [tid' [tcb' [HS'' _]]]].
      destruct HS'' as [_ [_ [HBP'' [HSP'' _]]]].
      rewrite HSP''; trivial.
    unfold Ret_Addr.
    exists (L_unblock +1+1+1+1+1).
    simplS.
    split.
      intros pc''.
      apply nextPC_retw with (sp -2-2-2); trivial.
      destruct HS'' as [HS'' | HS''].
        destruct HS'' as [_ HS''].
        destruct HS'' as [HM'' _].
        sep on.
        subst mQ.
        sep lift 5 in HM'.
        sep trivial.
      destruct HS'' as [Q' [tid' [tcb' [HS'' _]]]].
      destruct HS'' as [HM'' _].
        sep on.
        subst mQ.
        sep lift 5 in HM'.
        sep trivial.
    intros pc''.
    clear HM'.
    destruct HS'' as [HS'' | HS''].
      destruct HS'' as [_ HS''].
      destruct HS'' as [HM'' [HAX'' [HBP'' HSP'']]].
      apply nextPC_retw with (sp -2-2-2); trivial.
      sep on.
      subst mQ.
      sep lift 5 in HM''.
      sep trivial.
    destruct HS'' as [Q' [tid' [tcb' [HS'' _]]]].
    destruct HS'' as [HM'' [HAX'' [HBP'' [HSP'' _]]]].
    apply nextPC_retw with (sp -2-2-2); trivial.
    sep on.
    subst mQ.
    sep lift 6 in HM''.
    sep trivial.
  split.
    (* p => p' *)
    intros S' H.
    unfold conjPG, conjG, linkPG in H.
    destruct H as [S [HS HnextSS']].
    unfold p_unblock_4 in HS.
    unfold p_deQueue.
    destruct HS as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk HS]]]]]]].
    exists T; exists BlkQ; exists l_ptr; exists (S_RF S' BP); 
      exists (sp -2-2-2).
    set (mQ := (embed_RdyQ T RdyQ)
        *^* (stkSpacePrec (sp -2-2-2-2) (_UNBLK_Stk -2-2-2-2))
        *^* (sp -2-2-2 |->w (L_unblock +1+1+1+1+1))
        *^* (sp - 2 |->w bp)
        *^* (sp + 2 |-> l_ptr) 
        *^* (sp + 3 |-> jk)).
    assert (HS' : Ast_deQueue_pre T BlkQ l_ptr (S_RF S' BP) (sp-2-2-2)
      (mQ *^* mTrue) S').
      eapply Hpp'; eauto.
    clear Hpp''' Hpp' HS HnextSS'.
    unfold Ast_deQueue_pre in * .
    destruct HS' as [HM' H].
    split_r.
    sep on.
    sep matchH HM' with 10.
  split.
    (* p => p''' *)
    intros pc' S'''.
    unfold linkPG, wrap_fun, conjG.
    intros H.
    destruct H as [S H].
    destruct H as [HS H].
    destruct H as [S'' [H HS''S''']].
    destruct H as [S' [HSS' HS'S'']].
    unfold p_unblock_4, p_unblock_5 in * .
    destruct HS as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk HS]]]]]]].
    exists T; exists RdyQ; exists BlkQ; exists l_ptr.
    exists bp; exists sp; exists jk.
    eapply Hpp'''; eauto.
  (* g'' => g *)
  intros pc' S S'''' H.
  unfold conjPG, wrap_fun, conjG, linkPG in H.
  destruct H as [S''' [HS HS'''S'''']].
  destruct HS as [HSx [S'' [H HS'']]].
  destruct H as [S' [HSS' HS'S'']].
  unfold g_unblock_4, g_unblock_5 in * .
  intros mP T RdyQ BlkQ l_ptr bp sp jk HS.
  substH HS'''S'''' with (HS'''S'''' mP T RdyQ BlkQ l_ptr bp sp jk).
  assert (H : Ast_unblock_5_a T RdyQ BlkQ l_ptr bp sp jk mP S''' \/
           (exists BlkQ' : WQ,
              exists tid' : TID,
                exists tcb' : Thread,
                  Ast_unblock_5_b T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk
                    mP S''')).
    eapply Hpp'''; eauto.
  clear Hpp' Hpp'''.
  destruct HS'''S'''' as [H1 H2].
  destruct H as [HS''' | [BlkQ' [tid' [tcb' HS''']]]].
    left; auto.
  right.
  substH H2 with (H2 BlkQ' tid' tcb' HS''').
  destruct H2 as [T' [RdyQ' H2]].
  exists T'; exists RdyQ'; exists BlkQ'; exists tid'; exists tcb'.
  trivial.
Qed.

(* <unblock+5> :        addw    $2, %sp *)

Lemma SCAP_WFcmd_5_addw :
  SCAP_WFcmd (SCAP_consCT (L_unblock +1+1+1+1+1+1) 
    (Ast p_unblock_6 g_unblock_6 sctag) psi_unblock)
     (L_unblock +1+1+1+1+1) (Ast p_unblock_5 g_unblock_5 sctag)
     (comml (_addw   (Opr_imm 2 ) SP)).
Proof.
  assert (Hpp'a : forall T RdyQ BlkQ l_ptr bp sp jk mP S S',
    Ast_unblock_5_a T RdyQ BlkQ l_ptr bp sp jk mP S
    -> nextS (L_unblock+1+1+1+1+1) (comml (_addw   (Opr_imm 2 ) SP)) S S'
    -> Ast_unblock_6_a T RdyQ BlkQ l_ptr bp sp jk mP S').
    intros T RdyQ BlkQ l_ptr bp sp jk mP S S' HS HSS'.
    unfold Ast_unblock_5_a, Ast_unblock_6_a in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    simplS in * .
    destruct HS as [HM [HAX [HSP [HBP [Hsp HQ]]]]].
    subst M' R' isr'.
    split.
      sep on.
      sep matchH HM with 10.
      sep stksplit.
      sep matchH HM with 2.
    reg simpl.
    unfold opr_eval.
    rewrite HSP.
    split_l.
    split.
      generalize Hsp; clearall; omega.
    repeat split_l; trivial.
  assert (Hpp'b : forall T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S S',
    Ast_unblock_5_b T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S
    -> nextS (L_unblock+1+1+1+1+1) (comml (_addw   (Opr_imm 2 ) SP)) S S'
    -> Ast_unblock_6_b T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S').
    intros T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S S' HS HSS'.
    unfold Ast_unblock_5_b, Ast_unblock_6_b in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    simplS in * .
    destruct HS as [HM [HA [HSP [HBP [Htid' [Hsp [HBT [Hmto' [Hin HBlkQ]]]]]]]]].
    subst M' R' isr'.
    split.
      sep on.
      sep matchH HM with 10.
      generalize Hsp HM; clearall; intros.
      sep stksplit.
      sep matchH HM with 4.
    reg simpl.
    unfold opr_eval.
    split_l.
    split.
      rewrite HSP.
      generalize Hsp; clearall; intros.
      omega.
    repeat split_l; trivial.
  simpl.
  exists p_unblock_6; exists g_unblock_6.
  split_l.
  split.
    (* enable addw *)
    apply trivial_enable_addw.
  split. (* p ==> p' *)
    unfold p_unblock_5, p_unblock_6.
    intros S' H.
    unfold linkPG in H.
    destruct H as [S H].
    simpl in H.
    destruct H as [H HnextS].
    destruct H as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk H]]]]]]].
    exists T; exists RdyQ; exists BlkQ; exists l_ptr;
      exists bp; exists sp; exists jk.
    destruct H.
      left; apply Hpp'a with S; trivial.
    right.
    destruct H as [BlkQ' [tid' [tcb' HS]]].
    exists BlkQ'; exists tid'; exists tcb'.
    apply Hpp'b with S; trivial.
  (* gimp *)
  intros S S'' H.
  unfold conjPG, conjG in H.
  destruct H as [S' [[HSx HnextS] HS'S'']].
  unfold g_unblock_5, g_unblock_6 in * .
  intros mP T RdyQ BlkQ l_ptr bp sp jk.
  substH HS'S'' with (HS'S'' mP T RdyQ BlkQ l_ptr bp sp jk).
  destruct HS'S'' as [HS'S''a HS'S''b].
  split.
    intros HS.
    eapply HS'S''a; eauto.
  intros BlkQ' tid' tcb' HS.
  eapply HS'S''b; eauto.
Qed.

(* <unblock+6> :        cmpw    $0, %ax *)

Lemma SCAP_WFcmd_6_cmpw :
  SCAP_WFcmd (SCAP_consCT (L_unblock +1+1+1+1+1+1+1) 
    (Ast p_unblock_7 g_unblock_7 sctag) psi_unblock)
    (L_unblock +1+1+1+1+1+1) (Ast p_unblock_6 g_unblock_6 sctag)
     (comml (_cmpw   (Opr_imm 0 ) AX)).
Proof.
  assert (Hpp'a : forall T RdyQ BlkQ l_ptr bp sp jk mP S S',
    Ast_unblock_6_a T RdyQ BlkQ l_ptr bp sp jk mP S
    -> nextS (L_unblock+1+1+1+1+1+1) (comml (_cmpw   (Opr_imm 0 ) AX)) S S'
    -> Ast_unblock_7_a T RdyQ BlkQ l_ptr bp sp jk mP S').
    intros T RdyQ BlkQ l_ptr bp sp jk mP S S' HS HSS'.
    unfold Ast_unblock_6_a, Ast_unblock_7_a in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    simplS in * .
    destruct HS as [HM [HAX [HSP [HBP [Hsp HQ]]]]].
    subst M' R' isr'.
    split.
      sep on.
      sep matchH HM with 10.
    destruct Hfl' as [Hfl' | Hfl']; simpl in Hfl'.
      destruct Hfl' as [_ Hfl'].
      subst fl'.
      unfold Set_Flag.
      simpl.
      repeat split_l; trivial.
    destruct Hfl' as [Hfl' _].
    rewrite HAX in Hfl'.
    destruct (Hfl' (refl_equal _)).
  assert (Hpp'b : forall T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S S',
    Ast_unblock_6_b T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S
    -> nextS (L_unblock+1+1+1+1+1+1) (comml (_cmpw   (Opr_imm 0 ) AX)) S S'
    -> Ast_unblock_7_b T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S').
    intros T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S S' HS HSS'.
    unfold Ast_unblock_6_b, Ast_unblock_7_b in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    simplS in * .
    destruct HS as [HM [HAX [HSP [HBP [Hsp [Htid' [HBT [Hmto' [Hin HBlkQ]]]]]]]]].
    subst M' R' isr'.
    assert (Hzf' : fl' _ZF = false).
      simpl in Hfl'.
      destruct Hfl'.
      destruct H.
      rewrite HAX in H.
      rewrite <- H in Htid'.
      destruct (Htid' (refl_equal _)).
      destruct H; subst fl'.
      simpl; trivial.
    split.
      sep on.
      sep matchH HM with 10.
    repeat split_l; trivial.
  simpl.
  exists p_unblock_7; exists g_unblock_7.
  split_l.
  split.
    (* enable addw *)
    apply trivial_enable_cmpw.
  split. (* p ==> p' *)
    unfold p_unblock_6, p_unblock_7.
    intros S' H.
    unfold linkPG in H.
    destruct H as [S H].
    simpl in H.
    destruct H as [H HnextS].
    destruct H as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk H]]]]]]].
    exists T; exists RdyQ; exists BlkQ; exists l_ptr;
      exists bp; exists sp; exists jk.
    destruct H.
      left; apply Hpp'a with S; trivial.
    right.
    destruct H as [BlkQ' [tid' [tcb' HS]]].
    exists BlkQ'; exists tid'; exists tcb'.
    apply Hpp'b with S; trivial.
  (* gimp *)
  intros S S'' H.
  unfold conjPG, conjG in H.
  destruct H as [S' [[HSx HnextS] HS'S'']].
  unfold g_unblock_6, g_unblock_7 in * .
  intros mP T RdyQ BlkQ l_ptr bp sp jk.
  substH HS'S'' with (HS'S'' mP T RdyQ BlkQ l_ptr bp sp jk).
  destruct HS'S'' as [HS'S''a HS'S''b].
  split.
    intros HS.
    eapply HS'S''a; eauto.
  intros BlkQ' tid' tcb' HS.
  eapply HS'S''b; eauto.
Qed.

(* <unblock+7> :   jne     wakeup *)

Lemma SCAP_WFcmd_7_jne :
  SCAP_WFcmd (SCAP_consCT (L_unblock +1+1+1+1+1+1+1+1) 
    (Ast p_unblock_8 g_unblock_8 sctag) psi_unblock)
    (L_unblock +1+1+1+1+1+1+1) (Ast p_unblock_7 g_unblock_7 sctag)
     (comml (_jne    L_wakeup)).
Proof.
  assert (Hpp' : forall T RdyQ BlkQ l_ptr bp sp jk mP S S',
    Ast_unblock_7_a T RdyQ BlkQ l_ptr bp sp jk mP S
    -> nextS (L_unblock+1+1+1+1+1+1+1) (comml (_jne L_wakeup)) S S'
    -> Ast_unblock_8 T RdyQ BlkQ l_ptr bp sp jk mP S').
    intros T RdyQ BlkQ l_ptr bp sp jk mP S S' HS HSS'.
    unfold Ast_unblock_7_a, Ast_unblock_8 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    simplS in * .
    destruct HS as [HM [HZF [HAX [HSP [HBP [Hsp HQ]]]]]].
    subst M' R' isr'.
    repeat split_l; trivial.
  assert (Hpp'' : forall T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S S',
    Ast_unblock_7_b T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S
    -> nextS (L_unblock+1+1+1+1+1+1) (comml (_jne L_wakeup)) S S'
    -> Ast_unblock_10 T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S').
    intros T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S S' HS HSS'.
    unfold Ast_unblock_7_b, Ast_unblock_10 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    simplS in * .
    destruct HS as [HM [HZF [HAX [HSP [HBP [Hsp [Htid' [HBT [Hmto' [Hni HBlkQ']]]]]]]]]].
    subst M' R' isr'.
    repeat split_l; trivial.
  simpl.
  exists p_unblock_10; exists p_unblock_8.
  exists g_unblock_10; exists g_unblock_8.
  split_l.
    unfold SCAP_lookupCT, SCAP_consCT.
    apply SPSpec.lookup_merge_intro_r.
    apply L_wakeup_in_psi_unblock.
  split_l.
  split. (* p ==> p'' *)
    unfold p_unblock_7, p_unblock_8.
    intros S' H.
    unfold conjP, linkPG in H.
    destruct H as [H HZF].
    destruct H as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk H]]]]]]].
    exists T; exists RdyQ; exists BlkQ; exists l_ptr;
      exists bp; exists sp; exists jk.
    unfold ZF_cl in HZF.
    destruct H.
      unfold Ast_unblock_7_a in H.
      destruct H as [_ [HZF' _]].
      rewrite HZF' in HZF.
      discriminate.
    destruct H as [BlkQ' [tid' [tcb' HS']]].
    exists BlkQ'; exists tid'; exists tcb'.
    apply Hpp'' with S'; trivial.
    destruct S' as [[[M' R'] fl'] isr'].
    apply nextS_jne.
  split. (* g'' ==> g *)
    unfold g_unblock_10, g_unblock_7.
    intros S S'' H.
    unfold conjPG, conjG in H.
    destruct H as [HSx [HZF H]].
    intros mP T RdyQ BlkQ l_ptr bp sp jk.
    split.
      intros HS.
      unfold Ast_unblock_7_a in HS.
      unfold ZF_cl in HZF.
      destruct HS as [HM [HZF' [HAX [HSP [HBP [Hsp HQ]]]]]].
      rewrite HZF in HZF'.
      discriminate.
    intros BlkQ' tid' tcb' HS.
    eapply H; eauto.
    eapply Hpp''; eauto.
    destruct S as [[[M R] fl] isr].
    apply nextS_jne.
  split. (* p ==> p' *)
    unfold p_unblock_7, p_unblock_8.
    intros S' H.
    unfold conjP, linkPG in H.
    destruct H as [H HZF].
    destruct H as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk H]]]]]]].
    exists T; exists RdyQ; exists BlkQ; exists l_ptr;
      exists bp; exists sp; exists jk.
    unfold ZF_cl in HZF.
    destruct H.
      apply Hpp' with S'; trivial.
      destruct S' as [[[M' R'] fl'] isr'].
      apply nextS_jne.
    destruct H as [BlkQ' [tid' [tcb' HS']]].
    unfold Ast_unblock_7_b in HS'.
    destruct HS' as [_ [HZF' _]].
    unfold ZF_st in HZF.
    rewrite HZF' in HZF.
    discriminate.
  (* g' ==> g  *)
  unfold g_unblock_8, g_unblock_7.
  intros S S'' H.
  unfold conjPG, conjG in H.
  destruct H as [HSx [HZF H]].
  intros mP T RdyQ BlkQ l_ptr bp sp jk.
  split.
    intros HS.
    apply H.
    eapply Hpp'; eauto.
    destruct S as [[[M R] fl] isr].
    apply nextS_jne.
  intros BlkQ' tid' tcb' HS.
  unfold Ast_unblock_7_b in HS.
  destruct HS as [_ [HZF' _]].
  unfold ZF_st in HZF.
  rewrite HZF' in HZF.
  discriminate.
Qed.

(* <unblock+8> :        popw    %bp *)

Lemma SCAP_WFcmd_8_popw :
  SCAP_WFcmd (SCAP_consCT (L_unblock +1+1+1+1+1+1+1+1+1) 
    (Ast p_unblock_9 g_unblock_9 sctag) psi_unblock)
     (L_unblock +1+1+1+1+1+1+1+1) (Ast p_unblock_8 g_unblock_8 sctag)
     (comml (_popw   BP)).
Proof.
  assert (Hpp' : forall T RdyQ BlkQ l_ptr bp sp jk mP S S',
    Ast_unblock_8 T RdyQ BlkQ l_ptr bp sp jk mP S
    -> nextS (L_unblock+1+1+1+1+1+1+1+1) (comml (_popw BP)) S S'
    -> Ast_unblock_9 T RdyQ BlkQ l_ptr bp sp jk mP S').
    intros T RdyQ BlkQ l_ptr bp sp jk mP S S' HS HSS'.
    unfold Ast_unblock_8, Ast_unblock_9 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HR' as [bpx [Hlds HR']].
    simplS in * .
    destruct HS as [HM [HAX [HSP [HBP [Hsp HQ]]]]].
    assert (Heq : bp = bpx).
      sep on.
      sep lift 3 in HM.
      rewrite HSP in Hlds.
      sep inj.
      auto.
    subst bpx.
    subst M' fl' R' isr'.
    split.
      sep on.
      sep matchH HM with 12.
      generalize Hsp HM; clearall; intros.
      sep stksplit.
      sep matchH HM with 2.
    rewrite HSP.
    reg simpl.
    split_l.
    split.
      generalize Hsp; clearall; omega.
    auto.
  simpl.
  exists p_unblock_9; exists g_unblock_9.
  split_l.
  split.
  (* enable *)
    apply enable_popw.
    intros S HS.
    destruct S as [[[M R] fl] isr].
    unfold p_unblock_8 in HS.
    destruct HS as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk HS]]]]]]].
    unfold Ast_unblock_8 in HS.
    destruct HS as [HM [HAX [HSP [HBP [Hsp HQ]]]]].
    simplS in * .
    rewrite HSP.
    sep on.
    sep lift 3 in HM.
    sep trivial.
  split. (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold p_unblock_8, p_unblock_9.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk HS]]]]]]].
    exists T; exists RdyQ; exists BlkQ; exists l_ptr; 
      exists bp; exists sp; exists jk.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HSx HnextSS'] HS'S''].
  unfold g_unblock_8, g_unblock_9 in * .
  intros mP T RdyQ BlkQ l_ptr bp sp jk HS.
  eapply HS'S''; eauto.
Qed.

(* <unblock+9> :        retw                            # block_done *)

Lemma SCAP_WFcmd_9_retw :
  SCAP_WFcmd psi_unblock
    (L_unblock +1+1+1+1+1+1+1+1+1) (Ast p_unblock_9 g_unblock_9 sctag)
     (comms (_retw)).
Proof.
  unfold p_unblock_9, g_unblock_9.
  simpl.
  intros S H.
  intros.
  unfold Ast_unblock_9, Ast_unblock_post_a in * .
  InvertAll.
  repeat split_l.
  trivial.
Qed.

(* <unblock+10> :        movw    %ax, %bx *)

Lemma SCAP_WFcmd_10_movw :
  SCAP_WFcmd (SCAP_consCT (L_unblock +1+1+1+1+1+1+1+1+1+1+1) 
    (Ast p_unblock_11 g_unblock_11 sctag) psi_unblock)
     (L_unblock +1+1+1+1+1+1+1+1+1+1) (Ast p_unblock_10 g_unblock_10 sctag)
     (comml (_movwp  (Opr_reg AX) BX)).
Proof.
  assert (Hpp' : forall T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S S',
    Ast_unblock_10 T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S
    -> nextS (L_unblock +1+1+1+1 +1+1+1+1 +1+1) (comml (_movwp  (Opr_reg AX) BX)) S S'
    -> Ast_unblock_11 T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S').
    intros T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S S' HS HSS'.
    unfold Ast_unblock_10, Ast_unblock_11 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    simplS in * .
    destruct HS as [HM [HAX [HSP [HBP [Hsp [Htid' [HBT [Hmto [Hin HBlkQ']]]]]]]]].
    subst M' R' fl' isr'.
    simpl.
    reg simpl.
    rewrite HAX; rewrite HSP; rewrite HBP.
    split.
      simpl in HM.
      sep on.
      sep matchH HM with 8.
    repeat split_l; trivial.
  simpl.
  exists p_unblock_11; exists g_unblock_11.
  split_l.
  split.
    (* enable movw *)
    unfold enable, p_unblock_10.
    intros S H.
    destruct H as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk [BlkQ' [tid' [tcb' HS]]]]]]]]]].
    unfold Ast_unblock_10 in HS.
    destruct HS as [HM [HAX [HSP [HBP [Hsp [Htid' [Hmto [Hin HBlkQ']]]]]]]].
    destruct S as [[[M R] fl] isr].
    simplS in * .
    set (R' := RF_update R BX tid').
    exists (M, R', fl, isr).
    apply nextS_movwp with tid'; auto.
  split. (* p ==> p' *)
    unfold p_unblock_10, p_unblock_11.
    intros S' H.
    unfold linkPG in H.
    destruct H as [S H].
    destruct H as [H HnextS].
    destruct H as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk [BlkQ' [tid' [tcb' HS]]]]]]]]]].
    exists T; exists RdyQ; exists BlkQ; exists l_ptr.
    exists bp; exists sp; exists jk; exists BlkQ'.
    exists tid'; exists tcb'.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold conjPG, conjG in H.
  destruct H as [S' [[_ HnextS] HS'S'']].
  unfold g_unblock_10, g_unblock_11 in * .
  intros mP T RdyQ BlkQ l_ptr bp sp jk BlkQ' tid' tcb' HS.
  eapply HS'S''; eauto.
Qed.

(* <unblock+11> :         movw    $READY, TCB_STATUS(%bx) *)

Lemma SCAP_WFcmd_11_movwst :
  SCAP_WFcmd (SCAP_consCT (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1+1) 
    (Ast p_unblock_12 g_unblock_12 sctag) psi_unblock)
     (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1) (Ast p_unblock_11 g_unblock_11 sctag)
     (comml (_movwst (Opr_imm 1) (A_reg TCB_STATUS BX))).
Proof.
  assert (Hpp' : forall T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S S',
    Ast_unblock_11 T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S
    -> nextS (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1)
      (comml (_movwst (Opr_imm 1) (A_reg TCB_STATUS BX))) S S'
    -> Ast_unblock_12 T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S').
    intros T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S S' HS HSS'.
    unfold Ast_unblock_11, Ast_unblock_12 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hlds HM'].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simpl in Hlds.
    simplS in * .
    destruct HS as [HM [HAX [HBX [HSP [HBP [Hsp [Htid' [HBT [Hmto [Hin HBlkQ']]]]]]]]]].
    subst M' R' fl' isr'.
    split.
      sep on.
      unfold TCB_STATUS.
      unfold embed_TCB in * .
      rewrite HBX.
      sep lift 4 in HM.
      sep lift 4.
      apply ptoTCB_intro1 with 1.
        substH HM with (ptoTCB_elim1 HM).
        destruct HM as [tst [HM Htst]].
        sep lift 1 in HM.
        sep lift 1.
        arith_rep (tid' + 4 + 2) with (tid' + 6).
        arith_rep (tid' + 4 + 2) with (tid' + 6) in HM.
        sep upd.
        destruct tcb' as [[spx statusx] pcx].
        unfold TCB_SP, TCB_Status_Upd in * .
        sep matchH HM with 8.
      destruct tcb' as [[spx statusx] pcx].
      unfold TCB_SP, TCB_Status, TCB_Status_Upd in * .
      simpl; trivial.
    repeat split_l; trivial.
  simpl.
  exists p_unblock_12; exists g_unblock_12.
  split_l.
  split.
    (* enable movw *)
    apply enable_movwst.
    intros S HS.
    unfold p_unblock_11 in HS.
    destruct HS as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk [BlkQ' [tid' [tcb' HS]]]]]]]]]].
    unfold Ast_unblock_11 in HS.
    simpl.
    destruct HS as [HM [HAX [HBX [HSP [HBP [Hsp [Htid' [Hmto [Hin HBlkQ']]]]]]]]].
    destruct S as [[[M R] fl] isr].
    simplS in * .
    set (M' := MemUpdW M (tid' + 4 + 2) 1 0).
    unfold TCB_STATUS.
    rewrite HBX.
    unfold embed_TCB in HM.
    sep on.
    sep lift 4 in HM.
    substH HM with (ptoTCB_elim1 HM).
    destruct HM as [tst [HM Htst]].
    sep lift 1 in HM.
    arith_rep (tid' + 4 + 2) with (tid' + 6) in HM.
    sep trivial.
  split. (* p ==> p' *)
    unfold p_unblock_11, p_unblock_12.
    intros S' H.
    unfold linkPG in H.
    destruct H as [S H].
    destruct H as [H HnextS].
    destruct H as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk [BlkQ' [tid' [tcb' HS]]]]]]]]]].
    exists T; exists RdyQ; exists BlkQ; exists l_ptr.
    exists bp; exists sp; exists jk; exists BlkQ'; exists tid'; exists tcb'.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold conjPG, conjG in H.
  destruct H as [S' [[_ HnextS] HSS']].
  unfold g_unblock_11, g_unblock_12 in * .
  intros mP T RdyQ BlkQ l_ptr bp sp jk BlkQ' tid' tcb' HS.
  eapply HSS'; eauto.
Qed.

(* <unblock+12> :    pushw   $readyQ *)


Lemma SCAP_WFcmd_12_pushwp :
  SCAP_WFcmd (SCAP_consCT (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1+1 +1) 
    (Ast p_unblock_13 g_unblock_13 sctag) psi_unblock)
     (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1+1) (Ast p_unblock_12 g_unblock_12 sctag)
     (comml (_pushwp (Opr_imm ready_Q))).
Proof.
  assert (Hpp' : forall T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S S',
    Ast_unblock_12 T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S
    -> nextS (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1+1) (comml (_pushwp (Opr_imm ready_Q))) S S'
    -> Ast_unblock_13 T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S').
    intros T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S S' HS HSS'.
    unfold Ast_unblock_12, Ast_unblock_13 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspgt [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simplS in * .
    destruct HS as [HM [HBX [HSP [HBP [Hsp [Htid' [HBT [Hmto [Hin HBlkQ']]]]]]]]].
    subst M' R' fl' isr'.
    reg simpl.
    rewrite HSP.
    rewrite HBP.
    rewrite HBX.
    split.
      sep on.
      sep lift 4.
      sep upd.
      sep matchH HM with 13.
      sep stksplit in HM.
      trivial.
    repeat split_l; trivial.
  simpl.
  exists p_unblock_13; exists g_unblock_13.
  split_l.
  split.
  (* enable *)
    apply enable_pushwp with 8; simpl.
    intros S HS.
    unfold p_unblock_12 in HS.
    destruct HS as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk [BlkQ' [tid' [tcb' HS]]]]]]]]]].
    unfold Ast_unblock_12 in HS.
    destruct S as [[[M R] fl] isr].
    destruct HS as [HM [HBX [HSP [HBP [Hsp [Htid' [HBT [Hmto [Hin HBlkQ']]]]]]]]].
    simplS in * .
    rewrite HSP.
    sep on.
    sep lift 3 in HM.
    simpl in HM.
    sep trivial.
    clearall; omega.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold p_unblock_12, p_unblock_13.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk [BlkQ' [tid' [tcb' HS]]]]]]]]]].
    exists T; exists RdyQ; exists BlkQ; exists l_ptr; exists bp; exists sp;
      exists jk; exists BlkQ'; exists tid'; exists tcb'.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HS'S''].
  unfold g_unblock_12, g_unblock_13 in * .
  intros mP T RdyQ BlkQ l_ptr bp sp jk BlkQ' tid' tcb' HS.
  eapply HS'S''; eauto.
Qed.

(* <unblock+13> :        pushw   %bx *)

Lemma SCAP_WFcmd_13_pushwp :
  SCAP_WFcmd (SCAP_consCT (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1+1 +1+1) 
    (Ast p_unblock_14 g_unblock_14 sctag) psi_unblock)
     (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1+1 +1) (Ast p_unblock_13 g_unblock_13 sctag)
     (comml (_pushwp (Opr_reg BX))).
Proof.
  assert (Hpp' : forall T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S S',
    Ast_unblock_13 T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S
    -> nextS (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1+1 +1) (comml (_pushwp (Opr_reg BX))) S S'
    -> Ast_unblock_14 T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S').
    intros T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S S' HS HSS'.
    unfold Ast_unblock_13, Ast_unblock_14 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspgt [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simplS in * .
    destruct HS as [HM [HBX [HSP [HBP [Hsp [Htid' [HBT [Hmto [Hin HBlkQ']]]]]]]]].
    subst M' R' fl' isr'.
    reg simpl.
    rewrite HSP.
    rewrite HBP.
    rewrite HBX.
    split.
      sep on.
      sep lift 4.
      sep upd.
      sep matchH HM with 13.
      sep stksplit in HM.
      trivial.
    repeat split_l; trivial.
  simpl.
  exists p_unblock_14; exists g_unblock_14.
  split_l.
  split.
  (* enable *)
    apply enable_pushwp with 6; simpl.
    intros S HS.
    unfold p_unblock_13 in HS.
    destruct HS as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk [BlkQ' [tid' [tcb' HS]]]]]]]]]].
    unfold Ast_unblock_13 in HS.
    destruct S as [[[M R] fl] isr].
    destruct HS as [HM [HBX [HSP [HBP [Hsp [Htid' [HBT [Hmto [Hin HBlkQ']]]]]]]]].
    simplS in * .
    rewrite HSP.
    sep on.
    sep lift 3 in HM.
    simpl in HM.
    sep trivial.
    clearall; omega.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold p_unblock_13, p_unblock_14.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk [BlkQ' [tid' [tcb' HS]]]]]]]]]].
    exists T; exists RdyQ; exists BlkQ; exists l_ptr; exists bp; exists sp;
      exists jk; exists BlkQ'; exists tid'; exists tcb'.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HS'S''].
  unfold g_unblock_13, g_unblock_14 in * .
  intros mP T RdyQ BlkQ l_ptr bp sp jk BlkQ' tid' tcb' HS.
  eapply HS'S''; eauto.
Qed.

(* <unblock+14> :        call    enQueue *)

Lemma SCAP_WFcmd_14_call :
  SCAP_WFcmd (SCAP_consCT (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1+1 +1+1+1) 
    (Ast p_unblock_15 g_unblock_15 sctag) psi_unblock)
     (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1+1 +1+1) (Ast p_unblock_14 g_unblock_14 sctag)
     (comml (_call    L_enQueue)).
Proof.
  assert(Hpp' : forall T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mQ mP S S',
    Ast_unblock_14 T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S
    -> nextS (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1+1 +1+1) (comml (_call L_enQueue)) S S'
    -> mQ = ((embed_BlkQ l_ptr T BlkQ')
        *^* (sp - 2 - 2 - 2 - 2 |->w (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1+1 +1+1+1))
        *^* (sp - 2 |->w bp)
        *^* (sp + 2 |-> l_ptr)
        *^* (sp + 3 |-> jk))
    -> Ast_enQueue_pre T RdyQ ready_Q tid' (TCB_Status_Upd tcb' READY) (sp - 2)
    (sp -2-2-2-2) (mQ *^* mP) S').
    intros T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mQ mP S S'.
    intros HS HSS' HmQ.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    unfold Ast_unblock_14 in HS.
    destruct HS as [HM [HBX [HSP [HBP [Hsp [Htid' [HBT [Hmto [Hin HBlkQ']]]]]]]]].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspx [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    subst M' R' fl' isr'.
    unfold Ast_enQueue_pre.
    simplS in * .
    reg simpl.
    rewrite HSP.
    split.
      sep on.
      assert (H1 : sp - 2 - 2 - 2 - 2 + 2 = sp - 2 - 2 - 2).
        assert (H2 : sp - 2 - 2 - 2 = sp - 6).
          generalize Hsp; clearall; omega.
        rewrite H2.
        generalize Hsp; clearall; omega.
      rewrite H1.
      clear H1.
      assert (H1 : sp - 2 - 2 - 2 - 2 + 4 = sp - 2 - 2).
        assert (H2 : sp - 2 - 2 = sp - 4).
          generalize Hsp; clearall; omega.
        rewrite H2; clear H2.
        assert (H2 : sp - 4 - 2 = sp - 6).
          generalize Hsp; clearall; omega.
        rewrite H2; clear H2.
        assert (H2 : sp - 6 - 2 = sp - 8).
          generalize Hsp; clearall; omega.
        rewrite H2; clear H2.
        generalize Hsp; clearall; omega.
      rewrite H1; clear H1.
      subst mQ.
      sep lift 6.
      sep upd.
      unfold embed_RdyQ in HM.
      sep matchH HM with 10.
      sep stksplit in HM.
      simpl in HM.
      trivial.
    split.
      generalize Htid'; clearall. 
      unfold TID in * ; omega.
    split_lr.
  assert (Hpp''' : forall T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk pc' mP S S' S'' S''',
    Ast_unblock_14 T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S
    -> nextS (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1+1 +1+1) (comml (_call L_enQueue)) S S'
    -> g_enQueue S' S''
    -> nextS pc' (comms (_retw)) S'' S'''
    -> exists T', exists RdyQ',
      Ast_unblock_15 T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S''').
    intros T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk pc' mP.
    intros S S' S'' S''' HS HSS' HS'S'' HS''S'''.
    exists (TQMod.put T tid' (TCB_Status_Upd tcb' READY)).
    exists (addThrd RdyQ tid').
    unfold g_enQueue in HS'S''.
    set (mQ := (embed_BlkQ l_ptr T BlkQ')
        *^* (sp - 2 - 2 - 2 - 2 |->w (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1+1 +1+1+1))
        *^* (sp - 2 |->w bp)
        *^* (sp + 2 |-> l_ptr)
        *^* (sp + 3 |-> jk)).
    assert (HS' : Ast_enQueue_pre T RdyQ ready_Q tid' (TCB_Status_Upd tcb' READY) (sp - 2)
      (sp -2-2-2-2) (mQ *^* mP) S').
      eapply Hpp'; eauto.
    clear Hpp'.
    substH HS'S'' with (HS'S'' _ _ _ _ _ _ _ _ HS') into HS''.
    destruct HS'' as [T' [RdyQ' [HS'' [HT' HRdyQ']]]].
    subst mQ.
    unfold Ast_enQueue_post in HS''.
    unfold Ast_unblock_15.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct S'' as [[[M'' R''] fl''] isr''].
    destruct S''' as [[[M''' R'''] fl'''] isr'''].
    unfold Ast_unblock_14 in HS.
    destruct HS as [HM [HBX [HSP [HBP [Hsp [Htid' [HBT [Hmto [Hin HBlkQ']]]]]]]]]. 
    simplS in * .
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct (nextS_inversion HS''S''') as [HM''' [HR''' [Hfl''' Hisr''']]].
    destruct HR''' as [pcx [Hlds HR''']].
    destruct HS'' as [HM'' [HBP'' HSP'']].
    assert (Hpcx : pcx = L_unblock +1+1+1+1 +1+1+1+1 +1+1+1+1 +1+1+1).
      subst R' M''' R'''.
      clear HBP HSP Hsp HM HS' HS''S''' HM'.
      clear HT' HRdyQ' HSS'.
      subst fl' isr' fl''' isr'''.
      rewrite HSP'' in Hlds; clear HSP''.
      sep on.
      sep lift 5 in HM''.
      sep_inj; auto.
    subst pcx; clear Hlds.
    subst M''' R''' fl''' isr'''.
    reg simpl.
    rewrite HSP''.
    assert (H1 : sp - 2 - 2 - 2 - 2 + 2 = sp - 2 - 2 - 2).
      assert (H2 : sp - 2 - 2 - 2 = sp - 6).
        generalize Hsp; clearall; omega.
      rewrite H2; clear H2.
      generalize Hsp; clearall; omega.
    rewrite H1; rewrite H1 in HM''; clear H1.
    assert (H1 : sp - 2 - 2 - 2 - 2 + 4 = sp - 2 - 2).
      assert (H2 : sp - 2 - 2 = sp - 4).
        generalize Hsp; clearall; omega.
      rewrite H2; clear H2.
      assert (H2 : sp - 4 - 2 = sp - 6).
        generalize Hsp; clearall; omega.
      rewrite H2; clear H2.
      assert (H2 : sp - 6 - 2 = sp - 8).
        generalize Hsp; clearall; omega.
      rewrite H2; clear H2.
      generalize Hsp; clearall; omega.
    rewrite H1 in HM''; clear H1.
    split.
      unfold embed_RdyQ in * .
      unfold embed_BlkQ in *.
      rewrite <- HRdyQ'.
      rewrite <- HT'.
      rewrite (filterTQ_prev2 _ _ _ _ _ _ HBT HBlkQ' HT').
      sep on.
      sep matchH HM'' with 9.
      generalize Hsp HM''; clearall.
      sep stksplit.
      simpl.
      sep matchH HM'' with 2.
    repeat split_l; trivial.
  simpl.
  exists p_enQueue; exists p_unblock_15.
  exists g_enQueue; exists g_unblock_15.
  split.
    unfold SCAP_lookupCT, SCAP_consCT.
    apply SPSpec.lookup_merge_intro_r.
    apply L_enQueue_in_psi_unblock.
  split_l.
  split.
    clear Hpp' Hpp'''.
    unfold enable, p_unblock_14.
    intros S HS.
    destruct HS as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk [BlkQ' [tid' [tcb' HS]]]]]]]]]].
    unfold Ast_unblock_14 in HS.
    destruct HS as [HM [HBX [HSP [HBP [Hsp _]]]]].
    (* enable call *)
    destruct S as [[[M R] fl] isr].
    simplS in * .
    exists (MemUpdW M (sp-2-2-2-2)
      (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1+1 +1+1+1) 0, RF_update R SP (sp-2-2-2-2), fl, isr).
    apply nextS_call.
    unfold pushW.
    rewrite HSP.
    split.
      apply stkSpace_weaken with 4; 
        [idtac | generalize Hsp; clearall; omega].
      sep on.
      sep lift 3 in HM.
      simpl in HM.
      sep trivial.
    split_r.
    apply MemUpdW_MemUpd16 with 0; trivial.
  split.
    (* g' => fun_frame *)
    clear Hpp'''.
    unfold sp_call_pred, fun_frame.
    unfold linkPG, conjPG.
    intros S' S'' HS'S''.
    destruct HS'S'' as [H HS'S''].
    destruct H as [S [HS HSS']].
    unfold g_enQueue in HS'S''.
    unfold p_unblock_14 in HS.
    destruct HS as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk [BlkQ' [tid' [tcb'  HS]]]]]]]]]].
    set (mQ := (embed_BlkQ l_ptr T BlkQ')
        *^* (sp - 2 - 2 - 2 - 2 |->w (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1+1 +1+1+1))
        *^* (sp - 2 |->w bp)
        *^* (sp + 2 |-> l_ptr)
        *^* (sp + 3 |-> jk)).
    assert (HS' : Ast_enQueue_pre T RdyQ ready_Q tid' (TCB_Status_Upd tcb' READY) (sp - 2)
      (sp -2-2-2-2) (mQ *^* mTrue) S').
      eapply Hpp'; eauto.
    clear Hpp' HS HSS'.
    substH HS'S'' with (HS'S'' _ _ _ _ _ _ _ _ HS').
    destruct HS'S'' as [T' [RdyQ' HS'']].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct S'' as [[[M'' R''] fl''] isr''].
    simplS in * .
    destruct HS'' as [HS'' _].
    unfold Ast_enQueue_pre, Ast_enQueue_post in * .
    destruct HS' as [HM' [Htid [HBP' HSP']]].
    destruct HS'' as [HM'' [HBP'' HSP'']].
    simplS in * .
    split.
      rewrite HSP'; rewrite HSP''; trivial.
    unfold Ret_Addr.
    exists (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1+1 +1+1+1).
    simplS.
    split.
      intros pc''.
      apply nextPC_retw with (sp-2-2-2-2); trivial.
      clear HM''.
      sep on.
      subst mQ.
      sep lift 6 in HM'.
      sep trivial.
    intros pc''.
    apply nextPC_retw with (sp-2-2-2-2); trivial.
    clear HM'.
    sep on.
    subst mQ.
    sep lift 5 in HM''.
    sep trivial.
  split.
    (* p => p' *)
    intros S' H.
    unfold conjPG, conjG, linkPG in H.
    destruct H as [S [HS HnextSS']].
    unfold p_unblock_14 in HS.
    unfold p_enQueue.
    destruct HS as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk [BlkQ' [tid' [tcb' HS]]]]]]]]]].
    exists T; exists RdyQ; exists ready_Q; exists tid'. 
    exists (TCB_Status_Upd tcb' READY); exists (sp - 2); exists (sp -2-2-2-2). 
    set (mQ := (embed_BlkQ l_ptr T BlkQ')
        *^* (sp - 2 - 2 - 2 - 2 |->w (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1+1 +1+1+1))
        *^* (sp - 2 |->w bp)
        *^* (sp + 2 |-> l_ptr)
        *^* (sp + 3 |-> jk)).
    assert (HS' : Ast_enQueue_pre T RdyQ ready_Q tid' (TCB_Status_Upd tcb' READY) (sp - 2)
      (sp -2-2-2-2) (mQ *^* mTrue) S').
      eapply Hpp'; eauto.
    unfold Ast_enQueue_pre in * .
    destruct HS' as [HM' [Htid' [HBP' HSP']]].
    rewrite HBP'; rewrite HSP'.
    split.
      sep on.
      sep matchH HM' with 10.
    auto.
  split.
    (* p => p''' *)
    intros pc' S'''.
    unfold linkPG, wrap_fun, conjG.
    intros H.
    destruct H as [S H].
    destruct H as [HS H].
    destruct H as [S'' [H HS''S''']].
    destruct H as [S' [HSS' HS'S'']].
    unfold p_unblock_14, p_unblock_15 in * .
    destruct HS as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk [BlkQ' [tid' [tcb' HS]]]]]]]]]].
    assert (H : exists T' : TQ,
             exists RdyQ' : WQ,
               Ast_unblock_15 T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp
                 sp jk mTrue S''').
      eapply Hpp'''; eauto.
    exists T; exists RdyQ; exists BlkQ; exists l_ptr.
    exists bp; exists sp; exists jk.
    destruct H as [T' [RdyQ' HS''']].
    exists T'; exists RdyQ'; exists BlkQ'; exists tid'; exists tcb'.
    trivial.
  (* g''' => g *)
  intros pc' S S'''' H.
  unfold conjPG, wrap_fun, conjG, linkPG in H.
  destruct H as [S''' [HS HS'''S'''']].
  destruct HS as [HSx [S'' [Hx HnextS'']]].
  destruct Hx as [S' [HSS' HS'S'']].
  unfold g_unblock_14, g_unblock_15 in * .
  intros mP T RdyQ BlkQ l_ptr bp sp jk BlkQ' tid' tcb' HS.
  assert (H : exists T' : TQ, exists RdyQ' : WQ,
    Ast_unblock_15 T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S''').
    eapply Hpp'''; eauto.
  destruct H as [T' [RdyQ' HS']].
  exists T'; exists RdyQ'; trivial.
  eapply HS'''S''''; eauto.
Qed.

(* <unblock+15> :       popw    %ax *)

Lemma SCAP_WFcmd_15_popw :
  SCAP_WFcmd (SCAP_consCT (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1+1 +1+1+1+1) 
    (Ast p_unblock_16 g_unblock_16 sctag) psi_unblock)
     (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1+1 +1+1+1) (Ast p_unblock_15 g_unblock_15 sctag)
     (comml (_popw   AX)).
Proof.
  assert (Hpp' : forall T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S S',
    Ast_unblock_15 T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S
    -> nextS (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1+1 +1+1+1)
      (comml (_popw   AX)) S S'
    -> Ast_unblock_16 T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S').
  intros  T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S S' HS HSS'.
  unfold Ast_unblock_15, Ast_unblock_16 in * .
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HR' as [x [HMx HR']].
  assert (x = tid').
  simpl in HS.
  destruct HS as [HSM [HSP HS]].
  sep on.
  sep lift 3 in HSM.
  rewrite HSP in HMx.
  sep inj; auto.
    destruct HS as [HM [HSP [Hsp [Htid' [Hmto [Hin [HBlkQ' [HT' HRdyQ']]]]]]]].
    subst M' R' fl' isr' x.
  simpl in * .
    split.
      sep on.
      sep matchH HM with 10.
      sep stksplit.
      arith_rep (6 - 2) with 4.
      sep matchH HM with 2.
    repeat split_l; trivial.
    rewrite HSP.
    assert (Ha1 : sp - 2 - 2 - 2 + 2 = sp - 2 - 2).
    generalize Hsp; clearall.
    omega.
    rewrite Ha1;clear Ha1.
    trivial.
  exists p_unblock_16; exists g_unblock_16.
  split_l.
  split.
   (* enable popw *)
    apply enable_popw.
    intros S HS.
    unfold p_unblock_15 in HS.
    destruct HS as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk [T' [RdyQ' [BlkQ' [tid' [tcb' HS]]]]]]]]]]]].
    unfold Ast_unblock_15 in HS.
    simpl.
    destruct HS as [HM [HSP [Hsp [Htid' [Hmto [Hin [HBlkQ' [HT' HRdyQ']]]]]]]].
    destruct S as [[[M R] fl] isr].
    simplS in * .
    rewrite HSP.
    sep on.
    sep lift 3 in HM.
    sep trivial.
  split. (* p ==> p' *)
    unfold p_unblock_15, p_unblock_16.
    intros S' H.
    unfold linkPG in H.
    destruct H as [S H].
    destruct H as [H HnextS].
    destruct H as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk [T' [RdyQ' [BlkQ' [tid' [tcb' HS]]]]]]]]]]]].
    exists T; exists RdyQ; exists BlkQ; exists l_ptr.
    exists bp; exists sp; exists jk. 
    exists T'; exists RdyQ'; exists BlkQ'; exists tid'; exists tcb'.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold conjPG, conjG in H.
  destruct H as [S' [[_ HnextS] HSS']].
  unfold g_unblock_15, g_unblock_16 in * .
  intros.
  eapply HSS'; eauto.
Qed.

(* <unblock+16> :        addw    $2, %sp *)

Lemma SCAP_WFcmd_16_addw :
  SCAP_WFcmd (SCAP_consCT (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1+1 +1+1+1+1 +1) 
    (Ast p_unblock_17 g_unblock_17 sctag) psi_unblock)
     (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1+1 +1+1+1+1) (Ast p_unblock_16 g_unblock_16 sctag)
     (comml (_addw (Opr_imm 2) SP)).
Proof.
  assert (Hpp' : forall T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S S',
    Ast_unblock_16 T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S
    -> nextS (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1+1 +1+1+1+1)
      (comml (_addw (Opr_imm 2) SP)) S S'
    -> Ast_unblock_17 T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S').
  intros  T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S S' HS HSS'.
  unfold Ast_unblock_16, Ast_unblock_17 in * .
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS as [HM [HAX [HSP [Hsp [Htid' [Hmto [Hin [HBlkQ' [HT' HRdyQ']]]]]]]]].
    subst M' R'  isr'.
  simpl in * .
    split.
      sep on.
      sep matchH HM with 10.
      sep stksplit.
      arith_rep (8 - 2) with 6.
      sep matchH HM with 2.
    repeat split_l; trivial.
    rewrite HSP.
    arith_rep (S (S (sp -2 -2))) with (sp-2).
  exists p_unblock_17; exists g_unblock_17.
  split_l.
  split.
   (* enable addw *)
    apply trivial_enable_addw.
  split. (* p ==> p' *)
    unfold p_unblock_16, p_unblock_17.
    intros S' H.
    unfold linkPG in H.
    destruct H as [S H].
    simpl in H.
    destruct H as [H HnextS].
    destruct H as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk [T' [RdyQ' [BlkQ' [tid' [tcb' H]]]]]]]]]]]] .
    exists T; exists RdyQ; exists BlkQ; exists l_ptr;
      exists bp; exists sp; exists jk.
    exists T'; exists RdyQ'; exists BlkQ'; exists tid'; exists tcb'.
    apply Hpp' with S; trivial.
  (* gimp *)
  intros S S'' H.
  unfold conjPG, conjG in H.
  destruct H as [S' [[HSx HnextS] HS'S'']].
  unfold g_unblock_16, g_unblock_17 in * .
  intros mP T RdyQ BlkQ l_ptr bp sp jk T' RdyQ' BlkQ' tid' tcb' HS.
  substH HS'S'' with (HS'S'' mP T RdyQ BlkQ l_ptr bp sp jk T' RdyQ'  BlkQ' tid' tcb').  
  eapply HS'S''; eauto.
Qed.

(* <unblock+17> :        popw    %bp *)

Lemma SCAP_WFcmd_17_popw :
  SCAP_WFcmd (SCAP_consCT (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1+1 +1+1+1+1 +1+1) 
    (Ast p_unblock_18 g_unblock_18 sctag) psi_unblock)
     (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1+1 +1+1+1+1 +1) (Ast p_unblock_17 g_unblock_17 sctag)
     (comml (_popw   BP)).
Proof.
  assert (Hpp' : forall T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S S',
    Ast_unblock_17 T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S
    -> nextS (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1+1 +1+1+1+1 +1)
      (comml (_popw   BP)) S S'
    -> Ast_unblock_18 T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S').
  intros  T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S S' HS HSS'.
  unfold Ast_unblock_17, Ast_unblock_18 in * .
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HR' as [x [HMx HR']].
  assert (x = bp).
  simpl in HS.
  destruct HS as [HSM [HAX [HSP HS]]].
  sep on.
  sep lift 3 in HSM.
  rewrite HSP in HMx.
  sep inj; auto.
    destruct HS as [HM [HAX [HSP [Hsp [Htid' [Hmto [Hin [HBlkQ' [HT' HRdyQ']]]]]]]]].
    subst M' R' fl' isr' x.
  simpl in * .
    split.
      sep on.
      sep matchH HM with 10.
      sep stksplit.
      arith_rep (10 - 2) with 8.
      sep matchH HM with 2.
    repeat split_l; trivial.
    rewrite HSP.
    arith_rep (sp -2 +2) with sp.
  exists p_unblock_18; exists g_unblock_18.
  split_l.
  split.
   (* enable popw *)
    apply enable_popw.
    intros S HS.
    unfold p_unblock_17 in HS.
    destruct HS as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk [T' [RdyQ' [BlkQ' [tid' [tcb' HS]]]]]]]]]]]].
    unfold Ast_unblock_17 in HS.
    simpl.
    destruct HS as [HM [HAX [HSP [Hsp [Htid' [Hmto [Hin [HBlkQ' [HT' HRdyQ']]]]]]]]].
    destruct S as [[[M R] fl] isr].
    simplS in * .
    rewrite HSP.
    sep on.
    sep lift 3 in HM.
    sep trivial.
  split. (* p ==> p' *)
    unfold p_unblock_17, p_unblock_18.
    intros S' H.
    unfold linkPG in H.
    destruct H as [S H].
    destruct H as [H HnextS].
    destruct H as [T [RdyQ [BlkQ [l_ptr [bp [sp [jk [T' [RdyQ' [BlkQ' [tid' [tcb' HS]]]]]]]]]]]].
    exists T; exists RdyQ; exists BlkQ; exists l_ptr.
    exists bp; exists sp; exists jk. 
    exists T'; exists RdyQ'; exists BlkQ'; exists tid'; exists tcb'.
    apply Hpp' with S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold conjPG, conjG in H.
  destruct H as [S' [[_ HnextS] HSS']].
  unfold g_unblock_17, g_unblock_18 in * .
  intros.
  eapply HSS'; eauto.
Qed.

(* <unblock+18> :        retw *)

Lemma SCAP_WFcmd_18_retw :
  SCAP_WFcmd psi_unblock
    (L_unblock +1+1+1+1 +1+1+1+1 +1+1+1+1 +1+1+1+1 +1+1) (Ast p_unblock_18 g_unblock_18 sctag)
     (comms (_retw)).
Proof.
  unfold p_unblock_18, g_unblock_18.
  simpl.
  intros S H.
  intros.
  unfold Ast_unblock_18, Ast_unblock_post_b in * .
  InvertAll.
  repeat split_l.
  trivial.
Qed.

Lemma SCAP_WFiseq_wakeup :
  SCAP_WFiseq psi_unblock L_wakeup (Ast p_unblock_10 g_unblock_10 sctag) I_wakeup.
Proof.
  unfold I_wakeup, L_wakeup.
  apply scap_isq_seq with p_unblock_11 g_unblock_11.
  apply SCAP_WFcmd_10_movw.
  apply scap_isq_seq with p_unblock_12 g_unblock_12. 
  apply SCAP_WFcmd_11_movwst.
  apply scap_isq_seq with p_unblock_13 g_unblock_13.
  apply SCAP_WFcmd_12_pushwp.
  apply scap_isq_seq with p_unblock_14 g_unblock_14.
  apply SCAP_WFcmd_13_pushwp.
  apply scap_isq_seq with p_unblock_15 g_unblock_15.
  apply SCAP_WFcmd_14_call.
  apply scap_isq_seq with p_unblock_16 g_unblock_16.
  apply SCAP_WFcmd_15_popw.
  apply scap_isq_seq with p_unblock_17 g_unblock_17.
  apply SCAP_WFcmd_16_addw.
  apply scap_isq_seq with p_unblock_18 g_unblock_18.
  apply SCAP_WFcmd_17_popw.
  apply scap_isq_end.
  apply SCAP_WFcmd_18_retw.
Qed.

Lemma WFiseq_unblock :
  SCAP_WFiseq psi_unblock L_unblock spec_unblock I_unblock.
Proof.
  unfold L_unblock, spec_unblock, I_unblock.
  apply scap_isq_seq with p_unblock_1 g_unblock_1.
  apply SCAP_WFcmd_0_pushbp.
  apply scap_isq_seq with p_unblock_2 g_unblock_2. 
  apply SCAP_WFcmd_1_movwp.
  apply scap_isq_seq with p_unblock_3 g_unblock_3.
  apply SCAP_WFcmd_2_movwld.
  apply scap_isq_seq with p_unblock_4 g_unblock_4.
  apply SCAP_WFcmd_3_pushwp.
  apply scap_isq_seq with p_unblock_5 g_unblock_5.
  apply SCAP_WFcmd_4_call.
  apply scap_isq_seq with p_unblock_6 g_unblock_6.
  apply SCAP_WFcmd_5_addw.
  apply scap_isq_seq with p_unblock_7 g_unblock_7.
  apply SCAP_WFcmd_6_cmpw.
  apply scap_isq_seq with p_unblock_8 g_unblock_8.
  apply SCAP_WFcmd_7_jne.
  apply scap_isq_seq with p_unblock_9 g_unblock_9.
  apply SCAP_WFcmd_8_popw.
  apply scap_isq_end.
  apply SCAP_WFcmd_9_retw.
Qed.
