(***************************************************************)
(* Coq code for AIM                                            *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

Add LoadPath "..".
Add LoadPath "../CAP0".
Add LoadPath "../SCAP".

Require Import Coqlib.

Require Import Arith.
Require Import Machine.
Require Import Assertion.
Require Import Assertion_Lemmas.
Require Import Assertion_Lemmas2.
Require Import Assertion_Lemmas4.
Require Import Assertion_Lemmas5.
Require Import Operational_Semantics.

Require Import SCAP.
Require Import AIM.
Require Import QLib.
Require Import Linking.

Require Import Kernel.

Require Export EnQueueSpec.
Require Export DeQueueSpec.

Definition Ast_unblock_pre (T : TQ) (RdyQ BlkQ : WQ) (l_ptr : Label)
  (bp sp jk : nat) (mP : MPred) (S : State) : Prop :=
  ((embed_RdyQ T RdyQ) 
    *^* (embed_BlkQ l_ptr T BlkQ) 
    *^* (stkSpacePrec sp _UNBLK_Stk) 
    *^* (sp + 2 |-> l_ptr) 
    *^* (sp + 3 |-> jk) 
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp
  /\ S_RF S BP = bp
  /\ sp >= _UNBLK_Stk
  /\ subsetQ_TQ RdyQ T
  /\ subsetQ_TQ BlkQ T
  /\ WQMod.disjoint RdyQ BlkQ
  /\ WF_TQ T.

Definition Ast_unblock_post_a (T : TQ) (RdyQ BlkQ : WQ) (l_ptr : Label)
  (bp sp jk : nat) (mP : MPred) (S' : State) : Prop :=
  ((embed_RdyQ T RdyQ) 
    *^* (embed_BlkQ l_ptr T BlkQ) 
    *^* (stkSpacePrec sp _UNBLK_Stk) 
    *^* (sp + 2 |-> l_ptr) 
    *^* (sp + 3 |-> jk)
    *^* mP) (S_Mem S')
  /\ S_RF S' SP = sp
  /\ S_RF S' BP = bp
  /\ S_RF S' AX = 0
  /\ (forall l, (filterTQ T BlkQ) l = None).

Definition Ast_unblock_post_b (T T' : TQ) (RdyQ BlkQ RdyQ' BlkQ' : WQ) (l_ptr : Label)
  (tid' : TID) (tcb' : Thread) (bp sp jk : nat) (mP : MPred) (S' : State) : Prop :=
  ((embed_RdyQ T' RdyQ') 
    *^* (embed_BlkQ l_ptr T' BlkQ') 
    *^* (stkSpacePrec sp _UNBLK_Stk) 
    *^* (sp + 2 |-> l_ptr) 
    *^* (sp + 3 |-> jk)
    *^* mP) (S_Mem S')
  /\ S_RF S' SP = sp
  /\ S_RF S' BP = bp
  /\ S_RF S' AX = tid'
  /\ tid' <> 0
  /\ TQMod.maps_to T tid' tcb'
  /\ T' = TQMod.put T tid' (TCB_Status_Upd tcb' READY)
  /\ WQMod.in_dom tid' BlkQ
  /\ BlkQ' = WQMod.remove BlkQ tid'
  /\ RdyQ' = addThrd RdyQ tid'.

Definition p_unblock (S : State) : Prop :=
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists bp, exists sp, exists jk,
      Ast_unblock_pre T RdyQ BlkQ l_ptr bp sp jk mTrue S.

Definition g_unblock (S S' : State) : Prop := forall mP,
  forall T RdyQ BlkQ l_ptr bp sp jk,
    Ast_unblock_pre T RdyQ BlkQ l_ptr bp sp jk mP S
    -> Ast_unblock_post_a T RdyQ BlkQ l_ptr bp sp jk mP S'
    \/ exists T', exists RdyQ', exists BlkQ', exists tid', exists tcb',
      Ast_unblock_post_b T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S'.

Lemma p_unblock_sound : forall S,
  Linking.UNBLK_Prd S -> p_unblock S.
Proof.
  unfold Linking.UNBLK_Prd.
  unfold p_unblock.
  intros S H.
  destruct H as [T [RdyQ [BlkQ [sp [l [[junk HM] [HRdyQ [HBlkQ [HSP [Hdisj HT]]]]]]]]]]. 
  exists T; exists RdyQ; exists BlkQ; exists l; exists (S_RF S BP).
  exists sp; exists junk.
  unfold Ast_unblock_pre.
  unfold UNBLK_MPre in HM.
  unfold embed_BlkQ.
  intuition.
  sep on.
  sep matchH HM with 10.
  sep on.
  sep lift 4 in HM.
  assert (S_Mem S ||- stkSpace sp 10).
  sep trivial.
  apply stkSpace_sp_ge_sz with (S_Mem S); assumption.
Qed.

Lemma g_unblock_sound : forall S S',
  g_unblock S S' -> Linking.UNBLK_grt S S'.
Proof.
  unfold g_unblock, Linking.UNBLK_grt.
  intros S S' H.
  intros M0 M1 HdisU T RdyQ BlkQ sp l junk HP.
  unfold UNBLK_Prd_aux in HP.
  unfold Ast_unblock_pre in H.
  destruct HP as [HM [HRdyQ [HBlkQ [Hsp [Hdisj HT]]]]].
  unfold UNBLK_MPre in HM.
  unfold embed_BlkQ in H.
  assert ((embed_RdyQ T RdyQ *^* ptoTQ l (filterTQ T BlkQ) *^*
       stkSpacePrec sp 10 *^* sp + 2 |-> l *^* 
       sp + 3 |-> junk *^* (mEq M1)) (S_Mem S)).
  exists M0; exists M1.
  unfold mEq.
  intuition.
  simpl in HM.
  sep on.
  sep matchH HM with 10.
  assert (Ast_unblock_post_a T RdyQ BlkQ l (S_RF S BP) sp junk (mEq M1) S' \/
      (exists T' : TQ,
         exists RdyQ' : WQ,
           exists BlkQ' : WQ,
             exists tid' : TID,
               exists tcb' : Thread,
                 Ast_unblock_post_b T T' RdyQ BlkQ RdyQ' BlkQ' l tid'
                   tcb' (S_RF S BP) sp junk (mEq M1) S')).
  apply H.
  intuition.
  sep on.
  sep lift 4 in HM.
  assert (S_Mem (S_Mem_Upd S M0) ||- stkSpace sp 10).
  sep trivial.
  apply stkSpace_sp_ge_sz with (S_Mem (S_Mem_Upd S M0)); assumption.
  clear H.
  destruct H1.
  unfold Ast_unblock_post_a in H.
  destruct H as [HM' [HSP [HBP [HAX HQ]]]].
  destruct HM' as [M0' [M2 [HdisjU' [HM0' HM2]]]].
  unfold mEq in HM2.
  subst M1.
  exists M0'.
  unfold UNBLK_Post_aux.
  intuition.
  left.
  unfold UNBLK_MPost.
  unfold UNBLK_MPre.
  simpl.
  split.
  apply TQMod.extensional_eqMap.
  intro x.
  substH HQ with (HQ x).
  rewrite HQ.
  unfold TQMod.emp.
  trivial.
  split.
  sep on.
  unfold embed_BlkQ in HM0'.
  sep matchH HM0' with 10.
  assumption.
  unfold Ast_unblock_post_b in H.
  destruct H as [T' [RdyQ' [BlkQ' [tid' [tcb' [HM' [HSP [HBP [HAX [Htid' [Hmaps [HT' [Hindom [HBlkQ' HRdyQ']]]]]]]]]]]]]].
  destruct HM' as [M0' [M2 [HdisjU' [HM0' HM2]]]].
   unfold mEq in HM2.
  subst M1.
  exists M0'.
  unfold UNBLK_Post_aux.
  intuition.
  right.
  exists tid'; exists tcb'.
  simpl.
  rewrite HAX.
  unfold TQMod.maps_to in Hmaps.
  intuition.
  unfold UNBLK_MPost.
  unfold UNBLK_MPre.
  subst T' BlkQ' RdyQ'.
  unfold embed_BlkQ in HM0'.
  destruct tcb' as ((tsp, tst), tpc).
  simpl in * .
  sep on.
  sep matchH HM0' with 10.
Qed.

(* <unblock + 0>:      pushw   %bp *)

Definition Ast_unblock_1 (T : TQ) (RdyQ BlkQ : WQ) (l_ptr : Label)
  (bp sp jk : nat) (mP : MPred) (S : State) : Prop :=
  ((embed_RdyQ T RdyQ) 
    *^* (embed_BlkQ l_ptr T BlkQ) 
    *^* (stkSpacePrec (sp - 2) (_UNBLK_Stk - 2)) 
    *^* (sp - 2 |->w bp) 
    *^* (sp + 2 |-> l_ptr) 
    *^* (sp + 3 |-> jk) 
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 2
  /\ S_RF S BP = bp
  /\ sp >= _UNBLK_Stk
  /\ subsetQ_TQ RdyQ T
  /\ subsetQ_TQ BlkQ T
  /\ WQMod.disjoint RdyQ BlkQ
  /\ WF_TQ T.

Definition p_unblock_1 (S : State) : Prop :=
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists bp, exists sp, exists jk,
      Ast_unblock_1 T RdyQ BlkQ l_ptr bp sp jk mTrue S.

Definition g_unblock_1 (S S' : State) : Prop := forall mP,
  forall T RdyQ BlkQ l_ptr bp sp jk,
    Ast_unblock_1 T RdyQ BlkQ l_ptr bp sp jk mP S
    -> Ast_unblock_post_a T RdyQ BlkQ l_ptr bp sp jk mP S'
    \/ exists T', exists RdyQ', exists BlkQ', exists tid', exists tcb',
      Ast_unblock_post_b T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S'.

(* <unblock+1>:        movw    %sp, %bp *)

Definition Ast_unblock_2 (T : TQ) (RdyQ BlkQ : WQ) (l_ptr : Label)
  (bp sp jk : nat) (mP : MPred) (S : State) : Prop :=
  ((embed_RdyQ T RdyQ) 
    *^* (embed_BlkQ l_ptr T BlkQ) 
    *^* (stkSpacePrec (sp - 2) (_UNBLK_Stk - 2)) 
    *^* (sp - 2 |->w bp) 
    *^* (sp + 2 |-> l_ptr) 
    *^* (sp + 3 |-> jk) 
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 2
  /\ S_RF S BP = sp - 2
  /\ sp >= _UNBLK_Stk
  /\ subsetQ_TQ RdyQ T
  /\ subsetQ_TQ BlkQ T
  /\ WQMod.disjoint RdyQ BlkQ
  /\ WF_TQ T.

Definition p_unblock_2 (S : State) : Prop :=
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists bp, exists sp, exists jk,
      Ast_unblock_2 T RdyQ BlkQ l_ptr bp sp jk mTrue S.

Definition g_unblock_2 (S S' : State) : Prop := forall mP,
  forall T RdyQ BlkQ l_ptr bp sp jk,
    Ast_unblock_2 T RdyQ BlkQ l_ptr bp sp jk mP S
    -> Ast_unblock_post_a T RdyQ BlkQ l_ptr bp sp jk mP S'
    \/ exists T', exists RdyQ', exists BlkQ', exists tid', exists tcb',
      Ast_unblock_post_b T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S'.

(* <unblock+2> :   movw    4(%bp), %bx             # the pointer to the block queue *)

Definition Ast_unblock_3 (T : TQ) (RdyQ BlkQ : WQ) (l_ptr : Label)
  (bp sp jk : nat) (mP : MPred) (S : State) : Prop :=
  ((embed_RdyQ T RdyQ) 
    *^* (embed_BlkQ l_ptr T BlkQ) 
    *^* (stkSpacePrec (sp - 2) (_UNBLK_Stk - 2)) 
    *^* (sp - 2 |->w bp) 
    *^* (sp + 2 |-> l_ptr) 
    *^* (sp + 3 |-> jk) 
    *^* mP) (S_Mem S)
  /\ S_RF S BX = l_ptr
  /\ S_RF S SP = sp - 2
  /\ S_RF S BP = sp - 2
  /\ sp >= _UNBLK_Stk
  /\ subsetQ_TQ RdyQ T
  /\ subsetQ_TQ BlkQ T
  /\ WQMod.disjoint RdyQ BlkQ
  /\ WF_TQ T.

Definition p_unblock_3 (S : State) : Prop :=
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists bp, exists sp, exists jk,
      Ast_unblock_3 T RdyQ BlkQ l_ptr bp sp jk mTrue S.

Definition g_unblock_3 (S S' : State) : Prop := forall mP,
  forall T RdyQ BlkQ l_ptr bp sp jk,
    Ast_unblock_3 T RdyQ BlkQ l_ptr bp sp jk mP S
    -> Ast_unblock_post_a T RdyQ BlkQ l_ptr bp sp jk mP S'
    \/ exists T', exists RdyQ', exists BlkQ', exists tid', exists tcb',
      Ast_unblock_post_b T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S'.

(* <unblock+3> :        pushw   %bx *)

Definition Ast_unblock_4 (T : TQ) (RdyQ BlkQ : WQ) (l_ptr : Label)
  (bp sp jk : nat) (mP : MPred) (S : State) : Prop :=
  ((embed_RdyQ T RdyQ) 
    *^* (embed_BlkQ l_ptr T BlkQ) 
    *^* (stkSpacePrec (sp - 2 - 2) (_UNBLK_Stk - 2 - 2)) 
    *^* (sp - 2 - 2 |->w l_ptr) 
    *^* (sp - 2 |->w bp) 
    *^* (sp + 2 |-> l_ptr) 
    *^* (sp + 3 |-> jk) 
    *^* mP) (S_Mem S)
  /\ S_RF S BX = l_ptr
  /\ S_RF S SP = sp - 2 - 2
  /\ S_RF S BP = sp - 2
  /\ sp >= _UNBLK_Stk
  /\ subsetQ_TQ RdyQ T
  /\ subsetQ_TQ BlkQ T
  /\ WQMod.disjoint RdyQ BlkQ
  /\ WF_TQ T.

Definition p_unblock_4 (S : State) : Prop :=
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists bp, exists sp, exists jk,
      Ast_unblock_4 T RdyQ BlkQ l_ptr bp sp jk mTrue S.

Definition g_unblock_4 (S S' : State) : Prop := forall mP,
  forall T RdyQ BlkQ l_ptr bp sp jk,
    Ast_unblock_4 T RdyQ BlkQ l_ptr bp sp jk mP S
    -> Ast_unblock_post_a T RdyQ BlkQ l_ptr bp sp jk mP S'
    \/ exists T', exists RdyQ', exists BlkQ', exists tid', exists tcb',
      Ast_unblock_post_b T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S'.

(* <unblock+4> :        call    deQueue                 # not change bp *)

Definition Ast_unblock_5_a (T : TQ) (RdyQ BlkQ : WQ) (l_ptr : Label)
  (bp sp jk : nat) (mP : MPred) (S : State) : Prop :=
  ((embed_RdyQ T RdyQ) 
    *^* (embed_BlkQ l_ptr T BlkQ) 
    *^* (stkSpacePrec (sp - 2 - 2) (_UNBLK_Stk - 2 - 2)) 
    *^* (sp - 2 - 2 |->w l_ptr) 
    *^* (sp - 2 |->w bp) 
    *^* (sp + 2 |-> l_ptr) 
    *^* (sp + 3 |-> jk) 
    *^* mP) (S_Mem S)
  /\ S_RF S AX = 0
  /\ S_RF S SP = sp - 2 - 2
  /\ S_RF S BP = sp - 2
  /\ sp >= _UNBLK_Stk
  /\ (forall l, (filterTQ T BlkQ) l = None).

Definition Ast_unblock_5_b (T : TQ) (RdyQ BlkQ BlkQ' : WQ) (l_ptr : Label)
  (tid' : TID) (tcb' : Thread) (bp sp jk : nat) (mP : MPred) (S : State) : Prop :=
  ((embed_RdyQ T RdyQ)
    *^* (embed_BlkQ l_ptr T BlkQ')
    *^* (embed_TCB tid' tcb')
    *^* (stkSpacePrec (sp - 2 - 2) (_UNBLK_Stk - 2 - 2)) 
    *^* (sp - 2 - 2 |->w l_ptr)
    *^* (sp - 2 |->w bp)
    *^* (sp + 2 |-> l_ptr)
    *^* (sp + 3 |-> jk) 
    *^* mP) (S_Mem S)
  /\ S_RF S AX = tid'
  /\ S_RF S SP = sp - 2 - 2
  /\ S_RF S BP = sp - 2
  /\ tid' <> 0
  /\ sp >= _UNBLK_Stk
  /\ subsetQ_TQ BlkQ T
  /\ TQMod.maps_to T tid' tcb' 
  /\ WQMod.in_dom tid' BlkQ
  /\ BlkQ' = WQMod.remove BlkQ tid'.

Definition p_unblock_5 (S : State) : Prop :=
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists bp, exists sp, exists jk,
      Ast_unblock_5_a T RdyQ BlkQ l_ptr bp sp jk mTrue S
      \/ (exists BlkQ', exists tid', exists tcb', 
        Ast_unblock_5_b T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mTrue S).

Definition g_unblock_5 (S S' : State) : Prop := forall mP,
  forall T RdyQ BlkQ l_ptr bp sp jk,
    (Ast_unblock_5_a T RdyQ BlkQ l_ptr bp sp jk mP S
    -> Ast_unblock_post_a T RdyQ BlkQ l_ptr bp sp jk mP S')
    /\ forall BlkQ' tid' tcb',
      (Ast_unblock_5_b T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S
        -> exists T', exists RdyQ', 
          Ast_unblock_post_b T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S').

(* <unblock+5> :        addw    $2, %sp *)

Definition Ast_unblock_6_a (T : TQ) (RdyQ BlkQ : WQ) (l_ptr : Label)
  (bp sp jk : nat) (mP : MPred) (S : State) : Prop :=
  ((embed_RdyQ T RdyQ) 
    *^* (embed_BlkQ l_ptr T BlkQ) 
    *^* (stkSpacePrec (sp - 2) (_UNBLK_Stk - 2)) 
    *^* (sp - 2 |->w bp) 
    *^* (sp + 2 |-> l_ptr) 
    *^* (sp + 3 |-> jk) 
    *^* mP) (S_Mem S)
  /\ S_RF S AX = 0
  /\ S_RF S SP = sp - 2
  /\ S_RF S BP = sp - 2
  /\ sp >= _UNBLK_Stk
  /\ (forall l, (filterTQ T BlkQ) l = None).

Definition Ast_unblock_6_b (T : TQ) (RdyQ BlkQ BlkQ' : WQ) (l_ptr : Label)
  (tid' : TID) (tcb' : Thread) (bp sp jk : nat) (mP : MPred) (S : State) : Prop :=
  ((embed_RdyQ T RdyQ)
    *^* (embed_BlkQ l_ptr T BlkQ')
    *^* (embed_TCB tid' tcb')
    *^* (stkSpacePrec (sp - 2) (_UNBLK_Stk - 2)) 
    *^* (sp - 2 |->w bp)
    *^* (sp + 2 |-> l_ptr)
    *^* (sp + 3 |-> jk) 
    *^* mP) (S_Mem S)
  /\ S_RF S AX = tid'
  /\ S_RF S SP = sp - 2
  /\ S_RF S BP = sp - 2
  /\ sp >= _UNBLK_Stk
  /\ tid' <> 0
  /\ subsetQ_TQ BlkQ T
  /\ TQMod.maps_to T tid' tcb' 
  /\ WQMod.in_dom tid' BlkQ
  /\ BlkQ' = WQMod.remove BlkQ tid'.

Definition p_unblock_6 (S : State) : Prop :=
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists bp, exists sp, exists jk,
      Ast_unblock_6_a T RdyQ BlkQ l_ptr bp sp jk mTrue S
      \/ (exists BlkQ', exists tid', exists tcb', 
        Ast_unblock_6_b T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mTrue S).

Definition g_unblock_6 (S S' : State) : Prop := forall mP,
  forall T RdyQ BlkQ l_ptr bp sp jk,
    (Ast_unblock_6_a T RdyQ BlkQ l_ptr bp sp jk mP S
    -> Ast_unblock_post_a T RdyQ BlkQ l_ptr bp sp jk mP S')
    /\ forall BlkQ' tid' tcb',
      (Ast_unblock_6_b T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S
        -> exists T', exists RdyQ', 
          Ast_unblock_post_b T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S').

(* <unblock+6> :        cmpw    $0, %ax *)

Definition Ast_unblock_7_a (T : TQ) (RdyQ BlkQ : WQ) (l_ptr : Label)
  (bp sp jk : nat) (mP : MPred) (S : State) : Prop :=
  ((embed_RdyQ T RdyQ) 
    *^* (embed_BlkQ l_ptr T BlkQ) 
    *^* (stkSpacePrec (sp - 2) (_UNBLK_Stk - 2)) 
    *^* (sp - 2 |->w bp) 
    *^* (sp + 2 |-> l_ptr) 
    *^* (sp + 3 |-> jk) 
    *^* mP) (S_Mem S)
  /\ S_Flag S _ZF = true
  /\ S_RF S AX = 0
  /\ S_RF S SP = sp - 2
  /\ S_RF S BP = sp - 2
  /\ sp >= _UNBLK_Stk
  /\ (forall l, (filterTQ T BlkQ) l = None).

Definition Ast_unblock_7_b (T : TQ) (RdyQ BlkQ BlkQ' : WQ) (l_ptr : Label)
  (tid' : TID) (tcb' : Thread) (bp sp jk : nat) (mP : MPred) (S : State) : Prop :=
  ((embed_RdyQ T RdyQ)
    *^* (embed_BlkQ l_ptr T BlkQ')
    *^* (embed_TCB tid' tcb')
    *^* (stkSpacePrec (sp - 2) (_UNBLK_Stk - 2)) 
    *^* (sp - 2 |->w bp)
    *^* (sp + 2 |-> l_ptr)
    *^* (sp + 3 |-> jk) 
    *^* mP) (S_Mem S)
  /\ S_Flag S _ZF = false
  /\ S_RF S AX = tid'
  /\ S_RF S SP = sp - 2
  /\ S_RF S BP = sp - 2
  /\ sp >= _UNBLK_Stk 
  /\ tid' <> 0
  /\ subsetQ_TQ BlkQ T
  /\ TQMod.maps_to T tid' tcb' 
  /\ WQMod.in_dom tid' BlkQ
  /\ BlkQ' = WQMod.remove BlkQ tid'.

Definition p_unblock_7 (S : State) : Prop :=
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists bp, exists sp, exists jk,
      Ast_unblock_7_a T RdyQ BlkQ l_ptr bp sp jk mTrue S
      \/ (exists BlkQ', exists tid', exists tcb', 
        Ast_unblock_7_b T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mTrue S).

Definition g_unblock_7 (S S' : State) : Prop := forall mP,
  forall T RdyQ BlkQ l_ptr bp sp jk,
    (Ast_unblock_7_a T RdyQ BlkQ l_ptr bp sp jk mP S
    -> Ast_unblock_post_a T RdyQ BlkQ l_ptr bp sp jk mP S')
    /\ forall BlkQ' tid' tcb',
      (Ast_unblock_7_b T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S
        -> exists T', exists RdyQ', 
          Ast_unblock_post_b T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S').

(* <unblock+7> :   jne     wakeup *)

Definition Ast_unblock_8 (T : TQ) (RdyQ BlkQ : WQ) (l_ptr : Label)
  (bp sp jk : nat) (mP : MPred) (S : State) : Prop :=
  ((embed_RdyQ T RdyQ) 
    *^* (embed_BlkQ l_ptr T BlkQ) 
    *^* (stkSpacePrec (sp - 2) (_UNBLK_Stk - 2)) 
    *^* (sp - 2 |->w bp) 
    *^* (sp + 2 |-> l_ptr) 
    *^* (sp + 3 |-> jk) 
    *^* mP) (S_Mem S)
  /\ S_RF S AX = 0
  /\ S_RF S SP = sp - 2
  /\ S_RF S BP = sp - 2
  /\ sp >= _UNBLK_Stk
  /\ (forall l, (filterTQ T BlkQ) l = None).

Definition p_unblock_8 (S : State) : Prop :=
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists bp, exists sp, exists jk,
      Ast_unblock_8 T RdyQ BlkQ l_ptr bp sp jk mTrue S.

Definition g_unblock_8 (S S' : State) : Prop := forall mP,
  forall T RdyQ BlkQ l_ptr bp sp jk,
    (Ast_unblock_8 T RdyQ BlkQ l_ptr bp sp jk mP S
    -> Ast_unblock_post_a T RdyQ BlkQ l_ptr bp sp jk mP S').

(* <unblock+8> :        popw    %bp *)

Definition Ast_unblock_9 (T : TQ) (RdyQ BlkQ : WQ) (l_ptr : Label)
  (bp sp jk : nat) (mP : MPred) (S : State) : Prop :=
  ((embed_RdyQ T RdyQ) 
    *^* (embed_BlkQ l_ptr T BlkQ) 
    *^* (stkSpacePrec sp _UNBLK_Stk) 
    *^* (sp + 2 |-> l_ptr) 
    *^* (sp + 3 |-> jk) 
    *^* mP) (S_Mem S)
  /\ S_RF S AX = 0
  /\ S_RF S SP = sp
  /\ S_RF S BP = bp
  /\ sp >= _UNBLK_Stk
  /\ (forall l, (filterTQ T BlkQ) l = None).

Definition p_unblock_9 (S : State) : Prop :=
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists bp, exists sp, exists jk,
      Ast_unblock_9 T RdyQ BlkQ l_ptr bp sp jk mTrue S.

Definition g_unblock_9 (S S' : State) : Prop := forall mP,
  forall T RdyQ BlkQ l_ptr bp sp jk,
    (Ast_unblock_9 T RdyQ BlkQ l_ptr bp sp jk mP S
    -> Ast_unblock_post_a T RdyQ BlkQ l_ptr bp sp jk mP S').

(* <unblock+9> :        retw                            # block_done *)

Definition Ast_unblock_10 (T : TQ) (RdyQ BlkQ BlkQ' : WQ) (l_ptr : Label)
  (tid' : TID) (tcb' : Thread) (bp sp jk : nat) (mP : MPred) (S : State) : Prop :=
  ((embed_RdyQ T RdyQ)
    *^* (embed_BlkQ l_ptr T BlkQ')
    *^* (embed_TCB tid' tcb')
    *^* (stkSpacePrec (sp - 2) (_UNBLK_Stk - 2)) 
    *^* (sp - 2 |->w bp)
    *^* (sp + 2 |-> l_ptr)
    *^* (sp + 3 |-> jk) 
    *^* mP) (S_Mem S)
  /\ S_RF S AX = tid'
  /\ S_RF S SP = sp - 2
  /\ S_RF S BP = sp - 2
  /\ sp >= _UNBLK_Stk
  /\ tid' <> 0
  /\ subsetQ_TQ BlkQ T
  /\ TQMod.maps_to T tid' tcb' 
  /\ WQMod.in_dom tid' BlkQ
  /\ BlkQ' = WQMod.remove BlkQ tid'.

Definition p_unblock_10 (S : State) : Prop :=
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists bp, exists sp, exists jk,
      exists BlkQ', exists tid', exists tcb', 
        Ast_unblock_10 T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mTrue S.

Definition g_unblock_10 (S S' : State) : Prop := forall mP,
  forall T RdyQ BlkQ l_ptr bp sp jk BlkQ' tid' tcb',
    (Ast_unblock_10 T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S
      -> exists T', exists RdyQ', 
        Ast_unblock_post_b T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S').

(* <unblock+10> :        movw    %ax, %bx *)

Definition Ast_unblock_11 (T : TQ) (RdyQ BlkQ BlkQ' : WQ) (l_ptr : Label)
  (tid' : TID) (tcb' : Thread) (bp sp jk : nat) (mP : MPred) (S : State) : Prop :=
  ((embed_RdyQ T RdyQ)
    *^* (embed_BlkQ l_ptr T BlkQ')
    *^* (embed_TCB tid' tcb')
    *^* (stkSpacePrec (sp - 2) (_UNBLK_Stk - 2)) 
    *^* (sp - 2 |->w bp)
    *^* (sp + 2 |-> l_ptr)
    *^* (sp + 3 |-> jk) 
    *^* mP) (S_Mem S)
  /\ S_RF S AX = tid'
  /\ S_RF S BX = tid'
  /\ S_RF S SP = sp - 2
  /\ S_RF S BP = sp - 2
  /\ sp >= _UNBLK_Stk
  /\ tid' <> 0
  /\ subsetQ_TQ BlkQ T
  /\ TQMod.maps_to T tid' tcb' 
  /\ WQMod.in_dom tid' BlkQ
  /\ BlkQ' = WQMod.remove BlkQ tid'.

Definition p_unblock_11 (S : State) : Prop :=
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists bp, exists sp, exists jk,
      exists BlkQ', exists tid', exists tcb', 
        Ast_unblock_11 T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mTrue S.

Definition g_unblock_11 (S S' : State) : Prop := forall mP,
  forall T RdyQ BlkQ l_ptr bp sp jk BlkQ' tid' tcb',
    (Ast_unblock_11 T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S
      -> exists T', exists RdyQ', 
        Ast_unblock_post_b T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S').

(* <unblock+11> :         movw    $READY, TCB_STATUS(%bx) *)

Definition Ast_unblock_12 (T : TQ) (RdyQ BlkQ BlkQ' : WQ) (l_ptr : Label)
  (tid' : TID) (tcb' : Thread) (bp sp jk : nat) (mP : MPred) (S : State) : Prop :=
  ((embed_RdyQ T RdyQ)
    *^* (embed_BlkQ l_ptr T BlkQ')
    *^* (embed_TCB tid' (TCB_Status_Upd tcb' READY))
    *^* (stkSpacePrec (sp - 2) (_UNBLK_Stk - 2)) 
    *^* (sp - 2 |->w bp)
    *^* (sp + 2 |-> l_ptr)
    *^* (sp + 3 |-> jk) 
    *^* mP) (S_Mem S)
  /\ S_RF S BX = tid'
  /\ S_RF S SP = sp - 2
  /\ S_RF S BP = sp - 2
  /\ sp >= _UNBLK_Stk 
  /\ tid' <> 0
  /\ subsetQ_TQ BlkQ T
  /\ TQMod.maps_to T tid' tcb' 
  /\ WQMod.in_dom tid' BlkQ
  /\ BlkQ' = WQMod.remove BlkQ tid'.

Definition p_unblock_12 (S : State) : Prop :=
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists bp, exists sp, exists jk,
      exists BlkQ', exists tid', exists tcb', 
        Ast_unblock_12 T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mTrue S.

Definition g_unblock_12 (S S' : State) : Prop := forall mP,
  forall T RdyQ BlkQ l_ptr bp sp jk BlkQ' tid' tcb',
    (Ast_unblock_12 T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S
      -> exists T', exists RdyQ', 
        Ast_unblock_post_b T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S').

(* <unblock+12> :    pushw   $readyQ *)

Definition Ast_unblock_13 (T : TQ) (RdyQ BlkQ BlkQ' : WQ) (l_ptr : Label)
  (tid' : TID) (tcb' : Thread) (bp sp jk : nat) (mP : MPred) (S : State) : Prop :=
  ((embed_RdyQ T RdyQ)
    *^* (embed_BlkQ l_ptr T BlkQ')
    *^* (embed_TCB tid' (TCB_Status_Upd tcb' READY))
    *^* (stkSpacePrec (sp - 2 - 2) (_UNBLK_Stk - 2 - 2)) 
    *^* (sp - 2 - 2 |->w ready_Q)
    *^* (sp - 2 |->w bp)
    *^* (sp + 2 |-> l_ptr)
    *^* (sp + 3 |-> jk) 
    *^* mP) (S_Mem S)
  /\ S_RF S BX = tid'
  /\ S_RF S SP = sp - 2 - 2
  /\ S_RF S BP = sp - 2
  /\ sp >= _UNBLK_Stk
  /\ tid' <> 0
  /\ subsetQ_TQ BlkQ T
  /\ TQMod.maps_to T tid' tcb' 
  /\ WQMod.in_dom tid' BlkQ
  /\ BlkQ' = WQMod.remove BlkQ tid'.

Definition p_unblock_13 (S : State) : Prop :=
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists bp, exists sp, exists jk,
      exists BlkQ', exists tid', exists tcb', 
        Ast_unblock_13 T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mTrue S.

Definition g_unblock_13 (S S' : State) : Prop := forall mP,
  forall T RdyQ BlkQ l_ptr bp sp jk BlkQ' tid' tcb',
    (Ast_unblock_13 T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S
      -> exists T', exists RdyQ', 
        Ast_unblock_post_b T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S').

(* <unblock+13> :        pushw   %bx *)

Definition Ast_unblock_14 (T : TQ) (RdyQ BlkQ BlkQ' : WQ) (l_ptr : Label)
  (tid' : TID) (tcb' : Thread) (bp sp jk : nat) (mP : MPred) (S : State) : Prop :=
  ((embed_RdyQ T RdyQ)
    *^* (embed_BlkQ l_ptr T BlkQ')
    *^* (embed_TCB tid' (TCB_Status_Upd tcb' READY))
    *^* (stkSpacePrec (sp - 2 - 2 - 2) (_UNBLK_Stk - 2 - 2 - 2)) 
    *^* (sp - 2 - 2 - 2 |->w tid')
    *^* (sp - 2 - 2 |->w ready_Q)
    *^* (sp - 2 |->w bp)
    *^* (sp + 2 |-> l_ptr)
    *^* (sp + 3 |-> jk) 
    *^* mP) (S_Mem S)
  /\ S_RF S BX = tid'
  /\ S_RF S SP = sp - 2 - 2 - 2
  /\ S_RF S BP = sp - 2
  /\ sp >= _UNBLK_Stk
  /\ tid' <> 0
  /\ subsetQ_TQ BlkQ T
  /\ TQMod.maps_to T tid' tcb' 
  /\ WQMod.in_dom tid' BlkQ
  /\ BlkQ' = WQMod.remove BlkQ tid'.

Definition p_unblock_14 (S : State) : Prop :=
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists bp, exists sp, exists jk,
      exists BlkQ', exists tid', exists tcb', 
        Ast_unblock_14 T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mTrue S.

Definition g_unblock_14 (S S' : State) : Prop := forall mP,
  forall T RdyQ BlkQ l_ptr bp sp jk BlkQ' tid' tcb',
    (Ast_unblock_14 T RdyQ BlkQ BlkQ' l_ptr tid' tcb' bp sp jk mP S
      -> exists T', exists RdyQ', 
        Ast_unblock_post_b T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S').

(* <unblock+14> :        call    enQueue *)

Definition Ast_unblock_15 (T T' : TQ) (RdyQ BlkQ RdyQ' BlkQ' : WQ) (l_ptr : Label)
  (tid' : TID) (tcb' : Thread) (bp sp jk : nat) (mP : MPred) (S : State) : Prop :=
  ((embed_RdyQ T' RdyQ')
    *^* (embed_BlkQ l_ptr T' BlkQ')
    *^* (stkSpacePrec (sp - 2 - 2 - 2) (_UNBLK_Stk - 2 - 2 - 2)) 
    *^* (sp - 2 - 2 - 2 |->w tid')
    *^* (sp - 2 - 2 |->w ready_Q)
    *^* (sp - 2 |->w bp)
    *^* (sp + 2 |-> l_ptr)
    *^* (sp + 3 |-> jk) 
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp - 2 - 2 - 2
  /\ sp >= _UNBLK_Stk
  /\ tid' <> 0
  /\ TQMod.maps_to T tid' tcb' 
  /\ WQMod.in_dom tid' BlkQ
  /\ BlkQ' = WQMod.remove BlkQ tid'
  /\ T' = TQMod.put T tid' (TCB_Status_Upd tcb' READY)
  /\ RdyQ' = addThrd RdyQ tid'.

Definition p_unblock_15 (S : State) : Prop :=
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists bp, exists sp, exists jk,
      exists T', exists RdyQ', exists BlkQ', exists tid', exists tcb', 
        Ast_unblock_15 T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mTrue S.

Definition g_unblock_15 (S S' : State) : Prop := forall mP,
  forall T RdyQ BlkQ l_ptr bp sp jk T' RdyQ' BlkQ' tid' tcb',
    (Ast_unblock_15 T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S
      -> Ast_unblock_post_b T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S').

(* <unblock+15> :       popw    %ax *)

Definition Ast_unblock_16 (T T' : TQ) (RdyQ BlkQ RdyQ' BlkQ' : WQ) (l_ptr : Label)
  (tid' : TID) (tcb' : Thread) (bp sp jk : nat) (mP : MPred) (S : State) : Prop :=
  ((embed_RdyQ T' RdyQ')
    *^* (embed_BlkQ l_ptr T' BlkQ')
    *^* (stkSpacePrec (sp - 2 - 2) (_UNBLK_Stk - 2 - 2)) 
    *^* (sp - 2 - 2 |->w ready_Q)
    *^* (sp - 2 |->w bp)
    *^* (sp + 2 |-> l_ptr)
    *^* (sp + 3 |-> jk) 
    *^* mP) (S_Mem S)
  /\ S_RF S AX = tid'
  /\ S_RF S SP = sp - 2 - 2
  /\ sp >= _UNBLK_Stk
  /\ tid' <> 0
  /\ TQMod.maps_to T tid' tcb' 
  /\ WQMod.in_dom tid' BlkQ
  /\ BlkQ' = WQMod.remove BlkQ tid'
  /\ T' = TQMod.put T tid' (TCB_Status_Upd tcb' READY)
  /\ RdyQ' = addThrd RdyQ tid'.

Definition p_unblock_16 (S : State) : Prop :=
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists bp, exists sp, exists jk,
      exists T', exists RdyQ', exists BlkQ', exists tid', exists tcb', 
        Ast_unblock_16 T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mTrue S.

Definition g_unblock_16 (S S' : State) : Prop := forall mP,
  forall T RdyQ BlkQ l_ptr bp sp jk T' RdyQ' BlkQ' tid' tcb',
    (Ast_unblock_16 T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S
      -> Ast_unblock_post_b T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S').

(* <unblock+16> :        addw    $2, %sp *)

Definition Ast_unblock_17 (T T' : TQ) (RdyQ BlkQ RdyQ' BlkQ' : WQ) (l_ptr : Label)
  (tid' : TID) (tcb' : Thread) (bp sp jk : nat) (mP : MPred) (S : State) : Prop :=
  ((embed_RdyQ T' RdyQ')
    *^* (embed_BlkQ l_ptr T' BlkQ')
    *^* (stkSpacePrec (sp - 2) (_UNBLK_Stk - 2))
    *^* (sp - 2 |->w bp)
    *^* (sp + 2 |-> l_ptr)
    *^* (sp + 3 |-> jk) 
    *^* mP) (S_Mem S)
  /\ S_RF S AX = tid'
  /\ S_RF S SP = sp - 2
  /\ sp >= _UNBLK_Stk
  /\ tid' <> 0
  /\ TQMod.maps_to T tid' tcb' 
  /\ WQMod.in_dom tid' BlkQ
  /\ BlkQ' = WQMod.remove BlkQ tid'
  /\ T' = TQMod.put T tid' (TCB_Status_Upd tcb' READY)
  /\ RdyQ' = addThrd RdyQ tid'.

Definition p_unblock_17 (S : State) : Prop :=
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists bp, exists sp, exists jk,
      exists T', exists RdyQ', exists BlkQ', exists tid', exists tcb', 
        Ast_unblock_17 T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mTrue S.

Definition g_unblock_17 (S S' : State) : Prop := forall mP,
  forall T RdyQ BlkQ l_ptr bp sp jk T' RdyQ' BlkQ' tid' tcb',
    (Ast_unblock_17 T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S
      -> Ast_unblock_post_b T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S').

(* <unblock+17> :        popw    %bp *)

Definition Ast_unblock_18 (T T' : TQ) (RdyQ BlkQ RdyQ' BlkQ' : WQ) (l_ptr : Label)
  (tid' : TID) (tcb' : Thread) (bp sp jk : nat) (mP : MPred) (S : State) : Prop :=
  ((embed_RdyQ T' RdyQ')
    *^* (embed_BlkQ l_ptr T' BlkQ')
    *^* (stkSpacePrec (sp) (_UNBLK_Stk))
    *^* (sp + 2 |-> l_ptr)
    *^* (sp + 3 |-> jk) 
    *^* mP) (S_Mem S)
  /\ S_RF S AX = tid'
  /\ S_RF S BP = bp
  /\ S_RF S SP = sp
  /\ sp >= _UNBLK_Stk
  /\ tid' <> 0
  /\ TQMod.maps_to T tid' tcb' 
  /\ WQMod.in_dom tid' BlkQ
  /\ BlkQ' = WQMod.remove BlkQ tid'
  /\ T' = TQMod.put T tid' (TCB_Status_Upd tcb' READY)
  /\ RdyQ' = addThrd RdyQ tid'.

Definition p_unblock_18 (S : State) : Prop :=
  exists T, exists RdyQ, exists BlkQ, exists l_ptr, 
    exists bp, exists sp, exists jk,
      exists T', exists RdyQ', exists BlkQ', exists tid', exists tcb', 
        Ast_unblock_18 T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mTrue S.

Definition g_unblock_18 (S S' : State) : Prop := forall mP,
  forall T RdyQ BlkQ l_ptr bp sp jk T' RdyQ' BlkQ' tid' tcb',
    (Ast_unblock_18 T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S
      -> Ast_unblock_post_b T T' RdyQ BlkQ RdyQ' BlkQ' l_ptr tid' tcb' bp sp jk mP S').

(* <unblock+18> :        retw *)


Definition spec_unblock := Ast p_unblock g_unblock sctag.

Definition psi_unblock  :=
  (SCAP_consCT L_enQueue (Ast p_enQueue g_enQueue sctag)
  (SCAP_consCT L_deQueue (Ast p_deQueue g_deQueue sctag)
  (SCAP_consCT L_unblock (Ast p_unblock g_unblock sctag)
  (SCAP_consCT L_wakeup (Ast p_unblock_10 g_unblock_10 sctag)
    SCAP_emptCT)))).

Lemma L_deQueue_in_psi_unblock : 
  SCAP_lookupCT psi_unblock L_deQueue (Ast p_deQueue g_deQueue sctag).
Proof.
  unfold SCAP_lookupCT, psi_unblock.
  unfold SCAP_consCT.
  apply SPSpec.lookup_merge_intro_r.
  simpl.
  SPSpec.lookup_solve.
Qed.

Lemma L_enQueue_in_psi_unblock : 
  SCAP_lookupCT psi_unblock L_enQueue (Ast p_enQueue g_enQueue sctag).
Proof.
  unfold SCAP_lookupCT, psi_unblock.
  unfold SCAP_consCT.
  simpl.
  SPSpec.lookup_solve.
Qed.

Lemma L_wakeup_in_psi_unblock : 
  SCAP_lookupCT psi_unblock L_wakeup (Ast p_unblock_10 g_unblock_10 sctag).
Proof.
  unfold SCAP_lookupCT, psi_unblock.
  unfold SCAP_consCT.
  apply SPSpec.lookup_merge_intro_r.
  simpl.
  unfold SPSpec.lookup.
  simpl.
  right. right. left; trivial.
Qed.
