(***************************************************************)
(* Coq code for AIM                                            *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

Add LoadPath "..".
Add LoadPath "../CAP0".
Add LoadPath "../SCAP".

Set Undo 400.

Require Import Coqlib.

Require Import Arith.
Require Import Machine.
Require Import Mem_Lemmas.
Require Import Assertion.
Require Import Assertion_Lemmas.
Require Import Operational_Semantics.
Require Import OpSem_Lemmas.

Require Import AIM.
Require Import AIM_Lemmas.
Require Import SCAP.
Require Import SCAP_READY.
Require Import QLib.
Require Import Linking.

Require Import Assertion_Lemmas4.
Require Import Assertion_Lemmas5.
Require Import Reg.

Require Import Kernel.

Require Import YieldSpec.

(********Source Code *************************** 

yield:
        cli
        pushw   %ax
        pushw   %bx
        pushw   %cx
        pushw   %dx
        pushw   %di
        pushw   %si
        pushw   %bp
        pushfw
        call    scheduler
        popfw
        popw    %bp
        popw    %si
        popw    %di
        popw    %dx
        popw    %cx
        popw    %bx
        popw    %ax
        sti
        retw

************************************************)


Lemma NH_WFcmd_yield_0_cli :
  NH_WFcmd (SCAP_consCT (L_yield+1) (Ast L_yield_1_p L_yield_1_g tag_yield) psi_yield)
    (L_yield) (Ast L_yield_0_p L_yield_0_g tag_yield) 
    (comml (_cli)).
Proof.
  simpl.
  split.
  exists L_yield_1_p; exists L_yield_1_g.
  split_l.
  assert (Hpp': forall sp pc mP S S', 
    Ast_yield_pre sp pc mP S
    -> gcli S S'
    -> Ast_yield_1 sp pc mP S').
    intros sp pc mP S S' HS HSS'.
    unfold linkPG, conjPG, conjG.
    unfold Ast_yield_pre, gcli, Ast_yield_1 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct HS as [HM [Hsp Hitr]].
    destruct HSS' as [Hshift [Hif' [Hisr' HR']]].
    assert (Hshift' : prepost Aemp (INV0 *^* INV1) (M, R, fl, isr) (M', R', fl', isr')).
      apply (prepost_weaken_post (mQ' := INV0 *^* INV1) Hshift).
      intros Mx H.
      apply (choice_elim_left H); trivial.
    split.
      sep on.
      assert (HMx : S_Mem (M, R, fl, isr) ||- 
        (Aemp) *^* (stkSpacePrec sp _STK_SIZE_yield *^* sp |->w pc *^* mP)).
        sep matchH HM with 4.
      substH HM with (prepost_elim Hshift' precise_emp HMx).
      sep on.
      sep matchH HM with 5. 
    split.
      simplS in * .
      subst R'; trivial.
    split.
      simplS in * .
      assert (stkSpace sp _STK_SIZE_yield M).
        sep on.
        sep lift 0 in HM.
        sep trivial.
      destruct H; trivial.
    split_l.
      simplS in * .
      subst isr'; destruct Hitr; trivial.
  split.
  (* p ==> p' *)
  unfold L_yield_0_p, L_yield_1_p.
  intros S' H.
  unfold linkPG in H.
  destruct H as [S [HS HSS']].
  destruct HS as [sp [pc HpS]].
  exists sp; exists pc.
  apply Hpp' with S; trivial.
  (* g' ==> g *)
  unfold L_yield_0_g, L_yield_1_g.
  unfold linkPG, conjPG, conjG.
  intros S S'' H.
  destruct H as [S' H].
  destruct H as [[HpS HSS'] HS'S''].
  intros mP sp pc HS.
  substH Hpp' with (Hpp' sp pc mP S S' HS HSS').
  substH HS'S'' with (HS'S'' mP sp pc Hpp').
  destruct HS'S'' as [HS'' HR''].
  split_l.
  unfold gcli in HSS'.
  destruct HSS' as [Hshift [Hif' [Hisr' HR']]].
  rewrite HR'.
  rewrite HR''.
  trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_yield_0_p in HS.
    destruct HS as [sp [pc HS]].
    unfold Ast_yield_pre in HS.
    destruct HS as [HM [HSP [Hif Hisr]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    rewrite HSP.
    assert (Hsp : sp >= _STK_SIZE_yield).
      assert (stkSpace sp _STK_SIZE_yield M).
      sep on.
      sep lift 0 in HM.
      sep trivial.
      destruct H; trivial.
    sep on.
    sep lift 0 in HM.
    sep stksplit (IH_Stk + 6) in HM.
    sep lift 0 in HM.
    destruct HM as [M1 [M2 [HJM [HM1 HM2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    split_l.
    intros M1' HM1'.
    unfold L_yield_0_p.
    exists sp; exists pc.
    unfold Ast_yield_pre.
    simplS in * .
    rsplit.
      repeat split_l; trivial.
    sep on.
    sep lift 0.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      sep off.
      split.
        disjoint rep_l M1.
        eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    split_l.
    sep on.
    sep matchH HM2 with 4.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_yield_0_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  apply HSS'.
  clear HSS'.
  unfold Ast_yield_pre in * .
  simplS in * .
  destruct HS as [HM [HSP [Hif Hisr]]].
  rsplit.
    repeat split_l; trivial.
  sep on.
  assert (Hsp : sp >= _STK_SIZE_yield).
    assert (stkSpace sp _STK_SIZE_yield (MemMod.merge M1 M2')).
    sep on.
    sep lift 0 in HM.
    sep trivial.
    destruct H; trivial.
  sep lift 0 in HM.
  sep stksplit (IH_Stk + 6) in HM.
  sep lift 0 in HM.
  destruct HM as [M2x [M1x [HJM12 [HM2x HM1x]]]].
  sep lift 0.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  assert (M2x = M2' /\ M1x = M1).
    sep off.
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    rewrite HSP in HM2'.
    assert (Hx := prec_disjUnion_inv H H0 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  rewrite HSP in HM2.
  split_l.
  sep on.
  sep matchH HM1x with 3.
Qed.

Lemma NH_WFcmd_yield_1_pushwp :
  NH_WFcmd (SCAP_consCT 
    (L_yield+1+1)
      (Ast L_yield_2_p L_yield_2_g tag_yield) psi_yield)
    (L_yield+1) 
      (Ast L_yield_1_p L_yield_1_g tag_yield) 
    (comml (_pushwp (Opr_reg AX))).
Proof.
  simpl.
  split.
  exists L_yield_2_p; exists L_yield_2_g.
  split_l.
  assert (Hpp' : forall sp pc mP S S',
    Ast_yield_1 sp pc mP S
    -> nextS (L_yield+1) (comml (_pushwp (Opr_reg AX))) S S'
    -> Ast_yield_2 sp pc (S_RF S AX) mP S').
    intros sp pc mP S S' HS HSS'.
    unfold Ast_yield_1, Ast_yield_2 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspgt [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simplS in * .
    destruct HS as [HM [Hsp [Hspgt' [Hif Hisr]]]].
    split.
      subst M' R' sp.
      sep on.
      sep lift 3.
      reg simpl.
      sep upd.
      assert (Hx : M ||-
        stkSpacePrec (R SP - 2) (_STK_SIZE_yield - 2) 
        *^* R SP - 2 |->w_
        *^* R SP |->w pc 
        *^* INV1 *^* INV0 *^* mP).
        assocr.
        apply stkSpacePrec_splitW1.
        arith_rep (R SP - 2 + 2) with (R SP).
        arith_rep (_STK_SIZE_yield - 2 + 2) with (_STK_SIZE_yield).
        sep matchH HM with 4.
      sep matchH Hx with 5.
      subst R' fl' isr' sp; reg simpl.
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_yield_1_p in HpS.
    destruct HpS as [sp [pc HS]]. 
    exists (MemUpdW M (sp-2) (R AX) 0, RF_update R SP (sp - 2), fl, isr).
    apply nextS_pushwp with  (R AX); trivial.
    unfold pushW.
    destruct HS as [HM [Hsp [Hspgt [Hif Hisr]]]].
    simplS in * .
    split.
      apply stkSpace_weaken with _STK_SIZE_yield;
        [idtac | omega].
      sep on.
      sep lift 2 in HM.
      subst sp.
      sep trivial.
    split.
      subst sp; eapply MemUpdW_MemUpd16; eauto.
    subst sp; trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_yield_1_p, L_yield_2_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [sp [pc HS]].
    exists sp; exists pc; exists (S_RF S AX).
    apply Hpp'; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_yield_1_g, L_yield_2_g in * .
  intros mP sp pc HS.
  substH Hpp' with (Hpp' _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [_ [Hsp _]].
  simplS in * .
  subst sp R' fl'.
  reg simpl.
  apply (extensionality reg nat R'' R).
  induction x; auto.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_yield_1_p in HS.
    destruct HS as [sp [pc HS]].
    unfold Ast_yield_1 in HS.
    destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_yield_1_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_yield_1 in HS.
  simplS in * .
  destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma NH_WFcmd_yield_2_pushwp :
  NH_WFcmd (SCAP_consCT 
    (L_yield+2+1)
      (Ast L_yield_3_p L_yield_3_g tag_yield) psi_yield)
    (L_yield+2) 
      (Ast L_yield_2_p L_yield_2_g tag_yield) 
    (comml (_pushwp (Opr_reg BX))).
Proof.
  simpl.
  split.
  exists L_yield_3_p; exists L_yield_3_g.
  split_l (* lookup *).
  assert (Hpp' : forall sp pc ax mP S S',
    Ast_yield_2 sp pc ax mP S
    -> nextS (L_yield+2) (comml (_pushwp (Opr_reg BX))) S S'
    -> Ast_yield_3 sp pc ax (S_RF S BX) mP S').
    intros sp pc ax mP S S' HS HSS'.
    unfold Ast_yield_2, Ast_yield_3 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspgt [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simplS in * .
    destruct HS as [HM [Hax [Hsp [Hspgt' [Hif Hisr]]]]].
    split.
      subst M' R'.
      sep on.
      sep lift 3.
      rewrite Hsp.
      arith_rep (sp - 2 -2) with (sp-4).
      sep upd.
      assert (Hx : M ||-
        stkSpacePrec (sp - 4) (_STK_SIZE_yield - 4)
        *^* sp - 4 |->w_
        *^* sp - 2 |->w ax
        *^* sp |->w pc 
        *^* INV1 *^* INV0 *^* mP).
        arith_rep (sp - 4) with (sp - 2 -2).
        arith_rep (_STK_SIZE_yield - 4) with
          (_STK_SIZE_yield - 2 - 2).
        assocr.
        apply stkSpacePrec_splitW1.
        subst ax.
        arith_rep (sp - 2 - 2 + 2) with (sp - 2).
        arith_rep (_STK_SIZE_yield - 2 - 2 + 2) with (_STK_SIZE_yield - 2).
        sep matchH HM with 5.
      sep matchH Hx with 6.
    subst R' fl' isr'; reg simpl.
    rewrite Hsp.
    split_l. 
    split_l by omega. 
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_yield_2_p in HpS.
    destruct HpS as [sp [pc [ax HS]]].
    exists (MemUpdW M (sp-4) (R BX) 0, RF_update R SP (sp - 4), fl, isr).
    apply nextS_pushwp with  (R BX); trivial.
    unfold pushW.
    destruct HS as [HM [Hax [Hsp [Hspgt [Hif Hisr]]]]].
    simplS in * .
    rewrite Hsp.
    arith_rep (sp-2-2) with (sp - 4).
    split.
      apply stkSpace_weaken with (_STK_SIZE_yield - 2);
        [idtac | omega].
      sep on.
      sep lift 2 in HM.
      sep trivial.
    split.
      apply MemUpdW_MemUpd16 with 0; trivial.
    trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_yield_2_p, L_yield_3_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [sp [pc [ax HS]]].
    exists sp; exists pc; exists ax; exists (S_RF S BX).
    apply Hpp'; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_yield_2_g, L_yield_3_g in * .
  intros mP sp pc ax HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [_ [Hax [Hsp _]]].
  simplS in * .
  subst ax sp R' fl'.
  reg simpl.
  repeat split_l.
  trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_yield_2_p in HS.
    destruct HS as [sp [pc [ax HS]]].
    unfold Ast_yield_2 in HS.
    destruct HS as [HM [HAX [HSP [Hsp [Hif Hisr]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_yield_2_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP sp pc ax HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_yield_2 in HS.
  simplS in * .
  destruct HS as [HM [HAX [HSP [Hsp [Hif Hisr]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma NH_WFcmd_yield_3_pushwp :
  NH_WFcmd (SCAP_consCT 
    (L_yield+3+1)
      (Ast L_yield_4_p L_yield_4_g tag_yield) psi_yield)
    (L_yield+3) 
      (Ast L_yield_3_p L_yield_3_g tag_yield) 
    (comml (_pushwp (Opr_reg CX))).
Proof.
  simpl.
  split.
  exists L_yield_4_p; exists L_yield_4_g.
  split_l (* lookup *).
  assert (Hpp' : forall sp pc ax bx mP S S',
    Ast_yield_3 sp pc ax bx mP S
    -> nextS (L_yield+3) (comml (_pushwp (Opr_reg CX))) S S'
    -> Ast_yield_4 sp pc ax bx (S_RF S CX) mP S').
    intros sp pc ax bx mP S S' HS HSS'.
    unfold Ast_yield_3, Ast_yield_4 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspgt [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simplS in * .
    destruct HS as [HM [Hbx [Hsp [Hspgt' [Hif Hisr]]]]].
    split.
      subst M' R'.
      sep on.
      sep lift 3.
      rewrite Hsp.
      arith_rep (sp - 4 -2) with (sp-6).
      sep upd.
      assert (Hx : M ||-
        stkSpacePrec (sp - 6) (_STK_SIZE_yield - 6)
        *^* sp - 6 |->w_
        *^* sp - 4 |->w bx
        *^* sp - 2 |->w ax
        *^* sp |->w pc 
        *^* INV1 *^* INV0 *^* mP).
        arith_rep (sp - 6) with (sp - 4 -2).
        arith_rep (_STK_SIZE_yield - 6) with
          (_STK_SIZE_yield - 4 - 2).
        assocr.
        apply stkSpacePrec_splitW1.
        subst bx.
        arith_rep (sp - 4 - 2 + 2) with (sp - 4).
        arith_rep (_STK_SIZE_yield - 4 - 2 + 2) with (_STK_SIZE_yield - 4).
        sep matchH HM with 5.
      sep matchH Hx with 7.
    subst R' fl' isr'; reg simpl.
    rewrite Hsp.
    split_l. 
    split_l by omega. 
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_yield_3_p in HpS.
    destruct HpS as [sp [pc [ax [bx HS]]]].
    exists (MemUpdW M (sp-6) (R CX) 0, RF_update R SP (sp - 6), fl, isr).
    apply nextS_pushwp with  (R CX); trivial.
    unfold pushW.
    destruct HS as [HM [Hbx [Hsp [Hspgt [Hif Hisr]]]]].
    simplS in * .
    rewrite Hsp.
    arith_rep (sp-4-2) with (sp - 6).
    split.
      apply stkSpace_weaken with (_STK_SIZE_yield - 4);
        [idtac | omega].
      sep on.
      sep lift 2 in HM.
      sep trivial.
    split.
      apply MemUpdW_MemUpd16 with 0; trivial.
    trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_yield_3_p, L_yield_4_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [sp [pc [ax [bx HS]]]].
    exists sp; exists pc; exists ax; exists bx; exists (S_RF S CX).
    apply Hpp'; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_yield_3_g, L_yield_4_g in * .
  intros mP sp pc ax bx HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [_ [Hax [Hsp _]]].
  simplS in * .
  subst ax bx sp R' fl'.
  reg simpl.
  repeat split_l.
  trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_yield_3_p in HS.
    destruct HS as [sp [pc [ax [bx HS]]]].
    unfold Ast_yield_3 in HS.
    destruct HS as [HM [HBX [HSP [Hsp [Hif Hisr]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_yield_3_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP sp pc ax bx HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_yield_3 in HS.
  simplS in * .
  destruct HS as [HM [HBX [HSP [Hsp [Hif Hisr]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma NH_WFcmd_yield_4_pushwp :
  NH_WFcmd (SCAP_consCT 
    (L_yield+4+1)
      (Ast L_yield_5_p L_yield_5_g tag_yield) psi_yield)
    (L_yield+4) 
      (Ast L_yield_4_p L_yield_4_g tag_yield) 
    (comml (_pushwp (Opr_reg DX))).
Proof.
  simpl.
  split.
  exists L_yield_5_p; exists L_yield_5_g.
  split_l (* lookup *).
  assert (Hpp' : forall sp pc ax bx cx mP S S',
    Ast_yield_4 sp pc ax bx cx mP S
    -> nextS (L_yield+4) (comml (_pushwp (Opr_reg DX))) S S'
    -> Ast_yield_5 sp pc ax bx cx (S_RF S DX) mP S').
    intros sp pc ax bx cx mP S S' HS HSS'.
    unfold Ast_yield_4, Ast_yield_5 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspgt [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simplS in * .
    destruct HS as [HM [Hcx [Hsp [Hspgt' [Hif Hisr]]]]].
    split.
      subst M' R'.
      sep on.
      sep lift 3.
      rewrite Hsp.
      arith_rep (sp - 6 -2) with (sp-8).
      sep upd.
      assert (Hx : M ||-
        stkSpacePrec (sp - 8) (_STK_SIZE_yield - 8)
        *^* sp - 8 |->w_
        *^* sp - 6 |->w cx
        *^* sp - 4 |->w bx
        *^* sp - 2 |->w ax
        *^* sp |->w pc 
        *^* INV1 *^* INV0 *^* mP).
        arith_rep (sp - 8) with (sp - 6 -2).
        arith_rep (_STK_SIZE_yield - 8) with
          (_STK_SIZE_yield - 6 - 2).
        assocr.
        apply stkSpacePrec_splitW1.
        subst cx.
        arith_rep (sp - 6 - 2 + 2) with (sp - 6).
        arith_rep (_STK_SIZE_yield - 6 - 2 + 2) with (_STK_SIZE_yield - 6).
        sep matchH HM with 5.
      sep matchH Hx with 8.
    subst R' fl' isr'; reg simpl.
    rewrite Hsp.
    split_l. 
    split_l by omega. 
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_yield_4_p in HpS.
    destruct HpS as [sp [pc [ax [bx [cx HS]]]]].
    exists (MemUpdW M (sp-8) (R DX) 0, RF_update R SP (sp - 8), fl, isr).
    apply nextS_pushwp with (R DX); trivial.
    unfold pushW.
    destruct HS as [HM [Hcx [Hsp [Hspgt [Hif Hisr]]]]].
    simplS in * .
    rewrite Hsp.
    arith_rep (sp-6-2) with (sp - 8).
    split.
      apply stkSpace_weaken with (_STK_SIZE_yield - 6);
        [idtac | omega].
      sep on.
      sep lift 2 in HM.
      sep trivial.
    split.
      apply MemUpdW_MemUpd16 with 0; trivial.
    trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_yield_4_p, L_yield_5_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [sp [pc [ax [bx [cx HS]]]]].
    exists sp; exists pc; exists ax; exists bx; exists cx; exists (S_RF S DX).
    apply Hpp'; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_yield_4_g, L_yield_5_g in * .
  intros mP sp pc ax bx cx HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [_ [Hcx [Hsp _]]].
  simplS in * .
  subst ax bx cx sp R' fl'.
  reg simpl.
  repeat split_l.
  trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_yield_4_p in HS.
    destruct HS as [sp [pc [ax [bx [cx HS]]]]].
    unfold Ast_yield_4 in HS.
    destruct HS as [HM [HCX [HSP [Hsp [Hif Hisr]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_yield_4_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP sp pc ax bx cx HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_yield_4 in HS.
  simplS in * .
  destruct HS as [HM [HCX [HSP [Hsp [Hif Hisr]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma NH_WFcmd_yield_5_pushwp :
  NH_WFcmd (SCAP_consCT 
    (L_yield+5+1)
      (Ast L_yield_6_p L_yield_6_g tag_yield) psi_yield)
    (L_yield+5) 
      (Ast L_yield_5_p L_yield_5_g tag_yield) 
    (comml (_pushwp (Opr_reg DI))).
Proof.
  simpl.
  split.
  exists L_yield_6_p; exists L_yield_6_g.
  split_l (* lookup *).
  assert (Hpp' : forall sp pc ax bx cx dx mP S S',
    Ast_yield_5 sp pc ax bx cx dx mP S
    -> nextS (L_yield+5) (comml (_pushwp (Opr_reg DI))) S S'
    -> Ast_yield_6 sp pc ax bx cx dx (S_RF S DI) mP S').
    intros sp pc ax bx cx dx mP S S' HS HSS'.
    unfold Ast_yield_5, Ast_yield_6 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspgt [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simplS in * .
    destruct HS as [HM [Hdx [Hsp [Hspgt' [Hif Hisr]]]]].
    split.
      subst M' R'.
      sep on.
      sep lift 3.
      rewrite Hsp.
      arith_rep (sp - 8 - 2) with (sp-10).
      sep upd.
      assert (Hx : M ||-
        stkSpacePrec (sp - 10) (_STK_SIZE_yield - 10)
        *^* sp - 10 |->w_
        *^* sp - 8 |->w dx
        *^* sp - 6 |->w cx
        *^* sp - 4 |->w bx
        *^* sp - 2 |->w ax
        *^* sp |->w pc 
        *^* INV1 *^* INV0 *^* mP).
        arith_rep (sp - 10) with (sp - 8 -2).
        arith_rep (_STK_SIZE_yield - 10) with
          (_STK_SIZE_yield - 8 - 2).
        assocr.
        apply stkSpacePrec_splitW1.
        subst dx.
        arith_rep (sp - 8 - 2 + 2) with (sp - 8).
        arith_rep (_STK_SIZE_yield - 8 - 2 + 2) with (_STK_SIZE_yield - 8).
        sep matchH HM with 6.
      sep matchH Hx with 9.
    subst R' fl' isr'; reg simpl.
    rewrite Hsp.
    split_l. 
    split_l by omega.
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_yield_5_p in HpS.
    destruct HpS as [sp [pc [ax [bx [cx [dx HS]]]]]].
    exists (MemUpdW M (sp-10) (R DI) 0, RF_update R SP (sp - 10), fl, isr).
    apply nextS_pushwp with (R DI); trivial.
    unfold pushW.
    destruct HS as [HM [Hdx [Hsp [Hspgt [Hif Hisr]]]]].
    simplS in * .
    rewrite Hsp.
    arith_rep (sp-8-2) with (sp - 10).
    split.
      apply stkSpace_weaken with (_STK_SIZE_yield - 8);
        [idtac | omega].
      sep on.
      sep lift 2 in HM.
      sep trivial.
    split.
      apply MemUpdW_MemUpd16 with 0; trivial.
    trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_yield_5_p, L_yield_6_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [sp [pc [ax [bx [cx [dx HS]]]]]].
    exists sp; exists pc; exists ax; exists bx; exists cx; 
      exists dx; exists (S_RF S DI).
    apply Hpp'; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_yield_5_g, L_yield_6_g in * .
  intros mP sp pc ax bx cx dx HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [_ [Hdx [Hsp _]]].
  simplS in * .
  subst ax bx cx dx sp R' fl'.
  reg simpl.
  repeat split_l.
  trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_yield_5_p in HS.
    destruct HS as [sp [pc [ax [bx [cx [dx HS]]]]]].
    unfold Ast_yield_5 in HS.
    destruct HS as [HM [HDX [HSP [Hsp [Hif Hisr]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_yield_5_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP sp pc ax bx cx dx HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_yield_5 in HS.
  simplS in * .
  destruct HS as [HM [HDX [HSP [Hsp [Hif Hisr]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma NH_WFcmd_yield_6_pushwp :
  NH_WFcmd (SCAP_consCT 
    (L_yield+6+1)
      (Ast L_yield_7_p L_yield_7_g tag_yield) psi_yield)
    (L_yield+6) 
      (Ast L_yield_6_p L_yield_6_g tag_yield) 
    (comml (_pushwp (Opr_reg SI))).
Proof.
  simpl.
  split.
  exists L_yield_7_p; exists L_yield_7_g.
  split_l (* lookup *).
  assert (Hpp' : forall sp pc ax bx cx dx di mP S S',
    Ast_yield_6 sp pc ax bx cx dx di mP S
    -> nextS (L_yield+6) (comml (_pushwp (Opr_reg SI))) S S'
    -> Ast_yield_7 sp pc ax bx cx dx di (S_RF S SI) mP S').
    intros sp pc ax bx cx dx di mP S S' HS HSS'.
    unfold Ast_yield_6, Ast_yield_7 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspgt [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simplS in * .
    destruct HS as [HM [Hdi [Hsp [Hspgt' [Hif Hisr]]]]].
    split.
      subst M' R'.
      sep on.
      sep lift 3.
      rewrite Hsp.
      arith_rep (sp - 10 - 2) with (sp-12).
      sep upd.
      assert (Hx : M ||-
        stkSpacePrec (sp - 12) (_STK_SIZE_yield - 12)
        *^* sp - 12 |->w_
        *^* sp - 10 |->w di
        *^* sp - 8 |->w dx
        *^* sp - 6 |->w cx
        *^* sp - 4 |->w bx
        *^* sp - 2 |->w ax
        *^* sp |->w pc 
        *^* INV1 *^* INV0 *^* mP).
        arith_rep (sp - 12) with (sp - 10 -2).
        arith_rep (_STK_SIZE_yield - 12) with
          (_STK_SIZE_yield - 10 - 2).
        assocr.
        apply stkSpacePrec_splitW1.
        subst di.
        arith_rep (sp - 10 - 2 + 2) with (sp - 10).
        arith_rep (_STK_SIZE_yield - 10 - 2 + 2) with (_STK_SIZE_yield - 10).
        sep matchH HM with 7.
      sep matchH Hx with 10.
    subst R' fl' isr' di; reg simpl.
    rewrite Hsp.
    split_l. 
    split_l by omega.
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_yield_6_p in HpS.
    destruct HpS as [sp [pc [ax [bx [cx [dx [di HS]]]]]]].
    exists (MemUpdW M (sp-12) (R SI) 0, RF_update R SP (sp - 12), fl, isr).
    apply nextS_pushwp with (R SI); trivial.
    unfold pushW.
    destruct HS as [HM [Hdi [Hsp [Hspgt [Hif Hisr]]]]].
    simplS in * .
    rewrite Hsp.
    arith_rep (sp-10-2) with (sp - 12).
    split.
      apply stkSpace_weaken with (_STK_SIZE_yield - 10);
        [idtac | omega].
      sep on.
      sep lift 2 in HM.
      sep trivial.
    split.
      apply MemUpdW_MemUpd16 with 0; trivial.
    trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_yield_6_p, L_yield_7_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [sp [pc [ax [bx [cx [dx [di HS]]]]]]].
    exists sp; exists pc; exists ax; exists bx; exists cx; 
      exists dx; exists di; exists (S_RF S SI).
    apply Hpp'; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_yield_6_g, L_yield_7_g in * .
  intros mP sp pc ax bx cx dx di HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [_ [Hdi [Hsp _]]].
  simplS in * .
  subst ax bx cx dx di sp R' fl'.
  reg simpl.
  repeat split_l.
  trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_yield_6_p in HS.
    destruct HS as [sp [pc [ax [bx [cx [dx [di HS]]]]]]].
    unfold Ast_yield_6 in HS.
    destruct HS as [HM [HDI [HSP [Hsp [Hif Hisr]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_yield_6_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP sp pc ax bx cx dx di HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_yield_6 in HS.
  simplS in * .
  destruct HS as [HM [HDI [HSP [Hsp [Hif Hisr]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma NH_WFcmd_yield_7_pushwp :
  NH_WFcmd (SCAP_consCT 
    (L_yield+7+1)
      (Ast L_yield_8_p L_yield_8_g tag_yield) psi_yield)
    (L_yield+7) 
      (Ast L_yield_7_p L_yield_7_g tag_yield) 
    (comml (_pushwp (Opr_reg BP))).
Proof.
  simpl.
  split.
  exists L_yield_8_p; exists L_yield_8_g.
  split_l (* lookup *).
  assert (Hpp' : forall sp pc ax bx cx dx di si mP S S',
    Ast_yield_7 sp pc ax bx cx dx di si mP S
    -> nextS (L_yield+7) (comml (_pushwp (Opr_reg BP))) S S'
    -> Ast_yield_8 sp pc ax bx cx dx di si (S_RF S BP) mP S').
    intros sp pc ax bx cx dx di si mP S S' HS HSS'.
    unfold Ast_yield_7, Ast_yield_8 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspgt [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simplS in * .
    destruct HS as [HM [Hdi [Hsp [Hspgt' [Hif Hisr]]]]].
    split.
      subst M' R'.
      sep on.
      sep lift 3.
      rewrite Hsp.
      arith_rep (sp - 12 - 2) with (sp-14).
      sep upd.
      assert (Hx : M ||-
        stkSpacePrec (sp - 14) (_STK_SIZE_yield - 14)
        *^* sp - 14 |->w_
        *^* sp - 12 |->w si
        *^* sp - 10 |->w di
        *^* sp - 8 |->w dx
        *^* sp - 6 |->w cx
        *^* sp - 4 |->w bx
        *^* sp - 2 |->w ax
        *^* sp |->w pc 
        *^* INV1 *^* INV0 *^* mP).
        arith_rep (sp - 14) with (sp - 12 -2).
        arith_rep (_STK_SIZE_yield - 14) with
          (_STK_SIZE_yield - 12 - 2).
        assocr.
        apply stkSpacePrec_splitW1.
        subst si.
        arith_rep (sp - 12 - 2 + 2) with (sp - 12).
        arith_rep (_STK_SIZE_yield - 12 - 2 + 2) with (_STK_SIZE_yield - 12).
        sep matchH HM with 8.
      sep matchH Hx with 11.
    subst R' fl' isr' si; reg simpl.
    rewrite Hsp.
    split_l. 
    split_l by omega.
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_yield_7_p in HpS.
    destruct HpS as [sp [pc [ax [bx [cx [dx [di [si HS]]]]]]]].
    exists (MemUpdW M (sp-14) (R BP) 0, RF_update R SP (sp - 14), fl, isr).
    apply nextS_pushwp with (R BP); trivial.
    unfold pushW.
    destruct HS as [HM [Hsi [Hsp [Hspgt [Hif Hisr]]]]].
    simplS in * .
    rewrite Hsp.
    arith_rep (sp-12-2) with (sp - 14).
    split.
      apply stkSpace_weaken with (_STK_SIZE_yield - 12);
        [idtac | omega].
      sep on.
      sep lift 2 in HM.
      sep trivial.
    split.
      apply MemUpdW_MemUpd16 with 0; trivial.
    trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_yield_7_p, L_yield_8_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [sp [pc [ax [bx [cx [dx [di [si HS]]]]]]]].
    exists sp; exists pc; exists ax; exists bx; exists cx; 
      exists dx; exists di; exists si; exists (S_RF S BP).
    apply Hpp'; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_yield_6_g, L_yield_7_g in * .
  intros mP sp pc ax bx cx dx di si HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [_ [Hsi [Hsp _]]].
  simplS in * .
  subst ax bx cx dx di si sp R' fl'.
  reg simpl.
  repeat split_l.
  trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_yield_7_p in HS.
    destruct HS as [sp [pc [ax [bx [cx [dx [di [si HS]]]]]]]].
    unfold Ast_yield_7 in HS.
    destruct HS as [HM [HSI [HSP [Hsp [Hif Hisr]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_yield_7_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP sp pc ax bx cx dx di si HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_yield_7 in HS.
  simplS in * .
  destruct HS as [HM [HSI [HSP [Hsp [Hif Hisr]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma NH_WFcmd_yield_8_pushfw :
  NH_WFcmd (SCAP_consCT 
    (L_yield+8+1)
      (Ast L_yield_9_p L_yield_9_g tag_yield) psi_yield)
    (L_yield+8) 
      (Ast L_yield_8_p L_yield_8_g tag_yield) 
    (comml (_pushfw)).
Proof.
  simpl.
  split.
  exists L_yield_9_p; exists L_yield_9_g.
  split_l (* lookup *).
  assert (Hpp' : forall sp pc ax bx cx dx di si bp mP S S',
    Ast_yield_8 sp pc ax bx cx dx di si bp mP S
    -> nextS (L_yield+8) (comml (_pushfw)) S S'
    -> Ast_yield_9 sp pc ax bx cx dx di si bp (FR2nat (S_Flag S)) mP S').
    intros sp pc ax bx cx dx di si bp mP S S' HS HSS'.
    unfold Ast_yield_8, Ast_yield_9 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HM' as [Hspgt [Hlds HM']].
    substH HM' with (MemUpd16_MemUpdW HM').
    destruct HM' as [junk HM'].
    simpl in HM'.
    simplS in * .
    destruct HS as [HM [Hbp [Hsp [Hspgt' [Hif Hisr]]]]].
    split.
      subst M' R'.
      sep on.
      sep lift 3.
      rewrite Hsp.
      arith_rep (sp - 14 - 2) with (sp-16).
      sep upd.
      assert (Hx : M ||-
        stkSpacePrec (sp - 16) (_STK_SIZE_yield - 16)
        *^* sp - 16 |->w_
        *^* sp - 14 |->w bp
        *^* sp - 12 |->w si
        *^* sp - 10 |->w di
        *^* sp - 8 |->w dx
        *^* sp - 6 |->w cx
        *^* sp - 4 |->w bx
        *^* sp - 2 |->w ax
        *^* sp |->w pc 
        *^* INV1 *^* INV0 *^* mP).
        arith_rep (sp - 16) with (sp - 14 -2).
        arith_rep (_STK_SIZE_yield - 16) with
          (_STK_SIZE_yield - 14 - 2).
        assocr.
        apply stkSpacePrec_splitW1.
        subst bp.
        arith_rep (sp - 14 - 2 + 2) with (sp - 14).
        arith_rep (_STK_SIZE_yield - 14 - 2 + 2) with (_STK_SIZE_yield - 14).
        sep matchH HM with 9.
      sep matchH Hx with 12.
    subst R' fl' isr' bp; reg simpl.
    rewrite Hsp.
    split_l by omega.
    split_l.
    rewrite FR_nat_iso_left.
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_yield_8_p in HpS.
    destruct HpS as [sp [pc [ax [bx [cx [dx [di [si [bp HS]]]]]]]]].
    exists (MemUpdW M (sp-16) (FR2nat fl) 0, RF_update R SP (sp - 16), fl, isr).
    apply nextS_pushfw; trivial.
    unfold pushW.
    destruct HS as [HM [Hbp [Hsp [Hspgt [Hif Hisr]]]]].
    simplS in * .
    rewrite Hsp.
    arith_rep (sp-14-2) with (sp - 16).
    split.
      apply stkSpace_weaken with (_STK_SIZE_yield - 14);
        [idtac | omega].
      sep on.
      sep lift 2 in HM.
      sep trivial.
    split.
      apply MemUpdW_MemUpd16 with 0; trivial.
    trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_yield_8_p, L_yield_9_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [sp [pc [ax [bx [cx [dx [di [si [bp HS]]]]]]]]].
    exists sp; exists pc; exists ax; exists bx; exists cx; 
      exists dx; exists di; exists si; exists bp; exists (FR2nat (S_Flag S)).
    apply Hpp'; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_yield_8_g, L_yield_9_g in * .
  intros mP sp pc ax bx cx dx di si bp HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ _ _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [_ [Hsi [Hsp _]]].
  simplS in * .
  subst ax bx cx dx di si sp bp R' fl'.
  repeat split_l.
  trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_yield_8_p in HS.
    destruct HS as [sp [pc [ax [bx [cx [dx [di [si [bp HS]]]]]]]]].
    unfold Ast_yield_8 in HS.
    destruct HS as [HM [HBP [HSP [Hsp [Hif Hisr]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_yield_8_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP sp pc ax bx cx dx di si bp HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_yield_8 in HS.
  simplS in * .
  destruct HS as [HM [HBP [HSP [Hsp [Hif Hisr]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma NH_WFcmd_yield_9_call :
  NH_WFcmd (SCAP_consCT 
    (L_yield+9+1)
      (Ast L_yield_10_p L_yield_10_g tag_yield) psi_yield)
    (L_yield+9) 
      (Ast L_yield_9_p L_yield_9_g tag_yield)
    (comml (_call _SWITCH)).
Proof.
  simpl.
  split.
  assert (Hpp''' : forall sp pc ax bx cx dx di si bp flnum
    mP S S' S'' S''' pc',
    Ast_yield_9 sp pc ax bx cx dx di si bp flnum mP S
    -> nextS (L_yield+9) (comml (_call _SWITCH)) S S'
    -> gy S' S''
    -> nextS pc' (comms _retw) S'' S'''
    -> Ast_yield_10 sp pc ax bx cx dx di si bp flnum mP S''').
    intros sp pc ax bx cx dx di si bp flnum mP.
    intros S S' S'' S''' pc'.
    intros HS HSS' HS'S'' HS''S'''.
    unfold Ast_yield_9, Ast_yield_10 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct S'' as [[[M'' R''] fl''] isr''].
    destruct S''' as [[[M''' R'''] fl'''] isr'''].
    simplS in * .
    destruct HS as [HM [Hsp [Hspg [Hif [Hfl Hisr]]]]].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    unfold gy in HS'S''.
    destruct HS'S'' as [sp' [HM'' [Hif'' [Hisr'' HSP]]]].
    destruct (nextS_inversion HS''S''') as [HM''' [HR''' [Hfl''' Hisr''']]].
    destruct HR''' as [pc''' [Hlds''' HR''']].
    destruct HM' as [Hspg' [Hlds Hupd]].
    simplS in * .
    subst M''' R''' fl''' isr'''.
    subst isr'' R' fl' isr' isr.
    clear HSS' HS''S'''.
    reg simpl.
    destruct HSP as [Hsp' HSP].
    subst sp'.
    rewrite <- HSP.
    rewrite Hsp.
    rewrite Hsp in HM''.
    split.
      sep on.
      sep lift 2.
      apply stkSpacePrec_mergeW1; [idtac | omega | omega ].
      arith_rep (16+2+8-16-2) with 8.
      assert (Hx : M''
        ||- (INV0 *^* INV1 *^* stkSpacePrec (sp - 16 - 2) 8) *^*
          (stkSpacePrec (sp - 16 - 2 - 8) IH_Stk *^*
           sp - 16 - 2 |->w (L_yield + 9 + 1) *^*
           sp - 16 |->w flnum *^*
           sp - 14 |->w bp *^*
           sp - 12 |->w si *^*
           sp - 10 |->w di *^*
          sp - 8 |->w dx *^*
          sp - 6 |->w cx *^*
          sp - 4 |->w bx *^* 
          sp - 2 |->w ax *^* 
          sp |->w pc *^* mP)).
        replace M'' with (S_Mem (M'', R'', fl'', false));
          trivial.
        sep off.
        assert (forall sp size, precise (INV0 *^* INV1 *^* stkSpacePrec sp size)).
          intros spx sizex.
          apply mstar_presv_prec.
          apply Prec01.
          apply precise_stkSpacePrec.
        apply prepost_elim with (INV0 *^* INV1 *^* stkSpacePrec (sp - 16 - 2) 8)
          (M', RF_update R SP (sp - 16 - 2), fl, false);
          auto.
        simplS.
        substH Hupd with (MemUpd16_MemUpdW Hupd).
        destruct Hupd as [x HM'].
        rewrite Hsp in HM'.
        subst M'.
        sep on.
        sep lift 4.
        sep upd.
        sep lift 2 in HM.
        sep stksplit in HM.
        arith_rep (_STK_SIZE_yield - 16 - 2) with (8 + IH_Stk) in HM.
        sep stksplit 8 in HM.
        arith_rep (8 + IH_Stk - 8) with IH_Stk in HM.
        sep matchH HM with 14.
      arith_rep (_STK_SIZE_yield - 16 - 2) with (8 + IH_Stk).
      sep stksplit 8.
      arith_rep (8 + IH_Stk - 8) with IH_Stk.
      sep matchH Hx with 13.
      split_l by omega.
      subst fl.
      rewrite <- Hif''.
      auto.
  left.
  split.
    unfold prim_entry.
    left; trivial.
  split.
    constructor 1 with L_yield_9_p L_yield_10_p 
      L_yield_9_g L_yield_10_g tag_yield; trivial.
    (* enable *)
    unfold enable_sw.
    intros S HS.
    unfold L_yield_9_p in HS.
    destruct HS as [sp [pc [ax [bx [cx [dx [di [si [bp [flnum HS]]]]]]]]]].
    unfold pTrue, pmstar.
    unfold Ast_yield_9 in HS.
    destruct S as [[[M R] fl] isr].
    simplS in * .
    destruct HS as [HM [Hsp [Hspg HS]]].
    rewrite Hsp.
    assert (HMx : ((stkSpacePrec (sp - 16 - (8 + 2)) IH_Stk
               *^* sp - 16 |->w flnum 
               *^* sp - 14 |->w bp 
               *^* sp - 12 |->w si 
               *^* sp - 10 |->w di 
               *^* sp - 8 |->w dx 
               *^* sp - 6 |->w cx 
               *^* sp - 4 |->w bx 
               *^* sp - 2 |->w ax 
               *^* sp |->w pc *^* mTrue)
      *^* (INV0 *^* INV1 *^* stkSpacePrec (sp - 16) (8 + 2)))
        M).
      sep on.
      sep lift 2 in HM.
      sep stksplit (8+2) in HM.
      arith_rep (_STK_SIZE_yield - 16 - (8 + 2)) with IH_Stk in HM.
      sep matchH HM with 13.
    destruct HMx as [M1 [M2 [HJM [HM1 HM2]]]].
    exists M1; exists M2; split_l.
    auto.
    (* p ==> p' *)
    intros pc' S'''.
    unfold L_yield_9_p, L_yield_10_p.
    unfold wrap_fun.
    unfold linkPG, wrap_fun, conjG.
    intros H.
    destruct H as [S H].
    destruct H as [HS H].
    destruct H as [S'' [H HS''S''']].
    destruct H as [S' [HSS' HS'S'']].
    destruct HS as [sp [pc [ax [bx [cx [dx [di [si [bp [flnum HS]]]]]]]]]].
    exists sp; exists pc; exists ax; exists bx; exists cx; exists dx;
      exists di; exists si; exists bp; exists flnum.
    eapply Hpp'''; eauto.
    (* g''' ==> g *)
    intros pc' S S''''.
    unfold L_yield_9_g, L_yield_10_g.
    unfold wrap_fun.
    intros HSS''''.
    unfold conjG at 1 in HSS'''' .
    destruct HSS'''' as [S''' [HSS''' H]].
    destruct HSS''' as [HSx [S'' HSS''']].
    destruct HSS''' as [HSS'' HS''S'''].
    destruct HSS'' as [S' [HSS' HS'S'']].
    intros mP sp pc ax bx cx dx di si bp flnum HS.
    apply H with flnum.
    eapply Hpp'''; eauto.
    (* p ==> if=false *)
    intros S HS.
    destruct HS as [sp [pc [ax [bx [cx [dx [di [si [bp [flnum HS]]]]]]]]]].
    unfold Ast_yield_9 in HS.
    destruct HS as [HM [Hsp [Hspg [Hif [Hfl Hisr]]]]].
    rewrite <- Hfl in Hif.
    trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_yield_9_p in HS.
    destruct HS as [sp [pc [ax [bx [cx [dx [di [si [bp [fln HS]]]]]]]]]].
    unfold Ast_yield_9 in HS.
    destruct HS as [HM [HSP [Hsp [Hfln [Hif Hisr]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    rewrite Hfln in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_yield_9_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP sp pc ax bx cx dx di si bp fln HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_yield_9 in HS.
  simplS in * .
  destruct HS as [HM [HSP [Hsp [Hfln [Hif Hisr]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  rewrite Hfln in IHif.
  discriminate IHif.
Qed.

Lemma NH_WFcmd_yield_10_popfw :
  NH_WFcmd (SCAP_consCT
    (L_yield+10+1)
      (Ast L_yield_11_p L_yield_11_g tag_yield) psi_yield)
    (L_yield+10)
      (Ast L_yield_10_p L_yield_10_g tag_yield)
    (comml (_popfw)).
Proof.
  simpl.
  split.
  exists L_yield_11_p; exists L_yield_11_g.
  split_l (* lookup *).
  assert (Hpp' : forall sp pc ax bx cx dx di si bp flnum mP S S',
    Ast_yield_10 sp pc ax bx cx dx di si bp flnum mP S
    -> stk_IF_cl S
    -> gpopfw_cli S S'
    -> Ast_yield_11 sp pc ax bx cx dx di si bp mP S').
    intros sp pc ax bx cx dx di si bp flnum mP S S' HS Hcl HSS'.
    unfold Ast_yield_10, Ast_yield_11, stk_IF_cl in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    unfold gpopfw_cli in HSS'.
    simplS in * .
    destruct HSS' as [Hshift [Hif' [Hisr' HR']]].
    assert (HM' : M' = M).
      apply (prepost_elim_emp_M M R fl isr M' R' fl' isr').
      apply (prepost_weaken_post (mQ':=Aemp) Hshift).
      intros Mx H.
      apply (choice_elim_right H).
      intros [H1 H2].
      InvertAll.
      simplS in * .
      rewrite H5 in H1; discriminate H1.
    subst M' R' isr'.
    destruct HS as [HM [Hsp [Hspg [Hif [Hfl Hisr]]]]].
    split.
      sep on.
      sep matchH HM with 11.
      sep stksplit.
      arith_rep (sp - 14 - 2) with (sp - 16).
      arith_rep (_STK_SIZE_yield - 14 - 2) with (_STK_SIZE_yield - 16).
      sep matchH HM with 2.
    rewrite Hsp.
    reg simpl.
    split_l by omega.
    auto.
  assert (Hpp'_sti : forall sp pc ax bx cx dx di si bp flnum mP S S',
    Ast_yield_10 sp pc ax bx cx dx di si bp flnum mP S
    -> stk_IF_st S
    -> gpopfw_sti S S'
    -> Ast_yield_11 sp pc ax bx cx dx di si bp mP S').
    clear Hpp'.
    intros sp pc ax bx cx dx di si bp flnum mP S S' HS Hst HSS'.
    unfold stk_IF_st in * .
    destruct Hst as [flnumx [Hlds Hif]].
    unfold Ast_yield_10 in HS.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct HS as [HM [Hsp [Hspg [Hifx [Hfl Hisr]]]]].
    simplS in * .
    assert (Heq : flnum = flnumx).
      sep on.
      sep lift 3 in HM.
      rewrite Hsp in Hlds.
      sep inj.
      auto.
    subst flnumx.
    rewrite Hfl in Hif; discriminate Hif.
  split.
  (* enable popfw *)
    clear Hpp' Hpp'_sti.
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_yield_10_p in HpS.
    destruct HpS as [sp [pc [ax [bx [cx [dx [di [si [bp [flnum HS]]]]]]]]]].
    exists (M, RF_update R SP (sp - 14),(nat2FR flnum), isr).
    apply nextS_popfw with flnum; trivial.
    unfold popW.
    destruct HS as [HM [Hsp [Hspgt [Hif [Hfl Hisr]]]]].
    simplS in * .
    rewrite Hsp.
    split.
      sep on.
      sep lift 3 in HM.
      sep trivial.
    arith_rep (sp-16+2) with (sp - 14).
  split.
  (* enable sti *)
    unfold enable_popfw_sti, stk_IF_st, conjP.
    intros M R fl isr HS.
    destruct HS as [HS Hflnum].
    unfold L_yield_10_p in HS.
    destruct HS as [sp [pc [ax [bx [cx [dx [di [si [bp [flnum HS]]]]]]]]]].
    destruct Hflnum as [flnumx [Hlds Hif]].
    unfold Ast_yield_10 in HS. 
    destruct HS as [HM [Hsp [Hspg [Hifx [Hfl Hisr]]]]].
    assert (Heq : flnum = flnumx).
      sep on.
      sep lift 3 in HM.
      rewrite Hsp in Hlds.
      sep inj.
      auto.
    subst flnumx.
    rewrite Hfl in Hif; discriminate Hif.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_yield_10_p, L_yield_11_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HSS'].
    destruct HS as [HS Hst].
    destruct HS as [sp [pc [ax [bx [cx [dx [di [si [bp [flnum HS]]]]]]]]]].
    exists sp; exists pc; exists ax; exists bx; exists cx; 
      exists dx; exists di; exists si; exists bp.
    eapply Hpp'_sti; eauto.
  split.
  (* g' ==> g *)
    intros S S'' H.
    unfold linkPG, conjPG, conjG in H.
    destruct H as [S' H].
    destruct H as [[HpS HnextSS'] HgS'S''].
    destruct HpS as [HSx Hst].
    unfold L_yield_10_g, L_yield_11_g in * .
    intros mP sp pc ax bx cx dx di si bp flnum HS.
    substH Hpp'_sti with (Hpp'_sti _ _ _ _ _ _ _ _ _ _ _ _ _ HS Hst HnextSS').
    substH HgS'S'' with (HgS'S'' _ _ _ _ _ _ _ _ _ _ Hpp'_sti).
    destruct HgS'S'' as [HS'' HS'S''].
    split_l.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct S'' as [[[M'' R''] fl''] isr''].
    unfold gpopfw_sti in HnextSS'.
    destruct HnextSS' as [HM' [Hif' [Hisr' HR']]].
    destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
    destruct HS as [_ [Hsi [Hsp [Hflnum [Hflnumx _]]]]].
    simplS in * .
    subst ax bx cx dx di si sp bp R'.
    repeat split_l.
    trivial.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_yield_10_p, L_yield_11_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HSS'].
    destruct HS as [HS Hcl].
    destruct HS as [sp [pc [ax [bx [cx [dx [di [si [bp [flnum HS]]]]]]]]]].
    exists sp; exists pc; exists ax; exists bx; exists cx; 
      exists dx; exists di; exists si; exists bp.
    eapply Hpp'; eauto.
  (* g' ==> g *)
    intros S S'' H.
    unfold linkPG, conjPG, conjG in H.
    destruct H as [S' H].
    destruct H as [[HpS HnextSS'] HgS'S''].
    destruct HpS as [HSx Hst].
    unfold L_yield_10_g, L_yield_11_g in * .
    intros mP sp pc ax bx cx dx di si bp flnum HS.
    substH Hpp' with (Hpp' _ _ _ _ _ _ _ _ _ _ _ _ _ HS Hst HnextSS').
    substH HgS'S'' with (HgS'S'' _ _ _ _ _ _ _ _ _ _ Hpp').
    destruct HgS'S'' as [HS'' HS'S''].
    split_l.
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct S'' as [[[M'' R''] fl''] isr''].
    unfold gpopfw_cli in HnextSS'.
    destruct HnextSS' as [HM' [Hif' [Hisr' HR']]].
    destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
    destruct HS as [_ [Hsi [Hsp [Hflnum [Hflnumx _]]]]].
    simplS in * .
    subst ax bx cx dx di si sp bp R'.
    repeat split_l.
    trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_yield_10_p in HS.
    destruct HS as [sp [pc [ax [bx [cx [dx [di [si [bp [fln HS]]]]]]]]]].
    unfold Ast_yield_10 in HS.
    destruct HS as [HM [HSP [Hsp [Hif [Hfln Hisr]]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_yield_10_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP sp pc ax bx cx dx di si bp fln HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_yield_10 in HS.
  simplS in * .
  destruct HS as [HM [HSP [Hsp [Hif [Hfln Hisr]]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma NH_WFcmd_yield_11_popw :
  NH_WFcmd (SCAP_consCT 
    (L_yield+11+1)
      (Ast L_yield_12_p L_yield_12_g tag_yield) psi_yield)
    (L_yield+11)
      (Ast L_yield_11_p L_yield_11_g tag_yield)
    (comml (_popw BP)).
Proof.
  simpl.
  split.
  exists L_yield_12_p; exists L_yield_12_g.
  split_l (* lookup *).
  assert (Hpp' : forall sp pc ax bx cx dx di si bp mP S S',
    Ast_yield_11 sp pc ax bx cx dx di si bp mP S
    -> nextS (L_yield+11) (comml (_popw BP)) S S'
    -> Ast_yield_12 sp pc ax bx cx dx di si mP S').
    intros sp pc ax bx cx dx di si bp mP S S' HS HSS'.
    unfold Ast_yield_11, Ast_yield_12 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HR' as [bpx [Hlds HR']].
    simplS in * .
    destruct HS as [HM [Hsp [Hspgt' [Hif Hisr]]]].
    assert (Heq : bp = bpx).
      sep on.
      sep lift 3 in HM.
      rewrite Hsp in Hlds.
      sep inj.
      auto.
    subst bpx.
    subst M' fl' R' isr'.
    split.
      sep on.
      sep lift 2.
      sep lift 2 in HM.
      sep stksplit.
      arith_rep (sp - 12 - 2) with (sp-14).
      arith_rep (_STK_SIZE_yield - 12 - 2) with (_STK_SIZE_yield-14).
      sep matchH HM with 12.
    rewrite Hsp.
    reg simpl.
    split_l by omega.
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_yield_11_p in HpS.
    destruct HpS as [sp [pc [ax [bx [cx [dx [di [si [bp HS]]]]]]]]].
    exists (M, RF_update (RF_update R SP (sp - 12)) BP bp, fl, isr).
    apply nextS_popw with (RF_update R SP (sp - 12)) bp; trivial.
    unfold popW.
    destruct HS as [HM [Hsp [Hspgt [Hif Hisr]]]].
    simplS in * .
    rewrite Hsp.
    split.
      sep on.
      sep lift 3 in HM.
      sep trivial.
    arith_rep (sp - 14 + 2) with (sp - 12).
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_yield_11_p, L_yield_12_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [sp [pc [ax [bx [cx [dx [di [si [bp HS]]]]]]]]].
    exists sp; exists pc; exists ax; exists bx; exists cx; 
      exists dx; exists di; exists si.
    apply Hpp' with bp S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_yield_11_g, L_yield_12_g in * .
  intros mP sp pc ax bx cx dx di si bp HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [HM [Hsp _]].
  simplS in * .
  destruct HR' as [bpx [Hlds HR']].
  assert (Heq : bp = bpx).
    sep on.
    sep lift 3 in HM.
    rewrite Hsp in Hlds.
    sep inj.
    auto.
  subst bpx ax bx cx dx di si sp R' fl' isr'.
  reg simpl.
  repeat split_l; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_yield_11_p in HS.
    destruct HS as [sp [pc [ax [bx [cx [dx [di [si [bp HS]]]]]]]]].
    unfold Ast_yield_11 in HS.
    destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_yield_11_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP sp pc ax bx cx dx di si bp HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_yield_11 in HS.
  simplS in * .
  destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma NH_WFcmd_yield_12_popw :
  NH_WFcmd (SCAP_consCT 
    (L_yield+12+1)
      (Ast L_yield_13_p L_yield_13_g tag_yield) psi_yield)
    (L_yield+12)
      (Ast L_yield_12_p L_yield_12_g tag_yield) 
    (comml (_popw SI)).
Proof.
  simpl.
  split.
  exists L_yield_13_p; exists L_yield_13_g.
  split_l (* lookup *).
  assert (Hpp' : forall sp pc ax bx cx dx di si mP S S',
    Ast_yield_12 sp pc ax bx cx dx di si mP S
    -> nextS (L_yield+12) (comml (_popw SI)) S S'
    -> Ast_yield_13 sp pc ax bx cx dx di mP S').
    intros sp pc ax bx cx dx di si mP S S' HS HSS'.
    unfold Ast_yield_12, Ast_yield_13 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HR' as [six [Hlds HR']].
    simplS in * .
    destruct HS as [HM [Hsp [Hspgt' [Hif Hisr]]]].
    assert (Heq : si = six).
      sep on.
      sep lift 3 in HM.
      rewrite Hsp in Hlds.
      sep inj.
      auto.
    subst six.
    subst M' fl' R' isr'.
    split.
      sep on.
      sep lift 2.
      sep lift 2 in HM.
      sep stksplit.
      arith_rep (sp - 10 - 2) with (sp-12).
      arith_rep (_STK_SIZE_yield - 10 - 2) with (_STK_SIZE_yield-12).
      sep matchH HM with 12.
    rewrite Hsp.
    reg simpl.
    split_l by omega.
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_yield_12_p in HpS.
    destruct HpS as [sp [pc [ax [bx [cx [dx [di [si HS]]]]]]]].
    exists (M, RF_update (RF_update R SP (sp - 10)) SI si, fl, isr).
    apply nextS_popw with (RF_update R SP (sp - 10)) si; trivial.
    unfold popW.
    destruct HS as [HM [Hsp [Hspgt [Hif Hisr]]]].
    simplS in * .
    rewrite Hsp.
    split.
      sep on.
      sep lift 3 in HM.
      sep trivial.
    arith_rep (sp - 12 + 2) with (sp - 10).
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_yield_12_p, L_yield_13_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [sp [pc [ax [bx [cx [dx [di [si HS]]]]]]]].
    exists sp; exists pc; exists ax; exists bx; exists cx; 
      exists dx; exists di.
    apply Hpp' with si S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_yield_12_g, L_yield_13_g in * .
  intros mP sp pc ax bx cx dx di si HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [HM [Hsp _]].
  simplS in * .
  destruct HR' as [six [Hlds HR']].
  assert (Heq : si = six).
    sep on.
    sep lift 3 in HM.
    rewrite Hsp in Hlds.
    sep inj.
    auto.
  subst six ax bx cx dx di sp R' fl' isr'.
  reg simpl.
  repeat split_l; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_yield_12_p in HS.
    destruct HS as [sp [pc [ax [bx [cx [dx [di [si HS]]]]]]]].
    unfold Ast_yield_12 in HS.
    destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_yield_12_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP sp pc ax bx cx dx di si HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_yield_12 in HS.
  simplS in * .
  destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma NH_WFcmd_yield_13_popw :
  NH_WFcmd (SCAP_consCT 
    (L_yield+13+1)
      (Ast L_yield_14_p L_yield_14_g tag_yield) psi_yield)
    (L_yield+13)
      (Ast L_yield_13_p L_yield_13_g tag_yield) 
    (comml (_popw DI)).
Proof.
  simpl.
  split.
  exists L_yield_14_p; exists L_yield_14_g.
  split_l (* lookup *).
  assert (Hpp' : forall sp pc ax bx cx dx di mP S S',
    Ast_yield_13 sp pc ax bx cx dx di mP S
    -> nextS (L_yield+13) (comml (_popw DI)) S S'
    -> Ast_yield_14 sp pc ax bx cx dx mP S').
    intros sp pc ax bx cx dx di mP S S' HS HSS'.
    unfold Ast_yield_13, Ast_yield_14 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HR' as [dix [Hlds HR']].
    simplS in * .
    destruct HS as [HM [Hsp [Hspgt' [Hif Hisr]]]].
    assert (Heq : di = dix).
      sep on.
      sep lift 3 in HM.
      rewrite Hsp in Hlds.
      sep inj.
      auto.
    subst dix.
    subst M' fl' R' isr'.
    split.
      sep on.
      sep lift 2.
      sep lift 2 in HM.
      sep stksplit.
      arith_rep (sp - 8 - 2) with (sp-10).
      arith_rep (_STK_SIZE_yield - 8 - 2) with (_STK_SIZE_yield-10).
      sep matchH HM with 9.
    rewrite Hsp.
    reg simpl.
    split_l by omega.
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_yield_13_p in HpS.
    destruct HpS as [sp [pc [ax [bx [cx [dx [di HS]]]]]]].
    exists (M, RF_update (RF_update R SP (sp - 8)) DI di, fl, isr).
    apply nextS_popw with (RF_update R SP (sp - 8)) di; trivial.
    unfold popW.
    destruct HS as [HM [Hsp [Hspgt [Hif Hisr]]]].
    simplS in * .
    rewrite Hsp.
    split.
      sep on.
      sep lift 3 in HM.
      sep trivial.
      auto.
    arith_rep (sp - 10 + 2) with (sp - 8).
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_yield_13_p, L_yield_14_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [sp [pc [ax [bx [cx [dx [di HS]]]]]]].
    exists sp; exists pc; exists ax; exists bx; exists cx; 
      exists dx.
    apply Hpp' with di S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_yield_13_g, L_yield_14_g in * .
  intros mP sp pc ax bx cx dx di HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [HM [Hsp _]].
  simplS in * .
  destruct HR' as [dix [Hlds HR']].
  assert (Heq : di = dix).
    sep on.
    sep lift 3 in HM.
    rewrite Hsp in Hlds.
    sep inj.
    auto.
  subst dix ax bx cx dx sp R' fl' isr'.
  reg simpl.
  repeat split_l; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_yield_13_p in HS.
    destruct HS as [sp [pc [ax [bx [cx [dx [di HS]]]]]]].
    unfold Ast_yield_13 in HS.
    destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_yield_13_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP sp pc ax bx cx dx di HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_yield_13 in HS.
  simplS in * .
  destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma NH_WFcmd_yield_14_popw :
  NH_WFcmd (SCAP_consCT 
    (L_yield+14+1)
      (Ast L_yield_15_p L_yield_15_g tag_yield) psi_yield)
    (L_yield+14)
      (Ast L_yield_14_p L_yield_14_g tag_yield) 
    (comml (_popw DX)).
Proof.
  simpl.
  split.
  exists L_yield_15_p; exists L_yield_15_g.
  split_l (* lookup *).
  assert (Hpp' : forall sp pc ax bx cx dx mP S S',
    Ast_yield_14 sp pc ax bx cx dx mP S
    -> nextS (L_yield+14) (comml (_popw DX)) S S'
    -> Ast_yield_15 sp pc ax bx cx mP S').
    intros sp pc ax bx cx dx mP S S' HS HSS'.
    unfold Ast_yield_14, Ast_yield_15 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HR' as [dxx [Hlds HR']].
    simplS in * .
    destruct HS as [HM [Hsp [Hspgt' [Hif Hisr]]]].
    assert (Heq : dx = dxx).
      sep on.
      sep lift 3 in HM.
      rewrite Hsp in Hlds.
      sep inj.
      auto.
    subst dxx.
    subst M' fl' R' isr'.
    split.
      sep on.
      sep lift 2.
      sep lift 2 in HM.
      sep stksplit. 
      arith_rep (sp - 6 - 2) with (sp-8).
      arith_rep (_STK_SIZE_yield - 6 - 2) with (_STK_SIZE_yield-8).
      sep matchH HM with 8.
    rewrite Hsp.
    reg simpl.
    split_l by omega.
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_yield_14_p in HpS.
    destruct HpS as [sp [pc [ax [bx [cx [dx HS]]]]]].
    exists (M, RF_update (RF_update R SP (sp - 6)) DX dx, fl, isr).
    apply nextS_popw with (RF_update R SP (sp - 6)) dx; trivial.
    unfold popW.
    destruct HS as [HM [Hsp [Hspgt [Hif Hisr]]]].
    simplS in * .
    rewrite Hsp.
    split.
      sep on.
      sep lift 3 in HM.
      sep trivial.
    arith_rep (sp - 8 + 2) with (sp - 6).
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_yield_14_p, L_yield_15_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [sp [pc [ax [bx [cx [dx HS]]]]]].
    exists sp; exists pc; exists ax; exists bx; exists cx.
    apply Hpp' with dx S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_yield_14_g, L_yield_15_g in * .
  intros mP sp pc ax bx cx dx HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [HM [Hsp _]].
  simplS in * .
  destruct HR' as [dxx [Hlds HR']].
  assert (Heq : dx = dxx).
    sep on.
    sep lift 3 in HM.
    rewrite Hsp in Hlds.
    sep inj.
    auto.
  subst ax bx cx dxx sp R' fl' isr'.
  reg simpl.
  repeat split_l; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_yield_14_p in HS.
    destruct HS as [sp [pc [ax [bx [cx [dx HS]]]]]].
    unfold Ast_yield_14 in HS.
    destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_yield_14_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP sp pc ax bx cx dx HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_yield_14 in HS.
  simplS in * .
  destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma NH_WFcmd_yield_15_popw :
  NH_WFcmd (SCAP_consCT 
    (L_yield+15+1)
      (Ast L_yield_16_p L_yield_16_g tag_yield) psi_yield)
    (L_yield+15)
      (Ast L_yield_15_p L_yield_15_g tag_yield) 
    (comml (_popw CX)).
Proof.
  simpl.
  split.
  exists L_yield_16_p; exists L_yield_16_g.
  split_l (* lookup *).
  assert (Hpp' : forall sp pc ax bx cx mP S S',
    Ast_yield_15 sp pc ax bx cx mP S
    -> nextS (L_yield+15) (comml (_popw CX)) S S'
    -> Ast_yield_16 sp pc ax bx mP S').
    intros sp pc ax bx cx mP S S' HS HSS'.
    unfold Ast_yield_15, Ast_yield_16 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HR' as [cxx [Hlds HR']].
    simplS in * .
    destruct HS as [HM [Hsp [Hspgt' [Hif Hisr]]]].
    assert (Heq : cx = cxx).
      sep on.
      sep lift 3 in HM.
      rewrite Hsp in Hlds.
      sep inj.
      auto.
    subst cxx.
    subst M' fl' R' isr'.
    split.
      sep on.
      sep lift 2.
      sep lift 2 in HM.
      sep stksplit.
      arith_rep (sp - 4 - 2) with (sp-6).
      arith_rep (_STK_SIZE_yield - 4 - 2) with (_STK_SIZE_yield-6).
      sep matchH HM with 7.
    rewrite Hsp.
    reg simpl.
    split_l by omega.
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_yield_15_p in HpS.
    destruct HpS as [sp [pc [ax [bx [cx HS]]]]].
    exists (M, RF_update (RF_update R SP (sp - 4)) CX cx, fl, isr).
    apply nextS_popw with (RF_update R SP (sp - 4)) cx; trivial.
    unfold popW.
    destruct HS as [HM [Hsp [Hspgt [Hif Hisr]]]].
    simplS in * .
    rewrite Hsp.
    split.
      sep on.
      sep lift 3 in HM.
      sep trivial.
    arith_rep (sp - 6 + 2) with (sp - 4).
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_yield_15_p, L_yield_16_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [sp [pc [ax [bx [cx HS]]]]].
    exists sp; exists pc; exists ax; exists bx.
    apply Hpp' with cx S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_yield_15_g, L_yield_16_g in * .
  intros mP sp pc ax bx cx HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [HM [Hsp _]].
  simplS in * .
  destruct HR' as [cxx [Hlds HR']].
  assert (Heq : cx = cxx).
    sep on.
    sep lift 3 in HM.
    rewrite Hsp in Hlds.
    sep inj.
    auto.
  subst ax bx cxx sp R' fl' isr'.
  reg simpl.
  repeat split_l; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_yield_15_p in HS.
    destruct HS as [sp [pc [ax [bx [cx HS]]]]].
    unfold Ast_yield_15 in HS.
    destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_yield_15_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP sp pc ax bx cx HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_yield_15 in HS.
  simplS in * .
  destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma NH_WFcmd_yield_16_popw :
  NH_WFcmd (SCAP_consCT 
    (L_yield+16+1)
      (Ast L_yield_17_p L_yield_17_g tag_yield) psi_yield)
    (L_yield+16)
      (Ast L_yield_16_p L_yield_16_g tag_yield) 
    (comml (_popw BX)).
Proof.
  simpl.
  split.
  exists L_yield_17_p; exists L_yield_17_g.
  split_l (* lookup *).
  assert (Hpp' : forall sp pc ax bx mP S S',
    Ast_yield_16 sp pc ax bx mP S
    -> nextS (L_yield+16) (comml (_popw BX)) S S'
    -> Ast_yield_17 sp pc ax mP S').
    intros sp pc ax bx mP S S' HS HSS'.
    unfold Ast_yield_16, Ast_yield_17 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HR' as [bxx [Hlds HR']].
    simplS in * .
    destruct HS as [HM [Hsp [Hspgt' [Hif Hisr]]]].
    assert (Heq : bx = bxx).
      sep on.
      sep lift 3 in HM.
      rewrite Hsp in Hlds.
      sep inj.
      auto.
    subst bxx.
    subst M' fl' R' isr'.
    split.
      sep on.
      sep lift 2.
      sep lift 2 in HM.
      sep stksplit.
      arith_rep (sp - 2 - 2) with (sp-4).
      arith_rep (_STK_SIZE_yield - 2 - 2) with (_STK_SIZE_yield-4).
      sep matchH HM with 6.
    rewrite Hsp.
    reg simpl.
    split_l by omega.
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_yield_16_p in HpS.
    destruct HpS as [sp [pc [ax [bx HS]]]].
    exists (M, RF_update (RF_update R SP (sp - 2)) BX bx, fl, isr).
    apply nextS_popw with (RF_update R SP (sp - 2)) bx; trivial.
    unfold popW.
    destruct HS as [HM [Hsp [Hspgt [Hif Hisr]]]].
    simplS in * .
    rewrite Hsp.
    split.
      sep on.
      sep lift 3 in HM.
      sep trivial.
    arith_rep (sp - 4 + 2) with (sp - 2).
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_yield_16_p, L_yield_17_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [sp [pc [ax [bx HS]]]].
    exists sp; exists pc; exists ax.
    apply Hpp' with bx S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_yield_16_g, L_yield_17_g in * .
  intros mP sp pc ax bx HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS'S'' as [HAX [HBX [HCX [HDX [HDI [HSI [HBP HSP]]]]]]].
  destruct HS as [HM [Hsp _]].
  simplS in * .
  destruct HR' as [bxx [Hlds HR']].
  assert (Heq : bx = bxx).
    sep on.
    sep lift 3 in HM.
    rewrite Hsp in Hlds.
    sep inj.
    auto.
  subst ax bxx sp R' fl' isr'.
  reg simpl.
  repeat split_l; trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_yield_16_p in HS.
    destruct HS as [sp [pc [ax [bx HS]]]].
    unfold Ast_yield_16 in HS.
    destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_yield_16_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP sp pc ax bx HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_yield_16 in HS.
  simplS in * .
  destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma NH_WFcmd_yield_17_popw :
  NH_WFcmd (SCAP_consCT 
    (L_yield+17+1)
      (Ast L_yield_18_p L_yield_18_g tag_yield) psi_yield)
    (L_yield+17)
      (Ast L_yield_17_p L_yield_17_g tag_yield) 
    (comml (_popw AX)).
Proof.
  simpl.
  split.
  exists L_yield_18_p; exists L_yield_18_g.
  split_l (* lookup *).
  assert (Hpp' : forall sp pc ax mP S S',
    Ast_yield_17 sp pc ax mP S
    -> nextS (L_yield+17) (comml (_popw AX)) S S'
    -> Ast_yield_18 sp pc mP S').
    intros sp pc ax mP S S' HS HSS'.
    unfold Ast_yield_17, Ast_yield_18 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct (nextS_inversion HSS') as [HM' [HR' [Hfl' Hisr']]].
    destruct HR' as [axx [Hlds HR']].
    simplS in * .
    destruct HS as [HM [Hsp [Hspgt' [Hif Hisr]]]].
    assert (Heq : ax = axx).
      sep on.
      sep lift 3 in HM.
      rewrite Hsp in Hlds.
      sep inj.
      auto.
    subst axx.
    subst M' fl' R' isr'.
    split.
      sep on.
      sep lift 2.
      sep lift 2 in HM.
      sep stksplit.
      sep matchH HM with 6.
    rewrite Hsp.
    reg simpl.
    split_l by omega.
    auto.
  split.
  (* enable *)
    unfold enable.
    intros S HpS.
    destruct S as [[[M R] fl] isr].
    unfold L_yield_17_p in HpS.
    destruct HpS as [sp [pc [ax HS]]].
    exists (M, RF_update (RF_update R SP sp) AX ax, fl, isr).
    apply nextS_popw with (RF_update R SP sp) ax; trivial.
    unfold popW.
    destruct HS as [HM [Hsp [Hspgt [Hif Hisr]]]].
    simplS in * .
    rewrite Hsp.
    split.
      sep on.
      sep lift 3 in HM.
      sep trivial.
    arith_rep (sp - 2 + 2) with sp.
  split.
  (* p ==> p' *)
    unfold linkPG, conjPG, conjG.
    unfold L_yield_17_p, L_yield_18_p.
    intros S' H.
    destruct H as [S H].
    destruct H as [HS HnextS].
    destruct HS as [sp [pc [ax HS]]].
    exists sp; exists pc.
    apply Hpp' with ax S; trivial.
  (* g' ==> g *)
  intros S S'' H.
  unfold linkPG, conjPG, conjG in H.
  destruct H as [S' H].
  destruct H as [[HpS HnextSS'] HgS'S''].
  unfold L_yield_17_g, L_yield_18_g in * .
  intros mP sp pc ax HS.
  substH Hpp' with (Hpp' _ _ _ _ _ _ HS HnextSS').
  substH HgS'S'' with (HgS'S'' _ _ _ Hpp').
  destruct HgS'S'' as [HS'' HS'S''].
  split_l.
  destruct S as [[[M R] fl] isr].
  destruct S' as [[[M' R'] fl'] isr'].
  destruct S'' as [[[M'' R''] fl''] isr''].
  destruct (nextS_inversion HnextSS') as [HM' [HR' [Hfl' Hisr']]].
  destruct HS as [HM [Hsp HS]].
  simplS in * .
  destruct HR' as [axx [Hlds HR']].
  assert (Heq : ax = axx).
    sep on.
    sep lift 3 in HM.
    rewrite Hsp in Hlds.
    sep inj.
    auto.
  subst axx R'' R' fl' isr'.
  reg simpl.
  repeat split_l.
  destruct HS as [Hspgt _].
  rewrite Hsp.
  omega.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_yield_17_p in HS.
    destruct HS as [sp [pc [ax HS]]].
    unfold Ast_yield_17 in HS.
    destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_yield_17_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP sp pc ax HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_yield_17 in HS.
  simplS in * .
  destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma NH_WFcmd_yield_18_sti :
  NH_WFcmd (SCAP_consCT 
    (L_yield+18+1)
      (Ast L_yield_19_p L_yield_19_g tag_yield) psi_yield)
    (L_yield+18)
      (Ast L_yield_18_p L_yield_18_g tag_yield) 
    (comml (_sti)).
Proof.
  simpl.
  split.
  exists L_yield_19_p; exists L_yield_19_g.
  split_l. (* lookup *)
  assert (Hpp': forall sp pc mP S S', 
    Ast_yield_18 sp pc mP S
    -> gsti S S'
    -> Ast_yield_19 sp pc mP S').
    intros sp pc mP S S' HS HSS'.
    unfold linkPG, conjPG, conjG.
    unfold Ast_yield_18, gsti, Ast_yield_19 in * .
    destruct S as [[[M R] fl] isr].
    destruct S' as [[[M' R'] fl'] isr'].
    destruct HS as [HM [Hsp [Hspgt Hitr]]].
    destruct HSS' as [Hshift [Hif' [Hisr' HR']]].
    assert (Hshift' : prepost (INV0 *^* INV1) Aemp
      (M, R, fl, isr) (M', R', fl', isr')).
      apply (prepost_weaken_pre (mP' := INV0 *^* INV1) Hshift).
      intros Mx HMx.
      apply (choice_elim_right HMx).
      unfold itr_enabled.
      intros [H1 H2].
      simplS in * .
      destruct Hitr as [Hitr1 Hitr2].
      rewrite Hitr1 in H1; discriminate H1.
    clear Hshift; rename Hshift' into Hshift.
    split.
      sep on.
      assert (HMx : S_Mem (M, R, fl, isr) ||- 
        (INV0 *^* INV1) *^* (stkSpacePrec sp _STK_SIZE_yield *^* sp |->w pc *^* mP)).
        sep matchH HM with 4.
      substH HM with (prepost_elim Hshift Prec01 HMx).
      sep on.
      sep matchH HM with 6.
    simplS in * .
    subst R' isr'.
    split_l.
    split_l.
    destruct Hitr; trivial.
  split.
  (* enable *)
     unfold enable_sti.
     intros S HS.
     unfold L_yield_18_p in HS.
     unfold pTrue, pmstar, choice.
     destruct HS as [sp [pc HS]].
     unfold Ast_yield_18 in HS.
     destruct HS as [HM [Hsp [Hspg [Hif Hisr]]]].
     destruct S as [[[M R] fl] isr].
     simplS in * .
     assert (HMx : ((stkSpacePrec sp _STK_SIZE_yield *^* sp |->w pc *^* mTrue)
       *^* (INV0 *^* INV1)) M).
       sep on.
       sep matchH HM with 4.
     destruct HMx as [M1 [M2 [HJM [HM1 HM2]]]].
     exists M1; exists M2; split_l.
     split_l.
     right.
     split_r.
     intros [H1 H2].
     simplS in * .
     rewrite Hif in H1; discriminate H1.
  split.
  (* p ==> p' *)
  unfold L_yield_18_p, L_yield_19_p.
  intros S' H.
  unfold linkPG in H.
  destruct H as [S [HS HSS']].
  destruct HS as [sp [pc HpS]].
  exists sp; exists pc.
  apply Hpp' with S; trivial.
  (* g' ==> g *)
  unfold L_yield_18_g, L_yield_19_g.
  unfold linkPG, conjPG, conjG.
  intros S S'' H.
  destruct H as [S' H].
  destruct H as [[HpS HSS'] HS'S''].
  intros mP sp pc HS.
  substH Hpp' with (Hpp' sp pc mP S S' HS HSS').
  substH HS'S'' with (HS'S'' mP sp pc Hpp').
  destruct HS'S'' as [HS'' HR''].
  split_l.
  unfold gsti in HSS'.
  destruct HSS' as [Hshift [Hif' [Hisr' HR']]].
  rewrite HR'.
  rewrite HR''.
  trivial.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_yield_18_p in HS.
    destruct HS as [sp [pc HS]].
    unfold Ast_yield_18 in HS.
    destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    destruct IHS as [IHif IHisr].
    rewrite Hif in IHif.
    discriminate IHif.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_yield_18_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  unfold Ast_yield_18 in HS.
  simplS in * .
  destruct HS as [HM [HSP [Hsp [Hif Hisr]]]].
  destruct IHS as [IHif IHisr].
  rewrite Hif in IHif.
  discriminate IHif.
Qed.

Lemma NH_WFcmd_yield_19_retw :
  NH_WFcmd psi_yield
    (L_yield+19)
      (Ast L_yield_19_p L_yield_19_g tag_yield) 
    (comms (_retw)).
Proof.
  simpl.
  split.
  split_l.
  intros S HS.
  unfold L_yield_19_g.
  unfold Ast_yield_19, Ast_yield_post.
  intros mP sp pc H.
  destruct H as [H1 [H2 [H3 H4]]].
  split_r.
  split_l.
  split_lr.
  split.
  (* Prd_Itr_Stk *)
    unfold Prd_Itr_Stk.
    intros S HS IHS.
    unfold L_yield_19_p in HS.
    destruct HS as [sp [pc HS]].
    unfold Ast_yield_19 in HS.
    destruct HS as [HM [HSP [Hif Hisr]]].
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in IHS.
    simplS in * .
    rewrite HSP.
    assert (Hsp : sp >= _STK_SIZE_yield).
      assert (stkSpace sp _STK_SIZE_yield M).
      sep on.
      sep lift 0 in HM.
      sep trivial.
      destruct H; trivial.
    sep on.
    sep lift 0 in HM.
    sep stksplit (IH_Stk + 6) in HM.
    sep lift 0 in HM.
    destruct HM as [M1 [M2 [HJM [HM1 HM2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    split_l.
    intros M1' HM1'.
    unfold L_yield_19_p.
    exists sp; exists pc.
    unfold Ast_yield_19.
    simplS in * .
    rsplit.
      repeat split_l; trivial.
    sep on.
    sep lift 0.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      sep off.
      split.
        disjoint rep_l M1.
        eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    split_l.
    sep on.
    sep matchH HM2 with 4.
  (* Grt_Itr_Stk *)
  unfold Grt_Itr_Stk.
  intros S S' HSS' IHS.
  unfold L_yield_19_g in * .
  intros M1 M2 M2' HJM HM2 HM2'.
  intros mP sp pc HS.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in IHS.
  apply HSS'.
  clear HSS'.
  unfold Ast_yield_19 in * .
  simplS in * .
  destruct HS as [HM [HSP [Hif Hisr]]].
  rsplit.
    repeat split_l; trivial.
  sep on.
  assert (Hsp : sp >= _STK_SIZE_yield).
    assert (stkSpace sp _STK_SIZE_yield (MemMod.merge M1 M2')).
    sep on.
    sep lift 0 in HM.
    sep trivial.
    destruct H; trivial.
  sep lift 0 in HM.
  sep stksplit (IH_Stk + 6) in HM.
  sep lift 0 in HM.
  destruct HM as [M2x [M1x [HJM12 [HM2x HM1x]]]].
  sep lift 0.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  assert (M2x = M2' /\ M1x = M1).
    sep off.
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      eapply stkSpacePrec_stkSpacePrec_eqDom; eauto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    rewrite HSP in HM2'.
    assert (Hx := prec_disjUnion_inv H H0 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  rewrite HSP in HM2.
  split_l.
  sep on.
  sep matchH HM1x with 3.
Qed.

Lemma NH_WFiseq_yield :
  NH_WFiseq psi_yield L_yield spec_yield I_yield.
Proof.
unfold I_yield, spec_yield.
apply nh_isq_seq with L_yield_1_p L_yield_1_g.
apply NH_WFcmd_yield_0_cli.
apply nh_isq_seq with L_yield_2_p L_yield_2_g.
apply NH_WFcmd_yield_1_pushwp.
assert (H : L_yield+1+1 = L_yield+2); [omega | rewrite H; clear H].
apply nh_isq_seq with L_yield_3_p L_yield_3_g.
apply NH_WFcmd_yield_2_pushwp.
assert (H : L_yield+2+1 = L_yield+3); [omega | rewrite H; clear H].
apply nh_isq_seq with L_yield_4_p L_yield_4_g.
apply NH_WFcmd_yield_3_pushwp.
assert (H : L_yield+3+1 = L_yield+4); [omega | rewrite H; clear H].
apply nh_isq_seq with L_yield_5_p L_yield_5_g.
apply NH_WFcmd_yield_4_pushwp.
assert (H : L_yield+4+1 = L_yield+5); [omega | rewrite H; clear H].
apply nh_isq_seq with L_yield_6_p L_yield_6_g.
apply NH_WFcmd_yield_5_pushwp.
assert (H : L_yield+5+1 = L_yield+6); [omega | rewrite H; clear H].
apply nh_isq_seq with L_yield_7_p L_yield_7_g.
apply NH_WFcmd_yield_6_pushwp.
assert (H : L_yield+6+1 = L_yield+7); [omega | rewrite H; clear H].
apply nh_isq_seq with L_yield_8_p L_yield_8_g.
apply NH_WFcmd_yield_7_pushwp.
assert (H : L_yield+7+1 = L_yield+8); [omega | rewrite H; clear H].
apply nh_isq_seq with L_yield_9_p L_yield_9_g.
apply NH_WFcmd_yield_8_pushfw.
assert (H : L_yield+8+1 = L_yield+9); [omega | rewrite H; clear H].
apply nh_isq_seq with L_yield_10_p L_yield_10_g.
apply NH_WFcmd_yield_9_call.
assert (H : L_yield+9+1 = L_yield+10); [omega | rewrite H; clear H].
apply nh_isq_seq with L_yield_11_p L_yield_11_g.
apply NH_WFcmd_yield_10_popfw.
assert (H : L_yield+10+1 = L_yield+11); [omega | rewrite H; clear H].
apply nh_isq_seq with L_yield_12_p L_yield_12_g.
apply NH_WFcmd_yield_11_popw.
assert (H : L_yield+11+1 = L_yield+12); [omega | rewrite H; clear H].
apply nh_isq_seq with L_yield_13_p L_yield_13_g.
apply NH_WFcmd_yield_12_popw.
assert (H : L_yield+12+1 = L_yield+13); [omega | rewrite H; clear H].
apply nh_isq_seq with L_yield_14_p L_yield_14_g.
apply NH_WFcmd_yield_13_popw.
assert (H : L_yield+13+1 = L_yield+14); [omega | rewrite H; clear H].
apply nh_isq_seq with L_yield_15_p L_yield_15_g.
apply NH_WFcmd_yield_14_popw.
assert (H : L_yield+14+1 = L_yield+15); [omega | rewrite H; clear H].
apply nh_isq_seq with L_yield_16_p L_yield_16_g.
apply NH_WFcmd_yield_15_popw.
assert (H : L_yield+15+1 = L_yield+16); [omega | rewrite H; clear H].
apply nh_isq_seq with L_yield_17_p L_yield_17_g.
apply NH_WFcmd_yield_16_popw.
assert (H : L_yield+16+1 = L_yield+17); [omega | rewrite H; clear H].
apply nh_isq_seq with L_yield_18_p L_yield_18_g.
apply NH_WFcmd_yield_17_popw.
assert (H : L_yield+17+1 = L_yield+18); [omega | rewrite H; clear H].
apply nh_isq_seq with L_yield_19_p L_yield_19_g.
apply NH_WFcmd_yield_18_sti.
assert (H : L_yield+18+1 = L_yield+19); [omega | rewrite H; clear H].
apply nh_isq_end.
apply NH_WFcmd_yield_19_retw.
Qed.
