(***************************************************************)
(* Coq code for AIM                                            *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

Add LoadPath "..".
Add LoadPath "../CAP0".
Add LoadPath "../SCAP".

Require Import Coqlib.

Require Import Arith.
Require Import Machine.
Require Import Assertion.
Require Import Operational_Semantics.
Require Import AIM.
Require Import SCAP.
Require Import QLib.
Require Import Linking.

Require Import Kernel.

Definition Ast_yield_pre (sp pc : Label) (mP : MPred) (S : State) : Prop :=
  (stkSpacePrec sp _STK_SIZE_yield
    *^* sp |->w pc
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp
  /\ itr_enabled S.

Definition Ast_yield_post (sp pc : Label) (mP : MPred) (S' : State) : Prop :=
  (stkSpacePrec sp _STK_SIZE_yield
    *^* sp |->w pc
    *^* mP) (S_Mem S')
  /\ itr_enabled S'.
  
Definition L_yield_0_p (S : State) : Prop :=
  exists sp, exists pc, 
    (Ast_yield_pre sp pc mTrue S).

Definition L_yield_0_g (S S' : State) : Prop := forall mP,
  forall sp pc,
    (Ast_yield_pre sp pc mP S)
    -> (Ast_yield_post sp pc mP S')
    /\ S_RF S' = S_RF S.

(* <yield+0>:  cli           *)

Definition Ast_yield_1 (sp pc : Label) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec sp _STK_SIZE_yield
    *^* sp |->w pc
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp
  /\ sp >= _STK_SIZE_yield
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.

Definition L_yield_1_p (S : State) : Prop :=
  exists sp, exists pc, 
    Ast_yield_1 sp pc mTrue S.

Definition L_yield_1_g (S S' : State) : Prop := forall mP,
  forall sp pc,
    (Ast_yield_1 sp pc mP S)
    -> (Ast_yield_post sp pc mP S')
    /\ S_RF S' = S_RF S.

(* <yield+1>:  pushw   %ax   *)

Definition Ast_yield_2 (sp pc ax : nat) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec (sp-2)  (_STK_SIZE_yield - 2)
    *^* sp-2 |->w (S_RF S AX)
    *^* sp   |->w pc
    *^* mP) (S_Mem S)
  /\ S_RF S AX = ax
  /\ S_RF S SP = sp-2
  /\ sp >= _STK_SIZE_yield
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.
  
Definition L_yield_2_p (S : State) : Prop :=
  exists sp, exists pc, exists ax, 
    Ast_yield_2 sp pc ax mTrue S.

Definition L_yield_2_g (S S' : State) : Prop := forall mP,
  forall sp pc ax,
    Ast_yield_2 sp pc ax mP S
    -> Ast_yield_post sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = S_RF S BX
    /\ S_RF S' CX = S_RF S CX
    /\ S_RF S' DX = S_RF S DX
    /\ S_RF S' DI = S_RF S DI
    /\ S_RF S' SI = S_RF S SI
    /\ S_RF S' BP = S_RF S BP
    /\ S_RF S' SP = sp.

(* <yield+2>:  pushw   %bx   *)

Definition Ast_yield_3 (sp pc ax bx : nat) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec (sp-4) (_STK_SIZE_yield - 4)
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* mP) (S_Mem S)
  /\ S_RF S BX = bx
  /\ S_RF S SP = sp-4
  /\ sp >= _STK_SIZE_yield
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.
  
Definition L_yield_3_p (S : State) : Prop :=
  exists sp, exists pc, exists ax, exists bx,
    Ast_yield_3 sp pc ax bx mTrue S.

Definition L_yield_3_g (S S' : State) : Prop := forall mP,
  forall sp pc ax bx,
    Ast_yield_3 sp pc ax bx mP S
    -> Ast_yield_post sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = S_RF S CX
    /\ S_RF S' DX = S_RF S DX
    /\ S_RF S' DI = S_RF S DI
    /\ S_RF S' SI = S_RF S SI
    /\ S_RF S' BP = S_RF S BP
    /\ S_RF S' SP = sp.

(* <yield+3>:  pushw   %cx   *)

Definition Ast_yield_4 (sp pc ax bx cx : nat) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec (sp-6) (_STK_SIZE_yield - 6)
    *^* sp-6 |->w cx
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* mP) (S_Mem S)
  /\ S_RF S CX = cx
  /\ S_RF S SP = sp-6
  /\ sp >= _STK_SIZE_yield
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.
  
Definition L_yield_4_p (S : State) : Prop :=
  exists sp, exists pc, exists ax, exists bx, exists cx,
    Ast_yield_4 sp pc ax bx cx mTrue S.

Definition L_yield_4_g (S S' : State) : Prop := forall mP,
  forall sp pc ax bx cx,
    Ast_yield_4 sp pc ax bx cx mP S
    -> Ast_yield_post sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = cx
    /\ S_RF S' DX = S_RF S DX
    /\ S_RF S' DI = S_RF S DI
    /\ S_RF S' SI = S_RF S SI
    /\ S_RF S' BP = S_RF S BP
    /\ S_RF S' SP = sp.

(* <yield+4>:  pushw   %dx   *)

Definition Ast_yield_5 (sp pc ax bx cx dx : nat) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec (sp-8) (_STK_SIZE_yield - 8)
    *^* sp-8 |->w dx
    *^* sp-6 |->w cx
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* mP) (S_Mem S)
  /\ S_RF S DX = dx
  /\ S_RF S SP = sp-8
  /\ sp >= _STK_SIZE_yield
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.
  
Definition L_yield_5_p (S : State) : Prop :=
  exists sp, exists pc, exists ax, exists bx, exists cx, exists dx,
    Ast_yield_5 sp pc ax bx cx dx mTrue S.

Definition L_yield_5_g (S S' : State) : Prop := forall mP,
  forall sp pc ax bx cx dx,
    Ast_yield_5 sp pc ax bx cx dx mP S
    -> Ast_yield_post sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = cx
    /\ S_RF S' DX = dx    
    /\ S_RF S' DI = S_RF S DI
    /\ S_RF S' SI = S_RF S SI
    /\ S_RF S' BP = S_RF S BP
    /\ S_RF S' SP = sp.

(* <yield+5>:  pushw   %di   *)

Definition Ast_yield_6 (sp pc ax bx cx dx di : nat) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec (sp-10) (_STK_SIZE_yield - 10)
    *^* sp-10 |->w di
    *^* sp-8 |->w dx
    *^* sp-6 |->w cx
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* mP) (S_Mem S)
  /\ S_RF S DI = di
  /\ S_RF S SP = sp-10
  /\ sp >= _STK_SIZE_yield
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.
  
Definition L_yield_6_p (S : State) : Prop :=
  exists sp, exists pc, exists ax, exists bx, exists cx, exists dx,
  exists di,
    Ast_yield_6 sp pc ax bx cx dx di mTrue S.

Definition L_yield_6_g (S S' : State) : Prop := forall mP,
  forall sp pc ax bx cx dx di,
    Ast_yield_6 sp pc ax bx cx dx di mP S
    -> Ast_yield_post sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = cx
    /\ S_RF S' DX = dx    
    /\ S_RF S' DI = di
    /\ S_RF S' SI = S_RF S SI
    /\ S_RF S' BP = S_RF S BP
    /\ S_RF S' SP = sp.

(* <yield+6>:  pushw   %si   *)

Definition Ast_yield_7 (sp pc ax bx cx dx di si : nat) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec (sp-12) (_STK_SIZE_yield - 12)
    *^* sp-12 |->w si
    *^* sp-10 |->w di
    *^* sp-8 |->w dx
    *^* sp-6 |->w cx
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* mP) (S_Mem S)
  /\ S_RF S SI = si
  /\ S_RF S SP = sp-12
  /\ sp >= _STK_SIZE_yield
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.
  
Definition L_yield_7_p (S : State) : Prop :=
  exists sp, exists pc, exists ax, exists bx, exists cx, exists dx,
  exists di, exists si,
    Ast_yield_7 sp pc ax bx cx dx di si mTrue S.

Definition L_yield_7_g (S S' : State) : Prop := forall mP,
  forall sp pc ax bx cx dx di si,
    Ast_yield_7 sp pc ax bx cx dx di si mP S
    -> Ast_yield_post sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = cx
    /\ S_RF S' DX = dx    
    /\ S_RF S' DI = di
    /\ S_RF S' SI = si
    /\ S_RF S' BP = S_RF S BP
    /\ S_RF S' SP = sp.

(* <yield+7>:  pushw   %bp   *)

Definition Ast_yield_8 (sp pc ax bx cx dx di si bp : nat) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec (sp-14) (_STK_SIZE_yield - 14)
    *^* sp-14 |->w bp
    *^* sp-12 |->w si
    *^* sp-10 |->w di
    *^* sp-8 |->w dx
    *^* sp-6 |->w cx
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* mP) (S_Mem S)
  /\ S_RF S BP = bp
  /\ S_RF S SP = sp-14
  /\ sp >= _STK_SIZE_yield
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.
  
Definition L_yield_8_p (S : State) : Prop :=
  exists sp, exists pc, exists ax, exists bx, exists cx, exists dx,
  exists di, exists si, exists bp,
    Ast_yield_8 sp pc ax bx cx dx di si bp mTrue S.

Definition L_yield_8_g (S S' : State) : Prop := forall mP,
  forall sp pc ax bx cx dx di si bp,
    Ast_yield_8 sp pc ax bx cx dx di si bp mP S
    -> Ast_yield_post sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = cx
    /\ S_RF S' DX = dx
    /\ S_RF S' DI = di
    /\ S_RF S' SI = si
    /\ S_RF S' BP = bp
    /\ S_RF S' SP = sp.

(* <yield+8>:  pushfw       *)

Definition Ast_yield_9 (sp pc ax bx cx dx di si bp flnum : nat) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec (sp-16) (_STK_SIZE_yield - 16)
    *^* sp-16 |->w flnum
    *^* sp-14 |->w bp
    *^* sp-12 |->w si
    *^* sp-10 |->w di
    *^* sp-8 |->w dx
    *^* sp-6 |->w cx
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp-16
  /\ sp >= _STK_SIZE_yield
  /\ (nat2FR flnum) _IF = false
  /\ S_Flag S = nat2FR flnum
  /\ S_ISR S = false.
  
Definition L_yield_9_p (S : State) : Prop :=
  exists sp, exists pc, exists ax, exists bx, exists cx, exists dx,
  exists di, exists si, exists bp, exists flnum,
    Ast_yield_9 sp pc ax bx cx dx di si bp flnum mTrue S.

Definition L_yield_9_g (S S' : State) : Prop := forall mP,
  forall sp pc ax bx cx dx di si bp flnum,
    Ast_yield_9 sp pc ax bx cx dx di si bp flnum mP S
    -> Ast_yield_post sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = cx
    /\ S_RF S' DX = dx
    /\ S_RF S' DI = di
    /\ S_RF S' SI = si
    /\ S_RF S' BP = bp
    /\ S_RF S' SP = sp.

(* <yield+9>:  call    scheduler *)

Definition Ast_yield_10 (sp pc ax bx cx dx di si bp flnum : nat) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec (sp-16) (_STK_SIZE_yield - 16)
    *^* sp-16 |->w flnum
    *^* sp-14 |->w bp
    *^* sp-12 |->w si
    *^* sp-10 |->w di
    *^* sp-8 |->w dx
    *^* sp-6 |->w cx
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp-16
  /\ sp >= _STK_SIZE_yield
  /\ S_Flag S _IF = false
  /\ (nat2FR flnum) _IF = false
  /\ S_ISR S = false.

Definition L_yield_10_p (S : State) : Prop :=
  exists sp, exists pc, exists ax, exists bx, exists cx, exists dx,
  exists di, exists si, exists bp, exists flnum,
    Ast_yield_10 sp pc ax bx cx dx di si bp flnum mTrue S.

Definition L_yield_10_g (S S' : State) : Prop := forall mP,
  forall sp pc ax bx cx dx di si bp flnum,
    Ast_yield_10 sp pc ax bx cx dx di si bp flnum mP S
    -> Ast_yield_post sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = cx
    /\ S_RF S' DX = dx
    /\ S_RF S' DI = di
    /\ S_RF S' SI = si
    /\ S_RF S' BP = bp
    /\ S_RF S' SP = sp.

(* <yield+10>: popfw *)

Definition Ast_yield_11 (sp pc ax bx cx dx di si bp : nat) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec (sp-14) (_STK_SIZE_yield - 14)
    *^* sp-14 |->w bp
    *^* sp-12 |->w si
    *^* sp-10 |->w di
    *^* sp-8 |->w dx
    *^* sp-6 |->w cx
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp-14
  /\ sp >= _STK_SIZE_yield
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.
  
Definition L_yield_11_p (S : State) : Prop :=
  exists sp, exists pc, exists ax, exists bx, exists cx, exists dx,
  exists di, exists si, exists bp,
    Ast_yield_11 sp pc ax bx cx dx di si bp mTrue S.

Definition L_yield_11_g (S S' : State) : Prop := forall mP,
  forall sp pc ax bx cx dx di si bp,
    Ast_yield_11 sp pc ax bx cx dx di si bp mP S
    -> Ast_yield_post sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = cx
    /\ S_RF S' DX = dx
    /\ S_RF S' DI = di
    /\ S_RF S' SI = si
    /\ S_RF S' BP = bp
    /\ S_RF S' SP = sp.

(* <yield+11>: popw    %bp *)

Definition Ast_yield_12 (sp pc ax bx cx dx di si : nat) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec (sp-12) (_STK_SIZE_yield - 12)
    *^* sp-12 |->w si
    *^* sp-10 |->w di
    *^* sp-8 |->w dx
    *^* sp-6 |->w cx
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp-12
  /\ sp >= _STK_SIZE_yield
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.
  
Definition L_yield_12_p (S : State) : Prop :=
  exists sp, exists pc, exists ax, exists bx, exists cx, exists dx,
  exists di, exists si,
    Ast_yield_12 sp pc ax bx cx dx di si mTrue S.

Definition L_yield_12_g (S S' : State) : Prop := forall mP,
  forall sp pc ax bx cx dx di si,
    Ast_yield_12 sp pc ax bx cx dx di si mP S
    -> Ast_yield_post sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = cx
    /\ S_RF S' DX = dx    
    /\ S_RF S' DI = di
    /\ S_RF S' SI = si
    /\ S_RF S' BP = S_RF S BP
    /\ S_RF S' SP = sp.

(* <yield+12>: popw    %si *)

Definition Ast_yield_13 (sp pc ax bx cx dx di : nat) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec (sp-10) (_STK_SIZE_yield - 10)
    *^* sp-10 |->w di
    *^* sp-8 |->w dx
    *^* sp-6 |->w cx
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp-10
  /\ sp >= _STK_SIZE_yield
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.
  
Definition L_yield_13_p (S : State) : Prop :=
  exists sp, exists pc, exists ax, exists bx, exists cx, exists dx,
  exists di,
    Ast_yield_13 sp pc ax bx cx dx di mTrue S.

Definition L_yield_13_g (S S' : State) : Prop := forall mP,
  forall sp pc ax bx cx dx di,
    Ast_yield_13 sp pc ax bx cx dx di mP S
    -> Ast_yield_post sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = cx
    /\ S_RF S' DX = dx    
    /\ S_RF S' DI = di
    /\ S_RF S' SI = S_RF S SI
    /\ S_RF S' BP = S_RF S BP
    /\ S_RF S' SP = sp.

(* <yield+13>: popw    %di *)

Definition Ast_yield_14 (sp pc ax bx cx dx : nat) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec (sp-8) (_STK_SIZE_yield - 8)
    *^* sp-8 |->w dx
    *^* sp-6 |->w cx
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp-8
  /\ sp >= _STK_SIZE_yield
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.
  
Definition L_yield_14_p (S : State) : Prop :=
  exists sp, exists pc, exists ax, exists bx, exists cx, exists dx,
    Ast_yield_14 sp pc ax bx cx dx mTrue S.

Definition L_yield_14_g (S S' : State) : Prop := forall mP,
  forall sp pc ax bx cx dx,
    Ast_yield_14 sp pc ax bx cx dx mP S
    -> Ast_yield_post sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = cx
    /\ S_RF S' DX = dx    
    /\ S_RF S' DI = S_RF S DI
    /\ S_RF S' SI = S_RF S SI
    /\ S_RF S' BP = S_RF S BP
    /\ S_RF S' SP = sp.

(* <yield+14>: popw    %dx *)

Definition Ast_yield_15 (sp pc ax bx cx : nat) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec (sp-6) (_STK_SIZE_yield - 6)
    *^* sp-6 |->w cx
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp-6
  /\ sp >= _STK_SIZE_yield
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.
  
Definition L_yield_15_p (S : State) : Prop :=
  exists sp, exists pc, exists ax, exists bx, exists cx,
    Ast_yield_15 sp pc ax bx cx mTrue S.

Definition L_yield_15_g (S S' : State) : Prop := forall mP,
  forall sp pc ax bx cx,
    Ast_yield_15 sp pc ax bx cx mP S
    -> Ast_yield_post sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = cx
    /\ S_RF S' DX = S_RF S DX
    /\ S_RF S' DI = S_RF S DI
    /\ S_RF S' SI = S_RF S SI
    /\ S_RF S' BP = S_RF S BP
    /\ S_RF S' SP = sp.

(* <yield+15>: popw    %cx *)

Definition Ast_yield_16 (sp pc ax bx : nat) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec (sp-4) (_STK_SIZE_yield - 4)
    *^* sp-4 |->w bx
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp-4
  /\ sp >= _STK_SIZE_yield
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.
  
Definition L_yield_16_p (S : State) : Prop :=
  exists sp, exists pc, exists ax, exists bx,
    Ast_yield_16 sp pc ax bx mTrue S.

Definition L_yield_16_g (S S' : State) : Prop := forall mP,
  forall sp pc ax bx,
    Ast_yield_16 sp pc ax bx mP S
    -> Ast_yield_post sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = bx
    /\ S_RF S' CX = S_RF S CX
    /\ S_RF S' DX = S_RF S DX
    /\ S_RF S' DI = S_RF S DI
    /\ S_RF S' SI = S_RF S SI
    /\ S_RF S' BP = S_RF S BP
    /\ S_RF S' SP = sp.

(* <yield+16>: popw    %bx *)

Definition Ast_yield_17 (sp pc ax : nat) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec (sp-2)  (_STK_SIZE_yield - 2)
    *^* sp-2 |->w ax
    *^* sp   |->w pc
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp-2
  /\ sp >= _STK_SIZE_yield
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.
  
Definition L_yield_17_p (S : State) : Prop :=
  exists sp, exists pc, exists ax, 
    Ast_yield_17 sp pc ax mTrue S.

Definition L_yield_17_g (S S' : State) : Prop := forall mP,
  forall sp pc ax,
    Ast_yield_17 sp pc ax mP S
    -> Ast_yield_post sp pc mP S'
    /\ S_RF S' AX = ax
    /\ S_RF S' BX = S_RF S BX
    /\ S_RF S' CX = S_RF S CX
    /\ S_RF S' DX = S_RF S DX
    /\ S_RF S' DI = S_RF S DI
    /\ S_RF S' SI = S_RF S SI
    /\ S_RF S' BP = S_RF S BP
    /\ S_RF S' SP = sp.

(* <yield+17>: popw    %ax *)

Definition Ast_yield_18 (sp pc : Label) (mP : MPred) (S : State) : Prop :=
  (INV1 *^* INV0 
    *^* stkSpacePrec sp _STK_SIZE_yield
    *^* sp |->w pc
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp
  /\ sp >= _STK_SIZE_yield
  /\ S_Flag S _IF = false
  /\ S_ISR S = false.

Definition L_yield_18_p (S : State) : Prop :=
  exists sp, exists pc, 
    Ast_yield_18 sp pc mTrue S.

Definition L_yield_18_g (S S' : State) : Prop := forall mP,
  forall sp pc,
    (Ast_yield_18 sp pc mP S)
    -> (Ast_yield_post sp pc mP S')
    /\ S_RF S' = S_RF S.

(* <yield+18>: sti *)

Definition Ast_yield_19 (sp pc : Label) (mP : MPred) (S : State) : Prop :=
  (stkSpacePrec sp _STK_SIZE_yield
    *^* sp |->w pc
    *^* mP) (S_Mem S)
  /\ S_RF S SP = sp
  /\ itr_enabled S.
  
Definition L_yield_19_p (S : State) : Prop :=
  exists sp, exists pc, 
    (Ast_yield_19 sp pc mTrue S).

Definition L_yield_19_g (S S' : State) : Prop := forall mP,
  forall sp pc,
    (Ast_yield_19 sp pc mP S)
    -> (Ast_yield_post sp pc mP S')
    /\ S_RF S' = S_RF S.

(* <yield+19>: retw *)

Definition tag_yield := 1.

Definition p_yield := L_yield_0_p.

Definition g_yield := L_yield_0_g.

Definition spec_yield := Ast p_yield g_yield tag_yield.

Definition psi_yield  := SCAP_emptCT.
