(***************************************************************)
(* Coq code for AIM                                            *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

Add LoadPath "..".
Add LoadPath "../CAP0".
Add LoadPath "../SCAP".
Add LoadPath "../miniOS".

Require Import Coqlib.

Require Import List.
Require Import Arith.
Require Import Machine.
Require Import Assertion.
Require Import Assertion_Lemmas.
Require Import Assertion_Lemmas4.
Require Import Assertion_Lemmas5.
Require Import Operational_Semantics.
Require Import OpSem_Lemmas.
Require Import OpSem_Lemmas2.
Require Import Reg.
Require Import Mem_Lemmas.

Require Import AIM.
Require Import QLib.

Require Import SCAP.
Require Import SCAP_READY.
Require Import Linking.
Require Import Linking_Lemmas.
Require Import Omega.
Require Import Assembly.
Require Import SyncUtil.
Require Import Arith_Lemmas.


(* NOTE: in List lib, there is also a definition of "seq"   *)
(* so we have to import List earlier than Machine to shadow *)
(* the "seq" defined in List                                *)


(* the stack space required by l_release *)
Definition stk_l_release := 6 + _UNBLK_Stk.

Definition l_release_MPost (sp l pc : Label) (M : Mem) : Prop :=
    ((sp |->w pc) *^* (sp + 2 |->w l) *^* stkSpacePrec sp stk_l_release
       *^* INV1 *^* INV0) M /\ In l lockSet_H.

Definition l_release_MPre (sp l pc: Label) : MPred :=
  (sp |->w pc) *^* (sp + 2 |->w l) *^* stkSpacePrec sp stk_l_release
       *^* INV1 *^* INV0 *^* Gamma l.

Definition l_release_MPre_tmp (sp l pc: Label) : MPred :=
    (l_release_MPost sp l pc) *^* (Gamma l).

Definition l_release_Prd_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
      R SP = sp
      /\ R BP = bp
      /\ ((l_release_MPre sp l pc) *^* mP) M
      /\ In l lockSet_H
      /\ fl _IF = false
      /\ isr = false
  end.


Definition l_release_Prd_aux_tmp (sp l pc bp : Label) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
      R SP = sp
      /\ R BP = bp
      /\ (l_release_MPre_tmp sp l pc) M
      /\ fl _IF = false
      /\ isr = false
  end.

Definition l_release_Prd (S : State) : Prop :=
  exists sp, exists l, exists pc, exists bp, exists mP, l_release_Prd_aux sp l pc bp mP S.

Definition l_release_Prd_tmp (S : State) : Prop :=
  imprec (fun S => exists sp, exists l, exists pc, exists bp, l_release_Prd_aux_tmp sp l pc bp S) S.

Lemma adapt_Prd_0:
forall S, l_release_Prd S -> l_release_Prd_tmp S.
Proof.
  intros.
  induction S as (((M, R), fl), isr).
  unfold l_release_Prd in H.
  unfold l_release_Prd_tmp.
  unfold l_release_Prd_aux in H.
  InvertAll.
  destruct H1 as [M1 [M2 [HMsplit [Hmem Hframe]]]].
  unfold imprec. unfold pmstar.
  exists M1. exists M2. split_l.
  unfold mTrue. split_r.
  exists x. exists x0. exists x1. exists x2.
  unfold l_release_Prd_aux_tmp.
  UnfoldStateAll. split_l. split_l.
  split; auto.
  unfold l_release_MPre_tmp.
  unfold l_release_MPre in Hmem.
  unfold l_release_MPost.
  apply sep_pure_b.
  split; auto.
Qed.
 
Definition l_release_Post_aux (sp l pc bp : Label) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
      R SP = sp
      /\ R BP = bp
      /\ (l_release_MPost sp l pc) M
      /\ fl _IF = false
      /\ isr = false
  end.

Definition l_release_Post_aux_0 (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
      R SP = sp
      /\ R BP = bp
      /\ ((l_release_MPost sp l pc) *^* mP) M
      /\ fl _IF = false
      /\ isr = false
  end.


Definition l_release_grt_tmp (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ (l_release_Prd_aux_tmp sp l pc bp) (l_release_Post_aux sp l pc bp) S S'.

Definition l_release_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (l_release_Prd_aux sp l pc bp) (l_release_Post_aux_0 sp l pc bp) S S'.

Lemma adapt_Grt_0:
forall S S', 
  l_release_grt_tmp S S' -> l_release_grt S S'.
Proof.
  unfold l_release_grt_tmp.
  unfold l_release_grt.
  intros.
  unfold leadstoPQ2.
  intros.
  assert (leadstoPQ (l_release_Prd_aux_tmp sp l pc bp)
        (l_release_Post_aux sp l pc bp) S S').
  apply H.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  unfold leadstoPQ in H1.
  unfold l_release_Prd_aux in H0.
  InvertAll.
  destruct H3 as [M1 [M2 [HMsplit [Hmem Hframe]]]].
  assert (exists M1' : MemMod.Map,
         MemMod.disjUnion M1' M2 M' /\
         l_release_Post_aux sp l pc bp (M1', R', fl', isr')).
  apply H1 with M1; auto.
  unfold l_release_Prd_aux_tmp.
  split_l. split_l. split; auto.
  unfold l_release_MPre_tmp.
  unfold l_release_MPost.
  apply sep_pure_b. split; auto.
  InvertAll.
  unfold l_release_Post_aux in H8.
  InvertAll.
  unfold l_release_Post_aux_0.
  split_l. split_l. split; auto.
  exists x. exists M2. split; auto.
Qed.

Definition l_rel_done_MPre (sp l pc : Label) (bp : int) : MPred :=
    (sp |->w bp) *^* (sp+2 |->w pc) *^* (sp + 4 |->w l) *^* stkSpacePrec sp (stk_l_release - 2)
       *^* INV1 *^* INV0.

Definition l_rel_done_Prd_aux (sp l pc : Label) (bp : int) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R SP = sp
        /\ (l_rel_done_MPre (R SP) l pc bp *^* mP) M
        /\ In l lockSet_H
        /\ fl _IF = false
        /\ isr = false
  end.

Definition l_rel_done_Prd (S : State) : Prop :=
  exists sp, exists l, exists pc, exists bp, exists mP, l_rel_done_Prd_aux sp l pc bp mP S.


Definition l_rel_done_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (l_rel_done_Prd_aux sp l pc bp) (l_release_Post_aux_0 (sp+2) l pc bp) S S'.


Definition l_rel_done_spec := Ast l_rel_done_Prd l_rel_done_grt 1.

Definition psi := SCAP_consCT l_rel_done l_rel_done_spec SCAP_emptCT.

Lemma l_rel_done_in_psi:
  SCAP_lookupCT psi l_rel_done l_rel_done_spec.
Proof.
  unfold SCAP_lookupCT.
  unfold psi.
  unfold SCAP_consCT.
  unfold SPSpec.lookup.
  simpl. left. auto.
Qed.

Definition l_release_spec := Ast l_release_Prd l_release_grt 1.


(* spec for the next instruction *)

Definition l_release1_MPre (sp l pc bp : Label) :=
  fun (M : Mem) =>
    l_release_MPre_tmp (sp + 2) l pc M /\ (sp |~>w bp) M.

Definition l_release1_Prd_aux (sp l pc bp : Label) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R SP = sp
        /\ R BP = bp
        /\ l_release1_MPre (R SP) l pc (R BP) M
        /\ fl _IF = false
        /\ isr = false
  end.

Definition l_release1_Prd (S : State) :=
  imprec (fun S => exists sp, exists l, exists pc, exists bp, l_release1_Prd_aux sp l pc bp S) S.



Definition l_release1_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ (l_release1_Prd_aux sp l pc bp) (l_release_Post_aux (sp +2) l pc bp) S S'.

Lemma l_release_MPost_stkSpace:
  forall M sp l pc m, (l_release_MPost sp l pc *^* m) M -> stkSpace sp stk_l_release M.
Proof.
  intros.
  unfold l_release_MPost in H.
  assert ((sp |->w pc *^* sp + 2 |->w l *^* stkSpacePrec sp stk_l_release *^* INV1 *^* INV0 *^* m) M).
  unfold mstar at 1. unfold mstar at 1 in H. InvertAll.
  exists x. exists x0. split; auto.
  sep on.
  sep lift 2 in H0.
  sep trivial.
Qed.

Lemma l_release_Prd_aux_stkSpace:
  forall sp l pc bp S,
   l_release_Prd_aux_tmp sp l pc bp S -> stkSpace (S_RF S SP) stk_l_release (S_Mem S).
Proof.
  intros.
  unfold l_release_Prd_aux_tmp in H.
  induction S as (((M, R), fl), isr). UnfoldStateAll.
  InvertAll.
  unfold l_release_MPre_tmp in H1.
  apply l_release_MPost_stkSpace with l pc (Gamma l).
  rewrite H0. auto.
Qed.





Lemma stk_l_rel_gt_2:
  stk_l_release >= 2.
Proof.
  unfold stk_l_release. omega.
Qed.
Hint Resolve stk_l_rel_gt_2.

Lemma stk_l_rel_N2_gt_2:
  stk_l_release - 2 >= 2.
Proof.
  unfold stk_l_release.
  omega.
Qed.
Hint Resolve stk_l_rel_N2_gt_2.

Lemma stk_l_rel_match_UNBLK:
stk_l_release - 2 - 2 >=  _UNBLK_Stk + 2.
Proof.
  unfold stk_l_release. omega.
Qed.
Hint Resolve stk_l_rel_match_UNBLK.

Lemma l_release_step1_Prd:
forall sp l pc bp, pimp ((l_release_Prd_aux_tmp sp l pc bp) |> nextS l_release (comml (_pushwp (Opr_reg BP))))
                 (l_release1_Prd_aux (sp - 2) l pc bp).
Proof.
  intros.
  unfold pimp. intros.
  unfold linkPG in H. InvertAll.
  induction x as (((M0, R0), fl0), isr0).
  induction S as (((M, R), fl), isr).
  unfold l_release1_Prd_aux. unfold l_release1_MPre.
  unfold l_release_Prd_aux_tmp in H. InvertAll.
  destruct (nextS_inversion H1) as [HM' [HR' [Hfl' Hisr']]].
  rewrite H0 in HR'.
  rewrite HR'. split. reg simpl. auto.
  split. reg simpl. auto.
  split.
  unfold l_release_MPre_tmp.
  unfold l_release_MPre_tmp in H2.
  unfold l_release_MPost.
  unfold l_release_MPost in H2.
  substH H2 with (sep_pure_a 
       (sp |->w pc *^* sp + 2 |->w l *^* stkSpacePrec sp stk_l_release *^*
          INV1 *^* INV0) (In l lockSet_H) (Gamma l) M0 H2).
  split.
  apply sep_pure_b.
  inversion_clear H2. split; auto.
  InvertAll. 
  substH H9 with (MemUpd16_MemUpdW H9).
  destruct H9 as [junk H9]. simpl in H9.
  subst M.
  sep on.
  rewrite H. reg simpl.
  rewrite H0.
  rewrite H0 in H2.
  assert (sp - 2 + 2 = sp + 2 - 2). auto.
  assert (sp + 2 - 2 = sp). auto.
  rewrite H9 in H7. rewrite H7.
  sep lift 2.
  sep stk upd.
  sep lift 2 in H4.
  auto.

  reg simpl. rewrite H.
  InvertAll. simpl in H9.
  rewrite H0 in H9. rewrite H in H9.
  substH H9 with (MemUpd16_MemUpdW H9).
  destruct H9 as [junk H9]. subst M.
  sep on.
  sep upd.
  sep lift 2 in H4.
  assert (stk_l_release >= 2). auto.
  sep stksplit in H4.
  sep lift 1 in H4.
  sep_blur.
  subst fl isr. auto.
Qed.


Lemma l_release_sp_gt2:
  forall sp l pc bp S,
    l_release_Prd_aux_tmp sp l pc bp S -> sp - 2 + 2 = sp.
Proof.
  intros.
  unfold l_release_Prd_aux_tmp in H.
  induction S as (((M, R), fl), isr).
  unfold l_release_MPre_tmp in H. unfold l_release_MPost in H.
  InvertAll.
  substH H1 with (sep_pure_a 
       (sp |->w pc *^* sp + 2 |->w l *^* stkSpacePrec sp stk_l_release *^*
          INV1 *^* INV0) (In l lockSet_H) (Gamma l) M H1).
  inversion_clear H1.
  sep on.
  sep lift 2 in H3.
  assert (sp >= stk_l_release).
  apply stkSpacePrec_sp_ge_sz1 with M (INV1 *^* (INV0 *^* Gamma l) *^* (sp |->w pc *^* sp + 2 |->w l));
  auto.
  assert (stk_l_release >= 2). auto.
  assert (sp >= 2). apply ge_trans with stk_l_release; auto.
  assert (sp - 2 + 2 = sp + 2 - 2). auto.
  rewrite H8. auto.
Qed.
  
Lemma l_release_Prd_IF_false:
 forall S : State, l_release_Prd S -> S_Flag S _IF = false.
Proof.
    unfold l_release1_Prd.
    intros.
    induction S as (((M, R), fl), isr).
    unfold l_release_Prd in H.
    unfold l_release_Prd_aux in H.
    InvertAll.
    UnfoldStateAll. auto.
Qed.


Lemma l_release_Grt_Itr_Stk:
    Grt_Itr_Stk l_release_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  unfold l_release_grt.
  unfold leadstoPQ2.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  UnfoldStateAll.
  intros.
  unfold l_release_Prd_aux in H4.
  InvertAll.
  unfold itr_enabled in H0.
  UnfoldStateAll.
  inversion_clear H0.
  rewrite H9 in H8. inversion H8.
Qed.

Lemma wff_cmd_l_release1:
  NH_WFcmd
     (SCAP_consCT (l_release + 1) (Ast l_release1_Prd l_release1_grt 1) psi)
     l_release (Ast l_release_Prd l_release_grt 1)
     (comml (_pushwp (Opr_reg BP))).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists l_release1_Prd. exists l_release1_grt.
  split.
  auto.

  assert (forall sp l pc bp, enable (l_release_Prd_aux_tmp sp l pc bp) (nextS l_release (comml (_pushwp (Opr_reg BP))))).
  intros.
  apply enable_pushwp with stk_l_release; auto.
  apply l_release_Prd_aux_stkSpace.

  assert (enable l_release_Prd_tmp (nextS l_release (comml (_pushwp (Opr_reg BP)))) /\
   pimp (l_release_Prd_tmp |> nextS l_release (comml (_pushwp (Opr_reg BP))))
     l_release1_Prd /\
   gimp
     (l_release_Prd_tmp >o nextS l_release (comml (_pushwp (Opr_reg BP)))
      oo l_release1_grt) l_release_grt_tmp).
  unfold l_release_Prd_tmp.
  set (p := (fun S => exists sp, exists l, exists pc, exists bp, l_release_Prd_aux_tmp sp l pc bp S)).
  assert (enable p (nextS l_release (comml (_pushwp (Opr_reg BP))))).
  unfold p.
  apply forall_enable_exists_l_release; auto.
  assert (safe_mono (nextS l_release (comml (_pushwp (Opr_reg BP))))).
    apply safe_mono_pushwp; auto.
  
  split.
    unfold safe_mono in H1. apply H1; auto.

  assert (forall sp l pc bp, pimp ((l_release_Prd_aux_tmp sp l pc bp) |> nextS l_release (comml (_pushwp (Opr_reg BP))))
                 (l_release1_Prd_aux (sp - 2) l pc bp)).
  apply l_release_step1_Prd.

  split.

  unfold l_release1_Prd.
  set (p' := (fun S0 : State =>
         exists sp : nat,
           exists l : nat,
             exists pc : nat,
               exists bp : nat, l_release1_Prd_aux sp l pc bp S0)).
  apply forall_pimp_exists_l_release with l_release_Prd_aux_tmp l_release1_Prd_aux; auto.
    intros. exists (a - 2). exists b. exists c. exists d.
    apply H2; auto.
    apply frame_pushwp.

  apply gimp_l_release with 
    (fun x => x - 2) (fun x => x + 2) l_release_Prd_aux_tmp l_release1_Prd_aux l_release_Post_aux;
  auto.
  apply frame_pushwp.
    unfold gimp. unfold l_release_grt_tmp.
    intros.
    unfold leadstoPQ. intros. unfold leadstoPQ in H3.
    induction S as (((M, R), fl), isr).
    induction S' as (((M', R'), fl'), isr').
    intros.
    assert (exists M1' : MemMod.Map,
               MemMod.disjUnion M1' M2 M' /\
               l_release_Post_aux (sp - 2 + 2) l pc bp (M1', R', fl', isr')).
    apply H3 with M1; auto.
    InvertAll. exists x. split; auto.
    assert (sp - 2 + 2 = sp).
    apply l_release_sp_gt2 with l pc bp (M1, R, fl, isr); auto.
    rewrite H7 in H8. auto.

  InvertAll.
  split.
  unfold enable.
  unfold enable in H1.
  intros.
  apply H1. apply adapt_Prd_0; auto.
  split.
  unfold pimp.
  clear H1. unfold pimp in H0.
  intros.
  apply H0.
  unfold linkPG in H1. unfold linkPG.
  InvertAll. exists x. split; auto.
  apply adapt_Prd_0; auto.
  unfold gimp.
  intros.
  apply adapt_Grt_0.
  unfold gimp in H3.
  apply H3.
  clear H1 H0 H3.
  clear H.
  unfold conjG in *. unfold conjPG in *.
  InvertAll.
  exists x. split; auto.
  split; auto.
  apply adapt_Prd_0; auto.

  split. 
    apply IF_false_Prd_Itr_Stk.
    apply l_release_Prd_IF_false.
    apply l_release_Grt_Itr_Stk.
Qed.

Definition l_release2_Prd_aux (sp l pc bp : Label) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R BP = sp
        /\ l_release1_Prd_aux sp l pc bp (M, RF_update R BP bp, fl, isr)
  end.


Definition l_release2_Prd (S : State) :=
  imprec (fun S => exists sp, exists l, exists pc, exists bp, l_release2_Prd_aux sp l pc bp S) S.

Definition l_release2_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ (l_release2_Prd_aux sp l pc bp) (l_release_Post_aux (sp +2) l pc bp) S S'.

Lemma l_release_step2_Prd:
   forall sp l pc bp : nat,
   pimp
     (l_release1_Prd_aux sp l pc bp
      |> nextS (l_release + 1) (comml (_movwp (Opr_reg SP) BP)))
     (l_release2_Prd_aux sp l pc bp).
Proof.
  intros.
  unfold pimp. unfold l_release1_Prd_aux. unfold l_release2_Prd_aux.
  intros.
  induction S as (((M, R), fl), isr).
  unfold linkPG in H. InvertAll.
  induction x as (((M0, R0), fl0), isr0).
  destruct (nextS_inversion H1).
  InvertAll.
  simpl in H3.
  subst M R.
  split. reg simpl. auto.
  rewrite <- H.
  assert (RF_update (RF_update R0 BP (R0 SP)) BP (R0 BP) = R0).
  reg simpl. 
  apply RF_update_prev; auto.
  rewrite H0.
  rewrite <- H4.
  unfold l_release1_Prd_aux.
  split; auto.
  split; auto.
  split; auto.
  subst fl isr. auto.
Qed.

Lemma l_release1_Grt_Itr_Stk:
    Grt_Itr_Stk l_release1_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  unfold l_release1_grt.
  unfold leadstoPQ.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  UnfoldStateAll.
  intros.
  unfold l_release1_Prd_aux in H5.
  InvertAll.
  unfold itr_enabled in H0.
  UnfoldStateAll.
  inversion_clear H0.
  rewrite H9 in H8. inversion H8.
Qed.

Lemma l_release1_Prd_IF_false:
 forall S : State, l_release1_Prd S -> S_Flag S _IF = false.
Proof.
    unfold l_release1_Prd.
    intros.
    induction S as (((M, R), fl), isr).
    unfold imprec in H. unfold pmstar in H.
    InvertAll.
    unfold l_release1_Prd_aux in H.
    UnfoldStateAll. InvertAll. auto.
Qed.
Hint Resolve l_release1_Prd_IF_false.

Lemma wff_cmd_l_release2:
   NH_WFcmd
     (SCAP_consCT (l_release + 1 + 1) (Ast l_release2_Prd l_release2_grt 1)
        psi) (l_release + 1) (Ast l_release1_Prd l_release1_grt 1)
     (comml (_movwp (Opr_reg SP) BP)).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists l_release2_Prd. exists l_release2_grt.
  split. auto.
  split; auto. (* enable *)

  assert (forall sp l pc bp, pimp ((l_release1_Prd_aux sp l pc bp) |> nextS (l_release+1) (comml (_movwp (Opr_reg SP) BP)))
                 (l_release2_Prd_aux sp l pc bp)).
  apply l_release_step2_Prd.

  unfold l_release1_Prd.
  set (p := (fun S0 : State =>
         exists sp : nat,
           exists l : nat,
             exists pc : nat,
               exists bp : nat, l_release1_Prd_aux sp l pc bp S0)).

  split.
  unfold l_release2_Prd.
  set (p' := (fun S0 : State =>
         exists sp : nat,
           exists l : nat,
             exists pc : nat,
               exists bp : nat, l_release2_Prd_aux sp l pc bp S0)).
  apply frame_rule_1; auto.

    unfold pimp. intros.
    unfold linkPG in H0.
    unfold p in H0.
    InvertAll. rename x into S0.
    rename x0 into sp. rename x1 into l. rename x2 into pc. rename x3 into bp.
    unfold p'. exists sp. exists l. exists pc. exists bp.
    unfold pimp in H. apply H. unfold linkPG. exists S0.
    split; auto.
    apply frame_movwp.

  unfold gimp.
  unfold conjG. unfold conjPG.
  intros.
  InvertAll. rename x into S0.
  unfold l_release1_grt.
  intros.
  unfold l_release2_grt in H2.
  assert (leadstoPQ (l_release2_Prd_aux sp l pc bp)
             (l_release_Post_aux (sp + 2) l pc bp) S0 S').
  apply H2. clear H2.
  assert (pimp
         (l_release1_Prd_aux sp l pc bp
          |> nextS (l_release+1) (comml (_movwp (Opr_reg SP) BP)))
         (l_release2_Prd_aux sp l pc bp)).
  apply H. clear H.
  assert (gimp (nextS (l_release + 1) (comml (_movwp (Opr_reg SP) BP))
                  oo (leadstoPQ (l_release2_Prd_aux sp l pc bp) (l_release_Post_aux (sp+2) l pc bp)))
                   (leadstoPQ (l_release1_Prd_aux sp l pc bp) (l_release_Post_aux (sp+2) l pc bp))).
  apply leadstoPQ_sameQ_reverseP; auto.
  apply frame_movwp. unfold gimp in H.
  apply H. unfold conjG. exists S0.
  split; auto.
  split. 
    apply IF_false_Prd_Itr_Stk.
    auto.
    apply l_release1_Grt_Itr_Stk.
Qed.

(* _movwld (A_reg 4 BP) BX *)
Definition l_release3_Prd_aux (sp l pc bp : Label) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R BX = l
        /\ exists bx, l_release2_Prd_aux sp l pc bp (M, RF_update R BX bx, fl, isr)
  end.

Definition l_release3_Prd (S : State) :=
  imprec (fun S => exists sp, exists l, exists pc, exists bp, l_release3_Prd_aux sp l pc bp S) S.

Definition l_release3_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ (l_release3_Prd_aux sp l pc bp) (l_release_Post_aux (sp +2) l pc bp) S S'.

Lemma l_release2_Prd_1:
   forall sp l pc bp S0,
   l_release2_Prd_aux sp l pc bp S0 ->
   ((addr_eval (S_RF S0) (A_reg 4 BP)) |~>w l) (S_Mem S0).
Proof.
  unfold l_release2_Prd_aux.
  intros.
  induction S0 as (((M, R), fl), isr).
  unfold l_release1_Prd_aux in H.
  unfold l_release1_MPre in H.
  unfold S_RF. unfold S_Mem.
  reg simpl.
  unfold addr_eval.
  unfold l_release_MPre in H.
  unfold l_release_MPost in H.
  InvertAll.
  rewrite H0. rewrite H in H4.
  substH H4 with (sep_pure_a 
       (sp + 2 |->w pc *^* sp + 2 + 2 |->w l *^* stkSpacePrec (sp + 2) stk_l_release *^*
          INV1 *^* INV0) (In l lockSet_H) (Gamma l) M H4).
  inversion_clear H4.
  sep on.
  sep lift 1 in H2.
  assert (sp + 2 + 2 = sp + 4). auto.
  rewrite H4 in H2.
  sep_blur.
Qed.
  

Lemma l_release_step3_Prd:
   forall sp l pc bp : nat,
   pimp
     (l_release2_Prd_aux sp l pc bp
      |> nextS (l_release + 1 + 1) (comml (_movwld (A_reg 4 BP) BX)))
     (l_release3_Prd_aux sp l pc bp).
Proof.
  unfold pimp. unfold linkPG. unfold l_release2_Prd_aux. unfold l_release3_Prd_aux.
  intros.
  induction S as (((M, R), fl), isr).
  InvertAll.
  induction x as (((M0, R0), fl0), isr0).
  InvertAll.
  destruct (nextS_inversion H1).
  InvertAll.
  split.
  assert (((addr_eval (S_RF (M0, R0, fl0, isr0)) (A_reg 4 BP)) |~>w l) (S_Mem (M0, R0, fl0, isr0))).
  apply l_release2_Prd_1 with sp pc bp.
  unfold l_release2_Prd_aux. split; auto.
  unfold S_Mem in H5. unfold S_RF in H5.
  subst R. reg simpl.
  sep on.
  apply leadsToW_leadsToW_eq with M0 (addr_eval R0 (A_reg 4 BP)); auto.
  exists (R0 BX).
  subst R.
  assert (RF_update (RF_update R0 BX x) BX (R0 BX) = RF_update R0 BX (R0 BX)).
  apply set_set_eq; auto.
  rewrite H5.
  assert (RF_update R0 BX (R0 BX) = R0).
  apply RF_update_prev; auto.
  rewrite H7.
  unfold l_release2_Prd_aux.
  split; auto.
  subst M fl isr. auto.
Qed.

Lemma l_release2_Prd_IF_false:
 forall S : State, l_release2_Prd S -> S_Flag S _IF = false.
Proof.
    unfold l_release2_Prd.
    intros.
    induction S as (((M, R), fl), isr).
    unfold imprec in H. unfold pmstar in H.
    unfold l_release2_Prd_aux in H.
    UnfoldStateAll. InvertAll.
    unfold l_release1_Prd_aux in H3.
    InvertAll. auto.
Qed.
Hint Resolve l_release2_Prd_IF_false.

Lemma l_release2_Grt_Itr_Stk: Grt_Itr_Stk l_release2_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  unfold l_release2_grt.
  unfold leadstoPQ.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  UnfoldStateAll.
  intros.
  unfold l_release2_Prd_aux in H5.
  unfold l_release1_Prd_aux in H5.
  InvertAll.
  unfold itr_enabled in H0.
  UnfoldStateAll.
  inversion_clear H0.
  rewrite H10 in H9. inversion H9.
Qed.

Lemma wff_cmd_l_release3:
   NH_WFcmd
     (SCAP_consCT (l_release + 1 + 1 + 1)
        (Ast l_release3_Prd l_release3_grt 1) psi) 
     (l_release + 1 + 1) (Ast l_release2_Prd l_release2_grt 1)
     (comml (_movwld (A_reg 4 BP) BX)).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists l_release3_Prd. exists l_release3_grt.
  split. auto.

  (* enable *)
  assert (forall sp l pc bp, enable (l_release2_Prd_aux sp l pc bp) 
                                    (nextS (l_release + 1 + 1) (comml (_movwld (A_reg 4 BP) BX)))).
  intros.
  apply enable_movwld.
  intros.
  assert (((addr_eval (S_RF S) (A_reg 4 BP)) |~>w l) (S_Mem S)).
  apply l_release2_Prd_1 with sp pc bp; auto.
  apply leadsToW_leadsToxW with l; auto.

  unfold l_release2_Prd.
  set (p := (fun S0 : State =>
               exists sp : nat,
                 exists l : nat,
                   exists pc : nat,
                     exists bp : nat, l_release2_Prd_aux sp l pc bp S0)).
  assert (enable p (nextS (l_release+1+1) (comml (_movwld (A_reg 4 BP) BX)))).
    unfold enable. unfold enable in H.
    intros. unfold p in H0. InvertAll.
    apply H with x x0 x1 x2; auto.

  assert (safe_mono (nextS (l_release + 1 + 1) (comml (_movwld (A_reg 4 BP) BX)))).
    apply safe_mono_movwld; auto.
  split.
    unfold safe_mono in H1. apply H1; auto.


  assert (forall sp l pc bp, pimp ((l_release2_Prd_aux sp l pc bp) |> nextS (l_release+1+1) (comml (_movwld (A_reg 4 BP) BX)))
                 (l_release3_Prd_aux sp l pc bp)).
    apply l_release_step3_Prd.

  split.
  unfold l_release3_Prd.
  set (p' := (fun S0 : State =>
         exists sp : nat,
           exists l : nat,
             exists pc : nat,
               exists bp : nat, l_release3_Prd_aux sp l pc bp S0)).
  apply frame_rule_1; auto.

    unfold pimp. intros.
    unfold linkPG in H3.
    unfold p in H3.
    InvertAll. rename x into S0.
    rename x0 into sp. rename x1 into l. rename x2 into pc. rename x3 into bp.
    unfold p'. exists sp. exists l. exists pc. exists bp.
    unfold pimp in H2. apply H2. unfold linkPG. exists S0.
    split; auto.
    apply frame_movwld.

  unfold gimp.
  unfold conjG. unfold conjPG.
  intros.
  InvertAll. rename x into S0.
  unfold l_release2_grt.
  intros.
  unfold l_release3_grt in H5.
  assert (leadstoPQ (l_release3_Prd_aux sp l pc bp)
             (l_release_Post_aux (sp + 2) l pc bp) S0 S').
  apply H5. clear H5.
  assert (pimp
         (l_release2_Prd_aux sp l pc bp
          |> nextS (l_release+1+1) (comml (_movwld (A_reg 4 BP) BX)))
         (l_release3_Prd_aux sp l pc bp)).
  apply H2. clear H2.
  assert (gimp (nextS (l_release + 1 +1) (comml (_movwld (A_reg 4 BP) BX))
                  oo (leadstoPQ (l_release3_Prd_aux sp l pc bp) (l_release_Post_aux (sp+2) l pc bp)))
                   (leadstoPQ (l_release2_Prd_aux sp l pc bp) (l_release_Post_aux (sp+2) l pc bp))).
  apply leadstoPQ_sameQ_reverseP; auto.
  apply frame_movwld. unfold gimp in H2.
  apply H2. unfold conjG. exists S0.
  split; auto.

  split. 
    apply IF_false_Prd_Itr_Stk.
      auto.
    apply l_release2_Grt_Itr_Stk.
Qed.

(* spec for the next instruction *)
(* _pushwm (A_reg L_WAITING BX) *)


Definition l_release4_MPre (sp l pc bp : Label) :=
  fun (M : Mem) =>
    exists w, l_release_MPre_tmp (sp + 4) l pc M /\ ((sp +2 |->w bp) *^* (sp |->w w) *^* (l + L_WAITING |->w w) *^* mTrue) M.


Definition l_release4_Prd_aux (sp l pc bp : Label) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R SP = sp
        /\ R BP = sp + 2
        /\ l_release4_MPre (R SP) l pc bp M
        /\ fl _IF = false
        /\ isr = false
  end.

Definition l_release4_Prd (S : State) :=
  imprec (fun S => exists sp, exists l, exists pc, exists bp, l_release4_Prd_aux sp l pc bp S) S.

Definition l_release4_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ (l_release4_Prd_aux sp l pc bp) (l_release_Post_aux (sp +4) l pc bp) S S'.


Lemma l_release3_Prd_1:
   forall sp l pc bp S0,
   l_release3_Prd_aux sp l pc bp S0 ->
   exists w, ((addr_eval (S_RF S0) (A_reg L_WAITING BX)) |~>w w) (S_Mem S0).
Proof.
  intros.
  induction S0 as (((M, R), fl), isr).
  unfold l_release3_Prd_aux in H.
  unfold S_RF. unfold S_Mem.
  unfold addr_eval.
  inversion_clear H. rewrite H0.
  inversion_clear H1.
  unfold l_release2_Prd_aux in H.
  unfold l_release1_Prd_aux in H.
  unfold l_release1_MPre in H. unfold l_release_MPre in H.
  reg simpl.
  InvertAll.
  unfold l_release_MPost in H5.
  rewrite H in H5.
  substH H5 with (sep_pure_a 
       (sp + 2 |->w pc *^* sp + 2 + 2 |->w l *^* stkSpacePrec (sp + 2) stk_l_release *^*
          INV1 *^* INV0) (In l lockSet_H) (Gamma l) M H5).
  inversion_clear H5.
  sep on.
  sep lift 3 in H3.
  assert (forall M, INV1 M -> (l + L_WAITING |->w_ *^* mTrue) M).
  intros.
  apply INV1_lock_wid; auto.
  sep weaker_subst in H3 with (l + L_WAITING |->w_ *^* mTrue).
  sep on.
  sep lift 0 in H3.
  assert (exists w, M ||- l + L_WAITING |->w w *^*
           (mTrue *^*
            (INV0 *^* Gamma l *^*
             (sp + 2 |->w pc *^* sp + 2 + 2 |->w l *^*
              stkSpacePrec (sp + 2) stk_l_release)))).
  apply ptoxW_ptoW1. auto.
  clear H3.
  inversion_clear H10. exists x0.
  sep on. sep trivial.
Qed.

Lemma l_release3_Prd_enable:
   forall sp l pc bp : nat,
   enable (l_release3_Prd_aux sp l pc bp)
     (nextS (l_release + 3) (comml (_pushwm (A_reg L_WAITING BX)))).
Proof.
  unfold enable.
  intros.
  assert (exists w, ((addr_eval (S_RF S) (A_reg L_WAITING BX)) |~>w w) (S_Mem S)).
  apply l_release3_Prd_1 with sp l pc bp; auto.
  inversion_clear H0. rename x into w.
  induction S as (((M, R), fl), isr).
  exists (MemUpdW M (R SP - 2) w 0, RF_update R SP (R SP - 2), fl, isr).
  apply nextS_pushwm with (addr_eval R (A_reg L_WAITING BX)) w. auto.
  auto.
  unfold pushW.
  split.
  unfold l_release3_Prd_aux in H.
  InvertAll. unfold l_release2_Prd_aux in H.
  unfold l_release1_Prd_aux in H.
  reg simpl.
  InvertAll.
  unfold l_release1_MPre in H4.
  unfold l_release_MPre in H4.
  inversion_clear H4.
  unfold l_release_MPost in H6.
  rewrite H in H6.
  substH H6 with (sep_pure_a 
       (sp + 2 |->w pc *^* sp + 2 + 2 |->w l *^* stkSpacePrec (sp + 2) stk_l_release *^*
          INV1 *^* INV0) (In l lockSet_H) (Gamma l) M H6).
  inversion_clear H6.
  assert (stk_l_release >= 2). auto.
  sep on.
  sep lift 2 in H4.
  sep stksplit in H4.
  rewrite H.
  assert (sp + 2 - 2 = sp). auto.
  rewrite H10 in H4.
  assert (stk_l_release - 2 >= 2). auto.
  sep stksplit 2 in H4.
  sep lift 0 in H4.
  apply stkSpacePrec_stkSpace1 with (stkSpacePrec (sp - 2) (stk_l_release - 2 - 2) *^*
            (sp |->w_ *^*
             (INV1 *^* (INV0 *^* Gamma l) *^*
              (sp + 2 |->w pc *^* sp + 2 + 2 |->w l)))); auto.
  split; auto.
  apply  MemUpdW_MemUpd16 with 0; auto.
Qed.
  

Lemma l_release_step4_Prd:
   forall sp l pc bp : nat,
   pimp
     (l_release3_Prd_aux sp l pc bp
      |> nextS (l_release + 3) (comml (_pushwm (A_reg L_WAITING BX))))
     (l_release4_Prd_aux (sp - 2) l pc bp).
Proof.
  unfold pimp. unfold linkPG.
  intros.
  InvertAll.
  induction x as (((M0, R0), fl0), isr0).
  induction S as (((M, R), fl), isr).
  destruct (nextS_inversion H1).
  InvertAll. rename x into w0.
  substH H8 with (MemUpd16_MemUpdW H8).
  destruct H8 as [junk H8].
  subst M R fl isr.
  unfold l_release4_Prd_aux.
  unfold l_release3_Prd_aux in H.
  unfold l_release2_Prd_aux in H.
  unfold l_release1_Prd_aux in H.
  InvertAll.
  reg simpl.
  split; auto.
  rewrite H3.
  split.
  rewrite H in H4.
  assert (sp - 2 + 2 = sp + 2 - 2). auto.
  rewrite H9. auto.
  split; auto.
  unfold l_release4_MPre. clear H1.
  exists w0.
  rewrite H.
  rewrite H in H4.
  assert (sp - 2 + 2 = sp + 2 - 2); auto.
  assert (sp - 2 + 4 = sp + 4 - 2); auto.
  rewrite H1. rewrite H9.
  split.
  unfold l_release1_MPre in H7.
  unfold l_release_MPre_tmp in *.
  unfold l_release_MPost in *.
  inversion_clear H7.
  rewrite H in H11.
  substH H11 with (sep_pure_a 
       (sp + 2 |->w pc *^* sp + 2 + 2 |->w l *^* stkSpacePrec (sp + 2) stk_l_release *^*
          INV1 *^* INV0) (In l lockSet_H) (Gamma l) M0 H11).
  apply sep_pure_b.
  inversion_clear H11.
  split; auto.
  sep on.
  sep lift 2.
  (* unfold stk_l_release in *. *)
  sep stksplit.
  assert (sp + 4 - 2 - 2 = sp). auto. 
  rewrite H11.
  sep stk upd.
  assert (sp + 4 - 2 = sp + 2). auto.
  rewrite H14.
  sep lift 2 in H7.
  sep stksplit in H7.
  assert (sp + 2 - 2 = sp). auto.
  rewrite H15 in H7.
  auto.
  unfold l_release1_MPre in H7.
  unfold l_release_MPre_tmp in *.
  unfold l_release_MPost in *.
  inversion_clear H7.
  rewrite H in H11.
  substH H11 with (sep_pure_a 
       (sp + 2 |->w pc *^* sp + 2 + 2 |->w l *^* stkSpacePrec (sp + 2) stk_l_release *^*
          INV1 *^* INV0) (In l lockSet_H) (Gamma l) M0 H11).
  inversion_clear H11.
  sep on.
  sep lift 1.
  sep upd.
  sep lift 2 in H7.
  sep stksplit in H7.
  sep lift 1 in H7.
  assert (sp = sp + 2 - 2). auto.
  rewrite H in H12. rewrite H11 in H12.
  assert (M0
       ||- sp + 2 - 2 |->w bp *^*
           (INV1 *^* (INV0 *^* Gamma l) *^*
            (sp + 2 |->w pc *^* sp + 2 + 2 |->w l) *^*
            stkSpacePrec (sp + 2 - 2) (stk_l_release - 2))).
  apply leadsToW_ptoxW_unique. auto. auto.
  clear H7.
  rewrite <- H11 in *.
  unfold addr_eval in H0.
  rewrite H2 in H0.
  sep lift 6 in H14.
  sep stksplit in H14.
  sep lift 3 in H14.
  assert (forall M, INV1 M -> (l + L_WAITING |->w_ *^* mTrue) M).
  intros.
  apply INV1_lock_wid; auto.
  sep weaker_subst in H14 with (l + L_WAITING |->w_ *^* mTrue).
  sep on.
  sep lift 0 in H14.
  assert (M0
        ||- l + L_WAITING |->w w0 *^* (mTrue *^*
            (INV0 *^* Gamma l *^*
             (sp + 2 |->w pc *^* sp + 2 + 2 |->w l *^*
              (stkSpacePrec (sp - 2) (stk_l_release - 2 - 2) *^* sp - 2 |->w_ *^*
               sp |->w bp))))).
  apply leadsToW_ptoxW_unique; auto.
  clear H14.
  sep lift 1.
  sep match_left H16. auto.
  sep lift 1.
  sep lift 7 in H16.
  sep match_left H16; auto.
  sep lift 6 in H16.
  sep match_left H16; auto.
Qed.


Lemma l_release3_sp_gt2:
forall sp l pc bp S,
  l_release3_Prd_aux sp l pc bp S
  -> sp - 2 + 4 = sp + 2.
Proof.
  unfold l_release3_Prd_aux.
  intros.
  induction S as (((M, R), fl), isr).
  unfold l_release2_Prd_aux in H.
  unfold l_release1_Prd_aux in H.
  unfold l_release1_MPre in H.
  InvertAll.
  reg simpl.
  unfold l_release_MPre_tmp in H5.
  unfold l_release_MPost in H5.
  rewrite H in H5.
  substH H5 with (sep_pure_a 
       (sp + 2 |->w pc *^* sp + 2 + 2 |->w l *^* stkSpacePrec (sp + 2) stk_l_release *^*
          INV1 *^* INV0) (In l lockSet_H) (Gamma l) M H5).
  inversion_clear H5.
  sep on.
  sep lift 2 in H3.
  sep stksplit in H3.
  assert (sp + 2 - 2 >= stk_l_release - 2).
  apply stkSpacePrec_sp_ge_sz1 with M 
        (sp + 2 - 2 |->w_ *^*
            (INV1 *^* (INV0 *^* Gamma l) *^*
             (sp + 2 |->w pc *^* sp + 2 + 2 |->w l))); auto.
  assert (sp + 2 - 2 = sp). auto.
  rewrite H9 in H5.
  assert (sp >= 2). apply ge_trans with (stk_l_release - 2); auto.
  assert (sp - 2 + 4 = sp + 4 - 2). auto.
  rewrite H11. auto.
Qed.

Lemma l_release3_Grt_Itr_Stk:
  Grt_Itr_Stk l_release3_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  unfold l_release3_grt.
  unfold leadstoPQ.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  UnfoldStateAll.
  intros.
  unfold l_release3_Prd_aux in H5.
  unfold l_release2_Prd_aux in H5.
  unfold l_release1_Prd_aux in H5.
  InvertAll.
  unfold itr_enabled in H0.
  UnfoldStateAll.
  inversion_clear H0.
  rewrite H11 in H10. inversion H10.
Qed.



Lemma wff_cmd_l_release4:
   NH_WFcmd
     (SCAP_consCT (l_release + 1 + 1 + 1 + 1)
        (Ast l_release4_Prd l_release4_grt 1) psi) 
     (l_release + 1 + 1 + 1) (Ast l_release3_Prd l_release3_grt 1)
     (comml (_pushwm (A_reg L_WAITING BX))).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists l_release4_Prd. exists l_release4_grt.
  split. auto.

  assert (l_release + 1 + 1 + 1 = l_release + 3). omega.
  rewrite H. clear H.

  assert (forall sp l pc bp, enable (l_release3_Prd_aux sp l pc bp) 
                                    (nextS (l_release+3) (comml (_pushwm (A_reg L_WAITING BX))))).
  apply l_release3_Prd_enable.

  unfold l_release3_Prd.
  set (p :=
          fun S0 : State =>
          exists sp : nat,
            exists l : nat,
              exists pc : nat,
                exists bp : nat, l_release3_Prd_aux sp l pc bp S0).
  assert (enable p (nextS (l_release+3) (comml (_pushwm (A_reg L_WAITING BX))))).
  unfold p.
  apply forall_enable_exists_l_release; auto.
  assert (safe_mono (nextS (l_release+3) (comml (_pushwm (A_reg L_WAITING BX))))).
    apply safe_mono_pushwm; auto.
  
  split.
    unfold safe_mono in H1. apply H1; auto.

  assert (forall sp l pc bp, pimp ((l_release3_Prd_aux sp l pc bp) |> nextS (l_release+3) (comml (_pushwm (A_reg L_WAITING BX))))
                 (l_release4_Prd_aux (sp - 2) l pc bp)).
  apply l_release_step4_Prd.

  split.

  unfold l_release4_Prd.
  set (p' := 
         fun S0 : State =>
         exists sp : nat,
           exists l : nat,
             exists pc : nat,
               exists bp : nat, l_release4_Prd_aux sp l pc bp S0).
  apply forall_pimp_exists_l_release with l_release3_Prd_aux l_release4_Prd_aux; auto.
    intros. exists (a - 2). exists b. exists c. exists d.
    apply H2; auto.
    apply frame_pushwm.

  apply gimp_l_release with 
    (fun x => x - 2) (fun x => x + 4) l_release3_Prd_aux l_release4_Prd_aux l_release_Post_aux;
  auto.
  apply frame_pushwm.
    unfold gimp. unfold l_release3_grt.
    intros.
    unfold leadstoPQ. unfold leadstoPQ in H3.
    induction S as (((M, R), fl), isr).
    induction S' as (((M', R'), fl'), isr').
    intros.
    assert (exists M1' : MemMod.Map,
               MemMod.disjUnion M1' M2 M' /\
               l_release_Post_aux (sp - 2 + 4) l pc bp (M1', R', fl', isr')).
    apply H3 with M1; auto.
    InvertAll. exists x. split; auto.
    assert (sp - 2 + 4 = sp + 2).
    apply l_release3_sp_gt2 with l pc bp (M1, R, fl, isr); auto.
    rewrite H7 in H8. auto.

    split. unfold Prd_Itr_Stk.
    intros.
    induction S as (((M, R), fl), isr).
    unfold l_release3_Prd in H. unfold imprec in H. unfold pmstar in H.
    InvertAll. unfold l_release3_Prd_aux in H.
    UnfoldStateAll. InvertAll.
    unfold l_release2_Prd_aux in H. unfold l_release1_Prd_aux in H.
    InvertAll.
    unfold itr_enabled in H0.
    inversion_clear H0. unfold S_Flag in H8.
    rewrite H7 in H8. inversion H8.
    apply l_release3_Grt_Itr_Stk.
Qed.



Lemma l_release_MPost_Lock_INV:
forall sp l pc M,
  M ||- l_release_MPost sp l pc
    -> M ||- Lock_INV_H l *^* mTrue.
Proof.
  intros.
  unfold l_release_MPost in H.
  inversion_clear H.
  sep on.
  sep lift 3 in H0.
  assert (forall M, INV1 M -> (Lock_INV_H l *^* mTrue) M).
  intros.
  apply INV1_lockInv1; auto.
  sep weaker_subst in H0 with (Lock_INV_H l *^* mTrue).
  sep on.
  sep lift 0 in H0.
  sep match_left H0.
Qed.


Definition l_release_MPre_4 (sp l pc bp wid : Label) :=
     (sp |->w wid) *^* (sp + 2 |->w bp) *^* (sp + 4|->w pc) *^* (sp + 6 |->w l) *^* stkSpacePrec sp (stk_l_release - 4)
      *^* INV1 *^* INV0 *^* (Gamma l).


Definition l_release_Prd_4_aux (sp l pc bp : Label) (mP: MPred) (S : State) := 
    match S with
    (M, R, fl, isr) =>
        R SP = sp
        /\ (exists w, 
              (l_release_MPre_4 (R SP) l pc bp w *^* mP) M
              /\ (l + L_WAITING |~>w w) M)
        /\ In l lockSet_H
        /\ fl _IF = false
        /\ isr = false
    end.


Definition l_release_Prd_4 (S : State) :=
    exists sp, exists l, exists pc, exists bp, exists mP,
      l_release_Prd_4_aux sp l pc bp mP S.

Lemma adapt_Prd_4_aux:
  forall sp l pc bp M, l_release4_MPre sp l pc bp M -> exists w, l_release_MPre_4 sp l pc bp w M.
Proof.
  intros.
  unfold l_release_MPre_4.
  unfold l_release4_MPre in H.
  destruct H as [w [Hmem1 Hmem2]].
  exists w.
  unfold l_release_MPre_tmp in Hmem1.
  unfold l_release_MPost in Hmem1.
  substH Hmem1 with (sep_pure_a
    (sp + 4 |->w pc *^* sp + 4 + 2 |->w l *^*
          stkSpacePrec (sp + 4) stk_l_release *^* INV1 *^* INV0)
    (In l lockSet_H) (Gamma l) M Hmem1).
  destruct Hmem1 as [Hmem1 Hlock].
  sep on.
  sep lift 2 in Hmem1.
  sep stksplit in Hmem1.
  sep stksplit in Hmem1.
  assert (sp + 4 - 2 - 2 = sp). auto.
  rewrite H in Hmem1.
  assert (stk_l_release - 2 - 2 = stk_l_release - 4). auto with arith.
  rewrite H0 in Hmem1.
  sep lift 1 in Hmem1.
  sep lift 1 in Hmem2.
  sep mem_dec Hmem2 Hmem1.
  assert (sp + 4 - 2 = sp + 2). auto.
  rewrite H1 in Hmem1.
  sep lift 1 in Hmem1.
  sep lift 3 in Hmem2.
  sep mem_dec Hmem2 Hmem1.
  assert (sp + 4 + 2 = sp + 6). auto.
  rewrite H2 in Hmem1.
  sep matchH Hmem1 with 8.
Qed.




Lemma adapt_Prd_4:
  forall S, l_release4_Prd S -> l_release_Prd_4 S.
Proof.
  unfold l_release4_Prd.
  unfold l_release_Prd_4.
  intros.
  induction S as (((M, R), fl), isr).
  unfold imprec in H.
  unfold pmstar in H.
  InvertAll.
  exists x1. exists x2. exists x3. exists x4.
  exists mTrue.
  unfold l_release_Prd_4_aux.
  unfold l_release4_Prd_aux in H.
  UnfoldStateAll.
  InvertAll.
  split; auto.
  unfold l_release4_MPre in H3.
  destruct H3 as [w [Hmem1 Hmem2]].
  unfold l_release_MPre_tmp in Hmem1.
  unfold l_release_MPost in Hmem1.
  rename x1 into sp. rewrite H1.
  rewrite H1 in Hmem1, Hmem2.
  rename x3 into pc. rename x2 into l.
  substH Hmem1 with (sep_pure_a
    (sp + 4 |->w pc *^* sp + 4 + 2 |->w l *^*
          stkSpacePrec (sp + 4) stk_l_release *^* INV1 *^* INV0)
    (In l lockSet_H) (Gamma l) x Hmem1).
  destruct Hmem1 as [Hmem1 Hlock].
  split; auto.
  exists w.
  split.
  exists x. exists x0.
  split_l. split_r.
  unfold l_release_MPre_4.
  sep on.
  sep lift 2 in Hmem1.
  sep stksplit in Hmem1.
  sep stksplit in Hmem1.
  assert (sp + 4 - 2 - 2 = sp). auto.
  rewrite H3 in Hmem1.
  assert (stk_l_release - 2 - 2 = stk_l_release - 4). auto with arith.
  rewrite H5 in Hmem1.
  sep lift 1 in Hmem1.
  sep lift 1 in Hmem2.
  sep mem_dec Hmem2 Hmem1.
  assert (sp + 4 - 2 = sp + 2). auto.
  rewrite H7 in Hmem1.
  sep lift 1 in Hmem1.
  sep lift 3 in Hmem2.
  sep mem_dec Hmem2 Hmem1.
  assert (sp + 4 + 2 = sp + 6). auto.
  rewrite H8 in Hmem1.
  sep matchH Hmem1 with 8.
  assert ((l + L_WAITING |~>w w) x).
  sep on.
  sep lift 2 in Hmem2. sep trivial.
  apply leadsToW_Mono with x x0; auto.
Qed.


Definition l_release_grt_4 (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (l_release_Prd_4_aux sp l pc bp) (l_release_Post_aux_0 (sp + 4) l pc bp) S S'.

Lemma adapt_Grt_4:
forall S S',
  l_release_grt_4 S S' -> l_release4_grt S S'.
Proof.
  unfold l_release_grt_4. unfold l_release4_grt.
  intros.
  unfold leadstoPQ.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  intros.
  unfold leadstoPQ2 in H.
  assert (l_release_Post_aux_0 (sp+4) l pc bp (fun M => M = M2) (M', R', fl', isr')).
  apply H.
  unfold l_release4_Prd_aux in H1.
  unfold l_release_Prd_4_aux.
  InvertAll. split_l.
  rewrite H2 in H3.
  rewrite H2.

  unfold l_release4_MPre in H3.
  destruct H3 as [w [Hmem1 Hmem2]].
  unfold l_release_MPre_tmp in Hmem1.
  unfold l_release_MPost in Hmem1.
  substH Hmem1 with (sep_pure_a
    (sp + 4 |->w pc *^* sp + 4 + 2 |->w l *^*
          stkSpacePrec (sp + 4) stk_l_release *^* INV1 *^* INV0)
    (In l lockSet_H) (Gamma l) M1 Hmem1).
  destruct Hmem1 as [Hmem1 Hlock].
  split; auto.
  exists w.
  split.
  exists M1. exists M2.
  split_l. split_r.
  unfold l_release_MPre_4.
  sep on.
  sep lift 2 in Hmem1.
  sep stksplit in Hmem1.
  sep stksplit in Hmem1.
  assert (sp + 4 - 2 - 2 = sp). auto.
  rewrite H3 in Hmem1.
  assert (stk_l_release - 2 - 2 = stk_l_release - 4). auto with arith.
  rewrite H5 in Hmem1.
  sep lift 1 in Hmem1.
  sep lift 1 in Hmem2.
  sep mem_dec Hmem2 Hmem1.
  assert (sp + 4 - 2 = sp + 2). auto.
  rewrite H7 in Hmem1.
  sep lift 1 in Hmem1.
  sep lift 3 in Hmem2.
  sep mem_dec Hmem2 Hmem1.
  assert (sp + 4 + 2 = sp + 6). auto.
  rewrite H8 in Hmem1.
  sep matchH Hmem1 with 8.
  assert ((l + L_WAITING |~>w w) M1).
  sep on.
  sep lift 2 in Hmem2. sep trivial.
  apply leadsToW_Mono with M1 M2; auto.

  unfold l_release_Post_aux_0 in H2.
  InvertAll.
  unfold mstar in H4.
  InvertAll.
  exists x.
  subst x0.
  split; auto.
  unfold l_release_Post_aux. auto.
Qed.

(* _call _UNBLK *)
Definition l_release5_MPre (sp l pc bp ax : Label) : MPred :=
     (sp + 2 |->w bp) *^* (sp + 4|->w pc) *^* (sp + 6 |->w l) *^* stkSpacePrec (sp + 2) (stk_l_release - 2)
       *^* INV1 *^* INV0 *^* (choice (ax=0) (Gamma l) Aemp).


Definition l_release_Prd_5_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R SP = sp
        /\ (l_release5_MPre (R SP) l pc bp (R AX) *^* mP) M
        /\ In l lockSet_H
        /\ fl _IF = false
        /\ isr = false
  end.

Definition l_release_Prd_5 (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, l_release_Prd_5_aux sp l pc bp mP S.

Definition l_release_grt_5 (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (l_release_Prd_5_aux sp l pc bp) (l_release_Post_aux_0 (sp +4) l pc bp) S S'.

Lemma l_release4_enable_unblk_aux:
  forall sp l pc bp mP S,
    l_release_Prd_4_aux sp l pc bp mP S ->
        exists w, enable_unblk w (Gamma l) S.
Proof.
  intros.
  induction S as (((M, R), fl), isr).
  unfold l_release_Prd_4_aux in H.
  destruct H as [HRsp [Hmem [Hlock [Hfl Hisr]]]].
  destruct Hmem as [w [Hmem1 Hmem2]].
  exists w.
  unfold enable_unblk.
  unfold l_release_MPre_4 in Hmem1.
  rewrite HRsp in Hmem1.
  UnfoldStateAll. rewrite HRsp.
  sep on.
  sep lift 5 in Hmem1.
  assert (forall M, INV1 M -> (Lock_INV_H l *^* mTrue) M).
  intros. apply INV1_lockInv1; auto.
  sep weaker_subst in Hmem1 with (Lock_INV_H l *^* mTrue).
  sep on. sep lift 0 in Hmem1.
  assert (exists b, exists w0, 
          (M ||- (l + L_AVAIL |->w b) *^* (l + L_WAITING |->w w0) *^* (choice (b=0) Aemp (Gamma l)) *^* 
              (mTrue *^*
               (INV0 *^* (Gamma l *^* mP) *^*
                (sp |->w w *^* sp + 2 |->w bp *^* sp + 4 |->w pc *^*
                 sp + 6 |->w l *^* stkSpacePrec sp (stk_l_release - 4))))
          )
          /\ Delta w0 = SomeT (Gamma l)).
  apply decomp_ex_Lock_INV_H; auto.
  clear Hmem1.
  destruct H1 as [b [w0 [Hmem1 Hdelta]]].
  assert (w0 = w).
  sep lift 1 in Hmem1.
  apply  ptoW_leadsToW_eq1 with M (l + L_WAITING) 
              (choice (b = 0) Aemp (Gamma l) *^*
               (mTrue *^*
                (INV0 *^* (Gamma l *^* mP) *^*
                 (sp |->w w *^* sp + 2 |->w bp *^* sp + 4 |->w pc *^*
                  sp + 6 |->w l *^* stkSpacePrec sp (stk_l_release - 4)))) *^*
               l + L_AVAIL |->w b); auto.
  subst w0.
  split_l.
  assert (((sp |->w w *^*
     (Gamma l *^* stkSpacePrec sp (_UNBLK_Stk + 2))) *^* mTrue) M).
  sep on.
  sep lift 7 in Hmem1.
  sep lift 0.
  sep match_left Hmem1; auto.
  sep lift 0.
  sep lift 9 in Hmem1.
  sep match_left Hmem1; auto.
  sep lift 4 in Hmem1.
  assert (stk_l_release - 4 = _UNBLK_Stk + 2). auto.
  rewrite H1 in Hmem1.
  sep match_left Hmem1; auto.
  sep trivial.
  unfold mstar at 1 in H1. InvertAll.
  unfold pmstar. exists x0. exists x.
  UnfoldStateAll.
  split; auto with PMap.
Qed.


Lemma stk_l_release_ge6:
stk_l_release >= 6.
Proof.
  unfold stk_l_release.
  auto with arith.
Qed.
Hint Resolve stk_l_release_ge6.

Lemma l_release4_pgp_aux:
  forall sp l pc bp mP S f f',
    l_release_Prd_4_aux sp l pc bp mP S ->
      pimp (wrap_fun (gunblk (Gamma l)) f _UNBLK f' S) (l_release_Prd_5_aux sp l pc bp mP).
Proof.
  intros.
  rename H into Hpre.
  unfold pimp.
  intros S3 Hwrap.
  unfold l_release_Prd_4_aux in Hpre.
  induction S as (((M, R), fl), isr).
  unfold l_release_MPre_4 in Hpre.
  destruct Hpre as [HRsp [Hmem [Hlock [Hfl Hisr]]]].
  destruct Hmem as [w [Hmem1 Hmem2]].
  unfold l_release_Prd_5_aux.
  induction S3 as (((M3, R3), fl3), isr3).
  unfold wrap_fun in Hwrap.
  unfold conjG in Hwrap.
  destruct Hwrap as [S2 [Hcall Hret]].
  destruct Hcall as [S1 [Hcall Hunblk]].
  induction S1 as (((M1, R1), fl1), isr1).
  induction S2 as (((M2, R2), fl2), isr2).
  destruct (nextS_inversion Hcall).
  clear Hcall. destruct H as [HRspgt2 [Hstkpc HMM1]].
  destruct H0 as [HR1 [Hfl1 Hisr1]].
  subst R1.
  subst fl1 isr1.
  substH HMM1 with (MemUpd16_MemUpdW HMM1).
  destruct HMM1 as [junk HMM1].
  destruct (nextS_inversion Hret).
  subst M2.
  destruct H0 as [H1 [H2 H3]].
  destruct H1 as [pc' [HM3 HR3]].
  subst fl2 isr2. clear Hret.
  unfold gunblk in Hunblk.
  UnfoldStateAll.
  reg simpl.
  destruct Hunblk as [sp' [Hpp [HIFeq [Hisreq [HR0sp HRspeq]]]]].
  subst isr3 sp'.
  assert (R3 SP = sp).
  rewrite HR3.
  reg simpl.
  rewrite <- HRspeq.
  rewrite HRsp. rewrite HRsp in HRspgt2. auto.
  split_l.
  rewrite <- HIFeq.
  split; auto.
  assert (stk_l_release - 4 >= 2).
  unfold stk_l_release. omega.
  assert ((R SP |->w w *^* R SP + 2 |->w bp *^* R SP + 4 |->w pc *^*
           R SP + 6 |->w l *^* stkSpacePrec (R SP) (stk_l_release - 4) *^*
           INV1 *^* INV0 *^* Gamma l *^* mP) M1).
  sep on.
  sep lift 4.
  rewrite HMM1.
  sep stk upd.
  sep matchH Hmem1 with 9.
  rename H0 into HstkN4.
  rename H1 into HM1mem1.
  unfold l_release5_MPre.

  unfold prepost in Hpp.
  unfold S_Mem in Hpp.
  subst R3.
  reg simpl.
  destruct Hpp as [M1_1 [M1_F [M3_1 [HM1_1 [HM3_1 [HsplitM1 HsplitM3]]]]]].
  sep on.
  sep lift 4 in HM1mem1.
  sep stksplit in HM1mem1.
  assert (stk_l_release - 4 - 2 = _UNBLK_Stk).
  unfold stk_l_release. auto with arith.
  rewrite H0 in HM1mem1.
  assert (R2 AX = 0 \/ R2 AX <> 0).
  apply nat_eq_neq_dec.
  inversion_clear H1.
  (* R2 AX = 0 *)
  unfold mstar at 1 in HM1mem1.
  destruct HM1mem1 as
    [M1_1' [M1_F' [HsplitM1' [HM1_1' HM1_F']]]].
  unfold mstar in HM1_1.
  destruct HM1_1 as [M1_1_emp [M1_1_1 [HM1_1_split [Hchoice HM1_1_1]]]].
  unfold choice in Hchoice.
  inversion_clear Hchoice.
  destruct H1 as [Hax0 Hemp]. clear H2.
  assert (M1_1_emp = MemMod.emp). auto.
  rewrite H1 in HM1_1_split.
  assert (M1_1_1 = M1_1). auto with PMap.
  rewrite H2 in HM1_1_1.
  clear H1 H2 Hemp HM1_1_split M1_1_emp.
  assert (precise (stkSpacePrec (R SP - 2) _UNBLK_Stk)).
  apply precise_stkSpacePrec.
  assert (M1_1 = M1_1').
  unfold precise in H1.
  apply H1 with M1; auto.
  apply MemMod.disjUnion_subseteq_left with M1_F; auto.
  apply MemMod.disjUnion_subseteq_left with M1_F'; auto.
  assert (M1_F = M1_F').
  apply MemMod.disjUnion_unique with M1_1 M1_1' M1; auto.

  sep on.
  sep lift 3.
  sep stksplit.
  sep stksplit.
  rewrite H.
  assert (sp + 2 - 2 - 2 = sp - 2). auto with arith.
  rewrite H4. clear H4.
  assert (stk_l_release - 2 - 2 - 2 = _UNBLK_Stk).
  unfold stk_l_release. auto with arith.
  rewrite H4. clear H4.
  unfold mstar at 1.
  sep off.
  exists M3_1. exists M1_F.
  split; auto.
  split. rewrite <- HRsp. auto.
  sep on.
  rewrite H3.
  rewrite HRsp in HM1_F'.
  assert (sp + 2 - 2 = sp). auto.
  rewrite H4. clear H4.
  sep matchH HM1_F' with 9.
  sep off.
  unfold choice. left. split; auto.

  inversion H1. contradiction.

  (* R2 AX <> 0 *)
  assert ( M1 ||- stkSpacePrec (R SP - 2) _UNBLK_Stk *^* Gamma l *^*
                (R SP - 2 |->w_ *^*
                 (INV1 *^* (INV0 *^* mP) *^*
                  (R SP |->w w *^* R SP + 2 |->w bp *^* R SP + 4 |->w pc *^*
                   R SP + 6 |->w l)))).
  sep matchH HM1mem1 with 9.
  clear HM1mem1. rename H1 into HM1mem1.
  unfold mstar at 1 in HM1mem1.
  destruct HM1mem1 as
    [M1_1' [M1_F' [HsplitM1' [HM1_1' HM1_F']]]].
  assert ((stkSpacePrec (R SP - 2) _UNBLK_Stk *^* Gamma l) M1_1).
  unfold mstar in HM1_1.
  sep off.
  unfold mstar.
  InvertAll.
  exists x0. exists x.
  split; auto with PMap.
  split_l.
  unfold choice in H3.
  inversion_clear H3.
  inversion H4. contradiction.
  inversion H4. auto.
  assert (M1_F = M1_F').
  apply MemMod.disjUnion_unique with M1_1 M1_1' M1; auto.
  assert (precise (Gamma l *^* stkSpacePrec (R SP - 2) _UNBLK_Stk)).
  apply mstar_presv_prec.
  apply precGamma. apply precise_stkSpacePrec.
  unfold precise in H3.
  apply H3 with M1; auto.
  apply MemMod.disjUnion_subseteq_left with M1_F; auto.
  apply MemMod.disjUnion_subseteq_left with M1_F'; auto.

  sep on.
  sep lift 3.
  sep stksplit.
  sep stksplit.
  rewrite H.
  assert (sp + 2 - 2 - 2 = sp - 2). auto with arith.
  rewrite H4. clear H4.
  assert (stk_l_release - 2 - 2 - 2 = _UNBLK_Stk).
  unfold stk_l_release. auto with arith.
  rewrite H4. clear H4.
  unfold mstar at 1.
  sep off.
  exists M3_1. exists M1_F.
  split; auto.
  split. rewrite <- HRsp. auto.
  sep on.
  rewrite H3.
  rewrite HRsp in HM1_F'.
  assert (sp + 2 - 2 = sp). auto.
  rewrite H4. clear H4.
  sep matchH HM1_F' with 9.
  sep off.
  unfold mstar. exists MemMod.emp. exists h.
  split. auto with PMap.
  split_r.
  unfold choice. right. split; auto.
  unfold Aemp. auto.
Qed.

Lemma Prd_4_Unique_auxvar:
forall sp sp' l l' pc pc' bp bp' mP mP' S,
  l_release_Prd_4_aux sp l pc bp mP S -> l_release_Prd_4_aux sp' l' pc' bp' mP' S
    -> l = l'.
Proof.
  intros.
  induction S as (((M, R), fl), isr).
  unfold l_release_Prd_4_aux in *.
  InvertAll.
  unfold l_release_MPre_4 in *.
  sep on.
  sep lift 3 in H0.
  sep lift 3 in H.
  apply ptoW_ptoW_eq1 with M (R SP + 6) 
           (stkSpacePrec (R SP) (stk_l_release - 4) *^*
            (INV1 *^* (INV0 *^* (Gamma l *^* mP))) *^*
            (R SP |->w x *^* R SP + 2 |->w bp *^* R SP + 4 |->w pc))
            (stkSpacePrec (R SP) (stk_l_release - 4) *^*
             (INV1 *^* (INV0 *^* (Gamma l' *^* mP'))) *^*
             (R SP |->w x0 *^* R SP + 2 |->w bp' *^* R SP + 4 |->w pc')); auto.
Qed.

Lemma l_release4_pggg_aux:
  forall sp l pc bp mP S f f',
    l_release_Prd_4_aux sp l pc bp mP S
    -> pimp (wrap_fun (gunblk (Gamma l)) f _UNBLK f' S |> l_release_grt_5) (l_release4_grt S).
Proof.
  intros.
  rename H into Hpre.
  unfold pimp.
  intros S2 Hcomp.
  unfold linkPG in Hcomp.
  destruct Hcomp as [S1 [Hwrap Hgrt5]].
  cut (l_release_grt_4 S S2).
  apply adapt_Grt_4.
  unfold l_release_grt_4.
  unfold leadstoPQ2.
  intros sp0 l0 pc0 bp0 mP0 Hpre2.
  assert (l = l0).
  apply Prd_4_Unique_auxvar with sp sp0 pc pc0 bp bp0 mP mP0 S; auto.
  rewrite H in Hwrap.
  unfold l_release_grt_5 in Hgrt5.
  unfold leadstoPQ2 in Hgrt5.
  apply Hgrt5.
  assert (
     pimp (wrap_fun (gunblk (Gamma l0)) f _UNBLK f' S) (l_release_Prd_5_aux sp0 l0 pc0 bp0 mP0)).
  apply l_release4_pgp_aux; auto.
  unfold pimp in H0.
  apply H0; auto.
Qed.

Lemma l_release4_call_unblk:
   forall S0 : State,
   l_release4_Prd S0 ->
   exists w : nat,
     exists wqa : WQassert,
       enable_unblk w wqa S0 /\
       (forall pc' : nat,
        pimp
          (wrap_fun (gunblk wqa) (l_release + 1 + 1 + 1 + 1) _UNBLK pc' S0)
          l_release_Prd_5) /\
       (forall pc' : nat,
        pimp
          (wrap_fun (gunblk wqa) (l_release + 1 + 1 + 1 + 1) _UNBLK pc' S0
           |> l_release_grt_5) (l_release4_grt S0)).
Proof.
  intros.
  assert (l_release_Prd_4 S0).
  apply adapt_Prd_4; auto.
  clear H.
  unfold l_release_Prd_4 in H0.
  destruct H0 as [sp [l [pc [bp [mP Hpre]]]]].
  assert (exists w, enable_unblk w (Gamma l) S0).
  apply l_release4_enable_unblk_aux with sp pc bp mP; auto.
  InvertAll. rename x into w.
  exists w. exists (Gamma l). split_l.
  split.
  intros.
  unfold pimp.
  intros.
  unfold l_release_Prd_5.
  exists sp. exists l. exists pc. exists bp. exists mP.
  assert (pimp (wrap_fun (gunblk (Gamma l)) (l_release + 1 + 1 + 1 + 1) _UNBLK pc' S0) (l_release_Prd_5_aux sp l pc bp mP)).
  apply l_release4_pgp_aux; auto.
  unfold pimp in H1. apply H1; auto.
  intros.
  apply l_release4_pggg_aux with sp pc bp mP; auto.
Qed.

Lemma l_release4_Grt_Itr_Stk:
  Grt_Itr_Stk l_release4_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  unfold l_release4_grt.
  unfold leadstoPQ.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  UnfoldStateAll.
  intros.
  unfold l_release4_Prd_aux in H5.
  InvertAll.
  unfold itr_enabled in H0.
  UnfoldStateAll.
  inversion_clear H0.
  rewrite H9 in H8. inversion H8.
Qed.


Lemma l_release4_Prd_IF_false:
    forall S : State, l_release4_Prd S -> S_Flag S _IF = false.
Proof.
  intros. induction S as (((M, R), fl), isr).
  unfold l_release4_Prd in H. unfold imprec in H.
  unfold pmstar in H. InvertAll. unfold l_release4_Prd_aux in H.
  UnfoldStateAll. InvertAll. auto.
Qed.

Lemma wff_cmd_l_release5:
   NH_WFcmd
     (SCAP_consCT (l_release + 1 + 1 + 1 + 1 + 1)
        (Ast l_release_Prd_5 l_release_grt_5 1) psi)
     (l_release + 1 + 1 + 1 + 1) (Ast l_release4_Prd l_release4_grt 1)
     (comml (_call _UNBLK)).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  left.

  split; auto. unfold prim_entry. right.  right. auto.

  unfold NH_Call_Prim.
  assert (forall S : State, l_release4_Prd S -> S_Flag S _IF = false).
  apply l_release4_Prd_IF_false.
  split; auto.
  
  apply Call_UNBLK with l_release4_Prd l_release_Prd_5 l_release4_grt l_release_grt_5 1.
  auto. auto. auto.
  apply l_release4_call_unblk.

  split.
  apply IF_false_Prd_Itr_Stk.
  apply l_release4_Prd_IF_false.
  apply l_release4_Grt_Itr_Stk.
Qed.


(* _addw (Opr_imm 2) SP *)
Definition l_release_Prd_6_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R SP = sp
        /\ sp >= 2
        /\ l_release_Prd_5_aux (sp - 2) l pc bp mP (M, RF_update R SP (sp - 2), fl, isr)
  end.

Definition l_release_Prd_6 (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, l_release_Prd_6_aux sp l pc bp mP S.

Definition l_release_grt_6 (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (l_release_Prd_6_aux sp l pc bp) (l_release_Post_aux_0 (sp +2) l pc bp) S S'.

Lemma l_release5_Grt_Itr_Stk:
  Grt_Itr_Stk l_release_grt_5.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  unfold l_release_grt_5.
  unfold leadstoPQ2.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  UnfoldStateAll.
  intros.
  unfold l_release_Prd_5_aux in H4.
  InvertAll.
  unfold itr_enabled in H0.
  UnfoldStateAll.
  inversion_clear H0.
  rewrite H8 in H7. inversion H7.
Qed.

Lemma l_release5_Prd_IF_false:
   forall S : State, l_release_Prd_5 S -> S_Flag S _IF = false.
Proof.
    intros.
    unfold l_release_Prd_5 in H.
    InvertAll. unfold l_release_Prd_5_aux in H0.
    induction S as (((M, R), fl), isr). UnfoldStateAll. InvertAll.
    auto.
Qed.

Lemma l_release_step6_Prd:
forall f sp l pc bp mP S S',
  l_release_Prd_5_aux sp l pc bp mP S ->
    nextS f (comml (_addw (Opr_imm 2) SP)) S S'
    -> l_release_Prd_6_aux (sp + 2) l pc bp mP S'.
Proof.
  intros f sp l pc bp mP S S' Hpre Htrans.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion Htrans).
  clear Htrans. InvertAll.
  subst M' R' isr'.
  rename H0 into Hfl'.
  unfold trash_fl in Hfl'.
  unfold opr_eval.
  unfold l_release_Prd_6_aux.
  unfold l_release_Prd_5_aux in Hpre.
  destruct Hpre as [HRsp [Hmem [Hlock [Hfl Hisr]]]].
  split. reg simpl. rewrite HRsp. auto with arith.
  split. auto with arith.
  reg simpl.
  assert (sp + 2 - 2 = sp). auto.
  rewrite H.
  rewrite <- HRsp.
  assert (RF_update R SP (R SP) = R). 
  apply RF_update_prev.
  rewrite H0.
  unfold l_release_Prd_5_aux.
  split; auto.
  split; auto.
  split_l. split_r.
  inversion_clear Hfl'.
  subst fl'. unfold Set_Flag.
  simpl. auto.
  subst fl'. unfold Clear_Flag.
  simpl. auto.
Qed.

Lemma wff_cmd_l_release6:
   NH_WFcmd
     (SCAP_consCT (l_release + 1 + 1 + 1 + 1 + 1 + 1)
        (Ast l_release_Prd_6 l_release_grt_6 1) psi)
     (l_release + 1 + 1 + 1 + 1 + 1) (Ast l_release_Prd_5 l_release_grt_5 1)
     (comml (_addw (Opr_imm 2) SP)).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists l_release_Prd_6. exists l_release_grt_6.
  split. auto.

  generalize (l_release + 1 + 1 + 1 + 1 + 1).
  intro f.
  split. apply trivial_enable_addw.
  split.
  unfold pimp.
  unfold l_release_Prd_5.
  unfold linkPG.
  intros.
  destruct H as [S0 [Hpre Hgrt]].
  destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
  unfold l_release_Prd_6.
  exists (sp + 2). exists l. exists pc. exists bp. exists mP.
  apply l_release_step6_Prd with f S0; auto.
  unfold gimp.
  intros S S' Htmp.
  unfold conjG in Htmp.
  unfold conjPG in Htmp.
  destruct Htmp as [S0 [Htmp Hgrt6]].
  unfold l_release_grt_5.
  unfold leadstoPQ2.
  intros sp l pc bp mP Hpre.
  assert (sp + 2 + 2 = sp + 4); auto.
  rewrite <- H.
  unfold l_release_grt_6 in Hgrt6.
  unfold leadstoPQ2 in Hgrt6.
  apply Hgrt6.
  apply l_release_step6_Prd with f S; auto.
  inversion_clear Htmp; auto.

  split.
  apply IF_false_Prd_Itr_Stk.
  apply l_release5_Prd_IF_false.
  apply l_release5_Grt_Itr_Stk.
Qed.


(* _cmpw (Opr_imm 0) AX *)
Definition l_release_Prd_7_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        exists fl', l_release_Prd_6_aux sp l pc bp mP (M, R, fl', isr) /\
                    ((R AX = 0) /\ (fl = Set_Flag fl' _ZF)
                         \/ (R AX <> 0) /\ (fl = Clear_Flag fl' _ZF))
  end.

Definition l_release_Prd_7 (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, l_release_Prd_7_aux sp l pc bp mP S.

Definition l_release_grt_7 (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (l_release_Prd_7_aux sp l pc bp) (l_release_Post_aux_0 (sp + 2) l pc bp) S S'.


Lemma l_release6_Prd_IF_false:
   forall S : State, l_release_Prd_6 S -> S_Flag S _IF = false.
Proof.
   intros. unfold l_release_Prd_6 in H.
   InvertAll.
   unfold l_release_Prd_6_aux in H0. unfold l_release_Prd_5_aux in H0.
   induction S as (((M, R), fl), isr). UnfoldStateAll.
   InvertAll. auto.
Qed.

Lemma l_release_step7_Prd:
forall sp l pc bp mP S S' f,
  l_release_Prd_6_aux sp l pc bp mP S ->
    nextS f (comml (_cmpw (Opr_imm 0) AX)) S S'
    -> l_release_Prd_7_aux sp l pc bp mP S'.
Proof.
  intros.
  rename H into Hpre. rename H0 into Htrans.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion Htrans).
  InvertAll.
  subst M'. subst R' isr'.
  unfold l_release_Prd_7_aux.
  exists fl.
  unfold opr_eval in H0.
  split; auto.
  inversion_clear H0.
  left. 
  inversion H. split; auto.
  inversion H. right. split; auto.
Qed.

Lemma l_release6_Grt_Itr_Stk:
  Grt_Itr_Stk l_release_grt_6.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  unfold l_release_grt_6.
  unfold leadstoPQ2.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  UnfoldStateAll.
  intros.
  unfold l_release_Prd_6_aux in H4.
  unfold l_release_Prd_5_aux in H4.
  InvertAll.
  unfold itr_enabled in H0.
  UnfoldStateAll.
  inversion_clear H0.
  rewrite H10 in H9. inversion H9.
Qed.

Lemma wff_cmd_l_release7:
   NH_WFcmd
     (SCAP_consCT (l_release + 1 + 1 + 1 + 1 + 1 + 1 + 1)
        (Ast l_release_Prd_7 l_release_grt_7 1) psi)
     (l_release + 1 + 1 + 1 + 1 + 1 + 1)
     (Ast l_release_Prd_6 l_release_grt_6 1) (comml (_cmpw (Opr_imm 0) AX)).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists l_release_Prd_7. exists l_release_grt_7.
  split.
  auto.

  generalize (l_release + 1 + 1 + 1 + 1 + 1 + 1). intro f.
  split. apply trivial_enable_cmpw.

  split.
  unfold pimp. unfold linkPG.
  intros S1 Htmp.
  destruct Htmp as [S0 [Hpre Htrans]].
  unfold l_release_Prd_7.
  unfold l_release_Prd_6 in Hpre.
  destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
  exists sp. exists l. exists pc. exists bp. exists mP.
  apply l_release_step7_Prd with S0 f. auto. auto.
  unfold gimp.
  unfold conjG. unfold conjPG.
  intros.
  destruct H as [S0 [Htmp Hgrt7]].
  destruct Htmp as [Hpre Htrans].
  unfold l_release_grt_6.
  unfold leadstoPQ2.
  intros.
  unfold l_release_grt_7 in Hgrt7.
  unfold leadstoPQ2 in Hgrt7.
  apply Hgrt7.
  apply l_release_step7_Prd with S f; auto.
  split.
  apply IF_false_Prd_Itr_Stk.
  apply l_release6_Prd_IF_false.
  apply l_release6_Grt_Itr_Stk.
Qed.

(* _jne l_rel_done *)

Definition l_release_Prd_8_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        exists fl', l_release_Prd_6_aux sp l pc bp mP (M, R, fl', isr) /\
                    (R AX = 0) /\ (fl = Set_Flag fl' _ZF)
  end.

Definition l_release_Prd_8 (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, l_release_Prd_8_aux sp l pc bp mP S.

Definition l_release_grt_8 (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (l_release_Prd_8_aux sp l pc bp) (l_release_Post_aux_0 (sp +2) l pc bp) S S'.

Lemma l_release7_Grt_Itr_Stk: Grt_Itr_Stk l_release_grt_7.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  unfold l_release_grt_7.
  unfold leadstoPQ2.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  UnfoldStateAll.
  intros.
  unfold l_release_Prd_7_aux in H4.
  unfold l_release_Prd_6_aux in H4.
  unfold l_release_Prd_5_aux in H4.
  InvertAll.
  unfold itr_enabled in H0.
  UnfoldStateAll.
  inversion_clear H0.
  inversion_clear H6.
  inversion_clear H0.
  rewrite H14 in H11. unfold Set_Flag in H11.
  simpl in H11. rewrite H11 in H10. inversion H10.
  inversion_clear H0.
  rewrite H14 in H11. unfold Clear_Flag in H11.
  simpl in H11. rewrite H11 in H10. inversion H10.
Qed.

Lemma l_release7_Prd_IF_false:
   forall S : State, l_release_Prd_7 S -> S_Flag S _IF = false.
Proof.
  unfold l_release_Prd_7. intros.
  induction S as (((M, R), fl), isr). 
  InvertAll. unfold l_release_Prd_7_aux in H0.
  UnfoldStateAll. InvertAll.
  unfold l_release_Prd_6_aux in H0. unfold l_release_Prd_5_aux in H0.
  InvertAll. inversion_clear H1.
  inversion_clear H6. rewrite H8. unfold Set_Flag.
  simpl. auto. inversion_clear H6. rewrite H8.
  unfold Clear_Flag. simpl. auto.
Qed.

Lemma l_release_step8_a:
  forall sp l pc bp mP S,
    l_release_Prd_7_aux sp l pc bp mP S -> ZF_cl S -> l_rel_done_Prd_aux sp l pc bp mP S.
Proof.
  intros.
  unfold l_release_Prd_7_aux in H.
  induction S as (((M, R), fl), isr).
  destruct H as [fl' [Hpre6 Hcases]].
  unfold ZF_cl in H0.
  unfold S_Flag in H0.
  inversion_clear Hcases.
  inversion_clear H.
  rewrite H2 in H0.
  unfold Set_Flag in H0.
  simpl in H0.
  inversion H0.
  inversion_clear H.
  rename H1 into Hax0. rename H0 into Hzffalse.
  rename H2 into Hfl.
  unfold l_release_Prd_6_aux in Hpre6.
  unfold l_release_Prd_5_aux in Hpre6.
  reg simpl.
  destruct Hpre6 as [Hsp [Hspgt2 [Hjunk [Hmem [Hlock [Hfl' Hisr]]]]]].
  clear Hjunk.
  unfold l_rel_done_Prd_aux.
  split_l.
  split.
  unfold l_release5_MPre in Hmem.
  unfold l_rel_done_MPre.
  sep on.
  assert (sp - 2 + 2 = sp). auto.
  rewrite H in Hmem.
  clear H.
  assert (sp - 2 + 4 = sp + 2).
  assert (sp - 2 + 4 = sp + 4 - 2); auto.
  rewrite H in Hmem.
  assert (sp - 2 + 6 = sp + 6 - 2). auto.
  assert (sp + 6 - 2 = sp + 4). omega.
  rewrite H1 in H0. rewrite H0 in Hmem.
  clear H H0 H1.
  rewrite Hsp.
  sep matchH Hmem with 8.
  sep off.
  unfold mstar in Hmem.
  InvertAll.
  unfold choice in H0.
  inversion_clear H0.
  inversion H1. contradiction.
  inversion_clear H1. unfold Aemp in H3.
  rewrite H3 in H. assert (x0 = M). auto with PMap.
  rewrite H1 in H2. auto.
  split_l. split_r.
  subst fl. unfold Clear_Flag.
  simpl. auto.
Qed.


Lemma l_release_step8_b:
forall sp l pc bp mP S,
  l_release_Prd_7_aux sp l pc bp mP S -> ZF_st S 
    -> l_release_Prd_8_aux sp l pc bp mP S.
Proof.
  intros.
  unfold l_release_Prd_7_aux in H.
  induction S as (((M, R), fl), isr).
  destruct H as [fl' [Hpre6 Hcases]].
  unfold ZF_st in H0.
  unfold S_Flag in H0.
  inversion_clear Hcases.
  inversion_clear H.
  rename H1 into Hax0. rename H0 into Hzftrue.
  rename H2 into Hfl.
  unfold l_release_Prd_8_aux.
  exists fl'.
  split; auto.
  inversion_clear H.
  rewrite H2 in H0.
  unfold Clear_Flag in H0. simpl in H0.
  inversion H0.
Qed.

Lemma wff_cmd_l_release8:
   NH_WFcmd
     (SCAP_consCT (l_release + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
        (Ast l_release_Prd_8 l_release_grt_8 1) psi)
     (l_release + 1 + 1 + 1 + 1 + 1 + 1 + 1)
     (Ast l_release_Prd_7 l_release_grt_7 1) (comml (_jne l_rel_done)).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists l_rel_done_Prd. exists l_release_Prd_8. 
  exists l_rel_done_grt. exists l_release_grt_8.
  split. unfold SCAP_lookupCT. unfold SCAP_consCT.

  unfold SPSpec.lookup. unfold ListT.In. right.
  assert (SCAP_lookupCT psi l_rel_done l_rel_done_spec).
  apply l_rel_done_in_psi. auto.
  split; auto.

  split.
  unfold conjP.
  unfold pimp.
  intros.
  destruct H as [Hpre Hzf].
  unfold l_release_Prd_7 in Hpre.
  destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
  unfold l_rel_done_Prd.
  exists sp. exists l. exists pc. exists bp. exists mP.
  apply l_release_step8_a; auto.
  split.
  unfold gimp.
  unfold conjPG.
  intros.
  destruct H as [Hpre [Hzf Hgrtdone]].
  unfold l_release_grt_7.
  unfold leadstoPQ2.
  intros.
  unfold l_rel_done_grt in Hgrtdone.
  unfold leadstoPQ2 in Hgrtdone.
  apply Hgrtdone.
  apply l_release_step8_a. auto. auto.
  unfold conjP.
  split.
  unfold pimp.
  intros.
  destruct H as [Hpre Hzf].
  unfold l_release_Prd_7 in Hpre.
  destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
  unfold l_release_Prd_8.
  exists sp. exists l. exists pc. exists bp. exists mP.
  apply l_release_step8_b; auto.
  unfold gimp.
  unfold conjPG.
  intros.
  destruct H as [Hpre [Hzf Hgrtdone]].
  unfold l_release_grt_7.
  unfold leadstoPQ2.
  intros.
  unfold l_release_grt_8 in Hgrtdone.
  unfold leadstoPQ2 in Hgrtdone.
  apply Hgrtdone.
  apply l_release_step8_b; auto.
  split.
  apply IF_false_Prd_Itr_Stk.
  apply l_release7_Prd_IF_false.
  apply l_release7_Grt_Itr_Stk.
Qed.

(* _movwp (Opr_reg SP) BP; *)
Definition l_release_Prd_8b_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R BP = sp
        /\ exists bp0, l_release_Prd_8_aux sp l pc bp mP (M, RF_update R BP bp0, fl, isr)
  end.

Definition l_release_Prd_8b (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, l_release_Prd_8b_aux sp l pc bp mP S.

Definition l_release_grt_8b (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (l_release_Prd_8b_aux sp l pc bp) (l_release_Post_aux_0 (sp +2) l pc bp) S S'.

Lemma wff_cmd_l_release8b:
   NH_WFcmd
     (SCAP_consCT (l_release + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
        (Ast l_release_Prd_8b l_release_grt_8b 1) psi)
     (l_release + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
     (Ast l_release_Prd_8 l_release_grt_8 1) (comml (_movwp (Opr_reg SP) BP)).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists l_release_Prd_8b. exists l_release_grt_8b.
  split. auto.
  split; auto. (* enable *)

  generalize (l_release + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1).
  intro f.
Lemma l_release_step_8b:
forall f sp l pc bp mP S S',
  l_release_Prd_8_aux sp l pc bp mP S
  -> nextS f (comml (_movwp (Opr_reg SP) BP)) S S'
  -> l_release_Prd_8b_aux sp l pc bp mP S'.
Proof.
  intros.
  rename H0 into Htrans.
  unfold l_release_Prd_8_aux in H.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct H as [fl'' [Hpre6 [Hax0 Hfl]]].
  destruct (nextS_inversion Htrans).
  subst M'.
  destruct H0 as [HR' [Hfl' Hisr']].
  subst R' fl' isr'.
  unfold l_release_Prd_8b_aux.
  reg simpl. unfold opr_eval.
  split.
  unfold l_release_Prd_6_aux in Hpre6.
  InvertAll. auto.
  exists (R BP).
  assert (RF_update (RF_update R BP (R SP)) BP (R BP) = RF_update R BP (R BP)).
  apply RF_update_RF_update.
  rewrite H.
  assert (RF_update R BP (R BP) = R).
  apply RF_update_prev.
  rewrite H0.
  unfold l_release_Prd_8_aux.
  exists fl''.
  auto.
Qed.
  split.
  unfold linkPG.
  unfold pimp.
  intros.
  destruct H as [S0 [Hpre Htrans]].
  unfold l_release_Prd_8 in Hpre.
  InvertAll.
  unfold l_release_Prd_8b.
  exists x. exists x0. exists x1. exists x2. exists x3.
  apply l_release_step_8b with f S0; auto.
  unfold gimp.
  unfold conjPG.
  unfold conjG.
  intros.
  destruct H as [S'' [[Hpre Htrans] Hgrt8b]].
  unfold l_release_grt_8.
  unfold leadstoPQ2.
  intros.
  unfold l_release_grt_8b in Hgrt8b.
  unfold leadstoPQ2 in Hgrt8b.
  apply Hgrt8b.
  apply l_release_step_8b with f S; auto.
  split.
  apply IF_false_Prd_Itr_Stk.
Lemma l_release8_Prd_IF_false:
  forall S : State, l_release_Prd_8 S -> S_Flag S _IF = false.
Proof.
  unfold l_release_Prd_8. intro S.
  induction S as (((M, R), fl), isr).
  intros.
  InvertAll. unfold l_release_Prd_8_aux in H0.
  unfold l_release_Prd_6_aux in H0.
  unfold l_release_Prd_5_aux in H0.
  UnfoldStateAll. InvertAll.
  rewrite H2. unfold Set_Flag. simpl.
  auto.
Qed.
  apply l_release8_Prd_IF_false.
Lemma l_release8_Grt_Itr_Stk:
   Grt_Itr_Stk l_release_grt_8.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  unfold l_release_grt_8.
  unfold leadstoPQ2.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  UnfoldStateAll.
  intros.
  unfold l_release_Prd_8_aux in H4.
  unfold l_release_Prd_7_aux in H4.
  unfold l_release_Prd_6_aux in H4.
  unfold l_release_Prd_5_aux in H4.
  InvertAll.
  unfold itr_enabled in H0.
  UnfoldStateAll.
  inversion_clear H0.
  rewrite H7 in H12. unfold Set_Flag in H12.
  simpl in H12.
  rewrite H12 in H11. inversion H11.
Qed.
  apply l_release8_Grt_Itr_Stk.
Qed.


(* _movwld (A_reg 4 BP) BX *)
Definition l_release_Prd_9_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R BX = l
        /\ exists bx, l_release_Prd_8b_aux sp l pc bp mP (M, RF_update R BX bx, fl, isr)
  end.

Definition l_release_Prd_9 (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, l_release_Prd_9_aux sp l pc bp mP S.

Definition l_release_grt_9 (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (l_release_Prd_9_aux sp l pc bp) (l_release_Post_aux_0 (sp +2) l pc bp) S S'.

Lemma l_release8b_Prd_1:
   forall sp l pc bp mP S,
   l_release_Prd_8b_aux sp l pc bp mP S ->
   ((addr_eval (S_RF S) (A_reg 4 BP)) |~>w l) (S_Mem S).
Proof.
  intros.
  unfold l_release_Prd_8b_aux in H.
  induction S as (((M, R), fl), isr).
  unfold l_release_Prd_8_aux in H.
  unfold l_release_Prd_6_aux in H.
  unfold l_release_Prd_5_aux in H.
  unfold l_release5_MPre in H.
  InvertAll.
  reg simpl.
  unfold addr_eval.
  unfold S_RF. unfold S_Mem.
  rewrite H0.
  assert (sp - 2 + 6 = sp + 6 - 2). auto.
  assert (sp + 6 - 2 = sp + 4). omega.
  rewrite H10 in H8. rewrite H8 in H5.
  sep on.
  sep lift 2 in H5.
  sep trivial.
Qed.


Lemma l_release_step_9:
forall f sp l pc bp mP S S',
  l_release_Prd_8b_aux sp l pc bp mP S
  -> nextS f (comml (_movwld (A_reg 4 BP) BX)) S S'
  -> l_release_Prd_9_aux sp l pc bp mP S'.
Proof.
  intros.
  rename H0 into Htrans.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion Htrans).
  subst M'.
  destruct H1 as [[x [Hmemx HR']] [Hfl' Hisr']].
  assert (x = l).
  assert ((addr_eval (S_RF (M, R, fl, isr)) (A_reg 4 BP) |~>w l) (S_Mem (M, R, fl, isr))).
  apply l_release8b_Prd_1 with sp pc bp mP; auto.
  UnfoldStateAll.
  apply leadsToW_unique with (addr_eval R (A_reg 4 BP)) M; auto.
  subst x.
  subst R' fl' isr'.
  unfold l_release_Prd_9_aux.
  split. reg simpl. auto.
  exists (R BX).
  assert (RF_update (RF_update R BX l) BX (R BX) = R).
  assert (RF_update (RF_update R BX l) BX (R BX) = RF_update R BX (R BX)).
  apply RF_update_RF_update.
  rewrite H0.
  apply RF_update_prev.
  rewrite H0.
  auto.
Qed.

Lemma l_release8b_Prd_IF_false:
   forall S : State, l_release_Prd_8b S -> S_Flag S _IF = false.
Proof.
  intros.
  unfold l_release_Prd_8b in H.
  unfold l_release_Prd_8b_aux in H.
  induction S as (((M, R), fl), isr).
  unfold l_release_Prd_8_aux in H.
  unfold l_release_Prd_6_aux in H.
  unfold l_release_Prd_5_aux in H.
  InvertAll.
  reg simpl.
  unfold S_Flag.
  rewrite H3. unfold Set_Flag. simpl.
  auto.
Qed.

Lemma l_release8b_Grt_Itr_Stk: Grt_Itr_Stk l_release_grt_8b.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  unfold l_release_grt_8b.
  unfold leadstoPQ2.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  UnfoldStateAll.
  intros.
  unfold l_release_Prd_8b_aux in H4.
  unfold l_release_Prd_8_aux in H4.
  unfold l_release_Prd_6_aux in H4.
  unfold l_release_Prd_5_aux in H4.
  InvertAll. reg simpl.
  unfold itr_enabled in H0.
  UnfoldStateAll.
  inversion_clear H0.
  rewrite H8 in H13. 
  unfold Set_Flag in H13.
  simpl in H13.
  rewrite H13 in H12. inversion H12.
Qed.

Lemma wff_cmd_l_release9:
   NH_WFcmd
     (SCAP_consCT (l_release + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
        (Ast l_release_Prd_9 l_release_grt_9 1) psi)
     (l_release + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
     (Ast l_release_Prd_8b l_release_grt_8b 1) (comml (_movwld (A_reg 4 BP) BX)).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists l_release_Prd_9. exists l_release_grt_9.
  split. auto.
  generalize (l_release + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1). intro f.

  (* enable *)
  assert (forall sp l pc bp mP, enable (l_release_Prd_8b_aux sp l pc bp mP) 
                                    (nextS f (comml (_movwld (A_reg 4 BP) BX)))).
  intros.
  apply enable_movwld.
  intros.
  assert (((addr_eval (S_RF S) (A_reg 4 BP)) |~>w l) (S_Mem S)).
  apply l_release8b_Prd_1 with sp pc bp mP; auto.
  apply leadsToW_leadsToxW with l; auto.

  split.
  unfold enable in *.
  intros.
  unfold l_release_Prd_8b in H0.
  InvertAll.
  apply H with x x0 x1 x2 x3; auto.
  split.
  unfold pimp.
  unfold linkPG.
  intros.
  InvertAll.
  unfold l_release_Prd_8b in H0.
  InvertAll.
  unfold l_release_Prd_9.
  exists x0. exists x1. exists x2. exists x3. exists x4.
  apply l_release_step_9 with f x; auto.

  unfold gimp.
  unfold conjPG. unfold conjG.
  intros.
  InvertAll.
  unfold l_release_grt_8b.
  unfold leadstoPQ2.
  intros.
  unfold l_release_grt_9 in H2.
  unfold leadstoPQ2 in H2.
  apply H2.
  apply l_release_step_9 with f S; auto.
  split. 
    apply IF_false_Prd_Itr_Stk.
    apply l_release8b_Prd_IF_false; auto.
    apply l_release8b_Grt_Itr_Stk.
Qed.

Lemma l_release9_Prd_IF_false:
   forall S : State, l_release_Prd_9 S -> S_Flag S _IF = false.
Proof.
    unfold l_release_Prd_9.
    intros.
    induction S as (((M, R), fl), isr).
    unfold l_release_Prd_9_aux in H.
    unfold l_release_Prd_8b_aux in H.
    unfold l_release_Prd_8_aux in H.
    unfold l_release_Prd_6_aux in H.
    unfold l_release_Prd_5_aux in H.
    InvertAll. reg simpl.
    unfold S_Flag.
    rewrite H4. unfold Set_Flag. simpl.
    auto.
Qed.

Lemma l_release9_Grt_Itr_Stk: Grt_Itr_Stk l_release_grt_9.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  unfold l_release_grt_9.
  unfold leadstoPQ2.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  UnfoldStateAll.
  intros.
  unfold l_release_Prd_9_aux in H4.
  unfold l_release_Prd_8b_aux in H4.
  unfold l_release_Prd_8_aux in H4.
  unfold l_release_Prd_6_aux in H4.
  unfold l_release_Prd_5_aux in H4.
  InvertAll. reg simpl.
  unfold itr_enabled in H0.
  UnfoldStateAll.
  inversion_clear H0.
  rewrite H9 in H14. 
  unfold Set_Flag in H14.
  simpl in H14.
  rewrite H14 in H13. inversion H13.
Qed.

Lemma l_release9_Prd_1:
forall sp l pc bp mP S,
   l_release_Prd_9_aux sp l pc bp mP S ->
   (addr_eval (S_RF S) (A_reg L_AVAIL BX) |~>w_) (S_Mem S).
Proof.
  intros.
  unfold addr_eval.
  unfold l_release_Prd_9_aux in H.
  induction S as (((M, R), fl), isr).
  InvertAll.
  UnfoldStateAll.
  rewrite H0.
  unfold l_release_Prd_8b_aux in H.
  unfold l_release_Prd_8_aux in H.
  unfold l_release_Prd_6_aux in H.
  unfold l_release_Prd_5_aux in H.
  InvertAll. reg simpl.
  unfold l_release5_MPre in H6.
  sep on.
  sep lift 4 in H6.
  assert (forall M, INV1 M -> (Lock_INV_H l *^* mTrue) M).
  intros.
  apply INV1_lockInv1; auto.
  sep weaker_subst in H6 with (Lock_INV_H l *^* mTrue).
  sep on.
  sep lift 0 in H6.
  assert (exists b, exists w,
    (M ||- (l + L_AVAIL |->w b) *^* (l + L_WAITING |->w w) *^* (choice (b=0) Aemp (Gamma l))
           *^* (mTrue *^*
            (INV0 *^* (choice (R AX = 0) (Gamma l) Aemp *^* mP) *^*
             (sp - 2 + 2 |->w bp *^* sp - 2 + 4 |->w pc *^* sp - 2 + 6 |->w l *^*
              stkSpacePrec (sp - 2 + 2) (stk_l_release - 2)))))
    /\ Delta w = SomeT (Gamma l)).
  apply decomp_ex_Lock_INV_H; auto.
  clear H6. InvertAll.
  sep lift 0 in H6.
  sep trivial.
Qed.

Lemma l_release_step_a:
forall f sp l pc bp mP S S',
  l_release_Prd_9_aux sp l pc bp mP S
  -> nextS f (comml (_movwst (Opr_imm 1) (A_reg L_AVAIL BX))) S S'
  -> l_rel_done_Prd_aux sp l pc bp mP S'.
Proof.
  intros.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion H0).
  clear H0.
  InvertAll.
  subst R' fl' isr'.
  rename H3 into Htargetx.
  rename H5 into Hnewmem.
  unfold addr_eval in *.
  unfold opr_eval in *.
  substH Hnewmem with (MemUpd16_MemUpdW Hnewmem).
  destruct Hnewmem as [junk HM'].
  unfold l_rel_done_Prd_aux.
  unfold l_release_Prd_9_aux in H.
  destruct H as [HRbx [bx0 Hpre]].
  unfold l_release_Prd_8b_aux in Hpre.
  destruct Hpre as [Hbp [bp0 Hpre]].
  reg simpl.
  unfold l_release_Prd_8_aux in Hpre.
  destruct Hpre as [fl' [Hpre [Hax0 Hfl]]].
  reg simpl.
  unfold l_release_Prd_6_aux in Hpre.
  destruct Hpre as [HRsp [HRspgt2 Hpre]].
  reg simpl.
  split_l.
  unfold l_release_Prd_5_aux in Hpre.
  destruct Hpre as [Htmp [Hmem [Hlock [Hfl' Hisr]]]].
  clear Htmp.
  reg simpl.
  split.
  unfold l_rel_done_MPre.
  unfold l_release5_MPre in Hmem.
  sep on.
  rewrite HRsp.
  assert (exists m, leqvM INV1 (Lock_INV_H l *^* m)).
  apply INV1_decom_l; auto.
  destruct H as [mP' Hleqv].
  sep lift 4.
  sep leqv_subst with (Lock_INV_H l *^* mP').
  sep lift 4 in Hmem.
  sep leqv_subst in Hmem with (Lock_INV_H l *^* mP').
  sep on.
  assert (M
         ||- Lock_INV_H l *^* (Gamma l) *^* 
             (mP' *^* INV0 *^*  mP *^*
              (sp - 2 + 2 |->w bp *^* sp - 2 + 4 |->w pc *^*
               sp - 2 + 6 |->w l *^*
               stkSpacePrec (sp - 2 + 2) (stk_l_release - 2)))).
  sep matchH Hmem with 9.
  sep off. unfold choice in Hmem.
  inversion_clear Hmem. inversion H0; auto.
  inversion H0. contradiction.
  rewrite HRbx in HM'.
  assert (sp - 2 + 2 = sp).
  assert (sp - 2 + 2 = sp + 2 - 2).
  auto. assert (sp + 2 - 2 = sp). auto.
  rewrite H2 in H1; auto.
  rewrite H1 in H0. rewrite H1 in Hmem. clear H1.
  assert (sp - 2 + 4 = sp + 2).
  assert (sp - 2 + 4 = sp + 4 - 2); auto.
  assert (sp + 4 - 2 = sp + 2). auto.
  rewrite H1 in Hmem. auto. rewrite H1 in H0. clear H1.
  assert (sp - 2 + 6 = sp + 4).
  assert (sp - 2 + 6 = sp + 6 - 2); auto.
  assert (sp + 6 - 2 = sp + 4). omega.
  rewrite H1 in Hmem. rewrite H1 in H0.
  assert (M' ||- Lock_INV_H l *^* 
           (mP' *^* INV0 *^* mP *^*
            (sp |->w bp *^* sp + 2 |->w pc *^* sp + 4 |->w l *^*
             stkSpacePrec sp (stk_l_release - 2)))).
  subst M'.
  apply Lock_INV_convert_a; auto.
  sep matchH H4 with 8.
  split_l. split_r.
  subst fl. unfold Set_Flag. simpl.
  auto.
Qed.

Lemma wff_cmd_l_releasea:
   NH_WFcmd
     (SCAP_consCT (l_release + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
        (Ast l_rel_done_Prd l_rel_done_grt 1) psi)
     (l_release + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
     (Ast l_release_Prd_9 l_release_grt_9 1)
     (comml (_movwst (Opr_imm 1) (A_reg L_AVAIL BX))).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists l_rel_done_Prd. exists l_rel_done_grt.
  split. auto.

  generalize (l_release + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1). intro f.
  assert (forall sp l pc bp mP, enable (l_release_Prd_9_aux sp l pc bp mP) (nextS f (comml (_movwst (Opr_imm 1) (A_reg L_AVAIL BX))))).
  intros.
  apply enable_movwst.
  intros.
  apply l_release9_Prd_1 with sp l pc bp mP. auto.
  split.
  unfold enable in *.
  intros.
  unfold l_release_Prd_9 in H0.
  InvertAll.
  apply H with x x0 x1 x2 x3; auto.
  split.
  unfold pimp. unfold linkPG.
  intros.
  InvertAll.
  unfold l_rel_done_Prd.
  unfold l_release_Prd_9 in H0.
  InvertAll.
  exists x0. exists x1. exists x2. exists x3. exists x4.
  apply l_release_step_a with f x; auto.

  unfold gimp.
  unfold conjPG. unfold conjG.
  intros.
  InvertAll.
  unfold l_release_grt_9.
  unfold leadstoPQ2.
  intros.
  unfold l_rel_done_grt in H2.
  unfold leadstoPQ2 in H2.
  apply H2.
  apply l_release_step_a with f S; auto.

  split.
  apply IF_false_Prd_Itr_Stk.
  apply l_release9_Prd_IF_false.
  apply l_release9_Grt_Itr_Stk.
Qed.

Definition l_release_Post (S : State) : Prop :=
  exists sp, exists l, exists pc, exists bp, exists mP, l_release_Post_aux_0 sp l pc bp mP S.

Definition l_release_last_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (l_release_Post_aux_0 sp l pc bp) (l_release_Post_aux_0 sp l pc bp) S S'.


Lemma l_rel_done_Prd_IF_false:
   forall S : State, l_rel_done_Prd S -> S_Flag S _IF = false.
Proof.
  intros.
  induction S as (((M, R), fl), isr). unfold l_rel_done_Prd in H.
  InvertAll.
  unfold l_rel_done_Prd_aux in H0. UnfoldStateAll. InvertAll. auto.
Qed.

Lemma l_rel_done_Grt_Itr_Stk: Grt_Itr_Stk l_rel_done_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  unfold l_rel_done_grt.
  unfold leadstoPQ2.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  UnfoldStateAll.
  intros.
  unfold l_rel_done_Prd_aux in H4.
  InvertAll.
  unfold itr_enabled in H0.
  UnfoldStateAll.
  inversion_clear H0.
  rewrite H8 in H7. inversion H7.
Qed.

Lemma wff_cmd_l_rel_done_1:
   NH_WFcmd
     (SCAP_consCT (l_release + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
        (Ast l_release_Post l_release_last_grt 1) psi)
     (l_release + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
     (Ast l_rel_done_Prd l_rel_done_grt 1) (comml (_popw BP)).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists l_release_Post. exists l_release_last_grt.
  split. auto.

  generalize (l_release + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1). intro f.
  assert (forall sp l pc bp mP, enable (l_rel_done_Prd_aux sp l pc bp mP) (nextS f (comml (_popw BP)))).
  intros.
  apply enable_popw.
  intros. assert ((S_RF S SP |~>w bp) (S_Mem S)).
Lemma l_rel_done_Prd_1:
forall sp l pc bp mP S,
  l_rel_done_Prd_aux sp l pc bp mP S -> (S_RF S SP |~>w bp) (S_Mem S).
Proof.
  intros.
  unfold l_rel_done_Prd_aux in H.
  induction S as (((M, R), fl), isr).
  InvertAll.
  UnfoldStateAll.
  unfold l_rel_done_MPre in H.
  sep on.
  sep lift 0 in H.
  sep trivial.
Qed.

  apply l_rel_done_Prd_1 with sp l pc mP. auto.
  apply leadsToW_leadsToxW with bp; auto.
  split.
  unfold enable in *.
  intros.
  unfold l_rel_done_Prd in H0.
  InvertAll.
  apply H with x x0 x1 x2 x3; auto.

Lemma l_rel_done_step1:
forall f sp l pc bp mP S S',
  l_rel_done_Prd_aux sp l pc bp mP S  
  -> nextS f (comml (_popw BP)) S S'
  -> l_release_Post_aux_0 (sp + 2) l pc bp mP S'.
Proof.
  intros.
  rename H0 into Htrans.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion Htrans).
  subst M'.
  destruct H1 as [[x [Hmemx HR']] [Hfl' Hisr']].
  assert (x = bp).
  assert ((S_RF (M, R, fl, isr) SP |~>w bp) (S_Mem (M, R, fl, isr))).
  apply l_rel_done_Prd_1 with sp l pc mP; auto.
  UnfoldStateAll.
  apply leadsToW_unique with (R SP) M; auto.
  subst x.
  subst R' fl' isr'.
  unfold l_release_Post_aux_0.
  reg simpl.
  clear Htrans.
  unfold l_rel_done_Prd_aux in H.
  InvertAll.
  split. rewrite H0. auto.
  split_l.
  split; auto.
  rewrite H0 in H.
  unfold l_rel_done_MPre in H.
  unfold l_release_MPost.
  apply sep_pure_b.
  split; auto.
  sep on.
  sep lift 2.
  sep stksplit.
  assert (sp + 2 - 2 = sp). auto.
  rewrite H3.
  sep lift 1.
  sep matchH H with 7.
  assert (sp + 2 + 2 = sp + 4).
  auto with arith.
  rewrite H5. auto.
Qed.
  
  split.
  unfold pimp.
  unfold linkPG.
  intros.
  InvertAll.
  unfold l_rel_done_Prd in H0.
  InvertAll.
  unfold l_release_Post.
  exists (x0 + 2). exists x1. exists x2. exists x3. exists x4.
  apply l_rel_done_step1 with f x; auto.

  unfold gimp.
  unfold conjPG. unfold conjG.
  intros.
  InvertAll.
  unfold l_rel_done_grt.
  unfold leadstoPQ2.
  intros.
  unfold l_release_last_grt in H2.
  unfold leadstoPQ2 in H2.
  apply H2.
  apply l_rel_done_step1 with f S; auto.


  split.
  apply IF_false_Prd_Itr_Stk.
  apply l_rel_done_Prd_IF_false.
  apply l_rel_done_Grt_Itr_Stk.
Qed.


Lemma l_release_last_Grt_Itr_Stk: Grt_Itr_Stk l_release_last_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  unfold l_release_last_grt.
  unfold leadstoPQ2.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  UnfoldStateAll.
  intros.
  unfold l_release_Post_aux_0 in H4.
  InvertAll.
  unfold itr_enabled in H0.
  UnfoldStateAll.
  inversion_clear H0.
  rewrite H8 in H7. inversion H7.
Qed.

Lemma l_release_Post_IF_false:
   forall S : State, l_release_Post S -> S_Flag S _IF = false.
Proof.
  intros.
  induction S as (((M, R), fl), isr).
  unfold l_release_Post in H.
  InvertAll.
  unfold l_release_Post_aux_0 in H0. UnfoldStateAll.
  InvertAll. auto.
Qed.

Lemma wff_cmd_l_rel_done_2:
 NH_WFiseq psi (l_release + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
   (Ast l_release_Post l_release_last_grt 1) (sseq _retw).
Proof.
  apply  nh_isq_end.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  split; auto.
  intros.
  unfold l_release_last_grt.
  unfold leadstoPQ2. auto.
  split.
  apply IF_false_Prd_Itr_Stk.
  apply l_release_Post_IF_false.
  apply l_release_last_Grt_Itr_Stk.
Qed.

Lemma wff_l_rel_done:
   NH_WFiseq psi (l_release + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
     (Ast l_rel_done_Prd l_rel_done_grt 1) I_l_rel_done.
Proof.
   unfold I_l_rel_done.
   apply nh_isq_seq with l_release_Post l_release_last_grt.
   apply wff_cmd_l_rel_done_1.
   apply wff_cmd_l_rel_done_2.
Qed.

Lemma wff_l_release:
  NH_WFiseq psi l_release l_release_spec I_l_release.
Proof.
  unfold I_l_release.
  unfold l_release_spec.
  apply nh_isq_seq with l_release1_Prd l_release1_grt.
  apply wff_cmd_l_release1.
  apply nh_isq_seq with l_release2_Prd l_release2_grt.
  apply wff_cmd_l_release2.
  apply nh_isq_seq with l_release3_Prd l_release3_grt.
  apply wff_cmd_l_release3.
  apply nh_isq_seq with l_release4_Prd l_release4_grt.
  apply wff_cmd_l_release4.
  apply nh_isq_seq with l_release_Prd_5 l_release_grt_5.
  apply wff_cmd_l_release5.
  apply nh_isq_seq with l_release_Prd_6 l_release_grt_6.
  apply wff_cmd_l_release6.
  apply nh_isq_seq with l_release_Prd_7 l_release_grt_7.
  apply wff_cmd_l_release7.
  apply nh_isq_seq with l_release_Prd_8 l_release_grt_8.
  apply wff_cmd_l_release8.
  apply nh_isq_seq with l_release_Prd_8b l_release_grt_8b.
  apply wff_cmd_l_release8b.
  apply nh_isq_seq with l_release_Prd_9 l_release_grt_9.
  apply wff_cmd_l_release9.
  apply nh_isq_seq with l_rel_done_Prd l_rel_done_grt.
  apply wff_cmd_l_releasea.
  apply wff_l_rel_done.
Qed.


(* the stack space required by l_release *)
Definition stk_lock_release := 6 + IH_Stk + stk_l_release.

Definition lock_release_MPost (sp l pc : Label) : MPred :=
    (sp |->w pc) *^* (sp + 2 |->w l) *^* stkSpacePrec sp stk_lock_release.

Definition lock_release_MPre (sp l pc: Label) : MPred :=
    (lock_release_MPost sp l pc) *^* (Gamma l).

Definition lock_release_Prd_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
      R SP = sp
      /\ R BP = bp
      /\ ((lock_release_MPre sp l pc) *^* mP) M
      /\ In l lockSet_H
      /\ fl _IF = true
      /\ isr = false
  end.

Definition lock_release_Prd (S : State) : Prop :=
  exists sp, exists l, exists pc, exists bp, exists mP, lock_release_Prd_aux sp l pc bp mP S.

Definition lock_release_Post_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
      R SP = sp
      /\ R BP = bp
      /\ (lock_release_MPost sp l pc *^* mP) M
      /\ In l lockSet_H
      /\ fl _IF = true
      /\ isr = false
  end.

Definition lock_release_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (lock_release_Prd_aux sp l pc bp) (lock_release_Post_aux sp l pc bp) S S'.

Definition lock_release_psi := SCAP_consCT l_release (Ast l_release_Prd l_release_grt 1) SCAP_emptCT.

Lemma lock_release_in_psi:
  SCAP_lookupCT lock_release_psi l_release (Ast l_release_Prd l_release_grt 1).
Proof.
  unfold SCAP_lookupCT.
  unfold lock_release_psi.
  unfold SCAP_consCT.
  unfold SPSpec.lookup.
  simpl. left. auto.
Qed.
Hint Resolve lock_release_in_psi.

(* spec for the next instruction *)
(* _pushwp (Opr_reg BP) *)

Definition lock_release1_MPre (sp l pc bp : Label) : MPred :=
    (sp |->w bp) *^* (sp + 2 |->w pc) *^* (sp + 4 |->w l) *^* stkSpacePrec sp (stk_lock_release - 2) *^* (Gamma l).

Definition lock_release1_Prd_aux (sp l pc bp : Label) (mP: MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R SP = sp
        /\ R BP = bp
        /\ (lock_release1_MPre (R SP) l pc (R BP) *^* mP) M
        /\ In l lockSet_H
        /\ fl _IF = true
        /\ isr = false
  end.

Definition lock_release1_Prd (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, lock_release1_Prd_aux sp l pc bp mP S.

Definition lock_release1_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (lock_release1_Prd_aux sp l pc bp) (lock_release_Post_aux (sp +2) l pc bp) S S'.


Definition lock_release2_Prd_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R BP = sp
        /\ exists bp0, lock_release1_Prd_aux sp l pc bp mP (M, RF_update R BP bp0, fl, isr)
  end.


Definition lock_release2_Prd (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, lock_release2_Prd_aux sp l pc bp mP S.

Definition lock_release2_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (lock_release2_Prd_aux sp l pc bp) (lock_release_Post_aux (sp +2) l pc bp) S S'.

Definition lock_release3_MPre (sp l pc bp : Label) :=
    (sp |->w l) *^* (sp + 2 |->w bp) *^* (sp + 4 |->w pc) *^* (sp + 6 |->w l) *^* stkSpacePrec sp (stk_lock_release - 4) *^* (Gamma l).

Definition lock_release3_Prd_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R SP = sp
        /\ (lock_release3_MPre (R SP) l pc bp *^* mP) M
        /\ In l lockSet_H
        /\ fl _IF = true
        /\ isr = false
  end.

Definition lock_release3_Prd (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, lock_release3_Prd_aux sp l pc bp mP S.


Definition lock_release3_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (lock_release3_Prd_aux sp l pc bp) (lock_release_Post_aux (sp + 4) l pc bp) S S'.


Definition lock_release4_MPre (sp l pc bp : Label) :=
    ((lock_release3_MPre sp l pc bp) *^* (INV1 *^* INV0)).



Definition lock_release4_Prd_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R SP = sp
        /\ (lock_release4_MPre (R SP) l pc bp *^* mP) M
        /\ In l lockSet_H
        /\ fl _IF = false
        /\ isr = false
  end.

Definition lock_release4_Prd (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, lock_release4_Prd_aux sp l pc bp mP S.


Definition lock_release4_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (lock_release4_Prd_aux sp l pc bp) (lock_release_Post_aux (sp + 4) l pc bp) S S'.


Definition lock_release6_MPre (sp l pc bp : Label) :=
    (sp |->w_) *^* (sp + 2 |->w bp) *^* (sp + 4 |->w pc) *^* (sp + 6 |->w l) *^* stkSpacePrec sp (stk_lock_release - 4).

Definition lock_release5_MPre (sp l pc bp : Label) :=
  (lock_release6_MPre sp l pc bp) *^* (INV1 *^* INV0).

Definition lock_release5_Prd_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R SP = sp
        /\ (lock_release5_MPre (R SP) l pc bp *^* mP) M
        /\ In l lockSet_H
        /\ fl _IF = false
        /\ isr = false
  end.

Definition lock_release5_Prd (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, lock_release5_Prd_aux sp l pc bp mP S.


Definition lock_release5_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (lock_release5_Prd_aux sp l pc bp) (lock_release_Post_aux (sp + 4) l pc bp) S S'.


Definition lock_release6_Prd_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R SP = sp
        /\ (lock_release6_MPre (R SP) l pc bp *^* mP) M
        /\ In l lockSet_H
        /\ fl _IF = true
        /\ isr = false
  end.

Definition lock_release6_Prd (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, lock_release6_Prd_aux sp l pc bp mP S.


Definition lock_release6_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (lock_release6_Prd_aux sp l pc bp) (lock_release_Post_aux (sp + 4) l pc bp) S S'.

(* _addw (Opr_imm 2) SP *)
Definition lock_release7_Prd_aux (sp l pc bp : Label) (mP : MPred) (S : State) : Prop :=
  match S with
    (M, R, fl, isr) =>
        R SP = sp
        /\ sp >= 2
        /\ lock_release6_Prd_aux (sp - 2) l pc bp mP (M, RF_update R SP (sp - 2), fl, isr)
  end.


Definition lock_release7_Prd (S : State) :=
  exists sp, exists l, exists pc, exists bp, exists mP, lock_release7_Prd_aux sp l pc bp mP S.

Definition lock_release7_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (lock_release7_Prd_aux sp l pc bp) (lock_release_Post_aux (sp +2) l pc bp) S S'.

Definition lock_release_Post (S : State) : Prop :=
  exists sp, exists l, exists pc, exists bp, exists mP, lock_release_Post_aux sp l pc bp mP S.

Definition lock_release_last_grt (S S' : State) : Prop :=
  forall sp l pc bp,
    leadstoPQ2 (lock_release_Post_aux sp l pc bp) (lock_release_Post_aux sp l pc bp) S S'.



Lemma lock_release_stk_ge2: stk_lock_release >= 2.
Proof.
  unfold stk_lock_release. omega.
Qed.
Hint Resolve lock_release_stk_ge2.

Lemma lock_release_Prd_aux_stkSpace:
  forall sp l pc bp mP S,
   lock_release_Prd_aux sp l pc bp mP S -> stkSpace (S_RF S SP) stk_lock_release (S_Mem S).
Proof.
  intros.
  unfold lock_release_Prd_aux in H.
  induction S as (((M, R), fl), isr).
  unfold lock_release_MPre in H.
  unfold lock_release_MPost in H.
  InvertAll.
  sep on.
  sep lift 2 in H1.
  UnfoldStateAll.
  rewrite H0.
  sep trivial.
Qed.


Lemma lock_release_step1:
forall f sp l pc bp mP S S',
  lock_release_Prd_aux sp l pc bp mP S  
  -> nextS f (comml (_pushwp (Opr_reg BP))) S S'
  -> lock_release1_Prd_aux (sp - 2) l pc bp mP S'.
Proof.
  intros.
  rename H0 into Htrans.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion Htrans).
  destruct H0 as [Hspgt2 [Hmemblk HM']].
  destruct H1 as [HR' [Hfl' Hisr']].
  subst R' fl' isr'. clear Htrans.
  unfold opr_eval in HM'.
  unfold lock_release_Prd_aux in H.
  destruct H as [HRsp [HRbp [Hmem [Hlock [Hfl Hisr]]]]].
  unfold lock_release_MPre in Hmem.
  unfold lock_release_MPost in Hmem.
  unfold lock_release1_Prd_aux.
  reg simpl.
  rewrite HRsp.
  split_l; split_l.
  split; auto.
  unfold lock_release1_MPre.
  rewrite HRbp.
  rewrite HRsp in Hspgt2.
  assert (sp - 2 + 2 = sp).
  assert (sp - 2 + 2 = sp + 2 - 2). auto.
  assert (sp + 2 - 2 = sp). auto.
  rewrite H0 in H. auto.
  rewrite H. clear H.
  assert (sp - 2 + 4 = sp + 2).
  assert (sp - 2 + 4 = sp + 4 - 2). auto.
  assert (sp +  4 - 2 = sp + 2); auto.
  rewrite H. clear H.
  substH HM' with (MemUpd16_MemUpdW HM').
  destruct HM' as [junk HM'].
  subst M'.
  sep on.
  rewrite HRsp. rewrite HRbp.
  sep lift 0.
  sep upd2.
  sep lift 2 in Hmem.
  sep stksplit in Hmem.
  sep matchH Hmem with 6.
Qed.
  
Lemma lock_release_stk_geIH: stk_lock_release >= IH_Stk + 6.
Proof.
  unfold stk_lock_release. omega.
Qed.
Hint Resolve lock_release_stk_geIH.

Lemma lock_release_Prd_Itr_Stk: Prd_Itr_Stk lock_release_Prd.
Proof.
    unfold Prd_Itr_Stk.
    intros S Hpre Henable.
    unfold lock_release_Prd in Hpre.
    destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
    unfold lock_release_Prd_aux in Hpre.
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in Henable.
    simplS in * .
    destruct Hpre as [HRsp [HRbp [Hmem [Hlock [Hfl Hisr]]]]].
    rewrite HRsp.
    unfold lock_release_MPre in Hmem.
    unfold lock_release_MPost in Hmem.
    sep on.
    sep lift 2 in Hmem.
    sep stksplit (IH_Stk + 6) in Hmem.
    sep lift 0 in Hmem.
    destruct Hmem as [M1 [M2 [Hsplit [Hmem1 Hmem2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    split_l.
    intros M1' Hmem1'.
    unfold lock_release_Prd.
    mexists5 sp l pc bp mP.
    unfold lock_release_Prd_aux.
    unfold lock_release_MPre. unfold lock_release_MPost.
    split_l. split_l.
    split; auto.
    sep on.
    sep lift 2.
    assert (sp >= IH_Stk + 6).
    apply stkSpacePrec_sp_ge_sz with M1; auto.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      split.
        disjoint rep_l M1.
        apply stkSpacePrec_stkSpacePrec_eqDom with sp (IH_Stk + 6); auto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      apply stkSpacePrec_stkSpacePrec_eqDom with sp (IH_Stk + 6); auto.
    split_l.
    sep on.
    sep matchH Hmem2 with 6.
Qed.

Lemma lock_release_Grt_Itr_Stk: Grt_Itr_Stk lock_release_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros S S' Hgrt Henable.
  unfold lock_release_grt in * .
  intros M1 M2 M2' Hsplit HM2 HM2'.
  unfold leadstoPQ2 in *.
  intros sp l pc bp mP Hpre.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in Henable.
  apply Hgrt.
  clear Hgrt.
  unfold lock_release_Prd_aux in * .
  simplS in * .
  destruct Hpre as [HRsp [HRbp [Hmem [Hlock [Hfl Hisr]]]]].
  split_l. split_l.
  split; auto.
  unfold lock_release_MPre in *.
  unfold lock_release_MPost in *.
  sep on.
  sep lift 2 in Hmem.
  sep stksplit (IH_Stk + 6) in Hmem.
  sep lift 0 in Hmem.
  destruct Hmem as [M2x [M1x [Hsplit' [HM2x HM1x]]]].
  sep lift 2.
  assert (sp >= IH_Stk + 6).
  apply stkSpacePrec_sp_ge_sz with M2x; auto.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  assert (M2x = M2' /\ M1x = M1).
    sep off.
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      apply stkSpacePrec_stkSpacePrec_eqDom with (R SP) (IH_Stk + 6); auto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    rewrite HRsp in HM2'.
    assert (Hx := prec_disjUnion_inv H0 H1 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H0; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  rewrite HRsp in HM2.
  split_l.
  sep on.
  sep matchH HM1x with 3.
Qed.


Lemma wff_cmd_lock_release1:
   NH_WFcmd
     (SCAP_consCT (lock_release + 1)
        (Ast lock_release1_Prd lock_release1_grt 1) lock_release_psi)
     lock_release (Ast lock_release_Prd lock_release_grt 1)
     (comml (_pushwp (Opr_reg BP))).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists lock_release1_Prd. exists lock_release1_grt.
  split. auto.

  assert (forall sp l pc bp mP, enable (lock_release_Prd_aux sp l pc bp mP) (nextS lock_release (comml (_pushwp (Opr_reg BP))))).
  intros.
  apply enable_pushwp with stk_lock_release; auto.
  apply lock_release_Prd_aux_stkSpace.

  split.
  unfold enable in *.
  intros.
  unfold lock_release_Prd in H0.
  InvertAll.
  apply H with x x0 x1 x2 x3; auto.
  split.
  unfold pimp.
  unfold linkPG.
  intros.
  InvertAll.
  unfold lock_release_Prd in H0.
  InvertAll.
  unfold lock_release1_Prd.
  mexists5 (x0 - 2) x1 x2 x3 x4.
  apply lock_release_step1 with lock_release x; auto.

  unfold gimp.
  unfold conjPG. unfold conjG.
  intros.
  InvertAll.
  unfold lock_release_grt.
  unfold leadstoPQ2.
  intros.
  unfold lock_release1_grt in H2.
  unfold leadstoPQ2 in H2.
  assert (lock_release_Post_aux (sp - 2 + 2) l pc bp mP S').
  apply H2.
  apply lock_release_step1 with lock_release S; auto.
  induction S as (((M, R), fl), isr).
  induction x as (((M', R'), fl'), isr').
  destruct (nextS_inversion H3).
  InvertAll.
  unfold lock_release_Prd_aux in H0.
  InvertAll.
  rewrite H10 in H8.
  assert (sp - 2 + 2 = sp). auto.
  rewrite H15 in H4. auto.

  split.
  apply lock_release_Prd_Itr_Stk.
  apply lock_release_Grt_Itr_Stk.
Qed.

Lemma lock_release_step2:
forall f sp l pc bp mP S S',
  lock_release1_Prd_aux sp l pc bp mP S
  -> nextS f (comml (_movwp (Opr_reg SP) BP)) S S'
  -> lock_release2_Prd_aux sp l pc bp mP S'.
Proof.
  intros.
  rename H0 into Htrans.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion Htrans).
  subst M'.
  destruct H1 as [HR' [Hfl' Hisr']].
  unfold opr_eval in HR'.
  subst R' fl' isr'. clear Htrans.
  unfold lock_release1_Prd_aux in H.
  destruct H as [HRsp [HRbp [Hmem [Hlock [Hfl Hisr]]]]].
  unfold lock_release2_Prd_aux.
  reg simpl.
  rewrite HRsp.
  split_l.
  exists (R BP).
  assert (RF_update (RF_update R BP sp) BP (R BP) = R).
  assert (RF_update (RF_update R BP sp) BP (R BP) = RF_update R BP (R BP)).
  apply RF_update_RF_update.
  rewrite H.
  apply RF_update_prev.
  rewrite H. clear H.
  unfold lock_release1_Prd_aux.
  split_l; split_l.
  split; auto.
Qed.

Lemma stk_lock_release_N2_geIH:
  stk_lock_release - 2 >= (IH_Stk + 6).
Proof.
  unfold stk_lock_release.
  unfold stk_l_release.
  omega.
Qed.
Hint Resolve stk_lock_release_N2_geIH.

Lemma lock_release1_Prd_Itr_Stk: Prd_Itr_Stk lock_release1_Prd.
Proof.
    unfold Prd_Itr_Stk.
    intros S Hpre Henable.
    unfold lock_release1_Prd in Hpre.
    destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
    unfold lock_release1_Prd_aux in Hpre.
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in Henable.
    simplS in * .
    destruct Hpre as [HRsp [HRbp [Hmem [Hlock [Hfl Hisr]]]]].
    rewrite HRsp.
    rewrite HRsp in Hmem.
    unfold lock_release1_MPre in Hmem.
    sep on.
    sep lift 3 in Hmem.
    sep stksplit (IH_Stk + 6) in Hmem.
    sep lift 0 in Hmem.
    destruct Hmem as [M1 [M2 [Hsplit [Hmem1 Hmem2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    split_l.
    intros M1' Hmem1'.
    unfold lock_release1_Prd.
    mexists5 sp l pc bp mP.
    unfold lock_release1_Prd_aux.
    unfold lock_release1_MPre.
    split_l. split_l.
    split; auto.
    sep on.
    sep lift 3.
    assert (sp >= IH_Stk + 6).
    apply stkSpacePrec_sp_ge_sz with M1; auto.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      split.
        disjoint rep_l M1.
        apply stkSpacePrec_stkSpacePrec_eqDom with sp (IH_Stk + 6); auto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      apply stkSpacePrec_stkSpacePrec_eqDom with sp (IH_Stk + 6); auto.
    rewrite HRsp.
    split_l.
    sep on.
    sep matchH Hmem2 with 7.
Qed.

Lemma lock_release1_Grt_Itr_Stk: Grt_Itr_Stk lock_release1_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros S S' Hgrt Henable.
  unfold lock_release1_grt in * .
  intros M1 M2 M2' Hsplit HM2 HM2'.
  unfold leadstoPQ2 in *.
  intros sp l pc bp mP Hpre.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in Henable.
  apply Hgrt.
  clear Hgrt.
  unfold lock_release1_Prd_aux in * .
  simplS in * .
  destruct Hpre as [HRsp [HRbp [Hmem [Hlock [Hfl Hisr]]]]].
  split_l. split_l.
  split; auto.
  unfold lock_release1_MPre in *.
  sep on.
  sep lift 3 in Hmem.
  sep stksplit (IH_Stk + 6) in Hmem.
  sep lift 0 in Hmem.
  destruct Hmem as [M2x [M1x [Hsplit' [HM2x HM1x]]]].
  sep lift 3.
  assert (R SP >= IH_Stk + 6).
  apply stkSpacePrec_sp_ge_sz with M2x; auto.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  assert (M2x = M2' /\ M1x = M1).
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      apply stkSpacePrec_stkSpacePrec_eqDom with (R SP) (IH_Stk + 6); auto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    assert (Hx := prec_disjUnion_inv H0 H1 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H0; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  split_l.
  sep on.
  sep matchH HM1x with 6.
Qed.

Lemma wff_cmd_lock_release2:
   NH_WFcmd
     (SCAP_consCT (lock_release + 1 + 1)
        (Ast lock_release2_Prd lock_release2_grt 1) lock_release_psi)
     (lock_release + 1) (Ast lock_release1_Prd lock_release1_grt 1)
     (comml (_movwp (Opr_reg SP) BP)).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists lock_release2_Prd. exists lock_release2_grt.
  split. auto.
  split; auto. (* enable *)
  split.
  unfold pimp.
  unfold linkPG.
  intros.
  InvertAll.
  unfold lock_release1_Prd in H.
  InvertAll.
  unfold lock_release2_Prd.
  mexists5 x0 x1 x2 x3 x4.
  apply lock_release_step2 with (lock_release + 1) x; auto.

  unfold gimp.
  unfold conjPG. unfold conjG.
  intros.
  InvertAll.
  unfold lock_release1_grt.
  unfold leadstoPQ2.
  intros.
  unfold lock_release2_grt in H1.
  unfold leadstoPQ2 in H1.
  apply H1.
  apply lock_release_step2 with (lock_release + 1) S; auto.

  split. 
    apply lock_release1_Prd_Itr_Stk.
    apply lock_release1_Grt_Itr_Stk.
Qed.


Lemma stk_lock_release_N2_ge2:
  stk_lock_release - 2 >= 2.
Proof.
  unfold stk_lock_release. omega.
Qed.
Hint Resolve stk_lock_release_N2_ge2.



Lemma lock_release2_Prd_1:
   forall sp l pc bp mP S,
   lock_release2_Prd_aux sp l pc bp mP S ->
   ((addr_eval (S_RF S) (A_reg 4 BP)) |~>w l) (S_Mem S).
Proof.
  intros.
  induction S as (((M, R), fl), isr).
  unfold lock_release2_Prd_aux in H.
  unfold S_RF. unfold S_Mem.
  unfold addr_eval.
  inversion_clear H. rewrite H0.
  inversion_clear H1.
  unfold lock_release1_Prd_aux in H.
  unfold lock_release1_MPre in H.
  reg simpl.
  InvertAll.
  rewrite <- H1.
  sep on.
  sep lift 2 in H2.
  sep trivial.
Qed.

Lemma lock_release2_Prd_enable:
   forall f sp l pc bp mP,
   enable (lock_release2_Prd_aux sp l pc bp mP)
     (nextS f (comml (_pushwm (A_reg 4 BP)))).
Proof.
  unfold enable.
  intros.
  assert (((addr_eval (S_RF S) (A_reg 4 BP)) |~>w l) (S_Mem S)).
  apply lock_release2_Prd_1 with sp pc bp mP; auto.
  induction S as (((M, R), fl), isr).
  exists (MemUpdW M (R SP - 2) l 0, RF_update R SP (R SP - 2), fl, isr).
  apply nextS_pushwm with (addr_eval R (A_reg 4 BP)) l. auto.
  auto.
  unfold pushW.
  unfold lock_release2_Prd_aux in H.
  destruct H as [HRbp [junk Hpre]].
  unfold lock_release1_Prd_aux in Hpre.
  unfold lock_release1_MPre in Hpre.
  reg simpl.
  destruct Hpre as [HRsp [Hjunk [Hmem [Hlock [Hfl Hisr]]]]].
  split.
  sep on.
  sep lift 3 in Hmem.
  sep stksplit 2 in Hmem.
  sep lift 0 in Hmem.
  sep trivial.
  split; auto.
  apply  MemUpdW_MemUpd16 with 0; auto.
Qed.

Lemma lock_release_step3:
forall f sp l pc bp mP S S',
  lock_release2_Prd_aux sp l pc bp mP S  
  -> nextS f (comml (_pushwm (A_reg 4 BP))) S S'
  -> lock_release3_Prd_aux (sp - 2) l pc bp mP S'.
Proof.
  intros.
  rename H0 into Htrans.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion Htrans).
  destruct H0 as [w [Hsrc [Hspgt2 [Hmemblk HM']]]].
  destruct H1 as [HR' [Hfl' Hisr']].
  subst R' fl' isr'. clear Htrans.
  assert ((addr_eval (S_RF (M, R, fl, isr)) (A_reg 4 BP) |~>w l) (S_Mem (M, R, fl, isr))).
  apply lock_release2_Prd_1 with sp pc bp mP; auto.
  UnfoldStateAll.
  assert (w = l).
  apply leadsToW_leadsToW_eq with M (addr_eval R (A_reg 4 BP)); auto.
  subst w. clear H0.
  unfold addr_eval in Hsrc.
  unfold lock_release2_Prd_aux in H.
  destruct H as [HRbp [bp0 Hpre]].
  unfold lock_release1_Prd_aux in Hpre.
  unfold lock_release1_MPre in Hpre.
  reg simpl.
  destruct Hpre as [HRsp [Htmp [Hmem [Hlock [Hfl Hisr]]]]].
  unfold lock_release3_Prd_aux.
  reg simpl.
  split. rewrite HRsp. auto.
  split; auto.
  unfold lock_release3_MPre.
  substH HM' with (MemUpd16_MemUpdW HM').
  destruct HM' as [junk HM'].
  subst M'.
  sep on.
  sep lift 0.
  sep upd2.
  sep lift 3 in Hmem.
  sep stksplit in Hmem.
  sep lift 1 in Hmem.
  assert (R SP - 2 + 2 = R SP). auto.
  rewrite H. clear H.
  assert (R SP - 2 + 4 = R SP + 2).
  assert (R SP - 2 + 4 = R SP + 4 - 2). auto.
  assert (R SP +  4 - 2 = R SP + 2); auto.
  rewrite H. clear H.
  assert (R SP - 2 + 6 = R SP + 4).
  assert (R SP - 2 + 6 = R SP + 6 - 2). auto.
  assert (R SP + 6 - 2 = R SP + 4).
  auto.
  rewrite H0 in H. auto.
  rewrite H. clear H.
  rewrite Htmp in Hmem.
  assert (stk_lock_release - 2 - 2 = stk_lock_release - 4). auto with arith.
  rewrite H in Hmem.
  sep matchH Hmem with 7.
Qed.

Lemma lock_release2_Prd_Itr_Stk: Prd_Itr_Stk lock_release2_Prd.
Proof.
    unfold Prd_Itr_Stk.
    intros S Hpre Henable.
    unfold lock_release2_Prd in Hpre.
    destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
    unfold lock_release2_Prd_aux in Hpre.
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in Henable.
    simplS in * .
    destruct Hpre as [HRbp [bp0 Hpre]].
    unfold lock_release1_Prd_aux in Hpre.
    reg simpl.
    destruct Hpre as [HRsp [Hbp0 [Hmem [Hlock [Hfl Hisr]]]]].
    unfold lock_release1_MPre in Hmem.
    sep on.
    sep lift 3 in Hmem.
    sep stksplit (IH_Stk + 6) in Hmem.
    sep lift 0 in Hmem.
    destruct Hmem as [M1 [M2 [Hsplit [Hmem1 Hmem2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    split_l.
    intros M1' Hmem1'.
    unfold lock_release2_Prd.
    mexists5 sp l pc bp mP.
    unfold lock_release2_Prd_aux.
    split_l.
    exists bp0.
    unfold lock_release1_Prd_aux.
    unfold lock_release1_MPre.
    reg simpl.
    split_l. split_l.
    split; auto.
    sep on.
    sep lift 3.
    assert (R SP >= IH_Stk + 6).
    apply stkSpacePrec_sp_ge_sz with M1; auto.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      split.
        disjoint rep_l M1.
        apply stkSpacePrec_stkSpacePrec_eqDom with (R SP) (IH_Stk + 6); auto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      apply stkSpacePrec_stkSpacePrec_eqDom with (R SP) (IH_Stk + 6); auto.
    split_l.
    sep on.
    sep matchH Hmem2 with 6.
Qed.

Lemma lock_release2_Grt_Itr_Stk: Grt_Itr_Stk lock_release2_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros S S' Hgrt Henable.
  unfold lock_release2_grt in * .
  intros M1 M2 M2' Hsplit HM2 HM2'.
  unfold leadstoPQ2 in *.
  intros sp l pc bp mP Hpre.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in Henable.
  apply Hgrt.
  clear Hgrt.
  unfold lock_release2_Prd_aux in * .
  simplS in * .
  destruct Hpre as [HRbp [bp0 Hpre]].
  unfold lock_release1_Prd_aux in *.
  reg simpl.
  destruct Hpre as [HRsp [HRbp0 [Hmem [Hlock [Hfl Hisr]]]]].
  split_l. exists bp0.
  reg simpl.
  split_l. split_l.
  split; auto.
  unfold lock_release1_MPre in *.
  sep on.
  sep lift 3 in Hmem.
  sep stksplit (IH_Stk + 6) in Hmem.
  sep lift 0 in Hmem.
  destruct Hmem as [M2x [M1x [Hsplit' [HM2x HM1x]]]].
  sep lift 3.
  assert (R SP >= IH_Stk + 6).
  apply stkSpacePrec_sp_ge_sz with M2x; auto.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  assert (M2x = M2' /\ M1x = M1).
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      apply stkSpacePrec_stkSpacePrec_eqDom with (R SP) (IH_Stk + 6); auto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    assert (Hx := prec_disjUnion_inv H0 H1 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H0; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  split_l.
  sep on.
  sep matchH HM1x with 6.
Qed.

Lemma wff_cmd_lock_release3:
   NH_WFcmd
     (SCAP_consCT (lock_release + 1 + 1 + 1)
        (Ast lock_release3_Prd lock_release3_grt 1) lock_release_psi)
     (lock_release + 1 + 1) (Ast lock_release2_Prd lock_release2_grt 1)
     (comml (_pushwm (A_reg 4 BP))).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists lock_release3_Prd. exists lock_release3_grt.
  split. auto.

  generalize (lock_release + 1 + 1). intro f.

  split.
  assert (forall f sp l pc bp mP,
   enable (lock_release2_Prd_aux sp l pc bp mP)
     (nextS f (comml (_pushwm (A_reg 4 BP))))).
  apply lock_release2_Prd_enable.
  unfold enable in *.
  intros.
  unfold lock_release2_Prd in H0.
  InvertAll.
  apply H with x x0 x1 x2 x3; auto.

  split.
  unfold pimp.
  unfold linkPG.
  intros.
  InvertAll.
  unfold lock_release2_Prd in H.
  InvertAll.
  unfold lock_release3_Prd.
  mexists5 (x0 - 2) x1 x2 x3 x4.
  apply lock_release_step3 with f x; auto.

  unfold gimp.
  unfold conjPG. unfold conjG.
  intros.
  InvertAll.
  unfold lock_release2_grt.
  unfold leadstoPQ2.
  intros.
  unfold lock_release3_grt in H1.
  unfold leadstoPQ2 in H1.
  assert (lock_release_Post_aux (sp - 2 + 4) l pc bp mP S').
  apply H1.
  apply lock_release_step3 with f S; auto.
  induction S as (((M, R), fl), isr).
  induction x as (((M', R'), fl'), isr').
  destruct (nextS_inversion H2).
  InvertAll.
  unfold lock_release2_Prd_aux in H.
  unfold lock_release1_Prd_aux in H.
  InvertAll.
  reg simpl.
  rewrite <- H12.
  rewrite <- H12 in H3.
  assert (R SP - 2 + 4 = R SP + 2).
  assert (R SP - 2 + 4 = R SP + 4 - 2); auto.
  assert (R SP + 4 - 2 = R SP + 2); auto.
  rewrite H16 in H3. auto.
  
  split.
  apply lock_release2_Prd_Itr_Stk.
  apply lock_release2_Grt_Itr_Stk.
Qed.

Lemma lock_release_step4:
forall sp l pc bp mP S S',
  lock_release3_Prd_aux sp l pc bp mP S
  -> gcli S S'
  -> lock_release4_Prd_aux sp l pc bp mP S'.
Proof.
  intros.
  rename H0 into Htrans.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  unfold gcli in Htrans.
  UnfoldStateAll.
  destruct Htrans as [Hpp [Hfl' [Hisr' HR']]].
  subst R' isr'.
  unfold prepost in Hpp.
  destruct Hpp as [M1 [M2 [M1' [HM1 [HM1' [HsplitM HsplitM']]]]]].
  UnfoldStateAll.
  unfold lock_release4_Prd_aux.
  unfold lock_release3_Prd_aux in H.
  destruct H as [HRsp [Hmem [Hlock [Hfl Hisr]]]].
  split_l. split; auto.
  unfold lock_release4_MPre.
  assert ((INV1 *^* INV0 *^* (lock_release3_MPre (R SP) l pc bp *^* mP)) M').
  exists M1'. exists M2.
  split_l.
  unfold Aemp in HM1.
  rewrite HM1 in HsplitM.
  assert (M2 = M); auto with PMap.
  subst M2.
  split; auto.
  unfold choice in HM1'.
  inversion_clear HM1'.
  inversion H; auto.
  inversion H. assert (itr_enabled (M, R, fl, isr)).
  unfold itr_enabled. UnfoldStateAll. auto.
  contradiction.
  sep on.
  sep matchH H with 4.
Qed.

Lemma lock_release3_Prd_Itr_Stk: Prd_Itr_Stk lock_release3_Prd.
Proof.
    unfold Prd_Itr_Stk.
    intros S Hpre Henable.
    unfold lock_release3_Prd in Hpre.
    destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
    unfold lock_release3_Prd_aux in Hpre.
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in Henable.
    simplS in * .
    destruct Hpre as [HRsp [Hmem [Hlock [Hfl Hisr]]]].
    unfold lock_release3_MPre in Hmem.
    sep on.
    sep lift 4 in Hmem.
Lemma stk_lock_release_N4_geIH:
  stk_lock_release - 4 >= IH_Stk + 6.
Proof.
  unfold stk_lock_release.
  unfold stk_l_release. omega.
Qed.
Hint Resolve stk_lock_release_N4_geIH.
    sep stksplit (IH_Stk + 6) in Hmem.
    sep lift 0 in Hmem.
    destruct Hmem as [M1 [M2 [Hsplit [Hmem1 Hmem2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    split_l.
    intros M1' Hmem1'.
    unfold lock_release3_Prd.
    mexists5 sp l pc bp mP.
    unfold lock_release3_Prd_aux.
    split_l.
    split; auto.
    unfold lock_release3_MPre.
    sep on.
    sep lift 4.
    assert (R SP >= IH_Stk + 6).
    apply stkSpacePrec_sp_ge_sz with M1; auto.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      split.
        disjoint rep_l M1.
        apply stkSpacePrec_stkSpacePrec_eqDom with (R SP) (IH_Stk + 6); auto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      apply stkSpacePrec_stkSpacePrec_eqDom with (R SP) (IH_Stk + 6); auto.
    split_l.
    sep on.
    sep matchH Hmem2 with 7.
Qed.



Lemma lock_release3_Grt_Itr_Stk : Grt_Itr_Stk lock_release3_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros S S' Hgrt Henable.
  unfold lock_release3_grt in * .
  intros M1 M2 M2' Hsplit HM2 HM2'.
  unfold leadstoPQ2 in *.
  intros sp l pc bp mP Hpre.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in Henable.
  apply Hgrt.
  clear Hgrt.
  unfold lock_release3_Prd_aux in * .
  simplS in * .
  destruct Hpre as [HRsp [Hmem [Hlock [Hfl Hisr]]]].
  split_l. 
  unfold lock_release3_MPre in *.
  split; auto.
  sep on.
  sep lift 4 in Hmem.
  sep stksplit (IH_Stk + 6) in Hmem.
  sep lift 0 in Hmem.
  destruct Hmem as [M2x [M1x [Hsplit' [HM2x HM1x]]]].
  sep lift 4.
  assert (R SP >= IH_Stk + 6).
  apply stkSpacePrec_sp_ge_sz with M2x; auto.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  assert (M2x = M2' /\ M1x = M1).
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      apply stkSpacePrec_stkSpacePrec_eqDom with (R SP) (IH_Stk + 6); auto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    assert (Hx := prec_disjUnion_inv H0 H1 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H0; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  split_l.
  sep on.
  sep matchH HM1x with 7.
Qed.


Lemma wff_cmd_lock_release4:
   NH_WFcmd
     (SCAP_consCT (lock_release + 1 + 1 + 1 + 1)
        (Ast lock_release4_Prd lock_release4_grt 1) lock_release_psi)
     (lock_release + 1 + 1 + 1) (Ast lock_release3_Prd lock_release3_grt 1)
     (comml _cli).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  exists lock_release4_Prd. exists lock_release4_grt.
  split. auto.
  split.
  unfold pimp.
  unfold linkPG.
  intros.
  InvertAll.
  unfold lock_release3_Prd in H.
  InvertAll.
  unfold lock_release4_Prd.
  mexists5 x0 x1 x2 x3 x4.
  apply lock_release_step4 with x; auto.

  unfold gimp.
  unfold conjPG. unfold conjG.
  intros.
  InvertAll.
  unfold lock_release3_grt.
  unfold leadstoPQ2.
  intros.
  unfold lock_release4_grt in H1.
  unfold leadstoPQ2 in H1.
  apply H1.
  apply lock_release_step4 with S; auto.
  split.
  apply lock_release3_Prd_Itr_Stk.
  apply lock_release3_Grt_Itr_Stk.
Qed.

Parameter l_release_not_primentry:
  ~ prim_entry l_release.

Lemma lock_release4_enable:
   forall f sp l pc bp mP,
   enable (lock_release4_Prd_aux sp l pc bp mP)
     (nextS f (comml (_call l_release))).
Proof.
  intros.
  unfold enable.
  intros.
  unfold lock_release4_Prd_aux in H.
  induction S as (((M, R), fl), isr).
  exists (MemUpdW M (R SP - 2) (f+1) 0, RF_update R SP (R SP - 2), fl, isr).
  apply nextS_call. 
  unfold pushW.
  destruct H as [HRsp [Hmem [Hlock [Hfl Hisr]]]].
  split.
  unfold lock_release4_MPre in Hmem.
  unfold lock_release3_MPre in Hmem.
  sep on.
  sep lift 4 in Hmem.
  sep stksplit 2 in Hmem.
  sep lift 0 in Hmem.
  sep trivial.
  split; auto.
  apply  MemUpdW_MemUpd16 with 0; auto.
Qed.

Definition l_release_closure_step5 (sp l pc bp : int) (mP: MPred) : MPred :=
  (sp + 2 |->w bp *^* sp + 4 |->w pc *^* sp + 6 |->w l) *^* stkSpacePrec (sp - 2 - stk_l_release) IH_Stk *^* mP.

Lemma stk_lock_release_call1:
stk_lock_release - 4 - 2 - stk_l_release = IH_Stk.
Proof.
  unfold stk_lock_release. omega.
Qed.
Hint Resolve stk_lock_release_call1.

Lemma stk_lock_release_call2:
stk_lock_release - 4 - 2 >= stk_l_release.
Proof.
  unfold stk_lock_release. omega.
Qed.
Hint Resolve stk_lock_release_call2.

Lemma lock_release_step5_call3:
  forall f sp l pc bp mP S S', 
    lock_release4_Prd_aux sp l pc bp mP S
    -> nextS f (comml (_call l_release)) S S'
    -> l_release_Prd_aux (sp - 2) l (f + 1) (S_RF S' BP) (l_release_closure_step5 sp l pc bp mP) S'.
Proof.
  intros.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion H0).
  destruct H1 as [Hspgt2 [Hpcspace HM']].
  destruct H2 as [HR' [Hfl' Hisr']].
  substH HM' with (MemUpd16_MemUpdW HM').
  destruct HM' as [junk HM'].
  subst M' R' fl' isr'. clear H0.
  UnfoldStateAll.
  reg simpl.
  unfold lock_release4_Prd_aux in H.
  destruct H as [HRsp [Hmem [Hlock [Hfl Hisr]]]].
  rewrite HRsp in Hmem.
  unfold lock_release4_MPre in Hmem.
  unfold lock_release3_MPre in Hmem.
  unfold l_release_Prd_aux.
  reg simpl.
  rewrite HRsp. split_l. split_l.
  split; auto.
  unfold l_release_MPre.
  unfold l_release_closure_step5.
  sep on.
  sep lift 0.
  sep upd2.
  sep lift 4 in Hmem.
  sep stksplit in Hmem.
  sep stksplit stk_l_release in Hmem.
  assert (stk_lock_release - 4 - 2 - stk_l_release = IH_Stk).
  auto.
  rewrite H in Hmem.
  rewrite HRsp in Hspgt2.
  assert (sp - 2 + 2 = sp). auto.
  rewrite H0.
  sep matchH Hmem with 11.
Qed.

Lemma lock_release_step5_call2: 
  forall f, gimp (sp_call_pred lock_release4_Prd f l_release >o l_release_grt) fun_frame.
Proof.
  unfold gimp.
  unfold sp_call_pred. unfold conjPG. unfold linkPG.
  intros.
  destruct H as [[S0 [Hpre Htrans]] Hgrt].
  unfold lock_release4_Prd in Hpre.
  destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
  assert (l_release_Prd_aux (sp - 2) l (f + 1) (S_RF S BP) (l_release_closure_step5 sp l pc bp mP) S).
  apply lock_release_step5_call3 with S0; auto.
  rename H into HFpre.
  unfold l_release_grt in Hgrt.
  unfold leadstoPQ2 in Hgrt.
  assert (l_release_Post_aux_0 (sp - 2) l (f+1) (S_RF S BP) (l_release_closure_step5 sp l pc bp mP) S').
  apply Hgrt; auto.
  clear Hgrt.
  unfold fun_frame.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  unfold l_release_Prd_aux in HFpre.
  unfold l_release_Post_aux_0 in H.
  UnfoldStateAll.
  destruct HFpre as [HRsp [HRbp [Hmem [Hlock [Hfl Hisr]]]]].
  destruct H as [HRsp' [HRbp' [Hmem' [Hfl' Hisr']]]].
  split.
  rewrite HRsp. rewrite HRsp'. auto.
  exists (f + 1).
  unfold Ret_Addr.
  UnfoldStateAll.
  split.
  intro.
  apply nextPC_retw with (sp - 2). auto.
  unfold l_release_MPre in Hmem.
  sep on.
  sep lift 0 in Hmem.
  sep trivial.
  intro.
  apply nextPC_retw with (sp - 2); auto.
  unfold l_release_MPost in Hmem'.
  substH Hmem' with (sep_pure_a
    (sp - 2 |->w f + 1 *^* sp - 2 + 2 |->w l *^*
             stkSpacePrec (sp - 2) stk_l_release *^* INV1 *^* INV0)
    (In l lockSet_H) (l_release_closure_step5 sp l pc bp mP) M' Hmem').
  inversion_clear Hmem'.
  sep on.
  sep lift 0 in H.
  sep trivial.
Qed.

Lemma lock_release_step5_call4_aux:
forall f f' sp l pc bp mP S S',
  lock_release4_Prd_aux sp l pc bp mP S
  -> wrap_fun l_release_grt f l_release f' S S'
  -> lock_release5_Prd_aux sp l pc bp mP S'.
Proof.
  intros.
  rename H into Hpre.
  unfold wrap_fun in H0.
  unfold conjG in H0.
  destruct H0 as [S2 [[S1 [Hcall Hgrt]] Hretw]].
  assert (sp >= 2).
  unfold lock_release4_Prd_aux in Hpre.
  induction S as (((M, R), fl), isr).
  induction S1 as (((M', R'), fl'), isr').
  InvertAll.
  destruct (nextS_inversion Hcall).
  InvertAll.
  rewrite H in H7; auto.
  rename H into Hspge2.
  assert (l_release_Prd_aux (sp - 2) l (f + 1) (S_RF S1 BP) (l_release_closure_step5 sp l pc bp mP) S1).
  apply lock_release_step5_call3 with S; auto.
  clear Hpre.
  rename H into HFpre.
  unfold l_release_grt in Hgrt.
  unfold leadstoPQ2 in Hgrt.
  assert (l_release_Post_aux_0 (sp - 2) l (f+1) (S_RF S1 BP) (l_release_closure_step5 sp l pc bp mP) S2).
  apply Hgrt; auto.
  clear HFpre Hgrt Hcall.
  rename H into Hpre.
  induction S2 as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion Hretw).
  subst M'.
  destruct H0 as [[pc'' [HMpc HR']] [Hfl' Hisr']].
  subst fl' isr' R'. clear Hretw. clear HMpc.
  unfold l_release_Post_aux_0 in Hpre.
  destruct Hpre as [HRsp [HRbp [Hmem [Hfl Hisr]]]].
  unfold lock_release5_Prd_aux.
  reg simpl.
  unfold l_release_MPost in Hmem.
  substH Hmem with (sep_pure_a
    (sp - 2 |->w f + 1 *^* sp - 2 + 2 |->w l *^*
            stkSpacePrec (sp - 2) stk_l_release *^* INV1 *^* INV0)
    (In l lockSet_H) (l_release_closure_step5 sp l pc bp mP) M Hmem).
  destruct Hmem as [Hmem Hlock].
  assert (R SP + 2 = sp).
  rewrite HRsp.
  assert (sp - 2 + 2 = sp + 2 - 2). auto.
  rewrite H. auto.
  split_l.
  split; auto.
  rewrite H.
  unfold lock_release5_MPre.
  unfold lock_release6_MPre.
  unfold l_release_closure_step5 in Hmem.
  sep on.
  sep lift 4.
  sep stksplit.
  sep lift 2 in Hmem.
  assert (sp - 2 >= stk_l_release).
  eapply stkSpacePrec_sp_ge_sz1; eauto.
  assert (stk_lock_release - 4 - 2 >= stk_l_release).
  auto.
  sep stksplit stk_l_release.
  assert (stk_lock_release - 4 - 2 - stk_l_release = IH_Stk).
  auto.
  rewrite H2.
  assert (sp - 2 + 2 = sp). auto.
  rewrite H3 in Hmem.
  sep matchH Hmem with 10.
Qed.


Lemma lock_release_step5_call4:
forall f pc' : nat,
    pimp (lock_release4_Prd |> wrap_fun l_release_grt f l_release pc')
      lock_release5_Prd.
Proof.
  unfold pimp. unfold linkPG.
  intros.
  destruct H as [S0 [Hpre Htrans]].
  unfold lock_release4_Prd in Hpre.
  destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
  unfold lock_release5_Prd.
  mexists5 sp l pc bp mP.
  apply lock_release_step5_call4_aux with f pc' S0; auto.
Qed.


Lemma lock_release_step5_call5:
forall f pc' : nat,
    gimp
      (lock_release4_Prd >o wrap_fun l_release_grt f l_release pc'
       oo lock_release5_grt) lock_release4_grt.
Proof.
  intros.
  unfold gimp.
  unfold conjPG. unfold conjG.
  intros.
  destruct H as [S'' [[Hpre Hfun] Htrans5]].
  unfold lock_release4_grt.
  unfold leadstoPQ2.
  intros.
  unfold lock_release5_grt in Htrans5.
  unfold leadstoPQ2 in Htrans5.
  apply Htrans5.
  apply lock_release_step5_call4_aux with f pc' S; auto.
Qed.

Lemma lock_release4_Prd_IF_false:
   forall S : State, lock_release4_Prd S -> S_Flag S _IF = false.
Proof.
  intros.
  unfold lock_release4_Prd in H.
  unfold lock_release4_Prd_aux in H.
  induction S as (((M, R), fl), isr).
  InvertAll.
  unfold S_Flag. auto.
Qed.

Lemma lock_release4_Grt_Itr_Stk:
   Grt_Itr_Stk lock_release4_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  induction S as (((M, R), fl), isr).
  unfold lock_release4_grt. unfold leadstoPQ2.
  intros.
  unfold lock_release4_Prd_aux in H4.
  UnfoldStateAll.
  InvertAll.
  unfold itr_enabled in H0.
  inversion_clear H0.
  UnfoldStateAll.
  rewrite H8 in H7. inversion H7.
Qed.


Lemma wff_cmd_lock_release5:
   NH_WFcmd
     (SCAP_consCT (lock_release + 1 + 1 + 1 + 1 + 1)
        (Ast lock_release5_Prd lock_release5_grt 1) lock_release_psi)
     (lock_release + 1 + 1 + 1 + 1)
     (Ast lock_release4_Prd lock_release4_grt 1) (comml (_call l_release)).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux. right.
  split. apply l_release_not_primentry.
  unfold NH_Normal_Call.
  exists l_release_Prd. exists lock_release5_Prd.
  exists l_release_grt. exists lock_release5_grt.

  split. unfold SCAP_lookupCT. unfold SCAP_consCT.

  unfold SPSpec.lookup. unfold ListT.In. right.
  assert (SCAP_lookupCT lock_release_psi l_release (Ast l_release_Prd l_release_grt 1)).
  auto. auto. split; auto.
  generalize (lock_release + 1 + 1 + 1 + 1).
  intro f.

  split.
  assert (forall sp l pc bp mP, enable (lock_release4_Prd_aux sp l pc bp mP) (nextS f (comml (_call l_release)))).
  apply lock_release4_enable.
  unfold enable in *.
  intros.
  unfold lock_release4_Prd in H0.
  InvertAll.
  apply H with x x0 x1 x2 x3; auto.
  split. apply lock_release_step5_call2.
  split. unfold pimp. unfold linkPG.
  intros. InvertAll.
  unfold lock_release4_Prd in H.
  InvertAll.
  unfold l_release_Prd.
  mexists5 (x0 - 2) x1 (f+1) (S_RF S BP) (l_release_closure_step5 x0 x1 x2 x3 x4).
  apply lock_release_step5_call3 with x; auto.
  split. intro.
  apply lock_release_step5_call4.
  intro. apply lock_release_step5_call5.

  split. 
    apply IF_false_Prd_Itr_Stk.
    apply lock_release4_Prd_IF_false.
    apply lock_release4_Grt_Itr_Stk.
Qed.
  
Lemma lock_release_step6_enable:
  enable_sti lock_release5_Prd.
Proof.
  unfold lock_release5_Prd. unfold enable_sti.
  unfold lock_release5_Prd_aux.
  intros.
  induction S as (((M, R), fl), isr).
  destruct H as [sp [l [pc [bp [m [HRsp [Hmem [Hlock [Hfl Hisr]]]]]]]]].
  unfold lock_release5_MPre in Hmem.
  assert ((INV1 *^* INV0 *^* (lock_release6_MPre (R SP) l pc bp *^* m)) M).
  sep on.
  sep matchH Hmem with 4.
  clear Hmem. rename H into Hmem.
  destruct Hmem as [M1 [M2 [HMsplit [Hinv Hrest]]]].
  exists M2. exists M1.
  UnfoldStateAll.
  split; auto with PMap.
  split. unfold pTrue. auto.
  unfold choice. right.
  split; auto.
  intro. unfold itr_enabled in H.
  UnfoldStateAll. inversion H.
  rewrite H0 in Hfl. inversion Hfl.
Qed.


Lemma lock_release_step6_Prd_aux:
forall sp l pc bp mP S S',
  lock_release5_Prd_aux sp l pc bp mP S
  -> gsti S S'
  -> lock_release6_Prd_aux sp l pc bp mP S'.
Proof.
  unfold lock_release5_Prd_aux.
  intros.
  induction S as (((M, R), fl), isr).
  destruct H as [HRsp [Hmem [Hlock [Hfl Hisr]]]].
  rename H0 into Htrans.
  induction S' as (((M', R'), fl'), isr').
  unfold gsti in Htrans.
  UnfoldStateAll.
  destruct Htrans as [Hpp [Hfl' [Hisr' HR']]].
  subst R' isr'.
  unfold prepost in Hpp.
  destruct Hpp as [M1 [M2 [M1' [HM1 [Hm1' [HsplitM HsplitM']]]]]].
  UnfoldStateAll.
  assert (M2 = M').
  unfold Aemp in Hm1'.
  rewrite Hm1' in HsplitM'. auto with PMap.
  subst M'. clear HsplitM' Hm1'  M1'.
  unfold choice in HM1.
  inversion_clear HM1.
  inversion H. unfold itr_enabled in H0.
  inversion_clear H0. unfold S_Flag in H2.
  rewrite H2 in Hfl. inversion Hfl.
  inversion_clear H. clear H0. rename H1 into Hinv.
  unfold lock_release5_MPre in Hmem.
  assert ((INV1 *^* INV0 *^* (lock_release6_MPre (R SP) l pc bp *^* mP)) M).
  sep on. sep matchH Hmem with 4.
  clear Hmem. rename H into Hmem.
  destruct Hmem as [M1x [M2x [HMsplit' [Hinv' Hrest]]]].
  assert (M1 = M1x).
  assert (precise (INV1 *^* INV0)).
  apply mstar_presv_prec; auto.
  apply Prec1. apply Prec0.
  unfold precise in H.
  apply H with M; auto.
  apply MemMod.disjUnion_subseteq_left with M2; auto.
  apply MemMod.disjUnion_subseteq_left with M2x; auto.
  assert (M2 = M2x). apply MemMod.disjUnion_unique with M1 M1x M; auto.
  subst M2.
  unfold lock_release6_Prd_aux.
  split_l. split; auto.
Qed.

Lemma lock_release5_Prd_IF_false:
   forall S : State, lock_release5_Prd S -> S_Flag S _IF = false.
Proof.
  unfold lock_release5_Prd.
  intros.
  destruct H as [sp [l [oc [bp [mP Hpre]]]]].
  unfold lock_release5_Prd_aux in Hpre.
  induction S as (((M, R), fl), isr).
  InvertAll. auto.
Qed.

Lemma lock_release5_Grt_Itr_Stk:
   Grt_Itr_Stk lock_release5_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros.
  induction S as (((M, R), fl), isr).
  unfold lock_release5_grt. unfold leadstoPQ2.
  intros.
  unfold lock_release5_Prd_aux in H4.
  UnfoldStateAll.
  InvertAll.
  unfold itr_enabled in H0.
  inversion_clear H0.
  UnfoldStateAll.
  rewrite H8 in H7. inversion H7.
Qed.


Lemma wff_cmd_lock_release6:
   NH_WFcmd
     (SCAP_consCT (lock_release + 1 + 1 + 1 + 1 + 1 + 1)
        (Ast lock_release6_Prd lock_release6_grt 1) lock_release_psi)
     (lock_release + 1 + 1 + 1 + 1 + 1)
     (Ast lock_release5_Prd lock_release5_grt 1) (comml _sti).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  exists lock_release6_Prd. exists lock_release6_grt.
  split; auto.

  split. apply lock_release_step6_enable.

  split.
  unfold pimp. unfold linkPG.
  intros. 
  destruct H as [S0 [Hpre Htrans]].
  unfold lock_release5_Prd in Hpre.
  destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
  unfold lock_release6_Prd.
  mexists5 sp l pc bp mP.
  apply lock_release_step6_Prd_aux with S0; auto.

  unfold gimp. unfold conjG. unfold conjPG.
  intros.
  destruct H as [S'' [[Hpre Htrans] Hgrt6]].
  unfold lock_release5_grt.
  unfold lock_release6_grt in Hgrt6.
  unfold leadstoPQ2 in *.
  intros.
  apply Hgrt6.
  apply lock_release_step6_Prd_aux with S; auto.
  split. 
    apply IF_false_Prd_Itr_Stk.
    apply lock_release5_Prd_IF_false.
    apply lock_release5_Grt_Itr_Stk.
Qed.

Lemma lock_release_step7:
forall f sp l pc bp mP S S',
  lock_release6_Prd_aux sp l pc bp mP S  
  -> nextS f (comml (_addw (Opr_imm 2) SP)) S S'
  -> lock_release7_Prd_aux (sp + 2) l pc bp mP S'.
Proof.
  intros.
  rename H0 into Htrans.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion Htrans).
  subst M'.
  destruct H1 as [HR' [Hfl' Hisr']].
  unfold opr_eval in HR'.
  subst R' isr'.
  clear Htrans.
  unfold lock_release6_Prd_aux in H.
  destruct H as [HRsp [Hmem [Hlock [Hfl Hisr]]]].
  unfold lock_release7_Prd_aux.
  reg simpl.
  rewrite HRsp.
  split; auto. auto with arith.
  split; auto with arith.
  assert (sp + 2 - 2 = sp). auto.
  rewrite H.
  rewrite <- HRsp.
  assert (RF_update R SP (R SP) = R).
  apply RF_update_prev.
  rewrite H0.
  unfold lock_release6_Prd_aux.
  split_l. split; auto.
  split_l. split_r.
  unfold trash_fl in Hfl'.
  inversion Hfl'.
  subst fl'. unfold Set_Flag. simpl. auto.
  subst fl'. unfold Clear_Flag. simpl. auto.
Qed.

Lemma lock_release6_Prd_Itr_Stk: Prd_Itr_Stk lock_release6_Prd.
Proof.
    unfold Prd_Itr_Stk.
    intros S Hpre Henable.
    unfold lock_release6_Prd in Hpre.
    destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
    unfold lock_release6_Prd_aux in Hpre.
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in Henable.
    simplS in * .
    destruct Hpre as [HRsp [Hmem [Hlock [Hfl Hisr]]]].
    unfold lock_release6_MPre in Hmem.
    sep on.
    sep lift 4 in Hmem.
    sep stksplit (IH_Stk + 6) in Hmem.
    sep lift 0 in Hmem.
    destruct Hmem as [M1 [M2 [Hsplit [Hmem1 Hmem2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    split_l.
    intros M1' Hmem1'.
    unfold lock_release6_Prd.
    mexists5 sp l pc bp mP.
    unfold lock_release6_Prd_aux.
    unfold lock_release6_MPre.
    split_l.
    split; auto.
    sep on.
    sep lift 4.
    assert (R SP >= IH_Stk + 6).
    apply stkSpacePrec_sp_ge_sz with M1; auto.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      split.
        disjoint rep_l M1.
        apply stkSpacePrec_stkSpacePrec_eqDom with (R SP) (IH_Stk + 6); auto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      apply stkSpacePrec_stkSpacePrec_eqDom with (R SP) (IH_Stk + 6); auto.
    split_l.
    sep on.
    sep matchH Hmem2 with 6.
Qed.


Lemma lock_release6_Grt_Itr_Stk: Grt_Itr_Stk lock_release6_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros S S' Hgrt Henable.
  unfold lock_release6_grt in * .
  intros M1 M2 M2' Hsplit HM2 HM2'.
  unfold leadstoPQ2 in *.
  intros sp l pc bp mP Hpre.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in Henable.
  apply Hgrt.
  clear Hgrt.
  unfold lock_release6_Prd_aux in * .
  simplS in * .
  destruct Hpre as [HRsp [Hmem [Hlock [Hfl Hisr]]]].
  split_l.
  split; auto.
  unfold lock_release6_MPre in *.
  sep on.
  sep lift 4 in Hmem.
  sep stksplit (IH_Stk + 6) in Hmem.
  sep lift 0 in Hmem.
  destruct Hmem as [M2x [M1x [Hsplit' [HM2x HM1x]]]].
  sep lift 4.
  assert (R SP >= IH_Stk + 6).
  apply stkSpacePrec_sp_ge_sz with M2x; auto.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  assert (M2x = M2' /\ M1x = M1).
    sep off.
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      apply stkSpacePrec_stkSpacePrec_eqDom with (R SP) (IH_Stk + 6); auto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    assert (Hx := prec_disjUnion_inv H0 H1 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H0; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  split_l.
  sep on.
  sep matchH HM1x with 6.
Qed.

Lemma wff_cmd_lock_release7:
   NH_WFcmd
     (SCAP_consCT (lock_release + 1 + 1 + 1 + 1 + 1 + 1 + 1)
        (Ast lock_release7_Prd lock_release7_grt 1) lock_release_psi)
     (lock_release + 1 + 1 + 1 + 1 + 1 + 1)
     (Ast lock_release6_Prd lock_release6_grt 1)
     (comml (_addw (Opr_imm 2) SP)).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists lock_release7_Prd. exists lock_release7_grt.
  split. auto.

  generalize (lock_release + 1 + 1 + 1 + 1 + 1 + 1).
  intro f.
  split.
  apply trivial_enable_addw.


  split.
  unfold pimp. unfold linkPG.
  intros. 
  destruct H as [S0 [Hpre Htrans]].
  unfold lock_release6_Prd in Hpre.
  destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
  unfold lock_release7_Prd.
  mexists5 (sp + 2) l pc bp mP.
  apply lock_release_step7 with f S0; auto.

  unfold gimp. unfold conjG. unfold conjPG.
  intros.
  destruct H as [S'' [[Hpre Htrans] Hgrt7]].
  unfold lock_release6_grt.
  unfold lock_release7_grt in Hgrt7.
  unfold leadstoPQ2 in *.
  intros.
  assert (sp + 2 + 2 = sp + 4). auto.
  rewrite <- H0.
  apply Hgrt7.
  apply lock_release_step7 with f S; auto.

  split.

  apply lock_release6_Prd_Itr_Stk.
  apply lock_release6_Grt_Itr_Stk.
Qed.


Lemma wff_cmd_lock_release8:
   NH_WFcmd
     (SCAP_consCT (lock_release + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
        (Ast lock_release_Post lock_release_last_grt 1) lock_release_psi)
     (lock_release + 1 + 1 + 1 + 1 + 1 + 1 + 1)
     (Ast lock_release7_Prd lock_release7_grt 1) (comml (_popw BP)).
Proof.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  unfold SCAP_WFcmd.
  exists lock_release_Post. exists lock_release_last_grt.
  split. auto.

  generalize (lock_release + 1 + 1 + 1 + 1 + 1 + 1 + 1). intro f.
  assert (forall sp l pc bp mP, enable (lock_release7_Prd_aux sp l pc bp mP) (nextS f (comml (_popw BP)))).
  intros.
  apply enable_popw.
  intros. assert ((S_RF S SP |~>w bp) (S_Mem S)).
Lemma lock_release7_Prd_1:
  forall sp l pc bp mP S, lock_release7_Prd_aux sp l pc bp mP S -> (S_RF S SP |~>w bp) (S_Mem S).
Proof.
  intros.
  unfold lock_release7_Prd_aux in H.
  induction S as (((M, R), fl), isr).
  InvertAll.
  UnfoldStateAll.
  unfold lock_release6_Prd_aux in H2.
  unfold lock_release6_MPre in H2. reg simpl.
  InvertAll.
  sep on.
  assert (sp - 2 + 2 = sp). auto.
  rewrite H5 in H2.
  sep lift 1 in H2. rewrite H0.
  sep trivial.
Qed.

  apply lock_release7_Prd_1 with sp l pc mP. auto.
  apply leadsToW_leadsToxW with bp; auto.
  split.
  unfold enable in *.
  intros.
  unfold lock_release7_Prd in H0.
  InvertAll.
  apply H with x x0 x1 x2 x3; auto.

Lemma lock_release_step8:
forall f sp l pc bp mP S S',
  lock_release7_Prd_aux sp l pc bp mP S  
  -> nextS f (comml (_popw BP)) S S'
  -> lock_release_Post_aux (sp + 2) l pc bp mP S'.
Proof.
  intros.
  rename H0 into Htrans.
  induction S as (((M, R), fl), isr).
  induction S' as (((M', R'), fl'), isr').
  destruct (nextS_inversion Htrans).
  subst M'.
  destruct H1 as [[x [Hmemx HR']] [Hfl' Hisr']].
  assert (x = bp).
  assert ((S_RF (M, R, fl, isr) SP |~>w bp) (S_Mem (M, R, fl, isr))).
  apply lock_release7_Prd_1 with sp l pc mP; auto.
  UnfoldStateAll.
  apply leadsToW_unique with (R SP) M; auto.
  subst x.
  subst R' fl' isr'.
  unfold lock_release_Post_aux.
  reg simpl.
  clear Htrans.
  unfold lock_release7_Prd_aux in H.
  InvertAll.
  split. rewrite H0. auto.
  split_l.
  unfold lock_release6_Prd_aux in H2.
  reg simpl.
  InvertAll.
  split; auto.
  unfold lock_release6_MPre in H2.
  unfold lock_release_MPost.
  sep on.
  sep lift 2.
  sep stksplit.
  assert (sp + 2 - 2 = sp). auto.
  rewrite H5.
  sep stksplit.
  assert (sp - 2 + 2 = sp). auto.
  rewrite H7 in H2.
  assert (stk_lock_release - 2 - 2 = stk_lock_release - 4).
  auto. rewrite H8.
  assert (sp + 2 + 2 = sp + 4). auto.
  rewrite H9.
  assert (sp - 2 + 4 = sp + 4 - 2). auto.
  rewrite H10 in H2.
  assert (sp + 4 - 2 = sp + 2). auto.
  rewrite H11 in H2.
  assert (sp - 2 + 6 = sp + 6 - 2). auto.
  rewrite H12 in H2.
  assert (sp + 6 - 2 = sp + 4). auto.
  rewrite H13 in H2.
  sep matchH H2 with 6.
Qed.
  
  split.
  unfold pimp.
  unfold linkPG.
  intros.
  InvertAll.
  unfold lock_release7_Prd in H0.
  InvertAll.
  unfold lock_release_Post.
  exists (x0 + 2). exists x1. exists x2. exists x3. exists x4.
  apply lock_release_step8 with f x; auto.

  unfold gimp.
  unfold conjPG. unfold conjG.
  intros.
  InvertAll.
  unfold lock_release7_grt.
  unfold leadstoPQ2.
  intros.
  unfold lock_release_last_grt in H2.
  unfold leadstoPQ2 in H2.
  apply H2.
  apply lock_release_step8 with f S; auto.

Lemma lock_release7_Prd_Itr_Stk: Prd_Itr_Stk lock_release7_Prd.
Proof.
    unfold Prd_Itr_Stk.
    intros S Hpre Henable.
    unfold lock_release7_Prd in Hpre.
    destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
    unfold lock_release7_Prd_aux in Hpre.
    destruct S as [[[M R] fl] isr].
    unfold lock_release6_Prd_aux in Hpre.
    unfold itr_enabled in Henable.
    simplS in * .
    reg simpl.
    destruct Hpre as [HRsp [Hspgt2 [Htmp [Hmem [Hlock [Hfl Hisr]]]]]].
    unfold lock_release6_MPre in Hmem.
    sep on.
    sep lift 4 in Hmem.
    assert (stk_lock_release - 4 >= IH_Stk + 6 - 2).
    unfold stk_lock_release. unfold stk_l_release. omega.
    sep stksplit (IH_Stk + 6 - 2) in Hmem.
    sep lift 0 in Hmem.
    assert (M ||- stkSpacePrec (sp - 2) (IH_Stk + 6 - 2) *^* sp - 2 |->w_ *^* 
                  (stkSpacePrec (sp - 2 - (IH_Stk + 6 - 2))
                    (stk_lock_release - 4 - (IH_Stk + 6 - 2)) *^*
                    (mP *^*
                      (sp - 2 + 2 |->w bp *^* sp - 2 + 4 |->w pc *^*
                        sp - 2 + 6 |->w l)))).
    sep matchH Hmem with 8.
    clear Hmem. rename H0 into Hmem.
    destruct Hmem as [M1 [M2 [Hsplit [Hmem1 Hmem2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    split.
    rewrite HRsp.
    sep on.
    sep stksplit. auto.
    intros M1' Hmem1'.
    unfold lock_release7_Prd.
    mexists5 sp l pc bp mP.
    unfold lock_release7_Prd_aux.
    unfold lock_release6_Prd_aux.
    unfold lock_release6_MPre.
    reg simpl.
    split_l.
    split; auto. split_l. split; auto.
    sep on.
    sep lift 4.
    assert (sp - 2 >= IH_Stk + 6 - 2).
    apply stkSpacePrec_sp_ge_sz1 with M1 (sp - 2 |->w_); auto.
    sep stksplit (IH_Stk + 6 - 2).
    sep lift 1.
    assert (MemMod.merge M2 M1'
           ||- stkSpacePrec (sp - 2) (IH_Stk + 6 - 2) *^* sp - 2 |->w_ *^* 
             (mP *^*
               (sp - 2 + 2 |->w bp *^* sp - 2 + 4 |->w pc *^*
                 sp - 2 + 6 |->w l) *^*
               stkSpacePrec (sp - 2 - (IH_Stk + 6 - 2))
               (stk_lock_release - 4 - (IH_Stk + 6 - 2)))).
    exists M1'; exists M2.
    assert (stkSpacePrec (R SP) (IH_Stk + 6) M1).
    rewrite HRsp.
    sep on.
    sep stksplit. auto.
    split.
      split.
        disjoint rep_l M1.
        apply stkSpacePrec_stkSpacePrec_eqDom with (R SP) (IH_Stk + 6); auto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      apply stkSpacePrec_stkSpacePrec_eqDom with (R SP) (IH_Stk + 6); auto.
    split. 
    sep on. assert (M1' ||- stkSpacePrec (R SP) (IH_Stk + 6)). auto.
    rewrite HRsp in H2. 
    sep stksplit in H2. auto.
    sep on.
    sep matchH Hmem2 with 5. sep matchH H1 with 7.
Qed.

Lemma lock_release7_Grt_Itr_Stk: Grt_Itr_Stk lock_release7_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros S S' Hgrt Henable.
  unfold lock_release7_grt in * .
  intros M1 M2 M2' Hsplit HM2 HM2'.
  unfold leadstoPQ2 in *.
  intros sp l pc bp mP Hpre.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in Henable.
  apply Hgrt.
  clear Hgrt.
  unfold lock_release7_Prd_aux in * .
  unfold lock_release6_Prd_aux in * .
  simplS in * .
  reg simpl.
  destruct Hpre as [HRsp [Hspgt2 [Htmp [Hmem [Hlock [Hfl Hisr]]]]]].
  split_l. split_l. split_l.
  split; auto.
  unfold lock_release6_MPre in *.
  sep on.
  sep lift 4 in Hmem.
  assert (stk_lock_release - 4 >= IH_Stk + 6 - 2).
  unfold stk_lock_release. unfold stk_l_release. omega.
  sep stksplit (IH_Stk + 6 - 2) in Hmem.
  sep lift 0 in Hmem.
  assert (MemMod.merge M1 M2' ||- stkSpacePrec (sp - 2) (IH_Stk + 6 - 2) *^* sp - 2 |->w_ *^* 
    (stkSpacePrec (sp - 2 - (IH_Stk + 6 - 2))
      (stk_lock_release - 4 - (IH_Stk + 6 - 2)) *^*
      (mP *^*
        (sp - 2 + 2 |->w bp *^* sp - 2 + 4 |->w pc *^*
          sp - 2 + 6 |->w l)))).
  sep matchH Hmem with 8.
  clear Hmem. rename H0 into Hmem.
  destruct Hmem as [M2x [M1x [Hsplit' [HM2x HM1x]]]].
  sep lift 4.
  assert (sp - 2 >= IH_Stk + 6 - 2).
  apply stkSpacePrec_sp_ge_sz1 with M2x (sp - 2 |->w_); auto.
  sep stksplit (IH_Stk + 6 - 2).
  sep lift 1.
  assert (M
          ||- stkSpacePrec (sp - 2) (IH_Stk + 6 - 2) *^* sp - 2 |->w_ *^* 
            (mP *^*
              (sp - 2 + 2 |->w bp *^* sp - 2 + 4 |->w pc *^*
                sp - 2 + 6 |->w l) *^*
              stkSpacePrec (sp - 2 - (IH_Stk + 6 - 2))
              (stk_lock_release - 4 - (IH_Stk + 6 - 2)))).
  assert (M2x = M2' /\ M1x = M1).
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      apply stkSpacePrec_stkSpacePrec_eqDom with (R SP) (IH_Stk + 6); auto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    assert (M2x ||- stkSpacePrec (R SP) (IH_Stk + 6)).
    rewrite HRsp.
    sep stksplit. sep on. auto.
    assert (Hx := prec_disjUnion_inv H1 H2 HM2' H3 (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H1; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  split.
  assert (M2 ||- stkSpacePrec (R SP) (IH_Stk + 6)). auto.
  sep on.
  rewrite HRsp in H1.
  sep stksplit in H1. auto.
  sep on.
  sep matchH HM1x with 5.
  sep matchH H1 with 7.
Qed.

  split.
  apply lock_release7_Prd_Itr_Stk.
  apply lock_release7_Grt_Itr_Stk.
Qed.


Lemma lock_releasePost_Prd_Itr_Stk : Prd_Itr_Stk lock_release_Post.
Proof.
    unfold Prd_Itr_Stk.
    intros S Hpre Henable.
    unfold lock_release_Post in Hpre.
    destruct Hpre as [sp [l [pc [bp [mP Hpre]]]]].
    unfold lock_release_Post_aux in Hpre.
    destruct S as [[[M R] fl] isr].
    unfold itr_enabled in Henable.
    simplS in * .
    destruct Hpre as [HRsp [HRbp [Hmem [Hlock [Hfl Hisr]]]]].
    unfold lock_release_MPost in Hmem.
    sep on.
    sep lift 2 in Hmem.
    sep stksplit (IH_Stk + 6) in Hmem.
    sep lift 0 in Hmem.
    destruct Hmem as [M1 [M2 [Hsplit [Hmem1 Hmem2]]]].
    exists M2; exists M1.
    split.
      auto with PMap.
    rewrite HRsp.
    split_l.
    intros M1' Hmem1'.
    unfold lock_release_Post.
    mexists5 sp l pc bp mP.
    unfold lock_release_Post_aux.
    unfold lock_release_MPost.
    split_l. split_l.
    split; auto.
    sep on.
    sep lift 2.
    assert (sp >= IH_Stk + 6).
    apply stkSpacePrec_sp_ge_sz with M1; auto.
    sep stksplit (IH_Stk + 6).
    sep lift 1.
    exists M1'; exists M2.
    split.
      split.
        disjoint rep_l M1.
        apply stkSpacePrec_stkSpacePrec_eqDom with sp (IH_Stk + 6); auto.
      merge commut M2.
      trivial.
      disjoint rep_r M1.
      apply stkSpacePrec_stkSpacePrec_eqDom with sp (IH_Stk + 6); auto.
    split_l.
    sep on.
    sep matchH Hmem2 with 3.
Qed.

Lemma lock_releasePost_Grt_Itr_Stk : Grt_Itr_Stk lock_release_last_grt.
Proof.
  unfold Grt_Itr_Stk.
  intros S S' Hgrt Henable.
  unfold lock_release_last_grt in * .
  intros M1 M2 M2' Hsplit HM2 HM2'.
  unfold leadstoPQ2 in *.
  intros sp l pc bp mP Hpre.
  destruct S as [[[M R] fl] isr].
  unfold itr_enabled in Henable.
  apply Hgrt.
  clear Hgrt.
  unfold lock_release_Post_aux in * .
  simplS in * .
  destruct Hpre as [HRsp [HRbp [Hmem [Hlock [Hfl Hisr]]]]].
  split_l. split_l.
  split; auto.
  unfold lock_release_MPost in *.
  sep on.
  sep lift 2 in Hmem.
  sep stksplit (IH_Stk + 6) in Hmem.
  sep lift 0 in Hmem.
  destruct Hmem as [M2x [M1x [Hsplit' [HM2x HM1x]]]].
  sep lift 2.
  assert (sp >= IH_Stk + 6).
  apply stkSpacePrec_sp_ge_sz with M2x; auto.
  sep stksplit (IH_Stk + 6).
  sep lift 1.
  assert (M2x = M2' /\ M1x = M1).
    rewrite HRsp in HM2. rewrite HRsp in HM2'.
    assert (MemMod.disjUnion M1 M2' (MemMod.merge M1 M2')).
    split_r.
      disjoint rep_r M2.
      apply stkSpacePrec_stkSpacePrec_eqDom with sp (IH_Stk + 6); auto.
    assert (MemMod.disjUnion M1x M2x (MemMod.merge M1 M2')).
      auto with PMap.
    assert (Hx := prec_disjUnion_inv H0 H1 HM2' HM2x (@precise_stkSpacePrec _ _)).
    destruct Hx; auto.
  destruct H0; subst M1x M2x.
  exists M2; exists M1.
  split.
    auto with PMap.
  rewrite HRsp in HM2.
  split_l.
  sep on.
  sep matchH HM1x with 3.
Qed.


Lemma wff_cmd_lock_release9:
   NH_WFiseq lock_release_psi (lock_release + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1)
     (Ast lock_release_Post lock_release_last_grt 1) 
     (sseq _retw).
Proof.
  apply  nh_isq_end.
  unfold NH_WFcmd.
  split.
  unfold NH_WFcmd_aux.
  split; auto.
  intros.
  unfold lock_release_last_grt.
  unfold leadstoPQ2. auto.
  split.
  apply lock_releasePost_Prd_Itr_Stk.
  apply lock_releasePost_Grt_Itr_Stk.
Qed.


Lemma wff_lock_release:
  NH_WFiseq lock_release_psi lock_release (Ast lock_release_Prd lock_release_grt 1) I_lock_release.
Proof.
  unfold I_lock_release.
  apply nh_isq_seq with lock_release1_Prd lock_release1_grt.
  apply wff_cmd_lock_release1.
  apply nh_isq_seq with lock_release2_Prd lock_release2_grt.
  apply wff_cmd_lock_release2.
  apply nh_isq_seq with lock_release3_Prd lock_release3_grt.
  apply wff_cmd_lock_release3.
  apply nh_isq_seq with lock_release4_Prd lock_release4_grt.
  apply wff_cmd_lock_release4.
  apply nh_isq_seq with lock_release5_Prd lock_release5_grt.
  apply wff_cmd_lock_release5.
  apply nh_isq_seq with lock_release6_Prd lock_release6_grt.
  apply wff_cmd_lock_release6.
  apply nh_isq_seq with lock_release7_Prd lock_release7_grt.
  apply wff_cmd_lock_release7.
  apply nh_isq_seq with lock_release_Post lock_release_last_grt.
  apply wff_cmd_lock_release8.
  apply wff_cmd_lock_release9.
Qed.
